#!/usr/bin/perl
#
# summary_examine_complex.2_2D.raw.pl:
#	This slightly modified version of 'summary_examine_complex.2.raw.pl'
#	takes exclusively the 2D distributions of the properties.
#	It classifies the segments into two categories: A = 'too-long' + 'complex errors', and  B = 'non-complex errors' + 'correct'.
#	[CREATION STARTED on June 21 (Tue), 2016, ENDED on June dd (www), 2016.]
#
# summary_examine_complex.2.raw.pl:
#	The focus of this script is on the difference between the clusters of gapped segments
#	that were erroneously reconstructed and those that were correctly reconstructed.
#	This script was created by modifying 'summary_examine_complex.1.raw.pl'.
#	[CREATION STARTED on June 2 (Thu), 2016, ENDED on June 3 (Fri), 2016.]
#
# summary_examine_complex.1.raw.pl:
#	This Perl script summarizes the results of the Perl script 'examine_complex_via_mblks.2.pl',
#	which are converted into a single table via 'tabulate_examine_complex.pl'.
#	The script was created by modifying 'summary_examine_purge_ctrl.raw.pl'.
#	Especially, it imports (from 'bch_pbs_examine_purges.D.pl') the function to read in the list of relevant segments.
#	[CREATED on June 1 (Wed), 2016.]
#
# summary_examine_purge_ctrl.raw.pl:
#	This Perl script summarizes the results of the Perl script 'examine_purge_ctrls.2.pl',
#	which are converted into a single table via 'tabulate_examine_purge_ctrls.pl'.
#	The script was created by slightly modifying 'summary_examine_purge.raw.pl'.
#	[CREATED on May 16 (Mon), 2016.]
#
# summary_examine_purge.raw.pl:
#	This Perl script summarizes the results of the Perl script 'examine_purges_via_mblks.2.pl',
#	which are converted into a single table via 'tabulate_examine_purge.pl'.
#	The script was created by taking advantage of the framework of 'summary_classify_msa_errors1.raw.pl'.
#	[CREATION Started on May 11 (Wed), 2016, ENDED on Mmm dd (www), 2016.]
#
#	[MODIFIED on May 16 (Mon), 2016.]
#	Now, the script also examines more detailed distributions,
#	such as the distributions of the P-values for specified window sizes and specified branch properties.
#
# summary_classify_msa_errors1.raw.pl:
#	This script is a slightly modified version of 'summary_classify_msa_errors.raw.pl'.
#	The script now tallies the number of all composite-blocks in each erroneous segment as well.
#	[CREATED on Oct 26 (Mon), 2015]
#
# summary_classify_msa_errors.raw.pl:
#	This script does not pre-screen either in terms of whether or not the erroneous segment is on a MSA boudary or 
#	whether or not the segment contains non-parsimonious gapped segments.
#	It was created by slightly modifying the script 'summary_classify_msa_errors.bulk_pars.pl'.
#	(CREATION STARTED on Nov 17 (Mon), 2014, ENDED on Nov 18 (Tue), 2014.)
#
# summary_classify_msa_errors.bulk_pars.pl:
#	This Perl script summarizes the results of the Perl script 'classify_msa_errors_via_mblks.1.pl',
#	which are converted into a single table via 'tabulate_classify_msa_errors.pl'.
#	It excludes those erroneous segments on the MSA boundaries and those containing non-parsimonious gapped segments.
#	It was created by taking advantage of the framework of the script 'summary_cmp_ref_vs_rec_gpconf_stats.bulk_pars.pl'.
#	(CREATION STARTED on Nov 16 (Sun), 2014, ENDED on Nov 17 (Mon), 2014.)
#
# summary_cmp_ref_vs_rec_gpconf_stats.bulk_pars.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_gconf_stats.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_gconf_stats.pl'.
#	It excludes those erroneous segments on the MSA boundaries and those containing non-parsimonious gapped segments.
#	It was created by taking advantage of the framework of the script 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2_lite.pl'.
#	(CREATION STARTED on Nov 14 (Fri), 2014, ENDED on Nov 15 (Sat), 2014.)
#
# summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2_lite.pl:
#	This script was created by just stripping commented-out commands off 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2.pl'.
#	(CREATED on Nov 14 (Fri), 2014.)
#
# summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2.pl:
#	The only modifications from the previous version is that
#	this version uses boundaries whose results could easily be interpreted as 'round-off'.
#	(CREATED on Nov 13 (Thu), 2014.)
#
# summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_indel_cts.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_cts_indels.pl'.
#	It excludes those erroneous segments on the MSA boundaries and those containing non-parsimonious gapped segments.
#	It was created by modifying the script 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'.
#	(CREATED on Oct 9 (Thu), 2014.)
#
# summary_cmp_ref_vs_rec_complliment.bulk_pars.pl:
#	This revised version of 'summary_cmp_ref_vs_rec_complliment.bulk.pl'
#	further excludes erroneously reconstructed segments including non-parsimonious gapped segments
#	before tallying the bulk erroneous segments.
#	(CREATED on Oct 3 (Fri), 2014.)
#
# summary_cmp_ref_vs_rec_complliment.bulk.pl:
#	This revised version of 'summary_cmp_ref_vs_rec_complliment.pl' tallies the erroneously reconstructed segments
#	after excluding the segments that touch either of the MSA ends.
#	(CREATION STARTED on Oct 2 (Thu), 2014, ENDED on Oct 3 (Fri), 2014.)
#
#
# summary_cmp_ref_vs_rec_complliment.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_compllimment.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_compllimment.pl'.
#	It was created by modifying the script 'summary_cmp_ref_vs_rec_abs_scr.pl'.
#	(CREATED on Oct 2 (Thu), 2014.)
#
# summary_cmp_ref_vs_rec_abs_scr.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_abs_scoring.PhyML.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_abs_scoring.pl'.
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);
use MyMalnIO qw(
	extract_fasta extract_clw
	);
	## pr_clustal_like_lite_aa
	## pr_ODEN_input pr_phylip_aln
use MyArrayManipu qw( diff_array );

use MyMalnManipu qw( rm_gapcolumn2 );
use MyMalnManipu_UH qw(
	make_segid2segaln
	msa2len_grps msa2column_set extract_aln_mtrx2
	);
	## extract_aln_mtrx

use MyCluster_distribution qw(
	bin_initializer cluster_distributor
	batch_initializer batch_pr_file_distri batch_pr_file_distri_xl
	cluster_distributor_wgt
	init_2dim_distri distributor_2dim pr_file_2dim_distri pr_file_2dim_distri_xl
	create_key_intervals );
use MyContingency_table qw(
	ratios_contingency_tables add_ctbl_margins
	pr_juxtapose_ctbls_xl convert_ctbl_to_2d_distri
	);

sub broad_error_type ($);


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);


	### User-specified Parameters ###


		# MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #
our %INTYPE2CTSEQS = (
	'12primates' => 12,
	'15mammals' => 15,
	'9fe_mammals' => 9,
	'TESTs/Test1a' => 15
	);
##our  $CT_SEQS = 16; # Number of sequences in a MSA. # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #

	# If the scores of the reference and the reconstructed MSAs for a segment 
	# differ less than this value, they are regarded as 'equal' to each other.
##our $THRSH_EQUAL = 0.1;
##our $THRSH_EQUAL = 1E-03;
our $THRSH_EQUAL = 1E-05;

our $EPS = $THRSH_EQUAL * 1E-03; # An 'infinitesimal number' used to take statistics.


	# Level of the verboseness of the test outputs.
	# (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose.)

our $TEST_LEVEL = 5;

	## Specifying the fields of input/output files. ##

		# MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #

my @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}');

=For the predecessor 'summary_examine_purge_ctrl.raw.pl'.
		# MODIFIED in creation of 'summary_examine_purge_ctrl.raw.pl'. #

my @TAGS_IN_FLD = ( 'ID_Class', 'Session_ID', 'Segment_ID', 'Wd_size', 'indx_wd', 'br_ID', 'br_len', '#{ext_offs}', '#{diffs}_exp', '#{diffs}_obs', 'P-value');
=cut

=For the predecessor 'summary_examine_purge.raw.pl'
		# MODIFIED in creation of 'summary_examine_purge.raw.pl'. #

our @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Rlv_branch', 'len(rlv_branch)', '#{involved_seqs}', 'block_size',  '|shift|', '#{sbst_diffs}_exp', '#{sbst_diffs}_obs', 'P-value', 'log[Lk(sbst)]_[rec-ref]');
=cut

=For the predecessor 'summary_classify_msa_errors1.raw.pl'.
		# MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'.
our @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', '#{Equiv_indels}', 'Tot#{composite-blocks}', '#{cblks}_eff', 'Tot_len{cblks}_eff', 'Types;of;associated;errors', 'Lengths,of,effective,composite,blocks');
##our @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', '#{Equiv_indels}', 'Tot#{composite-blocks}', '#{cblks}_skipped', '#{cblks}_no_assoc', '#{cblks}_noneff', '#{cblks}_eff', 'Tot_len{cblks}_eff', 'Types;of;associated;errors', 'Lengths,of,effective,composite,blocks');
=cut

=For the predecessor 'summary_cmp_ref_vs_rec_gpconf_stats.bulk_pars.pl'.
	# MODIFIED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
our @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Tot_seg_length.ref', 'Tot_seg_length.rec', '#{Gapped_segments}.ref', '#{Gapped_segments}.rec', 'Tot#{gapped_columns}.ref', 'Tot#{gapped_columns}.rec', 'Max_gaplength.ref', 'Max_gaplength.rec', 'Gap_Content.ref', 'Gap_Content.rec');

our @TAGS_OUT_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Tot_seg_length.ref', '#{Gapped_segments}.ref', 'Tot#{gapped_columns}.ref', 'Max_gaplength.ref', 'Gap_Content.ref', '#{gap_characters}.ref', 'fDiff({Tot_seg_length})', 'fDiff(#{Gapped_segments})', 'fDiff(Tot#{gapped_columns})', 'fDiff(Max_gaplength)', 'fDiff(#{gap_characters})' );

## @tags_ct = ('Intervals', '#{gpd_segs}_ref', '#{gpd_clms}_ref', 'Max{gaplen}_ref', '{Tot_seglen}_ref');
## @tags_prpt ('Intervals', '{Gap_content}_ref');
## @tags_fdiff = ('Intervals', 'fDiff(#{gpd_segs})', 'fDiff(#{gpd_clms})', 'fDiff(Max#{gaplen})', 'fDiff({Tot_seglen})', 'fDiff({#{gap_characters}})');
=cut

=For the predecessor 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2_lite.pl'.
	# MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
our @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Total_Score.ref',  'Total_Score.rec', 'AvAv#{insertions}.ref', 'AvAv#{insertions}.rec', 'AvAv#{deletions}.ref', 'AvAv#{deletions}.rec', 'Av#{null_columns}.ref', 'Av#{null_columns}.rec');

our @TAGS_OUT_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'AvAv#{indels}.ref', 'Diff(AvAv#{indels})', 'L1_Dist(#{ins},#{dels})', 'Deletion_Bias');
=cut


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
our @TAGS_AUX_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', '#{nonpars_gpd_segs}'); # ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'. #

	# ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
our @TAGS_AUX3_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Total_Score.ref',  'Total_Score.rec', 'Av#{null_columns}.ref', 'Av#{null_columns}.rec');
=cut


##my $alignment_type = 'MAFFT/E_INS_i';
##my $alignment_type = 'MAFFT/G_INS_i';
##my $alignment_type = 'PRANK/Default';
my $alignment_type = 'PRANK/Best_Match';

	# ADDED in creation of 'summary_examine_purge.raw.pl'. #
#my $input_msa_type = '12primates';
#my $input_msa_type = '15mammals';
my $input_msa_type = '9fe_mammals';

#my $input_msa_type = 'TESTs/Test1a'; # For TESTING the script. #

our $CT_SEQS = $INTYPE2CTSEQS{$input_msa_type}; ## ADDED in creation of 'summary_examine_complex.1.raw.pl'. ##

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.

	# Auxiliary files. #
	# $auxfile = "$top_auxdir/$mid_auxdir_pref$id_class$mid_auxdir_suff/$mid_aux_path/$auxfile_pref$session_id$auxfile_suff",
	# where  $id_class and $sesion_id will be collected later.
	# (ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'.)

my $top_auxdir = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Errors/Erroneous_Regions/Equiv_Regions/$alignment_type/hogd_alt_lite_sbcl.1_lite/Outputs";
my $mid_auxdir_pref = 'hog_d';
my $mid_auxdir_suff = '';
my $mid_aux_path = '';
my $auxfile_pref = 'equiv_rgs_brief.';
my $auxfile_suff = '.txt';

		# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'. #
my $auxfile2 = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Errors/Erroneous_Regions/Equiv_Regions/$alignment_type/Inclusion_gpd_segs/hogd_alt_lite_sbcl.1/tbl_incl_gpd_segs.txt.gz";
=cut

=DISABLED in creation of 'summary_examine_purge.raw.pl'.

		# ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
##my $auxfile3 = "../../../Cts_Indels/Summary_tbl/hogd_alt_lite_sbcl.1/tbl_cmp_ref_vs_rec_cts_indels.txt.gz";
##my $auxfile3 = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Errors/Contrast_Analyses/Characteristics/Erroneous_Regions/$alignment_type/Cts_Indels/Summary_tbl/hogd_alt_lite_sbcl.1/tbl_cmp_ref_vs_rec_cts_indels.txt.gz";
my $auxfile3 = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Prcd_Errors/Contrast_Analyses/Characteristics/Erroneous_Regions/$alignment_type/Cts_Indels/Summary_tbl/hogd_alt_lite_sbcl.1/tbl_cmp_ref_vs_rec_cts_indels.txt.gz";

=cut

		# ADDED in creation of 'summary_examine_complex.1.raw.pl'. #
my $auxfile5 = "../../../Aux_Inputs/Tbl_Complex/tbl_errsegs_w_complex.$input_msa_type.txt.gz"; 

	### Input/Output files/directories ###

	# $infile1 = "$indir1/$infile_name1",
my $indir1 = "../../Summary_tbl/$input_msa_type"; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
##my $indir1 = "../../Summary_tbl/hogd_alt_lite_sbcl.1";
##my $infile_name1 = 'tbl_cmp_ref_vs_rec_compllimment.txt.gz';
##my $infile_name1 = 'tbl_cmp_ref_vs_rec_cts_indels.txt.gz'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
##my $infile_name1 = 'tbl_cmp_ref_vs_rec_gconf_stats.txt.gz'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'.
##my $infile_name1 = 'tbl_classify_msa_errors.txt.gz'; # MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'.
##my $infile_name1 = 'tbl_classify_msa_errors.1.txt.gz'; # MODIFIED in creation of 'summary_classify_msa_errors1.raw.pl'. #
##my $infile_name1 = 'tbl_examine_purge.txt.gz'; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
##my $infile_name1 = 'tbl_examine_purge_ctrl.txt.gz'; # MODIFIED in creation of 'summary_examine_purge_ctrl.raw.pl'. #

my $infile_name1 = 'tbl_examine_complex.txt.gz'; # MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #


	# $outfile = "$outdir/$outfile_name"
my $outdir =  $input_msa_type; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
##my $outdir =  'hogd_alt_lite_sbcl.1';
##my $outdir = "TESTs/Test1"; ### For TESTING the script ###
##my $outfile_name = 'tbl_diff_ref_vs_rec_compllimment.txt';
##my $outfile_name = 'tbl_diff_ref_vs_rec_cts_indels.txt'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
##my $outfile_name = 'tbl_fdiff_ref_vs_rec_cts_indels.txt'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
##my $outfile_name = 'tbl_classify_sbj_msa_errors.txt'; # MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
##my $outfile_name = 'tbl_classify_sbj_msa_errors.1.txt'; # MODIFIED in creation of 'summary_classify_msa_errors1.raw.pl'. #
##my $outfile_name = 'tbl_examine_sbj_purges.txt'; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
my $outfile_name = 'tbl_examine_sbj_purge_ctrls.txt'; # MODIFIED in creation of 'summary_examine_purge_ctrl.raw.pl'. #



	## Check files/directories ##

foreach my $dir ($indir1) {
##foreach my $dir ($intopdir1, $intopdir2) {
    unless (( -d $dir ) and ( -r $dir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable TOP INPUT DIRECTORY '$dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}

##=DISABLED in creation of 'summary_examine_purge.raw.pl'. ## RESURRECTED & MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. ##

foreach my $auxfile ($auxfile5) { # MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #
##foreach my $auxfile ($auxfile3) {...} # MODIFIED in creation of 'summary_classify_msa_errors.raw.pl'. #
##foreach my $auxfile ($auxfile2, $auxfile3) {...} # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
   unless ( ( -f $auxfile ) and ( -r $auxfile ) ) { # ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'. #
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable Auxiliary File '$auxfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
   }
}
##=cut

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.

		# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'. #

foreach my $dir ($top_auxdir) {
    unless (( -d $dir ) and ( -r $dir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable Top Auxiliary Directory '$dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}

my $list_id_classes = list_sessions_ext ($top_auxdir, $mid_auxdir_pref, $mid_auxdir_suff, 'dr');
unless (defined $list_id_classes) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the top auxiliary directory '$top_auxdir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

	# $auxfile = "$top_auxdir/$mid_auxdir_pref$id_class$mid_auxdir_suff/$mid_aux_path/$auxfile_pref$session_id$auxfile_suff",
		# END of "ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'." #

=cut # END of "DISABLED in creation of 'summary_classify_msa_errors.raw.pl'."


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
=cut


	### Declare the Beginning of the Script ###

print STDOUT "\n<< Program name: $program_name (path: $program_path) >>\n\n";

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.
print STDOUT "# This analysis will be performed after excluding the erroneous segments on the left-ends or the right-ends of MSAs, as well as the erroneous segments containing non-parsimonious gapped segments.\n\n";
=cut

print STDOUT '$alignment_type = ', $alignment_type , " .\n\n";

=For the predecessor 'summary_cmp_ref_vs_rec_abs_scr.pl'.
print STDOUT "Scoring method: $METHOD .\n\n";
=cut

print STDOUT '$input_msa_type = ', $input_msa_type , " .\n\n"; # ADDED in creation of 'summary_examine_purge.raw.pl'. #

print STDOUT "Input file:  {in_dir1}/$infile_name1,\n";
print STDOUT "       where  {in_dir1} = $indir1 .\n\n";


			# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'. #
print STDOUT "Auxiliary input files:\n\n";

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.
print STDOUT "   (for segments on MSA-ends)  {top_auxdir}/$mid_auxdir_pref", '{ID_Class}', "$mid_auxdir_suff/$mid_aux_path/$auxfile_pref", '{session_ID}', "$auxfile_suff ,\n";
print STDOUT "              where {top_auxdir} = $top_auxdir ,\n";
print STDOUT "                    {ID_Class} = ", join (', ', @{$list_id_classes}), " ,\n\n";
print STDOUT "                    {session_ID}'s will be collected in the class-specific manner ;\n\n";
	# $auxfile = "$top_auxdir/$mid_auxdir_pref$id_class$mid_auxdir_suff/$mid_aux_path/$auxfile_pref$session_id$auxfile_suff",

print STDOUT "    (for non-parsimonious segments) $auxfile2 ,\n\n"; # ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'.
=cut


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
print STDOUT "    (for complete-likelihoods and counts of null columns) $auxfile3 .\n\n"; # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
=cut

print STDOUT "    (for segments with 'Complex' errors or 'N/A') $auxfile5 .\n\n"; # ADDED in creation of 'summary_examine_complex.1.raw.pl'. #

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
print STDOUT "Output file: {out_dir}/$outfile_name,\n";
print STDOUT "       where {out_dir} = $outdir .\n\n\n";
=cut

print STDOUT "Parameters necessary for the calculation:\n\n";
##print STDOUT "DEFINITION:  'Diff' = 'Rec' - 'Ref' .\n\n";
print STDOUT '$THRSH_EQUAL = ', "$THRSH_EQUAL .\n";
print STDOUT '$EPS         = ', "$EPS .\n\n\n";


print STDOUT '#{Sequences} = ', $CT_SEQS , " .\n\n\n"; # ADDED in creation of 'summary_examine_complex.1.raw.pl'. #


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
	# ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
print STDOUT "** Acronyms Used:\n\n";
print STDOUT "  * '(i)CII'  = '(Incomplete-)Collapse-of-Independent-Insertions' ,\n";
print STDOUT "  * '(i)CEII' = '(Incomplete-)Creation-of-Erroneous-Independent-Insertions' ,\n";
print STDOUT "  * '(p)NB_Merge/Split' = '(Partial-)Neighboring-Branch-Merge/Split' .\n\n";
=cut



	### Arrays used to take Statistics ###

		### MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. ###
		### => FUrther MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. ###
		### => Yet Further MODIFIED in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

our @BOUNDS_CT0 = (0 .. 5);
our @BOUNDS_CT = (0 .. 11, 13, 15, 17, 19, 21, 26, 31, 41, 51);
our @BOUNDS_PNCT0 = (-7 .. 7);

=For the predecessor 'summary_examine_complex.2.raw.pl'.
our @BOUNDS_CT = (0 .. 11, 13, 15, 17, 19, 21, 26, 31, 41, 51, 76, 101, 151, 201, 301);
our @BOUNDS_PNCT = (-200, -100, -50, -40, -30, -25, -20, -15, -10 .. 11, 16, 21, 31, 41, 51, 101, 201);
##our @BOUNDS_CT = (0 .. 11, 13, 15, 17, 19, 21, 26, 31, 41, 51, 76, 101, 151, 201);
##our @BOUNDS_PNCT = (-100, -50, -40, -30, -25, -20, -15, -10 .. 11, 16, 21, 31, 41, 51, 101);
our @BOUNDS_PRPT = (0, 0.001, 0.005, 0.01, 0.025, 0.05, 0.1, 0.2, 0.5, 0.8, 0.9, 0.95, 0.975, 0.99, 0.995, 0.999, 1);
=cut

=For the predecessor 'summary_examine_purge.raw.pl'.
		### MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
our @BOUNDS_CT = (0 .. 10, 12, 14, 16, 18, 20, 25, 30, 40);
our @BOUNDS_BRLEN = (0, 0.02, 0.04, 0.06, 0.08, 0.1, 0.12, 0.14, 0.16, 0.18, 0.2, 0.25, 0.3);
our @BOUNDS_PRPT = (0, 0.001, 0.005, 0.01, 0.025, 0.05, 0.1, 0.2, 0.5, 0.8, 0.9, 0.95, 0.975, 0.99, 0.995, 0.999, 1); # CORRECTED on May 16, 2016. #
##our @BOUNDS_PRPT = (0, 0.001, 0.005, 0.01, 0.025, 0.05, 0.1, 0.2, 0.5, 0.8, 0.1, 0.95, 0.975, 0.99, 0.995, 0.999, 1);
##our @BOUNDS_DIFF_LNLK = (-20, -15, -12.5, -10, -7.5, -5, -4, -3, -2, -1, -0.5, -$EPS, +$EPS, +0.5, +1, +2, +3, +4, +5, +7.5, +10);
=cut

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
 	# ADDED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
our @BOUNDS_CT = (0, +$THRSH_EQUAL, 1, 2, 3, 5, 10, 20, 30, 50, 100, 200); # MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
##our @BOUNDS_CT = (0, +$THRSH_EQUAL, 1, 2, 3, 5, 10, 20, 30, 50, 100); # MODIFIED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
##our @BOUNDS_CT = (0, +$THRSH_EQUAL, 0.5, 1.5, 2.5, 3.5, 5.5, 10.5, 20.5); # MAJOR MODIFICATION in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2.pl'. #
##our @BOUNDS_CT = (0, +$THRSH_EQUAL, 1, 2, 3, 5, 10, 20);
##our @BOUNDS_CT = (0, +$EPS, 1, 2, 3, 5, 10, 20);

		# MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #

our @BOUNDS_PRPT = (0, +$THRSH_EQUAL, 0.05, 0.1, 0.2, 0.4, 0.6, 0.8, 0.9, 0.95, 1); 

=cut # END of "DISABLED in creation of 'summary_examine_purge.raw.pl'."


		### ADDED in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

(my $bins_max_totct_indels_vs_maxct_columns_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_maxct_columns_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_totlen_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_max_totlen_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_max_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

(my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_min_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");


=For the predecessor 'summary_examine_complex.2.raw.pl'. (I)

		### MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. ###
		### => Substantially MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. ###

my @tags_segtypes = ('Intervals', 'Too_long', 'Complex', 'Error(non-complex)', 'Correct');


		# Totally new. #
my @bins_ct_gpd_segs_tl = my @bins_ct_gpd_segs_cmplx = my @bins_ct_gpd_segs_err = my @bins_ct_gpd_segs_corr = ();
my @batch_ct_gpd_segs = (\@bins_ct_gpd_segs_tl, \@bins_ct_gpd_segs_cmplx, \@bins_ct_gpd_segs_err, \@bins_ct_gpd_segs_corr);

my @bins_span_cluster_tl = my @bins_span_cluster_cmplx = my @bins_span_cluster_err = my @bins_span_cluster_corr = ();
my @batch_span_cluster = (\@bins_span_cluster_tl, \@bins_span_cluster_cmplx, \@bins_span_cluster_err, \@bins_span_cluster_corr);

batch_initializer (@batch_ct_gpd_segs, @BOUNDS_CT);
batch_initializer (@batch_span_cluster, @BOUNDS_CT);

	## cluster_distributor ($ct_gpd_segs, @{$bins_ct_gpd_segs}, @BOUNDS_CT); # Totally new.
	## cluster_distributor ($span_cluster, @{$bins_span_cluster}, @BOUNDS_CT); # Totally new.

		# END of "Totally new." #


		# Modified (1) - Total counts -. #

my @bins_totct_clms_tl = my @bins_totct_clms_cmplx = my @bins_totct_clms_err = my @bins_totct_clms_corr = ();
my @batch_totct_clms = (\@bins_totct_clms_tl, \@bins_totct_clms_cmplx, \@bins_totct_clms_err, \@bins_totct_clms_corr);

##my @bins_prpt_gps_tl = my @bins_prpt_gps_cmplx = my @bins_prpt_gps_err = my @bins_prpt_gps_corr = ();
##my @batch_prpt_gps = (\@bins_prpt_gps_tl, \@bins_prpt_gps_cmplx, \@bins_prpt_gps_err, \@bins_prpt_gps_corr);

my @bins_tot_totct_indel_tl = my @bins_tot_totct_indel_cmplx = my @bins_tot_totct_indel_err = my @bins_tot_totct_indel_corr = ();
my @batch_tot_totct_indel = (\@bins_tot_totct_indel_tl, \@bins_tot_totct_indel_cmplx, \@bins_tot_totct_indel_err, \@bins_tot_totct_indel_corr);

my @bins_tot_diffct_indel_tl = my @bins_tot_diffct_indel_cmplx = my @bins_tot_diffct_indel_err = my @bins_tot_diffct_indel_corr = ();
my @batch_tot_diffct_indel = (\@bins_tot_diffct_indel_tl, \@bins_tot_diffct_indel_cmplx, \@bins_tot_diffct_indel_err, \@bins_tot_diffct_indel_corr);

my @bins_tot_totlen_indel_tl = my @bins_tot_totlen_indel_cmplx = my @bins_tot_totlen_indel_err = my @bins_tot_totlen_indel_corr = ();
my @batch_tot_totlen_indel = (\@bins_tot_totlen_indel_tl, \@bins_tot_totlen_indel_cmplx, \@bins_tot_totlen_indel_err, \@bins_tot_totlen_indel_corr);

my @bins_tot_difflen_indel_tl = my @bins_tot_difflen_indel_cmplx = my @bins_tot_difflen_indel_err = my @bins_tot_difflen_indel_corr = ();
my @batch_tot_difflen_indel = (\@bins_tot_difflen_indel_tl, \@bins_tot_difflen_indel_cmplx, \@bins_tot_difflen_indel_err, \@bins_tot_difflen_indel_corr);

batch_initializer (@batch_totct_clms, @BOUNDS_CT);
##batch_initializer (@batch_prpt_gps, @BOUNDS_PRPT);
batch_initializer (@batch_tot_totct_indel, @BOUNDS_CT);
batch_initializer (@batch_tot_diffct_indel, @BOUNDS_PNCT);
batch_initializer (@batch_tot_totlen_indel, @BOUNDS_CT);
batch_initializer (@batch_tot_difflen_indel, @BOUNDS_PNCT);

		# END of "Modified (1) - Total counts -." #


		# Modified (2) - Maxima -. #

my @bins_ct_clms_max_tl = my @bins_ct_clms_max_cmplx = my @bins_ct_clms_max_err = my @bins_ct_clms_max_corr = ();
my @batch_ct_clms_max = (\@bins_ct_clms_max_tl, \@bins_ct_clms_max_cmplx, \@bins_ct_clms_max_err, \@bins_ct_clms_max_corr);

my @bins_totct_indel_max_tl = my @bins_totct_indel_max_cmplx = my @bins_totct_indel_max_err = my @bins_totct_indel_max_corr = ();
my @batch_totct_indel_max = (\@bins_totct_indel_max_tl, \@bins_totct_indel_max_cmplx, \@bins_totct_indel_max_err, \@bins_totct_indel_max_corr);

my @bins_diffct_indel_max_tl = my @bins_diffct_indel_max_cmplx = my @bins_diffct_indel_max_err = my @bins_diffct_indel_max_corr = ();
my @batch_diffct_indel_max = (\@bins_diffct_indel_max_tl, \@bins_diffct_indel_max_cmplx, \@bins_diffct_indel_max_err, \@bins_diffct_indel_max_corr);

my @bins_totlen_indel_max_tl = my @bins_totlen_indel_max_cmplx = my @bins_totlen_indel_max_err = my @bins_totlen_indel_max_corr = ();
my @batch_totlen_indel_max = (\@bins_totlen_indel_max_tl, \@bins_totlen_indel_max_cmplx, \@bins_totlen_indel_max_err, \@bins_totlen_indel_max_corr);

my @bins_difflen_indel_max_tl = my @bins_difflen_indel_max_cmplx = my @bins_difflen_indel_max_err = my @bins_difflen_indel_max_corr = ();
my @batch_difflen_indel_max = (\@bins_difflen_indel_max_tl, \@bins_difflen_indel_max_cmplx, \@bins_difflen_indel_max_err, \@bins_difflen_indel_max_corr);

batch_initializer (@batch_ct_clms_max, @BOUNDS_CT);
batch_initializer (@batch_totct_indel_max, @BOUNDS_CT);
batch_initializer (@batch_diffct_indel_max, @BOUNDS_PNCT);
batch_initializer (@batch_totlen_indel_max, @BOUNDS_CT);
batch_initializer (@batch_difflen_indel_max, @BOUNDS_PNCT);

		# END of "Modified (2) - Maxima -." #


		# Modified (3) - 2nd maxima -. #

my @bins_ct_clms_2ndmax_tl = my @bins_ct_clms_2ndmax_cmplx = my @bins_ct_clms_2ndmax_err = my @bins_ct_clms_2ndmax_corr = ();
my @batch_ct_clms_2ndmax = (\@bins_ct_clms_2ndmax_tl, \@bins_ct_clms_2ndmax_cmplx, \@bins_ct_clms_2ndmax_err, \@bins_ct_clms_2ndmax_corr);

my @bins_totct_indel_2ndmax_tl = my @bins_totct_indel_2ndmax_cmplx = my @bins_totct_indel_2ndmax_err = my @bins_totct_indel_2ndmax_corr = ();
my @batch_totct_indel_2ndmax = (\@bins_totct_indel_2ndmax_tl, \@bins_totct_indel_2ndmax_cmplx, \@bins_totct_indel_2ndmax_err, \@bins_totct_indel_2ndmax_corr);

my @bins_diffct_indel_2ndmax_tl = my @bins_diffct_indel_2ndmax_cmplx = my @bins_diffct_indel_2ndmax_err = my @bins_diffct_indel_2ndmax_corr = ();
my @batch_diffct_indel_2ndmax = (\@bins_diffct_indel_2ndmax_tl, \@bins_diffct_indel_2ndmax_cmplx, \@bins_diffct_indel_2ndmax_err, \@bins_diffct_indel_2ndmax_corr);

my @bins_totlen_indel_2ndmax_tl = my @bins_totlen_indel_2ndmax_cmplx = my @bins_totlen_indel_2ndmax_err = my @bins_totlen_indel_2ndmax_corr = ();
my @batch_totlen_indel_2ndmax = (\@bins_totlen_indel_2ndmax_tl, \@bins_totlen_indel_2ndmax_cmplx, \@bins_totlen_indel_2ndmax_err, \@bins_totlen_indel_2ndmax_corr);

my @bins_difflen_indel_2ndmax_tl = my @bins_difflen_indel_2ndmax_cmplx = my @bins_difflen_indel_2ndmax_err = my @bins_difflen_indel_2ndmax_corr = ();
my @batch_difflen_indel_2ndmax = (\@bins_difflen_indel_2ndmax_tl, \@bins_difflen_indel_2ndmax_cmplx, \@bins_difflen_indel_2ndmax_err, \@bins_difflen_indel_2ndmax_corr);

batch_initializer (@batch_ct_clms_2ndmax, @BOUNDS_CT);
batch_initializer (@batch_totct_indel_2ndmax, @BOUNDS_CT);
batch_initializer (@batch_diffct_indel_2ndmax, @BOUNDS_PNCT);
batch_initializer (@batch_totlen_indel_2ndmax, @BOUNDS_CT);
batch_initializer (@batch_difflen_indel_2ndmax, @BOUNDS_PNCT);

		# END of "Modified (3) - 2nd maxima -." #

	## $raw_prpt_gaps = $ct_gaps/($CT_SEQS * $ct_clms);
	## $totct_indel = $ct_ins + $ct_del;
	## $diffct_indel = $ct_ins - $ct_del;
	## $totlen_indel = $totlen_ins + $totlen_del;
	## $difflen_indel = $totlen_ins - $totlen_del;


		# New - Minima and 2nd minima - .#

my @bins_diffct_indel_min_tl = my @bins_diffct_indel_min_cmplx = my @bins_diffct_indel_min_err = my @bins_diffct_indel_min_corr = ();
my @batch_diffct_indel_min = (\@bins_diffct_indel_min_tl, \@bins_diffct_indel_min_cmplx, \@bins_diffct_indel_min_err, \@bins_diffct_indel_min_corr);

my @bins_difflen_indel_min_tl = my @bins_difflen_indel_min_cmplx = my @bins_difflen_indel_min_err = my @bins_difflen_indel_min_corr = ();
my @batch_difflen_indel_min = (\@bins_difflen_indel_min_tl, \@bins_difflen_indel_min_cmplx, \@bins_difflen_indel_min_err, \@bins_difflen_indel_min_corr);

my @bins_diffct_indel_2ndmin_tl = my @bins_diffct_indel_2ndmin_cmplx = my @bins_diffct_indel_2ndmin_err = my @bins_diffct_indel_2ndmin_corr = ();
my @batch_diffct_indel_2ndmin = (\@bins_diffct_indel_2ndmin_tl, \@bins_diffct_indel_2ndmin_cmplx, \@bins_diffct_indel_2ndmin_err, \@bins_diffct_indel_2ndmin_corr);

my @bins_difflen_indel_2ndmin_tl = my @bins_difflen_indel_2ndmin_cmplx = my @bins_difflen_indel_2ndmin_err = my @bins_difflen_indel_2ndmin_corr = ();
my @batch_difflen_indel_2ndmin = (\@bins_difflen_indel_2ndmin_tl, \@bins_difflen_indel_2ndmin_cmplx, \@bins_difflen_indel_2ndmin_err, \@bins_difflen_indel_2ndmin_corr);

batch_initializer (@batch_diffct_indel_min, @BOUNDS_PNCT);
batch_initializer (@batch_difflen_indel_min, @BOUNDS_PNCT);
batch_initializer (@batch_diffct_indel_2ndmin, @BOUNDS_PNCT);
batch_initializer (@batch_difflen_indel_2ndmin, @BOUNDS_PNCT);


	#cluster_distributor ($diffct_indel_min, @{$bins_diffct_indel_min}, @BOUNDS_PNCT); # New.
	#cluster_distributor ($difflen_indel_min, @{$bins_difflen_indel_min}, @BOUNDS_PNCT); # New.
	#(defined $diffct_indel_2ndmin) and cluster_distributor ($diffct_indel_2ndmin, @{$bins_diffct_indel_2ndmin}, @BOUNDS_PNCT); # New.
	#(defined $difflen_indel_2ndmin) and cluster_distributor ($difflen_indel_2ndmin, @{$bins_difflen_indel_2ndmin}, @BOUNDS_PNCT); # New.

		# END of "New - Minima and 2nd minima - ." #

=cut # END of "For the predecessor 'summary_examine_complex.2.raw.pl'. (I)"


=For the predecessor 'summary_examine_purge.raw.pl'.
		### MODIFIED in creation of 'summary_examine_purge.raw.pl'. #

my @tags_ct = ('Intervals', 'block_size', '|shift|');

my @bins_size_block = my @bins_abs_shift = ();
my @batch_ct = (\@bins_size_block, \@bins_abs_shift);
batch_initializer (@batch_ct, @BOUNDS_CT);

my @bins_rlv_brlen = my @bins_pv = my @bins_diff_lnlk_respatt = ();
bin_initializer (@bins_rlv_brlen, @BOUNDS_BRLEN);
bin_initializer (@bins_pv, @BOUNDS_PRPT);
bin_initializer (@bins_diff_lnlk_respatt, @BOUNDS_DIFF_LNLK);

(my $bins_ct_invlvd_vs_brlen, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_BRLEN, "%5.3f");

		# ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016). #
(my $bins_blsize_vs_brlen, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_BRLEN, "%5.3f");
(my $bins_blsize_vs_ct_invlvd, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_CT, "%2d");
		# END of "ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016)." #

(my $bins_blsize_vs_pv, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_PRPT, "%4.2f");
(my $bins_ct_invlvd_vs_pv, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_PRPT, "%4.2f");
(my $bins_brlen_vs_pv, undef) = init_2dim_distri (@BOUNDS_BRLEN, @BOUNDS_PRPT, "%4.2f");
=cut

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
		# MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #

our @BROAD_TYPES = ('Shift', 'Complex', 'Merge', 'Purge', 'Split', 'Ex-nihilo', 'NB_Merge', 'NB_Split', 'CII', 'CEII', 'pNB_Merge', 'pNB_Split', 'iCII', 'iCEII', 'Other');
##our @BROAD_TYPES = ('Shift', 'Complex', 'Merge', 'Purge', 'Split', 'Ex-nihilo', 'NB_Merge', 'NB_Split', 'CII', 'CEII', 'Other');

	#  MODIFIED in creation of 'summary_classify_msa_errors1.raw.pl' (on Oct 26 (Mon), 2015). #
my @tags_ct = ('Intervals', '#{equiv_indels}', '#{cblks}_all', '#{cblks}_eff', 'Totlen{cblks}_eff', 'Len{cblk}_eff');
##my @tags_ct = ('Intervals', '#{equiv_indels}', '#{cblks}_eff', 'Totlen{cblks}_eff', 'Len{cblk}_eff');
my @tags_prpt = ('Intervals', @BROAD_TYPES);
my @tags_cblk_len = ('Intervals', @BROAD_TYPES);


	#  MODIFIED in creation of 'summary_classify_msa_errors1.raw.pl' (on Oct 26 (Mon), 2015). #
my @bins_ct_equiv_indels = my @bins_ct_all_cblks = my @bins_ct_eff_cblks = my @bins_totlen_eff_cblks = my @bins_len_eff_cblk = ();
my @batch_ct = (\@bins_ct_equiv_indels, \@bins_ct_all_cblks, \@bins_ct_eff_cblks, \@bins_totlen_eff_cblks, \@bins_len_eff_cblk);
##my @bins_ct_equiv_indels = my @bins_ct_eff_cblks = my @bins_totlen_eff_cblks = my @bins_len_eff_cblk = ();
##my @batch_ct = (\@bins_ct_equiv_indels, \@bins_ct_eff_cblks, \@bins_totlen_eff_cblks, \@bins_len_eff_cblk);
batch_initializer (@batch_ct, @BOUNDS_CT);

my @batch_prpt = my @batch_cblk_len = ();
my %brtype2bins_prpt;
my %brtype2bins_cblk_len;
my %brtype2bins_ct_vs_prpt_ecblk;
foreach my $brtype (@BROAD_TYPES) {
	my @bins_prpt = my @bins_cblk_len = ();
	push @batch_prpt, \@bins_prpt;
	$brtype2bins_prpt{$brtype} = \@bins_prpt;
	push @batch_cblk_len, \@bins_cblk_len;
	$brtype2bins_cblk_len{$brtype} = \@bins_cblk_len;

	(my $bins_ct_vs_prpt_ecblk, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_PRPT, "%1d");
	$brtype2bins_ct_vs_prpt_ecblk{$brtype} = $bins_ct_vs_prpt_ecblk;
}
batch_initializer (@batch_prpt, @BOUNDS_PRPT);
batch_initializer (@batch_cblk_len, @BOUNDS_CT);

=cut # END of "DISABLED in creation of 'summary_examine_purge.raw.pl'."




=DISABLED in creation of 'summary_examine_purge.raw.pl'.
			# ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
our @TAGS_OUT_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', '#{cblks}_eff');
foreach my $brtype (@BROAD_TYPES) { push @TAGS_OUT_FLD, "#{$brtype}"; }
=cut


=DISABLED in creation of 'summary_classify_msa_errors.raw.pl' (I). 
	### Preliminary processes. Store indices of segments touching the left-end or right-end of the MSA ###

my %session2end_seg_indices; # = ("$id_class:$session_id" => [$indx_seg_leftend, $indx_seg_rightend], ...)

foreach my $id_class (@{$list_id_classes}) { # Preliminary outer foreach-loop (over classes). #

   my $auxdir = join ('', $top_auxdir, '/', $mid_auxdir_pref, $id_class, $mid_auxdir_suff, '/', $mid_aux_path);
   unless (( -d $auxdir ) and ( -r $auxdir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable auxiliary directory '$auxdir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
   }

    my $sessions = list_sessions_ext ($auxdir, $auxfile_pref, $auxfile_suff, '-r');
    unless (defined $sessions) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the auxiliary directory '$auxdir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
    my $ct_sessions = @{$sessions};
    (print STDOUT "\nThe class '$id_class' has $ct_sessions sessions: ", join (', ', $sessions->[0], $sessions->[1], '...', $sessions->[$ct_sessions-1]), " .\n\n") if ($TEST_LEVEL>2);

    foreach my $session (@{$sessions}) { # preliminary middle foreach-loop (over sessions). #

   (print STDOUT "{ Session: '$session' }\n\n") if ($TEST_LEVEL>3);

	my $auxfile = join ('', $auxdir, '/', $auxfile_pref, $session, $auxfile_suff);

		# $auxfile = "$top_auxdir/$mid_auxdir_pref$id_class$mid_auxdir_suff/$mid_aux_path/$auxfile_pref$session_id$auxfile_suff",
	unless (( -f $auxfile ) and ( -r $auxfile )) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable auxiliary file '$auxfile' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ;
	}
	local $/ = "\n";

	my $fh_aux = open_input_file2 ($auxfile);
	unless (defined $fh_aux) {
	    print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the auxiliary file '$auxfile' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ;
	}
	my $aux_header = <$fh_aux>;

	my @seg_indices = ();
	while (my $line = <$fh_aux>) { # Preliminary inner while-loop (over unmatched segments). #
	   my ($indx) = ($line =~ m/^(\d+)\s/);
	   (defined $indx) and (push @seg_indices, $indx);
	}			# End of the preliminary inner while-loop (over unmatched segments). #
	close ($fh_aux);

	unless (@seg_indices>0) {
	print STDOUT "[WARNING] The session '$session' in the class '$id_class' has NO ERRONEOUS SEGMENTS: $! !\n";
	   next;
	}

	my ($indx_le, $indx_re) = @seg_indices[0, $#seg_indices];

	$session2end_seg_indices{"$id_class:$session"} = [$indx_le, $indx_re];
	

    }			# End of the preliminary middle foreach-loop (over sessions). #

}			# End of the preliminary outer foreach-loop (over classes). #

=cut # END of "DISABLED in creation of 'summary_classify_msa_errors.raw.pl' (I)."


=DISABLED in creation of 'summary_classify_msa_errors.raw.pl' (II).

	# Preliminary Process 2: Store information on the segments containing non-parsimonious gapped segments. #
	# (ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'.)
$/ = "\n"; # Use a newline character as an input record separator. #

my %seg2ct_nonpars; # ("$id_class:$session:$seg_id" => $ct_nonpars_gpd_segs, ...) for ONLY non-parsimonious segments. #

my $fh_aux2 = open_input_file2 ($auxfile2);
unless (defined $fh_aux2) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the auxiliary file '$auxfile2' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my $aux_header2 = <$fh_aux2>;

my $indices_fld_aux = tfm_labels2indices_gen (@TAGS_AUX_FLD, $aux_header2, "\t");
unless (defined $indices_fld_aux) {
	$aux_header2 =~ s/\s+$//; 
	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the auxiliary header:\n";
	print STDOUT " '$aux_header2' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
my $max_indx_aux = 0;
foreach my $indx (@{$indices_fld_aux}) { ($max_indx_aux <$indx) and ($max_indx_aux = $indx); }

while (my $in_aux = <$fh_aux2>) { # 2nd preliminary while-loop (over erroneous segments). #

	my @flds = split /\s+/, $in_aux;
	(@flds>$max_indx_aux) or next;

	my ($id_class, $session, $seg_id, $ct_nonpars_gpd_segs) = @flds[@{$indices_fld_aux}];

		# @TAGS_AUX_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', '#{nonpars_gpd_segs}');

	($ct_nonpars_gpd_segs>0) or next;

	$seg2ct_nonpars{"$id_class:$session:$seg_id"} = $ct_nonpars_gpd_segs;

}			# End of the 2nd preliminary while-loop (over erroneous segments). #

print STDOUT "\n ... Retrieved information on the ", scalar (keys %seg2ct_nonpars), " non-parsimonious erroneous segments ...\n\n\n";

=cut # END of "DISABLED in creation of 'summary_classify_msa_errors.raw.pl' (II)."


=DISABLED in creation of 'summary_examine_purge.raw.pl'.

	# Prelliminary Process 3: Store Complete-likelihoods of and #{null columns} in the reference & reconstructed MSAs of each segment.
	# (ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'.)

$/ = "\n";

my %seg_label2aux_info3; # = ( "$id_class:$session:$segment" => [$totsc_ref, $totsc_ref, $ct_null_ref, $ct_null_rec], ...) 

my $fh_aux3 = open_input_file2 ($auxfile3);
unless (defined $fh_aux3) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the auxiliary file '$auxfile3' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my $aux_header3 = <$fh_aux3>;

my $indices_fld_aux3 = tfm_labels2indices_gen (@TAGS_AUX3_FLD, $aux_header3, "\t");
unless (defined $indices_fld_aux3) {
	$aux_header3 =~ s/\s+$//; 
	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the auxiliary header:\n";
	print STDOUT " '$aux_header3' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
my $max_indx_aux3 = 0;
foreach my $indx (@{$indices_fld_aux3}) { ($max_indx_aux3 <$indx) and ($max_indx_aux3 = $indx); }

while (my $in_aux = <$fh_aux3>) { # 3rd preliminary while-loop (over erroneous segments). #

	my @flds = split /\s+/, $in_aux;
	(@flds>$max_indx_aux3) or next;

	my ($id_class, $session, $seg_id, $totsc_ref, $totsc_rec, $ct_null_ref, $ct_null_rec) = @flds[@{$indices_fld_aux3}];

		# @TAGS_AUX3_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Total_Score.ref',  'Total_Score.rec', 'Av#{null_columns}.ref', 'Av#{null_columns}.rec');

	$seg_label2aux_info3{"$id_class:$session:$seg_id"} = [$totsc_ref, $totsc_rec, $ct_null_ref, $ct_null_rec];

}			# End of the 3rd preliminary while-loop (over erroneous segments). #

print STDOUT "\n ... Retrieved complete-likelihoods and #{null_columns} of reference and reconstructed MSAs in the ", scalar (keys %seg_label2aux_info3), " erroneous segments ...\n\n\n";

	# END of "(ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'.)"

=cut # END of "DISABLED in creation of 'summary_examine_purge.raw.pl'."



	### Preliminary Process 5: Readin the list of subject segments. ###
	### (IMPORTED from 'bch_pbs_examine_purges.D.pl', and MODIFIED in creation of 'summary_examine_complex.1.raw.pl'.)

my %if_complex; # = ("$id_class:$in_msa_id:$segid" => 1, ...) for segments with complex errors.
my %if_na; # = ("$id_class:$in_msa_id:$segid" => 1, ...) for segments whose error types are 'N/A'.

if (1) {

   my $fh_aux5 = open_input_file2 ($auxfile5);
   unless (defined $fh_aux5) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the auxiliary input file '$auxfile5' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
   }
   local $/ = "\n"; # Use the newline character as the input-record-separator. #

   my $aux_header5 = <$fh_aux5>;

   my @TAGS_AUX5_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Types;of;associated;errors');

   my $indices_fld_aux5 = tfm_labels2indices_gen (@TAGS_AUX5_FLD, $aux_header5, "\t");
   unless (defined $indices_fld_aux5) {
   	$aux_header5 =~ s/\s+$//; 
   	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the auxiliary header:\n";
   	print STDOUT " '$aux_header5' !\n";
   	print STDERR "[ERROR '$program_name'] $!\n";
   	die ;
   }
   my $max_indx_aux5 = 0;
   foreach my $indx (@{$indices_fld_aux5}) { ($max_indx_aux5 <$indx) and ($max_indx_aux5 = $indx); }

   while (my $line = <$fh_aux5>) { # Preliminary while-loop. #

	my @fields = split /\t/, $line;
	(@fields>$max_indx_aux5) or next;
	##(@fields>3) or next;

	my ($id_class, $in_msa_id, $segid, $error_type) = @fields[@{$indices_fld_aux5}];
	##my ($id_class, $in_msa_id, $segid) = @fields[0,1,2];

	my $combo_id = "$id_class:$in_msa_id:$segid";

	if ($error_type eq 'N/A') {
	    $if_na{$combo_id} = 1;
	} elsif ($error_type =~ m/Complex/) {
	    $if_complex{$combo_id} = 1;
	}

=For the source 'bch_pbs_examine_purges.D.pl'
	my $combo_id = "$id_class:$in_msa_id";
	if (defined $class_msa_id2sbjct_segs{$combo_id}) {
	    push @{$class_msa_id2sbjct_segs{$combo_id}}, $segid;
	} else {
	    $class_msa_id2sbjct_segs{$combo_id} = [$segid];
	}
=cut

   }			# End of the preliminary while-loop. #
   close ($fh_aux5);
}

	### END of "Preliminary Process 5: Readin the list of subject segments." ###



	### Prepare the Input/Output files ###

my $infile1 = "$indir1/$infile_name1";
##my $outfile = "$outdir/$outfile_name"; ## DISABLED in creation of 'summary_examine_purge.raw.pl'.

my $fh_in1 = open_input_file2 ($infile1);
unless (defined $fh_in1) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the input file '$infile1' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
my $fh_out = open_output_file ($outfile, 1);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
=cut

	### Main Processes ###
$/ = "\n"; # Use a newline character as an input record separater.

my $in_header1 = <$fh_in1>;
my $indices_fld = tfm_labels2indices_gen (@TAGS_IN_FLD, $in_header1, "\t");
unless (defined $indices_fld) {
	$in_header1 =~ s/\s+$//; 
	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the input header:\n";
	print STDOUT " '$in_header1' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
my $max_indx = 0;
foreach my $indx (@{$indices_fld}) { ($max_indx <$indx) and ($max_indx = $indx); }


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
print $fh_out join ("\t", @TAGS_OUT_FLD), "\n\n";
=cut


my $totct_segs = my $ct_anom = my $ct_w_undef = my $ct_fail = my $ct_succ = 0;
my $ct_skipped1 = my $ct_skipped2 = my $ct_anom2 = 0;
my $ct_skipped3 = my $ct_skipped5  = 0;

my $ct_skipped_x = 0;

my $ct_anom3 = 0; # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
my $totct_all_cblks = 0; # ADDED in creation of 'summary_classify_msa_errors1.raw.pl' (on Oct 26 (Mon), 2015). #
my $totct_equiv_indels = my $totct_eff_cblks = 0; # ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
##my %type_error2sbtotct_ecblks; # = ($type_error => $sub_total_ct_eff_cblks, ...) # ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. 
=cut

=For the predecessor 'summary_examine_purge.raw.pl'.

	## ADDED in creation of 'summary_examine_purge.raw.pl'. ##

my $ct_w_fractional = 0;

my %br2freq_prgs; # = ($br_id => $abs_freq_purges, ...)
my %ct_invlvd2freq_prgs; # = ($ct_invlvd_seqs => $abs_freq_purges, ...)

	## END of "ADDED in creation of 'summary_examine_purge.raw.pl'." ##

	## ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016). ##

my %blsize2bins_ct_invlvd_vs_pv; # = ($block_size => \@bins_ct_invlvd_vs_pv_sp, ...) (Here, '_sp' indicates that the distribution is for a specific $block_size.) # 
my %blsize2bins_brlen_vs_pv; # = ($block_size => \@bins_brlen_vs_pv_sp, ...)

	## END Of "ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016)." ##

=cut # END of "For the predecessor 'summary_examine_purge.raw.pl'."

			### MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. ###

my @last_cluster = ();
while (my $in = <$fh_in1>) {
	($in =~ m/^\S/) or next;
	$in =~ s/\s+$//;
	my @in_flds = split /\t/, $in;
	unless (@in_flds > $max_indx) {
	    print STDOUT "[WARNING '$program_name'] An ANOMALOUS line was found:\n";
	    print STDOUT "    '$in' : $! !\n\n";
	    $ct_anom++;
	    next;
	}
	$totct_segs++;
	my @rlv_flds = @in_flds[@{$indices_fld}];
	push @last_cluster, \@rlv_flds;
	last;
}

my $ct_clusters = 0;
my $end_coord_last_cls ;
my $last_id_class = my $last_session = 'Pre-start'; # ADDED for DEBUGGING. #

while (@last_cluster>0) { # Outer while-loop (until the last cluster runs out). #

			### END of "MODIFIED in creation of 'summary_examine_complex.2.raw.pl'." ###

=DISABLED in creation of 'summary_examine_complex.2.raw.pl'.
while (my $in = <$fh_in1>) { # Outer while-loop (over erroneous MSA segments) #
=cut

	## ($totct_segs>=10) and last; ### For TESTING the script ###

	my @in_flds = (); # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

    while (my $in = <$fh_in1>) { # Middle preliminary while-loop (over erroneous MSA segments). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	($in =~ m/^\S/) or next;
	$in =~ s/\s+$//;

	##$totct_segs++;

	@in_flds = split /\t/, $in; # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #
	##my @in_flds = split /\t/, $in;
	unless (@in_flds > $max_indx) {
	    print STDOUT "[WARNING '$program_name'] An ANOMALOUS line was found:\n";
	    print STDOUT "    '$in' : $! !\n\n";
	    $ct_anom++;
	    @in_flds = (); # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #
	    next;
	}

	$totct_segs++; # MOVED from above. #

	last; # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

    } 		# End of the middle preliminary while-loop (over erroneous MSA segments). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	my $last_combo_id = join (':', @{$last_cluster[0]}[0,1,2]); # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	my @rlv_flds = (@in_flds>0) ? @in_flds[@{$indices_fld}] : () ; # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #


			# MAJOR MODIFICATION in creation of 'summary_examine_complex.1.raw.pl'. #

=To be moved below in creation of 'summary_examine_complex.2.raw.pl'.
	my ($id_class, $session, $segid, $type, $diagnosis, $ct_gpd_segs, $indx_gpd_seg, $beg_coord, $end_coord, $ct_clms, $ct_ins, $ct_del, $totlen_ins, $totlen_del, $ct_gaps) = @in_flds[@{$indices_fld}];

		## @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}');
=cut


		## Check if the current segment belongs to the last cluster or not. ##
		## (ADDED in creation of 'summary_examine_complex.2.raw.pl'.) ##

   ##if (@rlv_flds>0) { # DISABLED for DEBUGGING. #

	my $combo_id = (@rlv_flds>0) ? join (':', @rlv_flds[0,1,2]) : undef; # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #
	##my $combo_id = "$id_class:$session:$segid";

    if (defined $combo_id) { # ADDED for DEBUGGING. #
	if ($last_combo_id eq $combo_id) { # Incorporate the current segment into the last cluster.
	    push @last_cluster, \@rlv_flds;
	    next;
	}
    } # ADDED for DEBUGGING. #
    ##} # DISABLED for DEBUGGING. #

	$ct_clusters++;

		## END of "Check if the current segment belongs to the last cluster or not." ##

		## Initialize the important attributes of the last cluster. ##
		## (ADDED in creation of 'summary_examine_complex.2.raw.pl'.) ##

	my ($id_class, $session, $segid, $type, $diagnosis, $ct_gpd_segs, $beg_coord_cls) = @{$last_cluster[0]}[0..5,7];
	my $end_coord_cls = $last_cluster[$#last_cluster]->[8];

	unless (@last_cluster == $ct_gpd_segs) { # A Consistency check. #

	    print STDOUT "[CAUTION '$program_name'] The size of the last cluster (=", scalar (@last_cluster), ") CLASHES with #{gapped segments in it} (=$ct_gpd_segs) for (ID_Class, Session_ID, Segment_ID) = ($id_class, $session, $segid) : $! ! ... I will skip the analysis ...\n\n";
	    $ct_anom3++;
	    if (@rlv_flds>0) {
		@last_cluster = (\@rlv_flds);
		$end_coord_last_cls = $end_coord_cls;
		($last_id_class, $last_session) = ($id_class, $session); # ADDED for DEBUGGING. #
	    } else {
		@last_cluster = ();
	    }
	    next;
	}		# END of "A Consistency check." #


=For the predecessor 'summary_examine_complex.2.raw.pl'. (ii)
			# Calculate the span of this cluster of gapped segments. #

	my $beg_coord_next_cls = (@rlv_flds>0) ? $rlv_flds[7] : undef;

	my $span_cluster;

	my ($id_class1, $session1) = @rlv_flds[0,1]; # ADDED for DEBUGGING. #

	if (($last_id_class eq $id_class) and ($id_class eq $id_class1) and ($last_session eq $session) and ($session eq $session1) and (defined $end_coord_last_cls) and (defined $beg_coord_next_cls)) { # Further MODIFIED for DEBUGGING. #
	##if (($id_class eq $id_class) and ($session eq $session1) and (defined $end_coord_last_cls) and (defined $beg_coord_next_cls)) { # MODIFIED for DEBUGGING. # ...}
	##if ((defined $end_coord_last_cls) and (defined $beg_coord_next_cls)) {...}
	
	   my $coord_left_border = ($beg_coord_cls + $end_coord_last_cls)/2 ;
	   my $coord_right_border = ($end_coord_cls + $beg_coord_next_cls)/2 ;
	   $span_cluster = $coord_right_border - $coord_left_border ;

	   if ($span_cluster < 0) { # Alarm if the span is negative. #
		print STDOUT "[WARNING '$program_name'] The cluster '$combo_id' has a NEGATIVE SPAN (=$span_cluster):  (end_last_cls, beg_cls, end_cls, beg_next_cls) = ($end_coord_last_cls, $beg_coord_cls, $end_coord_cls, $beg_coord_next_cls): $! !\n\n";
	   }
	}

			# END of "Calculate the span of this cluster of gapped segments." #

=cut # END of "For the predecessor 'summary_examine_complex.2.raw.pl'. (ii)"

		### ADDED in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

	my $max_totct_indel = 0;
	my $maxct_clms = my $max_totlen_indel = 0;
	my ($max_diffct_indel, $min_diffct_indel);

    foreach my $segment_info (@last_cluster) { # Main middle foreach-loop (over segments in the last cluster). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	my ($id_class1, $session1, $segid1, $type1, $diagnosis1, $ct_gpd_segs1, $indx_gpd_seg, $beg_coord, $end_coord, $ct_clms, $ct_ins, $ct_del, $totlen_ins, $totlen_del, $ct_gaps) = @{$segment_info}; # MOVED from above in creation of 'summary_examine_complex.2.raw.pl'. #

	my $totct_indel = $ct_ins + $ct_del;
	my $diffct_indel = $ct_ins - $ct_del;
	my $totlen_indel = $totlen_ins + $totlen_del;
	my $difflen_indel = $totlen_ins - $totlen_del;

	if ($totct_indel > $max_totct_indel) {

		# Update the set of "largest" segments. #

	    $max_totct_indel = $totct_indel;

	    $maxct_clms = $ct_clms;
	    $max_totlen_indel = $totlen_indel;
	    $max_diffct_indel = $min_diffct_indel = $diffct_indel;

	} elsif ($totct_indel == $max_totct_indel) {

		# Incorporate the current segment into the set of "largest" segments. # 

	    ($ct_clms > $maxct_clms) and ($maxct_clms = $ct_clms);
	    ($totlen_indel > $max_totlen_indel) and ($max_totlen_indel = $totlen_indel);
	    ((defined $max_diffct_indel) and ($max_diffct_indel >= $diffct_indel)) or ($max_diffct_indel = $diffct_indel);
	    ((defined $min_diffct_indel) and ($min_diffct_indel <= $diffct_indel)) or ($min_diffct_indel = $diffct_indel);

	}

   }			# End of the main middle foreach-loop (over segments in the last cluster). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

=For the predecessor 'summary_examine_complex.2.raw.pl'. (IIa)

	my $totct_clms = my $tot_totct_indel = my $tot_diffct_indel = my $tot_totlen_indel = my $tot_difflen_indel = my $totct_gaps = 0;

	my @cts_clms = my @totcts_indel = my @diffcts_indel = my @totlens_indel = my @difflens_indel = my @cts_gaps = ();

		## END of "Initialize the important attributes of the last cluster." ##


    foreach my $segment_info (@last_cluster) { # Main middle foreach-loop (over segments in the last cluster). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	my ($id_class1, $session1, $segid1, $type1, $diagnosis1, $ct_gpd_segs1, $indx_gpd_seg, $beg_coord, $end_coord, $ct_clms, $ct_ins, $ct_del, $totlen_ins, $totlen_del, $ct_gaps) = @{$segment_info}; # MOVED from above in creation of 'summary_examine_complex.2.raw.pl'. #

	##my $raw_prpt_gaps = $ct_gaps/($CT_SEQS * $ct_clms); ##  For the predecessor 'summary_examine_complex.1.raw.pl'. #

	my $totct_indel = $ct_ins + $ct_del;
	my $diffct_indel = $ct_ins - $ct_del;
	my $totlen_indel = $totlen_ins + $totlen_del;
	my $difflen_indel = $totlen_ins - $totlen_del;

		# Tally the attributes of the constituent gapped segments. #
		# (ADDED in creation of 'summary_examine_complex.2.raw.pl'.) #

	$totct_clms += $ct_clms;
	$tot_totct_indel += $totct_indel;
	$tot_diffct_indel += $diffct_indel;
	$tot_totlen_indel += $totlen_indel;
	$tot_difflen_indel += $difflen_indel;
	$totct_gaps += $ct_gaps;

	push @cts_clms, $ct_clms;
	push @totcts_indel, $totct_indel;
	push @diffcts_indel, $diffct_indel;
	push @totlens_indel, $totlen_indel;
	push @difflens_indel, $difflen_indel;
	push @cts_gaps, $ct_gaps;

		# END of "Tally the attributes of the constituent gapped segments." #

   }			# End of the main middle foreach-loop (over segments in the last cluster). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

=cut # END of "For the predecessor 'summary_examine_complex.2.raw.pl'. (IIa)"

		## Update @last_cluster. ##
		## (ADDED in creation of 'summary_examine_complex.2.raw.pl'.)

	if (@rlv_flds>0) {
	    @last_cluster = (\@rlv_flds);
	    $end_coord_last_cls = $end_coord_cls;
	    ($last_id_class, $last_session) = ($id_class, $session); # ADDED for DEBUGGING. #
	} else {
	    @last_cluster = ();
	}
		## END of "Update @last_cluster." ##


=For the predecessor 'summary_examine_complex.2.raw.pl'. (IIb)
		## Calculate the maximum and the 2nd maximum of each attribute. ##
		## (ADDED in creation of 'summary_examine_complex.2.raw.pl'.)

	my @std_cts_clms = sort {$b <=> $a} @cts_clms;
	my ($ct_clms_max, $ct_clms_2ndmax) = @std_cts_clms[0,1];
	my @std_totcts_indel = sort {$b <=> $a} @totcts_indel;
	my ($totct_indel_max, $totct_indel_2ndmax) = @std_totcts_indel[0,1];
	my @std_totlens_indel = sort {$b <=> $a} @totlens_indel;
	my ($totlen_indel_max, $totlen_indel_2ndmax) = @std_totlens_indel[0,1];

	my @std_diffcts_indel = sort {$b <=> $a} @diffcts_indel;
	my ($diffct_indel_max, $diffct_indel_min) = @std_diffcts_indel[0, $#diffcts_indel];
	my ($diffct_indel_2ndmax, $diffct_indel_2ndmin) = (@diffcts_indel>2) ? @std_diffcts_indel[1, $#diffcts_indel -1] : (undef, undef) ;

	my @std_difflens_indel = sort {$b <=> $a} @difflens_indel;
	my ($difflen_indel_max, $difflen_indel_min) = @std_difflens_indel[0, $#difflens_indel];
	my ($difflen_indel_2ndmax, $difflen_indel_2ndmin) = (@difflens_indel>2) ? @std_difflens_indel[1, $#difflens_indel -1] : (undef, undef) ;

		## END of "Calculate the maximum and the 2nd maximum of each attribute." ##

=cut # END of "For the predecessor 'summary_examine_complex.2.raw.pl'. (IIb)"

=For the predecessor 'summary_examine_purge_ctrl.raw.pl'.
			# MAJOR MODIFICATION in creation of 'summary_examine_purge_ctrl.raw.pl' (on May 16, 2016). #

	my ($id_class, $session, $segment, $wd_size, $indx_wd, $rlv_br_id, $rlv_br_len, $ct_ext_offs, $d_sbst_exp, $d_sbst_obs, $pv) = @in_flds[@{$indices_fld}];

		## @TAGS_IN_FLD = ( 'ID_Class', 'Session_ID', 'Segment_ID', 'Wd_size', 'indx_wd', 'br_ID', 'br_len', '#{ext_offs}', '#{diffs}_exp', '#{diffs}_obs', 'P-value');

	my $size_block = $wd_size;
	my $ct_invlvd_seqs = $ct_ext_offs;
=cut


=DISABLED in creation of 'summary_examine_purge_ctrl.raw.pl'.
	unless (($size_block == int ($size_block) ) and ($abs_shift == int ($abs_shift)) ) {
	    $ct_w_fractional++;
	    next;
	}
=cut


			### MAJOR ADDITION in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

			# Default: Type B = 'non-complex errors' + 'correctly reconstructed'. #

	my ($bins_max_totct_indels_vs_maxct_columns, $bins_max_totct_indels_vs_max_totlen_indels,
		$bins_max_totct_indels_vs_max_diffct_indels, $bins_max_totct_indels_vs_min_diffct_indels)
	   = ($bins_max_totct_indels_vs_maxct_columns_B, $bins_max_totct_indels_vs_max_totlen_indels_B,
		$bins_max_totct_indels_vs_max_diffct_indels_B, $bins_max_totct_indels_vs_min_diffct_indels_B) ;

	if ($type eq 'Correct') {

		# Do nothing. #

	} elsif ((defined $if_complex{$last_combo_id}) or ((defined $if_na{$last_combo_id}) or ($diagnosis =~ m/TOO_LONG/)) ) {

		# Type A = 'Complex errors' + 'Too-long' #

	    ($bins_max_totct_indels_vs_maxct_columns, $bins_max_totct_indels_vs_max_totlen_indels,
		$bins_max_totct_indels_vs_max_diffct_indels, $bins_max_totct_indels_vs_min_diffct_indels)
	     = ($bins_max_totct_indels_vs_maxct_columns_A, $bins_max_totct_indels_vs_max_totlen_indels_A,
		$bins_max_totct_indels_vs_max_diffct_indels_A, $bins_max_totct_indels_vs_min_diffct_indels_A) ;

	} elsif ($type eq 'Erroneous') {

		# Do nothing. #

	} else {
	    print STDOUT "[CAUTION '$program_name'] (ID_Class, Session_ID, Segment_ID) = ($id_class, $session, $segid) has either ANOMALOUS TYPE (=$type) or ANOMALOUS DIAGNOSIS (=$diagnosis): $! ! ... I will skip tallying this segment ...\n\n";

	    $ct_anom2++;
	    next;
	}

			# Take the 2D distributions. #

	distributor_2dim ($max_totct_indel, $maxct_clms, @{$bins_max_totct_indels_vs_maxct_columns}, @BOUNDS_CT0, @BOUNDS_CT);
	distributor_2dim ($max_totct_indel, $max_totlen_indel, @{$bins_max_totct_indels_vs_max_totlen_indels}, @BOUNDS_CT0, @BOUNDS_CT);
	distributor_2dim ($max_totct_indel, $max_diffct_indel, @{$bins_max_totct_indels_vs_max_diffct_indels}, @BOUNDS_CT0, @BOUNDS_PNCT0);
	distributor_2dim ($max_totct_indel, $min_diffct_indel, @{$bins_max_totct_indels_vs_min_diffct_indels}, @BOUNDS_CT0, @BOUNDS_PNCT0);

			### RESTART FROM HERE ###

=Just a reminder... 
(my $bins_max_totct_indels_vs_maxct_columns_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_maxct_columns_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_totlen_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_max_totlen_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_max_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

(my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

=cut # END of "Just a reminder..."


=For the predecessor 'summary_examine_complex.2.raw.pl'. (III)

			### MAJOR ADDITION in creation of 'summary_examine_complex.1.raw.pl'. ###
			### => Substantially MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. ###

			# The default setting. #

	my ($bins_ct_gpd_segs, $bins_span_cluster) = (\@bins_ct_gpd_segs_err, \@bins_span_cluster_err); # Totally new. #

	my ($bins_totct_clms, $bins_tot_totct_indel, $bins_tot_diffct_indel, $bins_tot_totlen_indel, $bins_tot_difflen_indel)
	  = (\@bins_totct_clms_err, \@bins_tot_totct_indel_err, \@bins_tot_diffct_indel_err, \@bins_tot_totlen_indel_err, \@bins_tot_difflen_indel_err); # Modified.

	my ($bins_ct_clms_max, $bins_totct_indel_max, $bins_diffct_indel_max, $bins_totlen_indel_max, $bins_difflen_indel_max)
	  = (\@bins_ct_clms_max_err, \@bins_totct_indel_max_err, \@bins_diffct_indel_max_err, \@bins_totlen_indel_max_err, \@bins_difflen_indel_max_err); # Modified.

	my ($bins_ct_clms_2ndmax, $bins_totct_indel_2ndmax, $bins_diffct_indel_2ndmax, $bins_totlen_indel_2ndmax, $bins_difflen_indel_2ndmax)
	  = (\@bins_ct_clms_2ndmax_err, \@bins_totct_indel_2ndmax_err, \@bins_diffct_indel_2ndmax_err, \@bins_totlen_indel_2ndmax_err, \@bins_difflen_indel_2ndmax_err); # Modified.

	my ($bins_diffct_indel_min, $bins_diffct_indel_2ndmin, $bins_difflen_indel_min, $bins_difflen_indel_2ndmin)
	 = (\@bins_diffct_indel_min_err, \@bins_diffct_indel_2ndmin_err, \@bins_difflen_indel_min_err, \@bins_difflen_indel_2ndmin_err) ; # New.


	if ($type eq 'Correct') {

	    ##($bins_ct_clms, $bins_prpt_gps, $bins_totct_indel, $bins_diffct_indel, $bins_totlen_indel, $bins_difflen_indel)
	    ##= (\@bins_ct_clms_corr, \@bins_prpt_gps_corr, \@bins_totct_indel_corr, \@bins_diffct_indel_corr, \@bins_totlen_indel_corr, \@bins_difflen_indel_corr);

	    ($bins_ct_gpd_segs, $bins_span_cluster) = (\@bins_ct_gpd_segs_corr, \@bins_span_cluster_corr); # Totally new. #

	    ($bins_totct_clms, $bins_tot_totct_indel, $bins_tot_diffct_indel, $bins_tot_totlen_indel, $bins_tot_difflen_indel)
	    = (\@bins_totct_clms_corr, \@bins_tot_totct_indel_corr, \@bins_tot_diffct_indel_corr, \@bins_tot_totlen_indel_corr, \@bins_tot_difflen_indel_corr); # Modified.

	    ($bins_ct_clms_max, $bins_totct_indel_max, $bins_diffct_indel_max, $bins_totlen_indel_max, $bins_difflen_indel_max)
	    = (\@bins_ct_clms_max_corr, \@bins_totct_indel_max_corr, \@bins_diffct_indel_max_corr, \@bins_totlen_indel_max_corr, \@bins_difflen_indel_max_corr); # Modified.

	    ($bins_ct_clms_2ndmax, $bins_totct_indel_2ndmax, $bins_diffct_indel_2ndmax, $bins_totlen_indel_2ndmax, $bins_difflen_indel_2ndmax)
	    = (\@bins_ct_clms_2ndmax_corr, \@bins_totct_indel_2ndmax_corr, \@bins_diffct_indel_2ndmax_corr, \@bins_totlen_indel_2ndmax_corr, \@bins_difflen_indel_2ndmax_corr); # Modified.

	    ($bins_diffct_indel_min, $bins_diffct_indel_2ndmin, $bins_difflen_indel_min, $bins_difflen_indel_2ndmin)
	    = (\@bins_diffct_indel_min_corr, \@bins_diffct_indel_2ndmin_corr, \@bins_difflen_indel_min_corr, \@bins_difflen_indel_2ndmin_corr) ; # New.


	} elsif (defined $if_complex{$last_combo_id}) { # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #
	##} elsif (defined $if_complex{$combo_id}) {...}

	    ##($bins_ct_clms, $bins_prpt_gps, $bins_totct_indel, $bins_diffct_indel, $bins_totlen_indel, $bins_difflen_indel)
	    ##= (\@bins_ct_clms_cmplx, \@bins_prpt_gps_cmplx, \@bins_totct_indel_cmplx, \@bins_diffct_indel_cmplx, \@bins_totlen_indel_cmplx, \@bins_difflen_indel_cmplx);

	    ($bins_ct_gpd_segs, $bins_span_cluster) = (\@bins_ct_gpd_segs_cmplx, \@bins_span_cluster_cmplx); # Totally new. #

	    ($bins_totct_clms, $bins_tot_totct_indel, $bins_tot_diffct_indel, $bins_tot_totlen_indel, $bins_tot_difflen_indel)
	    = (\@bins_totct_clms_cmplx, \@bins_tot_totct_indel_cmplx, \@bins_tot_diffct_indel_cmplx, \@bins_tot_totlen_indel_cmplx, \@bins_tot_difflen_indel_cmplx); # Modified.

	    ($bins_ct_clms_max, $bins_totct_indel_max, $bins_diffct_indel_max, $bins_totlen_indel_max, $bins_difflen_indel_max)
	    = (\@bins_ct_clms_max_cmplx, \@bins_totct_indel_max_cmplx, \@bins_diffct_indel_max_cmplx, \@bins_totlen_indel_max_cmplx, \@bins_difflen_indel_max_cmplx); # Modified.

	    ($bins_ct_clms_2ndmax, $bins_totct_indel_2ndmax, $bins_diffct_indel_2ndmax, $bins_totlen_indel_2ndmax, $bins_difflen_indel_2ndmax)
	    = (\@bins_ct_clms_2ndmax_cmplx, \@bins_totct_indel_2ndmax_cmplx, \@bins_diffct_indel_2ndmax_cmplx, \@bins_totlen_indel_2ndmax_cmplx, \@bins_difflen_indel_2ndmax_cmplx); # Modified.

	    ($bins_diffct_indel_min, $bins_diffct_indel_2ndmin, $bins_difflen_indel_min, $bins_difflen_indel_2ndmin)
	    = (\@bins_diffct_indel_min_cmplx, \@bins_diffct_indel_2ndmin_cmplx, \@bins_difflen_indel_min_cmplx, \@bins_difflen_indel_2ndmin_cmplx) ; # New.


	} elsif ((defined $if_na{$last_combo_id}) or ($diagnosis =~ m/TOO_LONG/)) { # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #
	##} elsif ((defined $if_na{$combo_id}) or ($diagnosis =~ m/TOO_LONG/)) {...}

	    ##($bins_ct_clms, $bins_prpt_gps, $bins_totct_indel, $bins_diffct_indel, $bins_totlen_indel, $bins_difflen_indel)
	    ## = (\@bins_ct_clms_tl, \@bins_prpt_gps_tl, \@bins_totct_indel_tl, \@bins_diffct_indel_tl, \@bins_totlen_indel_tl, \@bins_difflen_indel_tl);

	    ($bins_ct_gpd_segs, $bins_span_cluster) = (\@bins_ct_gpd_segs_tl, \@bins_span_cluster_tl); # Totally new. #

	    ($bins_totct_clms, $bins_tot_totct_indel, $bins_tot_diffct_indel, $bins_tot_totlen_indel, $bins_tot_difflen_indel)
	    = (\@bins_totct_clms_tl, \@bins_tot_totct_indel_tl, \@bins_tot_diffct_indel_tl, \@bins_tot_totlen_indel_tl, \@bins_tot_difflen_indel_tl); # Modified.

	    ($bins_ct_clms_max, $bins_totct_indel_max, $bins_diffct_indel_max, $bins_totlen_indel_max, $bins_difflen_indel_max)
	    = (\@bins_ct_clms_max_tl, \@bins_totct_indel_max_tl, \@bins_diffct_indel_max_tl, \@bins_totlen_indel_max_tl, \@bins_difflen_indel_max_tl); # Modified.

	    ($bins_ct_clms_2ndmax, $bins_totct_indel_2ndmax, $bins_diffct_indel_2ndmax, $bins_totlen_indel_2ndmax, $bins_difflen_indel_2ndmax)
	    = (\@bins_ct_clms_2ndmax_tl, \@bins_totct_indel_2ndmax_tl, \@bins_diffct_indel_2ndmax_tl, \@bins_totlen_indel_2ndmax_tl, \@bins_difflen_indel_2ndmax_tl); # Modified.

	    ($bins_diffct_indel_min, $bins_diffct_indel_2ndmin, $bins_difflen_indel_min, $bins_difflen_indel_2ndmin)
	    = (\@bins_diffct_indel_min_tl, \@bins_diffct_indel_2ndmin_tl, \@bins_difflen_indel_min_tl, \@bins_difflen_indel_2ndmin_tl) ; # New.


	} elsif ($type eq 'Erroneous') {

		# Do nothing. #

	} else {

	    print STDOUT "[CAUTION '$program_name'] (ID_Class, Session_ID, Segment_ID) = ($id_class, $session, $segid) has either ANOMALOUS TYPE (=$type) or ANOMALOUS DIAGNOSIS (=$diagnosis): $! ! ... I will skip tallying this segment ...\n\n";

	    $ct_anom2++;
	    next;
	}

	cluster_distributor ($ct_gpd_segs, @{$bins_ct_gpd_segs}, @BOUNDS_CT); # Totally new.
	(defined $span_cluster) and cluster_distributor ($span_cluster, @{$bins_span_cluster}, @BOUNDS_CT); # Totally new.

	cluster_distributor ($totct_clms, @{$bins_totct_clms}, @BOUNDS_CT);
	##cluster_distributor ($raw_prpt_gaps, @{$bins_prpt_gps}, @BOUNDS_PRPT);
	cluster_distributor ($tot_totct_indel, @{$bins_tot_totct_indel}, @BOUNDS_CT);
	cluster_distributor ($tot_diffct_indel, @{$bins_tot_diffct_indel}, @BOUNDS_PNCT);
	cluster_distributor ($tot_totlen_indel, @{$bins_tot_totlen_indel}, @BOUNDS_CT);
	cluster_distributor ($tot_difflen_indel, @{$bins_tot_difflen_indel}, @BOUNDS_PNCT);

	cluster_distributor ($ct_clms_max, @{$bins_ct_clms_max}, @BOUNDS_CT);
	cluster_distributor ($totct_indel_max, @{$bins_totct_indel_max}, @BOUNDS_CT);
	cluster_distributor ($diffct_indel_max, @{$bins_diffct_indel_max}, @BOUNDS_PNCT);
	cluster_distributor ($totlen_indel_max, @{$bins_totlen_indel_max}, @BOUNDS_CT);
	cluster_distributor ($difflen_indel_max, @{$bins_difflen_indel_max}, @BOUNDS_PNCT);

	(defined $ct_clms_2ndmax) and cluster_distributor ($ct_clms_2ndmax, @{$bins_ct_clms_2ndmax}, @BOUNDS_CT);
	(defined $totct_indel_2ndmax) and cluster_distributor ($totct_indel_2ndmax, @{$bins_totct_indel_2ndmax}, @BOUNDS_CT);
	(defined $diffct_indel_2ndmax) and cluster_distributor ($diffct_indel_2ndmax, @{$bins_diffct_indel_2ndmax}, @BOUNDS_PNCT);
	(defined $totlen_indel_2ndmax) and cluster_distributor ($totlen_indel_2ndmax, @{$bins_totlen_indel_2ndmax}, @BOUNDS_CT);
	(defined $difflen_indel_2ndmax) and cluster_distributor ($difflen_indel_2ndmax, @{$bins_difflen_indel_2ndmax}, @BOUNDS_PNCT);

	cluster_distributor ($diffct_indel_min, @{$bins_diffct_indel_min}, @BOUNDS_PNCT); # New.
	cluster_distributor ($difflen_indel_min, @{$bins_difflen_indel_min}, @BOUNDS_PNCT); # New.
	(defined $diffct_indel_2ndmin) and cluster_distributor ($diffct_indel_2ndmin, @{$bins_diffct_indel_2ndmin}, @BOUNDS_PNCT); # New.
	(defined $difflen_indel_2ndmin) and cluster_distributor ($difflen_indel_2ndmin, @{$bins_difflen_indel_2ndmin}, @BOUNDS_PNCT); # New.

			### END of "=> Substantially MODIFIED in creation of 'summary_examine_complex.2.raw.pl'." ###

=cut # END of "For the predecessor 'summary_examine_complex.2.raw.pl'. (III)"


=Just a reminder...
my @bins_ct_clms_tl = my @bins_ct_clms_cmplx = my @bins_ct_clms_err = my @bins_ct_clms_corr = ();
my @batch_ct_clms = (\@bins_ct_clms_tl, \@bins_ct_clms_cmplx, \@bins_ct_clms_err, \@bins_ct_clms_corr);

my @bins_prpt_gps_tl = my @bins_prpt_gps_cmplx = my @bins_prpt_gps_err = my @bins_prpt_gps_corr = ();
my @batch_prpt_gps = (\@bins_prpt_gps_tl, \@bins_prpt_gps_cmplx, \@bins_prpt_gps_err, \@bins_prpt_gps_corr);

my @bins_totct_indel_tl = my @bins_totct_indel_cmplx = my @bins_totct_indel_err = my @bins_totct_indel_corr = ();
my @batch_totct_indel = (\@bins_totct_indel_tl, \@bins_totct_indel_cmplx, \@bins_totct_indel_err, \@bins_totct_indel_corr);

my @bins_diffct_indel_tl = my @bins_diffct_indel_cmplx = my @bins_diffct_indel_err = my @bins_diffct_indel_corr = ();
my @batch_diffct_indel = (\@bins_diffct_indel_tl, \@bins_diffct_indel_cmplx, \@bins_diffct_indel_err, \@bins_diffct_indel_corr);

my @bins_totlen_indel_tl = my @bins_totlen_indel_cmplx = my @bins_totlen_indel_err = my @bins_totlen_indel_corr = ();
my @batch_totlen_indel = (\@bins_totlen_indel_tl, \@bins_totlen_indel_cmplx, \@bins_totlen_indel_err, \@bins_totlen_indel_corr);

my @bins_difflen_indel_tl = my @bins_difflen_indel_cmplx = my @bins_difflen_indel_err = my @bins_difflen_indel_corr = ();
my @batch_difflen_indel = (\@bins_difflen_indel_tl, \@bins_difflen_indel_cmplx, \@bins_difflen_indel_err, \@bins_difflen_indel_corr);

batch_initializer (@batch_ct_clms, @BOUNDS_CT);
batch_initializer (@batch_prpt_gps, @BOUNDS_PRPT);
batch_initializer (@batch_totct_indel, @BOUNDS_CT);
batch_initializer (@batch_diffct_indel, @BOUNDS_PNCT);
batch_initializer (@batch_totlen_indel, @BOUNDS_CT);
batch_initializer (@batch_difflen_indel, @BOUNDS_PNCT);

=cut # END of "Just a reminder..."


=For the predecessor 'summary_examine_purge_ctrl.raw.pl'.
	(defined $br2freq_prgs{$rlv_br_id}) or  ($br2freq_prgs{$rlv_br_id} = 0);
	$br2freq_prgs{$rlv_br_id}++ ;

	(defined $ct_invlvd2freq_prgs{$ct_invlvd_seqs}) or ($ct_invlvd2freq_prgs{$ct_invlvd_seqs} = 0);
	$ct_invlvd2freq_prgs{$ct_invlvd_seqs}++ ;


	cluster_distributor ($rlv_br_len, @bins_rlv_brlen, @BOUNDS_BRLEN);
	cluster_distributor ($size_block, @bins_size_block, @BOUNDS_CT); 
	##cluster_distributor ($abs_shift, @bins_abs_shift, @BOUNDS_CT); # DISABLED in creation of 'summary_examine_purge_ctrl.raw.pl'. #
	cluster_distributor ($pv, @bins_pv, @BOUNDS_PRPT);
	##cluster_distributor ($diff_lnlk_respatt, @bins_diff_lnlk_respatt, @BOUNDS_DIFF_LNLK); # DISABLED in creation of 'summary_examine_purge_ctrl.raw.pl'. #

	distributor_2dim ($ct_invlvd_seqs, $rlv_br_len, @{$bins_ct_invlvd_vs_brlen}, @BOUNDS_CT, @BOUNDS_BRLEN);

	distributor_2dim ($size_block, $pv, @{$bins_blsize_vs_pv}, @BOUNDS_CT, @BOUNDS_PRPT);
	distributor_2dim ($ct_invlvd_seqs, $pv, @{$bins_ct_invlvd_vs_pv}, @BOUNDS_CT, @BOUNDS_PRPT);
	distributor_2dim ($rlv_br_len, $pv, @{$bins_brlen_vs_pv}, @BOUNDS_BRLEN, @BOUNDS_PRPT);


			# ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016). #

		# (my $bins_blsize_vs_brlen, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_BRLEN, "%5.3f");
		# (my $bins_blsize_vs_ct_invlvd, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_CT, "%5.3f");

	distributor_2dim ($size_block, $rlv_br_len, @{$bins_blsize_vs_brlen}, @BOUNDS_CT, @BOUNDS_BRLEN);
	distributor_2dim ($size_block, $ct_invlvd_seqs, @{$bins_blsize_vs_ct_invlvd}, @BOUNDS_CT, @BOUNDS_CT);


	my $bins_ct_invlvd_vs_pv_sp = $blsize2bins_ct_invlvd_vs_pv{$size_block};
	unless (defined $bins_ct_invlvd_vs_pv_sp) { # Initialize the 2D distribution for the specific block size. #
	    ($bins_ct_invlvd_vs_pv_sp, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_PRPT, "%4.2f");
	    $blsize2bins_ct_invlvd_vs_pv{$size_block} = $bins_ct_invlvd_vs_pv_sp;
	}
		## %blsize2bins_ct_invlvd_vs_pv; # = ($block_size => \@bins_ct_invlvd_vs_pv_sp, ...)

	distributor_2dim ($ct_invlvd_seqs, $pv, @{$bins_ct_invlvd_vs_pv_sp}, @BOUNDS_CT, @BOUNDS_PRPT);

	my $bins_brlen_vs_pv_sp = $blsize2bins_brlen_vs_pv{$size_block};
	unless (defined $bins_brlen_vs_pv_sp) { # Initialize the 2D distribution for the specific block size. #
	    ($bins_brlen_vs_pv_sp, undef) = init_2dim_distri (@BOUNDS_BRLEN, @BOUNDS_PRPT, "%4.2f");
	    $blsize2bins_brlen_vs_pv{$size_block} = $bins_brlen_vs_pv_sp;
	}
		## %blsize2bins_brlen_vs_pv; # = ($block_size => \@bins_brlen_vs_pv_sp, ...)

	distributor_2dim ($rlv_br_len, $pv, @{$bins_brlen_vs_pv_sp}, @BOUNDS_BRLEN, @BOUNDS_PRPT);

			# END of "ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016)." #

=cut # END of "For the predecessor 'summary_examine_purge_ctrl.raw.pl'."


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
		# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.pl'. #
	if (($id_class eq '2_4') or ($id_class eq '3_4')) {
	    $ct_skipped1++; # Skip the entries belonging to these classes (to avoid redundancy). #
	    next;
	}
=cut


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
		# ADDED in creation of 'summary_classify_msa_errors.raw.pl'. #

	unless (($ct_equiv_indels ne 'undef') and (uc($ct_equiv_indels) ne 'N/A')) {
	    print STDOUT '[CAUTION] ($id_class, $session, $segment) = ', "($id_class, $session, $segment) has ", '$ct_equiv_indels = ', "$ct_equiv_indels: $! !\n\n";
	    $ct_skipped_x++;
	    next;
	}
=cut


	$ct_succ++;

=DISABLED in creation of 'summary_examine_complex.2.raw.pl'.
}		# End of outer while-loop (over erroneous MSA segments) #
=cut

	 ## ($ct_succ>=1000) and last; ### For TESTING the script. ###

}		# End of outer while-loop (until the last cluster runs out). # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #

close ($fh_in1);
##close ($fh_out); ## DISABLED in creation of 'summary_examine_purge.raw.pl'. ##

print STDOUT "\n\n=== ALL the Processes ENDED !! ===\n\n";

print STDOUT "#{Anomalous input lines}    = $ct_anom .\n"; # MOVED from below in creation of 'summary_examine_complex.1.raw.pl'.
print STDOUT "Total#{input Segments}         = $totct_segs .\n";
print STDOUT "   #{Segments w/ anomalous type or diagnosis} = $ct_anom2 .\n"; # ADDED in creation of 'summary_examine_complex.1.raw.pl'.
print STDOUT "   #{Segments w/ inconsistent #{gapped segments}} = $ct_anom3 .\n\n"; # ADDED in creation of 'summary_examine_complex.2.raw.pl'.

=DISABLED in creation of 'summary_examine_complex.1.raw.pl'.
print STDOUT "#{Skipped segments (to avoid redundancy)} = $ct_skipped1 .\n";

		# MODIFIED in creation of 'summary_examine_purge.raw.pl'. #

print STDOUT '#{Skipped segments (w/ $size_block or $abs_shift fractional)} = ', $ct_w_fractional, " .\n";
=cut

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
			# ADDED in creation of 'summary_classify_msa_errors.raw.pl'. #
print STDOUT '#{Skipped segments (w $ct_equiv_indels unavailable)} = ', $ct_skipped_x, " .\n";
print STDOUT "#{Anomalous input lines}    = $ct_anom .\n"; # MOVED from above in creation of 'summary_classify_msa_errors.raw.pl'.
print STDOUT "#{Input segments containing 'undef' differences} = $ct_w_undef .\n";
=cut

print STDOUT "   #{Input segments failed to be processed} = $ct_fail .\n";
print STDOUT "   #{Input segments successfully processed} = $ct_succ .\n\n";

unless (0 == $ct_fail) {
	print STDOUT "[ERROR '$program_name'] FAILED to process $ct_fail inputs !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}


		### MAJOR ADDITION in creation of 'summary_examine_complex.2_2D.raw.pl'. ###


print STDOUT "\n\n[ (i) ] Categories of the erroneous and correct segments:\n\n";
print STDOUT "Category A:   'Complex' errors + segments that are 'Too-Long' ,\n\n";
print STDOUT "Category B:   'Non-complex' errors + Correctly reconstructed segments .\n\n";

my $section = 0;

$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and #{columns} (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_maxct_columns_A}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max#{columns}', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_maxct_columns_B}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max#{columns}', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_maxct_columns_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
# (my $bins_max_totct_indels_vs_maxct_columns_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and the total length of indels (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_totlen_indels_A}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max(Totlen{indels})', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_totlen_indels_B}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max(Totlen{indels})', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_max_totlen_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
# (my $bins_max_totct_indels_vs_max_totlen_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and Max(#{ins} - #{del}) (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_diffct_indels_A}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Max(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_diffct_indels_B}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Max(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_max_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
# (my $bins_max_totct_indels_vs_max_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and Min(#{ins} - #{del}) (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_min_diffct_indels_A}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Min(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_min_diffct_indels_B}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Min(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
# (my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");


print STDOUT "\n\n";


=For the predecessor 'summary_examine_complex.2.raw.pl'. (V)

		### MAJOR ADDITION in creation of 'summary_examine_complex.2.raw.pl'. ###

print STDOUT "\n\n[ i ] Distributions of #{gapped segments} in the clusters of segments of various types:\n\n";

batch_pr_file_distri_xl (@batch_ct_gpd_segs, @BOUNDS_CT, '#{gpd_segs}', *STDOUT, "%3d", @tags_segtypes);

##batch_initializer (@batch_ct_gpd_segs, @BOUNDS_CT);


print STDOUT "\n\n[ ii ] Distributions of the spans of the (clusters of) segments of various types:\n\n";

batch_pr_file_distri_xl (@batch_span_cluster, @BOUNDS_CT, '{span_cluster}', *STDOUT, "%3.0f", @tags_segtypes);
##batch_pr_file_distri_xl (@batch_span_cluster, @BOUNDS_CT, '{span_cluster}', *STDOUT, "%3d", @tags_segtypes);

print STDOUT "\n";

##batch_initializer (@batch_span_cluster, @BOUNDS_CT);

		### END of "MAJOR ADDITION in creation of 'summary_examine_complex.2.raw.pl'." ###

=cut # END of "For the predecessor 'summary_examine_complex.2.raw.pl'. (V)"


=For the predecessor 'summary_examine_complex.2.raw.pl'. (VI)

my $section = 0;

		### MAJOR ADDITION in creation of 'summary_examine_complex.1.raw.pl'. ###
		### => SUbstantially MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. ###

$section++;
print STDOUT "\n\n[$section] Distributions of #{columns} in the (clusters of) segments of various types:\n\n";

print STDOUT "{ Total }\n\n";

batch_pr_file_distri_xl (@batch_totct_clms, @BOUNDS_CT, 'Tot#{columns}', *STDOUT, "%3d", @tags_segtypes);

print STDOUT "\n\n{ Maximum }\n\n";

batch_pr_file_distri_xl (@batch_ct_clms_max, @BOUNDS_CT, 'Max#{columns}', *STDOUT, "%3d", @tags_segtypes);

print STDOUT "\n\n{ 2nd maximum }\n\n";

batch_pr_file_distri_xl (@batch_ct_clms_2ndmax, @BOUNDS_CT, '2nd_Max#{columns}', *STDOUT, "%3d", @tags_segtypes);

print STDOUT "\n";

##batch_initializer (@batch_ct_clms, @BOUNDS_CT);

#=For the predecessor 'summary_examine_complex.1.raw.pl'.
$section++;
print STDOUT "\n\n[$section] Distributions of the proportion of gaps in the segments of various types:\n\n";

batch_pr_file_distri_xl (@batch_prpt_gps, @BOUNDS_PRPT, 'prpt{gaps}', *STDOUT, "%5.3f", @tags_segtypes);

##batch_initializer (@batch_prpt_gps, @BOUNDS_PRPT);
#=cut


$section++;
print STDOUT "\n\n[$section] Distributions of #{indels} (Dollo-parsimony-inferred) in the (clusters of) segments of various types:\n\n";

print STDOUT "{ Total }\n\n";

batch_pr_file_distri_xl (@batch_tot_totct_indel, @BOUNDS_CT, 'Tot#{indels}', *STDOUT, "%3d", @tags_segtypes);

print STDOUT "\n\n{ Maximum }\n\n";

batch_pr_file_distri_xl (@batch_totct_indel_max, @BOUNDS_CT, 'Max#{indels}', *STDOUT, "%3d", @tags_segtypes);

print STDOUT "\n\n{ 2nd maximum }\n\n";

batch_pr_file_distri_xl (@batch_totct_indel_2ndmax, @BOUNDS_CT, '2nd_Max#{indels}', *STDOUT, "%3d", @tags_segtypes);
##batch_pr_file_distri_xl (@batch_totct_indel_max, @BOUNDS_CT, '2nd_Max#{indels}', *STDOUT, "%3d", @tags_segtypes); # A BUG fixed. #

print STDOUT "\n";

##batch_initializer (@batch_totct_indel, @BOUNDS_CT);


$section++;
print STDOUT "\n\n[$section] Distributions of (#{ins} - #{del}) (Dollo-parsimony-inferred) in the (clusters of) segments of various types:\n\n";

print STDOUT "{ Total }\n\n";

batch_pr_file_distri_xl (@batch_tot_diffct_indel, @BOUNDS_PNCT, 'Tot(#{ins} - #{del})', *STDOUT, "%4d", @tags_segtypes);

print STDOUT "\n\n{ Maximum }\n\n";

batch_pr_file_distri_xl (@batch_diffct_indel_max, @BOUNDS_PNCT, 'Max(#{ins} - #{del})', *STDOUT, "%4d", @tags_segtypes);

print STDOUT "\n\n{ Minimum }\n\n";

batch_pr_file_distri_xl (@batch_diffct_indel_min, @BOUNDS_PNCT, 'Min(#{ins} - #{del})', *STDOUT, "%4d", @tags_segtypes);

print STDOUT "\n\n{ 2nd maximum }\n\n";

batch_pr_file_distri_xl (@batch_diffct_indel_2ndmax, @BOUNDS_PNCT, '2nd_Max(#{ins} - #{del})', *STDOUT, "%4d", @tags_segtypes);

print STDOUT "\n\n{ 2nd minimum }\n\n";

batch_pr_file_distri_xl (@batch_diffct_indel_2ndmin, @BOUNDS_PNCT, '2nd_Min(#{ins} - #{del})', *STDOUT, "%4d", @tags_segtypes);

print STDOUT "\n";

##batch_initializer (@batch_diffct_indel, @BOUNDS_PNCT);


$section++;
print STDOUT "\n\n[$section] Distributions of the Total Length of indels (Dollo-parsimony-inferred) in the (clusters of) segments of various types:\n\n";

print STDOUT "{ Total }\n\n";

batch_pr_file_distri_xl (@batch_tot_totlen_indel, @BOUNDS_CT, 'Tot(Totlen{indels})', *STDOUT, "%3d", @tags_segtypes);

print STDOUT "\n\n{ Maximum }\n\n";

batch_pr_file_distri_xl (@batch_totlen_indel_max, @BOUNDS_CT, 'Max(Totlen{indels})', *STDOUT, "%3d", @tags_segtypes);

print STDOUT "\n\n{ 2nd maximum }\n\n";

batch_pr_file_distri_xl (@batch_totlen_indel_2ndmax, @BOUNDS_CT, '2nd_Max(Totlen{indels})', *STDOUT, "%3d", @tags_segtypes);

print STDOUT "\n";

##batch_initializer (@batch_totlen_indel, @BOUNDS_CT);


$section++;
print STDOUT "\n\n[$section] Distributions of (Totlen{ins} - Totlen{del}) (Dollo-parsimony-inferred) in the (clusters of) segments of various types:\n\n";

print STDOUT "{ Total }\n\n";

batch_pr_file_distri_xl (@batch_tot_difflen_indel, @BOUNDS_PNCT, 'Tot(Totlen{ins} - Totlen{del})', *STDOUT, "%4d", @tags_segtypes);

print STDOUT "\n\n{ Maximum }\n\n";

batch_pr_file_distri_xl (@batch_difflen_indel_max, @BOUNDS_PNCT, 'Max(Totlen{ins} - Totlen{del})', *STDOUT, "%4d", @tags_segtypes);

print STDOUT "\n\n{ Minimum }\n\n";

batch_pr_file_distri_xl (@batch_difflen_indel_min, @BOUNDS_PNCT, 'Min(Totlen{ins} - Totlen{del})', *STDOUT, "%4d", @tags_segtypes);

print STDOUT "\n\n{ 2nd maximum }\n\n";

batch_pr_file_distri_xl (@batch_difflen_indel_2ndmax, @BOUNDS_PNCT, '2nd_Max(Totlen{ins} - Totlen{del})', *STDOUT, "%4d", @tags_segtypes);

print STDOUT "\n\n{ 2nd minimum }\n\n";

batch_pr_file_distri_xl (@batch_difflen_indel_2ndmin, @BOUNDS_PNCT, '2nd_Min(Totlen{ins} - Totlen{del})', *STDOUT, "%4d", @tags_segtypes);

print STDOUT "\n";

##batch_initializer (@batch_difflen_indel, @BOUNDS_PNCT);

=cut # END of "For the predecessor 'summary_examine_complex.2.raw.pl'. (VI)"


=For the predecessor 'summary_examine_purge.raw.pl'.
		### ADDED in creation of 'summary_examine_purge.raw.pl'. ###

$section++;
print STDOUT "\n\n[$section] The frequencies of \"Purge\"s along specified branches:\n\n";

print STDOUT join ("\t", 'Branch_ID', 'Freq(Purges)'), "\n\n";

my $sum_freq = 0;
foreach my $br (sort {$a <=> $b} keys %br2freq_prgs) { # 1st trailing foreach-loop (over branches). #

	my $freq = $br2freq_prgs{$br};
	print STDOUT join ("\t", $br, $freq), "\n";
	# %br2freq_prgs; # = ($br_id => $abs_freq_purges, ...)

	$sum_freq += $freq;
}					# End of the 1st trailing foreach-loop (over branches). #
print STDOUT "\n";
print join ("\t", 'Total', $sum_freq), "\n\n";


$section++;
print STDOUT "\n\n[$section] The frequencies of \"Purge\"s involving particular numbers of sequences:\n\n";

print STDOUT join ("\t", '#{invlvd_seqs}', 'Freq(Purges)'), "\n\n";

my $sum_freq2 = 0;
foreach my $ct (sort {$a <=> $b} keys %ct_invlvd2freq_prgs) { # 2nd trailing foreach-loop (over #{involved sequences}). #

	my $freq = $ct_invlvd2freq_prgs{$ct};
	print STDOUT join ("\t", $ct, $freq), "\n";
	# %ct_invlvd2freq_prgs; # = ($ct_invlvd_seqs => $abs_freq_purges, ...)

	$sum_freq2 += $freq;
}						# End of the 2nd trailing foreach-loop (over #{involved sequences}). #
print STDOUT "\n";
print join ("\t", 'Total', $sum_freq2), "\n\n";


$section++;
print STDOUT "\n\n[$section] Distributions of block-sizes and |shift|'s for the \"purge\"-accommodated blocks:\n\n";

batch_pr_file_distri_xl (@batch_ct, @BOUNDS_CT, 'count', *STDOUT, "%2d", @tags_ct);

print STDOUT "\n\n[ NOTE: The shifts are not available for the control analyses... ]\n\n"; # ADDED in creation of 'summary_examine_purge_ctrl.raw.pl'. #

## @tags_ct = ('Intervals', 'block_size' '|shift|');
## @batch_ct = (\@bins_size_block, \@bins_abs_shift);

$section++;
print STDOUT "\n\n[$section] Frequencies of \"purge\"s classified by the length of branches separating the sheared blocks:\n\n";

my @batch_brlen = (\@bins_rlv_brlen);
my @tags_brlen = ('Intervals', 'branch_length');

batch_pr_file_distri_xl (@batch_brlen, @BOUNDS_BRLEN, 'value', *STDOUT, "%4.2f", @tags_brlen);

# bin_initializer (@bins_rlv_brlen, @BOUNDS_BRLEN);


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{involved sequences} (abscissa) and separating branch length (ordinates) of \"Purge\"s:\n\n";

pr_file_2dim_distri_xl (@{$bins_ct_invlvd_vs_brlen}, @BOUNDS_CT, @BOUNDS_BRLEN, '#{invlvd_seqs}', 'branch_length', *STDOUT, "%1d", "%4.2f");

## (my $bins_ct_invlvd_vs_brlen, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_BRLEN, "%5.3f");


		## ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016). ##

$section++;
print STDOUT "\n\n[$section] 2D-distribution of block-size (abscissa) and separating branch length (ordinates) of \"Purge\"s:\n\n";

pr_file_2dim_distri_xl (@{$bins_blsize_vs_brlen}, @BOUNDS_CT, @BOUNDS_BRLEN, 'block_size', 'branch_length', *STDOUT, "%1d", "%4.2f");

## (my $bins_blsize_vs_brlen, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_BRLEN, "%5.3f");

$section++;
print STDOUT "\n\n[$section] 2D-distribution of block-size (abscissa) and #{involved sequences} (ordinates) of \"Purge\"s:\n\n";

pr_file_2dim_distri_xl (@{$bins_blsize_vs_ct_invlvd}, @BOUNDS_CT, @BOUNDS_CT, 'block_size', '#{invlvd_seqs}', *STDOUT, "%1d", "%2d");

## (my $bins_blsize_vs_ct_invlvd, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_CT, "%1d");

		## END of "ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016)." ##

$section++;
print STDOUT "\n\n[$section] Frequencies of \"purge\"s classified by the P-values regarding the substitiutional differences:\n\n";

my @batch_pv = (\@bins_pv);
my @tags_pv = ('Intervals', 'P-value');

batch_pr_file_distri_xl (@batch_pv, @BOUNDS_PRPT, 'value', *STDOUT, "%5.3f", @tags_pv);

# bin_initializer (@bins_pv, @BOUNDS_PRPT);

$section++;
print STDOUT "\n\n[$section] Frequencies of \"purge\"s classified by the differences in the log-likelihoods of the residue patterns, between the reconstructed and reference MSAs:\n\n";

print STDOUT "\n[ NOTE: The log-likelihood differences are NOT AVAILABLE for the control analyses... ]\n\n"; # ADDED in creation of 'summary_examine_purge_ctrl.raw.pl'. #


$section++;
print STDOUT "\n\n[$section] 2D-distribution of block-sizes (abscissa) and P-values (ordinates) of \"Purge\"s:\n\n";

pr_file_2dim_distri_xl (@{$bins_blsize_vs_pv}, @BOUNDS_CT, @BOUNDS_PRPT, 'block_size', 'P-value', *STDOUT, "%1d", "%5.3f");

## (my $bins_blsize_vs_pv, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_PRPT, "%5.3f");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{involved sequences} (abscissa) and P-values (ordinates) of \"Purge\"s:\n\n";

pr_file_2dim_distri_xl (@{$bins_ct_invlvd_vs_pv}, @BOUNDS_CT, @BOUNDS_PRPT, '#{invlvd_seqs}', 'P-value', *STDOUT, "%1d", "%5.3f");

## (my $bins_ct_invlvd_vs_pv, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_PRPT, "%4.2f");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of separating branch lengths (abscissa) and P-values (ordinates) of \"Purge\"s:\n\n";

pr_file_2dim_distri_xl (@{$bins_brlen_vs_pv}, @BOUNDS_BRLEN, @BOUNDS_PRPT, 'branch_length', 'P-value', *STDOUT, "%3.2f", "%5.3f");

## (my $bins_brlen_vs_pv, undef) = init_2dim_distri (@BOUNDS_BRLEN, @BOUNDS_PRPT, "%4.2f");


			### ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016). ###

$section++;
print STDOUT "\n\n[$section] 2D-distributions, each for a specific block-size, of #{involved sequences} (abscissa) and P-values (ordinates) of \"Purge\"s:\n\n\n";

foreach my $size_block (sort {$a <=> $b} keys %blsize2bins_ct_invlvd_vs_pv) { # 3rd trailing foreach-loop (over block sizes). #

    print STDOUT "{ Block size = $size_block }\n\n";

    my $bins_ct_invlvd_vs_pv_sp = $blsize2bins_ct_invlvd_vs_pv{$size_block};
    pr_file_2dim_distri_xl (@{$bins_ct_invlvd_vs_pv_sp}, @BOUNDS_CT, @BOUNDS_PRPT, '#{invlvd_seqs}', 'P-value', *STDOUT, "%1d", "%5.3f");
    print STDOUT "\n\n";
}							# End of the 3rd trailing foreach-loop (over block sizes). #

## %blsize2bins_ct_invlvd_vs_pv; # = ($block_size => \@bins_ct_invlvd_vs_pv, ...)


$section++;
print STDOUT "\n\n[$section] 2D-distributions, each for a specific block-size, of separating branch lengths (abscissa) and P-values (ordinates) of \"Purge\"s:\n\n\n";

foreach my $size_block (sort {$a <=> $b} keys %blsize2bins_brlen_vs_pv) { # 5th trailing foreach-loop (over block sizes). #

    print STDOUT "{ Block size = $size_block }\n\n";

    my $bins_brlen_vs_pv_sp = $blsize2bins_brlen_vs_pv{$size_block};
    pr_file_2dim_distri_xl (@{$bins_brlen_vs_pv_sp}, @BOUNDS_BRLEN, @BOUNDS_PRPT, 'branch_length', 'P-value', *STDOUT, "%3.2f", "%5.3f");
    print STDOUT "\n\n";
}							# End of the 5th trailing foreach-loop (over block sizes). #

## %blsize2bins_brlen_vs_pv; # = ($block_size => \@bins_brlen_vs_pv, ...)


			### END of "ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016)." ###
=cut # END of "For the predecessor 'summary_examine_purge.raw.pl'."


print STDOUT "\n\n=== ENDED Printing Out the Statistics ===\n\n";


	### The DEFINITION(s) of the SUBROUTINE(s). ###

# broad_error_type:
#	This subroutine receives the name of an error-type,
#	and classifies it into a broad type.
#	(CREATED in creation of 'summary_classify_msa_errors.bulk_pars.pl' on Nov 17 (Mon), 2014.)
#
# INPUT: $type_error, which is the name of an error-type.
#
# OUTPUT: $br_type, which is the broad type accommodating $type_error.
#
sub broad_error_type ($) {

	my ($type_error) = @_;

	$type_error = uc($type_error);

	foreach my $brtype ('Shift', 'Complex', 'Merge', 'Purge', 'Split', 'Ex-nihilo') { # Main foreach-loop (over some broad error-types). #
	    if (substr ($type_error, 0, length ($brtype)) eq uc($brtype)) {
		return $brtype;
	    }
	} 				# End of the main foreach-loop (over some broad error-types). # 

	if ($type_error =~ m/^NEIGHBORING-BRANCH-MERGE/) {
	    return 'NB_Merge';

	} elsif ($type_error =~ m/^NEIGHBORING-BRANCH-SPLIT/) {
	    return 'NB_Split';

	} elsif ($type_error =~ m/^COLLAPSE-OF-INDEPENDENT-INSERTIONS/) {
	    return 'CII';

	} elsif ($type_error =~ m/^CREATION-OF-ERRONEOUS-INDEPENDENT-INSERTIONS/) {
	    return 'CEII';

	} elsif ($type_error =~ m/^INCOMPLETE-COLLAPSE-OF-INDEPENDENT-INSERTIONS/) {
	    return 'iCII';

	} elsif ($type_error =~ m/^INCOMPLETE-CREATION-OF-ERRONEOUS-INDEPENDENT-INSERTIONS/) {
	    return 'iCEII';

	} elsif ($type_error =~ m/^PARTIAL-NEIGHBORING-BRANCH-MERGE/) {
	    return 'pNB_Merge';

	} elsif ($type_error =~ m/^PARTIAL-NEIGHBORING-BRANCH-SPLIT/) {
	    return 'pNB_Split';

	} else {
	    return 'Other';
	}

		## @BROAD_TYPES = ('Shift', 'Complex', 'Merge', 'Purge', 'Split', 'Ex-nihilo', 'NB_Merge', 'NB_Split', 'CII', 'CEII', 'pNB_Merge', 'pNB_Split', 'iCII', 'iCEII', 'Other');

}		# End of the subroutine 'broad_error_type'. #


