#!/usr/bin/perl
#
# examine_complex_via_mblks.2aug.pl:
#	This augmented version of 'examine_complex_via_mblks.2.pl' outputs more detailed information of each gapped segment,
#	by appending the information of all indel events (inferred via the Dollo parsimony principle).
#	[CREATED on July 12 (Tue), 2016.]
#
# examine_complex_via_mblks.2.pl:
#	[This script examines erroneous segments each of which contains at least one 'Complex' error.] --- OBSOLETE.
#	This script examines both "correct" and "erroneous" segments.
#	The script was created by taking advantage of a part of the framework of 'examine_purges_via_mblks.2.pl'.
#
#	(NOTE: The output coordinates for the subjects are in the 0-based, full-closed convention.)
#
#	[CREATION STARTED on May 30 (Mon), 2016, ENDED on May 31 (Tue), 2016.]
#
# examine_purges_via_mblks.2.pl:
#	This script only examines erroneous segments that only contain a single 'Purge' each and nothing else.
#	The script was created by taking advantage of the part of the framework of 'classify_msa_errors_via_mblks.2.pl'.
#	[CREATION STARTED on May 8 (Sun), 2016, ENDED on Mmm dd (www), 2016.]
#
# classify_msa_errors_via_mblks.2.pl:
#	This script attempts to further classify some simple 'Complex' errors
#	that is either associated with a block "hard-link"ing another block
#	or associated with a pair of blocks.
#	The script was created by enhancing the script, 'classify_msa_errors_via_mblks.1.pl'.
#	[CREATION STARTED on Oct 5 (Tue), 2015, ENDED on Oct 23 (Fri), 2015.]
#
#	[MODOFIED on Oct 14 (Wed), 2015]
#	Now, the script will preview $commoner_shift_flank, and, if it is nonzero,
#	either the reference or reconstructed MSA will be shifted entirely,
#	to nullify $commoner_shift_flank before beginning the main processes.
#
#	[MODIFIED on Oct 15 (Thu), 2015]
#	The rather long blocks to display the position-shift map were now replaced by
#	the subroutine 'pr_map_shifts_respos'.
#
#	[MODIFIED on Oct 21 (Wed), 2015]
#	Now, the subroutine 'classify_error_assoc_w_sgl_cblk' outputs the penalty of the error as well.
#
#	[MODIFIED on Oct 24 (Sat), 2015]
#	BUGs were fixed in the subroutines, 'framework_cblock' and 'undo_mv_latter_of_cblock_pair'.
#
#	[MODIFIED on Oct 25 Sun), 2015]
#	(1) The subroutine, 'undo_sgl_bw_fw_move_on_2maps', was replaced with 'undo_sgl_bw_fw_move_on_2maps_2'.
#	The difference between the two in in how to handle "inpurities," i.e., gaps and sites with different values of shifts,
#	within the framework to be moved. (See the header of 'undo_sgl_bw_fw_move_on_2maps_2' for details.)
#	(2) The subroutine 'undo_mv_latter_of_cblock_pair' now handles any errors in satellite subroutines as
#		the situation hard to undo the move, instead of as its own error.
#	
# classify_msa_errors_via_mblks.1.pl:
#	This script uses the mini-block-organization of position shifts to classify the MSA errors.
#	It was created by disabling functions on the block-organization of position shifts.
#	(CREATION STARTED on Oct 24 (Fri), 2014, ENDED on Oct 10 (Mon), 2014.)
#
#	[MODIFIED on Oct 2 (Fri) through Oct 4 (Sun), 2015]
#	Now the script can handle a special case of iCII, which were overlooked before the modification.
#
# classify_msa_errors.1.pl:
#	This Perl script classifies MSA errors by comparing the reference MSA and a reconstructed MSA
#	in each segment that was erroneously reconstructed by an aligner.
#	It was created by taking advantage of the framework of 'compare_ref_vs_rec_prank_scores.stripped.pl'.
#	(CREATION STARTED on Oct 10 (Fri), 2014, ENDED on Oct xx (yyy), 2014.)
#
# compare_ref_vs_rec_prank_scores.stripped.pl:
#	This script is functionally identical to 'compare_ref_vs_rec_prank_scores.pl',
#	although it has many commented out commands and blocks stripped off.
#	(CREATED on Oct 7 (Tue), 2014.)
#
# compare_ref_vs_rec_prank_scores.pl:
#	This Perl script compares the reference MSA and a reconstructed MSA
#	in terms of the profile-profile pairwise score used in PRANK.
#	It was created by modifying 'compare_ref_vs_rec_mafft_consist.pl'.
#	(CREATION STARTED on Sep 18 (Thu), 2014, ENDED on Sep xx (yyy), 2014)
#
# compare_ref_vs_rec_mafft_consist.pl:
#	This Perl script compares the reference MSA and a reconstructed MSA
#	in terms of the consistency score used in (the refinement mode of) MAFFT.
#	It was created by modifying 'compare_ref_vs_rec_compllimment.pl'.
#	(CREATION STARTED on Sep 12 (Fri), 2014, ENDED on Sep xx (yyy), 2014)
#
# compare_ref_vs_rec_compllimment.pl:
#	This is a slightly modified version of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
#	The major difference from the predecessor is that it now accepts four input file paths,
#	instead of the path to an input directory as the predecessor did.
#	(CREATION STARTED on Aug 8 (Fri), 2014, ENDED on Aug 9 (Sat), 2014)
#
# compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl:
#	This revised version of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' uses the Perl package, "MyTreeMap_indels_spt_odr.pm,"
#	instead of "MyTreeMap_indels.pm" as the predecessor was using.
#	And it also underwent some additional minor modifications,
#	which were mainly imported from "br_infer_psm_indel_events_spt_odr_dawg.v2.pl"
#	in the "Proof_of_Principle2" session for the framework to enable indel likelihood analyses.
#	(CREATED on Aug 5 (Tue), 2014)
#
#	[REVISED on Aug 6 (Wed), 2014]
#	For the indel analyses (enumerating parsimonious indel scenarios
#	plus calculating the approximate probabilities of the gap configurations of the segments),
#	the script now uses the input tree as it is, 
#	even if it has a bifurcated root.
#
# compare_ref_vs_rec_abs_scoring.PhyML.pl:
#	This Perl script compares the reference MSA and a reconstructed MSA
#	in terms of the 'absolute score', which is the summation
#	of the log-likelihood of gap pattern and that of the residue pattern.
#	It was created by importing some parts from the script 'compare_ml_msa.bch.extJC.6t.PhyML.pl'.
#	

use strict;
use warnings;

	# MODIFIED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 28 (Tue), 2014). #
use MyArrayManipu_num qw(
	diff_nmarray
	nr_nmarray
	log_sum_exp
	binsearch_num
	);
##use MyArrayManipu_num qw( log_sum_exp );

	# MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 7 (Wed), 2015). #
use MyArrayManipu qw( diff_array binsearch_string );
##use MyArrayManipu qw( binsearch_string );

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext
	tfm_labels2indices
	);

use MyMalnIO qw(
	extract_fasta
	extract_alignment
	pr_ODEN_input
	pr_phylip_aln
	);
	## pr_clustal_like_lite_aa
use MyMalnManipu qw( rm_gapcolumn2 );

use MyTreeManipu_II qw(
	extract_sgltree_from_file
	input_tree_file output_tree
	);

	# ADDED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg
	);

	# ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl' (on Sep 18 (Thu), 2014). #
	# => MODIFIED in creation of 'classify_msa_errors.1.pl' (on Oct 19 (Sun), 2014). #
use MyTreeManipu_III qw( make_node2abs_depth  make_node2ct_exoffs );
##use MyTreeManipu_III qw( make_node2abs_depth );

use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );

use MyPAML_parser qw( retrieve_aa_freqs );
	##prepare_paml_ctrl 

use MyApBayes_Score_MSA_ExtJC qw(
	readin_set_trees3 log_akaike_weights
	);
	## apbayes_scoring_msa_extJC_f 

	### The following three modules are for calculating log-likelihoods of substitution patterns. ###

=for the predecessor 'compare_ml_msa.bch.extJC.6t.PhyML.pl'
use MyMalnManipu_UH2 qw( 
	parse_altaln_header extract_aln_mtrx3
	fetch_nonred_clms_in_msas_thr
	);
	##fetch_nonred_clms_in_msas_mdst
use MyTreeMap_MLf qw(
	msa_likelihood_extJC_fnr
	);
	## msa_likelihood_extJC_f
=cut


use MyPhyML_parser qw(
	est_resfreqs_frm_msa print_custom_sbst_model
	extract_phyml_statistics2
	fetch_clmwise_likelihoods_phyml
	);
	## extract_phyml_statistics


	### ADDED in creation of 'examine_purges_via_mblks.2.pl'. ###

use MyLikelihoods_respattern_msa_dna qw(
	calc_br2tr_mtrx_TN
	calc_node2ntfreqs_td
	calc_prob_descresstats_cond_ancstat2
	calc_clm_probs
	);

use MyDetect_purge_cands qw(
	sum_arrays_over_branches
	calc_br2cw_expcts_obscts_sbsts
	calc_pv_obscts_sbsts_wd_br
	);

	### END of "ADDED in creation of 'examine_purges_via_mblks.2.pl'." ###

=For the predecessor 'classify_msa_errors_via_mblks.2.pl'.
	### ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl'. ###
use MyPRANK_Scores qw(
	calc_br2tr_mtrx_TN
	calc_prob_descresstats_cond_ancstat
	calc_prob_descresstats
	set_prank_align_hmm
	prank_score_2profile_aln
	);
=cut

	### ADDED in creation of 'classify_msa_errors.1.pl'. ###

use MyMalnComparison qw(
	assign_resnums_to_clms
	map_shifts_respos_bw_2msas
	br_list_classes_shift_respos
	list_classes_shift_respos
	dpsm_infer_sep_patterns_for_classes
	split_unnatural_miniclasses
	reorganize_classes_shift_respos
	merge_miniclasses_shift_respos
	find_trivial_blocks
	lump_together_similar_blocks
	);
	## lump_together_similar_blocks_proto

use MyTreeMap_indels_spt_odr qw(
	extract_cls_gblocks
	msa_rep_gpatterns
	scan_gbcls_for_long_gaps
	create_initial_set_indels_spt_odr
	);
	## enum_psm_indel_scns_for_msa

use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);

use MyTaxonomy qw(
	fetch_ext_offsprings_nr
	fetch_common_ancestors
	); # ADDED on Oct 28 (Tue), 2014. #

	### END of "ADDED in creation of 'classify_msa_errors.1.pl'." ###

=for the predecessor 'compare_ref_vs_rec_compllimment.pl'.
	### For enumerating parsimonious indel scenarios. ###

	# A MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). #
use MyTreeMap_indels_spt_odr qw(
	enum_psm_indel_scns_for_msa
	);
##=for the predecessor 'compare_ref_vs_rec_abs_scoring.PhyML.pl'.
use MyTreeMap_indels qw(
	enum_psm_indel_scns_for_msa
	);
	## extract_cls_gblocks msa_rep_gpatterns
##=cut

=cut # END of "=for the predecessor 'compare_ref_vs_rec_compllimment.pl'."

sub bi2v_multifurcated_top (\%\%\%\%\@); # ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 9 (Mon), 2016). #

sub summarize_info_cblock ($\@\@\@\%\%); # A BUG fixed on Oct 23 (Fri), 2015.
##sub summarize_info_cblocks ($\@\@\@\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 22 (Thu), 2015). #
sub pr_map_shifts_respos (\*\@\@$$); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 15 (Thu), 2015). #
sub blockwise_framework ($$\@\@\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 12 (Mon), 2015). #
sub framework_cblock ($$\@\@\@\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 12 (Mon), 2015). #
sub scout_possible_undoing ($\%\@$); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 12 (Mon), 2015).

		# MODIFIED in DEBUGGING 'classify_msa_errors_via_mblks.2.pl' (on Oct 25 (Sun), 2015). #
sub undo_sgl_bw_fw_move_on_2maps_2 (\%$$$\@\@);
##sub undo_sgl_bw_fw_move_on_2maps (\%$$$\@\@); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 13 (Tue), 2015). #

sub rm_zero_shifts_at_ends (\@$\@\@\@\@\@\@\@$\$\@); # ADDED in creation of  'classify_msa_errors_via_mblks.2.pl' (on Oct 18 (Sun), 2015). #
sub undo_mv_latter_of_cblock_pair (\@$$$\@\@\%\%\%$); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 14 (Wed), 2015). #

sub sub_enum_end_sequences ($$$\@\@\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 10 (Sat), 2015). #
sub enum_end_sequences ($\@\@\%\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 10 (Sat), 2015). #
sub classify_error_assoc_w_sgl_cblk ($\@$$$\%$\%\%\%\@\@\@\@\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 6 (Tue), 2015). #

sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$\@); # MODIFIED on Oct 14 (Wed), 2015. #
##sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 9 (Fri), 2015). #

sub cal_init_psm_cands_for_twin_msas (\%\%\%\%\%\@\@\@$$); # ADDED in creation of 'classify_msa_errors.1.pl' (on Oct 20 (Mon), 2014). #
sub reorganize_set_indels (\@\@); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 25 (Sat), 2014). #
sub identify_equiv_indels (\%\%\@\@$$\%\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 28 (Tue), 2014). #
sub effective_distance ($$$\%\%\%$$); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 30 (Thu), 2014). #
sub find_ovlp_flanking_indels ($$\@\@\%$\%\%\%$$); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 29 (Wed)- 30 (Thu), 2014). #
sub register_association_cblk_and_indels ($$\@\%\%\@\@\@); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 31 (Fri), 2014). #
sub if_collapse_indep_ins (\@\@\%\%$); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Nov 4 (Tue), 2014). #
sub if_part_nbr_mrg_2ins (\@\@\@$$\%\%$); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Nov 12 (Wed), 2014). #


sub read_list_pw_lclhom (\$$); # ADDED in creation of 'compare_ref_vs_rec_mafft_consist.pl' (on Sep 15 (Mon), 2014). #

sub readin_list_segs2 (\$\@); # ADDED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##sub readin_list_segs (\$\%);
sub readin_simul_params (\$);
sub inspect_indel_events (\@\@$$$);

sub readin_dawg_tree2 (\$); # ADDED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##sub readin_dawg_tree (\$); # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011 
sub ln_ml_sgl_seq2 (\$\%); # ADDED in creation of 'compare_apbayes_msa.bch.extJC.6.pl' on Mar 7 (Mon), 2011 #
sub add_dummy_seq (\@\@\%\%\%\%\%\@); # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 12 (Tue), 2011 #
####sub even_brlen_2seq_tree (\%$);
####sub parse_paml_main_output (\$);
##sub fetch_clmwise_lnlks (\$$);

$| = 1; # Flush the buffer of standard output everytime data flows into it.

my $program_path = $0;
my ($program_name) = ($program_path =~ m/([^\/]+)$/); 


	### Fixed Parameters ###

	# INPORTED from 'ml_scoring_msa.bch.pl' on Mar 15 (Tue), 2011 #
(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;
our $DIS_LEN = 60; # Maximum length of the sequences to be displayed in a line

our $DIS_LEN_SHIFT = 50; # Maximum length of the run of shifts to be displayed in a line.

##our $GAP = '-'; # Character representing a gap
our $CHAR_ABS =  our $GAP  = '-';
our $CHAR_PRES = our $NGAP = 'X';

	# (ADDED on Nov 29 (Mon), 2010) #
our $TOP_NODE  = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;

	# (ADDED on Mar 7 (Mon), 2011) #
our $TOT_BOOT = -1; # Because bootstrap values are NOT necessary for this analysis. 
our @INDICES_ATTR = ($INDX_LENGTH, $INDX_BOOT, $TOT_BOOT);
our $COMMENT = '#';  # Indicator of the comment line in the tree file

our $INDX_STRING = undef;
	# Specifies the index for the reference to a string in the element array of a set of non-redundant columns.
	# = undef if the element itself is the reference to a string.

our %SBSTMODEL2FILENAME = ( # $sbst_model_name (for PhyML) => $dat_name (for PAML)  (relevant only when $seqtype eq 'AA')
	'LG' => 'lg.dat',
	'WAG' => 'wag.dat',
	'JTT' => 'jones.dat',
	'MtREV' => 'mtREV24.dat',
	'Dayhoff' => 'dayhoff.dat',
	'CpREV' => 'cpREV.dat',
	'MtMam' => 'mtmam.dat',
	'MtArt' => 'mtArt.dat',
	);

our %SUPPORTED_SBSTMODELS = ( # $sbst_model_name (for PhyML) => 1 (revelant only when $seqtype eq 'NT')
	'HKY85' => 1, # default #
	'JC69' => 1,
	'K80' => 1, 
	'F81' => 1,
	# 'F84' => 1,
	'TN93' => 1, # Resurrected in creation of 'compare_ref_vs_rec_prank_scores.pl'. #
	);


	### Parameters to be adjusted ###

	# The level of tests to examine the behavior of the script (ADDED on Sep 1 (Sun), 2012):
	# 0, nothing; 1, very brief; 2, a bit brief; 3, a bit verbose; 5, very verbose.

		# For the main program: #
##our $TEST_LEVEL = 1;
our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##$TEST_LEVEL = 6;

		# For the main program, the 2nd round. #
		# (ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' on Oct 8 (Thu), 2015.)
###our $TEST_LEVEL2 = 1;
our $TEST_LEVEL2 = 2;
###our $TEST_LEVEL2 = 3;
##$TEST_LEVEL2 = 6;

		# For the main program, the 3rd(?) round (to examine the purges). #
		# (ADDED in creation of 'examine_purges_via_mblks.2.pl' on May 10 (Tue), 2016.)
##our $TEST_LEVEL3 = 1;
our $TEST_LEVEL3 = 2;
#$TEST_LEVEL3 = 3;
##$TEST_LEVEL3 = 6;


		# For examining how new types of errors could be identified.
		# (ADDED on Oct 2 (Fri), 2015.)
our $TEST_LEVEL_newtype = 0;
##our $TEST_LEVEL_newtype = 5;

		# For the subroutine, 'cal_init_psm_cands_for_twin_msas':
		# (ADDED in creation of 'classify_msa_errors.1.pl' (on Oct 20 (Mon), 2014).) #
our $TEST_LEVEL_cipc = 0;
##our $TEST_LEVEL_cipc = 5;

		# For the subroutine, 'cal_init_psm_cands_for_sgl_msa':
		# (ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 9 (Fri), 2015).) #
our $TEST_LEVEL_cipc2 = 0;
##our $TEST_LEVEL_cipc2 = 5;

		# For the subroutine, 'undo_mv_latter_of_cblock_pair':
		# (ADDED on debugging of 'classify_msa_errors_via_mblks.2.pl' (on Oct 16 (Fri), 2015).) #
our $TEST_LEVEL_umlp = 0;
##our $TEST_LEVEL_umlp = 6;

	# The level of debugging output for 'extract_phyml_statistics2' (ADDED on Aug 26 (Sun), 2012)
our $DBG_LEVEL_EXPHYML = 0;

our $FLG_RMV_TMPOUT = 1; ## For actual (large-scale) analyses
##our $FLG_RMV_TMPOUT = 0; ## For TESTING the script ##
	# = 1 if you want to remove temporary directories and files under them,
	# = 0 if you want to keep them.


	# Parameters necessary for screening the erroneous segments before calculating the initial candidate of the parsimonious indel scenario(s)... To save the running time . #
	# (ADDED in creation of 'classify_msa_errors.1.pl'.)

		# Upper-bound on the accepted length of a gap-block cluster (i.e., gapped segment) in the reference/reconstructed MSA.
our $UPPER_BOUND_ANAL_LEN_GBCLS = $MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_LEN_GBCLS = 2000; # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
##our $UPPER_BOUND_ANAL_LEN_GBCLS = $MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_LEN_GBCLS = 150;

		# Upper-bound on the accepted length of a run of gaps in the reference/reconstructed MSA.
our $UPPER_BOUND_ANAL_GAPLEN = $MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN = 1000; # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
##our $UPPER_BOUND_ANAL_GAPLEN = $MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN = 120;


our $MIN_FREQ = 1E-06; # If the frequency of a residue is zero in the anchor MSA, this nominal frequency is used instead.


	# ADDED in creation of 'compare_ml_msa.bch.extJC.6t.I.pl' (on Mar 22 (Tue), 2011) #
our $SBST_MODEL_AA = 'JTT'; # Substitution model for amino acid sequences
our $SBST_MODEL_NT = 'HKY85'; # Substitution model for nucleotide sequences
##our $SBST_MODEL_NT = 'HKY'; # Substitution model for nucleotide sequences --- a BUG fixed on Apr 20 (Wed), 2011
our $CT_RCATE = 1; # #{substitution rate categories}
our $ALPHA =0; # Gamma shape parameter for the rate variation across sites ('': to be determined; 0: uniform rate model)
our $TOPO_SEARCH = 'NNI'; # Tree topology search method. ('NNI', 'SPR', or 'BEST')

	# ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' (on Mar 24 (Thu), 2011) #
	# REVISED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 13 (Wed), 2011
our $RESFREQ_AA = 'm'; # (To be changed to 'e' in the near future)
our $RESFREQ_NT = 'e'; # (To be changed to 'e' in the near future)
##our $RESFREQ = 'm'; # (To be changed to 'e' in the near future)
	# Option on what to do with residue frequencies.
	# = 'e' if you want to use the residue frequencies in the input anchor MSA;
	# = 'm' if you want to use the ML estimates of the nucleotide frequencies, or the amino-acid frequencies defined by the substitution model.

	## Global Variables Necessary for Calculating PRANK Scores. ##
	# (ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl'.) #

our $LOG_ZERO = $MyLikelihoods_respattern_msa_dna::LOG_ZERO;  # MODIFIED in creation of 'examine_purges_via_mblks.2.pl'. #
##our $LOG_ZERO = $MyPRANK_Scores::LOG_ZERO;


		### ADJUST THE FOLLOWING GLOBAL VARIABLES ONLY WHEN DEBUGGING ###

	## (MODIFIED in creation of 'classify_msa_errors.1.pl'.) ##

	## Level of test output for 'MyMalnComparison.pm'.
	## (0, silent; 1, very brief; 2, brief; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

		# For '(br_)list_classes_shift_respos'.
$MyMalnComparison::TEST_LEVEL_lc = 0;
##$MyMalnComparison::TEST_LEVEL_lc = 6;

	## Level of test output for 'MyTreeMap_D.pm'.
	## (Basically, believe in the behaviours of the subroutines in the package.)
	## (ADDED in creation of 'classify_msa_errors.1.pl'.)

		# For 'map_parsim_states_dollo_msa'.
$MyTreeMap_D::TEST_LEVEL = 0;


	## Control the level of verboseness of the test message for 'MyTreeMap_indels_spt_odr.pm'.
	## (Basically, believe in the behaviours of the subroutines in the package.)
	## [RESURRECTED in creation of 'classify_msa_errors.1.pl'.]

$MyTreeMap_indels_spt_odr::TEST_LEVEL1 = 0; # For 'extract_cls_glbocks' and 'msa_rep_gpatterns'.
$MyTreeMap_indels_spt_odr::TEST_LEVEL_SCN = 0; # For 'scan_gbcls_for_long_gaps'.
$MyTreeMap_indels_spt_odr::TEST_LEVEL2i = 0; # For 'create_initial_set_indels_spt_odr'.


		### END of "ADJUST THE FOLLOWING GLOBAL VARIABLES ONLY WHEN DEBUGGING" ###


	## IMPORTED from 'compare_ml_msa.bch.extJC.6t.PhyML.pl' (on Aug 26 (Sun), 2012) ##
	# Now, $main_bin_tree and $main_bin_clmlk are identical, renamed as $main_bin.
	# $src_ctrl and $src_dat are now optional.
	# Added another option "--resfreq", which specifies what are used for the residue frequencies.
	# [ Added another option "--sv_ratio", which specifies the transition/transversion ratio (on Apr 8 (Fri), 2011). ]
	# => Replaced with a pair of options, "--sv_ratio_r" and "--sv_ratio_y," which specify, respectively, the inter-purine (A<->G} and inter-pyrimidine (T<->C) transition/transversion ratios. (on Sep 17 (Thu), 2014).

	# MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_prank_scores.pl' (on Sep 17 (Thu), 2014). #
	# => Further modified, adding three new options, --gaprate, --gapext, and --matchext, on Sep 27 (Sat), 2014.
	# => Further added another new option, ---if_insopen_forever, on Sep 28 (Sun), 2014.

	# MAJOR MODIFICATION in creation of 'classify_msa_errors_via_mblks.2.BK1.pl' (on Oct 6 (Tue), 2015),
	# replacing (..., $out_summ, ...) with (..., $out_summ1, $out_summ2, ...).


		## MODIFIED on May 31 (Tue), 2016. ##

unless (@ARGV > 9) {
	print STDERR "\nUsage:  [perl] $program_name {main_binary} {sequence_type ('AA' for amino acid sequences, 'NT' for  nucleotide sequences)} {anchor_msa_file} {inaln_file_ref} {inaln_file_rec} {inlist_segs} {inlist_params} {out_top_dir} {out_summ_file} {log_dir}\n";
	print STDERR "          [ --sbst_model={substitution model name (default: $SBST_MODEL_AA for amino acids, $SBST_MODEL_NT for nucleotides)} --ct_rcate=\#{substitution rate categories (default: $CT_RCATE) }}  --alpha={gamma shape parameter for rate variation across sites (default: $ALPHA) } --topo_search={tree topology search method (default: $TOPO_SEARCH)} --resfreq={what to do with residue frequencies (default: $RESFREQ_AA for amino acids, $RESFREQ_NT for nucleotides)} --sv_ratio_r={inter-purine (A<->G) transition/transversion ratio (default: undef)} --sv_ratio_y={inter-pyrimidine (T<->C) transition/transversion ratio (default: undef)}  --ctrl={path to the control file (default: undef)} --sbst_mtrx={path to the file for substitution model (MANDATORY for amino-acids)} --intree={input_tree_file (default: undef)}  ";
	##print STDERR "--gaprate={gap opening rate (default: $GAPRATE_PRO for proteins, $GAPRATE_DNA for DNA)}  --gapext={gap extension probability (default: $GAPEXT_PRO for proteins, $GAPEXT_DNA for DNA)}  --matchext={match extension probability (default: $MATCHEXT)}  --if_insopen_forever={ 1 if 'confirm'ed insertions should be open forever, = 0 otherwise (default: $IF_INSOPEN_FOREVER)}"; # DISABLED in creation of 'examine_purges_via_mblks.2.pl'. #
	print STDERR " ]\n"; 
	die "$! .\n\n";
}
my ($main_bin, $seqtype, $anchor_msa_file, $inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params, $outtopdir, $out_summ, $logdir) = @ARGV[0 .. 9];


=DISABLED as of May 31 (Tue), 2016 (1).
unless (@ARGV > 10) {
	print STDERR "\nUsage:  [perl] $program_name {main_binary} {sequence_type ('AA' for amino acid sequences, 'NT' for  nucleotide sequences)} {anchor_msa_file} {inaln_file_ref} {inaln_file_rec} {inlist_segs} {inlist_params} {out_top_dir} {out_summ_file} {log_dir} {list,subject,segment,IDs}\n";
	print STDERR "          [ --sbst_model={substitution model name (default: $SBST_MODEL_AA for amino acids, $SBST_MODEL_NT for nucleotides)} --ct_rcate=\#{substitution rate categories (default: $CT_RCATE) }}  --alpha={gamma shape parameter for rate variation across sites (default: $ALPHA) } --topo_search={tree topology search method (default: $TOPO_SEARCH)} --resfreq={what to do with residue frequencies (default: $RESFREQ_AA for amino acids, $RESFREQ_NT for nucleotides)} --sv_ratio_r={inter-purine (A<->G) transition/transversion ratio (default: undef)} --sv_ratio_y={inter-pyrimidine (T<->C) transition/transversion ratio (default: undef)}  --ctrl={path to the control file (default: undef)} --sbst_mtrx={path to the file for substitution model (MANDATORY for amino-acids)} --intree={input_tree_file (default: undef)}  ";
	##print STDERR "--gaprate={gap opening rate (default: $GAPRATE_PRO for proteins, $GAPRATE_DNA for DNA)}  --gapext={gap extension probability (default: $GAPEXT_PRO for proteins, $GAPEXT_DNA for DNA)}  --matchext={match extension probability (default: $MATCHEXT)}  --if_insopen_forever={ 1 if 'confirm'ed insertions should be open forever, = 0 otherwise (default: $IF_INSOPEN_FOREVER)}"; # DISABLED in creation of 'examine_purges_via_mblks.2.pl'. #
	print STDERR " ]\n"; 
	die "$! .\n\n";
}
my ($main_bin, $seqtype, $anchor_msa_file, $inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params, $outtopdir, $out_summ, $logdir, $list_sbjct_segments) = @ARGV[0 .. 10];

my %segid2if_sbjct; # ($segment_ID => 1, ...) for ALL the subject segments.
foreach my $segid (split /\,/, $list_sbjct_segments) { $segid2if_sbjct{$segid} = 1; }

$list_sbjct_segments =~ s/\,/\,\ /g;
=cut # END of "DISABLED as of May 31 (Tue), 2016 (1)."


my $main_bin_tree = my $main_bin_clmlk = $main_bin; # These

$seqtype = uc $seqtype;
unless (($seqtype eq 'AA') or ($seqtype eq 'NT')) {
	print STDOUT "[ERROR '$program_name'] ILLEGAL INPUT: The sequence type '$seqtype' is not expected. Please specify either 'AA' or 'NT' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

unless ($seqtype eq 'NT') { # ADDED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
	print STDOUT "[ERROR '$program_name'] Currently, this script CANNOT HANDLE the sequence TYPE, '$seqtype', which is DIFFERENT FROM 'NT' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


	## END of "ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 6 (Tue), 2015)." #

my $sbst_model = ($seqtype eq 'AA') ? $SBST_MODEL_AA : $SBST_MODEL_NT;
my $ct_rcate = $CT_RCATE;
my $alpha = $ALPHA;
my $topo_search = $TOPO_SEARCH;


	# MODIFIED in creation of 'compare_ref_vs_rec_prank_scores.pl'. #
my ($sv_ratio_r, $sv_ratio_y) ;
##my $sv_ratio ;

my $resfreq = ($seqtype eq 'AA') ? $RESFREQ_AA : $RESFREQ_NT;

my ($intree, $src_ctrl, $src_dat); # Input tree file, control file specifying detailed options, and data file specifying the substitution model


for (my $k=10; $k<@ARGV; $k++) { # MODIFIED AS OF May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
##for (my $k=11; $k<@ARGV; $k++) {...} # MODIFIED in creation of 'examine_purges_via_mblks.2.pl' (on May 9 (Mon), 2016). #
##for (my $k=11; $k<@ARGV; $k++) {...} # MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 6 (Tue), 2015). #
##for (my $k=10; $k<@ARGV; $k++) {...} # MODIFIED in creation of 'compare_ref_vs_rec_prank_scores.pl' (on Sep 29 (Mon), 2014). #

	my $arg = $ARGV[$k];
	unless ($arg =~ s/^\-\-?//) {
	    print STDOUT "[WARNING '$program_name'] Found an UNEXPECTED command-line ARGUMENT '$arg': $! !\n";
	    print STDOUT "                 ---  I will just skip it ...\n\n";
	    next;
	}
	if ($arg =~ s/^intree\=// ) {
	    ($arg =~ m/\S/) and ($intree = $arg);

		# MODIFIED in creation of 'compare_ref_vs_rec_prank_scores.pl'. #
	} elsif ($arg =~ s/^sv_ratio_r\=//) {
	    ($arg =~ m/\S/) and ($sv_ratio_r = $arg);
	} elsif ($arg =~ s/^sv_ratio_y\=//) {
	    ($arg =~ m/\S/) and ($sv_ratio_y = $arg);

	##} elsif ($arg =~ s/^sv_ratio\=//) {
	##    ($arg =~ m/\S/) and ($sv_ratio = $arg);

	} elsif ($arg =~ s/^ctrl\=//) {
	    ($arg =~ m/\S/) and ($src_ctrl = $arg);
	} elsif ($arg =~ s/^sbst_mtrx\=//) {
	    ($arg =~ m/\S/) and ($src_dat = $arg);
	} elsif ($arg =~ s/^resfreq\=//) {
	     if (($arg =~ m/\S/ ) and ($arg ne $resfreq)) {
		print STDOUT "[INFO '$program_name'] The argument \$resfreq = '$resfreq' will be changed to '$arg' ...\n\n";
		$resfreq = $arg;
	     }

	} elsif ($arg =~ s/^sbst_model\=//) {
	     if (($arg =~ m/\S/ ) and ($arg ne $sbst_model)) {
		print STDOUT "[INFO '$program_name'] The argument \$sbst_model = '$sbst_model' will be changed to '$arg' ...\n\n";
		$sbst_model = $arg;
	     }
	} elsif ($arg =~ s/^ct_rcate\=// ) {
	    if (($arg =~ m/\S/ ) and ($arg ne $ct_rcate)) {
		print STDOUT "[INFO '$program_name'] The argument \$ct_rcate = '$ct_rcate' will be changed to '$arg' ...\n\n";
		$ct_rcate = $arg;
		##$sbst_model = $arg; # A BUG FIXED on March 23 (Wed), 2011 #
	    }
	} elsif ($arg =~ s/^alpha\=// ) {
	    if ($arg =~ m/\S/ ) {
		if ($arg ne $alpha) {
		   print STDOUT "[INFO '$program_name'] The argument \$alpha = '$alpha' will be changed to '$arg' ...\n\n";
		   $alpha = $arg;
		   ##$sbst_model = $arg; # A BUG FIXED on March 23 (Wed), 2011 #
		}
	    } else {
		print STDOUT "[INFO '$program_name'] I found \$alpha = '', I will regard this as an order to UNDEFINE it (, which means optimizing it) ...\n\n";
		$alpha = undef;
	    }
	} elsif ($arg =~ s/^topo_search\=// ) {
	    if (($arg =~ m/\S/ ) and ($arg ne $topo_search)) {
		print STDOUT "[INFO '$program_name'] The argument \$topo_search = '$topo_search' will be changed to '$arg' ...\n\n";
		$topo_search = $arg;
	    }

		# ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl' (on Sep 27 (Sat), 2014). #

	} else {
	    print STDOUT "[WARNING '$program_name'] Found an UNEXPECTED command-line ARGUMENT '--$arg': $! !\n";
	    print STDOUT "                 ---  I will just skip it ...\n\n";
	}
}

	# ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl'. #
my $sv_ratio = ((defined $sv_ratio_r) and ($sv_ratio_y)) ? ($sv_ratio_r + $sv_ratio_y)/2 : undef;

	## Check the consistency of the command-line arguments #

unless (defined $sbst_model) { # ADDED a long-overdue block on Sep 18 (Thu), 2014 (while creating  ompare_ref_vs_rec_prank_scores.pl').
	print STDOUT "[ERROR '$program_name'] The substitition model is UNSPECIFIED !!\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

if ($seqtype eq 'AA') {
    unless (($sbst_model eq 'custom') or (defined $SBSTMODEL2FILENAME{$sbst_model})) {
	print STDOUT "[ERROR '$program_name'] Sorry, the amino-acid substitution model '$sbst_model' CANNOT BE USED for this version !!\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

} else {

    unless (defined $SUPPORTED_SBSTMODELS{$sbst_model}) {
	print STDOUT "[ERROR '$program_name'] Sorry, the nucleotide substitution model '$sbst_model' is NOT SUPPORTED by this version !!\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

my $sbst_model_phyml = ($sbst_model eq 'TN93') ? 'HKY85' : $sbst_model ; # ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl'. #


if ($seqtype eq 'AA') { # DEBUGGED on Apr 20 (Wed), 2011 #
##if (($seqtype eq 'AA') and ($resfreq eq 'e')) {
    unless (defined $src_dat) {
	print STDOUT "[ERROR '$program_name'] The PATH to the FILE for the amino-acid substitution model '$sbst_model' is UNSPECIFIED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

if (($resfreq ne 'm') and ($resfreq ne 'e') and ($seqtype ne 'NT')) {
	print STDOUT "[ERROR '$program_name'] The residue frequency option '$resfreq' is NOT COMPATIBLE with the sequence type '$seqtype' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
if ((defined $intree) and ($intree =~ m/\S/) and ($seqtype eq 'NT') and ($resfreq eq 'm')) {
	print STDOUT "[ERROR '$program_name'] The option '-seqtype=NT -resfreq=m' is NOT COMPATIBLE with the use of an input tree ($intree), please use a different option for -resfreq !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
if ( (defined $sv_ratio) and (($sbst_model eq 'JC69') or ($sbst_model eq 'F81')) ) {
	print STDOUT "[WARNING '$program_name'] The option '--sv_ratio=$sv_ratio' is NOT NECESSARY for the substitution model '$sbst_model' !\n";
	print STDOUT "          --- I will UNDEFINE it ...\n\n";
	$sv_ratio = undef;

	$sv_ratio_y = $sv_ratio_r = undef; # ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl'. #
}
if (($sbst_model eq 'JC69') or ($sbst_model eq 'K80')) {
	print STDOUT "[WARNING '$program_name'] The substitution model '$sbst_model' has uniform equilibrium frequencies (fA = fC = fG = fT = 0.25), so I will SET \$resfreq = '' !\n";
	$resfreq = '';
}

if ((defined $intree) and !(defined $alpha)) {
	print STDOUT "[ERROR '$program_name'] Now \$alpha = undef, but I canNOT ML-estimate the shape parameter when an input tree file is specified (as '$intree') !\n";
	print STDOUT "[ERROR '$program_name'] $!\n\n";
	die ;
}


	# Out of command-line arguments,
	# build up options shared by the tree inference step and
	# the column-wise likelihood calculation step.

my $seqtype_lc = lc $seqtype;

our $OPTIONS_SHARED = " -d $seqtype_lc -c $ct_rcate ";
## (($ct_rcate > 1) and (defined $alpha) and ($alpha > 0)) and ($OPTIONS_SHARED .= " -a $alpha");



	### Input/Output files/directories ###

	## Interim files ##

my $name_interim_trees = "tree_whole_full.txt";
my $interim_trees = "$outtopdir/$name_interim_trees";

my $name_log_interim_trees = "log_$name_interim_trees";
my $log_interim_trees = "$logdir/$name_log_interim_trees";

my $interim_aln_name = 'input_aln.fa'; # $interim_aln = "$outdir/$interim_aln_name"
my $tmpfile_name = 'lnl_msa.txt';

=DISABLED in creation of 'examine_complex_via_mblks.2.pl' (1).
	# $int_clmlk_file_ref = "$top_clmlk_dir_ref/$pref_clmlk\{seg_id\}$suff_clmlk".
	# $int_clmlk_file_rec = "$top_clmlk_dir_rec/$pref_clmlk\{seg_id\}$suff_clmlk".
my $top_clmlk_dir_ref = "$outtopdir/Clm_lnlk/Ref";
my $top_clmlk_dir_rec = "$outtopdir/Clm_lnlk/Rec";
my $pref_clmlk = 'lnlk_clm.';
my $suff_clmlk = '.txt';

	# Top Temporary directory for PAML #
	# temporary directory (reference) = "$top_tmpdir_ref/s{seg_ID}/g{indx_gpattern}/t{indx_tree}"
	# temporary directory (reconstructed) = "$top_tmpdir_rec/s{seg_ID}/g{indx_gpattern}/t{indx_tree}"
my $top_tmpdir_ref = "$outtopdir/Tmp/Ref";
my $top_tmpdir_rec = "$outtopdir/Tmp/Rec";

=cut # END of "DISABLED in creation of 'examine_complex_via_mblks.2.pl' (1)."


	# Temporary directory for phylogenetic tree reconstruction #
	# $inaln_for_tree = "$tmpdir_phyml/$inaln_for_tree_name";
my $tmpdir_phyml = "$outtopdir/Infer_tree";

	# Master file for a custom substitution model (only for $seqtype eq 'AA' and $resfreq eq 'e') #
	# $master_sbst_file = "$outtopdir/$dat_name";
	# $master_ctrl_file = "$outtopdir/$ctrl_name";
my $master_sbst_file;
my $master_ctrl_file;


	## Check Necessary Files/Directories ##

foreach my $binary ($main_bin_tree) { # MODIFIED in creation of 'bch_pbs_cmp_mafft_consist.D.pl'. #
##foreach my $binary ($main_bin_tree, $main_bin_clmlk) {...}
    unless (( -f $binary ) and ( -x $binary )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND an executable PROGRAM '$binary' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

if (defined $src_ctrl) {
   unless (( -f $src_ctrl ) and ( -r $src_ctrl )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable PAML CONTROL FILE '$src_ctrl' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
   }
}
if (defined $src_dat) {
   unless (( -f $src_dat ) and ( -r $src_dat )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable PAML CONTROL FILE '$src_dat' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
   }
}

my ($ctrl_name) = (defined $src_ctrl) ? ($src_ctrl =~ m!([^/\s]+)\s*$!) : ('ctrl_phyml.txt');
my ($dat_name) = (defined $src_dat) ? ($src_dat =~ m!([^/\s]+)\s*$!) : ('undef');
my $log_name = (defined $src_ctrl) ? "log_$ctrl_name.txt" : "log_PhyML.txt";


if (($seqtype eq 'AA') and (defined $dat_name)) {
    my $dat_name0 = $SBSTMODEL2FILENAME{$sbst_model};
    unless (($dat_name eq 'custom') or ($dat_name eq $dat_name0)) {
	print STDOUT "[ERROR '$program_name'] The DATA FILE NAME for the amino-acid substitution model '$sbst_model' is NOT '$dat_name', as you specified, BUT '$dat_name0' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

if (($seqtype eq 'AA') and ($resfreq eq 'e')) {
	$dat_name =~ s/\.dat$//;
	##$dat_name =~ s/\.dat^//; # A BUG Fixed on Mar 29 (Tue), 2011 #

	$dat_name .= '_custom.dat';
	$master_sbst_file = "$outtopdir/$dat_name";
	$master_ctrl_file = "$outtopdir/$ctrl_name";
}


foreach my $file ($intree) {
    (defined $file) or next; ## ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PAML.pl' (on Mar 22 (Tue), 2011) ##

    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 8 (Fri), 2014). #

foreach my $infile ($inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params) { # MODIFIED in creation of 'compare_ref_vs_rec_prank_scores.pl'. #
##foreach my $infile ($inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params, $inlist_lclhom) {...} # MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
##foreach my $infile ($inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params) {...}
    unless (( -f $infile ) and ( -r $infile )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$infile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ; 
    }
}

foreach my $outdir ($outtopdir, $logdir, $tmpdir_phyml) { # MODIFIED in creation of 'examine_complex_via_mblks.2.pl'. #
##foreach my $outdir ($outtopdir, $logdir, $top_clmlk_dir_ref, $top_clmlk_dir_rec, $top_tmpdir_ref, $top_tmpdir_rec, $tmpdir_phyml) {...} # For the predecessor 'examine_purges_via_mblks.2.pl'. #
    unless ( -d $outdir ) {
	unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create a TOP OUTPUT DIRECTORY '$outdir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
    }
    unless (( -r $outdir ) and ( -w $outdir )) {
	unless (0 == system ("chmod u+rw $outdir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to CHANGE the permission MODES of the top output directory '$outdir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<<<<<< Program name: $program_name (path: $program_path) >>>>>>>>\n\n";

print STDOUT "Current working directory: $CURR_WORK_DIR .\n\n";


print STDOUT "Binary programs:\n";
print STDOUT "     (for tree inference)  $main_bin_tree ,\n";

print STDOUT "Source control file (for PhyML): ", ((defined $src_ctrl) ? $src_ctrl : 'Not used'), " .\n";
print STDOUT "Source file of substitution model (for PhyML): ", ((defined $src_dat) ?  $src_dat : 'Not used'), " .\n\n";


print STDOUT "Sequence type: ", (($seqtype eq 'AA') ? 'Amino Acid' : 'Nucleotide'), " .\n\n";

print STDOUT "Input files:\n";

if (defined $intree) { 
    print STDOUT "     (Tree)      $intree .\n\n";
}
print STDOUT "    (anchor MSA) $anchor_msa_file .\n\n";


	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 8 (Fri), 2014). #

print STDOUT "   (simulation parameters} $inlist_params,\n";
print STDOUT "   (reference MSA)         $inaln_file_ref,\n";
print STDOUT "   (reconstructed MSA)     $inaln_file_rec,\n";
print STDOUT "   (list of segments)      $inlist_segs .\n\n";

###print STDOUT "List of Subject Segments: $list_sbjct_segments .\n\n"; # ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 9 (Mon), 2016). # DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #

print STDOUT "Directory for interim output files for tree inference: $tmpdir_phyml .\n\n";

print STDOUT "Interim output list of candidate trees:  {topoutdir}/$name_interim_trees .\n";

if (defined $master_sbst_file) {
    print STDOUT "Master file for a custom  substitution model: {topoutdir}/$dat_name .\n";
}
if (defined $master_ctrl_file) {
    print STDOUT "Master control file for PhyML: {topoutdir}/$ctrl_name .\n";
}


=DISABLED in creation of 'examine_complex_via_mblks.2.pl' (2).

print STDOUT "Temporary directories for PAML (obsolete):  {top_tmpdir}/s{seg_id}/g{indx_gpattern}/t{indx_tree}/,\n"; # MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
##print STDOUT "Temporary directories for PAML:  {top_tmpdir}/s{seg_id}/g{indx_gpattern}/t{indx_tree}/,\n";
print STDOUT "   where {top_tmpdir} = $top_tmpdir_ref for the reference MSA,\n";
print STDOUT "                      = $top_tmpdir_rec for the reconstructed MSA.\n\n";

=cut # END of "DISABLED in creation of 'examine_complex_via_mblks.2.pl' (2)."

print STDOUT "Summary output: $out_summ .\n\n"; # RESURRECTED in creation of 'examine_purges_via_mblks.2.pl'.

print STDOUT "Directory for the log files: {log_dir} = $logdir .\n\n";
print STDOUT "Log files:\n";

print STDOUT "   (for candidate tree enumeration) {log_dir}/$name_log_interim_trees .\n\n";

		# MODIFIED in creation of 'classify_msa_errors.1.pl'. #
print STDOUT "   (for analyzing the reference & reconstructed MSAs for each segment) {log_dir}log_seg_anal.\{seg_id\}.txt .\n\n";


if ($FLG_RMV_TMPOUT) {
	print STDOUT "* All the temporary outputs will be removed .\n\n";
} else {
	print STDOUT "* All the temporary outputs will be kept .\n\n";
}


print STDOUT "Parameters for tree inference (via PhyML):\n\n";
print STDOUT "   Substitution model:  $sbst_model .\n";
print STDOUT "   (Model for PhyML: $sbst_model_phyml .)\n\n"; # ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl'. #

print STDOUT "   #{Substitution rate categories}: $ct_rcate .\n";
print STDOUT "   Gamma shape parameter for the rate variation across sites: ", ((defined $alpha) ? $alpha : 'undef'), " .\n";
print STDOUT "   Tree topology search method: $topo_search .\n";

print STDOUT "   Option on what to do with the residue frequencies: ", (($resfreq =~ m/\S/) ? $resfreq : "uniform (default for $sbst_model)"), " .\n\n\n";

if ($seqtype eq 'NT') {
		# MODIFIED in creation of 'compare_ref_vs_rec_prank_scores.pl'. #
	print STDOUT "   Transition/transversion ratios:\n";
	print STDOUT "        (for A <-> G) ", ((defined $sv_ratio_r) ? $sv_ratio_r : 'undef'), " ,\n";
	print STDOUT "        (for T <-> C) ", ((defined $sv_ratio_y) ? $sv_ratio_y : 'undef'), " ,\n";
	print STDOUT "       (average (to be used for PhyML (if applicable))) ", ((defined $sv_ratio) ? $sv_ratio : 'undef'), " .\n\n\n";
    ##print STDOUT "   Transition/transversion ratio: ", ((defined $sv_ratio) ? $sv_ratio : 'undef'), " .\n";
}

print STDOUT "\nOptions for PhyML shared by both tree inference and column-wise likelihood calculation:\n\n";
##print STDOUT "     '$OPTIONS_SHARED' .\n\n";
print STDOUT '   $OPTIONS_SHARED = ', "'$OPTIONS_SHARED' .\n\n";

print STDOUT "\nLevels of tests for debugging or behavior tests:\n\n";

print STDOUT '   $TEST_LEVEL        =  ', $TEST_LEVEL, " ,\n";

print STDOUT '   $TEST_LEVEL2       =  ', $TEST_LEVEL2, " ,\n"; # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 8 (Thu), 2015). #

print STDOUT '   $TEST_LEVEL3        =  ', $TEST_LEVEL3, " ,\n"; # ADDED in creation of 'examine_purges_via_mblks.2.pl'. #


print STDOUT '   $TEST_LEVEL_newtype   = ', $TEST_LEVEL_newtype,  " ,\n"; # ADDED on Oct 3 (Fri), 2015. #

print STDOUT '   $TEST_LEVEL_cipc   = ', $TEST_LEVEL_cipc,  " ,\n";

print STDOUT '   $TEST_LEVEL_cipc2   = ', $TEST_LEVEL_cipc2,  " ,\n"; # ADDED on Oct 9 (Fri), 2015. #
print STDOUT '   $TEST_LEVEL_umlp   = ', $TEST_LEVEL_umlp,  " ,\n"; # ADDED on Oct 16 (Fri), 2015.

print STDOUT '   $DBG_LEVEL_EXPHYML =  ', $DBG_LEVEL_EXPHYML, " ,\n";
print STDOUT '   $FLG_RMV_TMPOUT    =  ', $FLG_RMV_TMPOUT, " .\n\n";


	# MODOFIED in creation of 'classify_msa_errors.1.pl'. #

print STDOUT "\nParameters in the Perl package 'MyMalnComparison.pm':\n\n";

print STDOUT '$MyMalnComparison::TEST_LEVEL_lc  = ', $MyMalnComparison::TEST_LEVEL_lc , " .\n\n";

		# (Further ADDED on Oct 20 (Mon), 2014.) #

print STDOUT "\nParameters in the Perl package 'MyTreeMap_indels_spt_odr.pm':\n\n";

print STDOUT '$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_LEN_GBCLS = ', $UPPER_BOUND_ANAL_LEN_GBCLS, " ,\n";

print STDOUT '$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN = ', $UPPER_BOUND_ANAL_GAPLEN, " .\n\n\n";

	### Read in the list of segments. ###

print STDOUT "\n<<<<< Now, read in the list of segments >>>>>\n\n";

our @LABELS_SEG_ATTR = ('Indx_spacer', 'Start_in_Ref', 'End_in_ref', 'Start_in_Rec', 'End_in_Rec', 'Diagnosis'); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
##our @LABELS_SEG_ATTR = ('seg_type', 'ref_len', 'rec_len', '#len_grp', 'alt_state');

# @{$list_segments[$s]} = @attrs gives information on the $s th segment, including its ID.
#
my $list_segments = readin_list_segs2 ($inlist_segs, @LABELS_SEG_ATTR);# MODIFIED  in creation of 'compare_ref_vs_rec_compllimment.pl'. #

# @{$list_segments[$s]} = ($seg_ID, @attrs) gives information on the $s th segment.
#
##my $list_segments = readin_list_segs ($inlist_segs, %LABEL_SATTR2INDX);
unless (defined $list_segments) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_list_segs2' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_segments = @{$list_segments};

print STDOUT "\n#{segments} = $ct_segments\n\n";
print STDOUT join ("\t", 'Seg_ID', @LABELS_SEG_ATTR[1..$#LABELS_SEG_ATTR]), "\n\n"; # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
##print STDOUT join ("\t", 'Seg_ID', @LABELS_SEG_ATTR), "\n\n";

for (my $s=0; $s<$ct_segments; $s++) { print STDOUT join ("\t",  @{$list_segments->[$s]}), "\n"; }
print STDOUT "\n\n";

=set aside for later use
my $NEW_INDX_DIAG; # It points to the element of 'Diagnosis' in the segment information. #
for (my $i=0; $i<@LABELS_SEG_ATTR; $i++) { # Preliminary for-loop (over field labels). #
    ($LABELS_SEG_ATTR[$i] eq 'Diagnosis') or next;
    $NEW_INDX_DIAG = $i;
    last;
}				# End of the preliminary for-loop (over field labels). #
=cut


	### Read in the reference MSA ###
	### (ADDED in creation of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' on Aug 26 (Sun), 2012) ###

print STDOUT "\n<<<<< Now, read in the reference MSA >>>>>\n\n";

(my $ref_msa, my $sequence_ids, undef) = extract_alignment ($inaln_file_ref); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##(my $ref_msa, my $sequence_ids, undef, undef) = extract_fasta ($inaln_file_ref);
unless (defined $ref_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the reference MSA '$inaln_file_ref' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_sequences = @{$sequence_ids};
unless ($ct_sequences > 1) {
	print STDOUT "[ERROR '$program_name'] The number of sequences ($ct_sequences) is TOO SMALL ! I will stop here ...\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $totlen_refaln = length ${$ref_msa->[0]};

print STDOUT "#{sequences in MSA} =  $ct_sequences .\n";
print STDOUT "#{columns in MSA}   = $totlen_refaln .\n\n";

	### Read in the reconstructed MSA ###
	### (ADDED in creation of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' on Aug 26 (Sun), 2012) ###

print STDOUT "\n<<<<< Now, read in the reconstructed MSA >>>>>\n\n";

(my $rec_msa, my $seq_ids2, undef) = extract_alignment ($inaln_file_rec); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##(my $rec_msa, my $seq_ids2, undef, undef) = extract_fasta ($inaln_file_rec);
unless (defined $rec_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the reconstructed MSA '$inaln_file_rec' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


unless ($ct_sequences == @{$seq_ids2}) {
	print STDOUT "[ERROR '$program_name'] THe number of sequences in the core alignment (", scalar (@{$seq_ids2}), ") CLASHES with that of the reference alignment ($ct_sequences) !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
for (my $k=0; $k<$ct_sequences; $k++) {
	my $seqid = $sequence_ids->[$k];
	my $seqid2 = $seq_ids2->[$k];
	unless ($seqid eq $seqid2) {
	    print STDOUT "[ERROR '$program_name'] The ID of the $k th sequence in the reconstructed MSA ($seqid2) does NOT MATCH that in the reference MSA ($seqid) !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
}

my $totlen_recaln = length ${$rec_msa->[0]};

print STDOUT "#{columns in MSA} = $totlen_recaln .\n\n"; 


	### Extract a list of sequence names (IDs), as well as concatenated anchor MSAs, from the core alignment ###
	### (MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' on Aug 26 (Sun), 2012) ###

print STDOUT "\n<<<<< Now, read in the core alignment (to be fed into tree reconstruction, if necessary) >>>>>\n\n";

(my $anchor_msa, my $sequence_ids_core, undef, undef) = extract_alignment ($anchor_msa_file); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). # 
##(my $anchor_msa, my $sequence_ids_core, undef, undef) = extract_fasta ($anchor_msa_file);
unless (defined $sequence_ids_core) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the core alignment '$anchor_msa_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
unless ($ct_sequences == @{$sequence_ids_core}) {
	print STDOUT "[ERROR '$program_name'] THe number of sequences in the core alignment (", scalar (@{$sequence_ids_core}), ") CLASHES with that of the reference alignment ($ct_sequences) !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
for (my $k=0; $k<$ct_sequences; $k++) {
	my $seqid = $sequence_ids->[$k];
	my $seqid2 = $sequence_ids_core->[$k];
	unless ($seqid eq $seqid2) {
	    print STDOUT "[ERROR '$program_name'] The ID of the $k th sequence in the core alignment ($seqid2) does NOT MATCH that in the reference MSA ($seqid) !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
}

my $ancaln_len = length ${$anchor_msa->[0]};

print STDOUT "#{columns in MSA} = $ancaln_len .\n\n"; 


	### Give nominal frequencies to the residues. ###

print STDOUT "\n<<<<< Now, give nominal frequencies to the residues >>>>>\n\n";

our %RES2FREQ; # = ($residue => $equil_freq, ...)
our @RESIDUES = ($seqtype eq 'NT') ? qw( A C G T ) : @MyPAML_parser::AMINO_ACIDS ;
our $CT_RES = @RESIDUES;

our @RES_FREQS = ();

if ($resfreq =~ m/\S/) {
  if ($resfreq eq 'm') {
    if ($seqtype eq 'AA') {

      unless ( retrieve_aa_freqs ($src_dat, %RES2FREQ) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'retrieve_aa_freq' FAILED on the input data file '$src_dat' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
      }
      print STDOUT "\n --- Retrieved the Amino-acid frequencies from the substitution model ...\n\n";

    } else {
	# Readin the ML estimates of nucleotide frequencies

      print STDOUT "\n --- The nucleotide frequencies will be estimated via ML later ...\n\n";
    }

  } elsif ($resfreq eq 'e') {

	# Calculate the residue frequencies in the core alignment #

      my $fl_caseignore = 1; # = 1 to convert lower-case letters to upper-case letters, = 0 to keep lower-case letters as they are.
      my $fl_relative = 1; # = 1 to compute the relative frequencies, = 0 to compute the absolute frequencies (i.e. the counts), of the residues.
	
      my $totct_res = est_resfreqs_frm_msa (@{$anchor_msa}, $GAP, %RES2FREQ, $fl_caseignore, $fl_relative);
      unless (defined $totct_res) {
	print STDOUT "[ERROR '$program_name'] 'est_resfreqs_frm_msa' FAILED on the anchor MSA !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
      }
      unless ($totct_res > 0) {
	print STDOUT "[ERROR '$program_name'] NO RESIDUES other than gaps ($GAP) in the anchor MSA !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
      }
      print STDERR "\n --- estimated the residue frequencies out of the anchor MSA (total residue number = $totct_res) ...\n\n";

  } elsif ($seqtype eq 'NT') { # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 8 (Fri), 2011 #

      my @freqs = split /\,/, $resfreq;
      unless (@freqs==4) {
	print STDOUT "[ERROR '$program_name'] The number (", scalar (@freqs), ") of user-specified frequencies = (", join (', ', @freqs), ") CLASHES with 4, which is the number of residues (", join (', ', @RESIDUES), ") !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
      }
      for (my $r=0; $r<4; $r++) { $RES2FREQ{$RESIDUES[$r]} = $freqs[$r]; }
  }

} elsif (($sbst_model eq 'JC69') or ($sbst_model eq 'K80')) {
      for (my $r=0; $r<4; $r++) { $RES2FREQ{$RESIDUES[$r]} = 1/4; }
} else {
	print STDOUT "[ERROR '$program_name'] I do NOT SUPPORT the option \$resfreq = '$resfreq' for the substitution model '$sbst_model' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

	## Normalize the frequencies. ##

unless (($seqtype eq 'NT') and ($resfreq eq 'm')) {

    my $tot_resfreq = 0; # ADDED on Mar 29 (Tue), 2011 #
    foreach my $res (@RESIDUES) { # The 1st foreach-loop over residues #
	my $freq = $RES2FREQ{$res};
	unless (defined $freq) {
	   if ($resfreq eq 'm') {
	 	print STDERR "[ERROR '$program_name'] UNDEFINED FREQUENCY for the residue '$res' !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    } else {
		print STDOUT "[WARNING '$program_name'] The frequency for the residue '$res' is UNDEFINED : $! !\n";
		print STDOUT "        --- I will set it to $MIN_FREQ ...\n\n";
		##$freq = $RES2FREQ{$res} = 0;
		$freq = $RES2FREQ{$res} = $MIN_FREQ;
	    }
	}
	$tot_resfreq += $freq; # ADDED on Mar 29 (Tue), 2011 #
    }			# End of the 1st foreach-loop over residues #
    
		# Normalize the frequency (ADDED on Mar 29 (Tue), 2011) #

    foreach my $res (@RESIDUES) { # The 2nd foreach-loop over residues #
 	my $freq = $RES2FREQ{$res} = sprintf ("%9.7f", $RES2FREQ{$res} /$tot_resfreq );
	push @RES_FREQS, $freq;
    }			# End of the foreach-loop over residues #

    if ($resfreq eq 'm') {
	print STDOUT "\n\n<< Residue frequencies defined in the model >>\n\n";
    } else {
	print STDOUT "\n\n<< Residue frequencies estimated from the core alignment >>\n\n";
    }
    print STDOUT join ("\t", 'Residue', 'Frequency'), "\n\n";

    foreach my $res (@RESIDUES) { print STDOUT join ("\t", $res, $RES2FREQ{$res}), "\n"; }
    print STDOUT "\n\n";
}


	### Make a master file for a custom amino-acid substitution model ###

if (($seqtype eq 'AA') and ($resfreq eq 'e')) {

    print STDOUT "\n<<<<< Now, make master files for a custom amino-acid substitution model >>>>>\n\n";

    my $resct_in_line = 7; # Maximum number of residues whose frequencies are recorded in a line.
    my $endmark = ''; # A string that marks the end of the substitution model in the file #
    unless ( print_custom_sbst_model ($src_dat, $master_sbst_file, @RES_FREQS, $resct_in_line, $endmark) ) {
	print STDOUT "[ERROR '$program_name'] 'print_custom_sbst_model' FAILED to print out the pre-substitution matrix '$src_dat' combined with the empirical amino-acid frequencies:\n";
	print STDOUT join (', ', @RES_FREQS), " !!\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    my $fh_otmp = open_output_file ($master_ctrl_file, 0);
    unless (defined $fh_otmp) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the master control file '$master_ctrl_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    print $fh_otmp $dat_name, "\n";
    close ($fh_otmp);

    foreach my $master_file ($master_sbst_file, $master_ctrl_file) {
	unless (( -f $master_file ) and ( -r $master_file )) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a READABLE master file '$master_file' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
    }

    print STDOUT "(Master file of the model) $master_sbst_file ,\n";
    print STDOUT "(Master control file)      $master_ctrl_file .\n\n"; 
}


	### Either retrieve or reconstruct a phylogenetic tree that our main process is based on. ###

my @std_set_cand_trees = ();
our %STATS_PHYML;
	# = ($key => scalar_value ($key is the string between '.' and ':', which ' ' replaced with '_'),
	# 'discrete_gamma' => [$ncate, $alpha] (or undef),
	# 'nuc_freqs' => [$fA, $fC, $fG, $fT] (or undef),
	# 'rel_rates' => [$r_AC, $r_AG, $r_AT, $r_CG, $r_CT, $r_GT] (or undef),
	# 'rate_mtrx' => \@mtrx, with $mtrx[$i]->[$j] giving the instantaneous rate from the $i-th to the $j-th nucleotide (in the order: A, C, G, T) (or undef)
	# );
our @REL_RATES = ();

if ((defined $intree) and ($intree =~ m/\S/)) {

	## Read in the input tree (ESSENTIAL for the 'beta' version) ##

    print STDOUT "\n<<<<< Reading-in the Input Tree >>>>>\n\n";

    # @{$info_dawg_tree} = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr)
    #
    my $info_dawg_tree = readin_dawg_tree2 ($intree); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
    ##my $info_dawg_tree = readin_dawg_tree ($intree);
    unless (defined $info_dawg_tree) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'readin_dawg_tree2' FAILED on the input tree file '$intree' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    @std_set_cand_trees = (['NA', 'NA', 'NA', 0, 0, @{$info_dawg_tree}]);

} else {

	## Infer a phylogenetic tree from the core alignment ##

		# Prepare an input alignment #

    ## my $tmpdir_phyml = "$outtopdir/Infer_tree";
    my $inaln_for_tree_name = 'in_phyml.seq';
    my $inaln_for_tree = "$tmpdir_phyml/$inaln_for_tree_name";

    my $fh_tmpaln = open_output_file ($inaln_for_tree, 0);
    unless (defined $fh_tmpaln) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the input alignment file for the tree inference '$inaln_for_tree' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    print $fh_tmpaln "$ct_sequences $ancaln_len\n";

    unless ( pr_phylip_aln (@{$sequence_ids}, @{$anchor_msa}, $DIS_LEN, *${$fh_tmpaln}) ) {
	print STDOUT "[ERROR '$program_name'] 'pr_phylip_aln' FAILED to output the input alignment for the tree inference into the file '$inaln_for_tree' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    close ($fh_tmpaln);

		# Put together command-line options #

    ## my $sbst_model = ($seqtype eq 'AA') ? $SBST_MODEL_AA : $SBST_MODEL_NT;
    ## my $ct_rcate = $CT_RCATE;
    ## my $alpha = $ALPHA;
    ## my $topo_search = $TOPO_SEARCH;

    my $ct_bootstrap = 0; # 0: Neither approximate likelihood ratio test nor bootstrap values are computed.
			 # -1: Approximate likelihood ratio test returning aLRT statistics.
    my $opt_params = 'tlr'; # Tree topology (t), branch lengths (l), and substitution rate parameters (r) are optimized.


    my $options = " -i $inaln_for_tree_name $OPTIONS_SHARED -b $ct_bootstrap -s $topo_search -o $opt_params";
    (($ct_rcate > 1) and (defined $alpha) and ($alpha > 0)) and ($options .= " -a $alpha");

    (($seqtype eq 'NT') and (defined $sv_ratio)) and ($options .= " -t $sv_ratio");

		# Paths to the output/log files # 

    my $name_outtree = "${inaln_for_tree_name}_phyml_tree.txt";
    my $out_tree = "$tmpdir_phyml/$name_outtree";

	# A BUG Fixed on Mar 30 (Wed), 2011 #
    my $name_outstat = "${inaln_for_tree_name}_phyml_stats.txt"; # ADDED on Mar 29 (Tue), 2011 #
    ##my $name_outstat = "${inaln_for_tree_name}_phyml_stat.txt"; # ADDED on Mar 29 (Tue), 2011 #

    my $out_stat = "$tmpdir_phyml/$name_outstat"; # ADDED on Mar 29 (Tue), 2011 #
    my $name_log_tree = "log_phyml.txt";
    my $log_tree = "$tmpdir_phyml/$name_log_tree";

    if (($seqtype eq 'AA') and ($resfreq eq  'e')) {

	my $tmp_sbst_file = "$tmpdir_phyml/$dat_name";
	my $tmp_ctrl_file = "$tmpdir_phyml/$ctrl_name";

	unless (0 == system ("cp -p $master_sbst_file $tmp_sbst_file") ) {
	    print STDOUT "[ERROR '$program_name'] 'system (cp ...)' FAILED to copy the master file for the custom substitution model '$master_sbst_file' to the temporary directory '$tmpdir_phyml' : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	unless (0 == system ("cp -p $master_ctrl_file $tmp_ctrl_file") ) {
	    print STDOUT "[ERROR '$program_name'] 'system (cp ...)' FAILED to copy the master control file for PhyML '$master_ctrl_file' to the temporary directory '$tmpdir_phyml' : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	
	$options .= " -m custom -f m < $ctrl_name";

    } else {
		# This may have to be revised for $seqtype eq 'NT' #

	    # REVISED on Apr 13 (Wed), 2011 #
	    # => Further MODIFIED on Sep 18 (Thu), 2014. #
	$options .= " -m $sbst_model_phyml";
	##$options .= " -m $sbst_model";
	($resfreq =~ m/\S/) and ($options .= " -f $resfreq");
    }

    print STDOUT "\nOptions for PhyML tree inference:\n";
    print STDOUT "   '$options'\n\n";

   ### exit 0; ### For DEBUGGING the script ###

		# Execute the PhyML program #

    print STDOUT "\n<<< Start executing the PhyML program to infer the phylogenetic tree >>>\n\n";

    unless ( chdir $tmpdir_phyml ) {
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to change the directory to '$tmpdir_phyml' to infer the phylogenetic tree via PhyML !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    unless (0 == system ("$main_bin_tree $options > $name_log_tree 2>&1") ) {
	print STDOUT "[ERROR '$program_name'] FAILED 'system ('$main_bin_tree ...')' FAILED to infer the phylogenetic tree out of the core alignment '$inaln_for_tree': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    unless ( chdir $CURR_WORK_DIR ) {
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to MOVE BACK to the original working directory '$CURR_WORK_DIR' after inferring the phylogenetic tree out of the core alignment: $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

		# Read in the inferred tree #

    print STDOUT "\n ... Now that PhyML finished, read in the resulting output files...\n\n";

    unless (( -f $out_tree ) and ( -r $out_tree )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable OUTPUT TREE FILE '$out_tree' of PhyML for the tree inferred from the core alignment !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

	# Hashes to store tree information (see 'MyTreeManipu_II.pm' for their formats) #
    my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;

    unless ( extract_sgltree_from_file ($out_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr, $COMMENT) ) {
	print STDOUT "[ERROR '$program_name'] 'extract_sgltree_from_file' FAILED to extract a tree from the file '$out_tree' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    @std_set_cand_trees = (['NA', 'NA', 'NA', 0, 0, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr]);

		# Read in statistics file

    unless (( -f $out_stat ) and ( -r $out_stat )) {
	print STDERR "[ERROR '$program_name'] CanNOT FIND a readable OUTPUT STATISTICS FILE '$out_stat' of PhyML for the ML tree inference from the core alignment !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

	# REVISED on Apr 8 (Fri), 2011 #
	# => BUG fixed on Aug 26 (Sun), 2012. #
    unless (  extract_phyml_statistics2 ($out_stat, %STATS_PHYML, $DBG_LEVEL_EXPHYML) ) {
    ##unless (  extract_phyml_statistics2 ($out_stat, %STATS_PHYML) ) {
    ##unless (  extract_phyml_statistics ($out_stat, %STATS_PHYML) ) {
	print STDOUT "[ERROR '$program_name'] 'extract_phyml_statistics2' FAILED on the PhyML statistics file '$out_stat' concerning the core alignment !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}



	## Output the input tree to an interim file ##

print STDOUT "\n<<<<< Now, create an interim file recording the input trees >>>>>\n\n";
##print STDOUT "\n<<<<< Now, create an interim file recording the input tree >>>>>\n\n";

my $ct_cand_trees = @std_set_cand_trees;

unless ($ct_cand_trees == 1) { # ADDED in creation of 'classify_msa_errors.1.pl'. #

	print STDOUT "[ERROR '$program_name'] ILLEGAL INPUT!! ... There are $ct_cand_trees candidate trees, whereas this program accepts only a SINGLE INPUT TREE !!\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


my $fh_otree0  = open_output_file ($interim_trees, 0);
unless (defined $fh_otree0) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'open_output_file' FAILED on the interim tree file '$interim_trees' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
print $fh_otree0 "<< Candidate trees that are considerably likely (#{candidates} = $ct_cand_trees) >>\n\n";
##print $fh_otree0 "<< Candidate trees that are considerably likely (#{candidates} = 1) >>\n\n";

for (my $t=0; $t<$ct_cand_trees; $t++) { # Preliminary for-loop (over candidate trees). # ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 6 (Wed), 2014). #

	# REVISED in creation of 'compare_ml_msa.bch.extJC.6t.PAML.pl' (on Mar 22 (Tue), 2011) #
	# => Further REVISED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
	my ($ln_ml_msa, $node2pa0, $node2ch0, $name2id0, $id2name0, $br_attr0) = @{$std_set_cand_trees[$t]}[1,5..9];
	##my ($node2pa0, $node2ch0, $name2id0, $id2name0, $br_attr0) = @{$std_set_cand_trees[0]}[5..9];
	####my ($node2pa0, $node2ch0, $name2id0, $id2name0, $br_attr0) = @{$info_dawg_tree};
	my $topnode_id0 = $name2id0->{$TOP_NODE};

	print $fh_otree0 "< The $t th candidate >\n\n";
	##print $fh_otree0 "< The 0 th candidate >\n\n";
	print $fh_otree0 "{ RAW Phylogenetic Tree (top node ID: $topnode_id0) }\n\n"; # MODIFIED in creation of 'examine_purges_via_mblks.2.pl'. #
	##print $fh_otree0 "{ Phylogenetic Tree (top node ID: $topnode_id0) }\n\n";

	unless ( output_tree ($topnode_id0, %{$node2ch0}, %{$id2name0}, %{$br_attr0}, *${$fh_otree0}, 0, 0) ) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'output_tree' FAILED to output the dawg guide-tree onto the interim tree file '$interim_trees' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

			# ADDED in creation of 'examine_purges_via_mblks.2.pl'. #

	my $top_children = $node2ch0->{$topnode_id0};

	if (@{$top_children} == 2) {
	    print STDOUT "\n The top node has TWO CHILDREN !! ... make it virtually multifurcated, if possible ...\n\n";
	    my @indices_attr = ($INDX_LENGTH, -1, -1);
	    unless ( bi2v_multifurcated_top (%{$node2pa0}, %{$node2ch0}, %{$name2id0}, %{$br_attr0}, @indices_attr) ) {
		print STDOUT "[ERROR '$program_name'] Subroutine 'bi2v_multifurcated' FAILED !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	}

	print $fh_otree0 "\n{ Root-Moved Phylogenetic Tree (top node ID: $topnode_id0) }\n\n"; # MODIFIED in creation of 'examine_purges_via_mblks.2.pl'. #

	unless ( output_tree ($topnode_id0, %{$node2ch0}, %{$id2name0}, %{$br_attr0}, *${$fh_otree0}, 0, 0) ) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'output_tree' FAILED to output the dawg guide-tree onto the interim tree file '$interim_trees' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print $fh_otree0 "\n";

			# END of "ADDED in creation of 'examine_purges_via_mblks.2.pl'." #

		# FIXED on Mar 24 (Thu), 2011 #
		# => REVISED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
	print $fh_otree0 "\n{ Log-likelihood of the tree for the MSA: $ln_ml_msa }\n\n\n";
	##print $fh_otree0 "\n{ Log-likelihood of the MSA under the tree: 0 }\n\n";

}	# Preliminary for-loop (over candidate trees). # ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 6 (Wed), 2014). #

close ($fh_otree0);


	## Calculate the Akaike weights of the candidate trees

print STDOUT "\n<<<<< Now, calculate the Akaike weights of the candidate trees and necessary inputs  >>>>>\n\n";

# $ln_akaike_wgts->[$c] is the logarithmic Akaike weight factor of the $c th candidate tree.
#
my $ln_akaike_wgts = log_akaike_weights (@std_set_cand_trees, 3);
unless (defined $ln_akaike_wgts) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'log_akaike_weights' FAILED on the tree list extracted from '$interim_trees' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


	## Display the candidate trees, and their Akaike weights.
	## Also, assign absolute depths to the nodes of each tree. 
	## (MODIFIED in creation of 'compare_ref_vs_rec_prank_scores.pl'.)

print STDOUT "\n<<<<< Now, Output candidate trees, their Akaike weights, and their attributes  >>>>>\n\n";

my @set_node2depth = (); # %{$set_node2depth[$t]} = ($node_id => $depth, ...) for all nodes in the $t th candidate tree.
my @set_node2ct_exoffs = (); # %{$set_node2ct_exoffs[$t]} = ($node_id => #{exteior offsprings}, ...) for all nodes in the $t th candidate tree.
my @set_node2seq_indx = (); # %{$set_node2seq_indx[$t]} = ($node_id => {index in @{$sequence_ids}}, ...) for all exterior nodes for the $t th candidate tree. # (ADDED on Oct 28 (Tue), 2014.) #

my @set_node2trstats2score = (); # %{$set_node2trstats2score[$t]} = ($node_id => {"$state_bf:$state_af" => $score, ...) for all interior nodes in the $t th candidate tree.


for (my $t = 0; $t<$ct_cand_trees; $t++) { # Preliminary for-loop (over candidate trees). #

	my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[$t]};
	my $topnode_id = $name2id->{$TOP_NODE};
	my $ln_awgt = $ln_akaike_wgts->[$t];

	print STDOUT "\< The $t th candidate tree (topnode ID: $topnode_id\; Log_Akaike_weight = $ln_awgt) >\n\n";
	print STDOUT "Sum (-log(Pv)) = $sum_neg_log_pv\n";
	print STDOUT "log (Prob(MSA|Tree)_ML) = $ln_ml_msa\n";
	print STDOUT "#{free parameters} = $ct_free_params\n";
	print STDOUT "AIC = $aic\n";
	print STDOUT "BIC = $bic\n\n";

	my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	unless ( output_tree ($topnode_id, %{$node2ch}, %{$id2name}, %{$br_attr}, *STDOUT, $fl_bootlabel, $fl_nodeid) ) {
	    print STDOUT "[ERROR '$program_name'] 'output_tree' FAILED to output the $t th candidate tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n";

	##print STDERR "\n\nNode_ID --> Parent_ID\n\n";
	##foreach my $nodeid (sort keys %{$node2pa}) { print STDERR "$nodeid  --> ", $node2pa->{$nodeid}, "\n"; }
	##print STDERR "\n\n";

	print STDOUT join ("\t", 'Branch_ID', 'length', 'P-value', 'parent_node_ID'), "\n\n";
	foreach my $brid (sort {$a <=> $b} keys %{$br_attr}) { # 1st inner foreach-loop (over branches). #
	    my $attr = $br_attr->{$brid};
	    my ($brlen, $pvalue) = @{$attr}[$INDX_LENGTH, $INDX_BOOT];
	    (defined $brlen) or ($brlen = 'NA');
	    (defined $pvalue) or ($pvalue = 'NA');
	    my $pa = $node2pa->{$brid};
	    print STDOUT join ("\t", $brid, $brlen, $pvalue, $pa), "\n";
	}				# End of the 1st inner foreach-loop (over branches). #
	print STDOUT "\n\n";


		## ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl'. ##

		# Check if all the interior nodes (including the root) have 2 children each. #

	foreach my $node_id (sort {$a <=> $b} keys %{$node2ch}) { # 2nd inner foreach-loop (over interior nodes). #
	     my $children = $node2ch->{$node_id};
	     unless ((defined $children) and (@{$children}>0)) {
		print STDOUT "[WARNING '$program_name'] The node '$node_id' has an EMPTY (or UNDEFINED) children set: $! ! ... I will delete its entry in ", '%{$node2ch}', " ...\n\n";

		delete $node2ch->{$node_id};
		next;
	     }

	    unless (@{$children} == 2) {
		my $ct_ch = @{$children};
		print STDOUT "[ERROR '$program_name'] The node '$node_id' has $ct_ch children, instead of 2 children as required by this script: $! !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }
	}			# End of the 2nd inner foreach-loop (over interior nodes). #

	print STDOUT "\n ... Confirmed that all interior nodes of this tree have 2 child nodes each ...\n\n";


		# Assign depths to the nodes. #

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDOUT "[ERROR '$program_name'] 'make_node2abs_depth' FAILED on the $t th candidate tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n ... The subroutine 'make_node2abs_depth' Ended ...\n\n";

		# (MOVED to below with modification (on Oct 19 (Sun), 2014).) #
	##print STDOUT join ("\t", 'Node_ID', 'Depth'), "\n\n";
	##foreach my $node (sort {$a <=> $b} keys %{$node2depth}) { print STDOUT join ("\t", $node, $node2depth->{$node}), "\n"; }
	##print STDOUT "\n\n\n";

	$set_node2depth[$t] = $node2depth;


		# Assign #{exterior offsprings} to the nodes. #
		# (ADDED on Oct 19 (Sun), 2014.)

	my $node2ct_exoffs = make_node2ct_exoffs (%{$node2ch}, %{$node2depth});
	unless (defined $node2ct_exoffs) {
	    print STDOUT "[ERROR '$program_name'] 'make_node2ct_exoffs' FAILED on the $t th candidate tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n ... The subroutine 'make_node2ct_exoffs' Ended ...\n\n";

	$set_node2ct_exoffs[$t] = $node2ct_exoffs;


		# (MOVED from above with modification (on Oct 19 (Sun), 2014).) #

	print STDOUT join ("\t", 'Node_ID', 'Depth', '#{exteior_offsprings}'), "\n\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2depth}) { print STDOUT join ("\t", $node, $node2depth->{$node}, $node2ct_exoffs->{$node}), "\n"; }
	print STDOUT "\n\n\n";


		# Assign indices (for @{$sequence_ids}) to the exterior nodes. #
		# (ADDED on Oct 28 (Tue), 2014.) #

	print STDOUT join ("\t", 'Seq_indx', 'Seqname', 'Node_ID') , "\n\n";

	my %node2seq_indx;

	for (my $k=0; $k<$ct_sequences; $k++) { # 5th inner for-loop (over aligned sequences). #

	    my $seqname = $sequence_ids->[$k];
	    my $nodeid = $name2id->{$seqname};
	    unless (defined $nodeid) {
		print STDOUT "[ERROR '$program_name'] The $k th sequence (name: $seqname) has NO CORRESPONDING NODE in the $t th candidate tree !\n\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }

	    $node2seq_indx{$nodeid} = $k;

	   print STDOUT join ("\t", $k, $seqname, $nodeid), "\n";

	}			# End of the 5th inner for-loop (over aligned sequences). #
	print STDOUT "\n\n\n";

	$set_node2seq_indx[$t] = \%node2seq_indx;

		## END of "ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl'." ##

}			# end of the preliminary for-loop (over candidate trees). #

print STDOUT "\n";

### exit 0; ### For TESTING the Upper part of the script


###=For the predecessor 'compare_ref_vs_rec_prank_scores.stripped.pl' (V).
	# => RESURRECTED in creation of 'examine_purges_via_mblks.2.pl' (on May 9 (Mon), 2016). #

	### (MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_prank_scores.pl'.) ###

	## Assign a Residue Transition Probability Matrix to each branch. ##

			# MODIFIED in creation of 'examine_purges_via_mblks.2.pl' (on May 10 (Tue), 2016). #
print STDOUT "\n<<<<< Now, assign a Residue Transition Probability Matrix to each branch, and residue frequencies to each node >>>>>\n\n";
##print STDOUT "\n<<<<< Now, assign a Residue Transition Probability Matrix to each branch >>>>>\n\n";

##our @RESIDUES_PRANK = ($seqtype eq 'NT') ? @MyPRANK_Scores::NUCLEOTIDES : # MODIFIED in creation of 'examine_purges_via_mblks.2.pl'. #
our @RESIDUES_PRANK = ($seqtype eq 'NT') ? @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES : 
	@MyPAML_parser::AMINO_ACIDS ; # @MyPRANK_Scores::NUCLEOTIDES should be qw( A G C T ), as of Sep 12 (Fri), 2014. #
my @labels_res_weights = ();
foreach my $res (@RESIDUES_PRANK) { push @labels_res_weights, "wgt($res)"; }

my $ct_res_prank = @RESIDUES_PRANK;

my @resfreqs_prank = ();
for (my $c=0; $c<$ct_res_prank; $c++) { $resfreqs_prank[$c] = $RES2FREQ{$RESIDUES_PRANK[$c]}; }

print STDOUT "[ Residue frequencies for the PRANK analysis ]\n\n";

print STDOUT "Freq [", join (', ', @RESIDUES_PRANK), "] = (", join (', ', @resfreqs_prank), ") .\n\n";

my $sv_ratio_prank = (!(defined $sv_ratio) and (($sbst_model eq 'JC') or ($sbst_model eq 'F81'))) ? 1 : $sv_ratio ;
unless (defined $sv_ratio_prank) {
	print STDOUT "[ERROR '$program_name'] The transition/transversion ratio for the PRANK analysis is UNDEFINED !\n";
	print STDERR "[ERROR '$program_name'] $! !\n\n";
	die ;
}

my $sv_ratio_r_prank = ((defined $sv_ratio_r) and (defined $sv_ratio_y)) ? $sv_ratio_r : $sv_ratio_prank ;
my $sv_ratio_y_prank = ((defined $sv_ratio_r) and (defined $sv_ratio_y)) ? $sv_ratio_y : $sv_ratio_prank ;

print STDOUT "[ Transition/Transversion ratios for the PRANK analysis ]\n\n";

print STDOUT "       (for A <-> G)  $sv_ratio_r_prank ,\n";
print STDOUT "       (for T <-> C)  $sv_ratio_y_prank ,\n";
print STDOUT "     (average (used for PhyML)) $sv_ratio_prank .\n\n\n";


my @set_br2tr_mtrx_tn = (); # %{$set_br2tr_mtrx_tn[$t]} = ($br_id => \@tr_mtrx_tn, ...) for all branches, where @tr_mtrx_tn is the transition probability matrix calculated with the branch length of $br_id under the Tamura-Nei model.

my @set_node2ntfreqs = (); # %{$set_node2ntfreqs[$t]} = ($node_id => \@ntfreqs, ...) for all nodes, where $ntfreqs[$i] is the frequency of $RESIDUES_PRANK[$i] at the node, $node_id. # ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 10 (Tue), 2016). #

for (my $t=0; $t<$ct_cand_trees; $t++) { # Preliminary for-loop (over candidate trees).

	print STDOUT "< For the $t th candidate tree >\n\n";

	my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[$t]};

	my $br2tr_mtrx_tn = calc_br2tr_mtrx_TN (%{$br_attr}, @resfreqs_prank, $sv_ratio_r_prank, $sv_ratio_y_prank, 1);
	unless (defined $br2tr_mtrx_tn) {
	    print STDOUT "[ERROR '$program_name'] 'calc_br2tr_mtrx_TN' FAILED !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

   if ($TEST_LEVEL3>2) { # MODIFIED in creation of 'examine_purges_via_mblks.2.pl'. #
   ##if ($TEST_LEVEL>2) {...}
	print STDOUT "\n[ Transition matrices assigned to the branches ]\n\n";

	foreach my $br (sort {$a <=> $b} keys %{$br2tr_mtrx_tn}) { # Inner foreach-loop (over branches). #

	    print STDOUT "{ Branch ID: $br }\n\n";

	    my $tr_mtrx_tn = $br2tr_mtrx_tn->{$br};

	    print STDOUT join ("\t", 'From\To', @RESIDUES_PRANK), "\n\n";
	    for (my $c=0; $c<$ct_res_prank; $c++) {  print STDOUT join ("\t", $RESIDUES_PRANK[$c], @{$tr_mtrx_tn->[$c]}), "\n"; }
	    print STDOUT "\n\n";

	}		# End of the inner foreach-loop (over branches). #

   ##}	# END of "if ($TEST_LEVEL>2) {...}"
   }	# END of "if ($TEST_LEVEL3>2) {...}"

	##print STDOUT "\n";

	$set_br2tr_mtrx_tn[$t] = $br2tr_mtrx_tn;


		# ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 10 (Tue), 2016). #

	my $topnode_id = $name2id->{$TOP_NODE};
	my $node2ntfreqs = calc_node2ntfreqs_td (@resfreqs_prank, %{$br2tr_mtrx_tn}, %{$node2ch}, $topnode_id);
	unless (defined $node2ntfreqs) {
	    print STDOUT "[ERROR '$program_name'] 'calc_node2ntfreqs_td' FAILED !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

   if ($TEST_LEVEL3>2) {

	print STDOUT "\n[ Nucleotide frequencies assigned to the nodes ]\n\n";
	print STDOUT join ("\t", 'Node_ID', @RESIDUES_PRANK), "\n\n";

	foreach my $node_id (sort {$a <=> $b} keys %{$node2ntfreqs}) { # 2nd inner foreach-loop (over branches). #
	    print STDOUT join ("\t", $node_id, @{$node2ntfreqs->{$node_id}}), "\n";
	}					# End of the 2nd inner foreach-loop (over branches). #
	print STDOUT "\n";

   }	# END of "if ($TEST_LEVEL3>2) {...}"

	$set_node2ntfreqs[$t] = $node2ntfreqs;

	print STDOUT "\n";

		# END of "ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 10 (Tue), 2016)." #

}	# End of preliminary for-loop (over candidate trees). #

	### END of "(MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_prank_scores.pl'.)" ###

###=cut # END of "For the predecessor 'compare_ref_vs_rec_prank_scores.stripped.pl' (V).
	# => RESURRECTED in creation of 'examine_purges_via_mblks.2.pl' (on May 9 (Mon), 2016)." #


	### MAIN ANALYSES:  The Main Analyses: Examine the \"Purge\"s (isolated from other errors) by comparing Reference & Reconstructed segmental MSAs. ###
	### (MODIFIED in creation of 'examine_purges_via_mblks.2.pl'.)

###print STDOUT "\n\n<<<<<<< The Main Analyses: Examine the \"Purge\"s (isolated from other errors) by comparing Reference & Reconstructed segmental MSAs. >>>>>>>\n\n";
print STDOUT "\n\n<<<<<<< The Main Analyses: Examine the \"correct\" and \"erroneous\" segments of the reconstructed MSA. >>>>>>>\n\n";


		## RESURRECTED in creation of 'bch_pbs_examine_purges.D.pl' (with $fh_out modified to be $fh_out1). ##
my $fh_out1 = open_output_file ($out_summ, 0);
unless (defined $fh_out1) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$out_summ' !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_anom_segs = my $ct_in_segs = my $ct_succ_segs = my $ct_fail_segs = my $ct_skipped_segs = 0;
my $ct_skipped_segs2 = 0; # ADDED in DEBUGGING on Aug 11 (Mon), 2014. #
my $ct_skipped_segs3 = 0; # IMPORTED from 'examine_purge_ctrls.2.pl' in creation of 'examine_complex_via_mblks.2.pl'. (on May 30 (Mon), 2016). #

my $ct_succ_sbj = 0; # ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #

#my $list_segments = readin_list_segs ($inlist_segs, %LABEL_SATTR2INDX);
# my $ct_segments = @{$list_segments};

	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
my $last_ref_coord_end = my $last_rec_coord_end = 0;
##my $ref_coord_beg = my $rec_coord_beg = 0; # The 'beginning' coordinates of the segment in the reference MSA and reconstructed MSA, respectively. #

my $last_segid = 'LFLANK'; # ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #

$GAP = uc $GAP; # Make the gap character upper-case. (ADDED in creation of 'compare_ref_vs_rec_mafft_consist.pl'.) #

OUTER: # ADDED on Oct 14 (Wed), 2015. #
for (my $s = 0; $s<$ct_segments; $s++) { # Main outer for-loop (over MSA segments) #

		## MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). ##

			# (NOTE: Here, the coordinates follow the 0-based, half-open convention.) #
	my ($segid, $ref_coord_beg, $ref_coord_end, $rec_coord_beg, $rec_coord_end, $diagnosis) = @{$list_segments->[$s]};

	    # @LABELS_SEG_ATTR = ('Indx_spacer', 'Start_in_Ref', 'End_in_ref', 'Start_in_Rec', 'End_in_Rec', 'Diagnosis'); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
	    # @{$list_segments[$s]} = @attrs gives information on the $s th segment, including its ID.

	if (( uc ($diagnosis) =~ m/^EQUIV/ ) or ( uc ($diagnosis) =~ m/^IDENTICAL/) ) { # Skip segments NOT diagnosed as having reconstructed MSA 'DIFFERENT'  from the reference. # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
	##unless ( uc ($diagnosis) =~ m/DIFFERENT$/ ) {...} # Skip segments NOT diagnosed as having reconstructed MSA 'DIFFERENT'  from the reference.

   (print STDOUT "\n[INFORMATION] The $s th segment (ID: $segid) will be SKIPPED, because its reference MSA (in [$ref_coord_beg\, $ref_coord_end\) ) and reconstructed MSA (in [$rec_coord_beg\, $rec_coord_end\) ) are diagnosed as '$diagnosis' !\n\n") if ($TEST_LEVEL>0);

	    $ct_skipped_segs++;

	    next;
	}

	print STDOUT "<<<<<< Segment ID: $segid  (Diagnosis: $diagnosis) >>>>>>\n\n";

=DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
		# ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 9 (Mon), 2016). #

    unless (defined $segid2if_sbjct{$segid}) {
	print STDOUT "... This segment is irrelevant, thus its analysis is skipped ...\n\n";
	next;
    }

		# END of "ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 9 (Mon), 2016)." #
=cut # END of "DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'." #

=For TARGETTED DEBUGGING. ... For the predecessor 'classify_msa_errors_via_mblks.2.pl'. #
 
    ##if (($TEST_LEVEL_newtype>0) and ($segid ne '13')) {...}
    if ($segid ne '4') {
	print STDOUT "... This segment is irrelevant, thus its analysis is skipped ...\n\n";
	next;
    }
=cut

	###my $ct_lf_gapless_clms = my $ct_rf_gapless_clms = 'N/A'; ### ADDED in creation of 'examine_complex_via_mblks.2.pl'. ### DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #


	my @subject_segments = (); # element = [\@seg_msa, $baseline_coord, $label, $type, $diagnosis], where $label = $segid if $type = 'Erroneous', $label = "$last_segid\;$segid" if $type = 'Correct'. $diagnosis = 'Correct' if $type = 'Correct'. ## ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. ##

		# Remove the right flanking regions (of the last segment), if at all. #

	if ($last_ref_coord_end < $ref_coord_beg) {
	    my $len_rflnk = $ref_coord_beg - $last_ref_coord_end;

			# MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
	    for (my $k=0; $k < $ct_sequences; $k++) {
		my $tmp = uc substr (${$ref_msa->[$k]}, 0, $len_rflnk, '');
	    }
	    ##for (my $k=0; $k < $ct_sequences; $k++) { substr (${$ref_msa->[$k]}, 0, $len_rflnk, ''); }
	}

	if ($last_rec_coord_end < $rec_coord_beg) {
	    my $len_rflnk_rec = $rec_coord_beg - $last_rec_coord_end;
	    ##my $len_rflnk = $rec_coord_beg - $last_rec_coord_end;

	    my @tmp_lf = (); # ADDED in creation of 'examine_complex_via_mblks.2.pl'. #

			# MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
	    for (my $k=0; $k < $ct_sequences; $k++) {
		my $tmp = uc substr (${$rec_msa->[$k]}, 0, $len_rflnk_rec, '');
		##my $tmp = uc substr (${$rec_msa->[$k]}, 0, $len_rflnk, '');

		push @tmp_lf, \$tmp; # ADDED in creation of 'examine_complex_via_mblks.2.pl'. #
	    }
	    ##for (my $k=0; $k < $ct_sequences; $k++) { substr (${$rec_msa->[$k]}, 0, $len_rflnk, ''); }

		## ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. ##
	    my $if_gapped2 = 0;
	    for (my $k=0; $k < $ct_sequences; $k++) {
		my $cp = ${$tmp_lf[$k]};
		for (my $c=0; $c<$len_rflnk_rec; $c++) {
		##for (my $c=0; $c<$len_rflnk; $c++) {
		    my $res = substr ($cp, 0, 1, '');
		    if ($res eq $GAP) {
			$if_gapped2 = 1;
			last;
		    } 
		}
		($if_gapped2) and last;
	    }

	    if ($if_gapped2) {
		push @subject_segments, [\@tmp_lf, $last_rec_coord_end, "$last_segid\;$segid", 'Correct', 'Correct'];
	    }

		## END of "ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'." ##

=DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. (5)
		# ADDED in creation of 'examine_complex_via_mblks.2.pl'. #
	    my $ct_lf = 0;

	    for (my $c=0; $c<$len_rflnk; $c++) {
		my $if_gapless = 1;
		for (my $k=0; $k<$ct_sequences; $k++) {
		    my $res = substr (${$tmp_lf[$k]}, -1, 1, '');
		    if ($res eq uc($GAP)) {
			$if_gapless = 0;
			last;
		    }
		}
		$if_gapless and ($ct_lf++);
	    }

	    $ct_lf_gapless_clms = $ct_lf;
		# END of "ADDED in creation of 'examine_complex_via_mblks.2.pl'." #
=cut # END of "DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. (5)"

	} # END of "if ($last_rec_coord_end < $rec_coord_beg) {...}"

	print STDOUT "<<<< Prepare for segments of reference/reconstructed MSAs >>>>\n\n";

	my $ref_len = $ref_coord_end - $ref_coord_beg;
	my $rec_len = $rec_coord_end - $rec_coord_beg;

	my @ref_seg_msa = my @rec_seg_msa = ();

	for (my $k=0; $k<$ct_sequences; $k++) { #  1st middle for-loop #

			# CORRECTED in creation of 'examine_complex_via_mblks.2.pl' (on May 31 (Tue), 2016). #
	    my $seg_ref = uc substr (${$ref_msa->[$k]}, 0, $ref_len, '');
	    my $seg_rec = uc substr (${$rec_msa->[$k]}, 0, $rec_len, '');
	    ##my $seg_ref = substr (${$ref_msa->[$k]}, 0, $ref_len, '');
	    ##my $seg_rec = substr (${$rec_msa->[$k]}, 0, $rec_len, '');

	    push @ref_seg_msa, \$seg_ref;
	    push @rec_seg_msa, \$seg_rec;

	}	# End of the 1st middle for-loop #

	print STDOUT "(ref_len, rec_len) = ($ref_len, $rec_len)\n\n";

		## ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. ##

	my $if_gapped1 = 0;
	for (my $k=0; $k<$ct_sequences; $k++) {
	    my $cp = ${$rec_seg_msa[$k]};
	    for (my $c=0; $c<$rec_len; $c++) {
		my $res = substr ($cp, 0, 1, '');
		if ($res eq $GAP) {
		    $if_gapped1 = 1;
		    last;
		}
	    }
	    ($if_gapped1) and last;
	}

	if ($if_gapped1) {
	    push @subject_segments, [\@rec_seg_msa, $rec_coord_beg, $segid, 'Erroneous', $diagnosis];
	}

		## END of "ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'." ##

		## MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. ##

	if ( ($s == $ct_segments -1) and ($rec_coord_end < $totlen_recaln) ) {

	    my $len_rf = $totlen_recaln - $rec_coord_end;
            my @tmp_rf = ();
	    for (my $k=0; $k < $ct_sequences; $k++) {
		my $tmp = uc substr (${$rec_msa->[$k]}, 0, $len_rf);
		push @tmp_rf, \$tmp;
	    }

	    my $if_gapped3 = 0;
	    for (my $k=0; $k<$ct_sequences; $k++) {
		my $cp = ${$tmp_rf[$k]};
		for (my $c=0; $c<$len_rf; $c++) {
		    my $res = substr ($cp, 0, 1, '');
		    if ($res eq $GAP) {
			$if_gapped3 = 1;
			last;
		    }
		}
		($if_gapped3) and last;
	    }

	    if ($if_gapped3) {
		push @subject_segments, [\@tmp_rf, $rec_coord_end, "$segid:RFLANK", 'Correct', 'Correct'];
	    }

	} # END of "if ( ($s = $ct_segments) and ($rec_coord_end < $totlen_recaln) ) {...}"

=MOVED below for DEBUGGING on May 31 (Tue), 2016. #
	unless (@subject_segments>0) {

   (print STDOUT "\n[INFORMATION] The $s th segment (ID: $segid) will be SKIPPED, because it has NO SUBJECT SEGMENT to be examined ...\n\n\n") if ($TEST_LEVEL>0);

	    $ct_skipped_segs2++;

	    next;
	}
=cut

=DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'.

		# Count the number of right-franking gapless columns. #
		# (ADDED in creation of 'examine_complex_via_mblks.2.pl'.)

	my $rec_coord_beg_next = ($s+1 < $ct_segments) ? $list_segments->[$s+1][3] : $totlen_recaln ;
		# ($segid, $ref_coord_beg, $ref_coord_end, $rec_coord_beg, $rec_coord_end, $diagnosis) = @{$list_segments->[$s]};

	if ($rec_coord_end < $rec_coord_beg_next) {

	    my $len_rf = $rec_coord_beg_next - $rec_coord_end;

	    my @tmp_rf = ();
	    for (my $k=0; $k < $ct_sequences; $k++) {
		my $tmp = uc substr (${$rec_msa->[$k]}, 0, $len_rf);
		push @tmp_rf, \$tmp;
	    }

	    my $ct_rf = 0;
	    for (my $c=0; $c<$len_rf; $c++) {
		my $if_gapless = 1;
		for (my $k=0; $k<$ct_sequences; $k++) {
		    my $res = substr (${$tmp_rf[$k]}, 0, 1, '');
		    if ($res eq uc($GAP) ) {
			$if_gapless = 0;
			last;
		    }
		}
		$if_gapless and ($ct_rf++);
	    }

	    $ct_rf_gapless_clms = $ct_rf;
	}
		# END of "Count the number of right-franking gapless columns." #

=cut # END of "DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'."



		# Update the last end-coordinates. #

	$last_ref_coord_end = $ref_coord_end;
	$last_rec_coord_end = $rec_coord_end;
	$last_segid = $segid; # ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #

		## END of "MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014)." ##

		### MOVED from above for DEBUGGING on May 31 (Tue), 2016. ###
		### (The 'last' coordinates and segment ID MUST be updated before skipping the analysis.)
	unless (@subject_segments>0) {

   (print STDOUT "\n[INFORMATION] The $s th segment (ID: $segid) will be SKIPPED, because it has NO SUBJECT SEGMENT to be examined ...\n\n\n") if ($TEST_LEVEL>0);

	    $ct_skipped_segs2++;

	    next;
	}

		### END of "MOVED from above for DEBUGGING on May 31 (Tue), 2016." ###

=for targetted DEBUGGING (on Sep 22 (Sat) & 23 (Sun), 2012)
	##unless ($segid eq '0040') {...}
	unless ($segid eq '0014') {
	    $ref_coord_beg += $ref_len;
	    $rec_coord_beg += $rec_len;
	    $ct_skipped_segs++;
	    ##print STDOUT "\n --- Sgment with ID ne '0040' is skipped ---\n\n";
	    print STDOUT "\n --- Sgment with ID ne '0014' is skipped ---\n\n";
	    next;
	}
=cut

=DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #

			### Copy the segmental MSAs for the later use (examining the subject errors). ###
			### (ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 10 (Tue), 2016).) ###

	my @cp_ref_seg_msa = my @cp_rec_seg_msa = ();
	foreach my $seq (@ref_seg_msa) { my $cp = ${$seq}; push @cp_ref_seg_msa, \$cp; }
	foreach my $seq (@rec_seg_msa) { my $cp = ${$seq}; push @cp_rec_seg_msa, \$cp; }

			### END of "Copy the segmental MSAs for the later use (examining the subject errors)." ###

=cut # END of "DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'." #

		## Prepare for the files necessary for analyzing the segment. ##

		# MODIFIED in creation of 'classify_msa_errors.1.pl'. #

	my $logfile_seg_anal = "$logdir/log_seg_anal.$segid.txt";
	my $fh_log = open_output_file ($logfile_seg_anal, 0);
	unless (defined $fh_log) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the log file '$logfile_seg_anal' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}


		## FURTHER ADDED in creation of 'classify_msa_errors.1.pl'. #

		# Retrieve the tree information. #

	# ($sum_neg_log_pv through $bic are unnecessary for this program.) #
	#
	my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[0]};
	##my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[$0]}; # A BUG fixed on Oct 15 (Wed), 2014. #

	my $topnode_id = $name2id->{$TOP_NODE};
	my $node2depth = $set_node2depth[0];
	my $node2ct_exoffs = $set_node2ct_exoffs[0]; # ADDED on Oct 19 (Sun), 2014. #
	my $node2seq_indx = $set_node2seq_indx[0]; # ADDED on Oct 28 (Tue), 2014. # 


	print $fh_log "<<<<< Input Tree (Top_node = $topnode_id) >>>>>\n\n";


   if ($TEST_LEVEL>1) {

	my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	unless ( output_tree ($topnode_id, %{$node2ch}, %{$id2name}, %{$br_attr}, *${$fh_log}, $fl_bootlabel, $fl_nodeid) ) {
	    print STDOUT "[ERROR '$program_name'] 'output_tree' FAILED to output the input tree (for analysis of position shifts): $! !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print $fh_log "\n\n";
   }

   if ($TEST_LEVEL>3) {
	print $fh_log "< Depths assigned to the Nodes >\n\n";
	print $fh_log join ("\t", 'Node_ID', 'Depth', 'Name'), "\n\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2depth}) { print $fh_log join ("\t", $node, $node2depth->{$node}), "\n"; }
	print $fh_log "\n\n";
   }


		## ADDED in creation of 'classify_msa_errors.1.pl'. ##

	print $fh_log "<<<<< Input MSA (erroneous segment) >>>>>\n\n";
	##print $fh_log "<<<<< Input MSA >>>>>\n\n";

	print $fh_log "#{Sequences} = $ct_sequences .\n";
	print $fh_log "#{Sites in the segment}_ref = $ref_len ,\n";
	print $fh_log "#{Sites in the segment}_rec = $rec_len .\n\n\n";


=DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. (7)

		# ADDED in creation of 'examine_complex_via_mblks.2.pl'. #
    if ($TEST_LEVEL>1) {

	print $fh_log "< Sizes of gapless segments flanking the erroneous segment >\n\n";
	print $fh_log "Size(left-flanking)  = $ct_lf_gapless_clms ,\n";
	print $fh_log "Size(right-flanking) = $ct_rf_gapless_clms .\n\n\n";


    } # END of "if ($TEST_LEVEL>1) {...}"

		# END of "ADDED in creation of 'examine_complex_via_mblks.2.pl'." #

=cut # END of "DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. (7)"


		### MOVED from below, within the loop over gap patterns (on Sep 1 (Sat), 2012) ###

    if ($TEST_LEVEL3>5) { ## MODIFIED in creation of 'examine_complex_via_mblks.2.pl'. ##
    ##if ($TEST_LEVEL3>1) { ## MODIFIED in creation of 'examine_purges_via_mblks.2.pl'. ##
    ##if ($TEST_LEVEL>2) {
	##foreach my $fh_log ($fh_log_ref) { # MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
	####foreach my $fh_log ($fh_log_sref, $fh_log_gref) {...}
	    print $fh_log "<< Original Segment of the Reference Alignment: >>\n\n";
	    unless ( pr_ODEN_input (@{$sequence_ids}, @ref_seg_msa, 60, *${$fh_log}) ) {
		##print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reference MSA for the $indx_gp th gap pattern !\n";
		print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reference MSA !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    print $fh_log "\n\n";
	##}

	##foreach my $fh_log ($fh_log_rec) { # MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
	####foreach my $fh_log ($fh_log_srec, $fh_log_grec) {...}
	    print $fh_log "<< Original Segment of the Reconstructed Alignment: >>\n\n";
	    unless ( pr_ODEN_input (@{$sequence_ids}, @rec_seg_msa, 60, *${$fh_log}) ) {
		##print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reconstructed MSA for the $indx_gp th gap pattern !\n";
		print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reconstructed MSA !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    print $fh_log "\n\n";
	##}
    ##}
    ##}  # END of "if ($TEST_LEVEL3>1) {...}" ## MODIFIED in creation of 'examine_purges_via_mblks.2.pl'. ##
    } # END of "if ($TEST_LEVEL3>5) { ...}" ## MODIFIED in creation of 'examine_complex_via_mblks.2.pl'. ##

		### END of "MOVED from below, within the loop over gap patterns" ###




=DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. (8)

		### MODIFIED in creation of 'examine_complex_via_mblks.2.pl'. ###

	unless ($rec_len>0) {

	    print STDOUT "\n[INFORMATION] The reconstructed sub-MSA is NULL! ... I will skip further examination of this segment ...\n\n";
	    print $fh_log "\n[INFORMATION] The reconstructed sub-MSA is NULL! ... I will skip further examination of this segment ...\n\n";

	    print $fh_out1 "< Segment ID: $segid (length = 0; #{gapped segments} = 0) >\n\n";


	    ##print $fh_out1 join ("\t", 'Rlv_branch', 'len(rlv_branch)', '#{involved_seqs}', 'block_size', '|shift|', '#{sbst_diffs}_exp', '#{sbst_diffs}_obs', 'P-value', 'log[Lk(sbst)]_[rec-ref]'), "\n\n"; ## TO BE REPLACED ##
	    ##print $fh_out1 join ("\t", 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), "\n"; ## TO BE REPLACED ##

	    print $fh_out1 "\n\\\\\n\n";

	    close ($fh_log); # MODIFIED in creation of 'classify_msa_errors.1.pl'. #

	    next;

	} # END of "unless ($rec_len>0) {...}"

=cut # END of "DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. (8)"


=For the predecessor 'examine_purge_ctrls.2.pl'.

		## ADDED for DEBUGGING the script (on Aug 11 (Mon), 2014). ##

	unless (($ref_len>0) and ($rec_len>0)) {

	    print STDOUT "\n[INFORMATION] Either reference or reconstructed sub-MSA (or both) is NULL! ... I will skip the calculation of the MAFFT consistent scores for this segment pair... \n\n";

	    ##foreach my $fh_log ($fh_log_ref, $fh_log_rec) { # MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
	    ####foreach my $fh_log ($fh_log_sref, $fh_log_srec, $fh_log_gref, $fh_log_grec) {...}
		print $fh_log "\n[INFORMATION] Either reference or reconstructed sub-MSA (or both) is NULL! ... I will skip the calculation of the MAFFT consistent scores for this segment pair (ID: $segid) ... \n\n\n";
	    ##}

		# MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 6 (Tue), 2015). #
	    print $fh_out1 "< Segment ID: $segid >\n\n";
	    ##print $fh_out2 "< Segment ID: $segid >\n\n"; # DISABLED in creation of 'examine_purges_via_mblks.2.pl'. #
	    ####print $fh_out "< Segment ID: $segid >\n\n";


			# MODIFIED on Nov 9 (Sun), 2014. #
			# => FUrther MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 6 (Tue), 2015). #
			# => Further MODIFIED in creation of 'examine_purges_via_mblks.2.pl' (on May 11 (Wed), 2016). #

	    print $fh_out1 join ("\t", 'Rlv_branch', 'len(rlv_branch)', '#{involved_seqs}', 'block_size', '|shift|', '#{sbst_diffs}_exp', '#{sbst_diffs}_obs', 'P-value', 'log[Lk(sbst)]_[rec-ref]'), "\n\n";

	    print $fh_out1 join ("\t", 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), "\n";

		# MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 6 (Tue), 2015). #
	    print $fh_out1 "\n\\\\\n\n";
	    ##print $fh_out2 "\n\\\\\n\n"; # DISABLED in creation of 'examine_purges_via_mblks.2.pl'. #
	    ####print $fh_out "\n\\\\\n\n";

	    $ct_skipped_segs2++;


	    close ($fh_log); # MODIFIED in creation of 'classify_msa_errors.1.pl'. #
	    ##foreach my $fh_log ($fh_log_ref, $fh_log_rec) { close ($fh_log); } # ADDED on Sep 15 (Mon), 2014. #

	    next;
	}
		## END of "ADDED for DEBUGGING the script (on Aug 11 (Mon), 2014)." ##

=cut # END of "For the predecessor 'examine_purge_ctrls.2.pl'."


			### ADDED in creation of 'examine_complex_via_mblks.2.pl' (on May 31 (Tue), 2016). ###

	## my @subject_segments = (); # element = [\@seg_msa, $baseline_coord, $label, $type, $diagnosis], where $label = $segid if $type = 'Erroneous', $label = "$last_segid\;$segid" if $type = 'Correct'. $diagnosis = 'Correct' if $type = 'Correct'. ## ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. ##

	print $fh_log "<<<<<<< MAIN PROCESS: Examining the Subject Segments >>>>>>>\n\n";
	print STDOUT "<<<<<<< MAIN PROCESS: Examining the Subject Segments >>>>>>>\n\n";


    foreach my $subject (@subject_segments) { # Main outer-middle foreach-loop (over subjects). # ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #

		# ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #

	my ($seg_msa_sbj, $baseline_coord, $label_sbj, $type_sbj, $diagnosis_sbj) = @{$subject};

	print $fh_log "<<<<< Subject Segment ID: $label_sbj  >>>>>\n\n";
	print STDOUT "<<<<< Subject Segment ID: $label_sbj  >>>>>\n\n";

	print $fh_log "{Type} = $type_sbj ,\n";
	print $fh_log "{Diagnosis} = $diagnosis_sbj ,\n";
	print $fh_log "{Baseline} = $baseline_coord .\n\n\n";

	my @cp_seg_msa = ();
	foreach my $seq (@{$seg_msa_sbj}) { my $cp = ${$seq}; push @cp_seg_msa, \$cp; }


    if ($TEST_LEVEL3>2) {

	print $fh_log "<< The subject segment in the Reconstructed Alignment: >>\n\n";
	unless ( pr_ODEN_input (@{$sequence_ids}, @{$seg_msa_sbj}, 60, *${$fh_log}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reconstructed MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print $fh_log "\n\n";

    } # END of "if ($TEST_LEVEL3>2) {...}"

		# END of "ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'." #

		### RESTART FROM HERE (1) ###

			### IMPORTED from 'examine_purge_ctrls.2.pl' in creation of 'examine_complex_via_mblks.2.pl' (on May 30 (Mon), 2016). ###

			# MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
	print STDOUT "<<<< Preliminary (1): Identify gap-pattern blocks, calculate their Dollo parsimony scenarios, and the initial parsimony candidate scenario of each gapped segment in the subject. >>>>\n\n";
	print $fh_log "<<<< Preliminary (1): Identify gap-pattern blocks, calculate their Dollo parsimony scenarios, and the initial parsimony candidate scenario of each gapped segment in the subject. >>>>\n\n";
	##print STDOUT "<<<<< Preliminary (1): Identify gap-pattern blocks, calculate their Dollo parsimony scenarios, and the initial parsimony candidate scenario of each gapped segment in the erroneous segment (reconstructed). >>>>>\n\n";
	##print $fh_log "<<<<< Preliminary (1): Identify gap-pattern blocks, calculate their Dollo parsimony scenarios, and the initial parsimony candidate scenario of each gapped segment in the erroneous segmental (reconstructed). >>>>>\n\n";

	# @{$info_init_psm_scn_sbj} stores information on the initial parsimonious scenario candidate for each cluster (gapped segment) in the subject MSA;
	# @{$info_gpatterns} stores information on the gap-patterns, as well as their Dollo parsimony scenarios, that are found in the subject MSA. 
	#
	# For the contents of these variables, see below.
	#

	my @info_gpatterns0 = ([], [], []);
		# @info_gpatterns0 = (\@set_gpatterns0, \@set_node2dparsstat0, \@set_br2change0) gives information on the gap-patterns found in the reconstructed MSAs, and the Dollo parsimonious scenarios yielding these gap-patterns, where

	my ($info_init_psm_scn_sbj, $info_gpatterns) = cal_init_psm_cands_for_sgl_msa (%{$node2ch}, %{$node2pa}, %{$node2depth}, %{$node2ct_exoffs}, %{$name2id}, @{$sequence_ids}, @{$seg_msa_sbj}, $CHAR_ABS, $CHAR_PRES, @info_gpatterns0);

	unless (defined $info_init_psm_scn_sbj) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'cal_init_psm_cands_for_sgl_msa' FAILED for the subject, '$label_sbj', in the $s th section !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

		### ADDED to fix a BUG (on May 14 (Sat), 2016). ###

	unless ((@{$info_init_psm_scn_sbj}>2) and (@{$info_gpatterns}>2)) {

	    print STDOUT "[CAUTION '$program_name'] The subroutine 'cal_init_psm_cands_for_sgl_msa' could NOT handle the subject '$label_sbj' in the $s th section properly (probably because the segment is too long)... the further analyses of this segment will be suspended ...\n\n";
	    print $fh_log "\n[CAUTION '$program_name'] The subroutine 'cal_init_psm_cands_for_sgl_msa' could NOT handle the subject '$label_sbj' in the $s th section properly (probably because the segment is too long)... the further ANALYSES of this segment will be SUSPENDED ...\n\n\n";

	    print $fh_out1 "< Segment ID: $label_sbj >\n\n";
	    ##print $fh_out1 "< Segment ID: $segid >\n\n";

				# MODIFIED in creation of 'examine_complex_via_mblks.2aug.pl'. #'. #'. #'. #
	    print $fh_out1 join ("\t", 'Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}', '[br/stat_ue/beg/end/len]_ev1', '[br/stat_ue/beg/end/len]_ev2', '...'), "\n\n";
	    ##print $fh_out1 join ("\t", 'Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}'), "\n\n";
	    ####print $fh_out1 join ("\t", 'Wd_size', '#{wds}', 'indx_wd', 'br_ID', 'br_len', '#{ext_offs}', '#{diffs}_exp', '#{diffs}_obs', 'P-value'), "\n\n";

	    print $fh_out1 "...ANALYSES_SUSPENDED (segment_may_be_too_long)...\n";

	    print $fh_out1 "\n\\\\\n\n";

	    $ct_skipped_segs3++;


	    ##close ($fh_log); # DISABLED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #

	    next;

	}
		### END of "ADDED to fix a BUG (on May 14 (Sat), 2016)." ###

	# @{$set_cls_gblocks_rec->[$cl][$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the reconstructed MSA, with
	#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
	#	$len_gb being the length (in terms of #{columns}) of the block,
	#	and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);
	#
	# $init_totcts_indels_for_cls_rec->[$cl] is the total number of insertion/deletion events in the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reconstructed MSA;
	#
	# @{$init_psm_scn_for_cls_rec->[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reconstructed MSA, where 
	#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
	#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
	#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
	#	  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node.
	#
	my ($set_cls_gblocks_sbj, $init_totcts_indels_for_cls_sbj, $init_psm_scn_for_cls_sbj) = @{$info_init_psm_scn_sbj};

	# ${$set_gpatterns->[$indx_gp]} is the string representing the $ndx_gp th gap-pattern in the input MSA, created by concatenating $CHAR_PRES/ABS in the sequences;
	# %{$set_node2dparsstat->[$indx_gp]} = ($node_id => [$state, $ct_pres_ext], ...) is for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern;
	# %{$set_br2change->[$indx_gp]} = ($branch_id => [$state_ue, $state_le], ...) for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern.
	#
	my ($set_gpatterns, $set_node2dparsstat, $set_br2change) = @{$info_gpatterns};

	my $ct_gps = @{$set_gpatterns};

   if ($TEST_LEVEL3>2) { # (I) # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
   ##if ($TEST_LEVEL>2) { # (I)

	print $fh_log "<< Output of 'cal_init_psm_cands_for_sgl_msa' >>\n\n";

	print $fh_log "[[ Set of gap-patterns (#{patterns} = $ct_gps) ]]\n\n"; 

	print $fh_log join ("\t", 'Indx_gp', '#{indels}', "'gap_pattern(horizontal)'"), "\n\n";

	for (my $gp = 0; $gp<$ct_gps; $gp++) { # 1st outer testing for-loop (over gap-patterns). #
	   my $gpattern = ${$set_gpatterns->[$gp]};
	   my $ct_indels = scalar keys %{$set_br2change->[$gp]} ;
	   print $fh_log join ("\t", $gp, $ct_indels, "'$gpattern'"), "\n";
	}			# End of the 1st outer testing for-loop (over gap-patterns). #
	print $fh_log "\n\n";

     if ($TEST_LEVEL3>3) { # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
     ##if ($TEST_LEVEL>3) {

	print $fh_log "[ Dollo parsimonious scenario for each gap-pattern ]\n\n";

	for (my $gp=0; $gp<$ct_gps; $gp++) { # 2nd outer testing for-loop (over gap-patterns). #

		my $gpattern = ${$set_gpatterns->[$gp]};
		my $br2change = $set_br2change->[$gp];
		my $ct_indels = scalar keys %{$br2change};

		print $fh_log "{ For the $gp th pattern (horizontal: '$gpattern' ; #{indels} = $ct_indels) }\n\n";

		print $fh_log join ("\t", 'Branch_ID', 'state_ue', 'state_le'), "\n\n";

		foreach my $br (sort {$a <=> $b} keys %{$br2change}) { print $fh_log join ("\t", $br, @{$br2change->{$br}}), "\n"; }
		print $fh_log "\n\n";

      if ($TEST_LEVEL3>=5) { # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
      ##if ($TEST_LEVEL>=5) {
		my $node2dparsstat = $set_node2dparsstat->[$gp];

		print $fh_log ' ' x 3,  "( Map of presence(=$CHAR_PRES)/absence(=$CHAR_ABS) states onto each node )\n\n";
		print $fh_log ' ' x 3, join ("\t", 'Node', 'State'), "\n\n";

		foreach my $nodeid (sort {$a <=> $b} keys %{$node2dparsstat}) { print $fh_log ' ' x 3, join ("\t", $nodeid, $node2dparsstat->{$nodeid}[0]), "\n"; }
		print $fh_log "\n\n";
      ##}
      } # END of "if ($TEST_LEVEL3>=5) {...} # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'." #

	}		# End of the 2nd outer testing for-loop (over gap-patterns). #

     ##} # END of "if ($TEST_LEVEL>3) {...}"
     } # END of "if ($TEST_LEVEL3>3) {...} # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'." #


	my ($info_init_psm_scn, $msa_type) = ($info_init_psm_scn_sbj, $type_sbj); # Further MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
	##my ($info_init_psm_scn, $msa_type) = ($info_init_psm_scn_rec, 'Reconstructed'); # MODIFIED in creation of 'examine_complex_via_mblks.2.pl'. #
	##my ($info_init_psm_scn, $msa_type) = ($info_init_psm_scn_ref, 'Reference');
	####my ($info_init_psm_scn, $msa_type) = @{$set_info};
	my ($set_cls_gblocks, $init_totcts_indels_for_cls, $init_psm_scn_for_cls) = @{$info_init_psm_scn};
	my $ct_cls = @{$set_cls_gblocks};

	print $fh_log "[[ Gapped segments and the initial candidates of their parsimonious scenarios for '$msa_type' MSA (#{gapped segments} = $ct_cls)  ]]\n\n";

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Middle testing for-loop (over gapped segments). #

		my $gblks_in_cls = $set_cls_gblocks->[$cl];
		my $ct_blks = @{$gblks_in_cls};
		my $beg_cls = $gblks_in_cls->[0][2];
		my $end_cls = $gblks_in_cls->[$ct_blks-1][3];

		print $fh_log "{ The $cl th gapped segment  (region = \[$beg_cls, $end_cls] ; #{blocks} = $ct_blks ) }\n\n";

		print $fh_log "[ Gap-pattern blocks ]\n\n";

		print $fh_log join ("\t", 'Indx_blk', 'beg_blk', 'end_blk', 'length', 'indx_gp', "'gap-pattern'"), "\n\n";

	# @{$set_cls_gblocks_ref/rec->[$cl][$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA, with
	#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
	#	$len_gb being the length (in terms of #{columns}) of the block,
	#	and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);

		for (my $b=0; $b<$ct_blks; $b++) { # 1st inner testing for-loop (over gap-pattern blocks). #
		    my ($indx_gp, $len_gb, $beg_gb, $end_gb) = @{$gblks_in_cls->[$b]};
		    my $gpattern = ${$set_gpatterns->[$indx_gp]};

		    print $fh_log join ("\t", $b, $beg_gb, $end_gb, $len_gb, $indx_gp, "'$gpattern'"), "\n";
		} 			# End of the 1st inner testing for-loop (over gap-pattern blocks). #
		print $fh_log "\n\n";

		my $totct_indels0 = $init_totcts_indels_for_cls->[$cl];

	# $init_totcts_indels_for_cls_ref/rec->[$cl] is the total number of insertion/deletion events in the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA;

		my ($br2indels, $node2blwise_gstat) = @{$init_psm_scn_for_cls->[$cl]};

		print $fh_log "[ Initial candidate indel scenario (#{indels} = $totct_indels0) ]\n\n";

		print $fh_log join ("\t", 'Branch_ID', '(state_ue/state_le/indices,of,blocks,involved)_event1', '[similarly for event2]', '...'), "\n\n";

		foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # 2nd inner testing for-loop (over branches with indels). #
		    my @out = ($br);
		    foreach my $event (@{$br2indels->{$br}}) { # Inner-inner for-loop (over events along the branch). #
			my ($state_ue, $state_le, $indices_blk) = @{$event};
			my $cnct_indices_blk = join (',', @{$indices_blk});
			push @out, join ('/', $state_ue, $state_le, $cnct_indices_blk);
		    }					# End of the inner-inner for-loop (over events along the branch). #

		    print $fh_log join ("\t", @out), "\n";
		}				# End of the 2nd inner testing for-loop (over branches with indels). #
		print $fh_log "\n\n";


     if ($TEST_LEVEL3>3) { # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #
     ##if ($TEST_LEVEL>3) {

		print $fh_log "( Map of block-wise gap states onto the nodes )\n\n";

		print $fh_log join ("\t", 'Node_ID', "'block_wise_gap_states'"), "\n\n";

		foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat}) { # 3rd inner testing for-loop (over nodes). #

		    my $cnct_bw_gstats = join ('', @{$node2blwise_gstat->{$node}});

		    print $fh_log join ("\t", $node, "'$cnct_bw_gstats'"), "\n";

		}				# End of the 3rd inner testing for-loop (over nodes). #
		print $fh_log "\n\n";

     ##} # END of "if ($TEST_LEVEL>3) {...}"
     } # END of "if ($TEST_LEVEL3>3) {...} # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'." #


	# @{$init_psm_scn_for_cls_ref/rec->[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA, where 
	#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
	#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
	#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
	#	  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node.

	}			# End of the middle testing for-loop (over gapped segments). #

	print $fh_log "\n\n";

   ##} # END of "if ($TEST_LEVEL>2) { # (I) ...}"
   } # END of "if ($TEST_LEVEL3>2) { # (I) ...} # MODIFIED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'." #


			### END of "IMPORTED from 'examine_purge_ctrls.2.pl' in creation of 'examine_complex_via_mblks.2.pl' (on May 30 (Mon), 2016)." ###


		### ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. ###

	print $fh_out1 "< Segment ID: $label_sbj >\n\n";

				# MODIFIED in creation of 'examine_complex_via_mblks.2aug.pl'. #
	 print $fh_out1 join ("\t", 'Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}', '[br/stat_ue/beg/end/len]_ev1', '[br/stat_ue/beg/end/len]_ev2', '...'), "\n\n";
	##print $fh_out1 join ("\t", 'Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}'), "\n\n";

	##my ($set_cls_gblocks_sbj, $init_totcts_indels_for_cls_sbj, $init_psm_scn_for_cls_sbj) = @{$info_init_psm_scn_sbj};

	my $ct_gpd_segs = @{$set_cls_gblocks_sbj};

	my $last_end  = 0; # In the 0-based, half-open convention. #

	for (my $cl=0; $cl<$ct_gpd_segs; $cl++) { # Middle for-loop (over gapped segments). #

	    my $gblks_in_cls = $set_cls_gblocks_sbj->[$cl];
	    my $ct_blks = @{$gblks_in_cls};

			# NOTE: The coordinates are in the 0-based, full-closed convention. #
	    my $beg_cls = $gblks_in_cls->[0][2];
	    my $end_cls = $gblks_in_cls->[$ct_blks-1][3];
	    my $ct_clms = $end_cls - $beg_cls + 1;

			# Tally the insertions/deletions. #
	    my $ct_ins = my $ct_del = my $totlen_ins = my $totlen_del = 0;

	    my ($br2indels, $node2blwise_gstat) = @{$init_psm_scn_for_cls_sbj->[$cl]};

	    my @out_events = (); # ADDED in creation of 'examine_complex_via_mblks.2aug.pl'. #

	    foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # 1st inner for-loop (over branches with indels). #
		    foreach my $event (@{$br2indels->{$br}}) { # Inner-inner for-loop (over events along the branch). #
			my ($state_ue, $state_le, $indices_blk) = @{$event};
			my $len = 0;
		    			# ($indx_gp, $len_gb, $beg_gb, $end_gb) = @{$gblks_in_cls->[$b]};
			foreach my $indx (@{$indices_blk}) { $len += $gblks_in_cls->[$indx][1]; }

			if ($state_ue eq $CHAR_PRES) {
				# The event is a deletion. #
			    $ct_del++;
			    $totlen_del += $len;
			} else {
				# The event is an insertion. #
			    $ct_ins++;
			    $totlen_ins += $len;
			}

				# ADDED in creation of 'examine_complex_via_mblks.2aug.pl'. #
			my ($indx_lb, $indx_rb) = @{$indices_blk}[0, $#{$indices_blk}];
			my $beg_ev = $gblks_in_cls->[$indx_lb][2] + $baseline_coord;
			my $end_ev = $gblks_in_cls->[$indx_rb][3] + $baseline_coord;
			#my $beg_ev = $gblks_in_cls->[$indx_lb][2];
			#my $end_ev = $gblks_in_cls->[$indx_rb][3];

			push @out_events, join ('/', $br, $state_ue, $beg_ev, $end_ev, $len);

		    }					# End of the inner-inner for-loop (over events along the branch). #
	    }				# End of the 1st inner for-loop (over branches with indels). #

				### RESTART FROM HERE ###

			# Count gaps in the gapped segment. #
	    my $ct_gaps = 0;

	    my $len_lf = $beg_cls - $last_end;
	    for (my $k=0; $k<$ct_sequences; $k++) { # 2nd inner for-loop (over sequences). #
		my $lf = substr (${$cp_seg_msa[$k]}, 0, $len_lf, '');
		my $cp_seq = substr (${$cp_seg_msa[$k]}, 0, $ct_clms, '');
		for (my $r=0; $r<$ct_clms; $r++) { # Inner-inner for-loop (over columns in the segment). #
		    my $res = substr ($cp_seq, 0, 1, '');
		    ($res eq $GAP) and ($ct_gaps++);
		}			# End of the inner-inner for-loop (over columns in the segment). #
	    }				# End of the 2nd inner for-loop (over sequences). #

	    my $beg_coord = $baseline_coord + $beg_cls;
	    my $end_coord = $baseline_coord + $end_cls;

				# MODIFIED in creation of 'examine_complex_via_mblks.2aug.pl'. #
	    print $fh_out1 join ("\t", $type_sbj, $diagnosis_sbj, $ct_gpd_segs, $cl, $beg_coord, $end_coord, $ct_clms, $ct_ins, $ct_del, $totlen_ins, $totlen_del, $ct_gaps, @out_events), "\n";
	    ##print $fh_out1 join ("\t", $type_sbj, $diagnosis_sbj, $ct_gpd_segs, $cl, $beg_coord, $end_coord, $ct_clms, $ct_ins, $ct_del, $totlen_ins, $totlen_del, $ct_gaps), "\n";

	    $last_end = $end_cls + 1; # In the 0-based, half-open convension. #

	}				# End of the middle for-loop (over gapped segments). #

	print $fh_out1 "\n\\\\\n\n";

	$ct_succ_sbj++;

		### END of "ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'." ###

    }			# END of the outer-middle foreach-loop (over subjects). # ADDED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #




=DISABLED in creation of 'examine_complex_via_mblks.2.pl' (3).

# print STDERR "Temporary directories for PAML:  {top_tmpdir}/s{seg_id}/g{indx_gpattern}/t{indx_tree}/,\n";

		# The grand-parent directory of temporary working directories #

	my $grpa_tmp_workdir_ref = sprintf ("%s/s%s", $top_tmpdir_ref, $segid);
	my $grpa_tmp_workdir_rec = sprintf ("%s/s%s", $top_tmpdir_rec, $segid);

	foreach my $grpa_tmpdir ($grpa_tmp_workdir_ref, $grpa_tmp_workdir_rec) { # Preliminary middle foreach-loop #
	    unless ( -d $grpa_tmpdir) {
		unless (0 == system ("mkdir -p $grpa_tmpdir" ) ) {
		    print STDOUT "[ERROR '$program_name'] FAILED to create a grand-parent of temporary working directories '$grpa_tmpdir': $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
	    }
	    unless (( -r $grpa_tmpdir ) and ( -w $grpa_tmpdir )) {
		unless (0 == system ("chmod -r u+rw $grpa_tmpdir" ) ) {
		    print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the grand-parent of temporary working directories '$grpa_tmpdir': $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
	    }
	}			# End of the preliminary middle foreach-loop #

			### NEWLY ADDED in creation of 'classify_msa_errors.1.pl'. ###


		## Preliminary Process (0): Map the residue numbers (1-based) onto the reference & reconstructed MSAs. ##
			## (ADDED on Oct 28 (Tue), 2014.) ##

	print STDOUT "<<<<< Preliminary (0): Map the residue numbers onto the reference & reconstructed MSAs... >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (0): Map the residue numbers onto the reference & reconstructed MSAs... >>>>>\n\n";

	# $set_clminfo_ref/rec->[$c][$k] is the residue number (1-based) of the $k th sequence at the $c th column of the reference/reconstructed MSA.
	#		it is ($r + 1/2) if the column is occupied by a gap between the $r th and the ($r+1) th residues.
	#
	my $set_clminfo_ref = assign_resnums_to_clms (@ref_seg_msa, $GAP, 0);
	unless (defined $set_clminfo_ref) {
	    print STDOUT "[ERROR '$program_name'] 'assign_resnums_to_clms' FAILED on the reference MSA for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $set_clminfo_rec = assign_resnums_to_clms (@rec_seg_msa, $GAP, 0);
	unless (defined $set_clminfo_rec) {
	    print STDOUT "[ERROR '$program_name'] 'assign_resnums_to_clms' FAILED on the reconstructed MSA for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

   if ($TEST_LEVEL>3) { # (O)

	my $ct_clms_ref = @{$set_clminfo_ref};
	my $ct_clms_rec = @{$set_clminfo_rec};

	print $fh_log "<< Output of 'assign_resnums_to_clms' >>\n\n";

	print $fh_log "[ For reference MSA (#{columns} = $ct_clms_ref) ]\n\n";

	print $fh_log join ("\t", 'Clm_no.', @{$sequence_ids}), "\n\n";

	for (my $c=0; $c<$ct_clms_ref; $c++) { print $fh_log join ("\t", $c, @{$set_clminfo_ref->[$c]}), "\n"; }
	print $fh_log "\n\n";

	print $fh_log "[ For reconstructed MSA (#{columns} = $ct_clms_rec) ]\n\n";

	print $fh_log join ("\t", 'Clm_no.', @{$sequence_ids}), "\n\n";

	for (my $c=0; $c<$ct_clms_rec; $c++) { print $fh_log join ("\t", $c, @{$set_clminfo_rec->[$c]}), "\n"; }
	print $fh_log "\n\n\n";

   } # END of "if ($TEST_LEVEL>3) { # (O) ...}"

			## END of "(ADDED on Oct 28 (Tue), 2014.)" ##


		## Preliminary Process (1): Map the position shifts (from reference to reconstructed) onto the Reconstructed MSA.

	print STDOUT "<<<<< Preliminary (1): Map the position shifts (from reference to reconstructed) onto the Reconstructed MSA... >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (1): Map the position shifts (from reference to reconstructed) onto the Reconstructed MSA... >>>>>\n\n";

	# $map_shifts_respos->[$k][$s] is the shift in the residue position from the reference MSA to the reconstructed one
	# for the $s th site of the $k th sequence, in the reconstructed MSA. It is $gap if occupied by a gap;
	# $shift_lf is the shift at the left-flanking column of the reconstructed MSA. It must be 0 (zero) by default;
	# $shift_rf is the shift at the right-flanking column of the reconstructed MSA. It must be (len(reconstructed) - len(reference)) by default.
	#
	my ($map_shifts_respos, $shift_lf, $shift_rf) = map_shifts_respos_bw_2msas (@ref_seg_msa, @rec_seg_msa, $GAP);
	unless (defined $map_shifts_respos) {
	    print STDOUT "[ERROR '$program_name'] 'map_shifts_respos_bw_2msas' FAILED on the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

   if ($TEST_LEVEL>1) { # (I) 
	print $fh_log "<< Output of 'map_shifts_respos_bw_2msas' >>\n\n";

	print $fh_log '($shift_lf, $shift_rf) = ', "($shift_lf, $shift_rf) .\n\n";

	print $fh_log "[ Shifts in the Reconstructed MSA ]\n\n";

		# Improved the display of a long segment. #

	unless ( pr_map_shifts_respos (*${$fh_log}, @{$sequence_ids}, @{$map_shifts_respos}, $GAP, $DIS_LEN_SHIFT) ) {
	    print STDOUT "[ERROR '$program_name'] 'pr_map_shifts_respos' FAILED on the RAW map of position-shifts between the reference and reconstructed MSAs (onto the latter) !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ; 
	}

   } # END of "if ($TEST_LEVEL>1) { # (I) ...}"


	my $size_recpad = 0; # ADDED to fix a BUG (in Main Process (3)) on May 11 (Wed), 2016. #

		### NEW PRELIMINARY PROCESS (1hf):
		## PRE-Calculate the commoner flanking shift, and,
		# if the commoner_flanking shift is not zero,
		# nullify it by shifting the entire MSA (either reference or reconstructed).
		## (ADDED as a DRASTIC IMPROVEMENT on Oct 14 (Wed), 2015)

	my $freq_shift_lf = my $freq_shift_rf = 0;

	foreach my $row_shifts (@{$map_shifts_respos}) { # Preliminary outer foreach-loop (over sequences). #
	    foreach my $shift (@{$row_shifts}) { # Preliminary inner foreach-loop (over sites). #
		($shift eq $GAP) and next;
		if ($shift == $shift_lf) { $freq_shift_lf++;
		} elsif ($shift == $shift_rf) { $freq_shift_rf++;
		}
	    }			# End of the preliminary inner foreach-loop (over sites). #
	}		# End of the preliminary outer foreach-loop (over sequences). #

	my $commoner_shift_flank0 = ($freq_shift_lf >= $freq_shift_rf) ? $shift_lf : $shift_rf ;

	if ($commoner_shift_flank0 != 0) {

		# Subtract $commoner_shift_flank0 from all position-shifts in the map. #

	    foreach my $row_shifts (@{$map_shifts_respos}) { # 1st new outer foreach-loop (over sequences). #
		for (my $s=0; $s<$rec_len; $s++) { ($row_shifts->[$s] eq $GAP) or ($row_shifts->[$s] -= $commoner_shift_flank0); }
	    }				# End of the 1st new outer foreach-loop (over sequences). #

			# Prepare auxiliary variables. #

	    my $abs0 = ($commoner_shift_flank0>0)  ? $commoner_shift_flank0 : - $commoner_shift_flank0 ;
	    my $pad = $GAP x $abs0;

	    my @resnums_pad_clm = ();
	    for (1 .. $ct_sequences) { push @resnums_pad_clm, 1/2; }


	    if ($commoner_shift_flank0 < 0) {
		# Pad gap columns on the left-end of the reconstructed MSA.

	(print $fh_log "\n[INFORMATION] The original ", '$commoner_shift_flank', " = $commoner_shift_flank0. Thus, we will shift the entire reconstructed MSA ...\n\n") if ($TEST_LEVEL>1);

		foreach my $seq (@rec_seg_msa) { ${$seq} = $pad . ${$seq}; } 
		##foreach my $seq (@{$rec_msa}) { ${$seq} = $pad . ${$seq}; } # A BUG fixed on Oct 14 (Wed), 2015.

		for (1 .. $abs0) { # 3rd new outer for-loop (over padded columns). #
		   my @cp = @resnums_pad_clm;
		   unshift @{$set_clminfo_rec}, \@cp;
		}		# End of the 3rd new outer for-loop (over padded columns). #

			# In this case, pad the position-shift map as well.

		my @padded_le_gaps = ();
		for (1 .. $abs0) { push @padded_le_gaps, $GAP; }
		##for (1 .. $abs0) { push @padded_le_gaps; } # A BUG fixed on Oct 14 (Wed), 2015. #
		foreach my $row_shifts (@{$map_shifts_respos}) { # 5th new outer for-loop (over sequences). #
		    @{$row_shifts} = (@padded_le_gaps, @{$row_shifts});
		}			# End of the 5th new outer for-loop (over sequences). #

		$rec_len += $abs0;
		$shift_rf -= $commoner_shift_flank0;

		$size_recpad = $abs0; # ADDED to fix a BUG (in Main Process (3)) on May 11 (Wed), 2016. #

	    } else {
		# Pad gap columns on the left-end of the reference MSA.

	(print $fh_log "\n[INFORMATION] The original ", '$commoner_shift_flank', " = $commoner_shift_flank0. Thus, we will shift the entire reference MSA ...\n\n") if ($TEST_LEVEL>1);

		foreach my $seq (@ref_seg_msa) { ${$seq} = $pad . ${$seq}; }
		##foreach my $seq (@{$ref_msa}) { ${$seq} = $pad . ${$seq}; } # A BUG fixed on Oct 14 (Wed), 2015. 

		for (1 .. $abs0) { # 7th new outer for-loop (over padded columns). #
		    my @cp = @resnums_pad_clm;
		    unshift @{$set_clminfo_ref}, \@cp;
		}		# End of the 7th new outer for-loop (over padded columns). #

		$ref_len += $abs0;
		$shift_rf -= $commoner_shift_flank0;
	    }

   if ($TEST_LEVEL>1) { # (I') 

	print $fh_log "<< REVISED Output of 'map_shifts_respos_bw_2msas' >>\n\n";

	print $fh_log 'New ($shift_lf, $shift_rf) = ', "($shift_lf, $shift_rf) .\n\n";

	print $fh_log "[ New Shifts in the Reconstructed MSA ]\n\n";

	unless ( pr_map_shifts_respos (*${$fh_log}, @{$sequence_ids}, @{$map_shifts_respos}, $GAP, $DIS_LEN_SHIFT) ) {
	    print STDOUT "[ERROR '$program_name'] 'pr_map_shifts_respos' FAILED on the REVISED map of position-shifts between the reference and reconstructed MSAs (onto the latter) !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ; 
	}

   } # END of "if ($TEST_LEVEL>1) { # (I') ...}"


	} # END of "if ($commoner_shift_flank0 < 0) {...}"



		## Preliminary Process (2): Put together the mapped position shifts, to create classes of shifts.

	print STDOUT "<<<<< Preliminary (2): Put together the mapped position shifts into some Classes ... >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (2): Put together the mapped position shifts into some Classes ... >>>>>\n\n";

	# @{$set_classes_shift_respos->[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
	# is for the $cl th class of the run of shifts of residue positions, where
	# $shift is the value of shift,
	# $beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
	# @{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
	# @seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
	#
	# @{$set_miniclasses_shift_respos->[$mc]}, which follows the same format as @{$set_classes_shift_respos->[$cl]},
	# is for the $mc th mini-class of the run of shifts of residue positions.
	# Each mini-class consists of single runs of the same shift, beginning and ending at the same sites, and affecting one or more sequence(s).
	#
		# Further MODIFIED in the middle of creation of 'classify_msa_errors.1.pl' (on Oct 22 (Wed), 2016). #
	my ($set_classes_shift_respos, $set_miniclasses_shift_respos, $commoner_shift_flank) = br_list_classes_shift_respos (@{$map_shifts_respos}, $shift_lf, $shift_rf, $GAP);
	##	# MODIFIED in the middle of creation of 'classify_msa_errors.1.pl' (on Oct 16 (Thu), 2016). #
	##my ($set_classes_shift_respos, $commoner_shift_flank) = br_list_classes_shift_respos (@{$map_shifts_respos}, $shift_lf, $shift_rf, $GAP);
	####my ($set_classes_shift_respos, $commoner_shift_flank) = list_classes_shift_respos (@{$map_shifts_respos}, $shift_lf, $shift_rf, $GAP);
	unless (defined $set_classes_shift_respos) {
	    print STDOUT "[ERROR '$program_name'] 'br_list_classes_shift_respos' FAILED on the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $ct_classes_shift = @{$set_classes_shift_respos};
	my $ct_miniclasses_shift = @{$set_miniclasses_shift_respos}; # ADDED on Oct 22 (Wed), 2014. #

   if ($TEST_LEVEL>1) { # (IIa)

	print $fh_log "<< Output of 'br_list_classes_shift_respos' >>\n\n";
	##print $fh_log "<< Output of 'list_classes_shift_respos' >>\n\n";

	print $fh_log '$commoner_shift_flank = ', "$commoner_shift_flank .\n\n\n";

   } # END of "if ($TEST_LEVEL>1) { # (IIa) ...}"

   if ($TEST_LEVEL>1) { # (IIb)

			# ADDED on Oct 22 (Wed), 2014. (I) #

	print $fh_log "[ MINI-Classes of Shifts (#{classes} = $ct_miniclasses_shift) ]\n\n";

	print $fh_log join ("\t", 'Mini-class_no.', 'Shift', 'beg_site', 'end_site', 'beg_rg1,end_rg1;beg_rg2,end_rg2;...', 'IDs,of,affected,seqs'), "\n\n";

	for (my $mc=0; $mc<$ct_miniclasses_shift; $mc++) { # Outer testing for-loop (over mini-classes of shifts). #

	    my ($shift, $beg_site, $end_site, $affected_rgs, $indices_seqs_affected) = @{$set_miniclasses_shift_respos->[$mc]};

	    my $cnct_aff_rgs = '';
	    foreach my $region (@{$affected_rgs}) { $cnct_aff_rgs .= sprintf ("%d,%d;", @{$region}); } 

	    my @seqids_affected = ();
	    foreach my $indx (@{$indices_seqs_affected}) { push @seqids_affected, $sequence_ids->[$indx]; }
	    my $cnct_seqids_affected = join (',', @seqids_affected);

	    print $fh_log join ("\t", $mc, $shift, $beg_site, $end_site, $cnct_aff_rgs, $cnct_seqids_affected), "\n";
	    
	}				# End of the outer testing for-loop (over mini-classes of shifts). #
	print $fh_log "\n\n\n";

			# END of "ADDED on Oct 22 (Wed), 2014. (I)" #

   } # END of "if ($TEST_LEVEL>2) { # (IIb) ...}"


		## Preliminary Process (3'): For each class of MINI-shifts, parsimoniously infer the branches separating the affected sequences from the rest. #
		# (ADDED on Oct 22 (Wed), 2014. (II))

	print STDOUT "<<<<< Preliminary (3'): For each MINI-class of shifts, parsimoniously infer the branch(es) separating the affected sequences from the rest. >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (3'): For each MINI-class of shifts, parsimoniously infer the branch(es) separating the affected sequences from the rest. >>>>>\n\n";

	my $set_sep_patterns_for_miniclasses = dpsm_infer_sep_patterns_for_classes (@{$set_miniclasses_shift_respos}, 4, @{$sequence_ids}, %{$name2id}, %{$node2ch}, %{$node2pa}, %{$node2depth}, $CHAR_ABS, $CHAR_PRES);

	unless (defined $set_sep_patterns_for_miniclasses) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'dpsm_infer_sep_patterns_for_classes' FAILED on the MINI-classes for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

   if ($TEST_LEVEL>2) { # (III') 

	print $fh_log "<< Output of 'dpsm_infer_sep_patterns_for_classes', part 2: Content of ", '@{$set_sep_patterns_for_miniclasses}', " >>\n\n";

	for (my $mc=0; $mc<$ct_miniclasses_shift; $mc++) { # Outer testing for-loop (over mini-classes of position shifts). #

	    print $fh_log "< For the $mc th MINI-class >\n\n";

	    my $sep_patterns_for_class = $set_sep_patterns_for_miniclasses->[$mc];
	    unless (@{$sep_patterns_for_class} == 2) {
		my $ct_sep_patterns = @{$sep_patterns_for_class};
		print STDOUT "[ERROR '$program_name'] ILLEGAL OUTPUT: The number of separation patterns (=$ct_sep_patterns) is NOT the expected value, 2 !\n\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }

	    for (my $pt=1; $pt<2; $pt++) { # Middle testing for-lop (over separation patterns). #
	    ##for (my $pt=0; $pt<2; $pt++) {...} # Middle testing for-lop (over separation patterns). # ($pt = 0 is no longer necessary.) #

		my ($node2parsstat, $br2change, $ct_changes) = @{$sep_patterns_for_class->[$pt]};

		##my $assigned = $assigned_states[$pt];
		##print $fh_log "{ The $pt th separation pattern (when the state '$assigned' is assigned to the affected sequences) }\n\n";

		print $fh_log "[ Separating branches needed (#{branches} = $ct_changes) ]\n\n";

		print $fh_log join ("\t", 'Branch_ID', 'state_ue', 'state_le'), "\n\n";

		foreach my $br_id (sort {$a <=> $b} keys %{$br2change}) { print $fh_log join ("\t", $br_id, @{$br2change->{$br_id}}), "\n"; }

		print $fh_log "\n\n";
	    }			# End of the middle testing for-loop (over separation patterns). #

	    print $fh_log "\n";
	}			# End of the outer testing for-loop (over mini-classes of position shifts). #
	

   } # END of "if ($TEST_LEVEL>2) { # (III')  ...}"

		# END of "(ADDED on Oct 22 (Wed), 2014. (II))" 


		# ADDITIONAL Preliminary Process: Split mini-classes each of which consists of unnaturally remote groups of sequences. #
		# (ADDED on Oct 30 (Thu), 2014.) #

	print STDOUT "<<<<< ADDITIONAL Preliminary Process (3.5'): Split mini-classes each of which consists of unnaturally remote sequences... >>>>>\n\n";
	print $fh_log "<<<<< ADDITIONAL Preliminary Process (3.5'): Split mini-classes each of which consists of unnaturally remote sequences... >>>>>\n\n";

	# @{$set_classes_shift_respos->[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
	# is for the $cl th class of the run of shifts of residue positions, where
	# $shift is the value of shift,
	# $beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
	# @{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
	# @seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
	#
	# @{$set_miniclasses_shift_respos->[$mc]}, which follows the same format as @{$set_classes_shift_respos->[$cl]},
	# is for the $mc th mini-class of the run of shifts of residue positions.
	# Each mini-class consists of single runs of the same shift, beginning and ending at the same sites, and affecting one or more sequence(s).
	#
	# @{$set_sep_patterns_for_classes->[$cl][$pt]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $pt (=0 or 1) th pattern separating the sequences affected by the $cl th class of position shifts, where
	# %node2parsstat = ($node_id => [$state, $ct_pres_ext], ...),
	# %br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...), and
	# $ct_changes is the number of "changes" necessary for separating the affected sequences.
	# ($pt = 0 is when 'absence' is assigned to the affected sequences (i.e., they could be independent), and $pt = 1 is when 'presence' is assigned to them (i.e., they are regarded as sharing the same origin).)
	#
	my ($newset_miniclasses_shift_respos, $newset_sep_patterns_for_miniclasses) = split_unnatural_miniclasses (@{$set_miniclasses_shift_respos}, @{$set_sep_patterns_for_miniclasses}, $topnode_id, %{$node2pa}, %{$node2ch}, %{$node2depth}, $CHAR_PRES, $CHAR_ABS, %{$node2seq_indx});
	unless (defined $newset_miniclasses_shift_respos) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'split_unnatural_miniclasses' FAILED for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	my $newct_miniclasses_shift = @{$newset_miniclasses_shift_respos};


	if ($newct_miniclasses_shift == $ct_miniclasses_shift) {

	    print $fh_log "... NO CHANGES were made ...\n\n\n";

	} else {

		# Replace the old set of MINI-classes with the new one. #

	    ($set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses) = ($newset_miniclasses_shift_respos, $newset_sep_patterns_for_miniclasses);
	    my $diffct_miniclasses = $newct_miniclasses_shift - $ct_miniclasses_shift;
	    $ct_miniclasses_shift = $newct_miniclasses_shift;

	    print $fh_log "... The MINI-classes increased by $diffct_miniclasses compared to the old set!! ...\n\n\n";

   if ($TEST_LEVEL>2) { # (III'') 

	print $fh_log "<<< Output of 'split_unnatural_miniclasses' (New#{MINI-classes} = $ct_miniclasses_shift) >>>\n\n";

	print $fh_log "<< New MINI-Classes of Shifts (#{classes} = $ct_miniclasses_shift) >>\n\n";

	print $fh_log join ("\t", 'New-Mini-class_no.', 'Shift', 'beg_site', 'end_site', 'beg_rg1,end_rg1;beg_rg2,end_rg2;...', 'IDs,of,affected,seqs'), "\n\n";

	for (my $mc=0; $mc<$ct_miniclasses_shift; $mc++) { # Outer testing for-loop (over mini-classes of shifts). #

	    my ($shift, $beg_site, $end_site, $affected_rgs, $indices_seqs_affected) = @{$set_miniclasses_shift_respos->[$mc]};

	    my $cnct_aff_rgs = '';
	    foreach my $region (@{$affected_rgs}) { $cnct_aff_rgs .= sprintf ("%d,%d;", @{$region}); } 

	    my @seqids_affected = ();
	    foreach my $indx (@{$indices_seqs_affected}) { push @seqids_affected, $sequence_ids->[$indx]; }
	    my $cnct_seqids_affected = join (',', @seqids_affected);

	    print $fh_log join ("\t", $mc, $shift, $beg_site, $end_site, $cnct_aff_rgs, $cnct_seqids_affected), "\n";
	    
	}				# End of the outer testing for-loop (over mini-classes of shifts). #
	print $fh_log "\n\n\n";


	print $fh_log "<< New content of ", '@{$set_sep_patterns_for_miniclasses}', " >>\n\n";

	for (my $mc=0; $mc<$ct_miniclasses_shift; $mc++) { # Outer testing for-loop (over mini-classes of position shifts). #

	    print $fh_log "< For the $mc th new MINI-class >\n\n";

	    my $sep_patterns_for_class = $set_sep_patterns_for_miniclasses->[$mc];
	    unless (@{$sep_patterns_for_class} == 2) {
		my $ct_sep_patterns = @{$sep_patterns_for_class};
		print STDOUT "[ERROR '$program_name'] ILLEGAL OUTPUT: The number of separation patterns (=$ct_sep_patterns) is NOT the expected value, 2 !\n\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }

	    for (my $pt=1; $pt<2; $pt++) { # Middle testing for-lop (over separation patterns). #
	    ##for (my $pt=0; $pt<2; $pt++) {...} # Middle testing for-lop (over separation patterns). # ($pt = 0 is no longer necessary.) #

		my ($node2parsstat, $br2change, $ct_changes) = @{$sep_patterns_for_class->[$pt]};

		##my $assigned = $assigned_states[$pt];
		##print $fh_log "{ The $pt th separation pattern (when the state '$assigned' is assigned to the affected sequences) }\n\n";

		print $fh_log "[ Separating branches needed (#{branches} = $ct_changes) ]\n\n";

		print $fh_log join ("\t", 'Branch_ID', 'state_ue', 'state_le'), "\n\n";

		foreach my $br_id (sort {$a <=> $b} keys %{$br2change}) { print $fh_log join ("\t", $br_id, @{$br2change->{$br_id}}), "\n"; }

		print $fh_log "\n\n";
	    }			# End of the middle testing for-loop (over separation patterns). #

	    print $fh_log "\n";
	}			# End of the outer testing for-loop (over mini-classes of position shifts). #
	

   } # END of "if ($TEST_LEVEL>2) { # (III'')  ...}"

	} # END of "if ($newct_miniclasses_shift == $ct_miniclasses_shift) {...} else {...}"


		# END of "ADDED on Oct 30 (Thu), 2014." #


		## Preliminary Process (4'): Merge the MINI-classes of shifts. #
		## (ADDED on Oct 22 (Wed)- 23 (Thu), 2014.)

	print STDOUT "<<<<< Preliminary (4'): Merge the MINI-classes of shifts. >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (4): Merge the MINI-classes of shifts. >>>>>\n\n";

	# @{$set_miniblocks_shift_respos->[$mb]}, which follows the same format as @{$set_blocks_shift_respos->[$bl]},
	# is for the $mb th block created by merging the mini-classes of the run of shifts of residue positions.
	#
	my $set_miniblocks_shift_respos = merge_miniclasses_shift_respos (@{$set_miniclasses_shift_respos}, @{$set_sep_patterns_for_miniclasses}, @{$sequence_ids}, %{$name2id}, %{$id2name}, %{$node2ch}, %{$node2pa}, %{$node2depth}, $CHAR_ABS, $CHAR_PRES);
	unless (defined $set_miniblocks_shift_respos) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'merge_miniclasses_shift_respos' FAILED for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $ct_miniblocks_srp = @{$set_miniblocks_shift_respos};

   if ($TEST_LEVEL>1) { # (IV') 

	print $fh_log "<< Output of 'merge_miniclasses_shift_respos': Content of ", '@{$set_miniblocks_shift_respos}', " (#{blocks} = $ct_miniblocks_srp) >>\n\n";

	for (my $bl = 0; $bl<$ct_miniblocks_srp; $bl++) { # Outer testing for-loop (over mini-blocks). #

	    my ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, $seqs_involved, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_miniblocks_shift_respos->[$bl]};

	    print $fh_log "< The $bl th MINI-block (Shift = $shift; Region = \[$beg_blk, $end_blk\]; Total#{changes} = $ct_changes) >\n\n";
	    print $fh_log "{Indices of the source classes} = ", join (', ', @{$indices_src_classes}), " .\n\n";

	    my @seqids_invlv = ();
	    foreach my $indx (@{$seqs_involved}) { push @seqids_invlv, $sequence_ids->[$indx]; }

	    print $fh_log "{MRCA of involved sequences} = $id_node_mrca (depth=$depth_mrca) .\n\n";
	    print $fh_log "{Involved sequences} = ", join (', ', @seqids_invlv), " .\n\n";

	    my $ct_br = scalar keys %{$br2change};
	    print $fh_log "{ There are $ct_br branches separating the involved sequences from the rest: }\n\n";
	    print $fh_log join ("\t", 'Branch_ID', 'state_ue', 'state_le'), "\n\n";

	    foreach my $br (sort { $a <=> $b } keys %{$br2change}) { print $fh_log join ("\t", $br, @{$br2change->{$br}}), "\n"; }
	    ##foreach my $br (sort {$node2depth->{$a} <=> $node2depth->{$b}} keys %{$br2change}) { print $fh_log join ("\t", $br, @{$br2change->{$br}}), "\n"; }
	    print $fh_log "\n\n";

	    my $ct_excl = @{$excl_sbblocks};
	    print $fh_log "{ There are $ct_excl Excluded sub-blocks: }\n\n";
	    print $fh_log join ("\t", 'No.', 'beg_sbblk', 'end_sbblk', 'nodeid_mrca'), "\n\n";

	    for (my $sb=0; $sb<$ct_excl; $sb++) { print $fh_log join ("\t", $sb, @{$excl_sbblocks->[$sb]}), "\n"; }
	    print $fh_log "\n\n\n";

	}			# End of the outer testing for-loop (over mini-blocks). #

   } # END of "if ($TEST_LEVEL>1) { # (IV')  ...}"



		## Preliminary Process (5'): Identify 'trivial' MINI-blocks. ##

	print STDOUT "<<<<< Preliminary (5'): Identify 'trivial' MINI-blocks. >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (5'): Identify 'trivial' MINI-blocks. >>>>>\n\n";

	# %{$block2trivial} = ($indx_block => 1, ...) for the blocks judged as 'trivial'.
	#
	my $miniblock2trivial = find_trivial_blocks (@{$set_miniblocks_shift_respos}, @{$sequence_ids}, $commoner_shift_flank, $shift_lf, $shift_rf, @{$map_shifts_respos}, $GAP);
	unless (defined $miniblock2trivial) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'find_trivial_blocks' FAILED on the MINI-blocks for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

   if ($TEST_LEVEL>1) { # (V')

	print $fh_log "<< Output of 'find_trivial_blocks' on the MINI-blocks >>\n\n";

	my @trivial_miniblocks = sort {$a <=> $b} keys %{$miniblock2trivial};

	print $fh_log "{'Trivial' MINI-blocks} = ", join (', ', @trivial_miniblocks), " .\n\n\n";

   } # END of "if ($TEST_LEVEL>1) { # (V') ...}"


		## END of "(ADDED on Oct 22 (Wed) - 23 (Thu), 2014.)" #



		## Preliminary Process (6): Identify gap-pattern blocks, calculate their Dollo parsimony scenarios, and the initial parsimony candidate scenario of each gapped segment in the segmental MSAs (reference & reconstructed). ##

	print STDOUT "<<<<< Preliminary (6): Identify gap-pattern blocks, calculate their Dollo parsimony scenarios, and the initial parsimony candidate scenario of each gapped segment in the segmental MSAs (reference & reconstructed). >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (6): Identify gap-pattern blocks, calculate their Dollo parsimony scenarios, and the initial parsimony candidate scenario of each gapped segment in the segmental MSAs (reference & reconstructed). >>>>>\n\n";


	# @{$info_init_psm_scn_ref/rec} stores information on the initial parsimonious scenario candidate for each cluster (gapped segment) in the reference/reconstructed MSA;
	# @{$info_gpatterns} stores information on the gap-patterns, as well as their Dollo parsimony scenarios, that are found in the reference and reconstructed MSAs. 
	#
	# For the contents of these variables, see below.
	#
	my ($info_init_psm_scn_ref, $info_init_psm_scn_rec, $info_gpatterns) = cal_init_psm_cands_for_twin_msas (%{$node2ch}, %{$node2pa}, %{$node2depth}, %{$node2ct_exoffs}, %{$name2id}, @{$sequence_ids}, @ref_seg_msa, @rec_seg_msa, $CHAR_ABS, $CHAR_PRES);

	unless (defined $info_init_psm_scn_ref) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'cal_init_psm_cands_for_twin_msas' FAILED for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	unless (@{$info_gpatterns}>0) {

	    print STDOUT "[INFORMATION] SKIPPED the Calculation for the $s th segment, because either reference or reconstructed MSA contains (a) gap(s) or (a) gapped segment(s) that is (are) too long ...\n\n\n";

	    print $fh_log "SKIPPED the Calculation, because either reference or reconstructed MSA contains (a) gap(s) or (a) gapped segment(s) that is (are) too long ...\n\n\n";
	    close ($fh_log);


		##  MODIFIED (on Nov 9 (Sun), 2014). ##
 		## => Further MODIFIED in creation of 'classify_msa_errors_via_mblks.2.BK1.pl' (on Oct 7 (Wed), 2015). #

	  foreach my $fh_out ($fh_out1) { # ADDED on Oct 7 (Wed), 2015. # MODIFIED in creation of 'examine_purges_via_mblks.2.pl'. #
	  ##foreach my $fh_out ($fh_out1, $fh_out2) {...} # ADDED on Oct 7 (Wed), 2015. #
	    print $fh_out "< Segment ID: $segid >\n\n";
	  }

			# MODIFIED in creation of 'examine_purges_via_mblks.2.pl' (on May 11 (Wed), 2016). #

	    print $fh_out1 join ("\t", 'Rlv_branch', 'len(rlv_branch)', '#{involved_seqs}', 'block_size', '|shift|', '#{sbst_diffs}_exp', '#{sbst_diffs}_obs', 'P-value', 'log[Lk(sbst)]_[rec-ref]'), "\n\n";

	    print $fh_out1 join ("\t", 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), "\n";

	  foreach my $fh_out ($fh_out1) { # ADDED on Oct 7 (Wed), 2015. # MODIFIED in creation of 'examine_purges_via_mblks.2.pl'. #
	  ##foreach my $fh_out ($fh_out1, $fh_out2) {...} # ADDED on Oct 7 (Wed), 2015. #
	    print $fh_out "\n\\\\\n\n";
	  }

	    next;
	}

	# @{$set_cls_gblocks_ref/rec->[$cl][$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA, with
	#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
	#	$len_gb being the length (in terms of #{columns}) of the block,
	#	and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);
	#
	# $init_totcts_indels_for_cls_ref/rec->[$cl] is the total number of insertion/deletion events in the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA;
	#
	# @{$init_psm_scn_for_cls_ref/rec->[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA, where 
	#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
	#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
	#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
	#	  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node.
	#
	my ($set_cls_gblocks_ref, $init_totcts_indels_for_cls_ref, $init_psm_scn_for_cls_ref) = @{$info_init_psm_scn_ref};
	my ($set_cls_gblocks_rec, $init_totcts_indels_for_cls_rec, $init_psm_scn_for_cls_rec) = @{$info_init_psm_scn_rec};


	# ${$set_gpatterns->[$indx_gp]} is the string representing the $ndx_gp th gap-pattern in the input MSA, created by concatenating $CHAR_PRES/ABS in the sequences;
	# %{$set_node2dparsstat->[$indx_gp]} = ($node_id => [$state, $ct_pres_ext], ...) is for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern;
	# %{$set_br2change->[$indx_gp]} = ($branch_id => [$state_ue, $state_le], ...) for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern.
	#
	my ($set_gpatterns, $set_node2dparsstat, $set_br2change) = @{$info_gpatterns};

	my $ct_gps = @{$set_gpatterns};

   if ($TEST_LEVEL>2) { # (VII)

	print $fh_log "<< Output of 'cal_init_psm_cands_for_twin_msas' >>\n\n";

	print $fh_log "[[ Set of gap-patterns (#{patterns} = $ct_gps) ]]\n\n"; 

	print $fh_log join ("\t", 'Indx_gp', '#{indels}', "'gap_pattern(horizontal)'"), "\n\n";

	for (my $gp = 0; $gp<$ct_gps; $gp++) { # 1st outer testing for-loop (over gap-patterns). #
	   my $gpattern = ${$set_gpatterns->[$gp]};
	   my $ct_indels = scalar keys %{$set_br2change->[$gp]} ;
	   print $fh_log join ("\t", $gp, $ct_indels, "'$gpattern'"), "\n";
	}			# End of the 1st outer testing for-loop (over gap-patterns). #
	print $fh_log "\n\n";

     if ($TEST_LEVEL>3) {

	print $fh_log "[ Dollo parsimonious scenario for each gap-pattern ]\n\n";

	for (my $gp=0; $gp<$ct_gps; $gp++) { # 2nd outer testing for-loop (over gap-patterns). #

		my $gpattern = ${$set_gpatterns->[$gp]};
		my $br2change = $set_br2change->[$gp];
		my $ct_indels = scalar keys %{$br2change};

		print $fh_log "{ For the $gp th pattern (horizontal: '$gpattern' ; #{indels} = $ct_indels) }\n\n";

		print $fh_log join ("\t", 'Branch_ID', 'state_ue', 'state_le'), "\n\n";

		foreach my $br (sort {$a <=> $b} keys %{$br2change}) { print $fh_log join ("\t", $br, @{$br2change->{$br}}), "\n"; }
		print $fh_log "\n\n";

      if ($TEST_LEVEL>=5) {
		my $node2dparsstat = $set_node2dparsstat->[$gp];

		print $fh_log ' ' x 3,  "( Map of presence(=$CHAR_PRES)/absence(=$CHAR_ABS) states onto each node )\n\n";
		print $fh_log ' ' x 3, join ("\t", 'Node', 'State'), "\n\n";

		foreach my $nodeid (sort {$a <=> $b} keys %{$node2dparsstat}) { print $fh_log ' ' x 3, join ("\t", $nodeid, $node2dparsstat->{$nodeid}[0]), "\n"; }
		print $fh_log "\n\n";
      }

	}		# End of the 2nd outer testing for-loop (over gap-patterns). #

     } # END of "if ($TEST_LEVEL>3) {...}"

	for my $set_info (
		[$info_init_psm_scn_ref, 'Reference'],
		[$info_init_psm_scn_rec, 'Reconstructed'] ) { # 3rd outer testing for-loop (for reference & reconstructed MSAs). #

	    my ($info_init_psm_scn, $msa_type) = @{$set_info};
	    my ($set_cls_gblocks, $init_totcts_indels_for_cls, $init_psm_scn_for_cls) = @{$info_init_psm_scn};
	    my $ct_cls = @{$set_cls_gblocks};

	    print $fh_log "[[ Gapped segments and the initial candidates of their parsimonious scenarios for '$msa_type' MSA (#{gapped segments} = $ct_cls)  ]]\n\n";



	    for (my $cl=0; $cl<$ct_cls; $cl++) { # Middle testing for-loop (over gapped segments). #

		my $gblks_in_cls = $set_cls_gblocks->[$cl];
		my $ct_blks = @{$gblks_in_cls};
		my $beg_cls = $gblks_in_cls->[0][2];
		my $end_cls = $gblks_in_cls->[$ct_blks-1][3];

		print $fh_log "{ The $cl th gapped segment  (region = \[$beg_cls, $end_cls] ; #{blocks} = $ct_blks ) }\n\n";

		print $fh_log "[ Gap-pattern blocks ]\n\n";

		print $fh_log join ("\t", 'Indx_blk', 'beg_blk', 'end_blk', 'length', 'indx_gp', "'gap-pattern'"), "\n\n";

	# @{$set_cls_gblocks_ref/rec->[$cl][$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA, with
	#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
	#	$len_gb being the length (in terms of #{columns}) of the block,
	#	and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);

		for (my $b=0; $b<$ct_blks; $b++) { # 1st inner testing for-loop (over gap-pattern blocks). #
		    my ($indx_gp, $len_gb, $beg_gb, $end_gb) = @{$gblks_in_cls->[$b]};
		    my $gpattern = ${$set_gpatterns->[$indx_gp]};

		    print $fh_log join ("\t", $b, $beg_gb, $end_gb, $len_gb, $indx_gp, "'$gpattern'"), "\n";
		} 			# End of the 1st inner testing for-loop (over gap-pattern blocks). #
		print $fh_log "\n\n";

		my $totct_indels0 = $init_totcts_indels_for_cls->[$cl];

	# $init_totcts_indels_for_cls_ref/rec->[$cl] is the total number of insertion/deletion events in the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA;

		my ($br2indels, $node2blwise_gstat) = @{$init_psm_scn_for_cls->[$cl]};

		print $fh_log "[ Initial candidate indel scenario (#{indels} = $totct_indels0) ]\n\n";

		print $fh_log join ("\t", 'Branch_ID', '(state_ue/state_le/indices,of,blocks,involved)_event1', '[similarly for event2]', '...'), "\n\n";

		foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # 2nd inner testing for-loop (over branches with indels). #
		    my @out = ($br);
		    foreach my $event (@{$br2indels->{$br}}) { # Inner-inner for-loop (over events along the branch). #
			my ($state_ue, $state_le, $indices_blk) = @{$event};
			my $cnct_indices_blk = join (',', @{$indices_blk});
			push @out, join ('/', $state_ue, $state_le, $cnct_indices_blk);
		    }					# End of the inner-inner for-loop (over events along the branch). #

		    print $fh_log join ("\t", @out), "\n";
		}				# End of the 2nd inner testing for-loop (over branches with indels). #
		print $fh_log "\n\n";


     if ($TEST_LEVEL>3) {

		print $fh_log "( Map of block-wise gap states onto the nodes )\n\n";

		print $fh_log join ("\t", 'Node_ID', "'block_wise_gap_states'"), "\n\n";

		foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat}) { # 3rd inner testing for-loop (over nodes). #

		    my $cnct_bw_gstats = join ('', @{$node2blwise_gstat->{$node}});

		    print $fh_log join ("\t", $node, "'$cnct_bw_gstats'"), "\n";

		}				# End of the 3rd inner testing for-loop (over nodes). #
		print $fh_log "\n\n";

     } # END of "if ($TEST_LEVEL>3) {...}"


	# @{$init_psm_scn_for_cls_ref/rec->[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA, where 
	#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
	#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
	#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
	#	  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node.

	    }			# End of the middle testing for-loop (over gapped segments). #

	    print $fh_log "\n\n";

	}		# End of the 3rd outer testing for-loop (for reference & reconstructed MSAs). #


   } # END of "if ($TEST_LEVEL>2) { # (VII) ...}"


			## ADDED on Oct 23 (Thu), 2014. ##

		## Preliminary Process (7'): Lump together some neighboring MINI-blocks of the same type to form "composite MINI-blocks". ##

	print STDOUT "<<<<< Preliminary (7'): Lump together some neighboring MINI-blocks affecting the identical set of sequences. >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (7'): Lump together some neighboring MINI-blocks affecting the identical set of sequences. >>>>>\n\n";

	# @{$composite_blocks->[$cb]} = ($beg_site, $end_site, $nodeid_mrca, \%br2change, \@constituents, \@involved_seqs, \@merger_types), where
	# @constituents lists the indices of blocks that compose the $cb th composite block (including "composite blocks" each of which consists of a single block (except trivial blocks)).
	# @involved_seqs lists the indices (in @{$sequence_ids}) of the sequences involved,
	# $merger_types[$i] is the type of merger between the $i th and ($i+1) th constituent blocks
	# (= 0 for 'gap-mediated', = 1/2 for 'non-gap-mediated' with  the $i/($i+1) th block incorporated).
	#
			# (MODIFIED on Oct 20 (Mon) - 21 (Tue), 2014.) #
	my $composite_miniblocks = lump_together_similar_blocks (@{$set_miniblocks_shift_respos}, @{$sequence_ids}, %{$miniblock2trivial}, @{$map_shifts_respos}, $GAP, @{$set_cls_gblocks_ref}, @{$set_cls_gblocks_rec}, @{$set_br2change}, %{$node2pa}, %{$node2ch}, $topnode_id);
	##my $composite_blocks = lump_together_similar_blocks_proto (@{$set_blocks_shift_respos}, @{$sequence_ids}, %{$block2trivial}, @{$map_shifts_respos}, $GAP);
	unless (defined $composite_miniblocks) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'lump_together_similar_blocks' FAILED for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $ct_cmp_miniblks = @{$composite_miniblocks};

	my @set_pos_shifts = (); # @{$set_pos_shifts[$cb]} lists the position shifts of the $cb th composite-block. # ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 11 (Wed), 2016). #

   if ($TEST_LEVEL>1) { # (VII)

	print $fh_log "<< Output of 'lump_together_similar_blocks': Content of ", '@{$composite_miniblocks}', " (#{composite_miniblocks} = $ct_cmp_miniblks) >>\n\n";

	print $fh_log join ("\t", 'Indx_cmp_miniblock', 'beg_cmb', 'end_cmb', 'mrca', 'indices,constituent,miniblocks', 'list,position,shifts', 'merger,types', 'indices,involved,seqs'), "\n\n"; # MODIFIED on Oct 21 (Tue), 2014. #
	##print $fh_log join ("\t", 'Indx_cmp_block', 'beg_cb', 'end_cb', 'mrca', 'indices,constituent,blocks', 'merger,types', 'indices,involved,seqs'), "\n\n";

	for (my $cb=0; $cb<$ct_cmp_miniblks; $cb++) { # Testing for-loop (over composite mini-blocks). #

	    my ($beg_cb, $end_cb, $mrca, $br2change, $constituents, $involved_seqs, $merger_types) = @{$composite_miniblocks->[$cb]};
	    my $cnct_constituents = join (',', @{$constituents});
	    my $cnct_involved_seqs = join (',', @{$involved_seqs}); 
	    my $cnct_merger_types = (@{$merger_types}>0) ? join (',', @{$merger_types}) : 'n/a' ;

		# MODIFIED on Oct 21 (Tue), 2014. #

	    my @pos_shifts = ();
				# A BUG fixed on Oct 23 (Thu), 2014. #
	    foreach my $indx_blk (@{$constituents}) { push @pos_shifts, $set_miniblocks_shift_respos->[$indx_blk][0]; }
	    ##foreach my $indx_blk (@{$constituents}) { push @pos_shifts, $set_blocks_shift_respos->[$indx_blk][0]; }
	    my $cnct_pos_shifts = join (',', @pos_shifts);

	    $set_pos_shifts[$cb] = \@pos_shifts; ## ADDED in creation of 'examine_purges_via_mblks.2.pl'. ##

	    print $fh_log join ("\t", $cb, $beg_cb, $end_cb, $mrca, $cnct_constituents, $cnct_pos_shifts, $cnct_merger_types, $cnct_involved_seqs), "\n";
	    ##print $fh_log join ("\t", $cb, $beg_cb, $end_cb, $mrca, $cnct_constituents, $cnct_merger_types, $cnct_involved_seqs), "\n";
	    ####print $fh_log join ("\t", $cb, $beg_cb, $end_cb, $mrca, $br2change, $cnct_constituents, $cnct_merger_types, $cnct_involved_seqs), "\n"; # A BUG fixed. ($br2change was superfluous!!)

	}			# End of the testing for-loop (over composite mini-blocks). #
	print $fh_log "\n\n\n";

   } # END of "if ($TEST_LEVEL>1) { # (VII) ...}"

				## ADDED on Oct 24 (Fri), 2014). ##

		## Preliminary Process (8): Reorganize the list of insertions/deletions in the initial candidate of parsimonious scenarios, for reference and reconstructed MSAs.

	print STDOUT "<<<<< Preliminary (8): Reorganize the list of insertions/deletions in the initial candidate of parsimonious scenarios, for reference and reconstructed MSAs. >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (8): Reorganize the list of insertions/deletions in the initial candidate of parsimonious scenarios, for reference and reconstructed MSAs. >>>>>\n\n";


	print $fh_log "<<< (1) For Reference MSA >>>\n\n";

	# @{$linset_gblocks_ref->[$lgb]} = ($cl, $gb, $indx_gp, $len_gb, $beg_gb, $end_gb) is for the gap-pattern block in the $lgb th place in the linear set, which is the $gb th block in the $cl th cluster (i.e., gapped segment) in the refefence MSA, where
	# $indx_gp is the index of the gap-pattern in the @{$set_gpatterns}, and $beg/end_gb is the beginning/ending coordinate of the block (in the 0-based full-closed convention);
	#
	# %{$br2all_indels_ref} = ($branch_id => \@set_all_indels, ...), where
	# [ @{$set_all_indels[$e]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk) is for the $e th (in the spatial order) indel event along the branch, with ] --- OBSOLETE as of Nov 4 (Tue), 2014.
	# @{$set_all_indels[$e]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, $length_indel, \@indices_lin_blk) is for the $e th (in the spatial order) indel event along the branch, with # MODIFIED on Nov 4 (Tue), 2014.
	# $beg/end_site being the coordinate (in the 0-based full-closed convention) of the beginning/ending-site of the event, and
	# @indices_lin_blk storing the indices (in @{$linset_cls_gblocks_ref}) of the involved gap-pattern blocks;
	#
	# @{$set_indels_std_by_begsite_ref->[$i]} = ($beg_site, \@set_indels_beg_at_here) is for the $i th beginning site (in numerically ascending order), with
	# @{$set_indels_beg_at_here[$e1]} = (exactly in the same format as @{$set_all_indels[$e]}) is for the $e1 th (in the order of occurrences) indel event beginning at $beg_site;
	# @{$set_indels_std_by_endsite_ref->[$i]} = ($end_site, \@set_indels_end_at_here) is for the $i th ending site (in numerically ascending order), with
	# @{$set_indels_end_at_here[$e2]} = (exactly in the same format as @{$set_all_indels[$e]}) is for the $e2 th (in the order of occurrences) indel event ending at $end_site.
	#
	my ($linset_gblocks_ref, $br2all_indels_ref, $set_indels_std_by_begsite_ref, $set_indels_std_by_endsite_ref) = reorganize_set_indels (@{$set_cls_gblocks_ref}, @{$init_psm_scn_for_cls_ref}); 
	unless (defined $linset_gblocks_ref) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'reorganize_set_indels' FAILED for the Reference MSA of the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $totct_gblks_ref = @{$linset_gblocks_ref};
	my $totct_begsites_ref = @{$set_indels_std_by_begsite_ref};
	my $totct_endsites_ref = @{$set_indels_std_by_endsite_ref};

   if ($TEST_LEVEL>2) { # (VIIIa) 

	print $fh_log "[[ Outputs of 'reorganize_set_indels' ]]\n\n";

	print $fh_log '[ Content of @{$linset_gblocks_ref} (#{elements} = ', "$totct_gblks_ref) ]\n\n";

	print $fh_log join ("\t", 'Indx_lin_gblk', 'Indx_cl', 'Indx_gblk', 'Indx_gpatt', 'Len_gb', 'Beg_gb', 'End_gb') , "\n\n";

	for (my $lgb=0; $lgb<$totct_gblks_ref; $lgb++) { print $fh_log join ("\t", $lgb, @{$linset_gblocks_ref->[$lgb]}), "\n"; }
	print $fh_log "\n\n\n";

     if ($TEST_LEVEL>3) {

	print $fh_log '[ Content of %{$br2all_indels_ref} (#{entries} = ', scalar (keys %{$br2all_indels_ref}), ") ]\n\n";

	foreach my $br (sort {$a <=> $b} keys %{$br2all_indels_ref}) { # 2nd testing foreach-loop (over branches). #

	    print $fh_log "{ Branch ID: $br }\n\n";

	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014. #
	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";

	    foreach my $event (@{$br2all_indels_ref->{$br}}) {
		my ($br_id2, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
		##my ($br_id2, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
		print $fh_log ' ' x 3, join ("\t", $br_id2, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
		##print $fh_log ' ' x 3, join ("\t", $br_id2, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
	    }
	    print $fh_log "\n\n";
	}					# End of the 2nd testing foreach-loop (over branches). #
	print $fh_log "\n";

	print $fh_log '[ Content of @{$set_indels_std_by_begsite_ref} (#{elements} = ', "$totct_begsites_ref) ]\n\n";

	for (my $i=0; $i<$totct_begsites_ref; $i++) { # 3rd testing for-loop (over beginning sites). #

	    my ($beg_site0, $events) = @{$set_indels_std_by_begsite_ref->[$i]};

	    print $fh_log "{ The $i th beginning site (= $beg_site0) }\n\n";

	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014.
	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";

	    foreach my $event (@{$events}) {
		my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
		##my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
		print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
		##print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
	    }
	    print $fh_log "\n\n";

	}				# End of the 3rd testing for-loop (over beginning sites). #
	print $fh_log "\n";

	print $fh_log '[ Content of @{$set_indels_std_by_endsite_ref} (#{elements} = ', "$totct_endsites_ref) ]\n\n";

	for (my $i=0; $i<$totct_endsites_ref; $i++) { # 4th testing for-loop (over ending sites). #

	    my ($end_site0, $events) = @{$set_indels_std_by_endsite_ref->[$i]};

	    print $fh_log "{ The $i th ending site (= $end_site0) }\n\n";

	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014.
	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";

	    foreach my $event (@{$events}) {
		my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
		##my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
		print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
		##print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
	    }
	    print $fh_log "\n\n";

	}				# End of the 4th testing for-loop (over ending sites). #
	print $fh_log "\n";


     }   # END of "if ($TEST_LEVEL>3) {...}"


   } # END of "if ($TEST_LEVEL>2) { # (VIIIa) ...}"



	print $fh_log "<<< (2) For Reconstructed MSA >>>\n\n";

	# {Formats of the outputs are the identical to those for the reference MSA}
	#
	my ($linset_gblocks_rec, $br2all_indels_rec, $set_indels_std_by_begsite_rec, $set_indels_std_by_endsite_rec) = reorganize_set_indels (@{$set_cls_gblocks_rec}, @{$init_psm_scn_for_cls_rec}); 
	unless (defined $linset_gblocks_rec) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'reorganize_set_indels' FAILED for the Reconstructed MSA of the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $totct_gblks_rec = @{$linset_gblocks_rec};
	my $totct_begsites_rec = @{$set_indels_std_by_begsite_rec};
	my $totct_endsites_rec = @{$set_indels_std_by_endsite_rec};

   if ($TEST_LEVEL>2) { # (VIIIb) 

	print $fh_log "[[ Outputs of 'reorganize_set_indels' ]]\n\n";

	print $fh_log '[ Content of @{$linset_gblocks_rec} (#{elements} = ', "$totct_gblks_rec) ]\n\n";

	print $fh_log join ("\t", 'Indx_lin_gblk', 'Indx_cl', 'Indx_gblk', 'Indx_gpatt', 'Len_gb', 'Beg_gb', 'End_gb') , "\n\n";

	for (my $lgb=0; $lgb<$totct_gblks_rec; $lgb++) { print $fh_log join ("\t", $lgb, @{$linset_gblocks_rec->[$lgb]}), "\n"; }
	print $fh_log "\n\n\n";

     if ($TEST_LEVEL>3) {

	print $fh_log '[ Content of %{$br2all_indels_rec} (#{entries} = ', scalar (keys %{$br2all_indels_rec}), ") ]\n\n";

	foreach my $br (sort {$a <=> $b} keys %{$br2all_indels_rec}) { # 2nd testing foreach-loop (over branches). #

	    print $fh_log "{ Branch ID: $br }\n\n";

	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014.
	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";

	    foreach my $event (@{$br2all_indels_rec->{$br}}) {
		my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
		##my ($br_id2, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
		print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
		##print $fh_log ' ' x 3, join ("\t", $br_id2, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
	    }
	    print $fh_log "\n\n";
	}					# End of the 2nd testing foreach-loop (over branches). #
	print $fh_log "\n";

	print $fh_log '[ Content of @{$set_indels_std_by_begsite_rec} (#{elements} = ', "$totct_begsites_rec) ]\n\n";

	for (my $i=0; $i<$totct_begsites_rec; $i++) { # 3rd testing for-loop (over beginning sites). #

	    my ($beg_site0, $events) = @{$set_indels_std_by_begsite_rec->[$i]};

	    print $fh_log "{ The $i th beginning site (= $beg_site0) }\n\n";

	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014.
	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";

	    foreach my $event (@{$events}) {
		my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
		##my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
		print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
		##print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
	    }
	    print $fh_log "\n\n";

	}				# End of the 3rd testing for-loop (over beginning sites). #
	print $fh_log "\n";

	print $fh_log '[ Content of @{$set_indels_std_by_endsite_rec} (#{elements} = ', "$totct_endsites_rec) ]\n\n";

	for (my $i=0; $i<$totct_endsites_rec; $i++) { # 4th testing for-loop (over ending sites). #

	    my ($end_site0, $events) = @{$set_indels_std_by_endsite_rec->[$i]};

	    print $fh_log "{ The $i th ending site (= $end_site0) }\n\n";

	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014.
	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";

	    foreach my $event (@{$events}) {
		my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
		##my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
		print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
		##print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
	    }
	    print $fh_log "\n\n";

	}				# End of the 4th testing for-loop (over ending sites). #
	print $fh_log "\n";


     }   # END of "if ($TEST_LEVEL>3) {...}"


   } # END of "if ($TEST_LEVEL>2) { # (VIIIb) ...}"


		## Preliminary Process (9): Identify the pairs of "equivalent" indel events in the reference & reconstructed MSAs.
		## (ADDED on Oct 28 (Tue), 2014.)

	print STDOUT "<<<<< Preliminary (9): Identify the pairs of 'equivalent' indel events in the reference & reconstructed MSAs...  >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (9): Identify the pairs of 'equivalent' indel events in the reference & reconstructed MSAs...  >>>>>\n\n";

	# %{$ref_event2rec_event} = ("$br_ref:$beg_site_ref:$end_site_ref" => "$br_rec:$beg_site_rec:$end_site_rec", ...) for the pairs of "equivalent" indel events,
	# where $beg/end_site_ref/rec is the site coordinate (in the 0-based full-closed convention) of the beginning/ending site of the event in the reference/reconstructed MSA.
	# %{$rec_event2ref_event} is just the inverse of %{$ref_event2rec_event}.
	#
	my ($ref_event2rec_event, $rec_event2ref_event) = identify_equiv_indels (%{$br2all_indels_ref}, %{$br2all_indels_rec}, @{$set_clminfo_ref}, @{$set_clminfo_rec}, $CHAR_ABS, $ct_sequences, %{$node2ch}, %{$node2seq_indx}, %{$node2pa});
	unless (defined $ref_event2rec_event) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'identify_equiv_indels' FAILED on the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $ct_equiv_indels = scalar keys %{$ref_event2rec_event};

   if ($TEST_LEVEL>2) { # (IX)

	print $fh_log "[ Output of 'identify_equiv_indels' (#{equivalent indels} = $ct_equiv_indels) ]\n\n";

	print $fh_log join ("\t", 'Branch_ref:beg_site_ref:end_site_ref', '<=>', 'branch_rec:beg_site_rec:end_site_rec'), "\n\n";

	foreach my $ref_event (sort keys %{$ref_event2rec_event}) { print $fh_log join ("\t", $ref_event, '<=>', $ref_event2rec_event->{$ref_event}), "\n"; }
	print $fh_log "\n\n\n";

   } # END of "if ($TEST_LEVEL>2) { # (IX) ...}"

		## END of "(ADDED on Oct 28 (Tue), 2014.)" ##


				## ADDED on Oct 24 (Fri), 2014). ##

			# FINAL preliminary:
			# Choose whether to use the block-organization or the mini-block-organization. #
			# [ In this script, always use MINI-blocks. ]

	my ($name_tag, $set_classes_to_be_used, $set_seppatterns_for_classes_to_be_used, $set_blocks_to_be_used, $composite_blocks_to_be_used, $block2trivial_to_be_used)
		= ('MINI-Block', $set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses, $set_miniblocks_shift_respos, $composite_miniblocks, $miniblock2trivial);

		## my $set_miniblocks_shift_respos = merge_miniclasses_shift_respos (@{$set_miniclasses_shift_respos}, @{$set_sep_patterns_for_miniclasses}, @{$sequence_ids}, %{$name2id}, %{$id2name}, %{$node2ch}, %{$node2pa}, %{$node2depth}, $CHAR_ABS, $CHAR_PRES);

	my $ct_cblks_to_be_used = @{$composite_blocks_to_be_used};

				## END of "ADDED on Oct 24 (Fri), 2014)." ##


			## ADDED on Nov 10 (Mon), 2014. ##
	my %br2unchgd_indels_rec; # ($branch => [[$beg_site1, $end_site1], [$beg_site2, $end_site2], ...], ....) for the indels in the reconstructed MSA.
	foreach my $label_rec (sort keys %{$rec_event2ref_event}) { # Penultimate preliminary foreach-loop (over reconstructed indels equivalent to reference indels). #
	    my ($br, $beg, $end) = split /:/, $label_rec;
	    my $unchgd_indels = $br2unchgd_indels_rec{$br};
	    (defined $unchgd_indels) or ($unchgd_indels = $br2unchgd_indels_rec{$br} = []);
	    push @{$unchgd_indels}, [$beg, $end];
	}				# End of the penultimate preliminary foreach-loop (over reconstructed indels equivalent to reference indels). # 
			## END of "ADDED on Nov 10 (Mon), 2014." ##


			## ADDED on Oct 29 (Wed), 2014. ##

	my %br2sgl_chg_cbs; # = ($branch_id => [list of indices of composite-blocks with only a single change, which occurs along $branch_id], ...)
	my @cbwise_end_shifts = (); # The $cb th element = [$shift_le, $shift_re] for the $cb th composite-block. 

	for (my $cb=0; $cb<$ct_cblks_to_be_used; $cb++) { # Final prelliminary for-loop (over composite blocks). #

	    my ($beg_cb_rec, $end_cb_rec, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$composite_blocks_to_be_used->[$cb]};

	    my $ct_constituents = @{$constituent_blks};
	    my @shifts = ();
	    foreach my $bl (@{$constituent_blks}) { push @shifts, $set_blocks_to_be_used->[$bl][0]; }

	    my ($shift_le, $shift_re) = @shifts[0, $ct_constituents-1]; # The values of shift at the left-end and right-end.

	    $cbwise_end_shifts[$cb] = [$shift_le, $shift_re];


	    my @br = keys %{$br2change_cb};
	    (@br==1) or next;

	    my $br = $br[0];
	    my $sgl_chg_cbs = $br2sgl_chg_cbs{$br};
	    (defined $sgl_chg_cbs) or ($sgl_chg_cbs = $br2sgl_chg_cbs{$br} = []);

	    push @{$sgl_chg_cbs}, $cb;
	}				# End of the final preliminary for-loop (over composite blocks). #


	my %equiv_branches; # = ($branch_id => $equivalent_branch_id, ...)

	if (@{$node2ch->{$topnode_id}}==2) {
	   my ($br1, $br2) = @{$node2ch->{$topnode_id}};
	   $equiv_branches{$br1} = $br2;
	   $equiv_branches{$br2} = $br1;
	}

			## END of "ADDED on Oct 29 (Wed), 2014." ##


		### (i) MAIN PROCESS!!!: Associate each position-shift-block with (an) appropriate type(s) of MSA error(s). ###
		### (ADDED on Oct 29 (Wed), 2014. => MODIFIED on Oct 7 (Wed), 2015) ###

	print STDOUT "<<<<< (i) MAIN PROCESS (1st Round)!!!: Associate each Composite '$name_tag' with (an) appropriate type(s) of MSA error(s)... (#{composite blocks} = $ct_cblks_to_be_used) >>>>>\n\n";
	##print STDOUT "<<<<< (i) MAIN PROCESS!!!: Associate each Composite '$name_tag' with (an) appropriate type(s) of MSA error(s)... (#{composite blocks} = $ct_cblks_to_be_used) >>>>>\n\n";

	print $fh_log "<<<<< (i) MAIN PROCESS (1st Round)!!!: Associate each Composite '$name_tag' with (an) appropriate type(s) of MSA error(s)... (#{composite blocks} = $ct_cblks_to_be_used) >>>>>\n\n";
	##print $fh_log "<<<<< (i) MAIN PROCESS!!!: Associate each Composite '$name_tag' with (an) appropriate type(s) of MSA error(s)... (#{composite blocks} = $ct_cblks_to_be_used) >>>>>\n\n";


	my @cblk_wise_msa_errors = (); # @{$cblk_wise_msa_errors[$cb]->[$er]} = ($type, \@indels_ref, \@indels_rec) is for the $er th error associated with the $cb th composite block, where
					# $type is a string indicating the type of the error,
					# @{$indels_ref[$f]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk) is for the $f th involved indel in the reference MSA,
					# @{$indels_rec[$c]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk) is for the $c th involved indel in the reconstructed MSA.

	my @cblk_wise_cts_invlvd_indels = (); # $cblk_wise_cts_invlvd_indels[$cb] = [$ct_indels_rlv_ref, $ct_indels_rlv_rec, $ct_indels_rltd_ref, $ct_indels_rltd_rec, $ct_indels_other_invlvd_ref, $ct_indels_other_invlvd_rec] for the $cb th composite-block.

	my %cblk2if_skipped; # = ($indx_cblk => $reason, ...) for the composite-blocks skipped to be tested.

	my %indel_ref2assoc_cblks; # = ("$br:$beg:$end" => [list of composite-blocks associated with the indel event], ...) for indels in the reference MSA.
	my %indel_rec2assoc_cblks; # similarly, for indels in the reconstructed MSA.

	my @set_relv_branches = (); # $set_relv_branches[$cb] = [list of relvant branches] # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 5 (Tue), 2015). #


	for (my $cb=0; $cb < $ct_cblks_to_be_used; $cb++) { # Main middle for-loop (over composite blocks). #

	    my ($beg_cb_rec, $end_cb_rec, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$composite_blocks_to_be_used->[$cb]};

   (print STDOUT "<<< The $cb th composite '$name_tag' >>>\n\n") if ($TEST_LEVEL_newtype>0); # ADDED on Oct 3 (Fri), 2015.

	    my ($shift_le, $shift_re) = @{$cbwise_end_shifts[$cb]};

	    my $beg_cb_ref = $beg_cb_rec - $shift_le;
	    my $end_cb_ref = $end_cb_rec - $shift_re;

	    my @sep_branches = sort {$a <=> $b} keys %{$br2change_cb};

	    my @relv_branches = ();

	    if (@sep_branches==1) {

		@relv_branches = @sep_branches;

	    } else {
			# Choose the branches to be examined in the main process. #

		foreach my $br (@sep_branches) { # 1st preliminary inner foreach-loop (over separating branches). #


				# ADDED on Nov 10 (Mon), 2014. # Check if the branch "separates" the block via an event that remained unchanged from the reference to the reconstructed. #

		    my $unchgd_indels_rec = $br2unchgd_indels_rec{$br};
			## %br2unchgd_indels_rec; # ($branch => [[$beg_site1, $end_site1], [$beg_site2, $end_site2], ...], ....) for the indels in the reconstructed MSA.

		    if (defined $unchgd_indels_rec) {
			my $if_covered2 = 0;
			foreach my $indel_rec (@{$unchgd_indels_rec}) { # 1st inner-inner foreach-loop (over reconstructed indels with equivalents). #
			    my ($beg_ev_rec, $end_ev_rec) = @{$indel_rec};
			    ($end_cb_rec < $beg_ev_rec) and next; # The event is completely on the right of $cb.
			    ($end_ev_rec < $beg_cb_rec) and next; # The event is completely on the left of $cb.
			    (($beg_cb_rec < $beg_ev_rec) and ($end_ev_rec < $end_cb_rec)) and next; # The event resides completely within $cb.
			    $if_covered2 = 1;
			    last;
			}				# End of the 1st inner-inner foreach-loop (over reconstructed indels with equivalents). #
			($if_covered2) and next; # Skip if the 'separating branch' can be attributed to the event.
		    }


				# END of "ADDED on Nov 10 (Mon), 2014." #
		    

		    my $sgl_chg_cbs = $br2sgl_chg_cbs{$br};

		    unless (defined $sgl_chg_cbs) {

			my $equiv_br = $equiv_branches{$br};
			$sgl_chg_cbs = (defined $equiv_br) ? $br2sgl_chg_cbs{$equiv_br} : undef;

			unless (defined $sgl_chg_cbs) {
				    # No composite-blocks with only a single change that occurs along $br.
				    # (That is, no composite-blocks that are phylogenetically delimited by $br alone.)
			    push @relv_branches, $br;
			    next;
			}
		    }

		    # @{$sgl_chg_cbs} contains composite-blocks that are phylogenetically delimited by $br alone. #

		    my $if_covered = 0;
		    foreach my $cb2 (@{$sgl_chg_cbs}) { # Inner-inner foreach-loop (over relevant composite-blocks). #
			my ($beg_cb2_rec, $end_cb2_rec) = @{$composite_blocks_to_be_used->[$cb2]}[0,1];
			my ($shift2_le, $shift2_re) = @{$cbwise_end_shifts[$cb2]};

				# Calculate the boundaries of the maximum possible range of $cb2. #
			($shift2_le > $shift_le) and ($beg_cb2_rec -= $shift2_le - $shift_le); 
			($shift2_re < $shift_re) and ($end_cb2_rec -= $shift2_re - $shift_re);

			($end_cb_rec < $beg_cb2_rec) and next; # $cb2 is completely on the right of $cb.
			($end_cb2_rec < $beg_cb_rec) and next; # $cb2 is completely on the left of $cb.
			(($beg_cb_rec < $beg_cb2_rec) and ($end_cb2_rec < $end_cb_rec)) and next; # $cb2 resides completely within $cb.

			$if_covered = 1;
			last;
		    }				# End of the inner-inner foreach-loop (over relevant composite-blocks). #

		    ($if_covered) or (push @relv_branches, $br); # Only those not covered by any single-branch composite-block will be examined.

		}			# End of the 1st preliminary inner foreach-loop (over separating branches). #
	    }

	    $set_relv_branches[$cb] = \@relv_branches; # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 5 (Tue), 2015). #

	    unless (@relv_branches>0) {
		$cblk2if_skipped{$cb} = 'NO_RELEVANT_BRANCH';
		$cblk_wise_msa_errors[$cb] = undef;
		##$cblk_wise_msa_errors[$cb] = [];
		$cblk_wise_cts_invlvd_indels[$cb] = [];
		next;
	    }

	    if (@relv_branches>1) {
		my $ct_relv_branches = @relv_branches;
		$cblk2if_skipped{$cb} = "MULTIPLE_RELEVANT_BRANCHES(#{branches}=$ct_relv_branches)";
		$cblk_wise_msa_errors[$cb] = undef;
		$cblk_wise_cts_invlvd_indels[$cb] = [];
		next;
	    }

	    my $rlv_br = $relv_branches[0];

	    my @indices_indel_identifiers = (0, 1, 2); # Indices (for @{ ->[$e]}) specifying the branch ID, beginning site, and ending-site, respectively. # MOVED from below, within the candidate region of the new subroutine, (on Oct 6 (Tue), 2015). #

			# REPLACE the core block of the main routine with a new subroutine (plus a post-processing subroutine) !! #
			# (MODIFIED on Oct 6 (Tue), 2015.)

	    my ($type, $reason_skipped, $set_cts_invlvd_indels, $indels_invlvd_ref, $indels_invlvd_rec)
		= classify_error_assoc_w_sgl_cblk ($cb, @{$composite_blocks_to_be_used}, $shift_le, $shift_re, $rlv_br, %equiv_branches,
		    $topnode_id, %{$node2ch}, %{$node2pa}, %{$br_attr},
		    @{$set_indels_std_by_begsite_rec}, @{$set_indels_std_by_endsite_rec}, @{$set_indels_std_by_begsite_ref}, @{$set_indels_std_by_endsite_ref}, 
		    %{$rec_event2ref_event}, %{$ref_event2rec_event});

	    unless ((defined $type) or (defined $reason_skipped)) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'classify_error_assoc_w_sgl_cblk' FAILED on the $cb th composite shift-block for the $s th segment !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }

	    $cblk_wise_cts_invlvd_indels[$cb] = $set_cts_invlvd_indels;

	    unless (defined $type) {

			## The test of this block was SKIPPED. ##

		$cblk_wise_msa_errors[$cb] = undef;
		$cblk2if_skipped{$cb} = $reason_skipped;

		next;
	    }

	    unless ( register_association_cblk_and_indels ($type, $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @{$indels_invlvd_ref}, @{$indels_invlvd_rec}, @indices_indel_identifiers) ) {
		print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED after 'classify_error_assoc_w_sgl_cblk' on the $cb th composite shift-block for the $s th segment !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
			# END of "REPLACE the core block of the main routine with a new subroutine (plus a post-processing subroutine) !!" #


	}					# End of the main middle for-loop (over composite blocks). #

   (print STDOUT "\n... The MAIN PROCESS (1st Round) ENDED for the $s th segment ...\n\n\n") if ($TEST_LEVEL>1);

	my @purge_assoc_cbs = (); # List of indices of composite-blocks that are associated with "Purge"s. # ADDED in creation of 'examine_purges_via_mblks.2.pl'. #

	my @len_cblks = (); # $len_cblks[$cb] = [$len_cb_ref, $len_cb_rec] # ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 11 (Wed), 2016). #

			# ADDED on Nov 8 (Sat), 2014. #
	my @skipped_cblks = sort {$a <=> $b} keys %cblk2if_skipped;
	my $ct_skipped_cblks = @skipped_cblks;
			# END of "ADDED on Nov 8 (Sat), 2014." #

  if ($TEST_LEVEL>0) { # (X)

		# MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 6 (Tue), 2015). #
	print $fh_log "\n[[ Results of the Main Process (1st Round) ]]\n\n";
	##print $fh_log "\n[[ Results of the Main Process ]]\n\n";

    if ($TEST_LEVEL>1) { # (i) # This output is for testing each (sub)-step of the MAIN PROCESS. #

	print $fh_log '[ Contents of @cblk_wise_cts_invlvd_indels ]', "\n\n";

	print $fh_log join ("\t", 'Indx_cmp_blk', '#{rlv_indels}_ref', '#{rlv_indels}_rec', '#{rltd_indels}_ref', '#{rltd_indels}_rec', '#{other_involved}_ref', '#{other_involved}_rec'), "\n\n";

	for (my $cb=0; $cb<$ct_cblks_to_be_used; $cb++) { print $fh_log join ("\t", $cb, @{$cblk_wise_cts_invlvd_indels[$cb]}), "\n"; }
	print $fh_log "\n\n";

    } # END of "if ($TEST_LEVEL>1) { (i) ...}"

			# MODIFIED on Nov 8 (Sat), 2014. #
	print $fh_log "[ Skipped Composite-Blocks (#{cblocks} = $ct_skipped_cblks): ", join (', ', @skipped_cblks), " . ]\n\n\n";
	##print $fh_log '[ Skipped Composite-Blocks: ', join (', ', keys %cblk2if_skipped), " . ]\n\n\n";

	print $fh_log '[ Contents of @cblk_wise_msa_errors ]', "\n\n";

	print $fh_log join ("\t", 'Indx_cmp_blk', 'Indx_error', 'len_cblk_ref', 'len_cblk_rec', 'Type', 'br1:beg1:end1:stat_ue1/br2:beg2:end2:stat_ue2/...(ref)', 'br1:beg1:end1:stat_ue1/br2:beg2:end2:stat_ue2/...(rec)'), "\n\n"; # MODIFIED on Nov 9 (Sun), 2014. #
	##print $fh_log join ("\t", 'Indx_cmp_blk', 'Indx_error', 'Type', 'br1:beg1:end1:stat_ue1/br2:beg2:end2:stat_ue2/...(ref)', 'br1:beg1:end1:stat_ue1/br2:beg2:end2:stat_ue2/...(rec)'), "\n\n";

	for (my $cb=0; $cb<$ct_cblks_to_be_used; $cb++) { # Testing outer for-loop (over composite-blocks). #

	    my $msa_errors_assoc_cb = $cblk_wise_msa_errors[$cb];

	    unless (defined $msa_errors_assoc_cb) {
		my $reason = $cblk2if_skipped{$cb};
		my $type = (defined $reason) ? "Skipped!!($reason)" : 'Skipped!!(reason_unclear)';
		print $fh_log join ("\t", $cb, $type), "\n";
		next;
	    }
			# ADDED on Nov 9 (Sun), 2014. #
	    my ($beg_cb_rec, $end_cb_rec, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$composite_blocks_to_be_used->[$cb]};

	    my ($shift_le, $shift_re) = @{$cbwise_end_shifts[$cb]};

	    my $beg_cb_ref = $beg_cb_rec - $shift_le;
	    my $end_cb_ref = $end_cb_rec - $shift_re;

	    my $len_cb_rec = $end_cb_rec - $beg_cb_rec +1;
	    my $len_cb_ref = $end_cb_ref - $beg_cb_ref +1;
			# END of "ADDED on Nov 9 (Sun), 2014." #

	    $len_cblks[$cb] = [$len_cb_ref, $len_cb_rec]; # ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 11 (Wed), 2016). #

	    my $ct_msa_errors = @{$msa_errors_assoc_cb};

	    for (my $er=0; $er<$ct_msa_errors; $er++) { # Testing inner for-loop (over errors associated with $cb). #

		my ($type, $indels_ref, $indels_rec) = @{$msa_errors_assoc_cb->[$er]};

		my @br_indels_ref = ();
		foreach my $event (@{$indels_ref}) { push @br_indels_ref, join (':', @{$event}[0,1,2,3]); } 
		my $cnct_indels_ref = (@br_indels_ref>0) ? join ('/', @br_indels_ref) : 'None';

		my @br_indels_rec = ();
		foreach my $event (@{$indels_rec}) { push @br_indels_rec, join (':', @{$event}[0,1,2,3]); }
		my $cnct_indels_rec = (@br_indels_rec>0) ? join ('/', @br_indels_rec) : 'None';

		print $fh_log join ("\t", $cb, $er, $len_cb_ref, $len_cb_rec, $type, $cnct_indels_ref, $cnct_indels_rec), "\n"; # MODIFIED on Nov 9 (Sun), 2014. #
		##print $fh_log join ("\t", $cb, $er, $type, $cnct_indels_ref, $cnct_indels_rec), "\n";

		($type =~ m/^Purge/) and (push @purge_assoc_cbs, $cb); # ADDED in creation of 'examine_purges_via_mblks.2.pl'. #

	    }					# End of the testing inner for-loop (over errors associated with $cb). #
	}				# End of the testing outer for-loop (over composite-blocks). #
	print $fh_log "\n\n";

    if ($TEST_LEVEL>1) { # (ii)
	print $fh_log '[ Contents of %indel_ref2assoc_cblks ]', "\n\n";
	print $fh_log join ("\t", 'Br:beg:end(ref)', 'indices,of,associated,composite-blocks'), "\n\n";

	foreach my $info_set_indels (@{$set_indels_std_by_begsite_ref}) { # 2nd testing outer for-loop (over beginning sites of the reference indels). #
	    foreach my $event (@{$info_set_indels->[1]}) { # 2nd inner testing for-loop (over reference indels sharing the beginning-site).
		my $indel_ref = join (':', @{$event}[0,1,2]);
		my $indel_rec = $ref_event2rec_event->{$indel_ref};
		if (defined $indel_rec) {
		    print $fh_log join ("\t", $indel_ref, "{Equivalent to '$indel_rec'(rec)}"), "\n";
		    next;
		}
		my $assoc_cblks = $indel_ref2assoc_cblks{$indel_ref};
		my $cnct_assoc_cblks = ((defined $assoc_cblks) and (@{$assoc_cblks}>0)) ? join (',', @{$assoc_cblks}) : 'None' ;
		print $fh_log join ("\t", $indel_ref, $cnct_assoc_cblks), "\n";
	    }						# End of the 2nd inner testing for-loop (over reference indels sharing the beginning-site).
	}					# End of the 2nd testing outer for-loop (over beginning sites of the reference indels). #
	print $fh_log "\n\n";

	print $fh_log '[ Contents of %indel_rec2assoc_cblks ]', "\n\n";
	print $fh_log join ("\t", 'Br:beg:end(rec)', 'indices,of,associated,composite-blocks'), "\n\n";

	foreach my $info_set_indels (@{$set_indels_std_by_begsite_rec}) { # 2nd testing outer for-loop (over beginning sites of the reconstructed indels). #
	    foreach my $event (@{$info_set_indels->[1]}) { # 2nd inner testing for-loop (over reconstructed indels sharing the beginning-site).
		my $indel_rec = join (':', @{$event}[0,1,2]);
		my $indel_ref = $rec_event2ref_event->{$indel_rec};
		if (defined $indel_ref) {
		    print $fh_log join ("\t", $indel_rec, "{Equivalent to '$indel_ref'(ref)}"), "\n";
		    next;
		}
		my $assoc_cblks = $indel_rec2assoc_cblks{$indel_rec};
		my $cnct_assoc_cblks = ((defined $assoc_cblks) and (@{$assoc_cblks}>0)) ? join (',', @{$assoc_cblks}) : 'None' ;
		print $fh_log join ("\t", $indel_rec, $cnct_assoc_cblks), "\n";
	    }						# End of the 2nd inner testing for-loop (over reconstructed indels sharing the beginning-site).
	}					# End of the 2nd testing outer for-loop (over beginning sites of the reconstructed indels). #
	print $fh_log "\n\n";

    } # END of  "if ($TEST_LEVEL>1) { # (ii) ...}"

  } # END of "if ($TEST_LEVEL>0) { # (X) ...}"


		## Trailing Process: List the reference/reconstructed indel events NOT covered by any block (and without equivalent events). ##
		## ... Withheld ...


		## (ii) Main Process B: ... May not be necessary in this program ...


		## Final Outputs: 
		## (ADDED on Nov 8 (Sat) - 9 (Sun), 2014.)

	my $ct_cblks_wo_assoc_events = my $ct_noneff_cblks = 0; # The number of composite-blocks without associated indel events, and the number of composite-blocks 'not effectively' associated with any indel events.
	my $ct_skipped_cblks2 = 0;
	my @list_error_types = my @list_lengths_eff_cblks = ();
	my $totlength_eff_cblks = 0;

	for (my $cb=0; $cb<$ct_cblks_to_be_used; $cb++) { # Main trailing middle for-loop (over composite-blocks). #

	    my $msa_errors_assoc_cb = $cblk_wise_msa_errors[$cb];

	    unless (defined $msa_errors_assoc_cb) {
		$ct_skipped_cblks2++;
		next;
	    }
	    my $ct_msa_errors = @{$msa_errors_assoc_cb};

	    my ($beg_cb_rec, $end_cb_rec, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$composite_blocks_to_be_used->[$cb]};

	    my ($shift_le, $shift_re) = @{$cbwise_end_shifts[$cb]};

	    my $beg_cb_ref = $beg_cb_rec - $shift_le;
	    my $end_cb_ref = $end_cb_rec - $shift_re;

	    my $len_cb_rec = $end_cb_rec - $beg_cb_rec +1;
	    my $len_cb_ref = $end_cb_ref - $beg_cb_ref +1;
	    my $len_cb = ($len_cb_rec < $len_cb_ref) ? $len_cb_rec : $len_cb_ref ; # Take the length whichever is smaller. #

	    for (my $er=0; $er<$ct_msa_errors; $er++) { # Main trailing inner for-loop (over errors associated with $cb). #
		my ($type, $indels_ref, $indels_rec) = @{$msa_errors_assoc_cb->[$er]};

		if ($type =~ m/^NO_ASSOCIATED_EVENT/) {

		    $ct_cblks_wo_assoc_events++;
		    next;

		}
		push @list_error_types, $type;
		push @list_lengths_eff_cblks, $len_cb;
		$totlength_eff_cblks += $len_cb;
	    }					# End of the main trailing inner for-loop (over errors associated with $cb). #
	}				# End of the main trailing middle for-loop (over composite-blocks). #

	my $ct_eff_cblks = @list_error_types;

	my $tot_effct_cblks = $ct_eff_cblks + $ct_skipped_cblks2 + $ct_cblks_wo_assoc_events + $ct_noneff_cblks;

	my $cnct_error_types = join (';', @list_error_types);
	my $cnct_lens_eff_cblks = join (',', @list_lengths_eff_cblks);

	## $ct_cblks_wo_assoc_events = my $ct_noneff_cblks = 0; # The number of composite-blocks without associated indel events, and the number of composite-blocks 'not effectively' associated with any indel events.
	## $ct_skipped_cblks2 = 0;
	## $totlength_eff_cblks = 0;

		## END of "(ADDED on Nov 8 (Sat) - 9 (Sun), 2014.)"


			### END of "NEWLY ADDED in creation of 'classify_msa_errors.1.pl'." ###


			### NEWLY ADDED in creation of 'examine_purges_via_mblks.2.pl'. ###

	print $fh_log "\n<<<<< (iii) MAIN PROCESS (3rd Round)!!!: Examine the \"Purge\"... >>>>>\n\n";

	unless (@purge_assoc_cbs == 1) {
	    my $ct_pacbs = @purge_assoc_cbs;
	    print STDOUT "[ERROR '$program_name'] The number of purge-associated composite-blocks (=$ct_pacbs) DIFFERS FROM 1 !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $indx_pacb = $purge_assoc_cbs[0];
	my $relv_branches_purge = $set_relv_branches[$indx_pacb];

	unless (@{$relv_branches_purge}==1) {
	    my $ct_rlv_brs = @{$relv_branches_purge};
	    print STDOUT "[ERROR '$program_name'] The number (=$ct_rlv_brs) of branches relevant to the purge-associated composite-block (index: $indx_pacb) DIFFERS FROM 1 !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $rlv_br_prg = $relv_branches_purge->[0];
	my $rlv_brlen = $br_attr->{$rlv_br_prg}[$INDX_LENGTH];
	if (($rlv_brlen == 0) and (defined $equiv_branches{$rlv_br_prg})) {
	    $rlv_br_prg = $equiv_branches{$rlv_br_prg};
	    $rlv_brlen = $br_attr->{$rlv_br_prg}[$INDX_LENGTH];
	}

		# MODIFIED on May 11 (Wed), 2016. #
	my $prg_invlvd_seqindices = $composite_blocks_to_be_used->[$indx_pacb][5];
	my $ct_prg_invlvd_seqs = @{$prg_invlvd_seqindices};
	my @prg_invlvd_seqids = ();
	foreach my $indx (@{$prg_invlvd_seqindices}) { push @prg_invlvd_seqids, $sequence_ids->[$indx]; }
	##my $prg_invlvd_seqs = $composite_blocks_to_be_used->[$indx_pacb][5];
	##my $ct_prg_invlvd_seqs = @{$prg_invlvd_seqs};

		## MODIFIED on May 11 (Wed), 2016. ##

	my ($len_cb_ref, $len_cb_rec) = @{$len_cblks[$indx_pacb]};
	my $pos_shifts = $set_pos_shifts[$indx_pacb];

	my $size_block = $len_cb_ref;
	unless ($len_cb_ref == $len_cb_rec) {
	    print STDOUT "[WARNING '$program_name'] The purge-associated block has TWO SIZES (=$len_cb_ref in reference MSA, =$len_cb_rec in reconstructed MSA) !! ... I will use the average ...\n\n";
	    $size_block = ($len_cb_ref + $len_cb_rec)/2;
	}

	my $shift = $pos_shifts->[0];
	unless (@{$pos_shifts} == 1) {
	    print STDOUT "[WARNING '$program_name'] The purge-associated block has MORE THAN ONE position-shift !\n";
	    my $sum_shift = 0;
	    my $ct_discrepancies =0;
	    foreach my $ps (@{$pos_shifts}) {
		$sum_shift += $ps;
		($shift == $ps) or ($ct_discrepancies++);
	    }
	    unless ($ct_discrepancies==0) {
		print STDOUT "[CAUTION '$program_name'] There is(are) $ct_discrepancies position-shift(s) that DIFFER(S) from  $shift !! ... I will use the average ...\n\n";
		$shift = $sum_shift /scalar (@{$pos_shifts}) ;
	    }
	}
	my $abs_shift = ($shift>0) ? $shift : - $shift ;

	## @set_pos_shifts = (); # @{$set_pos_shifts[$cb]} lists the position shifts of the $cb th composite-block. # ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 11 (Wed), 2016). #
	## @len_cblks = (); # $len_cblks[$cb] = [$len_cb_ref, $len_cb_rec] # ADDED in creation of 'examine_purges_via_mblks.2.pl' (on May 11 (Wed), 2016). #


	print $fh_log "Indx(Purge-associated composite-block) = $indx_pacb .\n";
	print $fh_log "{Relevant branch ID} = $rlv_br_prg  (length = $rlv_brlen) .\n\n";

		# MODIFIED on May 11 (Wed), 2016. #
	print $fh_log "Involved sequences (# = $ct_prg_invlvd_seqs):  ", join (', ', @prg_invlvd_seqids), " .\n\n";
	##print $fh_log "Involved sequences (# = $ct_prg_invlvd_seqs):  ", join (', ', @{$prg_invlvd_seqs}), " .\n\n";
	print $fh_log "Block size = $size_block ,\n";
	print $fh_log "|Shift|    = $abs_shift .\n\n";


	print $fh_log "\n<<< Calculate the Expected and \"Observed\" Substitutional Differences >>>\n\n";

			# NOTE: Currently, we are using only a single tree (with $t = 0). #
	my $br2tr_mtrx  = $set_br2tr_mtrx_tn[0];
	my $node2ntfreqs = $set_node2ntfreqs[0];

	my ($br2set_cw_expcts_sbsts_rec, $br2set_cw_obscts_sbsts_rec)
	    = calc_br2cw_expcts_obscts_sbsts (@cp_rec_seg_msa, @{$sequence_ids}, %{$name2id}, %{$node2ch}, %{$node2pa}, %{$br2tr_mtrx}, %{$node2ntfreqs}, $GAP);
	unless (defined $br2set_cw_expcts_sbsts_rec) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'calc_br2cw_expcts_obscts_sbsts' FAILED on the reconstructed MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}


		# A BUG fixed on May 11 (Wed), 2016. # 
	my ($beg_pacb, $end_pacb) =  @{$composite_blocks_to_be_used->[$indx_pacb]}[0,1];
	my @window_rec = ($beg_pacb - $size_recpad .. $end_pacb - $size_recpad);
	##my @window_rec = (0 .. $size_block -1);

	print $fh_log "Columns in the window: ", join (', ', @window_rec), " .\n\n";

	my ($size_wd, $d_exp, $d_obs, $pv) = calc_pv_obscts_sbsts_wd_br (@window_rec, $rlv_br_prg, %{$br2set_cw_expcts_sbsts_rec}, %{$br2set_cw_obscts_sbsts_rec});
	unless (defined $pv) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'calc_pv_obscts_sbsts_wd_br' FAILED on the reconstructed MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	print $fh_log "#{diffs}_exp = $d_exp ,\n";
	print $fh_log "#{diffs}_obs = $d_obs ,\n";
	print $fh_log "P-value = $pv .\n\n";


	print $fh_log "\n<<< Calculate the Substitutional Likelihoods of the reference and reconstructed MSAs >>>\n\n";

	my ($node2set_sw_cond_probs_ref, $node2set_sw_extcond_probs_ref)
	    = calc_prob_descresstats_cond_ancstat2 (@cp_ref_seg_msa, @{$sequence_ids}, %{$name2id}, %{$node2ch}, %{$br2tr_mtrx}, $GAP);
	unless (defined $node2set_sw_cond_probs_ref) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'calc_prob_descresstats_cond_ancstats2' FAILED on the reference MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $set_clm_probs_ref = calc_clm_probs (@resfreqs_prank, %{$node2set_sw_cond_probs_ref}, $topnode_id);
	unless (defined $set_clm_probs_ref) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'calc_clm_probs' FAILED on the reference MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $loglk_sbst_ref = 0;
	foreach my $clm_prob (@{$set_clm_probs_ref}) { $loglk_sbst_ref += log ($clm_prob); }


	my ($node2set_sw_cond_probs_rec, $node2set_sw_extcond_probs_rec)
	    = calc_prob_descresstats_cond_ancstat2 (@cp_rec_seg_msa, @{$sequence_ids}, %{$name2id}, %{$node2ch}, %{$br2tr_mtrx}, $GAP);
	unless (defined $node2set_sw_cond_probs_rec) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'calc_prob_descresstats_cond_ancstats2' FAILED on the constructed MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $set_clm_probs_rec = calc_clm_probs (@resfreqs_prank, %{$node2set_sw_cond_probs_rec}, $topnode_id);
	unless (defined $set_clm_probs_rec) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'calc_clm_probs' FAILED on the reconstructed MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $loglk_sbst_rec = 0;
	foreach my $clm_prob (@{$set_clm_probs_rec}) { $loglk_sbst_rec += log ($clm_prob); }

	my $diff_loglk_sbst = $loglk_sbst_rec - $loglk_sbst_ref;

	print $fh_log "Log_likelihood(sbst)_rec = $loglk_sbst_rec ,\n";
	print $fh_log "Log_likelihood(sbst)_ref = $loglk_sbst_ref ,\n";
	print $fh_log "(Log_likelihood(sbst)_[rec - ref] = $diff_loglk_sbst .\n\n\n";


		## Finally, record the results of the examination into the output table. ##

	print $fh_out1 "< Segment ID: $segid >\n\n";

	print $fh_out1 join ("\t", 'Rlv_branch', 'len(rlv_branch)', '#{involved_seqs}', 'block_size', '|shift|', '#{sbst_diffs}_exp', '#{sbst_diffs}_obs', 'P-value', 'log[Lk(sbst)]_[rec-ref]'), "\n\n";

	print $fh_out1 join ("\t", $rlv_br_prg, $rlv_brlen, $ct_prg_invlvd_seqs, $size_block, $abs_shift, $d_exp, $d_obs, $pv, $diff_loglk_sbst), "\n";

	print $fh_out1 "\n\\\\\n\n";

			### END of "NEWLY ADDED in creation of 'examine_purges_via_mblks.2.pl'." ###


		## Close Temporary Filehandles, and Remove Unnecessary Temporary Files. ##
		## (MOVED from below with some modifications in creation of 'classify_msa_errors.1.pl'.)
		## ... Needs be Modified...

	close ($fh_log);
	##foreach my $fh_log ($fh_log_ref, $fh_log_rec) { close ($fh_log); }


	if ($FLG_RMV_TMPOUT) {
	    foreach my $grpa_tmp_workdir ($grpa_tmp_workdir_ref, $grpa_tmp_workdir_rec) {
		unless (0 == system ("rm -rf $grpa_tmp_workdir") ) {
		    print STDOUT "[ERROR '$program_name'] 'system ('rm -rf {outdir}')' FAILED to remove the grand-parent of temporary working directories '$grpa_tmp_workdir': $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
	    }
	}

=cut # END of "DISABLED in creation of 'examine_complex_via_mblks.2.pl' (3)."


	print STDOUT "\n --- Finished analysis for the segment $segid ! ---\n\n";

	close ($fh_log); # RESURRECTED as of May 31 (Tue), 2016, in creation of 'examine_complex_via_mblks.2.pl'. #

	$ct_succ_segs++;

}			# End of the main outer for-loop (over MSA segments) #

	# MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl'. (on Oct 6 (Tue), 2015). #
close ($fh_out1);
##close ($fh_out2); # DISABLED in creation of 'examine_purges_via_mblks.2.pl'. #

##close ($fh_out);

print STDOUT "\n\n === All the Analyses FINISHED !! ===\n\n";

print STDOUT "#{Sessions skipped (b/c they contain NO erroneous segment)} = $ct_skipped_segs .\n";
print STDOUT "#{Sessions skipped (b/c they have NO subject segment)} = $ct_skipped_segs2 .\n";
##print STDOUT "#{Segments skipped} = $ct_skipped_segs .\n";
##print STDOUT "#{Segments whose ComplLiMMent calculations skipped} = $ct_skipped_segs2 .\n";

print STDOUT "#{Segments skipped (b/c they are too long)}  = $ct_skipped_segs3 .\n"; # IMPORTED from 'examine_purge_ctrls.2.pl' in creation of 'examine_complex_via_mblks.2.pl' (on May 30 (Mon), 2016). #

print STDOUT "#{Anomalous segments} = $ct_anom_segs .\n";
##print STDOUT "#{Segments used as inputs} = $ct_in_segs .\n";

print STDOUT "#{Sessions successfully processed} = $ct_succ_segs .\n";
print STDOUT "#{Subject Segments successfully processed} = $ct_succ_sbj .\n";
##print STDOUT "#{Segments successfully processed} = $ct_succ_segs .\n";
print STDOUT "#{Segments failed to be processed} = $ct_fail_segs .\n\n";

unless (0 == $ct_fail_segs) {
	print STDOUT "[ERROR '$program_name'] FAILED to process $ct_fail_segs segments !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


	### DEFINITIONS of the subroutines added in creation of 'examine_purges_via_mblks.2.pl'. ###

# bi2v_multifurcated_top:
#	This subroutine converts a bufurcated top node into a "virtually" multifurcated one
#	by sliding the top node to the virtually the same position as that of its child node
#	with the shorter branch length.
#	If a tie situation occurs, the subroutine chooses the branch that comes first.
#	The subroutine was created by modifying 'bi2multifurcated_top' in the package, "MyTreeManipu_IV.pm."
#	[CREATED in creation of 'examine_purges_via_mblks.2.pl' (on May 9 (Mon), 2016).]
#
# bi2multifurcated_top:
#	This subroutine converts a bifurcated top node into a multifurcated one
#	by 'splicing out' the branch leading tothe lower node
#	containing the least number of children among the 'branching' lower nodes.
#	If a tie situation occurs, the subroutine chooses the branch
#	that comes first.
#
# INPUT: (\%node2parent, \%node2children, \%name2id, \%br_attr, \@indices_attr), where
#	@indices_attr = ($indx_blen, $indx_boot, $tot_boot), where
#	$br_attri{$node_id}->[$indx_blen] gives the branch length if $indx_blen >=0,
#	$br_attri{$node_id}->[$indx_boot] gives the bootstrap value if $indx_boot >=0.
#	$ct_boot gives the total count of bootstrap replicates.
#	It will take the summation of the lengths of the two branches,
#	and take the product of their bootstrap values, which is afterwards
#	divided by $tot_boot (>0).
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub bi2v_multifurcated_top (\%\%\%\%\@) {

	my ($node2parent, $node2children, $name2id, $br_attr, $indices_attr) = @_;

	my $top_id = $name2id->{$TOP_NODE};
	unless (defined $top_id) {
	    print STDERR "[CAUTION 'bi2v_multifurcated_top']  NO ID for the TOP NODE.\n"; 
	    return 0;
	}
	my $top_children = $node2children->{$top_id};

	unless (defined $top_children) {
	    print STDERR "[CAUTION 'bi2v_multifurcated_top'] NO CHILDREN for the TOP NODE (ID: $top_id).\n";
	    return 0;
	}

		# MODIFIED in creation of 'bi2v_multifurcated_top'. #

	unless (@{$top_children} == 2) {
	    print STDERR "[CAUTION 'bi2v_multifurcated_top'] The TOP NODE (ID: $top_id) is NOT BIFURCATED.\n";
	    return 0;
	}
=For the predecessor 'bi2multifurcated_top'.
		# It's already multifurcated,
		# so there is no need to change it.
	(@{$top_children}>2) and (return 1) ;
=cut

	my $target; # The child whose upper branch are to be shrinked
	my $minct_gkids = scalar keys %{$name2id};
	foreach my $child (@{$top_children}) {
	    my $gkids = $node2children->{$child};
	    (defined  $gkids) or next; # skip an external node
	    my $ct_gkids = @{$gkids};
	    ($ct_gkids >1)  or next; # skip a non-branching node
	    ($ct_gkids < $minct_gkids) or next; 
	    $target = $child;
	    $minct_gkids = $ct_gkids;
	}

		# It cannot be converted to a multifurcated tree.
	(defined $target) or (return 1); 

		# Convert the attributes #

	my ($indx_blen, $indx_boot, $tot_boot) = @{$indices_attr};

	my $partner ;
	foreach my $child (@{$top_children}) {
	    ($child == $target) or ($partner = $child);
	 }
	my $attr_tgt = $br_attr->{$target};
	my $attr_pt = (defined $partner) ? $br_attr->{$partner} : undef;

	if ( (defined $attr_tgt) and (defined $attr_pt) ) {
	    if ($indx_blen >=0) {
		my $tgt_blen = $attr_tgt->[$indx_blen];
		my $pt_blen = $attr_pt->[$indx_blen];
		if (defined $tgt_blen) {
		    if (defined $pt_blen) {
			$attr_pt->[$indx_blen] += $tgt_blen;
		    } else {
			# evacuate the target branch length #
			$attr_pt->[$indx_blen] = $tgt_blen;
		    }
		}
			# Set the branch length leading to the target node to be ZERO.
			# (MAJOR MODIFICATION in creation of 'bi2v_multifurcated_top'.)
		$attr_tgt->[$indx_blen] = 0; 
	    }

=For the predecessor 'bi2multifurcated_top'.
	    if (($indx_boot>=0) and ($tot_boot>1) ) {
		my $tgt_boot = $attr_tgt->[$indx_boot];
		my $pt_boot = $attr_pt->[$indx_boot];

			### For debugging ###
		### print STDERR "\nTot_boot = $tot_boot, TGT_boot = $tgt_boot, Pt_boot = $pt_boot.\n\n";
		if (defined $tgt_boot) {
		   if (defined $pt_boot) {
			my $cmp_boot = int ($tgt_boot * $pt_boot /$tot_boot);

			### For debugging ###
			### print STDERR "\nCMP_boot = $cmp_boot .\n\n"; 

			$attr_pt->[$indx_boot] = $cmp_boot;
		   } else {
			# evacuate the target bootstrap value #
			$attr_pt->[$indx_boot] = $tgt_boot;
		   }
		}
	    }

		## ADDED to avoid a confusion (on May 21 (Fri), 2010) ##
	    delete $br_attr->{$target};
=cut

		## ADDED to cope with a special situation (on May 21 (Fri), 2010) ##

	} elsif ( (defined $attr_tgt) and !(defined $attr_pt) ) {

		## Transfer all the attributes on the target branch to the partner branch. ##
		## (AND set the target branch length to ZERO (...ADDED in creation of 'bi2v_multifurcated_top'.) ##
	    my @cp_attr_tgt = @{$attr_tgt};
	    $br_attr->{$partner} = \@cp_attr_tgt;
	    $attr_tgt->[$indx_blen] = 0;

=For the predecessor 'bi2multifurcated_top'.
	    $br_attr->{$partner} = $attr_tgt;
	    delete $br_attr->{$target};
=cut
	}

=For the predecessor 'bi2multifurcated_top'.
		# Convert the topology #
	my $fl_sub = br_contract_branch ($target, %{$node2parent}, %{$node2children});

	unless ($fl_sub) {
	    print STDERR "[CAUTION 'bi2v_multifurcated_top'] The subroutine 'br_contract_branch' FAILED for the child '$target' of the top node '$top_id'.\n";
	    return 0;
	}
=cut

	return 1;

}		# End of the subroutine 'bi2v_multifurcated_top' #


		### DEFINITIONS of the subroutines ADDED in creation of 'classify_msa_errors_via_mblks.2.pl'. ###

# [ summarize_info_cblocks: ] --- An ERROR fixed on Oct 23 (Fri), 2015.
# summarize_info_cblock:
#	This subroutine creates some variables summarizing a subject composite-block,
#	out of some detailed information on it.
#	[CREATED on Oct 22 (Thu), 2015.]
#
# INPUT: ($indx_cb, \@set_composite_blocks, \@set_blocks_shift_respos, \@hard_linking_cbs, \%node2ch, \%node2seq_indx), where
#
#	$indx_cb is the index of the subject c-block (in @set_composite_blocks);
#
#	@{$set_composite_blocks[$cb]} = ($beg_cb, $end_cb, $nodeid_mrca_cb, \%br2change_cb, \@constituent_blks, \@invlvd_seqs_cb, \@merger_types) is for the $cb th composite block, with
#	@constituent_blks storing the indices of the blocks that constitutes this c-block;
#
#	@{$set_blocks_shift_respos[$bl]} = ($shift, $beg_bl, $end_bl, $nodeid_mrca_bl, $depth_mrca_bl, \@seqs_invlvd_bl, \%br2change, \@excl_sbblocks, $ct_changes_bl, \@indices_src_classes) is for the $bl th residue-position-shift block, with
#	@seqs_invlvd_bl storing the indices of the sequences involved, and
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) being for the $sb th sub-block excluded from this block;
#
#	@hard_linking_cbs stores the indices of the c-blocks that hard-llinks the $indx_cb th c-block.
#
# OUTPUT: ($beg_cb, $end_cb, $shift_le, $shift_re, \@set_invlvd_seqs_le, \@set_invlvd_seqs_re), where
#
#	$beg_cb and $end_cb are the horizontal positions of the beginning and ending of the subject c-block;
#	$shift_le/re is the shift of the left-/right-end of the subject c-block;
#
#	@set_invlvd_seqs_le/re stores the indices of the sequences involved with the c-block at its left-/right-end.
#
##sub summarize_info_cblocks ($\@\@\@\%\%) {...} # A BUG fixed on Oct 23 (Fri), 2015. #
sub summarize_info_cblock ($\@\@\@\%\%) {

	my ($indx_cb, $set_composite_blocks, $set_blocks_shift_respos, $hard_linking_cbs, $node2ch, $node2seq_indx) = @_;

	my ($beg_cb, $end_cb, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$set_composite_blocks->[$indx_cb]};

	my ($blk_le, $blk_re) = @{$constituent_blks}[0, $#{$constituent_blks}];

	my ($shift_le, $invlvd_seqs_le, $excl_sbbl_cands_le) = @{$set_blocks_shift_respos->[$blk_le]}[0, 5, 7];
	my ($shift_re, $invlvd_seqs_re, $excl_sbbl_cands_re) = @{$set_blocks_shift_respos->[$blk_re]}[0, 5, 7];

	my @setset_invlvd_seqs = (); # = (\@set_invlvd_seqs_le, \@set_invlvd_seqs_re);

	foreach my $info ([$beg_cb, $invlvd_seqs_le, $excl_sbbl_cands_le],
			[$end_cb, $invlvd_seqs_re, $excl_sbbl_cands_le]) { # Main outer foreach-loop (for left-end and right-end of the subject c-block). #

	    my ($pos, $set_invlvd_seqs, $excl_sbbl_cands) = @{$info};

			# Fill in the set of excluded sequences. #

	    my @set_invlvd_seqs = @{$set_invlvd_seqs};
	    my @excl_seqs = ();

	    foreach my $sbbl (@{$excl_sbbl_cands}) { # 1st middle foreach-loop (over excluded sub-blocks). #
		my ($beg_sbbl, $end_sbbl, $mrca_sbbl) = @{$sbbl};
		(($beg_sbbl <= $pos) and ($pos <= $end_sbbl)) or next; # Skip if the sub-block does not encompass the subject position. #

		my $exoffs = fetch_ext_offsprings_nr ($mrca_sbbl, %{$node2ch});
		foreach my $nodeid (@{$exoffs}) { push @excl_seqs, $node2seq_indx->{$nodeid}; }

	    }				# End of the 1st middle foreach-loop (over excluded sub-blocks). #

			# Examine the hard-linking blocks. #

	    foreach my $hl_cb (@{$hard_linking_cbs}) { # 2nd middle foreach-loop (over hard-linking c-blocks). #

		my ($beg_hl, $end_hl, $constituents_hl) = @{$set_composite_blocks->[$hl_cb]}[0,1,4];
		(($beg_hl <= $pos) and ($pos <= $end_hl)) or next; # Skip if the c-block does not encompass the subject position. #

		foreach my $hl_bl (@{$constituents_hl}) { # Inner foreach-loop (over constituent blocks of $hl_cb). #

		    my ($shift_hl, $beg_hl, $end_hl, $invlvd_seqs_hl, $excl_sbbl_cands_hl) = @{$set_blocks_shift_respos->[$hl_bl]}[0, 1, 2, 5, 7];
		    (($beg_hl <= $pos) and ($pos <= $end_hl)) or next; # Skip if the block does not encompoass the subject position. #

		    push @set_invlvd_seqs, @{$invlvd_seqs_hl};

		    foreach my $sbbl (@{$excl_sbbl_cands_hl}) { # Inner-inner foreach-loop (over excluded sub-blocks of $hl_bl). #
			my ($beg_sbbl, $end_sbbl, $mrca_sbbl) = @{$sbbl};
			(($beg_sbbl <= $pos) and ($pos <= $end_sbbl)) or next; # Skip if the sub-block does not encompass the subject position. #

			my $exoffs = fetch_ext_offsprings_nr ($mrca_sbbl, %{$node2ch});
			foreach my $nodeid (@{$exoffs}) { push @excl_seqs, $node2seq_indx->{$nodeid}; }

		    }				# End of the inner-inner foreach-loop (over excluded sub-blocks of $hl_bl). #
		    
		}			# End of the inner foreach-loop (over constituent blocks of $hl_cb). #

	    }				# End of the 2nd middle foreach-loop (over hard-linking c-blocks). #

	    my $nr_invlvd_seqs = (@set_invlvd_seqs>1) ? nr_nmarray (@set_invlvd_seqs) : \@set_invlvd_seqs ;
	    my $nr_excl_seqs = (@excl_seqs>1) ? nr_nmarray (@excl_seqs) : \@excl_seqs ;

	    my ($truly_excl_seqs, $truly_invlvd_seqs, $extra_excl_seqs)
		= (@excl_seqs>0) ? diff_nmarray (@{$nr_invlvd_seqs}, @{$nr_excl_seqs}) : ([], $nr_invlvd_seqs, []) ;

	    push @setset_invlvd_seqs, $truly_invlvd_seqs;

	}				# END of the main outer foreach-loop (for left-end and right-end of the subject c-block). #


	my ($set_invlvd_seqs_le, $set_invlvd_seqs_re) = @setset_invlvd_seqs;

	return ($beg_cb, $end_cb, $shift_le, $shift_re, $set_invlvd_seqs_le, $set_invlvd_seqs_re);

}	# END of the subroutine 'summarize_info_cblock'. #


# pr_map_shifts_respos:
#	This subroutine prints out a position-shift map
#	by formatting it so that the map will be decently visible.
#	[CREATED on Oct 15 (Thu), 2015.]
#
# INPUTS: (\*${$fh}, \@sequence_ids, \@map_shifts_respos, $char_gap, $dis_len_shift), where
#	$fh is the reference to a file-handle;
#	$sequence_ids[$k] is the ID of the sequence in the $k th row of the MSAs;
#	$map_shifts_respos[$k]->[$s] is the position-shift at the intersection of the $k th row and the $s th column of the mapped MSA;
#	$char_gap is the character representing a gap;
#	$dis_len_shift is the maximum length of the run of shifts to be displayed in a line.
#
# OUTPUTS: 1 if the subroutine succeeded,
#	   0 if it failed.
#
sub pr_map_shifts_respos (\*\@\@$$) {

	my ($fh, $sequence_ids, $map_shifts_respos, $char_gap, $dis_len_shift) = @_;

	my $ct_sequeunces = @{$sequence_ids};
	unless ($ct_sequences>0) {
	    print STDOUT "[INFORMATION 'pr_map_shifts_respos'] NO SEQUENCES in the alignment !\n\n";
	    return 1;
	}
	unless (@{$map_shifts_respos} == $ct_sequences) {
	    print STDOUT "[ERROR 'pr_map_shifts_respos'] The number of sequences in the map (=", scalar (@{$map_shifts_respos}), ") CLASHES with that in the set of sequences (=$ct_sequences): $! !\n\n";
	    return 0;
	}

	my $map_len = @{$map_shifts_respos->[0]}; 

	my $quotient = int (($map_len -1) / $dis_len_shift);

	for (my $r=0; $r<=$quotient; $r++) {  # 1st outer for-loop (over segmented displayes). #

	    my $beg_dis = $r * $dis_len_shift;
	    my $end_dis = ($r == $quotient) ? $map_len - 1 : ($r+1) * $dis_len_shift - 1 ;
	    my $format = "%-10s\t" . "%5d" x (($r==$quotient) ? $map_len - $quotient * $dis_len_shift : $dis_len_shift) . "\n";

	    printf $fh $format, 'Seq_ID', ($beg_dis .. $end_dis);
	    print $fh "\n";

	    for (my $k=0; $k<$ct_sequences; $k++) {  # Middle for-loop (over sequences). #

		my $out_string = sprintf ("%-10s\t", $sequence_ids->[$k]);

		my $row_shifts = $map_shifts_respos->[$k];
		my $ct_sites = @{$row_shifts};

		my $end_dis2 = ($ct_sites-1 < $end_dis) ? $ct_sites-1 : $end_dis;
		for (my $s=$beg_dis; $s<= $end_dis2; $s++) { # Inner testing for-loop (over sequences). #
		   my $shift = $row_shifts->[$s];
		   $out_string .= ($shift eq $char_gap) ? sprintf ("%5s", "'$char_gap'") : sprintf ("%5d", $shift) ;
		}				# End of the inner testing for-loop (over sequences). #

		print $fh $out_string, "\n";
	    
	    }			# End of the middle for-loop (over sequences). #
	    print $fh "\n\n\n";

	}			# End of the 1st outer for-loop (over segmented displayes). #

	return 1;

}		# End of the subroutine 'pr_map_shifts_respos'. #


# blockwise_framework:
#	This satellite subroutine of 'framework_cblock' creates
#	the framework of an input position-shift block.
#	[CREATED on Oct 12 (Mon), 2015.]
#
# INPUT: ($beg_bl, $end_bl, \@seqs_involved, \@excl_sbblockks, \%node2ch, \%node2seq_indx), where
#
#	$beg_bl and $end_bl are the beginning- and ending-positions of the input block;
#	@seqs_involved stores the indices of the sequences involved with the input block;
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block;
#	%node2ch = ($node_id => \@child_nodes, ...) for internal nodes;
#	%node2seq_indx = ($node_id => $indx_seq, ...) for external nodes;
#
# OUTPUT: (\%frwk_bl), where
#	$frwk_bl{$indx_seq} = [$beg_frwk, $end_frwk], which determines the beginning- and ending-positions of the framework of the input block in the $indx_seq th sequence (row),
#	 	     = undef if the input block does not involve the $indx_seq th sequence. 
#
sub blockwise_framework ($$\@\@\%\%) { 

	my ($beg_bl, $end_bl, $seqs_involved, $excl_sbblocks, $node2ch, $node2seq_indx) = @_;

	my %frwk_bl;

	foreach my $indx_seq (@{$seqs_involved}) { $frwk_bl{$indx_seq} = [$beg_bl, $end_bl]; }

	foreach my $info_sbbl (@{$excl_sbblocks}) { # 2nd foreach-loop (over excluded sub-blocks). #

	    my ($beg_sbbl, $end_sbbl, $mrca_sbbl) = @{$info_sbbl};

	    (($beg_bl < $beg_sbbl) and ($end_sbbl < $end_bl)) and next; # Skip if the sub-block is completely within the block, because the framework itself will not be influenced by such a sub-block. #

	    my $exoffs = fetch_ext_offsprings_nr ($mrca_sbbl, %{$node2ch}) ;

	    foreach my $nodeid_exoff (@{$exoffs}) { # Inner foreach-loop (over external offsprings). #
		my $seqid_exoff = $node2seq_indx->{$nodeid_exoff};
		my $frwk_exoff = $frwk_bl{$seqid_exoff};

		(defined $frwk_exoff) or next; # Skip if the sequence is not involved with the block. #

		if ($beg_sbbl <= $beg_bl) {
		    if ($end_bl <= $end_sbbl) {
			delete $frwk_bl{$seqid_exoff}; # In this case, the sequence is completely excluded.
		    } else {
			$frwk_exoff->[0] = $end_sbbl + 1;
		    }
		} elsif ($end_bl <= $end_sbbl) {
		    $frwk_exoff->[1] = $beg_sbbl-1;
		}
	    }					# End of the inner foreach-loop (over external offsprings). #

	}					# End of the 2nd foreach-loop (over excluded sub-blocks). #

	return \%frwk_bl;

}	# End of the subroutine 'blockwise_framework'. #


# framework_cblock:
#	This subroutine creates the framework of the subject composite-block.
#	[CREATION STARTED on Oct 11 (Sun), 2015, ENDED on Oct 12 (Mon), 2015.]
#
# INPUT: ($ct_seqs, $indx_cb, \@set_composite_blocks, \@hard_linking, \@set_blocks_shift_respos, \%node2ch, \%node2seq_indx), where
#
#	$indx_cb is the index of the subject c-block;
#	@{$set_composite_blocks[$cb]} = ($beg_cb, $end_cb, $nodeid_mrca_cb, \%br2change_cb, \@constituent_blks, \@involved_seqs_cb, \@merger_types) is for the $cb th composite block. Here
#	@constituent_blks stores indices of the position-shift blocks that constitute the composite block;
#	
#	@hard_linking stores indices of the c-blocks that hard links $indx_cb;
#
#	@{$set_blocks_shift_respos[$bl]} = ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes) is for the $bl th position-shift block. Here
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block;
#
#	%node2ch = ($node_id => \@child_nodes, ...) for internal nodes;
#	%node2seq_indx = ($node_id => $indx_seq, ...) for external nodes;
#
# OUTPUT: (\@frwk_cb, \@info_constituents, \@blwise_hard_linking), where
#
#	$frwk_cb[$i]->{$indx_seq} = [$beg_frwk, $end_frwk], which determines the beginning- and ending-positions of the framework of the $i th constituent block in the $indx_seq th sequence (row),
#			 = undef if the $i th constituent block does not involve the $indx_seq th sequence.
#
#	@{$info_constituents[$i]} = ($shift, $beg_bl, $end_bl) is for the $i th constituent block of the $indx_cb th c-block;
#
#	@{$blwise_hard_linking[$i]} is an array, each of whose elements is [$indx_cb2, $bl2] for the block that is hard_linking the $i th constituent block.
#
sub framework_cblock ($$\@\@\@\%\%) {

	my ($ct_seqs, $indx_cb, $set_composite_blocks, $hard_linking, $set_blocks_shift_respos, $node2ch, $node2seq_indx) = @_;

	my ($beg_cb, $end_cb, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$set_composite_blocks->[$indx_cb]}; 

	my $ct_constituents = @{$constituent_blks};

		# First, construct the framework of the main c-block. #

	my @info_constituents = (); # element = [$shift, $beg_bl, $end_bl] for the constituent block. #
	my @frwk_cb = ();

	for (my $i=0; $i<$ct_constituents; $i++) { # 1st main outer for-loop (over constituent blocks). #
	    my $bl = $constituent_blks->[$i];

	    my ($shift, $beg_bl, $end_bl, $nodeid_mrca_bl, $depth_mrca_bl, $seqs_involved, $br2change_bl, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_blocks_shift_respos->[$bl]};
	    (defined $excl_sbblocks) or ($excl_sbblocks = []);

	    $info_constituents[$i] = [$shift, $beg_bl, $end_bl];

	    my $frwk_bl = blockwise_framework ($beg_bl, $end_bl, @{$seqs_involved}, @{$excl_sbblocks}, %{$node2ch}, %{$node2seq_indx});
	    unless (defined $frwk_bl) {
		print STDOUT "[ERROR 'framework_cblock'] The subroutine 'blockwise_framework' FAILED on the $i th constituent block (Index: $bl) of the input c-block: $! !\n\n";
		return (undef, undef, undef);
	    }

	    $frwk_cb[$i] = $frwk_bl;

	}				# End of the 1st main outer for-loop (over constituent blocks). #


		# Second, augument the framework with those of the hard-linking c-blocks. #

	my @blwise_hard_linking = (); # @{$blwise_hard_linking[$i]} is an array of [$indx_cb2, $bl2] for the blocks that hard-link the $i th constituent block. #
	for (1 .. $ct_constituents) { push @blwise_hard_linking, []; }

	my $ct_hard_linking = @{$hard_linking};

	for (my $i2 = 0; $i2 <$ct_hard_linking; $i2++) { # 2nd main outer foreach-loop (over c-blocks hard-linking the input c-block). #

	    my $indx_cb2 = $hard_linking->[$i2];

	    my ($beg_cb2, $end_cb2, $nodeid_mrca_cb2, $br2change_cb2, $constituent_blks2, $involved_seqs_cb2, $merger_types2) = @{$set_composite_blocks->[$indx_cb2]}; 

	    foreach my $bl2 (@{$constituent_blks2}) { # Middle foreach-loop (over constituent blocks of hard-linking). #

		my ($shift2, $beg_bl2, $end_bl2, $nodeid_mrca_bl2, $depth_mrca_bl2, $seqs_involved2, $br2change_bl2, $excl_sbblocks2, $ct_changes2, $indices_src_classes2) = @{$set_blocks_shift_respos->[$bl2]};

		my $indx_accom;
		for (my $i=0; $i<$ct_constituents; $i++) { # 1st inner foreach-loop (over constituent blocks of input). #

		    my ($shift, $beg_bl, $end_bl) = @{$info_constituents[$i]};

		    ($shift == $shift2) or next;
		    ##($shift and $shift2) or next; # A BUG fixed on Oct 24 (Sat), 2015. #

		    (($beg_bl <= $beg_bl2) and ($end_bl2 <= $end_bl)) or next;
		    $indx_accom = $i;
		    last;
		}				# End of the 1st inner foreach-loop (over constituent blocks of input). #

		unless (defined $indx_accom) { # There is no input block accommodating the hard-likning block... This must not happen!! #
		    print STDOUT "[ERROR 'framework_cblock'] The $bl2 th block of the $indx_cb2 th c-block, which should hard-link the input c-block ($indx_cb), is actually NOT accommodated by any block of the input c-block: $! !\n\n";
		    return (undef, undef, undef);
		}

		push @{$blwise_hard_linking[$indx_accom]}, [$indx_cb2, $bl2];

		my $frwk_bl2 = blockwise_framework ($beg_bl2, $end_bl2, @{$seqs_involved2}, @{$excl_sbblocks2}, %{$node2ch}, %{$node2seq_indx});
		unless (defined $frwk_bl2) {
		   print STDOUT "[ERROR 'framework_cblock'] The subroutine 'blockwise_framework' FAILED on the $bl2 th block, which is a constituent of the hard-linking block (index: $indx_cb2): $! !\n\n";
		   return (undef, undef, undef);
		}

		my $frwk_bl = $frwk_cb[$indx_accom];

		foreach my $indx_seq (keys %{$frwk_bl2}) { $frwk_bl->{$indx_seq} = $frwk_bl2->{$indx_seq}; }

	    }				# End of the middle foreach-loop (over constituent blocks of hard-linking). #

	}			# End of the 2nd main outer foreach-loop (over c-blocks hard-linking the input c-block). #

	return (\@frwk_cb, \@info_constituents, \@blwise_hard_linking);

}	# END of the subroutine 'framework_cblock'. #

# scout_possible_undoing:
#	This satellite subroutine of 'undo_mv_latter_of_cblock_pair' examines
#	the extent to which the "undoing" of the specified block can be performed.
#	[CREATED on Oct 12 (Mon),2015.]
#	[MODIFIED on Oct 13 (Tue), 2015.] The subroutine now outputs \@indices_seqs_incomplete_mvs as well.
#	[DEBUGGED on Oct 16 (Fri), 2015.]
#
# INPUT: ($shift_bl, \%frwk_bl, \@map_shifts_respos, $char_abs), where
#	$shift_bl is the position-shift of the specified block;
#	%frwk_bl = ($indx_seq => [$beg_fw, $end_fw], ...) gives the beginning- and ending-positions of the framework in the $indx_seq th sequence;
#	@map_shifts_respos gives the position-shift map immediately before the move is undone. More specifically,
#	$map_shifts_respos[$indx_seq]->[$s]  gives the position-shift of the site with the horozontal position = $s in the $indx_seq th row of the MSA;
#
# OUTPUT: ($actual_shift_bl, \@indices_seqs_incomplete_mvs), where
#	$actual_shift_bl is the shift that can be actually undone;
#	@indices_seqs_incomplete_mvs stores indices of the sequences that cannot be undone the moves completely.
#
sub scout_possible_undoing ($\%\@$) {

	my ($shift_bl, $frwk_bl, $map_shifts_respos, $char_abs) = @_;

	my $actual_shift_bl = $shift_bl;
	my @indices_seqs_incomplete_mvs = ();

	foreach my $indx_seq (sort {$a <=> $b} keys %{$frwk_bl}) { # Main foreach-loop (over sequences involved with the block). #

	    my ($beg_fw, $end_fw) = @{$frwk_bl->{$indx_seq}};
	    my $shifts_in_seq = $map_shifts_respos->[$indx_seq];

		    # MODIFIED for DEBUGGING (on Oct 16 (Fri), 2015) #
	    my ($shift_beg, $shift_end) = @{$shifts_in_seq}[$beg_fw, $end_fw];
	    unless (($shift_beg ne $char_abs) and ($shift_bl == $shift_beg) and ($shift_end ne $char_abs) and ($shift_bl == $shift_end)) {
	    ##unless (($shift_bl eq $shifts_in_seq->[$beg_fw]) and ($shift_bl eq $shifts_in_seq->[$end_fw])) {
		print STDOUT "[ERROR 'scout_possible_undoing'] NOT BOTH of the positions $beg_fw and $end_fw in the $indx_seq th sequence have the position-shift = $shift_bl: $! !\n\n";
		return undef;
	    }
	    ## ($actual_shift_bl == 0) and next; # The actual shift cannot change any more. ## DISABLED on Oct 16 (Fri), 2015. #

	    my $shift2 = 0;

	    if ($shift_bl>0) { # The block moves leftward if undone.

		for (my $s = $beg_fw-1; $s>= $beg_fw - $shift_bl; $s--) { # 1st inner for-loop (downwards over sites in the desired range). #
		##for (my $s = $beg_fw-1; $s>= $beg_fw-$actual_shift_bl; $s--) {...} # 1st inner for-loop (downwards over sites in the possible range). # MODIFIED on Oct 13 (Tue), 2015. #

		    (($s < 0) or ($shifts_in_seq->[$s] ne $char_abs)) and last; # Found an obstacle (or the left-end)!! 
		    ($shifts_in_seq->[$s] eq $char_abs) or last; # Another possible BUG fixed on Oct 16 (Fri), 2015. #
		    ##($shifts_in_seq->[$s] eq $char_abs) and last; # A BUG fixed on Oct 16 (Fri), 2015. #
		    $shift2++;
		}				# End of the 1st inner for-loop (downwards over sites in the possible range). #
		($actual_shift_bl > $shift2) and ($actual_shift_bl = $shift2);

	    } else { # The block moves rightward if undone. #

		for (my $s = $end_fw+1; $s<= $end_fw - $shift_bl; $s++) { # 2nd inner for-loop (upwards over sites in the desired range). #
		##for (my $s = $end_fw+1; $s<= $end_fw + $shift_bl; $s++) {...} # 2nd inner for-loop (upwards over sites in the desired range). # Another BUG fixed on Oct 16 (Fri), 2015. #
		##for (my $s = $end_fw+1; $s<= $end_fw+$actual_shift_bl; $s++) { # 2nd inner for-loop (upwards over sites in the possible range). # MODIFIED on Oct 13 (Tue), 2015. #

		    ((defined $shifts_in_seq->[$s]) and ($shifts_in_seq->[$s] ne $char_abs)) and last; # Found an obstacle!!
		    ##($shifts_in_seq->[$s] eq $char_abs) or last; # Another BUG fixed on Oct 16 (Fri), 2015. #
		    ##($shifts_in_seq->[$s] eq $char_abs) and last; # A BUG fixed on Oct 16 (Fri), 2015. #
		    $shift2--;
		}				# End of the 1st inner for-loop (upwards over sites in the possible range). #
		($actual_shift_bl < $shift2) and ($actual_shift_bl = $shift2);
	    }

	    ($shift2 == $shift_bl) or (push @indices_seqs_incomplete_mvs, $indx_seq);

	}			# End of the main foreach-loop (over sequences involved with the block). #

		# MODIFIED on Oct 13 (Tue), 2015. #
	return ($actual_shift_bl, \@indices_seqs_incomplete_mvs);
	##return $actual_shift_bl;

}	# End of the subroutine 'scout_possible_undoing'. #


# undo_sgl_bw_fw_move_on_2maps:
#	This satellite subroutine of 'undo_mv_latter_of_cblock_pair'
#	undoes the move of the framework of a single (elementary but not composite) position-shift block.
#	The results will be reflected on two position-shift maps,
#	one between the reference and the intermediate MSAs (on the latter),
#	and the other between the intermediate and the reconstructed MSAs (on the latter).
#
#	This subroutine allows for up to one excluded region (per sequence) within the framework.
#
#	[CREATED on Oct 13 (Tue), 2015.]
#
#	[MODOFIED on Oct 14 (Wed), 2015]
#	When the undone move extends the intermediate MSA to the right,
#	the subroutine now fills in the extended part of the intermediate MSA with gaps
#	if the sequence is not involved in the undone move.
#
# INPUT: (\%frwk_bl, $char_abs, $full_shift, $actual_shift, \@map_shifts_respos_intm, \@map_shifts_respos_rec_af), where
#
#	%frwk_bl = ($indx_seq (in MSA) => [$beg_fw, $end_fw], ...) defines the framework of the subjecet block (in the intermediate MSA before the move of this block is undone);
#
#	$char_abs is the character representing the absence of the residue. (It is the same as the 'gap' character);
#	$full_shift is the full value of the shift of the subject block between the reference and the reconstructed MSA;
#	$actual_shift is the value of the shift by which the move of the subject block is actually undone (from the reconstructed MSA to the intermediate MSA);
#
#	@map_shifts_respos_intm is the position-shift map between the reference and the intermediate MSAs. It is under construction.
#	@map_shifts_respos_rec_af is the position-shift map between the intermediate and the reconstructed MSAs. It is also under construction.
#	($map_shifts_respos_xxx[$indx_seq]->[$s] is the position-shift at the site of the $indx_seq th sequence with horozontal position = $s.);
#
# OUTPUT: $if_success, where
#	$if_success = 1 if the subroutine ended successfully,
#		    = 0 if it failed.
#
sub undo_sgl_bw_fw_move_on_2maps (\%$$$\@\@) {

	my ($frwk_bl, $char_abs, $full_shift, $actual_shift, $map_shifts_respos_intm, $map_shifts_respos_rec_af) = @_;

	my $remaining_shift = $full_shift - $actual_shift;

	my $end_msa_intm = my $end_msa_intm0 = @{$map_shifts_respos_intm->[0]} -1; # ADDED on Oct 14 (Wed), 2015. #

	foreach my $indx_seq (keys %{$frwk_bl}) { # Main foreach-loop (over sequences involved with the subject block). #

		my ($beg_fw_outer, $end_fw_outer) = @{$frwk_bl->{$indx_seq}};
		my $shifts_in_seq_intm   = $map_shifts_respos_intm->[$indx_seq];
		my $shifts_in_seq_rec_af = $map_shifts_respos_rec_af->[$indx_seq];

		($end_msa_intm < $end_fw_outer - $actual_shift) and ($end_msa_intm = $end_fw_outer - $actual_shift); # ADDED on Oct 14 (Wed), 2015. #

			# Determine the boundaries of the "inner" framework, if at all. #

		my $beg_fw_inner = $beg_fw_outer;
		for (my $s=$beg_fw_outer+1; $s<=$end_fw_outer; $s++) { # 1st inner for-loop (over sites within the framework). #

		    my $shift = $shifts_in_seq_intm->[$s];
		    ##my $shift = $shifts_in_seq_rec_af->[$s]; # A BUG fixed on Oct 16 (Fri), 2015. #

		    (($shift eq $char_abs) or ($shift != $full_shift)) and last;
		    $beg_fw_inner++;
		}				# End of the 1st inner for-loop (over sites within the framework). #

		my $end_fw_inner;
		if ($beg_fw_inner < $end_fw_outer) {
		    $end_fw_inner = $end_fw_outer;
		    for (my $s = $end_fw_outer-1; $s>$beg_fw_inner; $s--) { # 2nd inner for-loop (over sites within the framework minus the shell at the beginning). #

			my $shift = $shifts_in_seq_intm->[$s];
			##my $shift = $shifts_in_seq_rec_af->[$s]; # A BUG fixed on Oct 16 (Fri), 2015. #

			(($shift eq $char_abs) or ($shift != $full_shift)) and last;
			$end_fw_inner--;
		    }				# End of the 2nd inner for-loop (over sites within the framework minus the shell at the beginning). #
		}

			# Extert the actual change. #

		for (my $s = $beg_fw_outer; $s <= $beg_fw_inner; $s++) { # 3rd inner for-loop (over sites within the shell at the beginning). #
			$shifts_in_seq_intm->[$s-$actual_shift] = $remaining_shift;
			$shifts_in_seq_rec_af->[$s] = $actual_shift;

		}				# End of the 3rd inner for-loop (over sites within the shell at the beginning). #

		my @sites_new_gaps = ($actual_shift>0) ? ($beg_fw_inner - $actual_shift+1 .. $beg_fw_inner) : ($beg_fw_outer .. $beg_fw_outer - $actual_shift - 1);

		for my $s (@sites_new_gaps) { $shifts_in_seq_intm->[$s] = $char_abs; }

		if (defined $end_fw_inner) {
		    for (my $s = $end_fw_inner; $s <= $end_fw_outer; $s++) { # 6th inner for-loop (over sites within the shell at the end). #
			$shifts_in_seq_intm->[$s - $actual_shift] = $remaining_shift;
			$shifts_in_seq_rec_af->[$s] = $actual_shift;
		    }				# End of the 6th inner for-loop (over sites within the shell at the end). #

		    my @sites_new_gaps2 = ($actual_shift>0) ? ($end_fw_outer -$actual_shift+1 .. $end_fw_outer) : ($end_fw_inner .. $end_fw_inner - $actual_shift -1);
		    for my $s (@sites_new_gaps2) { $shifts_in_seq_intm->[$s] = $char_abs; }
		}
		
	}			# End of the main foreach-loop (over sequences involved with the subject block). #

	if ($end_msa_intm > $end_msa_intm0) { # ADDED on Oct 14 (Wed), 2015. #

			# Fill in the new yet unoccupied right-end with gaps. #

	    foreach my $shifts_row (@{$map_shifts_respos_intm}) { # Trailing foreach-loop (over sequences in the MSA). #
		my $ct_sites = @{$shifts_row};
		for (my $s = $ct_sites; $s <= $end_msa_intm; $s++) { $shifts_row->[$s] = $char_abs; }
	    }					# End of the trailing foreach-loop (over sequences in the MSA). #
	}

	return 1;

}		# End of the subroutine 'undo_sgl_bw_move_on_2maps'. #


# undo_sgl_bw_fw_move_on_2maps_2:
#	This subroutine is a modified version of 'undo_sgl_bw_fw_move_on_2maps'.
#	The major difference from the old version is in how to handle "inpurities", i.e., gaps and/or sites with different values of the shift.
#	Whereas the old version attemped to keep the inpurities intact (unmoved),
#	the new version moves them along with the framework, by changing the values of the shift by the same amount (= $actual_shift).
#	[CREATED in DEBUGGING on Oct 25 (Sun), 2015.]
#
# undo_sgl_bw_fw_move_on_2maps:
#	This satellite subroutine of 'undo_mv_latter_of_cblock_pair'
#	undoes the move of the framework of a single (elementary but not composite) position-shift block.
#	The results will be reflected on two position-shift maps,
#	one between the reference and the intermediate MSAs (on the latter),
#	and the other between the intermediate and the reconstructed MSAs (on the latter).
#
#	This subroutine allows for up to one excluded region (per sequence) within the framework.
#
#	[CREATED on Oct 13 (Tue), 2015.]
#
#	[MODOFIED on Oct 14 (Wed), 2015]
#	When the undone move extends the intermediate MSA to the right,
#	the subroutine now fills in the extended part of the intermediate MSA with gaps
#	if the sequence is not involved in the undone move.
#
# INPUT: (\%frwk_bl, $char_abs, $full_shift, $actual_shift, \@map_shifts_respos_intm, \@map_shifts_respos_rec_af), where
#
#	%frwk_bl = ($indx_seq (in MSA) => [$beg_fw, $end_fw], ...) defines the framework of the subjecet block (in the intermediate MSA before the move of this block is undone);
#
#	$char_abs is the character representing the absence of the residue. (It is the same as the 'gap' character);
#	$full_shift is the full value of the shift of the subject block between the reference and the reconstructed MSA;
#	$actual_shift is the value of the shift by which the move of the subject block is actually undone (from the reconstructed MSA to the intermediate MSA);
#
#	@map_shifts_respos_intm is the position-shift map between the reference and the intermediate MSAs. It is under construction.
#	@map_shifts_respos_rec_af is the position-shift map between the intermediate and the reconstructed MSAs. It is also under construction.
#	($map_shifts_respos_xxx[$indx_seq]->[$s] is the position-shift at the site of the $indx_seq th sequence with horozontal position = $s.);
#
# OUTPUT: $if_success, where
#	$if_success = 1 if the subroutine ended successfully,
#		    = 0 if it failed.
#
sub undo_sgl_bw_fw_move_on_2maps_2 (\%$$$\@\@) {

	my ($frwk_bl, $char_abs, $full_shift, $actual_shift, $map_shifts_respos_intm, $map_shifts_respos_rec_af) = @_;

	my $remaining_shift = $full_shift - $actual_shift;

	my $end_msa_intm = my $end_msa_intm0 = @{$map_shifts_respos_intm->[0]} -1; # ADDED on Oct 14 (Wed), 2015. #

	foreach my $indx_seq (keys %{$frwk_bl}) { # Main foreach-loop (over sequences involved with the subject block). #

		my ($beg_fw_outer, $end_fw_outer) = @{$frwk_bl->{$indx_seq}};
		my $shifts_in_seq_intm   = $map_shifts_respos_intm->[$indx_seq];
		my $shifts_in_seq_rec_af = $map_shifts_respos_rec_af->[$indx_seq];

		($end_msa_intm < $end_fw_outer - $actual_shift) and ($end_msa_intm = $end_fw_outer - $actual_shift); # ADDED on Oct 14 (Wed), 2015. #

=DISABLED in creation of 'undo_sgl_bw_fw_move_on_2maps_2' (i).

			# Determine the boundaries of the "inner" framework, if at all. #

		my $beg_fw_inner = $beg_fw_outer;
		for (my $s=$beg_fw_outer+1; $s<=$end_fw_outer; $s++) { # 1st inner for-loop (over sites within the framework). #

		    my $shift = $shifts_in_seq_intm->[$s];
		    ##my $shift = $shifts_in_seq_rec_af->[$s]; # A BUG fixed on Oct 16 (Fri), 2015. #

		    (($shift eq $char_abs) or ($shift != $full_shift)) and last;
		    $beg_fw_inner++;
		}				# End of the 1st inner for-loop (over sites within the framework). #

		my $end_fw_inner;
		if ($beg_fw_inner < $end_fw_outer) {
		    $end_fw_inner = $end_fw_outer;
		    for (my $s = $end_fw_outer-1; $s>$beg_fw_inner; $s--) { # 2nd inner for-loop (over sites within the framework minus the shell at the beginning). #

			my $shift = $shifts_in_seq_intm->[$s];
			##my $shift = $shifts_in_seq_rec_af->[$s]; # A BUG fixed on Oct 16 (Fri), 2015. #

			(($shift eq $char_abs) or ($shift != $full_shift)) and last;
			$end_fw_inner--;
		    }				# End of the 2nd inner for-loop (over sites within the framework minus the shell at the beginning). #
		}

=cut # END of "DISABLED in creation of 'undo_sgl_bw_fw_move_on_2maps_2' (i)."

			# Extert the actual change. #

		my @sites_old_shifts = ($beg_fw_outer .. $end_fw_outer);
	 	($actual_shift < 0) and (@sites_old_shifts = reverse (@sites_old_shifts));

		for my $s (@sites_old_shifts) { # 1st inner for-loop (over sites within the framework.) #

		    my $old_shift = $shifts_in_seq_intm->[$s];
		    $shifts_in_seq_intm->[$s-$actual_shift] = ($old_shift eq $char_abs) ? $char_abs : $old_shift - $actual_shift; 
		    $shifts_in_seq_rec_af->[$s] = ($old_shift eq $char_abs) ? $char_abs : $actual_shift;

		}			# End of the 1st inner for-loop (over sites within the framework.) #


		my @sites_new_gaps = ($actual_shift>0) ? ($end_fw_outer - $actual_shift + 1 .. $end_fw_outer) : ($beg_fw_outer .. $beg_fw_outer - $actual_shift - 1);

		for my $s (@sites_new_gaps) { # The 2nd inner for-loop (over sites to be filled with new gaps.) #

		    $shifts_in_seq_intm->[$s] = $char_abs;

		} # End of the 2nd inner for-loop (over sites to be filled with new gaps.) #


=DISABLED in creation of 'undo_sgl_bw_fw_move_on_2maps_2' (ii).

		for (my $s = $beg_fw_outer; $s <= $beg_fw_inner; $s++) { # 3rd inner for-loop (over sites within the shell at the beginning). #
			$shifts_in_seq_intm->[$s-$actual_shift] = $remaining_shift;
			$shifts_in_seq_rec_af->[$s] = $actual_shift;

		}				# End of the 3rd inner for-loop (over sites within the shell at the beginning). #

		my @sites_new_gaps = ($actual_shift>0) ? ($beg_fw_inner - $actual_shift+1 .. $beg_fw_inner) : ($beg_fw_outer .. $beg_fw_outer - $actual_shift - 1);

		for my $s (@sites_new_gaps) { $shifts_in_seq_intm->[$s] = $char_abs; }

		if (defined $end_fw_inner) {
		    for (my $s = $end_fw_inner; $s <= $end_fw_outer; $s++) { # 6th inner for-loop (over sites within the shell at the end). #
			$shifts_in_seq_intm->[$s - $actual_shift] = $remaining_shift;
			$shifts_in_seq_rec_af->[$s] = $actual_shift;
		    }				# End of the 6th inner for-loop (over sites within the shell at the end). #

		    my @sites_new_gaps2 = ($actual_shift>0) ? ($end_fw_outer -$actual_shift+1 .. $end_fw_outer) : ($end_fw_inner .. $end_fw_inner - $actual_shift -1);
		    for my $s (@sites_new_gaps2) { $shifts_in_seq_intm->[$s] = $char_abs; }
		}
=cut # END of "DISABLED in creation of 'undo_sgl_bw_fw_move_on_2maps_2' (ii)."
		
	}			# End of the main foreach-loop (over sequences involved with the subject block). #

	if ($end_msa_intm > $end_msa_intm0) { # ADDED on Oct 14 (Wed), 2015. #

			# Fill in the new yet unoccupied right-end with gaps. #

	    foreach my $shifts_row (@{$map_shifts_respos_intm}) { # Trailing foreach-loop (over sequences in the MSA). #
		my $ct_sites = @{$shifts_row};
		for (my $s = $ct_sites; $s <= $end_msa_intm; $s++) { $shifts_row->[$s] = $char_abs; }
	    }					# End of the trailing foreach-loop (over sequences in the MSA). #
	}

	return 1;

}		# End of the subroutine 'undo_sgl_bw_move_on_2maps_2'. #


# rm_zero_shifts_at_ends:
#	This satellite subroutine of 'undo_mve_latter_of_cblocks_pair' removes blocks with zero shift
#	from both ends of a subject c-block.
#	The subroutine modifies the input variables accordingly, instead of outputting the results of its intended process.
#	[CREATED on Oct 18 (Sun), 2015.]
#
# INPUT: (\@set_blocks_shift_respos, $indx_shift, \@constituent_blks, \@cp_constituent_blks, \@info_constituents, \@frwks_cb, \@merger_types, \@cp_merger_types, \@bw_hard_linking, $indx_cb, \$ct_constituents, \@cbs_hard_linking), where
#
#	$set_blocks_shift_respos[$bl]->[$indx_shift] gives the shift of the $bl th block in the set;
#
#	@constituent_blks stores the indices (in @set_blocks_shift_respos) of the blocks that constitute the subject c-block,
#	and @cp_constituent_blks is its copy;
#
#	$info_constituents[$i] and $frwks_cb[$i] give some information and the framework, respectively, of the $i th block in @constituent_blks;
#
#	$merger_types[$i] gives the type of the merger between the $i th and the ($i+1) th constituent blocks,
#	and @cp_merger_types is a copy of @merger_types;
#
#	[ $bw_hard_linking[$i]->[$indx_cb] gives the index of the c-block accommodating the block that hard-links the $i th constituent block; ] --- It was WRONG!!
#	$bw_hard_linking[$i]->[$k][$indx_cb] gives the index of the c-block accommodating the $k th block that hard-links the $i th constituent block. (@{$bw_hard_linking[$i]} is empty if the $i th constituent block is NOT hard-linked by any block);

#
#	$ct_constituents is the number of constituents;
#
#	@cbs_hard_linking stores the indices of the c-blocks that hard-link the subject c-block.
#
# OUTPUT: 1 if the subroutine ended successfully,
#	  0 if it failed.
#
sub rm_zero_shifts_at_ends (\@$\@\@\@\@\@\@\@$\$\@) {

	my ($set_blocks_shift_respos, $indx_shift, $constituent_blks, $cp_constituent_blks, $info_constituents,
	   $frwks_cb, $merger_types, $cp_merger_types, $bw_hard_linking, $indx_cb, $rf_ct_constituents, $cbs_hard_linking) = @_;

	my @list_manipulated = ($constituent_blks, $cp_constituent_blks, $info_constituents, $frwks_cb, $merger_types, $cp_merger_types, $bw_hard_linking);

		# First, remove the blocks with zero shift from the left-end of the subject c-block. #

	while (@{$constituent_blks}>0) { # 1st main while-loop (as long as the c-block is non-empty). #

	    my $bl_le = $constituent_blks->[0];
	    my $shift_le = $set_blocks_shift_respos->[$bl_le][$indx_shift];
	    ($shift_le != 0) and last;

	    foreach my $manipulated (@list_manipulated) { shift @{$manipulated}; }

	}			# End of the 1st main while-loop (as long as the c-block is non-empty). #


		# Second, remove the blocks with zero shift from the right-end of the subject c-block. #

	while (@{$constituent_blks}>0) { # 2nd main while-loop (as long as the c-block is non-empty). #

	    my $bl_re = $constituent_blks->[$#{$constituent_blks}];
	    my $shift_re = $set_blocks_shift_respos->[$bl_re][$indx_shift];
	    ($shift_re != 0) and last;

	    foreach my $manipulated (@list_manipulated) { pop @{$manipulated}; }

	}			# End of the 2nd main while-loop (as long as the c-block is non-empty). #

	if (@{$constituent_blks} == ${$rf_ct_constituents}) {

	   return 1; # If no modification was made, end the subroutine successfully.
	}

		# Reflect the modification to the remaining variables. #

	${$rf_ct_constituents} = scalar (@{$constituent_blks}); # Update #{constituents}. #

	my @new_cbs_hard_linking = ();
	my %already; # = ($cb => 1, ...) for the c-blocks already appeared in @{$bw_hard_linking}.

		# An WRONG single-foreach-loop was CORRECTED to the DOUBLE-foreach-loops. #
	foreach my $hard_linking_to_bl (@{$bw_hard_linking}) { # 3rd main outer foreach-loop (over constituent blocks). #
	    (defined $hard_linking_to_bl) or next;
	    foreach my $info_hl_bl (@{$hard_linking_to_bl}) { # Inner foreach-loop (over hard-linking blocks). #
	##foreach my $info_hl_bl (@{$bw_hard_linking}) { # 3rd main foreach-loop (over hard-linking c-blocks). # # This was WRONG!!

		my $cb = $info_hl_bl->[$indx_cb];
		(defined $already{$cb}) and next;

		push @new_cbs_hard_linking, $cb;
		$already{$cb} = 1;
	##}				# End of the 3rd main foreach-loop (over hard-linking c-blocks). # # This was WRONG!!
	    }					 # End of the inner foreach-loop (over hard-linking blocks). #
	}				# End of the 3rd main outer foreach-loop (over constituent blocks). #

	@{$cbs_hard_linking} = @new_cbs_hard_linking; # Update the list of hard-likning c-blocks. #

	return 1; # End the subroutine successfully. #

}	# End of the subroutine 'rm_zero_shifts_at_ends'. #


# undo_mv_latter_of_cblock_pair:
#	This subroutine attempts to "undo" the move of a composite-blocks (the 2nd one in a specified pair),
#	and create a position-shift map between the reference MSA and the MSA resulting from the "undoing."
#
#	[RULES of "undoing"]
#	(1) Basically, a move is "undone" according to the following procedure:
#	   (i) Subtract the "baseline" shift from the shift at each site of the input position-shift map.
#		(If the "baseline" shift is negative, pad |"baseline" shift| columns onto the left of the map);
#	   (ii) "undo" the move of 2nd composite-block of the input pair;
#	   (iii) if the "baseline" shift is positive, add the "baseline" shift to the shift at each site of the resulting map.
#
#	[NOTE: The subroutien now refrains from performing (iii), because the process preceding this subroutine
#	has now designed to always make $baseline_shift zero. (on Oct 14 (Wed), 2015) ]
#
#	(2) The move of the 2nd composite-block is undone completely if there is no physical obstacle;
#	(3) If there are (physical) obstacles, the move is "undone" incompletely,
#	   and modify @cblock_2nd_mvd_af by reflecting such incomplete move,
#		and the remaining part of the 2nd c-block move will be incorporated into @cblock_1st_mvd_af.
#	[ (4) Unless @cblock_1st_mvd_af is expected to be undone completely, or unless the 2nd composite-block is expected to be undone at all,  the relevant one between \@cblock_1st/2nd_mvd_af will be replaced with undef. ] --- OBSOLETE
#	(4) Unless @cblock_1st_mvd_af is expected to be undone completely, or unless the 2nd composite-block is expected to be undone at all,  
#	   the subroutine returns a negative falue of $if_success (and undef for other output variables).
#
#	[CREATION STARTED on Oct 10 (Sat), 21015, ENDED on Oct 15 (Thu), 2015.]
#
#	[MODIFIED on Oct 16 (Fri), 2015]
#	Now, the subroutine returns the set of indices of c-blocks hard-linking the (new) 1st moved c-block,
#			as well as the set of indices of c-blocks hard-linking the (new) 2nd moved c-block,
#	instead of modifying the input %cb2hard_linking.
#
#	[MODIFIED in DEBUGGING on Oct 24 (Sat), 2015]
#	Now, when the subroutine encounters a c-block whose shift is zero at both ends,
#	the subroutine considers that the path cannot be undone,
#	instead of considering as the subroutine's own error.
#
#	[MODIFIED on Oct 25 (Sun), 2015]
#	Now, the subroutine handles any errors by its satellite subroutines as the situation where the unmoving is hard,
#	 instead of as its own error.
#
#
# INPUT: (\@map_shifts_respos, $baseline_shift, $indx_cb_1st_mvd, $indx_cb_2nd_mvd, \@set_composite_blocks, \@set_blocks_shift_respos, \%cb2hard_linking, \%node2ch, \%node2seq_indx, $char_abs), where
#
#	$map_shifts_respos[$k]->[$s] is the shift in the residue position from the reference MSA to the reconstructed one
#	for the site of the $k th sequence with the horizontal position = $s, in the reconstructed MSA. It is $gap if occupied by a gap;
#	$baseline_shift is the "baseline shift," which was originally the commoner shift between the left-flanking and the right-flanking ones;
#	$indx_cb_1st/2nd_mvd is the index of the composite-block that is interpreted as being moved 1st/2nd in the pair;
#
#	@{$set_composite_blocks[$cb]} = ($beg_cb_rec, $end_cb_rec, $nodeid_mrca_cb, \%br2change_cb, \@constituent_blks, \@involved_seqs_cb, \@merger_types) is for the $cb th composite block. Here
#	@constituent_blks stores indices of the position-shift blocks that constitute the composite block;
#
#	@{$set_blocks_shift_respos[$bl]} = ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes) is for the $bl th position-shift block. Here
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block;
#
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...), and
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-block excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block;
#
#	%cb2hard_linking = ($indx_cb => [c-blocks hard linking $indx_cb], ...);
#
#	%node2ch = ($node_id => \@child_nodes, ...) for internal nodes;
#	%node2seq_indx = ($node_id => $indx_seq, ...) for external nodes;
#
# [OUTPUT: ($if_success, \@map_shifts_respos_intm, \@map_shifts_respos_rec_af, $baseline_shift_intm, \@set_composite_blocks_intm, \@set_blocks_shift_respos_intm, \@set_composite_blocks_af, \@set_blocks_shift_respos_af), where ] --- OBSOLETE as of Oct 16 (Fri), 2015.
# OUTPUT: ($if_success, \@map_shifts_respos_intm, \@map_shifts_respos_rec_af, $baseline_shift_intm, \@set_composite_blocks_intm, \@set_blocks_shift_respos_intm, \@set_composite_blocks_af, \@set_blocks_shift_respos_af, \@cbs_hard_linking_1st_mvd_intm, \@cbs_hard_linking_2nd_mvd_rec_af), where
#
#	$if_success = 1 if the 2nd move is successfully undone (either completely or incompletely),
#		    = {a negative value} if we cannot undo the 2nd move or if it is expected that the subsequent 1st move cannot be undone completely,
#				because of the constraints by the surrounding of the relevant blocks (the specific value of $if_success tells what critetion was not satisfied),
#		    = undef if the subroutine failed;
#
#	@map_shifts_respos_intm is the map of position-shifts between the reference and the "intermediate" MSAs, (the latter of which resulted from "undoing' $indx_cb_2nd_mvd), onto the latter;
#	$baseline_shift_intm  is the "baseline shift" for the intermediate position-shift map;
#
#	@map_shifts_respos_rec_af is the map of position-shifts between the intermediate and the reconstructed MSAs, the latter of which underwent an overall shift (if $baseline_shift < 0);
#
#	@set_composite_blocks_intm must give the set of c-blocks between the reference and intermediate MSAs. Basically, it consists of the same components as @set_composite_blocks, which are modified to reflect the difference between the reference and the intermediate MSAs;
#	@set_blocks_shift_respos_intm must correctly give the constituent shift-blocks of the $indx_cb_1st th c-block, between the reference and the intermediate MSAs;
#
#	@set_composite_blocks_af must give the set of c-blocks between the intermediate and the reconstructed MSAs. Basically, it consists of the same components as @set_composite_blocks, which are modified to reflect the difference between the intermediate and the reconstructed MSAs;
#	@set_blocks_shift_respos_af must correctly give the constituent shift-blocks of the $indx_cb_2nd th c-block, between the reference and the intermediate MSAs;
#
#	@cbs_hard_linking_1st_mvd_intm stores indices of c-blocks that hard-links the (new) 1st moved c-block (between the reference and intermediate MSAs);
#	@cbs_hard_linking_2nd_mvd_rec_af stores indices of c-blocks that hard-links the (new) 2nd moved c-block (between the intermediate and reconstructed MSAs).
#
sub undo_mv_latter_of_cblock_pair (\@$$$\@\@\%\%\%$) {

	my ($map_shifts_respos, $baseline_shift, $indx_cb_1st_mvd, $indx_cb_2nd_mvd, $set_composite_blocks, $set_blocks_shift_respos, $cb2hard_linking, $node2ch, $node2seq_indx, $char_abs) = @_;

		# Prepare the 2ndary inputs. #

	my $ct_seqs = @{$map_shifts_respos};
	my $ct_clms_rec = @{$map_shifts_respos->[0]};

   (print STDOUT "  <<< Subroutine 'undo_mv_latter_of_cblock_pair' (#{sequences} = $ct_seqs; #{columns}_rec = $ct_clms_rec) >>>\n\n") if ($TEST_LEVEL_umlp>0);

	my $hard_linking1 = $cb2hard_linking->{$indx_cb_1st_mvd};
	my @cbs_hard_linking1 = (defined $hard_linking1) ? @{$hard_linking1} : (); # MODIFIED on Oct 16 (Fri), 2015. #
	##(defined $hard_linking1) or ($hard_linking1 = []);

	my $hard_linking2 = $cb2hard_linking->{$indx_cb_2nd_mvd};
	my @cbs_hard_linking2 = (defined $hard_linking2) ? @{$hard_linking2} : (); # MODIFIED on Oct 16 (Fri), 2015. #
	##(defined $hard_linking2) or ($hard_linking2 = []);

    if ($TEST_LEVEL_umlp>2) {
	print STDOUT "  [ (Primary and Secondary) Inputs ]\n\n";
	print STDOUT '  $baseline_shift = ', $baseline_shift, ",\n";
	print STDOUT '  $indx_cb_1st_mvd = ', $indx_cb_1st_mvd, ",\n";
	print STDOUT '  $indx_cb_2nd_mvd = ', $indx_cb_2nd_mvd, ",\n\n"; 
	print STDOUT '  @cbs_hard_linking1 = (', join (', ', @cbs_hard_linking1), ") ,\n";
	print STDOUT '  @cbs_hard_linking2 = (', join (', ', @cbs_hard_linking2), ") .\n\n";
    }

		# Initialize the outputs # 

	my $ct_clms_rec2 = $ct_clms_rec;
	my $baseline_shift_intm = $baseline_shift;

	if ($baseline_shift < 0) {
	    $ct_clms_rec2 -= $baseline_shift;
	    $baseline_shift_intm = 0;
	    ##for my $i (0, 1) { $cblock_1st_mvd_af[$i] -= $baseline_shift; $cblock_2nd_mvd_af -= $baseline_shift; }
	}

	my @set_composite_blocks_intm = my @set_composite_blocks_af = ();
		# Initially, both are identical to the input @{$set_composite_blocks},
		# except that it underwent an overall shift if $baseline_shift < 0).
	foreach my $info_cblk (@{$set_composite_blocks}) { # 1st preliminary foreach-loop (over input c-blocks). #

	    my ($beg_cb, $end_cb, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $invlvd_seqs_cb, $merger_types) = @{$info_cblk};
	    if ($baseline_shift < 0) { $beg_cb -= $baseline_shift; $end_cb -= $baseline_shift; }

	    my %br2change_cb1 = my %br2change_cb2 = %{$br2change_cb};
	    my @constituent_blks1 = my @constituent_blks2 = @{$constituent_blks};
	    my @invlvd_seqs_cb1 = my @invlvd_seqs_cb2 = @{$invlvd_seqs_cb}; 
	    my @merger_types1 = my @merger_types2 = @{$merger_types};

	    push @set_composite_blocks_intm, [$beg_cb, $end_cb, $nodeid_mrca_cb, \%br2change_cb1, \@constituent_blks1, \@invlvd_seqs_cb1, \@merger_types1];
	    push @set_composite_blocks_af, [$beg_cb, $end_cb, $nodeid_mrca_cb, \%br2change_cb2, \@constituent_blks2, \@invlvd_seqs_cb2, \@merger_types2];

	}				# End of the 1st preliminary foreach-loop (over input c-blocks). #

	my @set_blocks_shift_respos_intm = my @set_blocks_shift_respos_af = ();
		# Initially, both are identical to the input @{$set_blocks_shift_respos},
		# except that the shifts of the former are subtracted by $baseline_shift
		# and that the shifts of the latter are equated to zero,
		# (and except that it underwent an overall shift if $baseline_shift < 0).

	foreach my $info_block (@{$set_blocks_shift_respos}) { # 2nd preliminary foreach-loop (over input shift-blocks). #

	    my ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, $seqs_involved, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$info_block};
	    $shift -= $baseline_shift;
	    if ($baseline_shift < 0) { $beg_blk -= $baseline_shift; $end_blk -= $baseline_shift; }
	    my @seqs_involved1 = my @seqs_involved2 = @{$seqs_involved};
	    my %br2change1 = my %br2change2 = %{$br2change};
	    my @indices_src_classes1 = my @indices_src_classes2 = @{$indices_src_classes};

	    my @excl_sbblocks1 = my @excl_sbblocks2 = ();
	    foreach my $sbblk (@{$excl_sbblocks}) {
		my ($beg_site, $end_site, $mrca_node) = @{$sbblk};
		if ($baseline_shift < 0) { $beg_site -= $baseline_shift; $end_site -= $baseline_shift; }
		my @sbblk1 = my @sbblk2 = ($beg_site, $end_site, $mrca_node);
		push @excl_sbblocks1, \@sbblk1;
		push @excl_sbblocks2, \@sbblk2;
	    }

	    push @set_blocks_shift_respos_intm, [$shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved1, \%br2change1, \@excl_sbblocks1, $ct_changes, \@indices_src_classes1];
	    ##push @set_blocks_shift_respos_af, [$shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved2, \%br2change2, \@excl_sbblocks2, $ct_changes, \@indices_src_classes2];
	    push @set_blocks_shift_respos_af, [0, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved2, \%br2change2, \@excl_sbblocks2, $ct_changes, \@indices_src_classes2];
	}					# End of the 2nd preliminary foreach-loop (over input shift-blocks). #

	my @map_shifts_respos_intm = (); # Initially, it is just the input map (from reference to reconstructed), possibly undergoing an overall shift.
	my @map_shifts_respos_rec_af = (); # Initially, it is just the self-map from the reconstructed to the reconstructed (, which may be padded on the left-side).
	foreach my $shifts_in_seq (@{$map_shifts_respos}) { # 3rd preliminary foreach-loop (over sequences in the input position-shift map). #
	    my @cp1_shifts_in_seq = @{$shifts_in_seq};
	    if ($baseline_shift<0) {
		for (my $s=0; $s<$ct_clms_rec; $s++) { ($cp1_shifts_in_seq[$s] eq $char_abs) or ($cp1_shifts_in_seq[$s] -= $baseline_shift); }
		for (1 .. -$baseline_shift) { unshift @cp1_shifts_in_seq, $char_abs; }
	    }
	    push @map_shifts_respos_intm, \@cp1_shifts_in_seq;

	    my @cp2_shifts_in_seq = ();
	    foreach my $shift (@cp1_shifts_in_seq) { push @cp2_shifts_in_seq, (($shift eq $char_abs) ? $char_abs : 0); }
	    push @map_shifts_respos_rec_af, \@cp2_shifts_in_seq;

	}				# End of the 3rd preliminary foreach-loop (over sequences in the input position-shift map). #

   if ($TEST_LEVEL_umlp>5) {
	my @seq_ids = ();
	for my $i (0 .. $ct_seqs-1) { push @seq_ids, sprintf ("seq%04d", $i); }

	print STDOUT '   [ Initial value of @map_shifts_respos_intm ]', "\n\n"; 

	unless ( pr_map_shifts_respos (*STDOUT, @seq_ids, @map_shifts_respos_intm, $GAP, $DIS_LEN_SHIFT) ) {
	    print STDOUT "[ERROR 'undo_mv_latter_of_cblock_pair'] 'pr_map_shifts_respos' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}
	print STDOUT "\n";

	print STDOUT '   [ Initial value of @map_shifts_respos_rec_af ]', "\n\n"; 

	unless ( pr_map_shifts_respos (*STDOUT, @seq_ids, @map_shifts_respos_rec_af, $GAP, $DIS_LEN_SHIFT) ) {
	    print STDOUT "[ERROR 'undo_mv_latter_of_cblock_pair'] 'pr_map_shifts_respos' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}
	print STDOUT "\n";
   }

		# Again, prepare the 2ndary inputs. #

	my $info_cblock_1st_mvd = $set_composite_blocks_intm[$indx_cb_1st_mvd];
	my $info_cblock_2nd_mvd = $set_composite_blocks_intm[$indx_cb_2nd_mvd];

	my ($beg_cb1, $end_cb1, $nodeid_mrca_cb1, $br2change_cb1, $constituent_blks1, $invlvd_seqs_cb1, $merger_types1) = @{$info_cblock_1st_mvd};
	my ($beg_cb2, $end_cb2, $nodeid_mrca_cb2, $br2change_cb2, $constituent_blks2, $invlvd_seqs_cb2, $merger_types2) = @{$info_cblock_2nd_mvd};

	my $ct_constituents1 = @{$constituent_blks1};
	my $ct_constituents2 = @{$constituent_blks2};

			# ADDED on Oct 13 (Tue), 2015. # 
	my $info_cblock_1st_mvd_af = $set_composite_blocks_af[$indx_cb_1st_mvd];
	my $info_cblock_2nd_mvd_af = $set_composite_blocks_af[$indx_cb_2nd_mvd];
	my ($beg_cb1_af, $end_cb1_af, $nodeid_mrca_cb1_af, $br2change_cb1_af, $constituent_blks1_af, $invlvd_seqs_cb1_af, $merger_types1_af) = @{$info_cblock_1st_mvd_af};
	my ($beg_cb2_af, $end_cb2_af, $nodeid_mrca_cb2_af, $br2change_cb2_af, $constituent_blks2_af, $invlvd_seqs_cb2_af, $merger_types2_af) = @{$info_cblock_2nd_mvd_af};
			# END of "ADDED on Oct 13 (Tue), 2015." # 

		# Create Auxiliary Variables #

	# $frwk_cb_1st/2nd_mvd->[$i]{$indx_seq} = [$beg_frwk, $end_frwk], which determines the beginning- and ending-positions of the framework of the $i th constituent block in the $indx_seq th sequence (row),
	# 				= undef if the $i th constituent block does not involve the $indx_seq th sequence;
	# @{$info_constituents1/2->[$i]} = ($shift, $beg_bl, $end_bl) is for the $i th constituent block of the 1st/2nd moved c-block;
	# @{$bw_hard_linking1/2->[$i]} is an array, each of whose elements is [$indx_cb2, $bl2] for the block that is hard_linking the $i th constituent block of the 1st/2nd moved c-block. (It is empty if the $i th constituent is not hard-linked by any other block.)
	#
	##my ($frwk_cb_1st_mvd, $info_constituents1, $bw_hard_linking1) = framework_cblock ($ct_seqs, $indx_cb_1st_mvd, @set_composite_blocks_intm, @{$hard_linking1}, @set_blocks_shift_respos_intm, %{$node2ch}, %{$node2seq_indx}); # MODIFIED on Oct 16 (Fri), 2015. #
	my ($frwk_cb_1st_mvd, $info_constituents1, $bw_hard_linking1) = framework_cblock ($ct_seqs, $indx_cb_1st_mvd, @set_composite_blocks_intm, @cbs_hard_linking1, @set_blocks_shift_respos_intm, %{$node2ch}, %{$node2seq_indx});
	unless (defined $frwk_cb_1st_mvd) {
	    print STDOUT "[ERROR 'undo_mvd_latter_of_cblocks_pair'] Subroutine 'framkework_cblock' FAILED on the 1st c-block (index: $indx_cb_1st_mvd): $! !\n\n";
	    return (-1001, undef, undef, undef, undef, undef, undef, undef, undef, undef); # MODIFIED on Oct 25 (Sun), 2015. #
	    ##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}

   if ($TEST_LEVEL_umlp>3) {
	print STDOUT "   [ Output of 'framework_cblock' on the 1st-moved c-block ]\n\n";

	for (my $i=0; $i <$ct_constituents1; $i++) { # 1st outer testing for-loop (over constituent blocks of the 1st-moved). #

	    my ($shift_bl, $beg_bl, $end_bl) = @{$info_constituents1->[$i]};
	    print STDOUT "    { Framework of the $i th constituent (shift = $shift_bl, beg_bl = $beg_bl, end_bl = $end_bl) }\n\n"; 
	    my $frwk_bl = $frwk_cb_1st_mvd->[$i];

	    print STDOUT join ("\t", 'Indx_seq', 'beg_frwk', 'end_frwk'), "\n\n";
	    foreach my $indx_seq (sort {$a <=> $b} keys %{$frwk_bl}) { print STDOUT join ("\t", $indx_seq, @{$frwk_bl->{$indx_seq}}), "\n"; }

	    my $hard_linking_to_bl = $bw_hard_linking1->[$i];
	    my $ct_hl = @{$hard_linking_to_bl};

	    print STDOUT "\n   ( There are $ct_hl blocks hard-linking this block: )\n\n";
	    print STDOUT join ("\t", 'Indx_cb', 'indx_bl'), "\n\n";
	    foreach my $bl_info (@{$hard_linking_to_bl}) { print STDOUT join ("\t", @{$bl_info}), "\n"; }
	    print STDOUT "\n\n";

	}				# End of the 1st outer testing for-loop (over constituent blocks of the 1st-moved). #
	print STDOUT "\n";

   } # END of "if ($TEST_LEVEL_umlp>3) {...}"

	##my ($frwk_cb_2nd_mvd, $info_constituents2, $bw_hard_linking2) = framework_cblock ($ct_seqs, $indx_cb_2nd_mvd, @set_composite_blocks_intm, @{$hard_linking1}, @set_blocks_shift_respos_intm, %{$node2ch}, %{$node2seq_indx}); # MODIFIED on Oct 16 (Fri), 2015. #
	my ($frwk_cb_2nd_mvd, $info_constituents2, $bw_hard_linking2) = framework_cblock ($ct_seqs, $indx_cb_2nd_mvd, @set_composite_blocks_intm, @cbs_hard_linking2, @set_blocks_shift_respos_intm, %{$node2ch}, %{$node2seq_indx});
	unless (defined $frwk_cb_2nd_mvd) {
	    print STDOUT "[ERROR 'undo_mvd_latter_of_cblocks_pair'] Subroutine 'framkework_cblock' FAILED on the 2nd c-block (index: $indx_cb_2nd_mvd): $! !\n\n";
	    return (-1002, undef, undef, undef, undef, undef, undef, undef, undef, undef); # MODIFIED on Oct 25 (Sun), 2015. #
	    ##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}

   if ($TEST_LEVEL_umlp>3) {
	print STDOUT "   [ Output of 'framework_cblock' on the 2nd-moved c-block ]\n\n";

	for (my $i=0; $i <$ct_constituents2; $i++) { # 1st outer testing for-loop (over constituent blocks of the 2nd-moved). #

	    my ($shift_bl, $beg_bl, $end_bl) = @{$info_constituents2->[$i]};
	    print STDOUT "    { Framework of the $i th constituent (shift = $shift_bl, beg_bl = $beg_bl, end_bl = $end_bl) }\n\n"; 
	    my $frwk_bl = $frwk_cb_2nd_mvd->[$i];

	    print STDOUT join ("\t", 'Indx_seq', 'beg_frwk', 'end_frwk'), "\n\n";
	    foreach my $indx_seq (sort {$a <=> $b} keys %{$frwk_bl}) { print STDOUT join ("\t", $indx_seq, @{$frwk_bl->{$indx_seq}}), "\n"; }

	    my $hard_linking_to_bl = $bw_hard_linking2->[$i];
	    my $ct_hl = @{$hard_linking_to_bl};

	    print STDOUT "\n   ( There are $ct_hl blocks hard-linking this block: )\n\n";
	    print STDOUT join ("\t", 'Indx_cb', 'indx_bl'), "\n\n";
	    foreach my $bl_info (@{$hard_linking_to_bl}) { print STDOUT join ("\t", @{$bl_info}), "\n"; }
	    print STDOUT "\n\n";

	}				# End of the 1st outer testing for-loop (over constituent blocks of the 2nd-moved). #
	print STDOUT "\n";

   } # END of "if ($TEST_LEVEL_umlp>3) {...}"


	my $le_shift1 = $info_constituents1->[0][0];
	my $re_shift1 = $info_constituents1->[$ct_constituents1-1][0];

	my $le_shift2 = $info_constituents2->[0][0];
	my $re_shift2 = $info_constituents2->[$ct_constituents2-1][0];

	if (($le_shift2<=0) and ($re_shift2>=0)) {

			# MODIFIED in DEBUGGING on Oct 24 (Sat), 2015. #
	    print STDOUT "[WARNING 'undo_mvd_latter_of_cblocks_pair'] An UNEXPECTED situation: The 2nd moved c-block (index: $indx_cb_2nd_mvd) has a NON-POSITIVE left-end shift (=$le_shift2) and a NON-NEGATIVE right-end shift (=$re_shift2): $! !\n\n";
	    return (-101, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    ##print STDOUT "[ERROR 'undo_mvd_latter_of_cblocks_pair'] An UNEXPECTED situation: The 2nd moved c-block (index: $indx_cb_2nd_mvd) has a NON-POSITIVE left-end shift (=$le_shift2) and a NON-NEGATIVE right-end shift (=$re_shift2): $! !\n\n";
	    ##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}

	my @ordered_constituents2 = (0 .. $ct_constituents2-1);

	unless ($le_shift2>0) { @ordered_constituents2 = reverse (@ordered_constituents2); }


		## Main Process (1): 1st attempt to "undo" the 2nd move. ##

	my $indx_cannot_undo1 = -1;
	my @indices_seqs_incomplete_mvs = (); #
	my @set_incomplete1 = (); # element = [$indx_bl, $actual_shift, $shift - $actual_shift, \@indices_seqs_incomplete_mvs] for a constituent block whose move is incompletely undone.

	my @updated_frwk_cb_2nd_mvd = (); # Stores updated framework of the 2nd-moved c-block. # ADDED in DEBUGGING the subroutine on Oct 17 (Sat), 2015. #

	for (my $ii2=0; $ii2<$ct_constituents2; $ii2++) { # 1st main outer for-loop (over constituent blocks of the 2nd-moved c-block). #
	    my $i2 = $ordered_constituents2[$ii2];

	    my $bl2 = $constituent_blks2->[$i2];
	    my $frwk_bl2 = $frwk_cb_2nd_mvd->[$i2];
	    my ($shift_bl2, $beg_bl2, $end_bl2) = @{$info_constituents2->[$i2]};
	    my $set_hard_linking_bl2 = $bw_hard_linking2->[$i2];

	    my ($actual_shift_bl2, $indices_seqs_incomplete_mvs) = scout_possible_undoing ($shift_bl2, %{$frwk_bl2}, @map_shifts_respos_intm, $CHAR_ABS);
	    unless (defined $actual_shift_bl2) {
		print STDOUT "[ERROR 'undo_mv_latter_of_cblock_pair'] Subroutine 'scout_possible_undoing' FAILED on the $i2 th constituent block (index: $bl2) of the original 2nd-moved c-block: $! !\n\n";
		return (-1003, undef, undef, undef, undef, undef, undef, undef, undef, undef); # MODIFIED on Oct 25 (Sun), 2015. #
		##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }

	    if ($actual_shift_bl2 == 0) {
		$indx_cannot_undo1 = $ii2;
		@indices_seqs_incomplete_mvs = @{$indices_seqs_incomplete_mvs};
		last;
	    }

=REPLACED on Oct 13 (Tue), 2013.
	    if ($ii2 == 0) {
		($actual_shift_bl2 == 0) and last; # In this case, we cannot (properly) undo the 2nd move in any way(, unless the 1st move is undone together). #
		$if_undone1 = 1;
	    }
=cut

	    my $remaining_shift = $shift_bl2 - $actual_shift_bl2;

			# MODIFIED on Oct 13 (Tue), 2015.
	    ($remaining_shift == 0) or (push @set_incomplete1, [$i2, $actual_shift_bl2, $remaining_shift, $indices_seqs_incomplete_mvs]);
	    ##($remaining_shift == 0) or (push @set_incomplete1, [$i2, $remaining_shift]);

		## Actually undo the 2nd move. ##

		# Reflect it on the position-shift maps. #

			# MODIFIED in DEBUGGING on Oct 25 (Sun), 2015. # 
	    unless ( undo_sgl_bw_fw_move_on_2maps_2 (%{$frwk_bl2}, $char_abs, $shift_bl2, $actual_shift_bl2, @map_shifts_respos_intm, @map_shifts_respos_rec_af) ) {
	    ##unless ( undo_sgl_bw_fw_move_on_2maps (%{$frwk_bl2}, $char_abs, $shift_bl2, $actual_shift_bl2, @map_shifts_respos_intm, @map_shifts_respos_rec_af) ) { ...}
		print STDOUT "[ERROR 'undo_mv_latter_of_cblock_pair'] Subroutine 'undo_sgl_bw_fw_move_on_2maps_2' FAILED on the $i2 th constutuent block (index: $bl2): $! !\n\n";
		return (-1005, undef, undef, undef, undef, undef, undef, undef, undef, undef); # MODIFIED on Oct 25 (Sun), 2015. #
		##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }


		# Reflect it on the main position-shift blocks. #

	    my $block_shift_intm = $set_blocks_shift_respos_intm[$bl2];
	    for my $i (0, 1, 2) { $block_shift_intm->[$i] -= $actual_shift_bl2; }
	    my $excl_sbblocks_intm = $block_shift_intm->[7];
	    foreach my $info_sbblk (@{$excl_sbblocks_intm}) { $info_sbblk->[0] -= $actual_shift_bl2; $info_sbblk->[1] -= $actual_shift_bl2; }

	    my $block_shift_af = $set_blocks_shift_respos_af[$bl2];
	    $block_shift_af->[0] = $actual_shift_bl2;

		#  Reflect it on the hard-linking position-shift blocks. #

	    foreach my $info_blk (@{$set_hard_linking_bl2}) { # 2nd inner foreach-loop (over blocks hard-linking the main block). #
		my ($indx_cb3, $bl3) = @{$info_blk};
		my $block_shift_intm3 = $set_blocks_shift_respos_intm[$bl3];
		for my $i (0, 1, 2) { $block_shift_intm3->[$i] -= $actual_shift_bl2; }

		my $excl_sbblocks_intm3 = $block_shift_intm3->[7];
		foreach my $info_sbblk (@{$excl_sbblocks_intm3}) { $info_sbblk->[0] -= $actual_shift_bl2; $info_sbblk->[1] -= $actual_shift_bl2; }

		my $block_shift_af3 = $set_blocks_shift_respos_af[$bl3];
		$block_shift_af3->[0] = $actual_shift_bl2;
	    }				# End of the 2nd inner foreach-loop (over blocks hard-linking the main block). #

=DISABLED because it turned out to be unnecessary (on Oct 17 (Sat), 2015). #

		# Reflect it on the updated framework. # ADDED in DEBUGGING (on Oct 17 (Sat), 2015). #

	    my %updtd_frwk_bl2 ;
	    foreach my $indx_seq (keys %{$frwk_bl2}) { # 3rd inner foreach-loop (over sequences involved with the framework). #
		my ($beg_fw, $end_fw) = @{$frwk_bl2->{$indx_seq}};
		$updtd_frwk_bl2{$indx_seq} = [$beg_fw-$actual_shift_bl2; $end_fw-$actual_shift_bl2];
	    }				# End of the 3rd inner foreach-loop (over sequences involved with the framework). #
	    $updated_frwk_cb_2nd_mvd[$i2] = \%updtd_frwk_bl2;

=cut # END of "DISABLED because it turned out to be unnecessary (on Oct 17 (Sat), 2015)." #

	}			# End of the 1st main outer for-loop (over constituent blocks of the 2nd-moved c-block). #


	if ($indx_cannot_undo1>0) {

		# It would be better to give up undoing $indx_cb_2nd_mvd first. #

	    return (-1, undef, undef, undef, undef, undef, undef, undef, undef, undef);

	} elsif ($indx_cannot_undo1 == 0) {
	##unless ($if_undone1) {...} # REPLACED on Oct 13 (Tue), 2013.

		## Examine whether or not the 1st and 2nd moves can be simultaneously undone (by the same shift),
		## the 2nd move completely and the 1st move incompletely.

		# (1) Examine whether or not the incomplete moves of the "1st" block of the 2nd-moved c-block are all due to the "last" block of the 1st-moved c-block.

	    if ($le_shift2>0) { # When the 2nd move is undone, the left-end of the relevant c-block will move to the left.

		unless ($re_shift1 >= $le_shift2) {
			# In this case, the "last" block of the 1st-moved c-block cannot be the culprit.
		    return (-2, undef, undef, undef, undef, undef, undef, undef, undef, undef);
		}

		my $frwk_le2 = $frwk_cb_2nd_mvd->[0]; # Left-end (i.e., the "1st") framework of the 2nd-moved c-block. #
		my $frwk_re1 = $frwk_cb_1st_mvd->[$ct_constituents1-1]; # Right-end (i.e., the "last") framework of the 1st-moved c-block. #
		my $if_hampered = 1;
		my $if_hampered2 = 0; # ADDED in DEBUGGING on Oct 17 (Sat), 2015. #

		foreach my $indx_seq (@indices_seqs_incomplete_mvs) { # Foreach-loop (over sequences whose moves could be incompletely undone). #
		    my ($beg_le2, $end_le2) = @{$frwk_le2->{$indx_seq}};
		    ##my ($beg_le2, $end_le2) = @{$frwk_le2->[$indx_seq]}; # A BUG fixed on Oct 16 (Fri), 2015. #

		    my $shifts_in_seq = $map_shifts_respos_rec_af[$indx_seq]; # A possible BUG fixed AND MOVED FROM BELOW (on Oct 17  (Sat), 2015). #

			# A NEW conditional BLOCK REPLACING buggy blocks (on Oct 17 (Sat), 2015). # 

		    my ($beg_re1, $end_re1) = (defined $frwk_re1->{$indx_seq}) ? @{$frwk_re1->{$indx_seq}} : (undef, undef);

		    if (!(defined $beg_re1) or ($end_re1 >= $beg_le2) or ($end_re1 < $beg_le2 - $le_shift2) ) {
			 # Either the 1st framework does not involve the sequence, 
			 # it is on the wrong side of the 2nd framework,
			 # or there is enough room between the two frameworks. # 

			 # Thus, examine whether anything else block the move of the left-end of the 2nd-moved c-block. #

			for (my $s=$beg_le2-$le_shift2; $s<$beg_le2; $s++) { # Inner for-loop (over the range of the left-end of the 2nd-moved c-block). #
			    unless ($shifts_in_seq->[$s] eq $char_abs) {
				$if_hampered = 0;
				last;
			    }
			}				# End of the inner for-loop (over the range of the left-end of the 2nd moved c-block). #
			$if_hampered or last;
			next;
		    } # END of "if (!(defined $beg_re1) or ($end_re1 >= $beg_le2) or ($end_re1 < $beg_le2 - $le_shift2) ) {...}"

=DISABLED for DEBUGGING (on Oct 17 (Sat), 2015). #
		    unless (defined $frwk_re1->{$indx_seq}) {
		    ##unless (defined $frwk_re1->[$indx_seq]) {...} # A BUG fixed on Oct 16 (Fri), 2015. #
				# The 1st framework does not involve the sequence, thus cannot hamper the move of its portion in the 2nd framework. #
			$if_hampered = 0;
			last;
		    }
		    my ($beg_re1, $end_re1) = @{$frwk_re1->{$indx_seq}};
		    ##my ($beg_re1, $end_re1) = @{$frwk_re1->[$indx_seq]}; # A BUG fixed on Oct 16 (Fri), 2015. #

		    if (($end_re1 >= $beg_le2) or ($end_re1 < $beg_le2 - $le_shift2)) { # The 1st framework is on the wrong side of the 2nd framework, or there is enough roon between the two frameworks. # 
			$if_hampered = 0;
			last;
		    }
=cut # END of "DISABLED for DEBUGGING (on Oct 17 (Sat), 2015)." #


			# Look into the position-shift map,
			# and examine whether or not there is any intervening residue between the two frameworks. #

		    ##my $shifts_in_seq = $map_shifts_respos_intm[$indx_seq]; # A possible BUG fixed AND MOVED UPWARD (on Oct 17  (Sat), 2015). #

		    for (my $s = $end_re1+1; $s < $beg_le2; $s++) { # Inner for-loop (over intervening sites). #
			unless ($shifts_in_seq->[$s] eq $char_abs) {
			    $if_hampered = 0;
			    last;
			}
		    }					# End of the inner for-loop (over intervening sites). #
		    $if_hampered or last;

		    $if_hampered2 = 1; # ADDED in DEBUGGING (on Oct 17 (sat), 2015). #

		}				# End of the foreach-loop (over sequences whose moves could be incompletely undone). #

		($if_hampered and $if_hampered2) or ( return (-3, undef, undef, undef, undef, undef, undef, undef, undef, undef) );
		##$if_hampered or ( return (-3, undef, undef, undef, undef, undef, undef, undef, undef, undef) ); # A BUG fixed on Oct 17 (Sat), 2015. #

	    } else { # When the 2nd move is undone, the right-end of the relevant c-block will move to the right.

		unless ($le_shift1 <= $re_shift2) {
			# In this case, the "last" block of the 1st-moved c-block cannot be the culprit.
		    return (-5, undef, undef, undef, undef, undef, undef, undef, undef, undef);
		}

		my $frwk_re2 = $frwk_cb_2nd_mvd->[$ct_constituents2-1]; # Right-end (i.e., the "1st") framework of the 2nd-moved c-block. #
		my $frwk_le1 = $frwk_cb_1st_mvd->[0]; # Left-end (i.e., the "last") framework of the 1st-moved c-block. #

		my $if_hampered = 1;
		my $if_hampered2 = 0; # ADDED in DEBUGGING on Oct 17 (Sat), 2015. #
		foreach my $indx_seq (@indices_seqs_incomplete_mvs) { # Foreach-loop (over sequences whose moves could be incompletely undone). #
		    my ($beg_re2, $end_re2) = @{$frwk_re2->{$indx_seq}};
		    ##my ($beg_re2, $end_re2) = @{$frwk_re2->[$indx_seq]}; # A BUG fixed on Oct 16 (Fri), 2015. #

		    my $shifts_in_seq = $map_shifts_respos_rec_af[$indx_seq]; # A possible BUG fixed AND MOVED FROM BELOW (on Oct 17  (Sat), 2015). #

			# A NEW conditional BLOCK REPLACING buggy blocks (on Oct 17 (Sat), 2015). # 

		    my ($beg_le1, $end_le1) = (defined $frwk_le1->{$indx_seq}) ? @{$frwk_le1->{$indx_seq}} : (undef, undef);

		    if (!(defined $beg_le1) or ($end_re2 >= $beg_le1) or ($end_re2 -$re_shift2 < $beg_le1) ) {
			 # Either the 1st framework does not involve the sequence, 
			 # it is on the wrong side of the 2nd framework,
			 # or there is enough room between the two frameworks. # 

			 # Thus, examine whether anything else block the move of the left-end of the 2nd-moved c-block. #

			for (my $s=$end_re2+1; $s<=$end_re2-$re_shift2; $s++) { # Inner for-loop (over the range of the left-end of the 2nd-moved c-block). #
			    unless ($shifts_in_seq->[$s] eq $char_abs) {
				$if_hampered = 0;
				last;
			    }
			}				# End of the inner for-loop (over the range of the left-end of the 2nd moved c-block). #
			$if_hampered or last;
			next;
		    } # END of "if (!(defined $beg_re1) or ($end_re1 >= $beg_le2) or ($end_re1 < $beg_le2 - $le_shift2) ) {...}"

=DISABLED for DEBUGGING (on Oct 17 (Sat), 2015). #
		    unless (defined $frwk_le1->{$indx_seq}) {
		    ##unless (defined $frwk_le1->[$indx_seq]) {...} # A BUG fixed on Oct 16 (Fri), 2015. #
				# The 1st framework does not involve the sequence, thus cannot hamper the move of its portion in the 2nd framework. #
			$if_hampered = 0;
			last;
		    }
		    my ($beg_le1, $end_le1) = @{$frwk_le1->{$indx_seq}};
		    ##my ($beg_le1, $end_le1) = @{$frwk_le1->[$indx_seq]}; # A BUG fixed on Oct 16 (Fri), 2015. #

		    if (($end_re2 >= $beg_le1) or ($end_re2 - $re_shift2 < $beg_le1)) { # The 1st framework is on the wrong side of the 2nd framework, or there is enough room between the two frameworks. #
			$if_hampered = 0;
			last;
		    }
=cut # END of "DISABLED for DEBUGGING (on Oct 17 (Sat), 2015)." #


			# Look into the position-shift map,
			# and examine whether or not there is any intervening residue between the two frameworks. #

		    ##my $shifts_in_seq = $map_shifts_respos_intm[$indx_seq]; # A possible BUG fixed and MOVED UPWARD (on Oct 17 (Sat), 2015). #

		    for (my $s = $end_re2+1; $s < $beg_le1; $s++) { # Inner for-loop (over intervening sites). #
			unless ($shifts_in_seq->[$s] eq $char_abs) {
			    $if_hampered = 0;
			    last;
			}
		    }					# End of the inner for-loop (over intervening sites). #
		    $if_hampered or last;
		    $if_hampered2 = 1; # ADDED in DEBUGGING (on Oct 17 (sat), 2015). #

		}				# End of the foreach-loop (over sequences whose moves could be incompletely undone). #

		($if_hampered and $if_hampered2) or ( return (-6, undef, undef, undef, undef, undef, undef, undef, undef, undef) );
		##$if_hampered or ( return (-6, undef, undef, undef, undef, undef, undef, undef, undef, undef) ); # A BUG fixed (on Oct 17 (Sat), 2015).

	    }  # END of "if ($le_shift2>0) {...} else {...}"


		# (2) Incorporate the incomplete move of the 1st-moved c-block (by the full shift of the "1st" block of the 2nd-moved c-block) into the complete move of the 2nd-moved c-block.
		# (And the remaining move of the 1st-moved c-block will form the new 1st move.)

	    my @cp_frwk_cb_1st_mvd = ();
	    foreach my $frwk (@{$frwk_cb_1st_mvd}) { # 1st foreach-loop (over 1st moved frameworks). #
		my %cp_frwk;
		foreach my $indx_seq (keys %{$frwk}) { my ($beg_fw, $end_fw) = @{$frwk->{$indx_seq}}; $cp_frwk{$indx_seq} = [$beg_fw, $end_fw]; }
		push @cp_frwk_cb_1st_mvd, \%cp_frwk;
	    }				# End of the 1st foreach-loop (over 1st moved frameworks). #
	    my @cp_bw_hard_linking1 = ();
	    foreach my $hard_linking_to_bl (@{$bw_hard_linking1}) { # 2nd oreach-loop (over 1st moved blocks). #
		my @cp_hard_linking_to_bl = ();
		foreach my $info_bl (@{$hard_linking_to_bl}) { my ($indx_cb2, $bl2) = @{$info_bl}; push @cp_hard_linking_to_bl, [$indx_cb2, $bl2]; }
		push @cp_bw_hard_linking1, \@cp_hard_linking_to_bl;
	    }					# End of the 2nd foreach-loop (over 2nd moved blocks). #

	    my $coordinated_shift2; # The shift actually used for the coordinated undoing. # ADDED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #
	    my @one_or_two = (); # Indicates whether the constituent belonged to the 1st-moved (=1) or the 2nd-moved (=2) c-block. # ADDED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #

	    if ($le_shift2>0) { # The spatial order is: {1st mvd} {2nd mvd}.

			# Use $le_shift2 for the incomplete move of the 1st-moved c-block. #

		$coordinated_shift2 = $le_shift2; # ADDED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #

			# Prepare two sets of constituents, one is to be incorporated and the other represents the remaining move.
		my @info_constituents1_incpd = (); # Constituents of the 1st move that will be incorporated into the 2nd move.
		foreach my $constituent (@{$info_constituents1}) { # 3rd foreach-loop (over 1st moved blocks). #

 			# MODIFIED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #
		    my @cp_constituent = @{$constituent};
		    push @info_constituents1_incpd, \@cp_constituent;
		    ##push @info_constituents1_incpd, [$le_shift2, @{$constituent}[1,2] ];

		    for my $i (0, 1, 2) { $constituent->[$i] -= $le_shift2; }
		}				# End of the 3rd foreach-loop (over 1st moved blocks). #

			# Shift the original framework for the 1st move. #
		foreach my $frwk (@{$frwk_cb_1st_mvd}) { # 5th foreach-loop (over 1st moved frameworks). #
		    foreach my $indx_seq (keys %{$frwk}) {
			 for my $i (0, 1) { $frwk->{$indx_seq}[$i] -= $le_shift2; }
		    }
		}				# End of the 5th foreach-loop (over 1st moved frameworks). #

			# Incorporate the incomplete 1st move into the complete 2nd move. #

		for (1..$ct_constituents1) { push @one_or_two, 1; }  # ADDED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #
		for (1..$ct_constituents2) { push @one_or_two, 2; }  # ADDED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #

		$ct_constituents2 += $ct_constituents1;
		@{$constituent_blks2_af} = @{$constituent_blks2} = (@{$constituent_blks1}, @{$constituent_blks2});
		@{$info_constituents2} = (@info_constituents1_incpd, @{$info_constituents2});
		@{$frwk_cb_2nd_mvd} = (@cp_frwk_cb_1st_mvd, @{$frwk_cb_2nd_mvd});
		@{$bw_hard_linking2} = (@cp_bw_hard_linking1, @{$bw_hard_linking2});

			# ADDED on Oct 15 (Thu), 2015. #
			# => MODIFIED on Oct 16 (Fri), 2015. #
		@cbs_hard_linking2 = (@cbs_hard_linking1, @cbs_hard_linking2);
		##@{$hard_linking2} = (@{$hard_linking1}, @{$hard_linking2});
		##(@{$hard_linking2}>0) and ($cb2hard_linking->{$indx_cb_2nd_mvd} = $hard_linking2);

		@{$merger_types2_af} = @{$merger_types2} = (@{$merger_types1}, '1st+2nd->2nd', @{$merger_types2});

	    } else { # The spatial order is: {2nd mvd} {1st mvd}.

			# Use $re_shift2 for the incomplete move of the 1st-moved c-block. #

		$coordinated_shift2 = $re_shift2; # ADDED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #

			# Prepare two sets of constituents, one is to be incorporated and the other represents the remaining move.
		my @info_constituents1_incpd = ();
		foreach my $constituent (@{$info_constituents1}) { # 6th foreach-loop (over 1st moved blocks). #

 			# MODIFIED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #
		    my @cp_constituent = @{$constituent};
		    push @info_constituents1_incpd, \@cp_constituent;
		    ##push @info_constituents1_incpd, [$re_shift2, @{$constituent}[1,2] ];

		    for my $i (0, 1, 2) { $constituent->[$i] -= $re_shift2; }
		}				# End of the 6th foreach-loop (over 1st moved blocks). #

			# Shift the original framework for the 1st move. #
		foreach my $frwk (@{$frwk_cb_1st_mvd}) { # 7th foreach-loop (over 1st moved frameworks). #
		    foreach my $indx_seq (keys %{$frwk}) {
			 for my $i (0, 1) { $frwk->{$indx_seq}[$i] -= $re_shift2; }
		    }
		}				# End of the 7th foreach-loop (over 1st moved frameworks). #

			# Incorporate the incomplete 1st move into the complete 2nd move. #

		for (1..$ct_constituents2) { push @one_or_two, 2; }  # ADDED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #
		for (1..$ct_constituents1) { push @one_or_two, 1; }  # ADDED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #
		$ct_constituents2 += $ct_constituents1;
		@{$constituent_blks2_af} = @{$constituent_blks2} = (@{$constituent_blks2}, @{$constituent_blks1});
		@{$info_constituents2} = (@{$info_constituents2}, @info_constituents1_incpd);
		@{$frwk_cb_2nd_mvd} = (@{$frwk_cb_2nd_mvd}, @cp_frwk_cb_1st_mvd);
		@{$bw_hard_linking2} = (@{$bw_hard_linking2}, @cp_bw_hard_linking1);

			# ADDED on Oct 15 (Thu), 2015. #
			# => MODIFIED on Oct 16 (Fri), 2015. #
		@cbs_hard_linking2 = (@cbs_hard_linking2, @cbs_hard_linking1);
		##@{$hard_linking2} = (@{$hard_linking2}, @{$hard_linking1});
		##(@{$hard_linking2}>0) and ($cb2hard_linking->{$indx_cb_2nd_mvd} = $hard_linking2);

		@{$merger_types2_af} = @{$merger_types2} = (@{$merger_types2}, '2nd+1st->2nd', @{$merger_types1});

	    }  # END of "if ($le_shift2>0) {...} else {...}"

		# (3) Actually perform the new 2nd move.

	    my @new_ordered_constituents2 = (0 .. $ct_constituents2 -1);
	    unless ($le_shift2>0) { @new_ordered_constituents2 = reverse (@new_ordered_constituents2); }

	  for (my $ii2=0; $ii2<$ct_constituents2; $ii2++) { # 2nd main outer for-loop (over constituent blocks of the new 2nd-moved c-block). #
	    my $i2 = $new_ordered_constituents2[$ii2];

	    my $bl2 = $constituent_blks2->[$i2];
	    my $frwk_bl2 = $frwk_cb_2nd_mvd->[$i2];
	    my ($shift_bl2, $beg_bl2, $end_bl2) = @{$info_constituents2->[$i2]};
	    my $set_hard_linking_bl2 = $bw_hard_linking2->[$i2];

			# MODIFIED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #
	    my ($possible_shift_bl2, $indices_seqs_incomplete_mvs) = scout_possible_undoing ($shift_bl2, %{$frwk_bl2}, @map_shifts_respos_intm, $CHAR_ABS);
	    ##my ($actual_shift_bl2, $indices_seqs_incomplete_mvs) = scout_possible_undoing ($shift_bl2, %{$frwk_bl2}, @map_shifts_respos_intm, $CHAR_ABS);
	    unless (defined $possible_shift_bl2) {
	    ##unless (defined $actual_shift_bl2) {...}
		print STDOUT "[ERROR 'undo_mv_latter_of_cblock_pair'] Subroutine 'scout_possible_undoing_len' FAILED on the $i2 th constituent block (index: $bl2) of the new 2nd-moved c-block: $! !\n\n";
		return (-1006, undef, undef, undef, undef, undef, undef, undef, undef, undef); # MODIFIED on Oct 25 (Sun), 2015. #
		##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }

			# ADDED, then MODIFIED, in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #
	    my $actual_shift_bl2 = ($one_or_two[$i2] == 1) ? $coordinated_shift2 : $shift_bl2;

	    ##my $actual_shift_bl2 = ( (($coordinated_shift2>0) and ($shift_bl2 < $coordinated_shift2))
	    ##			   or (($coordinated_shift2<0) and ($shift_bl2 > $coordinated_shift2)) )
	    ##			? $shift_bl2 : $coordinated_shift2 ;


			# The condition was MODIFIED in DEBUGGING the subroutine (on Oct 16 (Fri), 2015). #
			# => Further MODIFIED (on Oct 16 (Fri), 2015). # 

	    unless ( (($actual_shift_bl2>0) and ($possible_shift_bl2>=$actual_shift_bl2)) 
			or (($actual_shift_bl2<0) and ($possible_shift_bl2<=$actual_shift_bl2)) ) {
	    ##unless ( (($actual_shift_bl2 == $shift_bl2) and ($possible_shift_bl2 == $actual_shift_bl2))
	    ##		or (($coordinated_shift2>0) and ($possible_shift_bl2>= $coordinated_shift2)) 
	    ##		or (($coordinated_shift2<0) and ($possible_shift_bl2<= $coordinated_shift2)) ) {
	    ####unless ($shift_bl2 == $actual_shift_bl2) {...} # This means that the new 2nd-moved c-block cannot be completely undone. #

		return (-7, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }

		## Actually undo the new 2nd move. ##

		# Reflect it on the position-shift maps. #

			# MODIFIED in DEBUGGING on Oct 25 (Sun), 2015. #
	    unless ( undo_sgl_bw_fw_move_on_2maps_2 (%{$frwk_bl2}, $char_abs, $shift_bl2, $actual_shift_bl2, @map_shifts_respos_intm, @map_shifts_respos_rec_af) ) {
	    ##unless ( undo_sgl_bw_fw_move_on_2maps (%{$frwk_bl2}, $char_abs, $shift_bl2, $actual_shift_bl2, @map_shifts_respos_intm, @map_shifts_respos_rec_af) ) { ...}
		print STDOUT "[ERROR 'undo_mv_latter_of_cblock_pair'] Subroutine 'undo_sgl_bw_fw_move_on_2maps_2' FAILED on the $i2 th constutuent block (index: $bl2) of the new 2nd-moved c-block: $! !\n\n";
		return (-1007, undef, undef, undef, undef, undef, undef, undef, undef, undef); # MODIFIED on Oct 25 (Sun), 2015. #
		##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }


		# Reflect it on the main position-shift blocks. #

	    my $block_shift_intm = $set_blocks_shift_respos_intm[$bl2];
	    for my $i (0, 1, 2) { $block_shift_intm->[$i] -= $actual_shift_bl2; }
	    my $excl_sbblocks_intm = $block_shift_intm->[7];
	    foreach my $info_sbblk (@{$excl_sbblocks_intm}) { $info_sbblk->[0] -= $actual_shift_bl2; $info_sbblk->[1] -= $actual_shift_bl2; }

	    my $block_shift_af = $set_blocks_shift_respos_af[$bl2];
	    $block_shift_af->[0] = $actual_shift_bl2;

		#  Reflect it on the hard-linking position-shift blocks. #

	    foreach my $info_blk (@{$set_hard_linking_bl2}) { # 2nd inner foreach-loop (over blocks hard-linking the main block). #
		my ($indx_cb3, $bl3) = @{$info_blk};
		my $block_shift_intm3 = $set_blocks_shift_respos_intm[$bl3];
		for my $i (0, 1, 2) { $block_shift_intm3->[$i] -= $actual_shift_bl2; }

		my $excl_sbblocks_intm3 = $block_shift_intm3->[7];
		foreach my $info_sbblk (@{$excl_sbblocks_intm3}) { $info_sbblk->[0] -= $actual_shift_bl2; $info_sbblk->[1] -= $actual_shift_bl2; }

		my $block_shift_af3 = $set_blocks_shift_respos_af[$bl3];
		$block_shift_af3->[0] = $actual_shift_bl2;
	    }				# End of the 2nd inner foreach-loop (over blocks hard-linking the main block). #

	  }			# End of the 2nd main outer for-loop (over constituent blocks of the new 2nd-moved c-block). #


	} elsif (@set_incomplete1>0) {

	 	## In this case, incorporate the remaining move of the 2nd moved c-block into the 1st move, if possible. ##

	    unless ($set_incomplete1[0]->[0] == $ordered_constituents2[0]) { # The incomplete unmoving starts with the 2nd or later block. #
		return (-8, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }

	    for (my $ii2 = 1; $ii2<$ct_constituents2; $ii2++) {
		(defined $set_incomplete1[$ii2]) or next; # ADDED for DEBUGGING on Oct 16 (Fri), 2015. #
		unless ($set_incomplete1[$ii2]->[0] == $ordered_constituents2[$ii2]) { # A completely unmoved block intervenes two incompletely unmoved blocks. 
		    return (-9, undef, undef, undef, undef, undef, undef, undef, undef, undef);
		}
	    }

		# (1) Examine whether or not the incomplete moves of the "1st" block of the 2nd-moved c-block are all due to the "last" block of the 1st-moved c-block. #

	    my $indices_seqs_incomplete_mvs0 = $set_incomplete1[0]->[3];

	    if ($le_shift2>0) { # When the 2nd move is undone, the left-end of the relevant c-block will move to the left. #

		unless ($re_shift1 < $le_shift2) {
			# In this case, it is difficult to incorporate the remaining 2nd move into the 1st move. #
		    return (-10, undef, undef, undef, undef, undef, undef, undef, undef, undef);
		}

		my $frwk_le2 = $frwk_cb_2nd_mvd->[0]; # Left-end (i.e., the "1st") framework of the 2nd-moved c-block. #
		my $frwk_re1 = $frwk_cb_1st_mvd->[$ct_constituents1-1]; # Right-end (i.e., the "last") framework of the 1st-moved c-block. #
		my $if_can_be_incorporated = 1;
		my $if_hampered = 0; # ADDED in DEBUGGING (on Oct 17 (Sat), 2015). #

		foreach my $indx_seq (@{$indices_seqs_incomplete_mvs0}) { # Foreach-loop (over sequences whose moves could be incompletely undone). #
		    my ($beg_le2, $end_le2) = @{$frwk_le2->{$indx_seq}};
		    ##my ($beg_le2, $end_le2) = @{$frwk_le2->[$indx_seq]}; # A BUG fixed on Oct 16 (Fri), 2015. #

		    my $shifts_in_seq = $map_shifts_respos_rec_af[$indx_seq]; # A BUG fixed then MOVED FROM BELOW (on Oct 17 (Sat), 2015).

				# A NEW conditional block REPLACING buggy ones (on Oct 17 (Sat), 2015). #

		    my ($beg_re1, $end_re1) = (defined $frwk_re1->{$indx_seq}) ? @{$frwk_re1->{$indx_seq}} : (undef, undef) ;

		    if (!(defined $beg_re1) or ($end_re1 >= $beg_le2) or ($end_re1 < $beg_le2 - $le_shift2) ) {
			# Either the 1st framework does not involve the sequence,
			# it is on the wrong side of the 2nd framework,
			# or there is enough roon between the two frameworks.

			# Thus, examine whether anything else has hampered the 2nd move. #

			for (my $s=$beg_le2 - $le_shift2; $s <=$beg_le2-1; $s++) { # Inner for-loop (over sites in the range of the left-end of the 2nd-moved c-block). #
			    unless ($shifts_in_seq->[$s] eq $char_abs) {
				$if_can_be_incorporated = 0;
				last;
			    }
			}				# End of the inner for-loop (over sites in the range of the left-end of the 2nd-moved c-block). #
			$if_can_be_incorporated or last;
			next;
		    }  # END of "if (!(defined $beg_re1) or ($end_re1 >= $beg_le2) or ($end_re1 < $beg_le2 - $le_shift2) ) {...}"

=DISABLED for DEBUGGING (on Oct 17 (Sat), 2015). #

		    unless (defined $frwk_re1->{$indx_seq}) {
		    ##unless (defined $frwk_re1->[$indx_seq]) {...} # A BUG fixed on Oct 16 (Fri), 2015. # 
				# The 1st framework does not involve the sequence, thus the move of its portion in the 2nd framework must have been hampered by something else. #
			$if_can_be_incorporated = 0;
			last;
		    }
		    my ($beg_re1, $end_re1) = @{$frwk_re1->{$indx_seq}};
		    ##my ($beg_re1, $end_re1) = @{$frwk_re1->[$indx_seq]}; # A BUG fixed on Oct 16 (Fri), 2015. #

		    if (($end_re1 >= $beg_le2) or ($end_re1 < $beg_le2 - $le_shift2)) { # The 1st framework is on the wrong side of the 2nd framework, or there is enough roon between the two frameworks, thus something else must have hampered the 2nd move. # 
			$if_can_be_incorporated = 0;
			last;
		    }

=cut # END of "DISABLED for DEBUGGING (on Oct 17 (Sat), 2015)." #

			# Look into the position-shift map,
			# and examine whether or not there is any intervening residue between the two frameworks. #

		    ##my $shifts_in_seq = $map_shifts_respos_intm[$indx_seq]; # A BUG fixed then MOVED UPWARD (on Oct 17 (Sat), 2015). #

		    for (my $s = $end_re1+1; $s < $beg_le2; $s++) { # Inner for-loop (over intervening sites). #
			unless ($shifts_in_seq->[$s] eq $char_abs) {
			    $if_can_be_incorporated = 0;
			    last;
			}
		    }					# End of the inner for-loop (over intervening sites). #
		    $if_can_be_incorporated or last;
		    $if_hampered = 1; # ADDED in DEBUGGING (on Oct 17 (Sat), 2015). #

		}				# End of the foreach-loop (over sequences whose moves could be incompletely undone). #

		($if_can_be_incorporated and $if_hampered) or ( return (-11, undef, undef, undef, undef, undef, undef, undef, undef, undef) );
		##$if_can_be_incorporated or ( return (-11, undef, undef, undef, undef, undef, undef, undef, undef, undef) ); # A BUG fixed (on Oct 17 (Sat), 2015). #

	    } else { # When the 2nd move is undone, the right-end of the relevant c-block will move to the right. #

		unless (-$le_shift1 < -$re_shift2) {
			# In this case, it is difficult to incorporate the remaining 2nd move into the 1st one. #
		    return (-12, undef, undef, undef, undef, undef, undef, undef, undef, undef);
		}

	    		# $updated_frwk_cb_2nd_mvd[$i2] = \%updtd_frwk_bl2; --- It was unnecessary...

		my $frwk_re2 = $frwk_cb_2nd_mvd->[$ct_constituents2-1]; # Right-end (i.e., the "1st") framework of the 2nd-moved c-block. #
		my $frwk_le1 = $frwk_cb_1st_mvd->[0]; # Left-end (i.e., the "last") framework of the 1st-moved c-block. #

		my $if_can_be_incorporated = 1;
		my $if_hampered = 0; # ADDED in DEBUGGING (on Oct 17 (Sat), 2015). # 
		foreach my $indx_seq (@{$indices_seqs_incomplete_mvs0}) { # Foreach-loop (over sequences whose moves could be incompletely undone). #
		    my ($beg_re2, $end_re2) = @{$frwk_re2->{$indx_seq}};
		    ##my ($beg_re2, $end_re2) = @{$frwk_re2->[$indx_seq]}; # A BUG fixed on Oct 16 (Fri), 2015. #

		    my $shifts_in_seq = $map_shifts_respos_rec_af[$indx_seq]; # A BUG fixed then MOVED FROM BELOW (on Oct 17 (Sat), 2015).

				# A NEW conditional block REPLACING buggy ones (on Oct 17 (Sat), 2015). #

		    my ($beg_le1, $end_le1) = (defined $frwk_le1->{$indx_seq}) ? @{$frwk_le1->{$indx_seq}} : (undef, undef) ;

		    if (!(defined $beg_le1) or ($end_re2 >= $beg_le1) or ($end_re2 - $re_shift2 < $beg_le1) ) {
			# Either the 1st framework does not involve the sequence,
			# it is on the wrong side of the 2nd framework,
			# or there is enough roon between the two frameworks.

			# Thus, examine whether anything else has hampered the 2nd move. #

			for (my $s=$end_re2+1; $s <=$end_re2-$re_shift2; $s++) { # Inner for-loop (over sites in the range of the right-end of the 2nd-moved c-block). #
			    unless ($shifts_in_seq->[$s] eq $char_abs) {
				$if_can_be_incorporated = 0;
				last;
			    }
			}				# End of the inner for-loop (over sites in the range of the right-end of the 2nd-moved c-block). #
			$if_can_be_incorporated or last;
			next;
		    }  # END of "if (!(defined $beg_le1) or ($end_re2 >= $beg_le1) or ($end_re2 - $re_shift2 < $beg_le1) ) {...}"

=DISABLED for DEBUGGING (on Oct 17 (Sat), 2015). #
		    unless (defined $frwk_le1->{$indx_seq}) {
		    ##unless (defined $frwk_le1->[$indx_seq]) { ...} # A BUG fixed on Oct 16 (Fri), 2015.#
				# The 1st framework does not involve the sequence, thus the move of its portion in the 2nd framework must have been hampered by something else. #
			$if_can_be_incorporated = 0;
			last;
		    }
		    my ($beg_le1, $end_le1) = @{$frwk_le1->{$indx_seq}};
		    ##my ($beg_le1, $end_le1) = @{$frwk_le1->[$indx_seq]}; # A BUG fixed on Oct 16 (Fri), 2015. #

		    if (($end_re2 >= $beg_le1) or ($end_re2 - $re_shift2 < $beg_le1)) { # The 1st framework is on the wrong side of the 2nd framework, or there is enough roon between the two frameworks, thus something else must have hampered the 2nd move. #
			$if_can_be_incorporated = 0;
			last;
		    }
=cut # END of "DISABLED for DEBUGGING (on Oct 17 (Sat), 2015)." #

			# Look into the position-shift map,
			# and examine whether or not there is any intervening residue between the two frameworks. #

		    ##my $shifts_in_seq = $map_shifts_respos_intm[$indx_seq]; # A BUG fixed then MOVED UPWARD (on Oct 17 (Sat), 2015).

		    for (my $s = $end_re2+1; $s < $beg_le1; $s++) { # Inner for-loop (over intervening sites). #
			unless ($shifts_in_seq->[$s] eq $char_abs) {
			    $if_can_be_incorporated = 0;
			    last;
			}
		    }					# End of the inner for-loop (over intervening sites). #
		    $if_can_be_incorporated or last;
		    $if_hampered = 1; # ADDED in DEBUGGING (on Oct 17 (Sat), 2015). #

		}				# End of the foreach-loop (over sequences whose moves could be incompletely undone). #

		($if_can_be_incorporated and $if_hampered) or ( return (-13, undef, undef, undef, undef, undef, undef, undef, undef, undef) );
		##$if_can_be_incorporated or ( return (-13, undef, undef, undef, undef, undef, undef, undef, undef, undef) ); # A BUG fixed (on Oct 17 (Sat), 2015). #

	    } # END of "if ($le_shift2>0) {...} else {...}"


		# (2) Incorporate the remaining move of the 2nd-moved c-block into the complete move of the 1st-moved c-block. #

	    my @bw_hard_linking2_incpd = ();
	    my @frwk_cb_2nd_mvd_incpd = ();
	    my @info_constituents2_incpd = ();
	    for (my $ii2=0; $ii2<$ct_constituents2; $ii2++) { # 1st for-loop (over the 2nd-moved blocks). #

		my $info_incomplete = $set_incomplete1[$ii2];
		my $i2 = $ordered_constituents2[$ii2];
		my ($full_shift2, $beg_bl2, $end_bl2) = @{$info_constituents2->[$i2]};
		my $actual_shift2 = (defined $info_incomplete) ? $info_incomplete->[1] : $full_shift2;

		$info_constituents2_incpd[$i2] = [$full_shift2 - $actual_shift2, $beg_bl2-$actual_shift2, $end_bl2-$actual_shift2];

		my $frwk2 = $frwk_cb_2nd_mvd->[$i2];
		my %frwk2_incpd ;
		foreach my $indx_seq (keys %{$frwk2}) { my ($beg, $end) = @{$frwk2->{$indx_seq}}; $frwk2_incpd{$indx_seq} = [$beg-$actual_shift2, $end-$actual_shift2]; }
		$frwk_cb_2nd_mvd_incpd[$i2] = \%frwk2_incpd;

		my $hard_linking_to_bl = $bw_hard_linking2->[$i2];
		my @cp_hard_linking_to_bl = ();
		foreach my $info_bl (@{$hard_linking_to_bl}) { my ($indx_cb2, $bl2) = @{$info_bl}; push @cp_hard_linking_to_bl, [$indx_cb2, $bl2]; }
		push @bw_hard_linking2_incpd, \@cp_hard_linking_to_bl;
	    }					# End of the 1st for-loop (over the 2nd-moved blocks). #

	    $ct_constituents1 += $ct_constituents2;

	    if ($le_shift2>0) { # The spatial order is: {1st mvd} {2nd mvd}.

		@{$constituent_blks1_af} = @{$constituent_blks1} = (@{$constituent_blks1}, @{$constituent_blks2});
		@{$info_constituents1} = (@{$info_constituents1}, @info_constituents2_incpd);
		@{$frwk_cb_1st_mvd} = (@{$frwk_cb_1st_mvd}, @frwk_cb_2nd_mvd_incpd);
		@{$bw_hard_linking1} = (@{$bw_hard_linking1}, @bw_hard_linking2_incpd);

			# ADDED on Oct 15 (Thu), 2015. #
			# => MODIFIED on Oct 16 (Fri), 2015. #
		@cbs_hard_linking1 = (@cbs_hard_linking1, @cbs_hard_linking2);
		##@{$hard_linking1} = (@{$hard_linking1}, @{$hard_linking2});
		##(@{$hard_linking1}>0) and ($cb2hard_linking->{$indx_cb_1st_mvd} = $hard_linking1);

		@{$merger_types1_af} = @{$merger_types1} = (@{$merger_types1}, '1st+2nd->1st', @{$merger_types2});

	    } else { # The spatial order is: {2nd mvd} {1st mvd}.

		@{$constituent_blks1_af} = @{$constituent_blks1} = (@{$constituent_blks2}, @{$constituent_blks1});
		@{$info_constituents1} = (@info_constituents2_incpd, @{$info_constituents1});
		@{$frwk_cb_1st_mvd} = (@frwk_cb_2nd_mvd_incpd, @{$frwk_cb_1st_mvd});
		@{$bw_hard_linking1} = (@bw_hard_linking2_incpd, @{$bw_hard_linking1});

			# ADDED on Oct 15 (Thu), 2015. #
			# => MODIFIED on Oct 16 (Fri), 2015. #
		@cbs_hard_linking1 = (@cbs_hard_linking2, @cbs_hard_linking1);
		##@{$hard_linking1} = (@{$hard_linking2}, @{$hard_linking1});
		##(@{$hard_linking1}>0) and ($cb2hard_linking->{$indx_cb_1st_mvd} = $hard_linking1);

		@{$merger_types1_af} = @{$merger_types1} = (@{$merger_types2}, '2nd+1st->1st', @{$merger_types1});

	    }   # END of  "if ($le_shift2>0) {...} else {...}"


	} # END of "if ($indx_cannot_undo1>0) {...} elsif ($indx_cannot_undo1 == 0) {...} elsif (@set_incomplete1>0) {...}"


		# Before the final examination, remove unmoving blocks from both ends of each of the 1st-moved and 2nd-moved c-blocks. #
		# (ADDED on IMPROVing the subroutine 'undo_mvd_latter_of_cblocks_pair' on Oct 18 (Sun), 2015.)

	unless ( rm_zero_shifts_at_ends (@set_blocks_shift_respos_intm, 0, @{$constituent_blks1}, @{$constituent_blks1_af}, @{$info_constituents1}, @{$frwk_cb_1st_mvd}, @{$merger_types1}, @{$merger_types1_af}, @{$bw_hard_linking1}, 0, $ct_constituents1, @cbs_hard_linking1) ) {
	    print STDOUT "[ERROR 'undo_mvd_latter_of_cblocks_pair]] Subroutine 'rm_zero_shifts_at_ends' FAILED on the 1st-moved c-block: $! !\n\n";
	    return (-1008, undef, undef, undef, undef, undef, undef, undef, undef, undef); # MODIFIED on Oct 25 (Sun), 2015. #
	    ##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}

	unless ( rm_zero_shifts_at_ends (@set_blocks_shift_respos_af, 0, @{$constituent_blks2}, @{$constituent_blks2_af}, @{$info_constituents2}, @{$frwk_cb_2nd_mvd}, @{$merger_types2}, @{$merger_types2_af}, @{$bw_hard_linking2}, 0, $ct_constituents2, @cbs_hard_linking2) ) {
	    print STDOUT "[ERROR 'undo_mvd_latter_of_cblocks_pair]] Subroutine 'rm_zero_shifts_at_ends' FAILED on the 2nd-moved c-block: $! !\n\n";
	    return (-1009, undef, undef, undef, undef, undef, undef, undef, undef, undef); # MODIFIED on Oct 25 (Sun), 2015. #
	    ##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}


		# END of "(ADDED on IMPROVing the subroutine 'undo_mvd_latter_of_cblocks_pair' on Oct 18 (Sun), 2015.)"


		# Check whether the relevant c-blocks became empty or not. #
		# (ADDED in DEBUGGING on Oct 24 (Sat), 2015.) #

	unless (@{$constituent_blks2_af}>0) { # 2nd-moved c-block in the reconstructed MSA (after its own move).

	    print STDOUT "[WARNING 'undo_mvd_latter_of_cblocks_pair'] An UNEXPECTED situation: The new 2nd moved c-block (index: $indx_cb_2nd_mvd) became EMPTY in the reconstructed MSA: $! !\n\n";
	    return (-103, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}

	unless (@{$constituent_blks1}>0) { # 1st-moved c-block in the intermediate MSA (after its own move).

	    print STDOUT "[WARNING 'undo_mvd_latter_of_cblocks_pair'] An UNEXPECTED situation: The new 1st moved c-block (index: $indx_cb_1st_mvd) became EMPTY in the intermediate MSA: $! !\n\n";
	    return (-105, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}

		# END of "(ADDED in DEBUGGING on Oct 24 (Sat), 2015.)" #


		# Finally, examine whether or not the move of the (possibly modified) 1st moved c-block can be undone completely. #


			# Prepare auxiliary variables. #

	my @map_shifts_respos_intm_intm = (); # Initially, it is just the intermediate map (from reference to intermediate).
	my @map_shifts_respos_intm_af = (); # Initially, it is just the self-map from the intermediate to itself.

	foreach my $shifts_in_seq (@map_shifts_respos_intm) { # 2nd trailing preliminary foreach-loop (over sequences in the intermediate position-shift map). #
	    my @cp1_shifts_in_seq = @{$shifts_in_seq};
	    push @map_shifts_respos_intm_intm, \@cp1_shifts_in_seq;

	    my @cp2_shifts_in_seq = ();
	    foreach my $shift (@cp1_shifts_in_seq) { push @cp2_shifts_in_seq, (($shift eq $char_abs) ? $char_abs : 0); }
	    push @map_shifts_respos_intm_af, \@cp2_shifts_in_seq;

	}				# End of the 2nd trailing preliminary foreach-loop (over sequences in the intermediate position-shift map). #


	$le_shift1 = $info_constituents1->[0][0];
	$re_shift1 = $info_constituents1->[$ct_constituents1-1][0];

	if (($le_shift1<=0) and ($re_shift1>=0)) {

			# MODIFIED in DEBUGGING on Oct 24 (Sat), 2015. #
	    print STDOUT "[WARNING 'undo_mvd_latter_of_cblocks_pair'] An UNEXPECTED situation: The (new) 1st moved c-block (index: $indx_cb_1st_mvd) has a NON-POSITIVE left-end shift (=$le_shift1) and a NON-NEGATIVE right-end shift (=$re_shift1): $! !\n\n";
	    return (-102, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    ##print STDOUT "[ERROR 'undo_mvd_latter_of_cblocks_pair'] An UNEXPECTED situation: The (new) 1st moved c-block (index: $indx_cb_1st_mvd) has a NON-POSITIVE left-end shift (=$le_shift1) and a NON-NEGATIVE right-end shift (=$re_shift1): $! !\n\n";
	    ##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}

	my @ordered_constituents1 = (0 .. $ct_constituents1-1);

	unless ($le_shift1>0) { @ordered_constituents1 = reverse (@ordered_constituents1); }


			# Now that the auxiliary variables are prepared, perform the final examination. #

	for (my $ii1=0; $ii1<$ct_constituents1; $ii1++) { # 3rd main outer for-loop (over constituent blocks of the 1st-moved c-block). #
	    my $i1 = $ordered_constituents1[$ii1];

	    my $bl1 = $constituent_blks1->[$i1];
	    my $frwk_bl1 = $frwk_cb_1st_mvd->[$i1];
	    my ($shift_bl1, $beg_bl1, $end_bl1) = @{$info_constituents1->[$i1]};
	    my $set_hard_linking_bl1 = $bw_hard_linking1->[$i1];

	    my ($actual_shift_bl1, $indices_seqs_incomplete_mvs) = scout_possible_undoing ($shift_bl1, %{$frwk_bl1}, @map_shifts_respos_intm_intm, $CHAR_ABS);
	    unless (defined $actual_shift_bl1) {
		print STDOUT "[ERROR 'undo_mv_latter_of_cblock_pair'] Subroutine 'scout_possible_undoing' FAILED on the $i1 th constituent block (index: $bl1) of the (new) 1st-moved c-block: $! !\n\n";
		return (-1010, undef, undef, undef, undef, undef, undef, undef, undef, undef); # MODIFOED on Oct 25 (Sun), 2015. #
		##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }

	    if ($actual_shift_bl1 != $shift_bl1) { # This means that the (new) 1st-moved c-block cannot be completely undone!! #
		return (-14, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }

	    my $remaining_shift = $shift_bl1 - $actual_shift_bl1;

		## Actually undo the 1st move: Reflect it on the position-shift maps. #

			# MODIFIED in DEBUGGING on Oct 25 (Sun), 2015. #
	    unless ( undo_sgl_bw_fw_move_on_2maps_2 (%{$frwk_bl1}, $char_abs, $shift_bl1, $actual_shift_bl1, @map_shifts_respos_intm_intm, @map_shifts_respos_intm_af) ) {
	    ##unless ( undo_sgl_bw_fw_move_on_2maps (%{$frwk_bl1}, $char_abs, $shift_bl1, $actual_shift_bl1, @map_shifts_respos_intm_intm, @map_shifts_respos_intm_af) ) {...}
		print STDOUT "[ERROR 'undo_mv_latter_of_cblock_pair'] Subroutine 'undo_sgl_bw_fw_move_on_2maps_2' FAILED on the $i1 th constutuent block (index: $bl1) of the (new) 1st-moved c-block: $! !\n\n";
		return (-1011, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef); # MODIFIED on Oct 25 (Sun), 2015. #
		##return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }

	}			# End of the 3rd main outer for-loop (over constituent blocks of the 1st-moved c-block). #

		## Put the Finishing Touch: Update the boundaries of the relevant c-blocks. ##
		## (ADDED on Oct 15 (Thu), 2015.) ##

	foreach my $pair ([\@set_composite_blocks_intm, \@set_blocks_shift_respos_intm],
			[\@set_composite_blocks_af, \@set_blocks_shift_respos_af]) { # Final outer foreach-loop. #

	    my ($set_cblocks, $set_shift_blocks) = @{$pair};

	    foreach my $indx_cb ($indx_cb_1st_mvd, $indx_cb_2nd_mvd, @cbs_hard_linking1, @cbs_hard_linking2) { # Final middle foreach-loop (over relevant indices). # MODIFIED on Oct 16 (Fri), 2015. #
	    ##foreach my $indx_cb ($indx_cb_1st_mvd, $indx_cb_2nd_mvd) {...} # Final middle foreach-loop (over relevant indices). #
		my $info_cb = $set_cblocks->[$indx_cb];
		my $constituents = $info_cb->[4];
		my $ct_constituents = @{$constituents};

			# Initialize the boundaries. #
		my ($beg_cb, $end_cb) = @{$set_shift_blocks->[$constituents->[0]]}[1,2];

		for (my $i = 1; $i<$ct_constituents; $i++) { # Final inner for-loop (over constituent blocks). #
		    my ($beg_bl, $end_bl) = @{$set_shift_blocks->[$constituents->[$i]]}[1,2];
			# Update the boundaries if necessary #
		    ($beg_bl < $beg_cb) and ($beg_cb = $beg_bl);
		    ($end_cb < $end_bl) and ($end_cb = $end_bl);
		}				# End of the final inner for-loop (over constituent blocks). #

		@{$info_cb}[0, 1] = ($beg_cb, $end_cb);

	    }					# End of the final middle foreach-loop (over relevant indices). #
	}				# End of the final outer foreach-loop. #

   (print STDOUT "  <<< END of the Subroutine 'undo_mv_latter_of_cblock_pair' >>>\n\n") if ($TEST_LEVEL_umlp>0);

			# MODIFIED on Oct 16 (Fri), 2015. #
	return (1, \@map_shifts_respos_intm, \@map_shifts_respos_rec_af, $baseline_shift_intm, \@set_composite_blocks_intm, \@set_blocks_shift_respos_intm, \@set_composite_blocks_af, \@set_blocks_shift_respos_af, \@cbs_hard_linking1, \@cbs_hard_linking2);
	##return (1, \@map_shifts_respos_intm, \@map_shifts_respos_rec_af, $baseline_shift_intm, \@set_composite_blocks_intm, \@set_blocks_shift_respos_intm, \@set_composite_blocks_af, \@set_blocks_shift_respos_af);

}		# End of the subroutine 'undo_mv_latter_of_cblock_pair'. #


# sub_enum_end_sequences:
#	This satellite subroutine of 'enum_end_sequences' enumerates sequences involved with a single c-block
#	at a specified pair of end-points.
#	[CREATED on Oct 10 (Sat), 2015]
#
# INPUT: ($beg, $end, $indx_cblk, \@set_composite_blocks, \@set_shift_blocks, \%$node2ch, \%node2seq_indx), where
#
#	$beg/$end is the specified left/right-end points;
#
#	@{$set_composite_blocks[$indx_cblk]} =  ($beg_cb, $end_cb, $nodeid_mrca_cb, \%br2change_cb, \@constituent_blks, \@involved_seqs_cb, \@merger_types) is for the $indx_cblk th composite-block of residue-position shifts;
#
#	@{$set_shift_blocks[$bl]} = ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes) is for the $bl th position-shift block, where
#	$beg/end_blk is the beginning/ending site (0-based full-closed) of the block,
#	@seqs_involved stores the (0-based) indices of the sequences involved with the block,
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
#	$ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
#	 @indices_src_classes stores the indices of the classes that are used to construct the block;
#
#	%node2ch = ($node_id => \@child_nodes, ...) for internal nodes;
#	%node2seq_indx = ($node_id => $indx_seq (in the MSA), ...) for external nodes.
#
# OUTPUT: (\@seqs_left_end, \@seqs_right_end), where
#	@seqs_left/right_end stores the indices of sequences involved in the subject c-block
#	at the block's left/right-end. (Each list may be redundant.)
#
sub sub_enum_end_sequences ($$$\@\@\%\%) { 

	my ($beg, $end, $indx_cblk, $set_composite_blocks, $set_shift_blocks, $node2ch, $node2seq_indx) = @_;

	my ($beg_cb, $end_cb, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$set_composite_blocks->[$indx_cblk]};

	my @seqs_left_end = ();

    if ($beg_cb <= $beg) {
	foreach my $bl (@{$constituent_blks}) { #  1st main foreach-loop (over constituent shift-blocks). #

 	    my ($shift, $beg_bl, $end_bl, $node_mrca_bl, $depth_mrca_bl, $involved_seqs_bl, $br2change_bl, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_shift_blocks->[$bl]};

	    ($beg_bl == $beg) or next;

	    my @excl_seqs = ();

	    if ((defined $excl_sbblocks) and (@{$excl_sbblocks}>0)) {

		foreach my $info_sbblk (@{$excl_sbblocks}) { # Inner foreach-loop (over excluded sub-blocks). #

		    my ($beg_sbbl, $end_sbbl, $mrca_sbbl) = @{$info_sbblk};

		    ($beg_sbbl == $beg) or next;

		    my $excl_seqs = fetch_ext_offsprings_nr ($mrca_sbbl, %{$node2ch});

		    foreach my $nodeid (@{$excl_seqs}) { push @excl_seqs, $node2seq_indx->{$nodeid}; }

		}				# End of the inner foreach-loop (over excluded sub-blocks). #
	    }

	    my ($excluded, $truly_involved, $extra_excl) = (@excl_seqs>0) ?
			diff_nmarray (@{$involved_seqs_bl}, @excl_seqs) : ([], $involved_seqs_bl, []);

	    push @seqs_left_end, @{$truly_involved};

	}				# End of the 1st main foreach-loop (over constituent shift-blocks). #

    } # END of "if ($beg_cb <= $beg) {...}"


	my @seqs_right_end = ();

    if ($end <= $end_cb) {

	foreach my $bl (@{$constituent_blks}) { #  2nd main foreach-loop (over constituent shift-blocks). #

 	    my ($shift, $beg_bl, $end_bl, $node_mrca_bl, $depth_mrca_bl, $involved_seqs_bl, $br2change_bl, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_shift_blocks->[$bl]};

	    ($end_bl == $end) or next;

	    my @excl_seqs = ();

	    if ((defined $excl_sbblocks) and (@{$excl_sbblocks}>0)) {

		foreach my $info_sbblk (@{$excl_sbblocks}) { # Inner foreach-loop (over excluded sub-blocks). #

		    my ($beg_sbbl, $end_sbbl, $mrca_sbbl) = @{$info_sbblk};

		    ($end_sbbl == $end) or next;

		    my $excl_seqs = fetch_ext_offsprings_nr ($mrca_sbbl, %{$node2ch});

		    foreach my $nodeid (@{$excl_seqs}) { push @excl_seqs, $node2seq_indx->{$nodeid}; }

		}				# End of the inner foreach-loop (over excluded sub-blocks). #
	    }

	    my ($excluded, $truly_involved, $extra_excl) = (@excl_seqs>0) ?
			diff_nmarray (@{$involved_seqs_bl}, @excl_seqs) : ([], $involved_seqs_bl, []);

	    push @seqs_right_end, @{$truly_involved};

	}				# End of the 2nd main foreach-loop (over constituent shift-blocks). #

    } # END of "if ($end <= $end_cb) {...}"


	return (\@seqs_left_end, \@seqs_right_end);

}	# End of the subroutine 'sub_enum_end_sequences'. #


# enum_end_sequences:
#	This subroutine enumerates sequences at the left-end and the right-end of a composite-block of residue-position shifts.
#	[CREATION STARTED on Oct 9 (Fri), 2015, ENDED on Oct 10 (Sat), 2015.]
#
# INPUT: ($indx_cblk, \@set_composite_blocks, \@set_shift_blocks, \%cb2hard_linking, \%node2ch, \%node2seq_indx), where
#
#	@{$set_composite_blocks[$indx_cblk]} =  ($beg_cb, $end_cb, $nodeid_mrca_cb, \%br2change_cb, \@constituent_blks, \@involved_seqs_cb, \@merger_types) is for the $indx_cblk th composite-block of residue-position shifts;
#
#	@{$set_shift_blocks[$bl]} = ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes) is for the $bl th position-shift block, where
#	$beg/end_blk is the beginning/ending site (0-based full-closed) of the block,
#	@seqs_involved stores the (0-based) indices of the sequences involved with the block,
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
#	$ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
#	 @indices_src_classes stores the indices of the classes that are used to construct the block;
#
#	%cb2hard_linking = ($indx_cb (for non-skipped) => [list of indices of (skipped and complex) composite-blocks that "hard-link" the $cb], ...);
#
#	%node2ch = ($node_id => \@child_nodes, ...) for internal nodes;
#	%node2seq_indx = ($node_id => $indx_seq (in the MSA), ...) for external nodes.
#
# OUTPUT: (\@seqs_left_end, \@seqs_right_end), where
#	@seqs_left/right_end stores the indices of sequences involved in the subject c-block
#	at the block's left/right-end.
#
sub enum_end_sequences ($\@\@\%\%\%) {

	my ($indx_cblk, $set_composite_blocks, $set_shift_blocks, $cb2hard_linking, $node2ch, $node2seq_indx) = @_;

	my ($beg_cb, $end_cb, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$set_composite_blocks->[$indx_cblk]};

	my ($seqs_left_end, $seqs_right_end) = sub_enum_end_sequences ($beg_cb, $end_cb, $indx_cblk, @{$set_composite_blocks}, @{$set_shift_blocks}, %{$node2ch}, %{$node2seq_indx});

	unless (defined $seqs_left_end) {
	    print STDOUT "[ERROR 'enum_end_sequences'] Satellite subroutine 'sub_enum_end_sequences' FAILED on the $indx_cblk th c-block: $! !\n\n";
	    return (undef, undef);
	}

	my $hard_linking = $cb2hard_linking->{$indx_cblk};

	if (defined $hard_linking) {
	    foreach my $cb2 (@{$hard_linking}) { # Foreach-loop (over c-blocks hard-linking $indx_cblk). #

		my ($seqs_left_end2, $seqs_right_end2) = sub_enum_end_sequences ($beg_cb, $end_cb, $cb2, @{$set_composite_blocks}, @{$set_shift_blocks}, %{$node2ch}, %{$node2seq_indx});
		unless (defined $seqs_left_end2) {
		    print STDOUT "[ERROR 'enum_end_sequences'] Satellite subroutine 'sub_enum_end_sequences' FAILED on the $cb2 th c-block (, which hard-links the $indx_cblk th c-block): $! !\n\n";
		    return (undef, undef);
		}

		(@{$seqs_left_end2}>0)  and (push @{$seqs_left_end}, @{$seqs_left_end2});
		(@{$seqs_right_end2}>0) and (push @{$seqs_right_end}, @{$seqs_right_end2});

	    }				# End of the foreach-loop (over c-blocks hard-linking $indx_cblk). #

	} # END of "if (defined $hard_linking) {...}"

	my $nr_seqs_left_end = nr_nmarray (@{$seqs_left_end});
	my $nr_seqs_right_end = nr_nmarray (@{$seqs_right_end});

	return ($nr_seqs_left_end, $nr_seqs_right_end);

}	# End of the subroutine 'enum_end_sequences'. #


# cal_init_psm_cands_for_sgl_msa:
#	This subroutine calculates the initial candidate of the parsimonious scenario(s) for each gap-block cluster (i.e., gapped segment)
#	of a single input MSA.
#	It also outputs the (non-redundant) list of gap-patterns and the information on their Dollo parsimonious scenarios.
#	NOTE that the subroutine eats up the input MSA.
#	It was created by modifying the subroutine 'cal_init_psm_cands_for_twin_msas'.
#	[CREATED on Oct 9 (Fri), 2015]
#
# cal_init_psm_cands_for_twin_msas:
#	This subroutine calculates the initial candidate of the parsimonious scenario(s) for each gap-block cluster (i.e., gapped segment)
#	of the reference and reconstructed MSAs,
#	it also outputs the (non-redundant) list of gap-patterns and the informaion on their Dollo parsimonious scenarios.
#	NOTE that the subroutine eats up the input MSAs.
#	(CREATED on Oct 19 (Sun), 2014)
#
# INPUT: (\%node2ch, \%node2pa, \%node2depth, \%node2ct_exoffs, \%name2id, \@seqnames, \@in_msa, $char_abs, $char_pres, \@info_gpatterns0), where
#	%node2ch = ($node_id => \@child_node_ids, ...) is for the all interior nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) is for all the nodes except the top node;
#
#	%node2depth = ($node_id => $absolute_depth, ...) is for all the nodes;
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) is for all the nodes;
#
#	%name2id = ($node_name => $node_id, ...) is for the exterior nodes and the top node (with $node_name = $TOP_NODE);
#	$seqnames[$k] is the name of the $k th sequence in the input MSA (reference/reconstructed)
#	(NOTE that these names must exactly match those used in %name2id);
#
#	${$in_msa$k]} is the $k th sequence embedded in the input MSA;
#
#	$char_pres/$char_abs is the character (or symbol) representing the 'presence'/'absence' state;
#
#	@info_gpatterns0 = (\@set_gpatterns0, \@set_node2dparsstat0, \@set_br2change0) gives information on the gap-patterns found in the reference & reconstructed MSAs, and the Dollo parsimonious scenarios yielding these gap-patterns, where
#	${$set_gpatterns0[$indx_gp]} is the string representing the $ndx_gp th gap-pattern in the input MSA, created by concatenating $CHAR_PRES/ABS in the sequences;
#	%{$set_node2dparsstat0[$indx_gp]} = ($node_id => [$state, $ct_pres_ext], ...) is for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern;
#	%{$set_br2change0[$indx_gp]} = ($branch_id => [$state_ue, $state_le], ...) for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern.
#
# OUTPUT: (\@info_init_psm_scn, \@info_gpatterns), where
#
#	@info_init_psm_scn = (\@set_cls_gblocks, \@init_totcts_indels_for_cls, \@init_psm_scn_for_cls) gives information on the initial candidate of the parsimonious scenario giving rise to the reference/reconstructed MSA gap configuration, where
#	
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$init_totcts_indels_for_cls[$cl] is the total number of insertion/deletion events in the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA,
#	
#	@{$init_psm_scn_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	@info_gpatterns =  (\@set_gpatterns, \@set_node2dparsstat, \@set_br2change) gives information on the gap-patterns found in the input MSA, and the Dollo parsimonious scenarios yielding these gap-patterns, (on top of those initially given,) where
#
#	@set_gpatterns follows the same format as @info_gpatterns0 in the input,
#	@set_node2dparsstat follows the same format as @set_node2dparsstat0 in the input,
#	@set_br2change follows the same format as @set_br2change0 in the input.
#
##sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$) { ....}
##	my ($node2ch, $node2pa, $node2depth, $node2ct_exoffs, $name2id, $seqnames, $in_msa, $char_abs, $char_pres) = @_;
#
#		... MODIFIED on Oct 14 (Wed), 2015...
#
sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$\@) {

	my ($node2ch, $node2pa, $node2depth, $node2ct_exoffs, $name2id, $seqnames, $in_msa, $char_abs, $char_pres, $info_gpatterns0) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ct_seqs = @{$seqnames};

    (print STDOUT "<< Subroutine 'cal_init_psm_cands_for_sgl_msa' (Top node ID: $topnode_id\; #{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL_cipc2>0);

		# Main Process 1: Extract gapped segments and their constituent gap-pattern blocks from the input MSA. #
	

			# MODIFIED on Oct 14 (Mon), 2015. #
	my ($set_gpatterns0, $set_node2dparsstat0, $set_br2change0) = @{$info_gpatterns0}; # Initial information of the gap-patterns and their Dollo parsimonious scenarios. #

	my %gpattern2indx;
	my $ct_gpatterns = my $ct_gpatterns0 = @{$set_gpatterns0};
	for (my $i=0; $i<$ct_gpatterns0; $i++) { $gpattern2indx{${$set_gpatterns0->[$i]}} = $i; }
	##my $ct_gpatterns = 0;

	my $set_cls_gblocks = extract_cls_gblocks (@{$in_msa}, $char_pres, $char_abs, %gpattern2indx, $ct_gpatterns) ;
	unless (defined $set_cls_gblocks) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'extract_cls_gblocks' FAILED: $! !\n\n";
	    return (undef, undef);
	}
	my $ct_cls = @{$set_cls_gblocks};
    (print STDOUT "\n --- 'extract_cls_gblocks' Ended (#{gap clusters} = $ct_cls) ...\n\n") if ($TEST_LEVEL_cipc2>1);

	# substr (${$sec_msa[$k]}, $indx_gp, 1) is the presence/absence state of the $k th sequence in the $indx_gp th gap pattern.
	# ${$set_gpatterns[$indx_gp]} is the string made by concatenating the states in the $indx_gp th gap pattern.
	#
	my ($sec_msa, $set_gpatterns) = msa_rep_gpatterns (%gpattern2indx, length ($char_abs), $ct_seqs);
	unless (defined $sec_msa) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'msa_rep_gpatterns' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
    (print STDOUT "\n --- 'msa_rep_gpatterns' Ended ...\n\n") if ($TEST_LEVEL_cipc2>1);


		# Scan the information of gapped segments for screening those with gapped segment(s) and/or gap(s) that are TOO LONG. #

	foreach my $cls_gblk (@{$set_cls_gblocks}) { # Screening foreach-loop (over gapped segments in the input MSA). #

	    my $ct_gblks = @{$cls_gblk};
	    my $start_cls = $cls_gblk->[0][2];
	    my $end_cls = $cls_gblk->[$ct_gblks-1][3];
	    my $cl_len = $end_cls - $start_cls+1;

	    unless ($cl_len <= $UPPER_BOUND_ANAL_LEN_GBCLS) {
		print STDOUT "[WARNING 'cal_init_psm_cands_for_sgl_msa'] Found a cluster of length $cl_len, which is LONGER THAN the upper-bound, $UPPER_BOUND_ANAL_LEN_GBCLS... I will skip analyzing this input MSA: $! !\n\n";
		return ([], [], []);
	    }

	    my $list_long_gaps = scan_gbcls_for_long_gaps (@{$cls_gblk}, @{$set_gpatterns}, $UPPER_BOUND_ANAL_GAPLEN, $char_abs);
	    unless (defined $list_long_gaps) {
		print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'scan_gbcls_for_long_gaps' FAILED: $! !\n\n";
		return (undef, undef, undef);
	    }
	    unless (@{$list_long_gaps} == 0) {
		my $ct_long_gaps = @{$list_long_gaps};
		print STDOUT "[WARNING 'cal_init_psm_cands_for_sgl_msa'] Found $ct_long_gaps gaps longer than $UPPER_BOUND_ANAL_GAPLEN. So, I will skip analyzing this input MSA: $! !\n\n"; 
		return ([], [], []);
	    }

	}		# End of the screening foreach-loop (over gapped segments in reference & reconstructed MSAs). #


		## Main Process 2: Calculate the Dollo parsimonious indel scenario for each gap-pattern. ##

 	my ($set_node2dparsstat, $set_br2change) = ([], []); # ADDED on Oct 14 (Wed), 2015. #
    if ($ct_gpatterns>$ct_gpatterns0) { # ADDED on Oct 14 (Wed), 2015. #

		# MODIFIED on Oct 14 (Wed), 2015. #

	foreach my $sec_seq (@{$sec_msa}) { substr (${$sec_seq}, 0, $ct_gpatterns0 * length ($char_abs), ''); } # Remove the first $ct_gpatterns0 gap-patterns, whose Dollo parsimonious scenarios were already calculated.

	my ($set_node2dparsstat1, $set_br2change1) = map_parsim_states_dollo_msa ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$sec_msa}, $char_abs, $char_pres, 1) ;
	##my ($set_node2dparsstat, $set_br2change) = map_parsim_states_dollo_msa ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$sec_msa}, $char_abs, $char_pres, 1) ;
	unless (defined $set_node2dparsstat1) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'map_parsim_states_dollo_msa' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
    (print STDOUT " --- 'map_parsim_states_dollo_msa' Ended ...\n\n") if ($TEST_LEVEL_cipc2>1);

		# ADDED on Oct 14 (Wed), 2015. #

	@{$set_node2dparsstat} = (@{$set_node2dparsstat0}, @{$set_node2dparsstat1});
	@{$set_br2change} = (@{$set_br2change0}, @{$set_br2change1});

   } else {
		# In this case, the initial (input) set exhausts the gap-patterns in the MSA. #
		# (ADDED on Oct 14 (Wed), 2015.) #

	$set_node2dparsstat = $set_node2dparsstat0;
	$set_br2change = $set_br2change0;

   } # END of "if ($ct_gpatterns>$ct_gpatterns0) {...} else {...}" # ADDED on Oct 14 (Wed), 2015. #


	my @info_gpatterns = ($set_gpatterns, $set_node2dparsstat, $set_br2change);


		## Main Process 3: Calculate the initial candidate of the parsimonious indel scenario resulting in the gap-configurations of each gapped segment in the input MSA. ##

			# Create a set of all nodes.
	my @set_all_nodes = sort keys %{$node2pa};
	unshift @set_all_nodes, $topnode_id;

			# Initialize the output. #
	my @info_init_psm_scn = ($set_cls_gblocks, [], []);

		# @info_init_psm_scn = (\@set_cls_gblocks, \@init_totcts_indels_for_cls, \@init_psm_scn_for_cls) gives information on the initial candidate of the parsimonious scenario giving rise to the input MSA gap configuration, where

	my $indx_indx_gp = 0; # Index to specify where the index of the gap-pattern is stored.

	foreach my $info_in_out (
		['input MSA', \@info_init_psm_scn]) { # Main outer foreach-loop (for the input MSA). #

	    my ($title_msa, $info_init_psm_scn) = @{$info_in_out};

	    my ($set_cls_gblocks, $init_totcts_indels_for_cls, $init_psm_scn_for_cls) = @{$info_init_psm_scn};

	    my $ct_cls = @{$set_cls_gblocks};

	    for (my $cl = 0; $cl<$ct_cls; $cl++) { # Main inner for-loop (over clusters (i.e., gapped segments) in each MSA. #

		my $blocks_in_cls = $set_cls_gblocks->[$cl];

		my ($ct_indels0, $node2blwise_gstat0, $br2indels0, $list_blank_blks) = create_initial_set_indels_spt_odr ($topnode_id, %{$node2pa}, %{$node2ch}, @set_all_nodes, $char_pres, $char_abs, @{$blocks_in_cls}, $indx_indx_gp, @{$set_node2dparsstat}, @{$set_br2change});
		unless (defined $ct_indels0) {
		    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'create_initial_set_indels' FAILED on the $cl th cluster of '$title_msa': $! !\n\n";
		    return (undef, undef, undef);
		}
    (print STDOUT "\n --- Subroutine 'create_initial_set_indels_spt_odr' Ended (with #{initial indels} = $ct_indels0) ...\n\n") if ($TEST_LEVEL_cipc2>1);

		$init_totcts_indels_for_cls->[$cl] = $ct_indels0;
		$init_psm_scn_for_cls->[$cl] = [$br2indels0, $node2blwise_gstat0];

	# @{$init_psm_scn_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 

	    }				# End of the main inner for-loop (over clusters (i.e., gapped segments) in each MSA. #

	}					# End of the main outer foreach-loop (for the input MSA). #

    (print STDOUT "<< END of the subroutine 'cal_init_psm_cand_for_sgl_msa' />>\n\n") if ($TEST_LEVEL_cipc2>0);

	return (\@info_init_psm_scn, \@info_gpatterns);

}	# End of the subroutine 'cal_init_psm_cands_for_sgl_msa'. #


# classify_error_assoc_w_sgl_cblk:
#	This gigantic subroutine attempts to classify an error associated with a single composite block.
#	The subroutine was created out of the core block of main routines in the predecessor script, "classify_msa_errors_via_mblks.1.pl."
#	(And it replaced the core block.)
#
#	[CREATED on Oct 6 (Tue), 2015]
#
#	[MODIFIED on Oct 21 (Wed), 2015]
#	The subroutine now outputs the penalty value of the error as well.
#
# INPUT: ($indx_cb, \@set_cblks, $shift_le, $shift_re, $rlv_br, \%equiv_branches,
#	    $topnode_id, \%node2ch, \%node2pa, \%br_attr,
#	\@set_indels_std_by_begsite_rec, \@set_indels_std_by_endsite_rec, \@set_indels_std_by_begsite_ref, \@set_indels_std_by_endsite_ref, \%rec_event2ref_event, \%ref_event2rec_event), where
#
#	@{$set_cblks[$indx_cb]} = ($beg_cb_rec, $end_cb_rec, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) is information on the target composite-block;
#
#	$shift_le/re is the position-shift at the left-end/right-end of the target composite-block;
#	$rlv_br is the relevant branch, that is, the "main" branch separating the target composite-block from the rest;
#	%equiv_branches = ($ch_root1 => $ch_root2, $ch_root2 => $ch_root1), where $ch_root1/2 is a child node of the bifurcated root node;
#
#	$topnode_id is the ID of the top node (i.e., root);
#	%node2ch = ($node_id => \@child_nodes, ...) for internal nodes;
#	%node2pa = ($node_id => $parent_node, ...) for all nodes except the root;
#	%br_attr = ($lower_end_node => \@attr, ...) for all branches;
#
#	@{$set_indels_std_by_begsite/endsite_rec/ref[$e]}
#	= ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, \@indices_lin_blk) 
#	is for the $e th indel event (sorted by the beginning/ending site) among those which are involved in the reconstructe/reference MSA;
#
#	%rec_event2ref_event = ("$br_rec:$beg_rec:$end_rec" => "$br_ref:$beg_ref:$end_ref", ...)
#	stores the relationships from the events involved in the reconstructed MSA to their equivalents involved in the reference MSA;
#	%ref_event2rec_event stores the reverse relationships, from the reference-involved events to the reconstructed-involved events.
#
#
# [ OUTPUT: ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec), where ] --- OBSOLETE as of Oct 21 (Wed), 2015.
# OUTPUT: ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty), where
#
#	$type is a string indicating the type of error (= undef if its test is skipped);
#	$reason_skipped is the reason why the test of the composite block is skipped (= undef if its test is performed);
#	@set_cts_invlvd_indels = ($ct_indels_rlv_ref, $ct_indels_rlv_rec, $ct_indels_rltd_ref, $ct_indels_rltd_rec, $ct_indels_other_invlvd_ref, $ct_indels_other_invlvd_rec);
#	@{$indels_invlvd_ref[$f]} = ($branch_id, $beg_site, $end_site, $state_le, \@indices_lin_blk) is for the $f th involved indel in the reference MSA;
#	@{$indels_invlvd_rec[$f]} = ($branch_id, $beg_site, $end_site, $state_le, \@indices_lin_blk) is for the $f th involved indel in the reconstructed MSA.
#
#	$penalty is the value of the penalty incurred by the error. # ADDED on Oct 21 (Wed), 2015. #
#
sub classify_error_assoc_w_sgl_cblk ($\@$$$\%$\%\%\%\@\@\@\@\%\%) {

	my ($indx_cb, $set_cblks, $shift_le, $shift_re, $rlv_br, $equiv_branches,
	    $topnode_id, $node2ch, $node2pa, $br_attr,
	    $set_indels_std_by_begsite_rec, $set_indels_std_by_endsite_rec,
	      $set_indels_std_by_begsite_ref, $set_indels_std_by_endsite_ref,
		$rec_event2ref_event, $ref_event2rec_event) = @_ ;

		# Restore some information hidden in the input variables. #
		# (ADDED on Oct 6 (Tue), 2015.)

	my ($beg_cb_rec, $end_cb_rec, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$set_cblks->[$indx_cb]};


	my $beg_cb_ref = $beg_cb_rec - $shift_le;
	my $end_cb_ref = $end_cb_rec - $shift_re;

	my $cb = $indx_cb; # This saves lots of re-writing the code.
	my %equiv_branches = %{$equiv_branches}; # Same here.

	my $equiv_rlv_br = $equiv_branches{$rlv_br}; # ADDED on Oct 31 (Fri), 2014. #

		# END of "Restore some information hidden in the input variables." #

		# Set output variables. #
		# (ADDED on Oct 6 (Tue), 2015.)

	my ($type, $reason_skipped);

	my @set_cts_invlvd_indels = ();

	my @cblk_wise_cts_invlvd_indels = (); # $cblk_wise_cts_invlvd_indels[$cb] = [$ct_indels_rlv_ref, $ct_indels_rlv_rec, $ct_indels_rltd_ref, $ct_indels_rltd_rec, $ct_indels_other_invlvd_ref, $ct_indels_other_invlvd_rec] for the $cb th composite-block.
	$cblk_wise_cts_invlvd_indels[$cb] = \@set_cts_invlvd_indels; # This saves lots of re-writing the code.

	my @indels_invlvd_ref = my @indels_invlvd_rec = ();

		# END of "Set output variables." #


		# "Downstream-related" & "Upstream-related" branches. #
		# (ADDED on Oct 31 (Fri), 2014.)

	    my @dw_rltd_brs = (defined $node2ch->{$rlv_br}) ? @{$node2ch->{$rlv_br}} : () ;
	    my @up_rltd_brs = ();

	    if (defined $equiv_rlv_br) {
		@up_rltd_brs = (defined $node2ch->{$equiv_rlv_br}) ? @{$node2ch->{$equiv_rlv_br}} : () ; # "Upper-related" branches. #
	    } else {
		my $pa_rlv_br = $node2pa->{$rlv_br};
		(defined $br_attr->{$pa_rlv_br}) and (push @up_rltd_brs, $pa_rlv_br);

		(defined $equiv_branches{$pa_rlv_br}) and (push @up_rltd_brs, $equiv_branches{$pa_rlv_br}); # ADDED on Nov 2 (Sun), 2014. #

		foreach my $sibl (@{$node2ch->{$pa_rlv_br}}) { ($sibl != $rlv_br) and (push @up_rltd_brs, $sibl); }
	    }
		# END of "(ADDED on Oct 31 (Fri), 2014.)"


	    # [ element of @{$indels_lf/rf/le/re/other_ovlp_rec} = [$branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk] ] --- OBSOLETE as of Nov 4 (Tue), 2014.
	    # element of @{$indels_lf/rf/le/re/other_ovlp_rec} = [$branch_id, $beg_site, $end_site, $state_ue, $state_le, $length_indel, \@indices_lin_blk] # MODIFIED on Nov 4 (Tue), 2014.
	    #
	    my ($ct_flnk_ovlp_indels_rec, $indels_lf_rec, $indels_rf_rec, $indels_le_rec, $indels_re_rec, $indels_other_ovlp_rec) = find_ovlp_flanking_indels ($beg_cb_rec, $end_cb_rec, @{$set_indels_std_by_begsite_rec}, @{$set_indels_std_by_endsite_rec}, %{$rec_event2ref_event}, $nodeid_mrca_cb, %{$br2change_cb}, %{$node2pa}, %{$node2ch}, $topnode_id, $CHAR_ABS);
	    unless (defined $ct_flnk_ovlp_indels_rec) {
		print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'find_ovlp_flanking_indels' FAILED on the $cb th block in the reconstructed MSA: $! !\n";
		return (undef, undef, undef, undef, undef);

		##print STDOUT "[ERROR '$program_name'] Subroutine 'find_ovlp_flanking_indels' FAILED on the $cb th composite shift-block for the reconstructed MSA of the $s th segment !\n";
		##print STDERR "[ERROR '$program_name'] $!\n\n";
		##die ;
	    }

	    my ($ct_flnk_ovlp_indels_ref, $indels_lf_ref, $indels_rf_ref, $indels_le_ref, $indels_re_ref, $indels_other_ovlp_ref) = find_ovlp_flanking_indels ($beg_cb_ref, $end_cb_ref, @{$set_indels_std_by_begsite_ref}, @{$set_indels_std_by_endsite_ref}, %{$ref_event2rec_event}, $nodeid_mrca_cb, %{$br2change_cb}, %{$node2pa}, %{$node2ch}, $topnode_id, $CHAR_ABS);
	    unless (defined $ct_flnk_ovlp_indels_ref) {
		print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'find_ovlp_flanking_indels' FAILED on the $cb th block in the reference MSA: $! !\n";
		return (undef, undef, undef, undef, undef);

		##print STDOUT "[ERROR '$program_name'] Subroutine 'find_ovlp_flanking_indels' FAILED on the $cb th composite shift-block for the reference MSA of the $s th segment !\n";
		##print STDERR "[ERROR '$program_name'] $!\n\n";
		##die ;
	    }

    if ($TEST_LEVEL_newtype>1) { # ADDED on Oct 4 (Sat), 2015.

	print STDOUT "\n[[[ Lists of indels flanking or overlapping the subject block ]]]\n\n";

	print STDOUT "[[ Indels resulting in Reference MSA (#{events} = $ct_flnk_ovlp_indels_ref) ]]\n\n";

	print STDOUT "[ Left-flanking (#{events} = ", scalar (@{$indels_lf_ref}), ") ]\n\n";
	if (@{$indels_lf_ref}>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@{$indels_lf_ref}) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}

	print STDOUT "[ Right-flanking (#{events} = ", scalar (@{$indels_rf_ref}), ") ]\n\n";
	if (@{$indels_rf_ref}>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@{$indels_rf_ref}) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}

	print STDOUT "[ Left-ending (#{events} = ", scalar (@{$indels_le_ref}), ") ]\n\n";
	if (@{$indels_le_ref}>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@{$indels_le_ref}) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}

	print STDOUT "[ Right-ending (#{events} = ", scalar (@{$indels_re_ref}), ") ]\n\n";
	if (@{$indels_re_ref}>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@{$indels_re_ref}) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}

	print STDOUT "[ Other overlapping (#{events} = ", scalar (@{$indels_other_ovlp_ref}), ") ]\n\n";
	if (@{$indels_other_ovlp_ref}>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@{$indels_other_ovlp_ref}) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}

	print STDOUT "\n";

	print STDOUT "[[ Indels resulting in Reconstructed MSA (#{events} = $ct_flnk_ovlp_indels_rec) ]]\n\n";

	print STDOUT "[ Left-flanking (#{events} = ", scalar (@{$indels_lf_rec}), ") ]\n\n";
	if (@{$indels_lf_rec}>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@{$indels_lf_rec}) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}

	print STDOUT "[ Right-flanking (#{events} = ", scalar (@{$indels_rf_rec}), ") ]\n\n";
	if (@{$indels_rf_rec}>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@{$indels_rf_rec}) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}

	print STDOUT "[ Left-ending (#{events} = ", scalar (@{$indels_le_rec}), ") ]\n\n";
	if (@{$indels_le_rec}>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@{$indels_le_rec}) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}

	print STDOUT "[ Right-ending (#{events} = ", scalar (@{$indels_re_rec}), ") ]\n\n";
	if (@{$indels_re_rec}>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@{$indels_re_rec}) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}

	print STDOUT "[ Other overlapping (#{events} = ", scalar (@{$indels_other_ovlp_rec}), ") ]\n\n";
	if (@{$indels_other_ovlp_rec}>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@{$indels_other_ovlp_rec}) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}

	print STDOUT "\n";

    } # End of "if ($TEST_LEVEL_newtype>1) { # ADDED on Oct 4 (Sat), 2015. ...}"


	    my @indels_rlv_lf_rec = my @indels_rlv_rf_rec = my @indels_rlv_le_rec = my @indels_rlv_re_rec = my @indels_rlv_other_ovlp_rec = ();
	    my @indels_rltd_lf_rec = my @indels_rltd_rf_rec = my @indels_rltd_le_rec = my @indels_rltd_re_rec = my @indels_rltd_other_ovlp_rec = (); # ADDED on Oct 31 (Fri), 2014. #
	    my %rlv_indels_rec;
	    my %rltd_indels_rec;

	    my @indels_rlv_rec = my @indels_rltd_rec = (); # ADDED on Oct 31 (Fri), 2014. #

	    foreach my $trio ([$indels_lf_rec, \@indels_rlv_lf_rec, \@indels_rltd_lf_rec],
			[$indels_rf_rec, \@indels_rlv_rf_rec, \@indels_rltd_rf_rec],
			[$indels_le_rec, \@indels_rlv_le_rec, \@indels_rltd_le_rec],
			[$indels_re_rec, \@indels_rlv_re_rec, \@indels_rltd_re_rec],
			[$indels_other_ovlp_rec, \@indels_rlv_other_ovlp_rec, \@indels_rltd_other_ovlp_rec]) { # 2nd preliminary inner foreach-loop (over indels (in the reconstructed MSA) of different positional relationships with the composite block). #

		my ($indels_cls, $indels_rlv_cls, $indels_rltd_cls) = @{$trio};
		foreach my $event_info (@{$indels_cls}) { # Inner foreach-loop (over events with a particular positional relationship). #

		    my ($br2, $beg2, $end2) = @{$event_info}[0,1,2];
		    my $label_event = "$br2:$beg2:$end2";

		    if (($br2 == $rlv_br) or ((defined $equiv_rlv_br) and ($br2 == $equiv_rlv_br))) {

			push @{$indels_rlv_cls}, $event_info;

			unless (defined $rlv_indels_rec{$label_event}) { # ADDED  on Oct 31 (Fri), 2014. # 
			    $rlv_indels_rec{$label_event} = 1;
			    push @indels_rlv_rec, $event_info;
			}
			next;
		    }

				# ADDED on Oct 31 (Fri), 2014. # 
		    foreach my $rltd_br (@dw_rltd_brs, @up_rltd_brs) { # Inner-inner foreach-loop (over relatives of the separating branch). #

			($br2 == $rltd_br) or next;

			push @{$indels_rltd_cls}, $event_info;

			unless (defined $rltd_indels_rec{$label_event}) {
			    $rltd_indels_rec{$label_event} = 1;
			    push @indels_rltd_rec, $event_info;
			}
			last;
		    }					# End of the inner-inner foreach-loop (over relatives of the separating branch). #

		}				# End of the inner foreach-loop (over events with a particular positional relationship). #

	    }				# End of the 2nd preliminary foreach-loop (over indels (in the reconstructed MSA) with different positional relationships with the composite block). #

	    my $ct_rlv_indels_rec = scalar (@indels_rlv_rec);
	    ##my $ct_rlv_indels_rec = scalar keys %rlv_indels_rec;

			# ADDED on Oct 31 (Fri), 2014. #
	    foreach my $event_info (@indels_rltd_other_ovlp_rec) { pop @indels_rltd_rec; }
	    my $ct_rltd_indels_rec = scalar (@indels_rltd_rec);
	    ##my $ct_rltd_indels_rec = scalar (keys %rltd_indels_rec);

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
	    @indels_invlvd_rec = (@indels_rlv_rec, @indels_rltd_rec);
	    ##my @indels_invlvd_rec = (@indels_rlv_rec, @indels_rltd_rec);



	    my @indels_rlv_lf_ref = my @indels_rlv_rf_ref = my @indels_rlv_le_ref = my @indels_rlv_re_ref = my @indels_rlv_other_ovlp_ref = ();
	    my @indels_rltd_lf_ref = my @indels_rltd_rf_ref = my @indels_rltd_le_ref = my @indels_rltd_re_ref = my @indels_rltd_other_ovlp_ref = (); # ADDED on Oct 31 (Fri), 2014. #
	    my %rlv_indels_ref;
	    my %rltd_indels_ref;

	    my @indels_rlv_ref = my @indels_rltd_ref = (); # ADDED on Oct 31 (Fri), 2014. #

	    foreach my $trio ([$indels_lf_ref, \@indels_rlv_lf_ref, \@indels_rltd_lf_ref],
			[$indels_rf_ref, \@indels_rlv_rf_ref, \@indels_rltd_rf_ref],
			[$indels_le_ref, \@indels_rlv_le_ref, \@indels_rltd_le_ref],
			[$indels_re_ref, \@indels_rlv_re_ref, \@indels_rltd_re_ref],
			[$indels_other_ovlp_ref, \@indels_rlv_other_ovlp_ref, \@indels_rltd_other_ovlp_ref]) { # 3rd preliminary inner foreach-loop (over indels (in the reference MSA) of different positional relationships with the composite block). #

		my ($indels_cls, $indels_rlv_cls, $indels_rltd_cls) = @{$trio};
		foreach my $event_info (@{$indels_cls}) { # Inner foreach-loop (over events with a particular positional relationship). #

		    my ($br2, $beg2, $end2) = @{$event_info}[0,1,2];
		    my $label_event = "$br2:$beg2:$end2";

		    if (($br2 == $rlv_br) or ((defined $equiv_rlv_br) and ($br2 == $equiv_rlv_br)) ) {

			push @{$indels_rlv_cls}, $event_info;

			unless (defined $rlv_indels_ref{$label_event}) { # ADDED on Oct 31 (Fri), 2014. #
			    $rlv_indels_ref{$label_event} = 1;
			    push @indels_rlv_ref, $event_info;
			}
			next;
		    }

				# ADDED on Oct 31 (Fri), 2014. #
		    foreach my $rltd_br (@dw_rltd_brs, @up_rltd_brs) { # Inner-inner foreach-loop (over relatives of the separating branch). #

			($br2 == $rltd_br) or next;

			push @{$indels_rltd_cls}, $event_info;

			unless (defined $rltd_indels_ref{$label_event}) {
			    $rltd_indels_ref{$label_event} = 1;
			    push @indels_rltd_ref, $event_info;
			}
			last;
		    }					# End of the inner-inner foreach-loop (over relatives of the separating branch). #

		}				# End of the inner foreach-loop (over events with a particular positional relationship). #

	    }				# End of the 3rd preliminary foreach-loop (over indels (in the reference MSA) with different positional relationships with the composite block). #

	    my $ct_rlv_indels_ref = scalar (@indels_rlv_ref);
	    ##my $ct_rlv_indels_ref = scalar keys %rlv_indels_ref;

		# ADDED on Oct 31 (Fri), 2014.
	    foreach my $event_info (@indels_rltd_other_ovlp_ref) { pop @indels_rltd_ref; }
	    my $ct_rltd_indels_ref = scalar (@indels_rltd_ref);
	    ##my $ct_rltd_indels_ref = scalar (keys %rltd_indels_ref);

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
	    @indels_invlvd_ref = (@indels_rlv_ref, @indels_rltd_ref);
	    ##my @indels_invlvd_ref = (@indels_rlv_ref, @indels_rltd_ref);


			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
			#
	    @set_cts_invlvd_indels = ($ct_rlv_indels_ref, $ct_rlv_indels_rec, $ct_rltd_indels_ref, $ct_rltd_indels_rec, 0, 0);
	    ##$cblk_wise_cts_invlvd_indels[$cb] = [$ct_rlv_indels_ref, $ct_rlv_indels_rec, $ct_rltd_indels_ref, $ct_rltd_indels_rec, 0, 0]; # ADDED on Nov 6 (Thu), 2014. #

    if ($TEST_LEVEL_newtype>1) { # ADDED on Oct 3 (Fri), 2015.

	print STDOUT "\n[[[ Lists of involved and surrounding indels ]]]\n\n";

	print STDOUT "[[ Relevant indels in Reference MSA (#{indels} = $ct_rlv_indels_ref) ]]\n\n";
	if ($ct_rlv_indels_ref >0) {

	    print STDOUT "[ Left-flanking (#{events} = ", scalar (@indels_rlv_lf_ref), ") ]\n\n";
	    if (@indels_rlv_lf_ref>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rlv_lf_ref) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }
	    
	    print STDOUT "[ Right-flanking (#{events} = ", scalar (@indels_rlv_rf_ref), ") ]\n\n";
	    if (@indels_rlv_rf_ref>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rlv_rf_ref) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    print STDOUT "[ Left-ending (#{events} = ", scalar (@indels_rlv_le_ref), ") ]\n\n";
	    if (@indels_rlv_le_ref>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rlv_le_ref) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    print STDOUT "[ Right-ending (#{events} = ", scalar (@indels_rlv_re_ref), ") ]\n\n";
	    if (@indels_rlv_re_ref>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rlv_re_ref) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    print STDOUT "[ Other overlapping (#{events} = ", scalar (@indels_rlv_other_ovlp_ref), ") ]\n\n";
	    if (@indels_rlv_other_ovlp_ref>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rlv_other_ovlp_ref) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    ##my @indels_rlv_lf_ref = my @indels_rlv_rf_ref = my @indels_rlv_le_ref = my @indels_rlv_re_ref = my @indels_rlv_other_ovlp_ref = ();
	}
	print STDOUT "\n";

	print STDOUT "[[ Related indels in Reference MSA (#{indels} = $ct_rltd_indels_ref) ]]\n\n";
	if ($ct_rltd_indels_ref>0) {

	    print STDOUT "[ Left-flanking (#{events} = ", scalar (@indels_rltd_lf_ref), ") ]\n\n";
	    if (@indels_rltd_lf_ref>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rltd_lf_ref) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }
	    
	    print STDOUT "[ Right-flanking (#{events} = ", scalar (@indels_rltd_rf_ref), ") ]\n\n";
	    if (@indels_rltd_rf_ref>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rltd_rf_ref) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    print STDOUT "[ Left-ending (#{events} = ", scalar (@indels_rltd_le_ref), ") ]\n\n";
	    if (@indels_rltd_le_ref>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rltd_le_ref) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    print STDOUT "[ Right-ending (#{events} = ", scalar (@indels_rltd_re_ref), ") ]\n\n";
	    if (@indels_rltd_re_ref>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rltd_re_ref) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    ##my @indels_rltd_lf_ref = my @indels_rltd_rf_ref = my @indels_rltd_le_ref = my @indels_rltd_re_ref = my @indels_rltd_other_ovlp_ref = (); # ADDED on Oct 31 (Fri), 2014. #
	}
	print STDOUT "\n";

	print STDOUT "[[ Other overlapping indels in Reference MSA (#{indels} = ", scalar (@indels_rltd_other_ovlp_ref), ") ]]\n\n";
	if (@indels_rltd_other_ovlp_ref>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@indels_rltd_other_ovlp_ref) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}
	print STDOUT "\n";

	print STDOUT "[[ Relevant indels in Reconstructed MSA (#{indels} = $ct_rlv_indels_rec) ]]\n\n";
	if ($ct_rlv_indels_rec>0) {

	    print STDOUT "[ Left-flanking (#{events} = ", scalar (@indels_rlv_lf_rec), ") ]\n\n";
	    if (@indels_rlv_lf_rec>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rlv_lf_rec) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }
	    
	    print STDOUT "[ Right-flanking (#{events} = ", scalar (@indels_rlv_rf_rec), ") ]\n\n";
	    if (@indels_rlv_rf_rec>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rlv_rf_rec) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    print STDOUT "[ Left-ending (#{events} = ", scalar (@indels_rlv_le_rec), ") ]\n\n";
	    if (@indels_rlv_le_rec>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rlv_le_rec) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    print STDOUT "[ Right-ending (#{events} = ", scalar (@indels_rlv_re_rec), ") ]\n\n";
	    if (@indels_rlv_re_rec>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rlv_re_rec) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    print STDOUT "[ Other overlapping (#{events} = ", scalar (@indels_rlv_other_ovlp_rec), ") ]\n\n";
	    if (@indels_rlv_other_ovlp_rec>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rlv_other_ovlp_rec) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }
	}
	print STDOUT "\n";

	print STDOUT "[[ Related indels in Reconstructed MSA (#{indels} = $ct_rltd_indels_rec) ]]\n\n";
	if ($ct_rltd_indels_rec>0) {

	    print STDOUT "[ Left-flanking (#{events} = ", scalar (@indels_rltd_lf_rec), ") ]\n\n";
	    if (@indels_rltd_lf_rec>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rltd_lf_rec) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }
	    
	    print STDOUT "[ Right-flanking (#{events} = ", scalar (@indels_rltd_rf_rec), ") ]\n\n";
	    if (@indels_rltd_rf_rec>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rltd_rf_rec) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    print STDOUT "[ Left-ending (#{events} = ", scalar (@indels_rltd_le_rec), ") ]\n\n";
	    if (@indels_rltd_le_rec>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rltd_le_rec) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	    print STDOUT "[ Right-ending (#{events} = ", scalar (@indels_rltd_re_rec), ") ]\n\n";
	    if (@indels_rltd_re_rec>0) {
		print STDOUT "Br:beg:end:stat_ue\n\n";
		foreach my $event (@indels_rltd_re_rec) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
		print STDOUT "\n\n";
	    }

	}
	print STDOUT "\n";

	print STDOUT "[[ Other overlapping indels in Reconstructed MSA (#{indels} = ", scalar (@indels_rltd_other_ovlp_rec), ") ]]\n\n";
	if (@indels_rltd_other_ovlp_rec>0) {
	    print STDOUT "Br:beg:end:stat_ue\n\n";
	    foreach my $event (@indels_rltd_other_ovlp_rec) { print STDOUT join (':', @{$event}[0..3]), "\n"; }
	    print STDOUT "\n\n";
	}
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_newtype>1) { # ADDED on Oct 3 (Fri), 2015. ...}"

    (print STDOUT "\n<< Now, Attempt to associate the block with an error of a definite type ... >>\n\n") if ($TEST_LEVEL_newtype>1);


	    ##my @indices_indel_identifiers = (0, 1, 2); # Indices (for @{ ->[$e]}) specifying the branch ID, beginning site, and ending-site, respectively. ## MOVED above, before the "CANDIDATE for the beginning of the new subroutine'," (on Oct 6 (Tue), 2015). ##

		#  Apparent lengths of the block in reference and reconstructed MSAs.
		# (ADDED for DEBUGGING on Oct 4 (Sun), 2015.)
	    my $applen_cb_ref = $end_cb_ref - $beg_cb_ref + 1;
	    my $applen_cb_rec = $end_cb_rec - $beg_cb_rec + 1;
		# END of "Apparent lengths of the block in reference and reconstructed MSAs."

			## Step 1: The simplest case, which does not involve indels along related branches.  ##

	    if (($ct_rltd_indels_rec == 0) and ($ct_rltd_indels_ref == 0)) {
	    ##if (($ct_flnk_ovlp_indels_rec == 1) and ($ct_flnk_ovlp_indels_ref == 1)) {...} # MODIFIED to a better condition (on Oct 31 (Fri), 2014). #

    (print STDOUT "< Step 1 >\n\n") if ($TEST_LEVEL_newtype>0);

		if (($ct_rlv_indels_ref == 0) and ($ct_rlv_indels_rec == 0)) {

			# Substep (o): The composite block is NOT directly associated with any errorneous indel event (???) #

    (print STDOUT "  { Substep (o) }\n\n") if ($TEST_LEVEL_newtype>1);

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
		    $reason_skipped = 'NO_ASSOCIATED_EVENT(???)';

		    my $penalty ; # MODIFIED for DEBUGGING on Oct 21 (Wed), 2015. #
		    ##$penalty = 200; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty); # MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    ##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

		    ##$cblk_wise_msa_errors[$cb] = undef;
		    ##$cblk2if_skipped{$cb} = 'NO_ASSOCIATED_EVENT(???)';
		    ##next;

		}


		my $type; # The type of MSA error(s). #
		my $penalty; # The penalty incurred by the error. # ADDED in REVISION on Oct 21 (Wed), 2015. #

			# Beginning of the long "if ... elsif ... elsif ... else" conditioning. #

		if ( ($ct_rlv_indels_ref == 1) and ($ct_rlv_indels_rec == 1) ) {
		##if ( ($ct_rlv_indels_ref == 1) and ($ct_rlv_indels_ref == 1) ) {...} # A FATAL BUG fixed on Nov 6 (Thu), 2014. #

			# Substep (i): Suspect a 'shift', an 'incomplete-collapse of independent insertions', or an 'incomplete-creation of erroneous indepenent insertions'. #

    (print STDOUT "  { Substep (i) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$indels_rlv_ref[0]}[0 .. 3, 5];
		    my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$indels_rlv_rec[0]}[0 .. 3, 5];
		    ##my ($br1, $state_ue1, $len1) = @{$indels_rlv_ref[0]}[0, 3, 5];
		    ##my ($br2, $state_ue2, $len2) = @{$indels_rlv_rec[0]}[0, 3, 5];

		if ((($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (($br1 != $br2) and ($state_ue1 ne $state_ue2)) ) { # A DEBUGGING measure on Nov 7 (Fri), 2014. #

		   if ($len1 == $len2) {

					# (a) Suspect a 'shift'. #

    (print STDOUT "     | (a) - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ( ((@indels_rlv_lf_rec ==1) and (@indels_rlv_rf_ref ==1)) or
			  	((@indels_rlv_rf_rec ==1) and (@indels_rlv_lf_ref ==1)) ) {
			  	##((@indels_rlv_rf_rec ==1) and (@indels_rlv_lf_rec ==1)) ) {...} # A BUG fixed on Nov 7 (Fri), 2014. 

				# It must be a clearly defined 'shift'. #

    (print STDOUT "     | (a) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Shift';
			    $penalty = 1; #  ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ( ((@indels_rlv_le_rec ==1) and (@indels_rlv_re_ref ==1) and (@indels_rlv_re_rec ==0) and (@indels_rlv_le_ref ==0)) or
					((@indels_rlv_re_rec ==1) and (@indels_rlv_le_ref ==1) and (@indels_rlv_le_rec ==0) and (@indels_rlv_re_ref ==0)) ) {

				# It should be a 'shift'... but I'm not completely sure... #

    (print STDOUT "     | (a) - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Shift(?)';
			    $penalty = 1.5; #  ADDED in REVISION on Oct 21 (Wed), 2015. #

			} else {
				# It may be a 'shift'. #

    (print STDOUT "     | (a) - 3 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Shift(???)';
			    $penalty = 2; #  ADDED in REVISION on Oct 21 (Wed), 2015. #
			}

		   } elsif (($len1 > $len2) and (@indels_rltd_other_ovlp_rec ==1) and (scalar(@{$indels_other_ovlp_rec}) - scalar(@indels_rlv_other_ovlp_rec) >=2) and (@indels_rlv_lf_rec == 0) and (@indels_rlv_rf_rec ==0)) { # This condition was ADDED on Nov 11 (Tue), 2014. #

					# (b) Suspect an 'incomplete-collapse' of independent insertions. #
					# (ADDED on Nov 11 (Tue), 2014.) #

    (print STDOUT "     | (b) - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

			my @indels_invlvd_ref1 = ();
			my @indels_invlvd_rec1 = ();
			my @indels_invlvd_rec1p = @indels_rlv_rec;

			my ($br3, $beg3, $end3, $state_ue3, $len3) = @{$indels_rltd_other_ovlp_rec[0]}[0 .. 3, 5];

			my $subtype;
			my $sbtype1;
			my $sbpenalty = my $sbpenalty1 = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			if ($end3 + 1 == $beg2) { # 'rltd_other_ovlp_rec' left-flanks 'rlv_rec'.

    (print STDOUT "     | (b) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = @indels_rltd_other_ovlp_rec;
			    foreach my $event (@{$indels_other_ovlp_rec}) { # 1st inner-inner foreach-loop (over 'other ovelapping' reconstructed indels). #
				my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event}[0 .. 3, 5];
				($end4 + 1 == $beg2) or next; # Accept only left-flanking events.
				(($br4 == $br3) or ($br4 == $br1)) and next;
				my $eq_br4 = $equiv_branches{$br4};
				((defined $eq_br4) and (($eq_br4 == $br3) or ($eq_br4 == $br1))) and next;
				push @indels_invlvd_rec1, $event;
			    }				# End of the 1st inner-inner foreach-loop (over 'other overlapping' reconstructed indels). #
			    if (@indels_invlvd_rec1 >= 2) { $sbtype1 = 'a'; }

			} elsif ($end2 + 1 == $beg3) { # 'rltd_other_ovlp_rec' right-flanks 'rlv_rec'.

    (print STDOUT "     | (b) - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = @indels_rltd_other_ovlp_rec;
			    foreach my $event (@{$indels_other_ovlp_rec}) { # 2nd inner-inner foreach-loop (over 'other ovelapping' reconstructed indels). #
				my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event}[0 .. 3, 5];
				($end2 + 1 == $beg4) or next; # Accept only right-flanking events.
				(($br4 == $br3) or ($br4 == $br1)) and next;
				my $eq_br4 = $equiv_branches{$br4};
				((defined $eq_br4) and (($eq_br4 == $br3) or ($eq_br4 == $br1))) and next;
				push @indels_invlvd_rec1, $event;
			    }				# End of the 2nd inner-inner foreach-loop (over 'other overlapping' reconstructed indels). #
			    if (@indels_invlvd_rec1 >= 2) { $sbtype1 = 'b'; }
			}

			if (defined $sbtype1) {

			    if (@indels_rlv_lf_ref ==1) { # An exceptional case (L). # 

    (print STDOUT "     | (b) - 3 |\n\n") if ($TEST_LEVEL_newtype>2);

				if ((@{$indels_other_ovlp_ref} >0) and ($sbtype1 = 'a')) {
				    @indels_invlvd_ref1 = @indels_rlv_lf_ref;
				    foreach my $event (@{$indels_other_ovlp_ref}) { # 3rd inner-inner foreach-loop (over 'other overlapping' refernce indels). #
					my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event}[0 .. 3, 5];
					($end1 + 1 == $beg4) or next; # Accept only  right-flanking events.
					(($br4 == $br3) or ($br4 == $br1)) and next;
					my $eq_br4 = $equiv_branches{$br4};
					((defined $eq_br4) and (($eq_br4 == $br3) or ($eq_br4 == $br1))) and next;
					push @indels_invlvd_ref1, $event;
				    }				# End of the 3rd inner-inner foreach-loop (over 'other overlapping' reference indels). #

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				    if (@indels_invlvd_ref1 ==2) { $subtype = '(+shift^2(??))'; $sbpenalty += 2; }
				    ##if (@indels_invlvd_ref1 ==2) { $subtype = '(+shift^2(??))'; }
				}

			    } elsif (@indels_rlv_rf_ref ==1) { # An exceptional case (R). #

    (print STDOUT "     | (b) - 4 |\n\n") if ($TEST_LEVEL_newtype>2);

				if ((@{$indels_other_ovlp_ref} >0) and ($sbtype1 = 'b')) {
				    @indels_invlvd_ref1 = @indels_rlv_rf_ref;
				    foreach my $event (@{$indels_other_ovlp_ref}) { # 4th inner-inner foreach-loop (over 'other overlapping' refernce indels). #
					my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event}[0 .. 3, 5];
					($end4 + 1 == $beg1) or next; # Accept only  left-flanking events.
					(($br4 == $br3) or ($br4 == $br1)) and next;
					my $eq_br4 = $equiv_branches{$br4};
					((defined $eq_br4) and (($eq_br4 == $br3) or ($eq_br4 == $br1))) and next;
					push @indels_invlvd_ref1, $event;
				    }				# End of the 4th inner-inner foreach-loop (over 'other overlapping' reference indels). #

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				    if (@indels_invlvd_ref1 ==2) { $subtype = '(+shift^2(??))'; $sbpenalty += 2; }
				    ##if (@indels_invlvd_ref1 ==2) { $subtype = '(+shift^2(??))'; }
				}

			    } else { # A normal case. #

    (print STDOUT "     | (b) - 5 |\n\n") if ($TEST_LEVEL_newtype>2);

				if (@{$indels_rf_ref} == 1) { # (X)

				    if (@indels_rlv_re_ref ==1) { # (i)
					$subtype = ($sbtype1 eq 'a') ? '(+slide(?))' : undef ;
					($sbtype1 eq 'a') and ($sbpenalty += 1);# ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif (@indels_rlv_le_ref ==1) { # (ii)
					$subtype = ($sbtype1 eq 'a') ? '(+overshoot(??))' : '(+shift(?))' ;
					$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif (@indels_rlv_other_ovlp_ref ==1) { # (iii)
					$subtype = ($sbtype1 eq 'a') ? '(+overshoot(??))' : undef;
					($sbtype1 eq 'a') and ($sbpenalty += 1) ; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				    @indels_invlvd_ref1 = (@indels_rlv_ref, @{$indels_rf_ref});

				} elsif (@{$indels_lf_ref} ==1) { # (Y)

				    if (@indels_rlv_le_ref ==1) { # (ii)
					$subtype = ($sbtype1 eq 'b') ? '(+slide(?))' : undef ;
					($sbtype1 eq 'b') and ($sbpenalty += 1);# ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif (@indels_rlv_re_ref ==1) { # (i)
					$subtype = ($sbtype1 eq 'b') ? '(+overshoot(??))' : '(+shift(?))' ;
					$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif (@indels_rlv_other_ovlp_ref ==1) { # (iii)
					$subtype = ($sbtype1 eq 'b') ? '(+overshoot(??))' : undef;
					($sbtype1 eq 'b') and ($sbpenalty += 1);# ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				    @indels_invlvd_ref1 = (@indels_rlv_ref, @{$indels_lf_ref});
				}
			    }

			}  # END of "if (defined $sbtype1) {...}"


		      if (defined $subtype) {

    (print STDOUT "     | (b) - 6 |\n\n") if ($TEST_LEVEL_newtype>2);

			 my ($if_cii, $indels_invlvd_rec2) = if_collapse_indep_ins (@indels_invlvd_ref1, @indels_invlvd_rec1, %{$node2pa}, %{$node2ch}, $CHAR_ABS);
			 unless (defined $if_cii) {
			    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i)-(b) of Step 1: $! !\n\n";
			    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);


			    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i)-(b) of Step 1 on the $cb th composite shift-block for the $s th segment !\n";
			    ##print STDERR "[ERROR '$program_name'] $!\n\n";
			    ##die ;
			 }

			 if ($if_cii) {
			    @indels_invlvd_ref = @indels_invlvd_ref1;
			    @indels_invlvd_rec = (@{$indels_invlvd_rec2}, @indels_invlvd_rec1p);
			    my $diffct = scalar (@indels_invlvd_rec) - @indels_invlvd_ref;

			    $cblk_wise_cts_invlvd_indels[$cb]->[5] = scalar (@{$indels_invlvd_rec2});

			    $type = "Incomplete-Collapse-of-Independent-Insertions$subtype(+$diffct)(?)";
			    $penalty = 15 + $sbpenalty; 
			 }
		      } # END of "if (defined $subtype) {...}"

					# END of "(b) Suspect an 'incomplete-collapse' of independent insertions." #

		   } elsif (($len1 < $len2) and (@indels_rltd_other_ovlp_ref ==1) and (scalar(@{$indels_other_ovlp_ref}) - scalar(@indels_rlv_other_ovlp_ref) >=2) and (@indels_rlv_lf_ref == 0) and (@indels_rlv_rf_ref ==0)) { # This condition was ADDED on Nov 11 (Tue), 2014. #

					# (c) Suspect an 'incomplete-creation' of erroneous independent insertions. #
					# (ADDED on Nov 11 (Tue), 2014.) #

    (print STDOUT "     | (c) - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

			my @indels_invlvd_rec1 = ();
			my @indels_invlvd_ref1 = ();
			my @indels_invlvd_ref1p = @indels_rlv_ref;

			my ($br3, $beg3, $end3, $state_ue3, $len3) = @{$indels_rltd_other_ovlp_ref[0]}[0 .. 3, 5];

			my $subtype;
			my $sbtype1;
			my $sbpenalty = my $sbpenalty1 = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			if ($end3 + 1 == $beg1) { # 'rltd_other_ovlp_ref' left-flanks 'rlv_ref'.

    (print STDOUT "     | (c) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = @indels_rltd_other_ovlp_ref;
			    foreach my $event (@{$indels_other_ovlp_ref}) { # 1st inner-inner foreach-loop (over 'other ovelapping' reference indels). #
				my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event}[0 .. 3, 5];
				($end4 + 1 == $beg1) or next; # Accept only left-flanking events.
				(($br4 == $br3) or ($br4 == $br1)) and next;
				my $eq_br4 = $equiv_branches{$br4};
				((defined $eq_br4) and (($eq_br4 == $br3) or ($eq_br4 == $br1))) and next;
				push @indels_invlvd_ref1, $event;
			    }				# End of the 1st inner-inner foreach-loop (over 'other overlapping' reference indels). #
			    if (@indels_invlvd_ref1 >= 2) { $sbtype1 = 'a'; }

			} elsif ($end1 + 1 == $beg3) { # 'rltd_other_ovlp_ref' right-flanks 'rlv_ref'.

    (print STDOUT "     | (c) - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = @indels_rltd_other_ovlp_ref;
			    foreach my $event (@{$indels_other_ovlp_ref}) { # 2nd inner-inner foreach-loop (over 'other ovelapping' reference indels). #
				my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event}[0 .. 3, 5];
				($end1 + 1 == $beg4) or next; # Accept only right-flanking events.
				(($br4 == $br3) or ($br4 == $br1)) and next;
				my $eq_br4 = $equiv_branches{$br4};
				((defined $eq_br4) and (($eq_br4 == $br3) or ($eq_br4 == $br1))) and next;
				push @indels_invlvd_rec1, $event;
			    }				# End of the 2nd inner-inner foreach-loop (over 'other overlapping' reference indels). #
			    if (@indels_invlvd_ref1 >= 2) { $sbtype1 = 'b'; }
			}

			if (defined $sbtype1) {
			    if (@indels_rlv_lf_rec ==1) { # An exceptional case (L). # 

    (print STDOUT "     | (c) - 3 |\n\n") if ($TEST_LEVEL_newtype>2);

				if ((@{$indels_other_ovlp_rec} >0) and ($sbtype1 = 'a')) {
				    @indels_invlvd_rec1 = @indels_rlv_lf_rec;
				    foreach my $event (@{$indels_other_ovlp_rec}) { # 3rd inner-inner foreach-loop (over 'other overlapping' reconstructed indels). #
					my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event}[0 .. 3, 5];
					($end2 + 1 == $beg4) or next; # Accept only  right-flanking events.
					(($br4 == $br3) or ($br4 == $br1)) and next;
					my $eq_br4 = $equiv_branches{$br4};
					((defined $eq_br4) and (($eq_br4 == $br3) or ($eq_br4 == $br1))) and next;
					push @indels_invlvd_rec1, $event;
				    }				# End of the 3rd inner-inner foreach-loop (over 'other overlapping' reconstructed indels). #

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. # 
				    if (@indels_invlvd_rec1 ==2) { $subtype = '(+shift^2(??))'; $sbpenalty += 2; }
				    ##if (@indels_invlvd_rec1 ==2) { $subtype = '(+shift^2(??))'; }
				}

			    } elsif (@indels_rlv_rf_rec ==1) { # An exceptional case (R). #

    (print STDOUT "     | (c) - 4 |\n\n") if ($TEST_LEVEL_newtype>2);

				if ((@{$indels_other_ovlp_rec} >0) and ($sbtype1 = 'b')) {
				    @indels_invlvd_rec1 = @indels_rlv_rf_rec;
				    foreach my $event (@{$indels_other_ovlp_rec}) { # 4th inner-inner foreach-loop (over 'other overlapping' reconstructed indels). #
					my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event}[0 .. 3, 5];
					($end4 + 1 == $beg1) or next; # Accept only  left-flanking events.
					(($br4 == $br3) or ($br4 == $br1)) and next;
					my $eq_br4 = $equiv_branches{$br4};
					((defined $eq_br4) and (($eq_br4 == $br3) or ($eq_br4 == $br1))) and next;
					push @indels_invlvd_rec1, $event;
				    }				# End of the 4th inner-inner foreach-loop (over 'other overlapping' reconstructed indels). #

					# ADDED in REVISION on Oct 21 (Wed), 2015. #
				    if (@indels_invlvd_rec1 ==2) { $subtype = '(+shift^2(??))'; $sbpenalty = 2; }
				    ##if (@indels_invlvd_rec1 ==2) { $subtype = '(+shift^2(??))'; }
				}

			    } else { # A normal case. #

    (print STDOUT "     | (c) - 5 |\n\n") if ($TEST_LEVEL_newtype>2);

				if (@{$indels_rf_rec} == 1) { # (X)

				    if (@indels_rlv_re_rec ==1) { # (i)
					$subtype = ($sbtype1 eq 'a') ? '(+slide(?))' : undef ;
					($sbtype1 eq 'a') and ($sbpenalty = 1); # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif (@indels_rlv_le_rec ==1) { # (ii)
					$subtype = ($sbtype1 eq 'a') ? '(+overshoot(??))' : '(+shift(?))' ;
					$sbpenalty = 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif (@indels_rlv_other_ovlp_rec ==1) { # (iii)
					$subtype = ($sbtype1 eq 'a') ? '(+overshoot(??))' : undef;
					($sbtype1 eq 'a') and ($sbpenalty = 1); # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				    @indels_invlvd_rec1 = (@indels_rlv_rec, @{$indels_rf_rec});

				} elsif (@{$indels_lf_rec} ==1) { # (Y)

				    if (@indels_rlv_le_rec ==1) { # (ii)
					$subtype = ($sbtype1 eq 'b') ? '(+slide(?))' : undef ;
					($sbtype1 eq 'b') and ($sbpenalty = 1); # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif (@indels_rlv_re_rec ==1) { # (i)
					$subtype = ($sbtype1 eq 'b') ? '(+overshoot(??))' : '(+shift(?))' ;
					$sbpenalty = 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif (@indels_rlv_other_ovlp_rec ==1) { # (iii)
					$subtype = ($sbtype1 eq 'b') ? '(+overshoot(??))' : undef;
					($sbtype1 eq 'b') and ($sbpenalty = 1); # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				    @indels_invlvd_rec1 = (@indels_rlv_rec, @{$indels_lf_rec});
				}
			    }

			}  # END of "if (defined $sbtype1) {...}"


		      if (defined $subtype) {

    (print STDOUT "     | (c) - 6 |\n\n") if ($TEST_LEVEL_newtype>2);

			 my ($if_cii, $indels_invlvd_ref2) = if_collapse_indep_ins (@indels_invlvd_rec1, @indels_invlvd_ref1, %{$node2pa}, %{$node2ch}, $CHAR_ABS);
			 unless (defined $if_cii) {
 
			    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i)-(c) of Step 1: $! !\n\n";
			    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

			    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i)-(c) of Step 1 on the $cb th composite shift-block for the $s th segment !\n";
			    ##print STDERR "[ERROR '$program_name'] $!\n\n";
			    ##die ;
			 }

			 if ($if_cii) {
			    @indels_invlvd_rec = @indels_invlvd_rec1;
			    @indels_invlvd_ref = (@{$indels_invlvd_ref2}, @indels_invlvd_ref1p);
			    my $diffct = scalar (@indels_invlvd_ref) - @indels_invlvd_rec;

			    $cblk_wise_cts_invlvd_indels[$cb]->[4] = scalar (@{$indels_invlvd_ref2});

			    $type = "Incomplete-Creation-of-Erroneous-Independent-Insertions$subtype(-$diffct)(?)";
			    $penalty = 15 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			 }
		      } # END of "if (defined $subtype) {...}"

					# END of "(c) Suspect an 'incomplete-creation' of erroneous independent insertions." #

    (print STDOUT "     | (c) - end |\n\n") if ($TEST_LEVEL_newtype>2);

		   } # END of "if ($len1 == $len2) {...} elsif (($len1 > $len2) and ...) {...} elsif (($len1 < $len2) and ...) {...}."

		} # END of "if ((($br1 == $br2) and ...) or ...) {...}"

					# ADDED on Nov 12 (Wed), 2014. #

		if (!(defined $type) and (@indels_rltd_other_ovlp_rec ==1) and (@indels_rltd_other_ovlp_ref ==1) 
			and ( ((@indels_rlv_rf_ref ==1) and (@indels_rlv_rf_rec ==1)) or ((@indels_rlv_lf_ref ==1) and (@indels_rlv_lf_rec ==1)) ) ) {

		    if ($indels_rlv_ref[0]->[5] > $indels_rlv_rec[0]->[5]) { # 'rlv_ref' is longer than 'rlv_rec'. #

				# (d): In this case, Suspect a 'Partial-Neighboring-Branch-Merge' (2ins) (Short-Block(exceptional)). #

    (print STDOUT "     | (d) - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

			my $event1 = $indels_rltd_other_ovlp_ref[0];
			my $event2 = $indels_rlv_ref[0];

			my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$event1}[0..3, 5];
			my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$event2}[0..3, 5];
			my $eq_br1 = $equiv_branches{$br1};
			my $eq_br2 = $equiv_branches{$br2};

			if ( ($len2 > $len1) and (($end1 + 1 < $beg2) or ($end2 + 1 < $beg1)) ) { # 'ref_rlv' is longer than 'ref_rltd', and the two events are separated.

    (print STDOUT "     | (d) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my $event3 = $indels_rltd_other_ovlp_rec[0];
			    my $event4 = $indels_rlv_rec[0];
			    my ($br3, $beg3, $end3, $state_ue3, $len3) = @{$event3}[0..3, 5];
			    my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event4}[0..3, 5];

					# See if 'rltd_rec' is on a different branch than 'rltd_ref'.
			    my $if_event3 = (($br1 != $br3) and !((defined $eq_br1) and ($eq_br1 == $br3)) ) ? 1 : 0;

					# See if 'rlv_rec' is of the same type as 'rlv_ref'.
			    my $if_event4 = ( (($br2 == $br4) and ($state_ue2 eq $state_ue4)) or (($br2 != $br4) and ($state_ue2 ne $state_ue4)) ) ? 1 : 0;

			    if (($if_event4) and ($if_event3) and ($len1 == $event3->[5]) and ($len2 == $len1 + $event4->[5])
				and (($end3 + 1 < $beg4) or ($end4 + 1 < $beg4)) ) {

    (print STDOUT "     | (d) - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

				my $if_part_nbr_mrg = if_part_nbr_mrg_2ins (@{$event1}, @{$event2}, @{$event3}, 0, 3, %equiv_branches, %{$node2pa}, $CHAR_ABS) ;
				unless (defined $if_part_nbr_mrg) {
				    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (i)-(d) of Step 1: $! !\n\n";
				    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);


				    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (i)-(d) of Step 1 on the $cb th composite shift-block for the $s th segment !\n";
				    ##print STDERR "[ERROR '$program_name'] $!\n\n";
				    ##die ;
				}

				if ($if_part_nbr_mrg) {

				    $type = "Partial-Neighboring-Branch-Merge(2ins)(+shift)(?)";
				    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				}
			    }

			} # END of "if (($len2 > $len1) and (...)) {...}"

		    } elsif ($indels_rlv_rec[0]->[5] > $indels_rlv_ref[0]->[5]) { # 'rlv_rec' is longer than 'rlv_ref'. #

				# (e): In this case, Suspect a 'Partial-Neighboring-Branch-Split' (2ins) (Short-Block(exceptional)). #

    (print STDOUT "     | (e) - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

			my $event1 = $indels_rltd_other_ovlp_rec[0];
			my $event2 = $indels_rlv_rec[0];

			my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$event1}[0..3, 5];
			my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$event2}[0..3, 5];
			my $eq_br1 = $equiv_branches{$br1};
			my $eq_br2 = $equiv_branches{$br2};

			if ( ($len2 > $len1) and (($end1 + 1 < $beg2) or ($end2 + 1 < $beg1)) ) { # 'rec_rlv' is longer than 'rec_rltd', and the two events are separated.

    (print STDOUT "     | (e) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my $event3 = $indels_rltd_other_ovlp_ref[0];
			    my $event4 = $indels_rlv_ref[0];
			    my ($br3, $beg3, $end3, $state_ue3, $len3) = @{$event3}[0..3, 5];
			    my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event4}[0..3, 5];

					# See if 'rltd_ref' is on a different branch than 'rltd_rec'.
			    my $if_event3 = (($br1 != $br3) and !((defined $eq_br1) and ($eq_br1 == $br3)) ) ? 1 : 0;

					# See if 'rlv_ref' is of the same type as 'rlv_rec'.
			    my $if_event4 = ( (($br2 == $br4) and ($state_ue2 eq $state_ue4)) or (($br2 != $br4) and ($state_ue2 ne $state_ue4)) ) ? 1 : 0;

			    if (($if_event4) and ($if_event3) and ($len1 == $event3->[5]) and ($len2 == $len1 + $event4->[5])
				and (($end3 + 1 < $beg4) or ($end4 + 1 < $beg4)) ) {

    (print STDOUT "     | (e) - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

				my $if_part_nbr_mrg = if_part_nbr_mrg_2ins (@{$event1}, @{$event2}, @{$event3}, 0, 3, %equiv_branches, %{$node2pa}, $CHAR_ABS) ;
				unless (defined $if_part_nbr_mrg) {
				    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (i)-(e) of Step 1: $! !\n\n";
				    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);


				    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (i)-(e) of Step 1 on the $cb th composite shift-block for the $s th segment !\n";
				    ##print STDERR "[ERROR '$program_name'] $!\n\n";
				    ##die ;
				}

				if ($if_part_nbr_mrg) {

				    $type = "Partial-Neighboring-Branch-Split(2ins)(+shift)(?)";
				    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				}
			    }

			} # END of "if (($len2 > $len1) and (...)) {...}"

		    }

		} # END of "if (!(defined $type) and (@indels_rltd_other_ovlp_rec ==1) and ...) {...}"

					# END of "ADDED on Nov 12 (Wed), 2014." #



				# I don't know.... May be a 'Complex' error? #

    (print STDOUT "     | (f) - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');



		} elsif (($ct_rlv_indels_ref == 1) and ($ct_rlv_indels_rec == 0)) {

			# Substep (ii): A rare case (a) ... may be involved in a complex error. #

    (print STDOUT "  { Substep (ii) }\n\n") if ($TEST_LEVEL_newtype>1);

					# ADDED on Nov 12 (Wed), 2014. #

		    if ((@indels_rltd_other_ovlp_ref ==1) and (@indels_rltd_other_ovlp_rec ==2) and (scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) ==1) ) {

				# A: In this case, Suspect a 'Partial-Neighboring-Branch-Merge' (2ins) (Long-Block(exceptional)). #

    (print STDOUT "     | A - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

			my $event1 = $indels_rlv_ref[0];
			my $event2 = $indels_rltd_other_ovlp_ref[0];
			my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$event1}[0..3, 5];
			my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$event2}[0..3, 5];
			my $eq_br2 = $equiv_branches{$br2};

			if ( ($len2 > $len1) and (($end1 + 1 < $beg2) or ($end2 + 1 < $beg1)) ) { # 'ref_rltd' is longer than 'ref_rlv', and the two events are separated.
			    my ($event3, $event4); # event 3 must be on the different branch, event 4 must be on the same branch and of the same type as event2.

    (print STDOUT "     | A - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    foreach my $event (@indels_rltd_other_ovlp_rec) { # Inner foreach-loop (over 'related' indels in reconstructed MSA).
				my ($br, $state_ue) = @{$event}[0, 3];
				if ($br2 == $br) {
				    ($state_ue2 eq $state_ue) and ($event4 = $event);

				} elsif ((defined $eq_br2) and ($eq_br2 == $br)) {
				    ($state_ue2 ne $state_ue) and ($event4 = $event);

				} else {
				    $event3 = $event;
				}
			    }				# End of the inner foreach-loop (over 'related' indels in reconstructed MSA).

			    if ((defined $event4) and (defined $event3) and ($len1 == $event3->[5]) and ($len2 == $len1 + $event4->[5])) {

    (print STDOUT "     | A - 3 |\n\n") if ($TEST_LEVEL_newtype>2);

				my $if_part_nbr_mrg = if_part_nbr_mrg_2ins (@{$event1}, @{$event2}, @{$event3}, 0, 3, %equiv_branches, %{$node2pa}, $CHAR_ABS) ;
				unless (defined $if_part_nbr_mrg) {
				    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (ii)-A of Step 1: $! !\n\n";
				    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);


				    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (ii)-A of Step 1 on the $cb th composite shift-block for the $s th segment !\n";
				    ##print STDERR "[ERROR '$program_name'] $!\n\n";
				    ##die ;
				}

				if ($if_part_nbr_mrg) {
				    my ($beg3, $end3) = @{$event3}[1,2];
				    my ($beg4, $end4) = @{$event4}[1,2];
				    my $subtype;
				    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    if (@indels_rlv_lf_ref == 1) {

					if ($end3 < $beg4) {
					    $subtype = '(+shift)(?)';
					    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif ($end4 < $beg3) {
					    $subtype = '(+shift+slide)(??)';
					    $sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}

				    } else { # @indels_rlv_rf_ref ==1. #

					if ($end4 < $beg3) {
					    $subtype = '(+shift)(?)';
					    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif ($end3 < $beg4) {
					    $subtype = '(+shift+slide)(??)';
					    $sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }

				    if (defined $subtype) {
					$type = "Partial-Neighboring-Branch-Merge(2ins)$subtype";
					$penalty = 5 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }

				} # END of "if ($if_part_nbr_mrg) {...}"
			    }

			} # END of "if (($len2 > $len1) and (...)) {...}"

					# END of "ADDED on Nov 12 (Wed), 2014." #


		    } elsif ((@indels_rltd_other_ovlp_ref>0) and (@indels_rltd_other_ovlp_ref>@indels_rltd_other_ovlp_rec) and (scalar (@indels_rlv_lf_ref) + scalar (@indels_rlv_rf_ref) ==1) ) {

				# B: In this case, Suspect a 'Purge' aligned with a run of gaps. #

    (print STDOUT "     | B - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

					# A DEBUGGING measure on Nov 7 (Fri), 2014. #
			my ($br1, $state_ue1, $len1) = @{$indels_rlv_ref[0]}[0, 3, 5];

			my @indels_invlvd_ref2 = ();
			foreach my $event_info (@indels_rltd_other_ovlp_ref) {

			    my ($br2, $state_ue2, $len2) = @{$event_info}[0, 3, 5];

			    if ($len2 == $len1) {
				if ($br2 == $node2pa->{$br1}) {
				    (($state_ue2 eq $CHAR_PRES) and ($state_ue1 eq $CHAR_ABS)) and (push @indels_invlvd_ref2, $event_info);
				} elsif ($br1 == $node2pa->{$br2}) {
				    (($state_ue1 eq $CHAR_PRES) and ($state_ue2 eq $CHAR_ABS)) and (push @indels_invlvd_ref, $event_info);
				} else {
				    (($state_ue1 eq $CHAR_ABS) and ($state_ue2 eq $CHAR_ABS)) and (push @indels_invlvd_ref2, $event_info);
				}
			    }
			}
			if (@indels_invlvd_ref2 ==1) {

    (print STDOUT "     | B - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    push @indels_invlvd_ref, @indels_invlvd_ref2;
			    $cblk_wise_cts_invlvd_indels[$cb]->[4] = 1;
			    $type = 'Purge(gap-aligned)(??)';
			    $penalty = 3 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			}
			##push @indels_invlvd_ref, @indels_rltd_other_ovlp_ref;
			##$cblk_wise_cts_invlvd_indels[$cb]->[4] = scalar (@indels_rltd_other_ovlp_ref); # ADDED on Nov 6 (Thu), 2014. #
			##$type = 'Purge(gap-aligned)(??)';

		    } # END of "if ((@indels_rltd_other_ovlp_ref ==1) and ...) {...} elsif ((@indels_rltd_other_ovlp_ref>0) and ...) {...}"


				# C: May be 'Complex' ???

    (print STDOUT "     | C - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

			# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} elsif (($ct_rlv_indels_rec == 1) and ($ct_rlv_indels_ref == 0)) {

			# Substep (iii): A rare case (b) ... may be involved in a complex error. #

    (print STDOUT "  { Substep (iii) }\n\n") if ($TEST_LEVEL_newtype>1);

					# ADDED on Nov 12 (Wed), 2014. #

		    if ((@indels_rltd_other_ovlp_rec ==1) and (@indels_rltd_other_ovlp_ref ==2) and (scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) ==1) ) {

				# A: In this case, Suspect a 'Partial-Neighboring-Branch-Split' (2ins) (Long-Block(exceptional)). #

    (print STDOUT "     | A - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

			my $event1 = $indels_rlv_rec[0];
			my $event2 = $indels_rltd_other_ovlp_rec[0];
			my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$event1}[0..3, 5];
			my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$event2}[0..3, 5];
			my $eq_br2 = $equiv_branches{$br2};

			if ( ($len2 > $len1) and (($end1 + 1 < $beg2) or ($end2 + 1 < $beg1)) ) { # 'rec_rltd' is longer than 'rec_rlv', and the two events are separated.

    (print STDOUT "     | A - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my ($event3, $event4); # event 3 must be on the different branch, event 4 must be on the same branch and of the same type as event2.

			    foreach my $event (@indels_rltd_other_ovlp_ref) { # Inner foreach-loop (over 'related' indels in reference MSA).
				my ($br, $state_ue) = @{$event}[0, 3];
				if ($br2 == $br) {
				    ($state_ue2 eq $state_ue) and ($event4 = $event);

				} elsif ((defined $eq_br2) and ($eq_br2 == $br)) {
				    ($state_ue2 ne $state_ue) and ($event4 = $event);

				} else {
				    $event3 = $event;
				}
			    }				# End of the inner foreach-loop (over 'related' indels in reference MSA).

			    if ((defined $event4) and (defined $event3) and ($len1 == $event3->[5]) and ($len2 == $len1 + $event4->[5])) {

    (print STDOUT "     | A - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

				my $if_part_nbr_mrg = if_part_nbr_mrg_2ins (@{$event1}, @{$event2}, @{$event3}, 0, 3, %equiv_branches, %{$node2pa}, $CHAR_ABS) ;
				unless (defined $if_part_nbr_mrg) {
				    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (iii)-A of Step 1: $! !\n\n";
				    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);


				    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (iii)-A of Step 1 on the $cb th composite shift-block for the $s th segment !\n";
				    ##print STDERR "[ERROR '$program_name'] $!\n\n";
				    ##die ;
				}

				if ($if_part_nbr_mrg) {
				    my ($beg3, $end3) = @{$event3}[1,2];
				    my ($beg4, $end4) = @{$event4}[1,2];
				    my $subtype;
				    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    if (@indels_rlv_lf_rec == 1) {

					if ($end3 < $beg4) {
					    $subtype = '(+shift)(?)';
					    $sbpenalty = 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif ($end4 < $beg3) {
					    $subtype = '(+shift+slide)(??)';
					    $sbpenalty = 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}

				    } else { # @indels_rlv_rf_rec ==1. #

					if ($end4 < $beg3) {
					    $subtype = '(+shift)(?)';
					    $sbpenalty = 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif ($end3 < $beg4) {
					    $subtype = '(+shift+slide)(??)';
					    $sbpenalty = 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }

				    if (defined $subtype) {
					$type = "Partial-Neighboring-Branch-Split(2ins)$subtype";
					$penalty = 5 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }

				} # END of "if ($if_part_nbr_mrg) {...}"
			    }

			} # END of "if (($len2 > $len1) and (...)) {...}"

					# END of "ADDED on Nov 12 (Wed), 2014." #


		    } elsif ((@indels_rltd_other_ovlp_rec>0) and (@indels_rltd_other_ovlp_rec>@indels_rltd_other_ovlp_ref) and (scalar (@indels_rlv_lf_rec) + scalar (@indels_rlv_rf_rec) ==1) ) {

				# B: In this case, Suspect an 'Ex-nihilo' aligned with a run of gaps. #

    (print STDOUT "     | B - 0 |\n\n") if ($TEST_LEVEL_newtype>2);


					# A DEBUGGING measure on Nov 7 (Fri), 2014. #
			my ($br1, $state_ue1, $len1) = @{$indels_rlv_rec[0]}[0, 3, 5];

			my @indels_invlvd_rec2 = ();
			foreach my $event_info (@indels_rltd_other_ovlp_rec) {

			    my ($br2, $state_ue2, $len2) = @{$event_info}[0, 3, 5];

			    if ($len2 == $len1) {
				if ($br2 == $node2pa->{$br1}) {
				    (($state_ue2 eq $CHAR_PRES) and ($state_ue1 eq $CHAR_ABS)) and (push @indels_invlvd_rec2, $event_info);
				} elsif ($br1 == $node2pa->{$br2}) {
				    (($state_ue1 eq $CHAR_PRES) and ($state_ue2 eq $CHAR_ABS)) and (push @indels_invlvd_rec, $event_info);
				} else {
				    (($state_ue1 eq $CHAR_ABS) and ($state_ue2 eq $CHAR_ABS)) and (push @indels_invlvd_rec, $event_info);
				}
			    }
			}
			if (@indels_invlvd_rec2 ==1) {

    (print STDOUT "     | B - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    push @indels_invlvd_rec, @indels_invlvd_rec2;
			    $cblk_wise_cts_invlvd_indels[$cb]->[5] = 1;
			    $type = 'Ex-nihilo(gap-aligned)(??)';
			    $penalty = 3 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			}

			##push @indels_invlvd_rec, @indels_rltd_other_ovlp_rec;
			##$cblk_wise_cts_invlvd_indels[$cb]->[5] = scalar (@indels_rltd_other_ovlp_rec); # ADDED on Nov 6 (Thu), 2014. #
			##$type = 'Ex-nihilo(gap-aligned)(??)';

		    } # END of "if ((@indels_rltd_other_ovlp_rec>0) and ...) {...}"


				# C: May be 'Complex' ???

    (print STDOUT "     | C - 0 |\n\n") if ($TEST_LEVEL_newtype>2);

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} elsif (($ct_rlv_indels_rec == 2) and ($ct_rlv_indels_ref == 0)) {

			# Substep (v): Suspect an 'Ex-nihilo'. #

    (print STDOUT "  { Substep (v) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my @indels_rlv1 = my @indels_rlv2 = ();

		    if ((@indels_rlv_lf_rec ==1) and (@indels_rlv_re_rec ==1)) {

    (print STDOUT "     | (v) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			@indels_rlv1 = @indels_rlv_lf_rec;
			@indels_rlv2 = @indels_rlv_re_rec;

		    } elsif ((@indels_rlv_le_rec ==1) and (@indels_rlv_rf_rec ==1)) {

    (print STDOUT "     | (v) - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

			@indels_rlv1 = @indels_rlv_rf_rec;
			@indels_rlv2 = @indels_rlv_le_rec;

		    } elsif ((scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) ==1) and (@indels_rlv_other_ovlp_rec ==1)) {

    (print STDOUT "     | (v) - 3 |\n\n") if ($TEST_LEVEL_newtype>2);

			@indels_rlv1 = (@indels_rlv_lf_rec>0) ? @indels_rlv_lf_rec : @indels_rlv_rf_rec;
			@indels_rlv2 = @indels_rlv_other_ovlp_rec;
		    }

		    if (@indels_rlv1>0) {

    (print STDOUT "     | (v) - 4 |\n\n") if ($TEST_LEVEL_newtype>2);

			my ($br1, $state_ue1, $len1) = @{$indels_rlv1[0]}[0, 3, 5];
			my ($br2, $state_ue2, $len2) = @{$indels_rlv2[0]}[0, 3, 5];

			if ( (($br1 == $br2) and ($state_ue1 ne $state_ue2)) or 
			    (($br1 != $br2) and ($state_ue1 eq $state_ue2)) ) {

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
			    if ($len1 == $len2) {
				$type = "Ex-nihilo(?)"; # It could be an authentic 'Ex-nihilo'. #
				$penalty = 3;
			    }
			    ##($len1 == $len2) and ($type = "Ex-nihilo(?)"); # It could be an authentic 'Ex-nihilo'. #
			}
		    }

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');



		} elsif (($ct_rlv_indels_ref == 2) and ($ct_rlv_indels_rec == 0)) {

			# Substep (vi): Suspect a 'Purge'. #

    (print STDOUT "  { Substep (vi) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my @indels_rlv1 = my @indels_rlv2 = ();

		    if ((@indels_rlv_lf_ref ==1) and (@indels_rlv_re_ref ==1)) {

    (print STDOUT "     | (vi) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			@indels_rlv1 = @indels_rlv_lf_ref;
			@indels_rlv2 = @indels_rlv_re_ref;

		    } elsif ((@indels_rlv_le_ref ==1) and (@indels_rlv_rf_ref ==1)) {

    (print STDOUT "     | (vi) - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

			@indels_rlv1 = @indels_rlv_rf_ref;
			@indels_rlv2 = @indels_rlv_le_ref;

		    } elsif ((scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) ==1) and (@indels_rlv_other_ovlp_ref ==1)) {

    (print STDOUT "     | (vi) - 3 |\n\n") if ($TEST_LEVEL_newtype>2);

			@indels_rlv1 = (@indels_rlv_lf_ref>0) ? @indels_rlv_lf_ref : @indels_rlv_rf_ref;
			@indels_rlv2 = @indels_rlv_other_ovlp_ref;
		    }


		    if (@indels_rlv1>0) {

    (print STDOUT "     | (vi) - 4 |\n\n") if ($TEST_LEVEL_newtype>2);

			my ($br1, $state_ue1, $len1) = @{$indels_rlv1[0]}[0, 3, 5];
			my ($br2, $state_ue2, $len2) = @{$indels_rlv2[0]}[0, 3, 5];

			if ( (($br1 == $br2) and ($state_ue1 ne $state_ue2)) or 
				(($br1 != $br2) and ($state_ue1 eq $state_ue2)) ) {

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
			    if ($len1 == $len2) {
				$type = 'Purge(?)'; # It could be an authentic 'Purge'. #
				$penalty = 3;
			    }
			    ##($len1 == $len2) and ($type = 'Purge(?)'); # It could be an authentic 'Purge'. #
			}
		    }

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} elsif (($ct_rlv_indels_rec == 2) and ($ct_rlv_indels_ref == 1)) {

			# Substep (vii): Suspect a 'Split'. #

    (print STDOUT "  { Substep (vii) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my ($br1, $state_ue1, $len1) = @{$indels_rlv_rec[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rlv_rec[1]}[0, 3, 5];
		    my ($br3, $state_ue3, $len3) = @{$indels_rlv_ref[0]}[0, 3, 5];

		  ##if ($len3 == $len1 + $len2) { # A BUG (1) fixed on Nov 7 (Fri), 2014. #

		    my $sub_type; # A DEBUGGING (1) measure on Nov 7 (Fri), 2014. #
		    ##my $sub_type = '???'; # = 'same-type', 'complementary', or '???';
		    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		  if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (($br1 != $br2) and ($state_ue1 ne $state_ue2)) ) {

    (print STDOUT "     | (vii) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

		     if ($len3 == $len1 + $len2) { # A DEBUGGING (1) measure on Nov 7 (Fri), 2014. #

    (print STDOUT "     | (vii) - 1a |\n\n") if ($TEST_LEVEL_newtype>2);

			if ( (($br3 == $br1) and ($state_ue3 eq $state_ue1)) or ($br3 != $br1) and ($state_ue3 ne $state_ue1) ) {

    (print STDOUT "     | (vii) - 1a - i |\n\n") if ($TEST_LEVEL_newtype>2);

					# MODIFIED on Nov 10 (Mon), 2014. #
			    if ( (@indels_rlv_lf_rec ==1) and (@indels_rlv_rf_rec ==1) and
				(scalar (@indels_rlv_lf_ref) + scalar (@indels_rlv_rf_ref) ==1) ) {
				$sub_type = '(same-type)';
			    } else {
				$sub_type = '(same-type)(+shift(^2))(??)';
				$sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }
			    ##( (@indels_rlv_lf_rec ==1) and (@indels_rlv_rf_rec ==1) and
			    ##	(scalar (@indels_rlv_lf_ref) + scalar (@indels_rlv_rf_ref) ==1) ) and ($sub_type = 'same-type');
			}
		     }

		  } elsif ( (($br1 == $br2) and ($state_ue1 ne $state_ue2)) or (($br1 != $br2) and ($state_ue1 eq $state_ue2)) ) {

    (print STDOUT "     | (vii) - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ( (($len3 == $len1 - $len2) and ((($br1 == $br3) and ($state_ue1 eq $state_ue3)) or (($br1 != $br3) and ($state_ue1 ne $state_ue3))))
				or (($len3 == $len2 - $len1) and ((($br2 == $br3) and ($state_ue2 eq $state_ue3)) or (($br2 != $br3) and ($state_ue2 ne $state_ue3)))) ) { # A DEBUGGING (1) measure on Nov 6 (Thu) - 7 (Fri), 2014. #

    (print STDOUT "     | (vii) - 2a |\n\n") if ($TEST_LEVEL_newtype>2);

				# A DEBUGGING (2) measure on Nov 7 (Fri), 2014. #

			  if (@indels_rlv_lf_rec ==1) {

    (print STDOUT "     | (vii) - 2a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			     if (@indels_rlv_lf_ref ==1) { # Short-block (normal).

    (print STDOUT "     | (vii) - 2a-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				(scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_le_rec) ==1) and ($sub_type = '(complementary)');

			     } elsif (@indels_rlv_rf_ref ==1) { # Short-block (+shift).

    (print STDOUT "     | (vii) - 2a-1b |\n\n") if ($TEST_LEVEL_newtype>2);

					# A DEBUGGING (3) measure on Nov 7 (Fri), 2014. #
				if (@indels_rlv_re_rec ==1) {
				    $sub_type = '(complementary)(+shift)(?)';
				    ##$sub_type = 'complementary(+shift(?))';
				    $sbpenalty += 1;  # ADDED in REVISION on Oct 21 (Wed), 2015. #
				} elsif (scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) ==1) {
				    $sub_type = '(complementary)(+shift^2)(?)';
				    ##$sub_type = 'complementary(+shift^2(?))';
				    $sbpenalty += 2;  # ADDED in REVISION on Oct 21 (Wed), 2015. #
				}
				##(@indels_rlv_re_rec ==1) and ($sub_type = 'complementary(+shift(?))');

			     } elsif (@indels_rlv_re_ref ==1) { # Long-block (normal, w/o 'right-bounding' column).

    (print STDOUT "     | (vii) - 2a-1c |\n\n") if ($TEST_LEVEL_newtype>2);

				(@indels_rlv_re_rec ==1) and ($sub_type = '(complementary)');
				##(@indels_rlv_re_rec ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_other_ovlp_ref ==1) { # Long-block (normal, with 'right-bounding' column).

    (print STDOUT "     | (vii) - 2a-1d |\n\n") if ($TEST_LEVEL_newtype>2);

				(scalar(@indels_rlv_other_ovlp_rec) + scalar (@indels_rlv_le_rec) ==1) and ($sub_type = '(complementary)');
				##(scalar(@indels_rlv_other_ovlp_rec) + scalar (@indels_rlv_le_rec) ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_le_ref ==1) { # Long-block (+shift).

    (print STDOUT "     | (vii) - 2a-1e |\n\n") if ($TEST_LEVEL_newtype>2);

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				if (scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_le_rec) ==1) {
				    $sub_type = '(complementary)(+shift)(?)';
				    $sbpenalty += 1;
				}
				##(scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_le_rec) ==1) and ($sub_type = '(complementary)(+shift)(?)');
				##(scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_le_rec) ==1) and ($sub_type = 'complementary(+shift(?))');
			     }

			  } elsif (@indels_rlv_rf_rec ==1) {

    (print STDOUT "     | (vii) - 2a-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			     if (@indels_rlv_rf_ref ==1) { # Short-block (normal).

    (print STDOUT "     | (vii) - 2a-2a |\n\n") if ($TEST_LEVEL_newtype>2);

				(scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_re_rec) ==1) and ($sub_type = '(complementary)');
				##(scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_re_rec) ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_lf_ref ==1) { # Short-block (+shift).

    (print STDOUT "     | (vii) - 2a-2b |\n\n") if ($TEST_LEVEL_newtype>2);

					# A DEBUGGING (3) measure on Nov 7 (Fri), 2014. #
				if (@indels_rlv_le_rec ==1) {
				    $sub_type = '(complementary)(+shift)(?)';
				    ##$sub_type = 'complementary(+shift(?))';
				    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				} elsif (scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) ==1) {
				    $sub_type = '(complementary)(+shift^2)(?)';
				    ##$sub_type = 'complementary(+shift^2(?))';
				    $sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				}
				##(@indels_rlv_le_rec ==1) and ($sub_type = 'complementary(+shift(?))');

			     } elsif (@indels_rlv_le_ref ==1) { # Long-block (normal, w/o 'right-bounding' column).

    (print STDOUT "     | (vii) - 2a-2c |\n\n") if ($TEST_LEVEL_newtype>2);

				(@indels_rlv_le_rec ==1) and ($sub_type = '(complementary)');
				##(@indels_rlv_le_rec ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_other_ovlp_ref ==1) { # Long-block (normal, with 'right-bounding' column).

    (print STDOUT "     | (vii) - 2a-2d |\n\n") if ($TEST_LEVEL_newtype>2);

				(scalar(@indels_rlv_other_ovlp_rec) + scalar (@indels_rlv_re_rec) ==1) and ($sub_type = '(complementary)');
				##(scalar(@indels_rlv_other_ovlp_rec) + scalar (@indels_rlv_re_rec) ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_re_ref ==1) { # Long-block (+shift).

    (print STDOUT "     | (vii) - 2a-2e |\n\n") if ($TEST_LEVEL_newtype>2);

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				if (scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_re_rec) ==1) {
				    $sub_type = '(complementary)(+shift)(?)';
				    $sbpenalty += 1;
				}
				##(scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_re_rec) ==1) and ($sub_type = '(complementary)(+shift)(?)');
				##(scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_re_rec) ==1) and ($sub_type = 'complementary(+shift(?))');
			     }
			  }

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
			  unless (defined $sub_type) {
				$sub_type = '(complementary)(+shift(^2))(??)';
				$sbpenalty += 2; 
			  }
			  ##(defined $sub_type) or ($sub_type = '(complementary)(+shift(^2))(??)'); # ADDED on Nov 10 (Mon), 2014. #

=A BUG (2) fixed on Nov 7 (Fri), 2014.
			if ( ((@indels_rlv_lf_rec ==1) and (@indels_rlv_re_rec ==1)) or
				((@indels_rlv_le_rec ==1) and (@indels_rlv_rf_rec ==1)) ) { # When the block is bounded by the gaps in reconstructed. #

			    (@indels_rlv_other_ovlp_ref ==0) and ($sub_type = '(complementary)');
			    ##(@indels_rlv_other_ovlp_ref ==0) and ($sub_type = 'complementary');

			} elsif ( ((@indels_rlv_lf_rec ==1) and (@indels_rlv_other_ovlp_rec ==1)) or
				((@indels_rlv_other_ovlp_rec ==1) and (@indels_rlv_rf_rec ==1)) ) { # When the block extends over the gap in reference.

			    (@indels_rlv_other_ovlp_ref ==1) and ($sub_type = '(complementary)');
			    ##(@indels_rlv_other_ovlp_ref ==1) and ($sub_type = 'complementary');
			}
=cut # END of "A BUG (2) fixed on Nov 7 (Fri), 2014."

		     } # END of "if ( (($len3 == $len1 - $len2) and ((($br1 == $br3) and ($state_ue1 eq $state_ue3)) or ())) or (...) ) {...}"
		  }

		  ##} else {...} # A BUG (1) fixed on Nov 7 (Fri), 2014. #

		  if (defined $sub_type) { # A DEBUGGING (1) measure on Nov 7 (Fri), 2014. #

		    $type = "Split$sub_type"; # MODIFIED on Nov 10 (Mon), 2014.
		    ##$type = "Split($sub_type)";
		    $penalty = 2 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		  } else {
			$type = 'Complex(???)';
			$penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #
		  }


		} elsif (($ct_rlv_indels_ref == 2) and ($ct_rlv_indels_rec == 1)) {

			# Substep (viii): Suspect a 'Merge'. #

    (print STDOUT "  { Substep (viii) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my ($br1, $state_ue1, $len1) = @{$indels_rlv_ref[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rlv_ref[1]}[0, 3, 5];
		    my ($br3, $state_ue3, $len3) = @{$indels_rlv_rec[0]}[0, 3, 5];

		  ##if ($len3 == $len1 + $len2) { # A BUG (1) fixed on Nov 6 (Thu), 2014. #

		    my $sub_type; # A DEBUGGING (1) measure on Nov 7 (Fri), 2014. #
		    ##my $sub_type = '???'; # = 'same-type', 'complementary', or '???';
		    my $sbpenalty = 0; # ADEED in REVISION on Oct 21 (Wed), 2015. #

		    if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (($br1 != $br2) and ($state_ue1 ne $state_ue2)) ) {

    (print STDOUT "     | (viii) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ($len3 == $len1 + $len2) { # A DEBUGGING (1) measure on Nov 6 (Thu), 2014. # 

    (print STDOUT "     | (viii) - 1a |\n\n") if ($TEST_LEVEL_newtype>2);

			    if ( (($br3 == $br1) and ($state_ue3 eq $state_ue1)) or ($br3 != $br1) and ($state_ue3 ne $state_ue1) ) {

    (print STDOUT "     | (viii) - 1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					# MODIFIED on Nov 10 (Mon), 2014. #
				if ( (@indels_rlv_lf_ref ==1) and (@indels_rlv_rf_ref ==1) and 
			           (scalar (@indels_rlv_lf_rec) + scalar (@indels_rlv_rf_rec) ==1) ) {
					$sub_type = '(same-type)';
				} else {
					$sub_type = '(same-type)(+shift(^2))(??)';
					$sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				}
				##( (@indels_rlv_lf_ref ==1) and (@indels_rlv_rf_ref ==1) and 
			        ##   (scalar (@indels_rlv_lf_rec) + scalar (@indels_rlv_rf_rec) ==1) ) and ($sub_type = 'same-type');
			    }
			}

		    } elsif ( (($br1 == $br2) and ($state_ue1 ne $state_ue2)) or (($br1 != $br2) and ($state_ue1 eq $state_ue2)) ) {

    (print STDOUT "     | (viii) - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ( (($len3 == $len1 - $len2) and ((($br1 == $br3) and ($state_ue1 eq $state_ue3)) or (($br1 != $br3) and ($state_ue1 ne $state_ue3))))
				or (($len3 == $len2 - $len1) and ((($br2 == $br3) and ($state_ue2 eq $state_ue3)) or (($br2 != $br3) and ($state_ue2 ne $state_ue3)))) ) { # A DEBUGGING (1) measure on Nov 6 (Thu) - 7 (Fri), 2014. #

    (print STDOUT "     | (viii) - 2a |\n\n") if ($TEST_LEVEL_newtype>2);

				# A DEBUGGING (2) measure on Nov 7 (Fri), 2014. #

			  if (@indels_rlv_lf_ref ==1) {

    (print STDOUT "     | (viii) - 2a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			     if (@indels_rlv_lf_rec ==1) { # Short-block (normal).

    (print STDOUT "     | (viii) - 2a-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				(scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_le_ref) ==1) and ($sub_type = '(complementary)');
				##(scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_le_ref) ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_rf_rec ==1) { # Short-block (+shift).

    (print STDOUT "     | (viii) - 2a-1b |\n\n") if ($TEST_LEVEL_newtype>2);

					# A DEBUGGING (3) measure on Nov 7 (Fri), 2014. #
				if (@indels_rlv_re_ref ==1) {
				    $sub_type = '(complementary)(+shift)(?)';
				    ##$sub_type = 'complementary(+shift(?))';
				    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				} elsif (scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) ==1) {
				    $sub_type = '(complementary)(+shift^2)(?)';
				    ##$sub_type = 'complementary(+shift^2(?))';
				    $sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				}
				##(@indels_rlv_re_ref ==1) and ($sub_type = 'complementary(+shift(?))');

			     } elsif (@indels_rlv_re_rec ==1) { # Long-block (normal, w/o 'right-bounding' column).

    (print STDOUT "     | (viii) - 2a-1c |\n\n") if ($TEST_LEVEL_newtype>2);

				(@indels_rlv_re_ref ==1) and ($sub_type = '(complementary)');
				##(@indels_rlv_re_ref ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_other_ovlp_rec ==1) { # Long-block (normal, with 'right-bounding' column).

    (print STDOUT "     | (viii) - 2a-1d |\n\n") if ($TEST_LEVEL_newtype>2);

				(scalar(@indels_rlv_other_ovlp_ref) + scalar (@indels_rlv_le_ref) ==1) and ($sub_type = '(complementary)');
				##(scalar(@indels_rlv_other_ovlp_ref) + scalar (@indels_rlv_le_ref) ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_le_rec ==1) { # Long-block (+shift).

    (print STDOUT "     | (viii) - 2a-1e |\n\n") if ($TEST_LEVEL_newtype>2);

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				if (scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_le_ref) ==1) {
				    $sub_type = '(complementary)(+shift)(?)';
				    $sbpenalty += 1;
				}
				##(scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_le_ref) ==1) and ($sub_type = '(complementary)(+shift)(?)');
				##(scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_le_ref) ==1) and ($sub_type = 'complementary(+shift(?))');
			     }

			  } elsif (@indels_rlv_rf_ref ==1) {

    (print STDOUT "     | (viii) - 2a-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			     if (@indels_rlv_rf_rec ==1) { # Short-block (normal).

    (print STDOUT "     | (viii) - 2a-2a |\n\n") if ($TEST_LEVEL_newtype>2);

				(scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_re_ref) ==1) and ($sub_type = '(complementary)');
				##(scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_re_ref) ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_lf_rec ==1) { # Short-block (+shift).

    (print STDOUT "     | (viii) - 2a-2b |\n\n") if ($TEST_LEVEL_newtype>2);

					# A DEBUGGING (3) measure on Nov 7 (Fri), 2014. #
				if (@indels_rlv_le_ref ==1) {
				    $sub_type = '(complementary)(+shift)(?)';
				    ##$sub_type = 'complementary(+shift(?))';
				    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				} elsif (scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) ==1) {
				    $sub_type = '(complementary)(+shift^2)(?)';
				    ##$sub_type = 'complementary(+shift^2(?))';
				    $sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				}
				##(@indels_rlv_le_ref ==1) and ($sub_type = 'complementary(+shift(?))');

			     } elsif (@indels_rlv_le_rec ==1) { # Long-block (normal, w/o 'right-bounding' column).

    (print STDOUT "     | (viii) - 2a-2c |\n\n") if ($TEST_LEVEL_newtype>2);

				(@indels_rlv_le_ref ==1) and ($sub_type = '(complementary)');
				##(@indels_rlv_le_ref ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_other_ovlp_rec ==1) { # Long-block (normal, with 'right-bounding' column).

    (print STDOUT "     | (viii) - 2a-2d |\n\n") if ($TEST_LEVEL_newtype>2);

				(scalar(@indels_rlv_other_ovlp_ref) + scalar (@indels_rlv_re_ref) ==1) and ($sub_type = '(complementary)');
				##(scalar(@indels_rlv_other_ovlp_ref) + scalar (@indels_rlv_re_ref) ==1) and ($sub_type = 'complementary');

			     } elsif (@indels_rlv_re_rec ==1) { # Long-block (+shift).

    (print STDOUT "     | (viii) - 2a-2e |\n\n") if ($TEST_LEVEL_newtype>2);

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. # 
				if (scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_re_ref) ==1) {
				    $sub_type = '(complementary)(+shift)(?)';
				    $sbpenalty += 1;
				}
				##(scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_re_ref) ==1) and ($sub_type = '(complementary)(+shift)(?)');
				##(scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_re_ref) ==1) and ($sub_type = 'complementary(+shift(?))');
			     }
			  }

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
			  unless (defined $sub_type) {
				$sub_type = '(complementary)(+shift(^2))(??)';
				$sbpenalty += 2;
			  }
			  ##(defined $sub_type) or ($sub_type = '(complementary)(+shift(^2))(??)'); # ADDED on Nov 10 (Mon), 2014. #


=A BUG (2) fixed on Nov 6 (Thu), 2014.
			    if ( ((@indels_rlv_lf_ref ==1) and (@indels_rlv_re_ref ==1)) or
				((@indels_rlv_le_ref ==1) and (@indels_rlv_rf_ref ==1)) ) { # When the block is bounded by the gaps in reference.

				(@indels_rlv_other_ovlp_rec ==0) and ($sub_type = 'complementary');

			    } elsif ( ((@indels_rlv_lf_ref ==1) and (@indels_rlv_other_ovlp_ref ==1)) or
				((@indels_rlv_other_ovlp_ref ==1) and (@indels_rlv_rf_ref ==1)) ) { # When the block extends over the gap in reconstructed.

				(@indels_rlv_other_ovlp_rec ==1) and ($sub_type = 'complementary');
			    }
=cut # END of "A BUG (2) fixed on Nov 6 (Thu), 2014."

			} # END of "if ( (($len3 == $len1 - $len2) and ((($br1 == $br3) and ($state_ue1 eq $state_ue3)) or (..) ) ) or ()) {...}"

		    }

		  #} else {...} # A BUG (1) fixed on Nov 6 (Thu), 2014. #

		  if (defined $sub_type) { # A DEBUGGING (1) measure on Nov 6 (Thu), 2014. #

				# MODIFIED on Nov 10 (Mon), 2014. #
		    $type = "Merge$sub_type"; 
		    ##$type = "Merge($sub_type)(?)";
		    $penalty = 2 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		  } else {
			$type = 'Complex(???)';
			$penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #
		  }


		} elsif (($ct_rlv_indels_ref == 2) and ($ct_rlv_indels_rec == 2)) {

			# Substep (ix): I don't know ... may be a 'Partial-Shift', or a 'Partial-Merge/Split' ? #

    (print STDOUT "  { Substep (ix) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my ($br1, $state_ue1, $len1) = @{$indels_rlv_ref[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rlv_ref[1]}[0, 3, 5];
		    my ($br3, $state_ue3, $len3) = @{$indels_rlv_rec[0]}[0, 3, 5];
		    my ($br4, $state_ue4, $len4) = @{$indels_rlv_rec[1]}[0, 3, 5];

		    ##$type = 'Complex(???)'; # = 'Complex(???)', 'Partial-Shift(?)', or 'Partial-Merge/Split(?)'. 

		    if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (($br1 != $br2) and ($state_ue1 ne $state_ue2)) ) {
    (print STDOUT "     | (ix) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ( (($br1 == $br3) and ($state_ue1 eq $state_ue3)) or (($br1 != $br3) and ($state_ue1 ne $state_ue3)) ) {

    (print STDOUT "     | (ix) - 1a |\n\n") if ($TEST_LEVEL_newtype>2);

			    if ( (($br3 == $br4) and ($state_ue3 eq $state_ue4)) or (($br3 != $br4) and ($state_ue3 ne $state_ue4)) ) {

    (print STDOUT "     | (ix) - 1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				# Examine whether the relevant indels are in the ideal positional relationships with the composite-block. #
				if ( (@indels_rlv_lf_rec ==1) and (@indels_rlv_rf_rec ==1) and (@indels_rlv_lf_ref ==1) and (@indels_rlv_rf_ref ==1) ) {
    (print STDOUT "     | (ix) - 1a-1a |\n\n") if ($TEST_LEVEL_newtype>2);

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				    if ($len1 + $len2 == $len3 + $len4) {
					$type = 'Partial-Shift(?)';
					$penalty = 3;
				    }
				    ##($len1 + $len2 == $len3 + $len4) and ($type = 'Partial-Shift(?)');
				}
			    }
			}

		    } elsif ( (($br1 == $br2) and ($state_ue1 ne $state_ue2)) or (($br1 != $br2) and ($state_ue1 eq $state_ue2)) ) {

    (print STDOUT "     | (ix) - 2 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ( (($br3 == $br4) and ($state_ue3 ne $state_ue4)) or (($br3 != $br4) and ($state_ue3 eq $state_ue4)) ) {

    (print STDOUT "     | (ix) - 2a |\n\n") if ($TEST_LEVEL_newtype>2);

			    my $diff_len_ref = ($len1 > $len2) ? $len1 - $len2 : $len2 - $len1;
			    my $diff_len_rec = ($len3 > $len4) ? $len3 - $len4 : $len4 - $len3;

			    if ($diff_len_ref == $diff_len_rec) { 

    (print STDOUT "     | (ix) - 2a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				# Examine whether the relevant indels are in the ideal positional relationships with the composite-block. #
			        if ((@indels_rlv_lf_ref ==1) and (scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) ==1) ) {

    (print STDOUT "     | (ix) - 2a-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				    if ((@indels_rlv_lf_rec ==1) and (scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) ==1)) {

    (print STDOUT "     | (ix) - 2a-1a1 |\n\n") if ($TEST_LEVEL_newtype>2);

					$type = ($len1 + $len2 > $len3 + $len4) ? 'Partial-Merge(?)' : 'Partial-Split(?)';
					$penalty = 3; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif ((scalar(@indels_rlv_le_rec) +scalar(@indels_rlv_other_ovlp_rec) ==1) and (@indels_rlv_rf_rec ==1)) {

    (print STDOUT "     | (ix) - 2a-1a2 |\n\n") if ($TEST_LEVEL_newtype>2);

					$type = 'Overshoot-Merge(??)';
					$penalty = 4; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }

			        } elsif ((scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) ==1) and (@indels_rlv_rf_ref ==1)) {

    (print STDOUT "     | (ix) - 2a-1b |\n\n") if ($TEST_LEVEL_newtype>2);

				    if ((scalar(@indels_rlv_le_rec) +scalar(@indels_rlv_other_ovlp_rec) ==1) and (@indels_rlv_rf_rec ==1)) {

    (print STDOUT "     | (ix) - 2a-1b1 |\n\n") if ($TEST_LEVEL_newtype>2);

					$type = ($len1 + $len2 > $len3 + $len4) ? 'Partial-Merge(?)' : 'Partial-Split(?)';
					$penalty = 3; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif ((@indels_rlv_lf_rec ==1) and (scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) ==1)) {

    (print STDOUT "     | (ix) - 2a-1b2 |\n\n") if ($TEST_LEVEL_newtype>2);

					$type = 'Overshoot-Merge(??)';
					$penalty = 4; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
			        }
			    } # END of "if ($diff_len_ref == $diff_len_rec) {...}"
			}
		    } # END of "if ( (($br1 == $br2) and ...) or ...) {...} elsif ( (($br1 == $br2) and ...) or ...) {...}"

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)' ;
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)') ;

		} else {

			# Substep (x): If the subject comes up to here, Regard the associated error as 'Complex'. #

    (print STDOUT "  { Substep (x) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'Complex';
		    $penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		}

				# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
				# => Further MODIFIED in REVISION on Oct 21 (Wed), 2015. #

		return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty);
		##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);


		##		# Finally, register the association relationship between the composite-block and the indel events considered to be involved. #

		##unless ( register_association_cblk_and_indels ($type, $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @indels_invlvd_ref, @indels_invlvd_rec, @indices_indel_identifiers) ) {

		##	print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED at the Step 1 on the $cb th composite shift-block for the $s th segment !\n";
		##	print STDERR "[ERROR '$program_name'] $!\n\n";
		##	die ;
		##}
		##next;

	    } # END of "if (($ct_rltd_indels_rec == 0) and ($ct_rltd_indels_ref == 0)) {...}"


			# Step 2: When involving only one event on a relative branch. #

	    if (($ct_rltd_indels_rec == 1) and ($ct_rltd_indels_ref == 0)) {

				# Step 2A. #

    (print STDOUT "< Step 2A >\n\n") if ($TEST_LEVEL_newtype>0);

		my $type; # The type of MSA error(s). #
		my $penalty; # The penalty incurred by the error. # ADDED in REVISION on Oct 21 (Wed), 2015. # 

		if (($ct_rlv_indels_ref == 0) and ($ct_rlv_indels_rec == 0)) {

			# Substep (o): The composite block is NOT directly associated with any errorneous indel event (???) #

    (print STDOUT "  { Substep (0) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'NO_ASSOCIATED_EVENT(???)';
		    $penalty = 200; # ADDED in REVISION on Oct 21 (Wed), 2015. #


		} elsif (($ct_rlv_indels_rec == 0) and ($ct_rlv_indels_ref == 1)) {

			# Substep (i): Suspect a 'Collapse of Independent Insertions'. #

    (print STDOUT "  { Substep (i) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my $subtype;
		    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #
		    my @indels_invlvd_ref1 = my @indels_invlvd_rec1 = ();

		    if (scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) == 0) {

    (print STDOUT "     | (i) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((@{$indels_le_rec} >= 2) and (@{$indels_lf_ref} ==1)) {

    (print STDOUT "     | (i) - 1a |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = (@{$indels_lf_ref}, @indels_rlv_ref);
			    @indels_invlvd_rec1 = @{$indels_le_rec};

			    if (scalar(@{$indels_rf_rec}) - scalar(@indels_rltd_rf_rec) > 0) {
				$subtype = '(+slide(?))';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			    } elsif (@indels_rlv_le_ref ==1) {
				$subtype = '';
			    } else {
				$subtype = '(+shift)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				if ($applen_cb_rec > $applen_cb_ref) {
				    $subtype .= '(+overshoot(?))';
				    $sbpenalty += 1;
				}
				##($applen_cb_rec > $applen_cb_ref) and ($subtype .= '(+overshoot(?))'); # ADDED for DEBUGGING on Oct 4 (Sun), 2015. #
			    }

			} elsif ((@{$indels_re_rec} >= 2) and (@{$indels_rf_ref} ==1)) {

    (print STDOUT "     | (i) - 1b |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = (@{$indels_rf_ref}, @indels_rlv_ref);
			    @indels_invlvd_rec1 = @{$indels_re_rec};

			    if (scalar(@{$indels_lf_rec}) - scalar(@indels_rltd_lf_rec) > 0) {
				$subtype = '(+slide(?))';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			    } elsif (@indels_rlv_re_ref ==1) {
				$subtype = '';
			    } else {
				$subtype = '(+shift)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. # 
				if ($applen_cb_rec > $applen_cb_ref) {
				    $subtype .= '(+overshoot(?))';
				    $sbpenalty += 1;
				}
				##($applen_cb_rec > $applen_cb_ref) and ($subtype .= '(+overshoot(?))'); # ADDED for DEBUGGING on Oct 4 (Sun), 2015. #
			    }

				# ADDED the following conditions on Nov 10 (Mon), 2014. #

			} elsif ((@{$indels_le_rec} >= 2) and (@{$indels_rf_ref} ==1)) {

    (print STDOUT "     | (i) - 1c |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = (@{$indels_rf_ref}, @indels_rlv_ref);
			    @indels_invlvd_rec1 = @{$indels_le_rec};
			    $subtype = (@indels_rlv_re_ref ==1) ? '(+shift^2(??))' : '(+overshoot(??))';
			    $sbpenalty += (@indels_rlv_re_ref==1) ? 2 : 1; # ADDED in REVISION on Oct 21 (Wed), 2015. # 

			} elsif ((@{$indels_re_rec} >= 2) and (@{$indels_lf_ref} ==1)) {

    (print STDOUT "     | (i) - 1d |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = (@{$indels_lf_ref}, @indels_rlv_ref);
			    @indels_invlvd_rec1 = @{$indels_re_rec};
			    $subtype = (@indels_rlv_le_ref ==1) ? '(+shift^2(??))' : '(+overshoot(??))';
			    $sbpenalty += (@indels_rlv_le_ref==1) ? 2 : 1; # ADDED in REVISION on Oct 21 (Wed), 2015. # 

				# END of "ADDED the following conditions on Nov 10 (Mon), 2014.} #
			}

		    } # END of "if (scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) ==0) {...}"

		    if (defined $subtype) {

    (print STDOUT "     | (i) - Test |\n\n") if ($TEST_LEVEL_newtype>2);

			my ($if_cii, $indels_invlvd_rec2) = if_collapse_indep_ins (@indels_invlvd_ref1, @indels_invlvd_rec1, %{$node2pa}, %{$node2ch}, $CHAR_ABS);
			unless (defined $if_cii) {
			    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i) of Step 2A: $! !\n\n";
			    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);


			    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i) of Step 2A on the $cb th composite shift-block for the $s th segment !\n";
			    ##print STDERR "[ERROR '$program_name'] $!\n\n";
			    ##die ;
			}

			if ($if_cii) {

    (print STDOUT "     | (i) - Test-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref = @indels_invlvd_ref1;
			    @indels_invlvd_rec = @{$indels_invlvd_rec2};
			    my $diffct = scalar (@indels_invlvd_rec) - @indels_invlvd_ref;

			    $cblk_wise_cts_invlvd_indels[$cb]->[5] = scalar (@{$indels_invlvd_rec2})-1; # ADDED on Nov 6 (Thu), 2014. #

			    my $len1 = $indels_invlvd_ref[0]->[5];
			    my $len2 = $indels_invlvd_rec[0]->[5];
			    if ($len1 == $len2) {
				$type = "Collapse-of-Independent-Insertions$subtype(+$diffct)(?)";
				$penalty = 10 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    } else {
				$diffct++;
				$type = "Incomplete-Collapse-of-Independent-Insertions$subtype(+$diffct)(?)";
				$penalty = 15 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }
			    
			}
		    }

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} elsif (($ct_rlv_indels_rec == 1) and ($ct_rlv_indels_ref == 0)) {

			# Substep (ii): Suspect an 'Ex-nihilo' aligned with a run of gaps. #

    (print STDOUT "  { Substep (ii) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my $len1 = $indels_rlv_rec[0]->[5];
		    my $len2 = $indels_rltd_rec[0]->[5];

		  if ($len1 == $len2) {

    (print STDOUT "     | (ii) - 1 |\n\n") if ($TEST_LEVEL_newtype>2);

		    if ( ((@indels_rlv_lf_rec ==1) and (@indels_rltd_re_rec ==1)) or 
				((@indels_rlv_rf_rec ==1) and (@indels_rltd_le_rec ==1)) ) {

    (print STDOUT "     | (ii) - 1a |\n\n") if ($TEST_LEVEL_newtype>2);

				# MODIFIED for DEBUGGING (on Nov 10 (Mon), 2014). 
		        my ($br1, $state_ue1) = @{$indels_rlv_rec[0]}[0, 3];
		        my ($br2, $state_ue2) = @{$indels_rltd_rec[0]}[0, 3];

			if ($br2 == $node2pa->{$br1}) {

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
			    if (($state_ue2 eq $CHAR_PRES) and ($state_ue1 eq $CHAR_ABS)) {
				$type = 'Ex-nihilo(gap-aligned)(?)';
				$penalty = 3 + 1;
			    }
			    ##(($state_ue2 eq $CHAR_PRES) and ($state_ue1 eq $CHAR_ABS)) and ($type = 'Ex-nihilo(gap-aligned)(?)');
			} elsif ($br1 == $node2pa->{$br2}) {

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
			    if (($state_ue1 eq $CHAR_PRES) and ($state_ue2 eq $CHAR_ABS)) {
				$type = 'Ex-nihilo(gap-aligned)(?)';
				$penalty = 3 + 1;
			    }
			    ##(($state_ue1 eq $CHAR_PRES) and ($state_ue2 eq $CHAR_ABS)) and ($type = 'Ex-nihilo(gap-aligned)(?)');
			} else {

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
			    if (($state_ue1 eq $CHAR_ABS) and ($state_ue2 eq $CHAR_ABS)) {
				$type = 'Ex-nihilo(gap-aligned)(?)';
				$penalty = 3 + 1;
			    }
			    ##(($state_ue1 eq $CHAR_ABS) and ($state_ue2 eq $CHAR_ABS)) and ($type = 'Ex-nihilo(gap-aligned)(?)');
			}
			##$type = 'Ex-nihilo(gap-aligned)(?)';

		    ##} elsif ( ((@indels_rlv_lf_rec ==1) and (@indels_rltd_other_ovlp_rec ==1)) or 
		    ##		((@indels_rlv_rf_rec ==1) and (@indels_rltd_other_ovlp_rec ==1)) ) {
		    ##	$type = 'Ex-nihilo(gap-aligned)(??)';
		    }

		  }

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} elsif (($ct_rlv_indels_rec == 1) and ($ct_rlv_indels_ref == 1)) {

			# Substep (iii): 

    (print STDOUT "  { Substep (iii) }\n\n") if ($TEST_LEVEL_newtype>1);

				# A: First, suspect an 'Incomplete-Collapse' of Independent Insertions. # 

		    my $subtype;
		    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. # 
		    my @indels_invlvd_ref1 = my @indels_invlvd_rec1 = my @indels_invlvd_rec1p = ();

		    if ((scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) == 0) and (scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) ==0) ) {

    (print STDOUT "     | (iii) - A1 |\n\n") if ($TEST_LEVEL_newtype>2);


			if ((@{$indels_le_rec} >= 2) and (@{$indels_lf_ref} ==1)) {

    (print STDOUT "     | (iii) - A1-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = (@{$indels_lf_ref}, @indels_rlv_ref);
			    @indels_invlvd_rec1 = @{$indels_le_rec};
			    @indels_invlvd_rec1p = @indels_rlv_rec;

			    if (@indels_rlv_le_ref ==1) {
				$subtype = '';
			    } else {
				$subtype = '(+shift)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }

			} elsif ((@{$indels_re_rec} >= 2) and (@{$indels_rf_ref} ==1)) {

    (print STDOUT "     | (iii) - A1-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = (@{$indels_rf_ref}, @indels_rlv_ref);
			    @indels_invlvd_rec1 = @{$indels_re_rec};
			    @indels_invlvd_rec1p = @indels_rlv_rec;

			    if (@indels_rlv_re_ref ==1) {
				$subtype = '';
			    } else {
				$subtype = '(+shift)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }

				# ADDED the following conditions (on Nov 10 (Mon), 2014). #

			} elsif ((@{$indels_le_rec} >= 2) and (@{$indels_rf_ref} ==1)) {

    (print STDOUT "     | (iii) - A1-3 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = (@{$indels_rf_ref}, @indels_rlv_ref);
			    @indels_invlvd_rec1 = @{$indels_le_rec};
			    @indels_invlvd_rec1p = @indels_rlv_rec;
			    $subtype = '(+slide(?))';
			    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@{$indels_re_rec} >= 2) and (@{$indels_lf_ref} ==1)) {

    (print STDOUT "     | (iii) - A1-4 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = (@{$indels_lf_ref}, @indels_rlv_ref);
			    @indels_invlvd_rec1 = @{$indels_re_rec};
			    @indels_invlvd_rec1p = @indels_rlv_rec;
			    $subtype = '(+slide(?))';
			    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				# END of "ADDED the following conditions (on Nov 10 (Mon), 2014)." #
			}


				# ADDED the following conditions (on Nov 10 (Mon), 2014). #

		    } elsif (scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) == 0) {

    (print STDOUT "     | (iii) - A2 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((@indels_rlv_lf_ref ==1) and (@{$indels_le_rec} >= 2) and (@{$indels_other_ovlp_ref} > 0)) {

    (print STDOUT "     | (iii) - A2-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = @indels_rlv_lf_ref;
			    @indels_invlvd_rec1 = @{$indels_le_rec};
			    @indels_invlvd_rec1p = @indels_rlv_rec;

			    foreach my $event (@{$indels_other_ovlp_ref}) {
				($event->[2] +1 == $beg_cb_ref) and (push @indels_invlvd_ref1, $event);
			    }

			} elsif ((@indels_rlv_rf_ref ==1) and (@{$indels_re_rec} >= 2) and (@{$indels_other_ovlp_ref} > 0)) {

    (print STDOUT "     | (iii) - A2-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref1 = @indels_rlv_rf_ref;
			    @indels_invlvd_rec1 = @{$indels_re_rec};
			    @indels_invlvd_rec1p = @indels_rlv_rec;

			    foreach my $event (@{$indels_other_ovlp_ref}) {
				($end_cb_ref + 1 == $event->[1]) and (push @indels_invlvd_ref1, $event);
			    }
			}

			if (@indels_invlvd_ref1 ==2) {
			    $subtype = '(+shift^2(??))';
			    $sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			}
				# END of "ADDED the following conditions (on Nov 10 (Mon), 2014)." #

		    } # END of "if ((scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) ==0) and (...)) {...} elsif (scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) == 0) {...}"

		    if (defined $subtype) {

    (print STDOUT "     | (iii) - A-Test |\n\n") if ($TEST_LEVEL_newtype>2);

			my ($if_cii, $indels_invlvd_rec2) = if_collapse_indep_ins (@indels_invlvd_ref1, @indels_invlvd_rec1, %{$node2pa}, %{$node2ch}, $CHAR_ABS);
			unless (defined $if_cii) {
			    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (iii)-A of Step 2A: $! !\n\n";
			    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

			    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (iii)-A of Step 2A on the $cb th composite shift-block for the $s th segment !\n";
			    ##print STDERR "[ERROR '$program_name'] $!\n\n";
			    ##die ;
			}

			if ($if_cii) {

    (print STDOUT "     | (iii) - A-Test-pass |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref = @indels_invlvd_ref1;
			    @indels_invlvd_rec = (@{$indels_invlvd_rec2}, @indels_invlvd_rec1p);
			    my $diffct = scalar (@indels_invlvd_rec) - @indels_invlvd_ref;

			    $cblk_wise_cts_invlvd_indels[$cb]->[5] = scalar (@{$indels_invlvd_rec2}) -1; # ADDED on Nov 6 (Thu), 2014. # 

			    my $len1 = $indels_invlvd_ref[0]->[5];
			    my $len2 = $indels_invlvd_rec[0]->[5];
			    if ($len1 == $len2) {
				$type = "Incomplete-Collapse-of-Independent-Insertions$subtype(+$diffct)(?)";
				$penalty = 15 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    } else {
				$diffct++;
				##$type = "Incomplete-Collapse-of-Independent-Insertions$subtype(+$diffct)(?)"; # I think it dubious...
			    }
			    
			}
		    }


					# ADDED on Nov 12 (Wed), 2014. #

		    if (!(defined $type) and (@indels_rltd_other_ovlp_ref ==1) and (scalar(@indels_rltd_lf_rec) + scalar(@indels_rltd_rf_rec) ==1) and (scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_re_ref) ==1) and (scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) == 0) ) {

				# B: In this case, Suspect a 'Partial-Neighboring-Branch-Split' (2ins) (Long-Block(normal)). #

    (print STDOUT "     | (iii) - B |\n\n") if ($TEST_LEVEL_newtype>2);

			my $event1 = $indels_rltd_rec[0];
			my $event2 = $indels_rlv_rec[0];

			my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$event1}[0..3, 5];
			my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$event2}[0..3, 5];
			my $eq_br1 = $equiv_branches{$br1};
			my $eq_br2 = $equiv_branches{$br2};

			if ( ($len2 > $len1) and (($end1 + 1 < $beg2) or ($end2 + 1 < $beg1)) ) { # 'rec_rlv' is longer than 'rec_rltd', and the two events are separated.

    (print STDOUT "     | (iii) - B-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my $event3 = $indels_rltd_other_ovlp_ref[0];
			    my $event4 = $indels_rlv_ref[0];
			    my ($br3, $beg3, $end3, $state_ue3, $len3) = @{$event3}[0..3, 5];
			    my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event4}[0..3, 5];

					# See if 'rltd_ref' is on a different branch than 'rltd_rec'.
			    my $if_event3 = (($br1 != $br3) and !((defined $eq_br1) and ($eq_br1 == $br3)) ) ? 1 : 0;

					# See if 'rlv_ref' is of the same type as 'rlv_rec'.
			    my $if_event4 = ( (($br2 == $br4) and ($state_ue2 eq $state_ue4)) or (($br2 != $br4) and ($state_ue2 ne $state_ue4)) ) ? 1 : 0;

			    if (($if_event4) and ($if_event3) and ($len1 == $event3->[5]) and ($len2 == $len1 + $event4->[5])) {

    (print STDOUT "     | (iii) - B-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				my $if_part_nbr_mrg = if_part_nbr_mrg_2ins (@{$event1}, @{$event2}, @{$event3}, 0, 3, %equiv_branches, %{$node2pa}, $CHAR_ABS) ;

				unless (defined $if_part_nbr_mrg) {
				    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (iii)-B of Step 2A: $! !\n\n";
				    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);


				    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (iii)-B of Step 2A on the $cb th composite shift-block for the $s th segment !\n";
				    ##print STDERR "[ERROR '$program_name'] $!\n\n";
				    ##die ;
				}

				if ($if_part_nbr_mrg) {

    (print STDOUT "     | (iii) - B-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				    my $subtype;
				    my $sbpenalty = 0;  # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    if (@indels_rlv_le_ref == 1) {

    (print STDOUT "     | (iii) - B-1a-1a |\n\n") if ($TEST_LEVEL_newtype>2);

					if ($end4 +1 == $beg3) {
					    $subtype = (@indels_rltd_lf_rec ==1) ? '(+overshoot)(?)' : '(+shift)(?)' ;
					    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}

				    } else { # @indels_rlv_re_ref ==1. #

    (print STDOUT "     | (iii) - B-1a-1b |\n\n") if ($TEST_LEVEL_newtype>2);

					if ($end3 +1 == $beg4) {
					    $subtype = (@indels_rltd_rf_rec ==1) ? '(+overshoot)(?)' : '(+shift)(?)' ;
					    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }

				    if (defined $subtype) {
					$type = "Partial-Neighboring-Branch-Split(2ins)$subtype";
					$penalty = 5 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }

				} # END of "if ($if_part_nbr_mrg) {...}"
			    }

			} # END of "if (($len2 > $len1) and (...)) {...}"

		    } # END of "if (!(defined $type) and (@indels_rltd_other_ovlp_ref ==1) and ...) {...}"

					# END of "ADDED on Nov 12 (Wed), 2014." #


		unless (defined $type) {

				# C: Then, Suspect a simple 'Shift', accompanied by an event along a related branch that happened to be near the composite block (in reconstructed MSA),
				# Or, maybe an 'Incomplete-split' or an 'Inter-branch-transfer'.

    (print STDOUT "     | (iii) - C |\n\n") if ($TEST_LEVEL_newtype>2);

		    my ($br1, $state_ue1, $len1) = @{$indels_rlv_ref[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rlv_rec[0]}[0, 3, 5];
		    my ($br3, $state_ue3) = @{$indels_rltd_rec[0]}[0, 3];

		    if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (($br1 != $br2) and ($state_ue1 ne $state_ue2)) ) {
				# At least the relevant events are related by a 'Shift'-like move. #

    (print STDOUT "     | (iii) - C-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ( ((@indels_rlv_lf_ref ==1) and (@indels_rlv_rf_rec ==1) and (@indels_rltd_lf_rec ==1)) or 
			     ((@indels_rlv_rf_ref ==1) and (@indels_rlv_lf_rec ==1) and (@indels_rltd_rf_rec ==1)) ) {

    (print STDOUT "     | (iii) - C-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				# Whether a 'Shift' or an 'Incomplete-spllit'.

			  if ($len1 == $len2) {

    (print STDOUT "     | (iii) - C-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			     if ($br3 == $node2pa->{$br1}) {

    (print STDOUT "     | (iii) - C-1a-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				$type = (($state_ue1 eq $CHAR_PRES) and ($state_ue3 eq $CHAR_ABS)) ? 'Shift(??)' : 'Incomplete-split(?)' ;
				$penalty = (($state_ue1 eq $CHAR_PRES) and ($state_ue3 eq $CHAR_ABS)) ? 1.5 : 2 ; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			     } elsif ($br1 == $node2pa->{$br3}) {

    (print STDOUT "     | (iii) - C-1a-1b |\n\n") if ($TEST_LEVEL_newtype>2);

				$type = (($state_ue3 eq $CHAR_PRES) and ($state_ue1 eq $CHAR_ABS)) ? 'Shift(??)' : 'Incomplete-split(?)' ;
				$penalty = (($state_ue3 eq $CHAR_PRES) and ($state_ue1 eq $CHAR_ABS)) ? 1.5 : 2 ; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			     } else {

    (print STDOUT "     | (iii) - C-1a-1c |\n\n") if ($TEST_LEVEL_newtype>2);

				$type = (($state_ue1 eq $CHAR_PRES) and ($state_ue3 eq $CHAR_PRES)) ? 'Shift(??)' : 'Incomplete-split(?)' ;
				$penalty = (($state_ue1 eq $CHAR_PRES) and ($state_ue3 eq $CHAR_PRES)) ? 1.5 : 2 ; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			     }
			  }

			} elsif ( ((@indels_rlv_lf_ref ==1) and (@indels_rlv_rf_rec ==1) and (@indels_rltd_le_rec ==1)) or 
			     ((@indels_rlv_rf_ref ==1) and (@indels_rlv_lf_rec ==1) and (@indels_rltd_re_rec ==1)) ) {

    (print STDOUT "     | (iii) - C-1b |\n\n") if ($TEST_LEVEL_newtype>2);

				# Maybe an 'Inter-branch-transfer' (#{indels}_rec>#{indels}_ref). #

			    $type = 'Inter-branch-transfer(rec>ref)(??)';
			    $penalty = 7; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			}
		    }
		} # END of "unless (defined $type) {...}"

		if (!(defined $type) and (@indels_rltd_other_ovlp_ref ==1)) {

				# D: Supect a 'Partial-Neighboring-Branch-Split' (Long-Block).
				# (ADDED on Nov 7 (Fri), 2014.) #

    (print STDOUT "     | (iii) - D |\n\n") if ($TEST_LEVEL_newtype>2);

		    my ($br1, $state_ue1, $len1) = @{$indels_rlv_ref[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rlv_rec[0]}[0, 3, 5];
		    my ($br3, $state_ue3, $len3) = @{$indels_rltd_rec[0]}[0, 3, 5];
		    my ($br4, $state_ue4, $len4) = @{$indels_rltd_other_ovlp_ref[0]}[0, 3, 5];

		    if ($len2 - $len3 == $len1) {

    (print STDOUT "     | (iii) - D1 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (($br1 != $br2) and ($state_ue1 ne $state_ue2))) { # The 'relevant' reference & reconstructed events are on the same branch and of the same type.
			##if ((($br1 == $br2) and ($state_ue1 == $state_ue2)) or (($br1 != $br2) and ($state_ue1 ne $state_ue2))) { # The 'relevant' reference & reconstructed events are on the same branch and of the same type. ...} # A BUG fixed on Nov 27 (Fri), 2015. #

    (print STDOUT "     | (iii) - D1-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    if (($br3 != $br4) and !((defined $equiv_branches{$br3}) and ($equiv_branches{$br3} == $br4))) { # The two 'related' events are on different branches. #

    (print STDOUT "     | (iii) - D1-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				if (@indels_rltd_lf_rec ==1) {

    (print STDOUT "     | (iii) - D1-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rlv_re_ref ==1) { # Normal (w/o 'bounding' column). # 

    (print STDOUT "     | (iii) - D1-1a-1a |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_le_rec ==1) {
					    $type = 'Partial-Neighborng-Branch-Split(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_re_rec) {
					    $type = 'Partial-Neighborng-Branch-Split(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    } elsif (@indels_rlv_other_ovlp_ref ==1) { # Normal (with 'bounding' column). #

    (print STDOUT "     | (iii) - D1-1a-1b |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_le_rec ==1) {
					    $type = 'Partial-Neighborng-Branch-Split(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_other_ovlp_rec ==1) {
					    $type = 'Partial-Neighborng-Branch-Split(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    } elsif (@indels_rlv_le_ref ==1) { # +Slide. #

    (print STDOUT "     | (iii) - D1-1a-1c |\n\n") if ($TEST_LEVEL_newtype>2);

					if (scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_re_rec) ==1) {
					    $type = 'Partial-Neighborng-Branch-Split(+slide(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }

				} elsif (@indels_rltd_rf_rec ==1) {

    (print STDOUT "     | (iii) - D1-1a-2 |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rlv_le_ref ==1) { # Normal (w/o 'bounding' column). # 

    (print STDOUT "     | (iii) - D1-1a-2a |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_re_rec ==1) {
					    $type = 'Partial-Neighborng-Branch-Split(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_le_rec) {
					    $type = 'Partial-Neighborng-Branch-Split(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    } elsif (@indels_rlv_other_ovlp_ref ==1) { # Normal (with 'bounding' column). #

    (print STDOUT "     | (iii) - D1-1a-2b |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_re_rec ==1) {
					    $type = 'Partial-Neighborng-Branch-Split(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_other_ovlp_rec ==1) {
					    $type = 'Partial-Neighborng-Branch-Split(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    } elsif (@indels_rlv_re_ref ==1) { # +Slide. #

    (print STDOUT "     | (iii) - D1-1a-2c |\n\n") if ($TEST_LEVEL_newtype>2);

					if (scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_le_rec) ==1) {
					    $type = 'Partial-Neighborng-Branch-Split(+slide(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
				}

			    } # END of "if (($br3 != $br4) and !(...)) {...}"
			}

			if (defined $type) {
			    push @indels_invlvd_ref, @indels_rltd_other_ovlp_ref;
			    $cblk_wise_cts_invlvd_indels[$cb]->[4] = 1;
			}

		    } # END of "if ($len2 - $len3 == $len1) {...}"


		} # END of "if (!(defined $type) and (@indels_rltd_other_ovlp_ref ==1)) {...}"

    (print STDOUT "     | (iii) - Z |\n\n") if ($TEST_LEVEL_newtype>2);

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');

			### NOTE: Probably, similar cases should be seen in other errors, such as 'merge', 'split', 'purge', and 'ex-nihilo', accompanied by a nearby 'related' event (in reconstructed MSA).
			### ... I WILL TAKE CARE OF THESE CASES as well LATER...

		} else {

			# Substep (x): If the subject comes up to here, Regard the associated error as 'Complex'. #

    (print STDOUT "  { Substep (x) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'Complex';
		    $penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		}

			## MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
			## => Further MODIFIED in REVISION on Oct 21 (Wed), 2015. #

		return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty);
		##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

		##	# Finally, register the association between the composite-block and indel events considered to be involved. #

		##unless ( register_association_cblk_and_indels ($type, $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @indels_invlvd_ref, @indels_invlvd_rec, @indices_indel_identifiers) ) {
		##	print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED at the Step 2A on the $cb th composite shift-block for the $s th segment !\n";
		##	print STDERR "[ERROR '$program_name'] $!\n\n";
		##	die ;
		##}
		##
		##next;

	    } elsif (($ct_rltd_indels_ref == 1) and ($ct_rltd_indels_rec == 0)) {

				# Step 2B. #

    (print STDOUT "< Step 2B >\n\n") if ($TEST_LEVEL_newtype>0);

		my $type; # The type of MSA error(s). #
		my $penalty; # The value of penalty incurred by the error. # ADDED in REVISIOn on Oct 21 (Wed), 2015. #

		if (($ct_rlv_indels_ref == 0) and ($ct_rlv_indels_rec == 0)) {

			# Substep (o): The composite block is NOT directly associated with any errorneous indel event (???) #

    (print STDOUT "  { Substep (o) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'NO_ASSOCIATED_EVENT(???)';
		    $penalty = 200; # ADDED in REVISION on Oct 21 (Wed), 2015. #


		} elsif (($ct_rlv_indels_ref == 0) and ($ct_rlv_indels_rec == 1)) {


			# Substep (i): Suspect a 'Creation of Erroneous Independent Insertions'. #

    (print STDOUT "  { Substep (i) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my $subtype;
		    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #
		    my @indels_invlvd_ref1 = my @indels_invlvd_rec1 = ();

		    if (scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) == 0) {

    (print STDOUT "     | (i) - A |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((@{$indels_le_ref} >= 2) and (@{$indels_lf_rec} ==1)) {

    (print STDOUT "     | (i) - A-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = (@{$indels_lf_rec}, @indels_rlv_rec);
			    @indels_invlvd_ref1 = @{$indels_le_ref};

			    if (scalar(@{$indels_rf_ref}) - scalar(@indels_rltd_rf_ref) > 0) {
				$subtype = '(+slide(?))';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			    } elsif (@indels_rlv_le_rec ==1) {
				$subtype = '';
			    } else {
				$subtype = '(+shift)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }

			} elsif ((@{$indels_re_ref} >= 2) and (@{$indels_rf_rec} ==1)) {

    (print STDOUT "     | (i) - A-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = (@{$indels_rf_rec}, @indels_rlv_rec);
			    @indels_invlvd_ref1 = @{$indels_re_ref};

			    if (scalar(@{$indels_lf_ref}) - scalar(@indels_rltd_lf_ref) > 0) {
				$subtype = '(+slide(?))';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			    } elsif (@indels_rlv_re_rec ==1) {
				$subtype = '';
			    } else {
				$subtype = '(+shift)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }

				# ADDED the following conditions on Nov 10 (Mon), 2014. #

			} elsif ((@{$indels_le_ref} >= 2) and (@{$indels_rf_rec} ==1)) {

    (print STDOUT "     | (i) - A-3 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = (@{$indels_rf_rec}, @indels_rlv_rec);
			    @indels_invlvd_ref1 = @{$indels_le_ref};
			    $subtype = (@indels_rlv_re_rec ==1) ? '(+shift^2(??))' : '(+overshoot(??))';
			    $sbpenalty += (@indels_rlv_re_rec ==1) ? 2 : 1 ; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@{$indels_re_ref} >= 2) and (@{$indels_lf_rec} ==1)) {

    (print STDOUT "     | (i) - A-4 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = (@{$indels_lf_rec}, @indels_rlv_rec);
			    @indels_invlvd_ref1 = @{$indels_re_ref};
			    $subtype = (@indels_rlv_le_rec ==1) ? '(+shift^2(??))' : '(+overshoot(??))';
			    $sbpenalty += (@indels_rlv_le_rec ==1) ? 2 : 1 ; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				# END of "ADDED the following conditions on Nov 10 (Mon), 2014.} #
			}

		    } # END of "if (scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) ==0) {...}"

		    if (defined $subtype) {

    (print STDOUT "     | (i) - A-Test |\n\n") if ($TEST_LEVEL_newtype>2);

			my ($if_cii, $indels_invlvd_ref2) = if_collapse_indep_ins (@indels_invlvd_rec1, @indels_invlvd_ref1, %{$node2pa}, %{$node2ch}, $CHAR_ABS);
			unless (defined $if_cii) {
			    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i) of Step 2B: $! !\n\n";
			    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

			    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i) of Step 2B on the $cb th composite shift-block for the $s th segment !\n";
			    ##print STDERR "[ERROR '$program_name'] $!\n\n";
			    ##die ;
			}

			if ($if_cii) {

    (print STDOUT "     | (i) - A-Test-pass |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec = @indels_invlvd_rec1;
			    @indels_invlvd_ref = @{$indels_invlvd_ref2};
			    my $diffct = scalar (@indels_invlvd_ref) - @indels_invlvd_rec;

			    $cblk_wise_cts_invlvd_indels[$cb]->[4] = scalar (@{$indels_invlvd_ref2}) -1; # ADDED on Nov 6 (Thu), 2014. #

			    my $len1 = $indels_invlvd_rec[0]->[5]; 
			    my $len2 = $indels_invlvd_rec[1]->[5]; 
			    if ($len1 == $len2) {
				$type = "Creation-of-Erroneous-Independent-Insertions$subtype(-$diffct)(?)";
				$penalty = 10 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    } else {
				$diffct++;
				$type = "Incomplete-Creation-of-Erroneous-Independent-Insertions$subtype(-$diffct)(?)";
				$penalty = 15 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }
			}
		    }

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} elsif (($ct_rlv_indels_ref == 1) and ($ct_rlv_indels_rec == 0)) {

			# Substep (ii): Suspect a 'Purge' aligned with a run of gaps. #

    (print STDOUT "  { Substep (ii) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my $len1 = $indels_rlv_ref[0]->[5];

		    my $len2 = $indels_rltd_ref[0]->[5];
		    ##my $len2 = $indels_rlv_ref[1]->[5]; # A BUG fixed on Nov 6 (Thu), 2014. #

		  if ($len1 == $len2) {

    (print STDOUT "     | (ii) - A |\n\n") if ($TEST_LEVEL_newtype>2);

				# MODIFIED for DEBUGGING (on Nov 10 (Mon), 2014). 
		    if ( ((@indels_rlv_lf_ref ==1) and (@indels_rltd_re_ref ==1)) or 
				((@indels_rlv_rf_ref ==1) and (@indels_rltd_le_ref ==1)) ) {

    (print STDOUT "     | (ii) - A-1 |\n\n") if ($TEST_LEVEL_newtype>2);

		        my ($br1, $state_ue1) = @{$indels_rlv_ref[0]}[0, 3];
		        my ($br2, $state_ue2) = @{$indels_rltd_ref[0]}[0, 3];

			if ($br2 == $node2pa->{$br1}) {

    (print STDOUT "     | (ii) - A-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
			    if (($state_ue2 eq $CHAR_PRES) and ($state_ue1 eq $CHAR_ABS)) {
				$type = 'Purge(gap-aligned)(?)';
				$penalty = 3 + 1;
			    }
			    ##(($state_ue2 eq $CHAR_PRES) and ($state_ue1 eq $CHAR_ABS)) and ($type = 'Purge(gap-aligned)(?)');
			} elsif ($br1 == $node2pa->{$br2}) {

    (print STDOUT "     | (ii) - A-1b |\n\n") if ($TEST_LEVEL_newtype>2);

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
			    if (($state_ue1 eq $CHAR_PRES) and ($state_ue2 eq $CHAR_ABS)) {
				$type = 'Purge(gap-aligned)(?)';
				$penalty = 3 + 1;
			    }
			    ##(($state_ue1 eq $CHAR_PRES) and ($state_ue2 eq $CHAR_ABS)) and ($type = 'Purge(gap-aligned)(?)');
			} else {

    (print STDOUT "     | (ii) - A-1c |\n\n") if ($TEST_LEVEL_newtype>2);

			    if (($state_ue1 eq $CHAR_ABS) and ($state_ue2 eq $CHAR_ABS)) {
				$type = 'Purge(gap-aligned)(?)';
				$penalty = 3 + 1;
			    }
			    ##(($state_ue1 eq $CHAR_ABS) and ($state_ue2 eq $CHAR_ABS)) and ($type = 'Purge(gap-aligned)(?)');
			}
		    }

		    ##if ((@indels_rlv_lf_ref ==1) and (@indels_rltd_re_ref ==1)) {
		    ##	$type = 'Purge(gap-aligned)(?)';
		    ##} elsif ((@indels_rlv_rf_ref ==1) and (@indels_rltd_le_ref ==1)) {
		    ##	$type = 'Purge(gap-aligned)(?)';

		    ##} elsif ( ((@indels_rlv_lf_ref ==1) and (@indels_rltd_other_ovlp_ref ==1)) or
		    ## 		((@indels_rlv_rf_ref ==1) and (@indels_rltd_other_ovlp_ref ==1)) ) {
		    ##	$type = 'Purge(gap-aligned)(??)';
		    ##}
		  }

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. # 
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} elsif (($ct_rlv_indels_ref == 1) and ($ct_rlv_indels_rec == 1)) {

			# Substep (iii): 

    (print STDOUT "  { Substep (iii) }\n\n") if ($TEST_LEVEL_newtype>1);

				# A: First, Suspect an 'Incomplete-Creation of Erroneous Independent Insertions'. #

		    my $subtype;
		    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #
		    my @indels_invlvd_ref1 = my @indels_invlvd_rec1 = my @indels_invlvd_ref1p = ();

		    if ((scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) == 0) and (scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) == 0)) {

    (print STDOUT "     | (iii) - A1 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((@{$indels_le_ref} >= 2) and (@{$indels_lf_rec} ==1)) {

    (print STDOUT "     | (iii) - A1-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = (@{$indels_lf_rec}, @indels_rlv_rec);
			    @indels_invlvd_ref1 = @{$indels_le_ref};
			    @indels_invlvd_ref1p = @indels_rlv_ref;

			    if (@indels_rlv_le_rec ==1) {
				$subtype = '';

			    } else {
				$subtype = '(+shift)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }

			} elsif ((@{$indels_re_ref} >= 2) and (@{$indels_rf_rec} ==1)) {

    (print STDOUT "     | (iii) - A1-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = (@{$indels_rf_rec}, @indels_rlv_rec);
			    @indels_invlvd_ref1 = @{$indels_re_ref};
			    @indels_invlvd_ref1p = @indels_rlv_ref;

			    if (@indels_rlv_re_rec ==1) {
				$subtype = '';
			    } else {
				$subtype = '(+shift)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }

				# ADDED the following conditions (on Nov 10 (Mon), 2014). #

			} elsif ((@{$indels_le_ref} >= 2) and (@{$indels_rf_rec} ==1)) {

    (print STDOUT "     | (iii) - A1-3 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = (@{$indels_rf_rec}, @indels_rlv_rec);
			    @indels_invlvd_ref1 = @{$indels_le_ref};
			    @indels_invlvd_ref1p = @indels_rlv_ref;
			    $subtype = '(+slide(?))';
			    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@{$indels_re_ref} >= 2) and (@{$indels_lf_rec} ==1)) {

    (print STDOUT "     | (iii) - A1-4 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = (@{$indels_lf_rec}, @indels_rlv_rec);
			    @indels_invlvd_ref1 = @{$indels_re_ref};
			    @indels_invlvd_ref1p = @indels_rlv_ref;
			    $subtype = '(+slide(?))';
			    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				# END of "ADDED the following conditions (on Nov 10 (Mon), 2014)." #
			}


				# ADDED the following conditions (on Nov 10 (Mon), 2014). #

		    } elsif (scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) == 0) {

    (print STDOUT "     | (iii) - A2 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((@indels_rlv_lf_rec ==1) and (@{$indels_le_ref} >= 2) and (@{$indels_other_ovlp_rec} > 0)) {

    (print STDOUT "     | (iii) - A2-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = @indels_rlv_lf_rec;
			    @indels_invlvd_ref1 = @{$indels_le_ref};
			    @indels_invlvd_ref1p = @indels_rlv_ref;

			    foreach my $event (@{$indels_other_ovlp_rec}) {
				($event->[2] +1 == $beg_cb_rec) and (push @indels_invlvd_rec1, $event);
			    }

			} elsif ((@indels_rlv_rf_rec ==1) and (@{$indels_re_ref} >= 2) and (@{$indels_other_ovlp_rec} > 0)) {

    (print STDOUT "     | (iii) - A2-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = @indels_rlv_rf_rec;
			    @indels_invlvd_ref1 = @{$indels_re_ref};
			    @indels_invlvd_ref1p = @indels_rlv_ref;

			    foreach my $event (@{$indels_other_ovlp_rec}) {
				($end_cb_rec + 1 == $event->[1]) and (push @indels_invlvd_rec1, $event);
			    }
			}

			if (@indels_invlvd_rec1 ==2) {
			    $subtype = '(+shift^2(??))';
			    $sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			}
				# END of "ADDED the following conditions (on Nov 10 (Mon), 2014)." #


		    } # END of "if ((scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) ==0) and (...)) {...} elsif (scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) == 0) {...}".

		    if (defined $subtype) {

    (print STDOUT "     | (iii) - A-Test |\n\n") if ($TEST_LEVEL_newtype>2);

			my ($if_cii, $indels_invlvd_ref2) = if_collapse_indep_ins (@indels_invlvd_rec1, @indels_invlvd_ref1, %{$node2pa}, %{$node2ch}, $CHAR_ABS);
			unless (defined $if_cii) {
			    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (iii)-A of Step 2B: $! !\n\n";
			    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

			    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (iii)-A of Step 2B on the $cb th composite shift-block for the $s th segment !\n";
			    ##print STDERR "[ERROR '$program_name'] $!\n\n";
			    ##die ;
			}

			if ($if_cii) {

    (print STDOUT "     | (iii) - A-Test-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec = @indels_invlvd_rec1;
			    @indels_invlvd_ref = (@{$indels_invlvd_ref2}, @indels_invlvd_ref1p);
			    my $diffct = scalar (@indels_invlvd_ref) - @indels_invlvd_rec;

			    $cblk_wise_cts_invlvd_indels[$cb]->[4] = scalar (@{$indels_invlvd_ref2}) -1;

			    my $len1 = $indels_invlvd_rec[0]->[5]; 
			    my $len2 = $indels_invlvd_rec[1]->[5]; 
			    if ($len1 == $len2) {
				$type = "Incomplete-Creation-of-Erroneous-Independent-Insertions$subtype(-$diffct)(?)";
				$penalty = 15 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    } else {
				$diffct++;
				##$type = "Incomplete-Creation-of-Erroneous-Independent-Insertions$subtype(-$diffct)(?)"; # I think it dubious...
			    }
			}
		    }

					# ADDED on Nov 12 (Wed), 2014. #

		    if (!(defined $type) and (@indels_rltd_other_ovlp_rec ==1) and (scalar(@indels_rltd_lf_ref) + scalar(@indels_rltd_rf_ref) ==1) and (scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_re_rec) ==1) and (scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) == 0) ) {

				# B: In this case, Suspect a 'Partial-Neighboring-Branch-Merge' (2ins) (Long-Block(normal)). #

    (print STDOUT "     | (iii) - B |\n\n") if ($TEST_LEVEL_newtype>2);

			my $event1 = $indels_rltd_ref[0];
			my $event2 = $indels_rlv_ref[0];

			my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$event1}[0..3, 5];
			my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$event2}[0..3, 5];
			my $eq_br1 = $equiv_branches{$br1};
			my $eq_br2 = $equiv_branches{$br2};

			if ( ($len2 > $len1) and (($end1 + 1 < $beg2) or ($end2 + 1 < $beg1)) ) { # 'ref_rlv' is longer than 'ref_rltd', and the two events are separated.

    (print STDOUT "     | (iii) - B1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my $event3 = $indels_rltd_other_ovlp_rec[0];
			    my $event4 = $indels_rlv_rec[0];
			    my ($br3, $beg3, $end3, $state_ue3, $len3) = @{$event3}[0..3, 5];
			    my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event4}[0..3, 5];

					# See if 'rltd_rec' is on a different branch than 'rltd_ref'.
			    my $if_event3 = (($br1 != $br3) and !((defined $eq_br1) and ($eq_br1 == $br3)) ) ? 1 : 0;

					# See if 'rlv_rec' is of the same type as 'rlv_ref'.
			    my $if_event4 = ( (($br2 == $br4) and ($state_ue2 eq $state_ue4)) or (($br2 != $br4) and ($state_ue2 ne $state_ue4)) ) ? 1 : 0;

			    if (($if_event4) and ($if_event3) and ($len1 == $event3->[5]) and ($len2 == $len1 + $event4->[5])) {

    (print STDOUT "     | (iii) - B1-1 |\n\n") if ($TEST_LEVEL_newtype>2);


				my $if_part_nbr_mrg = if_part_nbr_mrg_2ins (@{$event1}, @{$event2}, @{$event3}, 0, 3, %equiv_branches, %{$node2pa}, $CHAR_ABS) ;
				unless (defined $if_part_nbr_mrg) {
				    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (iii)-B of Step 2B: $! !\n\n";
				    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

				    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (iii)-B of Step 2B on the $cb th composite shift-block for the $s th segment !\n";
				    ##print STDERR "[ERROR '$program_name'] $!\n\n";
				    ##die ;
				}

				if ($if_part_nbr_mrg) {

    (print STDOUT "     | (iii) - B1-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				    my $subtype;
				    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    if (@indels_rlv_le_rec == 1) {

    (print STDOUT "     | (iii) - B1-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if ($end4 +1 == $beg3) {
					    $subtype = (@indels_rltd_lf_ref ==1) ? '(+overshoot)(?)' : '(+shift)(?)' ;
					    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}

				    } else { # @indels_rlv_re_rec ==1. #

    (print STDOUT "     | (iii) - B1-1a-2 |\n\n") if ($TEST_LEVEL_newtype>2);

					if ($end3 +1 == $beg4) {
					    $subtype = (@indels_rltd_rf_ref ==1) ? '(+overshoot)(?)' : '(+shift)(?)' ;
					    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }

				    if (defined $subtype) {
					$type = "Partial-Neighboring-Branch-Merge(2ins)$subtype";
					$penalty = 5 + $sbpenalty;
				    }

				} # END of "if ($if_part_nbr_mrg) {...}"
			    }

			} # END of "if (($len2 > $len1) and (...)) {...}"

		    } # END of "if (!(defined $type) and (@indels_rltd_other_ovlp_rec ==1) and ...) {...}"

					# END of "ADDED on Nov 12 (Wed), 2014." #



		unless (defined $type) {

				# C:  Then, Suspect a simple 'Shift', accompanied by an event along a related branch that happened to be near the composite block (in reconstructed MSA),
				# Or, maybe an 'Incomplete-merge' or an 'Inter-branch-transfer'.

    (print STDOUT "     | (iii) - C |\n\n") if ($TEST_LEVEL_newtype>2);

		    my ($br1, $state_ue1, $len1) = @{$indels_rlv_rec[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rlv_ref[0]}[0, 3, 5];
		    my ($br3, $state_ue3) = @{$indels_rltd_ref[0]}[0, 3];


		    if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (($br1 != $br2) and ($state_ue1 ne $state_ue2)) ) {
				# At least the relevant events are related by a 'Shift'-like move. #

    (print STDOUT "     | (iii) - C1 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ( ((@indels_rlv_lf_rec ==1) and (@indels_rlv_rf_ref ==1) and (@indels_rltd_lf_ref ==1)) or 
			     ((@indels_rlv_rf_rec ==1) and (@indels_rlv_lf_ref ==1) and (@indels_rltd_rf_ref ==1)) ) {

    (print STDOUT "     | (iii) - C1-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				# Whether a 'Shift' or an 'Incomplete-merge'.

			  if ($len1 == $len2) {

    (print STDOUT "     | (iii) - C1-1a |\n\n") if ($TEST_LEVEL_newtype>2);

			     if ($br3 == $node2pa->{$br1}) {

    (print STDOUT "     | (iii) - C1-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				$type = (($state_ue1 eq $CHAR_PRES) and ($state_ue3 eq $CHAR_ABS)) ? 'Shift(??)' : 'Incomplete-merge(?)' ;
				$penalty = (($state_ue1 eq $CHAR_PRES) and ($state_ue3 eq $CHAR_ABS)) ? 1.5 : 2 ; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			     } elsif ($br1 == $node2pa->{$br3}) {

    (print STDOUT "     | (iii) - C1-1a-2 |\n\n") if ($TEST_LEVEL_newtype>2);

				$type = (($state_ue3 eq $CHAR_PRES) and ($state_ue1 eq $CHAR_ABS)) ? 'Shift(??)' : 'Incomplete-merge(?)' ;
				$penalty = (($state_ue3 eq $CHAR_PRES) and ($state_ue1 eq $CHAR_ABS)) ? 1.5 : 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    } else {

    (print STDOUT "     | (iii) - C1-1a-3 |\n\n") if ($TEST_LEVEL_newtype>2);

				$type = (($state_ue1 eq $CHAR_PRES) and ($state_ue3 eq $CHAR_PRES)) ? 'Shift(??)' : 'Incomplete-merge(?)' ;
				$penalty = (($state_ue1 eq $CHAR_PRES) and ($state_ue3 eq $CHAR_PRES)) ? 1.5 : 2 ; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }
			  }

			} elsif ( ((@indels_rlv_lf_rec ==1) and (@indels_rlv_rf_ref ==1) and (@indels_rltd_le_ref ==1)) or 
			     ((@indels_rlv_rf_rec ==1) and (@indels_rlv_lf_ref ==1) and (@indels_rltd_re_ref ==1)) ) {

    (print STDOUT "     | (iii) - C1-2 |\n\n") if ($TEST_LEVEL_newtype>2);

				# Maybe an 'Inter-branch-transfer' (#{indels}_ref>#{indels}_rec). #

			    $type = 'Inter-branch-transfer(ref>rec)(??)';
			    $penalty = 7; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			}
		    }
		} # END of "unless (defined $type) {...}"


		if (!(defined $type) and (@indels_rltd_other_ovlp_rec ==1)) {

				# D: Supect a 'Partial-Neighboring-Branch-Merge' (Long-Block).
				# (ADDED on Nov 7 (Fri), 2014.) #

    (print STDOUT "     | (iii) - D |\n\n") if ($TEST_LEVEL_newtype>2);

		    my ($br1, $state_ue1, $len1) = @{$indels_rlv_rec[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rlv_ref[0]}[0, 3, 5];
		    my ($br3, $state_ue3, $len3) = @{$indels_rltd_ref[0]}[0, 3, 5];
		    my ($br4, $state_ue4, $len4) = @{$indels_rltd_other_ovlp_rec[0]}[0, 3, 5];

		    if ($len2 - $len3 == $len1) {

    (print STDOUT "     | (iii) - D1 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (($br1 != $br2) and ($state_ue1 ne $state_ue2))) { # The 'relevant' reference & reconstructed events are on the same branch and of the same type.
			##if ((($br1 == $br2) and ($state_ue1 == $state_ue2)) or (($br1 != $br2) and ($state_ue1 ne $state_ue2))) { # The 'relevant' reference & reconstructed events are on the same branch and of the same type. ...} # A BUG fixed on Novv 27 (Fri), 2015. #

    (print STDOUT "     | (iii) - D1-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    if (($br3 != $br4) and !((defined $equiv_branches{$br3}) and ($equiv_branches{$br3} == $br4))) { # The two 'related' events are on different branches. #

    (print STDOUT "     | (iii) - D1-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				if (@indels_rltd_lf_ref ==1) {

    (print STDOUT "     | (iii) - D1-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rlv_re_rec ==1) { # Normal (w/o 'bounding' column). # 

    (print STDOUT "     | (iii) - D1-1a-1a |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_le_ref ==1) {
					    $type = 'Partial-Neighborng-Branch-Merge(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #  

					} elsif (@indels_rlv_re_ref) {
					    $type = 'Partial-Neighborng-Branch-Merge(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #  
					}
				    } elsif (@indels_rlv_other_ovlp_rec ==1) { # Normal (with 'bounding' column). #

    (print STDOUT "     | (iii) - D1-1a-1b |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_le_ref ==1) {
					    $type = 'Partial-Neighborng-Branch-Merge(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #  

					} elsif (@indels_rlv_other_ovlp_ref ==1) {
					    $type = 'Partial-Neighborng-Branch-Merge(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #  
					}
				    } elsif (@indels_rlv_le_rec ==1) { # +Slide. #

    (print STDOUT "     | (iii) - D1-1a-1c |\n\n") if ($TEST_LEVEL_newtype>2);

					if (scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_re_ref) ==1) {
					    $type = 'Partial-Neighborng-Branch-Merge(+slide(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #  
					}
				    }

				} elsif (@indels_rltd_rf_ref ==1) {

    (print STDOUT "     | (iii) - D1-1a-2 |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rlv_le_rec ==1) { # Normal (w/o 'bounding' column). # 

    (print STDOUT "     | (iii) - D1-1a-2a |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_re_ref ==1) {
					    $type = 'Partial-Neighborng-Branch-Merge(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #  

					} elsif (@indels_rlv_le_ref) {
					    $type = 'Partial-Neighborng-Branch-Merge(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #  
					}
				    } elsif (@indels_rlv_other_ovlp_rec ==1) { # Normal (with 'bounding' column). #

    (print STDOUT "     | (iii) - D1-1a-2b |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_re_ref ==1) {
					    $type = 'Partial-Neighborng-Branch-Merge(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #  

					} elsif (@indels_rlv_other_ovlp_ref ==1) {
					    $type = 'Partial-Neighborng-Branch-Merge(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #  
					}
				    } elsif (@indels_rlv_re_rec ==1) { # +Slide. #

    (print STDOUT "     | (iii) - D1-1a-2c |\n\n") if ($TEST_LEVEL_newtype>2);

					if (scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_le_ref) ==1) {
					    $type = 'Partial-Neighborng-Branch-Merge(+slide(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #  
					}
				    }
				}

			    } # END of "if (($br3 != $br4) and !(...)) {...}"
			}

			if (defined $type) {

    (print STDOUT "     | (iii) - D1-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

			    push @indels_invlvd_rec, @indels_rltd_other_ovlp_rec;
			    $cblk_wise_cts_invlvd_indels[$cb]->[5] = 1;
			}

		    } # END of "if ($len2 - $len3 == $len1) {...}"


		} # END of "if (!(defined $type) and (@indels_rltd_other_ovlp_rec ==1)) {...}"
B
    (print STDOUT "     | (iii) - Z |\n\n") if ($TEST_LEVEL_newtype>2);

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');

			### NOTE: Probably, similar cases should be seen in other errors, such as 'merge', 'split', 'purge', and 'ex-nihilo', accompanied by a nearby 'related' event (in reconstructed MSA).
			### ... I WILL TAKE CARE OF THESE CASES as well LATER...


		} else {

			# Substep (x): If the subject comes up to here, Regard the associated error as 'Complex'. #

    (print STDOUT "  { Substep (x) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'Complex';
		    $penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		}

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
			# => Further MODIFIED in REVISION on Oct 21 (Wed), 2015. #

		return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty);
		##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

		##	# Finally, register the association between the composite-block and indel events considered to be involved. #
		##
		##unless ( register_association_cblk_and_indels ($type, $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @indels_invlvd_ref, @indels_invlvd_rec, @indices_indel_identifiers) ) {
		##	print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED at the Step 2B on the $cb th composite shift-block for the $s th segment !\n";
		##	print STDERR "[ERROR '$program_name'] $!\n\n";
		##	die ;
		##}
		##
		##next;

	    } # END of "if (($ct_rltd_indels_rec == 1) and ($ct_rltd_indels_ref == 0)) {...} elsif (($ct_rltd_indels_ref == 1) and ($ct_rltd_indels_rec == 0)) {...}"


			# Step 3: When involving two events on relative branches. #

	    if (($ct_rltd_indels_rec == 1) and ($ct_rltd_indels_ref == 1)) {

				# Step 3A. #

    (print STDOUT "< Step 3A >\n\n") if ($TEST_LEVEL_newtype>0);

		my $type; # The type of MSA error(s). #
		my $penalty; # The value of penalty incurred by the error. # ADDED in REVISION on Oct 21 (Wed), 2015. #

		if (($ct_rlv_indels_rec == 0) and ($ct_rlv_indels_ref == 0)) {

			# Substep (o): Suspect a 'Shift' aligned with a run of gaps....
			#		Otherwise, I can't come up with any other types, at least currently...

    (print STDOUT "  { Substep (o) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my ($br1, $state_ue1, $len1) = @{$indels_rltd_rec[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rltd_ref[0]}[0, 3, 5];

		  if ($len1 == $len2) {

    (print STDOUT "     | (o) - A |\n\n") if ($TEST_LEVEL_newtype>2);

		    if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or ((defined $equiv_branches{$br1}) and ($br2 == $equiv_branches{$br1}) and ($state_ue1 ne $state_ue2)) ) { 
		    ##if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or ((defined $equiv_branches{$br1}) and ($br2 == $equiv_branches{$br2}) and ($state_ue1 ne $state_ue2)) ) { # ..} A BUG fixed on Nov 27 (Fri), 2015. #

    (print STDOUT "     | (o) - A-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
			if ( ((@indels_rltd_lf_rec ==1) and (@indels_rltd_rf_ref ==1))
				or ((@indels_rltd_rf_rec ==1) and (@indels_rltd_lf_ref ==1)) ) {
			    $type = 'Shift(gap-aligned)(?)';
			    $penalty = 1.5 + 1;
			}
			##( ((@indels_rltd_lf_rec ==1) and (@indels_rltd_rf_ref ==1))
			##	or ((@indels_rltd_rf_rec ==1) and (@indels_rltd_lf_ref ==1)) ) and ($type = 'Shift(gap-aligned)(?)');

		    }
		  }

		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} elsif (($ct_rlv_indels_rec == 1) and ($ct_rlv_indels_ref == 0)) {

			# Substep (i): Suspect a 'Neighboring-branch-Split'...

    (print STDOUT "  { Substep (i) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my ($br1, $state_ue1, $len1) = @{$indels_rltd_ref[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rltd_rec[0]}[0, 3, 5];
		    my ($br3, $state_ue3, $len3) = @{$indels_rlv_rec[0]}[0, 3, 5];

		  if (($len1 == $len2) and ($len1 == $len3)) {
		  ##if ($len3 == $len1 + $len2) {...} # A FATAL BUG fixed on Nov 8 (Sat), 2014. #

    (print STDOUT "     | (i) - A |\n\n") if ($TEST_LEVEL_newtype>2);

		    unless (($br1 == $br2) or ((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2)) ) {

    (print STDOUT "     | (i) - A1 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((@indels_rltd_rf_ref ==1) and (@indels_rlv_lf_rec ==1) and (@indels_rltd_re_rec ==1)) {

    (print STDOUT "     | (i) - A1-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Split(2del)(?)'; # into two 'deletion's(?)
			    $penalty = 4; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@indels_rltd_lf_ref ==1) and (@indels_rlv_rf_rec ==1) and (@indels_rltd_le_rec ==1)) {

    (print STDOUT "     | (i) - A1-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Split(2del)(?)'; # into two 'deletion's(?)
			    $penalty = 4; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@indels_rltd_le_ref ==1) and (@indels_rlv_le_rec ==1) and (@indels_rltd_lf_rec ==1)) {

    (print STDOUT "     | (i) - A1-3 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Split(2ins)(?)'; # into two 'insertion's(?)
			    $penalty = 4; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@indels_rltd_re_ref ==1) and (@indels_rlv_re_rec ==1) and (@indels_rltd_rf_rec ==1)) {

    (print STDOUT "     | (i) - A1-4 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Split(2ins)(?)'; # into two 'insertion's(?)
			    $penalty = 4; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@indels_rltd_le_ref ==1) and (scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) ==1) and (@indels_rltd_lf_rec ==1)) {

    (print STDOUT "     | (i) - A1-5 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Split(2ins)(+shift)(?)'; # into two 'insertion's(?)
			    $penalty = 4 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@indels_rltd_re_ref ==1) and (scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) ==1) and (@indels_rltd_rf_rec ==1)) {

    (print STDOUT "     | (i) - A1-6 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Split(2ins)(+shift)(?)'; # into two 'insertion's(?)
			    $penalty = 4 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			}
		    } # END of "unless (($br1 == $br2) or ...) {...}"
		  }

		    (defined $type) or ($type = 'Complex(???)');

		    
		} elsif (($ct_rlv_indels_ref == 1) and ($ct_rlv_indels_rec == 0)) {

			# Substep (ii): Suspect a 'Neighboring-branch-Merge'...

    (print STDOUT "  { Substep (ii) }\n\n") if ($TEST_LEVEL_newtype>1);

		    my ($br1, $state_ue1, $len1) = @{$indels_rltd_ref[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rltd_rec[0]}[0, 3, 5];
		    my ($br3, $state_ue3, $len3) = @{$indels_rlv_ref[0]}[0, 3, 5];

		  if (($len1 == $len2) and ($len1 == $len3)) {

    (print STDOUT "     | (ii) - A |\n\n") if ($TEST_LEVEL_newtype>2);

		  ##if ($len3 == $len1 + $len2) {...} A FATAL BUG fixed on Nov 8 (Sat), 2014. #
		    if (($br1 != $br2) and !((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2)) ) {

    (print STDOUT "     | (ii) - A-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((@indels_rltd_rf_rec ==1) and (@indels_rlv_lf_ref ==1) and (@indels_rltd_re_ref ==1)) {

    (print STDOUT "     | (ii) - A-1a |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Merge(2del)(?)'; # of two 'deletion's(?)
			    $penalty = 4; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@indels_rltd_lf_rec ==1) and (@indels_rlv_rf_ref ==1) and (@indels_rltd_le_ref ==1)) {

    (print STDOUT "     | (ii) - A-1b |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Merge(2del)(?)'; # of two 'deletion's(?)
			    $penalty = 4; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@indels_rltd_le_rec ==1) and (@indels_rlv_le_ref ==1) and (@indels_rltd_lf_ref ==1)) {

    (print STDOUT "     | (ii) - A-1c |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Merge(2ins)(?)'; # of two 'insertion's(?)
			    $penalty = 4; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@indels_rltd_re_rec ==1) and (@indels_rlv_re_ref ==1) and (@indels_rltd_rf_ref ==1)) {

    (print STDOUT "     | (ii) - A-1d |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Merge(2ins)(?)'; # of two 'insertion's(?)
			    $penalty = 4; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@indels_rltd_le_rec ==1) and (scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) ==1) and (@indels_rltd_lf_ref ==1)) {

    (print STDOUT "     | (ii) - A-1e |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Merge(2ins)(+shift)(?)'; # of two 'insertion's(?)
			    $penalty = 4 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			} elsif ((@indels_rltd_re_rec ==1) and (scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) ==1) and (@indels_rltd_rf_ref ==1)) {

    (print STDOUT "     | (ii) - A-1f |\n\n") if ($TEST_LEVEL_newtype>2);

			    $type = 'Neighboring-branch-Merge(2ins)(+shift)(?)'; # of two 'insertion's(?)
			    $penalty = 4 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			}
		    } # END of "if (($br1 != $br2) and !(...)) {...}"
		  }

				# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} elsif (($ct_rlv_indels_rec == 1) and ($ct_rlv_indels_ref == 1)) {

			# Substep (iii): 

    (print STDOUT "  { Substep (iii) }\n\n") if ($TEST_LEVEL_newtype>1);

					# ADDED on Nov 12 (Wed), 2014. #

		my ($br_ref, $state_ref, $len_ref) = @{$indels_rlv_ref[0]}[0, 3, 5];
		my ($br_rec, $state_rec, $len_rec) = @{$indels_rlv_rec[0]}[0, 3, 5];

		if ( ( (($br_ref == $br_rec) and ($state_ref eq $state_rec)) or (($br_ref != $br_rec) and ($state_ref ne $state_ref)) )
and (@indels_rlv_lf_rec == 0) and (@indels_rlv_rf_rec == 0) and (@indels_rlv_lf_ref == 0) and (@indels_rlv_rf_ref == 0) ) {

    (print STDOUT "     | (iii) - A |\n\n") if ($TEST_LEVEL_newtype>2);

		    if ( ($len_ref > $len_rec) and (@indels_rltd_lf_rec ==0) and (@indels_rltd_rf_rec == 0) and (scalar(@indels_rltd_rf_ref) + scalar(@indels_rltd_rf_ref) ==1) ) { # 'rlv_ref' is longer than 'rlv_rec'. #

				# (a): In this case, Suspect a 'Partial-Neighboring-Branch-Merge' (2ins) (Long-Block(normal)). #

    (print STDOUT "     | (iii) - A-(a) |\n\n") if ($TEST_LEVEL_newtype>2);

			my $event1 = $indels_rltd_ref[0];
			my $event2 = $indels_rlv_ref[0];

			my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$event1}[0..3, 5];
			my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$event2}[0..3, 5];
			my $eq_br1 = $equiv_branches{$br1};
			my $eq_br2 = $equiv_branches{$br2};

			if ( ($len2 > $len1) and (($end1 + 1 < $beg2) or ($end2 + 1 < $beg1)) ) { # 'ref_rlv' is longer than 'ref_rltd', and the two events are separated.

    (print STDOUT "     | (iii) - A-(a)-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my $event3 = $indels_rltd_rec[0];
			    my $event4 = $indels_rlv_rec[0];
			    my ($br3, $beg3, $end3, $state_ue3, $len3) = @{$event3}[0..3, 5];
			    my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event4}[0..3, 5];

					# See if 'rltd_rec' is on a different branch than 'rltd_ref'.
			    my $if_event3 = (($br1 != $br3) and !((defined $eq_br1) and ($eq_br1 == $br3)) ) ? 1 : 0;

					# See if 'rlv_rec' is of the same type as 'rlv_ref'.
			    my $if_event4 = ( (($br2 == $br4) and ($state_ue2 eq $state_ue4)) or (($br2 != $br4) and ($state_ue2 ne $state_ue4)) ) ? 1 : 0;

			    if (($if_event4) and ($if_event3) and ($len1 == $event3->[5]) and ($len2 == $len1 + $event4->[5]) ) {

    (print STDOUT "     | (iii) - A-(a)-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				my $if_part_nbr_mrg = if_part_nbr_mrg_2ins (@{$event1}, @{$event2}, @{$event3}, 0, 3, %equiv_branches, %{$node2pa}, $CHAR_ABS) ;
				unless (defined $if_part_nbr_mrg) {
				    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (iii)-(a) of Step 3A: $! !\n\n";
				    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

				    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (iii)-(a) of Step 3A on the $cb th composite shift-block for the $s th segment !\n";
				    ##print STDERR "[ERROR '$program_name'] $!\n\n";
				    ##die ;
				}

				if ($if_part_nbr_mrg) {

    (print STDOUT "     | (iii) - A-(a)-1a-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

				    my $sbtype;
				    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    if (@indels_rltd_lf_ref ==1) {

    (print STDOUT "     | (iii) - A-(a)-1a-Pass-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if ( (@indels_rltd_le_rec ==1) and (@indels_rlv_le_ref ==0) and (@indels_rlv_le_rec ==0)) {

    (print STDOUT "     | (iii) - A-(a)-1a-Pass-1a |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
					    if (($end3 + 1 < $beg4) or ($end4 + 1 < $beg3)) {
						$sbtype = '(+shift)(?)';
						$sbpenalty += 1;
					    }
					    ##(($end3 + 1 < $beg4) or ($end4 + 1 < $beg3)) and ($sbtype = '(+shift)(?)');

					} elsif ( (@indels_rltd_re_rec ==1) and (@indels_rlv_re_ref ==1) and (@indels_rlv_other_ovlp_rec ==1) ) {

    (print STDOUT "     | (iii) - A-(a)-1a-Pass-1b |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
					    if (($end3 + 1 == $beg4) or ($end4 + 1 == $beg3)) {
						$sbtype = '(+shift+slide)(?)';
						$sbpenalty += 2;
					    }
					    ##(($end3 + 1 == $beg4) or ($end4 + 1 == $beg3)) and ($sbtype = '(+shift+slide)(?)');
					}
				    } else { # if (@indels_rltd_rf_ref ==1)

    (print STDOUT "     | (iii) - A-(a)-1a-Pass-2 |\n\n") if ($TEST_LEVEL_newtype>2);

					if ( (@indels_rltd_re_rec ==1) and (@indels_rlv_re_ref ==0) and (@indels_rlv_re_rec ==0)) {

    (print STDOUT "     | (iii) - A-(a)-1a-Pass-2a |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
					    if (($end3 + 1 < $beg4) or ($end4 + 1 < $beg3)) {
						$sbtype = '(+shift)(?)';
						$sbpenalty += 1;
					    }
					    ##(($end3 + 1 < $beg4) or ($end4 + 1 < $beg3)) and ($sbtype = '(+shift)(?)');

					} elsif ( (@indels_rltd_le_rec ==1) and (@indels_rlv_le_ref ==1) and (@indels_rlv_other_ovlp_rec ==1) ) {

    (print STDOUT "     | (iii) - A-(a)-1a-Pass-2b |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
					    if (($end3 + 1 == $beg4) or ($end4 + 1 == $beg3)) {
						$sbtype = '(+shift+slide)(?)';
						$sbpenalty += 2;
					    }
					    ##(($end3 + 1 == $beg4) or ($end4 + 1 == $beg3)) and ($sbtype = '(+shift+slide)(?)');
					}
				    }

				    if (defined $sbtype) {
					$type = "Partial-Neighboring-Branch-Merge(2ins)$sbtype";
					$penalty = 5 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				}

			    } # END of "if (($if_event4) and ...) {...}"

			} # END of "if (($len2 > $len1) and (...)) {...}"

		    } elsif ( ($len_rec > $len_ref) and (@indels_rltd_lf_ref ==0) and (@indels_rltd_rf_ref == 0) and (scalar(@indels_rltd_rf_rec) + scalar(@indels_rltd_rf_rec) ==1) ) { # 'rlv_rec' is longer than 'rlv_ref'. #

				# (b): In this case, Suspect a 'Partial-Neighboring-Branch-Split' (2ins) (Long-Block(normal)). #

    (print STDOUT "     | (iii) - A-(b) |\n\n") if ($TEST_LEVEL_newtype>2);

			my $event1 = $indels_rltd_rec[0];
			my $event2 = $indels_rlv_rec[0];

			my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$event1}[0..3, 5];
			my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$event2}[0..3, 5];
			my $eq_br1 = $equiv_branches{$br1};
			my $eq_br2 = $equiv_branches{$br2};

			if ( ($len2 > $len1) and (($end1 + 1 < $beg2) or ($end2 + 1 < $beg1)) ) { # 'rec_rlv' is longer than 'rec_rltd', and the two events are separated.

    (print STDOUT "     | (iii) - A-(b)-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my $event3 = $indels_rltd_ref[0];
			    my $event4 = $indels_rlv_ref[0];
			    my ($br3, $beg3, $end3, $state_ue3, $len3) = @{$event3}[0..3, 5];
			    my ($br4, $beg4, $end4, $state_ue4, $len4) = @{$event4}[0..3, 5];

					# See if 'rltd_ref' is on a different branch than 'rltd_rec'.
			    my $if_event3 = (($br1 != $br3) and !((defined $eq_br1) and ($eq_br1 == $br3)) ) ? 1 : 0;

					# See if 'rlv_ref' is of the same type as 'rlv_rec'.
			    my $if_event4 = ( (($br2 == $br4) and ($state_ue2 eq $state_ue4)) or (($br2 != $br4) and ($state_ue2 ne $state_ue4)) ) ? 1 : 0;

			    if (($if_event4) and ($if_event3) and ($len1 == $event3->[5]) and ($len2 == $len1 + $event4->[5])) {

    (print STDOUT "     | (iii) - A-(b)-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				my $if_part_nbr_mrg = if_part_nbr_mrg_2ins (@{$event1}, @{$event2}, @{$event3}, 0, 3, %equiv_branches, %{$node2pa}, $CHAR_ABS) ;
				unless (defined $if_part_nbr_mrg) {
				    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (iii)-(b) of Step 3A: $! !\n\n";
				    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

				    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (iii)-(b) of Step 3A on the $cb th composite shift-block for the $s th segment !\n";
				    ##print STDERR "[ERROR '$program_name'] $!\n\n";
				    ##die ;
				}

				if ($if_part_nbr_mrg) {

    (print STDOUT "     | (iii) - A-(b)-1a-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

				    my $sbtype;
				    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    if (@indels_rltd_lf_rec ==1) {

    (print STDOUT "     | (iii) - A-(b)-1a-Pass-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if ( (@indels_rltd_le_ref ==1) and (@indels_rlv_le_rec ==0) and (@indels_rlv_le_ref ==0)) {

    (print STDOUT "     | (iii) - A-(b)-1a-Pass-1a |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
					    if (($end3 + 1 < $beg4) or ($end4 + 1 < $beg3)) {
						$sbtype = '(+shift)(?)';
						$sbpenalty += 1;
					    }
					    ##(($end3 + 1 < $beg4) or ($end4 + 1 < $beg3)) and ($sbtype = '(+shift)(?)');

					} elsif ( (@indels_rltd_re_ref ==1) and (@indels_rlv_re_rec ==1) and (@indels_rlv_other_ovlp_ref ==1) ) {

    (print STDOUT "     | (iii) - A-(b)-1a-Pass-1b |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
					    if (($end3 + 1 == $beg4) or ($end4 + 1 == $beg3)) {
						$sbtype = '(+shift+slide)(?)';
						$sbpenalty += 2;
					    }
					    ##(($end3 + 1 == $beg4) or ($end4 + 1 == $beg3)) and ($sbtype = '(+shift+slide)(?)');
					}
				    } else { # if (@indels_rltd_rf_rec ==1)

    (print STDOUT "     | (iii) - A-(b)-1a-Pass-2 |\n\n") if ($TEST_LEVEL_newtype>2);

					if ( (@indels_rltd_re_ref ==1) and (@indels_rlv_re_rec ==0) and (@indels_rlv_re_ref ==0)) {

    (print STDOUT "     | (iii) - A-(b)-1a-Pass-2a |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
					    if (($end3 + 1 < $beg4) or ($end4 + 1 < $beg3)) {
						$sbtype = '(+shift)(?)';
						$sbpenalty += 1;
					    }
					    ##(($end3 + 1 < $beg4) or ($end4 + 1 < $beg3)) and ($sbtype = '(+shift)(?)');

					} elsif ( (@indels_rltd_le_ref ==1) and (@indels_rlv_le_rec ==1) and (@indels_rlv_other_ovlp_ref ==1) ) {

    (print STDOUT "     | (iii) - A-(b)-1a-Pass-2b |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
					    if (($end3 + 1 == $beg4) or ($end4 + 1 == $beg3)) {
						$sbtype = '(+shift+slide)(?)';
						$sbpenalty += 2;
					    }
					    ##(($end3 + 1 == $beg4) or ($end4 + 1 == $beg3)) and ($sbtype = '(+shift+slide)(?)');
					}
				    }

				    if (defined $sbtype) {
					$type = "Partial-Neighboring-Branch-Split(2ins)$sbtype";
					$penalty = 5 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				}

			    } # END of "if (($if_event4) and ...) {...}"

			} # END of "if (($len2 > $len1) and (...)) {...}"

		    }

		} # END of "if ( ( (($br_ref == $br_rec) and ...) or ...) and ...) {...}"

					# END of "ADDED on Nov 12 (Wed), 2014." #


		unless (defined $type) { # A BIG-UNLESS.

    (print STDOUT "     | (iii) - B |\n\n") if ($TEST_LEVEL_newtype>2);

		    my ($br1, $state_ue1, $len1) = @{$indels_rltd_ref[0]}[0, 3, 5];
		    my ($br2, $state_ue2, $len2) = @{$indels_rltd_rec[0]}[0, 3, 5];
		    my ($br3, $state_ue3, $len3) = @{$indels_rlv_ref[0]}[0, 3, 5];
		    my ($br4, $state_ue4, $len4) = @{$indels_rlv_rec[0]}[0, 3, 5];

		    if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or ((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2) and ($state_ue1 ne $state_ue2)) ) { # The 'related' events on reference and reconstructed are on the same branch and of the same type.

    (print STDOUT "     | (iii) - B1 |\n\n") if ($TEST_LEVEL_newtype>2);

				# First, Suspect a 'Partial-Merge/Split' aligned with a long gap.

			if ( (($br3 == $br4) and ($state_ue3 eq $state_ue4)) or (($br3 != $br4) and ($state_ue3 ne $state_ue4)) ) {

    (print STDOUT "     | (iii) - B1-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my $diff_len_ref = ($len1 > $len3) ? $len1 - $len3 : $len3 - $len1;
			    my $diff_len_rec = ($len2 > $len4) ? $len2 - $len4 : $len4 - $len2;

			    if ($diff_len_ref == $diff_len_rec) {

    (print STDOUT "     | (iii) - B1-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				if ((@indels_rltd_rf_ref ==1) and (scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) ==1)) {

    (print STDOUT "     | (iii) - B1-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				    if ((@indels_rltd_rf_rec ==1) and (scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) ==1)) {

    (print STDOUT "     | (iii) - B1-1a-1a |\n\n") if ($TEST_LEVEL_newtype>2);

					$type = ($len1 + $len3 > $len2 + $len4) ? 'Partial-Merge(gap-aligned)(?)' : 'Partial-Split(gap-aligned)(?)' ;
					$penalty = 3 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif ((@indels_rltd_lf_rec ==1) and (scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) ==1)) {

    (print STDOUT "     | (iii) - B1-1a-1b |\n\n") if ($TEST_LEVEL_newtype>2);

					$type = 'Overshoot-Merge(gap-aligned)(??)';
					$penalty = 4 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }

				} elsif ((@indels_rltd_lf_ref ==1) and (scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) ==1)) {

    (print STDOUT "     | (iii) - B1-1a-2 |\n\n") if ($TEST_LEVEL_newtype>2);

				    if ((@indels_rltd_lf_rec ==1) and (scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) ==1)) {

    (print STDOUT "     | (iii) - B1-1a-2a |\n\n") if ($TEST_LEVEL_newtype>2);

					$type = ($len1 + $len3 > $len2 + $len4) ? 'Partial-Merge(gap-aligned)(?)' : 'Partial-Split(gap-aligned)(?)' ;
					$penalty = 3 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    } elsif ((@indels_rltd_rf_rec ==1) and (scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) ==1)) {

    (print STDOUT "     | (iii) - B1-1a-2b |\n\n") if ($TEST_LEVEL_newtype>2);

					$type = 'Overshoot-Merge(gap-aligned)(??)';
					$penalty = 4 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				} # END of "if ((@indels_rltd_rf_rec ==1) and (...)) {...} elsif ((@indels_rltd_lf_ref ==1) and (...)) {...}"

			    } # END of "if ($diff_len_ref == $diff_len_rec) {...}"
			}

		    } elsif ( ($br1 != $br2) and !((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2)) ) { # The 'related' events are on different branches.

    (print STDOUT "     | (iii) - B2 |\n\n") if ($TEST_LEVEL_newtype>2);

				# Then, Suspect an 'Inter-branch-Transfer', or a 'Partial-Neighboring-branch-Merge/Split' (Long-block). 

			if ( (($br3 == $br4) and ($state_ue3 eq $state_ue4)) or (($br3 != $br4) and ($state_ue3 ne $state_ue4)) ) {

    (print STDOUT "     | (iii) - B2-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				# Examine the possibility of an 'Inter-branch-Transfer'.

			    if ((@indels_rlv_lf_ref ==1) and (@indels_rlv_lf_rec ==1)) {

    (print STDOUT "     | (iii) - B2-1a |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				if ( ((@indels_rltd_rf_ref ==1) and (@indels_rltd_re_rec ==1)) or
					((@indels_rltd_re_ref ==1) and (@indels_rltd_rf_rec ==1)) ) {
				    $type = 'Inter-branch-Transfer(del-based)(?)'; # 'Deletion'-based.
				    $penalty = 7;
				}
				##( ((@indels_rltd_rf_ref ==1) and (@indels_rltd_re_rec ==1)) or
				##	((@indels_rltd_re_ref ==1) and (@indels_rltd_rf_rec ==1)) ) and ($type = 'Inter-branch-Transfer(del-based)(?)'); # 'Deletion'-based.

			    } elsif ((@indels_rlv_rf_ref ==1) and (@indels_rlv_rf_rec ==1)) {

    (print STDOUT "     | (iii) - B2-1b |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				if ( ((@indels_rltd_lf_ref ==1) and (@indels_rltd_le_rec ==1)) or
					((@indels_rltd_le_ref ==1) and (@indels_rltd_lf_rec ==1)) ) {
				    $type = 'Inter-branch-Transfer(del-based)(?)'; # 'Deletion'-based.
				    $penalty = 7;
				}
				##( ((@indels_rltd_lf_ref ==1) and (@indels_rltd_le_rec ==1)) or
				##	((@indels_rltd_le_ref ==1) and (@indels_rltd_lf_rec ==1)) ) and ($type = 'Inter-branch-Transfer(del-based)(?)'); # 'Deletion'-based.

			    } elsif ((@indels_rlv_le_ref ==1) and (@indels_rlv_le_rec ==1)) {

    (print STDOUT "     | (iii) - B2-1c |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				if ( ((@indels_rltd_lf_ref ==1) and (@indels_rltd_le_rec ==1)) or
					((@indels_rltd_le_ref ==1) and (@indels_rltd_lf_rec ==1)) ) {
				    $type = 'Inter-branch-Transfer(ins-based)(?)'; # 'Insertion'-based.
				    $penalty = 7;
				}
				##( ((@indels_rltd_lf_ref ==1) and (@indels_rltd_le_rec ==1)) or
				##	((@indels_rltd_le_ref ==1) and (@indels_rltd_lf_rec ==1)) ) and ($type = 'Inter-branch-Transfer(ins-based)(?)'); # 'Insertion'-based.

			    } elsif ((@indels_rlv_re_ref ==1) and (@indels_rlv_re_rec ==1)) {

    (print STDOUT "     | (iii) - B2-1d |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				if ( ((@indels_rltd_lf_ref ==1) and (@indels_rltd_le_rec ==1)) or
					((@indels_rltd_le_ref ==1) and (@indels_rltd_lf_rec ==1)) ) {
				    $type = 'Inter-branch-Transfer(ins-based)(?)'; # 'Insertion'-based.
				    $penalty = 7;
				}
				##( ((@indels_rltd_lf_ref ==1) and (@indels_rltd_le_rec ==1)) or
				##	((@indels_rltd_le_ref ==1) and (@indels_rltd_lf_rec ==1)) ) and ($type = 'Inter-branch-Transfer(ins-based)(?)'); # 'Insertion'-based.
			    } # A DEBUGGING measure on Nov 7 (Fri), 2014.

			unless (defined $type) {

    (print STDOUT "     | (iii) - B2-1z |\n\n") if ($TEST_LEVEL_newtype>2);

				# Examine the possibility of a 'Partial-Neighboring-Branch-Merge/Split' (Long-Block). #

			    if ((@indels_rltd_lf_ref ==1) and ($len3 - $len1 == $len4)) { # Merge(?). 

    (print STDOUT "     | (iii) - B2-1z-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				if (@indels_rlv_re_rec ==1) { # Normal (w/o 'bounding' column). #

    (print STDOUT "     | (iii) - B2-1z-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_le_rec ==1) {

    (print STDOUT "     | (iii) - B2-1z-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_le_ref==1) {
					    $type = 'Partial-Neighboring-Branch-Merge(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_re_ref ==1) {
					    $type = 'Partial-Neighboring-Branch-Merge(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
				} elsif (@indels_rlv_other_ovlp_rec ==1) { # Normal (with 'bounding' column). #

    (print STDOUT "     | (iii) - B2-1z-1a-1b |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_le_rec ==1) {

    (print STDOUT "     | (iii) - B2-1z-1a-1b-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_le_ref==1) {
					    $type = 'Partial-Neighboring-Branch-Merge(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_other_ovlp_ref ==1) {
					    $type = 'Partial-Neighboring-Branch-Merge(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
			 	} elsif (@indels_rlv_le_rec ==1) { # (+Slide). #

    (print STDOUT "     | (iii) - B2-1z-1a-1c |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_re_rec ==1) {

    (print STDOUT "     | (iii) - B2-1z-1a-1c-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (scalar(@indels_rlv_le_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_re_ref) ==1) {
					    $type = 'Partial-Neighboring-Branch-Merge(+slide(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
				}

			    } elsif ((@indels_rltd_rf_ref ==1) and ($len3 - $len1 == $len4)) { # Merge(?).

    (print STDOUT "     | (iii) - B2-1z-2 |\n\n") if ($TEST_LEVEL_newtype>2);

				if (@indels_rlv_le_rec ==1) { # Normal (w/o 'bounding' column). #

    (print STDOUT "     | (iii) - B2-1z-2a |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_re_rec ==1) {

    (print STDOUT "     | (iii) - B2-1z-2a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_re_ref==1) {
					    $type = 'Partial-Neighboring-Branch-Merge(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_le_ref ==1) {
					    $type = 'Partial-Neighboring-Branch-Merge(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
				} elsif (@indels_rlv_other_ovlp_rec ==1) { # Normal (with 'bounding' column). #

    (print STDOUT "     | (iii) - B2-1z-2b |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_re_rec ==1) {

    (print STDOUT "     | (iii) - B2-1z-2b-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_re_ref==1) {
					    $type = 'Partial-Neighboring-Branch-Merge(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_other_ovlp_ref ==1) {
					    $type = 'Partial-Neighboring-Branch-Merge(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
			 	} elsif (@indels_rlv_re_rec ==1) { # (+Slide). #

    (print STDOUT "     | (iii) - B2-1z-2c |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_le_rec ==1) {

    (print STDOUT "     | (iii) - B2-1z-2c-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (scalar(@indels_rlv_re_ref) + scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_le_ref) ==1) {
					    $type = 'Partial-Neighboring-Branch-Merge(+slide(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
				}

			    } elsif ((@indels_rltd_lf_rec ==1) and ($len4 - $len2 == $len3)) { # Split(?).

    (print STDOUT "     | (iii) - B2-1z-3 |\n\n") if ($TEST_LEVEL_newtype>2);

				if (@indels_rlv_re_ref ==1) { # Normal (w/o 'bounding' column). #

    (print STDOUT "     | (iii) - B2-1z-3a |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_le_ref ==1) {

    (print STDOUT "     | (iii) - B2-1z-3a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_le_rec==1) {
					    $type = 'Partial-Neighboring-Branch-Split(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_re_rec ==1) {
					    $type = 'Partial-Neighboring-Branch-Split(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
				} elsif (@indels_rlv_other_ovlp_ref ==1) { # Normal (with 'bounding' column). #

    (print STDOUT "     | (iii) - B2-1z-3b |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_le_ref ==1) {

    (print STDOUT "     | (iii) - B2-1z-3b-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_le_rec==1) {
					    $type = 'Partial-Neighboring-Branch-Split(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_other_ovlp_rec ==1) {
					    $type = 'Partial-Neighboring-Branch-Split(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
			 	} elsif (@indels_rlv_le_ref ==1) { # (+Slide). #

    (print STDOUT "     | (iii) - B2-1z-3c |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_re_ref ==1) {

    (print STDOUT "     | (iii) - B2-1z-3c-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (scalar(@indels_rlv_le_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_re_rec) ==1) {
					    $type = 'Partial-Neighboring-Branch-Split(+slide(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
				}

			    } elsif ((@indels_rltd_rf_rec ==1) and ($len4 - $len2 == $len3)) { # Split(?).

    (print STDOUT "     | (iii) - B2-1z-4 |\n\n") if ($TEST_LEVEL_newtype>2);

				if (@indels_rlv_le_ref ==1) { # Normal (w/o 'bounding' column). #

    (print STDOUT "     | (iii) - B2-1z-4a |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_re_ref ==1) {

    (print STDOUT "     | (iii) - B2-1z-4a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_re_rec==1) {
					    $type = 'Partial-Neighboring-Branch-Split(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_le_rec ==1) {
					    $type = 'Partial-Neighboring-Branch-Split(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
				} elsif (@indels_rlv_other_ovlp_ref ==1) { # Normal (with 'bounding' column). #

    (print STDOUT "     | (iii) - B2-1z-4b |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_re_ref ==1) {

    (print STDOUT "     | (iii) - B2-1z-4b-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (@indels_rlv_re_rec==1) {
					    $type = 'Partial-Neighboring-Branch-Split(?)';
					    $penalty = 5; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					} elsif (@indels_rlv_other_ovlp_rec ==1) {
					    $type = 'Partial-Neighboring-Branch-Split(+shift(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
			 	} elsif (@indels_rlv_re_ref ==1) { # (+Slide). #

    (print STDOUT "     | (iii) - B2-1z-4c |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (@indels_rltd_le_ref ==1) {

    (print STDOUT "     | (iii) - B2-1z-4c-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if (scalar(@indels_rlv_re_rec) + scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_le_rec) ==1) {
					    $type = 'Partial-Neighboring-Branch-Split(+slide(?))(?)';
					    $penalty = 5 + 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					}
				    }
				}
			    }

			} # END of "unless (defined $type) {...}"

=A DEBUGGING measure on Nov 7 (Fri), 2014.
			    #{
			    } elsif ((@indels_rlv_le_ref==1) and (@indels_rltd_lf_ref==1) and (@indels_rltd_le_rec==1)) {
				(scalar (@indels_rlv_other_ovlp_rec) + scalar (@indels_rlv_re_rec) ==1) and ($type = 'Partial-Neighboring-Branch-Merge(?)');
			    } elsif ((@indels_rlv_re_ref==1) and (@indels_rltd_rf_ref==1) and (@indels_rltd_re_rec==1)) {
				(scalar (@indels_rlv_other_ovlp_rec) + scalar (@indels_rlv_le_rec) ==1) and ($type = 'Partial-Neighboring-Branch-Merge(?)');

			    } elsif ((@indels_rlv_le_rec==1) and (@indels_rltd_lf_rec==1) and (@indels_rltd_le_ref==1)) {
				(scalar (@indels_rlv_other_ovlp_ref) + scalar (@indels_rlv_re_ref) ==1) and ($type = 'Partial-Neighboring-Branch-Split(?)');
			    } elsif ((@indels_rlv_re_rec==1) and (@indels_rltd_rf_rec==1) and (@indels_rltd_re_ref==1)) {
				(scalar (@indels_rlv_other_ovlp_ref) + scalar (@indels_rlv_le_ref) ==1) and ($type = 'Partial-Neighboring-Branch-Split(?)');
			    }
=cut # END of "A DEBUGGING measure on Nov 7 (Fri), 2014."

			} # END of "if ( (($br3 == $br4) and ...) or ....) {...}"

		    } # End of "if (($br1 == $br2) and ...) {...} elsif (($br1 != $br2) and ...) {...}"

		} #  END of the BIG-UNLESS, "unless (defined $type) {...}".

    (print STDOUT "     | (iii) - Z |\n\n") if ($TEST_LEVEL_newtype>2);

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


			### NOTE: Probably, similar cases should be seen in other errors, such as ...
			### ... I WILL TAKE CARE OF THESE CASES as well LATER...

		} else {

			# Substep (x): If the subject comes up to here, Regard the associated error as 'Complex'. #

    (print STDOUT "  { Substep (x) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'Complex';
		    $penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		}

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2016). #
			# => FUrther MODIFIED in REVISION on Oct 21 (Wed), 2015. #

		return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty);
		##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

		##	# Finally, register the association between the composite-block and indel events considered to be involved. #
		##
		##unless ( register_association_cblk_and_indels ($type, $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @indels_invlvd_ref, @indels_invlvd_rec, @indices_indel_identifiers) ) {
		##	print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED at the Step 3A on the $cb th composite shift-block for the $s th segment !\n";
		##	print STDERR "[ERROR '$program_name'] $!\n\n";
		##	die ;
		##}
		##
		##next;

	    } elsif (($ct_rltd_indels_rec == 2) and ($ct_rltd_indels_ref == 0)) {

				# Step 3B. #

    (print STDOUT "< Step 3B >\n\n") if ($TEST_LEVEL_newtype>0);

		my $type; # The type of MSA error(s). #
		my $penalty; # The value of penalty incurred by the error. # ADDED in REVISION on Oct 21 (Wed), 2015. #


		if (0) {
			# So far, I cannot come up with any simple types of errors resulting in this case.

		} else {

			# Substep (x): If the subject comes up to here, Regard the associated error as 'Complex'. #

    (print STDOUT "  { Substep (x) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'Complex';
		    $penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		}

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
			# => Further MODIFIED in REVISION on Oct 21 (Wed), 2015. #

		return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty);
		##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

		##	# Finally, register the association between the composite-block and indel events considered to be involved. #
		##
		##unless ( register_association_cblk_and_indels ($type, $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @indels_invlvd_ref, @indels_invlvd_rec, @indices_indel_identifiers) ) {
		##	print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED at the Step 3B on the $cb th composite shift-block for the $s th segment !\n";
		##	print STDERR "[ERROR '$program_name'] $!\n\n";
		##	die ;
		##}
		##
		##next;

	    } elsif (($ct_rltd_indels_rec == 0) and ($ct_rltd_indels_ref == 2)) {

				# Step 3C. #

    (print STDOUT "< Step 3C >\n\n") if ($TEST_LEVEL_newtype>0);

		my $type; # The type of MSA error(s). #
		my $penalty; # The value of penalty incurred by the error. # ADDED in REVISION on Oct 21 (Wed), 2015. #

		if (0) {
			# So far, I cannot come up with any simple types of errors resulting in this case.

		} else {

			# Substep (x): If the subject comes up to here, Regard the associated error as 'Complex'. #

    (print STDOUT "  { Substep (x) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'Complex';
		    $penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #
		}

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
			# => Further MODIFIED in REVISION on Oct 21 (Wed), 2015. #

		return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty);
		##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

		##	# Finally, register the association between the composite-block and indel events considered to be involved. #
		##
		##unless ( register_association_cblk_and_indels ($type, $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @indels_invlvd_ref, @indels_invlvd_rec, @indices_indel_identifiers) ) {
		##	print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED at the Step 3C on the $cb th composite shift-block for the $s th segment !\n";
		##	print STDERR "[ERROR '$program_name'] $!\n\n";
		##	die ;
		##}
		##
		##next;

	    } # END of "if (($ct_rltd_indels_rec == 1) and ($ct_rltd_indels_ref == 1)) {... } elsif (($ct_rltd_indels_rec == 2) and ($ct_rltd_indels_ref == 0)) {... } elsif (($ct_rltd_indels_rec == 0) and ($ct_rltd_indels_ref == 2)) {...}"



			# Step 4: A case that involves three events on relative branches. #

	    if (($ct_rltd_indels_rec == 2) and ($ct_rltd_indels_ref == 1)) {

				# Step 4A. #

    (print STDOUT "< Step 4A >\n\n") if ($TEST_LEVEL_newtype>0);

		my $type; # The type of MSA error(s). #
		my $penalty; # The value of penalty incurred by the error. # ADDED in REVISION on Oct 21 (Wed), 2015. #

		if (($ct_rlv_indels_rec == 0) and ($ct_rlv_indels_ref == 1)) {

			# Substep (i): 

    (print STDOUT "  { Substep (i) }\n\n") if ($TEST_LEVEL_newtype>1);

				# A: First, suspect a 'Collapse of Independent Insertions' accompanied by a 'related' indel event flanking the reconstructed block. #

		    my $subtype;
		    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #
		    my @indels_invlvd_ref1 = my @indels_invlvd_rec1 = my @indels_invlvd_rec1p = my @indels_invlvd_ref1p = ();

		    if (scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) == 0) {

    (print STDOUT "     | (i) - A |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((@{$indels_le_rec} >= 2) and (@indels_rltd_lf_ref ==1) and (@indels_rltd_lf_rec==1) ) {

    (print STDOUT "     | (i) - A-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my ($br1, $state_ue1) = @{$indels_rltd_lf_ref[0]}[0, 3];
			    my ($br2, $state_ue2) = @{$indels_rltd_lf_rec[0]}[0, 3];

			    if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2)) and ($state_ue1 ne $state_ue2)) ) { # The left-flanking event on reference is of the same type as that on reconstructed.

    (print STDOUT "     | (i) - A-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				foreach my $event_info (@{$indels_other_ovlp_ref}) {
				    my $end = $event_info->[2];
				    ($end + 1 == $beg_cb_ref) and (push @indels_invlvd_ref1, $event_info);
				}
				if (@indels_invlvd_ref1 ==1) {

    (print STDOUT "     | (i) - A-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				    push @indels_invlvd_ref1, @indels_rlv_ref;
				    @indels_invlvd_rec1 = @{$indels_le_rec};
				    @indels_invlvd_ref1p = @indels_rltd_lf_ref;
				    @indels_invlvd_rec1p = @indels_rltd_lf_rec;

				    if (@indels_rlv_le_ref ==1) {
					$subtype = '';
				    } else {
					$subtype = '(+shift)';
					$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				}
			    }

			} elsif ((@{$indels_re_rec} >= 2) and (@indels_rltd_rf_ref ==1) and (@indels_rltd_rf_rec==1) ) {

    (print STDOUT "     | (i) - A-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my ($br1, $state_ue1) = @{$indels_rltd_rf_ref[0]}[0, 3];
			    my ($br2, $state_ue2) = @{$indels_rltd_rf_rec[0]}[0, 3];

			    if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2)) and ($state_ue1 ne $state_ue2)) ) { # The right-flaning event on reference is of the same type as that on reconstructed.

    (print STDOUT "     | (i) - A-2a |\n\n") if ($TEST_LEVEL_newtype>2);

				foreach my $event_info (@{$indels_other_ovlp_ref}) {
				    my $beg = $event_info->[1];
				    ($beg == $end_cb_ref + 1) and (push @indels_invlvd_ref1, $event_info);
				}
				if (@indels_invlvd_ref1 ==1) {

    (print STDOUT "     | (i) - A-2a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				    push @indels_invlvd_ref1, @indels_rlv_ref;
				    @indels_invlvd_rec1 = @{$indels_re_rec};
				    @indels_invlvd_ref1p = @indels_rltd_rf_ref;
				    @indels_invlvd_rec1p = @indels_rltd_rf_rec;
	
				    if (@indels_rlv_re_ref ==1) {
					$subtype = '';
				    } else {
					$subtype = '(+shift)';
					$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				}
			   }

			} # END of "if ((@{$indels_le_rec} >= 2) and (...) ) {..} elsif ((@{$indels_re_rec} >=2) and (...) ) {...}".

		    } # END of "if (scalar(@indels_rlv_lf_ref) + scalar(@indels_rlv_rf_ref) ==0) {...}"

		    if (defined $subtype) {

    (print STDOUT "     | (i) - A-Test |\n\n") if ($TEST_LEVEL_newtype>2);

			my ($if_cii, $indels_invlvd_rec2) = if_collapse_indep_ins (@indels_invlvd_ref1, @indels_invlvd_rec1, %{$node2pa}, %{$node2ch}, $CHAR_ABS);
			unless (defined $if_cii) {
			    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i)-A of Step 4A: $! !\n\n";
			    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

			    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i)-A of Step 4A on the $cb th composite shift-block for the $s th segment !\n";
			    ##print STDERR "[ERROR '$program_name'] $!\n\n";
			    ##die ;
			}

			if ($if_cii) {

    (print STDOUT "     | (i) - A-Test-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_ref = (@indels_invlvd_ref1, @indels_invlvd_ref1p);
			    @indels_invlvd_rec = (@{$indels_invlvd_rec2}, @indels_invlvd_rec1p);
			    my $diffct = scalar (@indels_invlvd_rec) - @indels_invlvd_ref;

			    $cblk_wise_cts_invlvd_indels[$cb]->[5] = scalar (@{$indels_invlvd_rec2}) - 1;

			    my $len1 = $indels_invlvd_ref[0]->[5];
			    my $len2 = $indels_invlvd_rec[0]->[5];
			    if ($len1 == $len2) {
				$type = "Collapse-of-Independent-Insertions$subtype(+$diffct)(flanking-related-event-accompanied)(?)";
				$penalty = 10 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    } else {
				$diffct++;
				##$type = "Incomplete-Collapse-of-Independent-Insertions$subtype(+$diffct)(flanking-related-event-accompanied)(?)"; # This is dubious...#
			    }
			    
			}
		    }

					# ADDED on Nov 12 (Wed), 2014. #

		    if (!(defined $type) and (scalar(@indels_rltd_lf_ref) + scalar(@indels_rltd_rf_ref) ==1) and (scalar(@indels_rltd_lf_rec) + scalar(@indels_rltd_rf_rec) ==1) ) {

				# B: In this case, Suspect a 'Partial-Neighboring-Branch-Merge' (2ins) (Short-Block(normal)). #

    (print STDOUT "     | (i) - B |\n\n") if ($TEST_LEVEL_newtype>2);

			my $event1 = $indels_rlv_ref[0];
			my $event2 = $indels_rltd_ref[0];
			my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$event1}[0..3, 5];
			my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$event2}[0..3, 5];
			my $eq_br2 = $equiv_branches{$br2};

			if ( ($len2 > $len1) and (($end1 + 1 < $beg2) or ($end2 + 1 < $beg1)) ) { # 'ref_rltd' is longer than 'ref_rlv', and the two events are separated.

    (print STDOUT "     | (i) - B-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my ($event3, $event4); # event 3 must be on the different branch, event 4 must be on the same branch and of the same type as event2.

			    foreach my $event (@indels_rltd_rec) { # Inner foreach-loop (over 'related' indels in reconstructed MSA).
				my ($br, $state_ue) = @{$event}[0, 3];
				if ($br2 == $br) {
				    ($state_ue2 eq $state_ue) and ($event4 = $event);

				} elsif ((defined $eq_br2) and ($eq_br2 == $br)) {
				    ($state_ue2 ne $state_ue) and ($event4 = $event);

				} else {
				    $event3 = $event;
				}
			    }				# End of the inner foreach-loop (over 'related' indels in reconstructed MSA).

			    if ((defined $event4) and (defined $event3) and ($len1 == $event3->[5]) and ($len2 == $len1 + $event4->[5])) {

    (print STDOUT "     | (i) - B-Test |\n\n") if ($TEST_LEVEL_newtype>2);

				my $if_part_nbr_mrg = if_part_nbr_mrg_2ins (@{$event1}, @{$event2}, @{$event3}, 0, 3, %equiv_branches, %{$node2pa}, $CHAR_ABS) ;
				unless (defined $if_part_nbr_mrg) {
				    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (i)-A of Step 4A: $! !\n\n";
				    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

				    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (i)-A of Step 4A on the $cb th composite shift-block for the $s th segment !\n";
				    ##print STDERR "[ERROR '$program_name'] $!\n\n";
				    ##die ;
				}

				if ($if_part_nbr_mrg) {

    (print STDOUT "     | (i) - B-Test-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

				    my ($br3, $beg3, $end3) = @{$event3}[0,1,2];
				    my ($br4, $beg4, $end4) = @{$event4}[0,1,2];
				    my $subtype;
				    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    if (@indels_rltd_lf_ref == 1) {

    (print STDOUT "     | (i) - B-Test-Pass-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					if ((@indels_rltd_lf_rec ==1) and ($indels_rltd_lf_rec[0]->[0] == $br4) and ((@indels_rlv_re_ref ==1) or (@indels_rlv_other_ovlp_ref ==1)) ) {

    (print STDOUT "     | (i) - B-Test-Pass-1a |\n\n") if ($TEST_LEVEL_newtype>2);

					    if (($end4 + 1 < $beg3) and (@indels_rltd_re_rec ==1)) {
						$subtype = '(+shift)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					    } elsif (($end4 + 1 == $beg3) and (@indels_rltd_le_rec ==1)) { 
						$subtype = '(+shift)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					    }

					} elsif ((@indels_rltd_rf_rec ==1) and ($indels_rltd_rf_rec[0]->[0] == $br4) and ($end3 + 1 == $beg4) and (@indels_rltd_re_rec ==1)) {

    (print STDOUT "     | (i) - B-Test-Pass-1b |\n\n") if ($TEST_LEVEL_newtype>2);

					    if (@indels_rlv_re_ref ==1) {
						$subtype = '(+overshoot)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					    } elsif ((@indels_rlv_rf_ref ==1) or ((@indels_rlv_other_ovlp_ref ==1) and ($end2 + 1 < $beg1)) ) {
						$subtype = '(+shift+slide)(?)';
						$sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					    }
					}

				    } else { # @indels_rltd_rf_ref ==1. #

    (print STDOUT "     | (i) - B-Test-Pass-2 |\n\n") if ($TEST_LEVEL_newtype>2);

					if ((@indels_rltd_rf_rec ==1) and ($indels_rltd_rf_rec[0]->[0] == $br4) and ((@indels_rlv_le_ref ==1) or (@indels_rlv_other_ovlp_ref ==1)) ) {

    (print STDOUT "     | (i) - B-Test-Pass-2a |\n\n") if ($TEST_LEVEL_newtype>2);

					    if (($end3 + 1 < $beg4) and (@indels_rltd_le_rec ==1)) {
						$subtype = '(+shift)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					    } elsif (($end3 + 1 == $beg4) and (@indels_rltd_re_rec ==1)) { 
						$subtype = '(+shift)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					    }

					} elsif ((@indels_rltd_lf_rec ==1) and ($indels_rltd_lf_rec[0]->[0] == $br4) and ($end4 + 1 == $beg3) and (@indels_rltd_le_rec ==1)) {

    (print STDOUT "     | (i) - B-Test-Pass-2b |\n\n") if ($TEST_LEVEL_newtype>2);

					    if (@indels_rlv_le_ref ==1) {
						$subtype = '(+overshoot)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					    } elsif ((@indels_rlv_lf_ref ==1) or ((@indels_rlv_other_ovlp_ref ==1) and ($end1 + 1 < $beg2)) ) {
						$subtype = '(+shift+slide)(?)';
						$sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					    }
					}
				    }

				    if (defined $subtype) {
					$type = "Partial-Neighboring-Branch-Merge(2ins)$subtype";
					$penalty = 5 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }

				} # END of "if ($if_part_nbr_mrg) {...}"
			    }

			} # END of "if (($len2 > $len1) and (...)) {...}"

		    } # END of "if (!(defined $type) and ...) {...}"

					# END of "ADDED on Nov 12 (Wed), 2014." #


		    unless (defined $type) {

				# C: Suspect a 'Partial-Neighboring-Branch-Merge' (Short-block). #

    (print STDOUT "     | (i) - C |\n\n") if ($TEST_LEVEL_newtype>2);

			my $subtype;
			my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			my @indels_invlvd_rec2 = ();
			if ((@indels_rltd_lf_rec==1) and (@indels_rltd_le_rec==1) and (@indels_rltd_lf_ref==1)) {

    (print STDOUT "     | (i) - C-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				# A BUG fixed on Nov 7 (Fri),2014. #
			    if (@indels_rlv_le_ref ==1) {
				$subtype = '(?)';
			    } elsif (scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_re_ref) ==1) {
				$subtype = '(+shift)(?)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }
			    ##$subtype = (@indels_rlv_le_ref==1) ? '(?)' : '(+shift)(?)';
			    @indels_invlvd_rec2 = (@indels_rltd_lf_rec, @indels_rltd_le_rec);

			} elsif ((@indels_rltd_rf_rec==1) and (@indels_rltd_re_rec==1) and (@indels_rltd_rf_ref==1)) {

    (print STDOUT "     | (i) - C-2 |\n\n") if ($TEST_LEVEL_newtype>2);

				# A BUG fixed on Nov 7 (Fri),2014. #
			    if (@indels_rlv_re_ref ==1) {
				$subtype = '(?)';
			    } elsif (scalar(@indels_rlv_other_ovlp_ref) + scalar(@indels_rlv_le_ref) ==1) {
				$subtype = '(+shift)(?)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }
			    ##$subtype = (@indels_rlv_re_ref==1) ? '(?)' : '(+shift)(?)';
			    @indels_invlvd_rec2 = (@indels_rltd_rf_rec, @indels_rltd_re_rec);

			} elsif ((@indels_rltd_lf_rec ==1) and (@indels_rltd_le_rec ==1) and (@indels_rltd_rf_ref ==1) and (@indels_rlv_le_ref ==1)) {
			##{...} elsif ((@indels_rltd_lf_rec ==1) and (@indels_rltd_le_rec ==1) and (@indels_rltd_rf_ref ==1) and (@indels_rlv_re_ref ==1)) {...} # A BUG fixed on Nov 7 (Fri), 2014.

    (print STDOUT "     | (i) - C-3 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $subtype = '(+slide)(?)';
			    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    @indels_invlvd_rec2 = (@indels_rltd_lf_rec, @indels_rltd_le_rec);

			} elsif ((@indels_rltd_rf_rec ==1) and (@indels_rltd_re_rec ==1) and (@indels_rltd_lf_ref ==1) and (@indels_rlv_re_ref ==1)) {
			##{...} elsif ((@indels_rltd_rf_rec ==1) and (@indels_rltd_re_rec ==1) and (@indels_rltd_lf_ref ==1) and (@indels_rlv_le_ref ==1)) {...} # A BUG fixed on Nov 7 (Fri), 2014. #

    (print STDOUT "     | (i) - C-4 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $subtype = '(+slide)(?)';
			    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    @indels_invlvd_rec2 = (@indels_rltd_rf_rec, @indels_rltd_re_rec);
			}

			if (defined $subtype) {

    (print STDOUT "     | (i) - C-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

			    my ($br1, $state_ue1, $len1) = @{$indels_invlvd_rec2[0]}[0, 3, 5];
			    my ($br2, $state_ue2, $len2) = @{$indels_invlvd_rec2[1]}[0, 3, 5];
			    my ($br3, $state_ue3, $len3) = @{$indels_rltd_ref[0]}[0, 3, 5];
			    my ($br4, $state_ue4, $len4) = @{$indels_rlv_ref[0]}[0, 3, 5];

			    if ((($br1 == $br3) and ($state_ue1 eq $state_ue3)) or ((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br3) and ($state_ue1 ne $state_ue3)) ) { # The flanking event on reconstructed must be of the same type as that on reference.

    (print STDOUT "     | (i) - C-Pass-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			        if (($br1 != $br2) and !((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2)) ) { # Flanking event and block-ending-event must be on different branches. #

    (print STDOUT "     | (i) - C-Pass-1a |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				    if ($len1 + $len4 == $len3) {
					$type = "Partial-Neighboring-Branch-Merge$subtype";
					$penalty = 5 + $sbpenalty;
				    }
				    ##($len1 + $len4 == $len3) and ($type = "Partial-Neighboring-Branch-Merge$subtype");

				    ##(($len1 + $len2 == $len3) and ($len2 == $len4)) and ($type = "Partial-Neighboring-Branch-Merge$subtype"); # A BUG fixed on Nov 7 (Fri), 2014. #
			        }
			    }
			}

		    }

    (print STDOUT "     | (i) - Z |\n\n") if ($TEST_LEVEL_newtype>2);

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} else {

			# Substep (x): If the subject comes up to here, Regard the associated error as 'Complex'. #

    (print STDOUT "  { Substep (x) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'Complex';
		    $penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		}

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
			# => Further MODIFIED in REVISION on Oct 21  (Wed), 2015. #

		return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty);
		##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

		##	# Finally, register the association between the composite-block and indel events considered to be involved. #
		##
		##unless ( register_association_cblk_and_indels ($type, $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @indels_invlvd_ref, @indels_invlvd_rec, @indices_indel_identifiers) ) {
		##	print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED at the Step 4A on the $cb th composite shift-block for the $s th segment !\n";
		##	print STDERR "[ERROR '$program_name'] $!\n\n";
		##	die ;
		##}
		##
		##next;

	    } elsif (($ct_rltd_indels_ref == 2) and ($ct_rltd_indels_rec == 1)) {

				# Step 4B. #

    (print STDOUT "< Step 4B >\n\n") if ($TEST_LEVEL_newtype>0);

		my $type; # The type of MSA error(s). #
		my $penalty; # The value of penalty incurred by the error. # ADDED in REVISIOn on Oct 21 (Wed), 2015. #

		if (($ct_rlv_indels_ref == 0) and ($ct_rlv_indels_rec == 1)) {

			# Substep (i): 

    (print STDOUT "  { Substep (i) }\n\n") if ($TEST_LEVEL_newtype>1);

				# A: First, suspect a 'Creation of Erroneous Independent Insertions' accompanied by a 'related' indel event flanking the reconstructed block. #

		    my $subtype;
		    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #
		    my @indels_invlvd_rec1 = my @indels_invlvd_ref1 = my @indels_invlvd_ref1p = my @indels_invlvd_rec1p = ();

		    if (scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) == 0) {

    (print STDOUT "     | (i) - A |\n\n") if ($TEST_LEVEL_newtype>2);

			if ((@{$indels_le_ref} >= 2) and (@indels_rltd_lf_rec ==1) and (@indels_rltd_lf_ref==1) ) {

    (print STDOUT "     | (i) - A-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my ($br1, $state_ue1) = @{$indels_rltd_lf_rec[0]}[0, 3];
			    my ($br2, $state_ue2) = @{$indels_rltd_lf_ref[0]}[0, 3];

			    if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2)) and ($state_ue1 ne $state_ue2)) ) { # The left-flanking event on reference is of the same type as that on reconstructed.

    (print STDOUT "     | (i) - A-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				foreach my $event_info (@{$indels_other_ovlp_rec}) {
				    my $end = $event_info->[2];
				    ($end + 1 == $beg_cb_rec) and (push @indels_invlvd_rec1, $event_info);
				}
				if (@indels_invlvd_rec1 ==1) {

    (print STDOUT "     | (i) - A-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				    push @indels_invlvd_rec1, @indels_rlv_rec;
				    @indels_invlvd_ref1 = @{$indels_le_ref};
				    @indels_invlvd_rec1p = @indels_rltd_lf_rec;
				    @indels_invlvd_ref1p = @indels_rltd_lf_ref;

				    if (@indels_rlv_le_rec ==1) {
					$subtype = '';
				    } else {
					$subtype = '(+shift)';
					$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				}
			    }

			} elsif ((@{$indels_re_ref} >= 2) and (@indels_rltd_rf_rec ==1) and (@indels_rltd_rf_ref==1) ) {

    (print STDOUT "     | (i) - A-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my ($br1, $state_ue1) = @{$indels_rltd_rf_rec[0]}[0, 3];
			    my ($br2, $state_ue2) = @{$indels_rltd_rf_ref[0]}[0, 3];

			    if ( (($br1 == $br2) and ($state_ue1 eq $state_ue2)) or (((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2)) and ($state_ue1 ne $state_ue2)) ) { # The right-flaning event on reference is of the same type as that on reconstructed.

    (print STDOUT "     | (i) - A-2a |\n\n") if ($TEST_LEVEL_newtype>2);

				foreach my $event_info (@{$indels_other_ovlp_rec}) {
				    my $beg = $event_info->[1];
				    ($beg == $end_cb_rec + 1) and (push @indels_invlvd_rec1, $event_info);
				}
				if (@indels_invlvd_rec1 ==1) {

    (print STDOUT "     | (i) - A-2a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				    push @indels_invlvd_rec1, @indels_rlv_rec;
				    @indels_invlvd_ref1 = @{$indels_re_ref};
				    @indels_invlvd_rec1p = @indels_rltd_rf_rec;
				    @indels_invlvd_ref1p = @indels_rltd_rf_ref;
	
				    if (@indels_rlv_re_rec ==1) {
					$subtype = '';
				    } else {
					$subtype = '(+shift)';
					$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }
				}
			   }

			} # END of "if ((@{$indels_le_ref} >= 2) and (...) ) {..} elsif ((@{$indels_re_ref} >=2) and (...) ) {...}".

		    } # END of "if (scalar(@indels_rlv_lf_rec) + scalar(@indels_rlv_rf_rec) ==0) {...}"

		    if (defined $subtype) {

    (print STDOUT "     | (i) - A-Test |\n\n") if ($TEST_LEVEL_newtype>2);

			my ($if_cii, $indels_invlvd_ref2) = if_collapse_indep_ins (@indels_invlvd_rec1, @indels_invlvd_ref1, %{$node2pa}, %{$node2ch}, $CHAR_ABS);
			unless (defined $if_cii) {
			    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i)-A of Step 4B: $! !\n\n";
			    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

			    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_collapse_indep_ins' FAILED at Substep (i)-A of Step 4B on the $cb th composite shift-block for the $s th segment !\n";
			    ##print STDERR "[ERROR '$program_name'] $!\n\n";
			    ##die ;
			}

			if ($if_cii) {

    (print STDOUT "     | (i) - A-Test-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec = (@indels_invlvd_rec1, @indels_invlvd_rec1p);
			    @indels_invlvd_ref = (@{$indels_invlvd_ref2}, @indels_invlvd_ref1p);
			    my $diffct = scalar (@indels_invlvd_ref) - @indels_invlvd_rec;

			    $cblk_wise_cts_invlvd_indels[$cb]->[4] = scalar (@{$indels_invlvd_ref2}) - 1;

			    my $len1 = $indels_invlvd_rec[0]->[5];
			    my $len2 = $indels_invlvd_ref[0]->[5];
			    if ($len1 == $len2) {
				$type = "Creation-of-Erroneous-Independent-Insertions$subtype(-$diffct)(flanking-related-event-accompanied)(?)";
				$penalty = 10 + $sbpenalty +1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    } else {
				$diffct++;
				##$type = "Incomplete-Creation-of-Erroneous-Independent-Insertions$subtype(-$diffct)(flanking-related-event-accompanied)(?)"; # This is dubious...#
			    }
			    
			}
		    }


					# ADDED on Nov 12 (Wed), 2014. #

		    if (!(defined $type) and (scalar(@indels_rltd_lf_rec) + scalar(@indels_rltd_rf_rec) ==1) and (scalar(@indels_rltd_lf_ref) + scalar(@indels_rltd_rf_ref) ==1) ) {

				# B: In this case, Suspect a 'Partial-Neighboring-Branch-Split' (2ins) (Short-Block(normal)). #

    (print STDOUT "     | (i) - B |\n\n") if ($TEST_LEVEL_newtype>2);

			my $event1 = $indels_rlv_rec[0];
			my $event2 = $indels_rltd_rec[0];
			my ($br1, $beg1, $end1, $state_ue1, $len1) = @{$event1}[0..3, 5];
			my ($br2, $beg2, $end2, $state_ue2, $len2) = @{$event2}[0..3, 5];
			my $eq_br2 = $equiv_branches{$br2};

			if ( ($len2 > $len1) and (($end1 + 1 < $beg2) or ($end2 + 1 < $beg1)) ) { # 'rec_rltd' is longer than 'rec_rlv', and the two events are separated.

    (print STDOUT "     | (i) - B-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    my ($event3, $event4); # event 3 must be on the different branch, event 4 must be on the same branch and of the same type as event2.

			    foreach my $event (@indels_rltd_ref) { # Inner foreach-loop (over 'related' indels in reference MSA).
				my ($br, $state_ue) = @{$event}[0, 3];
				if ($br2 == $br) {
				    ($state_ue2 eq $state_ue) and ($event4 = $event);

				} elsif ((defined $eq_br2) and ($eq_br2 == $br)) {
				    ($state_ue2 ne $state_ue) and ($event4 = $event);

				} else {
				    $event3 = $event;
				}
			    }				# End of the inner foreach-loop (over 'related' indels in reconstructed MSA).

			    if ((defined $event4) and (defined $event3) and ($len1 == $event3->[5]) and ($len2 == $len1 + $event4->[5])) {

    (print STDOUT "     | (i) - B-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				my $if_part_nbr_mrg = if_part_nbr_mrg_2ins (@{$event1}, @{$event2}, @{$event3}, 0, 3, %equiv_branches, %{$node2pa}, $CHAR_ABS) ;
				unless (defined $if_part_nbr_mrg) {
				    print STDOUT "[ERROR 'classify_error_assoc_w_sgl_cblk'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (i)-B of Step 4B: $! !\n\n";
				    return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

				    ##print STDOUT "[ERROR '$program_name'] Subroutine 'if_part_nbr_mrg' FAILED at Substep (i)-B of Step 4B on the $cb th composite shift-block for the $s th segment !\n";
				    ##print STDERR "[ERROR '$program_name'] $!\n\n";
				    ##die ;
				}

				if ($if_part_nbr_mrg) {

    (print STDOUT "     | (i) - B-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				    my ($br3, $beg3, $end3) = @{$event3}[0,1,2];
				    my ($br4, $beg4, $end4) = @{$event4}[0,1,2];
				    my $subtype;
				    my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #

				    if (@indels_rltd_lf_rec == 1) {

    (print STDOUT "     | (i) - B-1a-1a |\n\n") if ($TEST_LEVEL_newtype>2);

					if ((@indels_rltd_lf_ref ==1) and ($indels_rltd_lf_ref[0]->[0] == $br4) and ((@indels_rlv_re_rec ==1) or (@indels_rlv_other_ovlp_rec ==1)) ) {

    (print STDOUT "     | (i) - B-1a-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					    if (($end4 + 1 < $beg3) and (@indels_rltd_re_ref ==1)) {
						$subtype = '(+shift)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					    } elsif (($end4 + 1 == $beg3) and (@indels_rltd_le_ref ==1)) { 
						$subtype = '(+shift)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					    }

					} elsif ((@indels_rltd_rf_ref ==1) and ($indels_rltd_rf_ref[0]->[0] == $br4) and ($end3 + 1 == $beg4) and (@indels_rltd_re_ref ==1)) {

    (print STDOUT "     | (i) - B-1a-1a-2 |\n\n") if ($TEST_LEVEL_newtype>2);

					    if (@indels_rlv_re_rec ==1) {
						$subtype = '(+overshoot)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					    } elsif ((@indels_rlv_rf_rec ==1) or ((@indels_rlv_other_ovlp_rec ==1) and ($end2 + 1 < $beg1)) ) {
						$subtype = '(+shift+slide)(?)';
						$sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					    }
					}

				    } else { # @indels_rltd_rf_rec ==1. #

    (print STDOUT "     | (i) - B-1a-1b |\n\n") if ($TEST_LEVEL_newtype>2);

					if ((@indels_rltd_rf_ref ==1) and ($indels_rltd_rf_ref[0]->[0] == $br4) and ((@indels_rlv_le_rec ==1) or (@indels_rlv_other_ovlp_rec ==1)) ) {

    (print STDOUT "     | (i) - B-1a-1b-1 |\n\n") if ($TEST_LEVEL_newtype>2);

					    if (($end3 + 1 < $beg4) and (@indels_rltd_le_ref ==1)) {
						$subtype = '(+shift)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					    } elsif (($end3 + 1 == $beg4) and (@indels_rltd_re_ref ==1)) { 
						$subtype = '(+shift)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					    }

					} elsif ((@indels_rltd_lf_ref ==1) and ($indels_rltd_lf_ref[0]->[0] == $br4) and ($end4 + 1 == $beg3) and (@indels_rltd_le_ref ==1)) {

    (print STDOUT "     | (i) - B-1a-1b-2 |\n\n") if ($TEST_LEVEL_newtype>2);

					    if (@indels_rlv_le_rec ==1) {
						$subtype = '(+overshoot)(?)';
						$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #

					    } elsif ((@indels_rlv_lf_rec ==1) or ((@indels_rlv_other_ovlp_rec ==1) and ($end1 + 1 < $beg2)) ) {
						$subtype = '(+shift+slide)(?)';
						$sbpenalty += 2; # ADDED in REVISION on Oct 21 (Wed), 2015. #
					    }
					}
				    }

				    if (defined $subtype) {
					$type = "Partial-Neighboring-Branch-Split(2ins)$subtype";
					$penalty = 5 + $sbpenalty; # ADDED in REVISION on Oct 21 (Wed), 2015. #
				    }

				} # END of "if ($if_part_nbr_mrg) {...}"
			    }

			} # END of "if (($len2 > $len1) and (...)) {...}"

		    } # END of "if (!(defined $type) and ...) {...}"

					# END of "ADDED on Nov 12 (Wed), 2014." #


		    unless (defined $type) {

				# C: Suspect a 'Partial-Neighboring-Branch-Split' (Short-block). #

    (print STDOUT "     | (i) - C |\n\n") if ($TEST_LEVEL_newtype>2);

			my $subtype;
			my $sbpenalty = 0; # ADDED in REVISION on Oct 21 (Wed), 2015. #

			my @indels_invlvd_ref2 = ();
			if ((@indels_rltd_lf_ref==1) and (@indels_rltd_le_ref==1) and (@indels_rltd_lf_rec==1)) {

    (print STDOUT "     | (i) - C-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				# A BUG fixed on Nov 7 (Fri), 2014. #
			    if (@indels_rlv_le_rec ==1) {
				$subtype = '(?)';
			    } elsif (scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_re_rec) ==1) {
				$subtype = '(+shift)(?)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }
			    ##$subtype = (@indels_rlv_le_rec==1) ? '(?)' : '(+shift)(?)';
			    @indels_invlvd_ref2 = (@indels_rltd_lf_ref, @indels_rltd_le_ref);

			} elsif ((@indels_rltd_rf_ref==1) and (@indels_rltd_re_ref==1) and (@indels_rltd_rf_rec==1)) {

    (print STDOUT "     | (i) - C-2 |\n\n") if ($TEST_LEVEL_newtype>2);

				# A BUG fixed on Nov 7 (Fri), 2014. #
			    if (@indels_rlv_re_rec ==1) {
				$subtype = '(?)';
			    } elsif (scalar(@indels_rlv_other_ovlp_rec) + scalar(@indels_rlv_le_rec) ==1) {
				$subtype = '(+shift)(?)';
				$sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    }
			    ##$subtype = (@indels_rlv_re_rec==1) ? '(?)' : '(+shift)(?)';
			    @indels_invlvd_ref2 = (@indels_rltd_rf_ref, @indels_rltd_re_ref);

			} elsif ((@indels_rltd_lf_ref ==1) and (@indels_rltd_le_ref ==1) and (@indels_rltd_rf_rec ==1) and (@indels_rlv_le_rec ==1)) {
			##{...} elsif ((@indels_rltd_lf_ref ==1) and (@indels_rltd_le_ref ==1) and (@indels_rltd_rf_rec ==1) and (@indels_rlv_re_rec ==1)) {...} # A BUG fixed on Nov 7 (Fri), 2014. #

    (print STDOUT "     | (i) - C-3 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $subtype = '(+slide)(?)';
			    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    @indels_invlvd_ref2 = (@indels_rltd_lf_ref, @indels_rltd_le_ref);

			} elsif ((@indels_rltd_rf_ref ==1) and (@indels_rltd_re_ref ==1) and (@indels_rltd_lf_rec ==1) and (@indels_rlv_re_rec ==1)) {
			##{...} elsif ((@indels_rltd_rf_ref ==1) and (@indels_rltd_re_ref ==1) and (@indels_rltd_lf_rec ==1) and (@indels_rlv_le_rec ==1)) {...} #   A BUG fixed on Nov 7 (Fri), 2014.

    (print STDOUT "     | (i) - C-4 |\n\n") if ($TEST_LEVEL_newtype>2);

			    $subtype = '(+slide)(?)';
			    $sbpenalty += 1; # ADDED in REVISION on Oct 21 (Wed), 2015. #
			    @indels_invlvd_ref2 = (@indels_rltd_rf_ref, @indels_rltd_re_ref);
			}

			if (defined $subtype) {

    (print STDOUT "     | (i) - C-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

			    my ($br1, $state_ue1, $len1) = @{$indels_invlvd_ref2[0]}[0, 3, 5];
			    my ($br2, $state_ue2, $len2) = @{$indels_invlvd_ref2[1]}[0, 3, 5];
			    my ($br3, $state_ue3, $len3) = @{$indels_rltd_rec[0]}[0, 3, 5];
			    my ($br4, $state_ue4, $len4) = @{$indels_rlv_rec[0]}[0, 3, 5];

			    if ((($br1 == $br3) and ($state_ue1 eq $state_ue3)) or ((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br3) and ($state_ue1 ne $state_ue3)) ) { # The flanking event on reconstructed must be of the same type as that on reference.

    (print STDOUT "     | (i) - C-Pass-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			        if (($br1 != $br2) and !((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2)) ) { # Flanking event and block-ending-event must be on different branches. #

    (print STDOUT "     | (i) - C-Pass-1a |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				    if ($len1 + $len4 == $len3) {
					$type = "Partial-Neighboring-Branch-Merge$subtype";
					$penalty = 5 + $sbpenalty;
				    }
				    ##	# A BUG fixed on Nov 7 (Fri), 2014. #
				    ##($len1 + $len4 == $len3) and ($type = "Partial-Neighboring-Branch-Merge$subtype");
				    ####(($len1 + $len2 == $len3) and ($len2 == $len4)) and ($type = "Partial-Neighboring-Branch-Merge$subtype");
			        }
			    }
			}


		    }

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} else {

			# Substep (x): If the subject comes up to here, Regard the associated error as 'Complex'. #

    (print STDOUT "  { Substep (x) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'Complex';
		    $penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		}

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
			# => Further MODIFIED in REVISION on Oct 21 (Wed), 2015. #

		return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty);
		##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

		##	# Finally, register the association between the composite-block and indel events considered to be involved. #
		##
		##unless ( register_association_cblk_and_indels ($type, $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @indels_invlvd_ref, @indels_invlvd_rec, @indices_indel_identifiers) ) {
		##	print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED at the Step 4B on the $cb th composite shift-block for the $s th segment !\n";
		##	print STDERR "[ERROR '$program_name'] $!\n\n";
		##	die ;
		##}
		##
		##next;

	    } # END of "if (($ct_rltd_indels_rec == 2) and ($ct_rltd_indels_ref == 1)) {...} elsif (($ct_rltd_indels_rec == 1) and ($ct_rltd_indels_ref == 2)) {...}"


			# Step 5: When each of reference and reconstructed composite blocks involves two events on relative branches. #

	    if (($ct_rltd_indels_rec == 2) and ($ct_rltd_indels_ref == 2)) {

    (print STDOUT "< Step 5 >\n\n") if ($TEST_LEVEL_newtype>0);

		my $type; # The type of MSA error(s). #
		my $penalty; # The value of penalty incurred by the error. # ADDED in REVISION on Oct 21 (Wed), 2015. #

		if (($ct_rlv_indels_rec == 0) and ($ct_rlv_indels_ref == 0)) {

    (print STDOUT "  { Substep (i) }\n\n") if ($TEST_LEVEL_newtype>1);

		    if ((@indels_rltd_lf_rec ==1) and (@indels_rltd_rf_rec ==1) and (@indels_rltd_lf_ref ==1) and (@indels_rltd_rf_ref ==1)) {

				# Suspect a 'partial-shift' aligned with gap.

    (print STDOUT "     | (i) - A |\n\n") if ($TEST_LEVEL_newtype>2);

			my ($br1, $state_ue1, $len1) = @{$indels_rltd_lf_rec[0]}[0, 3, 5];
			my ($br2, $state_ue2, $len2) = @{$indels_rltd_rf_rec[0]}[0, 3, 5];
			my ($br3, $state_ue3, $len3) = @{$indels_rltd_lf_ref[0]}[0, 3, 5];
			my ($br4, $state_ue4, $len4) = @{$indels_rltd_rf_ref[0]}[0, 3, 5];

			if ((($br1 == $br2) and ($state_ue1 eq $state_ue2)) or ((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2) and ($state_ue1 ne $state_ue2)) ) {

    (print STDOUT "     | (i) - A-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    if ((($br3 == $br4) and ($state_ue3 eq $state_ue4)) or ((defined $equiv_branches{$br3}) and ($equiv_branches{$br3} == $br4) and ($state_ue3 ne $state_ue4)) ) {

    (print STDOUT "     | (i) - A-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				if ((($br1 == $br3) and ($state_ue1 eq $state_ue3)) or ((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br3) and ($state_ue1 ne $state_ue3)) ) {

    (print STDOUT "     | (i) - A-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
				    if ($len1+$len2 == $len3+$len4) {
					$type = 'Partial_Shift(gap-aligned)';
					$penalty = 3 + 1;
				    }
				    ##($len1+$len2 == $len3+$len4) and ($type = 'Partial_Shift(gap-aligned)');
				}
			    }
			}


		    } else {
				# Suspect a 'Double-Inter-Branch-Shift'. #

    (print STDOUT "     | (i) - B |\n\n") if ($TEST_LEVEL_newtype>2);

			my @indels_invlvd_rec1 = my @indels_invlvd_ref1 = ();

			if ((@indels_rltd_lf_rec == 1) and (@indels_rltd_re_rec ==1) and (@indels_rltd_le_ref ==1) and (@indels_rltd_rf_ref ==1)) {

    (print STDOUT "     | (i) - B-1 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = (@indels_rltd_lf_rec, @indels_rltd_re_rec);
			    @indels_invlvd_ref1 = (@indels_rltd_le_ref, @indels_rltd_rf_ref);

			} elsif ((@indels_rltd_le_rec == 1) and (@indels_rltd_rf_rec ==1) and (@indels_rltd_lf_ref ==1) and (@indels_rltd_re_ref ==1)) {

    (print STDOUT "     | (i) - B-2 |\n\n") if ($TEST_LEVEL_newtype>2);

			    @indels_invlvd_rec1 = (@indels_rltd_rf_rec, @indels_rltd_le_rec);
			    @indels_invlvd_ref1 = (@indels_rltd_re_ref, @indels_rltd_lf_ref);
			}

			if (@indels_invlvd_rec1>1) {

    (print STDOUT "     | (i) - B-Pass |\n\n") if ($TEST_LEVEL_newtype>2);

			    my ($br1, $state_ue1, $len1) = @{$indels_invlvd_rec1[0]}[0, 3, 5];
			    my ($br2, $state_ue2, $len2) = @{$indels_invlvd_rec1[1]}[0, 3, 5];
			    my ($br3, $state_ue3, $len3) = @{$indels_invlvd_ref1[0]}[0, 3, 5];
			    my ($br4, $state_ue4, $len4) = @{$indels_invlvd_ref1[1]}[0, 3, 5];

			    if ((($br1 == $br2) and ($state_ue1 ne $state_ue2)) or ((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br2) and ($state_ue1 eq $state_ue2)) ) { # The two events on reconstructed occur along the same branch but of opposite types.

    (print STDOUT "     | (i) - B-Pass-1 |\n\n") if ($TEST_LEVEL_newtype>2);

				if ((($br3 == $br4) and ($state_ue3 ne $state_ue4)) or ((defined $equiv_branches{$br3}) and ($equiv_branches{$br3} == $br4) and ($state_ue3 eq $state_ue4)) ) { # The two events on reference occur along the same branch but of opposite types.

    (print STDOUT "     | (i) - B-Pass-1a |\n\n") if ($TEST_LEVEL_newtype>2);

				    if (($br1 != $br3) and !((defined $equiv_branches{$br1}) and ($equiv_branches{$br1} == $br3)) ) { # The events on reference occur along a different branch from where the events on reconstructed occur. #

    (print STDOUT "     | (i) - B-Pass-1a-1 |\n\n") if ($TEST_LEVEL_newtype>2);

						# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
					if (($len1 == $len2) and ($len1 == $len3) and  ($len3 == $len4)) {
					    $type = 'Double-Inter-Branch-Shift';
					    $penalty = 6;
					}
					##(($len1 == $len2) and ($len1 == $len3) and  ($len3 == $len4)) and ($type = 'Double-Inter-Branch-Shift');
				    }
				}
			    }
			}

		    } # END of "if ((@indels_rltd_lf_rec ==1) and (@indels_rltd_rf_rec ==1) and (@indels_rltd_lf_ref ==1) and (@indels_rltd_rf_ref ==1)) {...} else {...}"

					# MODIFIED in REVISION on Oct 21 (Wed), 2015. #
		    unless (defined $type) {
			$type = 'Complex(???)';
			$penalty = 100;
		    }
		    ##(defined $type) or ($type = 'Complex(???)');


		} else {

			# Substep (x): If the subject comes up to here, Regard the associated error as 'Complex'. #

    (print STDOUT "  { Substep (x) }\n\n") if ($TEST_LEVEL_newtype>1);

		    $type = 'Complex';
		    $penalty = 100; # ADDED in REVISION on Oct 21 (Wed), 2015. #

		}

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
			# => Further MODIFIED in REVISION on Oct 21 (Wed), 2015. #

		return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty);
		##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

		##	# Finally, register the association between the composite-block and indel events considered to be involved. #
		##
		##unless ( register_association_cblk_and_indels ($type, $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @indels_invlvd_ref, @indels_invlvd_rec, @indices_indel_identifiers) ) {
		##	print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED at the Step 5 on the $cb th composite shift-block for the $s th segment !\n";
		##	print STDERR "[ERROR '$program_name'] $!\n\n";
		##	die ;
		##}
		##
		##next;

	    } # END of "if (($ct_rltd_indels_rec == 2) and ($ct_rltd_indels_ref == 2)) {...}"


			# Remainder: Otherwise, we will regard the associated error as 'Complex'. #

    (print STDOUT "< Step Z >\n\n") if ($TEST_LEVEL_newtype>0);

	$type = 'Complex'; # The type of MSA error(s). # ADDED for DEBUGGING on Oct 21 (Wed), 2015. #
	my $penalty = 100; # The value of penalty incurred by the error. # ADDED in REVISION on Oct 21 (Wed), 2015. #

			# MODIFIED in creation of 'classify_error_assoc_w_sgl_cblk' (on Oct 6 (Tue), 2015). #
			# => Further MODIFIED in REVISION on Oct 21 (Wed), 2015. #

	return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec, $penalty);
	##return ($type, $reason_skipped, \@set_cts_invlvd_indels, \@indels_invlvd_ref, \@indels_invlvd_rec);

	    ##unless ( register_association_cblk_and_indels ('Complex', $cb, @cblk_wise_msa_errors, %indel_ref2assoc_cblks, %indel_rec2assoc_cblks, @indels_invlvd_ref, @indels_invlvd_rec, @indices_indel_identifiers) ) {
		##print STDOUT "[ERROR '$program_name'] Subroutine 'register_association_cblk_and_indels' FAILED at the 'Remainder' step on the $cb th composite shift-block for the $s th segment !\n";
		##print STDERR "[ERROR '$program_name'] $!\n\n";
		##die ;
	    ##}

}	# END of the subroutine 'classify_error_assoc_w_sgl_cblk' #



	### DEFINITIONS of the subroutines ADDED in creation of 'classify_msa_errors.1.pl'. ###

# cal_init_psm_cands_for_twin_msas:
#	This subroutine calculates the initial candidate of the parsimonious scenario(s) for each gap-block cluster (i.e., gapped segment)
#	of the reference and reconstructed MSAs,
#	it also outputs the (non-redundant) list of gap-patterns and the informaion on their Dollo parsimonious scenarios.
#	NOTE that the subroutine eats up the input MSAs.
#	(CREATED on Oct 19 (Sun), 2014)
#
# INPUT: (\%node2ch, \%node2pa, \%node2depth, \%node2ct_exoffs, \%name2id, \@seqnames, \@in_msa_ref, \@in_msa_rec, $char_abs, $char_pres), where
#	%node2ch = ($node_id => \@child_node_ids, ...) is for the all interior nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) is for all the nodes except the top node;
#
#	%node2depth = ($node_id => $absolute_depth, ...) is for all the nodes;
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) is for all the nodes;
#
#	%name2id = ($node_name => $node_id, ...) is for the exterior nodes and the top node (with $node_name = $TOP_NODE);
#	$seqnames[$k] is the name of the $k th sequence in the input MSA (reference/reconstructed)
#	(NOTE that these names must exactly match those used in %name2id);
#
#	${$in_msa_ref/rec[$k]} is the $k th sequence embedded in the input reference/reconstructed MSA;
#
#	$char_pres/$char_abs is the character (or symbol) representing the 'presence'/'absence' state;
#
# OUTPUT: (\@info_init_psm_scn_ref, \@info_init_psm_scn_rec, \@info_gpatterns), where
#
#	@info_init_psm_scn_ref/rec = (\@set_cls_gblocks_ref/rec, \@init_totcts_indels_for_cls_ref/rec, \@init_psm_scn_for_cls_ref/rec) gives information on the initial candidate of the parsimonious scenario giving rise to the reference/reconstructed MSA gap configuration, where
#	
#	@{$set_cls_gblocks_ref/rec[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$init_totcts_indels_for_cls_ref/rec[$cl] is the total number of insertion/deletion events in the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA,
#	
#	@{$init_psm_scn_for_cls_ref/rec[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA, where 
#		%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	@info_gpatterns =  (\@set_gpatterns, \@set_node2dparsstat, \@set_br2change) gives information on the gap-patterns found in the reference & reconstructed MSAs, and the Dollo parsimonious scenarios yielding these gap-patterns, where
#
#	${$set_gpatterns[$indx_gp]} is the string representing the $ndx_gp th gap-pattern in the input MSA, created by concatenating $CHAR_PRES/ABS in the sequences;
#	%{$set_node2dparsstat[$indx_gp]} = ($node_id => [$state, $ct_pres_ext], ...) is for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern;
#	%{$set_br2change[$indx_gp]} = ($branch_id => [$state_ue, $state_le], ...) for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern.
#
sub cal_init_psm_cands_for_twin_msas (\%\%\%\%\%\@\@\@$$) {

	my ($node2ch, $node2pa, $node2depth, $node2ct_exoffs, $name2id, $seqnames, $in_msa_ref, $in_msa_rec, $char_abs, $char_pres) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ct_seqs = @{$seqnames};

    (print STDOUT "<< Subroutine 'cal_init_psm_cand_for_twin_msas' (Top node ID: $topnode_id\; #{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL_cipc>0);


		# Main Process 1: Extract gapped segments and their constituent gap-pattern blocks from the reference & reconstructed MSAs. #
	
	my %gpattern2indx;
	my $ct_gpatterns = 0;

	my $set_cls_gblocks_ref = extract_cls_gblocks (@{$in_msa_ref}, $char_pres, $char_abs, %gpattern2indx, $ct_gpatterns) ;
	unless (defined $set_cls_gblocks_ref) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_twin_msas'] The subroutine 'extract_cls_gblocks' FAILED on the reference MSA: $! !\n\n";
	    return (undef, undef, undef);
	}
	my $ct_cls_ref = @{$set_cls_gblocks_ref};
    (print STDOUT "\n --- 'extract_cls_gblocks' Ended on the reference MSA (#{gap clusters} = $ct_cls_ref) ...\n\n") if ($TEST_LEVEL_cipc>1);

	my $set_cls_gblocks_rec = extract_cls_gblocks (@{$in_msa_rec}, $char_pres, $char_abs, %gpattern2indx, $ct_gpatterns) ;
	unless (defined $set_cls_gblocks_rec) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_twin_msas'] The subroutine 'extract_cls_gblocks' FAILED on the reconstructed MSA: $! !\n\n";
	    return (undef, undef, undef);
	}
	my $ct_cls_rec = @{$set_cls_gblocks_rec};
    (print STDOUT "\n --- 'extract_cls_gblocks' Ended on the reconstructed MSA (#{gap clusters} = $ct_cls_rec) ...\n\n") if ($TEST_LEVEL_cipc>1);

	# substr (${$sec_msa[$k]}, $indx_gp, 1) is the presence/absence state of the $k th sequence in the $indx_gp th gap pattern.
	# ${$set_gpatterns[$indx_gp]} is the string made by concatenating the states in the $indx_gp th gap pattern.
	#
	my ($sec_msa, $set_gpatterns) = msa_rep_gpatterns (%gpattern2indx, length ($char_abs), $ct_seqs);
	unless (defined $sec_msa) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_twin_msas'] The subroutine 'msa_rep_gpatterns' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
    (print STDOUT "\n --- 'msa_rep_gpatterns' Ended ...\n\n") if ($TEST_LEVEL_cipc>1);


		# Scan the information of gapped segments for screening those with gapped segment(s) and/or gap(s) that are TOO LONG. #

	foreach my $cls_gblk (@{$set_cls_gblocks_ref}, @{$set_cls_gblocks_rec}) { # Screening foreach-loop (over gapped segments in reference & reconstructed MSAs). #

	    my $ct_gblks = @{$cls_gblk};
	    my $start_cls = $cls_gblk->[0][2];
	    my $end_cls = $cls_gblk->[$ct_gblks-1][3];
	    my $cl_len = $end_cls - $start_cls+1;

	    unless ($cl_len <= $UPPER_BOUND_ANAL_LEN_GBCLS) {
		print STDOUT "[WARNING 'cal_init_psm_cands_for_twin_msas'] Found a cluster of length $cl_len, which is LONGER THAN the upper-bound, $UPPER_BOUND_ANAL_LEN_GBCLS... I will skip analyzing this pair of reference & reconstructed MSAs: $! !\n\n";
		return ([], [], []);
	    }

	    my $list_long_gaps = scan_gbcls_for_long_gaps (@{$cls_gblk}, @{$set_gpatterns}, $UPPER_BOUND_ANAL_GAPLEN, $char_abs);
	    unless (defined $list_long_gaps) {
		print STDOUT "[ERROR 'cal_init_psm_cands_for_twin_msas'] The subroutine 'scan_gbcls_for_long_gaps' FAILED: $! !\n\n";
		return (undef, undef, undef);
	    }
	    unless (@{$list_long_gaps} == 0) {
		my $ct_long_gaps = @{$list_long_gaps};
		print STDOUT "[WARNING 'cal_init_psm_cands_for_twin_msas'] Found $ct_long_gaps gaps longer than $UPPER_BOUND_ANAL_GAPLEN. So, I will skip analyzing this pair of reference & reconstructed MSAs: $! !\n\n"; 
		return ([], [], []);
	    }

	}		# End of the screening foreach-loop (over gapped segments in reference & reconstructed MSAs). #


		## Main Process 2: Calculate the Dollo parsimonious indel scenario for each gap-pattern. ##

	my ($set_node2dparsstat, $set_br2change) = map_parsim_states_dollo_msa ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$sec_msa}, $char_abs, $char_pres, 1) ;
	unless (defined $set_node2dparsstat) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_twin_msas'] The subroutine 'map_parsim_states_dollo_msa' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
    (print STDOUT " --- 'map_parsim_states_dollo_msa' Ended ...\n\n") if ($TEST_LEVEL_cipc>1);

	my @info_gpatterns = ($set_gpatterns, $set_node2dparsstat, $set_br2change);


		## Main Process 3: Calculate the initial candidate of the parsimonious indel scenario resulting in the gap-configurations of each gapped segment in the reference or reconstructed MSA. ##

			# Create a set of all nodes.
	my @set_all_nodes = sort keys %{$node2pa};
	unshift @set_all_nodes, $topnode_id;

			# Initialize the outputs. #
	my @info_init_psm_scn_ref = ($set_cls_gblocks_ref, [], []);
	my @info_init_psm_scn_rec = ($set_cls_gblocks_rec, [], []);

		# @info_init_psm_scn_ref/rec = (\@set_cls_gblocks_ref/rec, \@init_totcts_indels_for_cls_ref/rec, \@init_psm_scn_for_cls_ref/rec) gives information on the initial candidate of the parsimonious scenario giving rise to the reference/reconstructed MSA gap configuration, where

	my $indx_indx_gp = 0; # Index to specify where the index of the gap-pattern is stored.

	foreach my $info_in_out (
		['(i) Reference MSA', \@info_init_psm_scn_ref], 
		['(ii) Reconstructed MSA', \@info_init_psm_scn_rec] ) { # Main outer foreach-loop (for reference & reconstructed MSAs). #

	    my ($title_msa, $info_init_psm_scn) = @{$info_in_out};

	    my ($set_cls_gblocks, $init_totcts_indels_for_cls, $init_psm_scn_for_cls) = @{$info_init_psm_scn};

	    my $ct_cls = @{$set_cls_gblocks};

	    for (my $cl = 0; $cl<$ct_cls; $cl++) { # Main inner for-loop (over clusters (i.e., gapped segments) in each MSA. #

		my $blocks_in_cls = $set_cls_gblocks->[$cl];

		my ($ct_indels0, $node2blwise_gstat0, $br2indels0, $list_blank_blks) = create_initial_set_indels_spt_odr ($topnode_id, %{$node2pa}, %{$node2ch}, @set_all_nodes, $char_pres, $char_abs, @{$blocks_in_cls}, $indx_indx_gp, @{$set_node2dparsstat}, @{$set_br2change});
		unless (defined $ct_indels0) {
		    print STDOUT "[ERROR 'cal_init_psm_cands_for_twin_msas'] The subroutine 'create_initial_set_indels' FAILED on the $cl th cluster of '$title_msa': $! !\n\n";
		    return (undef, undef, undef);
		}
    (print STDOUT "\n --- Subroutine 'create_initial_set_indels_spt_odr' Ended (with #{initial indels} = $ct_indels0) ...\n\n") if ($TEST_LEVEL_cipc>1);

		$init_totcts_indels_for_cls->[$cl] = $ct_indels0;
		$init_psm_scn_for_cls->[$cl] = [$br2indels0, $node2blwise_gstat0];

	# @{$init_psm_scn_for_cls_ref/rec[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA, where 

	    }				# End of the main inner for-loop (over clusters (i.e., gapped segments) in each MSA. #

	}					# End of the main outer foreach-loop (for reference & reconstructed MSAs). #

    (print STDOUT "<< END of the subroutine 'cal_init_psm_cand_for_twin_msas' />>\n\n") if ($TEST_LEVEL_cipc>0);

	return (\@info_init_psm_scn_ref, \@info_init_psm_scn_rec, \@info_gpatterns);

}	# End of the subroutine 'cal_init_psm_cands_for_twin_msas'. #


# reorganize_set_indels:
#	This subroutine receives a raw set of gap-pattern blocks and a candidate of the parsimonious indel scenario(s) for a MSA,
#	and returns a set of them reorganized to ease the search for a set of indel events with a particular identifier (either branch ID, beginning site, or ending site).
#	(CREATED on Oct 24 (Fri), 2014.)
#
#	[MODIFIED on Nov 4 (Tue), 2014.]
#	The output format somewhat changed. Now, the information on each indel event includes the length of the event.
#
# INPUT: (\@set_cls_gblocks, \@cand_psm_scn_for_cls), where
#
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the MSA, with
#	$indx_gp specifying the gap-pattern (stored in @set_gpatterns),
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);
#
#	@{$cand_psm_scn_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the MSA, where 
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices (for @{$set_cls_gblocks[$cl]}) of blocks involved), and where
#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#	  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node.
#
# OUTPUT: (\@linset_gblocks, \%br2all_indels, \@set_indels_std_by_begsite, \@set_indels_std_by_endsite), where
#
#	@{$linset_gblocks[$lgb]} = ($cl, $gb, $indx_gp, $len_gb, $beg_gb, $end_gb) is for the gap-pattern block in the $lgb th place in the linear set, which is the $gb th block in the $cl th cluster (i.e., gapped segment) in the MSA, where
#	$indx_gp is the index of the gap-pattern in the @{$set_gpatterns}, and $beg/end_gb is the beginning/ending coordinate of the block (in the 0-based full-closed convention);
#
#	%br2all_indels = ($branch_id => \@set_all_indels, ...), where
#	[ @{$set_all_indels[$e]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk) is for the $e th (in the spatial order) indel event along the branch, with ] --- OBSOLETE as of Nov 4 (Tue), 2014.
#	@{$set_all_indels[$e]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, $length_indel, \@indices_lin_blk) is for the $e th (in the spatial order) indel event along the branch, with # MODIFIED on Nov 4 (Tue), 2014.
#	$beg/end_site being the coordinate (in the 0-based full-closed convention) of the beginning/ending-site of the event, and
#	@indices_lin_blk storing the indices (in @linset_cls_gblocks}) of the involved gap-pattern blocks;
#
#	@{$set_indels_std_by_begsite[$i]} = ($beg_site, \@set_indels_beg_at_here) is for the $i th beginning site (in numerically ascending order), with
#	@{$set_indels_beg_at_here[$e1]} = (exactly in the same format as @{$set_all_indels[$e]}) is for the $e1 th (in the order of occurrences) indel event beginning at $beg_site;
#	@{$set_indels_std_by_endsite[$i]} = ($end_site, \@set_indels_end_at_here) is for the $i th ending site (in numerically ascending order), with
#	@{$set_indels_end_at_here[$e2]} = (exactly in the same format as @{$set_all_indels[$e]}) is for the $e2 th (in the order of occurrences) indel event ending at $end_site.
#
sub reorganize_set_indels (\@\@) {

	my ($set_cls_gblocks, $cand_psm_scn_for_cls) = @_;

	my $ct_cls = @{$set_cls_gblocks};

	my @linset_gblocks = my @set_indels_std_by_begsite = my @set_indels_std_by_endsite = ();
	my %br2all_indels;

	my $baseline = 0;

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Main outer for-loop (over clusters (i.e., gapped segments)). #

	    my $gblocks_in_cls = $set_cls_gblocks->[$cl];

	    my $ct_gblks = @{$gblocks_in_cls};

	    for (my $gb=0; $gb<$ct_gblks; $gb++) { # 1st main middle for-loop (over blocks). #

		my ($indx_gp, $len_gb, $beg_gb, $end_gb) = @{$gblocks_in_cls->[$gb]};
		$linset_gblocks[$baseline + $gb] = [$cl, $gb, $indx_gp, $len_gb, $beg_gb, $end_gb];

	    }				# End of the 1st main middle for-loop (over blocks). #

	    my %begsite2indels_in_cls;
	    my %endsite2indels_in_cls;

	    my $br2indels_in_cls = $cand_psm_scn_for_cls->[$cl][0];

	    foreach my $br (sort {$b <=> $a} keys %{$br2indels_in_cls}) { # 2nd main middle foreach-loop (over branches with indels). #

		my $list_indels = $br2indels_in_cls->{$br};
		my @lin_sblist_indels = ();

		foreach my $info_event (@{$list_indels}) { # Main inner foreach-loop (over events). #

		    my ($state_ue, $state_le, $indices_blk) = @{$info_event};

		    my ($indx_lm, $indx_rm) = @{$indices_blk}[0, $#{$indices_blk}];

		    my $beg_site = $gblocks_in_cls->[$indx_lm][2];
		    my $end_site = $gblocks_in_cls->[$indx_rm][3];

		    my @indices_lin_blk = ();

				# MODIFIED on Nov 4 (Tue), 2014. #

		    my $len_indel=0;
		    foreach my $indx (@{$indices_blk}) { # Inner-inner foreach-loop (over blocks involved in the event). #
			push @indices_lin_blk, $indx + $baseline;
			$len_indel += $gblocks_in_cls->[$indx][1];
		    }				# End of the inner-inner foreach-loop (over blocks involved in the event). #

		    my @lin_info_event = ($br, $beg_site, $end_site, $state_ue, $state_le, $len_indel, \@indices_lin_blk);

		    ##foreach my $indx (@{$indices_blk}) { push @indices_lin_blk, $indx + $baseline; }
		    ##my @lin_info_event = ($br, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk);

		    push @lin_sblist_indels, \@lin_info_event;

		    my $indels_in_cls1 = $begsite2indels_in_cls{$beg_site};
		    (defined $indels_in_cls1) or ($indels_in_cls1 = $begsite2indels_in_cls{$beg_site} = []);
		    push @{$indels_in_cls1}, \@lin_info_event;

		    my $indels_in_cls2 = $endsite2indels_in_cls{$end_site};
		    (defined $indels_in_cls2) or ($indels_in_cls2 = $endsite2indels_in_cls{$end_site} = []);
		    push @{$indels_in_cls2}, \@lin_info_event;

		}				# End of the main inner foreach-loop (over events). #

		my $all_indels = $br2all_indels{$br};
		(defined $all_indels) or ($all_indels = $br2all_indels{$br} = []);
		push @{$all_indels}, (sort { $a->[1] <=> $b->[1] } @lin_sblist_indels);

	    }						# End of the 2nd main middle foreach-loop (over branches with indels). #

	    foreach my $beg_site (sort {$a <=> $b} keys %begsite2indels_in_cls) { push @set_indels_std_by_begsite, [$beg_site, $begsite2indels_in_cls{$beg_site} ]; }

	    foreach my $end_site (sort {$a <=> $b} keys %endsite2indels_in_cls) { push @set_indels_std_by_endsite, [$end_site, $endsite2indels_in_cls{$end_site} ]; }


	    $baseline += $ct_gblks; # UPDATE the baseline. #

	}				# End of the main outer for-loop (over clusters (i.e., gapped segments)). #


	return (\@linset_gblocks, \%br2all_indels, \@set_indels_std_by_begsite, \@set_indels_std_by_endsite);

}	# End of the subroutine 'reorganize_set_indels'. #


# identify_equiv_indels:
#	This subroutine receives sets of indels resulting in two sets of MSAs (MSA 1 and 2),
#	and identifies the indel events for MSA 1 that have "equivalent" events for MSA 2.
#	(CREATED on Oct 28 (Tue), 2014)
#
#	[CRITERION for the "equivalence"]
#	The two events are considered as "equivalent" if they insert the identical set of sub-sequences (in terms of the residue numbers) (for equivalent insertions),
#		[or if the sets of "deleted" sub-sequences (in the complement of the deleted sequences) are identical (for equivalent deletions). ] --- Obsolete.
#	   or if the sets of "deleted" sub-sequences (in the sibling lineage, and, if any, in the uncle/aunt lineage, of the deleted lineage) are identical (for equivalent deletions).
#	The subroutine doesn't care about the positions they are inserted/deleted, i.e., the positions of the aligned gaps in the remaining sequences.
#
# INPUT: (\%br2all_indels1, \%br2all_indels2, \@set_clminfo1, \@set_clminfo2, $char_abs, $ct_seqs, \%node2ch, \%node2seq_indx, \%node2pa), where
#	%br2all_indels1/2 = ($branch_id => \@set_all_indels1/2, ...) is for the MSA 1/2, where
#	@{$set_all_indels1/2[$e]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk) is for the $e th (in the spatial order) indel event along the branch, with
#	$beg/end_site being the (0-based full-closed) coordinate of the beginning/ending site of the event in the MSA,
#	$state_ue/le being the state at the upper-/lower-end of the branch;
#
#	$set_clminfo1/2[$c]->[$k] is the residue number (1-based) of the $k th sequence at the $c th column of MSA 1/2.
#		it is ($r + 1/2) if the column is occupied by a gap between the $r th and the ($r+1) th residues.
#
#	$char_abs is the character representing the 'absence' state;
#
#	$ct_seqs is the number of sequences in the MSA (the two MSAs should be composed of the same sequences);
#
#	%node2ch = ($node_id => \@chilld_nodes, ...) is for the interior nodes;
#
#	%node2seq_indx = ($node_id => $seq_indx, ...) for the exterior nodes, with
#	$seq_indx specifying the index specifying the order of the sequence (or its corresponding exterior node) in the MSA;
#
#	%node2pa = ($node_id => $parent_id, ...) is for all nodes except the root.
#
# OUTPUT: (\%event1_2_event2, \%event2_2_event1), where
#
#	%event1_2_event2 = ("$br1:$beg1:$end1" => "$br2:$beg2:$end2", ...) with
#	$br1/2 being the branch along which the equivalent event in MSA 1/2 occurred,
#	$beg1/2 being the beginning site coordinate of the equivalent event in MSA 1/2,
#	$end1/2 beging the end site coordinate of the  equivalent event in MSA 1/2;
#
#	%event2_2_event1 is just the inverse of %event2_2_event1.
#
sub identify_equiv_indels (\%\%\@\@$$\%\%\%) {

	my ($br2all_indels1, $br2all_indels2, $set_clminfo1, $set_clminfo2, $char_abs, $ct_seqs, $node2ch, $node2seq_indx, $node2pa) = @_;

	my @indices_full_seqs = (0 .. $ct_seqs-1); 
	##my @indices_full_seqs = (0 .. $ct_seqs); # A BUG fixed (on Oct 28 (Tue), 2014). #

	my %event1_2_event2;
	my %event2_2_event1;

	foreach my $br (sort {$a <=> $b} keys %{$br2all_indels1}) { # Main outer foreach-loop (over branches where indels occur). #

	    my $all_indels1 = $br2all_indels1->{$br};
	   (@{$all_indels1}>0) or next;

	    my $all_indels2 = $br2all_indels2->{$br};
	   ((defined $all_indels2) and (@{$all_indels2}>0)) or next; # In this case, there can NEVER be equivalent events along this branch. #

	    my $exoffs = fetch_ext_offsprings_nr ($br, %{$node2ch});
	    my @offs_seq_indices = ();
	    foreach my $nodeid (@{$exoffs}) { push @offs_seq_indices, $node2seq_indx->{$nodeid}; }

	    @offs_seq_indices = sort {$a <=> $b} @offs_seq_indices;
	    my $cmpl_seq_indices;

	    foreach my $indel1 (@{$all_indels1}) { # Main middle foreach-loop (over indels in MSA 1 along $br). #

		my ($br1, $beg1, $end1, $state_ue1, $state_le1, $indices_blk1) = @{$indel1};

		my $label_event1 = "$br1:$beg1:$end1";

		my $resnums_beg1 = $set_clminfo1->[$beg1];
		my $resnums_end1 = $set_clminfo1->[$end1];

		my $relv_seq_indices;
		if ($state_le1 eq $char_abs) { # For a 'deletion'.

		    unless (defined $cmpl_seq_indices) {
	    		##(my $common, my $ex1, $cmpl_seq_indices) = diff_nmarray (@offs_seq_indices, @indices_full_seqs);
			my $node_pa = $node2pa->{$br};
			my $sibl;
			foreach my $node (@{$node2ch->{$node_pa}}) { # 1st preliminary inner foreach-loop (over siblings). #
			    if ($node != $br) { $sibl = $node; last; }
			} 					# End of the 1st preliminary inner foreach-loop (over siblings). #
			my $exoffs_sibl = fetch_ext_offsprings_nr ($sibl, %{$node2ch});
			my @cmpl_seq_indices = ();
			foreach my $nodeid (@{$exoffs_sibl}) { push @cmpl_seq_indices, $node2seq_indx->{$nodeid}; }
			my $node_grpa = $node2pa->{$node_pa};
			if (defined $node_grpa) {
			     my $uncle;
			     foreach my $node (@{$node2ch->{$node_grpa}}) { # 2nd preliminary inner foreach-loop (over siblings). #
				if ($node != $node_pa) { $uncle = $node; last; }
			     }					# End of the 2nd preliminary inner foreach-loop (over siblings). #
			     my $exoffs_uncle = fetch_ext_offsprings_nr ($uncle, %{$node2ch});
			     foreach my $nodeid (@{$exoffs_uncle}) { push @cmpl_seq_indices, $node2seq_indx->{$nodeid}; }
			}
			@cmpl_seq_indices = sort { $a <=> $b } @cmpl_seq_indices;
			$cmpl_seq_indices = \@cmpl_seq_indices;
		    }

		    $relv_seq_indices = $cmpl_seq_indices;

		} else {		# For an 'insertion'.

		    $relv_seq_indices = \@offs_seq_indices;
		}

		foreach my $indel2 (@{$all_indels2}) { # Main inner foreach-loop (over indels in MSA 2 along $br). #

		    my ($br2, $beg2, $end2, $state_ue2, $state_le2, $indices_blk2) = @{$indel2};

		    my $resnums_beg2 = $set_clminfo2->[$beg2];
		    my $resnums_end2 = $set_clminfo2->[$end2];

		    my $if_equiv = 1;
		    foreach my $seq_indx (@{$relv_seq_indices}) { # Innermost foreach-loop (over relevant sequences). #

			my $resnum_end1 = int ($resnums_end1->[$seq_indx]);
			my $resnum_end2 = int ($resnums_end2->[$seq_indx]);

			if ($resnum_end1 != $resnum_end2) {
			   $if_equiv = 0;
			   last;
			}

			my $resnum_beg1 = $resnums_beg1->[$seq_indx];
			($resnum_beg1 == int ($resnum_beg1)) or ($resnum_beg1 = int ($resnum_beg1) + 1 );
			my $resnum_beg2 = $resnums_beg2->[$seq_indx];
			($resnum_beg2 == int ($resnum_beg2)) or ($resnum_beg2 = int ($resnum_beg2) + 1 );

			if ($resnum_beg1 != $resnum_beg2) {
			    $if_equiv = 0;
			    last;
			}
		    }				# End of the innermost foreach-loop (over relevant sequences). #

		    if ($if_equiv) {
				# The two events were found to be "equivalent" to each other. #

			my $label_event2 = "$br2:$beg2:$end2";
			$event1_2_event2{$label_event1} = $label_event2;
			$event2_2_event1{$label_event2} = $label_event1;
			last;
		    }

		}				# End of the main inner foreach-loop (over indels in MSA 2 along $br). #

	    }				# End of the main middle foreach-loop (over indeld in MSA 1 along $br). #


	}				# End of the main outer foreach-loop (over branches where indels occur). #

	return (\%event1_2_event2, \%event2_2_event1);

}		# End of the subroutine 'identify_equiv_indels'. #


# effective_distance:
#	This satellite subroutine of 'find_ovlp_flanking_indels'
#	calculates the 'effective distance' between an indel event and a set of 'indel' events.
#	(CREATED on Oct 30 (Thu), 2014)
#
#	[DEFINITION]
#	The 'effective distance' is defined as the number of branches separating the two sets of nodes with the 'presence' state,
#	one set is defined by the indel event, and the other is defined by the set of 'indel' events.
#	If the two sets of nodes with the 'present' state overlap at some nodes, the 'effective distance' = 0.
#	If the two sets are complementary to each other (and both are singly-connected), the 'effective distance' = 1.
#	If, for example, the two sets resulted from two independent insertions along sibling branches,
#	the 'effective distance' = 2 usually, but = 1 if the sibling branches are children of the bifurcated top-node (i.e., root).
#	
# INPUT: ($br1, $state_ue1, $nodeid_mrca2, \%br2change2, \%node2pa, \%node2ch, $topnode_id, $char_abs), where
#
#	$br1 is the branch along which the 1st event occurred;
#	$state_ue1 is the state at the upper-end of $br1;
#
#	$nodeid_mrca2 is the MRCA of the nodes with the 'presence' state in the 2nd set;
#	%br2change2 = ($branch_id => [$state_ue, $state_le], ...) is for the changes ('indel' events) defining the 2nd set;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but the top-node (i.e., root);
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#	$topnode_id is the ID of the top-node (i.e., root);
#
#	$char_abs is the character representing a gap.
#
# OUTPUT: $eff_dist, which is the effective distance calculated.
#
sub effective_distance ($$$\%\%\%$$) {

	my ($br1, $state_ue1, $nodeid_mrca2, $br2change2, $node2pa, $node2ch, $topnode_id, $char_abs) = @_;

	if ($state_ue1 ne $char_abs) { # The 1st event is a deletion.

	    if ($nodeid_mrca2 == $topnode_id) {
		    # The two sets will certainly overlap. #
		return 0;
	    } elsif ($nodeid_mrca2 == $br1) {
		    # The 2nd set resulted from an 'insertion' along $br1. #
		return 1;
	    }
	} else { # The 1st event is an insertion.

	    if ($nodeid_mrca2 == $br1) {
		# The 2nd set resulted from an 'insertion' along $br1. => The two set will certainly overlap. #
		return 0;
	    }
	}

	my ($common_ancestors, $lineage1, $lineage2) = fetch_common_ancestors ($br1, $nodeid_mrca2, %{$node2pa});

	if ($state_ue1 ne $char_abs) { ## Again, the 1st event is a deletion.##

	    if (@{$lineage1}>0) {
		    # MRCA2 is also an ancestor of $br1. => The two set will overlap.
		return 0;
	    } else {
		   # $br1 is an ancestor of MRCA2. => The two sets will be separated by a series of branches leading to MRCA2.

		my $eff_dist = 1 + scalar (@{$lineage2});
		return $eff_dist;
	    }
	}

		## The 1st event is an insertion. ##

	if (@{$lineage1} == 0) {
		# MRCA2 is a 'descendant' of $br1. => The two sets will certainly overlap!!
	    return 0;
	}

	if (@{$lineage2}>0) {
		# MRCA2 and $br1 are on separate lineages. #
	    my $eff_dist = scalar (@{$lineage1}) + scalar (@{$lineage2});

	    if ((@{$common_ancestors}==1) and (@{$node2ch->{$topnode_id}}==2)) {
			# When the MRCA of MRCA2 and $br1 is the top-node with only 2 children, the two branches at the top will be regarded as a single branch. #
		$eff_dist--;
	    }

	    return $eff_dist;
	}

		# MRCA2 is an ancestor of $br1. #

	my $ct_lin1 = @{$lineage1};

	for (my $k=0; $k<$ct_lin1; $k++) { # Main for-loop (over descendant lineage of MRCA2 down to $br1). #
	    (defined $br2change2->{$lineage1->[0]}) and last; # A 'deletion' occurred. => The remaining branches separate the two sets. #
	    shift @{$lineage1};
	}			# End of the main for-loop (over descendant lineage of MRCA2 down to $br1). #

	my $eff_dist = @{$lineage1};

	return $eff_dist;


}	# End of the subroutine 'effective_distance'. #


	    # ($ct_flnk_ovlp_indels_rec, $indels_lf_rec, $indels_rf_rec, $indels_le_rec, $indels_re_rec, $indels_other_ovlp_rec) = find_ovlp_flanking_indels ($beg_cb_rec, $end_cb_rec, @{$set_indels_std_by_begsite_rec}, @{$set_indels_std_by_endsite_rec}, %{$rec_event2ref_event}, $nodeid_mrca_cb, %{$br2change_cb}, %{$node2pa}, %{$node2ch}, $topnode_id, $CHAR_ABS);

# find_ovlp_flanking_indels:
#	This subroutine searches for indel events that either flank or begin or end within a specified interval.
#	But those events with equivalent events in the other MSA will be excluded.
#	(CREATED in creation of 'classify_msa_errors_via_mblks.1.pl' on Oct 29 (Wed), 2014.)
#
#	[MODIFIED on Nov 5 (Wed), 2014]:
#	The elongation condition of the flanking events now takes account of the contiguity of the current and the next events.
#
# INPUT: ($beg_intvl, $end_intvl, \@set_indels_std_by_begsite, \@set_indels_std_by_endsite, \%event2equiv_event, $nodeid_mrca_cb, \%br2change_cb, \%node2pa, \%node2ch, $topnode_id, $char_abs), where
#
#	$beg/end_intvl is the coordinate of the beginning/ending site of the interval (in the 0-based full-closed convention);
#	
#	@{$set_indels_std_by_begsite_ref[$i]} = ($beg_site, \@set_indels_beg_at_here) is for the $i th beginning site (in numerically ascending order), with
#	[ @{$set_indels_beg_at_here[$e1]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk) ] --- OBSOLETE as of Nov 4 (Tue), 2014.
#	@{$set_indels_beg_at_here[$e1]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, $length_event, \@indices_lin_blk)  # MODIFIED as of Nov 4 (Tue), 2014.
#		is for the $e1 th (in the order of occurrences) indel event beginning at $beg_site;
#
#	@{$set_indels_std_by_endsite_ref[$i]} = ($end_site, \@set_indels_end_at_here) is for the $i th ending site (in numerically ascending order), with
#	@{$set_indels_end_at_here[$e2]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $e2 th (in the order of occurrences) indel event ending at $end_site.
#
#	%event2equiv_event = ("$br_id:$beg_site:$end_site" => "$br_id_eq:$beg_site_eq:$end_site_eq", ....), with
#	$br_id/$beg_site/$end_site being the branch ID/beginning site/ending site of the indel event, and
#	$br_id_eq/$beg_site_eq/$end_site_eq being the branch ID/beginning site/ending site of the equivalent indel event;
#
#	$nodeid_mrca_cb is the MRCA of the sequences involving the composite-block;
#
#	%br2change_cb = ($branch_id => [$state_ue, $state_le], ...) is for the branches separating sequences involving the composite-block from the rest;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but the top-node (i.e., root);
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#	$topnode_id is the node ID of the top-node (i.e., root);
#
#	$char_abs is the character representing the 'absence' state.
#
# OUTPUT: ($ct_flnk_ovlp_indels, \@indels_lf, \@indels_rf, \@indels_le, \@indels_re, \@indels_other_ovlp), where
#
#	$ct_flnk_ovlp_indels is the total number of indel events that either flank or begin/end within the interval;
#
#	@{$indels_lf[$lf]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $lf th indel event left-flanking the interval, i.e., ending exactly at ($beg_intvl-1);
#
#	@{$indels_rf[$rf]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $rf th indel event right-flanking the interval, i.e., beginning exactly at ($end_intvl+1);
#
#	@{$indels_le[$le]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $le th indel event beginning exactly at $beg_intvl;
#
#	@{$indels_re[$re]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $re th indel event ending exactly at $end_intvl;
#
#	@{$indels_other_ovlp[$ot]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $ot th other indel event beginning or ending within the interval.
#
#	NOTE that the first 4 sets can overlap with each other, whereas the last one (@indels_other_ovlp) does not overlap with the others.
#
sub find_ovlp_flanking_indels ($$\@\@\%$\%\%\%$$) {

	my ($beg_intvl, $end_intvl, $set_indels_std_by_begsite, $set_indels_std_by_endsite, $event2equiv_event, $nodeid_mrca_cb, $br2change_cb, $node2pa, $node2ch, $topnode_id, $char_abs) = @_;

	my $ct_begs = @{$set_indels_std_by_begsite};

	unless ($ct_begs>0) { # There is NO indel events anyway. #

	    return (0, [], [], [], [], []);
	}

	my $ct_ends = @{$set_indels_std_by_endsite};

	my @full_indices_beg = (0 .. $ct_begs -1);
	my @full_indices_end = (0 .. $ct_ends -1);

	my %event2if_ovlp; # = ("$branch_id:$beg_site:$end_site" => 1, ...) for the events either lanking the interval or beginning/ending within the interval.

		## First, look for events flanking the interval. ##
	my @indels_lf = my @indels_rf = ();

	my ($fl_lf, $indx2_lf, $indx_lf) = binsearch_num ($beg_intvl-1, @{$set_indels_std_by_endsite}, 0, @full_indices_end);
	if ($fl_lf) {
	    my $indx_eff_lf = $indx_lf;
	    my ($eff_dist, $set_sbj_indels);
	    for (; $indx_eff_lf >=0; $indx_eff_lf--) { # Preliminary for-loop (leftward down to the left-most endsite). #

			# Basically, skip a set of events that the composite-block can 'freely' path through. #

		$set_sbj_indels = $set_indels_std_by_endsite->[$indx_eff_lf][1];
		(@{$set_sbj_indels}>1) and last;

			# MODIFIED on Nov 5 (Wed), 2014. #
		my ($br, $beg_site, $state_ue) = @{$set_sbj_indels->[0]}[0, 1, 3];
		##my ($br, $state_ue) = @{$set_sbj_indels->[0]}[0, 3];
		$eff_dist = effective_distance ($br, $state_ue, $nodeid_mrca_cb, %{$br2change_cb}, %{$node2pa}, %{$node2ch}, $topnode_id, $char_abs);

		($eff_dist<3) and last;

		(($indx_eff_lf == 0) or ($set_indels_std_by_endsite->[$indx_eff_lf-1][0] < $beg_site-1)) and last; # If there is no contiguous events on the left. (ADDED on Nov 5 (Wed), 2014.)

	    }				# End of the preliminary for-loop (leftward down to the left-most endsite). #

	    (($indx_eff_lf >= 0) and ((@{$set_sbj_indels}>1) or ($eff_dist<3))) or ($indx_eff_lf = $indx_lf); # If no decent set of indels cannot be found, use the literally flanking indels. (ADDED on Nov 5 (Wed), 2014.)
	    
	    if ($indx_eff_lf >=0) {
		foreach my $indel_info (@{$set_indels_std_by_endsite->[$indx_eff_lf][1]}) { # 1st main foreach-loop (over left-flanking events). #
		    my $label = join (':', @{$indel_info}[0,1,2]);
		    (defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #

		    push @indels_lf, $indel_info;

		    $event2if_ovlp{$label} = 1;
		}				# End of the 1st main foreach-loop (over left-flanking events). #
	    }
		# MODIFIED to capture all events ending on the right of the 'left-flanking' events (on Nov 5 (WEd), 2014) 
	    $indx_lf = $indx_eff_lf;
	}
	$indx_lf++; # Increment the index by 1 to make the event-ending-site(s) on the right of $beg_intvl-1.

	my ($fl_rf, $indx2_rf, $indx_rf) = binsearch_num ($end_intvl+1, @{$set_indels_std_by_begsite}, 0, @full_indices_beg);
	if ($fl_rf) {
	    my $indx_eff_rf = $indx_rf;
	    my ($set_sbj_indels, $eff_dist);
	    for (; $indx_eff_rf < $ct_begs; $indx_eff_rf++) { # Preliminary for-loop (rightward up to the right-most beginning-site). #
			# Basically, skip a set of events that the composite-block can 'freely' path through. #

		$set_sbj_indels = $set_indels_std_by_begsite->[$indx_eff_rf][1];
		(@{$set_sbj_indels}>1) and last;

			# MODIFIED on Nov 5 (Wed), 2014. #
		my ($br, $end_site, $state_ue) = @{$set_sbj_indels->[0]}[0, 2, 3];
		##my ($br, $state_ue) = @{$set_sbj_indels->[0]}[0, 3];
		$eff_dist = effective_distance ($br, $state_ue, $nodeid_mrca_cb, %{$br2change_cb}, %{$node2pa}, %{$node2ch}, $topnode_id, $char_abs);

		($eff_dist<3) and last;

		(($indx_eff_rf == $ct_begs-1) or ($end_site + 1 < $set_indels_std_by_begsite->[$indx_eff_rf+1][0])) and last; # If there is no contiguous events on the right. (ADDED on Nov 5 (Wed), 2014.)

	    }				# End of the preliminary for-loop (rightward up to the right-most beginning-site). #

	    (($indx_eff_rf < $ct_begs) and ((@{$set_sbj_indels}>1) or ($eff_dist<3))) or ($indx_eff_rf = $indx_rf); # If no decent set of indels cannot be found, use the literally flanking indels. (ADDED on Nov 5 (Wed), 2014.)

	    if ($indx_eff_rf < $ct_begs) {
		foreach my $indel_info (@{$set_indels_std_by_begsite->[$indx_eff_rf][1]}) { # 2nd main foreach-loop  (over right-flanking events). #
		    my $label = join (':', @{$indel_info}[0,1,2]);
		    (defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #

		    push @indels_rf, $indel_info;

		    $event2if_ovlp{$label} = 1;
		}				# End of the 2nd main foreach-loop (over right-flanking events). #
	    }

		# MODIFIED to capture all events beginning on the left of the 'right-flanking' events (on Nov 5 (WEd), 2014) 
	    $indx_rf = $indx_eff_rf -1; 
	    ##$indx_rf--; # Decrement the index by 1 to make the event-beginning-site(s) on the left of $end_intvl+1.
	}


		## Second, look for events beginning/ending right at the beginning/ending site. ##
	my @indels_le = my @indels_re = ();

	my ($fl_le, $indx2_le, $indx_le) = binsearch_num ($beg_intvl, @{$set_indels_std_by_begsite}, 0, @full_indices_beg);
	if ($fl_le) {
	    foreach my $indel_info (@{$set_indels_std_by_begsite->[$indx_le][1]}) { # 3rd main foreach-loop (over left-ending events). #
		my $label = join (':', @{$indel_info}[0,1,2]);
		(defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #

		push @indels_le, $indel_info;

		$event2if_ovlp{$label} = 1;
	    }				# End of the 3rd main foreach-loop (over left-ending events). #
	}
	$indx_le++; # Increment the indx by 1 to make the event-beginning-site(s) on the right of $beg_intvl.

	my ($fl_re, $indx2_re, $indx_re) = binsearch_num ($end_intvl, @{$set_indels_std_by_endsite}, 0, @full_indices_end);
	if ($fl_re) {
	    foreach my $indel_info (@{$set_indels_std_by_endsite->[$indx_re][1]}) { # 4th main foreach-loop (over right-ending events). #
		my $label = join (':', @{$indel_info}[0,1,2]);
		(defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #

		push @indels_re, $indel_info;

		$event2if_ovlp{$label} = 1;
	    }				# End of the 4th main foreach-loop (over right-ending events). #

	    $indx_re--; # Decrement the index by 1 to make the event-ending-site(s) on the left of $end_intvl.
	}

		## Third, look for other events beginning/ending within the interval. ##
	my @indels_other_ovlp = ();

			# First, those beginning within the interval.

	for (my $i=$indx_le; $i<=$indx_rf; $i++) { # 5th main foreach-loop (over events beginning within the interval). #
	    foreach my $indel_info (@{$set_indels_std_by_begsite->[$i][1]}) { # Inner foreach-loop. #
		my $label = join (':', @{$indel_info}[0,1,2]);
		(defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #
		(defined $event2if_ovlp{$label}) and next; # Skip if it's already in any of the lists.

		push @indels_other_ovlp, $indel_info;

		$event2if_ovlp{$label} = 1;
	    }						# End of the inner foreach-loop. #
	}				# End of the 5th main foreach-loop (over events beginning within the interval). #

			# Then, those ending within the interval. 

	for (my $i=$indx_lf; $i<=$indx_re; $i++) { # 6th main foreach-loop (over events ending within the interval). #
	    foreach my $indel_info (@{$set_indels_std_by_endsite->[$i][1]}) { # Inner foreach-loop. #
		my $label = join (':', @{$indel_info}[0,1,2]);
		(defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #
		(defined $event2if_ovlp{$label}) and next; # Skip if it's already in any of the lists.

		push @indels_other_ovlp, $indel_info;

		$event2if_ovlp{$label} = 1;
	    }						# End of the inner foreach-loop. #
	}				# End of the 6th main foreach-loop (over events ending within the interval). #


		## Finally, Output the results. ##

	my $ct_flnk_ovlp_indels = scalar (keys %event2if_ovlp);

	return ($ct_flnk_ovlp_indels, \@indels_lf, \@indels_rf, \@indels_le, \@indels_re, \@indels_other_ovlp);

}		# End of the subroutine 'find_ovlp_flanking_indels'. #


# register_association_cblk_and_indels:
#	This subroutine registers the association relationships between a composite block of position-shifts and
#	indel events in the reference & reconstructed MSAs.
#	(CREATED on Oct 31 (Fri), 2014.)
# 
# INPUT: ($assoc_type, $indx_cb, \@cblk_wise_msa_errors, \%indel_ref2assoc_cblks, \%indel_rec2assoc_cblks, \@indels_assoc_ref, \@indels_assoc_rec, \@indices_indel_identifiers), where
#
#	$assoc_type is a string representing the type of association between the composite block and indel events;
#
#	$indx_cb is the index specifying the order of the subject composite-block;
#
#	@{$cblk_wise_msa_errors[$indx_cb]->[$e]} = ($assoc_type, \@indels_assoc_ref, \@indels_assoc_rec) (to be planned)
#		is for the $e th errors associated with the $indx_cb th composite block;
#
#	%indel_ref/rec2assoc_cblks = ("$br_id:$beg_site:$end_site" => \@cblks, ...) is for the indels in the reference/reconstructed MSA, with
#	$br_id/$beg_site/$end_site being the branch ID/beginning site/ending site of the indel event, and
#	@cblks listing the indices of blocks associated with the event;
#
#	@indels_assoc_ref/rec stores information on the reference/reconstructed indel events associated with the $indx_cb th composite block,
#	and @indices_indel_identifiers gives the indices of identifiers. More precisely,
#
#	@{$indels_assoc_ref/rec[$e]}[@indices_indel_identifiers] = ($br_id, $beg_site, $end_site) uniquely specify the $e th associated event in the reference/reconstructed MSA.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub register_association_cblk_and_indels ($$\@\%\%\@\@\@) {

	my ($assoc_type, $indx_cb, $cblk_wise_msa_errors, $indel_ref2assoc_cblks, $indel_rec2assoc_cblks, $indels_assoc_ref, $indels_assoc_rec, $indices_indel_identifiers) = @_;


		# Build up @{$cblk_wise_msa_errors}. #

	my $set_assoc_errors = $cblk_wise_msa_errors->[$indx_cb];
	(defined $set_assoc_errors) or ($set_assoc_errors = $cblk_wise_msa_errors->[$indx_cb] = []);
	push @{$set_assoc_errors}, [$assoc_type, $indels_assoc_ref, $indels_assoc_rec] ;


		# Build up %{$indel_ref_asoc_cblks}. #

	foreach my $event_info (@{$indels_assoc_ref}) { # 1st main foreach-loop (over associated indels in reference MSA). #

	    my $label_ref = join (':', @{$event_info}[@{$indices_indel_identifiers}]);
	    my $assoc_cblks_ref = $indel_ref2assoc_cblks->{$label_ref};
	    (defined $assoc_cblks_ref) or ($assoc_cblks_ref = $indel_ref2assoc_cblks->{$label_ref} = []);
	    push @{$assoc_cblks_ref}, $indx_cb;

	}				# End of the 1st main foreach-loop (over involved indels in reference MSA). #


		# Build up %{$indel_rec_asoc_cblks}. #

	foreach my $event_info (@{$indels_assoc_rec}) { # 2nd trailing inner foreach-loop (over involved indels in reconstructed MSA). #

		my $label_rec = join (':', @{$event_info}[@{$indices_indel_identifiers}]);
		my $assoc_cblks_rec = $indel_rec2assoc_cblks->{$label_rec};
		(defined $assoc_cblks_rec) or ($assoc_cblks_rec = $indel_rec2assoc_cblks->{$label_rec} = []);
		push @{$assoc_cblks_rec}, $indx_cb;

	}				# End of the 2nd trailing inner foreach-loop (over involved indels in reconstructed MSA). #

	return 1;

}	# End of the subroutine 'register_association_cblk_and_indels'. #


# if_collapse_indep_ins:
#	This subroutine examines whether the gap configuration of the input reconstructed MSA is likely due to
#	a 'Collapse' of independent insertions in the reference MSA.
#	(CREATED on Nov 4 (Tue), 2014.)
#	(=> MODIFIED on Nov 10 (Mon), 2014, for DEBUGGING.)
#
#	[MODIFIED on Oct 4 (Sun), 2015]
#	The subroutine now regards the error as (i)CII even if 1 spuious deletion is missing,
#	as long as it involves 2 or more spurious indels are involved.
#
# INPUT: (\@indels_ref, \@indels_rec, \%node2pa, \%node2ch, $char_abs), where
#
#	@{$indels_ref[$f]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, $length_indel, \@indices_lin_blk) is for the $f th indel event among those giving the reference MSA;
#	@{$indels_rec[$c]}, which follows the same format as @{$indels_ref[$f]}, is for the $c th indel event among those giving the reconstructed MSA;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but the top-node (i.e., root);
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#
#	$char_abs is a character representing an 'absence' state.
#
# OUTPUT: ($if_cii, \@indels_invlvd_rec), where
#	$if_cii = 1 if the input is likely explained by a 'Collapse of Independent Insertions' (CII),
#		= 0 otherwise, but if the subroutine successfully ended,
#		= undef if the subroutine failed.
#
#	@indels_invlvd_rec, which is basically the subset of @indels_rec, stores only those events
#	that were likely involved in the CII.
#
sub if_collapse_indep_ins (\@\@\%\%$) {

	my ($indels_ref, $indels_rec, $node2pa, $node2ch, $char_abs) = @_;

	unless (@{$indels_ref}==2) {
	    my $ct_indels_ref = @{$indels_ref};
	    print STDOUT "[ERROR 'if_collapse_indep_ins'] ILLEGAL INPUT: The number of indel events in the reference MSA is $ct_indels_ref, not 2 as expected: $! !\n\n";
	    return (undef, undef);
	}

	my ($br1, $state_ue1) = @{$indels_ref->[0]}[0, 3];
	my ($br2, $state_ue2) = @{$indels_ref->[1]}[0, 3];

	if (($state_ue1 ne $char_abs) and ($state_ue2 ne $char_abs)) {
		# Both events are deletions. Thus, a CII cannot occur.
	    return (0, []);
	}

	if ($br1 == $br2) {
		# A CII cannot occur. Even if the events collapsed, it should be called a 'purge'. #
	    return (0, []);
	}

	my ($common_anc, $lineage1, $lineage2) = fetch_common_ancestors ($br1, $br2, %{$node2pa});

	my @to_be_examined;
	my $mrca;
	if ($state_ue1 ne $char_abs) {
		# The 1st event is a deletion. => $br1 must be an ancestor of $br2.

	    if (@{$lineage1}>0) { # The two events can never collapse into the same block.
		return (0, []);
	    }
	    @to_be_examined = @{$lineage2};

	} elsif ($state_ue2 ne $char_abs) {
		# The 2nd event is a deletion. => $br2 must be an ancestor of $br1.

	   if (@{$lineage2}>0) { # The two events can never collapse into the same block.
		return (0, []);
	   }
	   @to_be_examined = @{$lineage1};

	} else {
		# The both events are insertions. => $br1 and $br2 must diverge into different lineages.

	    unless ((@{$lineage1}>0) and (@{$lineage2}>0)) {

			# MODIFIED for DEBUGGING on Nov 10 (Mon), 2014. #
			# (It turns out that, although rare, this case could actually occur.
			# ... In that case, the two events (that were examined) can never collapse into the same block.)
			#
		return (0, []);
		##print STDOUT "[ERROR 'if_collapse_indep_ins'] IMPOSSIBLE!!... The two events for reference MSA are insertions, yet the branches they occur does NOT seem to diverge into different lineages: $! !!\n\n";
		##return (undef, undef);
	    }

	   (@{$common_anc}>1) and ($mrca = $common_anc->[$#{$common_anc}]); # NOTE: $mrca is undef if MRCA is the top node (i.e., the root). #

	   shift @{$lineage1}; # The 1st branch after divergence must be removed.
	   shift @{$lineage2}; # Same here.
	   @to_be_examined = (@{$lineage1}, @{$lineage2});
	}

		# Preliminary process. #

	my %br2rec_event_info; # = ($branch_id => \@event_info, ...) for all events on the reconstructed MSA. #
	foreach my $event_info (@{$indels_rec}) { $br2rec_event_info{$event_info->[0]} = $event_info; }

		### Main process. ###

	my $maxct_indels_invlvd_rec = 0; # Maximum possible number of indels involved in the reconstructed MSA.  (ADDED for DEBUGGING on Oct 4 (Sun), 2015). #

	my @indels_invlvd_rec = ();

	if (defined $mrca) {
	   my $event_mrca = $br2rec_event_info{$mrca};
	   unless ((defined $event_mrca) and ($event_mrca->[3] eq $char_abs)) {
		# Either no event along the MRCA branch, or if the event is not an insertion. #
		return (0, []);
	   }
	   push @indels_invlvd_rec, $event_mrca;
	   $maxct_indels_invlvd_rec++; # ADDED for DEBUGGING on Oct 4 (Sun), 2015. #
	}

	foreach my $br (@to_be_examined) { # Main outer foreach-loop (over branches mediating the two nodes). # 
	    my $pa = $node2pa->{$br};
	    my $sibls = $node2ch->{$pa};
	    foreach my $sbl (@{$sibls}) { # Main inner foreach-loop (over siblings of $br). #

		($sbl == $br) and next;

		$maxct_indels_invlvd_rec++;

		my $event_sbl = $br2rec_event_info{$sbl};

		unless ((defined $event_sbl) and ($event_sbl->[3] ne $char_abs)) {
			# Either no event along the sibling branch, or if the event is not a deletion.

		    ##return (0, []); # This might have been too strict,
		    next; 		# hence was replaced with a more lenient condition (for DEBUGGING on Oct 4 (Sun), 2015).
		}
		push @indels_invlvd_rec, $event_sbl;

	    }			# End of the main inner foreach-loop (over siblings of $br). #
	}			# End of the main outer foreach-loop (over branches mediating the two nodes). #

		# Regard the error as (i)CII if 1 spurious deletion is missing.
		# (ADDED for DEBUGGING on Oct 4 (Sun), 2015.) #
	my $ct_missing_indels = $maxct_indels_invlvd_rec - scalar (@indels_invlvd_rec);
	unless (($ct_missing_indels == 0) or (($ct_missing_indels < 2) and (@indels_invlvd_rec>1)) ) { 
	    return (0, []);
	}
		# END of "# Regard the error as (i)CII if 1 spurious deletion is missing. #

	return (1, \@indels_invlvd_rec);

}	# End of the subroutine 'if_collapse_indep_ins'. #


# if_part_nbr_mrg_2ins:
#	This subroutine examines whether the three input events are consistent with
#	the error of the type 'partial-neighboring-branch-merge' of two insertions.
#	(CREATED on Nov 12 (Wed), 2014.)
#
#	CRITERIA:
#	(1) The three events must be on different branches (pre-guaranteed), 
#		and the branches must share a node.
#	(2) Starting from the shared node, event 1 and event 2 must be 'insertion's, and event 3 must be a 'deletion'.
#
# INPUT: (\@event1, \@event2, \@event3, $indx_br, $indx_state_ue,  \%equiv_branches, \%node2pa, $char_abs), where
#
#	($br1/2/3, $state_ue1/2/3) = @{$event1/2/3}[$indx_br, $indx_state_ue] are the branch along which the event 1/2/3 occurred,
#	and the state at the upper-end for the event 1/2/3;
#
#	%equib_branches = ($br_id => $equiv_br_id, ...) for the branches with their own 'equivalent's;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node (i.e., root);
#
#	$char_abs is a character representing an 'absence' state.
#
# OUTPUT: $if_part_nbr_mrg, which should be
#		= 1 if the events are consistent with the error type 'partial-neighboring-branch-merge',
#		= 0 otherwise,
#		= undef if the subroutine failed.
#
sub if_part_nbr_mrg_2ins (\@\@\@$$\%\%$) {

	my ($event1, $event2, $event3, $indx_br, $indx_state_ue, $equiv_branches, $node2pa, $char_abs) = @_;

	my ($br1, $state_ue1) = @{$event1}[$indx_br, $indx_state_ue];
	my ($br2, $state_ue2) = @{$event2}[$indx_br, $indx_state_ue];
	my ($br3, $state_ue3) = @{$event3}[$indx_br, $indx_state_ue];

	if ($br1 == $node2pa->{$br2}) { # $br1 must be the parent of $br2 and $br3.

	    ($br1 == $node2pa->{$br3}) or (return 0); 
	    if (($state_ue1 ne $char_abs) and ($state_ue2 eq $char_abs) and ($state_ue3 ne $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }

	} elsif ($br2 == $node2pa->{$br1}) { # $br2 must be the parent of $br1 and $br3.

	    ($br2 == $node2pa->{$br3}) or (return 0);
	    if (($state_ue1 eq $char_abs) and ($state_ue2 ne $char_abs) and ($state_ue3 ne $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }

	} elsif ($br3 == $node2pa->{$br1}) { # $br3 must be the parent of $br1 and $br2.

	    ($br3 == $node2pa->{$br2}) or (return 0);
	    if (($state_ue1 eq $char_abs) and ($state_ue2 eq $char_abs) and ($state_ue3 eq $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }

	}

	my $eq_br1 = $equiv_branches->{$br1};
	my $eq_br2 = $equiv_branches->{$br2};
	my $eq_br3 = $equiv_branches->{$br3};

		# NOTE: at most one of the equivalents will be defined. #

	if (defined $eq_br1) { # The equivalent of $br1 must be the parent of $br2 and $br3.

	    (($eq_br1 == $node2pa->{$br2}) and ($eq_br1 == $node2pa->{$br3})) or (return 0);
	    if (($state_ue1 eq $char_abs) and ($state_ue2 eq $char_abs) and ($state_ue3 ne $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }

	} elsif (defined $eq_br2) { # The equivalent of $br2 must be the parent of $br1 and $br3.

	    (($eq_br2 == $node2pa->{$br1}) and ($eq_br2 == $node2pa->{$br3})) or (return 0);
	    if (($state_ue1 eq $char_abs) and ($state_ue2 eq $char_abs) and ($state_ue3 ne $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }

	} elsif (defined $eq_br3) { # The equivalent of $br3 must be the parent of $br1 and $br2. 

	    (($eq_br3 == $node2pa->{$br1}) and ($eq_br3 == $node2pa->{$br2})) or (return 0);
	    if (($state_ue1 eq $char_abs) and ($state_ue2 eq $char_abs) and ($state_ue3 ne $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }
	}

		# The three events do NOT share a node. #
	return 0;

}		# End of the subroutine 'if_part_nbr_mrg_2ins'. #



	### DEFINITIONS of the Subroutines ADDED in creation of 'compare_ref_vs_rec_mafft_consist.pl' or later. ###

# read_list_pw_lclhom:
#	This subroutine reads a list of pairwise local homology information,
#	which was used to calculate the consistency score of MSAs.
#	(CREATED in creation of 'compare_ref_vs_rec_mafft_consist.pl' on Sep 15 (Mon), 2014.)
#
# INPUT: (\$inlist, $ct_seqs), where
#	$inlist specifies the path to the input list file;
#	$ct_seqs is the number of sequences in the MSA.
#
# OUTPUT: \@set_info_pw_lclhom, where
#	@info_pw_lclhom = @{$set_info_pw_lclhom[$i]->[$j]} ($i< $j) stores information on the pairwise local homology
#	between the $i th and the $j th sequences in the MSA. More specifically,
#	@{$info_pw_lclhom->[$n]} = ($start1, $end1, $start2, $end2, $importance) is for the $n the gapless segment
#	in the pairwise local alignment(s) between the two sequences.
#
sub read_list_pw_lclhom (\$$) {

	my ($inlist, $ct_seqs) = @_;
	local $/ = "\n";

	my $fh_tmp = open_input_file2 (${$inlist});
	unless (defined $fh_tmp) {
	    print STDOUT "[ERROR 'read_list_pw_lclhom'] 'open_input_file2' FAILED : $! !\n\n";
	    return undef;
	}

	my $header = <$fh_tmp>;

	my @set_info_pw_lclhom = ();
	for (my $i=0; $i<$ct_seqs-1; $i++) { $set_info_pw_lclhom[$i] = []; }

	while (my $line = <$fh_tmp>) { # Main while-loop (over input lines). #

	    my @fields = split /\s+/, $line;
	    (@fields>6) or next;

	    my ($i, $j, $start1, $end1, $start2, $end2, $importance) = @fields;

	    unless ($i < $j) {
		$line =~ s/\s+$//;
		print STDOUT "[WARNING 'read_list_pw_lclhom'] ILLEGAL INPUT: Indx_seq1 (=$i) is NOT LESS THAN Indx_seq2 (=$j) for the line:\n";
		print STDOUT "'$line' : $! !\n\n";
		next;
	    }

	    unless (($i>=0) and ($j <$ct_seqs)) {
		$line =~ s/\s+$//;
		print STDOUT "[WARNING 'read_list_pw_lclhom'] ILLEGAL INPUT: EITHER Indx_seq1 (=$i) is Negative OR Indx_seq2 (=$j) is NOT LESS THAN #{sequences} (=$ct_seqs) for the line:\n";
		print STDOUT "'$line' : $! !\n\n";
		next;
	    }

	    my $row_info = $set_info_pw_lclhom[$i];
	    (defined $row_info) or ($row_info = $set_info_pw_lclhom[$i] = []);

	    my $info_pw_lclhom = $row_info->[$j];
	    (defined $info_pw_lclhom) or ($info_pw_lclhom = $row_info->[$j] = []);

	    push @{$info_pw_lclhom}, [$start1, $end1, $start2, $end2, $importance];

	}			# End of the main while-loop (over input lines). #

	close ($fh_tmp);

	return \@set_info_pw_lclhom;

}	# End of the subroutine 'read_list_pw_lclhom'. #

	### DEFINITIONS OF THE NEW SUBROUTINES (ADDED since Aug 26 (Sun), 2012) ###

# readin_list_segs:
#	This subroutine reads in a list of segments.
#
# INPUT: (\$inlist, \%label_sattr2indx), where
#	$inlist is the path to the input file.
#	%label_sattr2indx = ($label_seg_attr => $indx, ...)
#	with $indx being the index in the output list.
#
# OUTPUT: \@list_segments, where
#	@{$list_segments[$s]} = ($seg_ID, @attributes)
#	is the information on the $s th segment, with
#	$seg_ID being the ID, and
#	@attributes storing the attributes of the segment in the same order as in @LABELS_ATTR.
#
sub readin_list_segs (\$\%) {

	my ($inlist, $label_sattr2indx) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$inlist});
	unless (defined $fh_tmpin) {
	   print STDOUT "[ERROR 'readin_list_segs'] 'open_input_file2' FAILED on the input file '${$inlist}' $! !\n";
	    return undef;
	}

	my $head1 = <$fh_tmpin>;
	my ($ct_seqs)  = ($head1 =~ m/^\#Sequences\s+(\d+)/);
	my $head2 = <$fh_tmpin>;
	my ($ct_segs) = ($head2 =~ m/^\#Segments\s+(\d+)/);
	unless ((defined $ct_seqs) and (defined $ct_segs)) {
	    (defined $ct_seqs) or ($ct_seqs = 'undef');
	    (defined $ct_segs) or ($ct_segs = 'undef');
	    print STDOUT "[ERROR 'readin_list_segs'] ILLEGAL HEADER: (\$ct_seqs, \$ct_segs) = ($ct_seqs, $ct_segs): $! !\n";
	    return undef;
	}

	my @list_segments = ();

	while (my $in = <$fh_tmpin>) { # Main while-loop #

	    my @flds = split /\s+/, $in;   

	    my $segid = shift @flds;
	    $segid =~ s/^seg\#//;
	    my @seg_info = ($segid);

	    while (my $label = shift @flds) { # Inner while-loop #
		$label =~ s/\:$//;
		my $indx = $label_sattr2indx->{$label};
		my $attr = shift @flds;
	 	unless (defined $indx) {
		    print STDERR "[CAUTION 'readin_list_segs'] NO INDEX corresponds to the attribute-label '$label' (with value '$attr') in the segment '$segid': $! !\n";
		    next;
		}
		$seg_info[$indx] = $attr; 
	    }		# End of the inner while-loop #

	    push @list_segments, \@seg_info;

	}		# End of the main while-loop #

	close ($fh_tmpin);

	return \@list_segments;

}	# End of the subroutine 'readin_list_segs' #


# readin_list_segs2:
#	This subroutine reads in a list of segments in a format different from the input for 'readin_list_segs'.
#	(CREATED in creation of 'compare_ref_vs_rec_compllimment.pl' on Aug 9 (Sat), 2014)
#
# INPUT: (\$inlist, \@labels_seg_attr), where
#	$inlist is the path to the input file.
#	@labels_seg_attr stores the labels of the fields for the segment attributes (including the ID (or index)).
#
# OUTPUT: \@list_segments, where
#	@{$list_segments[$s]} = @attributes
#	is the information on the $s th segment, with
#	@attributes storing the attributes of the segment in the same order as in @labels_seg_attr.
#
sub readin_list_segs2 (\$\@) {

	my ($inlist, $labels_seg_attr) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$inlist});
	unless (defined $fh_tmpin) {
	   print STDOUT "[ERROR 'readin_list_segs2'] 'open_input_file2' FAILED on the input file '${$inlist}' $! !\n";
	    return undef;
	}

	my $head = <$fh_tmpin>;

	my $indices_attr = tfm_labels2indices (@{$labels_seg_attr}, $head);
	unless (defined $indices_attr) {
	    $head =~ s/\s+$//;
	    print STDOUT "[ERROR 'readin_list_segs2'] 'tfm_labels2indices' FAILED on\n";
	    print STDOUT "      the labels  = (", join (', ', @{$labels_seg_attr}), " ),\n";
	    print STDOUT "  and the header  = '$head' : $! !\n\n";
	    return undef;
	}

=set aside for later use
	my $new_indx_diag;
	for (my $i=0; $i<@{$labels_seg_attr}; $i++) { # Preliminary for-loop (over field labels). #
	    ($labels_seg_attr->[$i] eq 'Diagnosis') or next;
	    $new_indx_diag = $i;
	    last;
	}				# End of the preliminary for-loop (over field labels). #
=cut

	my $max_indx_fld = 0;
	foreach my $indx (@{$indices_attr}) { ($max_indx_fld < $indx) and ($max_indx_fld = $indx); }

	my @list_segments = ();

	while (my $tmpin = <$fh_tmpin>) { # Main while-loop (over input lines). #

	    $tmpin =~ s/\s+$//;
	    my @fields = split /\t/, $tmpin;

	    (@fields > $max_indx_fld) or next; # Skip lines with an insufficient number of fields.

	    my @attributes = @fields[@{$indices_attr}];

	    push @list_segments, \@attributes;

	}		# End of the main while-loop (over input lines). #

	close ($fh_tmpin);

	return \@list_segments;

}	# End of the subroutine 'readin_list_segs2' #


# readin_simul_params:
#	This subroutine reads simulation parameters,
#	except the input tree and those in the comment line(s),
#	out of an input list.
#
# INPUT: (\$inlist), where
#	$inlist is the path to an input list file.
#
# OUTPUT: \%hash_simul_params, where
#	%hash_simul_params = ($label => $value, ...) for the simulation parameters.
#
sub readin_simul_params (\$) {

	my ($inlist) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$inlist});
	unless (defined $fh_tmpin) {
	    print STDOUT "[ERROR 'readin_simul_params'] 'open_input_file2' FAILED on the input list '${$inlist}': $! !\n";
	    return undef;
	}

	my %hash_simul_params;

	while (my $in = <$fh_tmpin>) { # Main while-loop #

	    ($in =~ m/^\s*\#/) and next; # Skip a comment line. #

		# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
	    ($in =~ m/^Tree\s/) and next; # Skip the input tree. #
	    ##($in =~ m/^Tree\.Tree/) and next; # Skip the input tree. #

	    ($in =~ s/\s+$//);

	    my ($label, $value) = split /\s*\=\s*/, $in;
	    unless ((defined $value) and ($value =~ m/\S/)) {
		print STDOUT "[CAUTION 'readin_simul_params'] NO VALUE in the input line:\n";
		print STDOUT "'$in': $! !\n\n";
		next;
	    }

		# ADDED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
	    if ($value =~ s/^\"//) { ($value =~ s/\"$//); }
	    if ($label eq 'GapParams') { $value =~ s/^\{//; $value =~ s/\}$//; }

	    $hash_simul_params{$label} = $value;

	}			# End of the main while-loop #
	close ($fh_tmpin);

	return \%hash_simul_params;

}		# End of the subroutine 'readin_simul_params' #


# inspect_indel_events:
#	This subroutine examines whether all the indel events are shorter than or equal to the cutoff-lengths.
#
# INPUT: (\@set_cls_gblocks, \@sets_psm_scns_for_cls, $cutoff_len_ins, $cutoff_len_del, $char_abs), where
#
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)
#	is for the $gb th block in the $cl th cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordiate of the block (in the 0-based convention);
#
#	@{$set_psm_scns_in_cls[$cl]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th of the parsimonious scenarios realizing the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) if the map of gap states onto the interior nodes,
#	with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id;
#	
#	$cutoff_len_ins/del is the cutoff length for the insertion/deletion events;
#
#	$char_abs is the character representing the 'absence' state.
#
# OUTPUT: \@illegal_indels, where
#	@{$illegal_indels[$il]} = ($cl, $ps, $br_id, $id, $length) is for the $il th illegal indel event, with
#	$cl specifying a cluster,
#	$ps specifying a parsimonious scenario in the cluster,
#	$br_id specifying a branch,
#	$id specifying an illegal event in the node,
#	and $length is the length of the event.
#
#	The subroutine returns undef if it fails.
#
sub inspect_indel_events (\@\@$$$) {

	my ($set_cls_gblocks, $sets_psm_scns_for_cls, $cutoff_len_ins, $cutoff_len_del, $char_abs) = @_;

    (print STDOUT ' ' x 3, "< Subroutine 'inspect_indel_events' >\n\n") if ($TEST_LEVEL>1);

	my $ct_cls = @{$set_cls_gblocks};
	unless (@{$sets_psm_scns_for_cls} == $ct_cls) {
	    print STDOUT "[ERROR 'inspect_indel_events'] ILLEGAL INPUT: The size of ", '@{$sets_psm_scns_for_cls} = (', scalar (@{$sets_psm_scns_for_cls}), ") CLASHES with the number of gap-block clusters (=$ct_cls): $! !\n";
	    return undef;
	}

	my @illegal_indels = ();

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Outer for-loop (over clusters) #

	    my $cls_gblock = $set_cls_gblocks->[$cl];
	    my $psm_scns_for_cls = $sets_psm_scns_for_cls->[$cl];
	    my $ct_psm_scns = @{$psm_scns_for_cls};

    (print STDOUT ' ' x 3, "[ The $cl th cluster has $ct_psm_scns parsimonious scenarios of indel events ]\n\n") if ($TEST_LEVEL>2);

	    for (my $ps=0; $ps<$ct_psm_scns; $ps++) { # Middle for-loop (over parsiminous scenarios) #

    (print STDOUT ' ' x 5, '( The $ps th parsimonious scenario )', "\n\n") if ($TEST_LEVEL>3);

		my ($br2indels, $node2bw_gstat) = @{$psm_scns_for_cls->[$ps]};

		foreach my $br_id (sort {$a <=> $b} keys %{$br2indels}) { # Inner-middle foreach-loop (over branches) #
		    my $indel_events = $br2indels->{$br_id};
		    my $ct_events = @{$indel_events};

    (print STDOUT ' ' x 7, "( The branch '$br_id' has $ct_events nominal indel events )\n") if ($TEST_LEVEL>3);

		    for (my $id=0; $id<$ct_events; $id++) { # Inner for-loop (over indel events) #
			my ($stat_pa, $stat_node, $indices_blk) = (defined $indel_events->[$id]) ? @{$indel_events->[$id]} : ();
			unless (defined $stat_pa) {
			    print STDOUT "[WARNING 'inspect_indel_events'] The $id th event on the branch '$br_id' is EMPTY  for the $ps th indel scenario for the $cl th cluster: $! !\n";
			    next;
			}
			my $cutoff_len = ($stat_pa eq $char_abs) ? $cutoff_len_ins : $cutoff_len_del;
			my $indel_len = 0;
			foreach my $indx_blk (@{$indices_blk}) { $indel_len += $cls_gblock->[$indx_blk][1] ; }

			unless ($indel_len <= $cutoff_len) {
			    print ' ' x 7, "\n[WARNING 'inspect_indel_events'] The $id th indel event along the $br_id th branch is EXCEEDINGLY LONG (length = $indel_len)  for the $ps th scenario in the $cl th gap-block cluster: $! !\n\n";
			    push @illegal_indels, [$cl, $ps, $br_id, $id, $indel_len];
			}

		    }				# End of the inner for-loop (over indel events) #

		} 		# End of the inner-middle foreach-loop (over branches) #

	    }			# End of the middle for-loop (over parsimonious scenarios) #

	}			# End of the outer for-loop (over clusters) #

    (print STDOUT ' ' x 3, "--- End of the Subroutine 'inspect_indel_events'  ---\n\n") if ($TEST_LEVEL>1);

	return \@illegal_indels;

}	# End of the subroutine 'inspect_indel_events' #


# readin_dawg_tree2:
#	This modified version of 'readin_dawg_tree' copes with the new format
#	of the Dawg control file.
#	(CREATED in creation of 'compare_ref_vs_rec_compllimment.pl' on Aug 9 (Sat), 2014)
#
# readin_dawg_tree:
#	This subroutine reads in the input tree used to guide
#	a simulation of sequence evolution via DAWG.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011) 
#
# INPUT: \$infile, where
#	$infile is the file recording input parameters of DAWG.
#
# OUTPUT: \@info_dawg_tree, where
#	@info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr).
#	See e.g. 'MyTreeManipu_II.pm' for the format of the output hash.
#
sub readin_dawg_tree2 (\$) {

	my ($infile) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$infile});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'open_input_file2' FAILED on the input file '${$infile}': $! \n";
	    return undef;
	}

	my $in_tree;
	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #

	    ($tmpin =~ s/^Tree\s+\=\s+//) or next; # MODIFIED in creation of 'readin_dawg_tree2'. #
	    ##($tmpin =~ s/^Tree\.Tree\s+\=\s+//) or next;
	    $in_tree = $tmpin;
	    last;
	}		# End of the main while-loop # 
	close ($fh_tmpin);

	unless ((defined $in_tree) and ($in_tree =~ m/\(/)) {
	    (defined $in_tree) or ($in_tree = 'undef');
	    $in_tree =~ s/\s+$//;
	    print STDERR "[ERROR 'readin_dawg_tree2'] NO TREE was extracted from the input file '${$infile}': $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

		## MODIFIED in creation of 'readin_dawg_tree2'. ##

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr; my %name2id_intnode; my $topnode_name;

	unless ( input_tree_file_dawg ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr, %name2id_intnode, $topnode_name) ) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'input_tree_file_dawg' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

		# (Just in case,) remove the interior node names (or bootstrap supports) from the list of branch attributes. #

	foreach my $br (keys %br_attr) {
	    my $attr = $br_attr{$br};
	    my @new_attr = ();
	    $new_attr[$INDX_LENGTH] = $attr->[$INDX_LENGTH];
	    $br_attr{$br} = \@new_attr;
	}

=for the predecessor 'readin_dawg_tree'
		# Process the input tree a bit #
	($in_tree =~ s/[^\)\;]+$//);
	($in_tree =~ m/\;$/) or ($in_tree .= ';');

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;
	unless ( input_tree_file ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr) ) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'input_tree_file' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}
=cut

	my @info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr);

	return \@info_dawg_tree;

}	# End of the subroutine 'readin_dawg_tree2' #


	### DEFINITIONS of the SUBROUTINES in the predecessor,
	### "../../../2nd_Stage.DNA/ML_ScoringF6t.PhyML/Scripts/compare_ml_msa.bch.extJC.6t.PhyML.pl".

# readin_dawg_tree:
#	This subroutine reads in the input tree used to guide
#	a simulation of sequence evolution via DAWG.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011) 
#
# INPUT: \$infile, where
#	$infile is the file recording input parameters of DAWG.
#
# OUTPUT: \@info_dawg_tree, where
#	@info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr).
#	See e.g. 'MyTreeManipu_II.pm' for the format of the output hash.
#
sub readin_dawg_tree (\$) {

	my ($infile) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$infile});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'readin_dawg_tree'] 'open_input_file2' FAILED on the input file '${$infile}': $! \n";
	    return undef;
	}

	my $in_tree;
	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #
	    ($tmpin =~ s/^Tree\.Tree\s+\=\s+//) or next;
	    $in_tree = $tmpin;
	    last;
	}		# End of the main while-loop # 
	close ($fh_tmpin);

	unless ((defined $in_tree) and ($in_tree =~ m/\(/)) {
	    (defined $in_tree) or ($in_tree = 'undef');
	    $in_tree =~ s/\s+$//;
	    print STDERR "[ERROR 'readin_dawg_tree'] NO TREE was extracted from the input file '${$infile}': $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}
		# Process the input tree a bit #
	($in_tree =~ s/[^\)\;]+$//);
	($in_tree =~ m/\;$/) or ($in_tree .= ';');

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;
	unless ( input_tree_file ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr) ) {
	    print STDERR "[ERROR 'readin_dawg_tree'] 'input_tree_file' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

	my @info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr);

	return \@info_dawg_tree;

}	# End of the subroutine 'readin_dawg_tree' #


# ln_ml_sgl_seq2:
#	This revised subroutine of 'ln_ml_sgl_seq' also outputs 'column-wise' log ML values.
#	(IMPORTED from 'MyApBayes_Score_MSA_ExtJC.pm' and MODIFIED in creation of
#	'compare_apbayes_msa.bch.extJC.6.pl' on Mar 7 (Mon), 2011)
#
# ln_ml_sgl_seq:
#	This subroutine estimates the log-likelihood of obtaining a single sequence.
#	The log-likelihood is simply calculated as:
#	$lnl = Sum_$r log ($freq[$r]) ,
#	where $freq[$r] is the equilibrium frequency of the $r th amino acid residue.
#
# INPUT: (\$seq, \%aa2freq), where
#	$seq is the sequence;
#	%aa2freq = ($amino_acid (a single capital letter) => $equilibrium_frequency, ...).
#
# OUTPUT: ($lnl, \@clmwise_lnl), where
#	$lnl is the log-likelihood of the whole alignment;
#	$clmwise_lnl[$l] is the log-likelihood of the $l th column.
#
#	The subroutine returns (undef, undef) if it failed.
#	
sub ln_ml_sgl_seq2 (\$\%) {

	my ($seq, $aa2freq) = @_;

	my $cp_seq = uc ${$seq};
	my $len = length $cp_seq;

	my @clmwise_lnl = ();
	my $lnl = 0;
	for (my $l=0; $l<$len; $l++) { # For-loop #
	    my $res = substr ($cp_seq, 0, 1, '');
	    my $frq = $aa2freq->{$res};
	    unless (defined $frq) {
		print STDERR "[ERROR 'ln_ml_sgl_seq2'] The frequency is UNDEFINED for the residue '$res' at the $l th site: $! !\n\n";
		return (undef, undef);
	    }
	    my $clmwise = log $frq;
	    $lnl += $clmwise;
	    $clmwise_lnl[$l] = $clmwise; # ADDED in creation of 'compare_apbayes_msa.bch.extJC.6.pl' (on Mar 7 (Mon), 2011) #
	}		# End of the for-loop #

	return ($lnl, \@clmwise_lnl);

}	# End of the subroutine 'ln_ml_sgl_seq2' #


# add_dummy_seq:
#	This subroutine adds a 'dummy' sequence to a tree and an MSA consisting of two sequences,
#	and makes a 'virtual' 3-sequence tree and MSA.
#	The dummy is the exact copy of the 2nd sequence, which 'diverge'd $eps above its exterior node,
#	where $eps is the minimum length permitted by PhyML.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 11 (Mon) & 12 (Tue), 2011)
#
# INPUT: (\@seqnames, \@msa, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr, \@indices_attr), where
#	$seqnames[$k] is the name of the $k th sequence;
#	${$msa[$k]} is the $k th sequence embedded in the alignment;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#	%name2id = ($name => $node_id, ...) for exterior nodes and the top node;
#	%id2name = ($node_id => $name, ...) for exterior nodes;
#	%br_attr = ($node_id => \@branch_attributes, ...) for the nodes immediately downstream of the branches,
#	with @branch_attributes[@indices_attr] = ($branch_length, $bootstrap_support, Total#{bootstrap resamplings}).
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub add_dummy_seq (\@\@\%\%\%\%\%\@) {

	my ($seqnames, $msa, $node2pa, $node2ch, $name2id, $id2name, $br_attr, $indices_attr) = @_;

	unless ((@{$seqnames} == 2) and (@{$msa} == 2)) {
	    print STDERR "[ERROR 'add_dummy_seq'] ILLEGAL INPUT: Either the size of MSA (", scalar (@{$msa}), ") or the number of sequence names (", scalar (@{$seqnames}), ") is NOT 2, which they are supposed to be: $! !\n";
	    return 0;
	}

		## Manipulate the input tree ##

	my $eps = 0.0000000001; # Minimum branch length permitted in PhyML #
	my ($indx_len, $indx_boot, $tot_boot) = @{$indices_attr};

	my $max_nodeid = 0;
	foreach my $nodeid (keys %{$node2ch}) { ($nodeid > $max_nodeid) and ($max_nodeid = $nodeid); }
	foreach my $nodeid (keys %{$id2name}) { ($nodeid > $max_nodeid) and ($max_nodeid = $nodeid); }

		# Change the identity of the second sequence #
	my $name_seq2 = $seqnames->[1];
	my $nid_seq2 = $name2id->{$name_seq2};
	delete $id2name->{$nid_seq2}; # $nid_seq2 is no longer an exterior node

		# Create two new exterior nodes, one real and the other virtual #
	my $nid_sbl1 = $max_nodeid + 1;
	my $nid_sbl2 = $max_nodeid + 2;

	##my $name_sbl1 = $name_seq2;
	my $name_sbl2 = 'dummy';

	$name2id->{$name_seq2} = $nid_sbl1;
	$name2id->{$name_sbl2} = $nid_sbl2;

	$id2name->{$nid_sbl1} = $name_seq2;
	$id2name->{$nid_sbl2} = $name_sbl2;

		# Establish the relationships between seq2, and the new siblings #
	$node2ch->{$nid_seq2} = [$nid_sbl1, $nid_sbl2];
	$node2pa->{$nid_sbl1} = $nid_seq2;
	$node2pa->{$nid_sbl2} = $nid_seq2;

		# Slightly change the length of the branch above seq2, which is now the 'virtual' parent node #
	my $attr_br2 = $br_attr->{$nid_seq2};
	my $len2 = $attr_br2->[$indx_len];
	$attr_br2->[$indx_len] = ($len2 > $eps) ? $len2 - $eps : 0 ;

		# Fix the lengths of the 'virtual' child branches as $eps #
	foreach my $nid ($nid_sbl1, $nid_sbl2) { # foreach-loop #
	    my @attr = ();
	    $attr[$indx_len] = $eps; 
	    $br_attr->{$nid} = \@attr;
	}			# End of the foreach-loop #

		## Manipulate the input alignment ##
	my $seq2 = ${$msa->[1]};
	$msa->[2] = \$seq2;
	$seqnames->[2] = $name_sbl2;

	return 1;

}	# End of the subroutine 'add_dummy_seq' #


