package MyANEX_IO;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyANEX_IO.pm:
#	This Perl package consists of subroutines for data inputs, (test) outputs,
#	and some other processings, that are used in:
#	'anex.verXXX.pl', 'anex_for_sgl_wd.verXXX.pl', 'anex_rev_purge_for_sgl_wd.verXXX.pl',
#	and 'detect_purge_cands.verXXX.pl',
#	which are the main scripts of 'ANEX_P'.
#	[CREATION STARTED on 2019/07/07, ENDED on 2020/01/27.]
#
#	[REVISED in creation of 'ANEX_P.ver0.7' (on 2020/04/26)]:
#	Now, the subroutine, 'print_out_br2log_mfacs', can print out
#	the log-multiplication factors for case-(iv) "gapped segment"s as well,
#	so that ANEX can take full advantage of the main outputs of LASTPIECE.
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	); # ADDED 'open_output_file' in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). #
	# list_sessions_ext
	# tfm_labels2indices

use MyMalnIO qw(
	extract_alignment
	pr_ODEN_input
	);

use MyTreeManipu_II qw(
	input_tree_file
	output_tree 
	br_output_tree
	);
	# make_node2parent
 
use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg
	br_output_tree_dawg
	);

use MyArrayManipu_num qw(
	diff_nmarray
	union_nmarray
	binsearch_num
	); # Added on 2019/02/16.
	# Added 'binsearch_num' on 2019/07/05.

use MyArrayManipu qw(
	multiple_sort
	); # Added on 2019/07/05.

	# IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (1) #

use MyTreeManipu_II qw( );
	# extract_sgltree_from_file output_tree 

use MyANEX_Main qw(
	nonrec_mlt_shifts_gapblocks
	printout_multidim_storage
	); # ADDED 'nonrec_mlt_shifts_gapblocks' in MAJOR REVISION of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/18). #


use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	readin_dawg_tree2 
	read_dawg_ctrl_file_sbst 
	read_dawg_ctrl_file3 
	readin_list_mgd_pcands 
	readin_set_clms 
	readin_list_gblks 
	revise_list_gblocks 

	mk_revised_list_gblocks

	readin_features_gblks

	print_out_input_tree 
	print_out_br2log_mfacs 
	print_out_tree_profile 
	print_out_cal_init_psm_cands_for_sgl_msa 
	print_out_bch_br_parsimonize_dollo_indel_histories 
	print_out_alignment_psmnzd 
	print_out_wrapper_detect_cmplx_error_cand_acls0 
	print_out_prepare_sliding_windows 
	print_out_list_mgd_pcands 
	print_out_chk_ovlp_pcands_vs_sliding_wds 
	print_out_set_clms 
	print_out_list_gblocks_in_wd 
	print_out_mk_pcc_sliding_windows 

	print_out_readin_list_gblks 
	print_out_readin_features_gblks 

	print_out_wrapper_mk_node2exoffs_msa_rows 
	print_out_mk_set_msa_rows2br_up_or_lw 
	print_out_br2tr_mtrx 
	print_out_node2ntfreqs 
	print_out_node2set_sw_cond_probs 
	print_out_br2set_sw_jnt_probs 
	print_out_set_clm_probs 
	print_out_orgs_bds_gb_coords 
	print_out_row_clm_info_gblocks 
	print_out_inter_block_relations 
	print_out_features_gblks 
	print_out_x_bin_hrels 
	print_in_nr_mshifts_glks 
	print_out_ln_prb_plus_msa 
	print_out_coord_frame 
	print_out_nr_mshifts_gblks 
	print_out_list_gblock_sizes

	print_out_purge_or_merge_cands
	print_out_merge_split_cands
	print_out_split_cands
	print_out_vsplit_cands

	print_out_revpurge_cand_sizes

	print_out_list_gpblocks
	print_out_list_cii_cands

	print_out_gblks_w_spec_bds
	print_out_cts_vincluding_vincluded_onx_gblks
	bch_print_out_other_gblk_attributes

	print_out_info_mvd_sbj
	print_out_info_chngs_by_purge
	print_out_info_chngs_by_cmerge
	print_out_info_chngs_by_smerge
	print_out_info_chngs_by_cvmerge
	print_out_info_chngs_by_merge_split
	print_out_info_chngs_by_cmerge_split_grow
	print_out_info_chngs_by_split
	print_out_info_chngs_by_vsplit 
	print_out_info_chngs_by_vsplit_plus_smerge
	print_out_info_chngs_by_ssplit_plus_vmerge

	print_out_sbsdry_outputs_rev_spec_cii_in_wd

	perform_mshifts_and_print_inputs_outputs

	select_sgn_wds_in_pcands 
	chk_ovlp_pcands_vs_sliding_wds 
	determine_limits_window 
	relative_coords_for_list_mgd_pcands 
	);


	# MOVED from 'anex.ver0.5.pl', 'anex_for_sgl_wd.ver0.5.pl' & 'anex_rev_purge_for_sgl_wd.ver0.5.pl'. #

sub readin_dawg_tree2 (\$); # IMPORTED to 'detect_purge_cands.ver0.5_OLD1.pl' in its creation (on 2019/06/16). #
sub read_dawg_ctrl_file_sbst (\$);
sub read_dawg_ctrl_file3 (\$);
sub readin_list_mgd_pcands (\$); # Added in creation of 'anex.ver0.5.pl' (on 2019/06/20). #
sub readin_set_clms (\$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/25). #
sub readin_list_gblks (\$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26). #
sub revise_list_gblocks (\@\%\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26-27). #

sub mk_revised_list_gblocks (\@\%\@); # CREATED in creation of 'coordinate_point_lcl_msa.ver0.7.pl' (on 2020/06/04). #

sub readin_features_gblks (\$); # CREATED in creation of 'coordinate_point_lcl_msa.ver0.7.pl' (on 2020/06/04). #

sub print_out_input_tree (\$$\%\%\%\%$);
sub print_out_br2log_mfacs (\$\%$$);
sub print_out_tree_profile (\$\%\%\%\%\@);
sub print_out_cal_init_psm_cands_for_sgl_msa (\$\@\@);

##sub print_out_bch_br_parsimonize_dollo_indel_histories (\$$$\@\@\@\@\@\@); # Modified on 2019/06/17.
sub print_out_bch_br_parsimonize_dollo_indel_histories (\$$$\@\@\@\@\@\@\@); # => Further Modified on 2019/07/07. #

sub print_out_alignment_psmnzd (\$$\@\@); # Added on 2019/06/17.
sub print_out_wrapper_detect_cmplx_error_cand_acls0 (\$\@\@\@\%\@\@); # Added on 2019/06/14.
sub print_out_prepare_sliding_windows (\$\$\@\@\@\@); # Added on 2019/06/14; Modified on 2019/06/14.
sub print_out_list_mgd_pcands (\$$\@); # Added in creation of 'anex.ver0.5.pl' (on 2019/06/20). #
sub print_out_chk_ovlp_pcands_vs_sliding_wds (\$\@\@\@\@); # Added in creation of 'anex.ver0.5.pl' (on 2019/06/21). #
sub print_out_set_clms (\$$\@$$$$); # Added in creation of 'anex.ver0.5.pl' (on 2019/06/23). #
sub print_out_list_gblocks_in_wd (\$$\@\%); # Added in creation of 'anex.ver0.5.pl' (on 2019/06/23). #
sub print_out_mk_pcc_sliding_windows (\$\$\@\@\@\@); # Added in creation of 'anex.ver0.5.pl' (on 2019/07/06). #

sub print_out_readin_list_gblks (\$$\@$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26). #
					   # => REVISED on 2019/06/27.

sub print_out_readin_features_gblks (\$$\@$); # CREATED in creation of 'coordinate_point_lcl_msa.ver0.7.pl' (on 2020/06/04). #

sub print_out_wrapper_mk_node2exoffs_msa_rows (\$$\%\%\@\%); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26). #
sub print_out_mk_set_msa_rows2br_up_or_lw (\$$\%); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26). #
sub print_out_br2tr_mtrx (\$$\%\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/27).
sub print_out_node2ntfreqs (\$$\%\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/27).
sub print_out_node2set_sw_cond_probs (\$$\%\@$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/27-28).
sub print_out_br2set_sw_jnt_probs (\$$\%\@$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/28).
sub print_out_set_clm_probs (\$$\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/29).
sub print_out_orgs_bds_gb_coords (\$$\@\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/29).
sub print_out_row_clm_info_gblocks (\$$\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/29).
sub print_out_inter_block_relations (\$$\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/29).
sub print_out_features_gblks (\$$\@$$\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/02).
sub print_out_x_bin_hrels (\$$\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/02).

##sub print_in_nr_mshifts_glks (\$\@$$\@\@\@\@\@\@$$$$\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/02).
sub print_in_nr_mshifts_glks (\$\@\@$$\@\@\@\@\@\@$$$$\@); # => REVISED on 2019/07/18.

sub print_out_ln_prb_plus_msa (\$$$$$$$$$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/02).
sub print_out_coord_frame (\$$\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/02).

##sub print_out_nr_mshifts_gblks (\$$$$$$$\@\@\@$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/02).
sub print_out_nr_mshifts_gblks (\$\@$$$$$$\@\@\@$); # => REVISED on 2019/07/18.

sub print_out_list_gblock_sizes (\$$\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/16).
sub print_out_purge_or_merge_cands (\$$\@\$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/16).
sub print_out_merge_split_cands (\$$\@\$\$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/08/01).
sub print_out_split_cands (\$$\@\$\$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/08/13). #
sub print_out_vsplit_cands (\$$\@\$\$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/08/24). # 

sub print_out_revpurge_cand_sizes (\$$\@$$); # CREATED in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl' (on 2019/09/16). #

sub print_out_list_gpblocks (\$$\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/10/01). #
sub print_out_list_cii_cands (\$$\@$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/10/01). #

sub print_out_gblks_w_spec_bds (\$$\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/17).
sub print_out_cts_vincluding_vincluded_onx_gblks (\$$\@\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/17).

##sub bch_print_out_other_gblk_attributes (\$\@\@\@\@\@\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/17).
sub bch_print_out_other_gblk_attributes (\$\@\@\@\@\@\@\@\@); # => REVISED on 2019/07/18.

sub print_out_info_mvd_sbj (\$$\@\@\@\@\@\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/17).

##sub print_out_info_chngs_by_purge (\$$$$\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/17).
sub print_out_info_chngs_by_purge (\$\@$$$\@\@); # => REVISED on 2019/07/18.
sub print_out_info_chngs_by_cmerge (\$\@$$$$$\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/20). #
sub print_out_info_chngs_by_smerge (\$\@$$$$\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/22). #
sub print_out_info_chngs_by_cvmerge (\$\@$$$$$$\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/29). #
##sub print_out_info_chngs_by_merge_split (\$\@$$$$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/08/01). #
sub print_out_info_chngs_by_merge_split (\$\@$$$$$); # => REVISED on 2019/08/11.
##sub print_out_info_chngs_by_cmerge_split_grow (\$\@$$$\@\%); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/08/11). #
sub print_out_info_chngs_by_cmerge_split_grow (\$\@$$$\@\%\@\@); # => REVISED on 2019/08/19.
##sub print_out_info_chngs_by_split (\$\@$$$$\@\@\@\%); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/08/14).
sub print_out_info_chngs_by_split (\$\@$$$$$$\@\@\@\@\%); # => REVISED on 2019/08/20.
##sub print_out_info_chngs_by_vsplit (\$\@$$$$$$\@\@\%); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/08/25). 
sub print_out_info_chngs_by_vsplit (\$\@$$$$\%$$\@\@\%); # => REVISED on 2019/08/30.
sub print_out_info_chngs_by_vsplit_plus_smerge (\$\@$$$$$\%$$\%$$$$$$\@\@\@\@\%); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/09/08).
sub print_out_info_chngs_by_ssplit_plus_vmerge (\$\@$$$$$$$\%$$$$$\%\@\@\@\@\%); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/10/17).

sub print_out_sbsdry_outputs_rev_spec_cii_in_wd (\$\@\@\@\@); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/10/01). #

##sub perform_mshifts_and_print_inputs_outputs ($\$\$\@\@\@\@\@\@\@$$\@\@$$\@\@\%\$\%\%\%\%\%\%\@\%\%\@\@$$\@\@\@\@\@\@\@\@$$$); # CREATED in MAJOR REVISION of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/18). #
sub perform_mshifts_and_print_inputs_outputs ($\$\$\@\@\@\@\@\@\@$$\@\@$$\@\@\%$\%\%\%\%\%\%\@\%\%\@\@$$\@\@\@\@\@\@\@\@$$$) ;  # => CORRECTED on 2020/05/08. #

sub select_sgn_wds_in_pcands (\@); # Added in creation of 'anex.ver0.5.pl' (on 2019/06/20). #
sub chk_ovlp_pcands_vs_sliding_wds (\@$$$\@$$\@\@$$); # Added in creation of 'anex.ver0.5.pl' (on 2019/06/20). #
sub determine_limits_window ($\@\@\@); # Added in creation of 'anex.ver0.5.pl' (on 2019/06/21). #
sub relative_coords_for_list_mgd_pcands (\@$); # CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/03). #

	# END of "MOVED from 'anex.ver0.5.pl', 'anex_for_sgl_wd.ver0.5.pl' & 'anex_rev_purge_for_sgl_wd.ver0.5.pl'." #


	# IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (2) #

	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our %NUCL2INDX; # = ($nucleotide => {index in @NUCLEOTIDES}, ...)
for (my $k=0; $k<$CT_NUCLS; $k++) { $NUCL2INDX{$NUCLEOTIDES[$k]} = $k; }

our $LOG_ZERO = $MyLikelihoods_respattern_msa_dna::LOG_ZERO;

	# END of "IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (2)" #


our $DIS_LEN = 60; # The maximum length of the (sub)sequence displayed in a line.

	# Testing the inputs/outputs of the processes that WERE going through the MAIN programs when defined.
our $TEST_LEVEL_MAIN = 0;

	# For some subroutines that had been defined in the scripts (and were moved to this package). #
our $TEST_LEVEL_SUB = 0;



### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.



	##### (I) DEFINITIONS of the Subroutines For Inputs. #####

# readin_dawg_tree2:
#	This modified version of 'readin_dawg_tree' copes with the new format
#	of the Dawg control file.
#	(CREATED in creation of 'compare_ref_vs_rec_compllimment.pl' on Aug 9 (Sat), 2014)
#	(=> IMPORTED from 'examine_purges_via_mblks.2.pl' to 'test_detect_purge_cands.pl' on Jan 11 (Thu), 2018.)
#	[=> Further IMPORTED to 'detect_purge_cands.ver0.5_OLD1.pl' in its creation (on 2019/06/16).]
#
# readin_dawg_tree:
#	This subroutine reads in the input tree used to guide
#	a simulation of sequence evolution via DAWG.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011) 
#
# INPUT: \$infile, where
#	$infile is the file recording input parameters of DAWG.
#
# OUTPUT: \@info_dawg_tree, where
#	@info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr).
#	See e.g. 'MyTreeManipu_II.pm' for the format of the output hash.
#
sub readin_dawg_tree2 (\$) {

	my ($infile) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$infile});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'open_input_file2' FAILED on the input file '${$infile}': $! \n";
	    return undef;
	}

	my $in_tree;
	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #

	    ($tmpin =~ s/^Tree\s+\=\s+//) or next; # MODIFIED in creation of 'readin_dawg_tree2'. #
	    ##($tmpin =~ s/^Tree\.Tree\s+\=\s+//) or next;
	    $in_tree = $tmpin;
	    last;
	}		# End of the main while-loop # 
	close ($fh_tmpin);

	unless ((defined $in_tree) and ($in_tree =~ m/\(/)) {
	    (defined $in_tree) or ($in_tree = 'undef');
	    $in_tree =~ s/\s+$//;
	    print STDERR "[ERROR 'readin_dawg_tree2'] NO TREE was extracted from the input file '${$infile}': $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

		## MODIFIED in creation of 'readin_dawg_tree2'. ##

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr; my %name2id_intnode; my $topnode_name;

	unless ( input_tree_file_dawg ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr, %name2id_intnode, $topnode_name) ) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'input_tree_file_dawg' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

		# (Just in case,) remove the interior node names (or bootstrap supports) from the list of branch attributes. #

	foreach my $br (keys %br_attr) {
	    my $attr = $br_attr{$br};
	    my @new_attr = ();
	    $new_attr[$INDX_LENGTH] = $attr->[$INDX_LENGTH];
	    $br_attr{$br} = \@new_attr;
	}

=for the predecessor 'readin_dawg_tree'
		# Process the input tree a bit #
	($in_tree =~ s/[^\)\;]+$//);
	($in_tree =~ m/\;$/) or ($in_tree .= ';');

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;
	unless ( input_tree_file ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr) ) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'input_tree_file' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}
=cut

	my @info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr);

	return \@info_dawg_tree;

}	# End of the subroutine 'readin_dawg_tree2' #


# read_dawg_ctrl_file_sbst:
#	This subroutine extracts some parameters for the substitution model
#	from a Dawg control file.
#	It was created by taking advantage of the framework of 'read_dawg_ctrl_file3'.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/11.]
#
#	NOTE: This version does NOT handle the parameters, "Width", "Scale", "Alpha", "Gamma", or "Iota".
#
#	[CORRECTED on 2020/05/11]:
#	Now, it can skip blank lines (consisting ONLY of white spaces).
#
# read_dawg_ctrl_file3:
#	This is a modified version of the subroutine 'read_dawg_ctrl_file'.
#	It can read in parameter settings of the insertion and deletion models
#	that are separately set.
#	[CREATED in creation of 'fa_lolipog_hs.alpha2.pl'.]
#	[=> IMPORTED from 'fa_lolipog_hs.alpha2.pl' in creation of 'anex.ver0.5.pl' (on 2019/06/11).]
#
# read_dawg_ctrl_file:
#	This subroutine reads a Dawg control file,
#	and retrieves indel parameters and a guide tree.
#
# INPUT: \$ctrl_file, where
#	$ctrl_file is the path to a Dawg control file.
#
# OUTPUTS: ($model_sbst, \@nt_freqs, \@sbst_params), where
#
#	$model_sbst is the code for the substitution model specified in the "Model" line;
#
#	@nt_freqs stores the base frequencies (A,C,G,T) specified in the "Freqs" line;
#	[NOTE: If $model_sbst eq 'JC', 'K2P', or 'K3P', this subroutine forcibly set
#		@nt_freqs = (1/4, 1/4, 1/4, 1/4)];
#
#	@sbst_params stores the substitution parameters specified in the "Params" line;
#	[NOTE2: The following is a quote from the "NOTES" in the "readme.txt" file of Dawg (ver. 1.2-release):
#	"The meaning of the "Params" vector is different for each substitution model.
#	  GTR: Substitution rates A-C, A-G, A-T, C-G, C-T, G-T
#	  JC:  Ignored
#	  K2P: Transition rate, Transversion rate
#	  K3P: Alpha (Transitions), Beta (A-T & G-C), Gamma (A-C & G-T)
#	  HKY: Transition rate, Transversion rate
#	  F81: Ignored
#	  F84: Kappa
#	  TN:  Alpha1 (A-G), Alpha2 (C-T), Beta (Transversions)"
#	]
#
#	If fails, the subroutine returns (undef, undef, undef).
#
sub read_dawg_ctrl_file_sbst (\$) {

	my ($ctrl_file) = @_;
	local $/ = "\n";

	my $fh_ctrl = open_input_file2 (${$ctrl_file});
	unless (defined $fh_ctrl) {
	    print STDOUT "[ERROR 'read_dawg_ctrl_file_sbst'] 'open_input_file2' FAILED on the control file '${$ctrl_file}': $! !\n";
	    return (undef, undef, undef);
	}

	my $model_sbst;
	my @nt_freqs = my @sbst_params = ();

		# Extract the substitition parameters from the control file. #

	while (my $line = <$fh_ctrl>) { # Main while-loop (over lines). #

	    ($line =~ m/\S/) or next; # Skip a blank line. (ADDED in CORRECTION on 2020/05/12.)

	    ($line =~ m/^\#/) and next; # Skip a comment line. #

	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;
	    $line =~ s/\;$//; # Added to prevent an error message (concerning $tot_indel_rate) on Mar 10 (Mon), 2014. #

	    my ($option_name, $value) = split /\s*\=\s*/, $line;

	    if ($option_name eq 'Model') {
		$value =~ s/^\"//;
		$value =~ s/\"$//;
		$model_sbst = $value;

	    } elsif ($option_name eq 'Freqs') {

		$value =~ s/^\{\s*//;
		$value =~ s/\s*\}$//;

		@nt_freqs = split /\s*\,\s*/, $value;

	    } elsif ($option_name eq 'Params') {

		$value =~ s/^\{\s*//;
		$value =~ s/\s*\}$//;

		@sbst_params = split /\s*\,\s*/, $value;
	    }

	}			# End of the main while-loop (over lines). #

	close ($fh_ctrl);

		# Check the consistencies between parameters.
		# If an inconsistency is found, 
		# EITHER forcibly correct it OR abort the subroutine, depending on the situations.

	unless (defined $model_sbst) {
	    print STDOUT "[INFORMATION: 'read_dawg_ctrl_file_sbst'] The substitution model is NOT specified in the control file. Thus I set \$model_sbst = $model_sbst ...\n\n";

	    $model_sbst = 'JC'; # Set it to the default value.
	}

	if (($model_sbst eq 'JC') or ($model_sbst eq 'K2P') or ($model_sbst eq 'K3P')) {
	    print STDOUT "[INFORMATION: 'read_dawg_ctrl_file_sbst'] Because \$model_sbst eq '$model_sbst', set all base frequencies to be 1/4, REGARDLESS of what values the control file specifies ...\n\n";

	    @nt_freqs = (1/4, 1/4, 1/4, 1/4); # These models assign frequency 1/4 to every base; if other values are specified in the control file, overwrite the values.

	} elsif  (($model_sbst eq 'GTR') or ($model_sbst eq 'HKY') or ($model_sbst eq 'F81') or ($model_sbst eq 'F84') or ($model_sbst eq 'TN')) {

	    my $ct_nt_freqs = @nt_freqs;
	    my $cnct_nt_freqs = join (', ', @nt_freqs);

	  if ($ct_nt_freqs == 4) {
	    my $tot_nt_freqs = 0;
	    foreach my $freq (@nt_freqs) { $tot_nt_freqs += $freq; }
	    unless ($tot_nt_freqs == 1) {
		print STDOUT "[ERROR 'read_dawg_ctrl_file_sbst'] The base frequencies (= $cnct_nt_freqs) specified by the control file WON'T SUM UP TO 1 (unity), but to $tot_nt_freqs: $! !\n\n";
		return (undef, undef, undef);
	    }
	  } elsif ($ct_nt_freqs == 0) {
		print STDOUT "[WARNING 'read_dawg_ctrl_file_sbst'] The base frequencies is UNSPECIFIED, while the substitution model is '$model_sbst'!! ... Are you sure of it? ... If so, I will use the average base frequencies of the input sequences as a substitite ...\n\n";
	  } else {
		print STDOUT "[ERROR 'read_dawg_ctrl_file_sbst'] The control file specifies an UNEXPECTED NUMBER (=$ct_nt_freqs) of base frequencies (= $cnct_nt_freqs ): $! !\n\n";
		return (undef, undef, undef);
	  }

	} else {
	    print STDOUT "[ERROR 'read_dawg_ctrl_file_sbst'] The control file specifies an UNKNOWN VALUE (= '$model_sbst') for the 'Model': $! !\n\n";
	    return (undef, undef, undef);
	}

	my $ct_sbst_params = @sbst_params;
	if (($model_sbst eq 'JC') or ($model_sbst eq 'F81')) {
	    if ($ct_sbst_params>0) {
		print STDOUT "[WARNING: 'read_dawg_ctrl_file_sbst'] Because \$model_sbst eq '$model_sbst', I will IGNORE the values, (", join (', ', @sbst_params), "), which is specified by the 'Params' line ... !\n\n";
	    }
	    @sbst_params = (); # Even if any values are specified by the control file, ignore these values.

	} elsif ($model_sbst eq 'GTR') {
	    unless ($ct_sbst_params == 6) {
		print STDOUT "[ERROR 'read_dawg_ctrl_file_sbst'] The number of the 'Param' parameters (=$ct_sbst_params) is NOT 6, which is required by the substitition model, '$model_sbst': $! !\n\n";
		return (undef, undef, undef);
	    }
	} elsif (($model_sbst eq 'K2P') or ($model_sbst eq 'HKY')) {
	    unless ($ct_sbst_params == 2) {
		print STDOUT "[ERROR 'read_dawg_ctrl_file_sbst'] The number of the 'Param' parameters (=$ct_sbst_params) is NOT 2, which is required by the substitition model, '$model_sbst': $! !\n\n";
		return (undef, undef, undef);
	    }
	} elsif (($model_sbst eq 'K3P') or ($model_sbst eq 'TN')) {
	    unless ($ct_sbst_params == 3) {
		print STDOUT "[ERROR 'read_dawg_ctrl_file_sbst'] The number of the 'Param' parameters (=$ct_sbst_params) is NOT 3, which is required by the substitition model, '$model_sbst': $! !\n\n";
		return (undef, undef, undef);
	    }
	} elsif ($model_sbst eq 'F84') {
	    unless ($ct_sbst_params == 1) {
		print STDOUT "[ERROR 'read_dawg_ctrl_file_sbst'] The number of the 'Param' parameters (=$ct_sbst_params) is NOT 1, which is required by the substitition model, '$model_sbst': $! !\n\n";
		return (undef, undef, undef);
	    }
	}

	return ($model_sbst, \@nt_freqs, \@sbst_params);

}	# End of the subroutine 'read_dawg_ctrl_file_sbst'. # 



# read_dawg_ctrl_file3:
#	This is a modified version of the subroutine 'read_dawg_ctrl_file'.
#	It can read in parameter settings of the insertion and deletion models
#	that are separately set.
#	[CREATED in creation of 'fa_lolipog_hs.alpha2.pl'.]
#	[=> IMPORTED from 'fa_lolipog_hs.alpha2.pl' in creation of 'anex.ver0.5.pl' (on 2019/06/11).]
#
#	[CORRECTED on 2020/05/11]:
#	Now, it can skip blank lines (consisting ONLY of white spaces).
#
# read_dawg_ctrl_file:
#	This subroutine reads a Dawg control file,
#	and retrieves indel parameters and a guide tree.
#
# INPUT: \$ctrl_file, where
#	$infile is the path to a Dawg control file.
#
# [OUTPUT: (\$tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate), where] --- For the predecessor, 'read_dawg_ctrl_file'.
# OUTPUT: (\$tree_string, $gapmodel_ins, $gapmodel_del, $gap_param_ins, $gap_param_del, $gaplen_cutoff_ins, $gaplen_cutoff_del, $ins_rate, $del_rate), where
#	$tree_string is a string representing the guide tree;
#	$gapmodel_ins/del specifies the gap model for the insertions/deletions;
#	$gap_param_ins/del is the parameter for the insertion/deletion model.
#	It is the power exponent if $gapmodel = '"PL"',
#	      the negative binomial parameter if $gapmodel = '"NB"';
#	$gaplen_cutoff_ins/del is the cut-off for the gap-length via an insertion/a deletion;
#	$ins/del_rate is the total insertion/deletion rate (per site).
#
sub read_dawg_ctrl_file3 (\$) {

	my ($ctrl_file) = @_;
	local $/ = "\n";

	my $fh_ctrl = open_input_file2 (${$ctrl_file});
	unless (defined $fh_ctrl) {
	    print STDOUT "[ERROR 'read_dawg_ctrl_file3'] 'open_input_file2' FAILED on the control file '${$ctrl_file}': $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

	my ($tree_string, $gapmodel_ins, $gapmodel_del, $gap_param_ins, $gap_param_del,
		$gaplen_cutoff_ins, $gaplen_cutoff_del, $ins_rate, $del_rate);

	while (my $line = <$fh_ctrl>) { # Main while-loop (over lines). #

	    ($line =~ m/\S/) or next; # Skip a blank line. (ADDED in CORRECTION on 2020/05/12.)

	    ($line =~ m/^\#/) and next; # Skip a comment line. #

	    $line =~ s/\s+$//;
	    $line =~ s/\;$//; # Added to prevent an error message (concerning $tot_indel_rate) on Mar 10 (Mon), 2014. #

	    my ($option_name, $value) = split /\s*\=\s*/, $line;

	    if ($option_name eq 'Tree') {
		$tree_string = $value;

	    } elsif ($option_name eq 'GapModel') {

			# MODIFIED in creation of 'read_dawg_ctrl_file3'. #
		if ($value =~ s/^\{\s*//) {
		   $value =~ s/\s*\}$//;
		   ($gapmodel_ins, $gapmodel_del) = split /\s*\,\s*/, $value;
		} else {
		    $gapmodel_ins = $gapmodel_del = $value;
		}
		## $gapmodel = $value;

	    } elsif ($option_name eq 'GapParams') {

		$value =~ s/^\{\s*//;
		$value =~ s/\s*\}$//;

			# MODIFIED in creation of 'read_dawg_ctrl_file3'. #
		if ($value =~ s/^\{\s*//) {

		   $value =~ s/\s*\}$//; 
		   ##my $value =~ s/\s*\}$//; # A BUG fixed in creation of 'fa_lolipog_hs.alpha2.pl'. #
		   my ($value_ins, $value_del) = split /\s*\}\s*\,\s*\{\s*/ , $value;
		   ($gap_param_ins, $gaplen_cutoff_ins) = split /\s*\,\s*/, $value_ins;
		   ($gap_param_del, $gaplen_cutoff_del) = split /\s*\,\s*/, $value_del;

		} else {

		   my ($gap_param, $gaplen_cutoff) = split /\s*\,\s*/, $value;
		   $gap_param_ins = $gap_param_del = $gap_param;
		   $gaplen_cutoff_ins = $gaplen_cutoff_del = $gaplen_cutoff;
		}
		## ($gap_param, $gaplen_cutoff) = split /\s*\,\s*/, $value;

	    } elsif ($option_name eq 'Lambda') {

			# MODIFIED in creation of 'read_dawg_ctrl_file3'. #
		if ($value =~ s/^\{\s*//) {
		   $value =~ s/\s*\}$//;
		   ($ins_rate, $del_rate) = split /\s*\,\s*/, $value;

		} else {
		   ($value =~ m/\d/) and ($ins_rate = $del_rate = $value/2);
		}
		## $tot_indel_rate = $value;
	    }

	}			# End of the main while-loop (over lines). #

	close ($fh_ctrl);

	foreach my $var ($tree_string, $gapmodel_ins, $gapmodel_del, $gap_param_ins, $gap_param_del, $gaplen_cutoff_ins, $gaplen_cutoff_del, $ins_rate, $del_rate) { # Trailing foreach-loop. #
	    (defined $var) and next;
	    return (undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}		# End of the trailing foreach-loop. #

	return (\$tree_string, $gapmodel_ins, $gapmodel_del, $gap_param_ins, $gap_param_del,
		$gaplen_cutoff_ins, $gaplen_cutoff_del, $ins_rate, $del_rate);

}	# End of the subroutine 'read_dawg_ctrl_file3'. # 


# readin_list_mgd_pcands:
#	This subroutine reads in a list of merged purge-candidate regions,
#	which is the main output of the script, 'detect_purge_cands.ver0.5.pl'.
#	[CREATED in creation of 'anex.ver0.5.pl' on 2019/06/20.]
#
# INPUT: (\$infile), where
#
#	$infile is a string specifying the path to the output file of 'detect_purge_cands.ver0.5.pl'.
#
# OUTPUTS: (\@list_mgd_pcands, $maxspan_mgd_pcand), where
#
#	@{$list_mgd_pcands[$pc]} = ($beg_pc, $end_pc, $br, \@msa_rows_under_br, \@constituent_sites, $pv, $pv0, \@sgn_wds_included) provides information on the $pc th (merged) purge-candidate region, where
#	$beg_pc & $end_pc are the sites at the beginning & end, respectively, of the region,
#	$br is the ID of the branch across which the "purge" is suspected within the region,
#	@msa_rows_under_br stores the indices of the MSA rows corresponding to the external offsprings of $br,
#	@constituent_sites stores the sites that constitute the region,
#	$pv is the P-value measuring the deviation from the specified substutution model,
#	$pv0 is the P-value measuring the deviation from the random-match model,
#	and @sgn_wds_included is the significant windows included in the region;
#	more precisely, 
#	@{$sgn_wds_included[$swd]} = (\@sites_in_wd, $pv_wd, $pv0_wd) provies information on the $swd th significant window, where
#	@sites_in_wd stores the sites in the window,
#	$pv_wd and $pv0_wd are the P-values for the window, measuring the same types of deviations as the aforementioned $pv and $pv0, respectively;
#
#	$maxspan_mgd_pcand is the maximum span among the merged purge-candidates.
#
sub readin_list_mgd_pcands (\$) {

	my ($rf_infile) = @_;

	my $infile = ${$rf_infile};

			# REVISED on 2020/05/06. #
	##unless (( -f $infile ) and ( -r $infile )){
	##   print STDOUT "[ERROR 'readin_list_mgd_pcands'] CanNOT FIND a readable input file '$infile': $! !\n";
	##   return (undef, undef);
	##}
	unless (( -f $infile ) and ( -r $infile )) {
	    my $infile_gz = "$infile.gz";
	    print STDOUT "[WARNING 'readin_list_mgd_pcands'] CanNOT FIND a readable input file '$infile': $! ... I will look for '$infile_gz' instead ... !\n";
	    unless (( -f $infile_gz) and ( -f $infile_gz)) {
		print STDOUT "[ERROR 'readin_list_mgd_pcands'] CanNOT FIND a readable input file '$infile_gz', EITHER: $! ! ... Sorry...\n";
		return (undef, undef);
	    }
	}
			# End of "REVISED on 2020/05/06." #

	local $/ = "\n"; # Use the newline as the input-record-separator.

	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    ##print STDOUT "[ERROR 'readin_list_mgd_pcands'] 'open_input_file2' FAILED on the input file '$infile': $! !\n";
	    print STDOUT "[ERROR 'readin_list_mgd_pcands'] 'open_input_file2' FAILED on the input file '$infile(.gz)': $! !\n";
	    return (undef, undef);
	}


	while (my $line = <$fh_in>) { # Remove decorative lines on the top. #
	    if (uc ($line) =~ m/^BEG_RG/) { last; }
	} # End of "Remove decorative lines on the top."

	my @list_mgd_pcands = ();
	my $maxspan_mgd_pcand = 0;

	while (my $line = <$fh_in>) { # Main while-loop (over input lines). #
	    if ($line =~ m/^\s+$/) { next; }

	    $line =~ s/^\s+$//;
	    $line =~ s/\s+$//;
	    my @fields = split /\t/, $line;
	    my $ct_flds = @fields;

	    if ($ct_flds<7) { # Modified so that it can also handle the reduced list, where no window may be listed in a line (on 2019/06/20). #
	    ##if ($ct_flds<8) {
		print STDOUT "[CAUTION 'readin_list_mgd_pcands'] An INSUFFICIENT number of FIELDS (=$ct_flds) in the line:\n";
		print STDOUT "\"$line\": $! !!\n\n";
		print STDOUT "   .... SKIP this line !! ...\n\n";
	    }

		# print $fh_out join ("\t", 'beg_rg', 'end_rg', 'br_ID', '[msa,rows,under,br]', '(sites,in,the,region)', 'pv', 'pv0', '{significant window included ([sites,in,wd]/pv_wd/pv0_wd)}_0', '{...}_1', '...'), "\n\n"; # Modified in revision (on 2019/06/19). # IMPORTED from 'detect_purge_cands.ver0.5.pl' for REFERENCE.

	    my ($beg_rg, $end_rg, $br, $msa_rows_ud_br, $constituent_sites, $pv, $pv0) = @fields[0 .. 6];

	    my $span_rg = $end_rg - $beg_rg + 1;
	    if ($maxspan_mgd_pcand < $span_rg) { $maxspan_mgd_pcand = $span_rg; }

	    $msa_rows_ud_br =~ s/^\[//;
	    $msa_rows_ud_br =~ s/\]$//;
	    my @msa_rows_ud_br = split /\,/, $msa_rows_ud_br;

	    $constituent_sites =~ s/^\(//;
	    $constituent_sites =~ s/\)$//;
	    my @constituent_sites = split /\,/, $constituent_sites;

	    my @sgn_wds_included = ();
	    for (my $f=7; $f<$ct_flds; $f++) { # Inner for-loop (over significant windows). #

		my ($sites_in_wd, $pv_wd, $pv0_wd) = split /\//, $fields[$f];
		$sites_in_wd =~ s/^\[//;
		$sites_in_wd =~ s/\]$//;
		my @sites_in_wd = split /\,/, $sites_in_wd;

		push @sgn_wds_included, [\@sites_in_wd, $pv_wd, $pv0_wd];

	    } # End of the inner for-loop (over significant windows). #

	    push @list_mgd_pcands, [$beg_rg, $end_rg, $br, \@msa_rows_ud_br, \@constituent_sites, $pv, $pv0, \@sgn_wds_included];
	    
	} # End of the main while-loop (over input lines). #

	close ($fh_in);

	return (\@list_mgd_pcands, $maxspan_mgd_pcand);

}	# End of the subroutine, 'readin_list_mgd_pcands'. #


# readin_set_clms:
#	This subroutine reads in a file
#	that records the set of columns in a (sub-)window & some information on the (sub-)window-boundaries.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl', on 2019/06/25.]
#	[REVISED on 2019/07/19]
#	Now, the subroutine can correctly handle the revision (on the same date) in 'print_out_set_clms'
#	regarding the record of $beg_wd and $end_wd.
#
# INPUT: (\$infile), where
#
#	$infile is a string specifying the path to the input file;
#
# OUTPUTS: ($beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, \@set_clms_in_wd), where
#
#	$beg/end_wd is the positions of the beginning/end of the window;
#
#	$if_on_le/re_msa = 1 if the window is on the left/right-end of the MSA,
#			 = 0 otherwise;
#
#	@column = @{$set_columns_in_wd[$c]} specifies the $c th column in the input MSA within the window.
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input alignment
#	that accommodates the corresponding site of the $k th "sequence" (or row);
#
sub readin_set_clms (\$) {

	my ($rf_infile) = @_;
	my $infile = ${$rf_infile};

	unless (( -f $infile ) and ( -r $infile )){
	   print STDOUT "[ERROR 'readin_set_clms'] CanNOT FIND a readable input file '$infile': $! !\n";
	   return (undef, undef, undef, undef, undef);
	}

	local $/ = "\n"; # Use the newline as the input-record-separator.

	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_set_clms'] 'open_input_file2' FAILED on the input file '$infile': $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

	my @indices_rows = ();
	my ($beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa);

	while (my $line = <$fh_in>) { # 1st main while-loop (to read in the header). #

	    $line =~ s/\s+$//;
	    $line = uc ($line);

	    if ($line =~ s/^COLUMN\\ROW\s+//) {
		@indices_rows = split /\t/, $line;
		last;
	    }

	    $line =~ s/^\#\s*//;
	    if ($line =~ s/^\[BEG\,\s+END\]\s*\=\s*//) {
		($beg_wd, $end_wd) = ($line =~ m/\[(\d+)\s*\,\s*(\d+)\]/);

		# ADDED in revision (on 2019/07/19). #
	    } elsif ($line =~ s/^\[BEG_WD\,\s+END_WD\]\S*\s*\=\s*//) {
		($beg_wd, $end_wd) = ($line =~ m/\[(\d+)\s*\,\s*(\d+)\]/);

			## print $fh_out "# [beg_wd, end_wd]_input_MSA = [$beg_wd, $end_wd] .\n"; # REVISED on 2019/07/19. # (<= This is a line in 'print_out_set_clms'). #
		# End of "ADDED in revision (on 2019/07/19)." #

	    } elsif ($line =~ s/IF_ON_LE_MSA\s*\=\s*//) {
		($if_on_le_msa) = ($line =~ m/(\S+)/);
	    } elsif ($line =~ s/IF_ON_RE_MSA\s*\=\s*//) {
		($if_on_re_msa) = ($line =~ m/(\S+)/);
	    }

	} # End of the 1st main while-loop (to read in the header). #

	my $ct_rows = @indices_rows;


	my @set_clms_in_wd = ();

	while (my $line = <$fh_in>) { # 2nd main while-loop (to read in the bulk). #

	    unless ($line =~ m/^\S/) { next; }

	    $line =~ s/\s+$//;
	    my @column = split /\t/, $line;
	    my $indx_clm = shift @column;

	    $set_clms_in_wd[$indx_clm] = \@column;

	} # End of the 2nd main while-loop (to read in the bulk). #

	close ($fh_in);

	return ($beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, \@set_clms_in_wd);

}	# End of the subroutine, 'readin_set_clms'. #


# readin_list_gblks:
#	This subroutine reads in a file
#	that lists gap-blocks, accompanied by their identifying information.
#	It was created by modifying 'readin_set_clms'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl', on 2019/06/26.]
#
# readin_set_clms:
#	This subroutine reads in a file
#	that records the set of columns in a (sub-)window & some information on the (sub-)window-boundaries.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl', on 2019/06/25.]
#
# INPUT: (\$infile), where
#
#	$infile is a string specifying the path to the input file;
#
# OUTPUTS: (\@list_gblks), where
#
#	@{$list_gblks[$indx_gb]} = ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, \@affd_exnodes_org, \@affd_msa_rows),
#	provides information on the $indx_gb th gap-block.
#	More precisely,
#	$br_org is the branch ID according to the original ID assignment of the tree nodes (used for the input),
#	$up_or_lw = 'U'/'L' if the gap-block is on the upper-/lower-side of the branch,
#	$lm/rm_clm is the left-/right-most column of the gap-block,
#	@affd_exnodes_org stores the IDs of the affected external nodes according to the (aforementioned) original ID assignment, and
#	$cnct_affd_exnodes_org = join (',', @affd_exnodes_org),
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block, and
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows).
#
sub readin_list_gblks (\$) {

	my ($rf_infile) = @_;

	my $infile = ${$rf_infile};
	unless (( -f $infile ) and ( -r $infile )){
	   print STDOUT "[ERROR 'readin_list_gblks'] CanNOT FIND a readable input file '$infile': $! !\n";
	   return (undef);
	}

	local $/ = "\n"; # Use the newline as the input-record-separator.

	my @list_gblks = ();

	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_list_gblks'] 'open_input_file2' FAILED on the input file '$infile': $! !\n";
	    return (undef);
	}

	while (my $line = <$fh_in>) { # 1st while-loop. #
	    if (uc ($line) =~ m/^INDX_GB/) { last; } # Skip the header.
	} # End of the 1st while-loop. #

	while (my $line = <$fh_in>) { # 2nd while-loop. #

	   unless ($line =~ m/^\S/) { next; }
	   $line =~ s/\s+$//;

	   my @fields = split /\t/, $line;

	   my ($indx_gb, $br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows) = @fields;

	   $cnct_affd_exnodes_org =~ s/^\(//;
	   $cnct_affd_exnodes_org =~ s/\)$//;

	   $cnct_affd_msa_rows =~ s/^\[//;
	   $cnct_affd_msa_rows =~ s/\]$//;

	   my @affd_exnodes_org = split /\,/, $cnct_affd_exnodes_org;
	   my @affd_msa_rows = split /\,/, $cnct_affd_msa_rows;

	   $list_gblks[$indx_gb] = [$br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, \@affd_exnodes_org, \@affd_msa_rows];

	} # End of the 2nd while-loop. #

	close ($fh_in);

	return (\@list_gblks);

}	# End of the subroutine, 'readin_list_gblks'. #


# readin_features_gblks:
#	This subroutine reads in a list of gap-blocks, usually named '(xxx_)features_gblocks.txt',
#	which follows a format only slightly different from that for the list read by 'readin_list_gblks'.
#	It was created by slightly modifying 'readin_list_gblks'.
#	[CREATED in creation of 'coordinate_point_lcl_msa.ver0.7.pl', on 2020/06/04.]
#
# readin_list_gblks:
#	This subroutine reads in a file, usually named 'list_gblocks(0).txt',
#	that lists gap-blocks, accompanied by their identifying information.
#	It was created by modifying 'readin_set_clms'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl', on 2019/06/26.]
#
# readin_set_clms:
#	This subroutine reads in a file
#	that records the set of columns in a (sub-)window & some information on the (sub-)window-boundaries.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl', on 2019/06/25.]
#
# INPUT: (\$infile), where
#
#	$infile is a string specifying the path to the input file;
#
# OUTPUTS: (\@features_gblks), where
#
#	@{$features_gblks[$indx_gb]} = ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_msa_rows, \@affd_msa_rows),
#	provides information on the $indx_gb th gap-block.
#	More precisely,
#	$br_org is the branch ID according to the original ID assignment of the tree nodes (used for the input),
#	$up_or_lw = 'U'/'L' if the gap-block is on the upper-/lower-side of the branch,
#	$lm/rm_clm is the left-/right-most column of the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block, and
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows).
#
sub readin_features_gblks (\$) {

	my ($rf_infile) = @_;

	my $infile = ${$rf_infile};
	unless (( -f $infile ) and ( -r $infile )){
	   print STDOUT "[ERROR 'readin_features_gblks'] CanNOT FIND a readable input file '$infile': $! !\n";
	   return (undef);
	}

	local $/ = "\n"; # Use the newline as the input-record-separator.

	my @features_gblks = ();

	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_features_gblks'] 'open_input_file2' FAILED on the input file '$infile': $! !\n";
	    return (undef);
	}

	while (my $line = <$fh_in>) { # 1st while-loop. #
	    if (uc ($line) =~ m/^INDX_GB/) { last; } # Skip the header.
	} # End of the 1st while-loop. #

	while (my $line = <$fh_in>) { # 2nd while-loop. #

	   unless ($line =~ m/^\S/) { next; }
	   $line =~ s/\s+$//;

	   my @fields = split /\t/, $line;

	   ##my ($indx_gb, $br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows) = @fields;
	   my ($indx_gb, $br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_msa_rows) = @fields; # Modified in creation of 'readin_features_gblks' (on 2020/06/04). (1) #

	   ##$cnct_affd_exnodes_org =~ s/^\(//; # Disabled in creation of 'readin_features_gblks' (on 2020/06/04). #
	   ##$cnct_affd_exnodes_org =~ s/\)$//; # Disabled in creation of 'readin_features_gblks' (on 2020/06/04). #

	   $cnct_affd_msa_rows =~ s/^\[//;
	   $cnct_affd_msa_rows =~ s/\]$//;

	   ##my @affd_exnodes_org = split /\,/, $cnct_affd_exnodes_org; # Disabled in creation of 'readin_features_gblks' (on 2020/06/04). #
	   my @affd_msa_rows = split /\,/, $cnct_affd_msa_rows;

	   ##$list_gblks[$indx_gb] = [$br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, \@affd_exnodes_org, \@affd_msa_rows];
	   $features_gblks[$indx_gb] = [$br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_msa_rows, \@affd_msa_rows]; # Modified in creation of 'readin_features_gblks' (on 2020/06/04). (2) #

	} # End of the 2nd while-loop. #

	close ($fh_in);

	return (\@features_gblks);

}	# End of the subroutine, 'readin_features_gblks'. #



	##### (II) DEFINITIONS of the Subroutines For (test) Outputs. #####

# print_out_input_tree:
#	This subroutine prints out an input tree,
#	possibly in various forms, if necessary.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/10]
#
# INPUTS: (\$outdir, $topnode_id, \%node2ch, \%node2pa, \%id2name, \%br_attr, $topnode_name), where
#
#	$outdir is a string specifying the path to the directory the output file will be placed in;
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but the root;
#	%id2name = ($node_id => $sequence_name, ...) for all external nodes;
#	%br_attr = ($node_id => [$length, $int_node_name, ...], ...) for all nodes at the lowe-ends of the branches;
#	$topnode_name is the name of the top-node (root).
#
# OUTPUTS: $flag_end, where
#
#	$flag_end = 1 if the subroutine ended without any conspicuous problems,
#		  = 0 if it failed.
#
sub print_out_input_tree (\$$\%\%\%\%$) {

	my ($rf_out_in_chk, $topnode_id, $node2ch, $node2pa, $id2name, $br_attr, $topnode_name) = @_;

	my $out_in_chk = ${$rf_out_in_chk};
	unless ( -d $out_in_chk )  {
	  unless (0 == system ("mkdir -p $out_in_chk") ) {
	    print STDOUT "[ERROR 'print_out_input_tree'] FAILED to create the directory \"$out_in_chk\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $out_in_chk ) {
	  unless (0 == system (" chmod -R u+w $out_in_chk " ) ) {
	    print STDOUT "[ERROR 'print_out_input_tree'] FAILED to change the permission mode of the output directory \"$out_in_chk\": $! !\n\n";
	    return 0;
	  }
	}

	my $chk_itree_file = "$out_in_chk/in_tree.txt";

	my $fh_ci_tree = open_output_file ($chk_itree_file, 0);
	unless (defined $fh_ci_tree) {
		print STDOUT "[ERROR 'print_out_input_tree'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$chk_itree_file\": $! !\n";
		return 0;
	}

	print STDOUT "[ Input Tree (showing interior node names, if at all) (Top_node = $topnode_id) ]\n\n";
	print $fh_ci_tree "[ Input Tree (showing interior node names, if at all) (Top_node = $topnode_id) ]\n\n";

	unless ( br_output_tree_dawg ($topnode_id, %{$node2ch}, %{$id2name}, %{$br_attr}, $topnode_name, *${$fh_ci_tree}, 1, 1) ) {
		print STDOUT "[ERROR 'print_out_input_tree'] Subroutine 'br_output_tree_dawg' FAILED to output the input tree into \"$chk_itree_file\": $! !\n";
		return 0;
	}
	print $fh_ci_tree "\n\n";

	print STDOUT "[ Input Tree (with node IDs shown) ]:\n\n";
	print $fh_ci_tree "[ Input Tree (with node IDs shown) ]:\n\n";
	my %br2blength;
	foreach my $br (keys %{$br_attr}) {
		my @attr = ();
		$attr[$INDX_LENGTH] = $br_attr->{$br}->[$INDX_LENGTH];
		$br2blength{$br} = \@attr;
	}

	unless ( br_output_tree ($topnode_id, %{$node2ch}, %{$id2name}, %br2blength, *${$fh_ci_tree}, 1, 1) ) {
		print STDOUT "[ERROR 'print_out_input_tree'] The subroutine 'output_tree' FAILED to print out the input tree into \"$chk_itree_file\": $! !\n";
		return 0;
	}
	print $fh_ci_tree "\n\n";

	print STDOUT "\n ... the input tree has been written into the file, \"$chk_itree_file\" ...\n\n";

  if ($TEST_LEVEL_MAIN>2) {

	print STDOUT "[ Relationships between Nodes ]\n\n";
	print $fh_ci_tree "[ Relationships between Nodes ]\n\n";
	print $fh_ci_tree join ("\t", 'Node_ID', 'parent', '[child,node,IDs]'), "\n\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2pa}, $topnode_id) {
		my $pa = $node2pa->{$node};
		(defined $pa) or ($pa = 'undef');
		my $children = $node2ch->{$node};
		my $cnct_children = (defined $children) ? join (',', @{$children}) : '';
		print $fh_ci_tree join ("\t", $node, $pa, "[$cnct_children]"), "\n";
	}
	print $fh_ci_tree "\n\n";

	print STDOUT "\n ... Some structures of the input tree have been appended to the file, \"$chk_itree_file\" ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>2) {...}"

	close ($fh_ci_tree);

	return 1;

}	# End of the subroutine, 'print_out_input_tree'. #


# print_out_br2log_mfacs:
#	This subroutine prints out the contents of a hash, %br2log_mfacs,
#	which is an output of 'assign_br2log_mfacs'.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/10]
#
#	[MAJOR REVISION in creation of 'ANEX_P.ver0.7' (on 2020/04/26)]:
#	Now, the subroutine can print out
#	the log-multiplication factors for case-(iv) "gapped segment"s as well,
#	so that ANEX can take full advantage of the main outputs of LASTPIECE.
#
# INPUTS: (\$out_in_chk, \%br2log_mfacs, $cutoff_ins_len, $cutoff_del_len), where
#
#	$out_in_chk is a string specifying the path to the directory that will store the files recording the contents of %br2log_mfacs;
#
#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#
#	[ @setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with ] ... for versions older than 'ANEX_P.ver0.7'.
#
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd, $log_mfac_noeffindels, $log_mfac_noeffindels_bd), with # REVISED in creation of 'ANEX_P.ver0.7' (on 2020/04/26). #
#
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,
#
#	$log_mfacs_ins/del_bd[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertions/deletions of $l sites on either boundary of the MSA,
#
#	$log_mfacs_pair[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites in the bulk of the MSA,
#
#	$log_mfacs_pair_bd[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites on either boundary of the MSA.
#	
#	$cutoff_ins_len and $cutoff_del_len are the cut-off values of the lengths of the (effective) insertions and deletions, respectively.
#
#		# (ADDED in creation of 'ANEX_P.ver0.7' (on 2020/04/26).) #

#	$log_mfac_noeffindels being the logarithm of the multiplication factor for a pair of PASs in the bulk of the MSA,
#	with NO effective insertions or deletions in between.
#
#	$log_mfac_noeffindels_bd being the logarithm of the multiplication factor for a PAS on either end of the MSA,
#	with NO effective insertions or deletions immediately outside.
#
#	[NOTE: TO keep the backward compatibility, $log_mfac_nodeffindels and $log_mfac_noeffindels_bd can be undefined.]
#		# End of "(ADDED in creation of 'ANEX_P.ver0.7' (on 2020/04/26).)" #
#
# OUTPUTS: $fl_end, where
#
#	$fl_end	= 1 if the subroutin ends without any particular problems,
#		= 0 if the subroutine failed.
#
sub print_out_br2log_mfacs (\$\%$$) {

	my ($rf_out_in_chk, $br2log_mfacs, $cutoff_ins_len, $cutoff_del_len) = @_;

	my $dir_chk_br2lmfacs = "${$rf_out_in_chk}/Br2log_mfacs";
	unless ( -d $dir_chk_br2lmfacs ) {
	  unless (0 == system ("mkdir -p $dir_chk_br2lmfacs") ) {
	    print STDOUT "[ERROR 'print_out_br2log_mfacs'] FAILED to create the directory '$dir_chk_br2lmfacs': $! ($?) !\n";
	    return 0;
	  }
	}
	unless ( -w $dir_chk_br2lmfacs ) {
	  unless (0 == system (" chmod -R u+w $dir_chk_br2lmfacs ") ) {
	    print STDOUT "[ERROR 'print_out_br2log_mfacs'] FAILED to change the permission mode of the output directory '$dir_chk_br2lmfacs': $!  ($?) !\n";
	    return 0;
	  }
	}

	my $pref_sgl_evs = "lmfacs_sgl_events";
	my $pref_idpairs_bulk = "lmfacs_insdelpairs_bulk";
	my $pref_idpairs_bound = "lmfacs_insdelpairs_bound";

	my $pref_no_eff_evs = "lmfacs_no_eff_events"; # ADDED in creation of 'ANEX_P.ver0.7' (on 2020/04/26.) #

	print STDOUT "[[ Output of 'assign_br2log_mfacs': Contents of ", '%{$br2log_mfacs}', " ]]\n\n";

	print STDOUT " ... will be recorded in the following files:\n\n";

	print STDOUT "Storing directory: {Dir_chk_br2lmfacs} = $dir_chk_br2lmfacs .\n\n";
	print STDOUT "Multiplication factors for single events (along branch B): {Dir_chk_br2lmfacs}/$pref_sgl_evs.B.txt .\n";
	print STDOUT "Multiplication factors for insertion-deletion pairs (in bulk) (along branch B): {Dir_chk_br2lmfacs}/$pref_idpairs_bulk.B.txt .\n";
	print STDOUT "Multiplication factors for insertion-deletion pairs (on either MSA boundary) (along branch B): {Dir_chk_br2lmfacs}/$pref_idpairs_bound.B.txt .\n";

	print STDOUT "Multiplication factors for NO effective-events (along branch B): {Dir_chk_br2lmfacs}/$pref_no_eff_evs.B.txt .\n"; # ADDED in creation of 'ANEX_P.ver0.7' (on 2020/04/26). #

	print STDOUT "\n\n";

	foreach my $br (sort {$a <=> $b} keys %{$br2log_mfacs}) { # Outer foreach-loop (over branches).

	    print STDOUT "{ Branch $br }\n\n";

	    ##my ($log_mfacs_ins, $log_mfacs_del, $log_mfacs_pair, $log_mfacs_ins_bd, $log_mfacs_del_bd, $log_mfacs_pair_bd) = @{$br2log_mfacs->{$br}};
	    my ($log_mfacs_ins, $log_mfacs_del, $log_mfacs_pair, $log_mfacs_ins_bd, $log_mfacs_del_bd, $log_mfacs_pair_bd, $log_mfac_no_eff_indels, $log_mfac_no_eff_indels_bd) = @{$br2log_mfacs->{$br}}; # MAJOR REVISION in creation of 'ANEX_P.ver0.7' (on 2020/04/26). 

	    my $chk_sgl_evs = "$dir_chk_br2lmfacs/$pref_sgl_evs.$br.txt";
	    ##my $fh_c_sgl = open_output_file ($chk_sgl_evs, 1);
	    my $fh_c_sgl = open_output_file ($chk_sgl_evs, 0); # Modified on 2020/04/26. #
	    unless (defined $fh_c_sgl) {
		print STDOUT "[ERROR 'print_out_br2log_mfacs'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$chk_sgl_evs\": $! !\n";
		return 0;
	    }

	    my $cutoff_max = ($cutoff_ins_len > $cutoff_del_len) ? $cutoff_ins_len : $cutoff_del_len ;

	    print $fh_c_sgl join ("\t", 'Length\log_mfacs', 'ins(bulk)', 'del(bulk)', 'ins(bound)', 'del(bound)'), "\n\n";

	    for (my $l=1; $l <= $cutoff_max; $l++) {
		my $lmfac_ins = $log_mfacs_ins->[$l];
		unless (defined $lmfac_ins) { $lmfac_ins = 'undef'; }
		my $lmfac_del = $log_mfacs_del->[$l];
		unless (defined $lmfac_del) { $lmfac_del = 'undef'; }
		my $lmfac_ins_bd = $log_mfacs_ins_bd->[$l];
		unless (defined $lmfac_ins_bd) { $lmfac_ins_bd = 'undef'; }
		my $lmfac_del_bd = $log_mfacs_del_bd->[$l];
		unless (defined $lmfac_del_bd) { $lmfac_del_bd = 'undef'; }

		print $fh_c_sgl join ("\t", $l, $lmfac_ins, $lmfac_del, $lmfac_ins_bd, $lmfac_del_bd), "\n";
	    }
	    print $fh_c_sgl "\n";
	    close ($fh_c_sgl);

	    my $chk_pairs_blk = "$dir_chk_br2lmfacs/$pref_idpairs_bulk.$br.txt";
	    my $fh_c_pairs_blk = open_output_file ($chk_pairs_blk, 1);
	    unless (defined $fh_c_pairs_blk) {
		print STDOUT "[ERROR 'print_out_br2log_mfacs'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$chk_pairs_blk\": $! !\n";
		return 0;
	    }
	    print $fh_c_pairs_blk join ("\t", 'L_A\L_D', 1 .. $cutoff_ins_len), "\n\n"; # DEBUGGED on 2019/07/07.
	    ##print $fh_c_pairs_blk join ("\t", 'L_A\L_D', 1 .. $CUTOFF_INS_LEN), "\n\n";

	    for (my $la=1; $la <= $cutoff_del_len; $la++) {
		my $row_lmfacs_pair = $log_mfacs_pair->[$la];
		unless ($row_lmfacs_pair) {
		    print $fh_c_pairs_blk $la, "\n";
		    next;
		}
		my @out_lmfacs = ();
		for (my $ld=1; $ld <= $cutoff_ins_len; $ld++) {
		    my $lmfac = $row_lmfacs_pair->[$ld];
		    push @out_lmfacs, ((defined $lmfac) ? $lmfac : 'undef');
		}
		print $fh_c_pairs_blk join ("\t", $la, @out_lmfacs), "\n";
	    }
	    print $fh_c_pairs_blk "\n";
	    close ($fh_c_pairs_blk);

	    my $chk_pairs_bd = "$dir_chk_br2lmfacs/$pref_idpairs_bound.$br.txt";
	    my $fh_c_pairs_bd = open_output_file ($chk_pairs_bd, 1);
	    unless (defined $fh_c_pairs_bd) {
		print STDOUT "[ERROR 'print_out_br2log_mfacs'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$chk_pairs_bd\": $! !\n";
		return 0;
	    }
	    print $fh_c_pairs_bd join ("\t", 'L_A\L_D', 1 .. $cutoff_ins_len), "\n\n"; # DEBUGGED on 2019/07/07.
	    ##print $fh_c_pairs_bd join ("\t", 'L_A\L_D', 1 .. $CUTOFF_INS_LEN), "\n\n";

	    for (my $la=1; $la <= $cutoff_del_len; $la++) {
		my $row_lmfacs_pair_bd = $log_mfacs_pair_bd->[$la];
		unless ($row_lmfacs_pair_bd) {
		    print $fh_c_pairs_bd $la, "\n";
		    next;
		}
		my @out_lmfacs = ();
		for (my $ld=1; $ld <= $cutoff_ins_len; $ld++) {
		    my $lmfac = $row_lmfacs_pair_bd->[$ld];
		    push @out_lmfacs, ((defined $lmfac) ? $lmfac : 'undef');
		}
		print $fh_c_pairs_bd join ("\t", $la, @out_lmfacs), "\n";
	    }
	    print $fh_c_pairs_bd "\n";
	    close ($fh_c_pairs_bd);


			# (ADDED in creation of 'ANEX_P.ver0.7' (on 2020/04/26).) (MAJOR REVISION 2) #

	    if ((defined $log_mfac_no_eff_indels) or (defined $log_mfac_no_eff_indels_bd)) { # This condition guarantees the backward compatibility of this subroutine. #

		my $chk_noeff_evs = "$dir_chk_br2lmfacs/$pref_no_eff_evs.$br.txt";
		my $fh_c_ne = open_output_file ($chk_noeff_evs, 0);
		unless (defined $fh_c_ne) {
		    print STDOUT "[ERROR 'print_out_br2log_mfacs'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$chk_noeff_evs\": $! !\n";
		    return 0;
		}

		print $fh_c_ne "[ Log-Multiplication Factors for cases where NO effective indels are in between a pair of PASs: ]\n\n";

		print $fh_c_ne ' (In the Bulk)   ', ((defined $log_mfac_no_eff_indels) ? $log_mfac_no_eff_indels : 'undef'), " .\n\n";

		print $fh_c_ne ' (On either end) ', ((defined $log_mfac_no_eff_indels_bd) ? $log_mfac_no_eff_indels_bd : 'undef'), " .\n\n";
		
		close ($fh_c_ne);

	    } # End of "if ((defined $log_mfac_no_eff_indels) or (defined $log_mfac_no_eff_indels_bd)) {...}"

			# END of "(ADDED in creation of 'ANEX_P.ver0.7' (on 2020/04/26).) (MAJOR REVISION 2)" #


	    print STDOUT "\n ... Ended recording All the log-multiplication factors assigned to branch $br !! ...\n\n"; 

	} # End of the outer foreach-loop (over branches). #

	print STDOUT "\n ... ENDED Recording ALL the log-multiplication factors assigned to ALL branches!! ...\n\n\n";

	return 1;

}	# End of the subroutine, 'print_out_br2log_mfacs'. #


# print_out_tree_profile:
#	This subroutine prints out some variables profiling a tree,
#	most of which are outputs of 'enum_psm_indel_hsts_for_msa_finer_PRE'.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/10]
#
# INPUTS: (\$out_interim, \%node2depth, \%node2ct_exoffs, \%br2eq, \%br2nn, \@set_all_nodes), where
#
#	$out_interim is a string specifying the path to the directory the output file will be placed in;
#
#	%node2depth = ($node_id => $depth, ...) for all nodes;
#	%node2ct_exoffs = ($node_id => $ct_exoffs, ...) for all nodes;
#	%br2eq = ($br_id => $equiv_br_id, ...) for branches with equivalents;
#	%br2nn = ($br_id => {$id_of_a_neighbor => 1, ...}, ...);
#	@set_all_nodes stores the IDs of all nodes;
#
# OUTPUTS: $flag_end, where
#
#	$flag_end = 1 if the subroutine ended without any conspicuous problem,
#		  = 0 if it failed.
#
sub print_out_tree_profile (\$\%\%\%\%\@) {

	my ($rf_out_interim, $node2depth, $node2ct_exoffs, $br2eq, $br2nn, $set_all_nodes) = @_;

	my $out_interim = ${$rf_out_interim};
	unless ( -d $out_interim )  {
	  unless (0 == system ("mkdir -p $out_interim") ) {
	    print STDOUT "[ERROR 'print_out_tree_profile'] FAILED to create the directory \"$out_interim\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $out_interim ) {
	  unless (0 == system (" chmod -R u+w $out_interim " ) ) {
	    print STDOUT "[ERROR 'print_out_tree_profile'] FAILED to change the permission mode of the output directory \"$out_interim\": $! !\n\n";
	    return 0;
	  }
	}
	my $file_tree_profile = "$out_interim/tree_profile.txt";

	my $fh_tr_prfl = open_output_file ($file_tree_profile, 0);
	unless (defined $fh_tr_prfl) {
		print STDOUT "[ERROR 'print_out_tree_profile'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$file_tree_profile\" !\n";
		print STDERR "[ERROR 'print_out_tree_profile'] $! !\n";
		die ;
	}

	print $fh_tr_prfl "[[ Outputs of 'enum_psm_indel_hsts_for_msa_finer_PRE' ]]\n\n";

	print $fh_tr_prfl '@{$set_all_nodes} = (', join (', ', @{$set_all_nodes}), ") .\n\n";

	print $fh_tr_prfl '[ Contents of %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq} & %{$br2nn}: ]', "\n\n";
	print $fh_tr_prfl join ("\t", 'Node/br_ID', 'depth', '#{external-offsprings}', 'equiv_br_ID', '[nearrest,neighboring,branches]'), "\n\n";

	foreach my $node_br (@{$set_all_nodes}) {

	    my $depth = $node2depth->{$node_br};
	    my $ct_exoffs = $node2ct_exoffs->{$node_br};
	    my $eqbr = $br2eq->{$node_br};
	    unless (defined $eqbr) { $eqbr = 'undef'; }
	    my $nnbrs2true = $br2nn->{$node_br};
	    my @nns = (defined $nnbrs2true) ? sort {$a <=> $b} keys %{$nnbrs2true} : () ;
	    my $cnct_nns = join (',', @nns);

	    print $fh_tr_prfl join ("\t", $node_br, $depth, $ct_exoffs, $eqbr, "[$cnct_nns]"), "\n";
	} 
	print $fh_tr_prfl "\n\n";

	close ($fh_tr_prfl);

	print STDOUT "\n ...  Outputs of 'enum_psm_indel_hsts_for_msa_finer_PRE' have been recorded into the file, '$file_tree_profile' ...\n\n\n";

	return 1;

}	# End of the subroutine, 'print_out_tree_profile'. #


# print_out_cal_init_psm_cands_for_sgl_msa:
#	This subroutine prints out the outputs of 'cal_init_psm_cands_for_sgl_msa'
#	into files in a directory under the top interim output directory specified.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/10.]
#
# INPUTS: (\$out_parent, \@info_gpatterns, \@info_init_psm_hst), where
#
#	$out_parent is a string specifying the path to the parent of the output directory;
#
#	(\@set_gpatterns, \@set_node2dparsstat, \@set_br2change) = @info_gpatterns ;
#	(\@set_cls_gpbs, \@init_totcts_indels_for_cls, \@init_psm_hsts_for_cls) = @info_init_psm_hst ;
#	[For details of these arrays, see the header of 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm".]
#
# OUTPUTS: $flag_end, where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#		  = 0 if it failed.
#
sub print_out_cal_init_psm_cands_for_sgl_msa (\$\@\@) {

	my ($out_parent, $info_gpatterns, $info_init_psm_hst) = @_;

	my ($set_cls_gpbs, $init_totcts_indels_for_cls, $init_psm_hsts_for_cls) = @{$info_init_psm_hst} ;
	my ($set_gpatterns, $set_node2dparsstat, $set_br2change) = @{$info_gpatterns} ;

	my $ct_gpd_segments = @{$set_cls_gpbs};
	my $ct_gp = @{$set_gpatterns};

	my $outdir = "${$out_parent}/Init_psm_cands";
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_cal_init_psm_cands_for_sgl_msa'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_cal_init_psm_cands_for_sgl_msa'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $pref_gpats = "gap_patterns";
	my $pref_gsegs_idhsts0 = "gsegs_indel_hsts0";
	my $out_gpatterns = "$outdir/$pref_gpats.txt";
	my $out_gsegs_idhsts0 = "$outdir/$pref_gsegs_idhsts0.txt";

	print STDOUT "[[[ Main Outputs of 'cal_init_psm_cands_for_sgl_msa' ]]]\n\n";
	print STDOUT " ... will be recorded in the following files:\n\n";
	print STDOUT "Storing directory: {Out_dir} = $outdir .\n";
	print STDOUT "Gap patterns: {Out_dir}/$pref_gpats.txt ,\n";
	print STDOUT "Gapped segments & initial indel histories: {Out_dir}/$pref_gsegs_idhsts0 .\n";
	print STDOUT "\n\n";

	##my $fh_gpats = open_output_file ($out_gpatterns, 1);
	my $fh_gpats = open_output_file ($out_gpatterns, 0); # Modified on 2020/05/06.
	unless (defined $fh_gpats) {
		print STDOUT "[ERROR 'print_out_cal_init_psm_cands_for_sgl_msa'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_gpatterns\": $! !\n";
		return 0;
	}


	print STDOUT "[1: List of gap patterns encountered (#{patterns} = $ct_gp) ]\n\n";
	print $fh_gpats "[1: List of gap patterns encountered (#{patterns} = $ct_gp) ]\n\n";

	print $fh_gpats join ("\t", 'Indx_gp', "'gap_patterns(horizontal_representation)'"), "\n\n";

    for (my $indx_gp=0; $indx_gp<$ct_gp; $indx_gp++) { # 1st preliminary outer for-loop (over the gap patterns) #

	my $gpattern = ${$set_gpatterns->[$indx_gp]};
	print $fh_gpats join ("\t", $indx_gp, "'$gpattern'"), "\n";

    }	# End of the 1st preliminary outer for-loop (over the gap patterns) #
	print $fh_gpats "\n\n";

	close ($fh_gpats);

	print STDOUT "\n ... Finished recording into \"$out_gpatterns\" ...\n\n";

	my $fh_gsegs_idhsts0 = open_output_file ($out_gsegs_idhsts0, 1);
	unless (defined $fh_gsegs_idhsts0) {
		print STDOUT "[ERROR 'print_out_cal_init_psm_cands_for_sgl_msa'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_gsegs_idhsts0\": $! !\n";
		return 0;
	}

	print STDOUT "\n[3: Gapped Segments = Clusters of gap-pattern-blocks (#{gapped segments} = $ct_gpd_segments) & the Initial Candidate of a Parsimonious History for each of the gapped segments ]\n\n\n";
	print $fh_gsegs_idhsts0 "\n[3: Gapped Segments = Clusters of gap-pattern-blocks (#{gapped segments} = $ct_gpd_segments) & the Initial Candidate of a Parsimonious History for each of the gapped segments ]\n\n\n";

    for (my $cl = 0; $cl<$ct_gpd_segments; $cl++) { # 2nd preliminary outer for-loop (over the gapped segments) #

	my $cls_gblks = $set_cls_gpbs->[$cl];
	my $ct_gblks = @{$cls_gblks};

	print $fh_gsegs_idhsts0 "< The $cl th gapped segment (#{gap-pattern-blocks} = $ct_gblks) >\n\n";

	print $fh_gsegs_idhsts0 "[ Gap-pattern-blocks in the cluster ]\n\n";

	print $fh_gsegs_idhsts0 join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal_representation)'"), "\n\n";

	for (my $bl=0; $bl<$ct_gblks; $bl++) { # 1st middle for-loop (over the gap-blocks) #
	    my ($indx_gp, $len_gb, $start, $end) = @{$cls_gblks->[$bl]};

	    my $gpattern = ${$set_gpatterns->[$indx_gp]};
	    print $fh_gsegs_idhsts0  join ("\t", $bl, $len_gb, $start, $end, $indx_gp, "'$gpattern'"), "\n";
	}			# End of the 1st middle for-loop (over the gap-blocks) #
	print $fh_gsegs_idhsts0 "\n\n";

## my ($set_cls_gpbs, $init_totcts_indels_for_cls, $init_psm_hsts_for_cls) = @{$info_init_psm_hst} ;

	my $init_totct_indel = $init_totcts_indels_for_cls->[$cl];
	my $init_psm_hst = $init_psm_hsts_for_cls->[$cl];

	print $fh_gsegs_idhsts0  "\n[ Initial Candidate Indel History (Total#{indel events}_init = $init_totct_indel) ]\n\n";

	my ($br2indels0, $node2bw_gstats0) = @{$init_psm_hst};

	print $fh_gsegs_idhsts0 "( Map of block-wise gap states onto the nodes )\n\n";
	print $fh_gsegs_idhsts0 join ("\t", 'Node_ID', "'block_wise_gap_states(concatenated)'"), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats0}) { # 1st inner foreach-loop #
	    my $cnct_bw_gstats = join ('', @{$node2bw_gstats0->{$node}});
	    print $fh_gsegs_idhsts0  join ("\t", $node, "'$cnct_bw_gstats'"), "\n";
	}				# End of the 1st inner foreach-loop #
	print $fh_gsegs_idhsts0  "\n\n";

	print $fh_gsegs_idhsts0  "( Map of indel events onto the branches )\n\n";
	print $fh_gsegs_idhsts0  join ("\t", 'Branch_ID', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2indels0}) { # 2nd inner foreach-loop #

	    my @out_indel_info = ();
	    foreach my $indel_info (@{$br2indels0->{$br}}) { # Innermost foreach-loop #
		my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
		(defined $stat_ue) or ($stat_ue = 'undef');
		(defined $stat_le) or ($stat_le = 'undef');
		my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef';
		push @out_indel_info, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
	    }			# End of the innermost foreach-loop #

	    print $fh_gsegs_idhsts0  join ("\t", $br, @out_indel_info), "\n";

	}				# End of the 2nd inner foreach-loop #
	print $fh_gsegs_idhsts0  "\n\n";

    }		# End of the 2nd preliminary outer for-loop (over the gapped segments) #

	close ($fh_gsegs_idhsts0);

	print STDOUT "\n ... Finished recording into \"$out_gsegs_idhsts0\" ...\n\n";

	return 1;

}	# End of the subroutine, 'print_out_cal_init_psm_cands_for_sgl_msa'. #


# print_out_bch_br_parsimonize_dollo_indel_histories
#	This subroutine prints out the outputs of 'bch_br_parsimonize_dollo_indel_histories'
#	into the files in a directory under a specified top directory.
#	[CREATED in creation of 'anex.ver0.5.pl, on 2019/06/11.]
#	[REVISED on 2019/06/17]: Now, the subroutine can specify the name of the output directory.
#	[DEBUGGED on 2019/07/07]: ADDED @{$set_gpatterns}, which had been missing from the list of arguments.
#
# [ INPUTS: (\$out_parent, \@seqnames, \@new_info_init_psm_hst, \@new_in_msa, \@set_new_lists_gblocks, \@sets_places_gpbs, \@sets_gpbs_at_spec_places), where ] ... OBSOLETE as of 2019/06/17.
# [ INPUTS: (\$out_parent, $outdir_name, $new_aln_filename, \@seqnames, \@new_info_init_psm_hst, \@new_in_msa, \@set_new_lists_gblocks, \@sets_places_gpbs, \@sets_gpbs_at_spec_places), where ] ... OBSOLETE as of 2019/07/07.
# INPUTS: (\$out_parent, $outdir_name, $new_aln_filename, \@seqnames, \@new_info_init_psm_hst, \@new_in_msa, \@set_new_lists_gblocks, \@sets_places_gpbs, \@sets_gpbs_at_spec_places, \@set_gpatterns), where # MODIFIED in DEBUGGING (on 2019/07/07).
#
#	$out_parent is a string specifying the path to the parent of the output directory;
#	$outdir_name is a string providing the name of the output directory; # Added on 2019/06/17.
#	$new_aln_filename is a string providing the name of the new alignment file; # Added on 2019/06/17.
#
#	@seqnames is an array storing the names of the sequences in the alignment;
#
#	The remaining 4 array-references are the outputs of 'bch_br_parsimonize_dollo_indel_histories'.
#	[For details, see the header of 'bch_br_parsimonize_dollo_indel_histories' (in "MyTreeMap_indels_spt_odr2.pm").]
#
#		# Added in DEBUGGING (on 2019/07/07). #
#	${$set_gpatterns[$indx_gp]} is the string representing the $indx_gp th gap-pattern;
#	[Typically, \@set_gpatterns is an output of 'cal_init_psm_cands_for_sgl_msa' (in "MyTreeMap_indels_spt_odr2.pm")];
#		# End of "Added in DEBUGGING (on 2019/07/07)." #
#
# OUTPUTS: $flag_end, where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#		  = 0 if it failed.
#
# [ sub print_out_bch_br_parsimonize_dollo_indel_histories (\$\@\@\@\@\@\@) {
#	my ($out_parent, $seqnames, $new_info_init_psm_hst, $new_in_msa, $set_new_lists_gblocks, $sets_places_gpbs, $sets_gpbs_at_spec_places) = @_; ] ... OBSOLETE as of 2019/06/17.
#
# [ sub print_out_bch_br_parsimonize_dollo_indel_histories (\$$$\@\@\@\@\@\@) {
#	my ($out_parent, $outdir_name, $new_aln_filename, $seqnames, $new_info_init_psm_hst, $new_in_msa, $set_new_lists_gblocks, $sets_places_gpbs, $sets_gpbs_at_spec_places) = @_; ] ... OBSOLETE as of 2019/07/07.
#
sub print_out_bch_br_parsimonize_dollo_indel_histories (\$$$\@\@\@\@\@\@\@) {

	my ($out_parent, $outdir_name, $new_aln_filename, $seqnames, $new_info_init_psm_hst, $new_in_msa, $set_new_lists_gblocks, $sets_places_gpbs, $sets_gpbs_at_spec_places, $set_gpatterns) = @_;

	my ($new_set_cls_gpbs, $new_init_totcts_indels_for_cls, $new_init_psm_hsts_for_cls) = @{$new_info_init_psm_hst} ;

	my $ct_gpd_segments = @{$new_set_cls_gpbs};
	##my $ct_gp = @{$set_gpatterns};

	my $outdir = "${$out_parent}/$outdir_name";
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_bch_br_parsimonize_dollo_indel_histories'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_bch_br_parsimonize_dollo_indel_histories'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	##my $new_aln_filename = "new_alignment.txt"; # DISABLED on 2019/06/17.
	my $pref_gsegs_idhsts0 = "gsegs_indel_hsts0_psmnzd";
	my $out_new_aln = "$outdir/$new_aln_filename";
	my $out_gsegs_idhsts0 = "$outdir/$pref_gsegs_idhsts0.txt";

	print STDOUT "[[[ Main Outputs of 'print_out_bch_br_parsimonize_dollo_indel_histories' ]]]\n\n";
	print STDOUT " ... will be recorded in the following files:\n\n";
	print STDOUT "Storing directory: {Out_dir} = $outdir .\n";
	print STDOUT "New alignments:  {Out_dir}/$new_aln_filename ,\n";
	print STDOUT "Gapped segments & initial indel histories, etc.: {Out_dir}/$pref_gsegs_idhsts0 .\n";
	print STDOUT "\n\n";

	my $fh_naln = open_output_file ($out_new_aln, 0); # Modified to Fix a BUG (on 2019/06/19).
	##my $fh_naln = open_output_file ($out_new_aln, 1);
	unless (defined $fh_naln) {
		print STDOUT "[ERROR 'print_out_bch_br_parsimonize_dollo_indel_histories'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_new_aln\": $! !\n";
		return 0;
	}

	print STDOUT "\n[1: Input Alignment after brief parsimonization. ]\n\n\n";
	## print $fh_naln "\n[1: Input Alignment after brief parsimonization. ]\n\n\n"; # DISABLED on 2019/06/17.

	unless ( pr_ODEN_input (@{$seqnames}, @{$new_in_msa}, $DIS_LEN, *${$fh_naln}) ) {
		print STDOUT "[ERROR 'print_out_bch_br_parsimonize_dollo_indel_histories'] The subroutine 'pr_ODEN_input'  FAILED to print out the NEW input MSA: $! !\n";
		return 0;
	}
	print STDOUT "\n ... The new alignment has been recorded into the file, \"$out_new_aln\" ...\n\n";
	## print $fh_naln "\n\n"; # DISABLED on 2019/06/17.

	close ($fh_naln);

	## $out_gsegs_idhsts0 = "$outdir/$pref_gsegs_idhsts0.txt";
	my $fh_gsegs_idhsts0 = open_output_file ($out_gsegs_idhsts0, 1);
	unless (defined $fh_gsegs_idhsts0) {
		print STDOUT "[ERROR 'print_out_bch_br_parsimonize_dollo_indel_histories'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_gsegs_idhsts0\": $! !\n";
		return 0;
	}


	print STDOUT "\n[3: New Gapped Segments = Clusters of gap-pattern-blocks (#{gapped segments} = $ct_gpd_segments) and the Parsimonized Initial Candidate of a Parsimonious History, as well as some auxiliary information, for each of the gapped segments ]\n\n\n";
	print $fh_gsegs_idhsts0 "\n[3: New Gapped Segments = Clusters of gap-pattern-blocks (#{gapped segments} = $ct_gpd_segments) and the Parsimonized Initial Candidate of a Parsimonious History, as well as some auxiliary information, for each of the gapped segments ]\n\n\n";


     for (my $cl = 0; $cl<$ct_gpd_segments; $cl++) { # Main outer for-loop (over the gapped segments) #


# my ($new_info_init_psm_hst, $new_in_msa, $set_new_lists_gblocks, $sets_places_gpbs, $sets_gpbs_at_spec_places)

	my $new_cls_gblks = $new_set_cls_gpbs->[$cl];
	my $ct_gblks = @{$new_cls_gblks};

	print $fh_gsegs_idhsts0  "< The $cl th new gapped segment (#{gap-pattern-blocks} = $ct_gblks) >\n\n";

	my $places_gpbs = $sets_places_gpbs->[$cl];
	my $gpbs_at_spec_plcs = $sets_gpbs_at_spec_places->[$cl];

	print $fh_gsegs_idhsts0  "[ Rearrangements of gap-pattern-blocks (compared to the input) ]\n\n";

	print $fh_gsegs_idhsts0  '@{$places_gpbs}       = (', join (', ', @{$places_gpbs}), "),\n\n";
	print $fh_gsegs_idhsts0  '@{$gpbs_at_spec_plcs} = (', join (', ', @{$gpbs_at_spec_plcs}), ").\n\n";

	print $fh_gsegs_idhsts0  "\n[ Gap-pattern-blocks in the new cluster ]\n\n";

	print $fh_gsegs_idhsts0  join ("\t", 'New_Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal_representation)'"), "\n\n";

	for (my $bl=0; $bl<$ct_gblks; $bl++) { # 1st middle for-loop (over the gap-blocks) #
	    my ($indx_gp, $len_gb, $start, $end) = @{$new_cls_gblks->[$bl]};

	    my $gpattern = ${$set_gpatterns->[$indx_gp]};

	    print $fh_gsegs_idhsts0 join ("\t", $bl, $len_gb, $start, $end, $indx_gp, "'$gpattern'"), "\n";
	}			# End of the 1st middle for-loop (over the gap-blocks) #
	print $fh_gsegs_idhsts0  "\n\n";

		# my ($new_set_cls_gpbs, $new_init_totcts_indels_for_cls, $new_init_psm_hsts_for_cls) = @{$new_info_init_psm_hst} ;

	my $new_init_totct_indel = $new_init_totcts_indels_for_cls->[$cl];
	my $new_init_psm_hst = $new_init_psm_hsts_for_cls->[$cl];

	print $fh_gsegs_idhsts0  "\n[ New Initial Candidate Indel History (Total#{indel events}_new_init = $new_init_totct_indel) ]\n\n";

	my ($new_br2indels0, $new_node2bw_gstats0) = @{$new_init_psm_hst};

			# CORRECTED on 2020/05/06. #
	##print STDOUT "( New Map of block-wise gap states onto the nodes )\n\n";
	##print STDOUT join ("\t", 'Node_ID', "'New_block_wise_gap_states(concatenated)'"), "\n\n";
	print $fh_gsegs_idhsts0  "( New Map of block-wise gap states onto the nodes )\n\n";
	print $fh_gsegs_idhsts0 join ("\t", 'Node_ID', "'New_block_wise_gap_states(concatenated)'"), "\n\n";
			# End of "CORRECTED on 2020/05/06." #

	foreach my $node (sort {$a <=> $b} keys %{$new_node2bw_gstats0}) { # 1st inner foreach-loop #
	    my $cnct_bw_gstats = join ('', @{$new_node2bw_gstats0->{$node}});

	    print $fh_gsegs_idhsts0  join ("\t", $node, "'$cnct_bw_gstats'"), "\n";
	}				# End of the 1st inner foreach-loop #
	print $fh_gsegs_idhsts0  "\n\n";

	print $fh_gsegs_idhsts0  "( New Map of indel events onto the branches )\n\n";
	print $fh_gsegs_idhsts0  join ("\t", 'Branch_ID', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$new_br2indels0}) { # 2nd inner foreach-loop #
	    my @out_indel_info = ();
	    foreach my $indel_info (@{$new_br2indels0->{$br}}) { # Innermost foreach-loop #
		my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
		(defined $stat_ue) or ($stat_ue = 'undef');
		(defined $stat_le) or ($stat_le = 'undef');
		my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef';
		push @out_indel_info, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
	    }			# End of the innermost foreach-loop #

	    print $fh_gsegs_idhsts0  join ("\t", $br, @out_indel_info), "\n";
	}				# End of the 2nd inner foreach-loop #
	print $fh_gsegs_idhsts0  "\n\n";


		# my ($new_info_init_psm_hst, $new_in_msa, $set_new_lists_gblocks, $sets_places_gpbs, $sets_gpbs_at_spec_places)

	my $new_list_gblocks = $set_new_lists_gblocks->[$cl];
	my $ct_gblocks = @{$new_list_gblocks};

	print $fh_gsegs_idhsts0  "\n[ List of New Gap-BLocks (#{gap-blocks} = $ct_gblocks) ]\n\n";

	print $fh_gsegs_idhsts0  join ("\t", 'Indx', 'branch_ID', 'upper(U)_or_lower(L)', 'lm_gpb', 'rm_gpb', '(involved,gap,pattern,blocks)', '[affected,external,nodes]', 'within_branch_indx'), "\n\n";

# @{$new_list_gblocks->[$gb]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e) 

	for (my $gb = 0; $gb < $ct_gblocks; $gb++) { # 3rd inner for-loop (over gap-blocks). #


	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs, $affected_exnodes, $e) = @{$new_list_gblocks->[$gb]};

	    my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs});
	    $cnct_invlvd_gpbs = "($cnct_invlvd_gpbs)";
	    my $cnct_affected_exnodes = join (',', @{$affected_exnodes});
	    $cnct_affected_exnodes = "[$cnct_affected_exnodes]";

	    print $fh_gsegs_idhsts0 join ("\t", $gb, $br, $up_or_lw, $lm_gpb, $rm_gpb, $cnct_invlvd_gpbs, $cnct_affected_exnodes), "\n";

	}	# End of the 3rd inner for-loop (over gap-blocks). #

	print $fh_gsegs_idhsts0  "\n\n\n";

    }		# End of the Main outer for-loop (over the gapped segments) #

	close ($fh_gsegs_idhsts0);

	print STDOUT "\n\n ... Information on the new gp-configurations after the parsimonization has been recorded into the file, \"$out_gsegs_idhsts0\" ...\n\n";

	return 1;

}	# End of the subroutine, 'print_out_bch_br_parsimonize_dollo_indel_histories'. #


# print_out_alignment_psmnzd:
#	This subroutine prints out the new MSA that resulted from the subroutine, 
#	'bch_br_parsimonize_dollo_indel_histories'.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/17]
#
# INPUTS: (\$outdir, $new_aln_filename, \@seqnames, \@new_in_msa), where
#
#	$outdir is a string specifying the path to the output directory;
#	$new_aln_filename is a string providing the name of the new alignment file; # Added on 2019/06/17.
#
#	@seqnames is an array storing the names of the sequences in the alignment;
#
#	@new_in_msa stores the alignment resulting from 'bch_br_parsimonize_dollo_indel_histories',
#	with ${$new_in_msa[$k]} is the $k th sequence embedded in the alignment;
#
# OUTPUT: $flag_end, where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#		  = 0 if it failed.
#
sub print_out_alignment_psmnzd (\$$\@\@) {

	my ($rf_outdir, $new_aln_filename, $seqnames, $new_in_msa) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	    unless (0 == system ("mkdir -p $outdir") ) {
		print STDOUT "[ERROR 'print_out_alignment_psmnzd'] FAILED to create the directory \"$outdir\": $! !\n\n";
		return 0;
	    }
	}
	unless ( -w $outdir ) {
	    unless (0 == system (" chmod -R u+w $outdir " ) ) {
		print STDOUT "[ERROR 'print_out_alignment_psmnzd'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
		return 0;
	    }
	}

	print STDOUT "\n[ Input Alignment after brief parsimonization, one of the Main Outputs of 'print_out_bch_br_parsimonize_dollo_indel_histories' ]\n\n";
	print STDOUT " ... will be recorded in the following file:\n\n";
	print STDOUT "  {Out_dir}/$new_aln_filename ,\n";
	print STDOUT "   where {Out_dir} = $outdir ...\n\n";

	my $out_new_aln = "$outdir/$new_aln_filename";
	my $fh_naln = open_output_file ($out_new_aln, 0); # Modified to Fix a BUG (on 2019/06/19). #
	##my $fh_naln = open_output_file ($out_new_aln, 1);
	unless (defined $fh_naln) {
		print STDOUT "[ERROR 'print_out_alignment_psmnzd'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_new_aln\": $! !\n";
		return 0;
	}

	unless ( pr_ODEN_input (@{$seqnames}, @{$new_in_msa}, $DIS_LEN, *${$fh_naln}) ) {
		print STDOUT "[ERROR 'print_out_alignment_psmnzd'] The subroutine 'pr_ODEN_input'  FAILED to print out the NEW input MSA: $! !\n";
		return 0;
	}
	print STDOUT "\n ... Subroutine, 'pr_ODEN_input', Ended ...\n\n";

	close ($fh_naln);

	return 1;

}	# End of the subroutine, 'print_out_alignment_psmnzd'. #


# print_out_wrapper_detect_cmplx_error_cand_acls0:
#	This subroutine prints out the outputs of 'wrapper_detect_cmplx_error_cand_acls0.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/14.]
#
# INPUTS: (\$out_parent, \@cmplx_cands, \@ncmplx_cands, \@artf_clusters, \%gs2prev_nn_events, \@set_lists_gblocks, \@set_cls_gpblocks), where
#
#	$out_parent is a string specifying the path to the parent of the output directory;
#
#	@{$cmplx_cands[$cc]} = ($indx_ac, $totct_indels, $diffct_indels) gives information on the $cc th artificial cluster likely to be involved in complex errors, where
#	$indx_ac is the index in @artf_cluster, $totct_indels = $totct_ins + $totct_dels, and  $diffct_indels = $totct_ins - $totct_dels in the artificial cluster;
#
#	@{$ncmplx_cands[$nc]} = ($indx_ac, $totct_indels, $diffct_indels) gives information on the $nc th artificial cluster UNlikely to be involved in complex errors;
#
#	[\@cmplx_cands and \@ncmplx_cands are the outputs of 'detect_cmplx_error_cand_acls0' above];
#
#
#	@{$atrf_clusters[$acl]} stores the indices of the gapped segments that constitute the $acl th artificial cluster;
#
#	%gs2prev_nn_events = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...),
#	    with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events).
#	    Incidentally, the suffixes, "_n" and "_c" indicate "next" and "current", respectively;
#
#	[\@artf_clusters and \%gs2prev_nn_events are the outputs of 'artf_cluster_gapped_segs(_len)' above];
#
#	@{$set_lists_gblocks[$gs]->[$gb]} provides information on the $gb th gap-block in the $gs th gapped segment. More precisely, 
#	($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@aff_exnodes, $e) = @{$set_lists_gblocks[$cl]->[$gb], where
#	the gap-block is separated by, and on the $up_or_lw (='U' or 'L') side of, branch $br,
#	$lm_gpb and $rm_gpb are the indices of the left-most and right-most gap-pattern blocks,
#	@invlvd_gpbs lists the gap-pattern-blocks involved;
#	@aff_exnodes lists the IDs of external nodes affected by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}};
#
#	[@br_list_gblocks = @{$set_lists_gblocks[$gs]} is the output of 'br_list_gblocks' (in "MyTreeMap_indels_spt_odr2.pm").];
#
#	@set_cls_gpblocks stores information on the gapped-segments, i.e., clusters of gap-pattern-blocks.
#	More precisely,
#	($indx_gp, $len, $beg, $end) = @{$set_cls_gpblocks->[$cls][$gpb]} are information on the $gpb th gap-pattern-block in the $cls th gapped segment, where
#	$indx_gp is the index of the gap-pattern,
#	$len, $beg, $end are the length, beginning coordinate, and end coordinate, respectively,
#	 of the gap-pattern-block;
#	[\@set_cls_gpblocks is an important input of 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm"];
#
# OUTPUT: $flag_end, where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#		  = 0 if it failed.
#
sub print_out_wrapper_detect_cmplx_error_cand_acls0 (\$\@\@\@\%\@\@) {

	my ($out_parent, $cmplx_cands, $ncmplx_cands, $artf_cls, $gs2prev_nn_events, $set_lists_gblocks, $set_cls_gpblocks) = @_;

	my $ct_ccands = @{$cmplx_cands};
	my $ct_nccands = @{$ncmplx_cands};

	my $outdir = "${$out_parent}/Cmplx_error_cands";
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_wrapper_detect_cmplx_error_cand_acls0'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_wrapper_detect_cmplx_error_cand_acls0'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $clist_filename = "cmplx_cands.txt";
	my $nclist_filename = "ncmplx_cands.txt";
	my $out_clist = "$outdir/$clist_filename";
	my $out_nclist = "$outdir/$nclist_filename";

	print STDOUT "[[[ Main Outputs of 'print_out_wrapper_detect_cmplx_error_cand_acls0' ]]]\n\n";
	print STDOUT " ... will be recorded in the following files:\n\n";
	print STDOUT "Storing directory: {Out_dir} = $outdir .\n";
	print STDOUT "Complex candidates:     {Out_dir}/$clist_filename ,\n";
	print STDOUT "Non-complex candidates: {Out_dir}/$nclist_filename .\n";
	print STDOUT "\n\n";

	##my $fh_clist = open_output_file ($out_clist, 1);
	my $fh_clist = open_output_file ($out_clist, 0); # Modified on 2020/05/06.
	unless (defined $fh_clist) {
		print STDOUT "[ERROR 'print_out_wrapper_detect_cmplx_error_cand_acls0'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_clist\": $! !\n";
		return 0;
	}

	print $fh_clist '[ The content of @{$cmplx_cands}, the 1st main output of ', "'detect_cmplx_error_cand_acls0'   (#{complex candidates} = $ct_ccands) ]\n\n";

	print $fh_clist join ("\t", 'Indx_acls', 'Tot#{indels}', 'Diff#{indels}', 'beg_acls', 'end_acls', '[contained,gapped,segments]'), "\n\n";
	foreach my $cand (@{$cmplx_cands}) { 

	    my $indx_acls = $cand->[0];
	    my $cntnd_gsegs = $artf_cls->[$indx_acls];
	    my ($lm_gseg, $rm_gseg) = @{$cntnd_gsegs}[0, $#{$cntnd_gsegs}];
	    my $gpbs_in_lm_gseg = $set_cls_gpblocks->[$lm_gseg];
	    my $gpbs_in_rm_gseg = $set_cls_gpblocks->[$rm_gseg];
	    my $beg_acls = $gpbs_in_lm_gseg->[0][2];
	    my $end_acls = $gpbs_in_rm_gseg->[$#{$gpbs_in_rm_gseg}][3];

		# $cls_gpblks = $set_cls_gpblocks->[$cl];
		# $ct_gpblks = @{$cls_gpblks};
	   	# ($indx_gp, $len, $beg, $end) = @{$cls_gpblks->[$bl]};

	    my $cnct_cntnd_gsegs = join (',', @{$cntnd_gsegs});
	    print $fh_clist join ("\t", @{$cand}, $beg_acls, $end_acls, "[$cnct_cntnd_gsegs]"), "\n";
	}
	print $fh_clist "\n\n\n";

	close ($fh_clist);

	print STDOUT "\n\n ... Information on the complex-error candidates has been recorded into the file, \"$out_clist\" ...\n\n";


	##my $fh_nclist = open_output_file ($out_nclist, 1);
	my $fh_nclist = open_output_file ($out_nclist, 0); # Modified on 2020/05/06.
	unless (defined $fh_nclist) {
		print STDOUT "[ERROR 'print_out_wrapper_detect_cmplx_error_cand_acls0'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_nclist\": $! !\n";
		return 0;
	}

	print $fh_nclist '[ The content of @{$ncmplx_cands}, the 2nd main output of ', "'detect_cmplx_error_cand_acls0'  (#{non-complex candidates} = $ct_nccands) ]\n\n";

	print $fh_nclist join ("\t", 'Indx_acls', 'Tot#{indels}', 'Diff#{indels}', 'beg', 'end', '[contained,gapped,segments]'), "\n\n";
	foreach my $ncand (@{$ncmplx_cands}) { 

	    my $indx_acls = $ncand->[0];
	    my $cntnd_gsegs = $artf_cls->[$indx_acls];
	    my ($lm_gseg, $rm_gseg) = @{$cntnd_gsegs}[0, $#{$cntnd_gsegs}];
	    my $gpbs_in_lm_gseg = $set_cls_gpblocks->[$lm_gseg];
	    my $gpbs_in_rm_gseg = $set_cls_gpblocks->[$rm_gseg];
	    my $beg_acls = $gpbs_in_lm_gseg->[0][2];
	    my $end_acls = $gpbs_in_rm_gseg->[$#{$gpbs_in_rm_gseg}][3];

		# $cls_gpblks = $set_cls_gpblocks->[$cl];
		# $ct_gpblks = @{$cls_gpblks};
	   	# ($indx_gp, $len, $beg, $end) = @{$cls_gpblks->[$bl]};

	    my $cnct_cntnd_gsegs = join (',', @{$cntnd_gsegs});
	    print $fh_nclist join ("\t", @{$ncand}, $beg_acls, $end_acls, "[$cnct_cntnd_gsegs]"), "\n";
	}
	print $fh_nclist "\n\n\n";

	close ($fh_nclist);

	print STDOUT "\n\n ... Information on the NON-complex-error candidates has been recorded into the file, \"$out_nclist\" ...\n\n";

	return 1;

}	# End of the subroutine, 'print_out_wrapper_detect_cmplx_error_cand_acls0'. #



# print_out_prepare_sliding_windows:
#	This subroutine prints out some information on the outputs of 'prepare_sliding_windows'.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/14; Modified on 2019/06/15.]
#
# [ INPUTS: (\$out_parent, \@window_frames, \@prohibited_zones, \@set_windows, \@set_lists_gblocks_in_wds), where ] ... OBSOLETE as of 2019/06/15.
# INPUTS: (\$out_parent, \$outdir_name, \@window_frames, \@prohibited_zones, \@set_windows, \@set_lists_gblocks_in_wds), where
#
#	$out_parent is a string specifying the path to the parent of the output directory;
#	$outdir_name is a string giving the name of the output directory; # Added in modification on 2019/06/15.
#
#	\@window_frames, \@prohibited_zones, \@set_windows & \@set_lists_gblocks_in_wds are the outputs of 'prepare_sliding_windows' (in "MyANEX_Supple.pm").
#	More precisely,
#
#	@{$window_frames[$fr]} = ($start_fr, $end_fr)
#	defines the $fr th frame for the window analysis,
#	that is, the frame starts at the $start_fr th column and ends at the $end_fr th column.
#	(Both are included in the frame.);
#
#	@{$prohibited_zones[$pz]} = ($start_pz, $end_pz)
#	defines the $pz th prohibited zone,
#	that is, the zone starts at the $start_pz th column and ends at the $end_pz th column.
#	(Both are included in the zone.);
#
#	@{$set_windows[$wd]} = ($beg_wd, $end_wd, \@invlvd_gpd_sgs, $ct_gbs, $fr) is for the $wd th sliding window, where
#	the window starts at the $beg_wd th column and ends at the $end_wd th column,
#	contains gapped segments listed in @invlvd_gpd_sgs,
#	$ct_gbs is the count of gap-blocks in the window,
#	and $fr is the index of the frame the window belongs to;
#
#	@{$set_lists_gblocks_in_wds[$wd]->[$gb_in_wd]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb_in_wd th gap-block in the $wd th window.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block extends on the upper/lower side of the branch,
#	$lm/rm_clm is the index of the left-most/right-most column affected by the gap-block,
#	@invlvd_gpbs lists the gap-pattern blocks (within the $gs th gapped segment) involved,
#	@affected_exnodes lists the IDs of external nodes affected by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}},
#	$gs is the index of the gapped segment accommodating the subject gap-block,
#	$gb_in_gs is the index of the subject gap-block within the $gs th gapped segment;
#
# OUTPUT: $flag_end, where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#		  = 0 if it failed.
#
sub print_out_prepare_sliding_windows (\$\$\@\@\@\@) {

	my ($out_parent, $outdir_name, $window_frames, $prohibited_zones, $set_windows, $set_lists_gblocks_in_wds) = @_;

	my $outdir = "${$out_parent}/${$outdir_name}";
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_prepare_sliding_windows'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_prepare_sliding_windows'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $wf_filename = "window_frames.txt";
	my $pz_filename = "prohibited_zones.txt";
	my $gbs_wds_filename = "gap_blocks_in_windows.txt";
	my $out_wf = "$outdir/$wf_filename";
	my $out_pz = "$outdir/$pz_filename";
	my $out_gbs_wds = "$outdir/$gbs_wds_filename";

	print STDOUT "\n\n[[[ The Outputs of the Subroutine, 'prepare_sliding_windows' ]]]\n\n";

	my $ct_pzs = @{$prohibited_zones};
	my $ct_frs = @{$window_frames};
	my $ct_wds = @{$set_windows};

	print STDOUT '#{prohibited zones} = ', $ct_pzs, " ,\n";
	print STDOUT '#{window frames}    = ', $ct_frs, " ,\n";
	print STDOUT '#{windows}          = ', $ct_wds, " .\n\n\n";

	print STDOUT " ... Further DETAILS will be recorded in the following files:\n\n";
	print STDOUT "Storing directory: {Out_dir} = $outdir .\n";
	print STDOUT "Window frames:     {Out_dir}/$wf_filename ,\n";
	print STDOUT "Prohibited zones:  {Out_dir}/$pz_filename ,\n";
	print STDOUT "Gap-blocks in windows: {Out_dir}/$gbs_wds_filename .\n";
	print STDOUT "\n\n";


	##my $fh_pz = open_output_file ($out_pz, 1);
	my $fh_pz = open_output_file ($out_pz, 0); # Modified on 2020/05/06.
	unless (defined $fh_pz) {
		print STDOUT "[ERROR 'print_out_prepare_sliding_windows'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_pz\": $! !\n";
		return 0;
	}

	print $fh_pz '[[ The contents of @{$prohibited_zones} ]]', "\n\n";

	print $fh_pz join ("\t", 'Indx_pz', 'beg_pz', 'end_pz'), "\n\n";

	for (my $pz = 0; $pz < $ct_pzs; $pz++) { # 1st for-loop (over zones). #
	    print $fh_pz  join ("\t", $pz, @{$prohibited_zones->[$pz]}), "\n";
	} # End of the 1st for-loop (over zones). #
	print $fh_pz "\n\n";

	close ($fh_pz);

	print STDOUT "\n\n ... Information on the Prohibited Zones has been recorded into the file, \"$out_pz\" ...\n\n";


	##my $fh_wf = open_output_file ($out_wf, 1);
	my $fh_wf = open_output_file ($out_wf, 0); # Modified on 2020/05/06.
	unless (defined $fh_wf) {
		print STDOUT "[ERROR 'print_out_prepare_sliding_windows'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_wf\": $! !\n";
		return 0;
	}

	print $fh_wf '[[ The contents of @{$window_frames} ]]', "\n\n";

	print $fh_wf join ("\t", 'Indx_fr', 'beg_fr', 'end_fr'), "\n\n";

	for (my $fr = 0; $fr < $ct_frs; $fr++) { # 2nd for-loop (over frames). #
	    print $fh_wf join ("\t", $fr, @{$window_frames->[$fr]}), "\n";
	} # End of the 2nd for-loop (over frames). #
	print $fh_wf "\n\n";

	close ($fh_wf);

	print STDOUT "\n\n ... Information on the Window Frames has been recorded into the file, \"$out_wf\" ...\n\n";

	my $fh_gbs_wds = open_output_file ($out_gbs_wds, 1);
	unless (defined $fh_gbs_wds) {
		print STDOUT "[ERROR 'print_out_prepare_sliding_windows'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_gbs_wds\": $! !\n";
		return 0;
	}

	print $fh_gbs_wds '[[ The contents of @{$set_windows} & @{$set_lists_gblocks_in_wds} ]]', "\n\n";

    for (my $wd = 0; $wd < $ct_wds; $wd++) { # 3rd outer for-loop (over windows). #

	my ($beg_wd, $end_wd, $invlvd_gpd_segs, $ct_gbs, $fr) = @{$set_windows->[$wd]};

	print $fh_gbs_wds "{ The $wd th window (belonging to the $fr th frame) }\n\n";
	print $fh_gbs_wds "[beg_wd, end_wd] = [$beg_wd, $end_wd] ,\n";
	print $fh_gbs_wds "(involved, gapped, segments) = (", join (', ', @{$invlvd_gpd_segs}), ") ,\n";
	print $fh_gbs_wds "#{gap-blocks} = $ct_gbs .\n\n\n";

	my $list_gbs_in_wd = $set_lists_gblocks_in_wds->[$wd];
	my $ct_gbs2 = @{$list_gbs_in_wd};
	unless ($ct_gbs == $ct_gbs2) {
	    print STDOUT "[WARNING] The size of the list of gap-blocks (=$ct_gbs2) CLASHES with the number of gap-blocks (=$ct_gbs): $! !\n\n";
	}

	print $fh_gbs_wds "[ The list of gap-blocks in the window (size = $ct_gbs2) ]\n\n";
	
	print $fh_gbs_wds ' ' x 2, join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '[involved,gap,pattern,blocks]', '(affected,external,nodes)', 'within_branch_index', 'Indx_gpd_seg', 'within_gpd_seg_index'), "\n\n"; # CORRECTED on 2019/06/23. #
	##print $fh_gbs_wds ' ' x 2, join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_gpb', 'rm_gpb', '[involved,gap,pattern,blocks]', '(affected,external,nodes)', 'within_branch_index', 'Indx_gpd_seg', 'within_gpd_seg_index'), "\n\n";
#	@{$set_lists_gblocks_in_wds[$wd]->[$gb_in_wd]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb_in_wd th gap-block in the $wd th window.

	for (my $gb = 0; $gb < $ct_gbs2; $gb++) {
	    my ($br, $up_or_lw, $lm_clm, $rm_clm, $invlvd_gpbs, $aff_exnodes, $e, $gs, $gb_in_gs) = @{$list_gbs_in_wd->[$gb]}; # CORRECTED on 2019/06/23. #
	    ##my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs, $aff_exnodes, $e, $gs, $gb_in_gs) = @{$list_gbs_in_wd->[$gb]};
	    my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs}); 
	    $cnct_invlvd_gpbs = "[$cnct_invlvd_gpbs]";
	    my $cnct_aff_exnodes = join (',', @{$aff_exnodes});
	    $cnct_aff_exnodes = "($cnct_aff_exnodes)";
	    print $fh_gbs_wds ' ' x 2, join ("\t", $gb, $br, $up_or_lw, $lm_clm, $rm_clm, $cnct_invlvd_gpbs, $cnct_aff_exnodes, $e, $gs, $gb_in_gs), "\n"; # CORRECTED on 2019/06/23. #
	    ##print $fh_gbs_wds ' ' x 2, join ("\t", $gb, $br, $up_or_lw, $lm_gpb, $rm_gpb, $cnct_invlvd_gpbs, $cnct_aff_exnodes, $e, $gs, $gb_in_gs), "\n";
	}
	print $fh_gbs_wds "\n\n\n";

    }   # End of the 3rd outer for-loop (over windows). #

	close ($fh_gbs_wds);

	print STDOUT "\n\n ... Information on the Gap-Blocks in each of the Windows  has been recorded into the file, \"$out_gbs_wds\" ...\n\n";

	return 1;

}	# End of the subroutine, 'print_out_prepare_sliding_windows'. #


# print_out_list_mgd_pcands:
#	This subroutine prints out a list of merged purge-candidate regions,
#	which is typically output by 'readin_list_mgd_pcands'
#	(, and possibly processed by 'select_sgn_wds_in_pcands').
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/20.]
#
# INPUTS: (\$outdir, $name_ofile, \@list_mgd_pcands), where
#
#	$outdir is a string specifying the path to the output directory;
#	$name_ofile is a string providing the name of the output file;
#
#	\@list_mgd_pcands is typically an output of 'readin_list_mgd_pcands'.
#	More precisely,
#	@{$list_mgd_pcands[$pc]} = ($beg_pc, $end_pc, $br, \@msa_rows_under_br, \@constituent_sites, $pv, $pv0, \@sgn_wds_included) provides information on the $pc th (merged) purge-candidate region, where
#	$beg_pc & $end_pc are the sites at the beginning & end, respectively, of the region,
#	$br is the ID of the branch across which the "purge" is suspected within the region,
#	@msa_rows_under_br stores the indices of the MSA rows corresponding to the external offsprings of $br,
#	@constituent_sites stores the sites that constitute the region,
#	$pv is the P-value measuring the deviation from the specified substutution model,
#	$pv0 is the P-value measuring the deviation from the random-match model,
#	and @sgn_wds_included is the significant windows included in the region;
#	more precisely, 
#	@{$sgn_wds_included[$swd]} = (\@sites_in_wd, $pv_wd, $pv0_wd) provies information on the $swd th significant window, where
#	@sites_in_wd stores the sites in the window,
#	$pv_wd and $pv0_wd are the P-values for the window, measuring the same types of deviations as the aforementioned $pv and $pv0, respectively;
#
# OUTPUT: $flag_end, where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problem,
#		  = 0 if it failed.
#
sub print_out_list_mgd_pcands (\$$\@) {

	my ($rf_outdir, $name_ofile, $list_mgd_pcands) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_list_mgd_pcands'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_list_mgd_pcands'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$name_ofile";

	##my $fh_out = open_output_file ($outfile, 1);
	my $fh_out = open_output_file ($outfile, 0); # MODIFIED on 2020/05/08, to facilitate the inter-connection between the master & slave scripts.
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_list_mgd_pcands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_mgd_pcands = @{$list_mgd_pcands};

	print $fh_out join ("\t", 'beg_rg', 'end_rg', 'br_ID', '[msa,rows,under,br]', '(sites,in,the,region)', 'pv', 'pv0', '{significant window included ([sites,in,wd]/pv_wd/pv0_wd)}_0', '{...}_1', '...'), "\n\n"; # IMPORTED from 'detect_purge_cands.ver0.5.pl' (on 2019/06/20).

	for (my $pc=0; $pc < $ct_mgd_pcands; $pc++) { # Main for-loop (over (merged) purge-candidate regions). #
	    my ($beg_pc, $end_pc, $br, $msa_rows_under_br, $constituent_sites, $pv, $pv0, $sgn_wds_incld) = @{$list_mgd_pcands->[$pc]};

	    my $cnct_msa_rows_under_br = join (',', @{$msa_rows_under_br});
	    my $cnct_constituent_sites = join (',', @{$constituent_sites});

	    my @out_sgn_wds_incld = ();
	    my $ct_sgn_wds_incld = @{$sgn_wds_incld};

	    for (my $swd=0; $swd < $ct_sgn_wds_incld; $swd++) { # Inner for-loop (over significant windows in the region). #
		my ($sites_in_wd, $pv_wd, $pv0_wd) = @{$sgn_wds_incld->[$swd]};
		my $cnct_sites_in_wd = join (',', @{$sites_in_wd});

		push @out_sgn_wds_incld, "[$cnct_sites_in_wd]/$pv_wd/$pv0_wd";

	    } # End of the inner for-loop (over significant windows in the region). #

	    print $fh_out join ("\t", $beg_pc, $end_pc, $br, "[$cnct_msa_rows_under_br]", "($cnct_constituent_sites)", $pv, $pv0, @out_sgn_wds_incld), "\n"; 
	
		# @{$list_mgd_pcands[$pc]} = ($beg_pc, $end_pc, $br, \@msa_rows_under_br, \@constituent_sites, $pv, $pv0, \@sgn_wds_included) provides information on the $pc th (merged) purge-candidate region, where
		# ...
		# @{$sgn_wds_included[$swd]} = (\@sites_in_wd, $pv_wd, $pv0_wd) provies information on the $swd th significant window, where
		# ...

	}  # End of the main for-loop (over (merged) purge-candidate regions). #

	close ($fh_out);

	return 1;

}	# End of the subroutine 'print_out_list_mgd_pcands'. #


# print_out_chk_ovlp_pcands_vs_sliding_wds:
#	This subroutine prints out the outputs of 'chk_ovlp_pcands_vs_sliding_wds'.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/21.]
#
# INPUTS: (\$outdir, \@list_mgd_pcands_ovlp_sbwds, \@off_sbwds_but_ovlp_wds, \@off_wds_but_ovlp_wframes, @off_wframes), where
#
#	$outdir is a string specifying the path to the output directory;
#
#	@{$list_mgd_pcands_ovlp_sbwds[$swd]} stores the indices of the merged purge-candidate regions overlapping the $swd th sub-window;
#
#	@{$off_sbwds_but_ovlp_wds[$wd}} stores the indices of the regions overlapping the $wd th window but NOT overlapping any sub-windows;
#
#	@{$off_wds_but_ovlp_wframes[$wf]} stores the indices of the regions overlapping the $wf th window-frame but NOT overlapping any windows;
#
#	@off_wframes stores the indices of the regions NOT overlapping any window-frames;
#
# OUTPUT: $flag_end, where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#		  = 0 if it failed.
#
sub print_out_chk_ovlp_pcands_vs_sliding_wds (\$\@\@\@\@) {

	my ($rf_outdir, $list_mgd_pcands_ovlp_sbwds, $off_sbwds_but_ovlp_wds, $off_wds_but_ovlp_wframes, $off_wframes) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_list_mgd_pcands'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_list_mgd_pcands'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $fname_ovlp_sbwds = 'mgd_cands_ovlp_sbwds.txt';
	my $fname_ovlp_wds = 'off_sbwds_but_ovlp_wds.txt';
	my $fname_ovlp_wfs = 'off_wds_but_ovlp_wframes.txt';
	##my $fname_remaining = 'off_wframes';
	my $fname_remaining = 'off_wframes.txt'; # CORRECTED on 2020/05/07.

	my $ofile_ovlp_sbwds = "$outdir/$fname_ovlp_sbwds";
	my $ofile_ovlp_wds = "$outdir/$fname_ovlp_wds";
	my $ofile_ovlp_wfs = "$outdir/$fname_ovlp_wfs";
	my $ofile_remaining = "$outdir/$fname_remaining";


	print STDOUT "\n[ Output files: ]\n\n";
	print STDOUT "Output directory: {Out_dir}=$outdir .\n";
	print STDOUT "Regions Overlapping w/ subwindows: {Out_dir}/$fname_ovlp_sbwds ,\n";
	print STDOUT "Rest Overlapping w/ windows:       {Out_dir}/$fname_ovlp_wds ,\n";
	print STDOUT "Rest Overlapping w/ window-frames: {Outdir}/$fname_ovlp_wfs ,\n";
	print STDOUT "Remaining regions:                 {Outdir}/$fname_remaining .\n";
	print STDOUT "\n\n";


		# (1) Purge-Candidate Regions Overlapping w/ sub-windows. #
	
	my $fh_ovlp_sbwds = open_output_file ($ofile_ovlp_sbwds, 0);
	unless (defined $fh_ovlp_sbwds) {
		print STDOUT "[ERROR 'print_out_list_mgd_pcands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$ofile_ovlp_sbwds\": $! !\n";
		return 0;
	}

	print $fh_ovlp_sbwds join ("\t", 'Indx_sbwd', '[purge,candidate,regions,overlapping,w/,sbwd]'), "\n\n";

	my $ct_sbwds = @{$list_mgd_pcands_ovlp_sbwds};
	for (my $sbwd=0; $sbwd < $ct_sbwds; $sbwd++) { # 1st for-loop (over sub-windows). #
	    my $cnct_pcands = join (',', @{$list_mgd_pcands_ovlp_sbwds->[$sbwd]});

	    print $fh_ovlp_sbwds join ("\t", $sbwd, "[$cnct_pcands]"), "\n";

	} # End of the 1st for-loop (over subwindows). #

	close ($fh_ovlp_sbwds);


		# (2) Remaining Overlapping w/ windows. #

	my $fh_ovlp_wds = open_output_file ($ofile_ovlp_wds, 0);
	unless (defined $fh_ovlp_wds) {
		print STDOUT "[ERROR 'print_out_list_mgd_pcands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$ofile_ovlp_wds\": $! !\n";
		return 0;
	}

	print $fh_ovlp_wds join ("\t", 'Indx_wd', '[purge,candidate,regions,overlapping,w/,wd]'), "\n\n"; # ADDED on 2020/05/08. (1)

	my $ct_wds = @{$off_sbwds_but_ovlp_wds};
	for (my $wd=0; $wd < $ct_wds; $wd++) { # 2nd for-loop (over windows). #
	    my $cnct_pcands = join (',', @{$off_sbwds_but_ovlp_wds->[$wd]});

	    print $fh_ovlp_wds join ("\t", $wd, "[$cnct_pcands]"), "\n";

	} # End of the 2nd for-loop (over windows). #

	close ($fh_ovlp_wds);


		# (3) Remaining Overlapping w/ window-frames. #

	my $fh_ovlp_wfs = open_output_file ($ofile_ovlp_wfs, 0);
	unless (defined $fh_ovlp_wfs) {
		print STDOUT "[ERROR 'print_out_list_mgd_pcands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$ofile_ovlp_wfs\": $! !\n";
		return 0;
	}

	print $fh_ovlp_wfs join ("\t", 'Indx_wframe', '[purge,candidate,regions,overlapping,w/,wframe]'), "\n\n"; # ADDED on 2020/05/08. (2)

	my $ct_wfs = @{$off_wds_but_ovlp_wframes};
	for (my $wf=0; $wf < $ct_wfs; $wf++) { # 3rd for-loop (over window-frames). #
	    my $cnct_pcands = join (',', @{$off_wds_but_ovlp_wframes->[$wf]});

	    ##print $fh_ovlp_wds join ("\t", $wf, "[$cnct_pcands]"), "\n"; # A BUG fixed on 2020/05/07.
	    print $fh_ovlp_wfs join ("\t", $wf, "[$cnct_pcands]"), "\n";

	} # End of the 2nd for-loop (over window-frames). #

	close ($fh_ovlp_wfs);


		# (4) Regions NOT overlapping w/ winfow-frames. #

	my $fh_remaining = open_output_file ($ofile_remaining, 0);
	unless (defined $fh_remaining) {
		print STDOUT "[ERROR 'print_out_list_mgd_pcands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$ofile_remaining\": $! !\n";
		return 0;
	}

 	my $ct_remaining = @{$off_wframes};

	print $fh_remaining "## There are $ct_remaining purge-candidate regions NOT overlapping any window-frames:\n\n";

	print $fh_remaining join (" ,\n", @{$off_wframes}), " .\n\n";

	close ($fh_remaining);

	return 1;

}	# End of the subroutine, 'print_out_chk_ovlp_pcands_vs_sliding_wds'. #


# print_out_set_clms:
#	This subroutine prints out the output of 'mk_init_set_columns0',
#	along with information on the boundaries of the subject window.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/23.]
#	[REVISED on 2019/07/19]:
#	From now on , $beg_wd and $end_wd MUST BE the values for the INITIAL INPUT MSA.
#
# INPUTS: (\$outdir, $ofname, \@set_clms, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@column = @{$set_columns[$c]} specifies the $c th column in the input local MSA. more precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input alignment
#	that accommodates the corresponding site of the $k th "sequence" (or row);
#
#	[ $beg/end_wd is the MSA site (column) at the beginning/end of the subject window; ] ... OBSOLETE as of 2019/07/19.
#	$beg/end_wd is the site (column) of the initial input MSA at the beginning/end of the subject window; # REVISED on 2019/07/19.
#
#	$if_on_le/re_msa = 1 if the subject window is on the left/right-end of the MSA,
#			 = 0 otherwise.
#
sub print_out_set_clms (\$$\@$$$$) {

	my ($rf_outdir, $ofname, $set_clms, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa) = @_;

	my $ct_clms = @{$set_clms};
	my $ct_rows = ($ct_clms>0) ? @{$set_clms->[0]} : 0;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_set_clms'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_set_clms'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_set_clms'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	print $fh_out "# [beg_wd, end_wd]_input_MSA = [$beg_wd, $end_wd] .\n"; # REVISED on 2019/07/19.
	##print $fh_out "# [beg, end] = [$beg_wd, $end_wd] .\n";

	print $fh_out "# if_on_le_msa = $if_on_le_msa ,\n";
	print $fh_out "# if_on_re_msa = $if_on_re_msa .\n";
	print $fh_out "#\n\n";

	print $fh_out join ("\t", 'Column\Row', 0 .. $ct_rows -1), "\n\n";

	for (my $c=0; $c < $ct_clms; $c++) { # Main for-loop (over the columns).

	    print $fh_out join ("\t", $c, @{$set_clms->[$c]}), "\n";

	} # End of the main for-loop (over the columns).
	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_set_clms'. #


# print_out_list_gblocks_in_wd:
#	This subroutine prints out information on the gap-blocks in a subject (sub-)window.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/23.]
#
#	NOTE: Before printing out, the subroutine "translate"s the indices of external nodes (under $br)
#		into the indices of the MSA rows (affected by the gap-block),
#		and it ignores pieces of information that are dispensable.
#
# INPUTS: (\$outdir, $ofname, \@list_gblocks_in_wd, %nodeid2msa_row), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@{$list_gblocks_in_wd[$gb]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb th gap-block in the subject window;
#
#	%nodeid2msa_row = ($node_id => $indx_msa_row, ...) for the nodes corresponding to the MSA rows;
#	
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_list_gblocks_in_wd (\$$\@\%) {

	my ($rf_outdir, $ofname, $list_gblocks_in_wd, $nodeid2msa_row) = @_;

	my $ct_gbs = @{$list_gblocks_in_wd};

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_list_gblocks_in_wd'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_list_gblocks_in_wd'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_list_gblocks_in_wd'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	print $fh_out "## The list of gap-blocks in the (sub-)window (#{blocks} = $ct_gbs).\n\n";
	
	print $fh_out  join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(affected,external,nodes)', '[msa,rows,affected]'), "\n\n";

	for (my $gb = 0; $gb < $ct_gbs; $gb++) { # Main for-loop (over gap-blocks) #

	    my ($br, $up_or_lw, $lm_clm, $rm_clm, $invlvd_gpbs, $affd_exnodes, $e, $gs, $gb_in_gs) = @{$list_gblocks_in_wd->[$gb]};

# @{$set_lists_gblocks_in_wds[$wd]->[$gb_in_wd]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb_in_wd th gap-block in the $wd th window.

	    my $cnct_affd_exnodes = join (',', @{$affd_exnodes});
	    $cnct_affd_exnodes = "($cnct_affd_exnodes)";

	    my @affd_msa_rows = ();
	    foreach my $node (@{$affd_exnodes}) { # Inner foreach-loop.
		my $msa_row = $nodeid2msa_row->{$node};
		unless (defined $msa_row) {
		    print STDOUT "[ERROR 'print_out_prepare_sliding_windows'] The \"external\" node, $node, has NO corresponding MSA row: $! !\n\n";
		    return 0;
		}
		push @affd_msa_rows, $msa_row;
	    } # End of the inner foreach-loop. 

	    my @std_affd_msa_rows = sort {$a <=> $b} @affd_msa_rows;

	    my $cnct_affd_msa_rows = join (',', @std_affd_msa_rows);
	    $cnct_affd_msa_rows = "[$cnct_affd_msa_rows]";

	    print $fh_out join ("\t", $gb, $br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows), "\n";

	} # End of the main for-loop (over gap-blocks). #

	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_list_gblocks_in_wd'. #


# print_out_mk_pcc_sliding_windows:
#	This subroutine prints out some information on the outputs of 'mk_pcc_sliding_windows'.
#	It was created by modifying the script, 'print_out_prepare_sliding_windows', above.
#	[CREATED in revision of 'anex.ver0.5.pl', on 2019/07/06.]
#	
# print_out_prepare_sliding_windows:
#	This subroutine prints out some information on the outputs of 'prepare_sliding_windows'.
#	[CREATED on 2019/06/14; Modified on 2019/06/15.]
#
# [ INPUTS: (\$out_parent, \$outdir_name, \@window_frames, \@prohibited_zones, \@set_windows, \@set_lists_gblocks_in_wds), where ] ... for the predecessor, 'print_out_prepare_sliding_windows'.
#
# INPUTS: (\$out_parent, \$outdir_name, \@window_frames, \@set_indices_mpcands, \@set_pcc_windows, \@set_lists_gblocks_in_pcc_wds), where
#
#	$out_parent is a string specifying the path to the parent of the output directory;
#	$outdir_name is a string giving the name of the output directory; # Added in modification on 2019/06/15.
#
#	\@window_frames is an important input of 'mk_pcc_sliding_windows'. More precisely,
#	@{$window_frames[$fr]} = ($start_fr, $end_fr)
#	defines the $fr th frame for the window analysis,
#	that is, the frame starts at the $start_fr th column and ends at the $end_fr th column.
#	(Both are included in the frame.);
#
#	\@set_indices_mpcands, \@set_pcc_windows & \@set_lists_gblocks_in_pcc_wds are the outputs of 'mk_pcc_sliding_windows' (in "MyANEX_Supple.pm").
#	More precisely,
#
#	$indx_mpcand = $set_indices_mpcands[$pwd] is the index of the (merged) purge-candidate region that is a main component of the $pwd th purge-candidate-containing (PCC) window;
#
#	@{$set_pcc_windows[$pwd]} = ($beg_pwd, $end_pwd, \@invlvd_gpd_sgs, $ct_gbs, $fr) is for the $pwd th PCC sliding window, where
#	the window starts at the $beg_pwd th column and ends at the $end_pwd th column,
#	contains gapped segments listed in @invlvd_gpd_sgs,
#	$ct_gbs is the count of gap-blocks in the window,
#	and $fr is the index of the frame the window belongs to;
#
#	@{$set_lists_gblocks_in_pcc_wds[$pwd]->[$gb_in_pwd]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb_in_pwd th gap-block in the $pwd th PCC window.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block extends on the upper/lower side of the branch,
#	$lm/rm_clm is the index of the left-most/right-most column affected by the gap-block,
#	@invlvd_gpbs lists the gap-pattern blocks (within the $gs th gapped segment) involved,
#	@affected_exnodes lists the IDs of external nodes affected by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}},
#	$gs is the index of the gapped segment accommodating the subject gap-block,
#	$gb_in_gs is the index of the subject gap-block within the $gs th gapped segment;
#
# OUTPUT: $flag_end, where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#		  = 0 if it failed.
#
sub print_out_mk_pcc_sliding_windows (\$\$\@\@\@\@) {

	my ($out_parent, $outdir_name, $window_frames, $set_indices_mpcands, $set_pcc_windows, $set_lists_gblocks_in_pcc_wds) = @_;

	my $outdir = "${$out_parent}/${$outdir_name}";
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_mk_pcc_sliding_windows'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_mk_pcc_sliding_windows'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $wf_filename = "window_frames.txt";
	my $impc_filename = "involved_mpcands.txt";
	my $gbs_pwds_filename = "gap_blocks_in_pcc_windows.txt";
	my $out_wf = "$outdir/$wf_filename";
	my $out_impc = "$outdir/$impc_filename";
	my $out_gbs_pwds = "$outdir/$gbs_pwds_filename";

	print STDOUT "\n\n[[[ The Inputs/Outputs of the Subroutine, 'mk_pcc_sliding_windows' ]]]\n\n";

	my $ct_frs = @{$window_frames};
	my $ct_pwds = @{$set_pcc_windows};

	print STDOUT '#{window frames}    = ', $ct_frs, " ,\n";
	print STDOUT '#{PCC windows}          = ', $ct_pwds, " .\n\n\n";

	print STDOUT " ... Further DETAILS will be recorded in the following files:\n\n";
	print STDOUT "Storing directory: {Out_dir} = $outdir .\n";
	print STDOUT "Window frames:     {Out_dir}/$wf_filename ,\n";
	print STDOUT "Invovled merged purge-candidates:  {Out_dir}/$impc_filename ,\n";
	print STDOUT "Gap-blocks in PCC windows: {Out_dir}/$gbs_pwds_filename .\n";
	print STDOUT "\n\n";


	my $fh_impc = open_output_file ($out_impc, 1);
	unless (defined $fh_impc) {
		print STDOUT "[ERROR 'print_out_mk_pcc_sliding_windows'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_impc\": $! !\n";
		return 0;
	}

	print $fh_impc '[[ The contents of @{$set_indices_mpcands} ]]', "\n\n";

	print $fh_impc join ("\t", 'Indx_pwd', 'involved_mgd_pcand'), "\n\n";

	for (my $pwd = 0; $pwd < $ct_pwds; $pwd++) { # 1st for-loop (over PCC windows). #
	    print $fh_impc  join ("\t", $pwd, $set_indices_mpcands->[$pwd]), "\n";
	} # End of the 1st for-loop (over PCC windows). #
	print $fh_impc "\n\n";

	close ($fh_impc);

	print STDOUT "\n\n ... Information on the involved Merged Purge-Candidates has been recorded into the file, \"$out_impc\" ...\n\n";


	my $fh_wf = open_output_file ($out_wf, 1);
	unless (defined $fh_wf) {
		print STDOUT "[ERROR 'print_out_mk_pcc_sliding_windows'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_wf\": $! !\n";
		return 0;
	}

	print $fh_wf '[[ The contents of @{$window_frames} ]]', "\n\n";

	print $fh_wf join ("\t", 'Indx_fr', 'beg_fr', 'end_fr'), "\n\n";

	for (my $fr = 0; $fr < $ct_frs; $fr++) { # 2nd for-loop (over frames). #
	    print $fh_wf join ("\t", $fr, @{$window_frames->[$fr]}), "\n";
	} # End of the 2nd for-loop (over frames). #
	print $fh_wf "\n\n";

	close ($fh_wf);

	print STDOUT "\n\n ... Information on the Window Frames has been recorded into the file, \"$out_wf\" ...\n\n";

	my $fh_gbs_pwds = open_output_file ($out_gbs_pwds, 1);
	unless (defined $fh_gbs_pwds) {
		print STDOUT "[ERROR 'print_out_mk_pcc_sliding_windows'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$out_gbs_pwds\": $! !\n";
		return 0;
	}

	print $fh_gbs_pwds '[[ The contents of @{$set_pcc_windows} & @{$set_lists_gblocks_in_pcc_wds} ]]', "\n\n";

    for (my $pwd = 0; $pwd < $ct_pwds; $pwd++) { # 3rd outer for-loop (over PCC windows). #

	my ($beg_pwd, $end_pwd, $invlvd_gpd_segs, $ct_gbs, $fr) = @{$set_pcc_windows->[$pwd]};

	print $fh_gbs_pwds "{ The $pwd th PCC window (belonging to the $fr th frame) }\n\n";
	print $fh_gbs_pwds "[beg_pwd, end_pwd] = [$beg_pwd, $end_pwd] ,\n";
	print $fh_gbs_pwds "(involved, gapped, segments) = (", join (', ', @{$invlvd_gpd_segs}), ") ,\n";
	print $fh_gbs_pwds "#{gap-blocks} = $ct_gbs .\n\n\n";

	my $list_gbs_in_pwd = $set_lists_gblocks_in_pcc_wds->[$pwd];
	my $ct_gbs2 = @{$list_gbs_in_pwd};
	unless ($ct_gbs == $ct_gbs2) {
	    print STDOUT "[WARNING] The size of the list of gap-blocks (=$ct_gbs2) CLASHES with the number of gap-blocks (=$ct_gbs) for the $pwd th PCC window: $! !\n\n";
	}

	print $fh_gbs_pwds "[ The list of gap-blocks in the PCC window (size = $ct_gbs2) ]\n\n";
	
	print $fh_gbs_pwds ' ' x 2, join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '[involved,gap,pattern,blocks]', '(affected,external,nodes)', 'within_branch_index', 'Indx_gpd_seg', 'within_gpd_seg_index'), "\n\n"; # CORRECTED on 2019/06/23. #

#	@{$set_lists_gblocks_in_wds[$wd]->[$gb_in_wd]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb_in_wd th gap-block in the $wd th window.

	for (my $gb = 0; $gb < $ct_gbs2; $gb++) {
	    my ($br, $up_or_lw, $lm_clm, $rm_clm, $invlvd_gpbs, $aff_exnodes, $e, $gs, $gb_in_gs) = @{$list_gbs_in_pwd->[$gb]};

	    my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs}); 
	    $cnct_invlvd_gpbs = "[$cnct_invlvd_gpbs]";
	    my $cnct_aff_exnodes = join (',', @{$aff_exnodes});
	    $cnct_aff_exnodes = "($cnct_aff_exnodes)";
	    print $fh_gbs_pwds ' ' x 2, join ("\t", $gb, $br, $up_or_lw, $lm_clm, $rm_clm, $cnct_invlvd_gpbs, $cnct_aff_exnodes, $e, $gs, $gb_in_gs), "\n";
	}
	print $fh_gbs_pwds "\n\n\n";

    }   # End of the 3rd outer for-loop (over PCC windows). #

	close ($fh_gbs_pwds);

	print STDOUT "\n\n ... Information on the Gap-Blocks in each of the PCC Windows has been recorded into the file, \"$out_gbs_pwds\" ...\n\n";

	return 1;

}	# End of the subroutine, 'print_out_mk_pcc_sliding_windows'. #


# print_out_readin_list_gblks:
#	This subroutine prints out the output of 'readin_list_gblks',
#	for the purpose of checking it.
#	The resulting file aims to almost perfectly reproduce the file read via 'readin_list_gblks',
#	except the header.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/26.]
#
#	[REVISED on 2019/06/27]:
#	It now also inputs $type, which indicates whether the input list of gap-blocks is 'original' or 'revised'.
#
# [ INPUTS: (\$outdir, $ofname, \@list_gblks), where ] ... OBSOLETE as of 2019/06/27.
# INPUTS: (\$outdir, $ofname, \@list_gblks, $type), where # Revised on 2019/06/27.
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@list_gblks is an output of 'readin_list_gblks'. More precisely,
#	@{$list_gblks[$indx_gb]} = ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, \@affd_exnodes_org, \@affd_msa_rows),
#	provides information on the $indx_gb th gap-block.
#	More precisely,
#	$br_org is the branch ID according to the original ID assignment of the tree nodes (used for the input),
#	$up_or_lw = 'U'/'L' if the gap-block is on the upper-/lower-side of the branch,
#	$lm/rm_clm is the left-/right-most column of the gap-block,
#	@affd_exnodes_org stores the IDs of the affected external nodes according to the (aforementioned) original ID assignment, and
#	$cnct_affd_exnodes_org = join (',', @affd_exnodes_org),
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block, and
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#
#	$type 	= 'org' if the @list_gblks is original (i.e., read-in from an input file),
#		= 'rvsd' if it is a revised one (most likely via 'revise_list_gblocks');
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 otherwise.
#
# [ sub print_out_readin_list_gblks (\$$\@) {
#	my ($rf_outdir, $ofname, $list_gblks) = @_; ] ... OBSOLETE as of 2019/06/27.
#
sub print_out_readin_list_gblks (\$$\@$) {

	my ($rf_outdir, $ofname, $list_gblks, $type) = @_; # Modified on 2019/06/27.

	my $ct_gbs = @{$list_gblks};

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_readin_list_gblks'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_readin_list_gblks'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_readin_list_gblks'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $full_type = ($type eq 'org') ? 'original' : 'revised'; # Added on 2019/06/27. #

	print $fh_out "## The list of \"$full_type\" gap-blocks in the (sub-)window (#{blocks} = $ct_gbs), read via 'readin_list_gblks'.\n\n"; # MODIFIED on 2019/06/27.
	##print $fh_out "## The list of gap-blocks in the (sub-)window (#{blocks} = $ct_gbs), read via 'readin_list_gblks'.\n\n";

	print $fh_out join ("\t", 'Indx_gb', "br_ID_$type", "up(U)_or_lw(L)_$type", 'lm_clm', 'rm_clm', "(affected,external,nodes)_$type", '[msa,rows,affected]'), "\n\n"; # MODIFIED on 2019/06/27.
	##print $fh_out join ("\t", 'Indx_gb', 'br_ID_org', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(affected,external,nodes)_org', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb <$ct_gbs; $gb++) { # Main for-loop (over gap-blocks). #

	    my ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, $set_affd_exnodes_org, $set_affd_msa_rows) = @{$list_gblks->[$gb]} ;

	    $cnct_affd_exnodes_org = "($cnct_affd_exnodes_org)";
	    $cnct_affd_msa_rows = "[$cnct_affd_msa_rows]";

	    print $fh_out join ("\t", $gb, $br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows), "\n";

	} # End of the main for-loop (over gap-blocks). #

	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_readin_list_gblks' #


# print_out_readin_features_gblks:
#	This subroutine prints out the output of 'readin_features_gblks',
#	for the purpose of checking it.
#	The resulting file aims to almost perfectly reproduce the file read via 'readin_features_gblks',
#	except the header.
#	[CREATED in creation of 'coordinate_point_lcl_msa.ver0.7.pl' on 2020/06/04.]
#
# print_out_readin_list_gblks:
#	This subroutine prints out the output of 'readin_list_gblks',
#	for the purpose of checking it.
#	The resulting file aims to almost perfectly reproduce the file read via 'readin_list_gblks',
#	except the header.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/26.]
#
#	[REVISED on 2019/06/27]:
#	It now also inputs $type, which indicates whether the input list of gap-blocks is 'original' or 'revised'.
#
# [ INPUTS: (\$outdir, $ofname, \@list_gblks), where ] ... OBSOLETE as of 2019/06/27.
# [ INPUTS: (\$outdir, $ofname, \@list_gblks, $type), where # Revised on 2019/06/27. ] ... OBSOLETE as of 2020/06/04.
#
# INPUTS: (\$outdir, $ofname, \@features_gblks, $type), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@features_gblks is an output of 'readin_features_gblks'. More precisely,
#	@{$features_gblks[$indx_gb]} = ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_msa_rows, \@affd_msa_rows),
#	provides information on the $indx_gb th gap-block.
#	More precisely,
#	$br_org is the branch ID according to the original ID assignment of the tree nodes (used for the input),
#	$up_or_lw = 'U'/'L' if the gap-block is on the upper-/lower-side of the branch,
#	$lm/rm_clm is the left-/right-most column of the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block, and
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#
#	$type 	= 'org' if the @list_gblks is original (i.e., read-in from an input file),
#		= 'rvsd' if it is a revised one (most likely via 'revise_list_gblocks');
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 otherwise.
#
# [ sub print_out_readin_list_gblks (\$$\@) {
#	my ($rf_outdir, $ofname, $list_gblks) = @_; ] ... OBSOLETE as of 2019/06/27.
#
sub print_out_readin_features_gblks (\$$\@$) {

	my ($rf_outdir, $ofname, $features_gblks, $type) = @_; # Modified on 2019/06/27. => Further Modified on 2020/06/04.

	my $ct_gbs = @{$features_gblks};

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_readin_features_gblks'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_readin_features_gblks'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_readin_features_gblks'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $full_type = ($type eq 'org') ? 'original' : 'revised'; # Added on 2019/06/27. #

	##print $fh_out "## The list of \"$full_type\" gap-blocks in the (sub-)window (#{blocks} = $ct_gbs), read via 'readin_list_gblks'.\n\n"; # MODIFIED on 2019/06/27.
	####print $fh_out "## The list of gap-blocks in the (sub-)window (#{blocks} = $ct_gbs), read via 'readin_list_gblks'.\n\n";
	print $fh_out "## The features of \"$full_type\" gap-blocks in the subject window (#{blocks} = $ct_gbs), read via 'readin_features_gblks'.\n\n"; # MODIFIED on 2020/06/04.

	##print $fh_out join ("\t", 'Indx_gb', "br_ID_$type", "up(U)_or_lw(L)_$type", 'lm_clm', 'rm_clm', "(affected,external,nodes)_$type", '[msa,rows,affected]'), "\n\n"; # MODIFIED on 2019/06/27.
	####print $fh_out join ("\t", 'Indx_gb', 'br_ID_org', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(affected,external,nodes)_org', '[msa,rows,affected]'), "\n\n";
	print $fh_out join ("\t", 'Indx_gb', "br_ID_$type", "up(U)_or_lw(L)_$type", 'lm_clm', 'rm_clm', '[msa,rows,affected]'), "\n\n"; # MODIFIED on 2020/06/04.

	for (my $gb=0; $gb <$ct_gbs; $gb++) { # Main for-loop (over gap-blocks). #

	    ##my ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, $set_affd_exnodes_org, $set_affd_msa_rows) = @{$list_gblks->[$gb]} ;
	    my ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_msa_rows, $set_affd_msa_rows) = @{$features_gblks->[$gb]} ;

	    ##$cnct_affd_exnodes_org = "($cnct_affd_exnodes_org)";
	    $cnct_affd_msa_rows = "[$cnct_affd_msa_rows]";

	    ##print $fh_out join ("\t", $gb, $br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows), "\n";
	    print $fh_out join ("\t", $gb, $br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_msa_rows), "\n";

	} # End of the main for-loop (over gap-blocks). #

	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_readin_features_gblks' #


# print_out_wrapper_mk_node2exoffs_msa_rows:
#	This subroutine prints out the outputs of the subroutine, 'wrapper_mk_node2exoffs_msa_rows'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/26.]
#
# INPUTS: (\$outdir, $ofname, \%node2exoffs_msa_rows, \%node2depth, \@nodeids_in_msa, \%nodeid2msa_row), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	The remaining arguments are the outputs of 'wrapper_mk_node2exoffs_msa_rows'.
#	More precisely,
#
#	%node2exoffs_msa_rows = ($node_id => \@exoffs_msa_rows, ...) for all nodes,
#	with @exoffs_msa_rows stores the indices of all MSA rows corresponding to the external offsprings of $node_id;
#
#	%node2depth = ($node_id => $abs_depth, ...) for all nodes;
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $node_ID.
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 otherwise.
#
sub print_out_wrapper_mk_node2exoffs_msa_rows (\$$\%\%\@\%) {

	my ($rf_outdir, $ofname, $node2exoffs_msa_rows, $node2depth2, $nodeids_in_msa, $nodeid2msa_row) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_wrapper_mk_node2exoffs_msa_rows'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_wrapper_mk_node2exoffs_msa_rows'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_wrapper_mk_node2exoffs_msa_rows'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	print $fh_out "## The outputs of 'wrapper_mk_node2exoffs_msa_rows'. ##\n\n";

	print $fh_out '%{$node2depth2} = ( # $node_ID => $depth,', "\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2depth2}) { # 1st foreach-loop. #
	    print $fh_out "\t", $node, ' => ', $node2depth2->{$node}, " ,\n";
	} # End of the 1st foreach-loop. #

	print $fh_out "\t\t) .\n";
	print $fh_out "\n\n";


	print $fh_out '@{$nodeids_in_msa} = (', join (', ', @{$nodeids_in_msa}), ") .\n";
	print $fh_out "\n\n";


	print $fh_out '%{$nodeid2msa_row} = ( # $node_ID => $indx_MSA_row,', "\n";

	foreach my $node (sort {$a <=> $b} keys %{$nodeid2msa_row}) { # 2nd foreach-loop. #
	    print $fh_out "\t", $node, ' => ', $nodeid2msa_row->{$node}, " ,\n";
	} # End of the 2nd foreach-loop. #

	print $fh_out "\t\t) .\n";
	print $fh_out "\n\n";

	print $fh_out '%{$node2exoffs_msa_rows} = ( # $node_ID => [MSA,rows,under,the,node],', "\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2exoffs_msa_rows}) { # 3rd foreach-loop. #
	    my $cnct_exoffs_msa_rows = join (',', @{$node2exoffs_msa_rows->{$node}});
	    print $fh_out "\t", $node, ' => ', "[$cnct_exoffs_msa_rows] ,\n";
	} # End of the 3rd foreach-loop. #

	print $fh_out "\t\t) .\n";
	print $fh_out "\n\n";
	
	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_wrapper_mk_node2exoffs_msa_rows'. #


# print_out_mk_set_msa_rows2br_up_or_lw:
#	This subroutine prints out the outputs of the subroutine, 'mk_set_msa_rows2br_up_or_lw'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/26.]
#
# INPUTS: (\$outdir, $ofname, \%set_msa_rows2br_up_or_lw), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_ID, $up_or_lw], ...), where
#	where $cnct_set_msa_rows = join (',', @set_msa_rows), and @set_msa_rows stores some MSA rows,
#	$br_ID is the branch that separates @set_msa_rows from the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_ID;
#	[\%set_msa_rows2br_up_or_lw is the main output of 'mk_set_msa_rows2br_up_or_lw'];
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 otherwise.
#
sub print_out_mk_set_msa_rows2br_up_or_lw (\$$\%) {

	my ($rf_outdir, $ofname, $set_msa_rows2br_up_or_lw) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_mk_set_msa_rows2br_up_or_lw'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_mk_set_msa_rows2br_up_or_lw'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_mk_set_msa_rows2br_up_or_lw'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	print $fh_out "## The outputs of 'mk_set_msa_rows2br_up_or_lw'. ##\n\n";

	print $fh_out '%{$set_msa_rows2br_up_or_lw} = ( # (concatenated,set,of,msa,rows) => [br, up(U)_or_lw(L)] ,', "\n";

	foreach my $cnct_set_msa_rows (sort keys %{$set_msa_rows2br_up_or_lw}) { # Main foreach-loop. #

	    my ($br, $up_or_lw) = @{$set_msa_rows2br_up_or_lw->{$cnct_set_msa_rows}};
	    print $fh_out "\t($cnct_set_msa_rows) => [$br, $up_or_lw] ,\n";

	} # End of the main foreach-loop. #

	print $fh_out "\t\t) .\n";
	print $fh_out "\n\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_mk_set_msa_rows2br_up_or_lw'. #


# print_out_br2tr_mtrx:
#	This subroutine prints out %br2tr_mtrx, which is the main output of 'calc_br2tr_mtrx_XX' (e.g., in ".
#	It was created based on the test-output block for "Preliminary 1" in "detect_purge_cands.ver0.5.pl".
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/27.]
#
# INPUTS: (\$outdir, $ofname, \%br2tr_mtrx, \@residues), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#       @{$br2tr_mtrx{$br_id}} is the transition matrix in a specified substitution model 
#       that describes the transition probabilities between both ends of the branch $br_id.
#       More specifically,
#       $br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $residues[$j]
#       given that the parent of $br_id shows $residues[$i].
#	
#	$residues[$i] is a character representing the $i th residue (nucleotide) in the "alphabet";
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_br2tr_mtrx (\$$\%\@) {

	my ($rf_outdir, $ofname, $br2tr_mtrx, $residues) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_br2tr_mtrx'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_br2tr_mtrx'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_br2tr_mtrx'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	print $fh_out '## Contents of %{br2tr_mtrx}: Substitution Transition Matrices Assigned to the Branches.', "\n\n";

	my $ct_residues = @{$residues};

	foreach my $br_id (sort {$a <=> $b} keys %{$br2tr_mtrx}) { # Main foreach-loop (over branches). #

	    my $tr_mtrx = $br2tr_mtrx->{$br_id};

	    print $fh_out "( Branch ID: $br_id )\n\n";
	    print $fh_out join ("\t", 'From\To', @{$residues}), "\n\n";
	    for (my $r=0; $r<$ct_residues; $r++) { print $fh_out  join ("\t", $residues->[$r], @{$tr_mtrx->[$r]}), "\n"; }
	    print $fh_out "\n\n";

	} # End of the main foreach-loop (over branches). #

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_br2tr_mtrx'. #


# print_out_node2ntfreqs:
#	This subroutine prints out %node2ntfreqs, which is the main output of 'calc_node2ntfreqs_td'.
#	It was created based on the test-output block for "Preliminary 2" of "detect_purge_cands.ver0.5.pl".
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/27.]
#
# INPUTS: (\$outdir, $ofname, \%node2ntfreqs, \@labels_resfreqs), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	%node2ntfreqs = ($node_id => \@ntfreqs, ...), where
#	$ntfreqs[$i] is the frequency of $NUCLEOTIDES[$i] at the node, $node_id;
#
#	$labels_resfreqs[$i] is a string labeling the frequency of $NUCLEOTIDES[$i].
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_node2ntfreqs (\$$\%\@) {

	my ($rf_outdir, $ofname, $node2ntfreqs, $labels_resfreqs) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_br2tr_mtrx'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_br2tr_mtrx'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_br2tr_mtrx'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}


	print $fh_out "## Contents of ", '%{$node2ntfreqs}', ", the output of 'calc_node2ntfreqs_td' .\n\n";

	print $fh_out join ("\t", 'Node_ID', @{$labels_resfreqs}), "\n\n";

	foreach my $node_id (sort {$a <=> $b} keys %{$node2ntfreqs}) { # Main foreach-loop (over nodes). #

	    my $ntfreqs = $node2ntfreqs->{$node_id};

	    print $fh_out join ("\t", $node_id, @{$ntfreqs}), "\n\n";

	}  # End of the main foreach-loop (over nodes). #

	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_node2ntfreqs'. #


# print_out_node2set_sw_cond_probs:
#	This subroutine prints out either of %node2set_sw_cond_probs and %node2set_sw_extcond_probs,
#	both output by 'calc_prob_descresstats_cond_ancstat2'.
#	It was created based on the test-output block for "(Preliminary 3)" of "detect_purge_cands.ver0.5.pl".
#	[CREATION (in creation of 'anex_for_sgl_wd.ver0.5.pl') STARTED on 2019/06/27, ENDED on 2019/06/28.]
#
# INPUTS: (\$outdir, $ofname, \%node2set_sw_cond_probs, \@residues, $label), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.;
#	( But $set_sw_cond_probs[$s] = [] (empty) if $node_id refers to an external node and if the node has its $s th site occupied by a gap);
#	OR
#	the %node2set_sw_cond_probs MAY actually BE:
#	%node2set_sw_extcond_probs = ($node_id => \@set_sw_extcond_probs, ...) for ALL nodes except the root, where
#	$set_sw_extcond_probs[$s]->[$i] is the probability of the same residue configuration was $set_sw_cond_probs[$s]->[$i],
#	but conditioned on that $NUCLEOTIDES[$i] occupies the $s th site of the PARENT OF the ancestral node, $node_id;
#	[BOTH \%node2set_sw_cond_probs and \%node2set_sw_extcond_probs are the main outputs of 'calc_prob_descresstats_cond_ancstat2' (in "MyLikelihoods_respattern_msa_dna.pm")];
#
#	@residues stores the characters representing the residues (or nucleotides);
#
#	$label (= '%{$node2set_sw_cond_probs}' or '%{$node2set_sw_extcond_probs}')
#	is a string indicating the type of the output.
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_node2set_sw_cond_probs (\$$\%\@$) {

	my ($rf_outdir, $ofname, $node2set_sw_cond_probs, $residues, $label) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_node2set_sw_cond_probs'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_node2set_sw_cond_probs'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0); unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_node2set_sw_cond_probs'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}


	print $fh_out "## Contents of '$label', a major output of 'calc_prob_descresstats_cond_ancstat2' .\n\n";

	foreach my $node_id (sort {$a <=> $b} keys %{$node2set_sw_cond_probs}) { # The main outer foreach-loop (over nodes). #

	    my $set_sw_cond_probs = $node2set_sw_cond_probs->{$node_id};
	    my $ct_sites = @{$set_sw_cond_probs};

	    print $fh_out "{ Node ID: $node_id }\n\n";
	    print $fh_out join ("\t", 'Site\Cond', @{$residues}), "\n\n";

	    for (my $s = 0; $s < $ct_sites; $s++) { # Inner for-loop (over sites).
		print $fh_out join ("\t", $s, @{$set_sw_cond_probs->[$s]}), "\n";
	    } # End of the inner for-loop (over sites).

	    print $fh_out "\n\n";

	} # End of the main outer foreach-loop (over nodes). #

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_node2set_sw_cond_probs'. #


# print_out_br2set_sw_jnt_probs:
#	This subroutine prints out either of %br2set_sw_jnt_probs and %br2set_sw_extjnt_probs,
#	both output by 'calc_joint_prob_ancresstats_and_ancstat2'.
#	It was created based on the test-output block for "(Preliminary 5)" of "detect_purge_cands.ver0.5.pl".
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/28.]
#
# INPUTS: (\$outdir, $ofname, \%br2set_sw_jnt_probs, \@residues, $label), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#       %br2set_sw_jnt_probs = ($br_id => \@set_sw_jnt_probs, ...) for ALL branches, where 
#       $set_sw_jnt_probs[$s]->[$i] is the probability that
#       all the extant sequences on the "ancestral side" of $br_id show the residue states that are actually found in the $s th site,
#       in addition to that $NUCLEOTIDES[$i] occupies the $s th site of the UPPER-end node of $br_id.
#	OR
#	the %br2set_sw_jnt_probs MAY actually BE:
#       %br2set_sw_extjnt_probs = ($br_id => \@set_sw_extjnt_probs, ...) for ALL branches, where 
#       $set_sw_extjnt_probs[$s]->[$i] is the probability very similar to $set_sw_jnt_probs[$s]->[$i],
#       with the only difference that the UPPER-end node of $br_id is now replaced with its LOWER-end node.
#	[BOTH \%br2set_sw_jnt_probs and \%br2set_sw_extjnt_probs are the main outputs of 'calc_joint_prob_ancresstats_and_ancstat2' (in "MyLikelihoods_respattern_msa_dna.pm")];
#
#	@residues stores the characters representing the residues (or nucleotides);
#
#	$label (= '%{$br2set_sw_jnt_probs}' or '%{$br2set_sw_extjnt_probs}')
#	is a string indicating the type of the output.
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_br2set_sw_jnt_probs (\$$\%\@$) {

	my ($rf_outdir, $ofname, $br2set_sw_jnt_probs, $residues, $label) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_br2set_sw_jnt_probs'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_br2set_sw_jnt_probs'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0); unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_br2set_sw_jnt_probs'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}


	print $fh_out "## Contents of '$label', a main output of 'calc_joint_prob_ancresstats_and_ancstat2' .\n\n";

	foreach my $br_id (sort {$a <=> $b} keys %{$br2set_sw_jnt_probs}) { # The main outer foreach-loop (over branches). #

	    my $set_sw_jnt_probs = $br2set_sw_jnt_probs->{$br_id};
	    my $ct_sites = @{$set_sw_jnt_probs};

	    print $fh_out "{ Branch ID: $br_id }\n\n";
	    print $fh_out join ("\t", 'Site\Jnt', @{$residues}), "\n\n";

	    for (my $s = 0; $s < $ct_sites; $s++) { # Inner for-loop (over sites).
		print $fh_out join ("\t", $s, @{$set_sw_jnt_probs->[$s]}), "\n";
	    } # End of the inner for-loop (over sites).

	    print $fh_out "\n\n";

	} # End of the main outer foreach-loop (over branches). #

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_br2set_sw_jnt_probs'. #


# print_out_set_clm_probs:
#	This subroutine prints out @set_clm_probs,
#	a main output of 'calc_clm_probs' (in "MyLikelihoods_respattern_msa_dna.pm").
#	It was created based on the test-output block for "(Preliminary 6)" of "detect_purge_cands.ver0.5.pl".
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/29.]
#
# INPUTS: (\$outdir, $ofname, \@set_clm_probs), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	$set_clm_probs[$s] is the occurrence probability of the $s th column
#	under the given substitution model.
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_set_clm_probs (\$$\@) {

	my ($rf_outdir, $ofname, $set_clm_probs) = @_;

	my $ct_clms = @{$set_clm_probs};

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_set_clm_probs'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_set_clm_probs'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0); unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_set_clm_probs'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	print $fh_out "## Contents of ", '@{$set_clm_probs}', " (#{columns} = $ct_clms), the main output of 'calc_clm_probs' .\n\n";

	print $fh_out join ("\t", 'Column', 'Probability'), "\n\n";

	for (my $c = 0; $c < $ct_clms; $c++) { # Main for-loop (over the columns).

	    print $fh_out join ("\t", $c, $set_clm_probs->[$c]), "\n";

	} # End of the main for-loop (over the columns).
	print $fh_out "\n\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_set_clm_probs'. #


# print_out_orgs_bds_gb_coords:
#	This subroutine prints out the outputs of 'mk_orgs_bds_gb_coords' (in "MyANEX_Supple.pm").
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/29.]
#
# INPUTS: (\$outdir, $ofname, \@orgs_gb_coords, \@bds_gb_coords, \@gb_coords_wd_frm), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	$org_coord = $orgs_gb_coords[$gb] gives the origin of the coordinate for the $gb th gap-block;
#
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords[$gb]} give the (accessible) left- and right-boundaries of the coordinate for the $gb th gap-block;
#
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]} give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window;
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_orgs_bds_gb_coords (\$$\@\@\@) {

	my ($rf_outdir, $ofname, $orgs_gb_coords, $bds_gb_coords, $gb_coords_wd_frm) = @_;

	my $ct_gbs = @{$orgs_gb_coords};

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_orgs_bds_gb_coords'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_orgs_bds_gb_coords'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0); unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_orgs_bds_gb_coords'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	print $fh_out '## Contents of @{$orgs_gb_coords} & @{$bds_gb_coords} & @{$gb_coords_wd_frm},', " main outputs of 'mk_orgs_bds_gb_coords'.\n\n";
	print $fh_out "## (#{gap-blocks} = $ct_gbs)\n\n";

	print $fh_out join ("\t", 'Indx_gb', 'org_coord', 'lb_coord', 'rb_coord', 'gb_coord_le_wd', 'gb_coord_re_wd'), "\n\n";

	for (my $gb=0; $gb < $ct_gbs; $gb++) { # Main for-loop (over gap-blocks). #

	    print $fh_out join ("\t", $gb, $orgs_gb_coords->[$gb], @{$bds_gb_coords->[$gb]}, @{$gb_coords_wd_frm->[$gb]}), "\n";

	} # End of the main for-loop (over gap-blocks).

	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_orgs_bds_gb_coords'. #


# print_out_row_clm_info_gblocks:
#	This subroutine prints out information on the rows & columns defining the gap-blocks.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/29.]
#
# INPUTS: (\$outdir, $ofname, \@bds_gblocksd, \@rows_affd_by_blks), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	($le_gb, $re_gb) = @{$bds_gblocks[$gb]} are the left-end and right-end of the $gb th gap-block;
#
#	@{$rows_affd_by_blks[$gb]} stores indices of the MSA rows affected by the $gb th gap-block;
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_row_clm_info_gblocks (\$$\@\@) {

	my ($rf_outdir, $ofname, $bds_gblocks, $rows_affd_by_blks) = @_;

	my $ct_gbs = @{$bds_gblocks};

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_row_clm_info_gblocks'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_row_clm_info_gblocks'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0); unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_row_clm_info_gblocks'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	print $fh_out '## Contents of @bds_gblocks & @rows_affd_by_blks.', "  (#{gap-blocks}=$ct_gbs)\n\n";

	print $fh_out join ("\t", 'Indx_gb', 'le_gb', 're_gb', '(MSA,rows,affected,by,gb)'), "\n\n";

	for (my $gb=0; $gb <$ct_gbs; $gb++) { # Main for-loop (over gap-blocks). #

	    my $cnct_rows_affd = join (',', @{$rows_affd_by_blks->[$gb]});
	    print $fh_out join ("\t", $gb, @{$bds_gblocks->[$gb]}, "($cnct_rows_affd)"), "\n";

	} # End of the main for-loop (over gap-blocks). #

	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_row_clm_info_gblocks'. #


# print_out_inter_block_relations:
#	This subroutine prints out @inter_block_relations,
#	the main output of 'mk_inter_block_relations' (in "MyGapPatternTopology.pm").
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/06/29.]
#
# INPUTS: (\$outdir, $ofname, \@inter_block_relations), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1, where
#	$relations is a string coding the relation;
#	[See the header of 'mk_inter_block_relations' for details];
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_inter_block_relations (\$$\@) {

	my ($rf_outdir, $ofname, $inter_block_relations) = @_;

	my $ct_gbs = @{$inter_block_relations};

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_inter_block_relations'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_inter_block_relations'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_inter_block_relations'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	print $fh_out '## Contents of @{$inter_block_relations},', " the main output of 'mk_inter_block_relations' .\n\n";

	print $fh_out join ("\t", 'Gb1\Gb2', 0 .. $ct_gbs-1), "\n\n";

	for (my $gb1=0; $gb1 <$ct_gbs; $gb1++) { # Main for-loop (over gap-blocks). #

	    my @out_rels_w_gb1 = @{$inter_block_relations->[$gb1]};
	    $out_rels_w_gb1[$gb1] = 'undef';

	    print $fh_out join ("\t", $gb1, @out_rels_w_gb1), "\n";

	}  # End of the main for-loop (over gap-blocks). #

	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_inter_block_relations'. #


# print_out_features_gblks:
#	This subroutine prints out main features of gap-blocks in a (sub-)window.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/02.]
#
#	NOTE: The print-out format is quite similar to but slightly different from that by 'print_out_readin_list_gblks' above.
#
# INPUTS: (\$outdir, $ofname, \@list_gblks, $indx_br, $indx_u_or_d, \@bds_gblks, \@rows_affd_by_gblks), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	($br_ID, $up_or_lw) = @{$list_gblks[$gb]}[$indx_br, $indx_u_or_d] provides information on the $gb th gap-block,
#	more precisely, the $gb th block is on the upper/lower-side of the branch (ID: $br_ID) if $up_or_lw = 'U'/'L';
#
#	($le_gb, $re_gb) = @{$bds_gblks[$gb]} are the left-end and right-end of the $gb th gap-block;
#	@{$rows_affd_by_blks[$gb]} stores indices of the MSA rows affected by the $gb th gap-block;
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_features_gblks (\$$\@$$\@\@) {

	my ($rf_outdir, $ofname, $list_gblks, $indx_br, $indx_u_or_d, $bds_gblks, $rows_affd_by_gblks) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_features_gblks'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_features_gblks'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_features_gblks'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_gbs = @{$list_gblks};

	print $fh_out "## #{gap-blocks} = $ct_gbs .\n\n";

	print $fh_out join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'le_gb', 're_gb', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb <$ct_gbs; $gb++) { # Main for-loop (over gap-blocks). #

	    my ($br, $up_or_lw) = @{$list_gblks->[$gb]}[$indx_br, $indx_u_or_d];
	    my ($le_gb, $re_gb) = @{$bds_gblks->[$gb]};

	    ##my $cnct_rows_affd = join (',', @{$rows_affd_by_gblks}); #  A BUG fixed on 2020/05/08. #
	    my $cnct_rows_affd = join (',', @{$rows_affd_by_gblks->[$gb]});

	    print $fh_out join ("\t", $gb, $br, $up_or_lw, $le_gb, $re_gb, "[$cnct_rows_affd]"), "\n";

	} # End of the main for-loop (over gap-blocks). #

	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_features_gblks'. #


# print_out_x_bin_hrels:
#	This subroutine prints out a list of prohibited binary horizontal relationships.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/02.]
#
# INPUTS: (\$outdir, $ofname, \@list_excluded_bin_hrels), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@list_excluded_bin_hrels lists the PROHIBITED binary horizontal relationships,
#	which is a main input of 'chk_excluded_bin_hrels';
#	More precisely,
#	($gb1, $gb2, $hrel) = @{$list_excluded_bin_hrels[$hr]} provide information on the $hr th horizontal relationship in the list, where
#	$gb1 and $gb2 are indices of the pair of gap-blocks in question,
#	and $hrel codes the horizontal relationships to be examined;
#	[See the header of 'nonrec_mlt_shifts_gapblocks' (in "MyANEX_Main.pm") for more details];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_x_bin_hrels (\$$\@) {

	my ($rf_outdir, $ofname, $list_x_bin_hrels) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_x_bin_hrels'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_x_bin_hrels'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_x_bin_hrels'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_xbhrels = @{$list_x_bin_hrels};

	print $fh_out "#{excluded binary horizontal relations} = $ct_xbhrels .\n\n";

	print $fh_out join ("\t", 'Indx_gb1', 'indx_gb2', 'hrel'), "\n\n";

	for (my $i=0; $i <$ct_xbhrels; $i++) { # Main for-loop (over relations).

	    print $fh_out join ("\t", @{$list_x_bin_hrels->[$i]}), "\n";

	} # End of the main for-loop (over relations).

	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_x_bin_hrels'. #


# print_in_nr_mshifts_glks:
#	This subroutine prints out the input data of 'nonrec_mlt_shifts_gapblocks'
#	that are specific to each session of the exploration.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/02.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine recieves a list of the output file names as well.
#	[REVISED on 2019/07/19]:
#	From now on, $beg_wd and $end_wd MUST BE the values for the INITIAL INPUT MSA.
#
# [ INPUTS: (\$outdir, \@list_gblks, $indx_br, $indx_u_or_d, \@bds_gblks, \@rows_affd_by_gblks, \@orgs_gb_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@set_columns, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, \@list_excluded_bin_hrels), where ] ... OBSOLETE as of 2019/07/18.
# INPUTS: (\$outdir, \@ofnames, \@list_gblks, $indx_br, $indx_u_or_d, \@bds_gblks, \@rows_affd_by_gblks, \@orgs_gb_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@set_columns, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, \@list_excluded_bin_hrels), where
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames lists the names of the 4 output files;
#
#	($br_ID, $up_or_lw) = @{$list_gblks[$gb]}[$indx_br, $indx_u_or_d] provides information on the $gb th gap-block,
#	more precisely, the $gb th block is on the upper/lower-side of the branch (ID: $br_ID) if $up_or_lw = 'U'/'L';
#
#	($le_gb, $re_gb) = @{$bds_gblks[$gb]} are the left-end and right-end of the $gb th gap-block;
#	@{$rows_affd_by_blks[$gb]} stores indices of the MSA rows affected by the $gb th gap-block;
#
#	$org_coord = $orgs_gb_coords[$gb] gives the origin of the coordinate for the $gb th gap-block;
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords[$gb]} give the (accessible) left- and right-boundaries of the coordinate for the $gb th gap-block;
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]} give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window;
#
#	@column = @{$set_columns[$c]} specifies the $c th column in the input local MSA. more precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input alignment
#	that accommodates the corresponding site of the $k th "sequence" (or row);
#
#	[ $beg/end_wd is the MSA site (column) at the beginning/end of the subject window; ] ... OBSOLETE as of 2019/07/19.
#	$beg/end_wd is site (column) of the initial INPUT MSA at the beginning/end of the subject window; # REVISED on 2019/07/19.
#
#	$if_on_le/re_msa = 1 if the subject window is on the left/right-end of the MSA,
#	                 = 0 otherwise;
#
#	@list_excluded_bin_hrels lists the PROHIBITED binary horizontal relationships,
#	which is a main input of 'chk_excluded_bin_hrels';
#	More precisely,
#	($gb1, $gb2, $hrel) = @{$list_excluded_bin_hrels[$hr]} provide information on the $hr th horizontal relationship in the list, where
#	$gb1 and $gb2 are indices of the pair of gap-blocks in question,
#	and $hrel codes the horizontal relationships to be examined;
#	[See the header of 'nonrec_mlt_shifts_gapblocks' (in "MyANEX_Main.pm") for more details];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
# [ sub print_in_nr_mshifts_glks (\$\@$$\@\@\@\@\@\@$$$$\@) {
#	my ($rf_outdir, $list_gblks, $indx_br, $indx_u_or_d, $bds_gblks, $rows_affd_by_gblks, $orgs_gb_coords, $bds_gb_coords, $gb_coords_wd_frm, $set_columns, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, $list_excluded_bin_hrels) = @_; ] ... OBSOLETE as of 2019/07/18.
#
sub print_in_nr_mshifts_glks (\$\@\@$$\@\@\@\@\@\@$$$$\@) { # REVISED on 2019/07/18.

	my ($rf_outdir, $ofnames, $list_gblks, $indx_br, $indx_u_or_d, $bds_gblks, $rows_affd_by_gblks, $orgs_gb_coords, $bds_gb_coords, $gb_coords_wd_frm, $set_columns, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, $list_excluded_bin_hrels) = @_; # REVISED on 2019/07/18.

	unless (@{$ofnames} == 4) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_purge'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 4 as desired: $! !\n\n";
	    return 0;
	}

	my $ct_gbs = @{$list_gblks};

   (print STDOUT "< Subroutine 'print_in_nr_mshifts_glks' (#{gap-blocks}=$ct_gbs) >\n\n") if ($TEST_LEVEL_SUB >0);

	my $outdir = ${$rf_outdir};

	my $ofname1 = $ofnames->[0]; # REVISED on 2019/07/18.
	##my $ofname1 = 'init_features_gblocks.txt';

	unless ( print_out_features_gblks ($outdir, $ofname1, @{$list_gblks}, $indx_br, $indx_u_or_d, @{$bds_gblks}, @{$rows_affd_by_gblks}) ) {
	    print STDOUT "[ERROR 'print_in_nr_mshifts_glks'] The subroutine 'print_out_features_gblks' FAILED on the output file \"$outdir/$ofname1\": $! !\n\n";
	    return 0;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_features_gblks', Ended on the output file, \"$outdir/$ofname1\" ...\n\n") if ($TEST_LEVEL_SUB >1);


	my $ofname2 = $ofnames->[1]; # REVISED on 2019/07/18.
	##my $ofname2 = 'orgs_bds_gb_coords.txt';

	unless ( print_out_orgs_bds_gb_coords ($outdir, $ofname2, @{$orgs_gb_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}) ) { 
	    print STDOUT "[ERROR 'print_in_nr_mshifts_glks'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the output file \"$outdir/$ofname2\": $! !\n\n";
	    return 0;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the output file, \"$outdir/$ofname2\" ...\n\n") if ($TEST_LEVEL_SUB >1);


	my $ofname3 = $ofnames->[2]; # REVISED on 2019/07/18.
	##my $ofname3 = 'init_set_columns.txt';

	unless ( print_out_set_clms ($outdir, $ofname3, @{$set_columns}, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa) ) {
	    print STDOUT "[ERROR 'print_in_nr_mshifts_glks'] The subroutine 'print_out_set_clms' FAILED on the output file \"$outdir/$ofname3\": $! !\n\n";
	    return 0;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_set_clms', Ended on the output file, \"$outdir/$ofname3\" ...\n\n") if ($TEST_LEVEL_SUB >1);


	my $ofname5 = $ofnames->[3]; # REVISED on 2019/07/18.
	##my $ofname5 = 'excluded_bin_hrels.txt';

	unless ( print_out_x_bin_hrels ($outdir, $ofname5, @{$list_excluded_bin_hrels}) ) {
	    print STDOUT "[ERROR 'print_in_nr_mshifts_glks'] The subroutine 'print_out_x_bin_hrels' FAILED on the output file \"$outdir/$ofname5\": $! !\n\n";
	    return 0;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_x_bin_hrels', Ended on the output file, \"$outdir/$ofname5\" ...\n\n") if ($TEST_LEVEL_SUB >1);


   (print STDOUT "< Subroutine 'print_in_nr_mshifts_glks' Ended />\n\n") if ($TEST_LEVEL_SUB >0);

	return 1;

}	# End of the subroutine, 'print_in_nr_mshifts_glks'. #


# print_out_ln_prb_plus_msa:
#	This subroutine prints out log-probabilities and some other quantities on an MSA (especially a local one).
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/02.]
#
# INPUTS: (\$outdir, $ofname, $ct_null_clms, $degeneracy, $minct_indels, $ln_prb_msa, $ln_prb_sbst, $ln_prb_indel, $string_undef), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	$ct_null_clms, $degeneracy, $minct_indels, $ln_prb_msa, $ln_prb_sbst, $ln_prb_indel are the quantities characterizing the subject MSA;
#	More precisely,
#	$ct_null_clms is the number of null columns,
#	$degeneracy is the number of occurrences of equivalent MSAs in the coordinate frame,
#	$minct_indels is the minimum number of indels,
#	$ln_prb_msa is the log-probability of the MSA,
#	$ln_prb_sbst is the substitution component of $ln_prb_msa,
#	$ln_prb_indel is the indel component of $ln_prb_msa.
#	[See the header of 'nonrec_mlt_shifts_gapblocks' (in "MyANEX_Main.pm") for more details];
#
#	$string_undef is the string designated for indicating that the component is UNDEFINED.
#	(Commonly used ones are: 'undef', 'n/a', and 'N/A'.)
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_ln_prb_plus_msa (\$$$$$$$$$) {

	my ($rf_outdir, $ofname, $ct_null_clms, $degeneracy, $minct_indels, $ln_prb_msa, $ln_prb_sbst, $ln_prb_indel, $string_undef) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_ln_prb_plus_msa'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_ln_prb_plus_msa'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_ln_prb_plus_msa'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my @outputs = ();
	foreach my $qt ($ct_null_clms, $degeneracy, $minct_indels, $ln_prb_msa, $ln_prb_sbst, $ln_prb_indel) {
	    push @outputs, ((defined $qt) ? $qt : $string_undef);
	}

	print $fh_out '$ct_null_clms = ', $outputs[0], " ,\n";
	print $fh_out '$degeneracy = ', $outputs[1], " ,\n";
	print $fh_out '$minct_indels = ', $outputs[2], " ,\n"; 
	print $fh_out '$ln_prb_msa = ', $outputs[3], " ,\n";
	print $fh_out '$ln_prb_sbst = ', $outputs[4], " ,\n";
	print $fh_out '$ln_prb_indel = ', $outputs[5], " .\n\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_ln_prb_plus_msa'. #


# print_out_coord_frame:
#	This subroutine prints out the coordinate frame in an UN-OFFSET format.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/02.]
#
# INPUTS: (\$outdir, $ofname, \@orgs_gb_coords, \@bds_gb_coords), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	$org_bl_coords[$bl] is the position of the "origin" of the coordinate of the $bl th block.
#	(Although the "origin" should actually be ZERO, it is SHIFTED to this value, SO THAT the range of the coordinate will become entirely non-negaitve.);
#
#	($lbd_coord, $rbd_coord) = @{$bds_bl_coords[$bl]}
#	are the left- and right-bounds of the coordinate of the $bl th block.
#	[ASSUME that 0 <= $lbd_coord <= $rbd_coord holds.]
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_coord_frame (\$$\@\@) {

	my ($rf_outdir, $ofname, $orgs_gb_coords, $bds_gb_coords) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_coord_frame'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_coord_frame'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_coord_frame'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_gbs = @{$orgs_gb_coords};

	print $fh_out "# The Coordinate Boundaries (inclusive) of the $ct_gbs Gap-Blocks.\n\n";

	print $fh_out join ("\t", 'Coordinate', 'lower_bound', 'upper_bound'), "\n\n";

	for (my $gb=0; $gb<$ct_gbs; $gb++) { # Inner for-loop (over gap-blocks). #

		my $org_coord = $orgs_gb_coords->[$gb];
		my ($lb_coord, $rb_coord) = @{$bds_gb_coords->[$gb]};
		$lb_coord -= $org_coord;
		$rb_coord -= $org_coord;

		print $fh_out join ("\t", "X$gb", $lb_coord, $rb_coord), "\n";

	} # End of the inner for-loop (over gap-blocks). #

	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_coord_frame'. #


# print_out_nr_mshifts_gblks:
#	This subroutine prints out the outputs of 'nonrec_mlt_shifts_gapblocks' (in "MyANEX_Main.pm").
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/02.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine recieves a list of the output file names as well.
#
# [ INPUTS: (\$outdir, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0,
#		 \@diffs_ln_prbs_plus, \@orgs_gb_coords, \@bds_gb_coords, $string_undef), where ] ... OBSOLETE as of 2019/07/18.
# INPUTS: (\$outdir, \@ofnames, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0,
#		 \@diffs_ln_prbs_plus, \@orgs_gb_coords, \@bds_gb_coords, $string_undef), where
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 8 output files;
#
#	$ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0
#	are the properties of the alignment at the "origin" of the multiple-"shift"s;
#	[See the header of 'nonrec_mlt_shifts_gapblocks' (in "MyANEX_Main.pm") for more details];
#
#	\@diffs_ln_prbs_plus is the MAIN output of 'nonrec_mlt_shifts_gapblocks' (in "MyANEX_Main.pm");
#	More precisely,
#	$diffs_ln_prbs_plus[$k_0]->...->[$k_B] = [$ct_null_clms, $degeneracy, $diff_minct_indels, $diff_ln_prb_msa, $diff_ln_prb_sbst, $diff_ln_prb_indel]
#	provide information on the local alignment resulting from the
#	shifts of the 0 th gap-block by ($k_0 - $org_bl_coords[0]) columns,
#	..., and the B (= #{blocks} -1) by ($k_B - $org_bl_coords[B]) columns.
#	Here,
#	$ct_null_clms is the number of null columns in the alignment,
#	$degeneracy is the degree of degeneracy of the alignment,
#	$diff_minct_indels is the difference of the minimum number of indel events (compared to $minct_indels0),
#	$diff_ln_prb_msa (= $dff_ln_prb_sbst + $diff_ln_prb_indel) is the difference of the total log-probability
#	of the alignment (with the degeneracy factor NOT subtracted) compared to $ln_prb_msa0,
#	$diff_ln_prb_sbst is the difference of the log-probability of the alignment under the given substitution model (compared to $ln_prb_sbst0),
#	$diff_ln_prb_indel is the difference of the log-probability of the alignment under the given indel model (compared to $ln_prb_indel0).
#
#	$org_bl_coords[$bl] is the position of the "origin" of the coordinate of the $bl th block.
#	(Although the "origin" should actually be ZERO, it is SHIFTED to this value, SO THAT the range of the coordinate will become entirely non-negaitve.);
#
#	($lbd_coord, $rbd_coord) = @{$bds_bl_coords[$bl]}
#	are the left- and right-bounds of the coordinate of the $bl th block.
#	[ASSUME that 0 <= $lbd_coord <= $rbd_coord holds.]
#
#	$string_undef is the string designated for indicating that the component is UNDEFINED.
#	(Commonly used ones are: 'undef', 'n/a', and 'N/A'.)
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
# [ sub print_out_nr_mshifts_gblks (\$$$$$$$\@\@\@$) {
#	my ($rf_outdir, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0,
#		$diffs_ln_prbs_plus, $orgs_gb_coords, $bds_gb_coords, $string_undef) = @_; ] ... OBSOLETE as of 2019/07/18.
#
sub print_out_nr_mshifts_gblks (\$\@$$$$$$\@\@\@$) { # REVISED on 2019/07/18.

	my ($rf_outdir, $ofnames, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0,
		$diffs_ln_prbs_plus, $orgs_gb_coords, $bds_gb_coords, $string_undef) = @_; # REVISED on 2019/07/18.

	my $ct_ofs = @{$ofnames}; # ADDED on 2019/07/18.
	unless ($ct_ofs == 8) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_purge'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofs elements, NOT 8 as desired: $! !\n\n";
	    return 0;
	}

	my $ct_gbs = @{$orgs_gb_coords};

   (print STDOUT "< Subroutine 'print_out_nr_mshifts_glks' (#{gap-blocks}=$ct_gbs) >\n\n") if ($TEST_LEVEL_SUB >0);

	my $outdir = ${$rf_outdir};

	my $ofname0 = $ofnames->[0]; # REVISED on 2019/07/18.
	##my $ofname0 = 'ln_prb_plus_msa_at_origin.txt';

	unless ( print_out_ln_prb_plus_msa ($outdir, $ofname0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $string_undef) ) {
	    print STDOUT "[ERROR 'print_out_nr_mshifts_glks'] The subroutine 'print_out_ln_prb_plus_msa' FAILED on the output file \"$outdir/$ofname0\": $! !\n\n";
	    return 0;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_ln_prb_plus_msa', Ended on the output file, \"$outdir/$ofname0\" ...\n\n") if ($TEST_LEVEL_SUB >1);


	my $ofname1 = $ofnames->[1]; # REVISED on 2019/07/18.
	##my $ofname1 = 'coord_frame.txt';

	unless ( print_out_coord_frame ($outdir, $ofname1, @{$orgs_gb_coords}, @{$bds_gb_coords}) ) {
	    print STDOUT "[ERROR 'print_out_nr_mshifts_glks'] The subroutine 'print_out_coord_frame' FAILED on the output file \"$outdir/$ofname1\": $! !\n\n";
	    return 0;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_coord_frame', Ended on the output file, \"$outdir/$ofname1\" ...\n\n") if ($TEST_LEVEL_SUB >1);


	##my @ofnames = ('ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # DISABLED on 2019/07/18.
		# $diffs_ln_prbs_plus[$k_0]->...->[$k_B] = [$ct_null_clms, $degeneracy, $diff_minct_indels, $diff_ln_prb_msa, $diff_ln_prb_sbst, $diff_ln_prb_indel]

	##my $ct_ofs = @ofnames; # DISABLED on 2019/07/18.

	for (my $indx_po =0; $indx_po <$ct_ofs-2; $indx_po++) { # Main for-loop (over printed-out elements). # REVISED on 2019/07/18.
	##for (my $indx_po =0; $indx_po <$ct_ofs; $indx_po++) { # Main for-loop (over printed-out elements). #

	    my $ofname = $ofnames->[$indx_po+2]; # REVISED on 2019/07/18.
	    ##my $ofname = $ofnames[$indx_po];
	    my $outfile = "$outdir/$ofname";
    
	    my $fh_out = open_output_file ($outfile, 1);
	    unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_nr_mshifts_gblks'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	    }

	    unless ( printout_multidim_storage (@{$orgs_gb_coords}, @{$bds_gb_coords}, @{$diffs_ln_prbs_plus}, $indx_po, *${$fh_out}, $string_undef) ) {
		print STDOUT "[ERROR 'print_out_nr_mshifts_glks'] The subroutine 'printout_multidim_storage' FAILED (to printout the $indx_po th elements) on the filehandle of the output file \"$outfile\": $! !\n\n";
		return 0;
	}
    (print STDOUT "\n ... The subroutine, 'printout_multidim_storage', Ended (printing-out the $indx_po th elements) on the filehandle of the output file, \"$outfile\" ...\n\n") if ($TEST_LEVEL_SUB >1);

	    close ($fh_out);

	} # End of the main for-loop (over printed-out elements).


   (print STDOUT "< Subroutine 'print_out_nr_mshifts_glks' Ended. />\n\n") if ($TEST_LEVEL_SUB >0);

	return 1;

}	# End of the subroutine, 'print_out_nr_mshifts_gblks'. #


# print_out_list_gblock_sizes:
#	This subroutine prints out \@gblock_sizes, which is the main output of the subroutine,
#	'mk_list_gblock_sizes' (in "MyGapPatternTopology.pm").
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/16.]
#
# INPUTS: (\$outdir, $ofname, \@gblock_sizes), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	$gblock_sizes[$gb] is the (horizontal) size of the $gb th gap-block;
#	[Typically, \@gblock_sizes is the main output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"];
#	
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_list_gblock_sizes (\$$\@) {

	my ($rf_outdir, $ofname, $gblock_sizes) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_list_gblock_sizes'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_list_gblock_sizes'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_list_gblock_sizes'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_gblks = @{$gblock_sizes};

	print $fh_out "# The Sizes of the Gap-Blocks in the Window (#{gblocks}=$ct_gblks):\n\n";

	print $fh_out join ("\t", 'Indx_gb', 'size'), "\n\n";

	for (my $gb=0; $gb <$ct_gblks; $gb++) {
	    print $fh_out join ("\t", $gb, $gblock_sizes->[$gb]), "\n";
	}
	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_list_gblock_sizes'. #


# print_out_purge_or_merge_cands: 
#	This subroutine prints out a list of gap-block pairs that are attempted to "purge" or "merge".
#	The list is typically an output of 'list_nn_purge_and_cmerge_candidates' or ''
#	(both in "MyANEX_Main.pm").
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/16.]
#
# INPUTS: (\$outdir, $ofname, \@candidate_pairs, \$label), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@{$candidate_pairs[$cp]} = ($gb1, $gb2, $dist) provides information on the $cp th pair,
#	where $gb1 and $gb2 are the indices of the gap-blocks in the pair,
#	$dist is the distance of $gb2 against $gb1;
#	[Typically, \@candidate_pairs is an output of 'list_nn_purge_and_cmerge_candidates' or ''
#	, both in "MyANEX_Main.pm"];
#
#	$label specifies what type the pairs in @candidate_pairs are;
#	
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_purge_or_merge_cands (\$$\@\$) {

	my ($rf_outdir, $ofname, $candidate_pairs, $rf_label) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_purge_or_merge_cands'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_purge_or_merge_cands'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_purge_or_merge_cands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_cps = @{$candidate_pairs};

	my $label = ${$rf_label};
	print $fh_out "# There are $ct_cps caididate gap-block pairs $label :\n\n";

	print $fh_out join ("\t", 'Indx_pair', 'indx_gb1', 'indx_gb2', 'distance'), "\n\n";

	for (my $cp=0; $cp <$ct_cps; $cp++) {

	    print $fh_out join ("\t", $cp, @{$candidate_pairs->[$cp]}), "\n";
	}
	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_purge_or_merge_cands'. #


# print_out_merge_split_cands: 
#	This subroutine prints out a list of gap-block pairs that are attempted to "merge + split".
#	The list is typically an output of 'list_smerge_split_candidates' (in "MyANEX_Main.pm").
#	It was created by modifying 'print_out_purge_or_merge_cands' above.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/01.]
#
# print_out_purge_or_merge_cands: 
#	This subroutine prints out a list of gap-block pairs that are attempted to "purge" or "merge".
#	The list is typically an output of 'list_nn_purge_and_cmerge_candidates' or ''
#	(both in "MyANEX_Main.pm").
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/16.]
#
# INPUTS: (\$outdir, $ofname, \@candidate_pairs, \$label, \$label_sizes), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@{$candidate_pairs[$cp]} = ($gb1, $gb2, $dist, \@sizes) provides information on the $cp th pair,
#	where $gb1 and $gb2 are the indices of the gap-blocks in the pair,
#	$dist is the distance of $gb2 against $gb1,
#	@sizes stores some sizes typical of the 'merge + split' process;
#	[Typically, \@candidate_pairs is an output of 'list_smerge_split_candidates' in "MyANEX_Main.pm"];
#
#	$label specifies what type the pairs in @candidate_pairs are;
#
#	$label_sizes labels the field for the contents of \@sizes;
#	
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_merge_split_cands (\$$\@\$\$) {

	my ($rf_outdir, $ofname, $candidate_pairs, $rf_label, $rf_label_sizes) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_merge_split_cands'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_merge_split_cands'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_merge_split_cands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_cps = @{$candidate_pairs};

	my $label = ${$rf_label};
	print $fh_out "# There are $ct_cps caididate gap-block pairs $label :\n\n";

	my $label_sizes = ${$rf_label_sizes};
	print $fh_out join ("\t", 'Indx_pair', 'indx_gb1', 'indx_gb2', 'distance', "[$label_sizes]"), "\n\n";

	for (my $cp=0; $cp <$ct_cps; $cp++) {

	    my $info_cpair = $candidate_pairs->[$cp];
	    my $cnct_sizes = join (',', @{$info_cpair->[3]});
	    print $fh_out join ("\t", $cp, @{$info_cpair}[0 .. 2], "[$cnct_sizes]"), "\n";
	}
	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_merge_split_cands'. #


# print_out_split_cands: 
#	This subroutine prints out a list of gap-blocks that are attempted to "split".
#	The list is typically an output of 'list_{s,c}split_candidates' (in "MyANEX_Main.pm").
#	It was created by modifying 'print_out_merge_split_cands' above.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/13.]
#
# print_out_merge_split_cands: 
#	This subroutine prints out a list of gap-block pairs that are attempted to "merge + split".
#	The list is typically an output of 'list_smerge_split_candidates' (in "MyANEX_Main.pm").
#	It was created by modifying 'print_out_purge_or_merge_cands' above.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/01.]
#
# print_out_purge_or_merge_cands: 
#	This subroutine prints out a list of gap-block pairs that are attempted to "purge" or "merge".
#	The list is typically an output of 'list_nn_purge_and_cmerge_candidates' or ''
#	(both in "MyANEX_Main.pm").
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/16.]
#
# INPUTS: (\$outdir, $ofname, \@candidate_gblocks, \$label_split, \$label_sizes), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@{$candidate_gblocks[$cg]} = ($indx_gb, $dist, \@sizes, \@indices_rlv_pcs, $indx_closest_rlv_pc, \@available_sites)
#	provides informaiton on the $cg th candidate gap-block to be "split" into two blocks,
#	where
#	$dist is the distance between the gap-block and the nearest (and relevant) purge-candidate region,
#	@sizes stores the sizes characterizing the "split", or the newly created gap-blocks,
#	@indices_rlv_pcs stores the indices of the relevant purge-candidate regions,
#	$indx_closest_rlv_pc is the index of the closest relevant purge-candidate region,
#	@available_sites stores the sites available for the complementary split (to be used, e.g., in 'csplit_gblock'); it is empty if NOT necessary;
#	[Typically, \@candidate_gblocks is an output of 'list_{s,c}split_candidates' in "MyANEX_Main.pm"];
#
#	$label_split specifies what type of "split" will be performed on the gap-blocks in @candidate_gblocks;
#
#	$label_sizes labels the field for the contents of \@sizes;
#	
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_split_cands (\$$\@\$\$) {

	my ($rf_outdir, $ofname, $candidate_gblocks, $rf_label_split, $rf_label_sizes) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_split_cands'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_split_cands'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_split_cands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_cgs = @{$candidate_gblocks};

	my $label_split = ${$rf_label_split};
	print $fh_out "# There are $ct_cgs caididate gap-blocks $label_split :\n\n";

	my $label_sizes = ${$rf_label_sizes};
	print $fh_out join ("\t", 'No.', 'indx_gb', 'distance_frm_pc', "[$label_sizes]", '(relv,purge,cands)', 'closest_pcand', '[sites,available,for,split]'), "\n\n";

	for (my $i=0; $i <$ct_cgs; $i++) {

	    my ($indx_gb, $dist, $sizes, $indices_rlv_pcs, $indx_closest_rlv_pc, $avail_sites) = @{$candidate_gblocks->[$i]};
	    my $cnct_sizes = join (',', @{$sizes});
	    my $cnct_rlv_pcs = join (',', @{$indices_rlv_pcs});
	    my $cnct_avail_sites = join (',', @{$avail_sites});

	    ##print STDOUT join ("\t", $i, $indx_gb, $dist, "[$cnct_sizes]", "($cnct_rlv_pcs)", $indx_closest_rlv_pc, "[$cnct_avail_sites]"), "\n"; # A BUG fixed on 2020/05/16.
	    print $fh_out join ("\t", $i, $indx_gb, $dist, "[$cnct_sizes]", "($cnct_rlv_pcs)", $indx_closest_rlv_pc, "[$cnct_avail_sites]"), "\n";

	}
	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_split_cands'. #


# print_out_vsplit_cands: 
#	This subroutine prints out a list of gap-blocks that are attempted to "vertically split".
#	The list is typically an output of 'list_(c)vsplit_candidates' (in "MyANEX_Main.pm").
#	It was created by modifying 'print_out_split_cands' above.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/24.]
#
# print_out_split_cands: 
#	This subroutine prints out a list of gap-blocks that are attempted to "split".
#	The list is typically an output of 'list_{s,c}split_candidates' (in "MyANEX_Main.pm").
#	It was created by modifying 'print_out_merge_split_cands' above.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/13.]
#
# print_out_merge_split_cands: 
#	This subroutine prints out a list of gap-block pairs that are attempted to "merge + split".
#	The list is typically an output of 'list_smerge_split_candidates' (in "MyANEX_Main.pm").
#	It was created by modifying 'print_out_purge_or_merge_cands' above.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/01.]
#
# print_out_purge_or_merge_cands: 
#	This subroutine prints out a list of gap-block pairs that are attempted to "purge" or "merge".
#	The list is typically an output of 'list_nn_purge_and_cmerge_candidates' or ''
#	(both in "MyANEX_Main.pm").
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/16.]
#
# INPUTS: (\$outdir, $ofname, \@candidate_gblocks, \$label_vsplit, \$suffix_new_sbj), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@{$candidate_gblocks[$cg]} = ($indx_gb, $dist, \@indices_rlv_pcs, $indx_closest_rlv_pc, \@info_new_sbj1, \@info_new_sbj2)
#	provides informaiton on the $cg th candidate gap-block to be "vertically split" into two blocks,
#	where
#	$dist is the distance between the gap-block and the nearest (and relevant) purge-candidate region,
#	@indices_rlv_pcs stores the indices of the relevant purge-candidate regions,
#	$indx_closest_rlv_pc is the index of the closest relevant purge-candidate region,
#	@info_new_sbj1/2 = ($br_nsbj, $up_or_lw_nsbj, \@exnodes_affd_by_nsbj, \@rows_affd_by_nsbj)
#	provides information on the 1st/2nd "new subject" resulting from the "vertical split" of $indx_gb, where
#	the new subject is on the $up_or_lw_nsbj side of the branch with ID $br_nsbj ,
#	@exnodes_affd_by_nsbj stores the indices of external nodes affected by the new subject,
#	and @rows_affd_by_nsbj stores the indices of the MSA rows affected by the new subject.
#	[Typically, \@candidate_gblocks is an output of 'list_(c)vsplit_candidates' in "MyANEX_Main.pm"];
#
#	$label_vsplit specifies what type of "vertical split" will be performed on the gap-blocks in @candidate_gblocks;
#
#	$suffix_new_sbj specifies the type of the new subjects;
#	
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_vsplit_cands (\$$\@\$\$) {

	my ($rf_outdir, $ofname, $candidate_gblocks, $rf_label_vsplit, $rf_suffix_nsbj) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_vsplit_cands'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_vsplit_cands'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_vsplit_cands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_cgs = @{$candidate_gblocks};

	my $label_vsplit = ${$rf_label_vsplit};
	print $fh_out "# There are $ct_cgs caididate gap-blocks $label_vsplit :\n\n";

	my $suffix_nsbj = ${$rf_suffix_nsbj};

	##print STDOUT join ("\t", 'No.', 'indx_gb', 'distance_frm_pc', '(relv,purge,cands)', 'closest_pcand', "[br/up(U)_or_lw(L)/(external,nodes,affected)/(msa,rows,affected)]_${suffix_nsbj}1", "[br/up(U)_or_lw(L)/(external,nodes,affected)/(msa,rows,affected)]_${suffix_nsbj}2"), "\n\n";
	print $fh_out join ("\t", 'No.', 'indx_gb', 'distance_frm_pc', '(relv,purge,cands)', 'closest_pcand', "[br/up(U)_or_lw(L)/(external,nodes,affected)/(msa,rows,affected)]_${suffix_nsbj}1", "[br/up(U)_or_lw(L)/(external,nodes,affected)/(msa,rows,affected)]_${suffix_nsbj}2"), "\n\n"; # CORRECTED on 2020/05/09.

	for (my $i=0; $i <$ct_cgs; $i++) {

	    my ($indx_gb, $dist, $indices_rlv_pcs, $indx_closest_rlv_pc, $info_effch1, $info_effch2) = @{$candidate_gblocks->[$i]};

	    my $cnct_rlv_pcs = join (',', @{$indices_rlv_pcs});

	    my ($br1, $up_or_lw1, $exnodes_affd1, $rows_affd1) = @{$info_effch1};
	    my ($br2, $up_or_lw2, $exnodes_affd2, $rows_affd2) = @{$info_effch2};

	    my $cnct_exnodes_affd1 = join (',', @{$exnodes_affd1});
	    my $cnct_exnodes_affd2 = join (',', @{$exnodes_affd2});
	    my $cnct_rows_affd1 = join (',', @{$rows_affd1});
	    my $cnct_rows_affd2 = join (',', @{$rows_affd2});

	    print $fh_out join ("\t", $i, $indx_gb, $dist, "($cnct_rlv_pcs)", $indx_closest_rlv_pc, "[$br1/$up_or_lw1/($cnct_exnodes_affd1)/($cnct_rows_affd1)]_${suffix_nsbj}1", "[$br2/$up_or_lw2/($cnct_exnodes_affd2)/($cnct_rows_affd2)]_${suffix_nsbj}2"), "\n";

	}
	print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_vsplit_cands'. #


# print_out_revpurge_cand_sizes:
#	This subroutine prints out a set of sizes of the "reverse-purge" candidate regions,
#	as well as some associated information.
#	[CREATED on 2019/09/16.]
#
# INPUTS: (\$outdir, $ofname, \@revpurge_cand_sizes, $xlevel, $max_rpc_size_for_xlevel2), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@revpurge_cand_sizes stores the sizes of the candidate regions to be "reverse-purge"d;
#	[Typically, \@revpurge_cand_sizes is the main output of 'list_revpurge_cand_sizes' in "MyANEX_Main.pm"];
#
#	$xlevel is an integer specifying the level of MSA-space exploration;
#	$max_rpc_size_for_xlevel2 is the maximum size of the "reverse-purge"d regions attempted when $xlevel = 2;
#	[$xlevel and $max_rpc_size_for_xlevel2 are inputs of 'list_revpurge_cand_sizes' in "MyANEX_Main.pm"];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_revpurge_cand_sizes (\$$\@$$) {

	my ($rf_outdir, $ofname, $revpurge_cand_sizes, $xlevel, $max_rpc_size_for_xlevel2) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_revpurge_cand_sizes'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_revpurge_cand_sizes'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_revpurge_cand_sizes'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_rpc_sizes = @{$revpurge_cand_sizes};

	print $fh_out "# Information on the Sizes of the Candidate-Regions to be \"Reverse-Purge\"d:\n";
	print $fh_out "#\n";
	print $fh_out "#   #{sizes} = $ct_rpc_sizes .\n";
	print $fh_out "#\n\n";

	print $fh_out "\$xlevel = $xlevel ,\n";
	print $fh_out "\$max_rpc_size_for_xlevel2 = $max_rpc_size_for_xlevel2 .\n\n";

	print $fh_out '@{$revpurge_cand_sizes} = (', join (', ', @{$revpurge_cand_sizes}), ") .\n\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_revpurge_cand_sizes'. #


# print_out_list_gpblocks:
#	This subroutine prints out a list of gap-pattern blocks in an MSA window,
#	which are a main ingredient when searching for CII-candidate regions.
#	The list is typically an output of 'list_cii_candidates' (in "MyANEX_MainX.pm").
#	It was created by taking advantage of the framework of 'print_out_list_cii_cands' below,
#	and by IMPORTING a block from 'test_reverse_spec_cii_in_wd.pl'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/10/01.]
#
# print_out_list_cii_cands: 
#	This subroutine prints out a list of CII-candidate regions on which the suspected CIIs are attempted to "reverse".
#	The list is typically an output of 'list_cii_candidates' or 'filter_cii_candidates'
#	(both in "MyANEX_MainX.pm").
#	It was created by taking advantage of the framework of 'print_out_vsplit_cands' above,
#	and by IMPORTING a block from 'test_reverse_spec_cii_in_wd.pl'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/10/01.]
#
# print_out_vsplit_cands: 
#	This subroutine prints out a list of gap-blocks that are attempted to "vertically split".
#	The list is typically an output of 'list_(c)vsplit_candidates' (in "MyANEX_Main.pm").
#	It was created by modifying 'print_out_split_cands' above.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/24.]
#
# INPUTS: (\$outdir, $ofname, \@set_gpblocks), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@set_gpblocks lists the gap-pattern blocks, each of which is the block of contiguous columns with the same set of MSA rows w/ the "presence" state;
#	More precisely,
#	@{$set_gpblocks[$gpb]} = ($le_clm2, $re_clm2, \@set_pres_rows, $class) provides information on the $gpb th gap-pattern block, where
#	$le_clm2 and $re_clm2 are the indices of the columns (in @set_clms above) at the left- and right-ends of the block,
#	@set_pres_rows stores the indices of the MSA rows with the "presence" state,
#	$class = 'F', 'E', 'Mono', and 'CII-cand', if the column is filled with the "presence" state, if it is filled with the "absence" state, if the rows with the "presence" state constitute a monophyletic group (or its complement), and if none of the above apply, respectively;
#	[Typically, \@set_gpblocks is a subsidiary output of 'list_cii_candidates', or a main output of 'partition_clms_into_gpbs' (both in "MyANEX_MainX.pm")];
#	
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_list_gpblocks (\$$\@) {

	my ($rf_outdir, $ofname, $set_gpblocks) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_list_cii_cands'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_list_cii_cands'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_list_cii_cands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_gpbs_in_wd = @{$set_gpblocks};

	print $fh_out "# There are $ct_gpbs_in_wd gap-pattern blocks (including gapless ('F') and null ('E') blocks) in this MSA window:\n\n";

	print $fh_out join ("\t", 'Indx_gpb', 'le_clm', 're_clm', 'class', '[rows,with,presence,state]'), "\n\n";

	for (my $gpb = 0; $gpb < $ct_gpbs_in_wd; $gpb++) { # Main for-loop (over gap-pattern blocks). #

	    my ($le_clm, $re_clm, $set_pres_rows, $class) = @{$set_gpblocks->[$gpb]};
	    my $cnct_pres_rows = join (',', @{$set_pres_rows});
	    print $fh_out join ("\t", $gpb, $le_clm, $re_clm, $class, "[$cnct_pres_rows]"), "\n";

	} # End of the Main for-loop (over gap-pattern blocks). #
	print STDOUT "\n\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_list_cii_cands'. #


# print_out_list_cii_cands: 
#	This subroutine prints out a list of CII-candidate regions on which the suspected CIIs are attempted to "reverse".
#	The list is typically an output of 'list_cii_candidates' or 'filter_cii_candidates'
#	(both in "MyANEX_MainX.pm").
#	It was created by taking advantage of the framework of 'print_out_vsplit_cands' above,
#	and by IMPORTING a block from 'test_reverse_spec_cii_in_wd.pl'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/10/01.]
#
# print_out_vsplit_cands: 
#	This subroutine prints out a list of gap-blocks that are attempted to "vertically split".
#	The list is typically an output of 'list_(c)vsplit_candidates' (in "MyANEX_Main.pm").
#	It was created by modifying 'print_out_split_cands' above.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/24.]
#
# INPUTS: (\$outdir, $ofname, \@list_cii_candidates, $cand_class), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@list_cii_cands lists the CII-candidate regions identified;
#	More precisely,
#	@{$list_cii_cands[$cc]} = ($le_clm, $re_clm, \@set_monophyl_sets_pres_rows, \@set_br_up_or_lw, $le_gpb, $re_gpb) provides information on the $cc th CII-candidate region, where
#	$le_clm and $re_clm are the IDs of the columns (in @set_clms above) at the left- and right-ends of the candidate region,
#	@{$set_monophyl_sets_pres_rows[$ms]} stores the indices of the MSA rows (with the "presence" state) in the $ms th monophyletic set;
#	@{$set_br_up_or_lw[$ms]} = ($br, $up_or_lw) indicates that the $ms th monophyletic set of the MSA rows (with the "presence" state) is on the $up_or_lw side of the branch $br,
#	$le_gpb and $re_gpb are the indices of the gap-pattern blocks (in @set_gpblocks below) at the left- and right-ends of the candidate region;
#	[NOTE that @set_monophyl_sets_msa_rows and @set_br_up_or_lw MUST be of the same size,
#	and that their size MUST be <= $maxct_monophyl_sets];
#	[Typically, \@list_cii_candidates is a main output of 'list_cii_candidates' or 'filter_cii_candidates' (both in "MyANEX_MainX.pm")];
#
#	$cand_class is a string specifying the class the candidate belongs to;
#	
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_list_cii_cands (\$$\@$) {

	my ($rf_outdir, $ofname, $list_cii_cands, $cand_class) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_list_cii_cands'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_list_cii_cands'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_list_cii_cands'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_cii_cands = @{$list_cii_cands};

	print $fh_out "# There are $ct_cii_cands CII-candidate regions ($cand_class):\n\n";


	print $fh_out join ("\t", 'Indx_cii_cand', 'le_clm', 're_clm', 'le_gpb', 're_gpb', '#{monophyl_sets}', '{br/up(U)_or_low(L)/rows,with,presence,state}_mnp_set1', '{...}_mnp_set2', '...'), "\n\n";

	for (my $cc = 0; $cc < $ct_cii_cands; $cc++) { # Main for-loop (over CII-candidates). #

	    my ($le_clm, $re_clm, $set_mnp_sets_pres_rows, $set_br_up_or_lw, $le_gpb, $re_gpb) = @{$list_cii_cands->[$cc]};

	    my $ct_mnp_sets = @{$set_mnp_sets_pres_rows};

	    my @printed_out = ($cc, $le_clm, $re_clm, $le_gpb, $re_gpb, $ct_mnp_sets);

	    for (my $k=0; $k < $ct_mnp_sets; $k++) {
	        my $cnct_pres_rows = join (',', @{$set_mnp_sets_pres_rows->[$k]});
	        my $cnct_info_mnp = join ('/', @{$set_br_up_or_lw->[$k]}, $cnct_pres_rows);
	        push @printed_out, "{$cnct_info_mnp}";
	    }

	    print $fh_out join ("\t", @printed_out), "\n";

	} # End of the Main for-loop (over CII-candidates). #
	print $fh_out "\n\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_list_cii_cands'. #



# print_out_gblks_w_spec_bds:
#	This subroutine prints out @gblks_w_spec_lb and @gblks_w_spec_rb,
#	which are the outputs of 'mk_lists_gblocks_w_spec_bounds' (in "MyGapPatternTopology.pm").
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#
# INPUTS: (\$outdir, $ofname, \@gblks_w_spec_lb, \@gblks_w_spec_rb), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_lb and \@gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' in "MyGapPatternTopology.pm"];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_gblks_w_spec_bds (\$$\@\@) {

	my ($rf_outdir, $ofname, $gblks_w_spec_lb, $gblks_w_spec_rb) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_gblks_w_spec_bds'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_gblks_w_spec_bds'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_gblks_w_spec_bds'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_clms = @{$gblks_w_spec_lb};

	print $fh_out "\n", '## Contents of @{$gblks_w_spec_lb} & @{$gblks_w_spec_rb}: ', "\n\n";

	print $fh_out join ("\t", 'Indx_clm(=c)', '[gap,blocks,w,lb=c]', '[gap,blocks,w,rb=c]'), "\n\n";

	for (my $c = 0; $c < $ct_clms; $c++) {
	    my $cnct_gblks_w_lb_eq_c = join (',', @{$gblks_w_spec_lb->[$c]});
	    my $cnct_gblks_w_rb_eq_c = join (',', @{$gblks_w_spec_rb->[$c]});
	    print $fh_out join ("\t", $c, "[$cnct_gblks_w_lb_eq_c]", "[$cnct_gblks_w_rb_eq_c]"), "\n";
	}
	print $fh_out "\n\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_gblks_w_spec_bds'. #


# print_out_cts_vincluding_vincluded_onx_gblks:
#	This subroutine prints out @cts_vincluding, @cts_vincluded, and @cts_onx,
#	which are the outputs of 'cts_vincluding_vincluded_onx_gblks' in "MyGapPatternTopology.pm".
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#
# INPUTS: (\$outdir, $ofname, \@cts_vincluding, \@cts_vincluded, \@cts_onx), where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	$cts_vincluding[$gb] is the number of gap-blocks vertically including the $gb th gap-block;
#	$cts_vincluded[$gb] is the number of gap-blocks vertically included in the $gb th gap-block;
#	$cts_onx[$gb] is the number of gap-blocks that are in the 'ONN' or 'ONCS' relationship with the $gb th gap-block;
#	[Typically, \@cts_vincluding, \@cts_vincluded, \@cts_onx are the outputs of 'cts_vincluding_vincluded_onx_gblks' in "MyGapPatternTopology.pm"];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
#
sub print_out_cts_vincluding_vincluded_onx_gblks (\$$\@\@\@) {

	my ($rf_outdir, $ofname, $cts_vincluding, $cts_vincluded, $cts_onx) = @_;

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_cts_vincluding_vincluded_onx_gblks'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_cts_vincluding_vincluded_onx_gblks'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_cts_vincluding_vincluded_onx_gblks'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_gblks = @{$cts_vincluding};

	print $fh_out "\n", '( Contents of @{$cts_vincluding} & @{$cts_vincluded} & @{$cts_onx} : )', "\n\n"; 

	print $fh_out join ("\t", 'Indx_gb', '#{vincluding_gbs}', '#{vincluded_gbs}', '#{gbs_w_onx_relations}'), "\n\n";

	for (my $gb = 0; $gb < $ct_gblks; $gb++) { # 3rd sub-main for-loop (over gap-blocks). #

	    print $fh_out join ("\t", $gb, $cts_vincluding->[$gb], $cts_vincluded->[$gb], $cts_onx->[$gb]), "\n";

	} # End of the 3rd sub-main for-loop (over gap-blocks). #
	print $fh_out "\n\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_cts_vincluding_vincluded_onx_gblks'. #


# bch_print_out_other_gblk_attributes:
#	This subroutine prints out some attributes of gap-blocks
#	that 'print_in_nr_mshifts_glks' does NOT print out.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine recieves a list of the output file names as well.
#
#
# [ INPUTS: (\$outdir, \@inter_block_relations, \@gblock_sizes, \@gblks_w_spec_lb, \@gblks_w_spec_rb, \@cts_vincluding, \@cts_vincluded, \@cts_onx), where ] ... OBSOLETE as of 2019/07/18.
# INPUTS: (\$outdir, \@ofnames, \@inter_block_relations, \@gblock_sizes, \@gblks_w_spec_lb, \@gblks_w_spec_rb, \@cts_vincluding, \@cts_vincluded, \@cts_onx), where
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames list the names of the 4 output files; # ADDED on 2019/07/18.
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1, where
#	$relations is a string coding the relation;
#	[See the header of 'mk_inter_block_relations' for details];
#
#	$gblock_sizes[$gb] is the (horizontal) size of the $gb th gap-block;
#	[Typically, \@gblock_sizes is the main output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"];
#	
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_lb and \@gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' in "MyGapPatternTopology.pm"];
#
#	$cts_vincluding[$gb] is the number of gap-blocks vertically including the $gb th gap-block;
#	$cts_vincluded[$gb] is the number of gap-blocks vertically included in the $gb th gap-block;
#	$cts_onx[$gb] is the number of gap-blocks that are in the 'ONN' or 'ONCS' relationship with the $gb th gap-block;
#	[Typically, \@cts_vincluding, \@cts_vincluded, \@cts_onx are the outputs of 'cts_vincluding_vincluded_onx_gblks' in "MyGapPatternTopology.pm"];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
# [ sub bch_print_out_other_gblk_attributes (\$\@\@\@\@\@\@\@) {
#	my ($rf_outdir, $inter_block_relations,
#	   $gblock_sizes, $gblks_w_spec_lb, $gblks_w_spec_rb,
#	   $cts_vincluding, $cts_vincluded, $cts_onx) = @_; ] ... OBSOLETE as of 2019/07/18.
#
sub bch_print_out_other_gblk_attributes (\$\@\@\@\@\@\@\@\@) { # REVISED on 2019/07/18.

	my ($rf_outdir, $ofnames, $inter_block_relations,
	   $gblock_sizes, $gblks_w_spec_lb, $gblks_w_spec_rb,
	   $cts_vincluding, $cts_vincluded, $cts_onx) = @_; # REVISED on 2019/07/18.

	unless (@{$ofnames} == 4) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_purge'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 4 as desired: $! !\n\n";
	    return 0;
	}

	my $ct_gbs = @{$gblock_sizes};
	my $ct_clms = @{$gblks_w_spec_lb};

   (print STDOUT "< Subroutine 'bch_print_out_other_gblk_attributes' (#{gap-blocks}=$ct_gbs, #{columns}=$ct_clms) >\n\n") if ($TEST_LEVEL_SUB >0);

	my $outdir = ${$rf_outdir};

	my $ofname1 = $ofnames->[0]; # REVISED on 2019/07/18.
	##my $ofname1 = 'inter_block_relations.txt';

	unless ( print_out_inter_block_relations ($outdir, $ofname1, @{$inter_block_relations}) ) {
	    print STDOUT "[ERROR 'bch_print_out_other_gblk_attributes'] The subroutine 'print_out_inter_block_relations' FAILED on the output file \"$outdir/$ofname1\": $! !\n\n";
	    return 0;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_inter_block_relations', Ended on the output file, \"$outdir/$ofname1\" ...\n\n") if ($TEST_LEVEL_SUB >1);


	my $ofname2 = $ofnames->[1]; # REVISED on 2019/07/18.
	##my $ofname2 = 'gblock_sizes.txt';

	unless ( print_out_list_gblock_sizes ($outdir, $ofname2, @{$gblock_sizes}) ) {
	    print STDOUT "[ERROR 'bch_print_out_other_gblk_attributes'] The subroutine 'print_out_list_gblock_sizes' FAILED on the output file \"$outdir/$ofname2\": $! !\n\n";
	    return 0;
	}
    (print STDOUT "\n ... The subroutine 'print_out_list_gblock_sizes', Ended on the output file, \"$outdir/$ofname2\" ...\n\n") if ($TEST_LEVEL_SUB >1);


	my $ofname3 = $ofnames->[2]; # REVISED on 2019/07/18.
	##my $ofname3 = 'gblks_w_spec_bds.txt';

	unless ( print_out_gblks_w_spec_bds ($outdir, $ofname3, @{$gblks_w_spec_lb}, @{$gblks_w_spec_rb}) ) {
	    print STDOUT "[ERROR 'bch_print_out_other_gblk_attributes'] The subroutine 'print_out_gblks_w_spec_bds' FAILED on the output file \"$outdir/$ofname3\": $! !\n\n";
	    return 0;
	}
    (print STDOUT "\n ... The subroutine 'print_out_gblks_w_spec_bds', Ended on the output file, \"$outdir/$ofname3\" ...\n\n") if ($TEST_LEVEL_SUB >1);


	my $ofname5 = $ofnames->[3]; # REVISED on 2019/07/18.
	##my $ofname5 = 'cts_vincluding_vincluded_onx_gblks.txt';

	unless ( print_out_cts_vincluding_vincluded_onx_gblks ($outdir, $ofname5, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}) ) {
	    print STDOUT "[ERROR 'bch_print_out_other_gblk_attributes'] The subroutine 'print_out_cts_vincluding_vincluded_onx_gblks' FAILED on the output file \"$outdir/$ofname5\": $! !\n\n";
	    return 0;
	}
    (print STDOUT "\n ... The subroutine 'print_out_cts_vincluding_vincluded_onx_gblks', Ended on the output file, \"$outdir/$ofname5\" ...\n\n") if ($TEST_LEVEL_SUB >1);


   (print STDOUT "< Subroutine 'bch_print_out_other_gblk_attributes' Ended />\n\n") if ($TEST_LEVEL_SUB >0);

	return 1;

}	# End of the subroutine, 'bch_print_out_other_gblk_attributes'. #


# print_out_info_mvd_sbj:
#	This subroutine prints out some information (attributes) of the gap-blocks
#	that are subject to a specified "move".
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#
# INPUTS: (\$outdir, $ofname, \@subjects, \@curr_gb_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_bds_gblks, \@gblock_sizes0, \@rows_affd_by_blks, where
#
#	$outdir is a string specifying the path to the output directory;
#	$ofname is a string providing the name of the output file;
#
#	@subjects stores the indices of the gap-blocks to be "move"d;
#
#	$curr_gb_coords[$gb] is the current coordinate value of the $gb th gap-block;
#
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords[$gb]} give the (accessible) left- and right-boundaries of the coordinate for the $gb th gap-block;
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]} give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window;
#	[Typically, \@bds_gb_coords and \@gb_coords_wd_frm are a part of the main outputs of 'mk_orgs_bds_gb_coords' in "MyANEX_Supple.pm"];
#
#	@{$curr_bds_gblks[$gb]} = ($curr_lbd_blk, $curr_rbd_blk) gives the current positions (within the window, 0-based) of the boundaries (inclusive) of the $gb th gap-block;
#	@{$rows_aff_by_blks[$gb]} lists the indices of the rows of the MSA affected by the $gb th gap-block;
#	[Typically, \@curr_bds_gblks and \@rows_affd_by_blks are the main outputs of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_info_mvd_sbj (\$$\@\@\@\@\@\@\@) {

	my ($rf_outdir, $ofname, $subjects, $curr_gb_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	    $curr_bds_gblks, $gblock_sizes, $rows_affd_by_blks) = @_;


	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_mvd_sbj'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_mvd_sbj'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $outfile = "$outdir/$ofname";

	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
		print STDOUT "[ERROR 'print_out_info_mvd_sbj'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile\": $! !\n";
		return 0;
	}

	my $ct_sbj = @{$subjects};

	print $fh_out "# There are $ct_sbj gap-blocks to be \"move\"d: (", join (', ', @{$subjects}), ") .\n\n";

	print $fh_out join ("\t", 'Indx_gb', 'curr_coord', 'lb_coord', 'rb_coord', 'coord_le_wd', 'coord_re_wd', 'lm_clm', 'rm_clm', 'block_size', '[msa,rows,affected]'), "\n\n";

	foreach my $gb (@{$subjects}) { # Inner foreach-loop (over subject gap-blocks). #
	    my $cnct_rows_affd = join (',', @{$rows_affd_by_blks->[$gb]});
	    print $fh_out join ("\t", $gb, $curr_gb_coords->[$gb], @{$bds_gb_coords->[$gb]}, @{$gb_coords_wd_frm->[$gb]}, @{$curr_bds_gblks->[$gb]}, $gblock_sizes->[$gb], "[$cnct_rows_affd]"), "\n";

	} # End of the inner foreach-loop (over subject gap-blocks). #

	print $fh_out "\n\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_out_info_mvd_sbj'. #


# print_out_info_chngs_by_purge:
#	This subroutine prints out the portions of the outputs of 'purge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"purge\" and the \"shift"s immediately before that.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine receives a list of output file names as well.
#
# [ INPUTS: (\$outdir, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, \@new_indices_gbs, \@old_indices_gbs), where ] ... OBSOLETE as of 2019/07/18.
# INPUTS: (\$outdir, \@ofnames, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, \@new_indices_old_gbs, \@old_indices_new_gbs), where
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 2 output files;
#
#	The other arguments are a part of the outputs of 'purge_gblock_pair' in "MyANEX_Main.pm";
#	More precisely,
#
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "purge" (= undef if an error occurs);
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "purge";
#	More precisely, the gap-block that had been the $gb th before the "purge"
#	has been assigned $new_indices_old_gbs[$gb] after the "purge"
#	(, and $new_indices_old_gbs[$gb1] and $new_indices_old_gbs[$gb2] are undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "purge";
#	More precisely, the gap-block that is the $gb th after the "purge"
#	had been assigned $old_indices_new_gbs[$gb] before the "purge";
# 
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
# [ sub print_out_info_chngs_by_purge (\$$$$\@\@) {
#	my ($rf_outdir, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, $new_indices_gbs, $old_indices_gbs) = @_; ] ... OBSOLETE as of 2019/07/18.
#
sub print_out_info_chngs_by_purge (\$\@$$$\@\@) { # REVISED on 2019/07/18.

	my ($rf_outdir, $ofnames, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, $new_indices_old_gbs, $old_indices_new_gbs) = @_; # REVISED on 2019/07/18.

	unless (@{$ofnames} == 2) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_purge'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 2 as desired: $! !\n\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_purge'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_purge'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $ofname1 = $ofnames->[0]; # REVISED on 2019/07/18.
	##my $ofname1 = 'info_serial_shifts.txt';
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_purge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	print $fh_out1 "{Subject gap-block \"shift\"ed} = $bl_sbj .\n";
	print $fh_out1 "{Increment of the subject block coordinate} = $incr_coord_sbj .\n";
	print $fh_out1 "#{Resulting null columns, which were removed during the \"purge\"} = $ct_rmvd_null_clms .\n\n";

	close ($fh_out1);


	my $ofname2 = $ofnames->[1]; # REVISED on 2019/07/18.
	##my $ofname2 = 'old_new_indices_gblks.txt';
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_purge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}

	my $ct_old_gbs = @{$new_indices_old_gbs};
        my $ct_new_gbs = @{$old_indices_new_gbs};

        print $fh_out2 "## Relationships between new and old indices of gap-blocks\n";
	print $fh_out2 "##  (#{gap-blocks}_old=$ct_old_gbs, #{gap-blocks}_new=$ct_new_gbs) :\n\n";

        my @out_new_indices_gbs = my @out_old_indices_gbs = ();
        foreach my $indx (@{$new_indices_old_gbs}) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$old_indices_new_gbs}) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

        print $fh_out2 '@{$new_indices_old_gbs} = (', join (', ', @out_new_indices_gbs), ") .\n";
        print $fh_out2 '@{$old_indices_new_gbs} = (', join (', ', @out_old_indices_gbs), ") .\n";
        print $fh_out2 "\n\n";

	close ($fh_out2);

	return 1;

}	# End of the subroutine, 'print_out_info_chngs_by_purge'. #


# print_out_info_chngs_by_cmerge:
#	This subroutine prints out the portions of the outputs of 'cmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (complementary-types)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_purge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/20.]
#
# print_out_info_chngs_by_purge:
#	This subroutine prints out the portions of the outputs of 'purge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"purge\" and the \"shift"s immediately before that.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine receives a list of output file names as well.
#
# [ INPUTS: (\$outdir, \@ofnames, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, \@new_indices_gbs, \@old_indices_gbs), where ] ... for the predecessor, 'print_out_info_chngs_by_purge'.
# INPUTS: (\$outdir, \@ofnames, $gb_rmvd, $gb_kpt, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, \@new_indices_old_gbs, \@old_indices_new_gbs), where
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 3 output files;
#
#	The other arguments are a part of the outputs of 'cmerge_gblock_pair' in "MyANEX_Main.pm";
#	More precisely,
#
#		# Added in creation of 'print_out_info_chngs_by_cmerge' on 2019/07/20. #
#	$gb_rmvd and $gb_kpt are the indices (in the set before the \"merge\") of the removed gap-block and that kept, respectively;
#		# End of "Added in creation of 'print_out_info_chngs_by_cmerge' on 2019/07/20." #
#
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "cmerge" (= undef if an error occurs);
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "cmerge";
#	More precisely, the gap-block that had been the $gb th before the "cmerge"
#	has been assigned $new_indices_old_gbs[$gb] after the "cmerge"
#	(, and $new_indices_old_gbs[$gb_rmvd] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "cmerge";
#	More precisely, the gap-block that is the $gb th after the "cmerge"
#	had been assigned $old_indices_new_gbs[$gb] before the "cmerge";
# 
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
sub print_out_info_chngs_by_cmerge (\$\@$$$$$\@\@) {

	my ($rf_outdir, $ofnames, $gb_rmvd, $gb_kpt, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, $new_indices_old_gbs, $old_indices_new_gbs) = @_;

	unless (@{$ofnames} == 3) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 3 as desired: $! !\n\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

		# ADDED in creation of 'print_out_info_chngs_by_cmerge' on 2019/07/20. #
	my $ofname0 = $ofnames->[0];
	my $outfile0 = "$outdir/$ofname0";

	my $fh_out0 = open_output_file ($outfile0, 0);
	unless (defined $fh_out0) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile0\": $! !\n";
		return 0;
	}

	print $fh_out0 "# Indices (in the old set) of Gap-blocks in the \"Merge\"d Pair:\n\n";

	print $fh_out0 "{Gap-block removed} = $gb_rmvd ,\n";
	print $fh_out0 "{Gap-block kept}    = $gb_kpt .\n\n";

	close ($fh_out0);
		# End of "ADDED in creation of 'print_out_info_chngs_by_cmerge' on 2019/07/20." #


	my $ofname1 = $ofnames->[1]; # Further REVISED on 2019/07/20.
	##my $ofname1 = $ofnames->[0]; # REVISED on 2019/07/18.
	####my $ofname1 = 'info_serial_shifts.txt';
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	print $fh_out1 "{Subject gap-block \"shift\"ed} = $bl_sbj .\n";
	print $fh_out1 "{Increment of the subject block coordinate} = $incr_coord_sbj .\n";
	print $fh_out1 "#{Resulting null columns, which were removed during the \"merge\"} = $ct_rmvd_null_clms .\n\n";

	close ($fh_out1);


	my $ofname2 = $ofnames->[2]; # Further REVISED on 2019/07/20.
	##my $ofname2 = $ofnames->[1]; # REVISED on 2019/07/18.
	####my $ofname2 = 'old_new_indices_gblks.txt';
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}

	my $ct_old_gbs = @{$new_indices_old_gbs};
        my $ct_new_gbs = @{$old_indices_new_gbs};

        print $fh_out2 "## Relationships between new and old indices of gap-blocks\n";
	print $fh_out2 "##  (#{gap-blocks}_old=$ct_old_gbs, #{gap-blocks}_new=$ct_new_gbs) :\n\n";

        my @out_new_indices_gbs = my @out_old_indices_gbs = ();
        foreach my $indx (@{$new_indices_old_gbs}) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$old_indices_new_gbs}) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

        print $fh_out2 '@{$new_indices_old_gbs} = (', join (', ', @out_new_indices_gbs), ") .\n";
        print $fh_out2 '@{$old_indices_new_gbs} = (', join (', ', @out_old_indices_gbs), ") .\n";
        print $fh_out2 "\n\n";

	close ($fh_out2);

	return 1;

}	# End of the subroutine, 'print_out_info_chngs_by_cmerge'. #


# print_out_info_chngs_by_smerge:
#	This subroutine prints out the portions of the outputs of 'smerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (same-type)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_cmerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/22.]
#
# print_out_info_chngs_by_cmerge:
#	This subroutine prints out the portions of the outputs of 'cmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (complementary-types)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_purge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/20.]
#
# print_out_info_chngs_by_purge:
#	This subroutine prints out the portions of the outputs of 'purge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"purge\" and the \"shift"s immediately before that.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine receives a list of output file names as well.
#
# [ INPUTS: (\$outdir, \@ofnames, $gb_rmvd, $gb_kpt, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, \@new_indices_gbs, \@old_indices_gbs), where ] ... for the predecessor, 'print_out_info_chngs_by_cmerge'.
# INPUTS: (\$outdir, \@ofnames, $gb_absbd, $gb_kpt, $bl_sbj, $incr_coord_sbj, \@new_indices_old_gbs, \@old_indices_new_gbs), where
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 3 output files;
#
#	The other arguments are a part of the outputs of 'smerge_gblock_pair' in "MyANEX_Main.pm";
#	More precisely,
#
#		# Added in creation of 'print_out_info_chngs_by_cmerge' on 2019/07/20. #
		#  => MODIFIED in creation of 'print_out_info_chngs_by_smerge' on 2019/07/22. #

#	$gb_absbd and $gb_kpt are the indices (in the set before the \"merge\") of the absorbed gap-block and that kept, respectively;
#		# End of "Added in creation of 'print_out_info_chngs_by_cmerge' on 2019/07/20. #
		#  => MODIFIED in creation of 'print_out_info_chngs_by_smerge' on 2019/07/22." #
#
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#	[ $ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "cmerge" (= undef if an error occurs); ] ... for the predecessor 'print_out_info_chngs_by_cmerge'.
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "smerge";
#	More precisely, the gap-block that had been the $gb th before the "smerge"
#	has been assigned $new_indices_old_gbs[$gb] after the "smerge"
#	(, and $new_indices_old_gbs[$gb_absbd] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "smerge";
#	More precisely, the gap-block that is the $gb th after the "smerge"
#	had been assigned $old_indices_new_gbs[$gb] before the "smerge";
# 
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
sub print_out_info_chngs_by_smerge (\$\@$$$$\@\@) {

	my ($rf_outdir, $ofnames, $gb_absbd, $gb_kpt, $bl_sbj, $incr_coord_sbj, $new_indices_old_gbs, $old_indices_new_gbs) = @_;

	unless (@{$ofnames} == 3) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_smerge'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 3 as desired: $! !\n\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_smerge'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_smerge'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

		# ADDED in creation of 'print_out_info_chngs_by_smerge' on 2019/07/20. #
	my $ofname0 = $ofnames->[0];
	my $outfile0 = "$outdir/$ofname0";

	my $fh_out0 = open_output_file ($outfile0, 0);
	unless (defined $fh_out0) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile0\": $! !\n";
		return 0;
	}

	print $fh_out0 "# Indices (in the old set) of Gap-blocks in the \"Merge\"d Pair:\n\n";

	print $fh_out0 "{Gap-block absorbed} = $gb_absbd ,\n"; # MODIFIED in creation of 'print_out_info_chngs_by_smerge' (on 2019/07/22). #
	##print $fh_out0 "{Gap-block removed} = $gb_rmvd ,\n";

	print $fh_out0 "{Gap-block kept}    = $gb_kpt .\n\n";

	close ($fh_out0);
		# End of "ADDED in creation of 'print_out_info_chngs_by_smerge' on 2019/07/20." #


	my $ofname1 = $ofnames->[1]; # Further REVISED on 2019/07/20.
	##my $ofname1 = $ofnames->[0]; # REVISED on 2019/07/18.
	####my $ofname1 = 'info_serial_shifts.txt';
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	my $ct_rmvd_null_clms = 0; # Always set this value to be zero. # ADDED in creation of 'print_out_info_chngs_by_smerge' (on 2019/07/22). #

	print $fh_out1 "{Subject gap-block \"shift\"ed} = $bl_sbj .\n";
	print $fh_out1 "{Increment of the subject block coordinate} = $incr_coord_sbj .\n";

	print $fh_out1 "#{Resulting null columns, which were removed during the \"merge\"} = $ct_rmvd_null_clms .\n\n";

	close ($fh_out1);


	my $ofname2 = $ofnames->[2]; # Further REVISED on 2019/07/20.
	##my $ofname2 = $ofnames->[1]; # REVISED on 2019/07/18.
	####my $ofname2 = 'old_new_indices_gblks.txt';
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}

	my $ct_old_gbs = @{$new_indices_old_gbs};
        my $ct_new_gbs = @{$old_indices_new_gbs};

        print $fh_out2 "## Relationships between new and old indices of gap-blocks\n";
	print $fh_out2 "##  (#{gap-blocks}_old=$ct_old_gbs, #{gap-blocks}_new=$ct_new_gbs) :\n\n";

        my @out_new_indices_gbs = my @out_old_indices_gbs = ();
        foreach my $indx (@{$new_indices_old_gbs}) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$old_indices_new_gbs}) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

        print $fh_out2 '@{$new_indices_old_gbs} = (', join (', ', @out_new_indices_gbs), ") .\n";
        print $fh_out2 '@{$old_indices_new_gbs} = (', join (', ', @out_old_indices_gbs), ") .\n";
        print $fh_out2 "\n\n";

	close ($fh_out2);

	return 1;

}	# End of the subroutine, 'print_out_info_chngs_by_smerge'. #


# print_out_info_chngs_by_cvmerge:
#	This subroutine prints out the portions of the outputs of 'cvmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"(incomplete) vertical-merge (complementary-siblings)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_cmerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/29.]
#
# print_out_info_chngs_by_cmerge:
#	This subroutine prints out the portions of the outputs of 'cmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (complementary-types)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_purge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/20.]
#
# print_out_info_chngs_by_purge:
#	This subroutine prints out the portions of the outputs of 'purge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"purge\" and the \"shift"s immediately before that.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine receives a list of output file names as well.
#
# [ INPUTS: (\$outdir, \@ofnames, $gb_rmvd, $gb_kpt, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, \@new_indices_gbs, \@old_indices_gbs), where ] ... for the predecessor 'print_out_info_chngs_by_cmerge'.
# INPUTS: (\$outdir, \@ofnames, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, \@new_indices_old_gbs, \@old_indices_new_gbs), where
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 3 output files;
#
#	The other arguments are a part of the outputs of 'cvmerge_gblock_pair' in "MyANEX_Main.pm";
#	More precisely,
#
#		# ADDED in creation of 'print_out_info_chngs_by_cvmerge' (on 2019/07/29). #
#	$if_complete = 1 if the "vertical-merge" is complete,
#		     = 0 if it is incomplete;
#
#	$gb_rplcd and $gb_rmvd_kpt, respectively, are the indices (in the set before the \"(incomplete) vertical-merge\") of the gap-block replaced with a new one and that either removed (when $if_complete=1) or kept (when $if_complete=0);
#		# END of "ADDED in creation of 'print_out_info_chngs_by_cvmerge' (on 2019/07/29)." #
#
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#	(NOTE that the "shift"s are undone after the "incomplete vertical-merge");
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "cvmerge" (= undef if an error occurs);
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "cvmerge";
#	More precisely, the gap-block that had been the $gb th before the "cvmerge"
#	has been assigned $new_indices_old_gbs[$gb] after the "cvmerge"
#	(, and $new_indices_old_gbs[$gb_rmvd_kpt] is undefined when $if_complete = 1);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "cvmerge";
#	More precisely, the gap-block that is the $gb th after the "cvmerge"
#	had been assigned $old_indices_new_gbs[$gb] before the "cvmerge";
# 
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
sub print_out_info_chngs_by_cvmerge (\$\@$$$$$$\@\@) {

	my ($rf_outdir, $ofnames, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, $new_indices_old_gbs, $old_indices_new_gbs) = @_;

	unless (@{$ofnames} == 3) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_cvmerge'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 3 as desired: $! !\n\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_cvmerge'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_cvmerge'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

		# ADDED in creation of 'print_out_info_chngs_by_cmerge' on 2019/07/20. #
	my $ofname0 = $ofnames->[0];
	my $outfile0 = "$outdir/$ofname0";

	my $fh_out0 = open_output_file ($outfile0, 0);
	unless (defined $fh_out0) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_cvmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile0\": $! !\n";
		return 0;
	}

			# ADDED in creation of 'print_out_info_chngs_by_cvmerge' (on 2019/07/29). #
	print $fh_out0 "# Whether the \"Vertical-Merge\" was complete(=1) or not (=0):\n\n";
	print $fh_out0 '$if_complete = ', $if_complete, " .\n\n";
			# END of "ADDED in creation of 'print_out_info_chngs_by_cvmerge' (on 2019/07/29)." #

	print $fh_out0 "# Indices (in the old set) of Gap-blocks in the \"(incomplete) Vertical-Merge\"d Pair:\n\n";

	print $fh_out0 "{Gap-block replaced} = $gb_rplcd ,\n";
	print $fh_out0 "{Gap-block removed/kept}    = $gb_rmvd_kpt .\n\n";

	close ($fh_out0);
		# End of "ADDED in creation of 'print_out_info_chngs_by_cmerge' on 2019/07/20." #


	my $ofname1 = $ofnames->[1]; # Further REVISED on 2019/07/20.
	##my $ofname1 = $ofnames->[0]; # REVISED on 2019/07/18.
	####my $ofname1 = 'info_serial_shifts.txt';
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_cvmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	print $fh_out1 "{Subject gap-block \"shift\"ed} = $bl_sbj .\n";
	print $fh_out1 "{Increment of the subject block coordinate BEFORE the \"(incomplete) vertical-merge\"} = $incr_coord_sbj .\n";
	print $fh_out1 "#{Null columns, which were produced & removed during the \"(incomplete) vertical-merge\"} = $ct_rmvd_null_clms .\n\n";

	close ($fh_out1);


	my $ofname2 = $ofnames->[2]; # Further REVISED on 2019/07/20.
	##my $ofname2 = $ofnames->[1]; # REVISED on 2019/07/18.
	####my $ofname2 = 'old_new_indices_gblks.txt';
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_cvmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}

	my $ct_old_gbs = @{$new_indices_old_gbs};
        my $ct_new_gbs = @{$old_indices_new_gbs};

        print $fh_out2 "## Relationships between new and old indices of gap-blocks\n";
	print $fh_out2 "##  (#{gap-blocks}_old=$ct_old_gbs, #{gap-blocks}_new=$ct_new_gbs) :\n\n";

        my @out_new_indices_gbs = my @out_old_indices_gbs = ();
        foreach my $indx (@{$new_indices_old_gbs}) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$old_indices_new_gbs}) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

        print $fh_out2 '@{$new_indices_old_gbs} = (', join (', ', @out_new_indices_gbs), ") .\n";
        print $fh_out2 '@{$old_indices_new_gbs} = (', join (', ', @out_old_indices_gbs), ") .\n";
        print $fh_out2 "\n\n";

	close ($fh_out2);

	return 1;

}	# End of the subroutine, 'print_out_info_chngs_by_cvmerge'. #


# print_out_info_chngs_by_merge_split:
#	This subroutine prints out the portions of the outputs of 'smerge_split_gblock_pair' or 'icmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge + split (same/complementary-type(s))\" and the \"shift"s immediately before and after that.
#	It was created by slightly modifying 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/01.]
#	[REVISED on 2019/08/11]:
#	Now, it receives and prints out $ct_rmvd_null_clms as well.
#
# print_out_info_chngs_by_smerge:
#	This subroutine prints out the portions of the outputs of 'smerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (same-type)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_cmerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/22.]
#
# print_out_info_chngs_by_cmerge:
#	This subroutine prints out the portions of the outputs of 'cmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (complementary-types)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_purge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/20.]
#
# print_out_info_chngs_by_purge:
#	This subroutine prints out the portions of the outputs of 'purge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"purge\" and the \"shift"s immediately before that.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine receives a list of output file names as well.
#
# [ INPUTS: (\$outdir, \@ofnames, $gb_absbd, $gb_kpt, $bl_sbj, $incr_coord_sbj, \@new_indices_old_gbs, \@old_indices_new_gbs), where ] ... for the predecessor, 'print_out_info_chngs_by_smerge'.
# [ INPUTS: (\$outdir, \@ofnames, $bl_sbj, $incr_coord_sbj, $new_bl_sbj, $incr_coord_new_sbj), where ] ... OBSOLETE as of 2019/08/11.
# INPUTS: (\$outdir, \@ofnames, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, $new_bl_sbj, $incr_coord_new_sbj), where # REVISED on 2019/08/11.
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 1 output file;
#
#	The other arguments are a part of the outputs of 'smerge_split_gblock_pair' or 'icmerge_gblock_pair', both in "MyANEX_Main.pm";
#	More precisely,
#
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed) before the "merge + split";
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed) before the "merge + split";
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "cmerge + split" (= undef if an error occurs);
#
#	$new_bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed) after the "merge + split";
#	$incr_coord_new_sbj is the increment of the coordinate of $bl_sbj resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed) after the "merge + split";
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
# [ sub print_out_info_chngs_by_merge_split (\$\@$$$$) {
#	my ($rf_outdir, $ofnames, $bl_sbj, $incr_coord_sbj, $new_bl_sbj, $incr_coord_new_sbj) = @_; ] ... OBSOLETE as of 2019/08/11.
#
sub print_out_info_chngs_by_merge_split (\$\@$$$$$) {

	my ($rf_outdir, $ofnames, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, $new_bl_sbj, $incr_coord_new_sbj) = @_; # REVISED on 2019/08/11.

	unless (@{$ofnames} == 1) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_merge_split'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 1 as desired: $! !\n\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_merge_split'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_merge_split'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $ofname1 = $ofnames->[0];
	####my $ofname1 = 'info_serial_shifts.txt';
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_merge_split'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	##my $ct_rmvd_null_clms = 0; # Always set this value to be zero. # ADDED in creation of 'print_out_info_chngs_by_smerge' (on 2019/07/22). #

	print $fh_out1 "{Subject gap-block \"shift\"ed BEFORE the \"merge + split\"} = $bl_sbj ,\n";
	print $fh_out1 "{Increment of the subject block coordinate before \"merge + split\"} = $incr_coord_sbj .\n";
	print $fh_out1 "\n";

	print $fh_out1 "{Subject gap-block \"shift\"ed AFTER the \"merge + split\"} = $new_bl_sbj ,\n";
	print $fh_out1 "{Increment of the subject block coordinate after \"merge + split\"} = $incr_coord_new_sbj .\n";
	print $fh_out1 "\n";

	print $fh_out1 "#{Null columns, which were produced & removed during the \"merge + split\"} = $ct_rmvd_null_clms .\n\n"; # ADDED in REVISION on 2019/08/11.

	close ($fh_out1);


=for the predecessor 'print_out_info_chngs_by_smerge'.
	my $ofname2 = $ofnames->[2]; # Further REVISED on 2019/07/20.
	##my $ofname2 = $ofnames->[1]; # REVISED on 2019/07/18.
	####my $ofname2 = 'old_new_indices_gblks.txt';
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}

	my $ct_old_gbs = @{$new_indices_old_gbs};
        my $ct_new_gbs = @{$old_indices_new_gbs};

        print $fh_out2 "## Relationships between new and old indices of gap-blocks\n";
	print $fh_out2 "##  (#{gap-blocks}_old=$ct_old_gbs, #{gap-blocks}_new=$ct_new_gbs) :\n\n";

        my @out_new_indices_gbs = my @out_old_indices_gbs = ();
        foreach my $indx (@{$new_indices_old_gbs}) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$old_indices_new_gbs}) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

        print $fh_out2 '@{$new_indices_old_gbs} = (', join (', ', @out_new_indices_gbs), ") .\n";
        print $fh_out2 '@{$old_indices_new_gbs} = (', join (', ', @out_old_indices_gbs), ") .\n";
        print $fh_out2 "\n\n";

	close ($fh_out2);
=cut # END of "for the predecessor 'print_out_info_chngs_by_smerge'."

	return 1;

}	# End of the subroutine, 'print_out_info_chngs_by_merge_split'. #


# print_out_info_chngs_by_cmerge_split_grow:
#	This subroutine prints out the portions of the outputs of 'cmerge_split_grow_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"(size-growing) merge + split (complementary types)\" and the \"shift"s immediately before and after that.
#	It was created by taking advantage of the frameworks of 'print_out_info_chngs_by_merge_split' and 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/11.]
#
# print_out_info_chngs_by_merge_split:
#	This subroutine prints out the portions of the outputs of 'smerge_split_gblock_pair' or 'icmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge + split (same/complementary-type(s))\" and the \"shift"s immediately before and after that.
#	It was created by slightly modifying 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/01.]
#	[REVISED on 2019/08/11]:
#	Now, it receives and prints out $ct_rmvd_null_clms as well.
#
# print_out_info_chngs_by_smerge:
#	This subroutine prints out the portions of the outputs of 'smerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (same-type)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_cmerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/22.]
#
# print_out_info_chngs_by_cmerge:
#	This subroutine prints out the portions of the outputs of 'cmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (complementary-types)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_purge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/20.]
#
# print_out_info_chngs_by_purge:
#	This subroutine prints out the portions of the outputs of 'purge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"purge\" and the \"shift"s immediately before that.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine receives a list of output file names as well.
#
# [ INPUTS: (\$outdir, \@ofnames, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, $new_bl_sbj, $incr_coord_new_sbj), where  ] ... for the predecessor, 'print_out_info_chngs_by_merge_split'.
#
# INPUTS: (\$outdir, \@ofnames, $flag_end_sbrtn, $gb_left, $gb_right, \@records_sessions, \%gb2incr_gb_coord_undo,
#		# IMPORTED from 'print_out_info_chngs_by_cvmerge' in revision on 2019/08/19. (1) #
#	\@new_indices_old_gbs, \@old_indices_new_gbs 
#		# END of "IMPORTED from 'print_out_info_chngs_by_cvmerge' in revision on 2019/08/19. (1)" #
#	), where
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 3 output files;
#
#	The other arguments are a part of the outputs of 'cmerge_split_grow_gblock_pair' in "MyANEX_Main.pm";
#	More precisely,
#
#	$flag_end_sbrtn = 1 if the subroutine 'cmerge_split_grow_gblock_pair' ended without any remarkable problems,
#	          	= 0 if it failed;
#		  	= -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; (ADDED on 2019/08/07.)
#
#	$gb_left and $gb_right are the indices of the members on the left and on the right, respectively, of the subject gap-block;
#
#	@{$records_sessions[$k]} = ($session_ID, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2)
#	provides the record of the $k th session, where
#	$session_ID is the ID of the session (integer),
#	$sbjct_site is the index of the column that is "split" in the session,
#	$incr_coord_sbj_smg1/2 is the increment of the coordinate of the gap-blocks
#	created by the "split" of the subject column necessary for them to be "absorb"ed to the $gb_right/$gb_left th gap-block;
#			
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#
#		# IMPORTED from 'print_out_info_chngs_by_cvmerge' in revision on 2019/08/19. (2) #
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "cvmerge";
#	More precisely, the gap-block that had been the $gb th before the "cvmerge"
#	has been assigned $new_indices_old_gbs[$gb] after the "cvmerge"
#	(, and $new_indices_old_gbs[$gb_rmvd_kpt] is undefined when $if_complete = 1);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "cvmerge";
#	More precisely, the gap-block that is the $gb th after the "cvmerge"
#	had been assigned $old_indices_new_gbs[$gb] before the "cvmerge";
#		# END of "IMPORTED from 'print_out_info_chngs_by_cvmerge' in revision on 2019/08/19. (2)" #
# 
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
# [ sub print_out_info_chngs_by_merge_split (\$\@$$$$) {
#	my ($rf_outdir, $ofnames, $bl_sbj, $incr_coord_sbj, $new_bl_sbj, $incr_coord_new_sbj) = @_; ] ... OBSOLETE as of 2019/08/11.
#
# [ sub print_out_info_chngs_by_merge_split (\$\@$$$$$) {
#	my ($rf_outdir, $ofnames, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, $new_bl_sbj, $incr_coord_new_sbj) = @_; # REVISED on 2019/08/11. ] ... for the predecessor, 'print_out_info_chngs_by_merge_split'. 
#
# [ sub print_out_info_chngs_by_cmerge_split_grow (\$\@$$$\@\%) {
#	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_left, $gb_right, $records_sessions, $gb2incr_gb_coord_undo) = @_; ] ... OBSOLETE as of 2019/08/19.
#
sub print_out_info_chngs_by_cmerge_split_grow (\$\@$$$\@\%\@\@) {

	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_left, $gb_right, $records_sessions, 
	$gb2incr_gb_coord_undo, $new_indices_old_gbs, $old_indices_new_gbs) = @_; # MODIFIED on 2019/08/19. #

	unless (@{$ofnames} == 4) { # MODIFIED on 2019/08/19.
	##unless (@{$ofnames} == 3) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge_split_grow'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 4 as desired: $! !\n\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge_split_grow'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge_split_grow'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $ofname1 = $ofnames->[0];
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge_split_grow'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	print $fh_out1 "{\$flag_end for 'cmerge_split_grow_gblock_pair'} = $flag_end_sbrtn .\n";
	print $fh_out1 "\n";

	print $fh_out1 "{LEFT member of the gap-block pair}  = $gb_left ,\n";
	print $fh_out1 "{RIGHT member of the gap-block pair} = $gb_right .\n";
	print $fh_out1 "\n";

	close ($fh_out1);


			### RESTART FROM HERE (on 2019/08/11). ###

	my $ofname2 = $ofnames->[1];
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge_split_grow'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}

	my $ct_sessions = @{$records_sessions};

	print $fh_out2 "# Records of the Sessions for the \"(size-growing) Merge + Split\" .\n";
	print $fh_out2 "#  #{Sessions} = $ct_sessions .\n";
	print $fh_out2 "\n";

	print $fh_out2 join ("\t", 'Session_ID', 'sbjct_site', 'incr_coord_sbj_smg_right', 'incr_coord_sbj_smg_left'), "\n";

	foreach my $record (@{$records_sessions}) { print $fh_out2 join ("\t", @{$record}), "\n"; }
	print $fh_out2 "\n";

	close ($fh_out2);


	my $ofname3 = $ofnames->[2];
	my $outfile3 = "$outdir/$ofname3";

	my $fh_out3 = open_output_file ($outfile3, 0);
	unless (defined $fh_out3) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_cmerge_split_grow'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile3\": $! !\n";
		return 0;
	}

	my $ct_rlv_gbs = scalar (keys %{$gb2incr_gb_coord_undo});

	print $fh_out3 "# Increments of the Coordinates of the Relevant Gap-Blocks When \"UNDO\"ing the Overlaps among them.\n";
	print $fh_out3 "#  #{Relevant Gap-Blocks} = $ct_rlv_gbs .\n";
	print $fh_out3 "\n";

	print $fh_out3 join ("\t", 'Indx_gb', 'incr_coord_gb'), "\n\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_gb_coord_undo}) {
	    print $fh_out3 join ("\t", $gb, $gb2incr_gb_coord_undo->{$gb}), "\n";
	}
	print $fh_out3 "\n";

	close ($fh_out3);


		# IMPORTED from 'print_out_info_chngs_by_cvmerge' in revision on 2019/08/19. (3) #
	my $ofname5 = $ofnames->[3];
	my $outfile5 = "$outdir/$ofname5";

	my $fh_out5 = open_output_file ($outfile5, 0);
	unless (defined $fh_out5) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_cvmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile5\": $! !\n";
		return 0;
	}

	my $ct_old_gbs = @{$new_indices_old_gbs};
        my $ct_new_gbs = @{$old_indices_new_gbs};

        print $fh_out5 "## Relationships between new and old indices of gap-blocks\n";
	print $fh_out5 "##  (#{gap-blocks}_old=$ct_old_gbs, #{gap-blocks}_new=$ct_new_gbs) :\n\n";

        my @out_new_indices_gbs = my @out_old_indices_gbs = ();
        foreach my $indx (@{$new_indices_old_gbs}) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$old_indices_new_gbs}) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

        print $fh_out5 '@{$new_indices_old_gbs} = (', join (', ', @out_new_indices_gbs), ") .\n";
        print $fh_out5 '@{$old_indices_new_gbs} = (', join (', ', @out_old_indices_gbs), ") .\n";
        print $fh_out5 "\n\n";

	close ($fh_out5);
		# END of "IMPORTED from 'print_out_info_chngs_by_cvmerge' in revision on 2019/08/19. (3)" #

	return 1;

}	# End of the subroutine, 'print_out_info_chngs_by_cmerge_split_grow'. #


# print_out_info_chngs_by_split:
#	This subroutine prints out the portions of the outputs of '{s,c}split_gblock' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"split\" and the \"shift\"s immediately before and after that.
#	It was created by taking advantage of the frameworks of 'print_out_info_chngs_by_cmerge_split_grow'
#	and importing a portion from 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/14.]
#	[REVISED on 2019/08/20]:
#	Now, the subroutine can accommodate a wider variety of output formats.
#
# print_out_info_chngs_by_cmerge_split_grow:
#	This subroutine prints out the portions of the outputs of 'cmerge_split_grow_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"(size-growing) merge + split (complementary types)\" and the \"shift"s immediately before and after that.
#	It was created by taking advantage of the frameworks of 'print_out_info_chngs_by_merge_split' and 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/11.]
#
# print_out_info_chngs_by_merge_split:
#	This subroutine prints out the portions of the outputs of 'smerge_split_gblock_pair' or 'icmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge + split (same/complementary-type(s))\" and the \"shift"s immediately before and after that.
#	It was created by slightly modifying 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/01.]
#	[REVISED on 2019/08/11]:
#	Now, it receives and prints out $ct_rmvd_null_clms as well.
#
# print_out_info_chngs_by_smerge:
#	This subroutine prints out the portions of the outputs of 'smerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (same-type)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_cmerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/22.]
#
# print_out_info_chngs_by_cmerge:
#	This subroutine prints out the portions of the outputs of 'cmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (complementary-types)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_purge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/20.]
#
# print_out_info_chngs_by_purge:
#	This subroutine prints out the portions of the outputs of 'purge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"purge\" and the \"shift"s immediately before that.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine receives a list of output file names as well.
#
# [ INPUTS: (\$outdir, \@ofnames, $flag_end_sbrtn, $gb_left, $gb_right, \@records_sessions, \%gb2incr_gb_coord_undo), where ] ... for the predecessor, 'print_out_info_chngs_by_cmerge_split_grow'.
#
# [ INPUTS: (\$outdir, \@ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj1_new, $gb_sbj2_new, \@new_indices_old_gbs, \@old_indices_new_gbs, \@records_split_sessions, \%gb2incr_gb_coord_undo), where ] ... OBSOLETE as of 2019/08/20.
#
# INPUTS: (\$outdir, \@ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj_new, $gb_cmpl_new, $new_gb_left, $new_gb_right, \@new_indices_old_gbs, \@old_indices_new_gbs, \@med_indices_new_gbs, \@records_split_sessions, \%gb2incr_gb_coord_undo), where # REVISED on 2019/08/20.
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 4 output files;
#
#	The other arguments are a part of the outputs of '{s,c}split_gblock' in "MyANEX_Main.pm";
#	More precisely,
#
#	$flag_end_sbrtn = 1 if the subroutine 'cmerge_split_grow_gblock_pair' ended without any remarkable problems,
#	          	= 0 if it failed;
#		  	= -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; (ADDED on 2019/08/07.)
#
#	$gb_sbj_old is the index (in the OLD gap-block set) of the "subject" gap-block to be "split";
#
#	$gb_sbj_NEW and $gb_cmpl_NEW are the indices (in the NEW gap-block set) of the "subject" gap-block and the newly created complementary gap-block, respectively;
#	$new_gb_left and $new_gb_right are the indices (in the NEW gap-block set) of the newly created gap-blocks (including the subject) that are on the left-side and on the right-side, respectively;
#	[[
#	$gb_sbj1_new and $gb_sbj2_new are the indices (in the NEW gap-block set) of the gap-blocks resulting from the "split";
#	]] ... OBSOLETE as of 2019/08/20.
#
#		# IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14). #
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "smerge";
#	More precisely, the gap-block that had been the $gb th before the "smerge"
#	has been assigned $new_indices_old_gbs[$gb] after the "smerge"
#	(, and $new_indices_old_gbs[$gb_absbd] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "smerge";
#	More precisely, the gap-block that is the $gb th after the "smerge"
#	had been assigned $old_indices_new_gbs[$gb] before the "smerge";
#		# END of "IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14)." #
#
#	@{$records_split_sessions[$k]} = ($session_ID, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2)
#	provides the record of the $k th session within the MAIN process of "split", where
#	$session_ID is the ID of the session (integer),
#	$sbjct_site is the index of the column that is "split" in the session,
#	$incr_coord_sbj_smg1/2 is the increment of the coordinate of the gap-blocks
#	created by the "split" of the subject column necessary for them to be "absorb"ed to the $gb_right/$gb_left th gap-block;
#	[NOTE: @records_split_sessions should usually be empty for, e.g., 'ssplit_gblocks'];
#			
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
#
# [ sub print_out_info_chngs_by_cmerge_split_grow (\$\@$$$\@\%) {
#	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_left, $gb_right, $records_sessions, $gb2incr_gb_coord_undo) = @_; ] ... for the predecessor, 'print_out_info_chngs_by_cmerge_split_grow'.
#
# [ sub print_out_info_chngs_by_split (\$\@$$$$\@\@\@\%) {
#	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj1_new, $gb_sbj2_new, $new_indices_old_gbs, $old_indices_new_gbs, $records_split_sessions, $gb2incr_gb_coord_undo) = @_; ] ... OBSOLETE as of 2019/08/20.
#
sub print_out_info_chngs_by_split (\$\@$$$$$$\@\@\@\@\%) { # REVISED on 2019/08/20.

	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj_NEW, $gb_cmpl_NEW, $new_gb_left, $new_gb_right, $new_indices_old_gbs, $old_indices_new_gbs, $med_indices_new_gbs, $records_split_sessions, $gb2incr_gb_coord_undo) = @_; # MODIFIED in REVISION on 2019/08/20.

	unless (@{$ofnames} == 4) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_split'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 4 as desired: $! !\n\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_split'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_split'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $ofname1 = $ofnames->[0];
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_split'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	print $fh_out1 "{\$flag_end for '{s,c}split_gblock'} = $flag_end_sbrtn .\n";
	print $fh_out1 "\n";

	print $fh_out1 "Index{gap-block to be \"split\"}_OLD = $gb_sbj_old .\n\n";

			# REVISED on 2019/08/20. (1) #
	print $fh_out1 "Index{gap-block to be \"split\"}_NEW  = ", ((defined $gb_sbj_NEW) ? $gb_sbj_NEW : 'undef'), " ,\n";
	print $fh_out1 "Index{complementary gap-block resulting from the \"split\"}_NEW  = $gb_cmpl_NEW .\n\n";
	print $fh_out1 "Index(left member of the newly created gap-block pair)_NEW  = $new_gb_left ,\n";
	print $fh_out1 "Index(right member of the newly created gap-block pair)_NEW = $new_gb_right .\n";
			# End of "REVISED on 2019/08/20. (1)" #
	print $fh_out1 "\n";

	close ($fh_out1);


		# IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14). #
	my $ofname2 = $ofnames->[1];
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_split'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}

	my $ct_old_gbs = @{$new_indices_old_gbs};
        my $ct_new_gbs = @{$old_indices_new_gbs};

        print $fh_out2 "## Relationships between new and old indices of gap-blocks\n";
	print $fh_out2 "##  (#{gap-blocks}_old=$ct_old_gbs, #{gap-blocks}_new=$ct_new_gbs) :\n\n";

        my @out_new_indices_gbs = my @out_old_indices_gbs = ();
        foreach my $indx (@{$new_indices_old_gbs}) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$old_indices_new_gbs}) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

		# ADDED in REVISION (on 2019/08/20). (1) #
	my @out_med_indices_new_gbs = ();
	foreach my $indx (@{$med_indices_new_gbs}) { push @out_med_indices_new_gbs, ((defined $indx) ? $indx : 'undef'); }
		# End of "ADDED in REVISION (on 2019/08/20). (1)" #

        print $fh_out2 '@{$new_indices_old_gbs} = (', join (', ', @out_new_indices_gbs), ") .\n";
        print $fh_out2 '@{$old_indices_new_gbs} = (', join (', ', @out_old_indices_gbs), ") .\n";

	print $fh_out2 '@{$med_indices_new_gbs} = (', join (', ', @out_med_indices_new_gbs), ") .\n"; #  ADDED in REVISION (on 2019/08/20). (2) #

        print $fh_out2 "\n\n";

	close ($fh_out2);
		# END of "IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14)." #


			### RESTART FROM HERE (on 2019/08/11). ###

	my $ofname3 = $ofnames->[2];
	my $outfile3 = "$outdir/$ofname3";

	my $fh_out3 = open_output_file ($outfile3, 0);
	unless (defined $fh_out3) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_split'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile3\": $! !\n";
		return 0;
	}

	my $ct_sessions = @{$records_split_sessions};

    if ($ct_sessions>0) { # Added the condition on 2019/08/15.

	print $fh_out3 "# Records of the Sessions for the \"Split\" .\n";
	print $fh_out3 "#  #{Sessions} = $ct_sessions .\n";
	print $fh_out3 "\n";

	print $fh_out3 join ("\t", 'Session_ID', 'sbjct_site', 'incr_coord_sbj_smg_right', 'incr_coord_sbj_smg_left'), "\n";

	foreach my $record (@{$records_split_sessions}) { print $fh_out3 join ("\t", @{$record}), "\n"; }
	print $fh_out3 "\n";

    } else {
 		# Added the block on 2019/08/15. #

	print $fh_out3 "# NOTE:\n";
	print $fh_out3 "# This \"Split\" process is NOT subdivided into repetitive sessions ...\n\n";
	print $fh_out3 "#\n";
	print $fh_out3 "\n";

    } # End of "if ($ct_sessions>0) {...}"

	close ($fh_out3);


	my $ofname5 = $ofnames->[3];
	my $outfile5 = "$outdir/$ofname5";

	my $fh_out5 = open_output_file ($outfile5, 0);
	unless (defined $fh_out5) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_split'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile5\": $! !\n";
		return 0;
	}

	my $ct_rlv_gbs = scalar (keys %{$gb2incr_gb_coord_undo});

	print $fh_out5 "# Increments of the Coordinates of the Relevant Gap-Blocks When \"UNDO\"ing the Overlaps among them.\n";
	print $fh_out5 "#  #{Relevant Gap-Blocks} = $ct_rlv_gbs .\n";
	print $fh_out5 "\n";

	print $fh_out5 join ("\t", 'Indx_gb', 'incr_coord_gb'), "\n\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_gb_coord_undo}) {
	    print $fh_out5 join ("\t", $gb, $gb2incr_gb_coord_undo->{$gb}), "\n";
	}
	print $fh_out5 "\n";

	close ($fh_out5);

	return 1;

}	# End of the subroutine, 'print_out_info_chngs_by_split'. #


# print_out_info_chngs_by_vsplit:
#	This subroutine prints out the portions of the outputs of '(c)vsplit_gblock' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"vertical split\" and the \"shift\"s immediately after that.
#	It was created by modifying 'print_out_info_chngs_by_split'
#	and importing a portion from 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/25.]
#	[REVISED on 2019/08/30]:
#	Now, it also receives %gb2incr_coord_1way_dsplc as an input.
#
# print_out_info_chngs_by_split:
#	This subroutine prints out the portions of the outputs of '{s,c}split_gblock' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"split\" and the \"shift\"s immediately before and after that.
#	It was created by taking advantage of the frameworks of 'print_out_info_chngs_by_cmerge_split_grow'
#	and importing a portion from 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/14.]
#	[REVISED on 2019/08/20]:
#	Now, the subroutine can accommodate a wider variety of output formats.
#
# print_out_info_chngs_by_cmerge_split_grow:
#	This subroutine prints out the portions of the outputs of 'cmerge_split_grow_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"(size-growing) merge + split (complementary types)\" and the \"shift"s immediately before and after that.
#	It was created by taking advantage of the frameworks of 'print_out_info_chngs_by_merge_split' and 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/11.]
#
# print_out_info_chngs_by_merge_split:
#	This subroutine prints out the portions of the outputs of 'smerge_split_gblock_pair' or 'icmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge + split (same/complementary-type(s))\" and the \"shift"s immediately before and after that.
#	It was created by slightly modifying 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/01.]
#	[REVISED on 2019/08/11]:
#	Now, it receives and prints out $ct_rmvd_null_clms as well.
#
# print_out_info_chngs_by_smerge:
#	This subroutine prints out the portions of the outputs of 'smerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (same-type)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_cmerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/22.]
#
# print_out_info_chngs_by_cmerge:
#	This subroutine prints out the portions of the outputs of 'cmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"merge (complementary-types)\" and the \"shift"s immediately before that.
#	It was created by slightly modifying 'print_out_info_chngs_by_purge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/20.]
#
# print_out_info_chngs_by_purge:
#	This subroutine prints out the portions of the outputs of 'purge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"purge\" and the \"shift"s immediately before that.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/17.]
#	[REVISED on 2019/07/18]:
#	Now, the subroutine receives a list of output file names as well.
#
# [ INPUTS: (\$outdir, \@ofnames, $flag_end_sbrtn, $gb_left, $gb_right, \@records_sessions, \%gb2incr_gb_coord_undo), where ] ... for the predecessor, 'print_out_info_chngs_by_cmerge_split_grow'.
#
# [ INPUTS: (\$outdir, \@ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj1_new, $gb_sbj2_new, \@new_indices_old_gbs, \@old_indices_new_gbs, \@records_split_sessions, \%gb2incr_gb_coord_undo), where ] ... OBSOLETE as of 2019/08/20.
#
# [ INPUTS: (\$outdir, \@ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj_new, $gb_cmpl_new, $new_gb_left, $new_gb_right, \@new_indices_old_gbs, \@old_indices_new_gbs, \@med_indices_new_gbs, \@records_split_sessions, \%gb2incr_gb_coord_undo), where # REVISED on 2019/08/20. ] ... for the predecessor 'print_out_info_chngs_by_split'. #
#
# [ INPUTS: (\$outdir, \@ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj1_new, $gb_sbj2_new, $gb_shftd_new, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_gb_coord_undo), where ] ... OBSOLETE as of 2019/08/30.
#
# INPUTS: (\$outdir, \@ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj1_new, $gb_sbj2_new, \%gb2incr_coord_1way_dsplc, $gb_shftd_new, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_gb_coord_undo), where
#
#	$outdir is a string specifying the path to the output directory;
#	[ @ofnames stores the names of the 4 output files; ] ... OBSOLETE as of 2019/08/30.
#	@ofnames stores the names of the 5 output files; # MODIFIED on 2019/08/30.
#
#	The other arguments are a part of the outputs of '{s,c}split_gblock' in "MyANEX_Main.pm";
#	More precisely,
#
#	$flag_end_sbrtn = 1 if the subroutine 'cmerge_split_grow_gblock_pair' ended without any remarkable problems,
#	          	= 0 if it failed;
#		  	= -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; (ADDED on 2019/08/07.)
#
#	$gb_sbj_old is the index (in the OLD gap-block set) of the "subject" gap-block to be "split";
#
#	$gb_sbj1_new and $gb_sbj2_new are the indices (in the NEW gap-block set) of the gap-blocks resulting from the "vertical split"; # RESURRECTED in creation of 'print_out_info_chngs_by_vsplit'.
#
#		# ADDED in revision (on 2019/08/30). #
#	\%gb2incr_coord_1way_dsplc = ($gb_old => $incr_coord, ..) stores information on the 1-way displacement before the "vertical-split" of $gb_sbj_old;
#		# End of "ADDED in revision (on 2019/08/30)." #
#
#		# ADDED in creation of 'print_out_info_chngs_by_vsplit'. #
#	$gb_shftd_new is the index (in the NEW gap-block set) of the gap-block that was "shift"ed to UN-NEST the resulting new subject blocks;
#	$incr_coord_shftd is the increment of the coordinate of $gb_shftd_new caused by the "shift";
#		# END of "ADDED in creation of 'print_out_info_chngs_by_vsplit'." #
#
#		# IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14). #
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "smerge";
#	More precisely, the gap-block that had been the $gb th before the "smerge"
#	has been assigned $new_indices_old_gbs[$gb] after the "smerge"
#	(, and $new_indices_old_gbs[$gb_absbd] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "smerge";
#	More precisely, the gap-block that is the $gb th after the "smerge"
#	had been assigned $old_indices_new_gbs[$gb] before the "smerge";
#		# END of "IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14)." #
#
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
#
# [ sub print_out_info_chngs_by_cmerge_split_grow (\$\@$$$\@\%) {
#	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_left, $gb_right, $records_sessions, $gb2incr_gb_coord_undo) = @_; ] ... for the predecessor, 'print_out_info_chngs_by_cmerge_split_grow'.
#
# [ sub print_out_info_chngs_by_split (\$\@$$$$\@\@\@\%) {
#	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj1_new, $gb_sbj2_new, $new_indices_old_gbs, $old_indices_new_gbs, $records_split_sessions, $gb2incr_gb_coord_undo) = @_; ] ... OBSOLETE as of 2019/08/20.
#
# [ sub print_out_info_chngs_by_split (\$\@$$$$$$\@\@\@\@\%) {
#	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj_NEW, $gb_cmpl_NEW, $new_gb_left, $new_gb_right, $new_indices_old_gbs, $old_indices_new_gbs, $med_indices_new_gbs, $records_split_sessions, $gb2incr_gb_coord_undo) = @_; # MODIFIED in REVISION on 2019/08/20. ] ... for the predecessor, 'print_out_info_chngs_by_split'.
#
# [ sub print_out_info_chngs_by_vsplit (\$\@$$$$$$\@\@\%) {
#	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj1_new, $gb_sbj2_new, $gb_shftd_new, $incr_coord_shftd, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_gb_coord_undo) = @_; ] ... OBSOLETE as of 2019/08/30.
#
sub print_out_info_chngs_by_vsplit (\$\@$$$$\%$$\@\@\%) {

	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj1_new, $gb_sbj2_new, 
	$gb2incr_coord_1way_dsplc, $gb_shftd_new, $incr_coord_shftd, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_gb_coord_undo) = @_;

	unless (@{$ofnames} == 5) { # MODIFIED on 2019/08/30.
	##unless (@{$ofnames} == 4) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 5 as desired: $! !\n\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $ofname1 = $ofnames->[0];
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	print $fh_out1 "{\$flag_end for '{s,c}split_gblock'} = $flag_end_sbrtn .\n";
	print $fh_out1 "\n";

	print $fh_out1 "Index{gap-block to be \"split\"}_OLD = $gb_sbj_old .\n\n";

			# MODIFIED in creation of 'print_out_info_chngs_by_vsplit'. #
	print $fh_out1 "Index{1st gap-block resulting from the \"split\"}_NEW  = $gb_sbj1_new ,\n";
	print $fh_out1 "Index{2nd gap-block resulting from the \"split\"}_NEW  = $gb_sbj2_new .\n\n";
			# End of "MODIFIED in creation of 'print_out_info_chngs_by_vsplit'." #

	print $fh_out1 "\n";

	close ($fh_out1);


			## ADDED in REVISION (on 2019/08/30). ##
	my $ofname2 = $ofnames->[1];
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}
	print $fh_out2 "# Increments of the Coordinates of the Relevant Gap-Blocks (w/ old indices) When \"DISPLAC\"ing them  before \"Vertically Split\"ting \$gb_sbj_old (=$gb_sbj_old).\n";

     if (defined $gb2incr_coord_1way_dsplc->{'N.A'}) {

	    print $fh_out2 "\n ... This information is NOT AVAILABLE, even in principle, for This Session ...\n\n";
     } else {

	my $ct_rlv_gbs = scalar (keys %{$gb2incr_coord_1way_dsplc});

	print $fh_out2 "#  #{Relevant Gap-Blocks} = $ct_rlv_gbs .\n";
	print $fh_out2 "\n";

	print $fh_out2 join ("\t", 'Indx_gb_old', 'incr_coord_gb'), "\n\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_1way_dsplc}) {
	    print $fh_out2 join ("\t", $gb, $gb2incr_coord_1way_dsplc->{$gb}), "\n";
	}
	print $fh_out2 "\n";

     } # End of "if (defined $gb2incr_coord_1way_dsplc->{'N.A'}) {...} else {...}"

	close ($fh_out2);

			## END of "ADDED in REVISION (on 2019/08/30)." ##

			### RESTART FROM HERE (on 2019/08/25). ###

		# IMPORTED (w/ modifications) from 'print_out_info_chngs_by_smerge' (on 2019/08/25). #

	my $ofname3 = $ofnames->[2];
	my $outfile3 = "$outdir/$ofname3";

	my $fh_out3 = open_output_file ($outfile3, 0);
	unless (defined $fh_out3) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile3\": $! !\n";
		return 0;
	}

	##my $ct_rmvd_null_clms = 0; # Always set this value to be zero. # ADDED in creation of 'print_out_info_chngs_by_smerge' (on 2019/07/22). #

	print $fh_out3 "{New subject gap-block \"shift\"ed} = $gb_shftd_new .\n";
	print $fh_out3 "{Increment of the block coordinate due to the \"shift\"} = $incr_coord_shftd .\n\n";

	##print $fh_out3 "#{Resulting null columns, which were removed during the \"merge\"} = $ct_rmvd_null_clms .\n\n";

	close ($fh_out3);
		# End of "IMPORTED (w/ modifications) from 'print_out_info_chngs_by_smerge' (on 2019/08/25)." #


		# IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14). #
	my $ofname5 = $ofnames->[3];
	##my $ofname2 = $ofnames->[1];
	my $outfile5 = "$outdir/$ofname5";

	my $fh_out5 = open_output_file ($outfile5, 0);
	unless (defined $fh_out5) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile5\": $! !\n";
		return 0;
	}

	my $ct_old_gbs = @{$new_indices_old_gbs};
        my $ct_new_gbs = @{$old_indices_new_gbs};

        print $fh_out5 "## Relationships between new and old indices of gap-blocks\n";
	print $fh_out5 "##  (#{gap-blocks}_old=$ct_old_gbs, #{gap-blocks}_new=$ct_new_gbs) :\n\n";

        my @out_new_indices_gbs = my @out_old_indices_gbs = ();
        foreach my $indx (@{$new_indices_old_gbs}) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$old_indices_new_gbs}) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

=For the predecessor 'print_out_info_chngs_by_split'. (1)"
		# ADDED in REVISION (on 2019/08/20). (1) #
	my @out_med_indices_new_gbs = ();
	foreach my $indx (@{$med_indices_new_gbs}) { push @out_med_indices_new_gbs, ((defined $indx) ? $indx : 'undef'); }
		# End of "ADDED in REVISION (on 2019/08/20). (1)" #
=cut # End of "For the predecessor 'print_out_info_chngs_by_split'. (1)"

        print $fh_out5 '@{$new_indices_old_gbs} = (', join (', ', @out_new_indices_gbs), ") .\n";
        print $fh_out5 '@{$old_indices_new_gbs} = (', join (', ', @out_old_indices_gbs), ") .\n";

	##print $fh_out5 '@{$med_indices_new_gbs} = (', join (', ', @out_med_indices_new_gbs), ") .\n"; #  ADDED in REVISION (on 2019/08/20). (2) # For the predecessor 'print_out_info_chngs_by_split'. (2) #

        print $fh_out5 "\n\n";

	close ($fh_out5);
		# END of "IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14)." #


=For the predecessor 'print_out_info_chngs_by_split'. (3)
	my $ofname3 = $ofnames->[2];
	my $outfile3 = "$outdir/$ofname3";

	my $fh_out3 = open_output_file ($outfile3, 0);
	unless (defined $fh_out3) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile3\": $! !\n";
		return 0;
	}

	my $ct_sessions = @{$records_split_sessions};

    if ($ct_sessions>0) { # Added the condition on 2019/08/15.

	print $fh_out3 "# Records of the Sessions for the \"Split\" .\n";
	print $fh_out3 "#  #{Sessions} = $ct_sessions .\n";
	print $fh_out3 "\n";

	print $fh_out3 join ("\t", 'Session_ID', 'sbjct_site', 'incr_coord_sbj_smg_right', 'incr_coord_sbj_smg_left'), "\n";

	foreach my $record (@{$records_split_sessions}) { print $fh_out3 join ("\t", @{$record}), "\n"; }
	print $fh_out3 "\n";

    } else {
 		# Added the block on 2019/08/15. #

	print $fh_out3 "# NOTE:\n";
	print $fh_out3 "# This \"Split\" process is NOT subdivided into repetitive sessions ...\n\n";
	print $fh_out3 "#\n";
	print $fh_out3 "\n";

    } # End of "if ($ct_sessions>0) {...}"

	close ($fh_out3);
=cut # End of "For the predecessor 'print_out_info_chngs_by_split'. (3)"


	my $ofname6 = $ofnames->[4];
	my $outfile6 = "$outdir/$ofname6";

	my $fh_out6 = open_output_file ($outfile6, 0);
	unless (defined $fh_out6) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile6\": $! !\n";
		return 0;
	}

	my $ct_rlv_gbs = scalar (keys %{$gb2incr_gb_coord_undo});

	print $fh_out6 "# Increments of the Coordinates of the Relevant Gap-Blocks When \"UNDO\"ing the Overlaps among them.\n";
	print $fh_out6 "#  #{Relevant Gap-Blocks} = $ct_rlv_gbs .\n";
	print $fh_out6 "\n";

	print $fh_out6 join ("\t", 'Indx_gb_NEW', 'incr_coord_gb'), "\n\n"; # MODIFIED in revision (on 2019/08/30). #
	##print $fh_out6 join ("\t", 'Indx_gb', 'incr_coord_gb'), "\n\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_gb_coord_undo}) {
	    print $fh_out6 join ("\t", $gb, $gb2incr_gb_coord_undo->{$gb}), "\n";
	}
	print $fh_out6 "\n";

	close ($fh_out6);

	return 1;

}	# End of the subroutine, 'print_out_info_chngs_by_vsplit'. #


# print_out_info_chngs_by_vsplit_plus_smerge:
#	This subroutine prints out the portions of the outputs of '(c)vsplit_plus_smerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"vertical split + merge (same-type)\" and the \"shift\"s accompanying it.
#	It was created by modifying 'print_out_info_chngs_by_vsplit'
#	and importing a portion from {'print_out_info_chngs_by_smerge'}.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/09/08.]
#	[Minor revision on 2019/09/12]: Changed 'vsplit' to '(c)vsplit', and similarly changed the printed-out strings.
#
# print_out_info_chngs_by_vsplit:
#	This subroutine prints out the portions of the outputs of '(c)vsplit_gblock' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"vertical split\" and the \"shift\"s immediately after that.
#	It was created by modifying 'print_out_info_chngs_by_split'
#	and importing a portion from 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/25.]
#	[REVISED on 2019/08/30]:
#	Now, it also receives %gb2incr_coord_1way_dsplc as an input.
#
# print_out_info_chngs_by_split:
#	This subroutine prints out the portions of the outputs of '{s,c}split_gblock' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"split\" and the \"shift\"s immediately before and after that.
#	It was created by taking advantage of the frameworks of 'print_out_info_chngs_by_cmerge_split_grow'
#	and importing a portion from 'print_out_info_chngs_by_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/08/14.]
#	[REVISED on 2019/08/20]:
#	Now, the subroutine can accommodate a wider variety of output formats.
#
#
# [ INPUTS: (\$outdir, \@ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj1_new, $gb_sbj2_new,
#	\%gb2incr_coord_1way_dsplc, $gb_shftd_new, $incr_coord_shftd, 
#	\@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_gb_coord_undo), where ] ... for the predecessor, 'print_out_info_chngs_by_vsplit'.
#
# INPUTS: (\$outdir, \@ofnames,
#	$flag_end_sbrtn, $gb_in_2B_vsplit_old, $gb_in_2B_smgd_old, $gb_out_unmgd_new, $gb_out_kpt_new, 
#	\%gb_old2incr_coord_1way_dsplc_bf_vspl,
#	$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, \%gb_intm2incr_coord_undo_af_vspl,
#	$gb_ch_2B_smgd_intm, $gb_in1_4smg_intm, $gb_in2_4smg_intm, $gb_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, 
#	 \@new_indices_old_gbs, \@old_indices_new_gbs, \@intm_indices_old_gbs, \@new_indices_intm_gbs,
#	\%gb_new2incr_coord_undo_af_smg), where
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 7 output files;
#
#	The other arguments are a part of the outputs of '(c)vsplit_plus_smerge_gblock_pair' in "MyANEX_Main.pm";
#	More precisely,
#
#	$flag_end_sbrtn = 1 if the subroutine 'cmerge_split_grow_gblock_pair' ended without any remarkable problems,
#	          	= 0 if it failed;
#		  	= -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; (ADDED on 2019/08/07.)
#
#	$gb_in_2B_vsplit_old is the index (in the OLD gap-block set) of the input "subject" gap-block to be "vertically split";
#	$gb_in_2B_smgd_old is the index (in the OLD gap-block set) of the input "subject" gap-block to be "(same-type) merge"d;
#
#	$gb_out_unmgd_new is the index (in the NEW gap-block set) of the "output" gap-block that was left UNmerged;
#	$gb_out_kpt_new is the index (in the NEW gap-block set) of the "output" gap-block that was KEPT by the "(same-type) merge";
#
#	\%gb_old2incr_coord_1way_dsplc = ($gb_old => $incr_coord, ..) stores information on the 1-way displacement before the "vertical-split" of $gb_in_2B_vsplit_old;
#
#	$gb_shftd_af_vspl_intm is the index (in the INTERMEDIATE gap-block set) of the gap-block that was "shift"ed to UN-NEST the (intermediate) subject blocks resulting from the "vertical split";
#	$incr_coord_shftd_af_vspl is the increment of the coordinate of $gb_shftd_af_vspl_intm caused by the "shift";
#
#	%gb_intm2incr_coord_undo_af_vspl = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "vertical split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#
#	$gb_ch_2B_smgd_intm is the index (in the INTERMEDIATE gap-block set) of the "child" gap-block resulting from the "verrical split" that is to be "smerge"d;
#
#	$gb_in1_4smg_intm and $gb_in2_4smg_intm are the indices (in the INTERMEDIATE gap-block set) of the 1st and 2nd input "subject" gap-blocks, respectively, for the "smerge_gblock_pair";
#
#	$gb_sbj_ssh4smg_intm is the index (in the INTERMEDIATE gap-block set) of the "subject" gap-block for the "serial shifts" necessary for the 'smerge_gblock_pair';
#	$incr_coord_sbj_ssh4smg is the coordinate increment of $gb_sbj_ssh4smg_intm by the aforementioned "serial shifts";
#	$gb_absbd_intm is the index (in the INTERMEDIATE gap-block set) of the gap-block that is "absorbed" when performing the 'smerge_gblock_pair';
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "vsplit + smerge";
#	More precisely, the gap-block that had been the $ogb th before the "vsplit + smerge"
#	has been assigned $new_indices_old_gbs[$ogb] after the "vsplit + smerge";
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "vsplit + smerge";
#	More precisely, the gap-block that is the $ngb th after the "vsplit + smerge"
#	had been assigned $old_indices_new_gbs[$ngb] before the "vsplit + smerge";
#
#	@intm_indices_old_gbs stores the intermediate indices assigned to the gap-blocks after the "vsplit" and before the "smerge";
#	More precisely, the gap-block that had been the $ogb th before the "vsplit"
#	has been assigned $intm_indices_old_gbs[$ogb] after the "vsplit";

#	@new_indices_intm_gbs stores the new indices assigned to the gap-blocks after the "smerge";
#	More precisely, the gap-block that had been the $igb th before the "smerge"
#	has been assigned $new_indices_intm_gbs[$igb] after the "smerge";
#
#	%gb_new2incr_coord_undo_af_smg = ($indx_gb_new => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "(same-type) merge"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_coord_undo_af_smg is one of the main outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' above];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
#
# [ sub print_out_info_chngs_by_vsplit (\$\@$$$$\%$$\@\@\%) {
#	my ($rf_outdir, $ofnames, $flag_end_sbrtn, $gb_sbj_old, $gb_sbj1_new, $gb_sbj2_new, 
#	$gb2incr_coord_1way_dsplc, $gb_shftd_new, $incr_coord_shftd, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_gb_coord_undo) = @_; ] ... for the predecessor, 'print_out_info_chngs_by_vsplit'.
#
sub print_out_info_chngs_by_vsplit_plus_smerge (\$\@$$$$$\%$$\%$$$$$$\@\@\@\@\%) {

	my ($rf_outdir, $ofnames, 
	$flag_end_sbrtn, $gb_in_2B_vsplit_old, $gb_in_2B_smgd_old, $gb_out_unmgd_new, $gb_out_kpt_new, 
	$gb_old2incr_coord_1way_dsplc_bf_vspl,
	$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $gb_intm2incr_coord_undo_af_vspl,
	$gb_ch_2B_smgd_intm, $gb_in1_4smg_intm, $gb_in2_4smg_intm, $gb_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, 
	 $new_indices_old_gbs, $old_indices_new_gbs, $intm_indices_old_gbs, $new_indices_intm_gbs,
	$gb_new2incr_coord_undo_af_smg) = @_;


	unless (@{$ofnames} == 7) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit_plus_smerge'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 7 as desired: $! !\n\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit_plus_smerge'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit_plus_smerge'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $ofname1 = $ofnames->[0];
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit_plus_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	print $fh_out1 "{\$flag_end for '(c)vsplit_plus_smerge_gblock_pair'} = $flag_end_sbrtn .\n";
	print $fh_out1 "\n";

	print $fh_out1 "Index{Input gap-block to be \"(c)vsplit\"}_OLD = $gb_in_2B_vsplit_old ,\n";
	print $fh_out1 "Index{Input gap-block to be \"smerge\"d}_OLD = $gb_in_2B_smgd_old .\n\n";

	print $fh_out1 "Index{Output gap-block that was left UNmerged}_NEW  = $gb_out_unmgd_new ,\n";
	print $fh_out1 "Index{Output gap-block that was kept by the \"(same-type) merge\"}_NEW  = $gb_out_kpt_new .\n\n";

	print $fh_out1 "\n";

	close ($fh_out1);


	my $ofname2 = $ofnames->[1];
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit_plus_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}
	print $fh_out2 "# Increments of the Coordinates of the Relevant Gap-Blocks (w/ old indices) When \"DISPLAC\"ing them  before \"(Complementarily-)Vertically Split\"ting \$gb_in_2B_vsplit_old (=$gb_in_2B_vsplit_old).\n";

	my $gb2incr_coord_1way_dsplc = $gb_old2incr_coord_1way_dsplc_bf_vspl;

     if (defined $gb2incr_coord_1way_dsplc->{'N.A'}) {

	    print $fh_out2 "\n ... This information is NOT AVAILABLE, even in principle, for This Session ...\n\n";
     } else {


	my $ct_rlv_gbs = scalar (keys %{$gb2incr_coord_1way_dsplc});

	print $fh_out2 "#  #{Relevant Gap-Blocks} = $ct_rlv_gbs .\n";
	print $fh_out2 "\n";

	print $fh_out2 join ("\t", 'Indx_gb_old', 'incr_coord_gb'), "\n\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_1way_dsplc}) {
	    print $fh_out2 join ("\t", $gb, $gb2incr_coord_1way_dsplc->{$gb}), "\n";
	}
	print $fh_out2 "\n";

     } # End of "if (defined $gb2incr_coord_1way_dsplc->{'N.A'}) {...} else {...}"

	close ($fh_out2);



		# IMPORTED (w/ modifications) from 'print_out_info_chngs_by_smerge' (on 2019/08/25). #

	my $ofname3 = $ofnames->[2];
	my $outfile3 = "$outdir/$ofname3";

	my $fh_out3 = open_output_file ($outfile3, 0);
	unless (defined $fh_out3) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit_plus_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile3\": $! !\n";
		return 0;
	}

	##my $ct_rmvd_null_clms = 0; # Always set this value to be zero. # ADDED in creation of 'print_out_info_chngs_by_smerge' (on 2019/07/22). #

	print $fh_out3 "{Gap-block \"shift\"ed after '(c)vsplit'}_INTM = $gb_shftd_af_vspl_intm .\n";
	print $fh_out3 "{Increment of the block coordinate due to the \"shift\" after '(c)vsplit'} = $incr_coord_shftd_af_vspl .\n\n";

	##print $fh_out3 "#{Resulting null columns, which were removed during the \"merge\"} = $ct_rmvd_null_clms .\n\n";

	close ($fh_out3);
		# End of "IMPORTED (w/ modifications) from 'print_out_info_chngs_by_smerge' (on 2019/08/25)." #


		# COPIED (w/ modifications) from below, in creation of 'print_out_info_chngs_by_vsplit_plus_smerge' (on 2019/09/08). #
	my $ofname5 = $ofnames->[3];
	my $outfile5 = "$outdir/$ofname5";

	my $fh_out5 = open_output_file ($outfile5, 0);
	unless (defined $fh_out5) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit_plus_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile5\": $! !\n";
		return 0;
	}

	my $ct_rlv_gbs_intm = scalar (keys %{$gb_intm2incr_coord_undo_af_vspl});

	print $fh_out5 "# Increments of the Coordinates of the Relevant Gap-Blocks When \"UNDO\"ing the Overlaps among them AFTER the '(c)vsplit'.\n";
	print $fh_out5 "#  #{Relevant Gap-Blocks} = $ct_rlv_gbs_intm .\n";
	print $fh_out5 "\n";

	print $fh_out5 join ("\t", 'Indx_gb_INTM', 'incr_coord_gb'), "\n\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb_intm2incr_coord_undo_af_vspl}) {
	    print $fh_out5 join ("\t", $gb, $gb_intm2incr_coord_undo_af_vspl->{$gb}), "\n";
	}
	print $fh_out5 "\n";

	close ($fh_out5);
		# End of "COPIED (w/ modifications) from below, in creation of 'print_out_info_chngs_by_vsplit_plus_smerge' (on 2019/09/08)." #


		# ADDED in creation of 'print_out_info_chngs_by_vsplit_plus_smerge' (on 2019/09/08). #
	my $ofname6 = $ofnames->[4];
	my $outfile6 = "$outdir/$ofname6";

	my $fh_out6 = open_output_file ($outfile6, 0);
	unless (defined $fh_out6) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit_plus_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile6\": $! !\n";
		return 0;
	}

	print $fh_out6 "Index{Gap-block product of 'vsplit' to be \"smerge\"d}_INTM = $gb_ch_2B_smgd_intm .\n\n";
	print $fh_out6 "Index{1st Input gap-block for \"smerge\"}_INTM = $gb_in1_4smg_intm ,\n";
	print $fh_out6 "Index{2nd Input gap-block for \"smerge\"}_INTM = $gb_in2_4smg_intm .\n\n";

	print $fh_out6 "Index{Input gap-block for \"smerge\" that was \"serially shift\"ed}_INTM  = $gb_sbj_ssh4smg_intm ,\n";
	print $fh_out6  "{Increment of its coordinate by the \"serial shift\"} = $incr_coord_sbj_ssh4smg .\n\n";
	print $fh_out6 "Index{Input gap-block for \"smerge\" that was \"absorb\"ed}_INTM = $gb_absbd_intm .\n\n";
	print $fh_out6 "\n";

	close ($fh_out6);
		# END of "ADDED in creation of 'print_out_info_chngs_by_vsplit_plus_smerge' (on 2019/09/08)." #


		# IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14). #
	my $ofname7 = $ofnames->[5];
	my $outfile7 = "$outdir/$ofname7";

	my $fh_out7 = open_output_file ($outfile7, 0);
	unless (defined $fh_out7) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit_plus_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile7\": $! !\n";
		return 0;
	}

	my $ct_old_gbs = @{$new_indices_old_gbs};
	my $ct_intm_gbs = @{$new_indices_intm_gbs};
        my $ct_new_gbs = @{$old_indices_new_gbs};

        print $fh_out7 "## Relationships between new, intermediate, and old indices of gap-blocks\n";
	print $fh_out7 "##  (#{gap-blocks}_old=$ct_old_gbs, #{gap-blocks}_intm=$ct_intm_gbs, #{gap-blocks}_new=$ct_new_gbs) :\n\n";

        my @out_new_indices_gbs = my @out_old_indices_gbs = ();
        foreach my $indx (@{$new_indices_old_gbs}) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$old_indices_new_gbs}) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

	my @out_intm_indices_old_gbs = my @out_new_indices_intm_gbs = ();
	foreach my $indx (@{$intm_indices_old_gbs}) { push @out_intm_indices_old_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$new_indices_intm_gbs}) { push @out_new_indices_intm_gbs, ((defined $indx) ? $indx : 'undef'); }

        print $fh_out7 '@{$new_indices_old_gbs} = (', join (', ', @out_new_indices_gbs), ") .\n";
        print $fh_out7 '@{$old_indices_new_gbs} = (', join (', ', @out_old_indices_gbs), ") .\n\n";

        print $fh_out7 '@{$intm_indices_old_gbs} = (', join (', ', @out_intm_indices_old_gbs), ") .\n";
        print $fh_out7 '@{$new_indices_intm_gbs} = (', join (', ', @out_new_indices_intm_gbs), ") .\n";
        print $fh_out7 "\n\n";

	close ($fh_out7);
		# END of "IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14)." #


	my $ofname8 = $ofnames->[6];
	my $outfile8 = "$outdir/$ofname8";

	my $fh_out8 = open_output_file ($outfile8, 0);
	unless (defined $fh_out8) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_vsplit_plus_smerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile8\": $! !\n";
		return 0;
	}

	# $gb_new2incr_coord_undo_af_smg) = @_;

	my $ct_rlv_gbs_new = scalar (keys %{$gb_new2incr_coord_undo_af_smg});

	print $fh_out8 "# Increments of the Coordinates of the Relevant Gap-Blocks When \"UNDO\"ing the Overlaps among them AFTER the 'smerge'.\n";
	print $fh_out8 "#  #{Relevant Gap-Blocks} = $ct_rlv_gbs_new .\n";
	print $fh_out8 "\n";

	print $fh_out8 join ("\t", 'Indx_gb_NEW', 'incr_coord_gb'), "\n\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb_new2incr_coord_undo_af_smg}) {
	    print $fh_out8 join ("\t", $gb, $gb_new2incr_coord_undo_af_smg->{$gb}), "\n";
	}
	print $fh_out8 "\n";

	close ($fh_out8);

	return 1;

}	# End of the subroutine, 'print_out_info_chngs_by_vsplit_plus_smerge'. #


# print_out_info_chngs_by_ssplit_plus_vmerge:
#	This subroutine prints out the portions of the outputs of 'ssplit_plus_vmerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"(horizontal) split (same-type) + vertical-merge\" and the \"shift\"s accompanying it.
#	It was created by modifying 'print_out_info_chngs_by_vsplit_plus_smerge'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/10/17-18.]
#
# print_out_info_chngs_by_vsplit_plus_smerge:
#	This subroutine prints out the portions of the outputs of '(c)vsplit_plus_smerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"vertical split + merge (same-type)\" and the \"shift\"s accompanying it.
#	It was created by modifying 'print_out_info_chngs_by_vsplit'
#	and importing a portion from {'print_out_info_chngs_by_smerge'}.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/09/08.]
#	[Minor revision on 2019/09/12]: Changed 'vsplit' to '(c)vsplit', and similarly changed the printed-out strings.
#
#	... OMITTED ...
#
# [ INPUTS: (\$outdir, \@ofnames,
#	$flag_end_sbrtn, $gb_in_2B_vsplit_old, $gb_in_2B_smgd_old, $gb_out_unmgd_new, $gb_out_kpt_new, 
#	\%gb_old2incr_coord_1way_dsplc_bf_vspl,
#	$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, \%gb_intm2incr_coord_undo_af_vspl,
#	$gb_ch_2B_smgd_intm, $gb_in1_4smg_intm, $gb_in2_4smg_intm, $gb_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, 
#	 \@new_indices_old_gbs, \@old_indices_new_gbs, \@intm_indices_old_gbs, \@new_indices_intm_gbs,
#	\%gb_new2incr_coord_undo_af_smg), where ] ... for the predecessor, 'print_out_info_chngs_by_vsplit_plus_smerge'.
#
# INPUTS: (\$outdir, \@ofnames,
#	$flag_ssvm, $gb_sbj_L_old, $gb_sbj_S_old, $gb_sbj1_unvmgd_NEW, $gb_mgd_NEW,
#	$gb_sbj1_left_intm, $gb_sbj1_right_intm,
#	\%gb_intm2incr_bl_coord_undo_ovlps_af_sspl,
#	$if_vmerged_w_lfrag, $if_shift_to_right, $gb_sbj_fxd_intm, $gb_sbj_shftd_intm, $incr_coord_sbj_shftd_B4_vmg,
#	\%gb_new2incr_coord_1way_dsplc_B4_vmg,
#	\@new_indices_old_gbs, \@old_indices_new_gbs, \@intm_indices_old_gbs, \@new_indices_intm_gbs,
#	\%gb_new2incr_coord_undo_af_vmg), where
#
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 7 output files;
#
#	The other arguments are a part of the outputs of 'ssplit_plus_vmerge_gblock_pair' in "MyANEX_Main.pm";
#	More precisely,
#
#	$flag_end_sbrtn = 1 if the subroutine 'ssplit_plus_vmerge_gblock_pair' ended without any remarkable problems,
#	          	= 0 if it failed;
#			= -1 if it failed BECAUSE the horizontal overlaps resulting from either 'ssplit' or 'vmerge' are NOT UN-DONE completely;
#
#	$gb_sbj_L_old and $gb_sbj_S_old are the indices (in the OLD gap-block set) of the subject gap-blocks
#	that are effective siblings to each other;
#	$gb_sbj_L_old is to be 'ssplit', while $gb_sbj_S_old will NOT be split;
#
#	$gb_sbj1_unvmgd_NEW is the NEW index of the fragment of $gb_sbj_L_old that was left UN-vertical-merged;
#	$gb_mgd_NEW is the NEW index of the gap-block newly created by the "vertical merge" of (a fragment of) $gb_sbj_L_old and (the entire) $gb_sbj_S_old;
#	[These are two of the MAIN-AUXILIARY outputs of 'ssplit_plus_vmerge_gblock_pair'];
#
#	$gb_sbj1_left_intm and $gb_sbj1_right_intm are the indices (in the INTERMEDIATE gap-block set) of the left- and right-fragments that resulted from the 'ssplit_gblock' of $gb_sbj_L_old;
#
#	%gb_intm2incr_bl_coord_undo_ovlps_af_sspl = ($indx_gb_intm => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "(horizontal) same-type split"}, ...);
#	(if $flag_end_sbrtn = -1, these 'UNDO'ing process is NOT actually executed);
#	[These are the AUXILIARY outputs of 'ssplit_gblock'];
#
#	$if_vmerged_w_lfrag = 1 if $gb_sbj_S_old was 'vmerge'd with $gb_sbj1_left_intm, = 0 if 'vmerge'd with $gb_sbj_right_intm;
#	$if_shift_to_right = 1 if a (serial) shift was performed to the right when effectively 'vmerge' the two gap-blocks, = 0 if performed to the left;
#	$gb_sbj_fxd_intm is the index (in the INTERMEDIATE gap-block set) of the gap-block to be 'vmerge'd whose coordinate was fixed;
#	$gb_sbj_shftd_intm if the index (in the INTERMEDIATE gap-block set) of the gap-block to be 'vmerge'd which was "shift"ed when effectively performing the 'vmerge';
#	[These were created in preparation of 'vmerge_gblock_pair'];
#
#	$incr_coord_sbj_shftd_B4_vmg is the increment of the coordinate of $gb_sbj_shftd_intm
#	necessary for its 'vmerge' with $gb_sbj_fxd_intm;
#
#	\%gb_new2incr_coord_1way_dsplc_B4_vmg = ($indx_gb_new => {increment of the block coordinate resulting from the DISPLACEMENT prior to the "complementary vertical split"}, ...) for the gap-blocks whose displacement canNOT be undone; (Such gap-blocks MUST NOT vertically include $gb_mgd_NEW); $indx_gb_new SHOULD BE the index in the NEW gap-block set;
#	[These are the AUXILIARY outputs of 'vmerge_gblock_pair'];
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "ssplit + vmerge";
#	More precisely, the gap-block that had been the $ogb th before the "ssplit + vmerge"
#	has been assigned $new_indices_old_gbs[$ogb] after the "ssplit + vmerge";
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "ssplit + vmerge";
#	More precisely, the gap-block that is the $ngb th after the "ssplit + vmerge"
#	had been assigned $old_indices_new_gbs[$ngb] before the "ssplit + vmerge";
#
#	@intm_indices_old_gbs stores the intermediate indices assigned to the gap-blocks after the "ssplit" and before the "vmerge";
#	More precisely, the gap-block that had been the $ogb th before the "ssplit"
#	has been assigned $intm_indices_old_gbs[$ogb] after the "ssplit";

#	@new_indices_intm_gbs stores the new indices assigned to the gap-blocks after the "vmerge";
#	More precisely, the gap-block that had been the $igb th before the "vmerge"
#	has been assigned $new_indices_intm_gbs[$igb] after the "vmerge";
#
#	%gb_new2incr_coord_undo_af_vmg = ($indx_gb_new => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "vertical-merge"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_coord_undo_af_vmg is one of the main outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' above];
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
#
# [ sub print_out_info_chngs_by_vsplit_plus_smerge (\$\@$$$$$\%$$\%$$$$$$\@\@\@\@\%) {
#
#	my ($rf_outdir, $ofnames, 
#	$flag_end_sbrtn, $gb_in_2B_vsplit_old, $gb_in_2B_smgd_old, $gb_out_unmgd_new, $gb_out_kpt_new, 
#	$gb_old2incr_coord_1way_dsplc_bf_vspl,
#	$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $gb_intm2incr_coord_undo_af_vspl,
#	$gb_ch_2B_smgd_intm, $gb_in1_4smg_intm, $gb_in2_4smg_intm, $gb_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, 
#	 $new_indices_old_gbs, $old_indices_new_gbs, $intm_indices_old_gbs, $new_indices_intm_gbs,
#	$gb_new2incr_coord_undo_af_smg) = @_; ] ... for the predecessor, 'print_out_info_chngs_by_vsplit_plus_smerge'.
#
sub print_out_info_chngs_by_ssplit_plus_vmerge (\$\@$$$$$$$\%$$$$$\%\@\@\@\@\%) {

	my ($rf_outdir, $ofnames, 
	    $flag_end_sbrtn, $gb_sbj_L_old, $gb_sbj_S_old, $gb_sbj1_unvmgd_NEW, $gb_mgd_NEW,
	    $gb_sbj1_left_intm, $gb_sbj1_right_intm,
	    $gb_intm2incr_bl_coord_undo_ovlps_af_sspl,
	    $if_vmerged_w_lfrag, $if_shift_to_right, $gb_sbj_fxd_intm, $gb_sbj_shftd_intm, $incr_coord_sbj_shftd_B4_vmg,
	    $gb_new2incr_coord_1way_dsplc_B4_vmg,
	    $new_indices_old_gbs, $old_indices_new_gbs, $intm_indices_old_gbs, $new_indices_intm_gbs,
	    $gb_new2incr_coord_undo_af_vmg) = @_;


	unless (@{$ofnames} == 7) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_info_chngs_by_ssplit_plus_vmerge'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 7 as desired: $! !\n\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_ssplit_plus_vmerge'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_info_chngs_by_ssplit_plus_vmerge'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $ofname1 = $ofnames->[0];
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_ssplit_plus_vmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	print $fh_out1 "{\$flag_end for 'ssplit_plus_vmerge_gblock_pair'} = $flag_end_sbrtn .\n";
	print $fh_out1 "\n";

	print $fh_out1 "Index{Input gap-block to be \"ssplit\"}_OLD = $gb_sbj_L_old ,\n";
	print $fh_out1 "Index{Input gap-block to be \"vmerge\"d}_OLD = $gb_sbj_S_old .\n\n";

	print $fh_out1 "Index{Output gap-block that was left UN-vmerged}_NEW  = $gb_sbj1_unvmgd_NEW ,\n";
	print $fh_out1 "Index{Output gap-block that resulted from the \"vertical-merge\"}_NEW  = $gb_mgd_NEW .\n\n";

	print $fh_out1 "\n";

	close ($fh_out1);

	##  $gb_sbj1_left_intm, $gb_sbj1_right_intm,



		# IMPORTED (w/ modifications) from 'print_out_info_chngs_by_smerge' (on 2019/08/25). #
		# => MODIFIED & Moved from $fh_out3 to $fh_out2, in creation of 'print_out_info_chngs_by_ssplit_plus_vmerge' (on 2019/10/17). #

	my $ofname2 = $ofnames->[1];
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_ssplit_plus_vmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}

	print $fh_out2 "Index{Output gap-block of \"ssplit\" on \$gb_sbj_L_old(=$gb_sbj_L_old) that is on the Left}_INTM = $gb_sbj1_left_intm .\n\n";
	print $fh_out2 "Index{Output gap-block of \"ssplit\" on \$gb_sbj_L_old(=$gb_sbj_L_old) that is on the Right}_INTM = $gb_sbj1_right_intm .\n\n";

=For the predecessor, 'print_out_info_chngs_by_vsplit_plus_smerge'.
	##my $ct_rmvd_null_clms = 0; # Always set this value to be zero. # ADDED in creation of 'print_out_info_chngs_by_smerge' (on 2019/07/22). #

	print $fh_out2 "{Gap-block \"shift\"ed after '(c)vsplit'}_INTM = $gb_shftd_af_vspl_intm .\n";
	print $fh_out2 "{Increment of the block coordinate due to the \"shift\" after '(c)vsplit'} = $incr_coord_shftd_af_vspl .\n\n";

	##print $fh_out2 "#{Resulting null columns, which were removed during the \"merge\"} = $ct_rmvd_null_clms .\n\n";
=cut # END of "For the predecessor, 'print_out_info_chngs_by_vsplit_plus_smerge'."

	close ($fh_out2);

		# End of "IMPORTED (w/ modifications) from 'print_out_info_chngs_by_smerge' (on 2019/08/25). #
		# => MODIFIED & Moved from $fh_out3 to $fh_out2, in creation of 'print_out_info_chngs_by_ssplit_plus_vmerge' (on 2019/10/17)." #


		# COPIED (w/ modifications) from below, in creation of 'print_out_info_chngs_by_vsplit_plus_smerge' (on 2019/09/08). #
		# => MODIFIED & Moved from $fh_out5 to $fh_out3, in creation of 'print_out_info_chngs_by_ssplit_plus_vmerge' (on 2019/10/17). #

	my $ofname3 = $ofnames->[2];
	my $outfile3 = "$outdir/$ofname3";

	my $fh_out3 = open_output_file ($outfile3, 0);
	unless (defined $fh_out3) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_ssplit_plus_vmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile3\": $! !\n";
		return 0;
	}

	my $ct_rlv_gbs_intm = scalar (keys %{$gb_intm2incr_bl_coord_undo_ovlps_af_sspl});

	print $fh_out3 "# Increments of the Coordinates of the Relevant Gap-Blocks When \"UNDO\"ing the Overlaps among them AFTER the 'ssplit'.\n";
	print $fh_out3 "#  #{Relevant Gap-Blocks} = $ct_rlv_gbs_intm .\n";
	print $fh_out3 "\n";

	print $fh_out3 join ("\t", 'Indx_gb_INTM', 'incr_coord_gb'), "\n\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb_intm2incr_bl_coord_undo_ovlps_af_sspl}) {
	    print $fh_out3 join ("\t", $gb, $gb_intm2incr_bl_coord_undo_ovlps_af_sspl->{$gb}), "\n";
	}
	print $fh_out3 "\n";

	close ($fh_out3);

		# End of "COPIED (w/ modifications) from below, in creation of 'print_out_info_chngs_by_vsplit_plus_smerge' (on 2019/09/08). #
		# => MODIFIED & Moved from $fh_out5 to $fh_out3, in creation of 'print_out_info_chngs_by_ssplit_plus_vmerge' (on 2019/10/17)." #


		# ADDED in creation of 'print_out_info_chngs_by_vsplit_plus_smerge' (on 2019/09/08). #
		# => MODIFIED & Moved from $fh_out6 to $fh_out5 in creation of 'print_out_info_chngs_by_ssplit_plus_vmerge' (on 2019/10/17). #
	my $ofname5 = $ofnames->[3];
	my $outfile5 = "$outdir/$ofname5";

	my $fh_out5 = open_output_file ($outfile5, 0);
	unless (defined $fh_out5) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_ssplit_plus_vmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile5\": $! !\n";
		return 0;
	}

	print $fh_out5 "\$if_vmerged_w_lfrag = $if_vmerged_w_lfrag ,\n\n";
	print $fh_out5 "\$if_shift_to_right   = $if_shift_to_right .\n\n";

	print $fh_out5 "Index{Input gap-block for \"vmerge\" whose (horizontal) position was fixed}_INTM = $gb_sbj_fxd_intm .\n\n";
	print $fh_out5 "Index{Input gap-block for \"vmerge\" that was \"serially shift\"ed}_INTM  = $gb_sbj_shftd_intm ,\n";
	print $fh_out5 "{Increment of its coordinate by the \"serial shift\" before the \"vmerge\"} = $incr_coord_sbj_shftd_B4_vmg .\n\n";
	
=For the predecessor, 'print_out_info_chngs_by_vsplit_plus_smerge'.
	print $fh_out5 "Index{Gap-block product of 'vsplit' to be \"smerge\"d}_INTM = $gb_ch_2B_smgd_intm .\n\n";
	print $fh_out5 "Index{1st Input gap-block for \"smerge\"}_INTM = $gb_in1_4smg_intm ,\n";
	print $fh_out5 "Index{2nd Input gap-block for \"smerge\"}_INTM = $gb_in2_4smg_intm .\n\n";

	print $fh_out5 "Index{Input gap-block for \"smerge\" that was \"serially shift\"ed}_INTM  = $gb_sbj_ssh4smg_intm ,\n";
	print $fh_out5  "{Increment of its coordinate by the \"serial shift\"} = $incr_coord_sbj_ssh4smg .\n\n";
	print $fh_out5 "Index{Input gap-block for \"smerge\" that was \"absorb\"ed}_INTM = $gb_absbd_intm .\n\n";
=cut # END of "For the predecessor, 'print_out_info_chngs_by_vsplit_plus_smerge'."

	print $fh_out5 "\n";

	close ($fh_out5);
		# END of "ADDED in creation of 'print_out_info_chngs_by_vsplit_plus_smerge' (on 2019/09/08). #
		#  => MODIFIED & Moved from $fh_out6 to $fh_out5 in creation of 'print_out_info_chngs_by_ssplit_plus_vmerge' (on 2019/10/17)." #


		# MOVED (w/ modifications) from $fh_out2 in the predecessor, 'print_out_info_chngs_by_vsplit_plus_smerge', in creation of 'print_out_info_chngs_by_ssplit_plus_vmerge' (on 2019/10/17). #

	my $ofname6 = $ofnames->[4];
	my $outfile6 = "$outdir/$ofname6";

	my $fh_out6 = open_output_file ($outfile6, 0);
	unless (defined $fh_out6) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_ssplit_plus_vmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile6\": $! !\n";
		return 0;
	}
	print $fh_out6 "# Increments of the Coordinates of the Relevant Gap-Blocks (w/ old indices) When \"DISPLAC\"ing them  before \"Vertically Merg\"ting \$gb_sbj_fxd_intm (=$gb_sbj_fxd_intm) & \$gb_sbj_shftd_intm (=$gb_sbj_shftd_intm) .\n";

	my $gb2incr_coord_1way_dsplc = $gb_new2incr_coord_1way_dsplc_B4_vmg;

     if (defined $gb2incr_coord_1way_dsplc->{'N.A'}) {

	    print $fh_out6 "\n ... This information is NOT AVAILABLE, even in principle, for This Session ...\n\n";
     } else {


	my $ct_rlv_gbs = scalar (keys %{$gb2incr_coord_1way_dsplc});

	print $fh_out6 "#  #{Relevant Gap-Blocks} = $ct_rlv_gbs .\n";
	print $fh_out6 "\n";

	print $fh_out6 join ("\t", 'Indx_gb_new', 'incr_coord_gb'), "\n\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_1way_dsplc}) {
	    print $fh_out6 join ("\t", $gb, $gb2incr_coord_1way_dsplc->{$gb}), "\n";
	}
	print $fh_out6 "\n";

     } # End of "if (defined $gb2incr_coord_1way_dsplc->{'N.A'}) {...} else {...}"

	close ($fh_out6);
		# END of "MOVED (w/ modifications) from $fh_out2 in the predecessor, 'print_out_info_chngs_by_vsplit_plus_smerge', in creation of 'print_out_info_chngs_by_ssplit_plus_vmerge' (on 2019/10/17)." #


		# IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14). #
	my $ofname7 = $ofnames->[5];
	my $outfile7 = "$outdir/$ofname7";

	my $fh_out7 = open_output_file ($outfile7, 0);
	unless (defined $fh_out7) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_ssplit_plus_vmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile7\": $! !\n";
		return 0;
	}

	my $ct_old_gbs = @{$new_indices_old_gbs};
	my $ct_intm_gbs = @{$new_indices_intm_gbs};
        my $ct_new_gbs = @{$old_indices_new_gbs};

        print $fh_out7 "## Relationships between new, intermediate, and old indices of gap-blocks\n";
	print $fh_out7 "##  (#{gap-blocks}_old=$ct_old_gbs, #{gap-blocks}_intm=$ct_intm_gbs, #{gap-blocks}_new=$ct_new_gbs) :\n\n";

        my @out_new_indices_gbs = my @out_old_indices_gbs = ();
        foreach my $indx (@{$new_indices_old_gbs}) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$old_indices_new_gbs}) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

	my @out_intm_indices_old_gbs = my @out_new_indices_intm_gbs = ();
	foreach my $indx (@{$intm_indices_old_gbs}) { push @out_intm_indices_old_gbs, ((defined $indx) ? $indx : 'undef'); }
        foreach my $indx (@{$new_indices_intm_gbs}) { push @out_new_indices_intm_gbs, ((defined $indx) ? $indx : 'undef'); }

        print $fh_out7 '@{$new_indices_old_gbs} = (', join (', ', @out_new_indices_gbs), ") .\n";
        print $fh_out7 '@{$old_indices_new_gbs} = (', join (', ', @out_old_indices_gbs), ") .\n\n";

        print $fh_out7 '@{$intm_indices_old_gbs} = (', join (', ', @out_intm_indices_old_gbs), ") .\n";
        print $fh_out7 '@{$new_indices_intm_gbs} = (', join (', ', @out_new_indices_intm_gbs), ") .\n";
        print $fh_out7 "\n\n";

	close ($fh_out7);
		# END of "IMPORTED from 'print_out_info_chngs_by_smerge' (on 2019/08/14)." #


	my $ofname8 = $ofnames->[6];
	my $outfile8 = "$outdir/$ofname8";

	my $fh_out8 = open_output_file ($outfile8, 0);
	unless (defined $fh_out8) {
		print STDOUT "[ERROR 'print_out_info_chngs_by_ssplit_plus_vmerge'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile8\": $! !\n";
		return 0;
	}

	# $gb_new2incr_coord_undo_af_smg) = @_;

	my $ct_rlv_gbs_new = scalar (keys %{$gb_new2incr_coord_undo_af_vmg});

	print $fh_out8 "# Increments of the Coordinates of the Relevant Gap-Blocks When \"UNDO\"ing the Overlaps among them AFTER the 'vmerge'.\n";
	print $fh_out8 "#  #{Relevant Gap-Blocks} = $ct_rlv_gbs_new .\n";
	print $fh_out8 "\n";

	print $fh_out8 join ("\t", 'Indx_gb_NEW', 'incr_coord_gb'), "\n\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb_new2incr_coord_undo_af_vmg}) {
	    print $fh_out8 join ("\t", $gb, $gb_new2incr_coord_undo_af_vmg->{$gb}), "\n";
	}
	print $fh_out8 "\n";

	close ($fh_out8);

	return 1;

}	# End of the subroutine, 'print_out_info_chngs_by_ssplit_plus_vmerge'. #


# print_out_sbsdry_outputs_rev_spec_cii_in_wd:
#	This subroutine prints out the subsidiary outputs of 'reverse_spec_cii_in_wd' (in "MyANEX_MainX.pm").
#	It was created by taking advantage of the framework of 'print_out_info_chngs_by_vsplit_plus_smerge' above,
#	and importing a few blocks from 'test_reverse_spec_cii_in_wd.pl'.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/10/01.]
#
# print_out_info_chngs_by_vsplit_plus_smerge:
#	This subroutine prints out the portions of the outputs of '(c)vsplit_plus_smerge_gblock_pair' (in "MyANEX_Main.pm")
#	that summarizes the changes resulting from the \"vertical split + merge (same-type)\" and the \"shift\"s accompanying it.
#	It was created by modifying 'print_out_info_chngs_by_vsplit'
#	and importing a portion from {'print_out_info_chngs_by_smerge'}.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/09/08.]
#	[Minor revision on 2019/09/12]: Changed 'vsplit' to '(c)vsplit', and similarly changed the printed-out strings.
#
# INPUTS: (\$outdir, \@ofnames,
#	# [ $flag_revcii, ] ... REMOVED (on 2019/10/01). #
#	\@new_set_cls_gpbs, \@sets_old_indices_new_gpbs, \@new_sets_gpatterns), where
#
#	$outdir is a string specifying the path to the output directory;
#	@ofnames stores the names of the 2 output files;
#
#	[[
#	$flag_revcii = 1 if the subroutine 'reverse_spec_cii_in_wd' ended without any particular problems,
#		  = 0 if it failed;
#	]] ... REMOVED (on 2019/10/01).
#
#	The other arguments, that is,
#	\@new_set_cls_gpbs, \@sets_old_indices_new_gpbs (=\@set_gpbs_at_spec_places), and \@new_sets_gpatterns,
#	are subsidiary outputs of 'reverse_spec_cii_in_wd';
#	Specifically,
#
#	@{$new_set_cls_gpbs[$gs]->[$ngpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $ngpb th gap-pattern block in the $gs th cluster (i.e., gapped segment) in the NEW input MSA (AFTER the brief-parsimonization), with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gpb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);
#
#	@{$sets_old_indices_new_gpbs[$gs]} is @old_indices_new_gpbs (=@gpbs_at_spec_places), which is one of the outputs of 'br_parsimonize_dollo_indel_histories' (in "MyMyTreeMap_indels_spt_odr2.pm") for the $gs th gapped segment, where
#	$gpbs_at_spec_places[$pl] is the OLD index of the gap-pattern block that is in the $pl th place AFTER the brief-parsimonization;
#	(In other words, $old_indices_new_gpbs[$ngpb] is the OLD index of the gap-pattern block that is the $ngpbs block AFTER the brief-parsimonization);
#
#	${$new_sets_gpatterns[$indx_gp]} gives a string representing the $indx_gp th gap-pattern that occurred when constructing gapped-segments;
#
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#	 
#
# [ sub print_out_sbsdry_outputs_rev_spec_cii_in_wd (\$\@$\@\@\@) { ] ... OBSOLETE as of 2019/10/01. #
#
sub print_out_sbsdry_outputs_rev_spec_cii_in_wd (\$\@\@\@\@) {

	my ($rf_outdir, $ofnames, 
	# [ $flag_revcii, ] ... REMOVED (on 2019/10/01). #
	$new_set_cls_gpbs, $sets_old_indices_new_gpbs, $new_set_gpatterns) = @_;


	unless (@{$ofnames} == 2) {
	    my $ct_ofnames = @{$ofnames};
	    print STDOUT "[ERROR 'print_out_sbsdry_outputs_rev_spec_cii_in_wd'] The ", '@{$ofnames} = (', join (', ', @{$ofnames}), ") has $ct_ofnames elements, NOT 2 as desired: $! !\n\n";
	    return 0;
	}

	my $ct_new_gsegs = @{$new_set_cls_gpbs};
	my $ct_new_gsegs2 = @{$sets_old_indices_new_gpbs};

	unless ($ct_new_gsegs == $ct_new_gsegs2) {
	    print STDOUT "[ERROR 'print_out_sbsdry_outputs_rev_spec_cii_in_wd'] The count of NEW gapped segments in ", '@{$new_set_cls_gpbs}', " (=$ct_new_gsegs) CLASHES with that in ", '@{$sets_old_indices_new_gpbs}', " (=$ct_new_gsegs2): $! !\n";
	    return 0;
	}

	my $totct_new_gpbs = ();
	foreach my $gseg (@{$new_set_cls_gpbs}) { $totct_new_gpbs += @{$gseg}; }
	my $totct_new_gpbs2 = 0;
	foreach my $old_indices_new_gpbs (@{$sets_old_indices_new_gpbs}) { $totct_new_gpbs2 += @{$old_indices_new_gpbs}; }

	unless ($totct_new_gpbs == $totct_new_gpbs2) {
	    print STDOUT "[ERROR 'print_out_sbsdry_outputs_rev_spec_cii_in_wd'] The total count of NEW gap-pattern blocks in ", '@{$new_set_cls_gpbs}', " (=$totct_new_gpbs) CLASHES with that in ", '@{$sets_old_indices_new_gpbs}', " (=$totct_new_gpbs2): $! !\n";
	    return 0;
	}

	my $outdir = ${$rf_outdir};
	unless ( -d $outdir )  {
	  unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR 'print_out_sbsdry_outputs_rev_spec_cii_in_wd'] FAILED to create the directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir ) {
	  unless (0 == system (" chmod -R u+w $outdir " ) ) {
	    print STDOUT "[ERROR 'print_out_sbsdry_outputs_rev_spec_cii_in_wd'] FAILED to change the permission mode of the output directory \"$outdir\": $! !\n\n";
	    return 0;
	  }
	}

	my $ofname1 = $ofnames->[0];
	my $outfile1 = "$outdir/$ofname1";

	my $fh_out1 = open_output_file ($outfile1, 0);
	unless (defined $fh_out1) {
		print STDOUT "[ERROR 'print_out_sbsdry_outputs_rev_spec_cii_in_wd'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile1\": $! !\n";
		return 0;
	}

	print $fh_out1 "# #{gapped-segments}_NEW = $ct_new_gsegs .\n";
	print $fh_out1 "# Total#{gap-pattern-blocks}_NEW = $totct_new_gpbs .\n\n";

	print $fh_out1 '[ Contents of @{$new_set_cls_gpbs}: ]', "\n\n";


	print $fh_out1 join ("\t", 'Indx_gseg', 'indx_gpb_in_gseg', 'indx_gpattern', 'len_gpb', 'le_gpb', 're_gpb', "'gap-pattern'"), "\n\n";

	for (my $gs = 0; $gs < $ct_new_gsegs; $gs++) {

	    my $cl_gpbs = $new_set_cls_gpbs->[$gs];
	    my $ct_gpbs = @{$cl_gpbs};

	    for (my $gpb = 0; $gpb < $ct_gpbs; $gpb++) {
		my $indx_gp = $cl_gpbs->[$gpb][0];
		my $gpattern = ${$new_set_gpatterns->[$indx_gp]};
	        print $fh_out1 join ("\t", $gs, $gpb, @{$cl_gpbs->[$gpb]}, "'$gpattern'"), "\n";
	    }
	    print $fh_out1 "\n";
	}
	print $fh_out1 "\n";

	close ($fh_out1);


	my $ofname2 = $ofnames->[1];
	my $outfile2 = "$outdir/$ofname2";

	my $fh_out2 = open_output_file ($outfile2, 0);
	unless (defined $fh_out2) {
		print STDOUT "[ERROR 'print_out_sbsdry_outputs_rev_spec_cii_in_wd'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$outfile2\": $! !\n";
		return 0;
	}

	my $sets_gpbs_at_spec_places = $sets_old_indices_new_gpbs;

	print $fh_out2 '[ Contents of @{$sets_old_indices_new_gpbs} (=@{$sets_gpbs_at_spec_places}): ]', "\n\n";

	print $fh_out2 join ("\t", 'Indx_gseg', '#{gap-pattern-blocks}', '[old,indices,for,new,gap,pattern,blocks]'), "\n\n";
	##print $fh_out2 join ("\t", 'Indx_gseg', '#{gap-pattern-blocks}', '[gap-pattern-blocks,at,specific,places]'), "\n\n";

	for (my $gs = 0; $gs < $ct_new_gsegs; $gs++) {

	    my $gpbs_at_spec_places = $sets_gpbs_at_spec_places->[$gs];

	    my $ct_gpbs = @{$gpbs_at_spec_places};
	    my $cnct_gpbs_at_spls = join (',', @{$gpbs_at_spec_places});

	    print $fh_out2  join ("\t", $gs, $ct_gpbs, "[$cnct_gpbs_at_spls]"), "\n";
	}
	print $fh_out2 "\n\n";

	close ($fh_out2);

	return 1;

}	# End of the subroutine, 'print_out_sbsdry_outputs_rev_spec_cii_in_wd'. #


# perform_mshifts_and_print_inputs_outputs:
#	This subroutine just performs the 4 subroutines,
#	'print_in_nr_mshifts_glks', 'bch_print_out_other_gblk_attributes', 'nonrec_mlt_shifts_gapblocks' & 'print_out_nr_mshifts_gblks',
#	in tandem.
#	[CREATED in MAJOR REVISION of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/18.]
#
# INPUTS: ($session, \$dir_out0, \$dir_out, \@ofnames1, \@ofnames2, \@ofnames3,
#	    \@orgs_gb_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@nodeids_in_msa, $char_abs, $char_pres,
#	    \@set_columns0, \@list_gblocks, $indx_br, $indx_u_or_d, \@bds_gblocks0, \@rows_affd_by_blks,
#           \%clm2ln_prb_sbst, $topnode_id, \%node2ch, \%node2pa, \%br_attr, \%node2depth, \%node2ct_exoffs, \%br2eq, 
#	    \@root_resfreqs, \%node2set_sw_extcond_probs0, \%br2tr_mtrx, 
#           \@set_ingredients_indel_hsts, \@set_ingredients_prb_msa_indel, $if_on_le_msa, $if_on_re_msa,
#           \@list_excluded_bin_hrels,
#	    \@inter_block_relations, \@gblock_sizes, \@gblks_w_spec_lb0, \@gblks_w_spec_rb0,
#	    \@cts_vincluding, \@cts_vincluded, \@cts_onx, $string_undef, $beg_wd, $end_wd), where
#
#	$session is the name of the present session;
#
#	$dir_out0 is the string providing the path to the output directory (mainly for the "origin", or the input MSA and gap-block configurations);
#	$dir_out is the path to the output directory (mainly for the output of 'nonrec_mlt_shifts_gapblocks');
#	@ofnames1 stores the names of the 4 output files for 'print_in_nr_mshifts_glks';
#	@ofnames2 stores the names of the 4 output files for 'bch_print_out_other_gblk_attributes';
#	@ofnames3 stores the names of the 8 output files for 'print_out_nr_mshifts_gblks';
#
#	\@orgs_gb_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@nodeids_in_msa, $char_abs, $char_pres,
#	\@set_columns0, \@list_gblocks, $indx_br, $indx_u_or_d, \@bds_gblocks0, \@rows_affd_by_blks,
#       \%clm2ln_prb_sbst, $topnode_id, \%node2ch, \%node2pa, \%br_attr, \%node2depth, \%node2ct_exoffs, \%br2eq, 
#	\@root_resfreqs, \%node2set_sw_extcond_probs0, \%br2tr_mtrx, 
#       \@set_ingredients_indel_hsts, \@set_ingredients_prb_msa_indel, $if_on_le_msa, $if_on_re_msa,
#       \@list_excluded_bin_hrels
#	are the inputs of 'nonrec_mlt_shifts_gapblocks' (& 'print_in_nr_mshifts_glks');
#	[For details, see the header of 'nonrec_mlt_shifts_gapblocks' in "MyANEX_Main.pm"];
#
#	\@inter_block_relations, \@gblock_sizes, \@gblks_w_spec_lb0, \@gblks_w_spec_rb0,
#	\@cts_vincluding, \@cts_vincluded, \@cts_onx,
#	are the inputs of 'bch_print_out_other_gblk_attributes';
#	[For details, see the header of 'bch_print_out_other_gblk_attributes' in "MyANEX_IO.pm"];
#
#	$string_undef is a string representing an undefined variable;
#
#	$beg_wd and $end_wd are the positions of the columns (in the input MSA)
#	at the beginning and end, respectively, of the window;
#	
#
# OUTPUTS: ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, \@mdimarray_diffs_ln_prbs_plus), where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed;
#
#	The other arguments are the outputs of 'nonrec_mlt_shifts_gapblocks';
#	[For details, see the header of 'nonrec_mlt_shifts_gapblocks' in "MyANEX_Main.pm"];
#
#
##sub perform_mshifts_and_print_inputs_outputs ($\$\$\@\@\@\@\@\@\@$$\@\@$$\@\@\%\$\%\%\%\%\%\%\@\%\%\@\@$$\@\@\@\@\@\@\@\@$$$) {
sub perform_mshifts_and_print_inputs_outputs ($\$\$\@\@\@\@\@\@\@$$\@\@$$\@\@\%$\%\%\%\%\%\%\@\%\%\@\@$$\@\@\@\@\@\@\@\@$$$) {  # CORRECTED on 2020/05/08. #


 	my ($session, $rf_dir_out0, $rf_dir_out, $ofnames1, $ofnames2, $ofnames3,
	    $orgs_gb_coords, $bds_gb_coords, $gb_coords_wd_frm, $nodeids_in_msa, $char_abs, $char_pres,
	    $set_columns0, $list_gblocks, $indx_br, $indx_u_or_d, $bds_gblocks0, $rows_affd_by_blks,
           $clm2ln_prb_sbst, $topnode_id, $node2ch, $node2pa, $br_attr, $node2depth, $node2ct_exoffs, $br2eq, 
	    $root_resfreqs, $node2set_sw_extcond_probs0, $br2tr_mtrx, 
           $set_ingredients_indel_hsts, $set_ingredients_prb_msa_indel, $if_on_le_msa, $if_on_re_msa,
           $list_excluded_bin_hrels,
	    $inter_block_relations, $gblock_sizes, $gblks_w_spec_lb0, $gblks_w_spec_rb0,
	    $cts_vincluding, $cts_vincluded, $cts_onx, $string_undef, $beg_wd, $end_wd) = @_;

	unless (@{$ofnames1} == 4) {
	    my $ct = @{$ofnames1};
	    print STDOUT "[ERROR 'perform_mshifts_and_print_inputs_outputs'] There are $ct, NOT 4, elements in ", '@{$ofnames1} = (', join (', ', @{$ofnames1}), "): $! !\n\n";

	    return (0, undef);
	}
	unless (@{$ofnames2} == 4) {
	    my $ct = @{$ofnames2};
	    print STDOUT "[ERROR 'perform_mshifts_and_print_inputs_outputs'] There are $ct, NOT 4, elements in ", '@{$ofnames2} = (', join (', ', @{$ofnames2}), "): $! !\n\n";

	    return (0, undef);
	}
	unless (@{$ofnames3} == 8) {
	    my $ct = @{$ofnames3};
	    print STDOUT "[ERROR 'perform_mshifts_and_print_inputs_outputs'] There are $ct, NOT 8, elements in ", '@{$ofnames3} = (', join (', ', @{$ofnames3}), "): $! !\n\n";

	    return (0, undef);
	}

   (print STDOUT "<< Subroutine, 'perform_mshifts_and_print_inputs_outputs' (w/ \$session = '$session'). >>\n\n") if ($TEST_LEVEL_MAIN >0);

	my $dir_out0 = ${$rf_dir_out0};
	my $dir_out  = ${$rf_dir_out};

			# (i) Print out Input Data. #

	unless ( print_in_nr_mshifts_glks ($dir_out0, @{$ofnames1}, 
		@{$list_gblocks}, $indx_br, $indx_u_or_d, @{$bds_gblocks0}, @{$rows_affd_by_blks},
		@{$orgs_gb_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, @{$set_columns0},
		$beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, @{$list_excluded_bin_hrels})
		) {
	    print STDOUT "[ERROR 'perform_mshifts_and_print_inputs_outputs'] The subroutine 'print_in_nr_mshifts_gblks' FAILED for $session on the input MSA: $! !\n";

	    return (0, undef);
	}
   (print STDOUT " ... Subroutine, 'print_in_nr_mshifts_gblks', Ended for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);

			# ADDED on 2019/07/17. (MI-1) #

	unless ( bch_print_out_other_gblk_attributes ($dir_out0, @{$ofnames2}, 
		@{$inter_block_relations}, @{$gblock_sizes}, @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0},
		@{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx})
		) {
	    print STDOUT "[ERROR 'perform_mshifts_and_print_inputs_outputs'] The subroutine 'print_out_other_gblk_attributes' FAILED for $session on the initial gap-blocks: $! !\n";

	    return (0, undef);
	}
   ##(print STDOUT " ... Subroutine, 'print_out_other_gblk_attributes', Ended for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT " ... Subroutine, 'bch_print_out_other_gblk_attributes', Ended for $session ...\n\n") if ($TEST_LEVEL_MAIN >0); # Corrected on 2020/05/09.

			# END of "ADDED on 2019/07/17. (MI-1)" #


			# (ii) Perform the Multiple-"Shift"s. #

	my ($if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	= nonrec_mlt_shifts_gapblocks (
	    @{$orgs_gb_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, @{$nodeids_in_msa}, $char_abs, $char_pres,
	    @{$set_columns0}, @{$list_gblocks}, $indx_br, $indx_u_or_d, @{$bds_gblocks0}, @{$rows_affd_by_blks},
            %{$clm2ln_prb_sbst}, $topnode_id, %{$node2ch}, %{$node2pa}, %{$br_attr}, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @{$root_resfreqs}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @{$set_ingredients_indel_hsts}, @{$set_ingredients_prb_msa_indel}, $if_on_le_msa, $if_on_re_msa,
            @{$list_excluded_bin_hrels}
		);
	unless ($if_success) {
	    print STDOUT "[ERROR 'perform_mshifts_and_print_inputs_outputs'] The subroutine 'nonrec_mlt_shifts_gapblocks' FAILED for $session on the input MSA: $! !\n";

	    return (0, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus);
	}
   (print STDOUT " ... Subroutine, 'nonrec_mlt_shifts_gapblocks', Ended for $session on the input MSA ...\n\n") if ($TEST_LEVEL_MAIN >0);


			# (iii) Print out the Outputs of 'nonrec_mlt_shifts_gapblocks'. #

	unless ( print_out_nr_mshifts_gblks ($dir_out, @{$ofnames3}, 
		$ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0,
		@{$mdimarray_diffs_ln_prbs_plus}, @{$orgs_gb_coords}, @{$bds_gb_coords}, $string_undef)
		) {
	    print STDOUT "[ERROR 'perform_mshifts_and_print_inputs_outputs'] The subroutine 'print_out_nr_mshifts_gblks' FAILED for $session on the input MSA: $! !\n";

	    return (0, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus);
	}
   (print STDOUT " ... Subroutine, 'print_out_nr_mshifts_gblks', Ended for $session ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

   (print STDOUT "<< Subroutine 'perform_mshifts_and_print_inputs_outputs' ENDED for $session. />/>\n\n") if ($TEST_LEVEL_MAIN >0);

	return (1, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus);

}	# END of the subroutine, 'perform_mshifts_and_print_inputs_outputs'. #


	##### (III) DEFINITIONS of the Subroutines For Other Processes. #####

# select_sgn_wds_in_pcands:
#	This subroutine receives a list of merged purge-candidate regions (e.g., output by 'readin_list_mgd_pcands'),
#	and selects some representative significant windows (for detecting purge-candidates) within each merged region.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/20.]
#
#	[Selection Criteria]:
#	A (subject) window (w/ $pv_sbj & $pv0_sbj) will be kept if:
#	there are NO windows (w/ $pv_wd & $pv0_wd) (including the merged region)
#	that include the subject window
#	and that also satisfy ($pv_wd <= $pv_sbj) and ($pv0_wd >= $pv0_sbj).
#
#	(As a corollary, if the merged region equals a window,
#	the window will be removed from the list of representatives,
#	to avoid redundancy in later processes.)
#
# INPUT: (\@list_mgd_pcands), where
#
#	@{$list_mgd_pcands[$pc]} = ($beg_pc, $end_pc, $br, \@msa_rows_under_br, \@constituent_sites, $pv, $pv0, \@sgn_wds_included) provides information on the $pc th (merged) purge-candidate region, where
#	$beg_pc & $end_pc are the sites at the beginning & end, respectively, of the region,
#	$br is the ID of the branch across which the "purge" is suspected within the region,
#	@msa_rows_under_br stores the indices of the MSA rows corresponding to the external offsprings of $br,
#	@constituent_sites stores the sites that constitute the region,
#	$pv is the P-value measuring the deviation from the specified substutution model,
#	$pv0 is the P-value measuring the deviation from the random-match model,
#	and @sgn_wds_included is the significant windows included in the region;
#	more precisely, 
#	@{$sgn_wds_included[$swd]} = (\@sites_in_wd, $pv_wd, $pv0_wd) provies information on the $swd th significant window, where
#	@sites_in_wd stores the sites in the window,
#	$pv_wd and $pv0_wd are the P-values for the window, measuring the same types of deviations as the aforementioned $pv and $pv0, respectively;
#
#	[NOTE: We ASSUME that the windows in @sgn_wds_included are sorted in ascending order of window sizes (in terms of the number of constituent sites)];
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
#	NOTE:
#	As a "side-effect",
#	this subroutine reduces the \@sgn_wds_included in each (merge) purge-candidate region,
#	by keeping only the representative windows.
#
sub select_sgn_wds_in_pcands (\@) {

	my ($list_mgd_pcands) = @_; 

	my $ct_mgd_pcs = @{$list_mgd_pcands};

	for (my $pc=0; $pc <$ct_mgd_pcs; $pc++) { # Main for-loop (over purge-candidate regions). #

	    my $info_mgd_pc = $list_mgd_pcands->[$pc];
	    my ($beg_pc, $end_pc, $br, $msa_rows_under_br, $constituent_sites, $pv, $pv0, $sgn_wds_included) = @{$info_mgd_pc};

	    my @rep_wds = ();
	    for (my $swd = $#{$sgn_wds_included}; $swd >=0; $swd--) { # Middle for-loop (over significant windows included in the region). #
		my $info_sbj = $sgn_wds_included->[$swd];
		my ($sites_in_sbj, $pv_sbj, $pv0_sbj) = @{$info_sbj};
		if (($pv <= $pv_sbj) and ($pv0 >= $pv0_sbj)) { next; } # The merged region is "better than" the subject.
		my ($le_sbj, $re_sbj) = @{$sites_in_sbj}[0, $#{$sites_in_sbj}];
		my $if_rep = 1;
		foreach my $info_wd (@rep_wds) { # Inner for-loop (over representative windows identified thus far). #
		    my ($sites_in_wd, $pv_wd, $pv0_wd) = @{$info_wd};
		    unless (($pv_wd <= $pv_sbj) and ($pv0_wd >= $pv0_sbj)) { next; }

			# Now, the window is "better than" the subject.

		    my ($le_wd, $re_wd) = @{$sites_in_wd}[0, $#{$sites_in_wd}];

		    unless (($le_wd <= $le_sbj) and ($re_sbj <= $re_wd)) { next; }

			# Now, the window includes the subject, at least interval-wise.

		    my ($intsc, $ext_sbj, $ext_wd) = diff_nmarray (@{$sites_in_sbj}, @{$sites_in_wd});

		    if (@{$ext_sbj}>0) { next; }

			# Now, the window COMPLETELY includes the subject.
			# => The subject should NOT be a representative.

		    $if_rep = 0;
		    last;

		}

		if ($if_rep) { unshift @rep_wds, $info_sbj; }
		
	    } # End of the middle for-loop (over siginificant windows included in the region). #

	    $info_mgd_pc->[7] = \@rep_wds;

	} # End of the main for-loop (over purge-candidate regions). #

	return 1;

}	# End of the subroutine 'select_sgn_wds_in_pcands'. #


# chk_ovlp_pcands_vs_sliding_wds:
#	This subroutine examines which sliding sub-windows overlap each (merged) purge-candidate regions.
#	For those purge-candidate regions NOT overlapping any sub-windows,
#	the subroutine examines whether the regions overlap any windows,
#	and if not again, whether they overlap any window-frames.
#	[CREATED in creation of 'anex.ver0.5.pl'. on 2019/06/20.]
#
#	NOTE: We ASSUME
#	  that the neihbboring window frames do NOT overlap each other,
#	  and that the neighboring windows (or sub-windows) can overlap, but NEVER be nested,
#	  with each other.
#
# INPUTS: (\@list_mgd_pcands, $indx_beg_pc, $indx_end_pc, $maxspan_mgd_pcand, \@window_frames, $indx_beg_wf, $indx_end_wf, \@set_windows, \@set_subwindows, $indx_beg_wd, $indx_end_wd), where
#
#	@{$list_mgd_pcands[$pc]}[$indx_beg_pc, $indx_end_pc] = ($beg_pc, $end_pc) are the sites at the beginning and end of the $pc th (merged) purge-candidate region;
#
#	$maxspan_mgd_pcand is the maximum among the spans of the merged purge-candidate regions;
#
#	[Typically, \@list_mgd_pcands & $maxspan_mgd_pcands are outputs of 'readin_list_mgd_pcands'];
#
#	
#	@{$window_frames[$fr]}[$indx_beg_wf, $indx_end_wf] = ($start_fr, $end_fr)
#	defines the $fr th frame for the window analysis,
#	that is, the frame starts at the $start_fr th column and ends at the $end_fr th column.
#	(Both are included in the frame.);
#
#	@{$set_windows[$wd]}[$indx_beg_wd, $indx_end_wd] = ($beg_wd, $end_wd) are the sites (or columns) at the beginning and end of the $wd th sliding window;
#	
#	@{$set_subwindows[$swd]}[$indx_beg_wd, $indx_end_wd] = ($beg_swd, $end_swd) are the sites (or columns) at the beginning and end of the $swd th sliding sub-window;
#
#	[Typically, \@window_frames, \@set_windows, and \@set_subwindows are outputs of 'prepare_sliding_windows'];
#
# OUTPUTS: (\@list_mgd_pcands_ovlp_sbwds, \@off_sbwds_but_ovlp_wds, \@off_wds_but_ovlp_wframes, @off_wframes), where
#
#	@{$list_mgd_pcands_ovlp_sbwds[$swd]} stores the indices of the merged purge-candidate regions overlapping the $swd th sub-window;
#
#	@{$off_sbwds_but_ovlp_wds[$wd}} stores the indices of the regions overlapping the $wd th window but NOT overlapping any sub-windows;
#
#	@{$off_wds_but_ovlp_wframes[$wf]} stores the indices of the regions overlapping the $wf th window-frame but NOT overlapping any windows;
#
#	@off_wframes stores the indices of the regions NOT overlapping any window-frames;
#
sub chk_ovlp_pcands_vs_sliding_wds (\@$$$\@$$\@\@$$) {

	my ($list_mgd_pcands, $indx_beg_pc, $indx_end_pc, $maxspan_mgd_pcand,
	$window_frames, $indx_beg_wf, $indx_end_wf,
	$set_windows, $set_subwindows, $indx_beg_wd, $indx_end_wd) = @_;

	my $ct_wfs = @{$window_frames};
	my $ct_wds = @{$set_windows};
	my $ct_swds = @{$set_subwindows};

	my @indices_wfs = (0 .. $ct_wfs -1);
	my @indices_wds = (0 .. $ct_wds -1);
	my @indices_swds = (0 .. $ct_swds-1);

		# Initialize the outputs. #
	my @list_mgd_pcands_ovlp_sbwds = my @off_sbwds_but_ovlp_wds = ();
	my @off_wds_but_ovlp_wframes = my @off_wframes = ();

	for (my $swd=0; $swd <$ct_swds; $swd++) { $list_mgd_pcands_ovlp_sbwds[$swd] = []; }
	for (my $wd=0; $wd <$ct_wds; $wd++) { $off_sbwds_but_ovlp_wds[$wd] = []; }
	for (my $wf=0; $wf <$ct_wfs; $wf++) { $off_wds_but_ovlp_wframes[$wf] = []; }

	my $ct_mgd_pcands = @{$list_mgd_pcands};
	for (my $pc = 0; $pc < $ct_mgd_pcands; $pc++) { # Main for-loop (over (merged) purge-candidates). #

	    my ($beg_pc, $end_pc) = @{$list_mgd_pcands->[$pc]}[$indx_beg_pc, $indx_end_pc];

			# Examine the overlapping sub-windows. #

	    my ($fl_exist_swd, $indx2_swd, $indx_swd) = binsearch_num ($beg_pc, @{$set_subwindows}, $indx_end_wd, @indices_swds);
			# $indx_swd is the index of the sub-window that is left-neighboring the purge-candidate. #
	    unless (defined $fl_exist_swd) {
		print STDOUT "[ERROR 'chk_ovlp_pcands_vs_sliding_wds'] The subroutine 'binsearch_num' FAILED on the $pc th purge-candidate (w/ \$beg_pc = $beg_pc) against the set of sub-windows: $! !\n\n";
		return (undef, undef, undef, undef);
	    }
	    my $ct_ovlp_swds =0;
	    if ($fl_exist_swd) {
		$ct_ovlp_swds++;
		push @{$list_mgd_pcands_ovlp_sbwds[$indx_swd]}, $pc;
	    }

	    for (my $swd = $indx_swd+1; $swd < $ct_swds; $swd++) {
		my $beg_swd = $set_subwindows->[$swd][$indx_beg_wd];
		if ($end_pc < $beg_swd) { last; } # The sub-window is already on the right of the purge-candidate.
		$ct_ovlp_swds++;
		push @{$list_mgd_pcands_ovlp_sbwds[$swd]}, $pc;
	    }
	    if ($ct_ovlp_swds>0) { next; }


			# Examine the overlapping windows. #

	    my ($fl_exist_wd, $indx2_wd, $indx_wd) = binsearch_num ($beg_pc, @{$set_windows}, $indx_end_wd, @indices_wds);
			# $indx_wd is the index of the window that is left-neighboring the purge-candidate. #
	    unless (defined $fl_exist_wd) {
		print STDOUT "[ERROR 'chk_ovlp_pcands_vs_sliding_wds'] The subroutine 'binsearch_num' FAILED on the $pc th purge-candidate (w/ \$beg_pc = $beg_pc) against the set of windows: $! !\n\n";
		return (undef, undef, undef, undef);
	    }
	    my $ct_ovlp_wds =0;
	    if ($fl_exist_wd) {
		$ct_ovlp_wds++;
		push @{$off_sbwds_but_ovlp_wds[$indx_wd]}, $pc;
	    }

	    for (my $wd = $indx_wd+1; $wd < $ct_wds; $wd++) {
		my $beg_wd = $set_windows->[$wd][$indx_beg_wd];
		if ($end_pc < $beg_wd) { last; } # The window is already on the right of the purge-candidate.
		$ct_ovlp_wds++;
		push @{$off_sbwds_but_ovlp_wds[$wd]}, $pc;
	    }
	    if ($ct_ovlp_wds>0) { next; }

			# Examine the overlapping widow-frames. #

	    my ($fl_exist_wf, $indx2_wf, $indx_wf) = binsearch_num ($beg_pc, @{$window_frames}, $indx_end_wf, @indices_wfs);
			# $indx_wf is the index of the window-frame that is left-neighboring the purge-candidate. #
	    unless (defined $fl_exist_wf) {
		print STDOUT "[ERROR 'chk_ovlp_pcands_vs_sliding_wds'] The subroutine 'binsearch_num' FAILED on the $pc th purge-candidate (w/ \$beg_pc = $beg_pc) against the set of window-frames: $! !\n\n";
		return (undef, undef, undef, undef);
	    }
	    my $ct_ovlp_wfs =0;
	    if ($fl_exist_wf) {
		$ct_ovlp_wfs++;
		push @{$off_wds_but_ovlp_wframes[$indx_wf]}, $pc;
	    }

	    for (my $wf = $indx_wf+1; $wf < $ct_wfs; $wf++) {
		my $beg_wf = $window_frames->[$wf][$indx_beg_wf];
		if ($end_pc < $beg_wf) { last; } # The window-frame is already on the right of the purge-candidate.
		$ct_ovlp_wfs++;
		push @{$off_wds_but_ovlp_wframes[$wf]}, $pc;
	    }
	    if ($ct_ovlp_wfs>0) { next; }


			# Now, the purge-candidate does NOT overlap any window-frames. #

	    push @off_wframes, $pc;


	} # End of the main for-loop (over (merged) purge-candidates). #


	return (\@list_mgd_pcands_ovlp_sbwds, \@off_sbwds_but_ovlp_wds, \@off_wds_but_ovlp_wframes, \@off_wframes);

}	# End of the subroutine, 'chk_ovlp_pcands_vs_sliding_wds'. #


# determine_limits_window:
#	This subroutine determines the left- and right-limits
#	as far as which the window can be expanded, at the maximum.
#	[CREATED in creation of 'anex.ver0.5.pl', on 2019/06/21.]
#
# INPUTS: ($wd, \@set_windows, \@window_frames, \@set_cls_gpbs), where
#
#	$wd is the index of the subject window;
#
#	@{$set_windows[$wd]} = ($beg_wd, $end_wd, \@invlvd_gpd_sgs, $ct_gbs, $fr) is for the $wd th sliding window,
#	where
#	the window starts at the $beg_wd th column and ends at the $end_wd th column,
#	contains gapped segments listed in @invlvd_gpd_sgs,
#	$ct_gbs is the count of gap-blocks in the window,
#	and $fr is the index of the frame the window belongs to;
#
#	@{$window_frames[$fr]} = ($start_fr, $end_fr)
#	defines the $fr th frame for the window analysis,
#	that is, the frame starts at the $start_fr th column and ends at the $end_fr th column.
#	(Both are included in the frame.)
#	[\@set_windows & \@window_frames are typically an output of 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	@{$set_cls_gpbs[$cl]->[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $gpb th gap-pattern block
#	in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#	$len_gpb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);
#	[\@set_windows is typically an output of 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm", 
#	  or 'enum_psm_indel_hsts_for_msa_finer_1ST_HF' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#	$beg_fr and $end_fr are the beginning and end of the window-frame that the $wd th window belongs to;
#
# OUTPUTS: ($left_limit_wd, $right_limit_wd), where
#
#	$left/right_limit_wd is the left/right-limit of the window.
#
sub determine_limits_window ($\@\@\@) {

	my ($wd, $set_windows, $window_frames, $set_cls_gpbs) = @_;

	my ($lm_wd, $rm_wd) = (0, $#{$set_windows});

	my ($beg_wd, $end_wd, $invlvd_gpd_sgs, $ct_gbs, $fr) = @{$set_windows->[$wd]};
	my ($beg_fr, $end_fr) = @{$window_frames->[$fr]};
	my ($lm_gseg, $rm_gseg) = @{$invlvd_gpd_sgs}[0, $#{$invlvd_gpd_sgs}];

	my $left_limit_wd;
	if ($wd == $lm_wd) { # There are no other windows on the left.
	    $left_limit_wd = $beg_fr;

	} elsif ($set_windows->[$wd-1][4] < $fr) { # The "left-neighboring" window belongs to a different frame.
	    $left_limit_wd = $beg_fr;

	} else { # The "left-neighboring" window also belongs to the same frame.

	    my $lflank_gseg = $set_cls_gpbs->[$lm_gseg-1];
	    my $lflank_gpb = $lflank_gseg->[$#{$lflank_gseg}];
	    my $lflank_gpd_clm = $lflank_gpb->[3];
	    $left_limit_wd = $lflank_gpd_clm +2; # At least one gapless column MUST separate the window from the gapped segment.
	} # End of "if ($wd == $lm_wd) {...} elsif (...) {...} else {...}".

	my $right_limit_wd;
	if ($wd == $rm_wd) { # There are no other windows on the right.
	    $right_limit_wd = $end_fr;

	} elsif ($fr < $set_windows->[$wd+1][4]) { # The "right-neighboring" window belongs to a different frame.
	    $right_limit_wd = $end_fr;

	} else { # The "right-neighboring" window also belongs to the same frame.

	    my $rflank_gseg = $set_cls_gpbs->[$rm_gseg+1];
	    my $rflank_gpb = $rflank_gseg->[0];
	    my $rflank_gpd_clm = $rflank_gpb->[2];
	    $right_limit_wd = $rflank_gpd_clm-2; # At least one gapless column MUST separate the window from the gapped segment.
	} # End of "if ($wd == $rm_wd) {...} elsif (...) {...} else {...}".

	return ($left_limit_wd, $right_limit_wd);

}	# End of the subroutine, 'determine_limits_window'. #


# revise_list_gblocks:
#	This subroutine revises @list_gblks_in_wd, which is the main output of 'readin_list_gblks' above,
#	using %set_msa_rows2br_up_or_lw & @nodeids_in_msa.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26-27).]
#
# INPUTS: (\@list_gblks, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa), where
#
#	@list_gblks is an output of 'readin_list_gblks' above.
#	More precisely,
#	@{$list_gblks[$indx_gb]} = ($br_org, $up_or_lw_org, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, \@affd_exnodes_org, \@affd_msa_rows),
#	provides information on the $indx_gb th gap-block;
#	(See the header of 'readin_list_gblks' for more details.);
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[\%set_msa_rows2br_up_or_lw is the output of 'mk_set_msa_rows2br_up_or_lw' (in "MyTreeMap_sppl.pm")];
#
#	$nodeids_in_msa[$k] is the ID of the node corresponding to the $k th row of the MSA;
#	[\@nodeids_in_msa is an output of 'wrapper_mk_node2exoffs_msa_rows' or of 'associate_nodeids_w_msa_rows', both in "MyTreeMap_sppl.pm"];
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
#	(NOTE: As a SIDE-EFFECT, this subroutine REVISES @list_gblks, if necessary.)
#
sub revise_list_gblocks (\@\%\@) {

	my ($list_gblks, $set_msa_rows2br_up_or_lw, $nodeids_in_msa) = @_;

	my $ct_gblks = @{$list_gblks};

	for (my $gb=0; $gb <$ct_gblks; $gb++) { # Main for-loop (over gap-blocks). #

	    my $info_gblk = $list_gblks->[$gb];
	    my ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, $affd_exnodes_org, $affd_msa_rows) = @{$info_gblk};

	    my $br_up_or_lw_rvsd = $set_msa_rows2br_up_or_lw->{$cnct_affd_msa_rows};
	    unless (defined $br_up_or_lw_rvsd) {
		print STDOUT "[ERROR 'revise_list_gblocks'] The $gb th gap-block SHOULD affect the MSA rows, ($cnct_affd_msa_rows), which has NO CORRESPONDING (branch & U/L status) pair: $! !!\n\n";
		return 0;
	    }
	    my ($br_rvsd, $up_or_lw_rvsd) = @{$br_up_or_lw_rvsd};

	    my @affd_exnodes_rvsd = ();
	    foreach my $row (@{$affd_msa_rows}) { push @affd_exnodes_rvsd, $nodeids_in_msa->[$row]; }
	    my @std_affd_exnodes_rvsd = sort {$a <=> $b} @affd_exnodes_rvsd;
	    my $cnct_affd_exnodes_rvsd = join (',', @std_affd_exnodes_rvsd);

	    $info_gblk->[0] = $br_rvsd;
	    $info_gblk->[1] = $up_or_lw_rvsd;
	    $info_gblk->[4] = $cnct_affd_exnodes_rvsd;
	    $info_gblk->[6] = \@std_affd_exnodes_rvsd;

	} # End of the main for-loop (over gap-blocks). #

	return 1;

}	# End of the subroutine, 'revise_list_gblocks'. #


# mk_revised_list_gblocks:
#	This subroutine creates a REVISED @list_gblks_in_wd
#	from @features_gblks_in_wd, which is the main output of 'readin_features_gblks' above,
#	with the assist of %set_msa_rows2br_up_or_lw & @nodeids_in_msa.
#	The subroutine is created by modifying 'revise_list_gblocks'.
#	[CREATED in creation of 'coordinate_point_lcl_msa.ver0.7.pl' (on 2020/06/05).]
#
# revise_list_gblocks:
#	This subroutine revises @list_gblks_in_wd, which is the main output of 'readin_list_gblks' above,
#	using %set_msa_rows2br_up_or_lw & @nodeids_in_msa.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26-27).]
#
# INPUTS: (\@features_gblks, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa), where
#
#	@features_gblks is an output of 'readin_features_gblks' above.
#	More precisely,
#	@{$features_gblks[$indx_gb]} = ($br_org, $up_or_lw_org, $lm_clm, $rm_clm, $cnct_affd_msa_rows, \@affd_msa_rows),
#	provides some defining features of the $indx_gb th gap-block;
#	(See the header of 'readin_features_gblks' for more details.);
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[\%set_msa_rows2br_up_or_lw is the output of 'mk_set_msa_rows2br_up_or_lw' (in "MyTreeMap_sppl.pm")];
#
#	$nodeids_in_msa[$k] is the ID of the node corresponding to the $k th row of the MSA;
#	[\@nodeids_in_msa is an output of 'wrapper_mk_node2exoffs_msa_rows' or of 'associate_nodeids_w_msa_rows', both in "MyTreeMap_sppl.pm"];
#
# OUTPUT: (\@list_gblks), where
#
#	@list_gblks is a REVISED list of gap-blocks,
#	which follows the same format as the output of 'readin_list_gblks' above.
#	More precisely,
#	@{$list_gblks[$indx_gb]} = ($br_org, $up_or_lw_org, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, \@affd_exnodes_org, \@affd_msa_rows),
#	provides information on the $indx_gb th gap-block;
#	(See the header of 'readin_list_gblks' for more details.);
#
#
sub mk_revised_list_gblocks (\@\%\@) {

	my ($features_gblks, $set_msa_rows2br_up_or_lw, $nodeids_in_msa) = @_;

	my $ct_gblks = @{$features_gblks};

	my @list_gblks = ();

	for (my $gb=0; $gb <$ct_gblks; $gb++) { # Main for-loop (over gap-blocks). #

	    my $feat_gblk = $features_gblks->[$gb];
	    my ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_msa_rows, $affd_msa_rows) = @{$feat_gblk};

	    my $br_up_or_lw_rvsd = $set_msa_rows2br_up_or_lw->{$cnct_affd_msa_rows};
	    unless (defined $br_up_or_lw_rvsd) {
		print STDOUT "[ERROR 'mk_revised_list_gblocks'] The $gb th gap-block SHOULD affect the MSA rows, ($cnct_affd_msa_rows), which has NO CORRESPONDING (branch & U/L status) pair: $! !!\n\n";
		return 0;
	    }
	    my ($br_rvsd, $up_or_lw_rvsd) = @{$br_up_or_lw_rvsd};

	    my @affd_exnodes_rvsd = ();
	    foreach my $row (@{$affd_msa_rows}) { push @affd_exnodes_rvsd, $nodeids_in_msa->[$row]; }
	    my @std_affd_exnodes_rvsd = sort {$a <=> $b} @affd_exnodes_rvsd;
	    my $cnct_affd_exnodes_rvsd = join (',', @std_affd_exnodes_rvsd);

	    ##$info_gblk->[0] = $br_rvsd;
	    ##$info_gblk->[1] = $up_or_lw_rvsd;
	    ##$info_gblk->[4] = $cnct_affd_exnodes_rvsd;
	    ##$info_gblk->[6] = \@std_affd_exnodes_rvsd;
	    ##my ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, $affd_exnodes_org, $affd_msa_rows) = @{$info_gblk};

	   $list_gblks[$gb] = [$br_rvsd, $up_or_lw_rvsd, $lm_clm, $rm_clm, $cnct_affd_exnodes_rvsd, $cnct_affd_msa_rows, \@std_affd_exnodes_rvsd, $affd_msa_rows];

	} # End of the main for-loop (over gap-blocks). #

	return \@list_gblks ;

}	# End of the subroutine, 'mk_revised_list_gblocks'. #




# relative_coords_for_list_mgd_pcands:
#	This subroutine switches from the absolute coordinates to the relative coordinates (against a reference position)
#	for all the positions regarding the (merged) purge-candidates.
#	[CREATED in creation of 'anex_for_sgl_wd.ver0.5.pl' on 2019/07/03.]
#
# INPUTS: (\@list_mgd_pcands, $ref_pos), where
#
#	\@list_mgd_pcands is typically an output of 'readin_list_mgd_pcands'.
#	More precisely,
#	@{$list_mgd_pcands[$pc]} = ($beg_pc, $end_pc, $br, \@msa_rows_under_br, \@constituent_sites, $pv, $pv0, \@sgn_wds_included) provides information on the $pc th (merged) purge-candidate region, where
#	$beg_pc & $end_pc are the sites at the beginning & end, respectively, of the region,
#	$br is the ID of the branch across which the "purge" is suspected within the region,
#	@msa_rows_under_br stores the indices of the MSA rows corresponding to the external offsprings of $br,
#	@constituent_sites stores the sites that constitute the region,
#	$pv is the P-value measuring the deviation from the specified substutution model,
#	$pv0 is the P-value measuring the deviation from the random-match model,
#	and @sgn_wds_included is the significant windows included in the region;
#	more precisely, 
#	@{$sgn_wds_included[$swd]} = (\@sites_in_wd, $pv_wd, $pv0_wd) provies information on the $swd th significant window, where
#	@sites_in_wd stores the sites in the window,
#	$pv_wd and $pv0_wd are the P-values for the window, measuring the same types of deviations as the aforementioned $pv and $pv0, respectively;
#	
#	$ref_pos is the coordinate of the reference position against which the relative coordinates will be set;
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub relative_coords_for_list_mgd_pcands (\@$) {

	my ($list_mgd_pcands, $ref_pos) = @_;

	my $ct_mpcands = @{$list_mgd_pcands};

	for (my $pc=0; $pc <$ct_mpcands; $pc++) { # Main outer for-loop (over purge-candidates). #

	    my $info_mpcand = $list_mgd_pcands->[$pc];
	    $info_mpcand->[0] -= $ref_pos; # $beg_pc,
	    $info_mpcand->[1] -= $ref_pos; # $end_pc,
	    my ($constituent_sites, $sgn_wds_included) = @{$info_mpcand}[4, 7];

	    my $ct_csites = @{$constituent_sites};
	    for (my $s=0; $s <$ct_csites; $s++) { $constituent_sites->[$s] -= $ref_pos; }

	    my $ct_sgn_wds = @{$sgn_wds_included};
	    for (my $w=0; $w <$ct_sgn_wds; $w++) { # 2nd middle for-loop (over significant windows). #

		my $sites_in_wd = $sgn_wds_included->[$w][0];

		my $ct_siw = @{$sites_in_wd};
		for (my $s=0; $s <$ct_siw; $s++) { $sites_in_wd->[$s] -= $ref_pos; }

	    } # End of the 2nd middle for-loop (over significant windows). #

	} # End of the main outer for-loop (over purge-candidates). #

	return 1;

}	# End of the subroutine, 'relative_coords_for_list_mgd_pcands'. #



1;		# a "spell" which makes the module "use"-able and/or "require"-able #


