#!/usr/bin/perl
#
# summary_examine_complex.2_3D.raw.pl:
#	This slightly modified version of 'summary_examine_complex.2_2D.raw.pl'
#	takes effectively 3D distributions of the properties.
#	As the predecessor does, it classifies the segments into two categories: A = 'too-long' + 'complex errors', and  B = 'non-complex errors' + 'correct'.
#	[CREATED on June 24 (Fri), 2016.]
#
# summary_examine_complex.2_2D.raw.pl:
#	This slightly modified version of 'summary_examine_complex.2.raw.pl'
#	takes exclusively the 2D distributions of the properties.
#	It classifies the segments into two categories: A = 'too-long' + 'complex errors', and  B = 'non-complex errors' + 'correct'.
#	[CREATION STARTED on June 21 (Tue), 2016, ENDED on June dd (www), 2016.]
#
# summary_examine_complex.2.raw.pl:
#	The focus of this script is on the difference between the clusters of gapped segments
#	that were erroneously reconstructed and those that were correctly reconstructed.
#	This script was created by modifying 'summary_examine_complex.1.raw.pl'.
#	[CREATION STARTED on June 2 (Thu), 2016, ENDED on June 3 (Fri), 2016.]
#
# summary_examine_complex.1.raw.pl:
#	This Perl script summarizes the results of the Perl script 'examine_complex_via_mblks.2.pl',
#	which are converted into a single table via 'tabulate_examine_complex.pl'.
#	The script was created by modifying 'summary_examine_purge_ctrl.raw.pl'.
#	Especially, it imports (from 'bch_pbs_examine_purges.D.pl') the function to read in the list of relevant segments.
#	[CREATED on June 1 (Wed), 2016.]
#
# summary_examine_purge_ctrl.raw.pl:
#	This Perl script summarizes the results of the Perl script 'examine_purge_ctrls.2.pl',
#	which are converted into a single table via 'tabulate_examine_purge_ctrls.pl'.
#	The script was created by slightly modifying 'summary_examine_purge.raw.pl'.
#	[CREATED on May 16 (Mon), 2016.]
#
# summary_examine_purge.raw.pl:
#	This Perl script summarizes the results of the Perl script 'examine_purges_via_mblks.2.pl',
#	which are converted into a single table via 'tabulate_examine_purge.pl'.
#	The script was created by taking advantage of the framework of 'summary_classify_msa_errors1.raw.pl'.
#	[CREATION Started on May 11 (Wed), 2016, ENDED on Mmm dd (www), 2016.]
#
#	[MODIFIED on May 16 (Mon), 2016.]
#	Now, the script also examines more detailed distributions,
#	such as the distributions of the P-values for specified window sizes and specified branch properties.
#
# summary_classify_msa_errors1.raw.pl:
#	This script is a slightly modified version of 'summary_classify_msa_errors.raw.pl'.
#	The script now tallies the number of all composite-blocks in each erroneous segment as well.
#	[CREATED on Oct 26 (Mon), 2015]
#
# summary_classify_msa_errors.raw.pl:
#	This script does not pre-screen either in terms of whether or not the erroneous segment is on a MSA boudary or 
#	whether or not the segment contains non-parsimonious gapped segments.
#	It was created by slightly modifying the script 'summary_classify_msa_errors.bulk_pars.pl'.
#	(CREATION STARTED on Nov 17 (Mon), 2014, ENDED on Nov 18 (Tue), 2014.)
#
# summary_classify_msa_errors.bulk_pars.pl:
#	This Perl script summarizes the results of the Perl script 'classify_msa_errors_via_mblks.1.pl',
#	which are converted into a single table via 'tabulate_classify_msa_errors.pl'.
#	It excludes those erroneous segments on the MSA boundaries and those containing non-parsimonious gapped segments.
#	It was created by taking advantage of the framework of the script 'summary_cmp_ref_vs_rec_gpconf_stats.bulk_pars.pl'.
#	(CREATION STARTED on Nov 16 (Sun), 2014, ENDED on Nov 17 (Mon), 2014.)
#
# summary_cmp_ref_vs_rec_gpconf_stats.bulk_pars.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_gconf_stats.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_gconf_stats.pl'.
#	It excludes those erroneous segments on the MSA boundaries and those containing non-parsimonious gapped segments.
#	It was created by taking advantage of the framework of the script 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2_lite.pl'.
#	(CREATION STARTED on Nov 14 (Fri), 2014, ENDED on Nov 15 (Sat), 2014.)
#
# summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2_lite.pl:
#	This script was created by just stripping commented-out commands off 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2.pl'.
#	(CREATED on Nov 14 (Fri), 2014.)
#
# summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2.pl:
#	The only modifications from the previous version is that
#	this version uses boundaries whose results could easily be interpreted as 'round-off'.
#	(CREATED on Nov 13 (Thu), 2014.)
#
# summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_indel_cts.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_cts_indels.pl'.
#	It excludes those erroneous segments on the MSA boundaries and those containing non-parsimonious gapped segments.
#	It was created by modifying the script 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'.
#	(CREATED on Oct 9 (Thu), 2014.)
#
# summary_cmp_ref_vs_rec_complliment.bulk_pars.pl:
#	This revised version of 'summary_cmp_ref_vs_rec_complliment.bulk.pl'
#	further excludes erroneously reconstructed segments including non-parsimonious gapped segments
#	before tallying the bulk erroneous segments.
#	(CREATED on Oct 3 (Fri), 2014.)
#
# summary_cmp_ref_vs_rec_complliment.bulk.pl:
#	This revised version of 'summary_cmp_ref_vs_rec_complliment.pl' tallies the erroneously reconstructed segments
#	after excluding the segments that touch either of the MSA ends.
#	(CREATION STARTED on Oct 2 (Thu), 2014, ENDED on Oct 3 (Fri), 2014.)
#
#
# summary_cmp_ref_vs_rec_complliment.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_compllimment.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_compllimment.pl'.
#	It was created by modifying the script 'summary_cmp_ref_vs_rec_abs_scr.pl'.
#	(CREATED on Oct 2 (Thu), 2014.)
#
# summary_cmp_ref_vs_rec_abs_scr.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_abs_scoring.PhyML.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_abs_scoring.pl'.
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);
use MyMalnIO qw(
	extract_fasta extract_clw
	);
	## pr_clustal_like_lite_aa
	## pr_ODEN_input pr_phylip_aln
use MyArrayManipu qw( diff_array );

use MyMalnManipu qw( rm_gapcolumn2 );
use MyMalnManipu_UH qw(
	make_segid2segaln
	msa2len_grps msa2column_set extract_aln_mtrx2
	);
	## extract_aln_mtrx

use MyCluster_distribution qw(
	bin_initializer cluster_distributor
	batch_initializer batch_pr_file_distri batch_pr_file_distri_xl
	cluster_distributor_wgt
	init_2dim_distri distributor_2dim pr_file_2dim_distri pr_file_2dim_distri_xl
	create_key_intervals );
use MyContingency_table qw(
	ratios_contingency_tables add_ctbl_margins
	pr_juxtapose_ctbls_xl convert_ctbl_to_2d_distri
	);

sub broad_error_type ($);


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);


	### User-specified Parameters ###


		# MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #
our %INTYPE2CTSEQS = (
	'12primates' => 12,
	'15mammals' => 15,
	'9fe_mammals' => 9,
	'TESTs/Test1a' => 15
	);
##our  $CT_SEQS = 16; # Number of sequences in a MSA. # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #

	# If the scores of the reference and the reconstructed MSAs for a segment 
	# differ less than this value, they are regarded as 'equal' to each other.
##our $THRSH_EQUAL = 0.1;
##our $THRSH_EQUAL = 1E-03;
our $THRSH_EQUAL = 1E-05;

our $EPS = $THRSH_EQUAL * 1E-03; # An 'infinitesimal number' used to take statistics.


	# Level of the verboseness of the test outputs.
	# (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose.)

our $TEST_LEVEL = 5;

	## Specifying the fields of input/output files. ##

		# MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #

my @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}');

=For the predecessor 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2_lite.pl'.
	# MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
our @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Total_Score.ref',  'Total_Score.rec', 'AvAv#{insertions}.ref', 'AvAv#{insertions}.rec', 'AvAv#{deletions}.ref', 'AvAv#{deletions}.rec', 'Av#{null_columns}.ref', 'Av#{null_columns}.rec');

our @TAGS_OUT_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'AvAv#{indels}.ref', 'Diff(AvAv#{indels})', 'L1_Dist(#{ins},#{dels})', 'Deletion_Bias');
=cut


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
our @TAGS_AUX_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', '#{nonpars_gpd_segs}'); # ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'. #

	# ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
our @TAGS_AUX3_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Total_Score.ref',  'Total_Score.rec', 'Av#{null_columns}.ref', 'Av#{null_columns}.rec');
=cut


##my $alignment_type = 'MAFFT/E_INS_i';
##my $alignment_type = 'MAFFT/G_INS_i';
##my $alignment_type = 'PRANK/Default';
my $alignment_type = 'PRANK/Best_Match';

	# ADDED in creation of 'summary_examine_purge.raw.pl'. #
#my $input_msa_type = '12primates';
my $input_msa_type = '15mammals';
#my $input_msa_type = '9fe_mammals';

#my $input_msa_type = 'TESTs/Test1a'; # For TESTING the script. #

our $CT_SEQS = $INTYPE2CTSEQS{$input_msa_type}; ## ADDED in creation of 'summary_examine_complex.1.raw.pl'. ##

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.

	# Auxiliary files. #
	# $auxfile = "$top_auxdir/$mid_auxdir_pref$id_class$mid_auxdir_suff/$mid_aux_path/$auxfile_pref$session_id$auxfile_suff",
	# where  $id_class and $sesion_id will be collected later.
	# (ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'.)

my $top_auxdir = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Errors/Erroneous_Regions/Equiv_Regions/$alignment_type/hogd_alt_lite_sbcl.1_lite/Outputs";
my $mid_auxdir_pref = 'hog_d';
my $mid_auxdir_suff = '';
my $mid_aux_path = '';
my $auxfile_pref = 'equiv_rgs_brief.';
my $auxfile_suff = '.txt';

		# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'. #
my $auxfile2 = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Errors/Erroneous_Regions/Equiv_Regions/$alignment_type/Inclusion_gpd_segs/hogd_alt_lite_sbcl.1/tbl_incl_gpd_segs.txt.gz";
=cut

=DISABLED in creation of 'summary_examine_purge.raw.pl'.

		# ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
##my $auxfile3 = "../../../Cts_Indels/Summary_tbl/hogd_alt_lite_sbcl.1/tbl_cmp_ref_vs_rec_cts_indels.txt.gz";
##my $auxfile3 = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Errors/Contrast_Analyses/Characteristics/Erroneous_Regions/$alignment_type/Cts_Indels/Summary_tbl/hogd_alt_lite_sbcl.1/tbl_cmp_ref_vs_rec_cts_indels.txt.gz";
my $auxfile3 = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Prcd_Errors/Contrast_Analyses/Characteristics/Erroneous_Regions/$alignment_type/Cts_Indels/Summary_tbl/hogd_alt_lite_sbcl.1/tbl_cmp_ref_vs_rec_cts_indels.txt.gz";

=cut

		# ADDED in creation of 'summary_examine_complex.1.raw.pl'. #
my $auxfile5 = "../../../Aux_Inputs/Tbl_Complex/tbl_errsegs_w_complex.$input_msa_type.txt.gz"; 

	### Input/Output files/directories ###

	# $infile1 = "$indir1/$infile_name1",
my $indir1 = "../../Summary_tbl/$input_msa_type"; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
##my $indir1 = "../../Summary_tbl/hogd_alt_lite_sbcl.1";
##my $infile_name1 = 'tbl_cmp_ref_vs_rec_compllimment.txt.gz';
##my $infile_name1 = 'tbl_cmp_ref_vs_rec_cts_indels.txt.gz'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
##my $infile_name1 = 'tbl_cmp_ref_vs_rec_gconf_stats.txt.gz'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'.
##my $infile_name1 = 'tbl_classify_msa_errors.txt.gz'; # MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'.
##my $infile_name1 = 'tbl_classify_msa_errors.1.txt.gz'; # MODIFIED in creation of 'summary_classify_msa_errors1.raw.pl'. #
##my $infile_name1 = 'tbl_examine_purge.txt.gz'; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
##my $infile_name1 = 'tbl_examine_purge_ctrl.txt.gz'; # MODIFIED in creation of 'summary_examine_purge_ctrl.raw.pl'. #

my $infile_name1 = 'tbl_examine_complex.txt.gz'; # MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #


	# $outfile = "$outdir/$outfile_name"
my $outdir =  $input_msa_type; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
##my $outdir =  'hogd_alt_lite_sbcl.1';
##my $outdir = "TESTs/Test1"; ### For TESTING the script ###
##my $outfile_name = 'tbl_diff_ref_vs_rec_compllimment.txt';
##my $outfile_name = 'tbl_diff_ref_vs_rec_cts_indels.txt'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
##my $outfile_name = 'tbl_fdiff_ref_vs_rec_cts_indels.txt'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
##my $outfile_name = 'tbl_classify_sbj_msa_errors.txt'; # MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
##my $outfile_name = 'tbl_classify_sbj_msa_errors.1.txt'; # MODIFIED in creation of 'summary_classify_msa_errors1.raw.pl'. #
##my $outfile_name = 'tbl_examine_sbj_purges.txt'; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
my $outfile_name = 'tbl_examine_sbj_purge_ctrls.txt'; # MODIFIED in creation of 'summary_examine_purge_ctrl.raw.pl'. #



	## Check files/directories ##

foreach my $dir ($indir1) {
##foreach my $dir ($intopdir1, $intopdir2) {
    unless (( -d $dir ) and ( -r $dir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable TOP INPUT DIRECTORY '$dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}

##=DISABLED in creation of 'summary_examine_purge.raw.pl'. ## RESURRECTED & MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. ##

foreach my $auxfile ($auxfile5) { # MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #
##foreach my $auxfile ($auxfile3) {...} # MODIFIED in creation of 'summary_classify_msa_errors.raw.pl'. #
##foreach my $auxfile ($auxfile2, $auxfile3) {...} # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
   unless ( ( -f $auxfile ) and ( -r $auxfile ) ) { # ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'. #
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable Auxiliary File '$auxfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
   }
}
##=cut

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.

		# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'. #

foreach my $dir ($top_auxdir) {
    unless (( -d $dir ) and ( -r $dir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable Top Auxiliary Directory '$dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}

my $list_id_classes = list_sessions_ext ($top_auxdir, $mid_auxdir_pref, $mid_auxdir_suff, 'dr');
unless (defined $list_id_classes) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the top auxiliary directory '$top_auxdir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

	# $auxfile = "$top_auxdir/$mid_auxdir_pref$id_class$mid_auxdir_suff/$mid_aux_path/$auxfile_pref$session_id$auxfile_suff",
		# END of "ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'." #

=cut # END of "DISABLED in creation of 'summary_classify_msa_errors.raw.pl'."


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
=cut


	### Declare the Beginning of the Script ###

print STDOUT "\n<< Program name: $program_name (path: $program_path) >>\n\n";

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.
print STDOUT "# This analysis will be performed after excluding the erroneous segments on the left-ends or the right-ends of MSAs, as well as the erroneous segments containing non-parsimonious gapped segments.\n\n";
=cut

print STDOUT '$alignment_type = ', $alignment_type , " .\n\n";

=For the predecessor 'summary_cmp_ref_vs_rec_abs_scr.pl'.
print STDOUT "Scoring method: $METHOD .\n\n";
=cut

print STDOUT '$input_msa_type = ', $input_msa_type , " .\n\n"; # ADDED in creation of 'summary_examine_purge.raw.pl'. #

print STDOUT "Input file:  {in_dir1}/$infile_name1,\n";
print STDOUT "       where  {in_dir1} = $indir1 .\n\n";


			# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'. #
print STDOUT "Auxiliary input files:\n\n";

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.
print STDOUT "   (for segments on MSA-ends)  {top_auxdir}/$mid_auxdir_pref", '{ID_Class}', "$mid_auxdir_suff/$mid_aux_path/$auxfile_pref", '{session_ID}', "$auxfile_suff ,\n";
print STDOUT "              where {top_auxdir} = $top_auxdir ,\n";
print STDOUT "                    {ID_Class} = ", join (', ', @{$list_id_classes}), " ,\n\n";
print STDOUT "                    {session_ID}'s will be collected in the class-specific manner ;\n\n";
	# $auxfile = "$top_auxdir/$mid_auxdir_pref$id_class$mid_auxdir_suff/$mid_aux_path/$auxfile_pref$session_id$auxfile_suff",

print STDOUT "    (for non-parsimonious segments) $auxfile2 ,\n\n"; # ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'.
=cut


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
print STDOUT "    (for complete-likelihoods and counts of null columns) $auxfile3 .\n\n"; # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
=cut

print STDOUT "    (for segments with 'Complex' errors or 'N/A') $auxfile5 .\n\n"; # ADDED in creation of 'summary_examine_complex.1.raw.pl'. #

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
print STDOUT "Output file: {out_dir}/$outfile_name,\n";
print STDOUT "       where {out_dir} = $outdir .\n\n\n";
=cut

print STDOUT "Parameters necessary for the calculation:\n\n";
##print STDOUT "DEFINITION:  'Diff' = 'Rec' - 'Ref' .\n\n";
print STDOUT '$THRSH_EQUAL = ', "$THRSH_EQUAL .\n";
print STDOUT '$EPS         = ', "$EPS .\n\n\n";


print STDOUT '#{Sequences} = ', $CT_SEQS , " .\n\n\n"; # ADDED in creation of 'summary_examine_complex.1.raw.pl'. #


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
	# ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
print STDOUT "** Acronyms Used:\n\n";
print STDOUT "  * '(i)CII'  = '(Incomplete-)Collapse-of-Independent-Insertions' ,\n";
print STDOUT "  * '(i)CEII' = '(Incomplete-)Creation-of-Erroneous-Independent-Insertions' ,\n";
print STDOUT "  * '(p)NB_Merge/Split' = '(Partial-)Neighboring-Branch-Merge/Split' .\n\n";
=cut



	### Arrays used to take Statistics ###

		### MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. ###
		### => FUrther MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. ###
		### => Yet Further MODIFIED in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

our @BOUNDS_CT0 = (0 .. 5);
our @BOUNDS_CT = (0 .. 11, 13, 15, 17, 19, 21, 26, 31, 41, 51);
##our @BOUNDS_PNCT0 = (-7 .. 7);
##our @BOUNDS_PNCT = (-30, -25, -20, -15, -10 .. 11, 16, 21, 31);

=For the predecessor 'summary_examine_complex.2.raw.pl'.
our @BOUNDS_CT = (0 .. 11, 13, 15, 17, 19, 21, 26, 31, 41, 51, 76, 101, 151, 201, 301);
our @BOUNDS_PNCT = (-200, -100, -50, -40, -30, -25, -20, -15, -10 .. 11, 16, 21, 31, 41, 51, 101, 201);
##our @BOUNDS_CT = (0 .. 11, 13, 15, 17, 19, 21, 26, 31, 41, 51, 76, 101, 151, 201);
##our @BOUNDS_PNCT = (-100, -50, -40, -30, -25, -20, -15, -10 .. 11, 16, 21, 31, 41, 51, 101);
our @BOUNDS_PRPT = (0, 0.001, 0.005, 0.01, 0.025, 0.05, 0.1, 0.2, 0.5, 0.8, 0.9, 0.95, 0.975, 0.99, 0.995, 0.999, 1);
=cut

		### MODIFIED in creation of 'summary_examine_complex.2_3D.raw.pl'. (I) ###

my @ctpairs = ( '(1,-1)', '(1,1)', '(2,-2)', '(2,0)', '(3,-3)', '(2,2)', '(3,-1)', '(3,1)', '(4,-2)', '(4,0)', 'Others');
my @tags_ct = ('Intervals \ X', @ctpairs);

my %ctpair2bins_maxct_columns_A; # = ($tag_ctpair => \@bins_maxct_columns_A, ...)
my %ctpair2bins_maxct_columns_B; # = ($tag_ctpair => \@bins_maxct_columns_B, ...)
my %ctpair2bins_max_totlen_indels_A; # = ($tag_ctpair => \@bins_max_totlen_indels_A, ...)
my %ctpair2bins_max_totlen_indels_B; # = ($tag_ctpair => \@bins_max_totlen_indels_B, ...)

my @batch_maxct_columns_A = ();
my @batch_maxct_columns_B = ();
my @batch_max_totlen_indels_A = ();
my @batch_max_totlen_indels_B = ();

foreach my $pair (@ctpairs) {

	my @bins_maxct_clms_A = ();
	my @bins_maxct_clms_B = ();
	my @bins_max_totlen_indels_A = ();
	my @bins_max_totlen_indels_B = ();

	$ctpair2bins_maxct_columns_A{$pair} = \@bins_maxct_clms_A;
	$ctpair2bins_maxct_columns_B{$pair} = \@bins_maxct_clms_B;
	$ctpair2bins_max_totlen_indels_A{$pair} = \@bins_max_totlen_indels_A;
	$ctpair2bins_max_totlen_indels_B{$pair} = \@bins_max_totlen_indels_B;

	push @batch_maxct_columns_A, \@bins_maxct_clms_A;
	push @batch_maxct_columns_B, \@bins_maxct_clms_B;
	push @batch_max_totlen_indels_A, \@bins_max_totlen_indels_A;
	push @batch_max_totlen_indels_B, \@bins_max_totlen_indels_B;
}

batch_initializer (@batch_maxct_columns_A, @BOUNDS_CT);
batch_initializer (@batch_maxct_columns_B, @BOUNDS_CT);
batch_initializer (@batch_max_totlen_indels_A, @BOUNDS_CT);
batch_initializer (@batch_max_totlen_indels_B, @BOUNDS_CT);


=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (I)
		### ADDED in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

(my $bins_max_totct_indels_vs_maxct_columns_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_maxct_columns_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_totlen_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_max_totlen_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_max_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

(my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_min_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (I)"


=For the predecessor 'summary_examine_purge.raw.pl'.
		### MODIFIED in creation of 'summary_examine_purge.raw.pl'. #

my @tags_ct = ('Intervals', 'block_size', '|shift|');

my @bins_size_block = my @bins_abs_shift = ();
my @batch_ct = (\@bins_size_block, \@bins_abs_shift);
batch_initializer (@batch_ct, @BOUNDS_CT);

my @bins_rlv_brlen = my @bins_pv = my @bins_diff_lnlk_respatt = ();
bin_initializer (@bins_rlv_brlen, @BOUNDS_BRLEN);
bin_initializer (@bins_pv, @BOUNDS_PRPT);
bin_initializer (@bins_diff_lnlk_respatt, @BOUNDS_DIFF_LNLK);

(my $bins_ct_invlvd_vs_brlen, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_BRLEN, "%5.3f");

		# ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016). #
(my $bins_blsize_vs_brlen, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_BRLEN, "%5.3f");
(my $bins_blsize_vs_ct_invlvd, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_CT, "%2d");
		# END of "ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016)." #

(my $bins_blsize_vs_pv, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_PRPT, "%4.2f");
(my $bins_ct_invlvd_vs_pv, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_PRPT, "%4.2f");
(my $bins_brlen_vs_pv, undef) = init_2dim_distri (@BOUNDS_BRLEN, @BOUNDS_PRPT, "%4.2f");
=cut




=DISABLED in creation of 'summary_examine_purge.raw.pl'.
			# ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
our @TAGS_OUT_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', '#{cblks}_eff');
foreach my $brtype (@BROAD_TYPES) { push @TAGS_OUT_FLD, "#{$brtype}"; }
=cut





	### Preliminary Process 5: Readin the list of subject segments. ###
	### (IMPORTED from 'bch_pbs_examine_purges.D.pl', and MODIFIED in creation of 'summary_examine_complex.1.raw.pl'.)

my %if_complex; # = ("$id_class:$in_msa_id:$segid" => 1, ...) for segments with complex errors.
my %if_na; # = ("$id_class:$in_msa_id:$segid" => 1, ...) for segments whose error types are 'N/A'.

if (1) {

   my $fh_aux5 = open_input_file2 ($auxfile5);
   unless (defined $fh_aux5) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the auxiliary input file '$auxfile5' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
   }
   local $/ = "\n"; # Use the newline character as the input-record-separator. #

   my $aux_header5 = <$fh_aux5>;

   my @TAGS_AUX5_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Types;of;associated;errors');

   my $indices_fld_aux5 = tfm_labels2indices_gen (@TAGS_AUX5_FLD, $aux_header5, "\t");
   unless (defined $indices_fld_aux5) {
   	$aux_header5 =~ s/\s+$//; 
   	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the auxiliary header:\n";
   	print STDOUT " '$aux_header5' !\n";
   	print STDERR "[ERROR '$program_name'] $!\n";
   	die ;
   }
   my $max_indx_aux5 = 0;
   foreach my $indx (@{$indices_fld_aux5}) { ($max_indx_aux5 <$indx) and ($max_indx_aux5 = $indx); }

   while (my $line = <$fh_aux5>) { # Preliminary while-loop. #

	my @fields = split /\t/, $line;
	(@fields>$max_indx_aux5) or next;
	##(@fields>3) or next;

	my ($id_class, $in_msa_id, $segid, $error_type) = @fields[@{$indices_fld_aux5}];
	##my ($id_class, $in_msa_id, $segid) = @fields[0,1,2];

	my $combo_id = "$id_class:$in_msa_id:$segid";

	if ($error_type eq 'N/A') {
	    $if_na{$combo_id} = 1;
	} elsif ($error_type =~ m/Complex/) {
	    $if_complex{$combo_id} = 1;
	}

=For the source 'bch_pbs_examine_purges.D.pl'
	my $combo_id = "$id_class:$in_msa_id";
	if (defined $class_msa_id2sbjct_segs{$combo_id}) {
	    push @{$class_msa_id2sbjct_segs{$combo_id}}, $segid;
	} else {
	    $class_msa_id2sbjct_segs{$combo_id} = [$segid];
	}
=cut

   }			# End of the preliminary while-loop. #
   close ($fh_aux5);
}

	### END of "Preliminary Process 5: Readin the list of subject segments." ###



	### Prepare the Input/Output files ###

my $infile1 = "$indir1/$infile_name1";
##my $outfile = "$outdir/$outfile_name"; ## DISABLED in creation of 'summary_examine_purge.raw.pl'.

my $fh_in1 = open_input_file2 ($infile1);
unless (defined $fh_in1) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the input file '$infile1' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
my $fh_out = open_output_file ($outfile, 1);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
=cut

	### Main Processes ###
$/ = "\n"; # Use a newline character as an input record separater.

my $in_header1 = <$fh_in1>;
my $indices_fld = tfm_labels2indices_gen (@TAGS_IN_FLD, $in_header1, "\t");
unless (defined $indices_fld) {
	$in_header1 =~ s/\s+$//; 
	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the input header:\n";
	print STDOUT " '$in_header1' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
my $max_indx = 0;
foreach my $indx (@{$indices_fld}) { ($max_indx <$indx) and ($max_indx = $indx); }


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
print $fh_out join ("\t", @TAGS_OUT_FLD), "\n\n";
=cut


my $totct_segs = my $ct_anom = my $ct_w_undef = my $ct_fail = my $ct_succ = 0;
my $ct_skipped1 = my $ct_skipped2 = my $ct_anom2 = 0;
my $ct_skipped3 = my $ct_skipped5  = 0;

my $ct_skipped_x = 0;

my $ct_anom3 = 0; # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
my $totct_all_cblks = 0; # ADDED in creation of 'summary_classify_msa_errors1.raw.pl' (on Oct 26 (Mon), 2015). #
my $totct_equiv_indels = my $totct_eff_cblks = 0; # ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
##my %type_error2sbtotct_ecblks; # = ($type_error => $sub_total_ct_eff_cblks, ...) # ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. 
=cut

=For the predecessor 'summary_examine_purge.raw.pl'.

	## ADDED in creation of 'summary_examine_purge.raw.pl'. ##

my $ct_w_fractional = 0;

my %br2freq_prgs; # = ($br_id => $abs_freq_purges, ...)
my %ct_invlvd2freq_prgs; # = ($ct_invlvd_seqs => $abs_freq_purges, ...)

	## END of "ADDED in creation of 'summary_examine_purge.raw.pl'." ##

	## ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016). ##

my %blsize2bins_ct_invlvd_vs_pv; # = ($block_size => \@bins_ct_invlvd_vs_pv_sp, ...) (Here, '_sp' indicates that the distribution is for a specific $block_size.) # 
my %blsize2bins_brlen_vs_pv; # = ($block_size => \@bins_brlen_vs_pv_sp, ...)

	## END Of "ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016)." ##

=cut # END of "For the predecessor 'summary_examine_purge.raw.pl'."

			### MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. ###

my @last_cluster = ();
while (my $in = <$fh_in1>) {
	($in =~ m/^\S/) or next;
	$in =~ s/\s+$//;
	my @in_flds = split /\t/, $in;
	unless (@in_flds > $max_indx) {
	    print STDOUT "[WARNING '$program_name'] An ANOMALOUS line was found:\n";
	    print STDOUT "    '$in' : $! !\n\n";
	    $ct_anom++;
	    next;
	}
	$totct_segs++;
	my @rlv_flds = @in_flds[@{$indices_fld}];
	push @last_cluster, \@rlv_flds;
	last;
}

my $ct_clusters = 0;
my $end_coord_last_cls ;
my $last_id_class = my $last_session = 'Pre-start'; # ADDED for DEBUGGING. #

while (@last_cluster>0) { # Outer while-loop (until the last cluster runs out). #

			### END of "MODIFIED in creation of 'summary_examine_complex.2.raw.pl'." ###

=DISABLED in creation of 'summary_examine_complex.2.raw.pl'.
while (my $in = <$fh_in1>) { # Outer while-loop (over erroneous MSA segments) #
=cut

	## ($totct_segs>=10) and last; ### For TESTING the script ###

	my @in_flds = (); # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

    while (my $in = <$fh_in1>) { # Middle preliminary while-loop (over erroneous MSA segments). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	($in =~ m/^\S/) or next;
	$in =~ s/\s+$//;

	##$totct_segs++;

	@in_flds = split /\t/, $in; # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #
	##my @in_flds = split /\t/, $in;
	unless (@in_flds > $max_indx) {
	    print STDOUT "[WARNING '$program_name'] An ANOMALOUS line was found:\n";
	    print STDOUT "    '$in' : $! !\n\n";
	    $ct_anom++;
	    @in_flds = (); # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #
	    next;
	}

	$totct_segs++; # MOVED from above. #

	last; # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

    } 		# End of the middle preliminary while-loop (over erroneous MSA segments). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	my $last_combo_id = join (':', @{$last_cluster[0]}[0,1,2]); # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	my @rlv_flds = (@in_flds>0) ? @in_flds[@{$indices_fld}] : () ; # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #


			# MAJOR MODIFICATION in creation of 'summary_examine_complex.1.raw.pl'. #

		## Check if the current segment belongs to the last cluster or not. ##
		## (ADDED in creation of 'summary_examine_complex.2.raw.pl'.) ##

   ##if (@rlv_flds>0) { # DISABLED for DEBUGGING. #

	my $combo_id = (@rlv_flds>0) ? join (':', @rlv_flds[0,1,2]) : undef; # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #
	##my $combo_id = "$id_class:$session:$segid";

    if (defined $combo_id) { # ADDED for DEBUGGING. #
	if ($last_combo_id eq $combo_id) { # Incorporate the current segment into the last cluster.
	    push @last_cluster, \@rlv_flds;
	    next;
	}
    } # ADDED for DEBUGGING. #
    ##} # DISABLED for DEBUGGING. #

	$ct_clusters++;

		## END of "Check if the current segment belongs to the last cluster or not." ##

		## Initialize the important attributes of the last cluster. ##
		## (ADDED in creation of 'summary_examine_complex.2.raw.pl'.) ##

	my ($id_class, $session, $segid, $type, $diagnosis, $ct_gpd_segs, $beg_coord_cls) = @{$last_cluster[0]}[0..5,7];
	my $end_coord_cls = $last_cluster[$#last_cluster]->[8];

	unless (@last_cluster == $ct_gpd_segs) { # A Consistency check. #

	    print STDOUT "[CAUTION '$program_name'] The size of the last cluster (=", scalar (@last_cluster), ") CLASHES with #{gapped segments in it} (=$ct_gpd_segs) for (ID_Class, Session_ID, Segment_ID) = ($id_class, $session, $segid) : $! ! ... I will skip the analysis ...\n\n";
	    $ct_anom3++;
	    if (@rlv_flds>0) {
		@last_cluster = (\@rlv_flds);
		$end_coord_last_cls = $end_coord_cls;
		($last_id_class, $last_session) = ($id_class, $session); # ADDED for DEBUGGING. #
	    } else {
		@last_cluster = ();
	    }
	    next;
	}		# END of "A Consistency check." #


=For the predecessor 'summary_examine_complex.2.raw.pl'. (ii)
			# Calculate the span of this cluster of gapped segments. #

	my $beg_coord_next_cls = (@rlv_flds>0) ? $rlv_flds[7] : undef;

	my $span_cluster;

	my ($id_class1, $session1) = @rlv_flds[0,1]; # ADDED for DEBUGGING. #

	if (($last_id_class eq $id_class) and ($id_class eq $id_class1) and ($last_session eq $session) and ($session eq $session1) and (defined $end_coord_last_cls) and (defined $beg_coord_next_cls)) { # Further MODIFIED for DEBUGGING. #
	##if (($id_class eq $id_class) and ($session eq $session1) and (defined $end_coord_last_cls) and (defined $beg_coord_next_cls)) { # MODIFIED for DEBUGGING. # ...}
	##if ((defined $end_coord_last_cls) and (defined $beg_coord_next_cls)) {...}
	
	   my $coord_left_border = ($beg_coord_cls + $end_coord_last_cls)/2 ;
	   my $coord_right_border = ($end_coord_cls + $beg_coord_next_cls)/2 ;
	   $span_cluster = $coord_right_border - $coord_left_border ;

	   if ($span_cluster < 0) { # Alarm if the span is negative. #
		print STDOUT "[WARNING '$program_name'] The cluster '$combo_id' has a NEGATIVE SPAN (=$span_cluster):  (end_last_cls, beg_cls, end_cls, beg_next_cls) = ($end_coord_last_cls, $beg_coord_cls, $end_coord_cls, $beg_coord_next_cls): $! !\n\n";
	   }
	}

			# END of "Calculate the span of this cluster of gapped segments." #

=cut # END of "For the predecessor 'summary_examine_complex.2.raw.pl'. (ii)"

		### ADDED in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

	my $max_totct_indel = 0;
	my $maxct_clms = my $max_totlen_indel = 0;
	my ($max_diffct_indel, $min_diffct_indel);

    foreach my $segment_info (@last_cluster) { # Main middle foreach-loop (over segments in the last cluster). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	my ($id_class1, $session1, $segid1, $type1, $diagnosis1, $ct_gpd_segs1, $indx_gpd_seg, $beg_coord, $end_coord, $ct_clms, $ct_ins, $ct_del, $totlen_ins, $totlen_del, $ct_gaps) = @{$segment_info}; # MOVED from above in creation of 'summary_examine_complex.2.raw.pl'. #

	my $totct_indel = $ct_ins + $ct_del;
	my $diffct_indel = $ct_ins - $ct_del;
	my $totlen_indel = $totlen_ins + $totlen_del;
	my $difflen_indel = $totlen_ins - $totlen_del;

	if ($totct_indel > $max_totct_indel) {

		# Update the set of "largest" segments. #

	    $max_totct_indel = $totct_indel;

	    $maxct_clms = $ct_clms;
	    $max_totlen_indel = $totlen_indel;
	    $max_diffct_indel = $min_diffct_indel = $diffct_indel;

	} elsif ($totct_indel == $max_totct_indel) {

		# Incorporate the current segment into the set of "largest" segments. # 

	    ($ct_clms > $maxct_clms) and ($maxct_clms = $ct_clms);
	    ($totlen_indel > $max_totlen_indel) and ($max_totlen_indel = $totlen_indel);
	    ((defined $max_diffct_indel) and ($max_diffct_indel >= $diffct_indel)) or ($max_diffct_indel = $diffct_indel);
	    ((defined $min_diffct_indel) and ($min_diffct_indel <= $diffct_indel)) or ($min_diffct_indel = $diffct_indel);

	}

   }			# End of the main middle foreach-loop (over segments in the last cluster). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

		## Update @last_cluster. ##
		## (ADDED in creation of 'summary_examine_complex.2.raw.pl'.)

	if (@rlv_flds>0) {
	    @last_cluster = (\@rlv_flds);
	    $end_coord_last_cls = $end_coord_cls;
	    ($last_id_class, $last_session) = ($id_class, $session); # ADDED for DEBUGGING. #
	} else {
	    @last_cluster = ();
	}
		## END of "Update @last_cluster." ##


=DISABLED in creation of 'summary_examine_purge_ctrl.raw.pl'.
	unless (($size_block == int ($size_block) ) and ($abs_shift == int ($abs_shift)) ) {
	    $ct_w_fractional++;
	    next;
	}
=cut

			### MAJOR MODIFICATION in creation of 'summary_examine_complex.2_3D.raw.pl'. (II) ###

			# Default: Type B = 'non-complex errors' + 'correctly reconstructed'. #

	my ($ctpair2bins_maxct_columns, $ctpair2bins_max_totlen_indels)
		= (\%ctpair2bins_maxct_columns_B,  \%ctpair2bins_max_totlen_indels_B);

	if ($type eq 'Correct') {

		# Do nothing. #

	} elsif ((defined $if_complex{$last_combo_id}) or ((defined $if_na{$last_combo_id}) or ($diagnosis =~ m/TOO_LONG/)) ) {

		# Type A = 'Complex errors' + 'Too-long' #

	    ($ctpair2bins_maxct_columns, $ctpair2bins_max_totlen_indels)
		= (\%ctpair2bins_maxct_columns_A,  \%ctpair2bins_max_totlen_indels_A);

	} elsif ($type eq 'Erroneous') {

		# Do nothing. #

	} else {
	    print STDOUT "[CAUTION '$program_name'] (ID_Class, Session_ID, Segment_ID) = ($id_class, $session, $segid) has either ANOMALOUS TYPE (=$type) or ANOMALOUS DIAGNOSIS (=$diagnosis): $! ! ... I will skip tallying this segment ...\n\n";

	    $ct_anom2++;
	    next;
	}

			# Take the 3D distributions. #

	my $tag_ctpair = sprintf ("(%d,%d)", $max_totct_indel, $max_diffct_indel);

	my $bins_maxct_columns = $ctpair2bins_maxct_columns->{$tag_ctpair};
	(defined $bins_maxct_columns) or ($bins_maxct_columns = $ctpair2bins_maxct_columns->{'Others'});

	my $bins_max_totlen_indels = $ctpair2bins_max_totlen_indels->{$tag_ctpair};
	(defined $bins_max_totlen_indels) or ($bins_max_totlen_indels = $ctpair2bins_max_totlen_indels->{'Others'});

	cluster_distributor ($maxct_clms, @{$bins_maxct_columns}, @BOUNDS_CT);
	cluster_distributor ($max_totlen_indel, @{$bins_max_totlen_indels}, @BOUNDS_CT);

			### RESTART FROM HERE ###

=Just a reminder ...

my @ctpairs = ( '(1,-1)', '(1,1)', '(2,-2)', '(2,0)', '(3,-3)', '(2,2)', '(3,-1)', '(3,1)', '(4,-2)', '(4,0)', 'Others');
my @tags_ct = ('Intervals \ X', @ctpairs);

my %ctpair2bins_maxct_columns_A; # = ($tag_ctpair => \@bins_maxct_columns_A, ...)
my %ctpair2bins_maxct_columns_B; # = ($tag_ctpair => \@bins_maxct_columns_B, ...)
my %ctpair2bins_max_totlen_indels_A; # = ($tag_ctpair => \@bins_max_totlen_indels_A, ...)
my %ctpair2bins_max_totlen_indels_B; # = ($tag_ctpair => \@bins_max_totlen_indels_B, ...)

my @batch_maxct_columns_A = ();
my @batch_maxct_columns_B = ();
my @batch_max_totlen_indels_A = ();
my @batch_max_totlen_indels_B = ();

foreach my $pair (@ctpairs) {

	my @bins_maxct_clms_A = ();
	my @bins_maxct_cmls_B = ();
	my @bins_max_totlen_indels_A = ();
	my @bins_max_totlen_indels_B = ();

	$ctpair2bins_maxct_columns_A{$pair} = \@bins_maxct_clms_A;
	$ctpair2bins_maxct_columns_B{$pair} = \@bins_maxct_clms_B;
	$ctpair2bins_max_totlen_indels_A{$pair} = \@bins_max_totlen_indels_A;
	$ctpair2bins_max_totlen_indels_B{$pair} = \@bins_max_totlen_indels_B;

	push @batch_maxct_columns_A, \@bins_maxct_clms_A;
	push @batch_maxct_columns_B, \@bins_maxct_clms_B;
	push @batch_max_totlen_indels_A, \@bins_max_totlen_indels_A;
	push @batch_max_totlen_indels_B, \@bins_max_totlen_indels_B;
}

batch_initializer (@batch_maxct_columns_A, @BOUNDS_CT);
batch_initializer (@batch_maxct_columns_B, @BOUNDS_CT);
batch_initializer (@batch_max_totlen_indels_A, @BOUNDS_CT);
batch_initializer (@batch_max_totlen_indels_B, @BOUNDS_CT);

=cut # END of "Just a reminder ..."

=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (II)

			### MAJOR ADDITION in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

			# Default: Type B = 'non-complex errors' + 'correctly reconstructed'. #

	my ($bins_max_totct_indels_vs_maxct_columns, $bins_max_totct_indels_vs_max_totlen_indels,
		$bins_max_totct_indels_vs_max_diffct_indels, $bins_max_totct_indels_vs_min_diffct_indels)
	   = ($bins_max_totct_indels_vs_maxct_columns_B, $bins_max_totct_indels_vs_max_totlen_indels_B,
		$bins_max_totct_indels_vs_max_diffct_indels_B, $bins_max_totct_indels_vs_min_diffct_indels_B) ;

	if ($type eq 'Correct') {

		# Do nothing. #

	} elsif ((defined $if_complex{$last_combo_id}) or ((defined $if_na{$last_combo_id}) or ($diagnosis =~ m/TOO_LONG/)) ) {

		# Type A = 'Complex errors' + 'Too-long' #

	    ($bins_max_totct_indels_vs_maxct_columns, $bins_max_totct_indels_vs_max_totlen_indels,
		$bins_max_totct_indels_vs_max_diffct_indels, $bins_max_totct_indels_vs_min_diffct_indels)
	     = ($bins_max_totct_indels_vs_maxct_columns_A, $bins_max_totct_indels_vs_max_totlen_indels_A,
		$bins_max_totct_indels_vs_max_diffct_indels_A, $bins_max_totct_indels_vs_min_diffct_indels_A) ;

	} elsif ($type eq 'Erroneous') {

		# Do nothing. #

	} else {
	    print STDOUT "[CAUTION '$program_name'] (ID_Class, Session_ID, Segment_ID) = ($id_class, $session, $segid) has either ANOMALOUS TYPE (=$type) or ANOMALOUS DIAGNOSIS (=$diagnosis): $! ! ... I will skip tallying this segment ...\n\n";

	    $ct_anom2++;
	    next;
	}

			# Take the 2D distributions. #

	distributor_2dim ($max_totct_indel, $maxct_clms, @{$bins_max_totct_indels_vs_maxct_columns}, @BOUNDS_CT0, @BOUNDS_CT);
	distributor_2dim ($max_totct_indel, $max_totlen_indel, @{$bins_max_totct_indels_vs_max_totlen_indels}, @BOUNDS_CT0, @BOUNDS_CT);
	distributor_2dim ($max_totct_indel, $max_diffct_indel, @{$bins_max_totct_indels_vs_max_diffct_indels}, @BOUNDS_CT0, @BOUNDS_PNCT0);
	distributor_2dim ($max_totct_indel, $min_diffct_indel, @{$bins_max_totct_indels_vs_min_diffct_indels}, @BOUNDS_CT0, @BOUNDS_PNCT0);



#=Just a reminder... 
(my $bins_max_totct_indels_vs_maxct_columns_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_maxct_columns_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_totlen_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_max_totlen_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_max_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

(my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

#=cut # END of "Just a reminder..."

=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (II)"


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
		# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.pl'. #
	if (($id_class eq '2_4') or ($id_class eq '3_4')) {
	    $ct_skipped1++; # Skip the entries belonging to these classes (to avoid redundancy). #
	    next;
	}
=cut


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
		# ADDED in creation of 'summary_classify_msa_errors.raw.pl'. #

	unless (($ct_equiv_indels ne 'undef') and (uc($ct_equiv_indels) ne 'N/A')) {
	    print STDOUT '[CAUTION] ($id_class, $session, $segment) = ', "($id_class, $session, $segment) has ", '$ct_equiv_indels = ', "$ct_equiv_indels: $! !\n\n";
	    $ct_skipped_x++;
	    next;
	}
=cut


	$ct_succ++;

=DISABLED in creation of 'summary_examine_complex.2.raw.pl'.
}		# End of outer while-loop (over erroneous MSA segments) #
=cut

	  ### ($ct_succ>=1000) and last; ### For TESTING the script. ###

}		# End of outer while-loop (until the last cluster runs out). # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #

close ($fh_in1);
##close ($fh_out); ## DISABLED in creation of 'summary_examine_purge.raw.pl'. ##

print STDOUT "\n\n=== ALL the Processes ENDED !! ===\n\n";

print STDOUT "#{Anomalous input lines}    = $ct_anom .\n"; # MOVED from below in creation of 'summary_examine_complex.1.raw.pl'.
print STDOUT "Total#{input Segments}         = $totct_segs .\n";
print STDOUT "   #{Segments w/ anomalous type or diagnosis} = $ct_anom2 .\n"; # ADDED in creation of 'summary_examine_complex.1.raw.pl'.
print STDOUT "   #{Segments w/ inconsistent #{gapped segments}} = $ct_anom3 .\n\n"; # ADDED in creation of 'summary_examine_complex.2.raw.pl'.

=DISABLED in creation of 'summary_examine_complex.1.raw.pl'.
print STDOUT "#{Skipped segments (to avoid redundancy)} = $ct_skipped1 .\n";

		# MODIFIED in creation of 'summary_examine_purge.raw.pl'. #

print STDOUT '#{Skipped segments (w/ $size_block or $abs_shift fractional)} = ', $ct_w_fractional, " .\n";
=cut

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
			# ADDED in creation of 'summary_classify_msa_errors.raw.pl'. #
print STDOUT '#{Skipped segments (w $ct_equiv_indels unavailable)} = ', $ct_skipped_x, " .\n";
print STDOUT "#{Anomalous input lines}    = $ct_anom .\n"; # MOVED from above in creation of 'summary_classify_msa_errors.raw.pl'.
print STDOUT "#{Input segments containing 'undef' differences} = $ct_w_undef .\n";
=cut

print STDOUT "   #{Input segments failed to be processed} = $ct_fail .\n";
print STDOUT "   #{Input segments successfully processed} = $ct_succ .\n\n";

unless (0 == $ct_fail) {
	print STDOUT "[ERROR '$program_name'] FAILED to process $ct_fail inputs !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}


		### MAJOR ADDITION in creation of 'summary_examine_complex.2_2D.raw.pl'. ###


print STDOUT "\n\n[ (i) ] Categories of the erroneous and correct segments:\n\n";
print STDOUT "Category A:   'Complex' errors + segments that are 'Too-Long' ,\n\n";
print STDOUT "Category B:   'Non-complex' errors + Correctly reconstructed segments .\n\n";

my $section = 0;

		### MAJOR MODIFICATION in creation of 'summary_examine_complex.2_3D.raw.pl'. (III) ###

##my @ctpairs = ( '(1,-1)', '(1,1)', '(2,-2)', '(2,0)', '(3,-3)', '(2,2)', '(3,-1)', '(3,1)', '(4,-2)', '(4,0)', 'Others');
##my @tags_ct = ('Intervals \ X', @ctpairs);

$section++;
print STDOUT "\n\n[$section] 3D-distributions of X = (Max#{indels}, Max(#{ins} - #{del}) (inferred with Dollo-parsimony, abscissa) and Max#{columns} (ordinates) for each of the segments of categories A and B:\n\n";

print STDOUT "{ Category A }  [ X = (Max#{indels}, Max(#{ins} - #{del}) ]\n\n";

batch_pr_file_distri_xl (@batch_maxct_columns_A, @BOUNDS_CT, 'Max#{columns}', *STDOUT, "%2d", @tags_ct);

print STDOUT "\n\n{ Category B }  [ X = (Max#{indels}, Max(#{ins} - #{del}) ]\n\n";

batch_pr_file_distri_xl (@batch_maxct_columns_B, @BOUNDS_CT, 'Max#{columns}', *STDOUT, "%2d", @tags_ct);

# batch_initializer (@batch_maxct_columns_A, @BOUNDS_CT);
# batch_initializer (@batch_maxct_columns_B, @BOUNDS_CT);


$section++;
print STDOUT "\n\n[$section] 3D-distributions of X = (Max#{indels}, Max(#{ins} - #{del}) (inferred with Dollo-parsimony, abscissa) and Max(Totlen{indels}) (ordinates) for each of the segments of categories A and B:\n\n";

print STDOUT "{ Category A }  [ X = (Max#{indels}, Max(#{ins} - #{del}) ]\n\n";

batch_pr_file_distri_xl (@batch_max_totlen_indels_A, @BOUNDS_CT, 'Max(Totlen{indels})', *STDOUT, "%2d", @tags_ct);

print STDOUT "\n\n{ Category B }  [ X = (Max#{indels}, Max(#{ins} - #{del}) ]\n\n";

batch_pr_file_distri_xl (@batch_max_totlen_indels_B, @BOUNDS_CT, 'Max(Totlen{indels})', *STDOUT, "%2d", @tags_ct);

# batch_initializer (@batch_max_totlen_indels_A, @BOUNDS_CT);
# batch_initializer (@batch_max_totlen_indels_B, @BOUNDS_CT);


=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (III)

$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and #{columns} (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_maxct_columns_A}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max#{columns}', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_maxct_columns_B}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max#{columns}', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_maxct_columns_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
# (my $bins_max_totct_indels_vs_maxct_columns_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and the total length of indels (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_totlen_indels_A}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max(Totlen{indels})', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_totlen_indels_B}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max(Totlen{indels})', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_max_totlen_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
# (my $bins_max_totct_indels_vs_max_totlen_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and Max(#{ins} - #{del}) (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_diffct_indels_A}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Max(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_diffct_indels_B}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Max(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_max_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
# (my $bins_max_totct_indels_vs_max_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and Min(#{ins} - #{del}) (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_min_diffct_indels_A}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Min(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_min_diffct_indels_B}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Min(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
# (my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (III)"

print STDOUT "\n\n";



print STDOUT "\n\n=== ENDED Printing Out the Statistics ===\n\n";



	### The DEFINITION(s) of the SUBROUTINE(s). ###

# broad_error_type:
#	This subroutine receives the name of an error-type,
#	and classifies it into a broad type.
#	(CREATED in creation of 'summary_classify_msa_errors.bulk_pars.pl' on Nov 17 (Mon), 2014.)
#
# INPUT: $type_error, which is the name of an error-type.
#
# OUTPUT: $br_type, which is the broad type accommodating $type_error.
#
sub broad_error_type ($) {

	my ($type_error) = @_;

	$type_error = uc($type_error);

	foreach my $brtype ('Shift', 'Complex', 'Merge', 'Purge', 'Split', 'Ex-nihilo') { # Main foreach-loop (over some broad error-types). #
	    if (substr ($type_error, 0, length ($brtype)) eq uc($brtype)) {
		return $brtype;
	    }
	} 				# End of the main foreach-loop (over some broad error-types). # 

	if ($type_error =~ m/^NEIGHBORING-BRANCH-MERGE/) {
	    return 'NB_Merge';

	} elsif ($type_error =~ m/^NEIGHBORING-BRANCH-SPLIT/) {
	    return 'NB_Split';

	} elsif ($type_error =~ m/^COLLAPSE-OF-INDEPENDENT-INSERTIONS/) {
	    return 'CII';

	} elsif ($type_error =~ m/^CREATION-OF-ERRONEOUS-INDEPENDENT-INSERTIONS/) {
	    return 'CEII';

	} elsif ($type_error =~ m/^INCOMPLETE-COLLAPSE-OF-INDEPENDENT-INSERTIONS/) {
	    return 'iCII';

	} elsif ($type_error =~ m/^INCOMPLETE-CREATION-OF-ERRONEOUS-INDEPENDENT-INSERTIONS/) {
	    return 'iCEII';

	} elsif ($type_error =~ m/^PARTIAL-NEIGHBORING-BRANCH-MERGE/) {
	    return 'pNB_Merge';

	} elsif ($type_error =~ m/^PARTIAL-NEIGHBORING-BRANCH-SPLIT/) {
	    return 'pNB_Split';

	} else {
	    return 'Other';
	}

		## @BROAD_TYPES = ('Shift', 'Complex', 'Merge', 'Purge', 'Split', 'Ex-nihilo', 'NB_Merge', 'NB_Split', 'CII', 'CEII', 'pNB_Merge', 'pNB_Split', 'iCII', 'iCEII', 'Other');

}		# End of the subroutine 'broad_error_type'. #


