package MyMalnComparison;
#
###################################################################################
#	
#	ComplLiMment_P: A package to calculate the Complete-likelihood for a given Multiple-sequence alignment
#			(almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.6.1: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyMalnComparison.pm:
#	This package contains subroutines that can be used
#	to compare multiple sequence alignments.
#	(CREATION STARTED on July 4 (Fri), 2014, ENDED on July xx (yyy), 2014)
#
#	[MODIFIED on Oct 7 (Wed), 2015]
#	A BUG was fixed in the subroutine 'exam_interfere_bw_gap_patterns'.
#

use strict;
use warnings;
use Exporter;

our @ISA = qw( Exporter);

our @EXPORT_OK = qw(
	tally_gaps_in_seqs_in_msa
	assign_resnums_gapchars_to_clms

	assign_resnums_to_clms
	raw_cmp_2msas_clmwise
	identify_gaponly_unmch_rgs
	mk_core_matched_segs

	map_shifts_respos_bw_2msas
	br_list_classes_shift_respos
	list_classes_shift_respos
	dpsm_infer_sep_patterns_for_classes

	br_search_new_mrcas
	fetch_ext_pres_offsprings_nr
	split_unnatural_miniclasses

	reorganize_classes_shift_respos
	merge_miniclasses_shift_respos
	merge_miniclasses_shift_respos_alpha
	find_trivial_blocks
	lump_together_similar_blocks_proto
	lump_together_similar_blocks
	exam_interfere_bw_gap_patterns
		);

use MyArrayManipu qw(
	diff_array_of_arrays
	);
	##union_array

use MyArrayManipu_num qw(
	union_nmarray
	diff_nmarray nr_nmarray
	);

use MyTreeManipu_II qw();
	## input_tree_file read_tree_file br_output_tree
	## make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree

use MyTreeMap_D qw(
	map_parsim_states_dollo_clm
	);
	## map_parsim_states_dollo_msamap_parsim_states_dollo_1clm

use MyTaxonomy qw(
	fetch_common_ancestors
	);

sub exam_interfere_bw_gap_patterns (\%\%$\%\%$);

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

	## Level of test output:
	## (0, silent; 1, very brief; 2, brief; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

	# For 'assign_resnums_gapchars_to_clms'.
our $TEST_LEVEL_as2 = 0;
##our $TEST_LEVEL_as2 = 1;

	# For 'assign_resnums_to_clms'.
our $TEST_LEVEL_as = 0;
##our $TEST_LEVEL_as = 1;

	# For 'raw_cmp_2msas_clmwise'.
our $TEST_LEVEL_rw = 0;
##our $TEST_LEVEL_rw = 1;

	# For 'identify_gaponly_unmch_rgs'.
our $TEST_LEVEL_id = 0;
##our $TEST_LEVEL_id = 1;

	# For 'mk_core_matched_segs'.
our $TEST_LEVEL_mk = 0;
##our $TEST_LEVEL_mk = 1;

	# For 'list_classes_shift_respos'.
our $TEST_LEVEL_lc = 0;
##our $TEST_LEVEL_lc = 1;

# tally_gaps_in_seqs_in_msa:
#	This subroutine counts runs of gaps in each of the sequences in an input MSA.
#	It assumes that the input MSA is flanked by gapless columns.
#	It leaves the input MSA intact.
#	(CREATED on Oct 7 (Tue), 2014)
#
# INPUT: (\@msa, $last_end_coord, $gap), where
#	$last_end_coord is the end coordinate of the last (i.e., left-flanking) region (in the 0-based half-open convention);
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	$gap is a character representing a gap.
#
# OUTPUT: \@set_gaps_in_seqs, where
#	@{$set_gaps_in_seqs[$k]->[$gp]} = ($beg_gap, $end_gap, $len_gap) is for the $gp th run of gaps in the $k th sequence, where
#	$beg/end_gap is the beginning/end-coordinate of the run of gap (in the 0-based half-open convention), and
#	$len_gap is the length of the run of gaps (in terms of the number of residues).
#
sub tally_gaps_in_seqs_in_msa (\@$$) {

	my ($msa, $last_end_coord, $gap) = @_;

	my $ct_seqs = @{$msa};
	unless ($ct_seqs>0) {
	    return [];
	}
	my $len_aln = length ${$msa->[0]};

	unless ($len_aln>0) {
	    my @out = ();
	    for (1..$ct_seqs) { push @out, []; }
	    return \@out;
	}

	$gap = uc ($gap);

	my @set_gaps_in_seqs = ();

	for (my $k=0; $k<$ct_seqs; $k++) { # Main outer for-loop (over sequences). #

	    my $cp_seq = uc ${$msa->[$k]};

	    my @gaps_in_seq = ();

	    for (my $s=0; $s<$len_aln; $s++) { # Main inner for-loop (over sites). #

		my $char = substr ($cp_seq, 0, 1, '');
		($char eq $gap) or next; # Skip unless the site is occupied by a gap.

		my $coord = $s + $last_end_coord; # Coordinate of the current site.

		unless (@gaps_in_seq>0) {
			# Create the first run of gaps. #
		   push @gaps_in_seq, [$coord, $coord+1, 1];
		   next;
		}

		my $last_gap = $gaps_in_seq[$#gaps_in_seq];
		my $end_last_gap = $last_gap->[1];

		if ($end_last_gap == $coord) {
			# Elongate the last run of gaps. #
		    $last_gap->[1]++;
		    $last_gap->[2]++;
		} else {
			# Create a new run of gaps. #
		   push @gaps_in_seq, [$coord, $coord+1, 1];
		}

	    }			# End of the main inner for-loop (over sites). #

	    $set_gaps_in_seqs[$k] = \@gaps_in_seq;

	}			# End of the main outer for-loop (over sequences). #

	return \@set_gaps_in_seqs;

}	# End of the subroutine 'tally_gaps_in_seqs_in_msa'. #


# assign_resnums_gapchars_to_clms:
#	This subroutine assigns each column the residue numbers (if the cell is occupied by a residue)
#	or the gap characters (if the cell is occupied by a gap).
#	It was created by modifying the subroutine 'assign_resnums_to_clms'.
#	(CREATED on July 11 (Fri), 2014)
#	
# assign_resnums_to_clms:
#	This satellite subroutine of 'raw_cmp_2msas_clmwise' receives an MSA,
#	and assigns each column the residue numbers (1-based) from the sequences embedded in the MSA.
#	(CREATED on Juy 4 (Fri), 2014)
#
# INPUT: (\@msa, $gap, $fl_clm_nos, $init_clm_number, \@last_resnums), where
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	$gap is a character representing a gap;
#	$fl_clm_nos = 1 if you want the column numbers (coordinates: 0-based) also be assigned (as the 0-th element),
#		    = 0 if you want only the residue numbers be assigned.
#	$init_clm_number is the column number to be assigned to the initial column (= undef if you don't care about it);
#	$last_resnums[$k] is the residue number of the last residue that the $k th sequence held upstream of this particular region
#	(= undef if you don't care about it).
#
# OUTPUT: \@set_clminfo, where
#
#	@{$set_clminfo[$c]} = ($column_coord, @residue_numbers) if $fl_clm_nos = 1,
#			    = @residue_numbers if $fl_clm_nos = 0,
#	for the $c th column.
#	Here $residue_numbers[$k] gives the residue number (1-based) of the $k th sequence in the MSA,
#	
#	[ and $residue_numbers[$k] = $r + 1/2 if the $k th row of the column is a gap between the $r th and the ($r+1) th residues of the $k th sequence (and = 1/2 if occupied by a gap on the left of the 1st residue). ] --- for the predecessor 'assign_resnums_to_clms'.
#	
sub assign_resnums_gapchars_to_clms (\@$$$\@) {

	my ($msa, $gap, $fl_clm_nos, $init_clm_number, $last_resnums) = @_;

	my $ct_seqs = @{$msa};

	unless ($ct_seqs>0) {
	    print STDOUT "[WARNING 'assign_resnums_gapchars_to_clms'] The input MSA consists of ZERO sequence: $! !\n";
	    return [];
	}

	my $len_msa = length ${$msa->[0]};

	unless ($len_msa>0) {
	    print STDOUT "[WARNING 'assign_resnums_gapchars_to_clms'] The input MSA consists of ZERO column: $! !\n";
	    return [];
	}

	(defined $init_clm_number) or ($init_clm_number = 0); # ADDED in creation of 'assign_resnums_gapchars_to_clms'. 

   (print STDOUT "  <  Subroutine 'assign_resnums_gapchars_to_clms' (#{sequences} = $ct_seqs; #{columns} = $len_msa; Gap = '$gap') >\n\n") if ($TEST_LEVEL_as2 > 0);

		# Initialize the output array. #
	my @set_clminfo = ();
	for (my $c=0; $c<$len_msa; $c++) { $set_clminfo[$c] = $fl_clm_nos ? [$init_clm_number + $c] : []; }
	##for (my $c=0; $c<$len_msa; $c++) { $set_clminfo[$c] = $fl_clm_nos ? [$c] : []; } # For the predecessor 'assign_resnums_to_clms'.


		## Assign residue numbers. ##

	for (my $k=0; $k<$ct_seqs; $k++) { # Main outer for-loop (over sequences). #

	    my $seq = ${$msa->[$k]}; # Copy the $k th sequence. #

	    my $resnum = (defined $last_resnums->[$k]) ? $last_resnums->[$k] : 0; # Initialize the residue-number (basically 1-based).
	    #my $resnum = 0; # Initialize the residue-number (1-based). # For the predecessor 'assign_resnums_to_clms'.

	    for (my $c=0; $c<$len_msa; $c++) { # Main inner for-loop (over MSA columns). #

		$resnum = int $resnum;
		my $char = substr ($seq, 0, 1, '');

			# MODIFIED in creation of 'assign_resnums_gapchars_to_clms'. #

		if ($char eq $gap) { # Assign a gap character if a gap occupies the site.
		    push @{$set_clminfo[$c]}, $gap;
		    next;
		}

		$resnum++; # Increment by 1 if a non-gap character resides.

		#$resnum += ($char eq $gap) ? 1/2 : 1; # Increment by 1 if a non-gap character resides, by 1/2 if a gap does. # For the predecessor 'assign_resnums_to_clms'.

		push @{$set_clminfo[$c]}, $resnum;

	    }			# End of the main inner for-loop (over MSA columns). #

	}			# End of the main outer for-loop (over sequences). #

	    
   (print STDOUT "  < End of the subroutine 'assign_resnums_gapchars_to_clms' >\n\n") if ($TEST_LEVEL_as2 > 0);

	return \@set_clminfo;

}	# End of the subroutine 'assign_resnums_gapchars_to_clms'. #


# assign_resnums_to_clms:
#	This satellite subroutine of 'raw_cmp_2msas_clmwise' receives an MSA,
#	and assigns each column the residue numbers (1-based) from the sequences embedded in the MSA.
#	(CREATED on Juy 4 (Fri), 2014)
#
# INPUT: (\@msa, $gap, $fl_clm_nos), where
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	$gap is a character representing a gap;
#	$fl_clm_nos = 1 if you want the column numbers (coordinates: 0-based) also be assigned (as the 0-th element),
#		    = 0 if you want only the residue numbers be assigned.
#
# OUTPUT: \@set_clminfo, where
#
#	@{$set_clminfo[$c]} = ($column_coord, @residue_numbers) if $fl_clm_nos = 1,
#			    = @residue_numbers if $fl_clm_nos = 0,
#	for the $c th column.
#	Here $residue_numbers[$k] gives the residue number (1-based) of the $k th sequence in the MSA,
#	and $residue_numbers[$k] = $r + 1/2 if the $k th row of the column is a gap between the $r th and the ($r+1) th residues of the $k th sequence (and = 1/2 if occupied by a gap on the left of the 1st residue).
#
sub assign_resnums_to_clms (\@$$) {

	my ($msa, $gap, $fl_clm_nos) = @_;

	my $ct_seqs = @{$msa};

	unless ($ct_seqs>0) {
	    print STDOUT "[WARNING 'assign_resnums_to_clms'] The input MSA consists of ZERO sequence: $! !\n";
	    return [];
	}

	my $len_msa = length ${$msa->[0]};

	unless ($len_msa>0) {
	    print STDOUT "[WARNING 'assign_resnums_to_clms'] The input MSA consists of ZERO column: $! !\n";
	    return [];
	}

   (print STDOUT "  <  Subroutine 'assign_resnums_to_clms' (#{sequences} = $ct_seqs; #{columns} = $len_msa; Gap = '$gap') >\n\n") if ($TEST_LEVEL_as > 0);

		# Initialize the output array. #
	my @set_clminfo = ();
	for (my $c=0; $c<$len_msa; $c++) { $set_clminfo[$c] = $fl_clm_nos ? [$c] : []; }


		## Assign residue numbers. ##

	for (my $k=0; $k<$ct_seqs; $k++) { # Main outer for-loop (over sequences). #

	    my $seq = ${$msa->[$k]}; # Copy the $k th sequence. #

	    my $resnum = 0; # Initialize the residue-number (1-based). #

	    for (my $c=0; $c<$len_msa; $c++) { # Main inner for-loop (over MSA columns). #

		$resnum = int $resnum;
		my $char = substr ($seq, 0, 1, '');
		$resnum += ($char eq $gap) ? 1/2 : 1; # Increment by 1 if a non-gap character resides, by 1/2 if a gap does.

		push @{$set_clminfo[$c]}, $resnum;
	    }			# End of the main inner for-loop (over MSA columns). #

	}			# End of the main outer for-loop (over sequences). #

	    
   (print STDOUT "  < End of the subroutine 'assign_resnums_to_clms' >\n\n") if ($TEST_LEVEL_as > 0);

	return \@set_clminfo;

}	# End of the subroutine 'assign_resnums_to_clms'. #



# raw_cmp_2msas_clmwise:
#	This subroutine compares two MSAs of the same set of sequences,
#	and returns the segments (or blocks of columns) in which the two MSAs perfectly matched,
#	as well as the segments that showed some differences between the two MSAs.
#	(CREATED on July 4 (Fri), 2014)
#
#	PREREQUISITE:
#	The two MSAs must consist of the identical regions of the same sequences,
#	and must arrange the sequences in the same order.
#
# INPUT: (\@msa1, \@msa2, $gap), where
#	${$msa1[$k]} is the $k th sequence embedded in the 1st MSA;
#	${$msa2[$k]} is the $k th sequence embedded in the 2nd MSA.
#	(${$msa1[$k]} and ${$msa2[$k]} must the the identical sequence,
#	possibly with different gap placements.)
#	$gap is a character representing a gap.
#
# OUTPUT: (\@matches1, \@matches2, \@ext1, \@ext2, \@summary_matches, \@summary_diffs), where
#
#	@column_info = @{$matches1/2[$sg]->[$c]} gives information on the $c th column
#	in the $sg th segment in the 1st/2nd MSA, where the two MSA perfectly matched;
#
#	@column_info = @{$ext1/2[$sg]->[$c]} gives information on the $c th column
#	in the region that MSA 1/2 has, but MSA 2/1 does not, between the ($sg-1) th and $sg th segments where both MSAs matched.
#
#	In the above descriptions,
#	@column_info = ($column_pos, @residue_numbers), where
#	$column_pos starts with 0 and ends with {MSA length} -1,
#	$residue_numbers[$k] is the residue number (1-based) of the $k th sequence in the respective MSA,
#	and $residue_numbers[$k] = $r + 1/2 if the $k th row of the column is occupied by a gap between the $r th residue and the ($r+1) th residue (= 1/2 if occupied by a gap on the left of the 1st residue).
#
#	@{$summary_matches[$sg]} = ($start1, $end1, $start2, $end2), where
#	$start1/2 and $end1/2 are the start and end coordinates (or column numbers) of the $sg th segment
#	in the 1st/2nd MSA where both MSAs perfectly matched.
#	The coordinates are given in the 0-based half-open convention;
#
#	Similarly, @{$summary_diffs[$sg]} = ($start1, $end1, $start2, $end2)
#	represents the region in MSA 1/2 between the ($sg-1) th and $sg th segments where both MSAs perfectly matched.
#
sub raw_cmp_2msas_clmwise (\@\@$) {

	my ($msa1, $msa2, $gap) = @_;

	my $ct_seqs = @{$msa1};
	unless ($ct_seqs == @{$msa2}) {
	    my $ct_seqs2 = @{$msa2};
	    print STDOUT "[ERROR 'raw_cmp_2msas_clmwise'] The two input MSAs consist of DIFFERENT NUMBERS of sequences ($ct_seqs for MSA1, $ct_seqs2 for MSA2): $! !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}
	unless ($ct_seqs>0) {
	    print STDOUT "[WARNING 'raw_cmp_2msas_clmwise'] Each of MSAs 1 and 2 consists of ZERO sequences: $! !\n";
	    return ([], [], [], [], [], []);
	}

	my $len_msa1 = length ${$msa1->[0]};
	my $len_msa2 = length ${$msa2->[0]};

   (print STDOUT "<  Subroutine 'raw_cmp_2msas_clmwise' (#{sequences} = $ct_seqs; #{columns}_MSA1 = $len_msa1; #{columns}_MSA2 = $len_msa2; Gap = '$gap') >\n\n") if ($TEST_LEVEL_rw > 0);

	my $clminfo_msa1 = assign_resnums_to_clms (@{$msa1}, $gap, 1); # The value 1 of the last argument specifies that each column information should start with the column number (coordinate) as the 0 th element, which should be followed by the residue numbers.
	unless (defined $clminfo_msa1) {
	    print STDOUT "[ERROR 'raw_cmp_2msas_clmwise'] The subroutine 'assign_resnums_to_clms' FAILED on the 1st MSA: $! !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

   if ($TEST_LEVEL_rw>3) {
	print STDOUT "[ Residue numbers in the 1st MSA (with $len_msa1 columns) ]\n\n";

	my @labels = ('Column_coord');
	foreach my $k (1..$ct_seqs) { push @labels, "seq$k"; }
	print STDOUT join ("\t", @labels), "\n\n";

	for (my $c=0; $c<$len_msa1; $c++) { print STDOUT join ("\t", @{$clminfo_msa1->[$c]}), "\n"; }
	print STDOUT "\n\n";
   }

	my $clminfo_msa2 = assign_resnums_to_clms (@{$msa2}, $gap, 1);
	unless (defined $clminfo_msa2) {
	    print STDOUT "[ERROR 'raw_cmp_2msas_clmwise'] The subroutine 'assign_resnums_to_clms' FAILED on the 2nd MSA: $! !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

   if ($TEST_LEVEL_rw>3) {
	print STDOUT "[ Residue numbers in the 2nd MSA (with $len_msa2 columns) ]\n\n";

	my @labels = ('Column_coord');
	foreach my $k (1..$ct_seqs) { push @labels, "seq$k"; }
	print STDOUT join ("\t", @labels), "\n\n";

	for (my $c=0; $c<$len_msa2; $c++) { print STDOUT join ("\t", @{$clminfo_msa2->[$c]}), "\n"; }
	print STDOUT "\n\n";
   }


		## Compare the two MSAs column-by-column. ##

	my @indices_cmp = (1..$ct_seqs);
	my @methods_cmp = ();
	for (@indices_cmp) { push @methods_cmp, '+'; }

	my ($common_clms1, $common_clms2, $ext_clms1, $ext_clms2) = diff_array_of_arrays (@{$clminfo_msa1}, @{$clminfo_msa2}, @indices_cmp, @methods_cmp);
	unless (defined $common_clms1) {
	    print STDOUT "[ERROR 'raw_cmp_2msas_clmwise'] The subroutine 'diff_arrays_of_arrays' FAILED to compare the two sets of column information, one for MSA 1 and the other for MSA 2: $! !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}
   (print STDOUT "\n   ... 'diff_array_of_arrays' successfully finished on the column information sets for MSAs 1 and 2 ...\n\n") if ($TEST_LEVEL_rw>1);

   if ($TEST_LEVEL_rw>3) {

	my @labels = ('Column_coord');
	foreach my $k (1..$ct_seqs) { push @labels, "seq$k"; }

	print STDOUT "[ Results of 'diff_array_of_arrays' ]\n\n";
	print STDOUT "( Columns only in MSA 1 )\n\n";

	print STDOUT join ("\t", @labels), "\n\n";

	foreach my $clm_info (@{$ext_clms1}) { print STDOUT join ("\t", @{$clm_info}), "\n"; } 

	print STDOUT "\n\n( Columns only in MSA 2 )\n\n";

	print STDOUT join ("\t", @labels), "\n\n";

	foreach my $clm_info (@{$ext_clms2}) { print STDOUT join ("\t", @{$clm_info}), "\n"; } 

	print STDOUT "\n\n";
   }


		## Partition the classified MSA columns into segments of contiguous columns. ##

	my @matches1 = my @matches2 = my @ext1 = my @ext2 = ();

	while (@{$common_clms1}>0) { # The main while-loop (over columns that both MSAs contain.) #

	    my @mch_seg1 = (shift @{$common_clms1});
	    my @mch_seg2 = (shift @{$common_clms2});

	    my $curr_clm1_coord = $mch_seg1[0]->[0];
	    my $curr_clm2_coord = $mch_seg2[0]->[0];

	    my @ext_seg1 = my @ext_seg2 = ();

		# Put all remaining 'extra' columns in MSA 1 on the left of @mch_seg1 into @ext_seg1. #
	    while ((@{$ext_clms1}>0) and ($ext_clms1->[0][0] < $curr_clm1_coord)) { push @ext_seg1, (shift @{$ext_clms1}); }
	
		# Put all remaining 'extra' columns in MSA 2 on the left of @mch_seg2 into @ext_seg2. #
	    while ((@{$ext_clms2}>0) and ($ext_clms2->[0][0] < $curr_clm2_coord)) { push @ext_seg2, (shift @{$ext_clms2}); }

	    push @ext1, \@ext_seg1;
	    push @ext2, \@ext_seg2;

		# Try to extend the current "matched" segments. #
	    while (@{$common_clms1}>0) { # The 3rd main inner while-loop (over columns that both MSA contain.) #

		my $nxt_clm1_coord = $common_clms1->[0][0];
		my $nxt_clm2_coord = $common_clms2->[0][0];

		(($nxt_clm1_coord == $curr_clm1_coord+1) and ($nxt_clm2_coord == $curr_clm2_coord+1)) or last; # The columns do NOT extend smoothly in BOTH MSAs.

		push @mch_seg1, (shift @{$common_clms1});
		push @mch_seg2, (shift @{$common_clms2});

		$curr_clm1_coord++;
		$curr_clm2_coord++;

	    }			# End of the 3rd main inner while-loop (over columns that both MSA contain.) #

	    push @matches1, \@mch_seg1;
	    push @matches2, \@mch_seg2;

	}		# END of the main while-loop (over columns that both MSAs contain.) #

	   # Put the sets of all remaining 'extra' columns in MSAs 1 and 2 into @ext1 and @ext2, respectively.
	push @ext1, $ext_clms1;
	push @ext2, $ext_clms2;


		## Summarize the information on the segments. ##

	my @summary_matches = my @summary_diffs = ();

		# Ihnitialize @summary_diffs. #

	if (@{$ext1[0]} > 0) {
	    my $lm_ex1 = $ext1[0]; # The left-most 'extra' segment from MSA 1.
	    push @summary_diffs, [ $lm_ex1->[0][0], $lm_ex1->[$#{$lm_ex1}][0] +1];
	} else {
	    push @summary_diffs, [0, 0];
	}
	if (@{$ext2[0]}>0) {
	    my $lm_ex2 = $ext2[0]; # The left-most 'extra' segment from MSA 2.
	    push @{$summary_diffs[0]}, $lm_ex2->[0][0], $lm_ex2->[$#{$lm_ex2}][0] + 1;
	} else {
	    push @{$summary_diffs[0]}, 0, 0;
	}

		# Handle the bulk segments (up to the right-most ones). #

	my $ct_mch_segs = @matches1;

	for (my $sg=0; $sg<$ct_mch_segs; $sg++) { # Main for-loop (over segments of perfect matches). #

	    my $mch_seg1 = $matches1[$sg];
	    my $mch_seg2 = $matches2[$sg];

	    my ($start_mch1, $end_mch1) = ($mch_seg1->[0][0], $mch_seg1->[$#{$mch_seg1}][0] + 1);
	    my ($start_mch2, $end_mch2) = ($mch_seg2->[0][0], $mch_seg2->[$#{$mch_seg2}][0] + 1);

	    $summary_matches[$sg] = [$start_mch1, $end_mch1, $start_mch2, $end_mch2];

	    my $rf_ex1 = $ext1[$sg+1]; # The right-flanking 'extra' segment from MSA 1.
	    my $rf_ex2 = $ext2[$sg+1]; # The right-flanking 'extra' segment from MSA 2.

	    my ($start_rf_ex1, $end_rf_ex1) = (@{$rf_ex1}>0) ? ($rf_ex1->[0][0], $rf_ex1->[$#{$rf_ex1}][0]+1) : ($end_mch1, $end_mch1);
	    my ($start_rf_ex2, $end_rf_ex2) = (@{$rf_ex2}>0) ? ($rf_ex2->[0][0], $rf_ex2->[$#{$rf_ex2}][0]+1) : ($end_mch2, $end_mch2);

	    $summary_diffs[$sg+1] = [$start_rf_ex1, $end_rf_ex1, $start_rf_ex2, $end_rf_ex2]; 

	}			# End of the main for-loop (over segments of perfect matches). #


   (print STDOUT "<  END of the subroutine 'raw_cmp_2msas_clmwise' >\n\n") if ($TEST_LEVEL_rw > 0);

	return (\@matches1, \@matches2, \@ext1, \@ext2, \@summary_matches, \@summary_diffs) ;

}	# End of the subroutine 'raw_cmp_2msas_clmwise'. #


# identify_gaponly_unmch_rgs:
#	This subroutine searches for the UNmatched regions
#	in each of which one MSA is empty and the other MSA consists only of gaps.
#	(CREATED on July 7 (Mon), 2014)
#
# INPUT: (\@ext1, \@ext2), where
#
#	@column_info = @{$ext1/2[$sg]->[$c]} gives information on the $c th column
#	in the region that MSA 1/2 has, but MSA 2/1 does not, between the ($sg-1) th and $sg th segments where both MSAs matched.
#	Here, @column_info = ($column_pos, @residue_numbers), where
#	$column_pos starts with 0 and ends with {MSA length} -1,
#	$residue_numbers[$k] is the residue number (1-based) of the $k th sequence in the respective MSA,
#	and $residue_numbers[$k] = $r + 1/2 if the $k th row of the column is occupied by a gap between the $r th residue and the ($r+1) th residue (= 1/2 if occupied by a gap on the left of the 1st residue).
#
# OUTPUT: \@indices_gaponly_unmch_rgs, where
#	@indices_gaponly_unmch_rgs stores the indices (such as $sg above) of the regions
#	in each of which one MSA is empty and the other MSA consists only of gaps.
#
sub identify_gaponly_unmch_rgs (\@\@) {

	my ($ext1, $ext2) = @_;

	my $ct_unmch_rgs = @{$ext1};

	unless (@{$ext2} == $ct_unmch_rgs) {
	    my $ct_unmch_rgs2 = @{$ext2};
	    print STDOUT "[ERROR 'identify_gaponly_unmch_rgs'] The number of regions in MSA 1 (=$ct_unmch_rgs) CLASHes with that in MSA 2 (=ct_unmch_rgs2): $! !\n";
	    return undef;
	}

   (print STDOUT "< Subroutine 'identify_gaponly_unmch_rgs' (#{UNmatched regions} = $ct_unmch_rgs) >\n\n") if ($TEST_LEVEL_id>0);

	my @indices_gaponly_unmch_rgs = ();

	for (my $rg=0; $rg<$ct_unmch_rgs; $rg++) { # Main outer for-loop (over UNmatched regions). #

	    my $rg1 = $ext1->[$rg];
	    my $rg2 = $ext2->[$rg];

	    my $ct_clms1 = @{$rg1};
	    my $ct_clms2 = @{$rg2};

	    (($ct_clms1>0) and ($ct_clms2>0)) and next; # If neither MSA is empty, the region is not what we want.

	    my $ct_clms = $ct_clms1+$ct_clms2;

	    ($ct_clms>0) or next; # Skip if both MSAs are empty. (This likely occurs on each end of the MSA.)

	    my $ct_gaponly_clms = 0;

	    foreach my $clm_info (@{$rg1}, @{$rg2}) { # Main middle for-loop (over columns). #

		my $size_info = @{$clm_info};
		my $ct_gaps = 0;
		for (my $k=1; $k<$size_info; $k++) { # Main inner for-loop (over aligned sequences). #

		    my $res_num = $clm_info->[$k];
		    ($res_num == int ($res_num) ) and last; # A residue was found.
		    $ct_gaps++;
		}			# End of the main inner for-loop (over aligned sequences). #

		($ct_gaps < $size_info-1) and last; # The column is NOT gap-only.
		$ct_gaponly_clms++;
	    }				# End of the main middle for-loop (over columns). #

	    ($ct_gaponly_clms == $ct_clms) and (push @indices_gaponly_unmch_rgs, $rg);

	}			# End of the main outer for-loop (over UNmatched regions). #

   (print STDOUT "\n< END of the subroutine 'identify_gaponly_unmch_rgs' >\n\n") if ($TEST_LEVEL_id>0);

	return \@indices_gaponly_unmch_rgs;

}	# End of the subroutine 'identify_gaponly_unmch_rgs'. #


# mk_core_matched_segs:
#	This subroutine recieves a set of segments in each of which the two MSAs match,
#	then constructs "core" segments by removing contiguous gap-containing columns on each end of
#	the matched segment.
#
# INPUT: (\@matches1, \@matches2, $len_msa1, $len_msa2), where
#
#	@column_info = @{$matches1/2[$sg]->[$c]} gives information on the $c th column
#	in the $sg th segment in the 1st/2nd MSA, where the two MSA perfectly matched.
#	Here,
#	@column_info = ($column_pos, @residue_numbers), where
#	$column_pos starts with 0 and ends with {MSA length} -1,
#	$residue_numbers[$k] is the residue number (1-based) of the $k th sequence in the respective MSA,
#	and $residue_numbers[$k] = $r + 1/2 if the $k th row of the column is occupied by a gap between the $r th residue and the ($r+1) th residue (= 1/2 if occupied by a gap on the left of the 1st residue).
#
#	$len_msa1/2 is the length of (i.e., the number of columns in) the 1st/2nd MSA.
#
# OUTPUT: (\@core_matches1, \@core_matches2, \@summary_core_mch_segs, \@summary_inter_core_rgs), where
#
#	@column_info = @{$core_matches1/2[$csg]->[$c]} gives information on the $c th column
#	in the $csg th "core" segment in the 1st/2nd MSA where the two MSA perfectly matched.
#	Here, @column_info follows the same format as in the input argument.
#
#	@{$summary_core_mch_segs[$csg]} = ($start1, $end1, $start2, $end2, $indx_orig), where
#	$start1/2 and $end1/2 are the start and end coordinates (or column numbers) of the $csg th "core" segment
#	in the 1st/2nd MSA where both MSAs perfectly matched.
#	The coordinates are given in the 0-based half-open convention.
#	$indx_orig is the index of the original matched segment the core is extracted from (i.e., $sg).
#
#	Similarly, @{$summary_inter_core_rgs[$csg]} = ($start1, $end1, $start2, $end2, \@indices_orig_diffs)
#	represents the region in MSA 1/2 between the ($csg-1) th and $csg th "core" segments where both MSAs perfectly matched.
#	@indices_orig_diffs stores the indices of the original UNmatched regions contained in the inter-core region.
#
sub mk_core_matched_segs (\@\@$$) {

	my ($matches1, $matches2, $len_msa1, $len_msa2) = @_;

	my $ct_mch_rgs = @{$matches1};
	unless (@{$matches2} == $ct_mch_rgs) {
	    my $ct_mch_rgs2 = @{$matches2};
	    print STDOUT "[ERROR 'mk_core_matched_segs'] The number of matched regions in MSA 1 (=$ct_mch_rgs) CLASHES with that in MSA 2 (=$ct_mch_rgs2): $! !\n";
	    return (undef, undef, undef, undef);
	}

   (print STDOUT "< Subroutine 'mk_core_matched_segs' (#{columns in MSA 1} = $len_msa1; #{columns in MSA 2} = $len_msa2; #{matched regions} = $ct_mch_rgs) >\n\n") if ($TEST_LEVEL_mk>0);

		## Make "core" matched regions. ##

	my @core_matches1 = my @core_matches2 = my @summary_core_mch_segs = ();

	for (my $sg=0; $sg<$ct_mch_rgs; $sg++) { # 1st main for-loop (over matched regions). #

		# Make the copies of the $sg th matched segments. #
	    my @mch_seg1 = @{$matches1->[$sg]};
	    my @mch_seg2 = @{$matches2->[$sg]};

	    my $ct_clms = @mch_seg1;
	    unless (@mch_seg2 == $ct_clms) {
		my $ct_clms2 = @mch_seg2;
		print STDOUT "[ERROR 'mk_core_matched_segs'] The number of columns in MSA 1 (=$ct_clms) CLASHES with that in MSA 2 (=$ct_clms2) in the $sg th matched region: $! !\n";
		return (undef, undef, undef, undef);
	    }

		# Remove gap-containing columns on the left-end of the region. #

	    while (@mch_seg1>0) { # 1st middle while-loop (forward over columns). #

		my $clm1_info = $mch_seg1[0];
		my $clm2_info = $mch_seg2[0];

		my $size_info = @{$clm1_info};
		my $if_gaps = 0;

		for (my $k=1; $k<$size_info; $k++) { # Inner for-loop (over sequences). #
		    my $resnum1 = $clm1_info->[$k];
		    my $resnum2 = $clm2_info->[$k];
		    (($resnum1 == int ($resnum1)) and ($resnum2 == int ($resnum2))) and next;
		    $if_gaps++;
		    last;
		}			# End of the inner for-loop (over sequences). #

		$if_gaps or last; # End the loop if we encounter a pair of columns with NO gaps.

			# Remove the left-end column from each segment. #
		shift @mch_seg1;
		shift @mch_seg2;

	    }		# End of the 1st middle while-loop (forward over columns). #

		# Remove gap-containing columns on the right-end of the region. #

	    while (@mch_seg1>0) { # 1st middle while-loop (forward over columns). #

		my $clm1_info = $mch_seg1[$#mch_seg1];
		my $clm2_info = $mch_seg2[$#mch_seg2];

		my $size_info = @{$clm1_info};
		my $if_gaps = 0;

		for (my $k=1; $k<$size_info; $k++) { # Inner for-loop (over sequences). #
		    my $resnum1 = $clm1_info->[$k];
		    my $resnum2 = $clm2_info->[$k];
		    (($resnum1 == int ($resnum1)) and ($resnum2 == int ($resnum2))) and next;
		    $if_gaps++;
		    last;
		}			# End of the inner for-loop (over sequences). #

		$if_gaps or last; # End the loop if we encounter a pair of columns with NO gaps.

			# Remove the right-end column from each segment. #
		pop @mch_seg1;
		pop @mch_seg2;

	    }		# End of the 1st middle while-loop (forward over columns). #

	    (@mch_seg1>0) or next; # Skip if the segment becomes empty.

	    push @core_matches1, \@mch_seg1;
	    push @core_matches2, \@mch_seg2;

	    my ($start1, $end1) = ($mch_seg1[0]->[0], $mch_seg1[$#mch_seg1]->[0]+1);
	    my ($start2, $end2) = ($mch_seg2[0]->[0], $mch_seg2[$#mch_seg2]->[0]+1);

	    push @summary_core_mch_segs, [$start1, $end1, $start2, $end2, $sg];

	}			# End of the 1st main for-loop (over matched regions). #


		## Make the list of inter-core regions. ##

	my @summary_inter_core_rgs = ();

	my $last_core_mch_seg = [0, 0, 0, 0, -1];

	foreach my $curr_core_mch_seg (@summary_core_mch_segs) { # 2nd main foreach-loop (over core matched regions). #
	    my ($end1_lf, $end2_lf, $indx_lf) = @{$last_core_mch_seg}[1,3,4];
	    my ($start1_rf, $start2_rf, $indx_rf) = @{$curr_core_mch_seg}[0,2,4];

	    my @indices_org_diffs = ($indx_lf+1 .. $indx_rf);

	    push @summary_inter_core_rgs, [$end1_lf, $start1_rf, $end2_lf, $start2_rf, \@indices_org_diffs];

	    $last_core_mch_seg = $curr_core_mch_seg; # Update the last "core" matched region.

	}					# End of the 2nd main foreach-loop (over core matched regions). #

		# Calculate information on the right-most "inter-core" region. #

	my ($end1_lf, $end2_lf, $indx_lf) = @{$last_core_mch_seg}[1,3,4];
	my @indices_org_diffs = ($indx_lf+1 .. $ct_mch_rgs);
	push @summary_inter_core_rgs, [$end1_lf, $len_msa1, $end2_lf, $len_msa2, \@indices_org_diffs];

   (print STDOUT "\n< End of the subroutine 'mk_core_matched_segs' >\n\n") if ($TEST_LEVEL_mk>0);

	return (\@core_matches1, \@core_matches2, \@summary_core_mch_segs, \@summary_inter_core_rgs) ;

}	# End of the subroutine 'mk_core_matched_segs'. #


# map_shifts_respos_bw_2msas:
#	This subroutine receives a reference MSA segment and its reconstructed counterpart,
#	and maps shifts in the residue positions (from the reference to the reconstructed)
#	onto the reconstructed MSA.
#	The subroutine leaves the input MSAs intact.
#	(CREATED on Oct 10 (Fri), 2014)
#
# INPUT: (\@msa_ref, \@msa_rec, $gap), where
#	${$msa_ref/rec[$k]} is the $k th sequence embedded in the reference/reconstructed MSAs;
#	$gap is a character representing a gap.
#
# OUTPUT: (\@map_shifts_respos, $shift_lf, $shift_rf), where
#	$map_shifts_respos[$k]->[$s] is the shift in the residue position from the reference MSA to the reconstructed one
#	for the $s th site of the $k th sequence, in the reconstructed MSA.
#	It is $gap if occupied by a gap;
#
#	$shift_lf is the shift at the left-flanking column of the reconstructed MSA. It must be 0 (zero) by default;
#	$shift_rf is the shift at the right-flanking column of the reconstructed MSA. It must be (len(reconstructed) - len(reference)) by default.
#
sub map_shifts_respos_bw_2msas (\@\@$) {

	my ($msa_ref, $msa_rec, $gap) = @_;

	my $ct_seqs = @{$msa_ref};

	unless (@{$msa_rec} == $ct_seqs) {
	    my $ct_seqs2 = @{$msa_rec};
	    print STDOUT "[ERROR 'map_shifts_respos_bw_2msas'] ILLEGAL INPUT: The reconstructed MSA has $ct_seqs2 sequences, which CLASHES with $ct_seqs, the number of sequences in the reference MSA: $! !\n";
	    return (undef, undef, undef);
	}
	unless ($ct_seqs>0) {
	    print STDOUT "[WARNING 'map_shifts_respos_bw_2msas'] The MSAs have NO SEQUENCES! ... I will return an empty array...: $! !\n";
	    return ([], undef, undef);
	}

	$gap = uc $gap;

	my $lenaln_ref = length ${$msa_ref->[0]};
	my $lenaln_rec = length ${$msa_rec->[0]};

	my $shift_lf = 0;
	my $shift_rf = $lenaln_rec - $lenaln_ref;

			## Main Process: Map the shifts onto the reconstructed MSA. ##

	my @map_shifts_respos = ();

	for (my $k=0; $k<$ct_seqs; $k++) { # Main outer for-loop (over sequences). #

	    my $cp_seq_ref = uc (${$msa_ref->[$k]}) ;

	    my @respos_ref = (); # The $r th element = the position of the $r th residue in the reference MSA. (Both 0-based)

	    my $resnum = 0;

	    for (my $s=0; $s<$lenaln_ref; $s++) { # 1st main inner for-loop (over sites in the reference MSA). #

		my $char = substr ($cp_seq_ref, 0, 1, '');
		($char eq $gap) and next; # Skip a gap.
		$respos_ref[$resnum] = $s;
		$resnum++; # Update the residue number.

	    }				# End of the 1st main inner for-loop (over sites in the reference MSA). #

	    my $cp_seq_rec = uc (${$msa_rec->[$k]}) ;

	    my @row_shifts = (); # The $s th element = the shift at the $s th site in the reconstructed MSA.

	    $resnum = 0;
	    for (my $s=0; $s<$lenaln_rec; $s++) { # 2nd main inner for-loop (over sites in the reconstructed MSA). #

		my $char = substr ($cp_seq_rec, 0, 1, '');
		if ($char eq $gap) {
		   $row_shifts[$s] = $gap;
		   next;
		}
		my $pos_ref = $respos_ref[$resnum];

		$row_shifts[$s] = $s - $pos_ref; # The shift at the $s th site of the reconstructed MSA.

		$resnum++; # Update the residue number.

	    }				# End of the 2nd main inner for-loop (over sites in the reconstructed MSA). #

	    $map_shifts_respos[$k] = \@row_shifts;

	}			# End of the main outer for-loop (over sequences). #


	return (\@map_shifts_respos, $shift_lf, $shift_rf);

}	# End of the subroutine 'map_shifts_respos_bw_2msas'. #


# br_list_classes_shift_respos:
#	This 'stripped-off' version of 'list_classes_shift_respos' no longer
#	sorts the raw classes into non-conflicging classes.
#	(CREATED on Oct 16 (Thu), 2014)
#	[MODIFIED on Oct 22 (Wed), 2014] The subroutine now returns the set of mini-classes of position-shifts as well.
#	Each mini-class consists of single runs of the same position-shift, beginning and ending at the same sites,
#	and affecting one or more sequence(s).
#
# list_classes_shift_respos:
#	This subroutine receives the output of 'map_shifts_respos_bw_2msas',
#	and sorts the shifts of residue positions into classes,
#	each consisting of the same type effecting the same regions (in the MSA) of diffferent sequences.
#	(CREATED on Oct 11 (Sat), 2014.)
#
#	NOTE:
#	(i) Basically, the subroutine does NOT tally those runs of shifts
#	   with what is commoner between $shift_lf and $shift_rf (or $shift_lf if they tie).
#	(ii) Each run could accommodate gaps if they are completely included (i.e., do not skick out of the boundaries).
#
# INPUT: (\@map_shifts_respos, $shift_lf, $shift_rf, $gap), where
#	$map_shifts_respos[$k]->[$s] is the shift in the residue position from the reference MSA to the reconstructed one
#	for the $s th site of the $k th sequence, in the reconstructed MSA.
#	It is $gap if occupied by a gap;
#
#	$shift_lf is the shift at the left-flanking column of the reconstructed MSA. It must be 0 (zero) by default;
#	$shift_rf is the shift at the right-flanking column of the reconstructed MSA. It must be (len(reconstructed) - len(reference)) by default.
#
#	$gap is a character representing a gap.
#
# OUTPUT: (\@set_classes_shift_respos, \@set_miniclasses_shift_respos, $commoner_shift_flank), where
# 	[ (\@set_classes_shift_respos, $commoner_shift_flank), where ] --- Obsolete as of Oct 22 (Wed), 2014. 
#	[ @{$set_classes_shift_respos[$cl]} = ($shift, $beg_site, $end_site, \@excluded_regions, \@seqs_affected) ] --- Obsoltete as of Oct 13 (Mon), 2014.
#	@{$set_classes_shift_respos[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $cl th class of the run of shifts of residue positions, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	[ @{$excluded_regions[$ex]} = ($shift_ex, $beg_ex, $end_ex) is for the $ex th region excluded from the class, where 
#	$shift_ex is its shift value, $beg/end_ex is its beginning/ending site, ] --- Obsolete as of Oct 13 (Mon), 2014.
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@{$set_miniclasses_shift_respos[$mc]}, which follows the same format as the @{$set_classes_shift_respos[$cl}},
#	is for the $mc th mini-class. # ADDED on Oct 22 (Wed), 2014. 
#
#	$commoner_shift_flank is whichever is commoner between $shift_lf and $shift_rf (or $shift_lf if they tie).
#
sub br_list_classes_shift_respos (\@$$$) {

	my ($map_shifts_respos, $shift_lf, $shift_rf, $gap) = @_;

	$gap = uc $gap;

	my $ct_seqs = @{$map_shifts_respos};

	unless ($ct_seqs>0) {
	    return ([], $shift_lf); # Return an empty array. #
	}
	my $alnlen = @{$map_shifts_respos->[0]};

   (print STDOUT "\n< Subtoutine 'br_list_classes_shift_respos' (", '$shift_lf=', "$shift_lf ; ", '$shift_rf = ', "$shift_rf ; ", '$gap = ', "'$gap') >\n\n") if ($TEST_LEVEL_lc>0);

		## First, count the flanking shifts. ##

	my $freq_shift_lf = my $freq_shift_rf = 0;

	for (my $k=0; $k<$ct_seqs; $k++) { # Preliminary outer for-loop (over sequences). #
	    my $row_shifts = $map_shifts_respos->[$k];
	    for (my $s=0; $s<$alnlen; $s++) { # Preliminary inner for-loop (over sites). #
		my $shift = $row_shifts->[$s];
		($shift eq $gap) and next;
		if ($shift == $shift_lf) { $freq_shift_lf++;
		} elsif ($shift == $shift_rf) { $freq_shift_rf++;
		}
	    }			# End of the preliminary inner for-loop (over sites). #
	}		# End of the preliminary outer for-loop (over sequences). #

	my $commoner_shift_flank = ($freq_shift_lf >= $freq_shift_rf) ? $shift_lf : $shift_rf ;

   if ($TEST_LEVEL_lc>2) {
	print STDOUT "[ Results of preliminary process ]\n\n";
	print STDOUT "#{Sequences} = $ct_seqs ,\n";
	print STDOUT "#{Columns} = $alnlen ,\n";
	print STDOUT '$commoner_shift_flank = ', $commoner_shift_flank, " .\n\n\n";
   }


		## Main Process: List the classes of the runs of shifts.   ##

   (print STDOUT "{{{ Main Process: List the classes of the runs of shifts. }}}\n\n") if ($TEST_LEVEL_lc>0);

	my @set_classes_shift_respos = ();
	my %class2indx; # = ("$shift:$beg_site:$end_site:$beg_rg1,$end_rg1;$beg_rg2,$end_rg2;..." => {index in @set_classes_shift_respos}, ...)

			# ADDED on Oct 22 (Wed), 2014. (i) #
	my @set_miniclasses_shift_respos = ();
	my %miniclass2indx; # = ("$shift:$beg_site:$end_site" => {index in @set_miniclasses_shift_respos}, ...)
			# END of "ADDED on Oct 22 (Wed), 2014. (i)" #

	for (my $k=0; $k<$ct_seqs; $k++) { # Main outer for-loop (over sequences). #

	    my $row_shifts = $map_shifts_respos->[$k];

   (print STDOUT "{{ The $k th sequence }}\n\n") if ($TEST_LEVEL_lc>1);

	    	## 1st Step: List all runs of shifts. ##

   (print STDOUT "(( 1st Step: List all runs of shifts ))\n\n") if ($TEST_LEVEL_lc>1);

	    my @list_raw_runs = (); # element = [$shift, $beg_run, $end_run].

	    for (my $s=0; $s<$alnlen; $s++) { # 1st main middle for-loop (over sites). #

		my $shift = $row_shifts->[$s];
		($shift eq $gap) and next;

		unless (@list_raw_runs>0) {
			# Create the first run. #
		    push @list_raw_runs, [$shift, $s, $s];
		    ##push @list_raw_runs, [$shift, $beg_run, $end_run]; # A BUG fixed on Oct 13 (Mon), 2014.
		    next;
		}

		my $last_run = $list_raw_runs[$#list_raw_runs];

		if ($last_run->[0] == $shift) {
			# Elongate the run. #
		    $last_run->[2] = $s;
		    ##$last_run->[2] = $shift; # A BUG fixed on Oct 13 (Mon), 2014.
		} else {
			# Create a new run. #
		    push @list_raw_runs, [$shift, $s, $s];
		    ##push @list_raw_runs, [$shift, $beg_run, $end_run]; # A BUG fixed on Oct 13 (Mon), 2014.
		}
	    }			# End of the 1st middle inner for-loop (over sites). #

		# Remove the runs of $commoner_shift_flank on both ends. #
	    if ((@list_raw_runs>0) and ($list_raw_runs[0]->[0] == $commoner_shift_flank)) {
		shift @list_raw_runs;
	    }
	    if ((@list_raw_runs>0) and ($list_raw_runs[$#list_raw_runs]->[0] == $commoner_shift_flank)) {
		pop @list_raw_runs;
	    }

	    my $ct_raw_runs = @list_raw_runs;

   if ($TEST_LEVEL_lc>2) {
	print STDOUT "\n[ Primary output of the 1st step: Content of ", '@list_raw_runs', " (#{raw runs} = $ct_raw_runs) ]\n\n";
	print STDOUT join ("\t", 'Indx_run', 'Shift', 'Beg_run', 'End_run'), "\n\n";
	for (my $r=0; $r<$ct_raw_runs; $r++) { print STDOUT join ("\t", $r, @{$list_raw_runs[$r]}), "\n"; } 
	print STDOUT "\n\n";
   }

	    ($ct_raw_runs>0) or next; # If nothing is left, end the process on this sequence. #

	    my %shift2run_set; # = ($shift => [[indices_of_runs_with $shift], $beg_span, $end_span, $span_width], ...)

	    for (my $r=0; $r<$ct_raw_runs; $r++) { # Preliminary middle for-loop (over raw runs). #

		my ($shift, $beg_run, $end_run) = @{$list_raw_runs[$r]};
		my $run_set = $shift2run_set{$shift};

			# ADDED on Oct 22 (Wed), 2014. (ii) # 
		my $shift_intvl = "$shift:$beg_run:$end_run";
		my $indx_miniclass = $miniclass2indx{$shift_intvl};
		if (defined $indx_miniclass) {

		    push @{$set_miniclasses_shift_respos[$indx_miniclass]->[4]}, $k; 
# @{$set_(mini)classes_shift_respos[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)

		} else {
		    $indx_miniclass = $miniclass2indx{$shift_intvl} = scalar (@set_miniclasses_shift_respos);
		    $set_miniclasses_shift_respos[$indx_miniclass] = [$shift, $beg_run, $end_run, [[$beg_run, $end_run]], [$k] ];
		}
			# END of "ADDED on Oct 22 (Wed), 2014. (ii)" # 

		unless (defined $run_set) {
		    $shift2run_set{$shift} = [[$r], $beg_run, $end_run, $end_run - $beg_run + 1];
		    next;
		}
			# If the shift is already found. #
		my ($indices_runs, $beg_span) = @{$run_set}[0,1];
		push @{$indices_runs}, $r;
		$run_set->[2] = $end_run;
		$run_set->[3] = $end_run - $beg_span + 1;
	    }			# End of the preliminary middle for-loop (over raw runs). #

   if ($TEST_LEVEL_lc>2) {
	print STDOUT "\n[ Secondary output of the 1st step: Content of ", '%shift2run_set', " ]\n\n";

	print STDOUT join ("\t", 'Shift', 'Beg_span', 'End_span', 'Width', 'Indices,of,runs,with,shift'), "\n\n";

	foreach my $shift (sort {$a <=> $b} keys %shift2run_set) { # Testing foreach-loop (over the values of shift). #
	    my ($indices_runs, $beg_span, $end_span, $width) = @{$shift2run_set{$shift}};
	    my $cnct_indices = join (',', @{$indices_runs}); 

	    print STDOUT join ("\t", $shift, $beg_span, $end_span, $width, $cnct_indices), "\n";

	}				# End of the testing foreach-loop (over the values of shift). #
	print STDOUT "\n\n";
   }


		## 2nd Step: Put together runs with the same shift value, to create a "maximum-spanning" composite run. ##

    (print STDOUT "(( 2nd Step: Put together runs with the same shift value. ))\n\n") if ($TEST_LEVEL_lc>1);

	    my @set_maxspan_cmp_runs = (); # element = [$shift, $beg_span, $end_span, \@indices_raw_runs, $span_width], where @indices_raw_runs stores the indices of the raw runs that makes up this composite run.
	    ##my @set_non_conflicting_cmp_runs = (); # element follows the same format as that in @set_maxspan_cmp_runs. # For the predecessor 'list_classes_shift_respos'.

	    foreach my $shift (sort { $shift2run_set{$a}->[1] <=> $shift2run_set{$b}->[1] } keys %shift2run_set) { # 2nd main middle foreach-loop (over shift values). #

		my ($indices_raw_runs, $beg_span, $end_span, $span_width) = @{$shift2run_set{$shift}};

			# DISABLED in creation of 'br_list_classes_shift_respos'. #
		##if (@{$indices_raw_runs}==1) {
		##
		##    push @set_non_conflicting_cmp_runs, [$shift, $beg_span, $end_span, $indices_raw_runs, $span_width];
		##
		##} else {

		    push @set_maxspan_cmp_runs, [$shift, $beg_span, $end_span, $indices_raw_runs, $span_width];
		##}
		
	    }			# End of the 2nd main middle foreach-loop (over shift values). #

	    my $ct_maxspan = @set_maxspan_cmp_runs;

   if ($TEST_LEVEL_lc>2) {

	print STDOUT "[ Output of the 2nd step: Content of ", '@set_maxspan_cmp_runs', " (#{composite runs} = $ct_maxspan) ]\n\n";

	print STDOUT  join ("\t", 'Indx_cmp_run', 'Shift', 'Beg_span', 'End_span', 'Span_width', 'indices,of,constituent,raw,runs'), "\n\n";

	for (my $c=0; $c<$ct_maxspan; $c++) { # Testing for-loop (over composite runs). #

	    my ($shift, $beg_span, $end_span, $indices_raw_runs, $span_width) = @{$set_maxspan_cmp_runs[$c]};
	    my $cnct_indices = join (',', @{$indices_raw_runs});

	    print STDOUT join ("\t", $c, $shift, $beg_span, $end_span, $span_width, $cnct_indices), "\n";

	}			# End of the testing for-loop (over composite runs). #
	print STDOUT "\n\n";

   }

=DISABLED in creation of 'br_list_classes_shift_respos' (I).

		## 3rd Step: Create a set of mutually "non-conflicting" composite runs, ##
		## by decomposing composite runs each of which is interrupted by either end of a larger composite run. ##

   (print STDOUT "(( 3rd Step: Create a set of mutually \"non-conflicting\" composite runs. ))\n\n") if ($TEST_LEVEL_lc>1);

	    my @setset_to_be_examined = (@set_maxspan_cmp_runs>0) ? (\@set_maxspan_cmp_runs) : ();

	    my $round = 0;

	    while (@setset_to_be_examined>0) { # 3rd main middle while-loop (until the sets of composite runs to be examined run out). #

		$round++;

		my $ct_sets = @setset_to_be_examined;

    (print STDOUT ' ' x 3, "|| The $round th round (#{sets to be examined} = $ct_sets): ||\n\n") if ($TEST_LEVEL_lc>2);

		my @new_setset = ();

		my $set_no = 0;

		while (my $set_to_be_examined = shift @setset_to_be_examined) { # Main inner while-loop (over the set of runs to be examined). #

		    $set_no++;

    (print STDOUT ' ' x 3, "| Set no. $set_no |\n\n") if ($TEST_LEVEL_lc>2);

		    my $ct_cmp_runs = @{$set_to_be_examined};
		    unless ($ct_cmp_runs >1) {
				# If the set consists only of one composite run, the run will NEVER conflict with any other composite runs.
			push @set_non_conflicting_cmp_runs, @{$set_to_be_examined};
			next;
		    }

   if ($TEST_LEVEL_lc>=5) {
	my $ct_non_cnfl = @set_non_conflicting_cmp_runs;

	print STDOUT ' ' x 3, '[ Content of @set_non_conflicting_cmp_runs up to here', "(size = $ct_non_cnfl) ]\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx', 'shift', 'beg_span', 'end_span', 'width', 'indices,constituent,raw,runs'), "\n\n";

	for (my $c=0; $c<$ct_non_cnfl; $c++) { # Testing for-loop (over non-conflicting composite runs). #
	    my ($shift, $beg_span, $end_span, $indices_raw_runs, $width) = @{$set_non_conflicting_cmp_runs[$c]};
	    my $cnct_indices = join (',', @{$indices_raw_runs});

	    print STDOUT ' ' x 3, join ("\t", $c, $shift, $beg_span, $end_span, $width, $cnct_indices), "\n";
	}			# End of the testing for-loop (over non-conflicting composite runs). #
	print STDOUT "\n\n";
   }

		    my @std_set_to_be_examined = sort {$a->[1] <=> $b->[1]} @{$set_to_be_examined};

    if ($TEST_LEVEL_lc>3) {
	print STDOUT ' ' x 3, "[ Set of composite runs to be examined: ", '@std_set_to_be_examined', " (#{composite runs} = $ct_cmp_runs) ]\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx', 'shift', 'beg_span', 'end_span', 'width', 'indices,constituent,raw,runs'), "\n\n";

	for (my $c=0; $c<$ct_cmp_runs; $c++) { # Testing for-loop (over composite runs). #
	    my ($shift, $beg_span, $end_span, $indices_raw_runs, $width) = @{$std_set_to_be_examined[$c]};
	    my $cnct_indices = join (',', @{$indices_raw_runs});

	    print STDOUT ' ' x 3, join ("\t", $c, $shift, $beg_span, $end_span, $width, $cnct_indices), "\n";
	}			# End of the testing for-loop (over composite runs). #
	print STDOUT "\n\n";
    }

		    my $indx_max_run = 0;
		    my $max_width = $std_set_to_be_examined[0]->[4];

		    for (my $c=1; $c<$ct_cmp_runs; $c++) { # Preliminary inner-inner for-loop (over the composite runs but the 1st). #
			my $width = $std_set_to_be_examined[$c]->[4];
			($width > $max_width) or next;
			$indx_max_run = $c;
			$max_width = $width;
		    }				# End of the preliminary inner-inner for-loop (over the composite runs but the 1st). # 

		    my $max_comp_run = $std_set_to_be_examined[$indx_max_run];
		    push @set_non_conflicting_cmp_runs, $max_comp_run; # By definition, the "widest" composite run no longer conflicts with any others.

				# "Sort" the composite runs according to their positions relative to the "widest" composite run. #

		    my ($shift0, $beg_span0, $end_span0, $indices_raw_runs0, $span_width0) = @{$max_comp_run};
		    my ($indx_left_bound, $indx_right_bound) = @{$indices_raw_runs0}[0, $#{$indices_raw_runs0}];

   if ($TEST_LEVEL_lc>3) {

	print STDOUT ' ' x 3, "[ The 'widest' composite run: ]\n\n";
	print STDOUT ' ' x 3, '{Index in @std_set_to_be_examined} = ', "$indx_max_run .\n";
	print STDOUT ' ' x 3, "{Index of left boundary run} = $indx_left_bound ,\n";
	print STDOUT ' ' x 3, "{Index of right boundary run} = $indx_right_bound .\n\n\n";
   }

   (print STDOUT ' ' x 3, "## Main Inner Process: 'Sort' the composite runs according to their positions relative to the 'widest' composite run. ##\n\n") if ($TEST_LEVEL_lc>3);
		    my @runs_left = my @runs_within = my @runs_right = ();

		    for (my $c=0; $c<$ct_cmp_runs; $c++) { # Main inner-inner for-loop (over the composite runs). #

			($c == $indx_max_run) and next; # Skip the "widest" composite run. #

			my $to_be_examined = $std_set_to_be_examined[$c];
			my ($shift, $beg_span, $end_span, $indices_raw_runs, $span_width) = @{$to_be_examined};
			my ($indx_leftmost, $indx_rightmost) = @{$indices_raw_runs}[0, $#{$indices_raw_runs}];

   (print STDOUT ' ' x 5, "{ The $c th composite run ", '($indx_leftmost, $indx_rightmost) = ', "($indx_leftmost, $indx_rightmost) }\n\n") if ($TEST_LEVEL_lc>=5);

			my $to_accommodate;
			if ($indx_rightmost < $indx_left_bound) {

			    $to_accommodate = \@runs_left;
   (print STDOUT ' ' x 5, " ... It's on the LEFT of the 'widest' one ...\n\n") if ($TEST_LEVEL_lc>=5);

			} elsif ($indx_right_bound < $indx_leftmost) {

			    $to_accommodate  = \@runs_right;
   (print STDOUT ' ' x 5, " ... It's on the RIGHT of the 'widest' one ...\n\n") if ($TEST_LEVEL_lc>=5);

			} elsif ( ($indx_left_bound < $indx_leftmost) and ($indx_rightmost < $indx_right_bound) ) {

   (print STDOUT ' ' x 5, " ... It's WITHIN the 'widest' one ...\n\n") if ($TEST_LEVEL_lc>=5);
			    $to_accommodate = \@runs_within;
			}

			if (defined $to_accommodate) {
			    push @{$to_accommodate}, $to_be_examined;
			    next;
			}

				# Now, we need to split the composite run. #

   (print STDOUT ' ' x 5, " ... It STRADDLES either boundary of the 'widest' one ... will be SPLITTED ...\n\n") if ($TEST_LEVEL_lc>=5);

			my $indx_bound = ($indx_left_bound < $indx_leftmost) ? $indx_right_bound: $indx_left_bound;

   (print STDOUT ' ' x 5, "{Index of the boundary run to be used} = $indx_bound .\n\n") if ($TEST_LEVEL_lc>=5);

   (print STDOUT ' ' x 5, "{Indices to be partitioned} = (", join (', ', @{$indices_raw_runs}), ") .\n\n") if ($TEST_LEVEL_lc>5);

			my @indices_left = my @indices_right = (); 
			foreach my $indx (@{$indices_raw_runs}) { ($indx < $indx_bound) ? (push @indices_left, $indx) : (push @indices_right, $indx); }

   if ($TEST_LEVEL_lc>=5) {
	print STDOUT ' ' x 5, "[ The resulting two subsets of indices: ]\n\n";
	print STDOUT ' ' x 5, '@indices_left = (', join (', ', @indices_left), ") ,\n";
	print STDOUT ' ' x 5, '@indices_right = (', join (', ', @indices_right), ") .\n";
	print STDOUT "\n\n";
   }

			unless ((@indices_left>0) and (@indices_right>0)) {
			    print STDOUT "[ERROR 'br_list_classes_shift_respos'] IMPOSSIBLE!! The composite run consisting of (", join (', ', @{$indices_raw_runs}), ") COULD NOT be SPLIT into two subsets by the boundary run with the index=$indx_bound: $! !\n\n";
			    return (undef, undef);
			}

			my ($indx_ll, $indx_rl) = @indices_left[0, $#indices_left];
			my ($indx_lr, $indx_rr) = @indices_right[0, $#indices_right];

   if ($TEST_LEVEL_lc>5) {
	print STDOUT ' ' x 5, "Indices of (leftmost run, rightmost run) of the left subset  = ($indx_ll, $indx_rl) .\n";
	print STDOUT ' ' x 5, "Indices of (leftmost run, rightmost run) of the right subset = ($indx_lr, $indx_rr) .\n";
	print STDOUT "\n\n";
   }

			my $beg_leftspan = $list_raw_runs[$indx_ll]->[1];
			my $end_leftspan = $list_raw_runs[$indx_rl]->[2];
			my $beg_rightspan = $list_raw_runs[$indx_lr]->[1];
			my $end_rightspan = $list_raw_runs[$indx_rr]->[2];

 	# @list_raw_runs = (); # element = [$shift, $beg_run, $end_run].

   if ($TEST_LEVEL_lc>5) {
	print STDOUT ' ' x 5, "Site that (begins, ends) the left subset =  ($beg_leftspan, $end_leftspan) .\n";
	print STDOUT ' ' x 5, "Site that (begins, ends) the right subset = ($beg_rightspan, $end_rightspan) .\n";
	print STDOUT "\n\n";
   }

			my @left_sub_cmp = ($shift, $beg_leftspan, $end_leftspan, \@indices_left, $end_leftspan - $beg_leftspan + 1);
			my @right_sub_cmp = ($shift, $beg_rightspan, $end_rightspan, \@indices_right, $end_rightspan - $beg_rightspan + 1);

   (print STDOUT ' ' x 5, "... The left subset is ... ") if ($TEST_LEVEL_lc>=5);

			if (@indices_left==1) {
			    push @set_non_conflicting_cmp_runs, \@left_sub_cmp;
   (print STDOUT " NON-CONFLICTING any longer ...\n\n") if ($TEST_LEVEL_lc>=5);

			} elsif ($indx_bound == $indx_left_bound) {
			    push @runs_left, \@left_sub_cmp;
   (print STDOUT " to the LEFT of the 'widest' set ...\n\n") if ($TEST_LEVEL_lc>=5);
			} else {
			    push @runs_within, \@left_sub_cmp;
   (print STDOUT " WITHIN the 'widest' set ...\n\n") if ($TEST_LEVEL_lc>=5);
			}

   (print STDOUT ' ' x 5, "... The right subset is ... ") if ($TEST_LEVEL_lc>=5);

			if (@indices_right==1) {
			    push @set_non_conflicting_cmp_runs, \@right_sub_cmp;
   (print STDOUT " NON-CONFLICTING any longer ...\n\n") if ($TEST_LEVEL_lc>=5);

			} elsif ($indx_bound == $indx_left_bound) {
			    push @runs_within, \@right_sub_cmp;
   (print STDOUT " WITHIN the 'widest' set ...\n\n") if ($TEST_LEVEL_lc>=5);
			} else {
			    push @runs_right, \@right_sub_cmp;
   (print STDOUT " to the RIGHT of the 'widest' set ...\n\n") if ($TEST_LEVEL_lc>=5);
			}

   (print STDOUT "\n") if ($TEST_LEVEL_lc>=5);

		    }			# End of the main inner-inner for-loop (over the composite runs). #

   (print STDOUT ' ' x 3, "[ Results of 'sorting' the composite runs ]\n\n") if ($TEST_LEVEL_lc>3);

		         # Store the subsets, each of which resulted from "sorting" the examined composite runs, into appropriate storages. #

		    foreach my $new_set_info ([\@runs_left, '@runs_left'], [\@runs_within, '@runs_within'], [\@runs_right, '@runs_right'] ) { # Trailing inner-inner foreach-loop (over the 3 subsets of the original set of runs). #

			my ($new_set, $name) = @{$new_set_info};
			my $size_new_set = @{$new_set};

   if ($TEST_LEVEL_lc>3) {
	print STDOUT ' ' x 3, "| Content of $name (size = $size_new_set): |\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx', 'shift', 'beg_span', 'end_span', 'width', 'indices,constituent,raw,runs'), "\n\n";

	for (my $c=0; $c<$size_new_set; $c++) { # Testing for-loop (over composite runs in the new set). #
	    my ($shift, $beg_span, $end_span, $indices_raw_runs, $width) = @{$new_set->[$c]};
	    my $cnct_indices = join (',', @{$indices_raw_runs});

	    print STDOUT ' ' x 3, join ("\t", $c, $shift, $beg_span, $end_span, $width, $cnct_indices), "\n";
	}			# End of the testing for-loop (over composite runs in the new set). #
	print STDOUT "\n\n";
   }

			if ($size_new_set>1) {
			    push @new_setset, $new_set;
		        } elsif ($size_new_set>0) {
				# The new set consists only of a single composite run. It will no longer conflict with any other by definition. #
			    push @set_non_conflicting_cmp_runs, @{$new_set};
		        }
				# (Do nothing when @{$new_set} is empty.) #

		    }				# End of the trailing inner-inner foreach-loop (over the 3 subsets of the original set of runs). #

		}			# End of the main inner while-loop (over the set of runs to be examined). #

			# Update the set of sets. #

		@setset_to_be_examined = @new_setset;

	    }		# End of the 3rd main middle while-loop (until the sets of composite runs to be examined run out). #


   if ($TEST_LEVEL_lc>2) {
	my $ct_non_cnfl = @set_non_conflicting_cmp_runs;

	print STDOUT ' ' x 3, '[ FINAL Content of @set_non_conflicting_cmp_runs ', "(size = $ct_non_cnfl) ]\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx', 'shift', 'beg_span', 'end_span', 'width', 'indices,constituent,raw,runs'), "\n\n";

	for (my $c=0; $c<$ct_non_cnfl; $c++) { # Testing for-loop (over non-conflicting composite runs). #
	    my ($shift, $beg_span, $end_span, $indices_raw_runs, $width) = @{$set_non_conflicting_cmp_runs[$c]};
	    my $cnct_indices = join (',', @{$indices_raw_runs});

	    print STDOUT ' ' x 3, join ("\t", $c, $shift, $beg_span, $end_span, $width, $cnct_indices), "\n";
	}			# End of the testing for-loop (over non-conflicting composite runs). #
	print STDOUT "\n\n";
   }

=cut # END Of "DISABLED in creation of 'br_list_classes_shift_respos' (I)."

=DISABLED in creation of 'br_list_classes_shift_respos' (II).
			# FINAL STEP: Incorporate the non-conflicting composite runs for this particular sequence into the set for ALL sequences. #

	    my $ct_non_cnfl = @set_non_conflicting_cmp_runs;

   (print STDOUT "(( FINAL STEP: Incorporate the $ct_non_cnfl composite runs for this particular sequence into the set for ALL sequences. ))\n\n") if ($TEST_LEVEL_lc>1);


	    for (my $c=0; $c<$ct_non_cnfl; $c++) { # Final main middle for-loop (over the non-conflicting composite runs). #

   (print STDOUT ' ' x 3, "| The $c th composite run |\n\n") if ($TEST_LEVEL_lc>2);

		my ($shift, $beg_site, $end_site, $indices_raw_runs, $span_width) = @{$set_non_conflicting_cmp_runs[$c]};
	# @set_maxspan_cmp_runs = (); # element = [$shift, $beg_span, $end_span, \@indices_raw_runs, $span_width], where @indices_raw_runs stores the indices of the raw runs that makes up this composite run.

=cut # END of "DISABLED in creation of 'br_list_classes_shift_respos' (II)."


			# FINAL STEP: Incorporate the maximum-spanning composite runs for this particular sequence into the set for ALL sequences. #

	    # $ct_maxspan = @set_maxspan_cmp_runs;

   (print STDOUT "(( FINAL STEP: Incorporate the $ct_maxspan composite runs for this particular sequence into the set for ALL sequences. ))\n\n") if ($TEST_LEVEL_lc>1);

	    for (my $c=0; $c<$ct_maxspan; $c++) { # Final main middle for-loop (over the non-conflicting composite runs). #

   (print STDOUT ' ' x 3, "| The $c th composite run |\n\n") if ($TEST_LEVEL_lc>2);

		my ($shift, $beg_site, $end_site, $indices_raw_runs, $span_width) = @{$set_maxspan_cmp_runs[$c]};
	# @set_maxspan_cmp_runs = (); # element = [$shift, $beg_span, $end_span, \@indices_raw_runs, $span_width], where @indices_raw_runs stores the indices of the raw runs that makes up this composite run.

		my $key = "$shift:$beg_site:$end_site:";

	# %class2indx; # = ("$shift:$beg_site:$end_site:$beg_rg1,$end_rg1;$beg_rg2,$end_rg2;..." => {index in @set_classes_shift_respos}, ...)
		my @affected_regions = ();

		foreach my $indx (@{$indices_raw_runs}) { # 1st inner foreach-loop (over constituent runs). #

		    my ($beg_run, $end_run) = @{$list_raw_runs[$indx]}[1,2];
 	# @list_raw_runs = (); # element = [$shift, $beg_run, $end_run].

		    push @affected_regions, [$beg_run, $end_run];
	# @{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),

		   $key .= "$beg_run,$end_run;";
		}			# End of the 1st inner foreach-loop (over constituent runs). #

   (print STDOUT ' ' x 3, "Key = '$key' .\n\n") if ($TEST_LEVEL_lc>=5); 

		my $indx_class = $class2indx{$key};
		if (defined $indx_class) {
			# The class of shift of residue positions already exists. => Let the sequence join the class. #

		    my $class_shift_respos = $set_classes_shift_respos[$indx_class];
		    push @{$class_shift_respos->[4]}, $k;

   (print STDOUT ' ' x 3, " ... Incorporated into the $indx_class th class ...\n\n\n") if ($TEST_LEVEL_lc>3);

		} else {
			# Create a new class of shift of residue positions. #
		    $indx_class = scalar (@set_classes_shift_respos);
		    $class2indx{$key} = $indx_class;
		    $set_classes_shift_respos[$indx_class] = [$shift, $beg_site, $end_site, \@affected_regions, [$k] ];

   (print STDOUT ' ' x 3, " ... Newly created the $indx_class th class ...\n\n\n") if ($TEST_LEVEL_lc>3);
		}

	# @{$set_classes_shift_respos[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
	# @seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;

	    }			# End of the final main middle for-loop (over the non-conflicitng composite runs). #


=FROZEN in the middle of writing a set of commands.

	    my @list_merged_runs = (); element = [$shift, $beg_run, $end_run, \@excluded, \@indices_runs]

	    	# First, tally those shifts used in only single runs. #

	    foreach my $shift (keys %shift2run_set) { # 2nd main inner foreach-loop (over shift values). #
		my ($indices_runs, $beg_span, $end_span, $span_width) = @{$shift2runs_set{$shift}};
		(@{$indices_runs}>1) and next;

		push @list_merged_runs, [$shift, $beg_span, $end_span, [], $indices_runs];

		delete $shift2run_set{$shift};
	    }			# End of the 2nd main inner foreach-loop (over shift values). #

		# Then, tally those shifts each of which is used in two or more runs. #

	    while (0< scalar (keys %shift2run_set)) { # 3rd main inner while-loop (over shift values used in 2 or more runs). #
	    }			# End of the 3rd main inner while-loop (over shift values used in 2 or more runs). #

=cut # END of "FROZEN in the middle of writing a set of commands."


	}		# End of the main outer for-loop (over sequences). #


   (print STDOUT "\n< End of the Subtoutine 'list_classes_shift_respos' />\n\n") if ($TEST_LEVEL_lc>0);

			# MODIFIED on Oct 22 (Wed), 2014. #
	return (\@set_classes_shift_respos, \@set_miniclasses_shift_respos, $commoner_shift_flank);
	##return (\@set_classes_shift_respos, $commoner_shift_flank);

}	# End of the subroutine 'br_list_classes_shift_respos'. #


# list_classes_shift_respos:
#	This subroutine receives the output of 'map_shifts_respos_bw_2msas',
#	and sorts the shifts of residue positions into classes,
#	each consisting of the same type effecting the same regions (in the MSA) of diffferent sequences.
#	(CREATED on Oct 11 (Sat), 2014.)
#
#	NOTE:
#	(i) Basically, the subroutine does NOT tally those runs of shifts
#	   with what is commoner between $shift_lf and $shift_rf (or $shift_lf if they tie).
#	(ii) Each run could accommodate gaps if they are completely included (i.e., do not skick out of the boundaries).
#
# INPUT: (\@map_shifts_respos, $shift_lf, $shift_rf, $gap), where
#	$map_shifts_respos[$k]->[$s] is the shift in the residue position from the reference MSA to the reconstructed one
#	for the $s th site of the $k th sequence, in the reconstructed MSA.
#	It is $gap if occupied by a gap;
#
#	$shift_lf is the shift at the left-flanking column of the reconstructed MSA. It must be 0 (zero) by default;
#	$shift_rf is the shift at the right-flanking column of the reconstructed MSA. It must be (len(reconstructed) - len(reference)) by default.
#
#	$gap is a character representing a gap.
#
# OUTPUT: (\@set_classes_shift_respos, $commoner_shift_flank), where
#	[ @{$set_classes_shift_respos[$cl]} = ($shift, $beg_site, $end_site, \@excluded_regions, \@seqs_affected) ] --- Obsoltete as of Oct 13 (Mon), 2014.
#	@{$set_classes_shift_respos[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $cl th class of the run of shifts of residue positions, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	[ @{$excluded_regions[$ex]} = ($shift_ex, $beg_ex, $end_ex) is for the $ex th region excluded from the class, where 
#	$shift_ex is its shift value, $beg/end_ex is its beginning/ending site, ] --- Obsolete as of Oct 13 (Mon), 2014.
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	$commoner_shift_flank is whichever is commoner between $shift_lf and $shift_rf (or $shift_lf if they tie).
#
sub list_classes_shift_respos (\@$$$) {

	my ($map_shifts_respos, $shift_lf, $shift_rf, $gap) = @_;

	$gap = uc $gap;

	my $ct_seqs = @{$map_shifts_respos};

	unless ($ct_seqs>0) {
	    return ([], $shift_lf); # Return an empty array. #
	}
	my $alnlen = @{$map_shifts_respos->[0]};

   (print STDOUT "\n< Subtoutine 'list_classes_shift_respos' (", '$shift_lf=', "$shift_lf ; ", '$shift_rf = ', "$shift_rf ; ", '$gap = ', "'$gap') >\n\n") if ($TEST_LEVEL_lc>0);

		## First, count the flanking shifts. ##

	my $freq_shift_lf = my $freq_shift_rf = 0;

	for (my $k=0; $k<$ct_seqs; $k++) { # Preliminary outer for-loop (over sequences). #
	    my $row_shifts = $map_shifts_respos->[$k];
	    for (my $s=0; $s<$alnlen; $s++) { # Preliminary inner for-loop (over sites). #
		my $shift = $row_shifts->[$s];
		($shift eq $gap) and next;
		if ($shift == $shift_lf) { $freq_shift_lf++;
		} elsif ($shift == $shift_rf) { $freq_shift_rf++;
		}
	    }			# End of the preliminary inner for-loop (over sites). #
	}		# End of the preliminary outer for-loop (over sequences). #

	my $commoner_shift_flank = ($freq_shift_lf >= $freq_shift_rf) ? $shift_lf : $shift_rf ;

   if ($TEST_LEVEL_lc>2) {
	print STDOUT "[ Results of preliminary process ]\n\n";
	print STDOUT "#{Sequences} = $ct_seqs ,\n";
	print STDOUT "#{Columns} = $alnlen ,\n";
	print STDOUT '$commoner_shift_flank = ', $commoner_shift_flank, " .\n\n\n";
   }


		## Main Process: List the classes of the runs of shifts.   ##

   (print STDOUT "{{{ Main Process: List the classes of the runs of shifts. }}}\n\n") if ($TEST_LEVEL_lc>0);

	my @set_classes_shift_respos = ();
	my %class2indx; # = ("$shift:$beg_site:$end_site:$beg_rg1,$end_rg1;$beg_rg2,$end_rg2;..." => {index in @set_classes_shift_respos}, ...)

	for (my $k=0; $k<$ct_seqs; $k++) { # Main outer for-loop (over sequences). #

	    my $row_shifts = $map_shifts_respos->[$k];

   (print STDOUT "{{ The $k th sequence }}\n\n") if ($TEST_LEVEL_lc>1);

	    	## 1st Step: List all runs of shifts. ##

   (print STDOUT "(( 1st Step: List all runs of shifts ))\n\n") if ($TEST_LEVEL_lc>1);

	    my @list_raw_runs = (); # element = [$shift, $beg_run, $end_run].

	    for (my $s=0; $s<$alnlen; $s++) { # 1st main middle for-loop (over sites). #

		my $shift = $row_shifts->[$s];
		($shift eq $gap) and next;

		unless (@list_raw_runs>0) {
			# Create the first run. #
		    push @list_raw_runs, [$shift, $s, $s];
		    ##push @list_raw_runs, [$shift, $beg_run, $end_run]; # A BUG fixed on Oct 13 (Mon), 2014.
		    next;
		}

		my $last_run = $list_raw_runs[$#list_raw_runs];

		if ($last_run->[0] == $shift) {
			# Elongate the run. #
		    $last_run->[2] = $s;
		    ##$last_run->[2] = $shift; # A BUG fixed on Oct 13 (Mon), 2014.
		} else {
			# Create a new run. #
		    push @list_raw_runs, [$shift, $s, $s];
		    ##push @list_raw_runs, [$shift, $beg_run, $end_run]; # A BUG fixed on Oct 13 (Mon), 2014.
		}
	    }			# End of the 1st middle inner for-loop (over sites). #

		# Remove the runs of $commoner_shift_flank on both ends. #
	    if ((@list_raw_runs>0) and ($list_raw_runs[0]->[0] == $commoner_shift_flank)) {
		shift @list_raw_runs;
	    }
	    if ((@list_raw_runs>0) and ($list_raw_runs[$#list_raw_runs]->[0] == $commoner_shift_flank)) {
		pop @list_raw_runs;
	    }

	    my $ct_raw_runs = @list_raw_runs;

   if ($TEST_LEVEL_lc>2) {
	print STDOUT "\n[ Primary output of the 1st step: Content of ", '@list_raw_runs', " (#{raw runs} = $ct_raw_runs) ]\n\n";
	print STDOUT join ("\t", 'Indx_run', 'Shift', 'Beg_run', 'End_run'), "\n\n";
	for (my $r=0; $r<$ct_raw_runs; $r++) { print STDOUT join ("\t", $r, @{$list_raw_runs[$r]}), "\n"; } 
	print STDOUT "\n\n";
   }

	    ($ct_raw_runs>0) or next; # If nothing is left, end the process on this sequence. #

	    my %shift2run_set; # = ($shift => [[indices_of_runs_with $shift], $beg_span, $end_span, $span_width], ...)

	    for (my $r=0; $r<$ct_raw_runs; $r++) { # Preliminary middle for-loop (over raw runs). #

		my ($shift, $beg_run, $end_run) = @{$list_raw_runs[$r]};
		my $run_set = $shift2run_set{$shift};

		unless (defined $run_set) {
		    $shift2run_set{$shift} = [[$r], $beg_run, $end_run, $end_run - $beg_run + 1];
		    next;
		}
			# If the shift is already found. #
		my ($indices_runs, $beg_span) = @{$run_set}[0,1];
		push @{$indices_runs}, $r;
		$run_set->[2] = $end_run;
		$run_set->[3] = $end_run - $beg_span + 1;
	    }			# End of the preliminary middle for-loop (over raw runs). #

   if ($TEST_LEVEL_lc>2) {
	print STDOUT "\n[ Secondary output of the 1st step: Content of ", '%shift2run_set', " ]\n\n";

	print STDOUT join ("\t", 'Shift', 'Beg_span', 'End_span', 'Width', 'Indices,of,runs,with,shift'), "\n\n";

	foreach my $shift (sort {$a <=> $b} keys %shift2run_set) { # Testing foreach-loop (over the values of shift). #
	    my ($indices_runs, $beg_span, $end_span, $width) = @{$shift2run_set{$shift}};
	    my $cnct_indices = join (',', @{$indices_runs}); 

	    print STDOUT join ("\t", $shift, $beg_span, $end_span, $width, $cnct_indices), "\n";

	}				# End of the testing foreach-loop (over the values of shift). #
	print STDOUT "\n\n";
   }


		## 2nd Step: Put together runs with the same shift value, to create a "maximum-spanning" composite run. ##

    (print STDOUT "(( 2nd Step: Put together runs with the same shift value. ))\n\n") if ($TEST_LEVEL_lc>1);

	    my @set_maxspan_cmp_runs = (); # element = [$shift, $beg_span, $end_span, \@indices_raw_runs, $span_width], where @indices_raw_runs stores the indices of the raw runs that makes up this composite run.
	    my @set_non_conflicting_cmp_runs = (); # element follows the same format as that in @set_maxspan_cmp_runs.

	    foreach my $shift (sort { $shift2run_set{$a}->[1] <=> $shift2run_set{$b}->[1] } keys %shift2run_set) { # 2nd main middle foreach-loop (over shift values). #

		my ($indices_raw_runs, $beg_span, $end_span, $span_width) = @{$shift2run_set{$shift}};

		if (@{$indices_raw_runs}==1) {

		    push @set_non_conflicting_cmp_runs, [$shift, $beg_span, $end_span, $indices_raw_runs, $span_width];

		} else {

		    push @set_maxspan_cmp_runs, [$shift, $beg_span, $end_span, $indices_raw_runs, $span_width];
		}
		
	    }			# End of the 2nd main middle foreach-loop (over shift values). #

   if ($TEST_LEVEL_lc>2) {
	my $ct_maxspan = @set_maxspan_cmp_runs;

	print STDOUT "[ Output of the 2nd step: Content of ", '@set_maxspan_cmp_runs', " (#{composite runs} = $ct_maxspan) ]\n\n";

	print STDOUT  join ("\t", 'Indx_cmp_run', 'Shift', 'Beg_span', 'End_span', 'Span_width', 'indices,of,constituent,raw,runs'), "\n\n";

	for (my $c=0; $c<$ct_maxspan; $c++) { # Testing for-loop (over composite runs). #

	    my ($shift, $beg_span, $end_span, $indices_raw_runs, $span_width) = @{$set_maxspan_cmp_runs[$c]};
	    my $cnct_indices = join (',', @{$indices_raw_runs});

	    print STDOUT join ("\t", $c, $shift, $beg_span, $end_span, $span_width, $cnct_indices), "\n";

	}			# End of the testing for-loop (over composite runs). #
	print STDOUT "\n\n";

   }

		## 3rd Step: Create a set of mutually "non-conflicting" composite runs, ##
		## by decomposing composite runs each of which is interrupted by either end of a larger composite run. ##

   (print STDOUT "(( 3rd Step: Create a set of mutually \"non-conflicting\" composite runs. ))\n\n") if ($TEST_LEVEL_lc>1);

	    my @setset_to_be_examined = (@set_maxspan_cmp_runs>0) ? (\@set_maxspan_cmp_runs) : ();

	    my $round = 0;

	    while (@setset_to_be_examined>0) { # 3rd main middle while-loop (until the sets of composite runs to be examined run out). #

		$round++;

		my $ct_sets = @setset_to_be_examined;

    (print STDOUT ' ' x 3, "|| The $round th round (#{sets to be examined} = $ct_sets): ||\n\n") if ($TEST_LEVEL_lc>2);

		my @new_setset = ();

		my $set_no = 0;

		while (my $set_to_be_examined = shift @setset_to_be_examined) { # Main inner while-loop (over the set of runs to be examined). #

		    $set_no++;

    (print STDOUT ' ' x 3, "| Set no. $set_no |\n\n") if ($TEST_LEVEL_lc>2);

		    my $ct_cmp_runs = @{$set_to_be_examined};
		    unless ($ct_cmp_runs >1) {
				# If the set consists only of one composite run, the run will NEVER conflict with any other composite runs.
			push @set_non_conflicting_cmp_runs, @{$set_to_be_examined};
			next;
		    }

   if ($TEST_LEVEL_lc>=5) {
	my $ct_non_cnfl = @set_non_conflicting_cmp_runs;

	print STDOUT ' ' x 3, '[ Content of @set_non_conflicting_cmp_runs up to here', "(size = $ct_non_cnfl) ]\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx', 'shift', 'beg_span', 'end_span', 'width', 'indices,constituent,raw,runs'), "\n\n";

	for (my $c=0; $c<$ct_non_cnfl; $c++) { # Testing for-loop (over non-conflicting composite runs). #
	    my ($shift, $beg_span, $end_span, $indices_raw_runs, $width) = @{$set_non_conflicting_cmp_runs[$c]};
	    my $cnct_indices = join (',', @{$indices_raw_runs});

	    print STDOUT ' ' x 3, join ("\t", $c, $shift, $beg_span, $end_span, $width, $cnct_indices), "\n";
	}			# End of the testing for-loop (over non-conflicting composite runs). #
	print STDOUT "\n\n";
   }

		    my @std_set_to_be_examined = sort {$a->[1] <=> $b->[1]} @{$set_to_be_examined};

    if ($TEST_LEVEL_lc>3) {
	print STDOUT ' ' x 3, "[ Set of composite runs to be examined: ", '@std_set_to_be_examined', " (#{composite runs} = $ct_cmp_runs) ]\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx', 'shift', 'beg_span', 'end_span', 'width', 'indices,constituent,raw,runs'), "\n\n";

	for (my $c=0; $c<$ct_cmp_runs; $c++) { # Testing for-loop (over composite runs). #
	    my ($shift, $beg_span, $end_span, $indices_raw_runs, $width) = @{$std_set_to_be_examined[$c]};
	    my $cnct_indices = join (',', @{$indices_raw_runs});

	    print STDOUT ' ' x 3, join ("\t", $c, $shift, $beg_span, $end_span, $width, $cnct_indices), "\n";
	}			# End of the testing for-loop (over composite runs). #
	print STDOUT "\n\n";
    }

		    my $indx_max_run = 0;
		    my $max_width = $std_set_to_be_examined[0]->[4];

		    for (my $c=1; $c<$ct_cmp_runs; $c++) { # Preliminary inner-inner for-loop (over the composite runs but the 1st). #
			my $width = $std_set_to_be_examined[$c]->[4];
			($width > $max_width) or next;
			$indx_max_run = $c;
			$max_width = $width;
		    }				# End of the preliminary inner-inner for-loop (over the composite runs but the 1st). # 

		    my $max_comp_run = $std_set_to_be_examined[$indx_max_run];
		    push @set_non_conflicting_cmp_runs, $max_comp_run; # By definition, the "widest" composite run no longer conflicts with any others.

				# "Sort" the composite runs according to their positions relative to the "widest" composite run. #

		    my ($shift0, $beg_span0, $end_span0, $indices_raw_runs0, $span_width0) = @{$max_comp_run};
		    my ($indx_left_bound, $indx_right_bound) = @{$indices_raw_runs0}[0, $#{$indices_raw_runs0}];

   if ($TEST_LEVEL_lc>3) {

	print STDOUT ' ' x 3, "[ The 'widest' composite run: ]\n\n";
	print STDOUT ' ' x 3, '{Index in @std_set_to_be_examined} = ', "$indx_max_run .\n";
	print STDOUT ' ' x 3, "{Index of left boundary run} = $indx_left_bound ,\n";
	print STDOUT ' ' x 3, "{Index of right boundary run} = $indx_right_bound .\n\n\n";
   }

   (print STDOUT ' ' x 3, "## Main Inner Process: 'Sort' the composite runs according to their positions relative to the 'widest' composite run. ##\n\n") if ($TEST_LEVEL_lc>3);
		    my @runs_left = my @runs_within = my @runs_right = ();

		    for (my $c=0; $c<$ct_cmp_runs; $c++) { # Main inner-inner for-loop (over the composite runs). #

			($c == $indx_max_run) and next; # Skip the "widest" composite run. #

			my $to_be_examined = $std_set_to_be_examined[$c];
			my ($shift, $beg_span, $end_span, $indices_raw_runs, $span_width) = @{$to_be_examined};
			my ($indx_leftmost, $indx_rightmost) = @{$indices_raw_runs}[0, $#{$indices_raw_runs}];

   (print STDOUT ' ' x 5, "{ The $c th composite run ", '($indx_leftmost, $indx_rightmost) = ', "($indx_leftmost, $indx_rightmost) }\n\n") if ($TEST_LEVEL_lc>=5);

			my $to_accommodate;
			if ($indx_rightmost < $indx_left_bound) {

			    $to_accommodate = \@runs_left;
   (print STDOUT ' ' x 5, " ... It's on the LEFT of the 'widest' one ...\n\n") if ($TEST_LEVEL_lc>=5);

			} elsif ($indx_right_bound < $indx_leftmost) {

			    $to_accommodate  = \@runs_right;
   (print STDOUT ' ' x 5, " ... It's on the RIGHT of the 'widest' one ...\n\n") if ($TEST_LEVEL_lc>=5);

			} elsif ( ($indx_left_bound < $indx_leftmost) and ($indx_rightmost < $indx_right_bound) ) {

   (print STDOUT ' ' x 5, " ... It's WITHIN the 'widest' one ...\n\n") if ($TEST_LEVEL_lc>=5);
			    $to_accommodate = \@runs_within;
			}

			if (defined $to_accommodate) {
			    push @{$to_accommodate}, $to_be_examined;
			    next;
			}

				# Now, we need to split the composite run. #

   (print STDOUT ' ' x 5, " ... It STRADDLES either boundary of the 'widest' one ... will be SPLITTED ...\n\n") if ($TEST_LEVEL_lc>=5);

			my $indx_bound = ($indx_left_bound < $indx_leftmost) ? $indx_right_bound: $indx_left_bound;

   (print STDOUT ' ' x 5, "{Index of the boundary run to be used} = $indx_bound .\n\n") if ($TEST_LEVEL_lc>=5);

   (print STDOUT ' ' x 5, "{Indices to be partitioned} = (", join (', ', @{$indices_raw_runs}), ") .\n\n") if ($TEST_LEVEL_lc>5);

			my @indices_left = my @indices_right = (); 
			foreach my $indx (@{$indices_raw_runs}) { ($indx < $indx_bound) ? (push @indices_left, $indx) : (push @indices_right, $indx); }

   if ($TEST_LEVEL_lc>=5) {
	print STDOUT ' ' x 5, "[ The resulting two subsets of indices: ]\n\n";
	print STDOUT ' ' x 5, '@indices_left = (', join (', ', @indices_left), ") ,\n";
	print STDOUT ' ' x 5, '@indices_right = (', join (', ', @indices_right), ") .\n";
	print STDOUT "\n\n";
   }

			unless ((@indices_left>0) and (@indices_right>0)) {
			    print STDOUT "[ERROR 'list_classes_shift_respos'] IMPOSSIBLE!! The composite run consisting of (", join (', ', @{$indices_raw_runs}), ") COULD NOT be SPLIT into two subsets by the boundary run with the index=$indx_bound: $! !\n\n";
			    return (undef, undef);
			}

			my ($indx_ll, $indx_rl) = @indices_left[0, $#indices_left];
			my ($indx_lr, $indx_rr) = @indices_right[0, $#indices_right];

   if ($TEST_LEVEL_lc>5) {
	print STDOUT ' ' x 5, "Indices of (leftmost run, rightmost run) of the left subset  = ($indx_ll, $indx_rl) .\n";
	print STDOUT ' ' x 5, "Indices of (leftmost run, rightmost run) of the right subset = ($indx_lr, $indx_rr) .\n";
	print STDOUT "\n\n";
   }

			my $beg_leftspan = $list_raw_runs[$indx_ll]->[1];
			my $end_leftspan = $list_raw_runs[$indx_rl]->[2];
			my $beg_rightspan = $list_raw_runs[$indx_lr]->[1];
			my $end_rightspan = $list_raw_runs[$indx_rr]->[2];

 	# @list_raw_runs = (); # element = [$shift, $beg_run, $end_run].

   if ($TEST_LEVEL_lc>5) {
	print STDOUT ' ' x 5, "Site that (begins, ends) the left subset =  ($beg_leftspan, $end_leftspan) .\n";
	print STDOUT ' ' x 5, "Site that (begins, ends) the right subset = ($beg_rightspan, $end_rightspan) .\n";
	print STDOUT "\n\n";
   }

			my @left_sub_cmp = ($shift, $beg_leftspan, $end_leftspan, \@indices_left, $end_leftspan - $beg_leftspan + 1);
			my @right_sub_cmp = ($shift, $beg_rightspan, $end_rightspan, \@indices_right, $end_rightspan - $beg_rightspan + 1);

   (print STDOUT ' ' x 5, "... The left subset is ... ") if ($TEST_LEVEL_lc>=5);

			if (@indices_left==1) {
			    push @set_non_conflicting_cmp_runs, \@left_sub_cmp;
   (print STDOUT " NON-CONFLICTING any longer ...\n\n") if ($TEST_LEVEL_lc>=5);

			} elsif ($indx_bound == $indx_left_bound) {
			    push @runs_left, \@left_sub_cmp;
   (print STDOUT " to the LEFT of the 'widest' set ...\n\n") if ($TEST_LEVEL_lc>=5);
			} else {
			    push @runs_within, \@left_sub_cmp;
   (print STDOUT " WITHIN the 'widest' set ...\n\n") if ($TEST_LEVEL_lc>=5);
			}

   (print STDOUT ' ' x 5, "... The right subset is ... ") if ($TEST_LEVEL_lc>=5);

			if (@indices_right==1) {
			    push @set_non_conflicting_cmp_runs, \@right_sub_cmp;
   (print STDOUT " NON-CONFLICTING any longer ...\n\n") if ($TEST_LEVEL_lc>=5);

			} elsif ($indx_bound == $indx_left_bound) {
			    push @runs_within, \@right_sub_cmp;
   (print STDOUT " WITHIN the 'widest' set ...\n\n") if ($TEST_LEVEL_lc>=5);
			} else {
			    push @runs_right, \@right_sub_cmp;
   (print STDOUT " to the RIGHT of the 'widest' set ...\n\n") if ($TEST_LEVEL_lc>=5);
			}

   (print STDOUT "\n") if ($TEST_LEVEL_lc>=5);

		    }			# End of the main inner-inner for-loop (over the composite runs). #

   (print STDOUT ' ' x 3, "[ Results of 'sorting' the composite runs ]\n\n") if ($TEST_LEVEL_lc>3);

		         # Store the subsets, each of which resulted from "sorting" the examined composite runs, into appropriate storages. #

		    foreach my $new_set_info ([\@runs_left, '@runs_left'], [\@runs_within, '@runs_within'], [\@runs_right, '@runs_right'] ) { # Trailing inner-inner foreach-loop (over the 3 subsets of the original set of runs). #

			my ($new_set, $name) = @{$new_set_info};
			my $size_new_set = @{$new_set};

   if ($TEST_LEVEL_lc>3) {
	print STDOUT ' ' x 3, "| Content of $name (size = $size_new_set): |\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx', 'shift', 'beg_span', 'end_span', 'width', 'indices,constituent,raw,runs'), "\n\n";

	for (my $c=0; $c<$size_new_set; $c++) { # Testing for-loop (over composite runs in the new set). #
	    my ($shift, $beg_span, $end_span, $indices_raw_runs, $width) = @{$new_set->[$c]};
	    my $cnct_indices = join (',', @{$indices_raw_runs});

	    print STDOUT ' ' x 3, join ("\t", $c, $shift, $beg_span, $end_span, $width, $cnct_indices), "\n";
	}			# End of the testing for-loop (over composite runs in the new set). #
	print STDOUT "\n\n";
   }

			if ($size_new_set>1) {
			    push @new_setset, $new_set;
		        } elsif ($size_new_set>0) {
				# The new set consists only of a single composite run. It will no longer conflict with any other by definition. #
			    push @set_non_conflicting_cmp_runs, @{$new_set};
		        }
				# (Do nothing when @{$new_set} is empty.) #

		    }				# End of the trailing inner-inner foreach-loop (over the 3 subsets of the original set of runs). #

		}			# End of the main inner while-loop (over the set of runs to be examined). #

			# Update the set of sets. #

		@setset_to_be_examined = @new_setset;

	    }		# End of the 3rd main middle while-loop (until the sets of composite runs to be examined run out). #


   if ($TEST_LEVEL_lc>2) {
	my $ct_non_cnfl = @set_non_conflicting_cmp_runs;

	print STDOUT ' ' x 3, '[ FINAL Content of @set_non_conflicting_cmp_runs ', "(size = $ct_non_cnfl) ]\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx', 'shift', 'beg_span', 'end_span', 'width', 'indices,constituent,raw,runs'), "\n\n";

	for (my $c=0; $c<$ct_non_cnfl; $c++) { # Testing for-loop (over non-conflicting composite runs). #
	    my ($shift, $beg_span, $end_span, $indices_raw_runs, $width) = @{$set_non_conflicting_cmp_runs[$c]};
	    my $cnct_indices = join (',', @{$indices_raw_runs});

	    print STDOUT ' ' x 3, join ("\t", $c, $shift, $beg_span, $end_span, $width, $cnct_indices), "\n";
	}			# End of the testing for-loop (over non-conflicting composite runs). #
	print STDOUT "\n\n";
   }


			# FINAL STEP: Incorporate the non-conflicting composite runs for this particular sequence into the set for ALL sequences. #

	    my $ct_non_cnfl = @set_non_conflicting_cmp_runs;

   (print STDOUT "(( FINAL STEP: Incorporate the $ct_non_cnfl composite runs for this particular sequence into the set for ALL sequences. ))\n\n") if ($TEST_LEVEL_lc>1);

	    for (my $c=0; $c<$ct_non_cnfl; $c++) { # Final main middle for-loop (over the non-conflicting composite runs). #

   (print STDOUT ' ' x 3, "| The $c th composite run |\n\n") if ($TEST_LEVEL_lc>2);

		my ($shift, $beg_site, $end_site, $indices_raw_runs, $span_width) = @{$set_non_conflicting_cmp_runs[$c]};
	# @set_maxspan_cmp_runs = (); # element = [$shift, $beg_span, $end_span, \@indices_raw_runs, $span_width], where @indices_raw_runs stores the indices of the raw runs that makes up this composite run.

		my $key = "$shift:$beg_site:$end_site:";

	# %class2indx; # = ("$shift:$beg_site:$end_site:$beg_rg1,$end_rg1;$beg_rg2,$end_rg2;..." => {index in @set_classes_shift_respos}, ...)
		my @affected_regions = ();

		foreach my $indx (@{$indices_raw_runs}) { # 1st inner foreach-loop (over constituent runs). #

		    my ($beg_run, $end_run) = @{$list_raw_runs[$indx]}[1,2];
 	# @list_raw_runs = (); # element = [$shift, $beg_run, $end_run].

		    push @affected_regions, [$beg_run, $end_run];
	# @{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),

		   $key .= "$beg_run,$end_run;";
		}			# End of the 1st inner foreach-loop (over constituent runs). #

   (print STDOUT ' ' x 3, "Key = '$key' .\n\n") if ($TEST_LEVEL_lc>=5); 

		my $indx_class = $class2indx{$key};
		if (defined $indx_class) {
			# The class of shift of residue positions already exists. => Let the sequence join the class. #

		    my $class_shift_respos = $set_classes_shift_respos[$indx_class];
		    push @{$class_shift_respos->[4]}, $k;

   (print STDOUT ' ' x 3, " ... Incorporated into the $indx_class th class ...\n\n\n") if ($TEST_LEVEL_lc>3);

		} else {
			# Create a new class of shift of residue positions. #
		    $indx_class = scalar (@set_classes_shift_respos);
		    $class2indx{$key} = $indx_class;
		    $set_classes_shift_respos[$indx_class] = [$shift, $beg_site, $end_site, \@affected_regions, [$k] ];

   (print STDOUT ' ' x 3, " ... Newly created the $indx_class th class ...\n\n\n") if ($TEST_LEVEL_lc>3);
		}

	# @{$set_classes_shift_respos[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
	# @seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;

	    }			# End of the final main middle for-loop (over the non-conflicitng composite runs). #


=FROZEN in the middle of writing a set of commands.

	    my @list_merged_runs = (); element = [$shift, $beg_run, $end_run, \@excluded, \@indices_runs]

	    	# First, tally those shifts used in only single runs. #

	    foreach my $shift (keys %shift2run_set) { # 2nd main inner foreach-loop (over shift values). #
		my ($indices_runs, $beg_span, $end_span, $span_width) = @{$shift2runs_set{$shift}};
		(@{$indices_runs}>1) and next;

		push @list_merged_runs, [$shift, $beg_span, $end_span, [], $indices_runs];

		delete $shift2run_set{$shift};
	    }			# End of the 2nd main inner foreach-loop (over shift values). #

		# Then, tally those shifts each of which is used in two or more runs. #

	    while (0< scalar (keys %shift2run_set)) { # 3rd main inner while-loop (over shift values used in 2 or more runs). #
	    }			# End of the 3rd main inner while-loop (over shift values used in 2 or more runs). #

=cut # END of "FROZEN in the middle of writing a set of commands."


	}		# End of the main outer for-loop (over sequences). #


   (print STDOUT "\n< End of the Subtoutine 'list_classes_shift_respos' />\n\n") if ($TEST_LEVEL_lc>0);

	return (\@set_classes_shift_respos, $commoner_shift_flank);

}	# End of the subroutine 'list_classes_shift_respos'. #


# dpsm_infer_sep_patterns_for_classes:
#	This subroutine receives a set of classes of position shifts,
#	and, for each class, to infer two patterns separating the sequences affected by the position shift
#	via Dollo parsimony.
#	(CREATED on Oct 15 (Wed), 2014.)
#
#	[METHOD]
#	For each class, the subroutine tries two Dollo parsimony patterns,
#	one (with $pt == 0) by assigning the 'absence' state to the affected sequences,
#	and the other (with $pt == 1) by assigning the 'presence' state to them.
#
# INPUT: (\@set_classes_shift_respos, $indx_affseqs, \@seqids, \%name2id, \%node2ch, \%node2pa, \%node2depth, $char_abs, $char_pres), where
#	@{$set_classes_shift_respos[$cl]->[$indx_affseqs]} (or @{$set_classes_shift_respos[$cl]} itself if $indx_affseqs is undefined)
#	lists the indices (for @seqids) of the sequences affected by (i.e., accommodating) the run of shifts defining the $cl th class;
#	$seqids[$k] is the ID of the $k th sequence in the MSA (w.r.t. which the shifts of residue positions are defined);
#	%name2id = ($sequence_ID => $node_id, ...) for the exterior nodes
#	(and $name2id{$TOP_NODE} = ID of the top-node (i.e., root));
#	%node2ch = ($node_id => \@child_nodes, ...) for the interior nodes;
#	%node2pa = ($node_id => $parent_node, ...) for all nodes except root;
#	%node2depth = ($node_id => $absolute_depth, ...) for all nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	$char_pres/abs is the character representing a 'presence'/'absence' state.
#
# OUTPUT: \@set_sep_patterns_for_classes, where
#	@{$set_sep_patterns_for_classes[$cl]->[$pt]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $pt th pattern separating the sequences affected by the $cl th class of position shifts, where
#	%node2parsstat = ($node_id => [$state, $ct_pres_ext], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.
#
sub dpsm_infer_sep_patterns_for_classes (\@$\@\%\%\%\%$$) {

	my ($set_classes_shift_respos, $indx_affseqs, $seqids, $name2id, $node2ch, $node2pa, $node2depth, $char_abs, $char_pres) = @_;

	my $ct_classes = @{$set_classes_shift_respos};
	my $ct_seqs = @{$seqids};
	my $topnode_id = $name2id->{$TOP_NODE};

	my $fl_full_mode = 1; # If you want the full output.

	my @set_sep_patterns_for_classes = ();

	for (my $cl=0; $cl<$ct_classes; $cl++) { # Main outer for-loop (over classes of position shifts). #

	    my $class_shift_respos = $set_classes_shift_respos->[$cl];
	    my $indices_affected = (defined $indx_affseqs) ? $class_shift_respos->[$indx_affseqs] : $class_shift_respos ;

	    my %seqindx2affected; # = ($indx => 1, ...) for the affected sequences.
	    foreach my $indx (@{$indices_affected}) { $seqindx2affected{$indx} = 1; }

	    my %exnode2abs_pres1; # It assigns 'absence' to the affected sequences.
	    my %exnode2abs_pres2; # It assigns 'presence' to the affected sequences.

	    for (my $k=0; $k<$ct_seqs; $k++) { # Preliminary inner for-loop (over sequences (i.e., exterior nodes)). #

		my $seqid = $seqids->[$k];
		my $nodeid = $name2id->{$seqid};
		my $if_affected = (defined $seqindx2affected{$k}) ? 1 : 0;

		$exnode2abs_pres1{$nodeid} = ($if_affected) ? $char_abs : $char_pres;
		$exnode2abs_pres2{$nodeid} = ($if_affected) ? $char_pres : $char_abs;

	    }			# End of the preliminary inner for-loop (over sequences (i.e., exterior nodes)). #

	    my @sep_patterns_for_class = ();

	    my @set_extnode2abs_pres = (\%exnode2abs_pres1, \%exnode2abs_pres2);

	    ##my $pt = 0; # A BUG fixed on Oct 15 (Wed), 2014. #
	    for (my $pt=0; $pt<2; $pt++) { # Main inner for-loop (over different 'presence'/'absence' assignments). #

		my $exnode2abs_pres = $set_extnode2abs_pres[$pt];
		##my $exnode2abs_pres = $set_extnode2abs_pres[$pt++]; # A BUG fixed on Oct 15 (Wed), 2014. #

		my ($node2parsstat, $br2change) = map_parsim_states_dollo_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %{$exnode2abs_pres}, $char_abs, $char_pres, $fl_full_mode) ;
		unless (defined $node2parsstat) {
		    print STDOUT "[ERROR 'dpsm_infer_sep_patterns_for_classes'] 'map_parsim_states_dollo_clm' FAILED for the $pt th separation pattern of the $cl th class: $! !\n\n";
		    return undef;
		}

		my $ct_changes = scalar keys %{$br2change};

		$sep_patterns_for_class[$pt] = [$node2parsstat, $br2change, $ct_changes];

	    }			# End of the main inner for-loop (over different 'presence'/'absence' assignments). #

	    $set_sep_patterns_for_classes[$cl] = \@sep_patterns_for_class;

	}			# End of the main outer for-loop (over classes of position shifts). #


	return \@set_sep_patterns_for_classes;

}	# End of the subroutine 'dpsm_infer_sep_patterns_for_classes'. #


# br_search_new_mrcas:
#	This satellite subroutine of 'split_unnatural_miniclasses' briefly searches for
#	new MRCAS of the affected seqeunces, which would give more parsimonious 'indel' scenarios,
#	by assigning the 'absence' state to the OLD MRCA.
#	(CREATED on Oct 30 (Thu), 2014.)
#
# INPUT: (\%old_br2change, $char_abs, $topnode_id, \%node2ch), where
#
#	%old_br2change = ($branch_id => [$state_ue, $state_le], ...) is for the OLD branches separating the 
#	sequences with the 'presence' state from the rest;
#	$char_abs is the character representing the 'absence' state;
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2ch = ($node_id => \@child_nodes, ...) is for the interior nodes;
#
# OUTPUT: \@set_new_mrcas, which lists the node IDs of new MRCAs.
#
sub br_search_new_mrcas (\%$$\%) {

	my ($old_br2change, $char_abs, $topnode_id, $node2ch) = @_;

	my $old_mrca;
	foreach my $br (keys %{$old_br2change}) { # Preliminary foreach-loop (over separating branches). #
	    ($old_br2change->{$br}[0] eq $char_abs) or next; # Skip 'deletion's.
	    $old_mrca = $br;
	    last;
	}				# End of the preliminary foreach-loop (over separating branches). #
	(defined $old_mrca) or ($old_mrca = $topnode_id); # When there is no 'insertion'.

	my $ch_old_mrca = $node2ch->{$old_mrca};

	unless (defined $ch_old_mrca) {
		# This is actually impossible. #
	    print STDOUT "[ERROR 'br_search_new_mrcas'] ILLEGAL INPUT: The OLD MRCA (ID: $old_mrca) has NO CHILDREN: $! !!\n\n";
	    return undef;
	}

	my @set_new_mrcas = ();

	my @subjects = @{$ch_old_mrca};

	while (1) { # Outer while-loop. #

	    my @new_sbjcts = ();
	    foreach my $cand (@subjects) { # Middle foreach-loop (over MRCA candidates). #

		my $ch_cand = $node2ch->{$cand};
		unless (defined $ch_cand) { # $cand is an exterior node. #
		   push @set_new_mrcas, $cand;
		   next;
		}
		my @undeleted = ();

		foreach my $ch (@{$ch_cand}) { (defined $old_br2change->{$ch}) or (push @undeleted, $ch); }

		if (@undeleted >1) {
			# Two or more children are undeleted. #
		    push @set_new_mrcas, $cand;

		} elsif (@undeleted>0) {
			# 1 child is undeleted. => Look deeper.
		    push @new_sbjcts, @undeleted;
		}

	    }			# End of the middle foreach-loop (over MRCA candidates). #

	    (@new_sbjcts>0) or last; # No candidates to further examine. #

	    @subjects = @new_sbjcts;

	}	# End of the outer while-loop. #

	return \@set_new_mrcas;

}	# End of the subroutine 'br_search_new_mrcas'. #


# fetch_ext_pres_offsprings_nr:
#	This satellite subroutine of 'split_unnatural_miniclasses' returns a list consisting only of
#	exterior offsprings with the 'presence' state.
#	It was created by slightly modifying 'fetch_ext_offsprings_nr' 
#	(CREATED on Oct 30 (Thu), 2014.)
#
# INPUT: ($nodeid, \%node2ch, \%br2change), where
#
#	$nodeid is the input MRCA node of the nodes with the 'presence' state;
#	%node2ch = ($node_id => \@child_nodes, ...) is for the interior nodes;
#	%br2change = ($branch_id => [$state_ue, $state_le], ...) for the branches along which an indel event occur.
#	(In this subroutine, we should never encounter an 'insertion' event,
#	and thus we can assume that all events we encounter are 'deletion's.)
#
# OUTPUT: \@ext_pres_offsprings, which lists the node IDs of the exterior offsprings with the 'presence' state.
#
sub fetch_ext_pres_offsprings_nr ($\%\%) {

	my ($nodeid, $node2ch, $br2change) = @_;

	my @subjects = ($nodeid);

	my @ext_pres_offsprings = ();

	while (1) { # Main outer while-loop. #

	    my @new_sbjcts = ();
	    foreach my $nodeid (@subjects) { # Main inner foreach-loop (over current subjects). #

		(defined $br2change->{$nodeid}) and next; # Skip if the node is 'delete'd. (Its offsprings must also be 'delete'd.) #

		my $children = $node2ch->{$nodeid};
		if (defined $children) {
			# The node is interior. #
		    push @new_sbjcts, @{$children};
		} else {
			# The node is exterior. #
		    push @ext_pres_offsprings, $nodeid;
		}
	    }				# End of the main inner foreach-loop (over current subjects). #

	    (@new_sbjcts>0) or last; # No more nodes to examine.

	    @subjects = @new_sbjcts;

	}	# End of the main outer while-loop. #

	return \@ext_pres_offsprings;

}	# End of the subroutine 'fetch_ext_pres_offsprings_nr'. #


# split_unnatural_miniclasses:
#	This subroutine splits miniclasses of residue-position-shifts each of which consists of unnaturally remote sequences.
#	(CREATED on Oct 30 (Thu), 2014.)
#
# INPUT: (\@set_miniclasses_shift_respos, \@set_sep_patterns_for_miniclasses, $topnode_id, \%node2pa, \%node2ch, \%node2depth, $char_pres, $char_abs, \%node2seq_indx), where
#
#	@{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th mini-class of the run of shifts of residue positions, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@{$set_sep_patterns_for_miniclasses[$mc]->[$pt]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $pt th pattern separating the sequences affected by the $mc th mini-class of position shifts, where
#	%node2parsstat = ($node_id => [$state, $ct_pres_ext], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the tree;
#	%node2pa = ($node_id => $parent_id, ...) is for all nodes except the top-node (i.e., root);
#	%node2ch = ($node_id => \@child_nodes, ...) is for the interior nodes;
#	%node2depth = ($node_id => $depth, ...) is for all nodes;
#	$char_pres/abs is the character resresenting the 'presence'/'absence' state;
#
#	%node2seq_indx = ($node_id => $indx_seq, ...) is for the exterior node, with
#		$indx_seq specifying the corresponding sequence in the MSA.
#
# OUTPUT: (\@newset_miniclasses_shift_respos, \@newset_sep_patterns_for_miniclasses), where
#
#	@newset_miniclasses_shift_respos, which follows the same format as @set_miniclasses_shift_respos,
#	lists the new mini-classes resulting from this subroutine;
#
#	@newset_sep_patterns_for_miniclasses, which follows the same format as @set_sep_patterns_for_miniclasses,
#	records the separation patterns for the elements of @newset_miniclasses_shift_respos.
#
sub split_unnatural_miniclasses (\@\@$\%\%\%$$\%) {

	my ($set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses, $topnode_id, $node2pa, $node2ch, $node2depth, $char_pres, $char_abs, $node2seq_indx) = @_;

	my $fl_full_mode = 1; # If you want the full output.
	my $indx_pt = 1; # Always consider the involving sequences as 'present'.

	my $ct_miniclasses = @{$set_miniclasses_shift_respos};
	my %master_exnode2abs_pres;
	foreach my $exnode (keys %{$node2seq_indx}) { $master_exnode2abs_pres{$exnode} = $char_abs; }

	my @newset_miniclasses_shift_respos = my @newset_sep_patterns_for_miniclasses = (); 

	for (my $mc=0; $mc<$ct_miniclasses; $mc++) { # Main outer for-loop (over mini-classes). # 

	    my $miniclass = $set_miniclasses_shift_respos->[$mc];

		# A BUG fixed on Oct 30 (Thu), 2014. => RETURNED to the old command, because it did NOT rectify the problem. #
	    ##my $sep_pattern = $set_sep_patterns_for_miniclasses->[$mc][$indx_pt];
	    my $sep_pattern = $set_sep_patterns_for_miniclasses->[$mc];

	    my ($shift, $beg_site, $end_site, $affected_regions, $seqs_affected)  = @{$miniclass};

		# A BUG truely fixed on Oct 30 (Thu), 2014. #
	    my ($node2parsstat, $br2change, $ct_changes) = @{$sep_pattern->[$indx_pt]};
	    ##my ($node2parsstat, $br2change, $ct_changes) = @{$sep_pattern};

	    if ($ct_changes <= @{$seqs_affected}) { # Torelate the old mini-class. #
		push @newset_miniclasses_shift_respos, $miniclass;
		push @newset_sep_patterns_for_miniclasses, $sep_pattern;
		next;
	    }

	    my $set_new_mrcas = br_search_new_mrcas (%{$br2change}, $char_abs, $topnode_id, %{$node2ch});

	    unless (@{$set_new_mrcas}>1) {
		my $ct_new_mrcas = @{$set_new_mrcas};
		print STDOUT "[ERROR 'split_unnatural_miniclasses'] Somehow, the subroutine 'br_search_new_mrcas' on the $mc th mini-class resulted in $ct_new_mrcas new MRCAs(= ", join (', ', @{$set_new_mrcas}), "): $! !\n\n";
		return (undef, undef);
	    }

	    my @union_new_seqs_affected = ();
	    foreach my $new_mrca (@{$set_new_mrcas}) { # Main middle for-loop (over new MRCAs of affected sequences). #

		my $exoffs_pres = fetch_ext_pres_offsprings_nr ($new_mrca, %{$node2ch}, %{$br2change});

		my @new_seqs_affected = ();
		my %exnode2abs_pres = %master_exnode2abs_pres;
		foreach my $exoff (@{$exoffs_pres}) { # Preliminary inner for-loop (over exterior offsprings of $new_mrca). #
		    push @new_seqs_affected, $node2seq_indx->{$exoff};
		    $exnode2abs_pres{$exoff} = $char_pres;
		}				# End of the preliminary inner for-loop (over exterior offsprings of $new_mrca). #

		my ($new_node2parsstat, $new_br2change) = map_parsim_states_dollo_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2abs_pres, $char_abs, $char_pres, $fl_full_mode) ;
		unless (defined $new_node2parsstat) {
		    print STDOUT "[ERROR 'split_unnatural_miniclasses'] 'map_parsim_states_dollo_clm' FAILED for the new MRCA (ID: $new_mrca) for the $mc th mini-class: $! !\n\n";
		    return (undef, undef);
		}
		my $newct_changes = scalar keys %{$new_br2change};

		(@new_seqs_affected>1) and (@new_seqs_affected = sort {$a <=> $b} @new_seqs_affected);

		push @newset_miniclasses_shift_respos, [$shift, $beg_site, $end_site, $affected_regions, \@new_seqs_affected];

			# A BUG fixed on Oct 30 (Thu), 2013. #
		my @dummy = ();
		my @set_sep_pattern_pres = ($new_node2parsstat, $new_br2change, $newct_changes);
		push @newset_sep_patterns_for_miniclasses, [\@dummy, \@set_sep_pattern_pres];
		##push @newset_sep_patterns_for_miniclasses, [$new_node2parsstat, $new_br2change, $newct_changes];

		push @union_new_seqs_affected, @new_seqs_affected;

	    }				# End of the main middle for-loop (over new MRCAs of affected sequences). #

			# Check the consistency (just in case). #
	    my ($common, $ex1, $ex2) = diff_nmarray (@{$seqs_affected}, @union_new_seqs_affected);
	    unless ((@{$ex1}==0) and (@{$ex2}==0)) {
		print STDOUT "[ERROR 'split_unnatural_miniclasses'] The New set of affected sequences CLASHES with the original set of affected sequences: $! !!\n";
		print STDOUT "              (In the OLD set only) ", join (', ', @{$ex1}), " \;\n";
		print STDOUT "              (In the NEW set only) ", join (', ', @{$ex1}), " .\n\n";
		return (undef, undef);
	    }

	}			# End of the main outer for-loop (over mini-classes). #


	return (\@newset_miniclasses_shift_respos, \@newset_sep_patterns_for_miniclasses);

}	# End of the subroutine 'split_unnatural_miniclasses'. #


# reorganize_classes_shift_respos:
#	This subroutine receives the main output of 'list_classes_shift_respos'
#	and the main output of 'dpsm_infer_sep_patterns_for_classes',
#	and reorganizes the classes of residue-position-shifts into (usually larger) blocks of residue-position-shifts.
#	(CREATED on Oct 16 (Thu), 2014.)
#
# INPUT: (\@set_classes_shift_respos, \@set_sep_patterns_for_classes, \@seqids, \%name2id, \%id2name, \%node2ch, \%node2pa, \%node2depth, $char_abs, $char_pres), where
#
#	@{$set_classes_shift_respos[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $cl th class of the run of shifts of residue positions, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	$shift_ex is its shift value, $beg/end_ex is its beginning/ending site, ] --- Obsolete as of Oct 13 (Mon), 2014.
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@{$set_sep_patterns_for_classes[$cl]->[$pt]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $pt th pattern separating the sequences affected by the $cl th class of position shifts, where
#	%node2parsstat = ($node_id => [$state, $ct_pres_ext], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.
#	($pt = 0 is for the Dollo parsimonious changes by assigning the 'absence' state to the affected sequences,
#	$pt = 1 is for the changes by assigning the 'presence' state to them.)
#
#	$seqids[$k] is the ID of the $k th sequence in the MSA (w.r.t. which the shifts of residue positions are defined);
#	%name2id = ($sequence_ID => $node_id, ...) for the exterior nodes
#	(and $name2id{$TOP_NODE} = ID of the top-node (i.e., root));
#	%id2name = ($node_id => $sequence_ID, ...) for the exterior nodes;
#	%node2ch = ($node_id => \@child_nodes, ...) for the interior nodes;
#	%node2pa = ($node_id => $parent_node, ...) for all nodes except root;
#	%node2depth = ($node_id => $absolute_depth, ...) for all nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	$char_pres/abs is the character representing a 'presence'/'absence' state.
#
# OUTPUT: \@set_blocks_shift_respos, where
#	@{$set_blocks_shift_respos[$bl]} = ($shift, $beg_site, $end_site, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes)
#	is for the $bl th block of the run of shifts of residue positions, where
#	$shift is the value of shift characterizing the block,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the block,
#	$node_mrca is the node of the most recent common ancestor of the sequences involved,
#	@seqs_affected stores the (0-based) indices of the sequences involved with the block,
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
#	$ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
#	 @indices_src_classes stores the indices of the classes that are used to construct the block.
#
sub reorganize_classes_shift_respos (\@\@\@\%\%\%\%\%$$) {

	my ($set_classes_shift_respos, $set_sep_patterns_for_classes, $seqids, $name2id, $id2name, $node2ch, $node2pa, $node2depth, $char_abs, $char_pres) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ct_seqs = @{$seqids};

	my $ct_classes = @{$set_classes_shift_respos};
	unless (@{$set_sep_patterns_for_classes} == $ct_classes) {
	    my $size_set = @{$set_sep_patterns_for_classes};
	    print STDOUT "[ERROR 'reorganize_classes_shift_respos'] ILLEGAL INPUT: The size of ", '@{$set_sep_patterns_for_classes', " (=$size_set) CLASHES with the number of classes (=$ct_classes): $! !\n\n";
	    return undef;
	}

	my %shift2classes; # = ($shift => \@indices_classes, ...) 

	for (my $c=0; $c<$ct_classes; $c++) { # Preliminary for-loop (over classes). #

	    my $shift = $set_classes_shift_respos->[$c][0];

	    my $classes = $shift2classes{$shift};
	    (defined $classes) or ($classes = $shift2classes{$shift} = []);
	    push @{$classes}, $c;

	}		# End of the preliminary for-loop (over classes). #

	my @affected_states = ($char_abs, $char_pres); # States assigned to the affected sequences.


		### Main Process ###

	my @set_blocks_shift_respos = ();

	foreach my $shift (sort { $a <=> $b} keys %shift2classes) { # Main Outer foreach-loop (over shifts). #

	    my $classes = $shift2classes{$shift};

	    my @info_classes = (); # element = [$indx_class, $indx_sep_pattern, $ct_changes, $id_mrca, $depth_mrca, \@indices_aff_rgs], where $indx_sep_pattern is the index for the separating pattern (0 or 1) chosen to be used for defining the block.

		# The rule(s) according to which separation pattern should be chosen.
		# [ (i) If both need only a single separation branch each, choose the 1th pattern (i.e., regard the shift as a 'presence' state).
		# (ii) If the numbers of necessary branches differ by >=2, choose the pattern requiring less branches.
		# (iii) Otherwise, consider also smaller-magnitude patterns (determined by the smallest branch depth), and 
		#     which can fit more such smaller patterns, and choose that with less (by 2 or more):
		#  #{required branches} - #{smaller patterns that can fit the subject pattern}.
		# (iv) Otherwise, consider both possibilities. (Generally, however, the 1th pattern would be more natural, considering the coherence of the shift patterns.) ] -- became OBSOLETE as of Oct 16 (Thu), 2014.
		#
		# => Now, we can UNCONDITIONALLY choose the 1th separation pattern (i.e., regard the shift as a 'presence' state),
		#   given that the states in each class became very coherent with each other, thanks to the use of 'br_list_classes_shift_respos'.

	    foreach my $indx_class (@{$classes}) { # Preliminary middle foreach-loop (over classes with a particular shift). #
		my $indx_sep_pt = 1;
		my ($node2parsstat, $br2change, $ct_changes) = @{$set_sep_patterns_for_classes->[$indx_class][$indx_sep_pt]};
		
		# @{$set_sep_patterns_for_classes[$cl]->[$pt]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $pt th pattern separating the sequences affected by the $cl th class of position shifts, where
		# %br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
		# $ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.
		# ($pt = 0 is for the Dollo parsimonious changes by assigning the 'absence' state to the affected sequences,
		# $pt = 1 is for the changes by assigning the 'presence' state to them.)


		my ($id_mrca, $depth_mrca);
		my $state_assigned = $affected_states[$indx_sep_pt];

		foreach my $br (keys %{$br2change}) { # Preliminary inner foreach-loop (over branches separating the affected sequences). #
		    my ($state_ue, $state_le) = @{$br2change->{$br}};
		    ($state_le eq $state_assigned) or next;

		    my $depth_br = $node2depth->{$br};
		    ((defined $id_mrca) and ($depth_mrca <= $depth_br)) or ( ($id_mrca, $depth_mrca) = ($br, $depth_br) );

		}			# End of the preliminary inner foreach-loop (over branches separating the affected sequences). #
		(defined $id_mrca) or ( ($id_mrca, $depth_mrca) = ($topnode_id, $node2depth->{$topnode_id}) );

		my $affected_regions = $set_classes_shift_respos->[$indx_class][3];
		my @indices_rlv_rgs = (0 .. $#{$affected_regions});

		push @info_classes, [$indx_class, $indx_sep_pt, $ct_changes, $id_mrca, $depth_mrca, \@indices_rlv_rgs];

	    }			# End of the preliminary middle foreach-loop (over classes with a particular shift). #

	    my @set_to_be_examined = sort {$a->[4] <=> $b->[4]} @info_classes; # Sort the classes by the depths of their MRCAs.

	    my @sbset_blocks_shift_respos = ();

	    while (my $to_be_examined = shift @set_to_be_examined ) { # Main middle while-loop (until the examined classes run out). #

		my ($indx_class, $indx_sep_pt, $ct_changes, $id_mrca, $depth_mrca, $indices_rlv_rgs) = @{$to_be_examined};
		my $state_assigned = $affected_states[$indx_sep_pt];
		my $ct_rlv_rgs = @{$indices_rlv_rgs};

		my $br2change = $set_sep_patterns_for_classes->[$indx_class][$indx_sep_pt][1];
		my ($shift1, $beg_site, $end_site, $affected_regions, $seqs_affected) = @{$set_classes_shift_respos->[$indx_class]};

		# @{$set_classes_shift_respos[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)	is for the $cl th class of the run of shifts of residue positions, where
		# @{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
		# @seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;

		my $if_incorporated = 0;

		my $curr_size_sbset = @sbset_blocks_shift_respos;

		for (my$bl = 0; $bl<$curr_size_sbset; $bl++) { # Main inner for-loop (over blocks constructed so far). #

		    my $block_shift_respos = $sbset_blocks_shift_respos[$bl];

		    my ($shift0, $beg_blk, $end_blk, $id_mrca0, $depth_mrca0, $seqs_involved0, $br2change0, $excl_sbblocks0, $ct_changes0, $indices_src_classes) = @{$block_shift_respos};

	# @{$set_blocks_shift_respos[$bl]} = ($shift, $beg_site, $end_site, $node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes)
	# is for the $bl th block of the run of shifts of residue positions, where
	# $shift is the value of shift characterizing the block,
	# $beg/end_site is the beginning/ending site (0-based full-closed) of the block,
	# $node_mrca is the node of the most recent common ancestor of the sequences involved,
	# @seqs_affected stores the (0-based) indices of the sequences involved with the block,
	# %br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
	# @{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
	# $ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
	#  @indices_src_classes stores the indices of the classes that are used to construct the block.


			## Screenings according to whether the examined class can be grafted into the block.

			# (1) In terms of sequence phylogeny. #

		    my $change0 = $br2change0->{$id_mrca};
		    (defined $change0) or next;

		    my ($state_ue, $state_le) = @{$change0};
		   ($state_ue eq $state_assigned) or next; # The state on the upper-end MUST match the state assigned to the affected sequences.

			# (2) In terms of the region in the MSA. #

		    my ($indx_rg_le, $indx_rg_re); # Indices (specifying the elements in @{$indices_rlv_rgs}) for the relevant regions giving the left-end and right-end of the block.

		    for (my $i=0; $i<$ct_rlv_rgs; $i++) { # 1st main inner-inner for-loop (over relevant regions for the examined class). #
			my $indx = $indices_rlv_rgs->[$i];

			my ($beg_rg, $end_rg) = @{$affected_regions->[$indx]};

			($beg_rg == $beg_blk) and ($indx_rg_le = $i);
			($end_rg == $end_blk) and ($indx_rg_re = $i);
			
		    }			# End of the 1st main inner-inner for-loop (over relevant regions for the examined class). #

		    ((defined $indx_rg_le) and (defined $indx_rg_re)) or next; # Each end of the block MUST match either end of a relevant regions.

			## Now, incorporate the examined class into the block. ##

		    $if_incorporated = 1;

		    push @{$indices_src_classes}, $indx_class;

			# Add excluded sub-blocks, if any. #

		    my $ct_new_excluded = 0;
		    for (my $i=$indx_rg_le; $i<$indx_rg_re; $i++) { # 2nd main inner-inner for-loop (over relevant regions covering the block). #
			my ($indx_bf, $indx_af) = @{$indices_rlv_rgs}[$i, $i+1];
			my $end_bf = $affected_regions->[$indx_bf][1];
			my $beg_af = $affected_regions->[$indx_af][0];
			push @{$excl_sbblocks0}, [$end_bf+1, $beg_af-1, $id_mrca];
			$ct_new_excluded++;
		    }					# End of the 2nd main inner-inner for-loop (over relevant regions covering the block). #

			# Merge the affected sequences into the involved sequences. #

		    my $seqs_involved_new = union_nmarray (@{$seqs_involved0}, @{$seqs_affected});
		    ##my $seqs_involved_new = union_array (@{$seqs_involved0}, @{$seqs_affected});
		    $block_shift_respos->[5] = $seqs_involved_new;

			# Edit the hash mapping changes onto the separating branches. #

		    delete $br2change0->{$id_mrca};
		    foreach my $br (keys %{$br2change}) { ($br == $id_mrca) or ($br2change0->{$br} = $br2change->{$br}); }

			# Update the number of changes forming the block. #

		    $ct_changes0 += scalar (keys %{$br2change}) - 2 + $ct_new_excluded ;

		    $block_shift_respos->[8] = $ct_changes0; 

			# If any, put the remaining regions in the class back to the set of classes to be examined. #

		    my @indices_left_margin = ($indx_rg_le>0) ? @{$indices_rlv_rgs}[0 .. $indx_rg_le-1] : ();
		    my @indices_right_margin = ($indx_rg_re < $ct_rlv_rgs-1) ? @{$indices_rlv_rgs}[$indx_rg_re+1 .. $ct_rlv_rgs-1] : ();

		    (@indices_right_margin>0) and (unshift @set_to_be_examined, [$indx_class, $indx_sep_pt, $ct_changes, $id_mrca, $depth_mrca, \@indices_right_margin]);
		    (@indices_left_margin>0) and (unshift @set_to_be_examined, [$indx_class, $indx_sep_pt, $ct_changes, $id_mrca, $depth_mrca, \@indices_left_margin]);

		    last;

		}				# End of the main inner for-loop (over blocks constructed so far). #

		unless ($if_incorporated) {

			# Create a new block out of the examined class. #

		    my @excl_sbblocks = ();

		    for (my $i=1; $i<$ct_rlv_rgs; $i++) { # 2nd main inner for-loop (over relevant regions). #
			my ($indx_bf, $indx_af) = @{$indices_rlv_rgs}[$i-1, $i];
			my $end_bf = $affected_regions->[$indx_bf][1];
			my $beg_af = $affected_regions->[$indx_af][0];
			push @excl_sbblocks, [$end_bf+1, $beg_af-1, $id_mrca];
		    }				# End of the 2nd main inner for-loop (over relevant regions). #

		    $ct_changes += @excl_sbblocks;

				# A BUG Fixed (on Oct 17 (Fri), 2014). #
		    my ($indx_lm, $indx_rm) = @{$indices_rlv_rgs}[0, $ct_rlv_rgs-1]; # Left-most and right-most relevant regions. #
		    my $beg_site_new = $affected_regions->[$indx_lm][0];
		    my $end_site_new = $affected_regions->[$indx_rm][1];
					# Another BUG fixed on Oct 22 (Wed), 2014. #
		    my %cp_br2change = %{$br2change};
		    my @cp_seqs_affected = @{$seqs_affected};
		    my @block = ($shift, $beg_site_new, $end_site_new, $id_mrca, $depth_mrca, \@cp_seqs_affected, \%cp_br2change, \@excl_sbblocks, $ct_changes, [$indx_class]);
		    ##my @block = ($shift, $beg_site_new, $end_site_new, $id_mrca, $depth_mrca, $seqs_affected, $br2change, \@excl_sbblocks, $ct_changes, [$indx_class]);
		    ##my @block = ($shift, $beg_site, $end_site, $id_mrca, $depth_mrca, $seqs_affected, $br2change, \@excl_sbblocks, $ct_changes, [$indx_class]);
		    push @sbset_blocks_shift_respos, \@block;
		}

	    }				# End of the main middle while-loop (until the examined classes run out). #

	    push @set_blocks_shift_respos, @sbset_blocks_shift_respos;

	}						# End of the main outer foreach-loop (over shifts). #

	return \@set_blocks_shift_respos;

}	# End of the subroutine 'reorganize_classes_shift_respos'. #



# merge_miniclasses_shift_respos_alpha:
#	AS OF Oct 24 (Fri), 2014, the first version of 'merge_miniclasses_shift_respos' was made an alpha version,
#	because it was found to have a couple of flaws:
#	(i) Two or more mini-blocks created in the preliminary process will never have a chance of merging with each other later.
#	(ii) It merges seemingly independent mini-blocks (or mini-classes).
#
# merge_miniclasses_shift_respos:
#	This subroutine merges mini-classes of runs of residue-position-shifts blocks.
#	It was created by taking advantage of the framework of the subroutine 'reorganize_classes_shift_respos'.
#	(CREATION STARTED on Oct 22 (Wed), 2014, ENDED on Oct 23 (Thu), 2014.)
#
#	[Method] Two mini-classes will be merged either if one of them completely encloses the other,
#	both phylogenetically and spatially, or if the MRCAs of the two share the same parent and
#	one completely encloses the other spatially.
#	In the latter case, the MRCA of the resulting block will be the shared parent.
#
# reorganize_classes_shift_respos:
#	This subroutine receives the main output of 'list_classes_shift_respos'
#	and the main output of 'dpsm_infer_sep_patterns_for_classes',
#	and reorganizes the classes of residue-position-shifts into (usually larger) blocks of residue-position-shifts.
#	(CREATED on Oct 16 (Thu), 2014.)
#
# INPUT: (\@set_miniclasses_shift_respos, \@set_sep_patterns_for_miniclasses, \@seqids, \%name2id, \%id2name, \%node2ch, \%node2pa, \%node2depth, $char_abs, $char_pres), where
#
#	@{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th mini-class of the run of shifts of residue positions, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@{$set_sep_patterns_for_miniclasses[$mc]->[$pt]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $pt th pattern separating the sequences affected by the $mc th miniclass of position shifts, where
#	%node2parsstat = ($node_id => [$state, $ct_pres_ext], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.
#	($pt = 0 is for the Dollo parsimonious changes by assigning the 'absence' state to the affected sequences,
#	$pt = 1 is for the changes by assigning the 'presence' state to them.)
#
#	$seqids[$k] is the ID of the $k th sequence in the MSA (w.r.t. which the shifts of residue positions are defined);
#	%name2id = ($sequence_ID => $node_id, ...) for the exterior nodes
#	(and $name2id{$TOP_NODE} = ID of the top-node (i.e., root));
#	%id2name = ($node_id => $sequence_ID, ...) for the exterior nodes;
#	%node2ch = ($node_id => \@child_nodes, ...) for the interior nodes;
#	%node2pa = ($node_id => $parent_node, ...) for all nodes except root;
#	%node2depth = ($node_id => $absolute_depth, ...) for all nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	$char_pres/abs is the character representing a 'presence'/'absence' state.
#
# OUTPUT: \@set_miniblocks_shift_respos, where
#	@{$set_blocks_shift_respos[$mb]} = ($shift, $beg_site, $end_site, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes)
#	is for the $mb th block created by merging mini-classes of the run of shifts of residue positions, where
#	$shift is the value of shift characterizing the block,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the block,
#	$node_mrca is the node of the most recent common ancestor of the sequences involved,
#	@seqs_affected stores the (0-based) indices of the sequences involved with the block,
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
#	$ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
#	 @indices_src_classes stores the indices of the classes that are used to construct the block.
#
sub merge_miniclasses_shift_respos_alpha (\@\@\@\%\%\%\%\%$$) {

	my ($set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses, $seqids, $name2id, $id2name, $node2ch, $node2pa, $node2depth, $char_abs, $char_pres) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ct_seqs = @{$seqids};

	my $ct_miniclasses = @{$set_miniclasses_shift_respos};
	unless (@{$set_sep_patterns_for_miniclasses} == $ct_miniclasses) {
	    my $size_set = @{$set_sep_patterns_for_miniclasses};
	    print STDOUT "[ERROR 'merge_miniclasses_shift_respos_alpha'] ILLEGAL INPUT: The size of ", '@{$set_sep_patterns_for_miniclasses', " (=$size_set) CLASHES with the number of mini-classes (=$ct_miniclasses): $! !\n\n";
	    return undef;
	}

	my %shift2miniclasses; # = ($shift => \@indices_miniclasses, ...) 
	my %shift_intvl2indx_mc; # = ("$shift:$beg_mc:$end_mc" => $indx_mc, ...)

	my @mrca_miniclasses = (); # element = [$nodeid_mrca, $depth_mrca] # ADDED in creation of 'merge_miniclasses_shift_respos'. #

	for (my $mc=0; $mc<$ct_miniclasses; $mc++) { # Preliminary for-loop (over mini-classes). #

	    my ($shift, $beg_mc, $end_mc) = @{$set_miniclasses_shift_respos->[$mc]}[0,1,2];
	    ##my $shift = $set_miniclasses_shift_respos->[$mc][0];

	    my $miniclasses = $shift2miniclasses{$shift};
	    (defined $miniclasses) or ($miniclasses = $shift2miniclasses{$shift} = []);
	    push @{$miniclasses}, $mc;

	    $shift_intvl2indx_mc{"$shift:$beg_mc:$end_mc"} = $mc;

		# ADDED in creation of 'merge_miniclasses_shift_respos'. #

	    my $br2change = $set_sep_patterns_for_miniclasses->[$mc][1][1];
	    ##my $br2change = $set_sep_patterns_for_miniclasses->[$mc][2]; # A BUG fixed on Oct 23 (Thu), 2014. #

	    my $mrca_mc ;
	    foreach my $br (sort {$b <=> $a} keys %{$br2change}) {
		($br2change->{$br}[0] eq $char_abs) or next; # Skip if the change is not a 'creation'.
		$mrca_mc = $br;
		last;
	    }
	    (defined $mrca_mc) or ($mrca_mc = $topnode_id); # If there is no 'creation', the top node must be the MRCA.
	    my $depth_mrca = $node2depth->{$mrca_mc};

	    $mrca_miniclasses[$mc] = [$mrca_mc, $depth_mrca];

		# END of "ADDED in creation of 'merge_miniclasses_shift_respos'." #

	}		# End of the preliminary for-loop (over miniclasses). #

	my @affected_states = ($char_abs, $char_pres); # States assigned to the affected sequences.


		### Main Process ###

	my @set_miniblocks_shift_respos = ();

	foreach my $shift (sort { $a <=> $b} keys %shift2miniclasses) { # Main Outer foreach-loop (over shifts). #

	    my $miniclasses = $shift2miniclasses{$shift};
	    my $sbct_mcs = @{$miniclasses};

		# The rule(s) according to which separation pattern should be chosen.
		#
		# => Now, we can UNCONDITIONALLY choose the 1th separation pattern (i.e., regard the shift as a 'presence' state),
		#   given that the states in each mini-class became very coherent with each other, thanks to the use of 'br_list_classes_shift_respos'.
	    my $indx_sep_pt = 1;
	    my $state_assigned = $affected_states[$indx_sep_pt];

	    ##		# Sort the mini-classes in the ascending order of the depth of MRCA. # --- DISABLED because it became meaningless..
	    ##my @std_miniclasses = ($sbct_mcs>1) ? sort { $mrca_miniclasses[$a]->[1] <=> $mrca_miniclasses[$b]->[1] } @{$miniclasses} : @{$miniclasses};

	    my @sbset_miniblocks_shift_respos = ();

			## Preliminary processes: Merge in advance those mini-classes which follow some simple rules. ##
			## (ADDED on Oct 23 (Thu), 2013.)

	    my @remaining_miniclasses = ();

	    my %mrca2indices_mc; # = ($mrca => \@indices_of_mini_classes_with_mrca, ...)
	    foreach my $mc (@{$miniclasses}) { # 1st preliminary middle foreach-loop (over mini-classes with $shift). #
		my $mrca = $mrca_miniclasses[$mc]->[0];
		my $indices_mc = $mrca2indices_mc{$mrca};
		(defined $indices_mc) or ($indices_mc = $mrca2indices_mc{$mrca} = []);
		push @{$indices_mc}, $mc;
	    }			# End of the 1st preliminary middle foreach-loop (over mini-classes with $shift). #

	    foreach my $mrca (sort { $node2depth->{$a} <=> $node2depth->{$b} } keys %mrca2indices_mc ) { # 2nd preliminary middle foreach-loop (over MRCAs of the mini-classes with $shift). #

		my $sbset_mc = $mrca2indices_mc{$mrca};
		unless (@{$sbset_mc}>1) {
		    push @remaining_miniclasses, @{$sbset_mc};
		    next;
		}

		my $depth_mrca = $node2depth->{$mrca};

		my %seqset2indices_mc; # = ({concatenated indices of sequences} => \@indices_mini_classes, ....)

		foreach my $mc (@{$sbset_mc}) { # 1st preliminary inner foreach-loop (over mini-classes with $mrca). #

		    my $cnct_seqs = join (';', @{$set_miniclasses_shift_respos->[$mc][4]});
			# @{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)

		    my $sbsbset_mc = $seqset2indices_mc{$cnct_seqs};
		    (defined $sbsbset_mc) or ($sbsbset_mc = $seqset2indices_mc{$cnct_seqs} = []);
		    push @{$sbsbset_mc}, $mc;

		}			# End of the 1st preliminary inner foreach-loop (over mini-classes with $mrca). #

		foreach my $seqset (sort keys %seqset2indices_mc) { # 2nd preliminary inner foreach-loop (over sequence sets). # 

		    my $sbsbset_mc = $seqset2indices_mc{$seqset};

		    my $size_sbsbset = @{$sbsbset_mc};

		    unless ($size_sbsbset>1) {
			push @remaining_miniclasses, @{$sbsbset_mc};
		 	next;
		    }

		    my $indx_mc0 = $sbsbset_mc->[0]; # The 0th mini-class in the sub-sub-set. #

		    my $seqs_involved = $set_miniclasses_shift_respos->[$indx_mc0][4];
		    my $br2change = $set_sep_patterns_for_miniclasses->[$indx_mc0][$indx_sep_pt][1];


			# Sort the mini-classes in ascending order of the beginning coordinate. #

		    my $tmp = $set_miniclasses_shift_respos;
		    my @std_sbsbset_mc = sort { $tmp->[$a][1] <=> $tmp->[$b][1] } @{$sbsbset_mc};

		    my $indxindx_le_mc = 0;

		    while ($indxindx_le_mc < $size_sbsbset-1) { # 3rd preliminary inner while-loop (over the left-ends of the possible arrays of mini-classes). #

			my $indx_le_mc = $std_sbsbset_mc[$indxindx_le_mc];
			my $beg_blk_cand = $tmp->[$indx_le_mc][1];
			my $end_blk_cand;

			my $indxindx_re_mc;
			for (my $i=$indxindx_le_mc+1; $i<$size_sbsbset; $i++) {  # Preliminary inner-inner for-loop (over possible right-ends). #

			    my $indx_mc1 = $std_sbsbset_mc[$i];
			    $end_blk_cand = $tmp->[$indx_mc1][2];
			    my $indx_cmpl_cand = $shift_intvl2indx_mc{"$shift:$beg_blk_cand:$end_blk_cand"};
			    (defined $indx_cmpl_cand) or next;

			    my ($mrca2, $depth_mrca2) = @{$mrca_miniclasses[$indx_cmpl_cand]};
			    my $br2change2 = $set_sep_patterns_for_miniclasses->[$indx_cmpl_cand][$indx_sep_pt][1];

			   if ($depth_mrca2 < $depth_mrca) { # The complement candidate might phylogenetically encompass the subject mini-classes. #
				my $change2 = $br2change2->{$mrca};
				((defined $change2) and ($change2->[0] eq $state_assigned)) and ($indxindx_re_mc = $i);

			   } elsif ($depth_mrca2 > $depth_mrca) { # The subject mini-classes might phylogenetically encompass the subject mini-classes. #
				my $change = $br2change->{$mrca2};
				((defined $change) and ($change->[0] eq $state_assigned)) and ($indxindx_re_mc = $i);

			   } elsif ($mrca != $topnode_id) { # The complement candidate and the subject mini-classes might merge together to form a larger block.
				($node2pa->{$mrca} == $node2pa->{$mrca2}) and ($indxindx_re_mc = $i);
			   }

			   (defined $indxindx_re_mc) and last;

			}				# End of the preliminary inner-inner for-loop (over possible right-ends). #

			unless (defined $indxindx_re_mc) {

			   push @remaining_miniclasses, $indx_le_mc; # In this case, the left-end candidate cannot pair with any other mini-class.
			
			   $indxindx_le_mc++;
			   next;
			}

				# Create a new block of mini-classes. #

			my @indices_src_miniclasses = @std_sbsbset_mc[$indxindx_le_mc .. $indxindx_re_mc];
			my $ct_src_mc = @indices_src_miniclasses;

			my @cp_seqs_involved = @{$seqs_involved};
			my %cp_br2change = %{$br2change};

			my @excl_sbblks = ();

			for (my $i=1; $i<$ct_src_mc; $i++) { # Trailing inner-inner for-loop (over pairs of mini-classes in the new block). # 
			    my ($indx_bf, $indx_af) = @indices_src_miniclasses[$i-1, $i];
			    my $end_bf = $tmp->[$indx_bf][2];
			    my $beg_af = $tmp->[$indx_af][1];
			    push @excl_sbblks, [$end_bf+1, $beg_af-1, $mrca];
			}			# End of trailing inner-inner for-loop (over pairs of mini-classes in the new block). #

			my $ct_changes = scalar (@excl_sbblks) + scalar (keys %cp_br2change);

			push @sbset_miniblocks_shift_respos, [$shift, $beg_blk_cand, $end_blk_cand, $mrca, $depth_mrca, \@cp_seqs_involved, \%cp_br2change, \@excl_sbblks, $ct_changes, \@indices_src_miniclasses];

		# ($shift0, $beg_blk, $end_blk, $mrca0, $depth_mrca0, $seqs_involved0, $br2change0, $excl_sbblocks0, $ct_changes0, $indices_src_miniclasses) = @{$block_shift_respos};

				# Update the possible left-end of the new block. #

			$indxindx_le_mc = $indxindx_re_mc+1;

		    }				# End of the 3rd preliminary inner while-loop (over the left-ends of the possible arrays of mini-classes). #

		    if ($indxindx_le_mc<$size_sbsbset) {
			push @remaining_miniclasses, @std_sbsbset_mc[$indxindx_le_mc .. $size_sbsbset-1];
		    }

		}				# End of the 2nd preliminary inner foreach-loop (over sequence sets). #


	    }					# End of the 2nd preliminary middle foreach-loop (over MRCAs of the mini-classes with $shift). #

	    my @std_miniclasses = @remaining_miniclasses;

			## END of "Preliminary processes: Merge in advance those mini-classes which follow some simple rules." ##


			## MAIN PROCESS ##

	    foreach my $indx_mc (@std_miniclasses) { # Main middle foreach-loop (over mini-classes with $shift). #

		my ($shift1, $beg_mc, $end_mc, $affected_regions, $seqs_affected) = @{$set_miniclasses_shift_respos->[$indx_mc]};
		my ($node2dparsstat_mc, $br2change_mc, $ct_changes_mc) = @{$set_sep_patterns_for_miniclasses->[$indx_mc][$indx_sep_pt]};
		my ($mrca_mc, $depth_mrca_mc) = @{$mrca_miniclasses[$indx_mc]};

		##my $state_assigned = $affected_states[$indx_sep_pt];

		# @{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention), --- Here, this is just a formality, and does not play any effective roles.
		# @seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;

		my $if_incorporated = 0;

		my $curr_size_sbset = @sbset_miniblocks_shift_respos;

		for (my $bl = 0; $bl<$curr_size_sbset; $bl++) { # Main inner for-loop (over blocks constructed so far). #

		    my $block_shift_respos = $sbset_miniblocks_shift_respos[$bl];

		    my ($shift0, $beg_blk, $end_blk, $mrca0, $depth_mrca0, $seqs_involved0, $br2change0, $excl_sbblocks0, $ct_changes0, $indices_src_miniclasses) = @{$block_shift_respos};

	# @{$set_blocks_shift_respos[$bl]} = ($shift, $beg_site, $end_site, $node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes)
	# is for the $bl th block of the run of shifts of residue positions, where
	# $shift is the value of shift characterizing the block,
	# $beg/end_site is the beginning/ending site (0-based full-closed) of the block,
	# $node_mrca is the node of the most recent common ancestor of the sequences involved,
	# @seqs_affected stores the (0-based) indices of the sequences involved with the block,
	# %br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
	# @{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
	# $ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
	#  @indices_src_classes stores the indices of the classes that are used to construct the block.


			## Screenings according to whether the examined class can be grafted into the block,
			# in terms of both phylogenetic and spatial relationships.

		    ($mrca_mc == $mrca0) and next; # If the MRCAs match, the mini-classes will never be merged.

		    my $inclusion_type; # = 1 if the block encompasses the mini-class,
					# = -1 if the block is encompassed by the mini-class,
					# = 0 if their MRCAs share the same parent.

		    my $pa_mrca;

		    if ($depth_mrca0 < $depth_mrca_mc) {
		 	my $change0 = $br2change0->{$mrca_mc};
			(defined $change0) or next;

			my ($state_ue, $state_le) = @{$change0};
			($state_ue eq $state_assigned) or next; # The state on the upper-end MUST match the state assigned to the affected sequences.
				# The block may encompass the mini-class... now examine the spatial relationship. #

			(($beg_blk <= $beg_mc) and ($end_mc <= $end_blk)) or next;

			$inclusion_type = 1;

		    } elsif ($depth_mrca0 > $depth_mrca_mc) {

			my $change_mc = $br2change_mc->{$mrca0};
			(defined $change_mc) or next;

			my ($state_ue, $state_le) = @{$change_mc};
			($state_ue eq $state_assigned) or next; # The state on the upper-end MUST match the state assigned to the affected sequences.
				# The block may be encompassed by the mini-class.... now examine the spatial relationship. #

			(($beg_mc <= $beg_blk) and ($end_blk <= $end_mc)) or next;

			$inclusion_type = -1;

		    } else {
				# The MRCAs of the block and the mini-class may share the same parent. #

			my $pa_mrca_mc = $node2pa->{$mrca_mc};
			my $pa_mrca0   = $node2pa->{$mrca0};
			($pa_mrca_mc == $pa_mrca0) or next; # The parents did NOT match. #

			$pa_mrca = $pa_mrca0;

				# Now, examine the spatial relationship. #
			
			( (($beg_blk <= $beg_mc) and ($end_mc <= $end_blk)) or 
			  (($beg_mc <= $beg_blk) and ($end_blk <= $end_mc)) )  or next;

			$inclusion_type = 0;
		    }

		    (defined $inclusion_type) or next; # The mini-class cannot be merged with this block. #


			## Now, incorporate the examined class into the block. ##
			## => Now, UPDATE the mini-block. (MODIFIED in creation of 'merge_miniclasses_shift_respos'.)

		    $if_incorporated = 1;

		    push @{$indices_src_miniclasses}, $indx_mc;

		    my $beg_new_blk = ($beg_blk < $beg_mc) ? $beg_blk : $beg_mc ;
		    my $end_new_blk = ($end_blk > $end_mc) ? $end_blk : $end_mc ;

		    my $new_mrca0;
		    my %new_br2change0;
		    my @new_excl_sbblocks0 = ();


		   if ($inclusion_type > 0) { # Incorporate the mini-class into the block. #

			$new_mrca0 = $mrca0;

				# Fill out @new_excl_sbblocks0. #

			my $if_reduce_excluded=0;
			foreach my $ex_sbblk (@{$excl_sbblocks0}) { # 1st main inner foreach-loop (over sub-blocks excluded from the current block). #
			    my ($beg_sb, $end_sb, $mrca_sb) = @{$ex_sbblk};
			    unless (($mrca_sb == $mrca_mc) and ($beg_sb <= $beg_mc) and ($end_mc <= $end_sb) ) {
					# The sub-block does NOT include the mini-class. #
				push @new_excl_sbblocks0, $ex_sbblk;
				next;
			    }
				# The sub-block must encompass the mini-class. #
				# => subtract the mini-class from the sub-block. #
			    $if_reduce_excluded = 1;

			    ($beg_sb < $beg_mc) and (push @new_excl_sbblocks0, [$beg_sb, $beg_mc-1, $mrca_mc]);
			    ($end_mc < $end_sb) and (push @new_excl_sbblocks0, [$end_mc+1, $end_sb, $mrca_mc]);

			}				# End of the 1st main inner foreach-loop (over sub-blocks excluded from the current block). #
			unless ($if_reduce_excluded) { # If no excluded region encompasses the mini-class.
			    ($beg_new_blk < $beg_mc) and (push @new_excl_sbblocks0, [$beg_new_blk, $beg_mc-1, $mrca_mc]);
			    ($end_mc < $end_new_blk) and (push @new_excl_sbblocks0, [$end_mc+1, $end_new_blk, $mrca_mc]);
			}

				# Fill out %new_br2change0. # (Remove entries with the key = $mrca_mc.)

			%new_br2change0 = %{$br2change0};
			(defined $new_br2change0{$mrca_mc}) and (delete $new_br2change0{$mrca_mc});
			foreach my $br (keys %{$br2change_mc}) { ($br == $mrca_mc) or ($new_br2change0{$br} = $br2change_mc->{$br}); }

		   } elsif ($inclusion_type < 0) { # Incorporate the block into the mini-class. #

		 	$new_mrca0 = $mrca_mc;

				# Fill out @new_excl_sbblocks0. #

			@new_excl_sbblocks0 = @{$excl_sbblocks0};
			($beg_new_blk < $beg_blk) and (push @new_excl_sbblocks0, [$beg_new_blk, $beg_blk-1, $mrca0]);
			($end_blk < $end_new_blk) and (push @new_excl_sbblocks0, [$end_blk+1, $end_new_blk, $mrca0]);

				# Fill out %new_br2change0. # (Remove entries with the key = $mrca0.)

			%new_br2change0 = %{$br2change_mc};
			(defined $new_br2change0{$mrca0}) and (delete $new_br2change0{$mrca0});
			foreach my $br (keys %{$br2change0}) { ($br == $mrca0) or ($new_br2change0{$br} = $br2change0->{$br}); }

		   } else { # Merge the block and the mini-class to create a bigger block. #

				# In this case, replace $mrca_mc and $mrca0 with $pa_mrca. #

			$new_mrca0 = $pa_mrca;

				# Fill out @new_excl_sbblocks0. #

			@new_excl_sbblocks0 = @{$excl_sbblocks0};
			($beg_new_blk < $beg_blk) and (push @new_excl_sbblocks0, [$beg_new_blk, $beg_blk-1, $mrca0]);
			($end_blk < $end_new_blk) and (push @new_excl_sbblocks0, [$end_blk+1, $end_new_blk, $mrca0]);
			($beg_new_blk < $beg_mc) and (push @new_excl_sbblocks0, [$beg_new_blk, $beg_mc-1, $mrca_mc]);
			($end_mc < $end_new_blk) and (push @new_excl_sbblocks0, [$end_mc+1, $end_new_blk, $mrca_mc]);

				# Fill out %new_br2change0. # (Remove entries with the key = $mrca_mc or $mrca0, and introduce 'creation' with the key = $pa_mrca instead.) 

			$new_br2change0{$pa_mrca} = [$char_abs, $char_pres];
			foreach my $br (keys %{$br2change_mc}) { ($br == $mrca_mc) or ($new_br2change0{$br} = $br2change_mc->{$br}); }
			foreach my $br (keys %{$br2change0}) { ($br == $mrca0) or ($new_br2change0{$br} = $br2change0->{$br}); }

		   }


			# Merge the affected sequences into the involved sequences. #

		    my $new_seqs_involved = union_nmarray (@{$seqs_involved0}, @{$seqs_affected});

			# Update other variables. #

		    my $new_ct_changes0 = scalar (@new_excl_sbblocks0) + scalar (keys %new_br2change0) ;
		    my $depth_new_mrca0 = $node2depth->{$new_mrca0};

			# Finally, exert the update. #

		    $sbset_miniblocks_shift_respos[$bl] = [$shift0, $beg_new_blk, $end_new_blk, $new_mrca0, $depth_new_mrca0, $new_seqs_involved, \%new_br2change0, \@new_excl_sbblocks0, $new_ct_changes0, $indices_src_miniclasses];

		    ##  ($shift0, $beg_blk, $end_blk, $mrca0, $depth_mrca0, $seqs_involved0, $br2change0, $excl_sbblocks0, $ct_changes0, $indices_src_miniclasses) = @{$block_shift_respos};

		    last;

		}				# End of the main inner for-loop (over blocks constructed so far). #

		unless ($if_incorporated) {

				### RESTART FROM HERE ### 

			# Create a new block out of the examined mini-class. #

		    my @excl_sbblocks = ();

		    ##$ct_changes_mc += @excl_sbblocks;

		    my %cp_br2change = %{$br2change_mc};
		    my @cp_seqs_affected = @{$seqs_affected};
		    my @block = ($shift, $beg_mc, $end_mc, $mrca_mc, $depth_mrca_mc, \@cp_seqs_affected, \%cp_br2change, \@excl_sbblocks, $ct_changes_mc, [$indx_mc]);
	    ##my @block = ($shift, $beg_site_new, $end_site_new, $id_mrca, $depth_mrca, $seqs_affected, $br2change, \@excl_sbblocks, $ct_changes, [$indx_class]);

		    push @sbset_miniblocks_shift_respos, \@block;
		}

	    }				# End of the main middle foreach-loop (over mini-classes with $shift). #

	    push @set_miniblocks_shift_respos, @sbset_miniblocks_shift_respos;

	}						# End of the main outer foreach-loop (over shifts). #

	return \@set_miniblocks_shift_respos;


}	# End of the subroutine 'merge_miniclasses_shift_respos_alpha'. #


# merge_miniclasses_shift_respos:
#	This subroutine merges mini-classes of runs of residue-position-shifts blocks.
#	It was created by taking advantage of the framework of the subroutine 'reorganize_classes_shift_respos'.
#	(CREATION STARTED on Oct 22 (Wed), 2014, ENDED on Oct 23 (Thu), 2014.)
#
#	MODIFIED on Oct 24 (Fri), 2014:
#	To rectify a couple of flaws (explained in the header of 'merge_miniclasses_shift_respos_alpha'),
#	a few modifications were made:
#	(i) Now, preliminary processes creates preliminary mini-blocks even from mini-classes that failed to be pre-merged,
#	  and all the preliminary mini-blocks will be tried to be merged in the main process;
#	(ii) The merger conditions in the main process were made more strict, more precisely, the preliminary mini-block
#	    and the pre-formed mini-block will be merged ONLY IF they share left- and right-ends.  
#
#	[Method] Two mini-classes will be merged either if one of them completely encloses the other,
#	both phylogenetically and spatially, or if the MRCAs of the two share the same parent and
#	one completely encloses the other spatially.
#	In the latter case, the MRCA of the resulting block will be the shared parent.
#
# reorganize_classes_shift_respos:
#	This subroutine receives the main output of 'list_classes_shift_respos'
#	and the main output of 'dpsm_infer_sep_patterns_for_classes',
#	and reorganizes the classes of residue-position-shifts into (usually larger) blocks of residue-position-shifts.
#	(CREATED on Oct 16 (Thu), 2014.)
#
# INPUT: (\@set_miniclasses_shift_respos, \@set_sep_patterns_for_miniclasses, \@seqids, \%name2id, \%id2name, \%node2ch, \%node2pa, \%node2depth, $char_abs, $char_pres), where
#
#	@{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th mini-class of the run of shifts of residue positions, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@{$set_sep_patterns_for_miniclasses[$mc]->[$pt]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $pt th pattern separating the sequences affected by the $mc th miniclass of position shifts, where
#	%node2parsstat = ($node_id => [$state, $ct_pres_ext], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.
#	($pt = 0 is for the Dollo parsimonious changes by assigning the 'absence' state to the affected sequences,
#	$pt = 1 is for the changes by assigning the 'presence' state to them.)
#
#	$seqids[$k] is the ID of the $k th sequence in the MSA (w.r.t. which the shifts of residue positions are defined);
#	%name2id = ($sequence_ID => $node_id, ...) for the exterior nodes
#	(and $name2id{$TOP_NODE} = ID of the top-node (i.e., root));
#	%id2name = ($node_id => $sequence_ID, ...) for the exterior nodes;
#	%node2ch = ($node_id => \@child_nodes, ...) for the interior nodes;
#	%node2pa = ($node_id => $parent_node, ...) for all nodes except root;
#	%node2depth = ($node_id => $absolute_depth, ...) for all nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	$char_pres/abs is the character representing a 'presence'/'absence' state.
#
# OUTPUT: \@set_miniblocks_shift_respos, where
#	@{$set_blocks_shift_respos[$mb]} = ($shift, $beg_site, $end_site, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes)
#	is for the $mb th block created by merging mini-classes of the run of shifts of residue positions, where
#	$shift is the value of shift characterizing the block,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the block,
#	$node_mrca is the node of the most recent common ancestor of the sequences involved,
#	@seqs_affected stores the (0-based) indices of the sequences involved with the block,
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
#	$ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
#	 @indices_src_classes stores the indices of the classes that are used to construct the block.
#
sub merge_miniclasses_shift_respos (\@\@\@\%\%\%\%\%$$) {

	my ($set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses, $seqids, $name2id, $id2name, $node2ch, $node2pa, $node2depth, $char_abs, $char_pres) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ct_seqs = @{$seqids};

	my $ct_miniclasses = @{$set_miniclasses_shift_respos};
	unless (@{$set_sep_patterns_for_miniclasses} == $ct_miniclasses) {
	    my $size_set = @{$set_sep_patterns_for_miniclasses};
	    print STDOUT "[ERROR 'merge_miniclasses_shift_respos'] ILLEGAL INPUT: The size of ", '@{$set_sep_patterns_for_miniclasses', " (=$size_set) CLASHES with the number of mini-classes (=$ct_miniclasses): $! !\n\n";
	    return undef;
	}

	my %shift2miniclasses; # = ($shift => \@indices_miniclasses, ...) 
	my %shift_intvl2indx_mc; # = ("$shift:$beg_mc:$end_mc" => $indx_mc, ...)

	my @mrca_miniclasses = (); # element = [$nodeid_mrca, $depth_mrca] # ADDED in creation of 'merge_miniclasses_shift_respos'. #

	for (my $mc=0; $mc<$ct_miniclasses; $mc++) { # Preliminary for-loop (over mini-classes). #

	    my ($shift, $beg_mc, $end_mc) = @{$set_miniclasses_shift_respos->[$mc]}[0,1,2];
	    ##my $shift = $set_miniclasses_shift_respos->[$mc][0];

	    my $miniclasses = $shift2miniclasses{$shift};
	    (defined $miniclasses) or ($miniclasses = $shift2miniclasses{$shift} = []);
	    push @{$miniclasses}, $mc;

	    $shift_intvl2indx_mc{"$shift:$beg_mc:$end_mc"} = $mc;

		# ADDED in creation of 'merge_miniclasses_shift_respos'. #

	    my $br2change = $set_sep_patterns_for_miniclasses->[$mc][1][1];
	    ##my $br2change = $set_sep_patterns_for_miniclasses->[$mc][2]; # A BUG fixed on Oct 23 (Thu), 2014. #

	    my $mrca_mc ;
	    foreach my $br (sort {$b <=> $a} keys %{$br2change}) {
		($br2change->{$br}[0] eq $char_abs) or next; # Skip if the change is not a 'creation'.
		$mrca_mc = $br;
		last;
	    }
	    (defined $mrca_mc) or ($mrca_mc = $topnode_id); # If there is no 'creation', the top node must be the MRCA.
	    my $depth_mrca = $node2depth->{$mrca_mc};

	    $mrca_miniclasses[$mc] = [$mrca_mc, $depth_mrca];

		# END of "ADDED in creation of 'merge_miniclasses_shift_respos'." #

	}		# End of the preliminary for-loop (over miniclasses). #

	my @affected_states = ($char_abs, $char_pres); # States assigned to the affected sequences.


		### Main Process ###

	my @set_miniblocks_shift_respos = ();

	foreach my $shift (sort { $a <=> $b} keys %shift2miniclasses) { # Main Outer foreach-loop (over shifts). #

	    my $miniclasses = $shift2miniclasses{$shift};
	    my $sbct_mcs = @{$miniclasses};

		# The rule(s) according to which separation pattern should be chosen.
		#
		# => Now, we can UNCONDITIONALLY choose the 1th separation pattern (i.e., regard the shift as a 'presence' state),
		#   given that the states in each mini-class became very coherent with each other, thanks to the use of 'br_list_classes_shift_respos'.
	    my $indx_sep_pt = 1;
	    my $state_assigned = $affected_states[$indx_sep_pt];

	    ##		# Sort the mini-classes in the ascending order of the depth of MRCA. # --- DISABLED because it became meaningless..
	    ##my @std_miniclasses = ($sbct_mcs>1) ? sort { $mrca_miniclasses[$a]->[1] <=> $mrca_miniclasses[$b]->[1] } @{$miniclasses} : @{$miniclasses};

	    my @sbset_miniblocks_shift_respos = ();

			## Preliminary processes: Merge in advance those mini-classes which follow some simple rules. ##
			## (ADDED on Oct 23 (Thu), 2013.)
			## (=> MODIFIED on Oct 24 (Fri), 2014.)

	    my @pre_sbset_miniblocks = ();
	    ##my @remaining_miniclasses = ();

	    my %mrca2indices_mc; # = ($mrca => \@indices_of_mini_classes_with_mrca, ...)
	    foreach my $mc (@{$miniclasses}) { # 1st preliminary middle foreach-loop (over mini-classes with $shift). #
		my $mrca = $mrca_miniclasses[$mc]->[0];
		my $indices_mc = $mrca2indices_mc{$mrca};
		(defined $indices_mc) or ($indices_mc = $mrca2indices_mc{$mrca} = []);
		push @{$indices_mc}, $mc;
	    }			# End of the 1st preliminary middle foreach-loop (over mini-classes with $shift). #

	    foreach my $mrca (sort { $node2depth->{$a} <=> $node2depth->{$b} } keys %mrca2indices_mc ) { # 2nd preliminary middle foreach-loop (over MRCAs of the mini-classes with $shift). #

		my $sbset_mc = $mrca2indices_mc{$mrca};
		unless (@{$sbset_mc}>1) {

			# MOFIFIED on Oct 24 (Fri), 2014. #
		    my $indx_mc = $sbset_mc->[0];
		    my ($mrca, $depth_mrca) = @{$mrca_miniclasses[$indx_mc]};
		    my ($shift1, $beg_mc, $end_mc, $aff_rgs, $seqs_involved) = @{$set_miniclasses_shift_respos->[$indx_mc]};
		    my $br2change = $set_sep_patterns_for_miniclasses->[$indx_mc][$indx_sep_pt][1];
		    my @cp_seqs_involved = @{$seqs_involved};
		    my %cp_br2change = %{$br2change};
		    my $ct_changes = scalar (keys %cp_br2change);
		    my @excl_sbblks = ();
		    push @pre_sbset_miniblocks, [$shift, $beg_mc, $end_mc, $mrca, $depth_mrca, \@cp_seqs_involved, \%cp_br2change, \@excl_sbblks, $ct_changes, $sbset_mc];
		    ##push @remaining_miniclasses, @{$sbset_mc};
		    next;
		}

		my $depth_mrca = $node2depth->{$mrca};

		my %seqset2indices_mc; # = ({concatenated indices of sequences} => \@indices_mini_classes, ....)

		foreach my $mc (@{$sbset_mc}) { # 1st preliminary inner foreach-loop (over mini-classes with $mrca). #

		    my $cnct_seqs = join (';', @{$set_miniclasses_shift_respos->[$mc][4]});
			# @{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)

		    my $sbsbset_mc = $seqset2indices_mc{$cnct_seqs};
		    (defined $sbsbset_mc) or ($sbsbset_mc = $seqset2indices_mc{$cnct_seqs} = []);
		    push @{$sbsbset_mc}, $mc;

		}			# End of the 1st preliminary inner foreach-loop (over mini-classes with $mrca). #

		foreach my $seqset (sort keys %seqset2indices_mc) { # 2nd preliminary inner foreach-loop (over sequence sets). # 

		    my $sbsbset_mc = $seqset2indices_mc{$seqset};

		    my $size_sbsbset = @{$sbsbset_mc};

		    unless ($size_sbsbset>1) {

			    # MOFIFIED on Oct 24 (Fri), 2014. #
			my $indx_mc = $sbsbset_mc->[0];
			my ($mrca, $depth_mrca) = @{$mrca_miniclasses[$indx_mc]};
			my ($shift1, $beg_mc, $end_mc, $aff_rgs, $seqs_involved) = @{$set_miniclasses_shift_respos->[$indx_mc]};
			my $br2change = $set_sep_patterns_for_miniclasses->[$indx_mc][$indx_sep_pt][1];
			my @cp_seqs_involved = @{$seqs_involved};
			my %cp_br2change = %{$br2change};
			my $ct_changes = scalar (keys %cp_br2change);
			my @excl_sbblks = ();
			push @pre_sbset_miniblocks, [$shift, $beg_mc, $end_mc, $mrca, $depth_mrca, \@cp_seqs_involved, \%cp_br2change, \@excl_sbblks, $ct_changes, $sbsbset_mc];
			##push @remaining_miniclasses, @{$sbsbset_mc};
		 	next;
		    }

		    my $indx_mc0 = $sbsbset_mc->[0]; # The 0th mini-class in the sub-sub-set. #

		    my $seqs_involved = $set_miniclasses_shift_respos->[$indx_mc0][4];
		    my $br2change = $set_sep_patterns_for_miniclasses->[$indx_mc0][$indx_sep_pt][1];


			# Sort the mini-classes in ascending order of the beginning coordinate. #

		    my $tmp = $set_miniclasses_shift_respos;
		    my @std_sbsbset_mc = sort { $tmp->[$a][1] <=> $tmp->[$b][1] } @{$sbsbset_mc};

		    my $indxindx_le_mc = 0;

		    while ($indxindx_le_mc < $size_sbsbset-1) { # 3rd preliminary inner while-loop (over the left-ends of the possible arrays of mini-classes). #

			my $indx_le_mc = $std_sbsbset_mc[$indxindx_le_mc];
			my $beg_blk_cand = $tmp->[$indx_le_mc][1];
			my $end_blk_cand;

			my $indxindx_re_mc;
			for (my $i=$indxindx_le_mc+1; $i<$size_sbsbset; $i++) {  # Preliminary inner-inner for-loop (over possible right-ends). #

			    my $indx_mc1 = $std_sbsbset_mc[$i];
			    $end_blk_cand = $tmp->[$indx_mc1][2];
			    my $indx_cmpl_cand = $shift_intvl2indx_mc{"$shift:$beg_blk_cand:$end_blk_cand"};
			    (defined $indx_cmpl_cand) or next;

			    my ($mrca2, $depth_mrca2) = @{$mrca_miniclasses[$indx_cmpl_cand]};
			    my $br2change2 = $set_sep_patterns_for_miniclasses->[$indx_cmpl_cand][$indx_sep_pt][1];

			   if ($depth_mrca2 < $depth_mrca) { # The complement candidate might phylogenetically encompass the subject mini-classes. #
				my $change2 = $br2change2->{$mrca};
				((defined $change2) and ($change2->[0] eq $state_assigned)) and ($indxindx_re_mc = $i);

			   } elsif ($depth_mrca2 > $depth_mrca) { # The subject mini-classes might phylogenetically encompass the subject mini-classes. #
				my $change = $br2change->{$mrca2};
				((defined $change) and ($change->[0] eq $state_assigned)) and ($indxindx_re_mc = $i);

			   } elsif ($mrca != $topnode_id) { # The complement candidate and the subject mini-classes might merge together to form a larger block.
				($node2pa->{$mrca} == $node2pa->{$mrca2}) and ($indxindx_re_mc = $i);
			   }

			   (defined $indxindx_re_mc) and last;

			}				# End of the preliminary inner-inner for-loop (over possible right-ends). #

			unless (defined $indxindx_re_mc) {

				# MOFIFIED on Oct 24 (Fri), 2014. #
			    my $indx_mc = $indx_le_mc;
			    my ($mrca, $depth_mrca) = @{$mrca_miniclasses[$indx_mc]};
			    my ($shift1, $beg_mc, $end_mc, $aff_rgs, $seqs_involved) = @{$set_miniclasses_shift_respos->[$indx_mc]};
			    my $br2change = $set_sep_patterns_for_miniclasses->[$indx_mc][$indx_sep_pt][1];
			    my @cp_seqs_involved = @{$seqs_involved};
			    my %cp_br2change = %{$br2change};
			    my $ct_changes = scalar (keys %cp_br2change);
			    my @excl_sbblks = ();
			    push @pre_sbset_miniblocks, [$shift, $beg_mc, $end_mc, $mrca, $depth_mrca, \@cp_seqs_involved, \%cp_br2change, \@excl_sbblks, $ct_changes, [$indx_mc]];
			   ##push @remaining_miniclasses, $indx_le_mc; # In this case, the left-end candidate cannot pair with any other mini-class.
			
			   $indxindx_le_mc++;
			   next;
			}

				# Create a new block of mini-classes. #

			my @indices_src_miniclasses = @std_sbsbset_mc[$indxindx_le_mc .. $indxindx_re_mc];
			my $ct_src_mc = @indices_src_miniclasses;

			my @cp_seqs_involved = @{$seqs_involved};
			my %cp_br2change = %{$br2change};

			my @excl_sbblks = ();

			for (my $i=1; $i<$ct_src_mc; $i++) { # Trailing inner-inner for-loop (over pairs of mini-classes in the new block). # 
			    my ($indx_bf, $indx_af) = @indices_src_miniclasses[$i-1, $i];
			    my $end_bf = $tmp->[$indx_bf][2];
			    my $beg_af = $tmp->[$indx_af][1];
			    push @excl_sbblks, [$end_bf+1, $beg_af-1, $mrca];
			}			# End of trailing inner-inner for-loop (over pairs of mini-classes in the new block). #

			my $ct_changes = scalar (@excl_sbblks) + scalar (keys %cp_br2change);

				# MODIFIED on Oct 24 (Fri), 2014. #
			push @pre_sbset_miniblocks, [$shift, $beg_blk_cand, $end_blk_cand, $mrca, $depth_mrca, \@cp_seqs_involved, \%cp_br2change, \@excl_sbblks, $ct_changes, \@indices_src_miniclasses];
			##push @sbset_miniblocks_shift_respos, [$shift, $beg_blk_cand, $end_blk_cand, $mrca, $depth_mrca, \@cp_seqs_involved, \%cp_br2change, \@excl_sbblks, $ct_changes, \@indices_src_miniclasses];

		# ($shift0, $beg_blk, $end_blk, $mrca0, $depth_mrca0, $seqs_involved0, $br2change0, $excl_sbblocks0, $ct_changes0, $indices_src_miniclasses) = @{$block_shift_respos};

				# Update the possible left-end of the new block. #

			$indxindx_le_mc = $indxindx_re_mc+1;

		    }				# End of the 3rd preliminary inner while-loop (over the left-ends of the possible arrays of mini-classes). #

		    if ($indxindx_le_mc<$size_sbsbset) {

				# MOFIFIED on Oct 24 (Fri), 2014. #

			foreach my $indx_mc (@std_sbsbset_mc[$indxindx_le_mc .. $size_sbsbset-1]) {
			    my ($mrca, $depth_mrca) = @{$mrca_miniclasses[$indx_mc]};
			    my ($shift1, $beg_mc, $end_mc, $aff_rgs, $seqs_involved) = @{$set_miniclasses_shift_respos->[$indx_mc]};
			    my $br2change = $set_sep_patterns_for_miniclasses->[$indx_mc][$indx_sep_pt][1];
			    my @cp_seqs_involved = @{$seqs_involved};
			    my %cp_br2change = %{$br2change};
			    my $ct_changes = scalar (keys %cp_br2change);
			    my @excl_sbblks = ();
			    push @pre_sbset_miniblocks, [$shift, $beg_mc, $end_mc, $mrca, $depth_mrca, \@cp_seqs_involved, \%cp_br2change, \@excl_sbblks, $ct_changes, [$indx_mc]];
			}
			##push @remaining_miniclasses, @std_sbsbset_mc[$indxindx_le_mc .. $size_sbsbset-1];
		    }

		}				# End of the 2nd preliminary inner foreach-loop (over sequence sets). #


	    }					# End of the 2nd preliminary middle foreach-loop (over MRCAs of the mini-classes with $shift). #

	    ##my @std_miniclasses = @remaining_miniclasses; # DISABLED on Oct 24 (Fri), 2014. #

			## END of "Preliminary processes: Merge in advance those mini-classes which follow some simple rules." ##


			## MAIN PROCESS ##

			# MODIFIED on Oct 24 (Fri), 2014. #
	    foreach my $pre_miniblock (@pre_sbset_miniblocks) { # Main middle foreach-loop (over pre-formed mini-blocks). #

			# NOTE: Interpret the prefix '_mc' as indicating the values for the pre-formed mini-block (as of Oct 24 (fri), 2014). #

		my ($shift1, $beg_mc, $end_mc, $mrca_mc, $depth_mrca_mc, $seqs_affected, $br2change_mc, $excl_sbblks_mc, $ct_changes_mc, $indices_src_mcs_mc) = @{$pre_miniblock};

=Became OBSOLETE as of Oct 24 (Fri), 2014.
	    foreach my $indx_mc (@std_miniclasses) {...} # Main middle foreach-loop (over mini-classes with $shift). #

		my ($shift1, $beg_mc, $end_mc, $affected_regions, $seqs_affected) = @{$set_miniclasses_shift_respos->[$indx_mc]};
		my ($node2dparsstat_mc, $br2change_mc, $ct_changes_mc) = @{$set_sep_patterns_for_miniclasses->[$indx_mc][$indx_sep_pt]};
		my ($mrca_mc, $depth_mrca_mc) = @{$mrca_miniclasses[$indx_mc]};

		##my $state_assigned = $affected_states[$indx_sep_pt];

		# @{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention), --- Here, this is just a formality, and does not play any effective roles.
		# @seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
=cut

		my $if_incorporated = 0;

		my $curr_size_sbset = @sbset_miniblocks_shift_respos;

		for (my $bl = 0; $bl<$curr_size_sbset; $bl++) { # Main inner for-loop (over blocks constructed so far). #

		    my $block_shift_respos = $sbset_miniblocks_shift_respos[$bl];

		    my ($shift0, $beg_blk, $end_blk, $mrca0, $depth_mrca0, $seqs_involved0, $br2change0, $excl_sbblocks0, $ct_changes0, $indices_src_miniclasses) = @{$block_shift_respos};

	# @{$set_blocks_shift_respos[$bl]} = ($shift, $beg_site, $end_site, $node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes)
	# is for the $bl th block of the run of shifts of residue positions, where
	# $shift is the value of shift characterizing the block,
	# $beg/end_site is the beginning/ending site (0-based full-closed) of the block,
	# $node_mrca is the node of the most recent common ancestor of the sequences involved,
	# @seqs_affected stores the (0-based) indices of the sequences involved with the block,
	# %br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
	# @{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
	# $ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
	#  @indices_src_classes stores the indices of the classes that are used to construct the block.


			## Screenings according to whether the examined class can be grafted into the block,
			# in terms of both phylogenetic and spatial relationships.

		    ($mrca_mc == $mrca0) and next; # If the MRCAs match, the mini-classes will never be merged.

		    my $inclusion_type; # = 1 if the block encompasses the mini-class,
					# = -1 if the block is encompassed by the mini-class,
					# = 0 if their MRCAs share the same parent.

		    my $pa_mrca;

		    if ($depth_mrca0 < $depth_mrca_mc) {
		 	my $change0 = $br2change0->{$mrca_mc};
			(defined $change0) or next;

			my ($state_ue, $state_le) = @{$change0};
			($state_ue eq $state_assigned) or next; # The state on the upper-end MUST match the state assigned to the affected sequences.
				# The block may encompass the mini-class... now examine the spatial relationship. #

					# MODIFIED on Oct 24 (Fri), 2014. #
			(($beg_blk == $beg_mc) and ($end_mc == $end_blk)) or next;
			##(($beg_blk <= $beg_mc) and ($end_mc <= $end_blk)) or next;

			$inclusion_type = 1;

		    } elsif ($depth_mrca0 > $depth_mrca_mc) {

			my $change_mc = $br2change_mc->{$mrca0};
			(defined $change_mc) or next;

			my ($state_ue, $state_le) = @{$change_mc};
			($state_ue eq $state_assigned) or next; # The state on the upper-end MUST match the state assigned to the affected sequences.
				# The block may be encompassed by the mini-class.... now examine the spatial relationship. #

					# MODIFIED on Oct 24 (Fri), 2014. #
			(($beg_mc == $beg_blk) and ($end_blk == $end_mc)) or next;
			##(($beg_mc <= $beg_blk) and ($end_blk <= $end_mc)) or next;

			$inclusion_type = -1;

		    } else {
				# The MRCAs of the block and the mini-class may share the same parent. #

			my $pa_mrca_mc = $node2pa->{$mrca_mc};
			my $pa_mrca0   = $node2pa->{$mrca0};
			($pa_mrca_mc == $pa_mrca0) or next; # The parents did NOT match. #

			$pa_mrca = $pa_mrca0;

				# Now, examine the spatial relationship. #
			
					# MODIFIED on Oct 24 (Fri), 2014. #
			(($beg_mc == $beg_blk) and ($end_blk == $end_mc)) or next;
			##( (($beg_blk <= $beg_mc) and ($end_mc <= $end_blk)) or 
			##  (($beg_mc <= $beg_blk) and ($end_blk <= $end_mc)) )  or next;

			$inclusion_type = 0;
		    }

		    (defined $inclusion_type) or next; # The mini-class cannot be merged with this block. #


			## Now, incorporate the examined class into the block. ##
			## => Now, UPDATE the mini-block. (MODIFIED in creation of 'merge_miniclasses_shift_respos'.)

		    $if_incorporated = 1;

				# MODIFIED on Oct 24 (Fri), 2014. #
		    push @{$indices_src_miniclasses}, @{$indices_src_mcs_mc};
		    ##push @{$indices_src_miniclasses}, $indx_mc;

		    my $beg_new_blk = ($beg_blk < $beg_mc) ? $beg_blk : $beg_mc ;
		    my $end_new_blk = ($end_blk > $end_mc) ? $end_blk : $end_mc ;

		    my $new_mrca0;
		    my %new_br2change0;
		    my @new_excl_sbblocks0 = ();


		   if ($inclusion_type > 0) { # Incorporate the mini-class into the block. #

			$new_mrca0 = $mrca0;

				# Fill out @new_excl_sbblocks0. #

			my $if_reduce_excluded=0;
			foreach my $ex_sbblk (@{$excl_sbblocks0}) { # 1st main inner foreach-loop (over sub-blocks excluded from the current block). #
			    my ($beg_sb, $end_sb, $mrca_sb) = @{$ex_sbblk};
			    unless (($mrca_sb == $mrca_mc) and ($beg_sb <= $beg_mc) and ($end_mc <= $end_sb) ) {
					# The sub-block does NOT include the mini-class. #
				push @new_excl_sbblocks0, $ex_sbblk;
				next;
			    }
				# The sub-block must encompass the mini-class. #
				# => subtract the mini-class from the sub-block. #
			    $if_reduce_excluded = 1;

			    ($beg_sb < $beg_mc) and (push @new_excl_sbblocks0, [$beg_sb, $beg_mc-1, $mrca_mc]);
			    ($end_mc < $end_sb) and (push @new_excl_sbblocks0, [$end_mc+1, $end_sb, $mrca_mc]);

			}				# End of the 1st main inner foreach-loop (over sub-blocks excluded from the current block). #
			unless ($if_reduce_excluded) { # If no excluded region encompasses the mini-class.
			    ($beg_new_blk < $beg_mc) and (push @new_excl_sbblocks0, [$beg_new_blk, $beg_mc-1, $mrca_mc]);
			    ($end_mc < $end_new_blk) and (push @new_excl_sbblocks0, [$end_mc+1, $end_new_blk, $mrca_mc]);
			}

			push @new_excl_sbblocks0, @{$excl_sbblks_mc}; # ADDED in modification on Oct 24 (Fri), 2014. #

				# Fill out %new_br2change0. # (Remove entries with the key = $mrca_mc.)

			%new_br2change0 = %{$br2change0};
			(defined $new_br2change0{$mrca_mc}) and (delete $new_br2change0{$mrca_mc});
			foreach my $br (keys %{$br2change_mc}) { ($br == $mrca_mc) or ($new_br2change0{$br} = $br2change_mc->{$br}); }

		   } elsif ($inclusion_type < 0) { # Incorporate the block into the mini-class. #

		 	$new_mrca0 = $mrca_mc;

				# Fill out @new_excl_sbblocks0. #

					# MODIFIED on Oct 24 (Fri), 2014. #
			@new_excl_sbblocks0 = (@{$excl_sbblks_mc}, @{$excl_sbblocks0});
			##@new_excl_sbblocks0 = @{$excl_sbblocks0};
			($beg_new_blk < $beg_blk) and (push @new_excl_sbblocks0, [$beg_new_blk, $beg_blk-1, $mrca0]);
			($end_blk < $end_new_blk) and (push @new_excl_sbblocks0, [$end_blk+1, $end_new_blk, $mrca0]);

				# Fill out %new_br2change0. # (Remove entries with the key = $mrca0.)

			%new_br2change0 = %{$br2change_mc};
			(defined $new_br2change0{$mrca0}) and (delete $new_br2change0{$mrca0});
			foreach my $br (keys %{$br2change0}) { ($br == $mrca0) or ($new_br2change0{$br} = $br2change0->{$br}); }

		   } else { # Merge the block and the mini-class to create a bigger block. #

				# In this case, replace $mrca_mc and $mrca0 with $pa_mrca. #

			$new_mrca0 = $pa_mrca;

				# Fill out @new_excl_sbblocks0. #

					# MODIFIED on Oct 24 (Fri), 2014. #
			@new_excl_sbblocks0 = (@{$excl_sbblocks0}, @{$excl_sbblks_mc});
			##@new_excl_sbblocks0 = @{$excl_sbblocks0};
			($beg_new_blk < $beg_blk) and (push @new_excl_sbblocks0, [$beg_new_blk, $beg_blk-1, $mrca0]);
			($end_blk < $end_new_blk) and (push @new_excl_sbblocks0, [$end_blk+1, $end_new_blk, $mrca0]);
			($beg_new_blk < $beg_mc) and (push @new_excl_sbblocks0, [$beg_new_blk, $beg_mc-1, $mrca_mc]);
			($end_mc < $end_new_blk) and (push @new_excl_sbblocks0, [$end_mc+1, $end_new_blk, $mrca_mc]);

				# Fill out %new_br2change0. # (Remove entries with the key = $mrca_mc or $mrca0, and introduce 'creation' with the key = $pa_mrca instead.) 

			$new_br2change0{$pa_mrca} = [$char_abs, $char_pres];
			foreach my $br (keys %{$br2change_mc}) { ($br == $mrca_mc) or ($new_br2change0{$br} = $br2change_mc->{$br}); }
			foreach my $br (keys %{$br2change0}) { ($br == $mrca0) or ($new_br2change0{$br} = $br2change0->{$br}); }

		   }


			# Merge the affected sequences into the involved sequences. #

		    my $new_seqs_involved = union_nmarray (@{$seqs_involved0}, @{$seqs_affected});

			# Update other variables. #

		    my $new_ct_changes0 = scalar (@new_excl_sbblocks0) + scalar (keys %new_br2change0) ;
		    my $depth_new_mrca0 = $node2depth->{$new_mrca0};

			# Finally, exert the update. #

		    $sbset_miniblocks_shift_respos[$bl] = [$shift0, $beg_new_blk, $end_new_blk, $new_mrca0, $depth_new_mrca0, $new_seqs_involved, \%new_br2change0, \@new_excl_sbblocks0, $new_ct_changes0, $indices_src_miniclasses];

		    ##  ($shift0, $beg_blk, $end_blk, $mrca0, $depth_mrca0, $seqs_involved0, $br2change0, $excl_sbblocks0, $ct_changes0, $indices_src_miniclasses) = @{$block_shift_respos};

		    last;

		}				# End of the main inner for-loop (over blocks constructed so far). #

				### RESTART FROM HERE ### 

		unless ($if_incorporated) {

			# Store the examined pre-formed mini-block as it is. #
			# (MODIFIED on Oct 24 (Fri), 2014.) 

		    push @sbset_miniblocks_shift_respos, $pre_miniblock;

=OBSOLETE as of Oct 24 (Fri), 2014. 
			# Create a new block out of the examined mini-class. #

		    my @excl_sbblocks = ();

		    ##$ct_changes_mc += @excl_sbblocks;

		    my %cp_br2change = %{$br2change_mc};
		    my @cp_seqs_affected = @{$seqs_affected};
		    my @block = ($shift, $beg_mc, $end_mc, $mrca_mc, $depth_mrca_mc, \@cp_seqs_affected, \%cp_br2change, \@excl_sbblocks, $ct_changes_mc, [$indx_mc]);
	    ##my @block = ($shift, $beg_site_new, $end_site_new, $id_mrca, $depth_mrca, $seqs_affected, $br2change, \@excl_sbblocks, $ct_changes, [$indx_class]);

		    push @sbset_miniblocks_shift_respos, \@block;
=cut

		}

				# MODIFIED on Oct 24 (Fri), 2014. #
	    }				# End of the main middle foreach-loop (over pre-formed mini-blocks). #
	    ##{...}				# End of the main middle foreach-loop (over mini-classes with $shift). #

	    push @set_miniblocks_shift_respos, @sbset_miniblocks_shift_respos;

	}						# End of the main outer foreach-loop (over shifts). #

	return \@set_miniblocks_shift_respos;


}	# End of the subroutine 'merge_miniclasses_shift_respos'. #



# find_trivial_blocks:
#	This subroutine receives a set of blocks of residue-position-shifts, which was output by 'reorganize_classes_shift_respos',
#	and searches for blocks that are 'trivial'.
#
#	A 'trivial' block is defined as a block with the commoner flanking shift,
#	which is NEITHER "isolated" from runs of the commoner flanking shift reaching the appropriate boundary
#	NOR included in an excluded sub-block of another block.
# 	(CREATED on Oct 17 (Fri), 2014.)
#
# INPUT: (\@set_blocks_shift_respos, \@seqids, $commoner_shift_flank, $shift_lf, $shift_rf, \@map_shifts_respos, $gap), where
#
#	@{$set_blocks_shift_respos[$bl]} = ($shift, $beg_site, $end_site, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes)
#	is for the $bl th block of the run of shifts of residue positions, where
#	$shift is the value of shift characterizing the block,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the block,
#	$node_mrca is the node of the most recent common ancestor of the sequences involved,
#	@seqs_affected stores the (0-based) indices of the sequences involved with the block,
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
#	$ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
#	 @indices_src_classes stores the indices of the classes that are used to construct the block;
#
#	$seqids[$k] is the ID of the $k th sequence in the MSA;
#
#	$shift_lf/rf is the value of shift at the left-/right-flanking column,
#	$commoner_shift_flank is the commoner one between them;
#
#	$map_shifts_respos[$k]->[$s] is the shift in the residue position from the reference MSA to the reconstructed one
#	for the $s th site of the $k th sequence, in the reconstructed MSA.
#	It is $gap if occupied by a gap;
#
#	$gap is the character representing a gap.
#
# OUTPUT: \%block2trivial, where
#	%block2trivial = ($indx_block => 1, ...) for blocks judged as "trivial."
#
sub find_trivial_blocks (\@\@$$$\@$) {

	my ($set_blocks_shift_respos, $seqids, $commoner_shift_flank, $shift_lf, $shift_rf, $map_shifts_respos, $gap) = @_;

	my $ct_seqs = @{$seqids};
	my @full_seq_indices = (0 .. $ct_seqs - 1);

	my $ct_blocks = @{$set_blocks_shift_respos};
	my $len_aln = @{$map_shifts_respos->[0]};

	my %excl_mrca2blocks; # ($mrca => \@blocks, ...), where @blocks stores [$bl, $sbblk], with $bl being an index of a block which has $mrca as the MRCA of one of their excluded sub-block, whose refernce is $sbblk. #

	for (my $bl=0; $bl<$ct_blocks; $bl++) { # Preliminary outer for-loop (over blocks). #

	    my $excl_sbblocks = $set_blocks_shift_respos->[$bl][7];

	    foreach my $sbblk (@{$excl_sbblocks}) { # Preliminary inner for-loop (over excluded blocks). #
		my $mrca = $sbblk->[2];
		my $blocks = $excl_mrca2blocks{$mrca};
		(defined $blocks) or ($blocks = $excl_mrca2blocks{$mrca} = []);
		push @{$blocks}, [$bl, $sbblk];
	    }				# End of the preliminary inner for-loop (over excluded blocks). #

	}			# End of the preliminary outer for-loop (over blocks). #

	my %block2trivial;

	for (my $bl=0; $bl<$ct_blocks; $bl++) { # Main outer for-loop (over blocks). #

	     my ($shift, $beg_site, $end_site, $nodeid_mrca, $depth_mrca, $seqs_invlvd, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_blocks_shift_respos->[$bl]};

	    ($shift == $commoner_shift_flank) or next; # Skip unless $shift is the common flanking value.

		# Test 1: Examine whether the block is within an excluded sub-block of another block. #

	    my $pot_accomm_blks = $excl_mrca2blocks{$nodeid_mrca}; # Set of potentially accommodating blocks. #
	    if (defined $pot_accomm_blks) {
		my $if_included = 0;
		foreach my $info_pablk (@{$pot_accomm_blks}) { # 1st main middle for-loop (over potentially accommodating blocks). #
		    my ($indx_pablk, $excl_sbblk) = @{$info_pablk};
		    my ($left_bound, $right_bound) = @{$excl_sbblk}[0,1];

			# Skip if the block sticks out of the sub-block. # 
		    ($beg_site < $left_bound) and next;
		    ($right_bound < $end_site) and next;

		    $if_included = 1;
		    last;
		}				# End of the 1st main middle for-loop (over potentially accommodating blocks). #
		$if_included and next; # If included in an excluded sub-block of another block, the block is NOT regarded as 'trivial'.
	    }

		# Test 2: Examine whether there is a run of $shift connecting either end of this block to the appropriate MSA end, in any sequence NOT involved in this block.

	    my ($common, $not_involved, $ex2) = diff_nmarray (@full_seq_indices, @{$seqs_invlvd});

	    my $if_isolated = 1;

	    foreach my $indx (@{$not_involved}) { # 2nd main middle for-loop (over sequences NOT involved in the block). #

		my $row_shifts_respos = $map_shifts_respos->[$indx];

		if ($commoner_shift_flank == $shift_lf) { # Examine the region on the left of the block. #
		    my $if_stretched = 0;
		    for (my $s=$beg_site; $s<=$end_site; $s++) { # 1st inner for-loop (forward over sites in the block). #
			my $shift1 = $row_shifts_respos->[$s];
			($shift1 eq $gap) and next;
			($shift == $commoner_shift_flank) or last;
			$if_stretched = 1;
			last;
		    }				# End of the 1st inner for-loop (forward over sites in the block). #
		    if ($if_stretched) {
			for (my $s=0; $s<$beg_site; $s++) { # 2nd inner for-loop (over sites on the left of the block). #
			    my $shift1 = $row_shifts_respos->[$s];
			    ($shift1 eq $gap) and next;
			    ($shift1 == $commoner_shift_flank) and next;
			    $if_stretched = 0;
			    last;
			}			# End of the 2nd inner for-loop (over sites on the left of the block). #
		    }
		    if ($if_stretched) {
			$if_isolated = 0;
			last;
		    }
		}

	 	if ($commoner_shift_flank == $shift_rf) { # Examine the region on the right of the block. #
		    my $if_stretched = 0;
		    for (my $s=$end_site; $s>=$beg_site; $s--) { # 3rd inner for-loop (backward over sites in the block). #
			my $shift1 = $row_shifts_respos->[$s];
			($shift1 eq $gap) and next;
			($shift == $commoner_shift_flank) or last;
			$if_stretched = 1;
			last;
		    }				# End of the 3rd inner for-loop (backward over sites in the block). #
		    if ($if_stretched) {
			for (my $s=$end_site+1; $s<$len_aln; $s++) { # 4th inner for-loop (over sites on the right of the block). #
			    my $shift1 = $row_shifts_respos->[$s];
			    ($shift1 eq $gap) and next;
			    ($shift1 == $commoner_shift_flank) and next;
			    $if_stretched = 0;
			    last;
			}			# End of the 4th inner for-loop (over sites on the right of the block). #
		    }
		    if ($if_stretched) {
			$if_isolated = 0;
			last;
		    }
		}
	    }			# End of the 2nd main middle for-loop (over sequences NOT involved in the block). #

	    ($if_isolated) and next; # If "isolated," the block is NOT regarded as "trivial."

	    $block2trivial{$bl} = 1;

	}			# End of the main outer for-loop (over blocks). #


	return \%block2trivial;

}	# End of the subroutine 'find_trivial_blocks'. #


# lump_together_similar_blocks_proto:
#	This is an imperfect version of 'lump_together_similar_blocks'.
#	It uses a simpler criterion for identifying gap-mediated merger of two blocks.
#	(DEVELOPMENT FROZEN on Oct 17 (Fri), 2014.)
#
# lump_together_similar_blocks:
#	This subroutine receives a set of blocks of residue-position-shifts, which was output by 'reorganize_classes_shift_respos',
#	and lumps together neighboring blocks affecting the identical set of sequences, if they satisfy some conditions.
#	(CREATION STARTED on Oct 17 (Fri), 2014, ENDED on Oct 21 (Tue), 2014.)
#
# INPUT: (\@set_blocks_shift_respos, \@seqids, \%block2trivial, \@map_shifts_respos, $gap), where
#
#	@{$set_blocks_shift_respos[$bl]} = ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes)
#	is for the $bl th block of the run of shifts of residue positions, where
#	$shift is the value of shift characterizing the block,
#	$beg/end_blk is the beginning/ending coordinate (0-based full-closed) of the block,
#	$node_mrca is the node of the most recent common ancestor of the sequences involved,
#	@seqs_affected stores the (0-based) indices of the sequences involved with the block,
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
#	$ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
#	 @indices_src_classes stores the indices of the classes that are used to construct the block;
#
#	$seqids[$k] is the ID of the $k th sequence in the MSA;
#
#	%block2trivial = ($indx_block => 1, ...) for blocks judged as "trivial";
#
#	$map_shifts_respos[$k]->[$s] is the shift in the residue position from the reference MSA to the reconstructed one
#	for the $s th site of the $k th sequence, in the reconstructed MSA.
#	It is $gap if occupied by a gap;
#
#	$gap is the character representing a gap.
#
# OUTPUT: \@composite_blocks, where
#	@{$composite_blocks[$cb]} = ($beg_cb, $end_cb, $nodeid_mrca, \%br2change, \@constituents, \@seqs_involved, \@merger_types) is for the $cb th composite block, where
#	$beg/end_cb is the beginning/ending coordinate of the composite block (in the 0-based full-closed convention),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...) for the branches necessary for separating the involved sequences,
#	@constituents stores the indices of the blocks that makes up the composite block,
#	@seqs_involved stores the indices of the sequences involved in the constituent blocks.
#	$merger_types[$i] dictates the type of merger between the $i th and ($i+1) th constituent blocks
#	( = 0 for 'gap-mediated', = 1/2 for 'non-gap-mediated' with the $i/($i+1) th block incorporated).
#
sub lump_together_similar_blocks_proto (\@\@\%\@$) {

	my ($set_blocks_shift_respos, $seqids, $block2trivial, $map_shifts_respos, $gap) = @_;

	my $ct_blocks = @{$set_blocks_shift_respos};
	if ($ct_blocks==0) {
	    return []; # Return an empty set.

	} elsif ($ct_blocks==1) {
	    my ($shift, $beg_blk, $end_blk, $nodeid_mrca, $depth_mrca, $seqs_involved, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_blocks_shift_respos->[0]};

			# A BUG Fixed (on Oct 18 (Sat), 2014). #
	    my @cmp_blk = ($beg_blk, $end_blk, $nodeid_mrca, $br2change, [0], $seqs_involved, []);
	    ##my @cmp_blk = ($beg_blk, $end_blk, $nodeid_mrca, $br2change, [0], $seqs_involved);
	    return [ \@cmp_blk ];
	}

	$gap = uc $gap;

	my $ct_seqs = @{$seqids};
	my $alnlen = @{$map_shifts_respos->[0]};
	my @full_seq_indices = (0 .. $ct_seqs-1);

	my %seqset2blocks; # = ("$indx_seq1;$indx_seq2;..." => \@indices_blocks, ...)
	my %shift_interval2blocks; # = ("$shift:$beg_blk:$end_blk" => \@indices_blocks, ...)

	for (my $bl=0; $bl<$ct_blocks; $bl++) { # Preliminary for-loop (over blocks). #

	    (defined $block2trivial->{$bl}) and next; # Skip a "trivial" block. #

	    my ($shift, $beg_blk, $end_blk, $nodeid_mrca, $depth_mrca, $seqs_involved, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_blocks_shift_respos->[$bl]};

	    my $cnct_seqs = join (';', @{$seqs_involved});
	    my $blocks1 = $seqset2blocks{$cnct_seqs};
	    (defined $blocks1) or ($blocks1 = $seqset2blocks{$cnct_seqs} = []);
	    push @{$blocks1}, $bl;

	    my $shift_interval = "$shift:$beg_blk:$end_blk";
	    my $blocks2 = $shift_interval2blocks{$shift_interval};
	    (defined $blocks2) or ($blocks2 = $shift_interval2blocks{$shift_interval} = []);
	    push @{$blocks2}, $bl;

	}			# End of the preliminary for-loop (over blocks). #

		## MAIN PROCESS: Lump together neighboring blocks involving the identical set of sequences, if they satisfy some conditions. ##

	my @composite_blocks = ();

	foreach my $seqset (sort keys %seqset2blocks) { # Main outer for-loop (over sets of blocks classified by their involved sequences). #
	    my $sbj_blocks = $seqset2blocks{$seqset};

			# Sort the subject blocks in the spatial order. #

	    my $ct_sbj_blocks = @{$sbj_blocks};
	    my @std_sbj_blocks = ($ct_sbj_blocks>1) ? sort {$set_blocks_shift_respos->[$a][1] <=> $set_blocks_shift_respos->[$b][1]} @{$sbj_blocks} : @{$sbj_blocks} ;
	

	    my @sbset_composite_blocks = ();

			# A BUG fixed on Oct 18 (Sat), 2014. #
	    for (my $i=0; $i<$ct_sbj_blocks; $i++) { # Main middle for-loop (over blocks in the set). #
	    ##while (my $bl = shift @std_sbj_blocks) {...} # Main middle while-loop (over blocks in the set). #

		my $bl = $std_sbj_blocks[$i];
		
		my ($shift, $beg_blk, $end_blk, $nodeid_mrca, $depth_mrca, $seqs_involved, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_blocks_shift_respos->[$bl]};

		my @tentative_cmp_blk = ($beg_blk, $end_blk, $nodeid_mrca, $br2change, [$bl], $seqs_involved, []);

		unless (@sbset_composite_blocks>0) {
			# Create a new composite block. #
		    push @sbset_composite_blocks, \@tentative_cmp_blk;
		    next;
		}

			# Retrieve the last (i.e., left-neighboring) block. #
		my $last_cmp_blk = $sbset_composite_blocks[$#sbset_composite_blocks];
		my $last_constituents = $last_cmp_blk->[4];
		my $last_blk = $last_constituents->[$#{$last_constituents}];

		my ($shift0, $beg_blk0, $end_blk0, $nodeid_mrca0, $depth_mrca0, $seqs_involved0, $br2change0, $excl_sbblocks0, $ct_changes0, $indices_src_classes0) = @{$set_blocks_shift_respos->[$last_blk]};

			# (By definition, $nodeid_mrca, $seqs_involved, and $br2changed must be identical to $nodeid_mrca0, $seqs_involved0, and $br2changed0, respectively.) #

		my $gapsize = $beg_blk - ($end_blk0 +1);

		if ($gapsize>0) {
			## When the two blocks are NOT contiguous. ##

			# Pre-test: Examine the consistency between the gap size and the shift difference.

		    my $diff_shift = $shift - $shift0;
		    unless ($gapsize == $diff_shift) {
				# Create a new composite block. #
			push @sbset_composite_blocks, \@tentative_cmp_blk;
			next;
		    }

			# Test 1a: Examine whether the mediating sites of the involved sequences are occupied by gaps alone. #

		    my $if_nongap = 0;
		    foreach my $seqid (@{$seqs_involved}) { # 1st main inner foreach-loop (over sequences involved). #

			my $row_shifts_respos = $map_shifts_respos->[$seqid];

			for (my $s=$end_blk0+1; $s<$beg_blk; $s++) { # Inner-inner for-loop (over sites between the blocks). #
			    ($row_shifts_respos->[$s] eq $gap) and next;
			    $if_nongap = 1;
			    last;
			}				# End of the inner-inner for-loop (over sites between the blocks). #
			$if_nongap and last;
		    }				# End of the 1st main inner foreach-loop (over sequences involved). #

		    if ($if_nongap) {
				# Create a new composite block. #
			push @sbset_composite_blocks, \@tentative_cmp_blk;
			next;
		    }

			# Test 1b: Examine whether at least one non-involved sequences has the mediating sites occupied by gaps alone. #
		    my ($common, $seqs_non_involved, $ex2) = diff_nmarray (@full_seq_indices, @{$seqs_involved});
			
		    my $if_gaprun = 0;
		    foreach my $seqid (@{$seqs_non_involved}) { # 2nd main inner foreach-loop (over sequences NOT involved). #

			my $row_shifts_respos = $map_shifts_respos->[$seqid];

			my $if_nongap2=0;
			for (my $s=$end_blk0+1; $s<$beg_blk; $s++) { # Inner-inner for-loop (over sites between the blocks). #
			    ($row_shifts_respos->[$s] eq $gap) and next;
			    $if_nongap2 = 1;
			}				# End of the inner-inner for-loop (over sites between the blocks). #
			$if_nongap2 and next;
			$if_gaprun = 1;
			last;
		    }				# End of the 2nd main inner foreach-loop (over sequences NOT involved). #

		    unless ($if_gaprun) {
				# Create a new composite block. #
			push @sbset_composite_blocks, \@tentative_cmp_blk;
			next;
		    }

				# Incorporate the subject block into the last composite block. #

		    $last_cmp_blk->[1] = $end_blk;
		    push @{$last_constituents}, $bl;
		    push @{$last_cmp_blk->[6]}, 0;

		    next;

		} # END of "if ($gapsize>0) {...}"


			# When the two blocks sit contiguously. #
			# => Examine whether intermediate candidate block can be incorporated into another block.

		my $intm_incp_type = 0; # = 1/2 if the last/current block was incorporated.

		foreach my $intm_cand (
			[$shift0 - $shift, $beg_blk0 - $shift, $end_blk0 - $shift, 1],
			[$shift - $shift0, $beg_blk - $shift0, $end_blk - $shift0, 2] ) { # 3rd main inner foreach-loop (over candidates of the intermediate state of the block). #

		    my ($shift_cand, $beg_cand, $end_cand, $type_cand) = @{$intm_cand};

		    my $shift_intvl_cand = "$shift_cand:$beg_cand:$end_cand";
		    my $blocks_cand = $shift_interval2blocks{$shift_intvl_cand};

		    (defined $blocks_cand) or next;

		    foreach my $blk_c (@{$blocks_cand}) { # Inner-inner foreach-loop (over blocks whose shift and region matched with those of the intermediate candidate). #

			my ($shift_c, $beg_blk_c, $end_blk_c, $nodeid_mrca_c, $depth_mrca_c, $seqs_involved_c, $br2change_c, $excl_sbblocks_c, $ct_changes_c, $indices_src_classes_c) = @{$set_blocks_shift_respos->[$blk_c]};

				# Examine whether the intermediate state candidate clicks together with the candidate block. #

			foreach my $br (sort {$a <=> $b} keys %{$br2change}) { # Innermost foreach-loop (over branches separating the involved sequences from the rest). #
			    my $change_c = $br2change_c->{$br};
			    (defined $change_c) or next;

			    my ($state_ue, $state_le) = @{$br2change->{$br}};
			    my ($state_ue_c, $state_le_c) = @{$change_c};

			    ($state_ue eq $state_le_c) or next;

				# The two blocks clicked together. #

			    $intm_incp_type = $type_cand;
			    last;
			}				# End of the innermost foreach-loop (over branches separating the involved sequences from the rest). #

		        ($intm_incp_type) and last;
		    }				# End of the inner-inner foreach-loop (over blocks whose shift and region matched with those of the intermediate candidate). #

		    ($intm_incp_type) and last;
		}					# End of the 3rd main inner foreach-loop (over candidates of the intermediate state of the block). #

		if ($intm_incp_type) {
			# If an intermediate candidate succeeded in being incorporated in another block, #
			# => incorporate the subject block into the last composite block. #

		    $last_cmp_blk->[1] = $end_blk;
		    push @{$last_constituents}, $bl;
		    push @{$last_cmp_blk->[6]}, $intm_incp_type;

		} else {
			# Otherwise, create a new composite block. #
		    push @sbset_composite_blocks, \@tentative_cmp_blk;
		    next;
		}

			# A BUG fixed (on Oct 18 (Sat), 2014). #
	    }					# End of the main middle for-loop (over blocks in the set). #
	    ##}					# End of the main middle while-loop (over blocks in the set). #

	    push @composite_blocks, @sbset_composite_blocks;

	}				# End of the main outer for-loop (over sets of blocks classified by their involved sequences). #

		### RESTART FROM HERE ###

		# Finally, sort the composite blocks according to their beginning coordinates. #

	@composite_blocks = sort { $a->[0] <=> $b->[0] } @composite_blocks;

	return \@composite_blocks;

}	# End of the subroutine 'lump_together_similar_blocks_proto'. # 


# lump_together_similar_blocks:
#	This subroutine receives a set of blocks of residue-position-shifts, which was output by 'reorganize_classes_shift_respos',
#	and lumps together neighboring blocks affecting the identical set of sequences, if they satisfy some conditions.
#	(CREATION STARTED on Oct 17 (Fri), 2014, ENDED on Oct 21 (Tue), 2014.)
#
# INPUT: (\@set_blocks_shift_respos, \@seqids, \%block2trivial, \@map_shifts_respos, $gap, \@set_cls_gblocks_ref, \@set_cls_gblocks_rec, \@set_br2change_for_gps, \%node2pa, \%node2ch, $topnode_id), where
#
#	@{$set_blocks_shift_respos[$bl]} = ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes)
#	is for the $bl th block of the run of shifts of residue positions, where
#	$shift is the value of shift characterizing the block,
#	$beg/end_blk is the beginning/ending coordinate (0-based full-closed) of the block,
#	$node_mrca is the node of the most recent common ancestor of the sequences involved,
#	@seqs_affected stores the (0-based) indices of the sequences involved with the block,
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
#	$ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
#	 @indices_src_classes stores the indices of the classes that are used to construct the block;
#
#	$seqids[$k] is the ID of the $k th sequence in the MSA;
#
#	%block2trivial = ($indx_block => 1, ...) for blocks judged as "trivial";
#
#	$map_shifts_respos[$k]->[$s] is the shift in the residue position from the reference MSA to the reconstructed one
#	for the $s th site of the $k th sequence, in the reconstructed MSA.
#	It is $gap if occupied by a gap;
#
#	$gap is the character representing a gap;
#
#	@{$set_cls_gblocks_ref/rec[$cl]->[$gb]} = ($indx_gp, $len_gb, $start_gb, $end_gb) is for the $gb th gap-pattern block in the $cl th cluster (i.e., gapped segment) in the reference/reconstructed MSA,
#	with $indx_gp being the index specifying the gap-pattern of the block ;
#
#	%{$set_br2change_for_gps->[$indx_gp]} = ($branch_id => [$state_ue, $state_le], ...) for the $indx_gp th gap-pattern;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but root of the tree;
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the tree.
#
# OUTPUT: \@composite_blocks, where
#	@{$composite_blocks[$cb]} = ($beg_cb, $end_cb, $nodeid_mrca, \%br2change, \@constituents, \@seqs_involved, \@merger_types) is for the $cb th composite block, where
#	$beg/end_cb is the beginning/ending coordinate of the composite block (in the 0-based full-closed convention),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...) for the branches necessary for separating the involved sequences,
#	@constituents stores the indices of the blocks that makes up the composite block,
#	@seqs_involved stores the indices of the sequences involved in the constituent blocks.
#	$merger_types[$i] dictates the type of merger between the $i th and ($i+1) th constituent blocks
#	( = 0 for 'gap-mediated', = 1/2 for 'non-gap-mediated' with the $i/($i+1) th block incorporated).
#
sub lump_together_similar_blocks (\@\@\%\@$\@\@\@\%\%$) {

	my ($set_blocks_shift_respos, $seqids, $block2trivial, $map_shifts_respos, $gap,
		$set_cls_gblocks_ref, $set_cls_gblocks_rec, $set_br2change_for_gps, $node2pa, $node2ch, $topnode_id) = @_;

	my $ct_blocks = @{$set_blocks_shift_respos};
	if ($ct_blocks==0) {
	    return []; # Return an empty set.

	} elsif ($ct_blocks==1) {
	    my ($shift, $beg_blk, $end_blk, $nodeid_mrca, $depth_mrca, $seqs_involved, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_blocks_shift_respos->[0]};

			# A BUG Fixed (on Oct 18 (Sat), 2014). #
	    my @cmp_blk = ($beg_blk, $end_blk, $nodeid_mrca, $br2change, [0], $seqs_involved, []);
	    ##my @cmp_blk = ($beg_blk, $end_blk, $nodeid_mrca, $br2change, [0], $seqs_involved);
	    return [ \@cmp_blk ];
	}

	$gap = uc $gap;

	my $ct_seqs = @{$seqids};
	my $alnlen = @{$map_shifts_respos->[0]};
	my @full_seq_indices = (0 .. $ct_seqs-1);

		# 1st preliminary process: Make quick reference lists for the blocks of residue-position shifts. #

	my %seqset2blocks; # = ("$indx_seq1;$indx_seq2;..." => \@indices_blocks, ...)
	my %shift_interval2blocks; # = ("$shift:$beg_blk:$end_blk" => \@indices_blocks, ...)

	for (my $bl=0; $bl<$ct_blocks; $bl++) { # 1st preliminary for-loop (over blocks). #

	    (defined $block2trivial->{$bl}) and next; # Skip a "trivial" block. #

	    my ($shift, $beg_blk, $end_blk, $nodeid_mrca, $depth_mrca, $seqs_involved, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_blocks_shift_respos->[$bl]};

	    my $cnct_seqs = join (';', @{$seqs_involved});
	    my $blocks1 = $seqset2blocks{$cnct_seqs};
	    (defined $blocks1) or ($blocks1 = $seqset2blocks{$cnct_seqs} = []);
	    push @{$blocks1}, $bl;

	    my $shift_interval = "$shift:$beg_blk:$end_blk";
	    my $blocks2 = $shift_interval2blocks{$shift_interval};
	    (defined $blocks2) or ($blocks2 = $shift_interval2blocks{$shift_interval} = []);
	    push @{$blocks2}, $bl;

	}			# End of the 1st preliminary for-loop (over blocks). #

		# 2nd preliminary process: Make quick reference lists for the gap-pattern blocks in the reference/reconstructed MSAs. #

	my %interval2gblk_ref; # = ("$beg_blk:$end_blk" => [$indx_cl, $indx_gb, $indx_gp], ...) for the gap-pattern blocks in the reference MSA.
	my %interval2gblk_rec; # = ("$beg_blk:$end_blk" => [$indx_cl, $indx_gb, $indx_gp], ...) for the gap-pattern blocks in the reconstructed MSA.

	my $ct_cls_ref = @{$set_cls_gblocks_ref};
	for (my $cl=0; $cl<$ct_cls_ref; $cl++) { # 2nd preliminary for-loop (over gapped segments in reference). #
	    my $gblks_in_cls = $set_cls_gblocks_ref->[$cl];
	    my $ct_blks = @{$gblks_in_cls};
	    for (my $gb=0; $gb<$ct_blks; $gb++) { # Inner for-loop (over gap-pattern blocks). #
		my ($indx_gp, $len_gb, $beg_gb, $end_gb) = @{$gblks_in_cls->[$gb]};
		$interval2gblk_ref{"$beg_gb:$end_gb"} = [$cl, $gb, $indx_gp];
	    }				# End of the inner for-loop (over gap-pattern blocks). #
	}			# End of the 2nd preliminary for-loop (over gapped segments in reference). #

	my $ct_cls_rec = @{$set_cls_gblocks_rec};
	for (my $cl=0; $cl<$ct_cls_rec; $cl++) { # 2nd preliminary for-loop (over gapped segments in reconstructed). #
	    my $gblks_in_cls = $set_cls_gblocks_rec->[$cl];
	    my $ct_blks = @{$gblks_in_cls};
	    for (my $gb=0; $gb<$ct_blks; $gb++) { # Inner for-loop (over gap-pattern blocks). #
		my ($indx_gp, $len_gb, $beg_gb, $end_gb) = @{$gblks_in_cls->[$gb]};
		$interval2gblk_rec{"$beg_gb:$end_gb"} = [$cl, $gb, $indx_gp];
	    }				# End of the inner for-loop (over gap-pattern blocks). #
	}			# End of the 2nd preliminary for-loop (over gapped segments in reconstructed). #



		## MAIN PROCESS: Lump together neighboring blocks involving the identical set of sequences, if they satisfy some conditions. ##

	my @composite_blocks = ();

	foreach my $seqset (sort keys %seqset2blocks) { # Main outer for-loop (over sets of blocks classified by their involved sequences). #
	    my $sbj_blocks = $seqset2blocks{$seqset};

			# Sort the subject blocks in the spatial order. #

	    my $ct_sbj_blocks = @{$sbj_blocks};
	    my @std_sbj_blocks = ($ct_sbj_blocks>1) ? sort {$set_blocks_shift_respos->[$a][1] <=> $set_blocks_shift_respos->[$b][1]} @{$sbj_blocks} : @{$sbj_blocks} ;
	

	    my @sbset_composite_blocks = ();

			# A BUG fixed on Oct 18 (Sat), 2014. #
	    for (my $i=0; $i<$ct_sbj_blocks; $i++) { # Main middle for-loop (over blocks in the set). #
	    ##while (my $bl = shift @std_sbj_blocks) {...} # Main middle while-loop (over blocks in the set). #

		my $bl = $std_sbj_blocks[$i];
		
		my ($shift, $beg_blk, $end_blk, $nodeid_mrca, $depth_mrca, $seqs_involved, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_blocks_shift_respos->[$bl]};

		my @tentative_cmp_blk = ($beg_blk, $end_blk, $nodeid_mrca, $br2change, [$bl], $seqs_involved, []);

		unless (@sbset_composite_blocks>0) {
			# Create a new composite block. #
		    push @sbset_composite_blocks, \@tentative_cmp_blk;
		    next;
		}

			# Retrieve the last (i.e., left-neighboring) block. #
		my $last_cmp_blk = $sbset_composite_blocks[$#sbset_composite_blocks];
		my $last_constituents = $last_cmp_blk->[4];
		my $last_blk = $last_constituents->[$#{$last_constituents}];

		my ($shift0, $beg_blk0, $end_blk0, $nodeid_mrca0, $depth_mrca0, $seqs_involved0, $br2change0, $excl_sbblocks0, $ct_changes0, $indices_src_classes0) = @{$set_blocks_shift_respos->[$last_blk]};

			# (By definition, $nodeid_mrca, $seqs_involved, and $br2changed must be identical to $nodeid_mrca0, $seqs_involved0, and $br2changed0, respectively.) #

			## First, examine whether a gap-mediated merger is possible or not. ##
			## (MODIFIED in further development beyond 'lump_together_similar_blocks_proto'.) ##

		my $if_gap_mediated = 1;

			# Examine consistency with the reconstructed MSA. #
		my $gapsize_rec = $beg_blk - ($end_blk0 +1);

		if ($gapsize_rec<0) { # This condition was introduced to avoid the loss of some non-trivial position-shifts (on Oct 22 (Wed), 2014). #
				# Create a new composite block. #
		    push @sbset_composite_blocks, \@tentative_cmp_blk;
		    next;
		}

		if ($gapsize_rec>0) {

			# See if the mediating region clicks with the gap in reconstructed MSA.

		    my ($beg_gap_rec, $end_gap_rec) = ($end_blk0+1, $beg_blk-1);

			# A BUG fixed on Oct 21 (Tue), 2014. #
		    my $info_gap_rec = $interval2gblk_rec{"$beg_gap_rec:$end_gap_rec"};
		    ##my $info_gap_rec = $interval2gblk_ref{"$beg_gap_rec:$end_gap_rec"};

	# %interval2gblk_rec; # = ("$beg_blk:$end_blk" => [$indx_cl, $indx_gb, $indx_gp], ...) for the gap-pattern blocks in the reconstructed MSA.

		    if (defined $info_gap_rec) {
			my $indx_gp = $info_gap_rec->[2];
			my $br2change_for_gp = $set_br2change_for_gps->[$indx_gp];

				# See if the gap-pattern in the mediating region interferes with the 'presence' pattern of the shift. #

			# $if_interfere = 1 if the two patterns are 'complementary' to each other,
			#               = 2 if they overlap each other,
			#               = 0 otherwise.
			#
			# $diff_level	= 0 if the two MRCAs either equal each other or are in different lineages,
			#		= -1 if the MRCA of the shift-block is the descendant of the MRCA for the gap-pattern,
			#		= +1 if the former is the ancestor of the latter.
			#
			my ($if_interfere, $mrca_sh_bl, $mrca_gp, $diff_level) = exam_interfere_bw_gap_patterns (%{$br2change}, %{$br2change_for_gp}, $gap, %{$node2pa}, %{$node2ch}, $topnode_id);

			if ($if_interfere) { $if_gap_mediated = 0; } # If the gap-pattern interfere with the shift-pattern, the two blocks can never merge in a gap-mediated manner.

		    } else {
			$if_gap_mediated = 0;
		    }

		} # END of "if ($gapsize_rec>0) {...}" #

			# Examine consistency with the reference MSA. #
		my $beg_blk_ref = $beg_blk - $shift;
		my $end_blk0_ref = $end_blk0 - $shift0;
		my $gapsize_ref = $beg_blk_ref - ($end_blk0_ref+1);

		if ($if_gap_mediated and ($gapsize_ref>0)) {

			# See if the mediating region clicks with the gap in reference MSA.

		    my ($beg_gap_ref, $end_gap_ref) = ($end_blk0_ref + 1, $beg_blk_ref-1);
		    my $info_gap_ref = $interval2gblk_ref{"$beg_gap_ref:$end_gap_ref"};
	# %interval2gblk_ref; # = ("$beg_blk:$end_blk" => [$indx_cl, $indx_gb, $indx_gp], ...) for the gap-pattern blocks in the reference MSA.

		    if (defined $info_gap_ref) {
			my $indx_gp = $info_gap_ref->[2];
			my $br2change_for_gp = $set_br2change_for_gps->[$indx_gp];

				# See if the gap-pattern in the mediating region interferes with the 'presence' pattern of the shift. #

			my ($if_interfere, $mrca_sh_bl, $mrca_gp, $diff_level) = exam_interfere_bw_gap_patterns (%{$br2change}, %{$br2change_for_gp}, $gap, %{$node2pa}, %{$node2ch}, $topnode_id);

			if ($if_interfere) { $if_gap_mediated = 0; } # If the gap-pattern interfere with the shift-pattern, the two blocks can never merge in a gap-mediated manner.

		    } else {
			$if_gap_mediated = 0;
		    }

		} # END of "if ($if_gap_mediated and ($gapsize_ref>0)) {...}" #

				### RESTART FROM HERE ###

		if ($if_gap_mediated) {
				# Incorporate the subject block into the last composite block. #

		    $last_cmp_blk->[1] = $end_blk;
		    push @{$last_constituents}, $bl;
		    push @{$last_cmp_blk->[6]}, 0;
		    next;

		} elsif ($gapsize_rec>0) {

				# Create a new composite block. #
		    push @sbset_composite_blocks, \@tentative_cmp_blk;
		    next;
		}

=For the predecessor 'lump_together_similar_blocks_proto'.
		if ($gapsize>0) {
			## When the two blocks are NOT contiguous. ##

			# Pre-test: Examine the consistency between the gap size and the shift difference.

		    my $diff_shift = $shift - $shift0;
		    unless ($gapsize == $diff_shift) {
				# Create a new composite block. #
			push @sbset_composite_blocks, \@tentative_cmp_blk;
			next;
		    }

			# Test 1a: Examine whether the mediating sites of the involved sequences are occupied by gaps alone. #

		    my $if_nongap = 0;
		    foreach my $seqid (@{$seqs_involved}) { # 1st main inner foreach-loop (over sequences involved). #

			my $row_shifts_respos = $map_shifts_respos->[$seqid];

			for (my $s=$end_blk0+1; $s<$beg_blk; $s++) { # Inner-inner for-loop (over sites between the blocks). #
			    ($row_shifts_respos->[$s] eq $gap) and next;
			    $if_nongap = 1;
			    last;
			}				# End of the inner-inner for-loop (over sites between the blocks). #
			$if_nongap and last;
		    }				# End of the 1st main inner foreach-loop (over sequences involved). #

		    if ($if_nongap) {
				# Create a new composite block. #
			push @sbset_composite_blocks, \@tentative_cmp_blk;
			next;
		    }

			# Test 1b: Examine whether at least one non-involved sequences has the mediating sites occupied by gaps alone. #
		    my ($common, $seqs_non_involved, $ex2) = diff_nmarray (@full_seq_indices, @{$seqs_involved});
			
		    my $if_gaprun = 0;
		    foreach my $seqid (@{$seqs_non_involved}) { # 2nd main inner foreach-loop (over sequences NOT involved). #

			my $row_shifts_respos = $map_shifts_respos->[$seqid];

			my $if_nongap2=0;
			for (my $s=$end_blk0+1; $s<$beg_blk; $s++) { # Inner-inner for-loop (over sites between the blocks). #
			    ($row_shifts_respos->[$s] eq $gap) and next;
			    $if_nongap2 = 1;
			}				# End of the inner-inner for-loop (over sites between the blocks). #
			$if_nongap2 and next;
			$if_gaprun = 1;
			last;
		    }				# End of the 2nd main inner foreach-loop (over sequences NOT involved). #

		    unless ($if_gaprun) {
				# Create a new composite block. #
			push @sbset_composite_blocks, \@tentative_cmp_blk;
			next;
		    }

				# Incorporate the subject block into the last composite block. #

		    $last_cmp_blk->[1] = $end_blk;
		    push @{$last_constituents}, $bl;
		    push @{$last_cmp_blk->[6]}, 0;

		    next;

		} # END of "if ($gapsize>0) {...}"

=cut # END of "For the predecessor 'lump_together_similar_blocks_proto'."


			# When the two (non-gap-mediated) blocks sit contiguously. #
			# => Examine whether the intermediate state candidate of either block can be incorporated into yet another block.

		my $intm_incp_type = 0; # = 1/2 if the last/current block was incorporated.

		foreach my $intm_cand (
			[$shift0 - $shift, $beg_blk0 - $shift, $end_blk0 - $shift, 1],
			[$shift - $shift0, $beg_blk - $shift0, $end_blk - $shift0, 2] ) { # 3rd main inner foreach-loop (over candidates of the intermediate state of the block). #

		    my ($shift_cand, $beg_cand, $end_cand, $type_cand) = @{$intm_cand};

		    my $shift_intvl_cand = "$shift_cand:$beg_cand:$end_cand";
		    my $blocks_cand = $shift_interval2blocks{$shift_intvl_cand};

		    (defined $blocks_cand) or next;

		    foreach my $blk_c (@{$blocks_cand}) { # Inner-inner foreach-loop (over blocks whose shift and region matched with those of the intermediate candidate). #

			my ($shift_c, $beg_blk_c, $end_blk_c, $nodeid_mrca_c, $depth_mrca_c, $seqs_involved_c, $br2change_c, $excl_sbblocks_c, $ct_changes_c, $indices_src_classes_c) = @{$set_blocks_shift_respos->[$blk_c]};

				# Examine whether the intermediate state candidate clicks together with the candidate block. #

			foreach my $br (sort {$a <=> $b} keys %{$br2change}) { # Innermost foreach-loop (over branches separating the involved sequences from the rest). #
			    my $change_c = $br2change_c->{$br};
			    (defined $change_c) or next;

			    my ($state_ue, $state_le) = @{$br2change->{$br}};
			    my ($state_ue_c, $state_le_c) = @{$change_c};

			    ($state_ue eq $state_le_c) or next;

				# The two blocks clicked together. #

			    $intm_incp_type = $type_cand;
			    last;
			}				# End of the innermost foreach-loop (over branches separating the involved sequences from the rest). #

		        ($intm_incp_type) and last;
		    }				# End of the inner-inner foreach-loop (over blocks whose shift and region matched with those of the intermediate candidate). #

		    ($intm_incp_type) and last;
		}					# End of the 3rd main inner foreach-loop (over candidates of the intermediate state of the block). #

		if ($intm_incp_type) {
			# If an intermediate candidate succeeded in being incorporated in another block, #
			# => incorporate the subject block into the last composite block. #

		    $last_cmp_blk->[1] = $end_blk;
		    push @{$last_constituents}, $bl;
		    push @{$last_cmp_blk->[6]}, $intm_incp_type;

		} else {
			# Otherwise, create a new composite block. #
		    push @sbset_composite_blocks, \@tentative_cmp_blk;
		    next;
		}

			# A BUG fixed (on Oct 18 (Sat), 2014). #
	    }					# End of the main middle for-loop (over blocks in the set). #
	    ##}					# End of the main middle while-loop (over blocks in the set). #

	    push @composite_blocks, @sbset_composite_blocks;

	}				# End of the main outer for-loop (over sets of blocks classified by their involved sequences). #


		# Finally, sort the composite blocks according to their beginning coordinates. #

	@composite_blocks = sort { $a->[0] <=> $b->[0] } @composite_blocks;

	return \@composite_blocks;

}	# End of the subroutine 'lump_together_similar_blocks'. # 


# exam_interfere_bw_gap_patterns:
#	This subroutine receives two sets of binary changes ('presence' <-> 'absence') across a tree,
#	and examines whether the resulting maps of 'presence' states interfere with each other.
#	(CREATED on Oct 21 (Tue), 2014.)
#
#	PREREQUISITE:
#	The map of 'presence' states on the nodes must be "phylogenetically correct."
#	That is, if any pair of nodes show the 'presence' state, all nodes along the path connecting them must also show the 'presence' state.
#
# INPUT: (\%br2change1, \%br2change2, $char_abs, \%node2pa, \%node2ch, $topnode_id), where
#
#	%br2change1/2 = ($branch_id => [$state_ue, $state_le], ...) for all branches with changes in the set 1/2,
#	with $state_ue/le being the state ('presence'/'absence') at the upper-end/lower-end of the branch;
#
#	$char_abs is the character representing the 'absence' state;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but root;
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#	$topnode_id is the ID of the top node (i.e., root) of the tree.
#
# OUTPUT: ($if_interfere, $mrca1, $mrca2, $diff_level), where
#
#	$if_interfere = 1 if the two maps of 'presence' on the nodes are largely 'complementary' to each other,
#		      = 2 if the two maps overlap each other,
#		      = 0 otherwise (i.e., they do not at all interfere with each other);
#
#	$diff_level	= 0 if either $mrca1 equals $mrca2 or they are in different lineages,
#			= -1 if $mrca1 is the descendant of $mrca2,
#			= +1 if $mrca1 is the ancestor of $mrca2,
#
sub exam_interfere_bw_gap_patterns (\%\%$\%\%$) {

	my ($br2change1, $br2change2, $char_abs, $node2pa, $node2ch, $topnode_id) = @_;

		# Preliminary Process: Identify the MRCA of the nodes with 'presence' state in each of sets 1 & 2. #

	my ($mrca1, $mrca2);

	foreach my $br (keys %{$br2change1}) { # 1st preliminary foreach-loop (over branches with change2 in set 1). #
	    my ($state_ue, $state_le) = @{$br2change1->{$br}};
	    if ($state_ue eq $char_abs) {
		$mrca1 = $br;
		last;
	    }
	}			# End of the 1st preliminary foreach-loop (over branches with changes in set 1). #
	(!(defined $mrca1) and (0<scalar (keys %{$br2change1}))) and ($mrca1 = $topnode_id); # If there is no 'insertion' (or 'creation') but is(are) some 'deletion(s)' (or 'annihilation(s)'), the top-node must be the MRCA.
	##(defined $mrca1) or ($mrca1 = $topnode_id); # If there is no 'insertion' (or 'creation'), the top-node must be the MRCA.

	foreach my $br (keys %{$br2change2}) { # 2nd preliminary foreach-loop (over branches with change2 in set 2). #
	    my ($state_ue, $state_le) = @{$br2change2->{$br}};
	    if ($state_ue eq $char_abs) {
		$mrca2 = $br;
		last;
	    }
	}			# End of the 2nd preliminary foreach-loop (over branches with changes in set 2). #
	(!(defined $mrca2) and (0<scalar (keys %{$br2change2}))) and ($mrca2 = $topnode_id); # If there is no 'insertion' (or 'creation') but is(are) some 'deletion(s)' (or 'annihilation(s)'), the top-node must be the MRCA.
	##(defined $mrca2) or ($mrca2 = $topnode_id); # If there is no 'insertion' (or 'creation'), the top-node must be the MRCA.


		## MAIN PROCESS: Examine the relationship between the two MRCAs. #

	unless ((defined $mrca1) and (defined $mrca2)) {
		# An EXCEPTIONAL case, where either (or both) of the maps is(are) null. => Consider them as 'non-interfering'. #
		# (ADDED on Oct 21 (Tue), 2014.) #

	    return (0, $mrca1, $mrca2, 0);
	}


	if ($mrca1 == $mrca2) {
	    return (2, $mrca1, $mrca2, 0); # The two 'presence' state maps must overlap each other, in one way or another.
	}

	my $pa_mrca1 = $node2pa->{$mrca1};
	my $pa_mrca2 = $node2pa->{$mrca2};

	if ((defined $pa_mrca1) and (defined $pa_mrca2) and ($pa_mrca1 == $topnode_id) and ($pa_mrca2 == $topnode_id) and (@{$node2ch->{$topnode_id}} == 2)) {
	##if (($pa_mrca1 == $topnode_id) and ($pa_mrca2 == $topnode_id) and (@{$node2ch->{$topnode_id}} == 2)) {...} # A BUG fixed on Oct 7 (Wed), 2015. #
	    return (1, $mrca1, $mrca2, 0); # The two 'presence' state maps must be largely complementary.
	}

	my ($common_ancestors, $lineage1, $lineage2) = fetch_common_ancestors ($mrca1, $mrca2, %{$node2pa});

	if ((@{$lineage1}>0) and (@{$lineage2}>0)) {
	    return (0, $mrca1, $mrca2, 0); # The two 'presence' state maps must have emerged independently, and thus must never interfere with each other.
	}

		# One of the MRCAs must be a 'descendant' of the other. #

	my ($bridging_nodes, $br2change_upper_map, $diff_level) = (@{$lineage1}>0) ? ($lineage1, $br2change2, -1) : ($lineage2, $br2change1, +1) ;

	my $mrca_lower_map = pop @{$bridging_nodes};

	my $chg_um_at_mrca_lm = $br2change_upper_map->{$mrca_lower_map};
	if (defined $chg_um_at_mrca_lm) {
	    my ($state_ue, $state_le) = @{$chg_um_at_mrca_lm};

	    if ($state_le eq $char_abs) {
			# If the 'upper-map' shows a 'deletion' along the branch jut above the MRCA of the 'lower-map',
			# that means that the two maps are largely complementary.
		return (1, $mrca1, $mrca2, $diff_level);
	    } else {
		my ($mrca_up, $mrca_lw) = ($diff_level>0) ? ($mrca1, $mrca2) : ($mrca2, $mrca1);
		print STDOUT "[ERROR 'exam_inferfere_by_gap_patterns'] IMPOSSIBLE! Found an 'insertion' along the branch above the lower MRCA '$mrca_lower_map' (the upper MRCA is '$mrca_up'): $! !\n\n";
		return (undef, $mrca1, $mrca2, $diff_level);
	    }
	}

	foreach my $bridge (@{$bridging_nodes}) { # Main foreach-loop (over branches bridging the two MRCAs). #

	    my $chg_um_brg = $br2change_upper_map->{$bridge};
	    if (defined $chg_um_brg) {
		my ($state_ue, $state_le) = @{$chg_um_brg};

		if ($state_le eq $char_abs) {
			# If the 'upper-map' shows a 'deletion' along a bridging branch (except the one above the lower MRCA),
			# that means that the two maps should never interfere with each other.
		    return (0, $mrca1, $mrca2, $diff_level);
		} else {
		    my ($mrca_up, $mrca_lw) = ($diff_level>0) ? ($mrca1, $mrca2) : ($mrca2, $mrca1);
		    print STDOUT "[ERROR 'exam_inferfere_by_gap_patterns'] IMPOSSIBLE! Found an 'insertion' along the branch '$bridge', which bridges the lower MRCA (='$mrca_lw') and the upper MRCA (='$mrca_up'): $! !\n\n";
		    return (undef, $mrca1, $mrca2, $diff_level);
		}
	    }
	}			# End of the main foreach-loop (over branches bridging the two MRCAs). #


	return (2, $mrca1, $mrca2, $diff_level); # Otherwise, the two maps must overlap each other.

}	# End of the subroutine 'exam_interfere_bw_gap_patterns'. #


1;   # a 'spell' which enables the module 'use'-able and 'require'-able #



