package MyTreeMap_IImlt;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeMap_IImlt.pm:
#	This revised version of 'MyTreeMap_II.pm' specializes in mapping
#	the histories of character changes at a site,
#	while (partially) resolving multifurcating nodes.
#	(CREATION BEGAN on Nov 27 (Sun), 2011, minimally essential part ENDED on Jan 9 (Mon), 2012)
#
#	The subroutines in this module work together to implement our newly devised parsimony algorithm,
#	which is detailed in the Additional file 2 of the article:
#	"Detecting negative selection on recurrent mutations using gene genealogy",
#	by Ezawa K, Landan G, Graur D, submitted to BMC Genetics.
#	(The article and additional files are available
#	on request to the first author at the above e-mail address.)
#
# MyTreeMap_II.pm:
#	This package performs more refined algorithms than 'MyTreeMap.pm',
#	to map various properties, or their changes,
#	onto the nodes or the branches of a phylogenetic tree.
#	(CREATED on June 16 (Wed), 2010)
#
# MyTreeMap.pm:
#	This package maps various properties, or their changes,
#	 onto the nodes or the branches
#	of a (molecular) phylogenetic tree.
#	It mainly uses the parsimony principle,
#	but sometimes uses other methods, such as the maximum likelihood.

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array nr_array);
use MyArrayManipu_num qw( diff_nmarray );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

use MyTreeManipu_III qw(
	make_node2abs_depth
	);

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	compare_varsize_nmarrays make_nonredundant_combinations
	search_minpen_prim_to_sec sort_secondary_characters
	sort_remaining_nodes_per_secchars 
	enumerate_core_maxdroppen_combi_secchars
	part_resolve_mlt_node_bu
	min_penalties_sankoff_ptrslv_bo
	enum_parsim_scenarios_sankoff_ptrslv_td
	map_parsim_changes_sankoff_ptrslv_1clm
	);

=for the predecessor 'MyTreeMap_II.pm'
our @EXPORT_OK = qw(
	min_penalties_sankoff_bo enum_node2parsstat_sankoff_td
	map_parsim_states_sankoff_clm
	map_parsim_states_sankoff_msa map_parsim_states_sankoff_1clm
	map_parsim_gapstates_sankoff locate_parsim_changes
	);
=cut

sub compare_varsize_nmarrays (\@\@$);
sub make_nonredundant_combinations (\@$);

sub assign_uniq_characters (\@\@\%);
sub search_minpen_prim_to_sec ($\@\@\@);
sub minpen_among_specchars_for_sbnodes ($\@\@\%);
sub diff_penalties_specchars_for_sbnodes (\@\@\%\@);
sub sort_secondary_characters (\@\@);
sub sort_remaining_nodes_per_secchars (\@\@);
sub make_initcand_max_droppen_combi_sc ($\@\@);
sub copy_cand_maxdroppen_combi_secchars (\@);
sub incorporate_secchar_into_seccombi (\@$\@\@\@\@);
sub search_downhill_cand_maxdroppen_sc (\@$$\@\@\@\@\@$$);
sub search_back_forth_lin_maxdroppen_sc (\@$\@\@\@\@\@);
sub dress_core_maxdroppen_combi_secchars (\@\@\@\@\@);
sub attach_keeping_secs_to_maxdroppen_combi (\@\@\@\@\@);
sub enumerate_maxdroppen_char_patterns (\@\@\@\@);
sub make_parsim_scenarios_fm_pattern (\@\@\%$$\$\%\@\%);
sub add_parsim_scenarios_by_chng_uniq_chars (\@\%\%$$\$\@\%);

	# This specifies the level of tests on main subroutines.
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;
	# This specifies the level of tests on the main satellite subroutine 'part_resolve_mlt_node_bu'.
our $TEST_LEVEL1 = 0;
##our $TEST_LEVEL1 = 2;
##our $TEST_LEVEL1 = 5;
	# This specifies the level of tests on important satellite subroutines.
our $TEST_LEVEL2 = 0;
##our $TEST_LEVEL2 = 2;
##our $TEST_LEVEL2 = 5;

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.

	## These two indices specify where the information is stored
	## in %nodee2pars_stats = ($node_id => \@stat_info, ...) #
our $INDX_STAT = 0; # The state (or the character) mapped onto the node.
our $INDX_SBTOT_PEN = 1; # The subtotal of penalties on the subtree under $node_id. 

	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the NEW SUBROUTINES ###

# compare_varsize_nmarrays:
#	This subroutine compared two arrays whose sizes might be different (but might also be the same),
#	and returns the result.
#	ALGORITHM:
#	The two arrays will be compared as follows, starting with $k=0:
#	(i) numerically compare the $k th components of the two arrays,
#	   if the components differ from each other,
#		judge the array with the smaller component as 'smaller',
#	   if the components equal to each other, increment $k by 1 and return to the beginning of the loop;
#	(ii) When only one of the array exhausts the component, that array is judged as 'smaller'.
#	    When both arrays exhausts the component at the same time, they are judged as 'equal' to each other.
#
# INPUT: (\@src_array1, \@src_array2, $indx), where
#	@{$src_array1->[$indx}} and @{$src_array2->[$indx]},
#	or @{$src_array1} and @{$src_array2} if $indx = undef,
#	are the two arrays to be compared.
#
# OUTPUT: $sgn, where
#	$sgn = -1 if the first array is 'smaller than' the second array,
#	     = 1 if the first array is 'larger than' the second array,
#	     = 0 if the two arrays are equal to each other.
#
sub compare_varsize_nmarrays (\@\@$) {

	my ($src_array1, $src_array2, $indx) = @_;

	my $array1 = (defined $indx) ? $src_array1->[$indx] : $src_array1;
	my $array2 = (defined $indx) ? $src_array2->[$indx] : $src_array2;
	unless ((defined $array1) and (defined $array2)) {
	   (defined $array1) or ($array1 = 'undef');
	   (defined $array2) or ($array2 = 'undef');
	    print STDOUT "[ERROR 'compare_varsize_nmarrays'] ILLEGAL INPUT: FOund (an) UNDEFINED ARRAY(s): (\$rf_array1, \$rf_array2) = ($array1, $array2): $! !\n";
	    return undef;
	}

	my $size1 = @{$array1};
	my $size2 = @{$array2};
	my $min_size = ($size1 < $size2) ? $size1 : $size2;

	for (my $k=0; $k<$min_size; $k++) { # Main for-loop #
	    my $ele1 = $array1->[$k];
	    my $ele2 = $array2->[$k];

	    if ($ele1 < $ele2) {
		return -1; # The 1st array is 'smaller'.

	    } elsif ($ele1 > $ele2) {
		return 1; # The 1st array is 'larger'.
	    }
	}		# End of the main for-loop #

		# Either or both of the arrays exhausted the components.

	if ($size1 < $size2) {
	    return -1; # The 1st array is 'smaller than' the 2nd array.
	} elsif ($size1 > $size2) {
	    return 1; # The 1st array is 'larger than' the 2nd one.
	} else {
	    return 0; # The two arrays are equal to each other.
	}

}		# End of the subroutine 'compare_varsize_nmarrays' #


# make_nonredundant_combinations:
#	This quite general subroutine makes a non-redundant set of combinations,
#	each of which should be specified by a set of indices of the elements (in an array, say, @elements)
#	that constitute the combination.
#
# INPUT: (\@set_combinations, $indx_combi), where
#	@set_combinations is a set of combinations, with
#	@{$set_combinations[$k]->[$indx_combi]},
#	or @{$set_combinations->[$k]} if $indx_combi = undef,
#	listing the indices of all the elements making up the $k th combination.
#
# OUTPUT: \@nr_set_combinations,
#	which is the non-redundant set.
#
sub make_nonredundant_combinations (\@$) {

	my ($set_combinations, $indx_combi) = @_;

		# Make a new set whose elements each has a sorted set of indices. #
	my @new_set_combinations = ();
	foreach my $combination (@{$set_combinations}) { # 1st foreach-loop #
	    my @array_combi;
	    if (defined $indx_combi) {
	        @array_combi = @{$combination};
		my @indices = sort {$a <=> $b} @{$combination->[$indx_combi]};
		$array_combi[$indx_combi] = \@indices;
	    } else {
	        @array_combi = sort {$a <=> $b} @{$combination};
	    }
	    push @new_set_combinations, \@array_combi;
	}			# End of the 1st foreach-loop #

	my @std_set_combinations = sort { compare_varsize_nmarrays (@{$a}, @{$b}, $indx_combi) } @new_set_combinations;

	my @nr_set_combinations = (shift @std_set_combinations);
	my $last_combi = $nr_set_combinations[0];

	while (my $combi = shift @std_set_combinations) { # 2nd foreach-loop #
	    my $sgn = compare_varsize_nmarrays (@{$last_combi}, @{$combi}, $indx_combi);
	    if ($sgn>0) {
		# This canNOT happen logically, because the combinations are sorted in advance using 'compare_varsize_nmarrays'. #

		my @last_indices = (defined $indx_combi) ? @{$last_combi->[$indx_combi]} : @{$last_combi};
		my @curr_indices = (defined $indx_combi) ? @{$combi->[$indx_combi]} : @{$combi};
		print STDOUT "[ERROR 'make_nonredundant_combinations'] It's IMPOSSIBLE ! The current combination \[= (", join (', ', @curr_indices), ") \] is 'SMALLER THAN' the last combination \[= (", join (', ', @last_indices), ") \]: $! !\n";
		return undef;

	    } elsif ($sgn < 0) {
		# In this case, $combi is different from $last_combi, so add it to the non-redundant set.
		push @nr_set_combinations, $combi;
		$last_combi = $combi;

	    } else {
		# In this case, $combi should be identical to $last_combi, so discard $combi.
		next;
	    }
	}		# End of the 2nd foreach-loop #

	return \@nr_set_combinations;

}		# End of the subroutine 'make_nonredundant_combinations' #


# assign_uniq_characters:
#	This satellite subroutine of the subroutine 'part_resolve_mlt_node_bu'
#	tries to assign a 'unique character' to each node.
#	A 'unique character' is defined as the character that gives the minimum 
#	total penalty (along the branch from the parent to the node in question and along the subtree under the node) whatever character the parent node has.
#	(NOTE that the 'unique character' does not necessarily have to be literally uniquely minimum-penalized character, but it could tie the other minimum penalized character(s) when the parent node does not have the 'unique character'.)
#
# INPUT: (\@nodes, \@pen_mtrx, \%node2min_penalties), where
#	@nodes lists the IDs of the nodes to be examined;
#	@children lists the node ID of the child nodes;
#	$pen_mtrx[$i]->[$j] is the penalty imposed on the change
#	from the $i th character to the $j th character;
#	%node2min_penalties = ($node_id => \@min_penalties, ...) for ALL nodes,
#	with $min_penalties[$i] being the minimum summed penalty
#	on the subtree under $node_id under the condition that the node has
#	the $i th character in the matrix @pen_mtrx.
#	($min_penalties[$i] is undef if it is known in advance
#	that the $i th character will NEVER be included in ANY parsimonious scenarios.)
#	
# OUTPUT: (\%node2uniq_char, \%uniq_char2nodes, \@remain_nodes), where
#	%node2uniq_char = ($node_id => [$uniq_char_id, $node2min_penalties{$node_id}->[$uniq_char_id]], ...) is for the nodes with unique characters assigned;
#	%uniq_char2nodes = ($char_id => \@node_ids, ...) for the characters which is assigned as 'uniquely parsimonious' character to some nodes;
#	@remain_nodes lists all the nodes NOT assugned unique characters.
#
sub assign_uniq_characters (\@\@\%) {

	my ($nodes, $pen_mtrx, $node2min_penalties) = @_;

	my $ct_chars = @{$pen_mtrx};

	my %node2uniq_char; my %uniq_char2nodes;
	my @remain_nodes = ();

	foreach my $node_id (@{$nodes}) { # Main outer foreach-loop (over nodes) #

	    my $cond_min_pens = $node2min_penalties->{$node_id};
	    unless (defined $cond_min_pens) {
		print STDERR "[ERROR 'assign_uniq_characters'] NO MINIMUM PENALTY INFORMATION on the node '$node_id': $! !\n";
		return (undef, undef, undef);
	    }

	    my $min_pen ;
	    my @minpen_chars = (); # characters to give the minimum penalty #
	    for (my $i=0; $i<$ct_chars; $i++) { # 1st middle for-loop (over characters) #
		my $cond_minpen = $cond_min_pens->[$i];
		(defined $cond_minpen) or next;
	
		if (!(defined $min_pen) or ($min_pen>$cond_minpen)) { # Update the minimum penalty #
		    $min_pen = $cond_minpen;
		    @minpen_chars = ($i);
		} elsif ($min_pen<$cond_minpen) {
		    next; # Do nothing #
		} else {
		    push @minpen_chars, $i; # Expand the list #
		}
	    }			# End of the 1st middle for-loop (over characters) #

	    if (@minpen_chars < 1) {
		print STDERR "[ERROR 'assign_uniq_characters'] NO MINIMUM-PENALTY CHARACTER for the node '$node_id' : $! !\n";
		return (undef, undef, undef);
	    } elsif (@minpen_chars > 1) {
		push @remain_nodes, $node_id;
		next;
	    }

	    my $minpen_char = $minpen_chars[0];
	    my $if_uniq = 1;
	    for (my $i=0; $i<$ct_chars; $i++) { # 2nd middle for-loop (over parent characters) #
		($i == $minpen_char) and next; # When the parent has the minimum-penalty character, it should definitely give the minimum penalty. #
		my $pen_row = $pen_mtrx->[$i];
		my $sbjct_pen = $min_pen + $pen_row->[$minpen_char]; # The penalty when the parent has the $i th character and the subject has $mipen_char th character. #
		for (my $j=0; $j<$ct_chars; $j++) { # inner for-loop (over subject characters) #
		    ($j == $minpen_char) and next; # Avoid self-comparisons. #
		    my $pen = $cond_min_pens->[$j];
		    (defined $pen) or next; # Skip non-parsiminious characters. #
		    $pen += $pen_row->[$j];
		    if ($pen < $sbjct_pen) { # $minpen_char does NOT give the minimum penalty when the parent has the $i th character!! #
			$if_uniq = 0;
			last;
		    }
		    
		}		# End of the inner for-loop (over subject characters) #
		$if_uniq or last;
	    }		# End of the 2nd middle for-loop (over parent characters) #

	    unless ($if_uniq) { # The node does NOT have a 'unique character'.
		push @remain_nodes, $node_id;
		next;
	    }
		# $minpen_char is the 'unique character' of $node_id. #

	    $node2uniq_char{$node_id} = [$minpen_char, $min_pen];
	    if (defined $uniq_char2nodes{$minpen_char}) {
		push @{$uniq_char2nodes{$minpen_char}}, $node_id;
	    } else {
		$uniq_char2nodes{$minpen_char} = [$node_id];
	    }

	}			# End of the main outer foreach-loop (over nodes) #


	return (\%node2uniq_char, \%uniq_char2nodes, \@remain_nodes);

}	# End of the subroutine 'assign_uniq_characters' #


# search_minpen_prim_to_sec:
#	This satellite subroutine of the subroutine 'part_resolve_mlt_node_bu'
#	searches for the primary characters (either the parent character for 'unique_character's)
#	that gives the minimum penalty to each secondary character.
#
# INPUT: ($pchar, \@uniq_chars, \@sec_chars, \@pen_mtrx), where
#	$pchar is the index of the parent character;
#	@uniq_chars lists the indices of 'unique character's;
#	@sec_chars list the indices of secondary characters (which are neither parent nor unique characters);
#	$pen_mtrx[$i]->[$j] is the penalty for a change from the $i th character to the $j th character
#	(the order is in the whole set of characters);
#
# OUTPUT: (\@minpen_prim_to_sec, \@best_prims_for_sec), where
#	$minpen_prim_to_sec[$c] is the minimum penalty for a change from a primary character to $sec_chars[$c];
#	@{$best_prims_for_sec[$c]} lists the indices of the primary characters that give the minimum penalty to $sec_chars[$c].
#
sub search_minpen_prim_to_sec ($\@\@\@) {

	my ($pchar, $uniq_chars, $sec_chars, $pen_mtrx) = @_;

	my $ct_secs = @{$sec_chars};
	my @minpen_prim_to_sec = my @best_prims_for_sec = ();

	for (my $c=0; $c<$ct_secs; $c++) { # Main for-loop (over secondary characters) #

		# A BUG Fixed on Dec 21 (Wed), 2011 #
	    my $sec_char = $sec_chars->[$c];
	    my $minpen = $pen_mtrx->[$pchar][$sec_char];
	    ##my $minpen = $pen_mtrx->[$pchar][$c];

	    my @best_prims = ($pchar);

	    foreach my $prim (@{$uniq_chars}) { # Inner for-loop (over 'unique' characters) #

		# A BUG Fixed on Dec 21 (Wed), 2011 #
		my $pen = $pen_mtrx->[$prim][$sec_char];
		##my $pen = $pen_mtrx->[$prim][$c];

		($pen>$minpen) and next; # This primary character doesn't provide a minimum penalty. #
		if ($minpen > $pen) {
			# Update the minimum penalty. #
		    $minpen = $pen;
		    @best_prims = ($prim);
		} else {
			# Add the unique character to the set of minimum-penalty-giving primaries. #
		    push @best_prims, $prim;
		}
	    }		# End of the inner for-loop (over 'unique' characters) #

	    $minpen_prim_to_sec[$c] = $minpen;
	    $best_prims_for_sec[$c] = \@best_prims;

	}		# End of the main for-loop (over secondary characters) #

	return (\@minpen_prim_to_sec, \@best_prims_for_sec);

}		# End of the subroutine 'search_minpen_prim_to_sec' #


# minpen_among_specchars_for_sbnodes:
#	This satellite subroutine of the subroutine 'part_resolve_mlt_node_bu'
#	identifies the minimum penalty among those given by the specified characters
#	at each of the subject nodes.
#
# INPUT: ($first_char, \@spec_chars, \@sbjct_nodes, \%node2min_penalties), where
#	($first_char, @spec_chars) lists indices of all the specified characters (non-redundantly),
#	and $first_char could be a special one if you want;
#	@sbjct_nodes lists the IDs of the nodes to be examined;
#	%node2min_penalties = ($node_id => \@min_penalties, ...) for ALL nodes,
#	with $min_penalties[$i] being the minimum summed penalty
#	on the subtree under $node_id under the condition that the node has
#	the $i th character in the matrix @pen_mtrx.
#	($min_penalties[$i] is undef if it is known in advance
#	that the $i th character will NEVER be included in ANY parsimonious scenarios.)
#	
# OUTPUT: (\@spec_nodes, \@minpen_amng_spec_for_spnodes, \@minpen_spec_chars_for_spnodes, \@nonspec_nodes), where
#	@spec_nodes lists the IDs of the nodes for which some of the specified characters can give the minimum penalty;
#	$minpen_amng_spec_for_spnodes[$sp] is the minimum penalty for the subtree under the node with ID = $spec_nodes[$sp] among those given by the specified characters; and
#	@{$minpen_spec_chars_for_spnodes[$sp]} lists all such specified characters giving the minimum penalty;
#	@nonspec_nodes list the IDs of the nodes whose minimum penalties will NEVER be given by the specified characters;
#
sub minpen_among_specchars_for_sbnodes ($\@\@\%) {

	my ($first_char, $spec_chars, $sbjct_nodes, $node2min_penalties) = @_;

	my $ct_sbjcts = @{$sbjct_nodes};

	my @spec_nodes = my @nonspec_nodes = ();
	my @minpen_amng_spec_for_spnodes = my @minpen_spec_chars_for_spnodes = ();

	for (my $sb=0; $sb<$ct_sbjcts; $sb++) { # Main outer for-loop (over subject nodes) #
	    my $node_id = $sbjct_nodes->[$sb];
	    my $cond_min_pens = $node2min_penalties->{$node_id};
	    unless (defined $cond_min_pens) {
		print STDERR "[ERROR 'minpen_among_specchars_for_sbnodes'] The $sb th subject node (ID: $node_id) does NOT HAVE its own PENALTY INFORMATION: $! !\n";
		return (undef, undef);
	    }

	    my $minpen;
	    my @minpen_chars = ();
	    foreach my $char ($first_char, @{$spec_chars}) { # Inner for-loop (over specified characters) #
		my $pen = $cond_min_pens->[$char];
		(defined $pen) or next; # Skip characters that will NEVER contribute to parsiminious scenarios. #
		if (!(defined $minpen) or ($minpen > $pen)) { # Update the minimum penalty #
		    $minpen = $pen;
		    @minpen_chars = ($char);
		} elsif ($minpen < $pen) {

		   next; # Do nothing
		} else { # Incorporate the character into the set of minimum characters #
		    push @minpen_chars, $char;
		}
	    }		# End of the inner for-loop (over specified characters) #

	    unless (defined $minpen) { # NO CHANCE that the specified characters will give the minimum penalty. #
		push @nonspec_nodes, $node_id;
		next; # ADDED to fix a BUG on Dec 12 (Mon), 2011. #
	    }

		# Some of the specified characters may still give the minimum penalty. #

	    push  @spec_nodes, $node_id;
	    push @minpen_amng_spec_for_spnodes, $minpen;
	    push @minpen_spec_chars_for_spnodes, \@minpen_chars;

	}	# End of the main outer for-loop (over subject nodes)  #

	return (\@spec_nodes, \@minpen_amng_spec_for_spnodes, \@minpen_spec_chars_for_spnodes, \@nonspec_nodes);

}		# End of the subroutine 'minpen_among_specchars_for_sbnodes' #


# diff_penalties_specchars_for_sbnodes:
#	This satellite subroutine of the subroutine 'part_reslve_mlt_node_bu'
#	calculates the differences of the penalties on the specified characters
#	from the 'base-points' for each of the subject nodes.
#
# INPUT: (\@spec_chars, \@sbnodes, \%node2min_penalties, \@base_points), where
#	@spec_chars list the indices of the specified characters;
#	@sbnodes lists the IDs of the subject nodes;
#	%node2min_penalties = ($node_id => \@min_penalties, ...) for ALL nodes,
#	with $min_penalties[$i] being the minimum summed penalty
#	on the subtree under $node_id under the condition that the node has
#	the $i th character in the matrix @pen_mtrx.
#	($min_penalties[$i] is undef if it is known in advance
#	that the $i th character will NEVER be included in ANY parsimonious scenarios.);
#	$base_points[$sn] gives the 'base-point' with which the penalties on the node with ID = $sbnodes[$sn] will be compared.
#	It is undef when there is no 'base-point' so far.
#
# OUTPUT: (\@diffpens_specchars_for_sbnodes, \@tot_diffpens_specchars), where
#	$diffpens_specchars_for_sbnodes[$sp]->[$sn] is the differential penalty when the node with ID = $sbodes[$sn] has the $spec_chars[$sp] th character from the 'base-point'.
#	If $base_points[$sn] = undef, then the absolute penalty will be filled in.
#	(The entry will be set UNDEFINED if either the relevant penalty is undefined OR if the entry is going to be negative);
#	$tot_diffpens_specchars[$sp] 
#	 = summation of DEFINED $diffpens_specchars_for_sbnodes[$sp]->[$sn] over the nodes in @sbnodes for the character $spec_chars[$sp],
#	( = undef if ALL the $diffpens_specchars_for_sbnodes[$sp]->[$sn] are undefined.)
#
sub diff_penalties_specchars_for_sbnodes (\@\@\%\@) {

	my ($spec_chars, $sbnodes, $node2min_penalties, $base_points) = @_;

	my $ct_nodes = @{$sbnodes};
	my $ct_chars = @{$spec_chars};

	my @diffpens_specchars_for_sbnodes = ();
	my @tot_diffpens_specchars = ();

	if ($ct_nodes==0) {
	    for (my $sp=0; $sp<$ct_chars; $sp++) {
	    }
	}

	for (my $sp = 0; $sp<$ct_chars; $sp++) { # Main outer for-loop #

		# ADDED to cope with the case with empty @{$sbnodes}. (on Dec 27 (Tue), 2011) #
	    if ($ct_nodes==0) {
		    # In this case, total differential penalties should be zero by definition. #
		$diffpens_specchars_for_sbnodes[$sp] = [];
		$tot_diffpens_specchars[$sp] = 0;
		next;
	    }

	    my $char = $spec_chars->[$sp];
	    my @diffpens_for_sbnodes = ();
	    my $tot_diffpens;
	    for (my $sn = 0; $sn<$ct_nodes; $sn++) { # Main inner for-loop #
		my $node = $sbnodes->[$sn];
		my $cond_min_pens = $node2min_penalties->{$node};
	 	unless (defined $cond_min_pens) {
		    print STDERR "[ERROR 'diff_penalties_specchars_for_sbnodes'] NO CONDITIONAL MINIMUM PENALTY information for the $sn th subject node (ID: $node): $! !\n";
		    return (undef, undef);
		}
		my $cond_minpen = $cond_min_pens->[$char];
		unless (defined $cond_minpen) {
		    $diffpens_for_sbnodes[$sn] = undef;
		    next;
		}
		(defined $tot_diffpens) or ($tot_diffpens = 0);

		my $base = $base_points->[$sn];
		my $diffpen = (defined $base) ? $cond_minpen - $base : $cond_minpen;

		if ((defined $base) and ($diffpen > 0)) { # If no improvement (or unchange) cannot be expected, don't use the character for the node. #
		    $diffpens_for_sbnodes[$sn] = undef;
		    next;
		}
		$diffpens_for_sbnodes[$sn] = $diffpen;
		$tot_diffpens += $diffpen;
	    }			# End of the main inner for-loop #

	    $diffpens_specchars_for_sbnodes[$sp] = \@diffpens_for_sbnodes;
	    $tot_diffpens_specchars[$sp] = $tot_diffpens;

	}		# End of the main outer for-loop #

	return (\@diffpens_specchars_for_sbnodes, \@tot_diffpens_specchars);

}		# End of the subroutine 'diff_penalties_specchars_for_sbnodes' #


# sort_secondary_characters:
#	This satellite subroutine of the subroutine 'part_resolve_mlt_node_bu'
#	sorts out secondary characters according to the total penalty difference.
#
# INPUT: (\@tot_diffpens_secchars_for_rem, \@minpen_prim_to_sec), where
#	$tot_diffpens_secchars_for_rem[$sc] is the sum of the differntial penalties over all the remaining nodes $remain_nodes->[$r] that have $sec_chars[$sc] th character giving the penalty smaller than or equal to $minpen_amng_prim_for_rnodes->[$r], subtracted by $minpen_amng_prim_for_rnodes->[$r] per each remaining node;
#	$minpen_prim_to_sec[$sc] gives the minimum penalty for a change from a primary character (in ($pchar, @uniq_chars)) to $sec_chars[$sc];
#
# OUTPUT: (\@improving_secs, \@keeping_secs, \@no_chance, \@max_droppen_by_rem),
#	$sec_chars[$improving_secs[$i]] ($i=0,...,$#{$improving_secs}) gives the secondary characters that are bound to improve the total penalty over primary characters;
#	Secondary characters in @improving_secs are sorted in order of decreasing total penalty drop;
#	$sec_chars[$keeping_secs[$i]] ($i=0,...,$#{$keeping_secs}) gives the secondary character that can tie the primary nodes at the best;
#	$sec_chars[$no_chance[$i]] ($i=0,...,$#{$no_chance}) gives the secondary character that 
#	$max_droppen_by_rem[$sc] gives the maximum possible drop in the penalty when the character $sec_chars[$sc] is used for some of the 'remaining nodes'.
#
sub sort_secondary_characters (\@\@) {

	my ($tot_diffpens_secchars_for_rem, $minpen_prim_to_sec) = @_;

	my $ct_secchars = @{$tot_diffpens_secchars_for_rem};
	unless (@{$minpen_prim_to_sec} == $ct_secchars) {
	    print STDOUT "[ERROR 'sort_secondary_characters'] ILLEGAL INPUT: The size of \@\{\$minpen_prim_to_sec\} (=", scalar (@{$minpen_prim_to_sec}), ") CLASHES with that of \@\{$tot_diffpens_secchars_for_rem\} (=$ct_secchars): $! !\n";
	    return (undef, undef, undef, undef);
	}

	my @max_droppen_by_rem = ();
	my @improving_secs = my @keeping_secs = my @no_chance = ();
	for (my $sc=0; $sc<$ct_secchars; $sc++) { # Main for-loop (over secondary characters) #
	    my $max_droppen = $tot_diffpens_secchars_for_rem->[$sc] + $minpen_prim_to_sec->[$sc];
	    $max_droppen_by_rem[$sc] = $max_droppen;

	    if ($max_droppen<0) { # The character could improve the penalty.
		push @improving_secs, $sc;

	    } elsif ($max_droppen>0) { # The character canNOT even keep the penalty (by the primary characters).
		push @no_chance, $sc;
	    } else {
		push @keeping_secs, $sc;
	    }
	}		# End of the main for-loop (over secondary characters) #

	my @std_improving_secs = sort {$max_droppen_by_rem[$a] <=> $max_droppen_by_rem[$b] } @improving_secs;

	return (\@std_improving_secs, \@keeping_secs, \@no_chance, \@max_droppen_by_rem);

}	# End of the subroutine 'sort_secondary_characters' #


# sort_remaining_nodes_per_secchars:
#	This satellite subroutine of the subroutine 'part_resolve_mlt_node_bu'
#	sorts the remaining nodes into those where a specified secondary character
#	improves the penalty over primary characters,
#	and those where a specified secondary character ties the least-scoring primary character.
#
# INPUT: (\@spec_secchars, \@diffpens_secchars_for_rem), where
#	@spec_secchars lists the indices (for @sec_chars) of secondary characters to be examined;
#	$diffpens_secchars_for_rem[$sc]->[$r]is the differential penalty when the node with ID $remain_nodes->[$r] has the $sec_chars[$sc] character, subtracted by $minpen_amng_prim_for_pnodes->[$r];
#
# OUTPUT: \@key_marginal_nodes_for_spec, where
#	@{$key_marginal_nodes_for_spec[$spc]} = (\@key_nodes, \@marginal)
#	for the secondary character $sec_chars[$spec_secchars[$spc]], with
#	$remain_nodes->[$key_nodes[$kn]] being the ID of the $kn th remaining node
#	where the secondary character has a lower penalty than any primary characters, and with
#	$remain_nodes->[$marginal[$mg]] beging the ID of the $mg th remaining node
#	where the secondary character ties the lowest-penalty primary node.
#
sub sort_remaining_nodes_per_secchars (\@\@) {

	my ($spec_secchars, $diffpens_secchars_for_rem) = @_;

	my $ct_spec = @{$spec_secchars};

	my @key_marginal_nodes_for_spec = ();

	for (my $spc=0; $spc<$ct_spec; $spc++) { # Main outer for-loop #
	    my $indx_sec = $spec_secchars->[$spc];
	    my $diffpens_for_rem = $diffpens_secchars_for_rem->[$indx_sec];

	    my $ct_rem = @{$diffpens_for_rem};
	    my @key_nodes = my @marginal = ();
	    for (my $r = 0; $r<$ct_rem; $r++) { # Main inner for-loop #
		my $diffpen = $diffpens_for_rem->[$r];
		(!(defined $diffpen) or ($diffpen > 0)) and next; # The character will never reduce the penalty on the node.

		if ($diffpen < 0) { # The character could reduce the total penalty. #
		    push @key_nodes, $r;
		} else { # The character could maintain the total penalty. #
		    push @marginal, $r;
		}
	    }			# End of the main inner for-loop #

	    $key_marginal_nodes_for_spec[$spc] = [\@key_nodes, \@marginal];
	}		# End of the main outer for-loop #

	return \@key_marginal_nodes_for_spec;

}		# End of the subroutine 'sort_remaining_nodes_per_secchars' #


# make_initcand_max_droppen_combi_sc:
#	This satellite subroutine of 'enumerate_max_droppen_combi_secchars'
#	creates the main ingredients of the initial candidate of
#	a secondary character combination providing a max drop of penalty.
#
# INPUT: ($init_indx_sec, \@key_marginal_nodes_for_sec, \@diffpens_secchars_for_rem), where
#	$init_indx_sec is an index for @sec_chars that specifies the current initial candidate combination;
#	@{$key_marginal_nodes_for_sec[$sc]} = (\@key_nodes, \@marginal)
#	for the secondary character $sec_chars[$sc], with
#	$remain_nodes->[$key_nodes[$kn]] being the ID of the $kn th remaining node
#	where the secondary character has a lower penalty than any primary characters.
#	(Here \@marginal will not be necessary);
#	$diffpens_secchars_for_rem[$sc]->[$r]is the differential penalty when the node with ID $remain_nodes->[$r] has the $sec_chars[$sc] character, subtracted by $minpen_amng_prim_for_pnodes->[$r];
#	$max_droppen_by_sec_over_rem[$sc] gives the maximum possible drop in the penalty when the character $sec_chars[$sc] is used for some of the 'remaining nodes'.
#
# OUTPUT: (\@init_secchars_for_rnode, \@init_droppen_for_rnode), which
#	determines the initial candidate combination of 'improving' secondary characters on the 'remaining nodes' that gives the maximum drop in the penalty (compared with the 'best' combination of primary characters), with
#	@{$init_secchars_for_rnode[$r]} = ($init_indx_sec)
#	for the $remain_nodes->[$r] th node when the $init_indx_sec th secondary character improves the penalty (= undef if the character cannot improve the penalty at the node), with
#	$init_droppen_for_rnode[$r]  being the penalty drop achieved for the $remain_nodes->[$r] th node by the $init_indx_sec th secondary character ( = undef if the enalty at the node does NOT improve over the best primary).
#
sub make_initcand_max_droppen_combi_sc ($\@\@) {

	my ($init_indx_sec, $key_marginal_nodes_for_sec, $diffpens_secchars_for_rem) = @_;

	my $key_nodes = $key_marginal_nodes_for_sec->[$init_indx_sec][0];
	my $diffpens_for_rem = $diffpens_secchars_for_rem->[$init_indx_sec];
	my $ct_rnodes = @{$diffpens_for_rem};

	my @init_secchars_for_rnode = my @init_droppen_for_rnode = ();
	$init_secchars_for_rnode[$ct_rnodes-1] = undef;
	$init_droppen_for_rnode[$ct_rnodes-1] = undef;

	foreach my $indx_node (@{$key_nodes}) { # Main foreach-loop #
	    my $diffpen_node = $diffpens_for_rem->[$indx_node];
	    unless ((defined $diffpen_node) and ($diffpen_node < 0)) {
		print STDOUT "[ERROR 'make_initcand_max_droppen_combi_sc'] ILLEGAL INPUT: NON_NEGATIVE PENALTY DIFFERENCE (= $diffpen_node) for the $init_indx_sec th secondary character on the $indx_node th 'remaining' node: $! !\n";
		return (undef, undef);
	    }
	    $init_secchars_for_rnode[$indx_node] = [$init_indx_sec];
	    $init_droppen_for_rnode[$indx_node] = $diffpen_node;
	}		# End of the main foreach-loop # 

	return (\@init_secchars_for_rnode, \@init_droppen_for_rnode);

}	# End of the subroutine 'make_initcand_max_droppen_combi_sc' #


# copy_cand_maxdroppen_combi_secchars:
#	This satellite subroutine of 'enumerate_max_droppen_combi_secchars'
#	copies a candidate combination of secondary caracters with maximum drop in penalty.
#
# INPUT: \@cand_max_droppen_combi_secchars, where
#	@cand_max_droppen_combi_secchars = (\@secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $max_droppen_for_combi).
#	NOTE that @secchars_for_rnode is an ARRAY OF ARRAYS,
#	some of whose elements may be undefined. [--- ADDED to solicit caution on Dec 21 (Wed), 2011.]
#	
# OUTPUT: \@copy_cand, where
#	@copy_cand is an exact copy of @cand_max_droppen_combi_secchars
#	but references to arrays, for example, are replaced by
#	new references to the COPYES OF arrays.
#
sub copy_cand_maxdroppen_combi_secchars (\@) {

	my ($cand_max_droppen_combi_secchars) = @_;

	my ($secchar_indices, $secchars_for_rnode, $droppen_for_rnode, $cp_max_droppen_for_combi) = @{$cand_max_droppen_combi_secchars};

	my @cp_secchar_indices = @{$secchar_indices};
	my @cp_secchars_for_rnode = ();
	foreach my $set_secchars (@{$secchars_for_rnode}) {

		# ADDED to Fix a BUG on Dec 21 (Wed), 2011 #
	    unless (defined $set_secchars) {
		push @cp_secchars_for_rnode, undef;
		next;
	    }
	    my @cp_set_secchars = @{$set_secchars};
	    push @cp_secchars_for_rnode, \@cp_set_secchars;
	}
	my @cp_droppen_for_rnode = @{$droppen_for_rnode};

	my @cp_cand = (\@cp_secchar_indices, \@cp_secchars_for_rnode, \@cp_droppen_for_rnode, $cp_max_droppen_for_combi);

	return \@cp_cand;

}	# End of the subroutine 'copy_cand_maxdroppen_combi_secchars' #


# incorporate_secchar_into_seccombi:
#	This satellite subroutine of 'enumerate_max_droppen_combi_secchars'
#	incorporates an improving secondary character into
#	a combination of improving secondary character,
#	so that the new combination will achieve as large penalty drop as possible.
#
# INPUT: (\@in_maxdroppen_combi_sc, $indx_imprv_sec, \@improving_secs, \@key_marginal_nodes_for_sec, \@diffpens_secchars_for_rem, \@minpen_prim_to_sec), where
#	@in_maxdroppen_combi_sc = (\@imprv_secchar_indices_i, \@secchars_for_rnode_i, \@droppen_for_rnode_i, $maxdroppen_for_combi_i)
#	gives the input ombination of 'improving' secondary characters on the 'remaining nodes' that potentially gives the maximum drop in the penalty (compared with the 'best' combination of primary characters), with
#	@imprv_secchar_indices_i listing indices in @improving_secs for the 'improving' secondary characters that make up the combination,
#	@{$secchars_for_rnode_i[$r]} listing indices of characters in @sec_chars that give the maximum penalty drop for the $remain_nodes->[$r] th node (= undef if no secondary characters satisfy the condition), with
#	$droppen_for_rnode_i[$r]  being the maximum penalty drop achieved for the $remain_nodes->[$r] th node by the secondary characters in the combination ( = undef if the penalty at the node does NOT improve over the best primary), and
#	$maxdroppen_for_combi_i being the maximum penalty drop achieved by the combination given;
#
#	$indx_imprv_sec is the index, in @improving_secs, of the secondary character to be incorporated here;
#	@improving_secs lists the indices of secondary characters that can improve the score over the best primary-only combination, in order of decreasing penalty drop;
#	@{$key_marginal_nodes_for_sec[$sc]} = (\@key_nodes, \@marginal)
#	for the secondary character $sec_chars[$sc], with
#	$remain_nodes->[$key_nodes[$kn]] being the ID of the $kn th remaining node
#	where the secondary character has a lower penalty than any primary characters.
#	(Here \@marginal will not be necessary);
#	$diffpens_secchars_for_rem[$sc]->[$r]is the differential penalty when the node with ID $remain_nodes->[$r] has the $sec_chars[$sc] character, subtracted by $minpen_amng_prim_for_pnodes->[$r];
#	$minpen_prim_to_sec[$sc] gives the minimum penalty for a change from a primary character (in ($pchar, @uniq_chars)) to $sec_chars[$sc];
#
# OUTPUT: \@out_maxdroppen_combi_sc, where
#	@out_maxdroppen_combi_sc = (\@imprv_secchar_indices_o, \@secchars_for_rnode_o, \@droppen_for_rnode_o, $maxdroppen_for_combi_o), with
#	the elements following the same format as @in_maxdroppen_combi_sc,
#	but with the $indx_imprv_sec th improving secondary character incorporated.
#
sub incorporate_secchar_into_seccombi (\@$\@\@\@\@) {

	my ($in_maxdroppen_combi_sc, $indx_imprv_sec, $improving_secs, $key_marginal_nodes_for_sec, $diffpens_secchars_for_rem, $minpen_prim_to_sec) = @_;

		# Initialize the output and prepare other ingredients #

	my $indx_sec = $improving_secs->[$indx_imprv_sec];

	my $cp_in_maxdroppen_combi_sc = copy_cand_maxdroppen_combi_secchars (@{$in_maxdroppen_combi_sc});
	my ($imprv_secchar_indices_o, $secchars_for_rnode_o, $droppen_for_rnode_o, $max_droppen_for_combi_o)
		= @{$cp_in_maxdroppen_combi_sc};
	push @{$imprv_secchar_indices_o}, $indx_imprv_sec;
	$max_droppen_for_combi_o += $minpen_prim_to_sec->[$indx_sec];

	my $key_nodes = $key_marginal_nodes_for_sec->[$indx_sec][0];
	my $alt_diffpens_for_rem = $diffpens_secchars_for_rem->[$indx_sec];

	foreach my $indx_rnode (@{$key_nodes}) { # Main foreach-loop #

	    my $alt_diffpen = $alt_diffpens_for_rem->[$indx_rnode];
	    my $orig_droppen = $droppen_for_rnode_o->[$indx_rnode];

	    unless (defined $orig_droppen) {
			# Create a new 'site' for replacement #
		$max_droppen_for_combi_o += $alt_diffpen;
		$droppen_for_rnode_o->[$indx_rnode] = $alt_diffpen;
		$secchars_for_rnode_o->[$indx_rnode] = [$indx_sec];
		next;
	    }

	    if ($orig_droppen < $alt_diffpen) {
		   # The original is better, so no need to change characters. #
		next;
	    } elsif ($alt_diffpen < $orig_droppen) {
		   # The alternative is better, so replace the original with the alternative. #
		$max_droppen_for_combi_o += $alt_diffpen - $orig_droppen;
		$droppen_for_rnode_o->[$indx_rnode] = $alt_diffpen;
		$secchars_for_rnode_o->[$indx_rnode] = [$indx_sec];
	    } else {
		   # Incorporate the alternative into the original replacing character set (penalty itself doesn't change). #
		push @{$secchars_for_rnode_o->[$indx_rnode]}, $indx_sec;
	    }
	}		# End of the main foreach-loop #

	my @out_maxdroppen_combi_sc = ($imprv_secchar_indices_o, $secchars_for_rnode_o, $droppen_for_rnode_o, $max_droppen_for_combi_o);

	return \@out_maxdroppen_combi_sc;

}	# End of the subroutine 'incorporate_secchar_into_seccombi' #


# search_downhill_cand_maxdroppen_sc:
#	This satellite subroutine of 'enumerate_max_droppen_combi_secchars'
#	simply goes 'downhill' the list of 'improving' secondary characters,
#	starting from a candidate secondary character combination possibly providing
#	the maximum drop in the penalty.
#
# INPUT: (\@cand0_maxdroppen_combi_sc, $curr_maxdroppen_for_combi, $init_dh_indx, \@improving_secs, \@key_marginal_nodes_for_sec, \@diffpens_secchars_for_rem, \@minpen_prim_to_sec, \@max_droppen_by_sec_over_rem, $fl_lenient1, $fl_lenient2), where
#	@cand0_maxdroppen_combi_sc = (\@imprv0_secchar_indices, \@secchars0_for_rnode, \@droppen0_for_rnode, $maxdroppen0_for_combi)
#	gives the initial candidate combination of 'improving' secondary characters on the 'remaining nodes' that potentially gives the maximum drop in the penalty (compared with the 'best' combination of primary characters), with
#	@imprv0_secchar_indices listing indices in @improving_secs for the 'improving' secondary characters that make up the combination,
#	@{$secchars0_for_rnode[$r]} listing indices of characters in @sec_chars that give the maximum penalty drop for the $remain_nodes->[$r] th node (= undef if no secondary characters satisfy the condition), with
#	$droppen0_for_rnode[$r]  being the maximum penalty drop achieved for the $remain_nodes->[$r] th node by the secondary characters in the combination ( = undef if the penalty at the node does NOT improve over the best primary), and
#	$maxdroppen0_for_combi being the maximum penalty drop achieved by the combination given;
#
#	$curr_maxdroppen_for_combi is the current recorded maximum drop in the penalty;
#	$init_dh_indx is an index in @improving_secs specifying the initial point of the 'downhill' search;
#	@improving_secs lists the indices of secondary characters that can improve the score over the best primary-only combination, in order of decreasing penalty drop;
#	@{$key_marginal_nodes_for_sec[$sc]} = (\@key_nodes, \@marginal)
#	for the secondary character $sec_chars[$sc], with
#	$remain_nodes->[$key_nodes[$kn]] being the ID of the $kn th remaining node
#	where the secondary character has a lower penalty than any primary characters.
#	(Here \@marginal will not be necessary);
#	$diffpens_secchars_for_rem[$sc]->[$r]is the differential penalty when the node with ID $remain_nodes->[$r] has the $sec_chars[$sc] character, subtracted by $minpen_amng_prim_for_pnodes->[$r];
#	$minpen_prim_to_sec[$sc] gives the minimum penalty for a change from a primary character (in ($pchar, @uniq_chars)) to $sec_chars[$sc];
#	$max_droppen_by_sec_over_rem[$sc] gives the maximum possible drop in the penalty when the character $sec_chars[$sc] is used for some of the 'remaining nodes';
#	$fl_lenient1 = 1 if the potentially score-tying extensions are also tried,
#		    = 0 if only potentially score-improving extensions are pursued;
#	$fl_lenient2 = 1 if the score-tying combinations are also incorporated into the history;
#		    = 0 if only score-improving combinations are incorporated into the history.
#
# OUTPUT: \@hist_dh_cand_maxdroppen_combi_sc, where
#	@hist_dh_cand_maxdroppen_combi_sc = (\@cand0 = \@cand0_maxdroppen_combi_sc, \@cand1, \@cand2, ...) records the history of growing the candidates like a snowball, starting with the initial candidate down to the end of @improving_secs, with
#	@cand1/2/... following the same format as @cand0;
#
sub search_downhill_cand_maxdroppen_sc (\@$$\@\@\@\@\@$$) {

	my ($cand0_maxdroppen_combi_sc,$curr_maxdroppen_for_combi, $init_dh_indx, $improving_secs, $key_marginal_nodes_for_sec, $diffpens_secchars_for_rem, $minpen_prim_to_sec, $max_droppen_by_sec_over_rem, $fl_lenient1, $fl_lenient2)  = @_;

	my ($imprv0_secchar_indices, $secchars0_for_rnode, $droppen0_for_rnode, $maxdroppen0_for_combi) = @{$cand0_maxdroppen_combi_sc};

	my $ct_imprv_secs = @{$improving_secs};  # Upper-bound of the index for downhill.
	my $last_indx0 = $imprv0_secchar_indices->[$#{$imprv0_secchar_indices}];

		## A BUG Fixed on Dec 28 (Wed), 2011 ##
		## (Relaxed a condition on the input data, which had been too harsh.) ##
	unless ($init_dh_indx > $last_indx0) {
	   print STDOUT "[ERROR 'search_downhill_cand_maxdroppen_sc'] ILLEGAL INPUT: \$init_dh_indx = $init_dh_indx is NOT in the expected range, which is $last_indx0 < \$init_dh_indx : $! !\n";
	##unless(($init_dh_indx > $last_indx0) and ($init_dh_indx<$ct_imprv_secs)) {
	##   print STDOUT "[ERROR 'search_downhill_cand_maxdroppen_sc'] ILLEGAL INPUT: \$init_dh_indx = $init_dh_indx is NOT in the expected range, which is $last_indx0 < \$init_dh_indx < $ct_imprv_secs: $! !\n";
	    return undef;
	}

	my $pot_maxadd_maxdroppen_for_combi = 0; # Potential maximum addition to the maximum drop in the penalty.
	for (my $i=$init_dh_indx; $i<$ct_imprv_secs; $i++) { # Preliminary for-loop #
	    my $indx_sec = $improving_secs->[$i];
	    my $maxdroppen_for_sec = $max_droppen_by_sec_over_rem->[$indx_sec];
	    $pot_maxadd_maxdroppen_for_combi += $maxdroppen_for_sec;
	}			# End of the preliminary for-loop #

		# Initialize the history #

	my $cp_cand0_maxdroppen_combi_sc = copy_cand_maxdroppen_combi_secchars (@{$cand0_maxdroppen_combi_sc});
	my @hist_dh_cand_maxdroppen_combi_sc = ($cp_cand0_maxdroppen_combi_sc);
	my $curr_maxdroppen_dh = $maxdroppen0_for_combi;

	for (my $i=$init_dh_indx; $i<$ct_imprv_secs; $i++) { # Main outer for-loop (for going 'downhill') #

	    my $ub_tot_maxdroppen = $curr_maxdroppen_dh + $pot_maxadd_maxdroppen_for_combi; 
	    (($curr_maxdroppen_for_combi < $ub_tot_maxdroppen) or
	    (!$fl_lenient1 and ($curr_maxdroppen_for_combi ==$ub_tot_maxdroppen)))and last; # We cannot improve the current maximum drop in penalty any longer #

	    my $indx_sec = $improving_secs->[$i];
	    my $maxdroppen_for_sec = $max_droppen_by_sec_over_rem->[$indx_sec]; # Maximum drop in penalty by employing the subject secondary character.
	    $pot_maxadd_maxdroppen_for_combi -= $maxdroppen_for_sec; # reduce the potentially maximum addition by the maximum drop for the current 'improving' secondary character.

		# Try to augment the current 'best'
		# with the secondary character at the current point in the downhill process. #

	    my $curr_maxdroppen_combi_sc = $hist_dh_cand_maxdroppen_combi_sc[$#hist_dh_cand_maxdroppen_combi_sc];
	    my $curr_imprv_sc_indices = $curr_maxdroppen_combi_sc->[0];
	    # @{$cand_new_maxdroppen_combi_sc} = (\@imprv_secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $maxdroppen_for_combi),
	    #
	    my $cand_new_maxdroppen_combi_sc = incorporate_secchar_into_seccombi (@{$curr_maxdroppen_combi_sc}, $i, @{$improving_secs}, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem}, @{$minpen_prim_to_sec});
	    unless (defined $cand_new_maxdroppen_combi_sc) {
		print STDOUT "[ERROR 'search_downhill_cand_maxdroppen_sc'] 'incorporate_secchar_into_seccombi' FAILED to incorporate the $i th improving secondary character (or the $indx_sec th secondary character) into the set (", join (', ', @{$curr_imprv_sc_indices}), "): $! !\n";
		return undef;
	    }
	    my $maxdroppen_for_new_combi = $cand_new_maxdroppen_combi_sc->[3];

		# Update the stack of candidate combinations if the new combination is good enough. #

	    if ($maxdroppen_for_new_combi < $curr_maxdroppen_dh) {
			# The new combination improves the penalty drop. #
		push @hist_dh_cand_maxdroppen_combi_sc, $cand_new_maxdroppen_combi_sc;
		$curr_maxdroppen_dh = $maxdroppen_for_new_combi;

	    } elsif ($fl_lenient2 and ($maxdroppen_for_new_combi == $curr_maxdroppen_dh)) {
			# The new combination ties the old one in terms of penalty drop. #
		push @hist_dh_cand_maxdroppen_combi_sc, $cand_new_maxdroppen_combi_sc;
	    }
		# If the new candidate is not good enough, keep the current best candidate and go to the next round. #

	}			# End of the Main outer for-loop (for going 'downhill') #

	return \@hist_dh_cand_maxdroppen_combi_sc;

}	# End of the subroutine 'search_downhill_cand_maxdroppen_sc' #


# search_back_forth_lin_maxdroppen_sc:
#	This satellite subroutine of the subroutine 'enumerate_max_droppen_combi_secchars'
#	searches for the combinations of improving secondary characters
#	that provide the maximum drop in penalty among all possible combinations in a 'lineage' of improving secondary characters,
#	by traveling back and forth along the 'lineage'.
#	Briefly, the improvement-loop algorithm is as follows:
#	(1) First check if the history contains more than one combination.
#	(2) If this is the case, search for better combinations via ''
#	using the 'next best' combination (namely the second last one) as a base
#	and starting from the next improving character of the last character in the current best combination.
#	(3) Graft the resulting new history onto the remainder of the current history.
#	(4) The loop (1)~(3) continues until the size of history becomes 1 or less.
#	(1.5) Between (1) and (2), always check if the 'current best' (i.e. the last) combination
#	gives the best penalty drop or not.
#
#	NOTE that this subroutine eats up the input history, leaving only the initial 'combination'.
#
# INPUT: (\@hist_dh_cand_maxdroppen_combi_sc, $curr_maxdroppen_for_combi, \@improving_secs, \@key_marginal_nodes_for_sec, \@diffpens_secchars_for_rem, \@minpen_prim_to_sec, \@max_droppen_by_sec_over_rem), where
#	@hist_dh_cand_maxdroppen_combi_sc = (\@cand0 = \@cand0_maxdroppen_combi_sc, \@cand1, \@cand2, ...) records the history of growing the candidates like a snowball, starting with the initial candidate down to the end of @improving_secs, with
#	@cand1/2/... following the same format as @cand0, where
#	@cand0_maxdroppen_combi_sc = (\@imprv0_secchar_indices, \@secchars0_for_rnode, \@droppen0_for_rnode, $maxdroppen0_for_combi)
#	gives the initial candidate combination of 'improving' secondary characters on the 'remaining nodes' that potentially gives the maximum drop in the penalty (compared with the 'best' combination of primary characters), with
#	@imprv0_secchar_indices listing indices in @improving_secs for the 'improving' secondary characters that make up the combination,
#	@{$secchars0_for_rnode[$r]} listing indices of characters in @sec_chars that give the maximum penalty drop for the $remain_nodes->[$r] th node (= undef if no secondary characters satisfy the condition), with
#	$droppen0_for_rnode[$r]  being the maximum penalty drop achieved for the $remain_nodes->[$r] th node by the secondary characters in the combination ( = undef if the penalty at the node does NOT improve over the best primary), and
#	$maxdroppen0_for_combi being the maximum penalty drop achieved by the combination given;
#
#	$curr_maxdroppen_for_combi is the current recorded maximum drop in the penalty;
#	$init_dh_indx is an index in @improving_secs specifying the initial point of the 'downhill' search;
#	@improving_secs lists the indices of secondary characters that can improve the score over the best primary-only combination, in order of decreasing penalty drop;
#	@{$key_marginal_nodes_for_sec[$sc]} = (\@key_nodes, \@marginal)
#	for the secondary character $sec_chars[$sc], with
#	$remain_nodes->[$key_nodes[$kn]] being the ID of the $kn th remaining node
#	where the secondary character has a lower penalty than any primary characters.
#	(Here \@marginal will not be necessary);
#	$diffpens_secchars_for_rem[$sc]->[$r]is the differential penalty when the node with ID $remain_nodes->[$r] has the $sec_chars[$sc] character, subtracted by $minpen_amng_prim_for_pnodes->[$r];
#	$minpen_prim_to_sec[$sc] gives the minimum penalty for a change from a primary character (in ($pchar, @uniq_chars)) to $sec_chars[$sc];
#	$max_droppen_by_sec_over_rem[$sc] gives the maximum possible drop in the penalty when the character $sec_chars[$sc] is used for some of the 'remaining nodes';
#
# OUTPUT: \@set_cand_maxdroppen_combi_sc, where
#	@{$set_cand_maxdroppen_combi_sc[$cd]} = (\@secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $max_droppen_for_combi)
#	is the $cd th candidate combination of secondary characters,
#	which has the best penalty drop among the combinations starting with the $i th 'improving' secondary character.
#	The output could be empty if the best combination from this lineage does NOT improve or tie the input drop in penalty, $curr_maxdroppen_for_combi.
#
sub search_back_forth_lin_maxdroppen_sc (\@$\@\@\@\@\@) {

	my ($hist_dh_cand_maxdroppen_combi_sc, $curr_maxdroppen_for_combi, $improving_secs, $key_marginal_nodes_for_sec, $diffpens_secchars_for_rem, $minpen_prim_to_sec, $max_droppen_by_sec_over_rem) = @_;

    (print STDOUT "< Subroutine 'search_back_forth_lin_maxdroppen_sc' >\n\n") if ($TEST_LEVEL2>3);

	my @history = @{$hist_dh_cand_maxdroppen_combi_sc};
	if (@history<2) {
	    # There should be no chance of further improvement. #
	    # (This should include the cases with #{'improving' secondary characters} = 0,
	    # and the cases with #{'remaining' nodes} = 0.)

	    my $hist_size = @history;
   (print STDOUT "\n With the history size of $hist_size, NO FURTHER IMPROVEMENT will be EXPECTED. So I will leave the subroutine here ...\n\n") if ($TEST_LEVEL2>3);

	    return \@history;
	}
	my $last_indx_imprv_sec = $#{$improving_secs};

	my @set_cand_maxdroppen_combi_sc = ();
	my $curr_maxdroppen = $curr_maxdroppen_for_combi;

	my $ct_remnodes = scalar (@{$diffpens_secchars_for_rem->[0]});

    if ($TEST_LEVEL2>3) {
	print STDOUT "{ Input }\n\n";
	print STDOUT "#{'Remaining' nodes} = $ct_remnodes ,\n\n";
	print STDOUT "Current maximum penalty-drop:\n";
	print STDOUT '  $curr_maxdroppen = ', $curr_maxdroppen, ';', "\n";
	print STDOUT "IDs of 'improving' secondary characters:\n";
	print STDOUT '  @{$improving_secs} = (', join (', ', @{$improving_secs}), ');', "\n";
	print STDOUT "The index of the last 'improving' secondary character:\n";
	print STDOUT '  $last_indx_imprv_sec = ', $last_indx_imprv_sec, ' .', "\n\n";
    }

	while (@history>1) { # Main outer while-loop #

	    my $curr_best_cand = pop @history;
	    ##my $curr_best_droppen = $curr_best_cand->[3];
	    ##my $best_imprv_sc_indices = $curr_best_cand->[0];
	    my ($best_imprv_sc_indices, $scs_for_rnode, $droppen_for_rnode, $curr_best_droppen) = @{$curr_best_cand};

    if ($TEST_LEVEL2>3) {
	print STDOUT "{ The 'best' candidate in the current history }\n\n";
	print STDOUT 'Penalty-drop: $curr_best_droppen = ', $curr_best_droppen, " ,\n";
	print STDOUT "Indices of 'improving' secondary characters:\n";
	print STDOUT '@{$best_imprv_sc_indices} = (', join (', ',  @{$best_imprv_sc_indices}), ');', "\n\n";
	print STDOUT "Penalty drop and Secondary characters for each 'remaining' node:\n\n";
	print STDOUT join ("\t", 'Indx_rem_node', 'penalty-drop', 'IDs,best,sec,chars'), "\n\n";
	for (my $r=0; $r<$ct_remnodes; $r++) {
	    my $droppen = $droppen_for_rnode->[$r];
	    (defined $droppen) or ($droppen = 'undef');
	    my $best_sec_chars = $scs_for_rnode->[$r];
	    my $cnct_best_scs = (defined $best_sec_chars) ? join (',', @{$best_sec_chars}) : 'undef';
	    print STDOUT join ("\t", $r, $droppen, $cnct_best_scs), "\n";
	}
	print STDOUT "\n\n";
    }

	    if ($curr_best_droppen < $curr_maxdroppen) {
		# The penalty drop improved. -> Update the set of current maximum combinations. #
    (print STDOUT "\n ... The current maximum penalty-drop (= $curr_maxdroppen) is going to be UPDATED to \$curr_best_droppen = $curr_best_droppen, and accordingly the set of candidate 'best' combinations will also be ...\n\n") if ($TEST_LEVEL2>3);
		$curr_maxdroppen = $curr_best_droppen;
		@set_cand_maxdroppen_combi_sc = ($curr_best_cand);


	    } elsif ($curr_best_droppen == $curr_maxdroppen) {
		# The penalty drop tied the current maximum. -> Append the combination to the set of current maximum combinations. #
    (print STDOUT "\n ... The current best candidate TIES the current maximum penalty-drop (= $curr_maxdroppen), so it will be APPENDED to the set of candidate 'best' combinations ...\n\n") if ($TEST_LEVEL2>3);
		push @set_cand_maxdroppen_combi_sc, $curr_best_cand;
	    }
	    my $curr_last_indx = $best_imprv_sc_indices->[$#{$best_imprv_sc_indices}];

	    unless ($curr_last_indx < $last_indx_imprv_sec) {
		# We cannot try to improve the current next best any longer!! #
    (print STDOUT "\n ... The last indx of the current combination (=$curr_last_indx) is NOT SMALLER THAN the last index among ALL 'improving' secondary characters. So, This series canNOT IMPROVE any further ...\n\n") if ($TEST_LEVEL2>3);

		next;
	    }

	    my $curr_next_best_cand = pop @history;
	    my $next_best_indices = $curr_next_best_cand->[0];
   (print STDOUT "\n ... Now we try to improve the combination, (", join (', ', @{$next_best_indices}), "), which is the '1-step behind' the current 'best' combination ...\n\n") if ($TEST_LEVEL2>3);

	    my $new_dh_history = search_downhill_cand_maxdroppen_sc (@{$curr_next_best_cand}, $curr_maxdroppen, $curr_last_indx +1, @{$improving_secs}, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem}, @{$minpen_prim_to_sec}, @{$max_droppen_by_sec_over_rem}, 1, 0);
	    unless (defined $new_dh_history) {
		print STDOUT "[ERROR 'enumerate_max_droppen_combi_secchars'] Subroutine 'search_downhill_cand_maxdroppen_combi_secchars' FAILED starting with the 'initial combination' = (", join (', ', @{$next_best_indices}), ") and starting from ", $curr_last_indx + 1, " th improving secondary character: $! !\n";
		return undef;
	    }
	    my $size_new_history = @{$new_dh_history};

		# Modified to avoid confusion (on December 23 (Fri), 2011) #
    (print STDOUT "\n --- The subroutine 'search_downhill_cand_maxdroppen_sc' ENDED, producing ", $size_new_history -1, " new 'better' candidates than the '1-step behind' the 'current best' ...\n\n") if ($TEST_LEVEL2>3);
    ##(print STDOUT "\n --- The subroutine 'search_downhill_cand_maxdroppen_sc' ENDED, producing $size_new_history new 'better' candidates than the '1-step behind' the 'current best' ...\n\n") if ($TEST_LEVEL2>3);

	    push @history, @{$new_dh_history};

	}		# End of the main outer while-loop #

    (print STDOUT "--- END of 'search_back_forth_lin_maxdroppen_sc' ---\n\n") if ($TEST_LEVEL2>3);

	return \@set_cand_maxdroppen_combi_sc;

}	# End of the subroutine 'search_back_forth_lin_maxdroppen_sc' #


# enumerate_core_maxdroppen_comi_secchars:
#	This final main satellite subroutine of the subroutine 'part_resolve_mlt_node_bu' enumerates
#	the 'core' combinations of secondary characters that provide the maximum drop in the penalty (compared to the best combinarion(s) of primary characters).
#	The 'core' combination means that any of its constituent secondary characters
#	contribute a positive penalty drop.
#
# INPUT: (\@improving_secs, \@key_marginal_nodes_for_sec, \@diffpens_secchars_for_rem, \@minpen_prim_to_sec, \@max_droppen_by_sec_over_rem), where
#	@improving_secs lists the indices of secondary characters that can improve the score over the best primary-only combination, in order of decreasing penalty drop;
#	@{$key_marginal_nodes_for_sec[$sc]} = (\@key_nodes, \@marginal)
#	for the secondary character $sec_chars[$sc], with
#	$remain_nodes->[$key_nodes[$kn]] being the ID of the $kn th remaining node
#	where the secondary character has a lower penalty than any primary characters.
#	(Here \@marginal will not be necessary);
#	$diffpens_secchars_for_rem[$sc]->[$r]is the differential penalty when the node with ID $remain_nodes->[$r] has the $sec_chars[$sc] character, subtracted by $minpen_amng_prim_for_pnodes->[$r];
#	$minpen_prim_to_sec[$sc] gives the minimum penalty for a change from a primary character (in ($pchar, @uniq_chars)) to $sec_chars[$sc];
#	$max_droppen_by_sec_over_rem[$sc] gives the maximum possible drop in the penalty when the character $sec_chars[$sc] is used for some of the 'remaining nodes'.
#
# OUTPUT: \@core_maxdroppen_combi_secchars, where
#	@{$core_maxdroppen_combi_secchars[$cm]} = (\@imprv_secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $max_droppen_for_combi)
#	gives the $cm th combination of 'improving' secondary characters on the 'remaining nodes' that gives the maximum drop in the penalty (compared with the 'best' combination of primary characters), with
#	@imprv_secchar_indices listing indices in @improving_secs for the 'improving' secondary characters that make up the combination,
#	@{$secchars_for_rnode[$r]} listing indices of characters in @sec_chars that give the maximum penalty drop for the $remain_nodes->[$r] th node (= undef if no secondary characters satisfy the condition), with
#	$droppen_for_rnode[$r]  being the maximum penalty drop achieved for the $remain_nodes->[$r] th node by the secondary characters in the combination ( = undef if the penalty at the node does NOT improve over the best primary), and
#	$max_droppen_for_combi being the maximum penalty drop achieved by the combination given.
#
sub enumerate_core_maxdroppen_combi_secchars (\@\@\@\@\@) {

	my ($improving_secs, $key_marginal_nodes_for_sec, $diffpens_secchars_for_rem, $minpen_prim_to_sec, $max_droppen_by_sec_over_rem) = @_;

	my $ct_imprv_secs = @{$improving_secs};
	unless ($ct_imprv_secs>0) {
	    return []; # NO combination of secondary characters can improve the penalty over the primary characters. #
	}
    (print STDOUT "< Subroutine 'enumerate_core_maxdroppen_combi_secchars' >\n\n") if ($TEST_LEVEL2>0);

	my $ct_remnodes = @{$diffpens_secchars_for_rem->[0]};

    if ($TEST_LEVEL2>1) {
	print STDOUT "Indices of 'improving' secondary characters (# = $ct_imprv_secs):\n";
	print STDOUT '@{$improving_secs} = (', join (', ', @{$improving_secs}), ") .\n\n";
    } 
    (print STDOUT "[ Beginning of the Main for-loop (over the 'improving' secondary characters) ]\n\n") if ($TEST_LEVEL2>1);

	my @core_maxdroppen_combi_secchars = ();
	my $max_droppen_for_combi;

	for (my $i=0; $i<$ct_imprv_secs; $i++) { # Main outer for-loop #
	    my $indx_sec0 = $improving_secs->[$i];

    (print STDOUT "{ The $i th 'improving' secondary character (i.e. the $indx_sec0 th secondary character }\n\n") if ($TEST_LEVEL2>0);

		# Briefly examine if there is a chance that
		# the current session (starting with $i) can improve the current maximum drop in penalty. #

	    my $pot_max_droppen_line = 0; # Potentially maximum drop in penalty in the current 'line' (from $i through $ct_imprv_secs-1) #
	    for (my $j=$i; $j<$ct_imprv_secs; $j++) { $pot_max_droppen_line += $max_droppen_by_sec_over_rem->[$improving_secs->[$j]]; }

    if ($TEST_LEVEL2>2) {
	print STDOUT '$max_droppen_for_combi = ', ((defined $max_droppen_for_combi) ? $max_droppen_for_combi : 'undef'), ' ,', "\n";
	print STDOUT '$pot_max_droppen_line = ', $pot_max_droppen_line, " .\n\n";
    }

	    ((defined $max_droppen_for_combi) and ($max_droppen_for_combi < $pot_max_droppen_line)) and last; # We will never improve (or tie) the drop in penalty any further. #

		# Initialize the combination to maximize the drop in penalty in the current 'line' #
	
	    # @{$init_max_droppen_combi_secchars} = (\@imprv_secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $max_droppen_for_combi).
	    #
	    my ($init_secchars_for_rnode, $init_droppen_for_rnode) = make_initcand_max_droppen_combi_sc ($indx_sec0, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem});
	    unless (defined $init_secchars_for_rnode) {
		print STDOUT "[ERROR 'enumerate_core_maxdroppen_combi_secchars'] Subroutine 'make_initcand_max_droppen_combi_sc' FAILED on the $i th improving secondary character, which is the $indx_sec0 th secondary character: $! !\n";
		return undef;
	    }
    (print STDOUT "\n --- The subroutine 'make_initcand_max_droppen_combi_sc' ENDED ...\n\n") if ($TEST_LEVEL2>1);

	    my $init_max_droppen = $max_droppen_by_sec_over_rem->[$indx_sec0];
	    my $pot_maxadd_droppen = $pot_max_droppen_line - $init_max_droppen;

	    my @cand0_maxdroppen_combi_secchars = ([$i], $init_secchars_for_rnode, $init_droppen_for_rnode, $init_max_droppen);

    if ($TEST_LEVEL2>2) {
	print STDOUT "( Initial candidate maximum penalty-dropping combination starting with the $i th 'improving' secondary character ):\n\n";
	print STDOUT '@{$imprv_secchar_indices} = (', $i, ')', ' ;', "\n";
	print STDOUT 'Maximum penalty-drop = ', $init_max_droppen, ' ;', "\n";
	print STDOUT 'Content of @{$secchars_for_rnode} and @{$droppen_for_rnode}:', "\n\n";
	print STDOUT join ("\t", 'Indx_RemNode', 'droppen_by_sec', 'indices,best,sec,chars'), "\n\n";
	for (my $r=0; $r<$ct_remnodes; $r++) {
	    my $secchars = (defined $init_secchars_for_rnode->[$r]) ? join (',', @{$init_secchars_for_rnode->[$r]}) : 'undef';
	    my $droppen = $init_droppen_for_rnode->[$r];
	    (defined $droppen) or ($droppen = 'undef');
	    print STDOUT join ("\t", $r, $droppen, $secchars), "\n";
	}
	print STDOUT "\n\n";
    }

	    unless (defined $max_droppen_for_combi) { # Initialize the maximum penalty drop #
		$max_droppen_for_combi = $init_max_droppen;

     (print STDOUT 'Now, $max_droppen_for_combi has been initialized as ', $init_max_droppen, "...\n\n") if ($TEST_LEVEL2>1);

			# Disabled to incorporate the current initial combination,
			# because it will be recovered later anyway (if it remains the maximum penalty-dropping). #
		##my $cp_cand0 = copy_cand_maxdroppen_combi_secchars (@cand0_maxdroppen_combi_secchars);
		##@core_maxdroppen_combi_secchars = ($cp_cand0);
	    }

	    	# Make a series of candidate combinations
		# just by going 'downhill' the @{$improving_secs} from $i+1,
		# starting with @cand0_max_droppen_combi_secchars. # 

	    # @{$hist_dh_cand_max_droppen_combi_secchars} = (\@cand0_max_droppen_combi_secchars, \@cand1, \@cand2, ...), which records the history of growing the candidate in a simply downhill manner, starting with @cand0_max_droppen_combi_secchars.
	    #
	    my $hist_dh_cand_maxdroppen_combi_secchars = search_downhill_cand_maxdroppen_sc (@cand0_maxdroppen_combi_secchars, $max_droppen_for_combi, $i+1, @{$improving_secs}, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem}, @{$minpen_prim_to_sec}, @{$max_droppen_by_sec_over_rem}, 1, 0);
	    unless (defined $hist_dh_cand_maxdroppen_combi_secchars) {
		print STDOUT "[ERROR 'enumerate_core_maxdroppen_combi_secchars'] Subroutine 'search_downhill_cand_maxdroppen_combi_secchars' FAILED starting with the 'initial combination' = ($i), i.e. the $indx_sec0 th secondary character: $! !\n";
		return undef;
	    }
    (print STDOUT "\n --- The subroutine 'search_downhill_cand_maxdroppen_sc' ENDED ...\n\n") if ($TEST_LEVEL2>1);

	    my $size_history = @{$hist_dh_cand_maxdroppen_combi_secchars};
	    my $interim_cand_maxdroppen_combi_secchars = $hist_dh_cand_maxdroppen_combi_secchars->[$size_history-1];
	    my $interim_maxdroppen = $interim_cand_maxdroppen_combi_secchars->[3];

    if ($TEST_LEVEL2>2) {
	print STDOUT "[ Output of 'search_downhill_cand_maxdroppen_sc' ]\n\n";
	print STDOUT "#{combinations made during the 'downhill' process} = $size_history ,\n";
	print STDOUT "The 'best' combination achieved by the 'downhill':\n\n";
	my ($imprv_secchar_indices_dh, $secchars_for_rnode_dh, $droppen_for_rnode_dh, $max_droppen_dh) = @{$interim_cand_maxdroppen_combi_secchars};
	print STDOUT '@{$imprv_secchar_indices} = (', join (', ', @{$imprv_secchar_indices_dh}) , ')', ' ;', "\n";
	print STDOUT 'Maximum penalty-drop = ', $max_droppen_dh, ' ;', "\n";
	print STDOUT 'Content of @{$secchars_for_rnode} and @{$droppen_for_rnode}:', "\n\n";
	print STDOUT join ("\t", 'Indx_RemNode', 'droppen_by_sec', 'indices,best,sec,chars'), "\n\n";
	for (my $r=0; $r<$ct_remnodes; $r++) {
	    my $secchars = (defined $secchars_for_rnode_dh->[$r]) ? join (',', @{$secchars_for_rnode_dh->[$r]}) : 'undef';
	    my $droppen = $droppen_for_rnode_dh->[$r];
	    (defined $droppen) or ($droppen = 'undef');
	    print STDOUT join ("\t", $r, $droppen, $secchars), "\n";
	}
	print STDOUT "\n\n";
    }

	    unless ($size_history>1) {
		# There is no new candidate combination
		# that is better than the initial one (starting with $i),
		# which means there is no hope to improve in this 'lineage'. #

     (print STDOUT "\n--- The 'downhill' process could NOT produce ANY combination BETTER THAN the initial one (i.e. \($i\)).\n\n") if ($TEST_LEVEL2>1);
	
		if ($interim_maxdroppen < $max_droppen_for_combi) {
		    print STDOUT "[ERROR 'enumerate_core_maxdroppen_combi_secchars'] This is IMPOSSIBLE! The history starting from $i, representing the $indx_sec0 th secondary character, consists only of a single combination, yet its maximum drop penalty (=$interim_maxdroppen) is BETTER THAN the maximum drop penalty ever (=$max_droppen_for_combi): $! !\n";
		    return undef;
		} elsif ($interim_maxdroppen == $max_droppen_for_combi) {
		    my $cp_interim_cand = copy_cand_maxdroppen_combi_secchars (@{$interim_cand_maxdroppen_combi_secchars});

		    push @core_maxdroppen_combi_secchars, $cp_interim_cand;
     (print STDOUT " ... The 'best-downhill' combination ties the largest penalty-drop so far. So, it has been added to the set of 'core' maximum penalty-dropping combinations !...\n\n") if ($TEST_LEVEL2>1);
		}
     (print STDOUT " ... Now, Skip to the next initial character ...\n\n") if ($TEST_LEVEL2>1);
		next;
	    }

     (print STDOUT "\n--- The 'downhill' process produced a combination ", $size_history -1, " step better than the initial one (i.e. ($i)).\n\n") if ($TEST_LEVEL2>1);

		# There is still a hope to improve the penalty drop. #

	    if ($interim_maxdroppen < $max_droppen_for_combi) {
		    # Reset the set of 'core' combinations of secondary characters with maximum drop in penalty. #
		$max_droppen_for_combi = $interim_maxdroppen;
		## my $cp_interim_cand = copy_cand_maxdroppen_combi_secchars (@{$interim_cand_maxdroppen_combi_secchars});
		## @core_maxdroppen_combi_secchars = ($cp_interim_cand);
		@core_maxdroppen_combi_secchars = (); # Reset it to empty, because $cp_interim_cand will be recovered later if it remains maximum penalty-dropping. #

    (print STDOUT "\n", ' ... $max_droppen_for_combi has been UPDATED to ', $max_droppen_for_combi, " !\n\n") if ($TEST_LEVEL2>1);
	    }


		## Try to improve the combinations by going up step-by-step,
		## then try to further improve the penalty drop by going 'downhill' again.
		## (The process repeats untill no further improvement is expected.)

	    # @{$set_cand_maxdroppen_combi_sc->[$cd]} = (\@imprv_secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $max_droppen_for_combi)
	    #is the $cd th candidate combination of secondary characters, which has the best penalty drop among the combinations starting with the $i th 'improving' secondary character.
	    # [ This set could be empty if none of the combination either improve or tie the penalty drop (compared to $max_droppen_for_combi). ]
	    #
	    my $set_cand_maxdroppen_combi_sc = search_back_forth_lin_maxdroppen_sc (@{$hist_dh_cand_maxdroppen_combi_secchars}, $max_droppen_for_combi, @{$improving_secs}, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem}, @{$minpen_prim_to_sec}, @{$max_droppen_by_sec_over_rem});
	    unless (defined $set_cand_maxdroppen_combi_sc) {
		print STDOUT "[ERROR 'enumerate_core_maxdroppen_combi_secchars'] Subroutine 'search_back_forth_lin_maxdroppen_sc' FAILED on the $i th 'lineage' of improving secondary characters, which started with the $indx_sec0 th secondary character: $! !\n";
		return undef;
	    }
    (print STDOUT "\n --- The subroutine 'search_back_forth_lin_maxdroppen' ENDED ...\n\n") if ($TEST_LEVEL2>1);

	    unless (@{$set_cand_maxdroppen_combi_sc}>0) {
		# No combination in this 'lineage' improved
		# or tied the current maximum penalty drop.
    (print STDOUT "... NO combination improved or tied the current maximum penalty drop , when starting with the $i th 'improving' secondary character. Skip to the next 'improving' character  ...\n\n") if ($TEST_LEVEL2>1);
		next;
	    }
    if ($TEST_LEVEL2>2) {
	print STDOUT "[ Output of 'search_back_forth_lin_maxdroppen_sc' ]\n\n";
	my $ct_cand_mxdpen_cmbs = @{$set_cand_maxdroppen_combi_sc};
	print STDOUT 'Content of @{$set_cand_maxdroppen_combi_sc} (with ', $ct_cand_mxdpen_cmbs, ' elements):', "\n\n";
	for (my $cd=0; $cd<$ct_cand_mxdpen_cmbs; $cd++) {
	    print STDOUT "( The $cd th candidate combination )\n\n";
	    my ($imprv_secchar_indices_cd, $secchars_for_rnode_cd, $droppen_for_rnode_cd, $max_droppen_cd) = @{$set_cand_maxdroppen_combi_sc->[$cd]};
	    print STDOUT '@{$imprv_secchar_indices} = (', join (', ', @{$imprv_secchar_indices_cd}) , ')', ' ;', "\n";
	    print STDOUT 'Maximum penalty-drop = ', $max_droppen_cd, ' ;', "\n";
	    print STDOUT 'Content of @{$secchars_for_rnode} and @{$droppen_for_rnode}:', "\n\n";
	    print STDOUT join ("\t", 'Indx_RemNode', 'droppen_by_sec', 'indices,best,sec,chars'), "\n\n";
	    for (my $r=0; $r<$ct_remnodes; $r++) {
	    	my $secchars = (defined $secchars_for_rnode_cd->[$r]) ? join (',', @{$secchars_for_rnode_cd->[$r]}) : 'undef';
	    	my $droppen = $droppen_for_rnode_cd->[$r];
	    	(defined $droppen) or ($droppen = 'undef');
	    	print STDOUT join ("\t", $r, $droppen, $secchars), "\n";
	    }
	    print STDOUT "\n\n";
	}
    }

		# Compare the maximum penalty drop achieved in this lineage
		# with the maximum penalty drop ever. #
	    my $lin_maxdroppen = $set_cand_maxdroppen_combi_sc->[0][3];

	    if ($lin_maxdroppen > $max_droppen_for_combi) {
		# The lineage could NOT improve or tie the maximum penalty drop. #
     (print STDOUT "\n ... The lineage starting with the $i th 'improving' secondary character could NOT improve or tie the maximum penalty-drop. So NO CANDIDATES were ADDED  ...\n\n") if ($TEST_LEVEL2>1);
		next;
	    }

	    if ($lin_maxdroppen < $max_droppen_for_combi) {
		# The lineage maximum improved the maximum penalty drop. #
		$max_droppen_for_combi = $lin_maxdroppen;
		@core_maxdroppen_combi_secchars = @{$set_cand_maxdroppen_combi_sc};
    (print STDOUT "\n", ' ... $max_droppen_for_combi was UPDATED to the best achievable value in this lineage (=', $max_droppen_for_combi, '), and @core_maxdroppen_combi_secchars was also UPDATED accordingly ...', "\n\n") if ($TEST_LEVEL2>1);
	    } else {
		# The lineage maximum tied the maximum penalty drop. #
		push @core_maxdroppen_combi_secchars, @{$set_cand_maxdroppen_combi_sc};

    (print STDOUT "\n", ' ... THe BEST achievable combinations in this lineage TIED with the current maxium penalty-drop ! So they were APPENDED to @core_maxdroppen_combi_secchars ...', "\n\n") if ($TEST_LEVEL2>1);
	    }
    (print STDOUT "\n --- The processes starting with the $i th 'improving' secondary candidates ENDED ---\n\n") if ($TEST_LEVEL2>0);

	}		# End of the main outer for-loop #

    (print STDOUT " === END of the subroutine 'enumerate_core_maxdroppen_combi_secchars' ===\n\n") if ($TEST_LEVEL2>0);

	return \@core_maxdroppen_combi_secchars;

}		# End of the subroutine 'enumerate_core_maxdroppen_combi_secchars' #


# dress_core_maxdroppen_combi_secchars:
#	This satellite subroutine of the subroutine 'part_resolve_mlt_node_bu'
#	'dress'es each 'core' secondary-character combination with the improving secondary characters
#	NOT included in the core and yet NOT reducing the penalty drop.
#
#	NOTE that this subroutine eats up the input @core_maxdroppen_combi_secchars.
#	
#
# INPUT: (\@core_maxdroppen_combi_secchars, \@improving_secs, \@key_marginal_nodes_for_sec, \@diffpens_secchars_for_rem, \@minpen_prim_to_sec), where
#	@{$core_maxdroppen_combi_secchars[$cm]} = (\@imprv_secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $max_droppen_for_combi)
#	gives the $cm th combination of 'improving' secondary characters on the 'remaining nodes' that gives the maximum drop in the penalty (compared with the 'best' combination of primary characters), with
#	@imprv_secchar_indices listing indices in @improving_secs for the 'improving' secondary characters that make up the combination,
#	@{$secchars_for_rnode[$r]} listing indices of characters in @sec_chars that give the maximum penalty drop for the $remain_nodes->[$r] th node (= undef if no secondary characters satisfy the condition), with
#	$droppen_for_rnode[$r]  being the maximum penalty drop achieved for the $remain_nodes->[$r] th node by the secondary characters in the combination ( = undef if the penalty at the node does NOT improve over the best primary), and
#	$max_droppen_for_combi being the maximum penalty drop achieved by the combination given;
#
#	@improving_secs lists the indices of secondary characters that can improve the score over the best primary-only combination, in order of decreasing penalty drop;
#
#	@{$key_marginal_nodes_for_sec[$sc]} = (\@key_nodes, \@marginal)
#	for the secondary character $sec_chars[$sc], with
#	$remain_nodes->[$key_nodes[$kn]] being the ID of the $kn th remaining node
#	where the secondary character has a lower penalty than any primary characters.
#	(Here \@marginal will not be necessary);
#
#	$diffpens_secchars_for_rem[$sc]->[$r]is the differential penalty when the node with ID $remain_nodes->[$r] has the $sec_chars[$sc] character, subtracted by $minpen_amng_prim_for_pnodes->[$r];
#
#	$minpen_prim_to_sec[$sc] gives the minimum penalty for a change from a primary character (in ($pchar, @uniq_chars)) to $sec_chars[$sc];
#
# OUTPUT: \@dressed_core_maxdroppen_combi_secchars, where
#	@dressed_core_maxdroppen_combi_secchars stores
#	core max penalty-dropping combinations of 'improving' secondary characters
#	that are 'dressed' with the 'improving' secondary characters that does NOT reduce the penalty drop by the core combination.
#
#	Rest assured that this output set is non-redundant.
#
sub dress_core_maxdroppen_combi_secchars (\@\@\@\@\@) {

	my ($core_maxdroppen_combi_secchars, $improving_secs, $key_marginal_nodes_for_sec, $diffpens_secchars_for_rem, $minpen_prim_to_sec) = @_;

	unless (@{$core_maxdroppen_combi_secchars}>0) {
	    return []; # When the core is empty, return an empty set.
	}

	my $ct_improving = @{$improving_secs};
	my @full_imprv_indices = (0 .. $ct_improving-1);
	my @dressed_core_maxdroppen_combi_secchars = ();

	while (my $core_combi = shift @{$core_maxdroppen_combi_secchars}) { # Main outer while-loop (over 'core' maximum penalty-dropping combinations) #

	    my ($imprv_secchar_indices, $secchars_for_rnode, $droppen_for_rnode, $max_droppen_for_core) = @{$core_combi};

	    my @tried_imprv_secs;
	    if (@{$imprv_secchar_indices}>1) {
			# In this case, try all improving secondary characters NOT included in this particular core. #
		my ($common, $ex_full, $ex_core) = diff_nmarray (@full_imprv_indices, @{$imprv_secchar_indices});
		@tried_imprv_secs = @{$ex_full};
	    } else {
			# In this case, only try the 'downstream' characters, to avoid clear double counting.
		@tried_imprv_secs = ($imprv_secchar_indices->[0]+1 .. $ct_improving -1);
	    }

	    my @set_dressed_cores  = ($core_combi);
	    foreach my $indx_tried_imprv_sec (@tried_imprv_secs) { # Middle foreach-loop (over improving secondary characters to be tried) #
		my $indx_sec = $improving_secs->[$indx_tried_imprv_sec];
		my @new_set_dressed_cores = ();

		while (my $original = shift @set_dressed_cores) { # Inner while-loop (over elements of the current set of dressed cores) #

		    push @new_set_dressed_cores, $original; # Always keep the original combination.

		    my $org_imprv_sec_indices = $original->[0];
		    my $cand_new_dressed_core = incorporate_secchar_into_seccombi (@{$original}, $indx_tried_imprv_sec, @{$improving_secs}, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem}, @{$minpen_prim_to_sec});
		    unless (defined $cand_new_dressed_core) {
			print STDOUT "[ERROR 'dress_core_maxdroppen_combi_secchars'] 'incorporate_secchar_into_seccombi' FAILED to incorporate the $indx_tried_imprv_sec th improving secondary character (or the $indx_sec th secondary character) into the set (", join (', ', @{$org_imprv_sec_indices}), "): $! !\n";
			return undef;
		    }
		    my $cand_new_droppen = $cand_new_dressed_core->[3];

		    if ($cand_new_droppen < $max_droppen_for_core) { # This canNOT happen. #

			print STDOUT "[ERROR 'dress_core_maxdroppen_combi_secchars'] It's IMPOSSIBLE ! Somehow, the penalty drop IMPROVED (from $max_droppen_for_core to $cand_new_droppen) in the incorporation of the $indx_tried_imprv_sec th improving secondary character into the (dressed) core set (", join (', ', @{$org_imprv_sec_indices}), "): $! !\n";
			return undef;

		    } elsif ($cand_new_droppen > $max_droppen_for_core) {
			next; # The incorporated 'improving' secondary character reduced the core penalty drop. #
		    }
			# The 'core' penalty drop did NOT change through the incorporation. #
		    push @new_set_dressed_cores, $cand_new_dressed_core;

		}	# End of the inner while-loop (over elements of the current set of dressed cores) #

			# Update the set of dressed cores. #
		@set_dressed_cores = @new_set_dressed_cores;

	    }	# End of the middle foreach-loop (over improving secondary characters to be tried) #

	    push @dressed_core_maxdroppen_combi_secchars, @set_dressed_cores;

	}	# End of the main outer while-loop (over 'core' maximum penalty-dropping combinations) #

	my $nr_dressed_core_maxdroppen_combi_secchars = make_nonredundant_combinations (@dressed_core_maxdroppen_combi_secchars, 0);

	unless (defined $nr_dressed_core_maxdroppen_combi_secchars) {
	    print STDOUT "[ERROR 'dress_core_maxdroppen_combi_secchars'] The subroutine 'make_nonredundant_combinations' FAILED: $! !\n";
	}

	return $nr_dressed_core_maxdroppen_combi_secchars;

}	# End of the subroutine 'dress_core_maxdroppen_combi_secchars' #


# attach_keeping_secs_to_maxdroppen_combi:
#	This satellite subroutine of the subroutine 'part_resolve_mlt_node_bu'
#	attaches 'KEEPING' secondary characters to each 'dressed core' combination of secondary characters
#	and completes the set of all combinations of secondary characters that achieve the maximum drop in penalty.
#
#	The subroutine was created by slightly modifying the subroutine 'dress_core_maxdroppen_combi_secchars'.
#	Although there may be more efficient algorithms,
#	I made this choice to facilitate to fix potential bugs.
#
#	NOTE, however, that the indices defining each combination is of @sec_chars in this subroutine,
#	instead of @{$improving_secs} as in 'dress_core_maxdroppen_combi_secchars'.
#
# dress_core_maxdroppen_combi_secchars:
#	This satellite subroutine of the subroutine 'part_resolve_mlt_node_bu'
#	'dress'es each 'core' secondary-character combination with the improving secondary characters
#	NOT included in the core and yet NOT reducing the penalty drop.
#
#	NOTE that this subroutine eats up the input @core_maxdroppen_combi_secchars.
#	
#
# INPUT: (\@drssd_core_maxdroppen_combi_secchars, \@keeping_secs, \@key_marginal_nodes_for_sec, \@diffpens_secchars_for_rem, \@minpen_prim_to_sec), where
#	@{$drssd_core_maxdroppen_combi_secchars[$cm]} = (\@secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $max_droppen_for_combi)
#	gives the $cm th 'dressed core' combination of 'improving' secondary characters on the 'remaining nodes' that gives the maximum drop in the penalty (compared with the 'best' combination of primary characters), with
#	@secchar_indices listing indices in @sec_chars for the 'improving' secondary characters that make up the combination,
#	@{$secchars_for_rnode[$r]} listing indices of characters in @sec_chars that give the maximum penalty drop for the $remain_nodes->[$r] th node (= undef if no secondary characters satisfy the condition), with
#	$droppen_for_rnode[$r]  being the maximum penalty drop achieved for the $remain_nodes->[$r] th node by the secondary characters in the combination ( = undef if the penalty at the node does NOT improve over the best primary), and
#	$max_droppen_for_combi being the maximum penalty drop achieved by the combination given;
#
#	@keeping_secs lists the indices of secondary characters that can at best maintain the maximum score;
#
#	@{$key_marginal_nodes_for_sec[$sc]} = (\@key_nodes, \@marginal)
#	for the secondary character $sec_chars[$sc], with
#	$remain_nodes->[$key_nodes[$kn]] being the ID of the $kn th remaining node
#	where the secondary character has a lower penalty than any primary characters.
#	(Here \@marginal will not be necessary);
#
#	$diffpens_secchars_for_rem[$sc]->[$r]is the differential penalty when the node with ID $remain_nodes->[$r] has the $sec_chars[$sc] character, subtracted by $minpen_amng_prim_for_pnodes->[$r];
#
#	$minpen_prim_to_sec[$sc] gives the minimum penalty for a change from a primary character (in ($pchar, @uniq_chars)) to $sec_chars[$sc];
#
# OUTPUT: \@fullset_maxdroppen_combi_secchars, where
#	@fullset_maxdroppen_combi_secchars stores ALL THE
#	core max penalty-dropping combinations of secondary characters.
#
#	Rest assured that this output set is non-redundant.
#
sub attach_keeping_secs_to_maxdroppen_combi (\@\@\@\@\@) {

	my ($drssd_core_maxdroppen_combi_secchars, $keeping_secs, $key_marginal_nodes_for_sec, $diffpens_secchars_for_rem, $minpen_prim_to_sec) = @_;

	unless (@{$drssd_core_maxdroppen_combi_secchars}>0) {
	    print STDOUT "[ERROR 'attach_keeping_secs_to_maxdroppen_combi'] ILLEGAL INPUT: EMPTY SET of 'dressed core' secondary character combinations giving maximum penalty drop: $! !\n";
	    return undef;
	}

	my $ct_secchars = @{$minpen_prim_to_sec};
	my @full_secchar_indices = (0 .. $ct_secchars-1);
	my @full_maxdroppen_combi_secchars = ();

	while (my $maxdp_combi = shift @{$drssd_core_maxdroppen_combi_secchars}) { # Main outer while-loop (over 'dressed core' maximum penalty-dropping combinations) #

	    my ($secchar_indices, $secchars_for_rnode, $droppen_for_rnode, $max_droppen_for_combi) = @{$maxdp_combi};

	    my @set_maxdp_combs  = ($maxdp_combi);
	    foreach my $indx_kp_sec (@{$keeping_secs}) { # Middle foreach-loop (over keeping secondary characters). #
		## my $indx_sec = $improving_secs->[$indx_tried_imprv_sec];
		my @new_set_maxdp_combs = ();

		while (my $original = shift @set_maxdp_combs) { # Inner while-loop (over elements of the current set of max penalty-drop combinations) #

		    push @new_set_maxdp_combs, $original; # Always keep the original combination.

		    my $org_sec_indices = $original->[0];

			# The key revision when creating 'attach_keeping_secs_to_maxdroppen_combi'. #
			# (COULD be replaced with a more efficient subroutine (and the subsequent decision-making), if I have time.)

		    my $cand_new_maxdp_combi = incorporate_secchar_into_seccombi (@{$original}, $indx_kp_sec, @full_secchar_indices, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem}, @{$minpen_prim_to_sec});
		    ##my $cand_new_dressed_core = incorporate_secchar_into_seccombi (@{$original}, $indx_tried_imprv_sec, @{$improving_secs}, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem}, @{$minpen_prim_to_sec});
		    unless (defined $cand_new_maxdp_combi) {
			print STDOUT "[ERROR 'attach_keeping_secs_to_maxdroppen_combi'] 'incorporate_secchar_into_seccombi' FAILED to incorporate the $indx_kp_sec th secondary character, one of the 'keeping' characters', into the set (", join (', ', @{$org_sec_indices}), "): $! !\n";
			return undef;
		    }
		    my $cand_new_droppen = $cand_new_maxdp_combi->[3];

		    if ($cand_new_droppen < $max_droppen_for_combi) { # This canNOT happen. #

			print STDOUT "[ERROR 'attach_keeping_secs_to_maxdroppen_combi'] It's IMPOSSIBLE ! Somehow, the penalty drop IMPROVED (from $max_droppen_for_combi to $cand_new_droppen) in the incorporation of the $indx_kp_sec th secondary character, one of the 'keeping' secondary character,  into the set (", join (', ', @{$org_sec_indices}), "): $! !\n";
			return undef;

		    } elsif ($cand_new_droppen > $max_droppen_for_combi) {
			next; # The incorporated 'keeping' secondary character REDUCED the penalty drop. #
		    }
			# The 'core' penalty drop was KEPT through the incorporation. #
		    push @new_set_maxdp_combs, $cand_new_maxdp_combi;

		}	# End of the inner while-loop (over elements of the current set of max penalty-droppping combinations) #

			# Update the set of dressed cores. #
		@set_maxdp_combs = @new_set_maxdp_combs;

	    }	# End of the middle foreach-loop (over keeping secondary characters) #

	    push @full_maxdroppen_combi_secchars, @set_maxdp_combs;

	}	# End of the main outer while-loop (over 'dressed core' maximum penalty-dropping combinations) #

		# Sort the combinations (the redundancy removal may be unnecessary, though). #
	my $nr_full_maxdroppen_combi_secchars = make_nonredundant_combinations (@full_maxdroppen_combi_secchars, 0);

	unless (defined $nr_full_maxdroppen_combi_secchars) {
	    print STDOUT "[ERROR 'attach_keeping_secs_to_maxdroppen_combi'] The subroutine 'make_nonredundant_combinations' FAILED: $! !\n";
	    return undef;
	}

	return $nr_full_maxdroppen_combi_secchars;

}	# End of the subroutine 'attach_keeping_secs_to_maxdroppen_combi' #


# enumerate_maxdroppen_char_patterns:
#	This satellite subroutine of 'part_resolve_mlt_node_bu' makes ALL the character patterns
#	for the nodes withOUT unique characters,
#	starting from the set of ALL secondary character combinations that maximize the penalty drop.
#
#	ALGORITHM:
#	From each combination of secondary characters,
#	the character patterns are constructed by specifying each of ALL possible characters for each 'remaining' node, node by node.
#	The possible characters are assigned to each node as follows:
#	(i) If the node has a particular set of 'key' secondary characters, assign each of these characters;
#	(ii) Otherwise, assign each of 'best' primary characters or a 'marginal' secondary character among those defining the combination.
#
# INPUT: (\@set_maxdroppen_combi_secchars, \@minpen_prim_chars_for_rnodes, \@sec_chars, \@key_marginal_nodes_for_sec), where
#
#	@{$set_maxdroppen_combi_secchars[$md]} = (\@secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $max_droppen_for_combi)
#	specifies the $md th secondary-character combination that gives the maximum drop in penalty, with
#	@secchar_indices listing the indices for @sec_chars of the secondary characters in this combination, and
#	@{$secchars_for_rnode[$r]} lists the indices for @sec_chars of the secondary characters that should maximize the penalty drop
#	at the $r th 'remaining' node. $secchars_for_rnode is undef if there is no such character for the node.
#	(The remaining elements are unnecessary in this subroutine.);
#
#	@{$minpen_prim_chars_for_rnodes->[$r]} lists all primary characters giving the minimum penalty among primary at the node with ID = $remain_nodes->[$r] (and the subtree under it);
#	@{$key_marginal_nodes_for_sec[$sc]} = (\@key_nodes, \@marginal)
#	for the secondary character $sec_chars[$sc], with
#	$remain_nodes->[$key_nodes[$kn]] being the ID of the $kn th remaining node
#	where the secondary character has a lower penalty than any primary characters, and with
#	$remain_nodes->[$marginal[$mg]] beging the ID of the $mg th remaining node
#	where the secondary character ties the lowest-penalty primary node.
#
# OUTPUT: \@setset_maxdroppen_char_patterns, where
#	@char_pattern = split //, ${$setset_maxdroppen_char_patterns->[$md][$p]}
#	is the $p th character pattern made from the combination $fullset_maxdroppen_combi_secchars->[$md], with
#	$char_pattern[$r] is the index, for the  whole set of characters, of the character at the $r th remaining node.
#
sub enumerate_maxdroppen_char_patterns (\@\@\@\@) {

	my ($set_maxdroppen_combi_secchars, $minpen_prim_chars_for_rnodes, $sec_chars, $key_marginal_nodes_for_sec) = @_;

	my $ct_rnodes = @{$minpen_prim_chars_for_rnodes};
	my $ct_sec_chars = @{$sec_chars};

	my @setset_maxdroppen_char_patterns = ();

	foreach my $maxdp_combi (@{$set_maxdroppen_combi_secchars}) { # Main outer foreach-loop (over maximum penalty-dropping combinations) #

	    my ($secchar_indices, $secchars_for_rnode) = @{$maxdp_combi}[0,1];

	    my $if_undef=0;
	    for (my $r=0; $r<$ct_rnodes; $r++) { # Preliminary for-loop (I) #
		(defined $secchars_for_rnode->[$r]) and next;
		$if_undef = 1;
		last;
	    }			# End of the preliminary for-loop (I) #

	    my @marginal_sc_for_rnode = (); # The $r th element lists the IDs of the secondary character for which the $r th node is 'marginal'.
	    if ($if_undef>0) { # If there is any 'remaining' nodes to which NO secondary node is assigned.
		foreach my $indx_sc (@{$secchar_indices}) { # Middle preliminary foreach-loop (II) #
		    my $id_sc = $sec_chars->[$indx_sc];
		    my $marginal_rnode = $key_marginal_nodes_for_sec->[$indx_sc][1];
		    (defined $marginal_rnode) or ($marginal_rnode = []);
		    foreach my $indx_rnode (@{$marginal_rnode}) { # Inner foreach-loop #
			(defined $marginal_sc_for_rnode[$indx_rnode]) or ($marginal_sc_for_rnode[$indx_rnode] = []);
			push @{$marginal_sc_for_rnode[$indx_rnode]}, $id_sc;
		    }			# End of the inner foreach-loop #
		}		# End of the middle preliminary foreach-loop (II) #
	    }

	    my $nucleus = '';
	    my @set_mdpen_patterns = (\$nucleus);

	    for (my $r=0; $r<$ct_rnodes; $r++) { # Main middle for-loop (over 'remaining' nodes) #
		my @cand_chars = ();
		my $cand_secchars = $secchars_for_rnode->[$r];
		if ((defined $cand_secchars) and (@{$cand_secchars}>0) ) {
			# If the node is a 'key' node for some secondary characters, use them.#
		   @cand_chars = ();
		   foreach my $indx_sc (@{$cand_secchars}) { push @cand_chars, $sec_chars->[$indx_sc]; }
		} else {
			# Otherwise, use 'minimum penalty' primary characters
			# as well as secondary characters for which the node is 'marginal'.
		   @cand_chars = @{$minpen_prim_chars_for_rnodes->[$r]};
		   my $marginal = $marginal_sc_for_rnode[$r];
		   (defined $marginal) and (push @cand_chars, @{$marginal});
		}

		my @new_set_mdpen_patterns = ();
		while (my $curr_mdpen_pattern = shift @set_mdpen_patterns) { # Inner while-loop (over current portions of max penalty-dropping patterns (for the combination) #
		    foreach my $cand_char (@cand_chars) { # Innermost foreach-loop (over candidate characters) #
			my $new_mdpen_pattern = ${$curr_mdpen_pattern};
			$new_mdpen_pattern .= $cand_char;
			push @new_set_mdpen_patterns, \$new_mdpen_pattern;
		    }	    # End of the innermost foreach-loop (over candidate characters) #
		}	# End of the inner while-loop (over current portions of max penalty-dropping patterns (for the combination) #

			# Update the current set of maximum penalty-dropping patterns. #
		@set_mdpen_patterns = @new_set_mdpen_patterns;

	    }		# End of the main middle for-loop (over 'remaining' nodes) #

	    push @setset_maxdroppen_char_patterns, \@set_mdpen_patterns;

	}	# End of the main outer foreach-loop (over maximum penalty-dropping combinations) #

	return \@setset_maxdroppen_char_patterns;

}		# End of the subroutine 'enumerate_maxdroppen_char_patterns' #


# make_parsim_scenarios_fm_pattern:
#	This satellite subroutine of the subroutine 'part_resolve_mlt_node_bu'
#	constructs a parsimonious scenario from a set of maximum penalty-droppping character pattern (on the 'remaining' nodes),
#	accompanied by the nodes with unique characters.
#
#	NOTE: In this version, the following rules are used:
#	(i) The common ancestor of nodes with each unique character is assumed to be directly descended from the subject node;
#	(ii) The common ancestor of nodes with each secondary character will be
#	descended from the common ancestor of the primary character whose change to the secondary character suffers the minimum penalty;
#	(iii) If the character is assumed by a single node, or if some nodes assume the parent character, such nodes will be directly descended from the subject node.
#
# INPUT: (\@char_pattern, \@remain_nodes, \%uniq_char2nodes, $pchar, $sbjct_id, \$max_nodeid, \%sec2best_prims, \@pen_mtrx, \%node2min_penalties), where
#	@char_pattern specifies the character pattern, with $char_pattern[$r] representing
#	the character at the node with ID $remain_nodes[$r] ;
#	@remain_nodes stores IDs of 'remaining' nodes, which do NOT have their own unique characters;
#	%uniq_char2nodes = ($uniq_char => \@node_ids, ...) for the child nodes with unique characters;
#	$pchar is the character at the subject node (aka the 'parent' node);
#	$sbjct_id is the node ID of the multifurcated node in question;
#	$max_nodeid is the maximum node ID that has ever be assigned concerning this tree;
#	%sec2best_prims = ($id_sec_char => \@id_best_prim_chars, ...);
#	$pen_mtrx[$i]->[$j] is the penalty imposed on the change
#	from the $i th character to the $j th character;
#	%node2min_penalties = ($node_id => \@min_penalties, ...) for ALL lowermost nodes,
#	with $min_penalties[$i] being the minimum summed penalty
#	on the subtree under $node_id under the condition that the node has
#	the $i th character in the matrix @pen_mtrx.
#	($min_penalties[$i] is undef if the $i th character is NOT the character on the exterior node $node_id.
#
# [OUTPUT: \@set_parsim_scenarios, where
#	@{$set_parsim_scenarios[$s]} = (\%add_node2ch, \%add_node2min_penalty, \%add_char_cngs, \%chnode2char) ] --- Old version (became Obsolete as of Dec 24 (Sat), 2011)
# OUTPUT: (\@set_parsim_scenarios, \chnode2char), where
#	@{$set_parsim_scenarios[$s]} = (\%add_node2ch, \%add_node2min_penalty, \%add_char_cngs)
#	specifies the $s th parsimonious scenario from the pattern @char_pattern, with
#	%add_node2ch = ($additional_node_id => \@children, ...) giving the partial resolution (including $additonal_node_id = $sbjct_id),
#	%add_node2min_penalty = ($additional_node_id => $min_penalty, ...) giving the minimum summed penalty under $additional_node_id in this particular parsimonious scenario,
#	%add_char_chngs = ($branch_id (including additional branch) => [$char_up, $char_lw, $pen], ...) describing the changes (from $char_up to $char_lw, which incurrs the penalty of $pen) that comprizes the scenario;
#
#	%chnode2char = ($child_node_id => $character_id, ...) specifies the characters assigned to the child nodes. (This should be shared by ALL the parsimonious scenarios output in a session.)
#
sub make_parsim_scenarios_fm_pattern (\@\@\%$$\$\%\@\%) {

	my ($char_pattern, $remain_nodes, $uniq_char2nodes, $pchar, $sbjct_id, $max_nodeid, $sec2best_prims, $pen_mtrx, $node2min_penalties) = @_;

	my $ct_rnodes = @{$remain_nodes};
	unless (@{$char_pattern} == $ct_rnodes) {
	    print STDOUT "[ERROR 'make_parsim_scenarios_fm_pattern'] ILLEGAL INPUT: The size of the character pattern (=", scalar (@{$char_pattern}), ") CLASHES with the number of 'remaining' nodes (=$ct_rnodes): $! !\n";
	    return (undef, undef, undef);
	}

	my @prim_chars = ($pchar);

	my %char2nodes; # ($character_id => \@node_ids, ...) for ALL characters in the pattern.
	my %chnode2char;
	foreach my $char (keys %{$uniq_char2nodes}) { # 1st foreach-loop (over the unique characters) #
	    my @nodes = @{$uniq_char2nodes->{$char}};
	    $char2nodes{$char} = \@nodes;
	    ($char eq $pchar) or (push @prim_chars, $char);
	    foreach my $node (@nodes) { $chnode2char{$node} = $char; }
	} 		# End of the 1st foreach-loop (over the unique characters) #

	for (my $r=0; $r<$ct_rnodes; $r++) {  # 2nd for-loop (over 'remaining' nodes) #
	    my $char = $char_pattern->[$r];
	    my $node = $remain_nodes->[$r];
	    (defined $char2nodes{$char}) or ($char2nodes{$char} = []);
	    push @{$char2nodes{$char}}, $node;
	    $chnode2char{$node} = $char;
	}		# End of the 2nd for-loop (over 'remaining' nodes) #

	my %char2add_node = ($pchar => $sbjct_id); # An auxiliary hash #
	my %add_node2ch = ($sbjct_id => []);
	my %add_node2min_penalty = ($sbjct_id => 0);
	my %add_char_chngs;

		# First, deal with the primary character. #
	foreach my $char (@prim_chars) { # 3rd foreach-loop (over the primary characters) #
	    my $nodes = $char2nodes{$char};
	    ((defined $nodes) and (@{$nodes}>0)) or next;

	    my $sum_pen = 0;
	    foreach my $node (@{$nodes}) { $sum_pen += $node2min_penalties->{$node}[$char]; }
	    if ($char eq $pchar) { # The nodes assume the parent character.
		push @{$add_node2ch{$sbjct_id}}, @{$nodes};
		$add_node2min_penalty{$sbjct_id} += $sum_pen;
		next;

=DISABLED to cope with a primary character that has a single child node
 # AND that is a best primary for one or more secondary characters (on Dec 24 (Sat), 2011).

	    } elsif (@{$nodes} == 1) {# Only a single node assumes this character.
		my $node = $nodes->[0];
		push @{$add_node2ch{$sbjct_id}}, $node;
		my $penalty = $pen_mtrx->[$pchar][$char];
		$add_char_chngs{$node} = [$pchar, $char, $penalty];

		$add_node2min_penalty{$sbjct_id} += $sum_pen + $penalty;
		next;
=cut
	    }
		# Otherwise #
	    ${$max_nodeid}++;
	    my $common_anc = ${$max_nodeid};
	    my @nodes = @{$nodes};
	    $char2add_node{$char} = $common_anc;
	    $add_node2ch{$common_anc} = \@nodes;
	    $add_node2min_penalty{$common_anc} = $sum_pen;

	    push @{$add_node2ch{$sbjct_id}}, $common_anc;
	    my $penalty = $pen_mtrx->[$pchar][$char];
	    $add_char_chngs{$common_anc} = [$pchar, $char, $penalty];
	    $add_node2min_penalty{$sbjct_id} += $sum_pen + $penalty;

	}		# End of the 3rd foreach-loop (over the primary characters) #

		# Slightly MODIFIED on Dec 24 (Sat), 2011 #
	my @set_parsim_scenarios = ([\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs] );
	##my @set_parsim_scenarios = ([\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs, \%chnode2char]);

	foreach my $char (keys %{$sec2best_prims}) { # 4th outer foreach-loop (over the secondary characters) #
	     my $nodes = $char2nodes{$char};
	     ((defined $nodes) and (@{$nodes}>0)) or next;
	     my @nodes = @{$nodes};
	     my $sum_pen = 0;
	     foreach my $node (@nodes) { $sum_pen += $node2min_penalties->{$node}[$char]; }
	     my $common_anc ;
	     if (@nodes>1) {
		${$max_nodeid}++;
		$common_anc = ${$max_nodeid};
		$char2add_node{$char} = $common_anc;
		##$add_node2min_penalty{$common_anc} = $sum_pen;
		##$add_node2ch{$common_anc} = \@nodes;
	     } else {
		$common_anc = $nodes[0];
	     }

	     my @best_prims = @{$sec2best_prims->{$char}};
	     my @new_set_parsim_scenarios = ();

	     while (my $parsim_scenario = shift @set_parsim_scenarios) { # Middle while-loop (over current parsimonious scenarios) #

			# Slightly MODIFIED on Dec 24 (Sat), 2011 #
		my ($add_node2ch, $add_node2min_penalty, $add_char_chngs) = @{$parsim_scenario};
		##my ($add_node2ch, $add_node2min_penalty, $add_char_chngs, $chnode2char) = @{$parsim_scenario};
		foreach my $prim (@best_prims) { # Inner foreach-loop (over best primary characters) #
		    my %new_add_node2ch;
		    foreach my $node (keys %{$add_node2ch}) {
		    	my @children = @{$add_node2ch->{$node}};
		    	$new_add_node2ch{$node} = \@children;
		    }
		    my %new_add_node2min_penalty = %{$add_node2min_penalty};
		    my %new_add_char_chngs;
		    foreach my $node (keys %{$add_char_chngs}) {
		    	my @changes = @{$add_char_chngs->{$node}};
		    	$new_add_char_chngs{$node} = \@changes;
		    }
		    if (@nodes>1) { # Incorporate the set of nodes #
			my @cp_nodes = @nodes;
			$new_add_node2min_penalty{$common_anc} = $sum_pen;
			$new_add_node2ch{$common_anc} = \@cp_nodes;
		    }

			# Specify the parent node #
		    my $pnode = $char2add_node{$prim};
		    push @{$new_add_node2ch{$pnode}}, $common_anc;

			# Incorporate the character change #
		    my $new_pen = $pen_mtrx->[$prim][$char];
		    $new_add_char_chngs{$common_anc} = [$prim, $char, $new_pen];
		    $new_add_node2min_penalty{$pnode} += $sum_pen + $new_pen;

			# ADDED to fix a bug on Dec 24 (Sat), 2011 #
			# (This should be replaced with a more general measure
			# when the subroutine is to create general genealogical structures among the relevant characters.)
		    ($pnode eq $sbjct_id) or ($new_add_node2min_penalty{$sbjct_id} += $sum_pen + $new_pen);

			# Slightly MODIFIED on Dec 24 (Sat), 2011 #
		    push @new_set_parsim_scenarios, [\%new_add_node2ch, \%new_add_node2min_penalty, \%new_add_char_chngs];
		    ## push @new_set_parsim_scenarios, [\%new_add_node2ch, \%new_add_node2min_penalty, \%new_add_char_chngs, $chnode2char];

		}    # End of the inner foreach-loop (over best primary characters) #

	     }	# End of the middle while-loop (over current parsimonious scenarios) #
	     
			# Update the set of current parsimonious scenarios #
	     @set_parsim_scenarios = @new_set_parsim_scenarios;

	}	# End of the 4th outer foreach-loop (over the secondary characters) #

	   # Remove additional nodes that is assigned with a primary character (but NOT the parent character) AND that has only a single child. #
	   # (ADDED to cope with a primary character that has a single child node
 	   # AND that is a best primary for one or more secondary characters (on Dec 24 (Sat), 2011)).

	foreach my $parsim_scenario (@set_parsim_scenarios) { # 5th outer foreach-loop (over the parsimonious scenarios) #


			# Slightly MODIFIED on Dec 24 (Sat), 2011 #
	    my ($add_node2ch, $add_node2min_pen, $add_char_chngs) = @{$parsim_scenario};
	    ## my ($add_node2ch, $add_node2min_pen, $add_char_chngs, $chnode2char) = @{$parsim_scenario};

	    my %old2new; # = ($add_node => $child, ...) for the additional nodes each with a single child.
	    foreach my $prim (@prim_chars) { # Inner foreach-loop (over the primary characters) #
		($prim eq $pchar) and next; # Skip the parent character #
		my $add_node = $char2add_node{$prim};
		my $chnodes = $add_node2ch->{$add_node};
		((defined $chnodes) and (@{$chnodes}==1)) or next; # Skip unless the additional node has a single chlid node.

		my $child = $chnodes->[0];
		$old2new{$add_node} = $child;

		    # REVISED to Fix a BUG on Dec 26 (Mon), 2011) #
		unless (defined $node2min_penalties->{$child}) {
			# If $child is NOT a lowermost node  (i.e. a node recorded in %{$node2min_penalties}),
			# keep the minimum penalty of the $add_node (at $child). #
		    my $min_pen = $add_node2min_pen->{$add_node};
		    $add_node2min_pen->{$child} = $min_pen;
		}
		##my $min_pen = $add_node2min_pen->{$add_node};
		##$add_node2min_pen->{$child} = $min_pen;

		delete $add_node2min_pen->{$add_node};

		my $char_chng = $add_char_chngs->{$add_node};
		$add_char_chngs->{$child} = $char_chng;
		delete $add_char_chngs->{$add_node};

		delete $add_node2ch->{$add_node};

	    }	    # End of the inner foreach-loop (over the primary characters) #

	    if (scalar (keys %old2new) > 0) {
		my @new_sb_chnodes = ();
		foreach my $ch (@{$add_node2ch->{$sbjct_id}}) {
		    my $replace = $old2new{$ch};
		    push @new_sb_chnodes, ((defined $replace) ? $replace : $ch);
		}
		$add_node2ch->{$sbjct_id} = \@new_sb_chnodes;
	    }

	}   # End of the 5th outer foreach-loop (over the parsiminious scenarios) #

		# Slightly MODIFIED (on Dec 24 (Sat), 2011) #
	return (\@set_parsim_scenarios, \%chnode2char);
	##return \@set_parsim_scenarios;

}		# End of the subroutine 'make_parsim_scenarios_fm_pattern' #


# add_parsim_scenarios_by_chng_uniq_chars:
#	This satellite subroutine of the subroutine 'part_resolve_mlt_node_bu'
#	constructs additonal parsimonious scenarios from an input parsimonious scenario,
#	by changing 'unique' characters each of which is assigned to a single child node
#	and each of whose total penalty from the 'subject node' ties those of other character.
#
#	NOTE: In this version, the subroutine ONLY tries to replace each 'lone' unique character
#	with another character assigned either to the subject node or to one of its immediate children.
#	=> REVISED on Dec 26 (Mon), 2011:
#	The subroutine now tries to re-graft each 'lone' unique character
#	onto the subject node as well as the nodes whose characters were gained by the mutations immediately above them.
#
# INPUT: (\@info_parsim_scenario, \%chnode2char, \%sgl_node2uniq_char, $pchar, $sbjct_id, \$max_nodeid, \@pen_mtrx, \%node2min_penalties), where
#	@info_parsim_scenario = (\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs)
#	specifies the input parsimonious scenario, with
#	%add_node2ch = ($additional_node_id => \@children, ...) giving the partial resolution (including $additonal_node_id = $sbjct_id),
#	%add_node2min_penalty = ($additional_node_id => $min_penalty, ...) giving the minimum summed penalty under $additional_node_id in this particular parsimonious scenario,
#	%add_char_chngs = ($branch_id (including additional branch) => [$char_up, $char_lw, $pen], ...) describing the changes (from $char_up to $char_lw, which incurrs the penalty of $pen) that comprizes the scenario;
#
#	%chnode2char = ($child_node_id => $character_id, ...) specifies the characters assigned to the child nodes under the input parsimonious scenario;
#	%sgl_node2uniq_char = ($node_id => $character, ...) for the (child) nodes each of which is the only node with the unique character $character (excluding $pchar);
#	%uniq_char2nodes = ($uniq_char => \@node_ids, ...) for the child nodes with unique characters;
#	$pchar is the character at the subject node (aka the 'parent' node);
#	$sbjct_id is the node ID of the multifurcated node in question;
#	$max_nodeid is the maximum node ID that has ever be assigned concerning this tree;
#	$pen_mtrx[$i]->[$j] is the penalty imposed on the change
#	from the $i th character to the $j th character;
#	%node2min_penalties = ($node_id => \@min_penalties, ...) for ALL nodes,
#	with $min_penalties[$i] being the minimum summed penalty
#	on the subtree under $node_id under the condition that the node has
#	the $i th character in the matrix @pen_mtrx.
#	($min_penalties[$i] is undef if the $i th character is NOT the character on the exterior node $node_id.
#
# OUTPUT: \@set_info_new_psm_scns, where
#	@{$set_info_new_psm_scns[$aps]} = (\@info_new_parsim_scenario, \%new_chnode2char) specifies the $aps th additional parsimonious scenario,
#	with @info_new_parsim_scenario = (\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs)
#	following the same format as the input @info_parsim_scenario,
#	and with %new_chnode2char = ($child_node_id => $character, ...)
#	describing the character pattern accompanying this particular additional parsimonious scenario.
#
sub add_parsim_scenarios_by_chng_uniq_chars (\@\%\%$$\$\@\%) {

	my ($info_parsim_scenario, $chnode2char, $sgl_node2uniq_char, $pchar, $sbjct_id, $max_nodeid, $pen_mtrx, $node2min_penalties) = @_;

		# A quick screening #
	my @sgl_unq_nodes = ();
	foreach my $node (sort {$a <=> $b} keys %{$sgl_node2uniq_char}) { ($pchar == $sgl_node2uniq_char->{$node}) or (push @sgl_unq_nodes, $node); } 
	unless (@sgl_unq_nodes>0) {
	    return []; # There is NO CHANCE of finding additonal scenarios by this subroutine,
			# if there is no single (or 'lone') node assigned with a unique character. 
	}

	my ($add_node2ch, $add_node2minpen, $add_char_chngs) = @{$info_parsim_scenario};
	my @immediate_children = @{$add_node2ch->{$sbjct_id}};

	my %destn2char = ($sbjct_id => $pchar); # ($potential_regrafting_destination => $character_after_regrafting, ...)
	my %node2pa; # ($node_id => $parent_node_id, ...) (ADDED on Dec 26 (Mon), 2011)

		# REVISED on Dec 26 (Mon), 2011 #
	my @pa_examined = ($sbjct_id);
	while (@pa_examined>0) { # 1st preliminary outer while-loop (until the parents are exhausted) #
	    my @new_pa_examined = ();
	    while (my $pa = shift @pa_examined) {  # Preliminary middle while-loop (over the current parents) #
		my $examined = $add_node2ch->{$pa};
		((defined $examined) and (@{$examined}>0)) or next; # Skip if there is no children.
		push @new_pa_examined, @{$examined};
		foreach my $node (@{$examined}) { # Preliminary inner foreach-loop (over the nodes to be examined) #
		    $node2pa{$node} = $pa;
		    my $char_chng = $add_char_chngs->{$node};
		    (defined $char_chng) or next; # Only target the 'MRCA's of the characters. #
		    my $char = $char_chng->[1];
		    $destn2char{$node} = $char;
		}	# End of the preliminary inner foreach-loop (over the nodes to be examined) #
	    }		# End of the preliminary middle while-loop (until @pa_examined is emptied) # 
	    @pa_examined = @new_pa_examined;
	}	# End of the 1st preliminary outer while-loop (until the parents are exhausted) #

=It was REPLACED with a better version (on Dec 26 (Mon), 2011)
	foreach my $node (@immediate_children) { # 1st preliminary foreach-loop (over the immediate children of the subject) #
	    my $char = $chnode2char->{$node};
	    unless (defined $char) {
		my $char_chng = $add_char_chngs->{$node};
		unless (defined $char_chng) {
		    print STDOUT "[ERROR 'add_parsim_scenarios_by_chng_uniq_chars'] ILLEGAL INPUT: NO CHARACTER CHANGE accompanies the node '$node', which is NOT a 'child' of the subject '$sbjct_id' in the input multifurcated tree : $! !\n";
		    return undef;
		}
		$char = $char_chng->[1];
	    }
	    ($char == $pchar) and next; # Skip a child with the 'parent' character.
	    $destn2char{$node} = $char;
	}	# End of the 1st preliminary foreach-loop (over the immediate children of the subject) #
=cut
	my @pot_dests = sort keys %destn2char; # Potential destinations #

	my %sgl_unq_node2regraft_dests; # = ($single_unique_node => [destinations,for,re-grafting], ...)
	foreach my $node (@immediate_children) { # 2nd preliminary foreach-loop (over the immediate children of the subject) #
	    my $unq_char = $sgl_node2uniq_char->{$node};
	    (defined $unq_char) or next; # Skip if the node is not an only node assigned with a unique character.
	    ($unq_char == $pchar) and next; # Skip if the 'parent' character is assigned.

	    my $min_pens_for_node = $node2min_penalties->{$node};
	    my $org_pen = $min_pens_for_node->[$unq_char] + $pen_mtrx->[$pchar][$unq_char];
	    my @rg_dests = ();
	    foreach my $pot_dest (@pot_dests) { # Inner foreach-loop (over potential destinations) #

		($pot_dest == $node) and next; # Cannot be regrafted onto itself!

		my $pot_char = $destn2char{$pot_dest};
		my $pot_pen = $min_pens_for_node->[$pot_char];
		(defined $pot_pen) or next; # Skip if the penalty is not defined.
		($pot_pen > $org_pen) and next; # Skip if the penalty is worse than the original.
		if ($pot_pen < $org_pen) {
		    print STDOUT "[ERROR 'add_parsim_scenarios_by_chng_uniq_chars'] ILLEGAL INPUT: The penalty for the potential destination '$pot_dest' (character: $pot_char, penalty: $pot_pen) is SMALLER THAN the penalty, $org_pen, for the original 'unique' character ($unq_char) assigned to the single node '$node': $! !\n";
		    return undef;
		}
		push @rg_dests, $pot_dest;
	    }		# End of the inner foreach-loop (over potential destinations) #

	    (@rg_dests>0) or next; # Skip if there is no regrafting destination. #
	    $sgl_unq_node2regraft_dests{$node} = \@rg_dests;
		
	}	# End of the 2nd preliminary foreach-loop (over the immediate children of the subject) #

		# A set of single unique nodes to be re-grafted. #
	my @to_be_regrafted = sort {$a <=> $b} keys %sgl_unq_node2regraft_dests;

	unless (@to_be_regrafted>0) {
	    return []; # No chance of making an additional parsimonious scenario. #
	}

	my @init_cmbs_tbrg = ([]); # element = a combination of single unique nodes to be re-grafted. #
	foreach my $tbrg (@to_be_regrafted) { # 3rd preliminary foreach-loop (over single unique nodes to be re-grafted) #

	    my @new_cmbs_tbrg = ();
	    while (my $combi = shift @init_cmbs_tbrg) { # Inner while-loop (over the current combination. #
		my @cp_combi = @{$combi};
		push @cp_combi, $tbrg;
		push @new_cmbs_tbrg, $combi, \@cp_combi;
	    }		# End of the inner while-loop (over the current combination. #

	    @init_cmbs_tbrg = @new_cmbs_tbrg;

	}   # End of the 3rd preliminary foreach-loop  (over single unique nodes to be re-grafted) #

	my @fnl_cmbs_tbrg = (); # The final set of combinations of single unique nodes to be re-grafted. #
	my @cmbs_dest_sets = ();
		# element = a combination of sets of proper destinations.
		# Each element corresponds to an element of @fnl_cmbs_tbrg.

	while (my $cmb_tbrg = shift @init_cmbs_tbrg) { # 4th preliminary foreach-loop (over the combinations of nodes to be re-grafted) #
	    (@{$cmb_tbrg}>0) or next; # Skip an empty combination.
	    my %prohibited; # This enables a kind of 'taboo' search.
	    foreach my $tbrg (@{$cmb_tbrg}) { $prohibited{$tbrg} = 1; }

	    my @cmb_dest_sets = ();
	    foreach my $tbrg (@{$cmb_tbrg}) { # Middle foreach-loop (over nodes to be re-grafted) #
		my @proper_dests = ();
		foreach my $dest (@{$sgl_unq_node2regraft_dests{$tbrg}}) { # Inner foreach-loop (over potential destinations) #
		    (defined $prohibited{$dest}) and next;
		    push @proper_dests, $dest;
		}		# End of the inner foreach-loop (over potential destinations) #
		unless (@proper_dests>0) {
		   @cmb_dest_sets = ();
		   last;
		}
		push @cmb_dest_sets, \@proper_dests;
	    }		# End of the middle foreach-loop (over nodes to be re-grafted) #

	    (@cmb_dest_sets == @{$cmb_tbrg}) or next;  # Skip if some of the nodes to be re-grafted have NO destination.
	    push @fnl_cmbs_tbrg, $cmb_tbrg;
	    push @cmbs_dest_sets, \@cmb_dest_sets;


	}  # End of the 4th preliminary foreach-loop (over the combinations of nodes to be re-grafted) #

	my $ct_cmbs_tbrg = @fnl_cmbs_tbrg;
	unless ($ct_cmbs_tbrg>0) {
	    return []; # NO possibility of re-grafting.
	}

	my @set_info_new_psm_scns = ();

	my @org_sbjct_children = @{$add_node2ch->{$sbjct_id}};

	for (my $c=0; $c<$ct_cmbs_tbrg; $c++) { # Main outer for-loop (over the combinations of nodes to be re-grafted) #

	    my $cmb_tbrg = $fnl_cmbs_tbrg[$c];
	    my $cmb_dest_sets = $cmbs_dest_sets[$c];
	    my $ct_nodes_tbrg = @{$cmb_tbrg};

	    my @set_cmbs_dest_rg = ([]);
	    foreach my $dest_set (@{$cmb_dest_sets}) { # Preliminary middle for-loop (over the sets of destinations in a combination) #
		my @new_set_cmbs_dest_rg = ();
		while (my $cmb_dest_rg = shift @set_cmbs_dest_rg) { # Inner for-loop (over the current combinations of destinations) #

		    foreach my $dest (@{$dest_set}) { # Innermost for-loop (over the destinations) #
			my @cp_cmb_dest_rg = @{$cmb_dest_rg};
			push @cp_cmb_dest_rg, $dest;
			push @new_set_cmbs_dest_rg, \@cp_cmb_dest_rg;

		    }		# End of the innermost for-loop (over the destinations) #

		}    # End of the inner for-loop (over the current combinations of destinations) #

		@set_cmbs_dest_rg = @new_set_cmbs_dest_rg;

	    } 	# End of the preliminary middle for-loop (over the sets of destinations in a combination) #
	
			# Now, all the combinations of destinations have been created. #

	    my ($common, $new_sbjct_children, $must_be_empty) = diff_nmarray (@org_sbjct_children, @{$cmb_tbrg});
	    my $ct_new_sb_ch = @{$new_sbjct_children};

	    foreach my $cmb_dest_rg (@set_cmbs_dest_rg) { # Main middle for-loop (over the combinations of destinations) #

			# Initialize the output variables by copying the input variables. #

		my %new_chnode2char = %{$chnode2char};
		my %new_add_node2minpen = %{$add_node2minpen};

		my %new_add_node2ch;
		foreach my $node (keys %{$add_node2ch}) { # 1st inner preliminary foreach-loop #
		    my @cp_ch = ($node == $sbjct_id) ? @{$new_sbjct_children} : @{$add_node2ch->{$node}};
		    $new_add_node2ch{$node} = \@cp_ch;
		}	# End of the 1st inner preliminary foreach-loop #
		
		my %new_add_char_chngs;
		foreach my $brid (keys %{$add_char_chngs}) { # 2nd inner preliminary foreach-loop #
		    my @info_char_chng = @{$add_char_chngs->{$brid}};
		    $new_add_char_chngs{$brid} = \@info_char_chng;
		}	# End of the 2nd inner preliminary foreach-loop #

		my %new_parent; # = ($dest_rg => $new_parent_node, ...) for the destination of re-grafting that is a 'leaf' (i.e. a key of %{$chnode2char}).
		for (my $k=0; $k<$ct_nodes_tbrg; $k++) { # Main inner for-loop (over the nodes to be re-grafted) # 

		    my $node_tbrg = $cmb_tbrg->[$k];
		    my $dest_rg = $cmb_dest_rg->[$k];

		    my $dest_char = $chnode2char->{$dest_rg};
		    if (defined $dest_char) {
			# In this case, the destination was originally a 'leaf' .
			# Shift the destination to its 'parent',
			# which will be created if there isn't yet.

			my $pa_dest_rg = $new_parent{$dest_rg};
			unless (defined $pa_dest_rg) {
			    ${$max_nodeid}++;
			    $pa_dest_rg = ${$max_nodeid};
			    $new_parent{$dest_rg} = $pa_dest_rg;
			    $new_add_char_chngs{$pa_dest_rg} = $new_add_char_chngs{$dest_rg};
			    delete $new_add_char_chngs{$dest_rg};
			    $new_add_node2ch{$pa_dest_rg} = [$dest_rg];
			    $new_add_node2minpen{$pa_dest_rg} = $node2min_penalties->{$dest_rg}[$dest_char];
	## my ($info_parsim_scenario, $chnode2char, $sgl_node2uniq_char, $pchar, $sbjct_id, $max_nodeid, $pen_mtrx, $node2min_penalties) = @_;
				# ADDED on Dec 26 (Mon), 2011 #
			    my $old_pa = $node2pa{$dest_rg};
			    $node2pa{$pa_dest_rg} = $old_pa;
			    $node2pa{$dest_rg} = $pa_dest_rg;
			}
			$dest_rg = $pa_dest_rg;
			
		    } else {
			   # The destination should be an 'additional' node. #
			$dest_char = $destn2char{$dest_rg};
		    }
		    ##my $dest_char_chng = $new_add_char_chngs{$dest_rg};
		    ##(defined $dest_char) or ($dest_char = $dest_char_chng->[1]);

			# Re-graft $node_tbrg under $dest_rg. #

		    push @{$new_add_node2ch{$dest_rg}}, $node_tbrg;

				# REVISED on Dec 26 (Mon), 2011 #
		    my $upward = $dest_rg;
		    my $incl_minpen = $node2min_penalties->{$node_tbrg}[$dest_char];
		    while (defined $node2pa{$upward}) { # Inclement the minimum penalties for the ancestral nodes from $dest_rg up to (but EXCLUDING) $sbjct_id.
			(defined $new_add_node2minpen{$upward}) and ($new_add_node2minpen{$upward} += $incl_minpen);
			$upward = $node2pa{$upward};
		    }
		    ##($dest_rg == $sbjct_id) or ($new_add_node2minpen{$dest_rg} += $node2min_penalties->{$node_tbrg}[$dest_char]); # Because we know that the minimum penalty won't change at $sbjct_id, we keep it untouched.
		    $new_chnode2char{$node_tbrg} = $dest_char;

		    (defined $new_add_char_chngs{$node_tbrg}) and (delete $new_add_char_chngs{$node_tbrg});
		}	# End of the main inner for-loop (over the nodes to be re-grafted) #

			# Replace the children of $sbjct_id with their parent, whenever applicable. #
		my $new_sb_ch = $new_add_node2ch{$sbjct_id};
		for (my $i=0; $i<$ct_new_sb_ch; $i++) { # Trailing inner for-loop (over the new children of $sbjct_id) #
		    my $ch = $new_sb_ch->[$i];
		    my $pa = $new_parent{$ch};
		    (defined $pa) and ($new_sb_ch->[$i] = $pa);
		}	# End of the trailing inner for-loop (over the new children of $sbjct_id) #

	# %destn2char = ($sbjct_id => $pchar); # ($potential_regrafting_destination => $character_after_regrafting, ...);
	# ($add_node2ch, $add_node2minpen, $add_char_chngs) = @{$info_parsim_scenario};
	# %add_node2ch = ($additional_node_id => \@children, ...) giving the partial resolution (including $additonal_node_id = $sbjct_id),
	# %add_node2min_penalty = ($additional_node_id => $min_penalty, ...) giving the minimum summed penalty under $additional_node_id in this particular parsimonious scenario,
	# %add_char_chngs = ($branch_id (including additional branch) => [$char_up, $char_lw, $pen], ...) describing the changes (from $char_up to $char_lw, which incurrs the penalty of $pen) that comprizes the scenario;
	#
	# %chnode2char = ($child_node_id => $character_id, ...) specifies the characters assigned to the child nodes under the input parsimonious scenario;


		my @info_new_parsim_scenario = (\%new_add_node2ch, \%new_add_node2minpen, \%new_add_char_chngs); 
		push @set_info_new_psm_scns, [\@info_new_parsim_scenario, \%new_chnode2char];

	    }		# End of the main middle for-loop (over the conbinations of destinations) #

	}	# End of the main outer for-loop (over the combinations of nodes to be re-grafted) # 


	return \@set_info_new_psm_scns;

}		# End of the subroutine 'add_parsim_scenarios_by_chng_uniq_chars' #


# part_resolve_mlt_node_bu:
#	This satellite subroutine of the subroutine 'min_penalties_sankoff_ptrslv_bo'
#	tries to (partially) resolve a multifurcating node
#	according to Sankoff's parsimony principle applied to a site (or a locus)
#	that potentially undergoes recurrent mutations,
#	while enumerating parsimonious scenarios conditional on each character
#	on the node in question.
#	(CREATION STARTED on Nov 27 (Sun), 2011, ENDED on Dec 24 (Sat), 2011)
#
#	REVISED on Dec 24 (Sat), 2011:
#	The format of the output changed a little,
#	to more or less facilitate the downstream processes.
#
#	NOTE on Dec 29 (Thu), 2011:
#	The algorithm employed by the current version is not perfectly parsimonious,
#	because the algorithm restricts the space of 'better-resolved' trees to be explored as follows:
#	(a) The nodes assigned with 'unique' characters (which is NOT the character at $sbjct_id) are always assumed to descend directly from $sbjct_id;
#	(b) The MRCA of a 'secondary' character is assumed to descend directly
#	either from $sbjct_id or the MRCA of a 'unique' character.
#	These restrictions will prevent all possible parsimonious scenarios from being exhausted
#	when dealing with a site (or locus) with more than 2 characters.
#	Besides, the restriction may sometimes cause a failure to identify truly parsimonious scenarios when the penalty matrix is NOT uniform (i.e. different penalties are assigned to different classes of character changes).
#	[ Nevertheless, the current version will (almost) perfectly work
#	when we handle a site (or locus) consisting of 2 or less characters. ]
#	Extention of the search space to a more general (and yet reasonably simple) one
#	will be achieved by first identifying the 'least-penalized tree', accompanied by the 'back-bone penalty', under a given combination of characters,
#	and it is left as a future task.
#
# INPUT: ($sbjct_id, \$max_nodeid, \@children, \@pen_mtrx, \%node2min_penalties), where
#	$sbjct_id is the node ID of the multifurcated node in question;
#	$max_nodeid is the maximum node ID that has ever be assigned concerning this tree;
#	@children lists the node ID of the child nodes;
#	$pen_mtrx[$i]->[$j] is the penalty imposed on the change
#	from the $i th character to the $j th character;
#	%node2min_penalties = ($node_id => \@min_penalties, ...) for ALL nodes,
#	with $min_penalties[$i] being the minimum summed penalty
#	on the subtree under $node_id under the condition that the node has
#	the $i th character in the matrix @pen_mtrx.
#	($min_penalties[$i] is undef if it is known in advance
#	that the $i th character will NEVER be included in ANY parsimonious scenarios.)
#
#	PREREQUSITE: The penalty from a character to the same character must always be zero.
#
# OUTPUT: (\@cond_min_penalties, \@cond_parsim_scenarios, $minmin_penalty), where
#	$cond_min_penalties[$i] is the minimum penalty along the subtree (after partially resolved) under the multifurcated node in question
#	on condition that the node takes the $i th character;
#	[ @{$cond_parsim_scenarios[$i]->[$s]} = (\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs, \%chnode2char)
#	uniquely specifies the $s th scenario that provides the minimum penalty when the multifurcated node in question takes the $i th character, with ] --- OBSOLETE as of Dec 24 (Sat), 2011
#	@{$cond_parsim_scenarios[$i]->[$cp]} = (\@parsim_scenarios_with_chpattern, \%chnode2char)
#	gives the parsimonious scenarios providing the minimum penalty when the multifurcated node in question takes the $i th character, with
#	@{$parsim_scenarios_with_chpattern[$s]} = (\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs), 
#	uniquely specifying the $s th scenario with the character pattern designated with 
#	%chnod2char = ($child_node_id => $character_indx, ...), which uniquely assigns a character to each child node;
#	Here,
#	%add_node2ch = ($additional_node_id => \@children, ...) giving the partial resolution,
#	%add_node2min_penalty = ($additional_node_id => $min_penalty, ...) giving the minimum summed penalty imposed on the subtree under $additional_node_id in this particular parsiminous scenario;
#	and %add_char_chngs = ($branch_id (including additional branch) => [$char_up, $char_lw, $pen], ...) describing the changes (from $char_up to $char_lw, which incurrs the penalty of $pen) that comprizes the scenario,
#	$minmin_penalty is the minimum penalty among those in @cond_min_penalties.
#
sub part_resolve_mlt_node_bu ($\$\@\@\%) {

	my ($sbjct_id, $max_nodeid, $children, $pen_mtrx, $node2min_penalties) = @_;

	my $ct_chars = @{$pen_mtrx};
	my $ct_children = @{$children};

    (print STDOUT "<< Subroutine 'part_resolve_mlt_node_bu' (#{characters} = $ct_chars\; #{child nodes} = $ct_children) >>\n\n") if ($TEST_LEVEL1>0);

	# %{$node2uniq_char} = ($node_id => [$uniq_char_id, $node2min_penalties->{$node_id}[$uniq_char_id]], ...) is for the nodes with unique characters;
	# %{$uniq_char2nodes} = ($char_id => \@node_ids, ...) for the characters which is a 'uniquely parsimonious' character for some nodes.
	# @{$remain_nodes} lists all the nodes NOT assigned unique characters.
	#
	my ($node2uniq_char, $uniq_char2nodes, $remain_nodes) = assign_uniq_characters (@{$children}, @{$pen_mtrx}, %{$node2min_penalties});
	unless (defined $node2uniq_char) {
	    print STDERR "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'assign_uniq_characters' FAILED : $! !\n";
	    return (undef, undef, undef);
	}
    (print STDOUT "\n --- The subroutine 'assign_uniq_characters' ENDED ...\n\n") if ($TEST_LEVEL1>1);

	my $ct_remnodes = @{$remain_nodes};

    if ($TEST_LEVEL1>2) {
	print STDOUT "\n[ Output of the subroutine 'assign_uniq_characters' ]\n\n";
	print STDOUT "Nodes withOUT their own unique characters (#{such nodes} = $ct_remnodes):\n";
	print STDOUT '@{$remain_nodes}  = (', join (', ', @{$remain_nodes}) , ');', "\n\n";
	print STDOUT '%{$uniq_char2nodes} = (', "\n";
	foreach my $char (sort {$a <=> $b} keys %{$uniq_char2nodes}) {
	    my $nodes = $uniq_char2nodes->{$char};
	    print STDOUT "\t", $char, ' =>  [', join (', ', @{$nodes}), '],', "\n";
	}
	print STDOUT "\t)\n\n";
	if ($TEST_LEVEL1>3) {
	    print STDOUT '%{$node2uniq_char} = ( # node_ID => [$uniq_char_id, $penalty_under_node]', "\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2uniq_char}) { print STDOUT "\t", $node, ' =>  [', join (', ', @{$node2uniq_char->{$node}}), '],', "\n"; }
	    print STDOUT "\t)\n\n";
	}
	print STDOUT "\n";
    }

	    # Calculate the 'base' of the 'base' of the summed penalty,
	    # which is defined by the subtrees under child nodes with 'unique characters'.
	my $basebase_sum_pen = 0;
	my %uniq_char2sum_pen; # = ($unique_char => {summed penalty over subtrees under the nodes with the $unique_char assigned}, ...)
	##foreach my $char (keys %{$uniq_char2basebase_sum_pen}) { # Over 'unique-character's #
	foreach my $char (keys %{$uniq_char2nodes}) { # Over 'unique-character's #
	    my $sum_pen = 0;
	    foreach my $node (@{$uniq_char2nodes->{$char}}) { $sum_pen += $node2uniq_char->{$node}[1]; }
	    $uniq_char2sum_pen{$char} = $sum_pen;
	    $basebase_sum_pen += $sum_pen;
	}
	## foreach my $node (keys %{$node2uniq_char}) { $basebase_sum_pen += $node2uniq_char->{$node}[1]; }

    if ($TEST_LEVEL1>2) {
	print STDOUT "\n[ First set of Auxiliary Variables ]\n\n";
	print STDOUT '%uniq_char2sum_pen = (', "\n";
	foreach my $char (sort keys %uniq_char2sum_pen) { print STDOUT "\t", $char, ' => ', $uniq_char2sum_pen{$char}, "\,\n"; }
	print STDOUT "\t)\n\n";
	print STDOUT '$basebase_sum_pen = ', $basebase_sum_pen, "\n\n";
    }

	my @cond_min_penalties = my @cond_parsim_scenarios = ();
	my $minmin_penalty;

    (print STDOUT "\n[[ Start looping over the $ct_chars parent characters ]]\n\n") if ($TEST_LEVEL1>0);

	for (my $pchar=0; $pchar<$ct_chars; $pchar++) { # Main outer for-loop (over potential characters at the subject node (, which will sometimes be called the 'parent').) #

    (print STDOUT "[[ Parent character = $pchar ]]\n\n") if ($TEST_LEVEL1>0);

	    my @uniq_chars = (); # Set of 'unique character's other than the parent character.
	    my @sec_chars = (); # Set of 'secondary character's.
	    for (my $j=0; $j<$ct_chars; $j++) { # preliminary middle for-loop (over child characters) #
		($j == $pchar) and next;
		if (defined $uniq_char2nodes->{$j}) {
		    push @uniq_chars, $j;
		} else {
		    push @sec_chars, $j;
		}
	    }	# End of the preliminary middle for-loop (over child characters) #

	    my $ct_uniq = @uniq_chars;
	    my $ct_sec = @sec_chars;
	    my @indices_secchars = (0 .. $ct_sec-1); # The set of ALL indices for @sec_chars.

    if ($TEST_LEVEL1>2) {
	print STDOUT "Unique characters (other than $pchar) (# = $ct_uniq):\n";
	print STDOUT '@uniq_chars = (', join (', ', @uniq_chars), ')', " .\n";
	print STDOUT "Secondary characters (# = $ct_sec):\n";
	print STDOUT '@sec_chars = (',  join (', ', @sec_chars), ')', " .\n\n";
    }

	    my $row_pen_fm_pchar = $pen_mtrx->[$pchar]; # Penalties from the parent character.

		# First calculate the 'base' of the summed penalty,
		# which is defined by the subtrees under child nodes with 'unique characters' (already calculated),
		# as well as the penalties from the parent character to the 'unique characters'.
		# (Because we search for parsimonious scenarios, we assume that a change from the parent to a 'unique character' occurred ONLY ONCE.)
		# (THERE IS A ROOM FOR IMPROVEMENT, namely, we could consider more complex scenarios by allowing for changes from 'uqique character's to other 'unique character's !!)
	    my $base_sum_pen = $basebase_sum_pen;
	    foreach my $char (@uniq_chars) { $base_sum_pen += $row_pen_fm_pchar->[$char]; }

    if ($TEST_LEVEL1>2) {
	print STDOUT '$base_sum_pen (excluding ', "'remaining nodes'", ') =  ', $base_sum_pen , " .\n\n";
    }
	    
	    # $minpen_prim_to_sec->[$c] gives the minimum penalty for a change from a primary character (in ($pchar, @uniq_chars)) to $sec_chars[$c];
	    # @{$best_prims_for_sec[$c]} lists the indices of the primary characters that give the minimum penalty to $sec_chars[$c].
	    #
		# REVISED to simplify the workflow (on Dec 27 (Tue), 2011) #
	    my ($minpen_prim_to_sec, $best_prims_for_sec)
		= search_minpen_prim_to_sec ($pchar, @uniq_chars, @sec_chars, @{$pen_mtrx}) ;
	    ## my ($minpen_prim_to_sec, $best_prims_for_sec) = ($ct_remnodes>0) ?
	    ##	search_minpen_prim_to_sec ($pchar, @uniq_chars, @sec_chars, @{$pen_mtrx}) : ([], []);
	    unless (defined $minpen_prim_to_sec) {
		print STDERR "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'search_minpen_prim_to_sec' FAILED on the parent character '$pchar' : $! !\n";
		return (undef, undef, undef);
	    }
    (print STDOUT "\n --- The subroutine 'search_minpen_prim_to_sec' ENDED ...\n\n") if ($TEST_LEVEL1>1);

    if ($TEST_LEVEL1>2) {
	print STDOUT '[ Output of ', "'search_minpen_prim_to_sec'", ' ]', "\n\n";
	print STDOUT 'Contents of @{$minpen_prim_to_sec} and @{$best_prims_for_sec}', "\n";
	print STDOUT join ("\t", 'Indx_sec_char', 'ID_sec_char', 'minpen(prim->sec)', 'ID,best,prim,chars'), "\n\n";
	for (my $c=0; $c<$ct_sec; $c++) {
	    my $best_prims = $best_prims_for_sec->[$c];
	    my $cnct_best_prims = join (',', @{$best_prims});
	    print STDOUT join ("\t", $c, $sec_chars[$c], $minpen_prim_to_sec->[$c], $cnct_best_prims), "\n";
	}
	print STDOUT "\n\n";
    }

=REPLACED with a more simplified series of processes (on Nov 28 (Mon)) together with disabling the 'new function' of 'min_penalties_sankoff_bo'.
	    # @{$prim_nodes} lits the IDs of the 'remaining nodes' for which some of the primary nodes may still give the mimimum penalty,
	    # $minpen_amng_prim_for_pnodes->[$p] is the minimum penalty for the subtree under the node with ID = $prim_nodes->[$p] among those given by primary characters at the node; and
	    # @{$minpen_prim_chars_for_pnodes->[$p]} lists all such primary characters giving the minimum penalty;
	    # @{$sec_nodes} list the IDs of the 'secondary nodes'  for which NO primary nodes can give the minimum penalty.
	    #
	    my ($prim_nodes, $minpen_amng_prim_for_pnodes, $minpen_prim_chars_for_pnodes, $sec_nodes)
		= minpen_among_specchars_for_sbnodes ($pchar, @uniq_chars, @{$remain_nodes}, %{$node2min_penalties});
	    unless (defined $minpen_amng_prim_for_pnodes) {
		print STDERR "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'minpen_among_specchars_for_sbnodes' FAILED : $! !\n";
		return (undef, undef, undef);
	    }

	    # $pens_secchars_for_sec->[$sc][$sn] is the penalty when the node with ID $sec_nodes->[$sn] has the $sec_chars[$sc] th character;
	    # $totpens_secchars_for_sec->[$sc] is the total penalty over all the secondary nodes that could possibly have $sec_chars[$sc] th character give the minimum penalty.
	    #
	    my @dummy = ();
	    my ($pens_secchars_for_sec, $totpens_secchars_for_sec)
		= diff_penalties_specchars_for_sbnodes (@sec_chars, @{$sec_nodes}, %{$node2min_penalthes}, @dummy);
	    unless (defined $pens_secchars_for_sec) {
		print STDERR "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'diff_penalties_specchars_for_sbnodes' FAILED on the secondary child nodes for the node '$node_id': $! !\n";
		return (undef, undef, undef);
	    }

	    # $diffpens_secchars_for_prim->[$sc][$pn]is the differential penalty when the node with ID $prim_nodes->[$pn] has the $sec_chars[$sc] character, subtracted by $minpen_amng_prim_for_pnodes->[$pn];
	    # $tot_diffpens_secchars_for_prim is the sum of the differntial penalties over all the primary nodes $prim_nodes->[$pn] that have $sec_chars[$sc] th character give the penalty smaller than or equal to $minpen_amng_prim_for_pnodes->[$pn], subtracted by $minpen_amng_prim_for_pnodes->[$pn] per each primary node.
	    #
	    my ($diffpens_secchars_for_prim, $tot_diffpens_secchars_for_prim)
		= diff_penalties_specchars_for_sbnodes (@sec_chars, @{$prim_nodes}, %{$node2min_penalties}, @{$minpen_amng_prim_for_pnodes});
	    unless (defined $diffpens_secchars_for_prim) {
		print STDERR "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'diff_penalties_specchars_for_sbnodes' FAILED on the primary child nodes for the node '$node_id': $! !\n"; 
		return (undef, undef, undef);
	    }
=cut

	    # @{$remain_nodes2} SHOULD NOW EQUAL to @{$remain_nodes};
	    # $minpen_amng_prim_for_rnodes->[$r] is the minimum penalty for the subtree under the node with ID = $remain_nodes->[$r] among those given by primary characters at the node; and
	    # @{$minpen_prim_chars_for_rnodes->[$r]} lists all such primary characters giving the minimum penalty;
	    # @{$sec_nodes} SHOULD NOW BE EMPTY!!
	    #
	    my ($remain_nodes2, $minpen_amng_prim_for_rnodes, $minpen_prim_chars_for_rnodes, $sec_nodes)
		= minpen_among_specchars_for_sbnodes ($pchar, @uniq_chars, @{$remain_nodes}, %{$node2min_penalties});
	    unless (defined $minpen_amng_prim_for_rnodes) {
		print STDERR "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'minpen_among_specchars_for_sbnodes' FAILED with the $pchar th 'parent' character : $! !\n";
		return (undef, undef, undef);
	    }
	    unless (@{$sec_nodes} == 0) {
		print STDERR "[ERROR 'part_resolve_mlt_node_bu'] The output \@\{\$sec_nodes\} = (", join (', ', @{$sec_nodes}), "), which should have been empty, is NOT EMPTY for the node '$sbjct_id', with the $pchar th 'parent' character : $! !\n\n";
		return (undef, undef, undef);
	    }
    (print STDOUT "\n --- The subroutine 'minpen_among_specchars_for_sbnodes' ENDED ...\n\n") if ($TEST_LEVEL1>1);
    if ($TEST_LEVEL1>2) {
	print STDOUT "[ Output of 'minpen_among_specchars_for_sbnodes' ]\n\n";
	print STDOUT 'Contents of @{$minpen_amng_prim_for_rnodes} and @{$minpen_prim_chars_for_rnodes}:', "\n\n";
	print STDOUT join ("\t", 'Indx_rnode', 'ID_rnode', 'min_pen_prim', 'ID,best,prim,chars'), "\n\n";
	for (my $r=0; $r<$ct_remnodes; $r++) {
	    my $nodeid = $remain_nodes->[$r];
	    my $minpen = $minpen_amng_prim_for_rnodes->[$r];
	    my $best_prims = $minpen_prim_chars_for_rnodes->[$r];
	    my $cnct_best_prims = join (',', @{$best_prims});
	    print STDOUT join ("\t", $r, $nodeid, $minpen, $cnct_best_prims), "\n";
	}
	print STDOUT "\n\n";
    }

	    foreach my $pen (@{$minpen_amng_prim_for_rnodes}) { $base_sum_pen += $pen; }

    (print STDOUT "\n", '$base_sum_pen (INCLUDING the ', "'remaining'" , ' nodes) = ', $base_sum_pen, "\n\n") if ($TEST_LEVEL1>1);

	    # $diffpens_secchars_for_rem->[$sc][$r]is the differential penalty when the node with ID $remain_nodes->[$r] has the $sec_chars[$sc] character, subtracted by $minpen_amng_prim_for_pnodes->[$r];
	    # $tot_diffpens_secchars_for_rem->[$sc] is the sum of the differntial penalties over all the remaining nodes $remain_nodes->[$r] that have $sec_chars[$sc] th character giving the penalty smaller than or equal to $minpen_amng_prim_for_rnodes->[$r], subtracted by $minpen_amng_prim_for_rnodes->[$r] per each remaining node.
	    #
	    my ($diffpens_secchars_for_rem, $tot_diffpens_secchars_for_rem)
		= diff_penalties_specchars_for_sbnodes (@sec_chars, @{$remain_nodes}, %{$node2min_penalties}, @{$minpen_amng_prim_for_rnodes});
	    unless (defined $diffpens_secchars_for_rem) {
		print STDERR "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'diff_penalties_specchars_for_sbnodes' FAILED on the 'remaining' child nodes for the node '$sbjct_id' on the $pchar th 'parent' character: $! !\n"; 
		return (undef, undef, undef);
	    }
    (print STDOUT "\n --- The subroutine 'diff_penalties_specchars_for_sbnodes' ENDED...\n\n") if ($TEST_LEVEL1>1);
    if ($TEST_LEVEL1>2) {
	print STDOUT "[ Output of 'diff_penalties_specchars_for_sbnodes' ]\n\n";
	print STDOUT 'Content of @{$diffpens_penalties_specchars_for_sbnodes} and @{$tot_diffpens_secchars_for_rem}:', "\n\n";
	print STDOUT join ("\t", 'Indx_sec_char', 'ID_sec_char', 'Tot_diffpen');
	foreach my $nid (@{$remain_nodes}) { print STDOUT "\tdiffpen(node:$nid)"; }
	print STDOUT "\n\n";

	for (my $sc=0; $sc<$ct_sec; $sc++) {
	    my $id_sec = $sec_chars[$sc];
	    my $tot_diffpen = $tot_diffpens_secchars_for_rem->[$sc];
	    ##(defined $tot_diffpen) or ($tot_diffpen = 'undef'); # ADDED on Dec 27 (Tue), 2011 #
	    my @diffpens = ();
	    for (my $r=0; $r<$ct_remnodes; $r++) {
		my $diffpen = $diffpens_secchars_for_rem->[$sc][$r];
		push @diffpens, ((defined $diffpen) ? $diffpen : 'undef');
	    }
	    print STDOUT join ("\t", $sc, $id_sec, $tot_diffpen, @diffpens), "\n";
	}
	print STDOUT "\n\n";
    }

	    # $sec_chars[$improving_secs->[$i]] ($i=0,...,$#{$improving_secs}) gives the secondary characters that are bound to improve the total penalty over primary characters;
	    # Secondary characters in @{$improving_secs} are sorted in order of decreasing total penalty drop;
	    # $sec_chars[$keeping_secs->[$i]] ($i=0,...,$#{$keeping_secs}) gives the secondary character that can tie the primary characters at the best configuration;
	    # $sec_chars[$no_chance->[$i]] ($i=0,...,$#{$no_chance}) gives the secondary character that canNOT even tie the maximum-drop penalty;
	    # $max_droppen_by_sec_over_rem->[$sc] gives the maximum possible drop in the penalty when the character $sec_chars[$sc] is used for some of the 'remaining nodes'.
	    #
	    my ($improving_secs, $keeping_secs, $no_chance, $max_droppen_by_sec_over_rem)
		= sort_secondary_characters (@{$tot_diffpens_secchars_for_rem}, @{$minpen_prim_to_sec});
	    unless (defined $improving_secs) {
		print STDOUT "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'sort_secondary_characters' FAILED on the secondary characters (", join (', ', @sec_chars), ") for the node '$sbjct_id' when it assumes the $pchar th character: $! !\n";
		return (undef, undef, undef);
	    }
    (print STDOUT "\n --- The subroutine 'sort_secondary_characters' ENDED ...\n\n") if ($TEST_LEVEL1>1);
    if ($TEST_LEVEL1>2) {
	print STDOUT "[ Output of 'sort_secondary_characters' ]\n\n";
	print STDOUT 'Content of @{$max_droppen_by_sec_over_rem}:', "\n\n";
	print STDOUT join ("\t", 'Indx_sec_char', 'ID_sec_char', 'Max_droppen_by_sec'), "\n";
	for (my $sc=0; $sc<$ct_sec; $sc++) {
	    my $id_sec = $sec_chars[$sc];
	    my $max_droppen = $max_droppen_by_sec_over_rem->[$sc];
	    print STDOUT join ("\t", $sc, $id_sec, $max_droppen), "\n";
	}
	print STDOUT "\n\n{Indices of the 'improving' secondary characters}:\n";
	print STDOUT '@{$improving_secs} = (',  join (', ', @{$improving_secs}), "),\n\n";
	print STDOUT "{Indices of the 'keeping' secondary characters}:\n";
	print STDOUT '@{$keeping_secs} = (', join (', ', @{$keeping_secs}), "),\n\n";
	print STDOUT "{Indices of the 'no-chance' secondary characters}:\n";
	print STDOUT '@{$no_chance} = (', join (', ', @{$no_chance}), ") .\n\n\n";
    }

	    # @{$key_marginal_nodes_for_sec->[$sc]} = (\@key_nodes, \@marginal)
	    # for the secondary character $sec_chars[$sc], with
	    # $remain_nodes->[$key_nodes[$kn]] being the ID of the $kn th remaining node
	    # where the secondary character has a lower penalty than any primary characters, and with
	    # $remain_nodes->[$marginal[$mg]] beging the ID of the $mg th remaining node
	    # where the secondary character ties the lowest-penalty primary node.
	    #
	    my $key_marginal_nodes_for_sec = sort_remaining_nodes_per_secchars (@indices_secchars, @{$diffpens_secchars_for_rem});
	    unless (defined $key_marginal_nodes_for_sec) {
		print STDOUT "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'sort_remaining_nodes_per_secchars' FAILED on the secondary characters with indices (", join (', ', @indices_secchars), "), on the $pchar th 'parent' character : $! !\n";
		return (undef, undef, undef);
	    }
    (print STDOUT "\n --- The subroutine 'sort_remaining_nodes_per_secchars' ENDED...\n\n") if ($TEST_LEVEL1>1);
    if ($TEST_LEVEL1>2) {
	print STDOUT "[ Output of 'sort_remaining_nodes_per_secchars' ]\n\n";
	print STDOUT 'Content of @{$key_marginal_nodes_for_sec}:', "\n\n";
	print STDOUT join ("\t", 'Indx_sec_char', 'ID_sec_char', 'ID,key,remaining,nodes', 'ID,marginal,remaining,nodes'), "\n\n";
	for (my $sc=0; $sc<$ct_sec; $sc++) {
	    my $id_sec = $sec_chars[$sc];
	    my ($keys, $marginals) = @{$key_marginal_nodes_for_sec->[$sc]};
	    my @keys = my @marginals = ();
	    foreach my $rindx (@{$keys}) { push @keys, $remain_nodes->[$rindx]; }
	    foreach my $rindx (@{$marginals}) { push @marginals, $remain_nodes->[$rindx]; }
	    my $cnct_keys = join (',', @keys);
	    my $cnct_mgnls = join (',', @marginals);
	    print STDOUT join ("\t", $sc, $id_sec, $cnct_keys, $cnct_mgnls), "\n";
	}
	print STDOUT "\n\n";
    }

		## The Final Main Process: Enumerate the combinations of secondary characters
		## that provide the maximum drop in the penalty (compared to the best combination(s) of primary characters).

	    # @{$core_maxdroppen_combi_secchars->[$cm]} = (\@imprv_secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $max_droppen_for_combi)
	    # gives the $cm th combination of 'improving' secondary characters on the 'remaining nodes' that gives the maximum drop in the penalty (compared with the 'best' combination of primary characters), with
	    # @imprv_secchar_indices listing indices in @{$improving_secs} for the 'improving' secondary characters that make up the combination,
	    # @{$secchars_for_rnode[$r]} listing indices of characters in @sec_chars that give the maximum penalty drop for the $remain_nodes->[$r] th node (= undef if no secondary characters satisfy the condition), with
	    # $droppen_for_rnode[$r]  being the maximum penalty drop achieved for the $remain_nodes->[$r] th node by the secondary characters in the combination ( = undef if the penalty at the node does NOT improve over the best primary penalty), and
	    # $max_droppen_for_combi being the maximum penalty drop achieved by the combination given.
	    #
	    my $core_maxdroppen_combi_secchars
		= enumerate_core_maxdroppen_combi_secchars (@{$improving_secs}, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem}, @{$minpen_prim_to_sec}, @{$max_droppen_by_sec_over_rem});
	    unless (defined $core_maxdroppen_combi_secchars) {
		print STDOUT "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'enumerate_core_maxdroppen_combi_secchars' FAILED on the improving secondary characters with indices (", join (', ', @{$improving_secs}), "), on the $pchar th 'parent character' : $! !\n";
		return (undef, undef, undef);
	    }

    (print STDOUT "\n --- The subroutine 'enumerate_core_maxdroppen_combi_secchars' ENDED ...\n\n") if ($TEST_LEVEL1>1);
    if ($TEST_LEVEL1>2) {
	my $ct_mxdpen_cmbs = @{$core_maxdroppen_combi_secchars};
	print STDOUT "[ Output of 'enumerate_core_maxdroppen_combi_secchars' ]\n\n";
	print STDOUT 'Content of @{$core_maxdroppen_combi_secchars}', " (with $ct_mxdpen_cmbs elements):\n\n";
	for (my $mc=0; $mc<$ct_mxdpen_cmbs; $mc++) {
	    my $mxdpen_combi = $core_maxdroppen_combi_secchars->[$mc];
	    my ($imprv_sc_indices, $scchars_for_rnode, $droppen_for_rnode, $mxdpen_for_combi) = @{$mxdpen_combi};
	    my @imprv_sc_ids = ();
	    foreach my $indx (@{$imprv_sc_indices}) { push @imprv_sc_ids, $improving_secs->[$indx]; }
	    print STDOUT "{ The $mc th combination (maximum penalty-drop: $mxdpen_for_combi) }\n\n";
	    print STDOUT "Constituent indices in 'improving' secs:  ", join (', ', @{$imprv_sc_indices}), "\n";
	    print STDOUT "Constituent indices in secondary characters:  ", join (', ', @imprv_sc_ids), "\n\n";
	    print STDOUT '#', join ("\t", 'Indx_rnode', 'ID_rnode', 'penalty-drop', 'Indices,best,sec,chars'), "\n\n";
	    for (my $r=0; $r<$ct_remnodes; $r++) {
		my $id_rnode = $remain_nodes->[$r];
		my $dpen = $droppen_for_rnode->[$r];
		(defined $dpen) or ($dpen = 'undef');
		my $scchrs = $scchars_for_rnode->[$r];
		my $cnct_scchrs = (defined $scchrs) ? join (',', @{$scchrs}) : 'undef';
		print STDOUT join ("\t", $r, $id_rnode, $dpen, $cnct_scchrs), "\n";
	    }
	    print STDOUT "\n\n";
	}
    }

			# Now, the conditional minimum penalty can be calculated. #

	    my $minpen_under_pchar = $base_sum_pen; # Minimum penalty when the parent node assumes the character $pchar
	    if (@{$core_maxdroppen_combi_secchars}>0) {
			# Minimum penalty should be smaller than
			# the smallest penalty achieved by the primary characters alone.
		$minpen_under_pchar += $core_maxdroppen_combi_secchars->[0][3];
	    }

    (print STDOUT "\n{Minimum penalty for the subtree under the subject ID assuming the character '$pchar'} = $minpen_under_pchar .\n\n") if ($TEST_LEVEL1>1);

	    $cond_min_penalties[$pchar] = $minpen_under_pchar;
	    ((defined $minmin_penalty) and ($minmin_penalty <= $minpen_under_pchar)) or ($minmin_penalty = $minpen_under_pchar);


			# In a couple of processeds from here,
			# attempt to expand the set of secondary character combinations
			# that can achieve the conditional minimum penalty.

	    # @{$dressed_core_maxdroppen_combi_secchars} lists all possible maximum-penalty-dropping combinations of secondary characters, made from a 'core' combination 'dressed' with the improving secondary characters that do NOT worsen the penalty drop of the 'core' combination.
	    # Its elements follow the same format as the elements of @{$core_maxdroppen_combi_secchars}.
	    #
	    my $drssd_core_maxdroppen_combi_secchars
		= dress_core_maxdroppen_combi_secchars (@{$core_maxdroppen_combi_secchars},
		   @{$improving_secs}, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem}, @{$minpen_prim_to_sec});
##); ## DISABLED because it was judged as a 'syntax error', while I am not sure if it is necessary or not.
	    unless (defined $drssd_core_maxdroppen_combi_secchars) {
		print STDOUT "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'dress_core_maxdroppen_combi_secchars' FAILED on the improving secondary characters with indices (", join (', ', @{$improving_secs}), "), on the $pchar th 'parent character' : $! !\n";
		return (undef, undef, undef);
	    }
    (print STDOUT "\n --- The subroutine 'dress_core_maxdroppen_combi_secchars' ENDED ...\n\n") if ($TEST_LEVEL1>1);

	    my $ct_dressed_cores = @{$drssd_core_maxdroppen_combi_secchars};

    if ($TEST_LEVEL1>2) {
	print STDOUT "[ Output of 'dress_core_maxdroppen_combi_secchars' ]\n\n";
	print STDOUT 'Content of @{$drssd_core_maxdroppen_combi_secchars}', " (with $ct_dressed_cores elements):\n\n";
	for (my $mc=0; $mc<$ct_dressed_cores; $mc++) {
	    my $mxdpen_combi = $drssd_core_maxdroppen_combi_secchars->[$mc];
	    my ($imprv_sc_indices, $scchars_for_rnode, $droppen_for_rnode, $mxdpen_for_combi) = @{$mxdpen_combi};
	    my @imprv_sc_ids = ();
	    foreach my $indx (@{$imprv_sc_indices}) { push @imprv_sc_ids, $improving_secs->[$indx]; }
	    print STDOUT "{ The $mc th combination (maximum penalty-drop: $mxdpen_for_combi) }\n\n";
	    print STDOUT "Constituent indices in 'improving' secs:  ", join (', ', @{$imprv_sc_indices}), "\n";
	    print STDOUT "Constituent indices in secondary characters:  ", join (', ', @imprv_sc_ids), "\n\n";
	    print STDOUT '#', join ("\t", 'Indx_rnode', 'ID_rnode', 'penalty-drop', 'Indices,best,sec,chars'), "\n\n";
	    for (my $r=0; $r<$ct_remnodes; $r++) {
		my $id_rnode = $remain_nodes->[$r];
		my $dpen = $droppen_for_rnode->[$r];
		(defined $dpen) or ($dpen = 'undef');
		my $scchrs = $scchars_for_rnode->[$r];
		my $cnct_scchrs = (defined $scchrs) ? join (',', @{$scchrs}) : 'undef';
		print STDOUT join ("\t", $r, $id_rnode, $dpen, $cnct_scchrs), "\n";
	    }
	    print STDOUT "\n\n";
	}
    }

		# 'Translate' the indices in @{$improving_secs} of each secondary character combination into the indices in @sec_chars

	    for (my $cs=0; $cs<$ct_dressed_cores; $cs++) { # piping for-loop #
		my $dressed_core_combi = $drssd_core_maxdroppen_combi_secchars->[$cs];
		my $imprv_secchar_indices = $dressed_core_combi->[0];
		my @secchar_indices = ();
		foreach my $imprv_indx (@{$imprv_secchar_indices}) { push @secchar_indices, $improving_secs->[$imprv_indx]; }
		$dressed_core_combi->[0] = \@secchar_indices;
	    }				# End of the piping for-loop #

	    if ($ct_dressed_cores == 0) { # This can happen if @{$improving_secs} is empty. #
		# Make a 'dummy' dressed core set,
		# whose only element consists of NO improving character. #

	    	    # @{$core_maxdroppen_combi_secchars->[$cm]}
		    # = (\@imprv_secchar_indices, \@secchars_for_rnode, \@droppen_for_rnode, $max_droppen_for_combi).
		my @dummy_secchars_for_rnode = my @dummy_droppen_for_rnode = ();

		if ($ct_remnodes>0) { # ADDED to cope with the case with $ct_remnodes = 0. (on Dec 27 (Tue), 2011) #
		    $dummy_secchars_for_rnode[$ct_remnodes -1] = undef;
		    $dummy_droppen_for_rnode[$ct_remnodes -1] = undef;
		}	# ADDED to cope with the case with $ct_remonodes = 0. (on Dec 27 (Tue), 2011) #

		my @dummy_combi = ([], \@dummy_secchars_for_rnode, \@dummy_droppen_for_rnode, 0);
		$drssd_core_maxdroppen_combi_secchars = [\@dummy_combi];
	    }

		# Make a full set of maximum penalty-dropping combinations of secondary characters,
		# by incorporating the 'keeping' secondary characters to the 'dressed core' set.

	    # @{$fullset_maxdroppen_combi_secchars} lists literally all possible combinations of secondary characters that provides the maximum drop in penalty.
	    # The format of the elements is almost the same as that for @{$(drssed_)core_maxdroppen_combi_secchars}.
	    # The only difference is that the zeroth element now lists the indices in @sec_chars, instead of those in @{$improving_secs}.
	    #
	    my $fullset_maxdroppen_combi_secchars
		= attach_keeping_secs_to_maxdroppen_combi (@{$drssd_core_maxdroppen_combi_secchars}, @{$keeping_secs}, @{$key_marginal_nodes_for_sec}, @{$diffpens_secchars_for_rem}, @{$minpen_prim_to_sec});
	    unless (defined $fullset_maxdroppen_combi_secchars) {
		print STDOUT "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'attach_keeping_secs_to_maxdroppen_combi' FAILED on the 'keeping' secondary characters with indices (", join (', ', @{$keeping_secs}), ") with the $pchar th 'parent character': $! !\n";
		return (undef, undef, undef);
	    }
    (print STDOUT "\n --- The subroutine 'attach_keeping_secs_to_maxdroppen_combi' ENDED ...\n\n") if ($TEST_LEVEL1>1);

	    my $ct_mdpen_combi = @{$fullset_maxdroppen_combi_secchars};

    if ($TEST_LEVEL1>2) {
	my $ct_mdpen_combi = @{$fullset_maxdroppen_combi_secchars};
	print STDOUT "[ Output of 'attach_keeping_secs_to_maxdroppen_combi' ]\n\n";
	print STDOUT 'Content of @{$fullset_maxdroppen_combi_secchars}', " (with $ct_mdpen_combi elements):\n\n";
	for (my $mc=0; $mc<$ct_mdpen_combi; $mc++) {
	    my $mxdpen_combi = $fullset_maxdroppen_combi_secchars->[$mc];
	    my ($imprv_sc_indices, $scchars_for_rnode, $droppen_for_rnode, $mxdpen_for_combi) = @{$mxdpen_combi};
	    my @imprv_sc_ids = @{$imprv_sc_indices};
	    ##foreach my $indx (@{$imprv_sc_indices}) { push @imprv_sc_ids, $improving_secs->[$indx]; }
	    print STDOUT "{ The $mc th combination (maximum penalty-drop: $mxdpen_for_combi) }\n\n";
	    ## print STDOUT "Constituent indices in 'improving' secs:  ", join (', ', @{$imprv_sc_indices}), "\n";
	    print STDOUT "Constituent indices in secondary characters:  ", join (', ', @imprv_sc_ids), "\n\n";
	    print STDOUT '#', join ("\t", 'Indx_rnode', 'ID_rnode', 'penalty-drop', 'Indices,best,sec,chars'), "\n\n";
	    for (my $r=0; $r<$ct_remnodes; $r++) {
		my $id_rnode = $remain_nodes->[$r];
		my $dpen = $droppen_for_rnode->[$r];
		(defined $dpen) or ($dpen = 'undef');
		my $scchrs = $scchars_for_rnode->[$r];
		my $cnct_scchrs = (defined $scchrs) ? join (',', @{$scchrs}) : 'undef';
		print STDOUT join ("\t", $r, $id_rnode, $dpen, $cnct_scchrs), "\n";
	    }
	    print STDOUT "\n\n";
	}
    }


		# Flesh out each maximum penalty-dropping combination of secondary characters
		# by filling each yet empty 'remaining' node with either a primary character
		# or a secondary character for which the node is 'marginal'.

	    # @char_pattern = split //, ${$setset_maxdroppen_char_patterns->[$md][$p]}
	    # is the $p th character pattern made from the combination $fullset_maxdroppen_combi_secchars->[$md], with
	    # $char_pattern[$r] is the index, for the  whole set of characters, of the character at the $r th remaining node.
	    #
	    my $setset_maxdroppen_char_patterns = enumerate_maxdroppen_char_patterns (@{$fullset_maxdroppen_combi_secchars}, @{$minpen_prim_chars_for_rnodes}, @sec_chars, @{$key_marginal_nodes_for_sec});
	    unless (defined $setset_maxdroppen_char_patterns) {
		print STDOUT "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'enumerate_maxdroppen_char_patterns' FAILED on the $pchar th 'parent' character: $! !\n";
		return (undef, undef, undef);
	    }
    (print STDOUT "\n --- The subroutine 'enumerate_maxdroppen_char_patterns' ENDED ...\n\n") if ($TEST_LEVEL1>1);
    if ($TEST_LEVEL1>2) {
	print STDOUT "[ Output of 'enumerate_maxdroppen_char_patterns' ]\n\n";
	print STDOUT 'Content of @{$setset_maxdroppen_char_patterns}:', "\n\n";
	for (my $md=0; $md<$ct_mdpen_combi; $md++) {
	    my $set_mxdpen_char_patterns = $setset_maxdroppen_char_patterns->[$md];
	    my $ct_patterns_fm_combi = @{$set_mxdpen_char_patterns};
	    my $sc_indices = $fullset_maxdroppen_combi_secchars->[$md][0];
	    print STDOUT "{ The $md th secondary combination (with the secondary character indices: ", join (', ', @{$sc_indices}), ") created $ct_patterns_fm_combi character patterns }\n\n";
	    print STDOUT join ("\t", 'Indx_pattern', "'pattern'"), "\n\n";
	    for  (my $p=0; $p<$ct_patterns_fm_combi; $p++) {
		my $char_pattern = ${$set_mxdpen_char_patterns->[$p]};
		print STDOUT join ("\t", $p, "'$char_pattern'"), "\n"; 
	    }
	    print STDOUT "\n\n";
	}
	print STDOUT "\n";
    }

	    my %sec2best_prims; # = ($id_secchar => \@ids_best_prims, ...)
	    for (my $sc=0; $sc<$ct_sec; $sc++) { # Preliminary middle foreach-loop #
		my $id_secchar = $sec_chars[$sc];
		my $best_prims = $best_prims_for_sec->[$sc];
		$sec2best_prims{$id_secchar} = $best_prims;
	    }		# End of the preliminary middle foreach-loop #


    if ($TEST_LEVEL1>2) {
	print STDOUT "[ Interim Auxiliary Variable ]\n\n";
	print STDOUT '%sec2best_prims = ( # $id_secchar => \@ids_best_prim_chars, ', "\n";
	foreach my $id_sc (sort {$a <=> $b} keys %sec2best_prims) { print STDOUT "\t", $id_sc, ' =>  [', join (',', @{$sec2best_prims{$id_sc}}), '],', "\n"; }
	print STDOUT "\t", ').', "\n\n\n";
    }

    (print STDOUT "{{{ Final Main Middle-Loop }}}\n\n") if ($TEST_LEVEL1>1);


			## Prepare for the final main process. ##
			## (ADDED on Dec 25 (Sun), 2011) ##

		# %sgl_node2uniq_char = ($node_id => $character, ...) for the (child) nodes each of which is the only node with the unique character $character (excluding $pchar).
		# This is used as an input for 'add_parsim_scenarios_by_chng_uniq_chars'.
		#
	    my %sgl_node2uniq_char;
	    foreach my $unq_char (keys %{$uniq_char2nodes}) { # Auxiliary foreach-loop (over unique characters) #
		($unq_char == $pchar) and next;
		my $nodes = $uniq_char2nodes->{$unq_char};
		(@{$nodes} == 1) or next;
		my $sgl_node = $nodes->[0];
		$sgl_node2uniq_char{$sgl_node} = $unq_char;
	    }			# End of the auxiliary foreach-loop (over unique characters) #

			## THe FINAL MAIN PROCESS (middle for-loop) ##

	    my @parsim_scenarios_curr_cond = (); # lists the parsimonious scenarios under the current condition. 

	    # %add_chptn2psm_scns = ($string_character_pattern => [\@set_parsim_scenarios_fm_pattern, \%chnode2char], ...)
	    # with $string_character_pattern being the concatenation of the characters on the child nodes (in order appearing in @chnodes),
	    # and with @set_parsim_scenarios_fm_pattern and %chnode2char
	    # following the same formats as the outputs of 'make_parsim_scenarios_fm_pattern' do.
	    # (This will rearrange the output of 'add_parsim_scenarios_by_chng_uniq_chars'.)
	    #
	    my %add_chptn2psm_scns;
	    my $totct_add_psm_scns = 0;

	    for (my $md=0; $md<$ct_mdpen_combi; $md++) { # Main middle for-loop (over maximum penalty-dropping character combinations) #
		my $mdppen_combi_secchars = $fullset_maxdroppen_combi_secchars->[$md];
		my $set_mdpen_char_patterns = $setset_maxdroppen_char_patterns->[$md];
		my $ct_char_patterns = @{$set_mdpen_char_patterns};

    (print STDOUT "{ The $md th combination has $ct_char_patterns character patterns }\n\n") if ($TEST_LEVEL1>2);

		for (my $p=0; $p<$ct_char_patterns; $p++) { # Main inner for-loop (over maximum penalty-dropping character patterns) #
		    my @char_pattern = split //, ${$set_mdpen_char_patterns->[$p]};

    if ($TEST_LEVEL1>2) {
	print STDOUT "{ The $p th pattern from the $md th combination }\n\n";
	print STDOUT "Pattern: ", join ('', @char_pattern), "\n\n";
    }


			# Slightly MODIFIED on Dec 24 (Sat), 2011 #

		    # [ @{$set_parsim_scenarios_w_pattern->[$s]} = (\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs, \%chnode2char) ] --- OBSOLETE as of Dec 24 (Sat), 2011
		    # @{$set_parsim_scenarios_fm_pattern->[$s]} = (\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs)
		    # specifies the $s th parsimonious scenario with the character pattern @char_pattern, with
		    # %add_node2ch = ($additional_node_id => \@children, ...) giving the partial resolution (including $additonal_node_id = $sbjct_id);
		    # %add_node2min_penalty = ($additional_node_id => $min_penalty, ...) giving the minimum summed penalty imposed on the subtree under $additional_node_id in this particular parsimonious scenario.
		    # %add_char_chngs = ($branch_id (including additional branch) => [$char_up, $char_lw, $pen], ...) describing the changes (from $char_up to $char_lw, which incurrs the penalty of $pen) that comprizes the scenario;
		    #
		    # %chnode2char = ($child_node_id => $character_id, ...) uniquely assigns the characters to the child nodes.
		    #
		    ##my $set_parsim_scenarios_fm_pattern
		    my ($set_parsim_scenarios_fm_pattern, $chnode2char)
			= make_parsim_scenarios_fm_pattern (@char_pattern, @{$remain_nodes}, %{$uniq_char2nodes}, $pchar, $sbjct_id, ${$max_nodeid}, %sec2best_prims, @{$pen_mtrx}, %{$node2min_penalties});
		    unless (defined $set_parsim_scenarios_fm_pattern) {
			print STDOUT "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'make_parsim_scenarios_fm_pattern' FAILED on the $p th scenario for the $md th maximum penalty-dropping character pattern on the $pchar th 'parent' character: $! !\n";
			return (undef, undef, undef);
		    }
    (print STDOUT "\n --- The subroutine 'make_parsim_scenarios_fm_pattern' ENDED for the $p th character pattern from the $md th combination ...\n\n") if ($TEST_LEVEL1>1);

		    my $ct_psm_scenarios = @{$set_parsim_scenarios_fm_pattern};

    if ($TEST_LEVEL1>2) {
	print STDOUT "[ Output of 'make_parsim_scenarios_fm_pattern' ]\n\n";
	print STDOUT 'Content of @{$set_parsim_scenarios_fm_pattern} (#{scenarios} = ', $ct_psm_scenarios, '):', "\n\n";
	##my $chnode2char = ($ct_psm_scenarios>0) ? $set_parsim_scenarios_fm_pattern->[0][3] : {}; # OBSOLETE as of Dec 24 (Sat), 2011
	print STDOUT "( Shared by all patterns )\n\n";
	print STDOUT '%{$chnode2char} = ( # $child_node_ID => $character_ID #', "\n";
	foreach my $nodeid (sort {$a <=> $b} keys %{$chnode2char}) {  print STDOUT "\t", $nodeid, ' => ', $chnode2char->{$nodeid}, ",\n"; }
	print STDOUT "\t) .\n\n";

	for (my $sn = 0; $sn < $ct_psm_scenarios; $sn++) {
	    print STDOUT "\n( For the $sn th scenario )\n\n";

		# Slightly modified on Dec 24 (Sat), 2011 #
	    my ($add_node2ch, $add_node2minpen, $add_char_chngs) = @{$set_parsim_scenarios_fm_pattern->[$sn]};
	    ##my ($add_node2ch, $add_node2minpen, $add_char_chngs, $chnod2char2) = @{$set_parsim_scenarios_fm_pattern->[$sn]};

	    print STDOUT '%{$add_node2ch} = ( # $add_node_ID => \@child_nodes #', "\n";
	    foreach my $nodeid (sort {$a <=> $b} keys %{$add_node2ch}) { print STDOUT "\t", $nodeid, ' =>  [', join (', ', @{$add_node2ch->{$nodeid}}), '],', "\n"; }
	    print STDOUT "\t) \;\n\n";

	    print STDOUT '%{$add_node2min_penalty} = ( # $add_node_ID => $min_penalty_under_node #', "\n";
	    foreach my $nodeid (sort {$a <=> $b} keys %{$add_node2minpen}) { print STDOUT "\t", $nodeid, ' => ', $add_node2minpen->{$nodeid}, ',', "\n"; }
	    print STDOUT "\t) \;\n\n";

	    print STDOUT '%{$add_char_chngs} = ( # $branch_id => [$char_up, $char_lw, $penalty] #', "\n";
	    foreach my $brid (sort {$a <=> $b} keys %{$add_char_chngs}) { print STDOUT "\t", $brid, ' =>  [', join (', ', @{$add_char_chngs->{$brid}}), '],', "\n"; }
	    print STDOUT "\t) .\n\n";
	}
	print STDOUT "\n";
    }

			# Slightly MODIFIED on Dec 24 (Sat), 2011 #
		    push @parsim_scenarios_curr_cond, [$set_parsim_scenarios_fm_pattern, $chnode2char];
		    ##push @parsim_scenarios_curr_cond, @{$set_parsim_scenarios_fm_pattern};



			# ADDED to make the final output closer to perfect (on Dec 25 (Sun), 2011) #

		    for (my $sn=0; $sn<$ct_psm_scenarios; $sn++) { # Inner-inner for-loop (over parsimonious scenarios) #
			my $info_parsim_scenario = $set_parsim_scenarios_fm_pattern->[$sn];

			# @{$set_info_new_psm_scns->[$aps]} = (\@info_new_parsim_scenario, \%new_chnode2char) specifies the $aps th additional parsimonious scenario,
			# with @info_new_parsim_scenario = (\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs)
			# following the same format as @{$info_parsim_scenario},
			# and with %new_chnode2char = ($child_node_id => $character, ...) for the new character pattern accompanying this particular additional parsimonious scenario.
			#
			my $set_info_add_psm_scns = add_parsim_scenarios_by_chng_uniq_chars (@{$info_parsim_scenario}, %{$chnode2char}, %sgl_node2uniq_char, $pchar, $sbjct_id, ${$max_nodeid}, @{$pen_mtrx}, %{$node2min_penalties});
			unless (defined $set_info_add_psm_scns) {
			    print STDOUT "[ERROR 'part_resolve_mlt_node_bu'] The subroutine 'add_parsim_scenarios_by_chng_uniq_chars' FAILED on the $sn th scenario for the $p th maximum penalty-dropping character pattern from the $md th combination, on the $pchar th 'parent' character: $! !\n";
			    return (undef, undef, undef);
			}
    (print STDOUT "\n --- The subroutine 'add_parsim_scenarios_by_chng_uniq_chars' ENDED on the $sn th parsimonious scenario for the $p th character pattern from the $md th combination ...\n\n") if ($TEST_LEVEL1>1);


			my $ct_add_psm_scns = @{$set_info_add_psm_scns};
			$totct_add_psm_scns += $ct_add_psm_scns;

   if ($TEST_LEVEL1>2) {
	print STDOUT "\n[ Output of 'add_parsim_scenarios' ]\n\n";
	print STDOUT "#{Additional parsimonious scenarios created} = $ct_add_psm_scns .\n\n";
	if ($TEST_LEVEL1>3) {
	   print STDOUT 'Contents of @{$set_info_add_psm_scns}:', "\n\n";
	} else {
	   print STDOUT join ("\t", 'Indx_add_psm_scn', "'character_pattern'"), "\n\n";
	}
   }
			for (my $aps=0; $aps<$ct_add_psm_scns; $aps++) { # Innermost for-loop (over additional parsimonious scenarios) #
			    my ($new_parsim_scenario, $new_chnode2char) = @{$set_info_add_psm_scns->[$aps]};
			    my $str_charpattern = '';
			    foreach my $node (@{$children}) { $str_charpattern .= $new_chnode2char->{$node}; }
			    my $set_add_psm_scns = $add_chptn2psm_scns{$str_charpattern};
			    if (defined $set_add_psm_scns) {
				push @{$set_add_psm_scns->[0]}, $new_parsim_scenario;
			    } else {
				    # A BUG Fixed on Dec 26 (Mon), 2011 #
				$add_chptn2psm_scns{$str_charpattern} = [[$new_parsim_scenario], $new_chnode2char];
				##$add_chptn2psm_scns{$str_charpattern} = [[$new_parsim_scenario], $chnode2char];
			    }

   if ($TEST_LEVEL1>2) {
	if ($TEST_LEVEL1>3) {

	    print STDOUT "\n( For the $aps th additional scenario )\n\n";
	    print STDOUT '%{$new_chnode2char} = ( # $child_node_ID => $character_ID #', "\n";
	    foreach my $nodeid (sort {$a <=> $b} keys %{$new_chnode2char}) {  print STDOUT "\t", $nodeid, ' => ', $new_chnode2char->{$nodeid}, ",\n"; }
	    print STDOUT "\t) .\n\n";

	    my ($new_add_node2ch, $new_add_node2minpen, $new_add_char_chngs) = @{$new_parsim_scenario};

	    print STDOUT '%{$new_add_node2ch} = ( # $add_node_ID => \@child_nodes #', "\n";
	    foreach my $nodeid (sort {$a <=> $b} keys %{$new_add_node2ch}) { print STDOUT "\t", $nodeid, ' =>  [', join (', ', @{$new_add_node2ch->{$nodeid}}), '],', "\n"; }
	    print STDOUT "\t) \;\n\n";

	    print STDOUT '%{$new_add_node2minpen} = ( # $add_node_ID => $min_penalty_under_node #', "\n";
	    foreach my $nodeid (sort {$a <=> $b} keys %{$new_add_node2minpen}) { print STDOUT "\t", $nodeid, ' => ', $new_add_node2minpen->{$nodeid}, ',', "\n"; }
	    print STDOUT "\t) \;\n\n";

	    print STDOUT '%{$new_add_char_chngs} = ( # $branch_id => [$char_up, $char_lw, $penalty] #', "\n";
	    foreach my $brid (sort {$a <=> $b} keys %{$new_add_char_chngs}) { print STDOUT "\t", $brid, ' =>  [', join (', ', @{$new_add_char_chngs->{$brid}}), '],', "\n"; }
	    print STDOUT "\t) .\n\n";
	    print STDOUT "\n";
	} else {
	    print STDOUT join ("\t", $aps, "'$str_charpattern'"), "\n";
	}
   }
	    # %add_chptn2psm_scns = ($string_character_pattern => [\@set_parsim_scenarios_fm_pattern, \%chnode2char], ...)
			}	# End of the innermost for-loop (over additional parsimonious scenarios) # 
   if ($TEST_LEVEL1>2) {
	print STDOUT "\n";
   }

		    }		# End of the inner-inner for-loop (over parsimonious scenarios) #

		} # End of the main inner for-loop (over maximum penalty-dropping character patterns) #

	    }	# End of the main middle for-loop (over maximum penalty-dropping character combinations) #

		# ADDED to make the final output closer to perfect (on Dec 25 (Sun), 2011) #

    if ($TEST_LEVEL1>2) {
	print STDOUT "\n[ There are $totct_add_psm_scns additional parsimonious scenarios in total. ]\n\n";
    }

	    if ($totct_add_psm_scns>0) {
		foreach my $charpattern (sort keys %add_chptn2psm_scns) { # Innermost foreach-loop (over additional character patterns) #
		     my $info_add_parsim_scenarios = $add_chptn2psm_scns{$charpattern};
		     push @parsim_scenarios_curr_cond, $info_add_parsim_scenarios;

    if ($TEST_LEVEL1>2) {
	my ($addset_psm_scns, $new_chnode2char) = @{$info_add_parsim_scenarios};
	my $ct_add_psm_scns = @{$addset_psm_scns};
	print STDOUT "\n{ There are $ct_add_psm_scns additional parsimonious scenarios with the character pattern '$charpattern' }\n\n";
	print STDOUT "( Shared by the $ct_add_psm_scns patterns )\n\n";
	print STDOUT '%{$new_chnode2char} = ( # $child_node_ID => $character_ID #', "\n";
	foreach my $nodeid (sort {$a <=> $b} keys %{$new_chnode2char}) {  print STDOUT "\t", $nodeid, ' => ', $new_chnode2char->{$nodeid}, ",\n"; }
	print STDOUT "\t) .\n\n";

	for (my $sn = 0; $sn < $ct_add_psm_scns; $sn++) {
	    print STDOUT "\n( For the $sn th additional scenario )\n\n";

		# Slightly modified on Dec 24 (Sat), 2011 #
	    my ($new_add_node2ch, $new_add_node2minpen, $new_add_char_chngs) = @{$addset_psm_scns->[$sn]};

	    print STDOUT '%{$new_add_node2ch} = ( # $add_node_ID => \@child_nodes #', "\n";
	    foreach my $nodeid (sort {$a <=> $b} keys %{$new_add_node2ch}) { print STDOUT "\t", $nodeid, ' =>  [', join (', ', @{$new_add_node2ch->{$nodeid}}), '],', "\n"; }
	    print STDOUT "\t) \;\n\n";

	    print STDOUT '%{$new_add_node2min_penalty} = ( # $add_node_ID => $min_penalty_under_node #', "\n";
	    foreach my $nodeid (sort {$a <=> $b} keys %{$new_add_node2minpen}) { print STDOUT "\t", $nodeid, ' => ', $new_add_node2minpen->{$nodeid}, ',', "\n"; }
	    print STDOUT "\t) \;\n\n";

	    print STDOUT '%{$new_add_char_chngs} = ( # $branch_id => [$char_up, $char_lw, $penalty] #', "\n";
	    foreach my $brid (sort {$a <=> $b} keys %{$new_add_char_chngs}) { print STDOUT "\t", $brid, ' =>  [', join (', ', @{$new_add_char_chngs->{$brid}}), '],', "\n"; }
	    print STDOUT "\t) .\n\n";
	}
	print STDOUT "\n";

    }  # End of "if ($TEST_LEVEL1>2)"

		}	# End of the innermost foreach-loop (over additional character patterns) # 
	    } # End of "if ($ct_add_psm_scns>0)"


		# Streamline the list of parsimonious scenarios by removing
		# the scenarios where the subject node has only one child,
		# (because such a case can be realized by grafting the scenario where the subject has a different character onto the parent with the character in question.)
		# (ADDED on Jan 7 (Sat), 2012) #

	    my @sl_parsim_scenarios_curr_cond = ();
	    my $ct_psm_chpatterns = @parsim_scenarios_curr_cond;
	    for (my $cp = 0; $cp<$ct_psm_chpatterns; $cp++) { # Trailing middle for-loop (over character patterns resulsing from parsimonious scenarios) #
		my ($psm_scns_w_chpattern, $chnode2char) = @{$parsim_scenarios_curr_cond[$cp]};
		my @sl_parsim_scns_w_chpattern = ();
		my $ct_psm_scns = @{$psm_scns_w_chpattern};
		for (my $ps=0; $ps<$ct_psm_scns; $ps++) { # Trailing inner foreach-loop (over parsimonious scenarios) #
		    my $parsim_scenario = $psm_scns_w_chpattern->[$ps];
		    my $add_node2ch = $parsim_scenario->[0];
		    my $children_sbjct = $add_node2ch->{$sbjct_id};
		    unless (@{$children_sbjct}>1) {
			# Discard the scenario if $sbjct_id has only one (or zero) child #
	if ($TEST_LEVEL1>2) {
	    my $ct_children = @{$children_sbjct};
	    print STDOUT "[WARNING 'part_resolve_mlt_node_bu'] The subject ($sbjct_id) has ONLY $ct_children CHILD in the $ps th scenario giving the $cp th character pattern descended from the subject character '$pchar' : $! !\n";
	    print STDOUT '%{$add_node2ch} = ( # $node_id => [child,node,IDs] #', "\n";
	    foreach my $nodeid (sort {$a <=> $b} keys %{$add_node2ch}) { print STDOUT "\t", $nodeid, ' => [', join (', ', @{$add_node2ch->{$nodeid}}), '],', "\n"; }
	    print STDOUT "\t", ');', "\n\n";
	    print STDOUT '%{$chnode2char} = ( # $leave_id => $character_indx #' , "\n";
	    foreach my $nodeid (sort {$a <=> $b} keys %{$chnode2char}) { print STDOUT "\t", $nodeid, ' => ', $chnode2char->{$nodeid}, ',', "\n"; }
	    print STDOUT "\t", ').', "\n\n";
	}
			next;
		    }
		    push @sl_parsim_scns_w_chpattern, $parsim_scenario;
		}	# End of the trailing inner foreach-loop (over parsimonious scenarios) #

		unless (@sl_parsim_scns_w_chpattern>0) {
	    (print STDOUT "[WARNING 'part_resolve_mlt_node_bu'] The $cp th character pattern descended from the subject character '$pchar' has NO PARSIMONIOUS SCENARIOS letting the subject node ($sbjct_id) have two or more children, so the pattern will be discarded here: $! !\n\n") if ($TEST_LEVEL1>2);
		    next;
		}
		push @sl_parsim_scenarios_curr_cond, [\@sl_parsim_scns_w_chpattern, $chnode2char];
	    }	# End of the trailing middle for-loop (over character patterns resulsing from parsimonious scenarios) #


		# REVISED on Jan 7 (Sat), 2012 #
	    $cond_parsim_scenarios[$pchar] = \@sl_parsim_scenarios_curr_cond;
	    ##$cond_parsim_scenarios[$pchar] = \@parsim_scenarios_curr_cond;


    (print STDOUT "\n --- ALL the processes ENDED with Parent Character = $pchar ---\n\n") if ($TEST_LEVEL1>0);

	}	# End of the main outer for-loop (over potential characters at the subject node (, which will sometimes be called the 'parent').) #

    (print STDOUT "\n=== The subroutine 'part_resolve_mlt_node_bu' ENDED ===\n\n") if ($TEST_LEVEL1>0);

	return (\@cond_min_penalties, \@cond_parsim_scenarios, $minmin_penalty);

}	# End of the subroutine 'part_resolve_mlt_node_bu' #


	### DEFINITIONS of the SUBROUTINES that are revised versions
	### of those in the package 'MyTreeMap_II.pm'.

# min_penalties_sankoff_ptrslv_bo:
#	This satellite subroutine of 'map_parsim_states_sankoff_ptrslv_clm'
#	was created by slightly modifying 'min_penalties_sankoff_bo',
#	so that it can partly resolve unresolved (i.e. multifurcated) nodes of
#	the input tree. For resolving multifurcation,
#	it uses the subroutine 'part_resolve_mlt_node_bu'.
#	(CREATED on Jan 7 (Sat), 2012)
#
# min_penalties_sankoff_bo:
#	This satellite subroutine of 'map_parsim_states_sankoff_clm'
#	assigns minimum penalties on the subtree 
#	under the node in question under the condition that
#	the node  has particular characters.
#	(REVISED on Nov 27 (Sun), 2011)
#	A process is introduced to remove characters at each node that will NEVER be included
#	in any parsimonious states as soon as possible.
#	[ => DISABLED on Nov 28 (Mon), because it turned out to complicate the downstream analyses. ]
#
# INPUT: (\%node2children, \%node2depth, \%exnode2charindx, \@pen_mtrx), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%exnode2charindx = ($node_id => $indx of the character at the node, ...)
#	for the exterior nodes. The $indx conforms to @pen_mtrx;
#	$pen_mtrx[$i]->[$j] is the penalty imposed on the change
#	from the $i th character to the $j th character;
#
# [ OUTPUT: \%node2min_penalties, where ]  -- for the predecessor 'min_penalties_sankoff_bo'.
# OUTPUT: (\%node2min_penalties, \%mfnode2cond_parsim_scenarios), where
#	%node2min_penalties = ($node_id => \@min_penalties, ...) for ALL nodes,
#	with $min_penalties[$i] being the minimum summed penalty
#	on the subtree under $node_id under the condition that the node has
#	the $i th character in the matrix @pen_mtrx;
#	%mfnode2cond_parsim_scenarios = ($node_id => \@cond_parsim_scenarios, ...) for ALL multi-furcated nodes, where
#	@{$cond_parsim_scenarios[$i]->[$cp]} = (\@parsim_scenarios_with_chpattern, \%chnode2char)
#	gives the parsimonious scenarios providing the minimum penalty when the multifurcated node in question takes the $i th character, with
#	@{$parsim_scenarios_with_chpattern[$s]} = (\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs), 
#	uniquely specifying the $s th scenario with the character pattern designated with 
#	%chnod2char = ($child_node_id => $character_indx, ...), which uniquely assigns a character to each child node (in terms of the multifurcated input tree). Here,
#	%add_node2ch = ($additional_node_id => \@children, ...) giving the partial resolution,
#	%add_node2min_penalty = ($additional_node_id => $min_penalty, ...) giving the minimum summed penalty imposed on the subtree under $additional_node_id in this particular parsiminous scenario;
#	and %add_char_chngs = ($branch_id (including additional branch) => [$char_up, $char_lw, $pen], ...) describing the changes (from $char_up to $char_lw, which incurrs the penalty of $pen) that comprizes the scenario,
#	$minmin_penalty is the minimum penalty among those in @cond_min_penalties.
#
sub min_penalties_sankoff_ptrslv_bo (\%\%\%\@) {

	my ($node2ch, $node2depth, $exnode2charindx, $pen_mtrx) = @_;

    (print STDOUT "\n< Subroutine 'min_penalties_sankoff_ptrslv_bo' >\n\n") if ($TEST_LEVEL>0);
	my $ct_chars = @{$pen_mtrx};

	my %node2min_penalties;
	my %mfnode2cond_parsim_scenarios; # ADDED in creation of 'min_penalties_sankoff_ptrslv_bo' (on Jan 7 (Sat), 2012) #

	my @nodes = sort {$node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2depth};

	    # ADDED in creation of 'min_penalties_sankoff_ptrslv_bo' (on Jan 7 (Sat), 2012) #
	my $topnode_id = $nodes[$#nodes];
	my $max_nodeid = $topnode_id;
	foreach my $node (@nodes) { ($max_nodeid < $node) and ($max_nodeid = $node); }

	foreach my $node (@nodes) { # Main outer foreach-loop #

	    my $charindx = $exnode2charindx->{$node};
	    if (defined $charindx) { # An exterior node #
		my @min_penalties = ();
		$min_penalties[$charindx] = 0;
		$node2min_penalties{$node} = \@min_penalties;
		next;
	    }

	    my $children = $node2ch->{$node};
	    unless (defined $children) {
		print STDOUT "[ERROR 'min_penalties_sankoff_ptrslv_bo'] The 'interior' node '$node'  has NO CHILDREN: $! !\n";
		return (undef, undef);
	    }
	    my $ct_chnodes = @{$children};

		# Process multi-furcated nodes using 'part_resolve_mlt_node_bu'. #
		# (ADDED in creation of 'min_penalties_sankoff_ptrslv_bo' (on Jan 7 (Sat), 2012) #
	    unless ( ($ct_chnodes<3) or (($node == $topnode_id) and ($ct_chnodes<4)) ) {
		my ($cond_min_penalties, $cond_parsim_scenarios, $minmin_penalty)
		    =  part_resolve_mlt_node_bu ($node, $max_nodeid, @{$children}, @{$pen_mtrx}, %node2min_penalties);
		unless (defined $cond_min_penalties) {
		    print STDOUT "[ERROR 'min_penalties_sankoff_ptrslv_bo'] The subroutine 'part_resolve_mlt_node_bu' FAILED on the node '$node': $! !\n";
		    return (undef, undef);
		}

		$node2min_penalties{$node} = $cond_min_penalties;
		$mfnode2cond_parsim_scenarios{$node} = $cond_parsim_scenarios;

		next;

		# sub part_resolve_mlt_node_bu ($\$\@\@\%) {
		# INPUT: ($sbjct_id, \$max_nodeid, \@children, \@pen_mtrx, \%node2min_penalties), where
		# OUTPUT: (\@cond_min_penalties, \@cond_parsim_scenarios, $minmin_penalty), where
	    }

		# Process completely resolved nodes,
		# just as in the predecessor 'min_penalties_sankoff_bo'.

	    my @min_penalties = ();
	    my $minmin_penalty; # ADDED to simplify the downstream analyses on Nov 27 (Sun), 2011. #
	    for (my $i=0; $i<$ct_chars; $i++) { # Mid-outer for-loop (over the possible characters on the node) #
		my $min_penalty = 0;
		my $pen_row = $pen_mtrx->[$i];
		foreach my $ch (@{$children}) { # Mid-inner for-loop (over the children) #
		    my $min_pens_ch = $node2min_penalties{$ch};
		    unless (defined $min_pens_ch) {
			print STDOUT "[ERROR 'min_penalties_sankoff_ptrslv_bo'] The child node '$ch' of the node '$node' has NO INFORMATION on the MINIMUM PENALTIES: $! !\n";
			return undef; 
		    }

		    my $sb_min_pen;
		    for (my $j=0; $j<$ct_chars; $j++) { # Inner for-loop (over the possible characters on the child) #
			my $pen = $min_pens_ch->[$j];
			(defined $pen) or next;
			$pen += $pen_row->[$j];

			((defined $sb_min_pen) and ($sb_min_pen <= $pen)) and next;
			$sb_min_pen = $pen;
		    }				# End of the inner for-loop (over the possible characters on the child) #
		    $min_penalty += $sb_min_pen;

		}		# End of the mid-inner for-loop (over the children) #

		$min_penalties[$i] = $min_penalty;

=DISABLED because it turned out to further complicate the downstream analyses (on Nov 28 (Mon), 2011)
		(!(defined $minmin_penalty) or ($minmin_penalty > $min_penalty)) and ($minmin_penalty = $min_penalty); # ADDED to simplify the downstream analyses on Nov 27 (Sun), 2011. #
=cut

	    }		# End of the mid-outer for-loop (over the possible characters on the node) #

=DISABLED because it turned out to further complicate the downstream analyses (on Nov 28 (Mon), 2011)
	    	# ADDED to simplify the downstream analyses
		# by removing characters that canNOT be included in any parsimonious states (on Nov 27 (Sun), 2011). #
	    for (my $i=0; $i<$ct_chars; $i++) { # Trailing mid-outer for-loop (over possible characters on the node (switches to the parent node in the middle)) #
		my $cond_min_pen = $min_penalties[$i];
		((defined $cond_min_pen) and ($cond_min_pen> $minmin_penalty)) or next;
		my $pen_row = $pen_mtrx->[$i];
		my $if_ok = 1;
		for (my $j=0; $j<$ct_chars; $j++) { # Trailing inner for-loop (over possible characters on the node in question) #
		    my $alt_minpen = $min_penalties[$j];
		    (defined $alt_minpen) or next;
		    $alt_minpen += $pen_row->[$j];
		    if ($cond_min_pen > $alt_minpen) {
			$if_ok = 0;
			last;
		    }
		}		# End of the trailng inner for-loop (over possible characters on the node in question) #

		unless ($if_ok) { # Remove characters that will NEVER be included in parsiminious states #
		   $min_penalties[$i] = undef;
##=for TESTING the new function
	print STDOUT "[WARNING 'min_penalties_sankoff_ptrslv_bo'] The $i th character was removed on the node '$node' ... \n";
##=cut
		}
	    }		# End of the trailing mid-outer for-loop (over possible characters on the node (switches to the parent node in the middle)) #
=cut

	    $node2min_penalties{$node} = \@min_penalties;

	}			# End of the main outer foreach-loop #

    (print STDOUT "\n --- Subroutine 'min_penalties_sankoff_ptrslv_bo' ENDED ---\n\n") if ($TEST_LEVEL>0);

	 # REVISED in creation of 'min_penalties_sankoff_ptrslv_bo' (on Jan 7 (Sat), 2012) #
	return (\%node2min_penalties, \%mfnode2cond_parsim_scenarios);
	##return \%node2min_penalties;

}		# End of the subroutine 'min_penalties_sankoff_ptrslv_bo' #


# enum_parsim_scenarios_sankoff_ptrslv_td:
#	This satellite subroutine of 'map_parsim_states_sankoff_ptrslv_clm'
#	enumerates all possible parsimonious scenarios (within its capability)
#	based on the output of 'min_penalties_sankoff_ptrslv_bo'.
#	In terms of function this could be regarded as a modified version of 'enum_node2parsstat_sankoff_td' in 'MyTreeMap_II.pm'.
#	In terms of algorithm, however, these two are quite different.
#	(CREATED on Jan 8 (Sun), 2012)
#
# INPUT: ($topnode_id, \%node2ch, \%node2min_penalties, \%mfnode2cond_parsim_scenarios, \@pen_mtrx, \@charset, $fl_full_mode), where
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#
#	%node2min_penalties = ($node_id => \@min_penalties, ...) for ALL nodes,
#	with $min_penalties[$i] being the minimum summed penalty
#	on the subtree under $node_id under the condition that the node has
#	the character $charset[$i].
#
#	%mfnode2cond_parsim_scenarios = ($node_id => \@cond_parsim_scenarios, ...) for ALL multi-furcated nodes, where
#	@{$cond_parsim_scenarios[$i]->[$cp]} = (\@parsim_scenarios_with_chpattern, \%chnode2char)
#	gives the parsimonious scenarios providing the minimum penalty when the multifurcated node in question takes the $i th character, with
#	@{$parsim_scenarios_with_chpattern[$s]} = (\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs), 
#	uniquely specifying the $s th scenario with the character pattern designated with 
#	%chnod2char = ($child_node_id => $character_indx, ...), which uniquely assigns a character to each child node (in terms of the multifurcated input tree). Here,
#	%add_node2ch = ($additional_node_id => \@children, ...) giving the partial resolution,
#	%add_node2min_penalty = ($additional_node_id => $min_penalty, ...) giving the minimum summed penalty imposed on the subtree under $additional_node_id in this particular parsiminous scenario;
#	and %add_char_chngs = ($branch_id (including additional branch) => [$char_up, $char_lw, $pen], ...) describing the changes (from $charset[$char_up] to $charset[$char_lw], which incurrs the penalty of $pen) that comprizes the scenario;
#
#	$pen_mtrx[$i]->[$j] gives the penalty incurred on the change from the $charset[$i] to the $charset[$j];
#	@charset stores all the characters under consideration;
#
#	$fl_full_mode = 1 if the subroutine enumerate all possible parsimonious scenarios,
#		      = 0 if it only outputs the ancestral states.
#
# OUTPUT: ($min_tot_penalty, \@psm_anc_states, \@set_lists_psm_scenarios), where
#	$min_tot_penalty is the minimum total penalty achieved for the current state pattern (in a site or a locus);
#	@psm_anc_states lists all the ancestral states (i.e. elements of @charset at $topnode_id) that give $min_tot_penalty;
#
#	@{$set_lists_psm_scenarios[$as]->[$bps]} = (\%trunc_char_chngs, \%mfnode2psm_scn_w_chpattern)
#	gives information on the $bps th set of parsimonious scenarios
#	that have the ancestral state $psm_anc_states[$as]
#	and that appear identical on a 'broad' scale (i.e. along the input multifurcated tree)
#	but that can differ on a 'finer' scale (i.e. details on newly resolved nodes and associated branches).
#	More specifically,
#	%trunc_char_chngs = ($branch_id => [$char_up, $char_lw, $pen], ...)
#	records the changes from completely resolved nodes to their children (NOTE that $char_up and $char_lw are
#	character themselves here, instead of indices for @charset), and
#	%mfnode2psm_scn_w_chpattern = ($mfnode => \@br_parsim_scenarios_with_chpattern, ...)
#	is for changes from multi-furcated nodes to their children (on the input tree), with
#	@{$br_parsim_scenarios_w_chpattern[$s]} = (\%add_node2ch, \%add_char_chngs)
#	describes the $s th parsimonious scenario at the $mfnode, with
#	%add_node2ch = ($additional_node_id => \@new_children, ...) and
#	%add_char_chngs = ($additional_branch_id => [$char_up, $char_lw, $pen], ...)
#	are mostly the same as in the input data, EXCEPT that $char_up and $char_lw here are
#	characters themselves (i.e. elements of @charset), instead of indices for @charset.
#
#	NOTE: @{$set_lsits_psm_scenarios[$as]}'s are all empty if $fl_full_mode = 0.
#
sub enum_parsim_scenarios_sankoff_ptrslv_td ($\%\%\%\@\@$) {

	my ($topnode_id, $node2ch, $node2minpens, $mfnode2cond_psm_scns, $pen_mtrx, $charset, $fl_full_mode) = @_;

    (print STDOUT "\n< Subroutine 'enum_parsim_scenarios_sankoff_ptrslv_td' >\n\n") if ($TEST_LEVEL>0);

	my $ct_chars = @{$pen_mtrx};
	my $tot_min_cond_pens = $node2minpens->{$topnode_id};
	my $min_tot_penalty;
	my @psm_anc_states = ();

	for (my $i=0; $i<$ct_chars; $i++) { # 1st main for-loop #
	   my $min_condpen = $tot_min_cond_pens->[$i];
	   unless (defined $min_tot_penalty) {
		$min_tot_penalty = $min_condpen;
		@psm_anc_states = ($i);
		next;
	   }
	   ($min_tot_penalty < $min_condpen) and next; # No change is necessary.
	   if ($min_tot_penalty > $min_condpen) {
		$min_tot_penalty = $min_condpen;
		@psm_anc_states = ($i);
	   } else {
		push @psm_anc_states, $i;
	   }
	}	# End of the 1st main for-loop #

	unless ($fl_full_mode) { # A brief mode. #
	    my @set_lists_psm_scns = ();
	    my $ct_anc_states = @psm_anc_states;
	    for (my $s=0; $s<$ct_anc_states; $s++) {
		my $indx_anc_state = $psm_anc_states[$s];
		$psm_anc_states[$s] = $charset->[$indx_anc_state];
		push @set_lists_psm_scns, [];
	    }
	    return ($min_tot_penalty, \@psm_anc_states, \@set_lists_psm_scns);
	}

		## Hereafter the subroutine is in the full mode. ##

	my @set_lists_psm_scns = ();

	# @{$set_lists_psm_scenarios[$as]->[$bps]} = (\%trunc_char_chngs, \%mfnode2psm_scn_w_chpattern)

	foreach my $anc_state (@psm_anc_states) { # Outermost foreach-loop (over parsimonious ancestral states) #
		# Initialize arguments #
	    my @sbjct_nodes = ($topnode_id);
	    my @psm_states_on_sbjcts = ($anc_state);
	    my @list_psm_scns = ([{}, {}, \@psm_states_on_sbjcts]);
	    my @new_sbjct_nodes = ();

	    while (@sbjct_nodes>0) {  # Outer while-loop (until the subject nodes are exhausted) #

		my @new_sbjct_nodes = ();
		while (my $sbjct = shift @sbjct_nodes) { # Outer-middle while-loop (over the current set of subject nodes) #
		    my $children = $node2ch->{$sbjct};
		    unless ((defined $children) and (@{$children}>0)) { # Skip if the subject is an exterior node. #
			foreach my $info_psm_scn (@list_psm_scns) { shift @{$info_psm_scn->[2]}; } # This is necessary to synchronize @list_psm_scns with @sbjct_nodes. #
			next;
		    }
		    my $sets_cond_psm_scns = $mfnode2cond_psm_scns->{$sbjct};

		    my @char_cts = ();
		    for (1 .. $ct_chars) { push @char_cts, 0; }
		    foreach my $info_psm_scn (@list_psm_scns) { # Preliminary-preliminary middle foreach-loop #
			my $chars_on_sbjcts = $info_psm_scn->[2];
			my $charindx = $chars_on_sbjcts->[0];
			(defined $charindx) or next;
			$char_cts[$charindx]++;
		    }	# End of the preliminary-preliminary middle foreach-loop #

		    my @br_sets_cond_psm_scns = (); # element = \@br_cond_psm_scns_on_sbjct, whose element is [\@br_psm_scns_w_chpat, \%chnode2char].
		    my @set_chnode2psm_charindices = (); # element = {$chnode => \@psm_indices, ...}
		    if (defined $sets_cond_psm_scns) {

			for (my $i=0; $i<$ct_chars; $i++) { # 1st preliminary middle for-loop #

			    unless ($char_cts[$i]>0) { # Hope that this will save some memory space. #
				push @br_sets_cond_psm_scns, [];
				next;
			    }

			    my @br_cond_psm_scns_on_sbjct = ();
			    foreach my $info_psm_scns_w_chpat (@{$sets_cond_psm_scns->[$i]}) { # 1st preliminary inner-middle foreach-loop #
				my ($psm_scns_w_chpat, $chnode2char) = @{$info_psm_scns_w_chpat};
				my @br_psm_scns_w_chpat = ();
				foreach my $psm_scn (@{$psm_scns_w_chpat}) { # 1st preliminary inner foreach-loop #
				    my ($add_node2ch, $add_char_chngs) = @{$psm_scn}[0, 2];
				    my %trfm_add_char_chngs;
				    foreach my $br (keys %{$add_char_chngs}) { # Preliminary innermost foreach-loop #
					my ($indx_up, $indx_lw, $pen_br) = @{$add_char_chngs->{$br}};
					$trfm_add_char_chngs{$br} = [$charset->[$indx_up], $charset->[$indx_lw], $pen_br];
				
				    }		# End of the preliminary innermost foreach-loop #
				    my @br_psm_scn = ($add_node2ch, \%trfm_add_char_chngs);
				    push @br_psm_scns_w_chpat, \@br_psm_scn;
				} 	# End of the 1st preliminary inner foreach-loop #
				push @br_cond_psm_scns_on_sbjct, [\@br_psm_scns_w_chpat, $chnode2char];
			    }	# End of the 1st preliminary inner-middle foreach-loop #
			    push @br_sets_cond_psm_scns, \@br_cond_psm_scns_on_sbjct;
			}	# End of the 1st preliminary middle for-loop #
		    } else {

			for (my $i=0; $i<$ct_chars; $i++) { # 2nd preliminary middle for-loop (over characters) #

			    unless ($char_cts[$i]>0) { # Hope that this will save computational time. #
				push @set_chnode2psm_charindices, {};
				next;
			    }
			    my $row_pen = $pen_mtrx->[$i];

			    my %chnode2psm_charindices; # = ($chnode => \@psm_charindices, ...)
			    foreach my $chnode (@{$children}) { # 2nd preliminary inner-middle foreach-loop (over the child nodes) #
				my $cond_minpens = $node2minpens->{$chnode};
				my $sbtot_minpen ;
				my @psm_charindices = ();
				for (my $indx_lw=0; $indx_lw<$ct_chars; $indx_lw++) { # 2nd preliminary inner for-loop (over characters on the child node) #
				    my $cond_minpen = $cond_minpens->[$indx_lw];
				    (defined $cond_minpen) or next; # Skip if the conditional minimum penalty is undefined.
				    $cond_minpen += $row_pen->[$indx_lw];
				    unless (defined $sbtot_minpen) {
					$sbtot_minpen = $cond_minpen;
					@psm_charindices = ($indx_lw);
					next;
				    }
				    ($sbtot_minpen < $cond_minpen) and next;
				    if ($sbtot_minpen > $cond_minpen) {
					$sbtot_minpen = $cond_minpen;
					@psm_charindices = ($indx_lw);
				    } else {
					push @psm_charindices, $indx_lw;
				    }
				}	# End of 2nd preliminary inner for-loop (over characters on the child node) #
				$chnode2psm_charindices{$chnode} = \@psm_charindices;
			    }	# End of the 2nd preliminary inner-middle foreach-loop (over the child nodes) # 
			    push @set_chnode2psm_charindices, \%chnode2psm_charindices;
			}	# End of the 2nd preliminary middle for-loop (over characters) #

		    } 	# End of large if-else #

		    my @new_list_psm_scns = ();
		    while (my $info_psm_scn = shift @list_psm_scns) { # Middle while-loop (over the current list of parsimonious scenarios) #
			my ($trunc_char_chngs, $mfnode2psm_scn_w_chpat, $chars_on_sbjcts) = @{$info_psm_scn};
			my $indx_up = shift @{$chars_on_sbjcts};
			my $char_up = $charset->[$indx_up];
			my $row_pen = $pen_mtrx->[$indx_up];

			if (defined $sets_cond_psm_scns) { # The subject is a multi-furcated node. #

			    my $br_cond_psm_scns_on_sbjct = $br_sets_cond_psm_scns[$indx_up];
			    my $ct_psm_scns_on_sb = @{$br_cond_psm_scns_on_sbjct};
			    for (my $cp=0; $cp<$ct_psm_scns_on_sb; $cp++) { # 1st inner-middle for-loop (over the character patterns of the children of $sbjct) #
				my $br_info_psm_scns_w_chpat = $br_cond_psm_scns_on_sbjct->[$cp];
				my ($br_psm_scns_w_chpat, $chnode2char) = @{$br_info_psm_scns_w_chpat};

				my %new_mfnode2psm_scn_w_chpat = %{$mfnode2psm_scn_w_chpat};
				$new_mfnode2psm_scn_w_chpat{$sbjct} = $br_psm_scns_w_chpat;

				my @new_chars_on_sbjcts = @{$chars_on_sbjcts};
				foreach my $chnode (@{$children}) { push @new_chars_on_sbjcts, $chnode2char->{$chnode}; }

				push @new_list_psm_scns, [$trunc_char_chngs, \%new_mfnode2psm_scn_w_chpat, \@new_chars_on_sbjcts];

#	@{$cond_parsim_scenarios[$i]->[$cp]} = (\@parsim_scenarios_with_chpattern, \%chnode2char)
			    }	# End of the 1st inner-middle for-loop (over the character patterns of the children of $sbjct) #
			    next;
			}

			    ## The subject node is completely resolved. ##
			    # First, construct parsimonious scenarios describing changes from $sbjct to @{$children}

			my $chnode2psm_charindices = $set_chnode2psm_charindices[$indx_up];

			my @sets_trunc_char_chngs = ($trunc_char_chngs);
			my @sets_chars_on_sbjcts = ($chars_on_sbjcts);
			foreach my $chnode (@{$children}) { # 2nd inner-middle foreach-loop (over the child nodes) #

			    my $psm_charindices = $chnode2psm_charindices->{$chnode};

			    my @newsets_trunc_char_chngs = my @newsets_chars_on_sbjcts = ();

			    my $ct_trnc_char_chngs = @sets_trunc_char_chngs;
			    my $ct_psm_chars = @{$psm_charindices};
			    for (my $i=0; $i<$ct_psm_chars; $i++) { # 2nd inner for-loop (over parsimonious characters on the child node) #
				my $indx_lw = $psm_charindices->[$i];
				my $char_lw = $charset->[$indx_lw];
				my $pen = $row_pen->[$indx_lw];
				my @char_chng = ($char_up, $char_lw, $pen);

				for (my $tc=0; $tc<$ct_trnc_char_chngs; $tc++) { # Innermost for-loop (over the current trunc character changes) #
				    my $trunc_char_chngs = $sets_trunc_char_chngs[$tc];
				    my %new_trunc_char_chngs = %{$trunc_char_chngs};

					# Record if the branch immediately above $chnode actually underwent a character change.
				    ($indx_up == $indx_lw) or ($new_trunc_char_chngs{$chnode} = \@char_chng);

				    my $chars_on_sbjcts = $sets_chars_on_sbjcts[$tc];
				    my @new_chars_on_sbjcts = @{$chars_on_sbjcts};
				    push @new_chars_on_sbjcts, $indx_lw;

				    push @newsets_trunc_char_chngs, \%new_trunc_char_chngs;
				    push @newsets_chars_on_sbjcts, \@new_chars_on_sbjcts;
				}		# End of the innermost for-loop (over the current trunc character changes) #

			    }		# End of the 2nd inner for-loop (over parsimonous characters on the child node) #

				# Update the subset of parsimonious scenarios #
			    @sets_trunc_char_chngs = @newsets_trunc_char_chngs;
			    @sets_chars_on_sbjcts = @newsets_chars_on_sbjcts;

			}		# End of the 2nd inner-middle foreach-loop (over the child nodes) #


				# Now, incorporate the newly constructed parsimonious scenarios into the lists. #

			my $ct_new_trnc_char_chngs = @sets_trunc_char_chngs;
			for (my $ntc=0; $ntc<$ct_new_trnc_char_chngs; $ntc++) { # 3rd inner-middle for-loop (over the new trunc parsimonious scenarios) #
			    my $new_trunc_char_chngs = $sets_trunc_char_chngs[$ntc];
			    my $new_chars_on_sbjcts = $sets_chars_on_sbjcts[$ntc];

			    push @new_list_psm_scns, [$new_trunc_char_chngs, $mfnode2psm_scn_w_chpat, $new_chars_on_sbjcts];
			    ## ($trunc_char_chngs, $mfnode2psm_scn_w_chpat, $chars_on_sbjcts) = @{$info_psm_scn};

			}		# End of the 3rd inner-middle for-loop (over the new trunc parsimonious scenarios) #

			#	@{$parsim_scenarios_with_chpattern[$s]} = (\%add_node2ch, \%add_node2min_penalty, \%add_char_chngs), 
			#	%trunc_char_chngs = ($branch_id => [$char_up, $char_lw, $pen], ...)
			#	%mfnode2psm_scn_w_chpattern = ($mfnode => \@br_parsim_scenarios_with_chpattern, ...), with
			#	@{$br_parsim_scenarios_w_chpattern[$s]} = (\%add_node2ch, \%add_char_chngs), with
			#	%add_node2ch = ($additional_node_id => \@new_children, ...) and
			#	%add_char_chngs = ($additional_branch_id => [$char_up, $char_lw, $pen], ...)

		    }		# End of the middle while-loop (over the current list of parsimonious scenarios) #

		    push @new_sbjct_nodes, @{$children};

			# Update the list of parsimonious scenarios. #
		    @list_psm_scns = @new_list_psm_scns;

		}	# End of the outer-middle while-loop (over the current set of subject nodes) #

			# Update the list of subject nodes. #
		@sbjct_nodes = @new_sbjct_nodes;

	    }	# End of the outer while-loop (until the subject nodes are exhausted) #


		# Reformat the output array set by removing the auxiliary array, @chars_on_sbjcts . #

	    my $ct_broad_psm_scns = @list_psm_scns;
	    for (my $bps=0; $bps<$ct_broad_psm_scns; $bps++) { # Outer trailing for-loop #
		my ($trunc_char_chngs, $mfnode2psm_scn_w_chpat, $chars_on_sbjcts)  = @{$list_psm_scns[$bps]};
		$list_psm_scns[$bps] = [$trunc_char_chngs, $mfnode2psm_scn_w_chpat];

		# @{$set_lists_psm_scenarios[$as]->[$bps]} = (\%trunc_char_chngs, \%mfnode2psm_scn_w_chpattern)

	    }		# End of the outer trailing for-loop #

	    push @set_lists_psm_scns, \@list_psm_scns;

	}	# End of the outermost foreach-loop (over parsimonious ancestral states) #


	my $ct_anc_states = @psm_anc_states;
	for (my $s=0; $s<$ct_anc_states; $s++) { # Outermost trailing for-loop #
	    my $indx_anc_state = $psm_anc_states[$s];
	    $psm_anc_states[$s] = $charset->[$indx_anc_state];
	}		# End of the outermost trailing for-loop #

   (print STDOUT "\n --- 'enum_parsim_scenarios_sankoff_ptrslv_td' Ended ---\n\n") if ($TEST_LEVEL>0);

	return ($min_tot_penalty, \@psm_anc_states, \@set_lists_psm_scns);

}	# End of the subroutine 'enum_parsim_scenarios_sankoff_ptrslv_td' #


=for the predecessor 'MyTreeMap_II.pm'
# map_parsim_states_sankoff_clm:
#	This satellite subroutine of 'map_parsim_states_sankoff_msa'
#	performs Sankoff's parsimony algorithm on a column of MSA,
#	and find (a) map(s) of states onto nodes that are the most parsimonious
#	under the given penalty matrix.
#
# INPUT: (\%node2children, \%node2parent, \%node2depth, \%exnode2charindx, \@pen_mtrx, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id,...) for the non-top nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%exnode2charindx = ($node_id => $indx of the character at the node, ...)
#	for the exterior nodes. The $indx conforms to @pen_mtrx;
#	$pen_mtrx[$i]->[$j] is the penalty imposed on the change
#	from the $i th character to the $j th character;
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: \@set_node2parsstat, where
#
#	%{$set_node2parsstat[$opt]} = ($node_id => \@state_info, ...) for the $opt th
#	option of the parsimonious states of the site (or column) mapped onto the nodes,
#	with
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#	$sum_penal is the minimum summation of the penalties across the subtree under the node.
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
sub map_parsim_states_sankoff_clm (\%\%\%\%\@$) {

	my ($node2ch, $node2pa, $node2depth, $exnode2charindx, $pen_mtrx, $fl_full_mode) = @_;

#=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'map_parsim_states_sankoff_clm' >\n\n";
	print STDERR '%{$exnode2charindx} = (', "\n";
	foreach my $node (sort {$a <=> $b} keys %{$exnode2charindx}) {
	    print STDERR $node, '  =>  ', $exnode2charindx->{$node}, ",\n";
	}
	print STDERR ")\n\n";
	print STDERR '@{$pen_mtrx} = (', "\n";
	foreach my $row (@{$pen_mtrx}) { print STDERR "\t\[", join (', ', @{$row}), "\]\,\n"; }
	print STDERR "\t)\n\n";
#=cut

		## Assign conditional minimum penalties to the nodes
		## in a bottom-up manner

	# %{$node2min_penalties} = ($node_id => \@min_penals, ...) for all nodes, where
	# $min_penals[$i] is the minimum summed penalty
	# imposed on the subtree under the node in question
	# under the condition that the node has the $i th character.
	#
	my $node2min_penalties = min_penalties_sankoff_bo (%{$node2ch}, %{$node2depth}, %{$exnode2charindx}, @{$pen_mtrx});
	unless (defined $node2min_penalties) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_clm'] The subroutine 'min_penalties_sankoff_bo' FAILED: $! !\n";
	    return undef;
	}

#=for DEBUGGING the subroutine
	print STDERR "\n{ Output of 'min_penalties_sankoff_bo' }\n\n"; 
	print STDERR '%{$node2min_penalties} = ( # Node_ID => [minimin, penalties, conditional, on, the, state]', "\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2min_penalties}) {
	    print STDERR "\t", $node, '  =>   [', join (', ', @{$node2min_penalties->{$node}}), '],', "\n";
	}
	print STDERR "\t)\n\n";
#=cut

		## Enumerate the maps from the nodes to the parsimonious states
		## in a top-down manner

	my $set_node2parsstat = enum_node2parsstat_sankoff_td (%{$node2pa}, %{$node2depth}, %{$node2min_penalties}, @{$pen_mtrx}, $fl_full_mode);
	unless (defined $set_node2parsstat) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_clm'] The subroutine 'enum_node2parsstat_sankoff_td' FAILED: $! !\n";
	    return undef;
	}
#=for DEBUGGING the subroutine
	my $ct_sets = @{$set_node2parsstat};
	print STDERR "\n{ Output of 'enum_node2parsstat_sankoff_td' ($ct_sets maps of states} }\n\n";
	for (my $m = 0; $m <$ct_sets; $m++) {
	    my $node2parsstat = $set_node2parsstat->[$m];
	    print STDERR "( The $m th map )\n\n";
	    print STDERR join ("\t", '#Node_ID', 'State', '#{changes}'), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2parsstat}) {
		print STDERR join ("\t", $node, @{$node2parsstat->{$node}}), "\n";
	    }
	    print STDERR "\n\n";
	}
#=cut

#=for DEBUGGING the subroutine
	print STDERR " --- 'map_parsim_states_sankoff_cml' ENDED ---\n\n";
#=cut

	return $set_node2parsstat;

}	# End of the subroutine 'map_parsim_states_sankoff_clm' # 
=cut


=for the predecessor 'MyTreeMap_II.pm', but SET IT ASIDE,
    # because it may be modified into a generalized version of 'map_parsim_scenarios_sankoff_ptrslv_1clm' !!

# map_parsim_states_sankoff_msa:
#	This subroutine performs Sankoff's parsimony algorithm on a MSA,
#	and serially find a set of parsimonious maps of states onto nodes for
#	each of the columns in the input MSA.
#
#	NOTE that this subroutine eats up the input MSA.
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@msa, \@charlist, \@pen_mtrx, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	@charlist lists the possible characters;
#	$pen_mtrx[$i]->[$j] gives the penalty for the change from $charlist[$i] to $charlist[$j];
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: \@setset_node2parsstat, where
#	%{$setset_node2parsstat[$l]->[$opt]} = ($node_id => \@state_info, ...) for the $opt th
#	option of the parsimonious states of the $l th site mapped onto the nodes,
#	with
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#	$sum_penal is the minimum summation of the penalties across the subtree under the node.
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
sub map_parsim_states_sankoff_msa ($\%\%\%\@\@\@\@$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $msa, $charlist, $pen_mtrx, $fl_full_mode) = @_;

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] NO SEQUENCE in the input MSA: $! !\n";
	    return undef;
	}
	unless ($ct_seqs == @{$msa}) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] The number of sequences (", scalar (@{$msa}), ") in the MSA CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return undef;
	}
	my $ct_clms = length ${$msa->[0]};
	($ct_clms>0) or (return []);

	my $ct_chars = @{$charlist};
	unless ($ct_chars == @{$pen_mtrx}) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] The number of rows in the penalty matrix (", scalar (@{$pen_mtrx}), ") CLASHES with the number of characters ($ct_chars): $! !\n";
	    return undef;
	}

		# Prepare auxiliary arrays/hashes #

	# %{$node2depth} = ($node_id => #{nodes from the top node to itself}, ...);
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return undef;
	}

	my %char2indx; # ($character => $index in @{$charlist}, ...)
	for (my $i = 0; $i<$ct_chars; $i++) { $char2indx{$charlist->[$i]} = $i; }

	my @node_ids; # $node_ids[$k] is the exterior node ID corresponding to the $k th sequence in @{$msa}
	for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] NODE ID is UNDEFINED for the $k th sequence (name: $name): $! !\n";
		return undef;
	    }
	    $node_ids[$k] = $node_id;
	}			# End of the 2nd preliminary for-loop #

		## Main Process ##

	my @setset_node2parsstat = ();

	for (my $l=0; $l<$ct_clms; $l++) { # The main outer for-loop #

	    my %exnode2charindx; # ($node_id => $index of the character, ...) for the exterior nodes
	    my @char_in_clm = (); ## For a debugging purpose ##
	    for (my $k=0; $k<$ct_seqs; $k++) { # The 1st inner for-loop #

		my $node_id = $node_ids[$k];

		my $char = substr (${$msa->[$k]}, 0, 1, '');
		my $indx = $char2indx{$char};
		unless (defined $indx) {
		    print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] INDEX is UNDEFINED for the character '$char' in the $l th site of the $k th sequence: $! !\n";
		    return undef;
		}
		$exnode2charindx{$node_id} = $indx;
		push @char_in_clm, $char; ## For debugging ##
	    }				# End of the 1st inner for-loop #

	    my $set_node2parsstat =  map_parsim_states_sankoff_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2charindx, @{$pen_mtrx}, $fl_full_mode) ;
	    unless (defined $set_node2parsstat) {
		print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] 'map_marsim_states_sankoff_msa' FAILED on the $l th site: $! !\n";
		return undef;
	    }

		## Translate the character index back to the character ##

	    foreach my $node2parsstat (@{$set_node2parsstat}) { # The 2nd inner foreach-loop #
		foreach my $nodeid (keys %{$node2parsstat}) { # The inner-most foreach-loop #
		    my $parsstat = $node2parsstat->{$nodeid};
		    my $charindx = $parsstat->[0];
		    my $char = $charlist->[$charindx];
		    unless (defined $char) {
			print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] NO CHARACTER CORRESPONDING to the character index '$charindx' for the node ID $nodeid at the $l th column: $! !\n";
			return undef;
		    }
		    $parsstat->[0] = $char;
		}			# End of the inner-most foreach-loop #
	    }			# End of the 2nd inner foreach-loop #

	    push @setset_node2parsstat, $set_node2parsstat;

	}			# End of the main outer for-loop #

	return \@setset_node2parsstat;

}	# End of the subroutine 'map_parsim_states_sankoff_msa' #
=cut


# map_parsim_changes_sankoff_ptrslv_1clm:
#	This Perl script was created by substantially modifying the script 'map_parsim_states_sankoff_1clm'.
#	Different from the predecessor, which used input tree as it is even when the tree is incompletely resolved,
#	this script tries to (partially) resolve the unresolved part of the tree so that
#	the total penalty will be minimized
#	for the changes at the site (or the locus) represented by the input character column.
#	(But it (partially) resolves the tree only when it is necessary.)
#	Another major difference is that this subroutine outputs maps of character changes onto the branches,
#	instead of maps of character states onto the nodes, in order to save as much memory as possible.
#	(A minor difference is that this script no longer uses %node2parent.)
#
#	The subroutine heavily depends on two satellite subroutines,
#	'min_penalties_sankoff_ptrslv_bo', which in turn heavily depends on its own satellite named 'part_resolve_mlt_node_bu',
#	and  'enum_parsim_scenarios_sankoff_ptrslv_td'.
#	(CREAED on Jan 9 (Mon), 2012)
#
# map_parsim_states_sankoff_1clm:
#	This variant of 'map_parsim_states_sankoff_msa' specializes in
#	handling just a single column in a MSA,
#	namely comparing just a single trait across taxa.
#
# map_parsim_states_sankoff_msa:
#	This subroutine performs Sankoff's parsimony algorithm on a MSA,
#	and serially find a set of parsimonious maps of states onto nodes for
#	each of the columns in the input MSA.
#
#	NOTE that this subroutine eats up the input MSA.
#
# [INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@clm, \@charlist, \@pen_mtrx, $fl_full_mode), where ] -- for the predecessor 'map_parsim_states_sankoff_1clm'.
# INPUT: ($topnode_id, \%node2children, \%name2id, \@seqnames, \@clm, \@charlist, \@pen_mtrx, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	[ %node2parent = ($node_id => $parent_id, ...) for the non-top nodes; ] --- for the predecessor 'map_parsim_states_sankoff_1clm'
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	$clm[$k] is the character of the $k th sequence (or taxon);
#	@charlist lists the possible characters;
#	$pen_mtrx[$i]->[$j] gives the penalty for the change from $charlist[$i] to $charlist[$j];
#	$fl_full_mode = 1 if you want information on all the parsimonious scenarios,
#		      = 0 if you only need possible ancestral states giving the minimum penalty.
#
# [OUTPUT: \@set_node2parsstat, where
#	%{$set_node2parsstat[$opt]} = ($node_id => \@state_info, ...) for the $opt th
#	option of the parsimonious states of the clumn mapped onto the nodes,
#	with
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#	$sum_penal is the minimum summation of the penalties across the subtree under the node.
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given. ] -- for the predecessor 'map_parsim_states_sankoff_1clm'
#
# OUTPUT: ($min_tot_penalty, \@psm_anc_states, \@set_lists_psm_scenarios), where
#	$min_tot_penalty is the minimum total penalty achieved for the current state pattern (in a site or a locus);
#	@psm_anc_states lists all the ancestral states (i.e. elements of @charlist at $topnode_id) that give $min_tot_penalty;
#
#	@{$set_lists_psm_scenarios[$as]->[$bps]} = (\%trunc_char_chngs, \%mfnode2psm_scn_w_chpattern)
#	gives information on the $bps th set of parsimonious scenarios
#	that have the ancestral state $psm_anc_states[$as]
#	and that appear identical on a 'broad' scale (i.e. along the input multifurcated tree)
#	but that can differ on a 'finer' scale (i.e. details on newly resolved nodes and associated branches).
#	More specifically,
#	%trunc_char_chngs = ($branch_id => [$char_up, $char_lw, $pen], ...)
#	records the changes from completely resolved nodes to their children (NOTE that $char_up and $char_lw are
#	character themselves here, instead of indices for @charlist), and
#	%mfnode2psm_scn_w_chpattern = ($mfnode => \@br_parsim_scenarios_with_chpattern, ...)
#	is for changes from multi-furcated nodes to their children (on the input tree), with
#	@{$br_parsim_scenarios_w_chpattern[$s]} = (\%add_node2ch, \%add_char_chngs)
#	describes the $s th parsimonious scenario at the $mfnode, with
#	%add_node2ch = ($additional_node_id => \@new_children, ...) and
#	%add_char_chngs = ($additional_branch_id => [$char_up, $char_lw, $pen], ...),
#	with $char_up and $char_lw being characters themselves (i.e. elements of @charlist),
#	instead of indices for @charlist.
#
#	NOTE: @{$set_lsits_psm_scenarios[$as]}'s are all empty if $fl_full_mode = 0.
#
##sub map_parsim_states_sankoff_1clm ($\%\%\%\@\@\@\@$) {
#	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $clm, $charlist, $pen_mtrx, $fl_full_mode) = @_;
#
sub map_parsim_changes_sankoff_ptrslv_1clm ($\%\%\@\@\@\@$) {

	my ($topnode_id, $node2ch, $name2id, $seqnames, $clm, $charlist, $pen_mtrx, $fl_full_mode) = @_;

    (print STDOUT "\n< Subroutine 'map_parsim_changes_sankoff_ptrslv_1clm' >\n\n") if ($TEST_LEVEL>0);

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>0) {
	    print STDOUT "[ERROR 'map_parsim_changes_sankoff_ptrslv_1clm'] NO SEQUENCE in the input MSA: $! !\n";
	    return (undef, undef, undef);
	}
	unless ($ct_seqs == @{$clm}) {
	    print STDOUT "[ERROR 'map_parsim_changes_sankoff_ptrslv_1clm'] The number of sequences (", scalar (@{$clm}), ") in the column CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return (undef, undef, undef);
	}
	my $len_char  = length $clm->[0];
	unless ($len_char>0) {
	    print STDOUT "[ERROR 'map_parsim_changes_sankoff_ptrslv_1clm'] We do NOT HANDLE NULL CHARACTERS: $! !\n";
	    return (undef, undef, undef);
	}
	if ($len_char>1) {
	    print STDOUT "[WARNING 'map_parsim_changes_sankoff_ptrslv_1clm'] The character seems be of length $len_char, instead of one, but I will go on anyway ...\n\n";
	}

	my $ct_chars = @{$charlist};
	unless ($ct_chars == @{$pen_mtrx}) {
	    print STDOUT "[ERROR 'map_parsim_changes_sankoff_ptrslv_1clm'] The number of rows in the penalty matrix (", scalar (@{$pen_mtrx}), ") CLASHES with the number of characters ($ct_chars): $! !\n";
	    return (undef, undef, undef);
	}

		# Prepare auxiliary arrays/hashes #

	# %{$node2depth} = ($node_id => #{nodes from the top node to itself}, ...);
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDOUT "[ERROR 'map_parsim_changes_sankoff_ptrslv_1clm'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return (undef, undef, undef);
	}

	my %char2indx; # ($character => $index in @{$charlist}, ...)
	for (my $i = 0; $i<$ct_chars; $i++) { $char2indx{$charlist->[$i]} = $i; }

	my @node_ids; # $node_ids[$k] is the exterior node ID corresponding to the $k th sequence in @{$msa}
	for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDOUT "[ERROR 'map_parsim_changes_sankoff_ptrslv_1clm'] NODE ID is UNDEFINED for the $k th sequence (name: $name): $! !\n";
		return (undef, undef, undef);
	    }
	    $node_ids[$k] = $node_id;
	}			# End of the 2nd preliminary for-loop #

		# A key auxiliary hash #

	my %exnode2charindx; # ($node_id => $index of the character, ...) for the exterior nodes
	for (my $k=0; $k<$ct_seqs; $k++) { # The 1st main for-loop #

	    my $node_id = $node_ids[$k];

	    my $char = $clm->[$k];
	    my $indx = $char2indx{$char};
	    unless (defined $indx) {
	        print STDOUT "[ERROR 'map_parsim_changes_sankoff_ptrslv_1clm'] INDEX is UNDEFINED for the character '$char' of the $k th sequence: $! !\n";
		return (undef, undef, undef);
	    }
	    $exnode2charindx{$node_id} = $indx;
	}				# End of the 1st main for-loop #

		### Main Process ###

		# MODIFIED in creation of 'map_parsim_changes_sankoff_ptrslv_1clm' (on Jan 9 (Mon), 2012).
		# (This version does away with (simple modifications of) 'map_parsim_states_sankoff_clm',
		# which just performs two subroutines,
		# 'min_penalties_sankoff_bo' and 'enum_node2parsstat_sankoff_td'.)

	# %{$node2min_penalties} = ($node_id => \@cond_min_pens, ...) with $cond_min_pens[$i] is the minimum penalty on condition that $node_id has the character $charlist->[$i];
	# %{$mfnode2cond_parsim_scenarios} = ($node_id => \@info_cond_parsim_scenarios, ...)
	# for the multi-furcated nodes. (Refer to 'min_penalties_sankoff_ptrslv_bo' for detailes.)
	#
	my ($node2min_penalties, $mfnode2cond_parsim_scenarios) = min_penalties_sankoff_ptrslv_bo (%{$node2ch}, %{$node2depth}, %exnode2charindx, @{$pen_mtrx});
	unless (defined $node2min_penalties) {
	    print STDOUT "[ERROR 'map_parsim_changes_sankoff_ptrslv_1clm'] The subroutine 'min_penalties_sankoff_ptrslv_bo' FAILED: $! !\n";
	    return (undef, undef, undef);
	}

   if ($TEST_LEVEL>2) {

	print STDOUT "[[ Outputs of 'min_penalties_sankoff_ptrslv_bo' ]]\n\n";

	print STDOUT '[ Conditional Minimum penalties (i.e. contents of %{$node2min_penalties}) ]', "\n\n";
	print STDOUT 'Node_ID';
	foreach my $char (@{$charlist}) { print STDOUT "\t", 'pen(', $char, ')'; }
	print STDOUT "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2min_penalties}) { # 5th middle foreach-loop #
	    my $min_penalties = $node2min_penalties->{$node};
	    my @out = ();
	    for (my $i=0; $i<$ct_chars; $i++) {
		my $minpen = $min_penalties->[$i];
		push @out, ((defined $minpen) ? $minpen : 'undef');
	    }
	    print STDOUT join ("\t", $node, @out), "\n";
	}	# End of the 1st auxiliary foreach-loop #
	print STDOUT "\n";

	print STDOUT '[ Conditional parsimonious scenarios for multfurcated nodes (i.e. contents of %{$mfnode2cond_parsim_scenarios}) ]', "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$mfnode2cond_parsim_scenarios}) { # 2nd auxiliary foreach-loop (over mutlifurcated nodes) #
	    my $sets_cond_parsim_scenarios = $mfnode2cond_parsim_scenarios->{$node};
	    print STDOUT "{{ Node ID: $node }}\n\n";

	    for (my $i=0; $i<$ct_chars; $i++) { # Inner-middle for-loop (over characters) #
		my $char = $charlist->[$i];
		my $cond_parsim_scenarios = $sets_cond_parsim_scenarios->[$i];
		my $ct_chpatterns = @{$cond_parsim_scenarios};
		print STDOUT "{ For the $i th character, '$char' }\n\n";
		print STDOUT '#{character_patterns} = ', $ct_chpatterns, "\n\n";

		for (my $cp=0; $cp<$ct_chpatterns; $cp++) { # Inner for-loop (over character patterns) #
		    my ($psm_scns_w_chpat, $chnode2char) = @{$cond_parsim_scenarios->[$cp]};
		    print STDOUT "(( The $cp th character pattern ))\n\n";

		    print STDOUT '( Condent of %{$chnode2char} )', "\n\n";
		    print STDOUT join ("\t", 'Child_node', 'charindx', 'char'), "\n\n";
		    foreach my $chnode (sort {$a <=> $b} keys %{$chnode2char}) { # 1st inner-inner foreach-loop (over child nodes) #
			my $charindx = $chnode2char->{$chnode};
			my $char = $charlist->[$charindx];
			print STDOUT join ("\t", $chnode, $charindx, $char), "\n";
		    }	# End of the 1st inner-inner foreach-loop (over child node) #
		    print STDOUT "\n";

		    my $ct_psm_scns = @{$psm_scns_w_chpat};
		    print STDOUT '#{Parsimonious scenarios} = ', $ct_psm_scns, " .\n\n";

		    for (my $ps=0; $ps<$ct_psm_scns; $ps++) { # Innermost for-loop (over parsimonious scenarios) #
			my ($add_node2ch, $add_node2minpen, $add_char_chngs) = @{$psm_scns_w_chpat->[$ps]};
			print STDOUT "( The $ps th scenario )\n\n";

			print STDOUT 'Contents of %{$add_node2ch} and %{$add_node2minpen}:', "\n\n";
			print STDOUT join ("\t", 'Add_node', 'min_pen', 'child,node,IDs'), "\n";
			# = ( # $additional_node_ID => [child,node,IDs] #', "]n";
			foreach my $node (sort {$a <=> $b} keys %{$add_node2ch}) { # 1st auxiliary foreach-loop #
			   my $minpen = $add_node2minpen->{$node};
			   (defined $minpen) or ($minpen = 'undef');
			   my $cnct_children = join (',', @{$add_node2ch->{$node}});
			   print STDOUT join ("\t", $node, $minpen, $cnct_children), "\n";
			   ##print STDOUT "\t", $node, ' =>  [', join (', ', @{$add_node2ch->{$node}}), '],', "\n";
			}	# End of the 1st auxiliary foreach-loop #
			print STDOUT "\n";

			print STDOUT 'Content of %{$add_char_chngs}:', "\n\n";
			print STDOUT join ("\t", 'Br_ID', 'char(before)', 'char(after)', 'pen'), "\n\n";
			foreach my $br (sort {$a <=> $b} keys %{$add_char_chngs}) { # 2nd auxiliary foreach-loop #
			    my ($indx_bf, $indx_af, $pen) = @{$add_char_chngs->{$br}};
			    my $char_bf = $charlist->[$indx_bf];
			    my $char_af = $charlist->[$indx_af];
			    print STDOUT join ("\t", $br, $char_bf, $char_af, $pen), "\n";
			} 	# End of the 2nd auxiliary foreach-loop #
			print STDOUT "\n";
			
		    }		# End of the innermost for-loop (over parsimonious scenarios) #
		    print STDOUT "\n";

		}	# End of the inner for-loop (over character patterns) #

	    }	# End of the inner-middle for-loop (over characters) #
	    print STDOUT "\n";

	}	# End of the 2nd auxiliary foreach-loop (over multi_furcated nodes) #

   }	# End of 'if ($TEST_LEVEL>2) '

	# @{$psm_anc_states} lists all the ancestral states giving $min_tot_penalty;
	# @{$set_lists_psm_scenarios->[$as][$bps]} = (\%trunc_char_chngs, \%mfnode2psm_scn_w_chpat)
	# records parsimonious scenarios in the $bps th 'broad' category when the ancestral state is $psm_anc_states->[$as]. (Refer to 'enum_scenarios_sankoff_ptrslv' for datails.)
	#
	my ($min_tot_penalty, $psm_anc_states, $set_lists_psm_scenarios) = enum_parsim_scenarios_sankoff_ptrslv_td ($topnode_id, %{$node2ch}, %{$node2min_penalties}, %{$mfnode2cond_parsim_scenarios}, @{$pen_mtrx}, @{$charlist}, $fl_full_mode) ;
	unless (defined $min_tot_penalty) {
	    print STDOUT "[ERROR 'map_parsim_changes_sankoff_ptrslv_1clm'] The subroutine 'enum_parsim_scenarios_sankoff_ptrslv_td' FAILED: $! !\n";
	    return (undef, undef, undef);
	}

		### RESTART FROM HERE ###

=for the predecessor 'map_parsim_states_sankoff_1clm'
	my $set_node2parsstat =  map_parsim_states_sankoff_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2charindx, @{$pen_mtrx}, $fl_full_mode) ;
	unless (defined $set_node2parsstat) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_1clm'] 'map_marsim_states_sankoff_msa' FAILED: $! !\n";
	    return undef;
	}
=cut

=for the predecessor 'map_parsim_states_sankoff_1clm'. This was DISABLED because it is no longer necessary.
		## Translate the character index back to the character ##
	foreach my $node2parsstat (@{$set_node2parsstat}) { # The 2nd main foreach-loop #
	    foreach my $nodeid (keys %{$node2parsstat}) { # The inner foreach-loop #
		my $parsstat = $node2parsstat->{$nodeid};
		my $charindx = $parsstat->[0];
		my $char = $charlist->[$charindx];
		unless (defined $char) {
		    print STDERR "[ERROR 'map_parsim_states_sankoff_1clm] NO CHARACTER CORRESPONDING to the character index '$charindx' for the node ID $nodeid: $! !\n";
		    return undef;
		}
		$parsstat->[0] = $char;
	    }			# End of the inner foreach-loop #
	}			# End of the 2nd main foreach-loop #
=cut

    (print STDOUT "\n --- The subroutine 'map_parsim_changes_sankoff_ptrslv_1clm' ENDED ---\n\n") if ($TEST_LEVEL>0);

		# MODIFIED in creation of 'map_parsim_changes_sankoff_ptrslv_1clm' (on Jan 9 (Mon), 2012) #
	return ($min_tot_penalty, $psm_anc_states, $set_lists_psm_scenarios);
	##return $set_node2parsstat;

}	# End of the subroutine 'map_parsim_changes_sankoff_ptrslv_1clm' #


=for the predecessor 'MyTreeMap_II.pm'
# locate_parsim_changes:
#	This subroutine receives a map of states (or characters) onto nodes,
#	and locates the branches that exhibit changes
#	under the assumption of parsimony.
#
# INPUT: (\%node2pa, \%node2stat, $indx_stat), where
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2stat = ($node_id => $stat_info, ...) for all nodes, with
#	$stat_info->[$indx_stat] if $indx_stat is defined,
#	or $stat_info if $indx_stat = undef, being the state (character)
#	mapped onto the node $node_id.
#
# OUTPUT: (\%br2change, $ct_changes), where
#	%br2change = ($branch_id => [$stat_parent, $stat_node], ...), where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch.
#	$stat_node is the state (or character) of the node, and
#	$stat_parent is the state (or character) of the parent node.
#
sub locate_parsim_changes (\%\%$) {

	my ($node2pa, $node2stat, $indx_stat) = @_;

	my $ct_changes = 0;
	my %br2change;

	foreach my $node (keys %{$node2pa}) {
	    my $parent = $node2pa->{$node};
	    (defined $parent) or next;

	    my $stat_info = $node2stat->{$node};
	    unless (defined $stat_info) {
		print STDERR "[ERROR 'locate_parsim_changes'] NO STATE INFORMATION on the node '$node': $! !\n";
		return (undef, undef);
	    }
	    my $pa_stat_info = $node2stat->{$parent};
	    unless (defined $pa_stat_info) {
		print STDERR "[ERROR 'locate_parsim_changes'] NO STATE INFORMATION on the parent '$parent' of the node '$node': $! !\n";
		return (undef, undef);
	    }
	    my $stat_node = (defined $indx_stat) ? $stat_info->[$indx_stat] : $stat_info ;
	    my $stat_pa = (defined $indx_stat) ? $pa_stat_info->[$indx_stat] : $pa_stat_info ;
	    ($stat_node eq $stat_pa) and next;
	    $br2change{$node} = [$stat_pa, $stat_node];
	    $ct_changes++;
	}

	return (\%br2change, $ct_changes);

}	# End of the subroutine 'locate_parsim_changes' #
=cut


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


