#!/usr/bin/perl
#
# summary_examine_complex.6p_aug_len.raw.pl:
#	This slightly modified version of 'summary_examine_complex.6p_aug.raw.pl'
#	artificially clusters gapped segment under a more lenient criterion.
#	More precisely, it regards two gapped segments as interacting
#	if they underwent indels along nearest-neighbor branches,
#	even if the indels do not form a 'triangle' with an indel along a yet another nearest-neighboring branch.
#	[CREATED on July 18 (Mon), 2016.]
#
# summary_examine_complex.6p_aug.raw.pl:
#	This "augmented" version of 'summary_examine_complex.6p.raw.pl' artificially clusters
#	gapped segments taking account of phylogenetic relationships between the branches
#	where the indels are inferred to have taken place.
#	[CREATION STARTED on July 13 (Wed), 2016, ENDED on July dd (www), 2016.]
#
# summary_examine_complex.6p.raw.pl:
#	This slightly modified version of 'summary_examine_complex.5p.raw.pl' classifies
#	the artificial clusters of gapped segments in terms of %{category-A} (for abscissa)
#	and (Tot_Tot#{indels}, Tot_diff#{indels}) (for ordinates).
#	[CREATION STARTED on July 1 (Fri), 2016, ENDED on July 2 (Sat), 2016.]
#
# summary_examine_complex.5p.raw.pl:
#	This slightly modified version of 'summary_examine_complex.5.raw.pl' counts the gapped segments,
#	instead of the artificial clusters as the predecessor counts.
#	[CREATED on June 30 (Thu), 2016.]
#
# summary_examine_complex.5.raw.pl:
#	This script artificially clusters the gapped segments mutually separated by distances shorter than
#	a given threshold, and tallies the total number of indels in each artificial cluster.
#	The script was created by modifying the framework of 'summary_examine_complex.2_2D.raw.pl'.
#	[CREATION STARTED on June 29 (Wed), 2016, ENDED on June 30 (Thu), 2016.]
#
# summary_examine_complex.2_2D.raw.pl:
#	This slightly modified version of 'summary_examine_complex.2.raw.pl'
#	takes exclusively the 2D distributions of the properties.
#	It classifies the segments into two categories: A = 'too-long' + 'complex errors', and  B = 'non-complex errors' + 'correct'.
#	[CREATION STARTED on June 21 (Tue), 2016, ENDED on June dd (www), 2016.]
#
# summary_examine_complex.2.raw.pl:
#	The focus of this script is on the difference between the clusters of gapped segments
#	that were erroneously reconstructed and those that were correctly reconstructed.
#	This script was created by modifying 'summary_examine_complex.1.raw.pl'.
#	[CREATION STARTED on June 2 (Thu), 2016, ENDED on June 3 (Fri), 2016.]
#
# summary_examine_complex.1.raw.pl:
#	This Perl script summarizes the results of the Perl script 'examine_complex_via_mblks.2.pl',
#	which are converted into a single table via 'tabulate_examine_complex.pl'.
#	The script was created by modifying 'summary_examine_purge_ctrl.raw.pl'.
#	Especially, it imports (from 'bch_pbs_examine_purges.D.pl') the function to read in the list of relevant segments.
#	[CREATED on June 1 (Wed), 2016.]
#
# summary_examine_purge_ctrl.raw.pl:
#	This Perl script summarizes the results of the Perl script 'examine_purge_ctrls.2.pl',
#	which are converted into a single table via 'tabulate_examine_purge_ctrls.pl'.
#	The script was created by slightly modifying 'summary_examine_purge.raw.pl'.
#	[CREATED on May 16 (Mon), 2016.]
#
# summary_examine_purge.raw.pl:
#	This Perl script summarizes the results of the Perl script 'examine_purges_via_mblks.2.pl',
#	which are converted into a single table via 'tabulate_examine_purge.pl'.
#	The script was created by taking advantage of the framework of 'summary_classify_msa_errors1.raw.pl'.
#	[CREATION Started on May 11 (Wed), 2016, ENDED on Mmm dd (www), 2016.]
#
#	[MODIFIED on May 16 (Mon), 2016.]
#	Now, the script also examines more detailed distributions,
#	such as the distributions of the P-values for specified window sizes and specified branch properties.
#
# summary_classify_msa_errors1.raw.pl:
#	This script is a slightly modified version of 'summary_classify_msa_errors.raw.pl'.
#	The script now tallies the number of all composite-blocks in each erroneous segment as well.
#	[CREATED on Oct 26 (Mon), 2015]
#
# summary_classify_msa_errors.raw.pl:
#	This script does not pre-screen either in terms of whether or not the erroneous segment is on a MSA boudary or 
#	whether or not the segment contains non-parsimonious gapped segments.
#	It was created by slightly modifying the script 'summary_classify_msa_errors.bulk_pars.pl'.
#	(CREATION STARTED on Nov 17 (Mon), 2014, ENDED on Nov 18 (Tue), 2014.)
#
# summary_classify_msa_errors.bulk_pars.pl:
#	This Perl script summarizes the results of the Perl script 'classify_msa_errors_via_mblks.1.pl',
#	which are converted into a single table via 'tabulate_classify_msa_errors.pl'.
#	It excludes those erroneous segments on the MSA boundaries and those containing non-parsimonious gapped segments.
#	It was created by taking advantage of the framework of the script 'summary_cmp_ref_vs_rec_gpconf_stats.bulk_pars.pl'.
#	(CREATION STARTED on Nov 16 (Sun), 2014, ENDED on Nov 17 (Mon), 2014.)
#
# summary_cmp_ref_vs_rec_gpconf_stats.bulk_pars.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_gconf_stats.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_gconf_stats.pl'.
#	It excludes those erroneous segments on the MSA boundaries and those containing non-parsimonious gapped segments.
#	It was created by taking advantage of the framework of the script 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2_lite.pl'.
#	(CREATION STARTED on Nov 14 (Fri), 2014, ENDED on Nov 15 (Sat), 2014.)
#
# summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2_lite.pl:
#	This script was created by just stripping commented-out commands off 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2.pl'.
#	(CREATED on Nov 14 (Fri), 2014.)
#
# summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2.pl:
#	The only modifications from the previous version is that
#	this version uses boundaries whose results could easily be interpreted as 'round-off'.
#	(CREATED on Nov 13 (Thu), 2014.)
#
# summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_indel_cts.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_cts_indels.pl'.
#	It excludes those erroneous segments on the MSA boundaries and those containing non-parsimonious gapped segments.
#	It was created by modifying the script 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'.
#	(CREATED on Oct 9 (Thu), 2014.)
#
# summary_cmp_ref_vs_rec_complliment.bulk_pars.pl:
#	This revised version of 'summary_cmp_ref_vs_rec_complliment.bulk.pl'
#	further excludes erroneously reconstructed segments including non-parsimonious gapped segments
#	before tallying the bulk erroneous segments.
#	(CREATED on Oct 3 (Fri), 2014.)
#
# summary_cmp_ref_vs_rec_complliment.bulk.pl:
#	This revised version of 'summary_cmp_ref_vs_rec_complliment.pl' tallies the erroneously reconstructed segments
#	after excluding the segments that touch either of the MSA ends.
#	(CREATION STARTED on Oct 2 (Thu), 2014, ENDED on Oct 3 (Fri), 2014.)
#
#
# summary_cmp_ref_vs_rec_complliment.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_compllimment.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_compllimment.pl'.
#	It was created by modifying the script 'summary_cmp_ref_vs_rec_abs_scr.pl'.
#	(CREATED on Oct 2 (Thu), 2014.)
#
# summary_cmp_ref_vs_rec_abs_scr.pl:
#	This Perl script summarizes the results of the Perl script 'compare_ref_vs_rec_abs_scoring.PhyML.pl',
#	which are converted into a single table via 'tabulate_cmp_ref_vs_rec_abs_scoring.pl'.
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);
use MyMalnIO qw(
	extract_fasta extract_clw
	);
	## pr_clustal_like_lite_aa
	## pr_ODEN_input pr_phylip_aln
use MyArrayManipu qw( diff_array );

use MyMalnManipu qw( rm_gapcolumn2 );
use MyMalnManipu_UH qw(
	make_segid2segaln
	msa2len_grps msa2column_set extract_aln_mtrx2
	);
	## extract_aln_mtrx

use MyCluster_distribution qw(
	bin_initializer cluster_distributor cluster_distributor_wgt
	batch_initializer batch_pr_file_distri batch_pr_file_distri_xl
	cluster_distributor_wgt
	init_2dim_distri distributor_2dim pr_file_2dim_distri pr_file_2dim_distri_xl
	create_key_intervals );
	## bin_initializer cluster_distributor  # REPLACED in creation of 'summary_examine_complex.6p.raw.pl'. #

use MyContingency_table qw(
	ratios_contingency_tables add_ctbl_margins
	pr_juxtapose_ctbls_xl convert_ctbl_to_2d_distri
	);

		### MAJOR ADDITION in creation of 'summary_examine_complex.5p.raw.pl'. ###
use MyCluster_distribution_II qw(
	distributor_2dim_wgt
	);

	## The following modules were ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. ###

use MyTreeManipu_II qw(
	extract_sgltree_from_file input_tree_file output_tree
	);

use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg
	);

use MyTreeManipu_III qw( make_node2abs_depth  make_node2ct_exoffs );

use MyTaxonomy qw( fetch_common_ancestors );

	## END of "The following modules were ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'." ###

sub readin_dawg_tree2 (\$); # IMPORTED from '../../../Full_Analyses/Scripts/examine_complex_via_mblks.2.pl' in creation of 'summary_examine_complex.6p_aug.raw.pl'. #
sub selfcompare_indel_hosting_branches (\@\%\%); # ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #
sub compare_indel_hosting_branches (\@\@\%\%$\%); # ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #

sub broad_error_type ($);


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);


	### Fixed parameters ###
	### (IMPORTED from '../../../Full_Analyses/Scripts/examine_complex_via_mblks.2.pl' in creation of 'summary_examine_complex.6p_aug.raw.pl'.) ###

our $CHAR_ABS =  our $GAP  = '-';
our $CHAR_PRES = our $NGAP = 'X';

our $TOP_NODE  = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;

	### User-specified Parameters ###

	# If the {spacer size} <= {an element of @THRSHS_SPSIZE},
	# then the two gapped segments will be put into the same cluster. #
	# (ADDED in creation of 'summary_examine_complex.5.raw.pl'.)
##our @THRSHS_SPSIZE = (1 .. 5);
##our @THRSHS_SPSIZE = (1 .. 8);
#our @THRSHS_SPSIZE = (4, 6, 8, 10);
##our @THRSHS_SPSIZE = (4, 6, 8, 10, 12);
our @THRSHS_SPSIZE = (2, 3, 4, 5, 6);
our $CT_THRSHS_SPSIZE = @THRSHS_SPSIZE;


	# (ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'.) #

	# Upper-bound of the spacer size that could mediate interacting gapped segments. #
#our $UPPER_BOUND_SPSIZE = 10;
our $UPPER_BOUND_SPSIZE = 15;
#our $UPPER_BOUND_SPSIZE = 20;

#our $UPPER_BOUND_SPSIZE = 25;
#our $UPPER_BOUND_SPSIZE = 30;
##our $UPPER_BOUND_SPSIZE = 40;
#our $UPPER_BOUND_SPSIZE = 50;

	# [ If two gapped segments mediated by the spacer of size $spsize (<= $UPPER_BOUND_SPSIZE),
	# and the larger and the smaller of them  have the sizes larger than $spsize * $MAG_THRSH
	# and $spsize / $MAG_THRSH, respectively, then the segments could interact with each other. ] --- OBSOLETE (because these conditions are NOT what I really meant.)
	# => When $THRSH_SPSIZE < Ls = {total spacer size} <= $UPPER_BOUND_SPSIZE,
	# the two gapped segments will be clustered if the following two conditions are satisfied:
	# (i) min {size_gs1, size_gs2} >= Ls /$MAG_THRSH,
	# (ii) max {size_gs1, size_gs2} >= Ls * $MAG_THRSH.
our $MAG_THRSH = 2;

	# END of "(ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'.)" #


		# MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #
our %INTYPE2CTSEQS = (
	'12primates' => 12,
	'15mammals' => 15,
	'9fe_mammals' => 9,
	'TESTs/Test1a' => 15
	);
##our  $CT_SEQS = 16; # Number of sequences in a MSA. # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #

	# If the scores of the reference and the reconstructed MSAs for a segment 
	# differ less than this value, they are regarded as 'equal' to each other.
##our $THRSH_EQUAL = 0.1;
##our $THRSH_EQUAL = 1E-03;
our $THRSH_EQUAL = 1E-05;

our $EPS = $THRSH_EQUAL * 1E-03; # An 'infinitesimal number' used to take statistics.


	# Level of the verboseness of the test outputs.
	# (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose.)
our $TEST_LEVEL = 1;
#our $TEST_LEVEL = 5;

our $TEST_LEVEL_DBG = 0; # Special outputs for Debugging. #
#our $TEST_LEVEL_DBG = 5; # Special outputs for Debugging. #

	## Specifying the fields of input/output files. ##

		# MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #
		# => Further MODIFIED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #

my @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}', '[br/stat_ue/beg/end/len]_ev1');
##my @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}');

=For the predecessor 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.2_lite.pl'.
	# MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
our @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Total_Score.ref',  'Total_Score.rec', 'AvAv#{insertions}.ref', 'AvAv#{insertions}.rec', 'AvAv#{deletions}.ref', 'AvAv#{deletions}.rec', 'Av#{null_columns}.ref', 'Av#{null_columns}.rec');

our @TAGS_OUT_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'AvAv#{indels}.ref', 'Diff(AvAv#{indels})', 'L1_Dist(#{ins},#{dels})', 'Deletion_Bias');
=cut


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
our @TAGS_AUX_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', '#{nonpars_gpd_segs}'); # ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'. #

	# ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
our @TAGS_AUX3_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Total_Score.ref',  'Total_Score.rec', 'Av#{null_columns}.ref', 'Av#{null_columns}.rec');
=cut


##my $alignment_type = 'MAFFT/E_INS_i';
##my $alignment_type = 'MAFFT/G_INS_i';
##my $alignment_type = 'PRANK/Default';
my $alignment_type = 'PRANK/Best_Match';

	# ADDED in creation of 'summary_examine_purge.raw.pl'. #
#my $input_msa_type = '12primates';
my $input_msa_type = '15mammals';
#my $input_msa_type = '9fe_mammals';

#####my $input_msa_type = 'TESTs/Test1a'; # For TESTING the script. #

our $CT_SEQS = $INTYPE2CTSEQS{$input_msa_type}; ## ADDED in creation of 'summary_examine_complex.1.raw.pl'. ##

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.

	# Auxiliary files. #
	# $auxfile = "$top_auxdir/$mid_auxdir_pref$id_class$mid_auxdir_suff/$mid_aux_path/$auxfile_pref$session_id$auxfile_suff",
	# where  $id_class and $sesion_id will be collected later.
	# (ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'.)

my $top_auxdir = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Errors/Erroneous_Regions/Equiv_Regions/$alignment_type/hogd_alt_lite_sbcl.1_lite/Outputs";
my $mid_auxdir_pref = 'hog_d';
my $mid_auxdir_suff = '';
my $mid_aux_path = '';
my $auxfile_pref = 'equiv_rgs_brief.';
my $auxfile_suff = '.txt';

		# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'. #
my $auxfile2 = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Errors/Erroneous_Regions/Equiv_Regions/$alignment_type/Inclusion_gpd_segs/hogd_alt_lite_sbcl.1/tbl_incl_gpd_segs.txt.gz";
=cut

=DISABLED in creation of 'summary_examine_purge.raw.pl'.

		# ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
##my $auxfile3 = "../../../Cts_Indels/Summary_tbl/hogd_alt_lite_sbcl.1/tbl_cmp_ref_vs_rec_cts_indels.txt.gz";
##my $auxfile3 = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Errors/Contrast_Analyses/Characteristics/Erroneous_Regions/$alignment_type/Cts_Indels/Summary_tbl/hogd_alt_lite_sbcl.1/tbl_cmp_ref_vs_rec_cts_indels.txt.gz";
my $auxfile3 = "/Users/kezawa3/Ezawa_works_UH/Main_Project/Secondary_Projects/ANALYSES/MSA_Prcd_Errors/Contrast_Analyses/Characteristics/Erroneous_Regions/$alignment_type/Cts_Indels/Summary_tbl/hogd_alt_lite_sbcl.1/tbl_cmp_ref_vs_rec_cts_indels.txt.gz";

=cut

		# ADDED in creation of 'summary_examine_complex.1.raw.pl'. #
##my $auxfile5 = "../../../Aux_Inputs/Tbl_Complex/tbl_errsegs_w_complex.$input_msa_type.txt.gz"; 
my $auxfile5 = "../../../Aux_Inputs/Tbl_Complex_wo_ud4/tbl_errsegs_w_complex.$input_msa_type.txt.gz"; 

		# ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #
my $intree = "../../Dawg_ctrl_templates/dawg_inp_$input_msa_type.txt";


	### Input/Output files/directories ###

	# $infile1 = "$indir1/$infile_name1",
my $indir1 = "../../Summary_tbl/$input_msa_type"; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
##my $indir1 = "../../Summary_tbl/hogd_alt_lite_sbcl.1";
##my $infile_name1 = 'tbl_cmp_ref_vs_rec_compllimment.txt.gz';
##my $infile_name1 = 'tbl_cmp_ref_vs_rec_cts_indels.txt.gz'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
##my $infile_name1 = 'tbl_cmp_ref_vs_rec_gconf_stats.txt.gz'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'.
##my $infile_name1 = 'tbl_classify_msa_errors.txt.gz'; # MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'.
##my $infile_name1 = 'tbl_classify_msa_errors.1.txt.gz'; # MODIFIED in creation of 'summary_classify_msa_errors1.raw.pl'. #
##my $infile_name1 = 'tbl_examine_purge.txt.gz'; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
##my $infile_name1 = 'tbl_examine_purge_ctrl.txt.gz'; # MODIFIED in creation of 'summary_examine_purge_ctrl.raw.pl'. #

my $infile_name1 = 'tbl_examine_complex.txt.gz'; # MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. #


	# $outfile = "$outdir/$outfile_name"
#my $outdir =  $input_msa_type; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
##my $outdir =  'hogd_alt_lite_sbcl.1';
##my $outdir = "TESTs/Test1"; ### For TESTING the script ###
my $outdir = "TESTs/Test1b"; ### For TESTING the script ###

##my $outfile_name = 'tbl_diff_ref_vs_rec_compllimment.txt';
##my $outfile_name = 'tbl_diff_ref_vs_rec_cts_indels.txt'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
##my $outfile_name = 'tbl_fdiff_ref_vs_rec_cts_indels.txt'; # MODIFIED in creation of 'summary_cmp_ref_vs_rec_cts_indels.bulk_pars.pl'. #
##my $outfile_name = 'tbl_classify_sbj_msa_errors.txt'; # MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
##my $outfile_name = 'tbl_classify_sbj_msa_errors.1.txt'; # MODIFIED in creation of 'summary_classify_msa_errors1.raw.pl'. #
##my $outfile_name = 'tbl_examine_sbj_purges.txt'; # MODIFIED in creation of 'summary_examine_purge.raw.pl'. #
my $outfile_name = 'tbl_examine_sbj_purge_ctrls.txt'; # MODIFIED in creation of 'summary_examine_purge_ctrl.raw.pl'. #



	## Check files/directories ##

foreach my $dir ($indir1) {
##foreach my $dir ($intopdir1, $intopdir2) {
    unless (( -d $dir ) and ( -r $dir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable TOP INPUT DIRECTORY '$dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}

##=DISABLED in creation of 'summary_examine_purge.raw.pl'. ## RESURRECTED & MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. ##

foreach my $auxfile ($auxfile5, $intree) { # MODIFIED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #
##foreach my $auxfile ($auxfile5) { # MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. # ...}
   unless ( ( -f $auxfile ) and ( -r $auxfile ) ) { # ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'. #
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable Auxiliary File '$auxfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
   }
}
##=cut

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.

		# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'. #

foreach my $dir ($top_auxdir) {
    unless (( -d $dir ) and ( -r $dir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable Top Auxiliary Directory '$dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}

my $list_id_classes = list_sessions_ext ($top_auxdir, $mid_auxdir_pref, $mid_auxdir_suff, 'dr');
unless (defined $list_id_classes) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the top auxiliary directory '$top_auxdir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

	# $auxfile = "$top_auxdir/$mid_auxdir_pref$id_class$mid_auxdir_suff/$mid_aux_path/$auxfile_pref$session_id$auxfile_suff",
		# END of "ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'." #

=cut # END of "DISABLED in creation of 'summary_classify_msa_errors.raw.pl'."


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
=cut


	### Declare the Beginning of the Script ###

print STDOUT "\n<<<<< Program name: $program_name (path: $program_path) >>>>>\n\n";

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.
print STDOUT "# This analysis will be performed after excluding the erroneous segments on the left-ends or the right-ends of MSAs, as well as the erroneous segments containing non-parsimonious gapped segments.\n\n";
=cut

print STDOUT '$alignment_type = ', $alignment_type , " .\n\n";

=For the predecessor 'summary_cmp_ref_vs_rec_abs_scr.pl'.
print STDOUT "Scoring method: $METHOD .\n\n";
=cut

print STDOUT '$input_msa_type = ', $input_msa_type , " .\n\n"; # ADDED in creation of 'summary_examine_purge.raw.pl'. #

print STDOUT "Input file:  {in_dir1}/$infile_name1,\n";
print STDOUT "       where  {in_dir1} = $indir1 .\n\n";


			# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.2.pl'. #
print STDOUT "Auxiliary input files:\n\n";

=DISABLED in creation of 'summary_classify_msa_errors.raw.pl'.
print STDOUT "   (for segments on MSA-ends)  {top_auxdir}/$mid_auxdir_pref", '{ID_Class}', "$mid_auxdir_suff/$mid_aux_path/$auxfile_pref", '{session_ID}', "$auxfile_suff ,\n";
print STDOUT "              where {top_auxdir} = $top_auxdir ,\n";
print STDOUT "                    {ID_Class} = ", join (', ', @{$list_id_classes}), " ,\n\n";
print STDOUT "                    {session_ID}'s will be collected in the class-specific manner ;\n\n";
	# $auxfile = "$top_auxdir/$mid_auxdir_pref$id_class$mid_auxdir_suff/$mid_aux_path/$auxfile_pref$session_id$auxfile_suff",

print STDOUT "    (for non-parsimonious segments) $auxfile2 ,\n\n"; # ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.bulk_pars.pl'.
=cut


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
print STDOUT "    (for complete-likelihoods and counts of null columns) $auxfile3 .\n\n"; # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #
=cut

print STDOUT "    (for phylogenetic tree)  $intree .\n\n"; # ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #


print STDOUT "    (for segments with 'Complex' errors or 'N/A') $auxfile5 .\n\n"; # ADDED in creation of 'summary_examine_complex.1.raw.pl'. #

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
print STDOUT "Output file: {out_dir}/$outfile_name,\n";
print STDOUT "       where {out_dir} = $outdir .\n\n\n";
=cut

print STDOUT "Parameters necessary for the calculation:\n\n";
##print STDOUT "DEFINITION:  'Diff' = 'Rec' - 'Ref' .\n\n";
print STDOUT '$THRSH_EQUAL = ', "$THRSH_EQUAL .\n";
print STDOUT '$EPS         = ', "$EPS .\n\n\n";


print STDOUT '#{Sequences} = ', $CT_SEQS , " .\n\n\n"; # ADDED in creation of 'summary_examine_complex.1.raw.pl'. #


print STDOUT '@THRSHS_SPSIZE = (', join (', ', @THRSHS_SPSIZE) , ") .\n\n\n"; # ADDED in creation of 'summary_examine_complex.5.raw.pl'. #

		# ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #

print STDOUT '$UPPER_BOUND_SPSIZE = ', $UPPER_BOUND_SPSIZE, " ,\n";
print STDOUT '$MAG_THRSH = ', $MAG_THRSH , " .\n\n";

print STDOUT '# NOTE: When $THRSH_SPSIZE < Ls = {total spacer size} <= $UPPER_BOUND_SPSIZE,', "\n";
print STDOUT '# the two gapped segments will be clustered if the following two conditions are satisfied:', "\n";
print STDOUT '# (i) min {size_gs1, size_gs2} >= Ls /$MAG_THRSH,', "\n";
print STDOUT '# (ii) max {size_gs1, size_gs2} >= Ls * $MAG_THRSH.', "\n\n\n";

print STDOUT '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";
print STDOUT '$TEST_LEVEL_DBG = ', $TEST_LEVEL_DBG, " .\n\n\n";

		# END of "ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'." #



	### Arrays used to take Statistics ###


		### MODIFIED in creation of 'summary_examine_complex.6p.raw.pl'. (Ia) ###

our @BOUNDS_PCA = ($EPS, 20+$EPS, 40+$EPS, 60, 80, 100); # MODIFIED in creation of 'summary_examine_complex.6p.raw.pl'. #
my $CT_BOUNDS_PCA = @BOUNDS_PCA;

our @TOTCTS_INDELS = (1 .. 7);

our @CTPAIRS = (); # Element = sprintf ("(%d,%d)", $tot_totct_indel, $tot_diffct_indel)

foreach my $totct (@TOTCTS_INDELS) {
    foreach my $x (0 .. $totct) {
	my $diffct = $totct - 2 * $x;
	push @CTPAIRS, sprintf ("(%d,%d)", $totct, $diffct);
    }
}

(undef, my $tags_intvl_pcA) = init_2dim_distri (@BOUNDS_PCA, @BOUNDS_PCA, "%1.0f");

		# MODIFIED to fix a bug. #
our @TAGS_INTVL_PCA = @{$tags_intvl_pcA}[1 .. $#{$tags_intvl_pcA}];
##our @TAGS_INTVL_PCA = @{$tags_intvl_pcA}; # A BUG fixed. #

=For the predecessor 'summary_examine_complex.5p.raw.pl'.

		### MODIFIED in creation of 'summary_examine_complex.1.raw.pl'. ###
		### => FUrther MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. ###
		### => Yet Further MODIFIED in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

	# MODIFIED in creation of 'summary_examine_complex.5.raw.pl'. #
our @BOUNDS_CT0 = (0 .. 6, 8, 10);
our @BOUNDS_CT = (0 .. 11, 13, 15, 17, 19, 21, 26, 31, 41, 51, 76, 101);
##our @BOUNDS_CT0 = (0 .. 5);
##our @BOUNDS_CT = (0 .. 11, 13, 15, 17, 19, 21, 26, 31, 41, 51);
our @BOUNDS_PNCT0 = (-7 .. 7);

our @BOUNDS_PCA = (0, $EPS, 20+$EPS, 40+$EPS, 60, 80, 100); # ADDED in creation of 'summary_examine_complex.5.raw.pl'. #

=cut

=For the predecessor 'summary_examine_complex.2.raw.pl'.
our @BOUNDS_CT = (0 .. 11, 13, 15, 17, 19, 21, 26, 31, 41, 51, 76, 101, 151, 201, 301);
our @BOUNDS_PNCT = (-200, -100, -50, -40, -30, -25, -20, -15, -10 .. 11, 16, 21, 31, 41, 51, 101, 201);
##our @BOUNDS_CT = (0 .. 11, 13, 15, 17, 19, 21, 26, 31, 41, 51, 76, 101, 151, 201);
##our @BOUNDS_PNCT = (-100, -50, -40, -30, -25, -20, -15, -10 .. 11, 16, 21, 31, 41, 51, 101);
our @BOUNDS_PRPT = (0, 0.001, 0.005, 0.01, 0.025, 0.05, 0.1, 0.2, 0.5, 0.8, 0.9, 0.95, 0.975, 0.99, 0.995, 0.999, 1);
=cut

		### MODIFIED in creation of 'summary_examine_complex.6p.raw.pl'. (Ib) ###

my @set_ctpair2bins_pcA = ();
my @set_bins_pcA_total = ();

for (my $i_thr = 0; $i_thr <$CT_THRSHS_SPSIZE; $i_thr++) {

    my %ctpair2bins_pcA;

    my @batch_pcA = ();
    foreach my $ctpair (@CTPAIRS, 'Others') {
	my @bins_pcA = ();
	push @batch_pcA, \@bins_pcA;
	$ctpair2bins_pcA{$ctpair} = \@bins_pcA;
    }
    my @bins_pcA_total = ();
    push @batch_pcA, \@bins_pcA_total;

    batch_initializer (@batch_pcA, @BOUNDS_PCA);

    push @set_ctpair2bins_pcA, \%ctpair2bins_pcA;
    push @set_bins_pcA_total, \@bins_pcA_total;
}


=For the predecessor 'summary_examine_complex.5p.raw.pl'. #

		### ADDED in creation of 'summary_examine_complex.5.raw.pl'. (I) ###

my @set_bins_pcA_vs_totct_clms = my @set_bins_pcA_vs_tot_totct_indel = ();

for (my $i_thr = 0; $i_thr <$CT_THRSHS_SPSIZE; $i_thr++) {

	(my $bins_pcA_vs_totct_clms, undef) = init_2dim_distri (@BOUNDS_PCA, @BOUNDS_CT, "%1.0f");
	(my $bins_pcA_vs_tot_totct_indel, undef) = init_2dim_distri (@BOUNDS_PCA, @BOUNDS_CT0, "%1.0f");

	$set_bins_pcA_vs_totct_clms[$i_thr] = $bins_pcA_vs_totct_clms;
	$set_bins_pcA_vs_tot_totct_indel[$i_thr] = $bins_pcA_vs_tot_totct_indel;

}

=cut


=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (I) ##

		### ADDED in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

(my $bins_max_totct_indels_vs_maxct_columns_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_maxct_columns_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_totlen_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_max_totlen_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_max_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

(my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_min_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (I)" ##


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
		# MODIFIED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #

our @BROAD_TYPES = ('Shift', 'Complex', 'Merge', 'Purge', 'Split', 'Ex-nihilo', 'NB_Merge', 'NB_Split', 'CII', 'CEII', 'pNB_Merge', 'pNB_Split', 'iCII', 'iCEII', 'Other');
##our @BROAD_TYPES = ('Shift', 'Complex', 'Merge', 'Purge', 'Split', 'Ex-nihilo', 'NB_Merge', 'NB_Split', 'CII', 'CEII', 'Other');

	#  MODIFIED in creation of 'summary_classify_msa_errors1.raw.pl' (on Oct 26 (Mon), 2015). #
my @tags_ct = ('Intervals', '#{equiv_indels}', '#{cblks}_all', '#{cblks}_eff', 'Totlen{cblks}_eff', 'Len{cblk}_eff');
##my @tags_ct = ('Intervals', '#{equiv_indels}', '#{cblks}_eff', 'Totlen{cblks}_eff', 'Len{cblk}_eff');
my @tags_prpt = ('Intervals', @BROAD_TYPES);
my @tags_cblk_len = ('Intervals', @BROAD_TYPES);


	#  MODIFIED in creation of 'summary_classify_msa_errors1.raw.pl' (on Oct 26 (Mon), 2015). #
my @bins_ct_equiv_indels = my @bins_ct_all_cblks = my @bins_ct_eff_cblks = my @bins_totlen_eff_cblks = my @bins_len_eff_cblk = ();
my @batch_ct = (\@bins_ct_equiv_indels, \@bins_ct_all_cblks, \@bins_ct_eff_cblks, \@bins_totlen_eff_cblks, \@bins_len_eff_cblk);
##my @bins_ct_equiv_indels = my @bins_ct_eff_cblks = my @bins_totlen_eff_cblks = my @bins_len_eff_cblk = ();
##my @batch_ct = (\@bins_ct_equiv_indels, \@bins_ct_eff_cblks, \@bins_totlen_eff_cblks, \@bins_len_eff_cblk);
batch_initializer (@batch_ct, @BOUNDS_CT);

my @batch_prpt = my @batch_cblk_len = ();
my %brtype2bins_prpt;
my %brtype2bins_cblk_len;
my %brtype2bins_ct_vs_prpt_ecblk;
foreach my $brtype (@BROAD_TYPES) {
	my @bins_prpt = my @bins_cblk_len = ();
	push @batch_prpt, \@bins_prpt;
	$brtype2bins_prpt{$brtype} = \@bins_prpt;
	push @batch_cblk_len, \@bins_cblk_len;
	$brtype2bins_cblk_len{$brtype} = \@bins_cblk_len;

	(my $bins_ct_vs_prpt_ecblk, undef) = init_2dim_distri (@BOUNDS_CT, @BOUNDS_PRPT, "%1d");
	$brtype2bins_ct_vs_prpt_ecblk{$brtype} = $bins_ct_vs_prpt_ecblk;
}
batch_initializer (@batch_prpt, @BOUNDS_PRPT);
batch_initializer (@batch_cblk_len, @BOUNDS_CT);

=cut # END of "DISABLED in creation of 'summary_examine_purge.raw.pl'."




=DISABLED in creation of 'summary_examine_purge.raw.pl'.
			# ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
our @TAGS_OUT_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', '#{cblks}_eff');
foreach my $brtype (@BROAD_TYPES) { push @TAGS_OUT_FLD, "#{$brtype}"; }
=cut

	### Preliminary Process 0: Readin the input tree,
	### and pre-process some phylogenetic information.

print STDOUT "\n<<< Preliminary 0: Reading-in the Input Tree, and pre-process some phylogenetic information  >>>\n\n";

	### (IMPORTED from '../../../Full_Analyses/Scripts/examine_complex_via_mblks.2.pl'
	###    in creation of 'summary_examine_complex.6p_aug.raw.pl'.)

# @{$info_dawg_tree} = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr)
#
my $info_dawg_tree = readin_dawg_tree2 ($intree);
##my $info_dawg_tree = readin_dawg_tree ($intree);
unless (defined $info_dawg_tree) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'readin_dawg_tree2' FAILED on the input tree file '$intree' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my ($node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$info_dawg_tree};

	## ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. ##

my $topnode_id = $name2id->{$TOP_NODE};

my %equiv_branches; # = ($branch_id => $equivalent_branch_id, ...)

if (@{$node2ch->{$topnode_id}}==2) {
   my ($br1, $br2) = @{$node2ch->{$topnode_id}};
   $equiv_branches{$br1} = $br2;
   $equiv_branches{$br2} = $br1;
}


		# Assign depths to the nodes. #

my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
unless (defined $node2depth) {
    print STDOUT "[ERROR '$program_name'] 'make_node2abs_depth' FAILED on the shared phylogenetic tree !\n";
    print STDERR "[ERROR '$program_name'] $!\n\n";
    die ;
}
print STDOUT "\n ... The subroutine 'make_node2abs_depth' Ended ...\n\n";


		# Assign #{exterior offsprings} to the nodes. #

my $node2ct_exoffs = make_node2ct_exoffs (%{$node2ch}, %{$node2depth});
unless (defined $node2ct_exoffs) {
    print STDOUT "[ERROR '$program_name'] 'make_node2ct_exoffs' FAILED on the shared phylogenetic tree !\n";
    print STDERR "[ERROR '$program_name'] $!\n\n";
    die ;
}
print STDOUT "\n ... The subroutine 'make_node2ct_exoffs' Ended ...\n\n";

		# Assign the nearest neighbors to each branch. #

my %br2nn; # = ($br_id => {$id_nearest_neighbor => 1, ...}, ...)

foreach my $br (keys %{$br_attr}) { # Preliminary foreach-loop (over branches). #

	my %nn; # = ($id_nearest_neighbor => 1, ...)
	$br2nn{$br} = \%nn;

	my $children = $node2ch->{$br};
	if (defined $children) {
	    foreach my $ch (@{$children}) { $nn{$ch} = 1; }
	}

	my $eq_br = $equiv_branches{$br};
	if (defined $eq_br) { # The upper-end is the top node.
	    my $children_eq = $node2ch->{$eq_br};
	    if (defined $children_eq) {
		foreach my $ch (@{$children_eq}) { $nn{$ch} = 1; }
	    }
	    next;
	}

	my $pa = $node2pa->{$br};
	if (defined $br_attr->{$pa}) {
	    $nn{$pa} = 1;
	    my $eq_pa = $equiv_branches{$pa};
	    if (defined $eq_pa) { $nn{$eq_pa} = 1; }
	}
	my $sibls = $node2ch->{$pa};
	foreach my $sbl (@{$sibls}) { ($sbl == $br) or ($nn{$sbl} =1); }

}				# End of the preliminary foreach-loop (over branches). #

if ($TEST_LEVEL_DBG>2) {
    print STDOUT "\n", '[ Content of %br2nn ]', "\n\n";
    print STDOUT 'Br_ID', "\t", 'IDs,of,nearest,neighbor,branches', "\n\n";

    foreach my $br (sort {$a <=> $b} keys %br2nn) {
	my @nns = sort {$a <=> $b} keys %{$br2nn{$br}};
	print STDOUT $br, "\t", join (',', @nns), "\n";
    }
} # END of "if ($TEST_LEVEL_DBG>2) {...}"

	## END of "ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'." ##

			### RESTART FROM HERE (0) ###


	### Preliminary Process 5: Readin the list of subject segments. ###
	### (IMPORTED from 'bch_pbs_examine_purges.D.pl', and MODIFIED in creation of 'summary_examine_complex.1.raw.pl'.)

my %if_complex; # = ("$id_class:$in_msa_id:$segid" => 1, ...) for segments with complex errors.
my %if_na; # = ("$id_class:$in_msa_id:$segid" => 1, ...) for segments whose error types are 'N/A'.

if (1) {

   print STDOUT "\n<<< Preliminary 5: Reading-in the list of sugject segments  >>>\n\n";

   my $fh_aux5 = open_input_file2 ($auxfile5);
   unless (defined $fh_aux5) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the auxiliary input file '$auxfile5' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
   }
   local $/ = "\n"; # Use the newline character as the input-record-separator. #

   my $aux_header5 = <$fh_aux5>;

   my @TAGS_AUX5_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Types;of;associated;errors');

   my $indices_fld_aux5 = tfm_labels2indices_gen (@TAGS_AUX5_FLD, $aux_header5, "\t");
   unless (defined $indices_fld_aux5) {
   	$aux_header5 =~ s/\s+$//; 
   	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the auxiliary header:\n";
   	print STDOUT " '$aux_header5' !\n";
   	print STDERR "[ERROR '$program_name'] $!\n";
   	die ;
   }
   my $max_indx_aux5 = 0;
   foreach my $indx (@{$indices_fld_aux5}) { ($max_indx_aux5 <$indx) and ($max_indx_aux5 = $indx); }

   while (my $line = <$fh_aux5>) { # Preliminary while-loop. #

	my @fields = split /\t/, $line;
	(@fields>$max_indx_aux5) or next;
	##(@fields>3) or next;

	my ($id_class, $in_msa_id, $segid, $error_type) = @fields[@{$indices_fld_aux5}];
	##my ($id_class, $in_msa_id, $segid) = @fields[0,1,2];

	my $combo_id = "$id_class:$in_msa_id:$segid";

	if ($error_type eq 'N/A') {
	    $if_na{$combo_id} = 1;
	} elsif ($error_type =~ m/Complex/) {
	    $if_complex{$combo_id} = 1;
	}

=For the source 'bch_pbs_examine_purges.D.pl'
	my $combo_id = "$id_class:$in_msa_id";
	if (defined $class_msa_id2sbjct_segs{$combo_id}) {
	    push @{$class_msa_id2sbjct_segs{$combo_id}}, $segid;
	} else {
	    $class_msa_id2sbjct_segs{$combo_id} = [$segid];
	}
=cut

   }			# End of the preliminary while-loop. #
   close ($fh_aux5);
}

	### END of "Preliminary Process 5: Readin the list of subject segments." ###



	### Prepare the Input/Output files ###

my $infile1 = "$indir1/$infile_name1";
##my $outfile = "$outdir/$outfile_name"; ## DISABLED in creation of 'summary_examine_purge.raw.pl'.

my $fh_in1 = open_input_file2 ($infile1);
unless (defined $fh_in1) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the input file '$infile1' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
my $fh_out = open_output_file ($outfile, 1);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
=cut

	### Main Processes ###

print STDOUT "\n\n<<< MAIN PROCESS:  Artificially Cluster the Gapped Segments and Examine Their Properties >>>\n\n\n"; # ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #

$/ = "\n"; # Use a newline character as an input record separater.

my $in_header1 = <$fh_in1>;
my $indices_fld = tfm_labels2indices_gen (@TAGS_IN_FLD, $in_header1, "\t");
unless (defined $indices_fld) {
	$in_header1 =~ s/\s+$//; 
	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the input header:\n";
	print STDOUT " '$in_header1' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
my $max_indx = 0;
foreach my $indx (@{$indices_fld}) { ($max_indx <$indx) and ($max_indx = $indx); }

		# ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #
my @indices_fld1 = ();
foreach my $indx (@{$indices_fld}) { ($indx == $max_indx) or (push @indices_fld1, $indx); }
		# END of "ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'." #


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
print $fh_out join ("\t", @TAGS_OUT_FLD), "\n\n";
=cut


my $totct_segs = my $ct_anom = my $ct_w_undef = my $ct_fail = my $ct_succ = 0;
my $ct_skipped1 = my $ct_skipped2 = my $ct_anom2 = 0;
my $ct_skipped3 = my $ct_skipped5  = 0;

my $ct_skipped_x = 0;

my $ct_anom3 = 0; # ADDED in creation of 'summary_cmp_ref_vs_rec_ğconf_stats.bulk_pars.pl'. #

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
my $totct_all_cblks = 0; # ADDED in creation of 'summary_classify_msa_errors1.raw.pl' (on Oct 26 (Mon), 2015). #
my $totct_equiv_indels = my $totct_eff_cblks = 0; # ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. #
##my %type_error2sbtotct_ecblks; # = ($type_error => $sub_total_ct_eff_cblks, ...) # ADDED in creation of 'summary_classify_msa_errors.bulk_pars.pl'. 
=cut

=For the predecessor 'summary_examine_purge.raw.pl'.

	## ADDED in creation of 'summary_examine_purge.raw.pl'. ##

my $ct_w_fractional = 0;

my %br2freq_prgs; # = ($br_id => $abs_freq_purges, ...)
my %ct_invlvd2freq_prgs; # = ($ct_invlvd_seqs => $abs_freq_purges, ...)

	## END of "ADDED in creation of 'summary_examine_purge.raw.pl'." ##

	## ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016). ##

my %blsize2bins_ct_invlvd_vs_pv; # = ($block_size => \@bins_ct_invlvd_vs_pv_sp, ...) (Here, '_sp' indicates that the distribution is for a specific $block_size.) # 
my %blsize2bins_brlen_vs_pv; # = ($block_size => \@bins_brlen_vs_pv_sp, ...)

	## END Of "ADDED when modifying 'summary_examine_purge.raw.pl' (on May 16, 2016)." ##

=cut # END of "For the predecessor 'summary_examine_purge.raw.pl'."

			### MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. ###

	## MODIFIED in creation of 'summary_examine_complex.5.raw.pl'. ##
my @first_gpd_seg = ();
##my @last_cluster = ();

my @first_eventset = (); # element = [$br, $stat_ue, $beg, $end, $len] # ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #

while (my $in = <$fh_in1>) {
	($in =~ m/^\S/) or next;
	$in =~ s/\s+$//;
	my @in_flds = split /\t/, $in;
	unless (@in_flds > $max_indx) {
	    print STDOUT "[WARNING '$program_name'] An ANOMALOUS line was found:\n";
	    print STDOUT "    '$in' : $! !\n\n";
	    $ct_anom++;
	    next;
	}
	$totct_segs++;
	my @rlv_flds = @in_flds[@{$indices_fld}];

		## MODIFIED in creation of 'summary_examine_complex.5.raw.pl'. ##
	@first_gpd_seg = @rlv_flds;
	##push @last_cluster, \@rlv_flds;

	foreach my $ev (@in_flds[$max_indx .. $#in_flds]) { # ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #
	    my @evinfo = split /\//, $ev;
	    push @first_eventset, \@evinfo;
	} 					# END of "ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'." #

	last;
}

my $ct_clusters = 0;
my $end_coord_last_cls ;
##my $last_id_class = my $last_session = 'Pre-start'; # ADDED for DEBUGGING. #

	## MODIFIED in creation of 'summary_examine_complex.5.raw.pl'. ##
while (@first_gpd_seg>0) { # Outer while-loop (until the gapped segments run out). #
##while (@last_cluster>0) { # Outer while-loop (until the last cluster runs out). #

			### END of "MODIFIED in creation of 'summary_examine_complex.2.raw.pl'." ###

=DISABLED in creation of 'summary_examine_complex.2.raw.pl'.
while (my $in = <$fh_in1>) { # Outer while-loop (over erroneous MSA segments) #
=cut

		### MAJOR MODIFICATION in creation of 'summary_examine_complex.5.raw.pl'. (IIa) ###

	my ($id_class0, $session0) = @first_gpd_seg[0,1];
	my @cp_first_gpd_seg = @first_gpd_seg;
	my @gpd_segs_in_msa = (\@cp_first_gpd_seg);
	@first_gpd_seg = ();

		# ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #
	my @cp_first_eventset = @first_eventset;
	my @eventsets_in_msa = (\@cp_first_eventset);
	@first_eventset = ();
		# END of "ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'." #

	while (my $in = <$fh_in1>) { # 1st middle preliminary while-loop (over gapped segments). #

	    ($in =~ m/^\S/) or next;
	    $in =~ s/\s+$//;

	    my @in_flds = split /\t/, $in;

	    unless (@in_flds > $max_indx) {
		print STDOUT "[WARNING '$program_name'] An ANOMALOUS line was found:\n";
		print STDOUT "    '$in' : $! !\n\n";
		$ct_anom++;
		@in_flds = (); # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #
		next;
	    }

	    $totct_segs++;

	    my @rlv_flds = @in_flds[@{$indices_fld}];
	    my ($id_class, $session) = @rlv_flds[0,1];

		# ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #
	    my @eventset = (); # element = [$br, $stat_ue, $beg, $end, $len]
	    foreach my $ev (@in_flds[$max_indx .. $#in_flds]) {
		my @evinfo = split /\//, $ev;
		push @eventset, \@evinfo;
	    }
		# END of "ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'." #

	    if (($id_class eq $id_class0) and ($session eq $session0)) {

		push @gpd_segs_in_msa, \@rlv_flds;
		push @eventsets_in_msa, \@eventset; # ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #

	    } else {

		@first_gpd_seg = @rlv_flds;
		@first_eventset = @eventset; # ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #
		last;
	    }

	} 		# End of the 1st middle preliminary while-loop (over gapped segments). #




			# Pre-process the records of gapped segments in the subject MSA. #

			# => Each element of @gpd_segs_in_msa will become:
			# [$category, $beg_coord, $end_coord, $ct_clms, $totct_indel, $diffct_indel, $totlen_indel, $difflen_indel]
	my $ct_gpd_segs = @gpd_segs_in_msa;

   (print STDOUT "<< Class: $id_class0 \; MSA ID: $session0 \; #{gapped segments} = $ct_gpd_segs >>\n\n") if ($TEST_LEVEL> 0);

	for (my $s = 0; $s <$ct_gpd_segs; $s++) { # 2nd middle for-loop (over gapped segments in the subject MSA). #
	    my $rlv_flds = $gpd_segs_in_msa[$s];
 	    my ($id_class, $session, $segid, $type, $diagnosis, $ct_gpd_segs, $indx_gpd_seg, $beg_coord, $end_coord, $ct_clms, $ct_ins, $ct_del, $totlen_ins, $totlen_del, $ct_gaps) = @{$rlv_flds};


				# ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'. #

	    foreach my $evinfo (@{$eventsets_in_msa[$s]}) { # Inner for-loop (over events in the subject gapped segment). #

			# Switch the branch to its equivalent if the former has more external offsprings than the latter. #

		my $br = $evinfo->[0];
		my $eq_br = $equiv_branches{$br};

		if ((defined $eq_br) and ($node2ct_exoffs->{$br} > $node2ct_exoffs->{$eq_br})) {
		    my $len_ev = $evinfo->[4];
		    $evinfo->[0] = $eq_br;
		    if ($evinfo->[1] eq $CHAR_ABS) {

				# Switch an insertion into a deletion. #
			$evinfo->[1] = $CHAR_PRES ;
			$ct_ins--;
			$ct_del++;
			$totlen_ins -= $len_ev;
			$totlen_del += $len_ev;

		    } else {
				# Switch a deletion into an insertion. #
			$evinfo->[1] = $CHAR_ABS ;
			$ct_ins++;
			$ct_del--;
			$totlen_ins += $len_ev;
			$totlen_del -= $len_ev;
		    }
		}
	    }				# End of the inner for-loop (over events in the subject gapped segment). #

				# END of "ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'." #


	    my $totct_indel = $ct_ins + $ct_del;
	    my $diffct_indel = $ct_ins - $ct_del;
	    my $totlen_indel = $totlen_ins + $totlen_del;
	    my $difflen_indel = $totlen_ins - $totlen_del;

	    my $combo_id = join (':', @{$rlv_flds}[0,1,2]);

	    my $category = 'B'; # Default category. #

	    if ($type eq 'Correct') {

		# Do nothing. #

	    } elsif ((defined $if_complex{$combo_id}) or ((defined $if_na{$combo_id}) or ($diagnosis =~ m/TOO_LONG/)) ) {

		# Type A = 'Complex errors' + 'Too-long' #

		$category = 'A';
		##$category = 'B'; # A BUG fixed in creation of 'summary_examine_complex.5.raw.pl'. #

	    } elsif ($type eq 'Erroneous') {

		# Do nothing. #

	    } else {
		print STDOUT "[CAUTION '$program_name'] (ID_Class, Session_ID, Segment_ID) = ($id_class, $session, $segid) has either ANOMALOUS TYPE (=$type) or ANOMALOUS DIAGNOSIS (=$diagnosis): $! ! ... I will skip processing this MSA ...\n\n";

		$ct_anom2++;
		next OUTER;
	    }

	    $gpd_segs_in_msa[$s] = [$category, $beg_coord, $end_coord, $ct_clms, $totct_indel, $diffct_indel, $totlen_indel, $difflen_indel];

	}				# End of the 2nd middle for-loop (over gapped segments in the subject MSA). #


   if ($TEST_LEVEL>2) {
	print STDOUT "\n", '[ Content of @gpd_segs_in_msa ]', "\n\n";
	if ($TEST_LEVEL>3) {
	    print STDOUT join ("\t", 'Indx', 'A/B', 'Beg', 'End', '#{columns}', 'Tot#{indels}', '#{ins}-#{del}', 'Totlen{indels}', 'Len{ins}-Len{del}', '[br/stat_ue/beg/end/len]_event1', '...'), "\n\n";
	} else {
	    print STDOUT join ("\t", 'Indx', 'A/B', 'Beg', 'End', '#{columns}', 'Tot#{indels}', '#{ins}-#{del}', 'Totlen{indels}', 'Len{ins}-Len{del}'), "\n\n";
	}

	for (my $gs = 0; $gs<$ct_gpd_segs; $gs++) {
	    my @out_seginfo = @{$gpd_segs_in_msa[$gs]} ;
	    if ($TEST_LEVEL>3) {
	        foreach my $evinfo (@{$eventsets_in_msa[$gs]}) { push @out_seginfo, join ('/', @{$evinfo}); }
	    }
	    print STDOUT join ("\t", $gs, @out_seginfo), "\n";
	}
	print STDOUT "\n\n";
   }


=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (IIa)

	## ($totct_segs>=10) and last; ### For TESTING the script ###

	my @in_flds = (); # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

    while (my $in = <$fh_in1>) { # Middle preliminary while-loop (over erroneous MSA segments). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	($in =~ m/^\S/) or next;
	$in =~ s/\s+$//;

	##$totct_segs++;

	@in_flds = split /\t/, $in; # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #
	##my @in_flds = split /\t/, $in;
	unless (@in_flds > $max_indx) {
	    print STDOUT "[WARNING '$program_name'] An ANOMALOUS line was found:\n";
	    print STDOUT "    '$in' : $! !\n\n";
	    $ct_anom++;
	    @in_flds = (); # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #
	    next;
	}

	$totct_segs++; # MOVED from above. #

	last; # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

    } 		# End of the middle preliminary while-loop (over erroneous MSA segments). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	my $last_combo_id = join (':', @{$last_cluster[0]}[0,1,2]); # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	my @rlv_flds = (@in_flds>0) ? @in_flds[@{$indices_fld}] : () ; # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (IIa)"


			# MAJOR MODIFICATION in creation of 'summary_examine_complex.1.raw.pl'. #

=To be moved below in creation of 'summary_examine_complex.2.raw.pl'.
	my ($id_class, $session, $segid, $type, $diagnosis, $ct_gpd_segs, $indx_gpd_seg, $beg_coord, $end_coord, $ct_clms, $ct_ins, $ct_del, $totlen_ins, $totlen_del, $ct_gaps) = @in_flds[@{$indices_fld}];

		## @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}');
=cut



=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (IIb)

		## Check if the current segment belongs to the last cluster or not. ##
		## (ADDED in creation of 'summary_examine_complex.2.raw.pl'.) ##

   ##if (@rlv_flds>0) { # DISABLED for DEBUGGING. #

	my $combo_id = (@rlv_flds>0) ? join (':', @rlv_flds[0,1,2]) : undef; # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #
	##my $combo_id = "$id_class:$session:$segid";

    if (defined $combo_id) { # ADDED for DEBUGGING. #
	if ($last_combo_id eq $combo_id) { # Incorporate the current segment into the last cluster.
	    push @last_cluster, \@rlv_flds;
	    next;
	}
    } # ADDED for DEBUGGING. #
    ##} # DISABLED for DEBUGGING. #

	$ct_clusters++;

		## END of "Check if the current segment belongs to the last cluster or not." ##

		## Initialize the important attributes of the last cluster. ##
		## (ADDED in creation of 'summary_examine_complex.2.raw.pl'.) ##

	my ($id_class, $session, $segid, $type, $diagnosis, $ct_gpd_segs, $beg_coord_cls) = @{$last_cluster[0]}[0..5,7];
	my $end_coord_cls = $last_cluster[$#last_cluster]->[8];

	unless (@last_cluster == $ct_gpd_segs) { # A Consistency check. #

	    print STDOUT "[CAUTION '$program_name'] The size of the last cluster (=", scalar (@last_cluster), ") CLASHES with #{gapped segments in it} (=$ct_gpd_segs) for (ID_Class, Session_ID, Segment_ID) = ($id_class, $session, $segid) : $! ! ... I will skip the analysis ...\n\n";
	    $ct_anom3++;
	    if (@rlv_flds>0) {
		@last_cluster = (\@rlv_flds);
		$end_coord_last_cls = $end_coord_cls;
		($last_id_class, $last_session) = ($id_class, $session); # ADDED for DEBUGGING. #
	    } else {
		@last_cluster = ();
	    }
	    next;
	}		# END of "A Consistency check." #

=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (IIb)"


=For the predecessor 'summary_examine_complex.2.raw.pl'. (ii)
			# Calculate the span of this cluster of gapped segments. #

	my $beg_coord_next_cls = (@rlv_flds>0) ? $rlv_flds[7] : undef;

	my $span_cluster;

	my ($id_class1, $session1) = @rlv_flds[0,1]; # ADDED for DEBUGGING. #

	if (($last_id_class eq $id_class) and ($id_class eq $id_class1) and ($last_session eq $session) and ($session eq $session1) and (defined $end_coord_last_cls) and (defined $beg_coord_next_cls)) { # Further MODIFIED for DEBUGGING. #
	##if (($id_class eq $id_class) and ($session eq $session1) and (defined $end_coord_last_cls) and (defined $beg_coord_next_cls)) { # MODIFIED for DEBUGGING. # ...}
	##if ((defined $end_coord_last_cls) and (defined $beg_coord_next_cls)) {...}
	
	   my $coord_left_border = ($beg_coord_cls + $end_coord_last_cls)/2 ;
	   my $coord_right_border = ($end_coord_cls + $beg_coord_next_cls)/2 ;
	   $span_cluster = $coord_right_border - $coord_left_border ;

	   if ($span_cluster < 0) { # Alarm if the span is negative. #
		print STDOUT "[WARNING '$program_name'] The cluster '$combo_id' has a NEGATIVE SPAN (=$span_cluster):  (end_last_cls, beg_cls, end_cls, beg_next_cls) = ($end_coord_last_cls, $beg_coord_cls, $end_coord_cls, $beg_coord_next_cls): $! !\n\n";
	   }
	}

			# END of "Calculate the span of this cluster of gapped segments." #

=cut # END of "For the predecessor 'summary_examine_complex.2.raw.pl'. (ii)"

=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (III)

		### ADDED in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

	my $max_totct_indel = 0;
	my $maxct_clms = my $max_totlen_indel = 0;
	my ($max_diffct_indel, $min_diffct_indel);

    foreach my $segment_info (@last_cluster) { # Main middle foreach-loop (over segments in the last cluster). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

	my ($id_class1, $session1, $segid1, $type1, $diagnosis1, $ct_gpd_segs1, $indx_gpd_seg, $beg_coord, $end_coord, $ct_clms, $ct_ins, $ct_del, $totlen_ins, $totlen_del, $ct_gaps) = @{$segment_info}; # MOVED from above in creation of 'summary_examine_complex.2.raw.pl'. #

	my $totct_indel = $ct_ins + $ct_del;
	my $diffct_indel = $ct_ins - $ct_del;
	my $totlen_indel = $totlen_ins + $totlen_del;
	my $difflen_indel = $totlen_ins - $totlen_del;

	if ($totct_indel > $max_totct_indel) {

		# Update the set of "largest" segments. #

	    $max_totct_indel = $totct_indel;

	    $maxct_clms = $ct_clms;
	    $max_totlen_indel = $totlen_indel;
	    $max_diffct_indel = $min_diffct_indel = $diffct_indel;

	} elsif ($totct_indel == $max_totct_indel) {

		# Incorporate the current segment into the set of "largest" segments. # 

	    ($ct_clms > $maxct_clms) and ($maxct_clms = $ct_clms);
	    ($totlen_indel > $max_totlen_indel) and ($max_totlen_indel = $totlen_indel);
	    ((defined $max_diffct_indel) and ($max_diffct_indel >= $diffct_indel)) or ($max_diffct_indel = $diffct_indel);
	    ((defined $min_diffct_indel) and ($min_diffct_indel <= $diffct_indel)) or ($min_diffct_indel = $diffct_indel);

	}

   }			# End of the main middle foreach-loop (over segments in the last cluster). # ADDED in creation of 'summary_examine_complex.2.raw.pl'. #

=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (III)"


=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (IV)

		## Update @last_cluster. ##
		## (ADDED in creation of 'summary_examine_complex.2.raw.pl'.)

	if (@rlv_flds>0) {
	    @last_cluster = (\@rlv_flds);
	    $end_coord_last_cls = $end_coord_cls;
	    ($last_id_class, $last_session) = ($id_class, $session); # ADDED for DEBUGGING. #
	} else {
	    @last_cluster = ();
	}
		## END of "Update @last_cluster." ##

=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (IV)"


=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (V)

			### MAJOR ADDITION in creation of 'summary_examine_complex.2_2D.raw.pl'. ###

			# Default: Type B = 'non-complex errors' + 'correctly reconstructed'. #

	my ($bins_max_totct_indels_vs_maxct_columns, $bins_max_totct_indels_vs_max_totlen_indels,
		$bins_max_totct_indels_vs_max_diffct_indels, $bins_max_totct_indels_vs_min_diffct_indels)
	   = ($bins_max_totct_indels_vs_maxct_columns_B, $bins_max_totct_indels_vs_max_totlen_indels_B,
		$bins_max_totct_indels_vs_max_diffct_indels_B, $bins_max_totct_indels_vs_min_diffct_indels_B) ;

	if ($type eq 'Correct') {

		# Do nothing. #

	} elsif ((defined $if_complex{$last_combo_id}) or ((defined $if_na{$last_combo_id}) or ($diagnosis =~ m/TOO_LONG/)) ) {

		# Type A = 'Complex errors' + 'Too-long' #

	    ($bins_max_totct_indels_vs_maxct_columns, $bins_max_totct_indels_vs_max_totlen_indels,
		$bins_max_totct_indels_vs_max_diffct_indels, $bins_max_totct_indels_vs_min_diffct_indels)
	     = ($bins_max_totct_indels_vs_maxct_columns_A, $bins_max_totct_indels_vs_max_totlen_indels_A,
		$bins_max_totct_indels_vs_max_diffct_indels_A, $bins_max_totct_indels_vs_min_diffct_indels_A) ;

	} elsif ($type eq 'Erroneous') {

		# Do nothing. #

	} else {
	    print STDOUT "[CAUTION '$program_name'] (ID_Class, Session_ID, Segment_ID) = ($id_class, $session, $segid) has either ANOMALOUS TYPE (=$type) or ANOMALOUS DIAGNOSIS (=$diagnosis): $! ! ... I will skip tallying this segment ...\n\n";

	    $ct_anom2++;
	    next;
	}

=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (V)"


			## Prepare an essential tool for the clustering. ##
			## (ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'.) ##

	my @set_self_nn_events = (); # %{$set_self_nn_events[$gs]} = ($br2 => {$br1 => \@set_eventpairs, ...}, ..), where @{$set_eventpairs[$ep]} = ($gs, $indx_ev1, $gs, $indx_ev2) for the $ep th pair of events occurring in the nearest-neighbor branch pair, ($br1, $br2). #

=DISABLED in creation of 'summary_examine_complex.6p_aug_len.raw.pl'. (I)
	for (my $gs = 0; $gs < $ct_gpd_segs; $gs++) { # Preliminary outer-middle for-loop (over gapped segments in the MSA). #

	    my $eventset = $eventsets_in_msa[$gs];

		# @{$on_coinc_brs[$k]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $k th pair of events along the same branch ($br1 or $br2), with
		# @{$on_nn_brs[$m]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $m th pair of events along the nearest-neighboring branches ($br1 and $br2), with
		#
	    my ($on_coinc_brs, $on_nn_brs) = selfcompare_indel_hosting_branches (@{$eventset}, %equiv_branches, %br2nn);

	    my %self_nn_events;
	    foreach my $pairinfo0 (@{$on_nn_brs}) { # Middle for-loop (over event pairs). #
		my ($br01, $br02, $indx_ev01, $indx_ev02) = @{$pairinfo0};
		##foreach my $pairinfo ([$br01, $br02, $indx_ev01, $indx_ev02]) {
		foreach my $pairinfo ([$br01, $br02, $indx_ev01, $indx_ev02],
				[$br02, $br01, $indx_ev02, $indx_ev01]) {

		    my ($br1, $br2, $indx_ev1, $indx_ev2) = @{$pairinfo};
		    my $hash = $self_nn_events{$br2};
		    (defined $hash) or ($hash = $self_nn_events{$br2} = {});
		    my $array = $hash->{$br1};
		    (defined $array) or ($array = $hash->{$br1} = []);
		    push @{$array}, [$gs, $indx_ev1, $gs, $indx_ev2];
		}
		##}
	    }				# End of the middle for-loop (over eventpairs). #

	    $set_self_nn_events[$gs] = \%self_nn_events;

	}				# End of the preliminary outer-middle for-loop (over gapped segments in the MSA). #
=cut # END of "DISABLED in creation of 'summary_examine_complex.6p_aug_len.raw.pl'. (I)"


			## END of "(ADDED in creation of 'summary_examine_complex.6p_aug.raw.pl'.)" ##

			### MAJOR MODIFICATION in creation of 'summary_examine_complex.5.raw.pl'. (VI) ###
			### => Further  MAJOR MODIFICATION in creation of 'summary_examine_complex.6p.raw.pl'. (II) ###

			# Take statistics on the artificially clustered gapped segments 
			# with different THRESHOLDS for the spacer sizes. #

	for (my $i_thr = 0; $i_thr < $CT_THRSHS_SPSIZE; $i_thr++) { # 3rd main outer-middle for-loop (over the threshoulds for the spacer sizes). #

			# MODIFIED in creation of 'summary_examine_complex.6p.raw.pl'. (IIa) #
	    my $ctpair2bins_pcA = $set_ctpair2bins_pcA[$i_thr];
	    my $bins_pcA_total = $set_bins_pcA_total[$i_thr];
	    ##my $bins_pcA_vs_totct_clms = $set_bins_pcA_vs_totct_clms[$i_thr];
	    ##my $bins_pcA_vs_tot_totct_indel = $set_bins_pcA_vs_tot_totct_indel[$i_thr];

				### MAJOR MODIFICATION in creation of 'summary_examine_complex.6p_aug.raw.pl'. ###

	    my $thrsh_spsize0 = $THRSHS_SPSIZE[$i_thr];
			# If the {total spacer size} <= $thrsh_spsize,
			# then the two gapped segments will be put into the same cluster.
			# Here, the {total spacer size} is the summation of the number of intervening gapless columns
			# and the number of columns only undergoing irrelevant deletions.

   (print STDOUT "<  Clustering with {Basic Spacer-Size Threshold} = $thrsh_spsize0 >\n\n") if ($TEST_LEVEL>1);

			# MOVED from within the 1st middle while-loop, to fix a BUG. #
	    my %gs2prev_nn_events; # = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...), with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events). #
			# END of "MOVED from within the 1st middle while-loop, to fix a BUG." #

	    my $last_cluster = [0];

=For DEBUGGING (1)
	    $last_cluster = [14];
	    $ct_gpd_segs = 22;
=cut
=For DEBUGGING (2)
	    $last_cluster = [35];
	    $ct_gpd_segs = 42;
=cut

	    my @clusters = ($last_cluster);

	    while ($last_cluster->[$#{$last_cluster}] < $ct_gpd_segs-1) { # 1st middle while-loop (untill all gapped segments are examined). #

   if ($TEST_LEVEL_DBG>0) {
	print STDOUT "\n{{{ Initial 'last' cluster = (", join (', ', @{$last_cluster}), ") }}}\n\n";
   }

=MOVED to before the 1st middle while-loop, to fix a BUG.
		my %gs2prev_nn_events; # = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...), with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events). #
=cut

		my $indx_c_gs = 0;
		while ($indx_c_gs < @{$last_cluster}) { # Inner-middle while-loop (until the "last" cluster is exhausted). #


	            my $c_gs = $last_cluster->[$indx_c_gs]; # Current gapped segment. #
	            my ($c_beg_coord, $c_end_coord) = @{$gpd_segs_in_msa[$c_gs]}[1,2];
		    my $c_gs_size = $c_end_coord - $c_beg_coord + 1;

   if ($TEST_LEVEL_DBG>0) {
	print STDOUT "\n{{ The 'Current' segment = the $indx_c_gs th segment in the 'last' cluster (= the $c_gs th segment in the entire MSA) ; [beg, end] = [$c_beg_coord, $c_end_coord] ; size = $c_gs_size }}}\n\n";
   }
		    my $c_eventset = $eventsets_in_msa[$c_gs];

			# push @eventsets_in_msa, \@eventset; element of @eventset = [$br, $stat_ue, $beg, $end, $len] 

		    my $c_self_nn_events = $set_self_nn_events[$c_gs]; # = {$br2 => {$br1 => \@set_eventpairs, ...}, ...}, where @{$set_eventpairs[$ep]} = ($c_gs, $indx_ev1, $c_gs, $indx_ev2) for the $ep-th pair of nearest-neighboring events. #

=DISABLED because these conditions are NOT what I meant. (1)
		    my $thrsh_spsize_sht = ($thrsh_spsize0 > $c_gs_size / $MAG_THRSH) ? $thrsh_spsize0 : $c_gs_size / $MAG_THRSH ;
		    ($thrsh_spsize_sht > $UPPER_BOUND_SPSIZE) and ($thrsh_spsize_sht = $UPPER_BOUND_SPSIZE);
		    my $thrsh_spsize_lng = ($thrsh_spsize0 > $c_gs_size * $MAG_THRSH) ? $thrsh_spsize0 : $c_gs_size * $MAG_THRSH ;
		    ($thrsh_spsize_lng > $UPPER_BOUND_SPSIZE) and ($thrsh_spsize_lng = $UPPER_BOUND_SPSIZE);
=cut

		    my $tot_spsize = 0;

		    my $last_n_end_coord = $c_end_coord; # A BUG-FIXATION measure (A-1). #

		    for (my $n_gs = $c_gs+1; $n_gs < $ct_gpd_segs; $n_gs++) { # Inner for-loop (over "next" gapped segments). #

			my ($n_beg_coord, $n_end_coord) = @{$gpd_segs_in_msa[$n_gs]}[1,2];
			my $n_gs_size = $n_end_coord - $n_beg_coord + 1;

     if ($TEST_LEVEL_DBG>0) {
	print STDOUT "\n{ The 'Next' segment = the $n_gs th segment in the entire MSA ;  [beg, end] = [$n_beg_coord, $n_end_coord] ; size = $n_gs_size }}\n\n";
     }

				# A BUG-FIXATION measure (A-2). #
			$tot_spsize += $n_beg_coord - $last_n_end_coord - 1;
			$last_n_end_coord = $n_end_coord;
			##$tot_spsize += $n_beg_coord - $c_end_coord - 1;
			####my $tot_spsize += $n_beg_coord - $c_end_coord - 1; # A BUG fixed on July 15 (Fri), 2016. #

    if ($TEST_LEVEL_DBG>1) {
	print STDOUT "(Total spacer size) = $tot_spsize .\n\n";
    }

				## Replace the conditions with what I really meant. (1) ##
			($tot_spsize > $UPPER_BOUND_SPSIZE) and last; # In this case, we give up extending the "last" cluster based on the current gapped segment. #
			## ($tot_spsize > $thrsh_spsize_lng) and last; # In this case, we give up extending the "last" cluster based on the current gapped segment. # ... OBSOLETE...

			my $n_eventset = $eventsets_in_msa[$n_gs];
			    # push @eventsets_in_msa, \@eventset; element of @eventset = [$br, $stat_ue, $beg, $end, $len] 

			# @{$on_coinc_brs->[$k]} = ($br_c, $br_n, $indx_c_ev, $indx_n_ev) for the $k th pair of events along the same branch.
			# @{$on_nn_brs->[$m]} = ($br_c, $br_n, $indx_c_ev, $indx_n_ev) for the $m th pair of events along the nearest-neighboring branches.
			# @{$irrlv_insertions_n} stores the indices of the insertions in the "next" gapped segment along the branches that neither coincide with, nor is the nearest-neighbors of, any "current" branches.
			# 
			my ($on_coinc_brs, $on_nn_brs, $irrlv_insertions_n)
				= compare_indel_hosting_branches (@{$c_eventset}, @{$n_eventset}, %equiv_branches, %br2nn, $CHAR_ABS, %{$node2pa});

				# %br2nn; # = ($br_id => {$id_nearest_neighbor => 1, ...}, ...)

			my $incl_spsize = $n_gs_size; # The increment of the spacer size. #

			foreach my $indx_ev (@{$irrlv_insertions_n}) { $incl_spsize -= $n_eventset->[$indx_ev][4]; }

   if ($TEST_LEVEL_DBG>2) {
	print STDOUT "(Increment of the spacer-size that will be originated from the 'next' segment) = $incl_spsize .\n\n";
   }

			##$tot_spsize += $incl_spsize; # The "next" branch will be incorporated into the "spacer", after excluding the irrelevant insertions. # A BUG ... it was too early to incorporate it into the "spacer"!!

				## Replace the conditions with what I really meant. (2) ##

			my ($max_gs_size, $min_gs_size) = ($n_gs_size > $c_gs_size) ?
					($n_gs_size, $c_gs_size) : ($c_gs_size, $n_gs_size) ;

			unless (($tot_spsize <= $thrsh_spsize0) or 
				(($max_gs_size >= $tot_spsize * $MAG_THRSH) and 
				($min_gs_size >= $tot_spsize / $MAG_THRSH)) ) {
				##($max_gs_size >= $tot_spsize * $MAG_THRSH) or
				##($min_gs_size >= $tot_spsize / $MAG_THRSH) ) {

=DISABLED because these conditions are NOT what I meant. (2)
			unless (($tot_spsize <= $thrsh_spsize0) or ($thrsh_spsize_lng * $MAG_THRSH <= $n_gs_size) or 
			    (($thrsh_spsize_sht / $MAG_THRSH <= $n_gs_size) and ($tot_spsize <= $thrsh_spsize_sht)) ) {
=cut

				# In this case, this "next" segment is NEVER considered as interacting with the current segment.

			    $tot_spsize += $incl_spsize; # A measure to fix the BUG which is one block above. #

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (A) ...\n\n") if ($TEST_LEVEL_DBG>2);

			    next;
			}

			$tot_spsize += $incl_spsize; # A measure to fix the BUG which is two blocks above. #

=DISABLED in creation of 'summary_examine_complex.6p_aug_len.raw.pl'. (II)
			if (@{$on_nn_brs}>0) {

				# Store the information on the interacting event pairs for later use. #

			    my $br2br2nn_events = $gs2prev_nn_events{$n_gs};
			    (defined $br2br2nn_events) or ($br2br2nn_events = $gs2prev_nn_events{$n_gs} = {});

			    foreach my $info_evpair (@{$on_nn_brs}) { # 2nd inner-inner foreach-loop (over nn event pairs). #
				my ($br_c, $br_n, $indx_c_ev, $indx_n_ev) = @{$info_evpair};
				my $br2nn_events = $br2br2nn_events->{$br_n};
				(defined $br2nn_events) or ($br2nn_events = $br2br2nn_events->{$br_n} = {});
				my $nn_events = $br2nn_events->{$br_c};
				(defined $nn_events) or ($nn_events = $br2nn_events->{$br_c} = []);
				push @{$nn_events}, [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev];
			    
			    }			# End of the 2nd inner-inner foreach-loop (over nn event pairs). #
				## my %gs2prev_nn_events; # = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...), with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events). #

			} # END of "if (@{$on_nn_brs}>0) {...}" #
=cut # END of "DISABLED in creation of 'summary_examine_complex.6p_aug_len.raw.pl'. (II)"


			my $last_gs_in_cls = $last_cluster->[$#{$last_cluster}];

			unless ($last_gs_in_cls < $n_gs) {

   (print STDOUT " ... The 'next' segment is already incorporated in the cluster. Thus, skip the subsequent analyses ...\n\n") if ($TEST_LEVEL_DBG>2);

			    next; # In this case, $n_gs is already incorporated in the cluster, thus we don't have to examine further. #
			}

				### Examine whether or not this "next" segment can interact with the current segment. ###

				## REPLACED the type (II) conditions with a more lenient one,
				## in creation of 'summary_examine_complex.6p_aug_len.raw.pl'. ##*

			if (@{$on_coinc_brs}>0) {

				# Interaction between $n_gs and $c_gs (I). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (I) ...\n\n") if ($TEST_LEVEL_DBG>2);

			    push @{$last_cluster}, ($last_gs_in_cls + 1 .. $n_gs);

   if ($TEST_LEVEL_DBG>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

			    next;

			} elsif (@{$on_nn_brs}>0) {

				# Interaction between $n_gs and $c_gs (II-len). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (II-len) ...\n\n") if ($TEST_LEVEL_DBG>2);

			    push @{$last_cluster}, ($last_gs_in_cls + 1 .. $n_gs);

   if ($TEST_LEVEL_DBG>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

			    next;

			} else {

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (C) ...\n\n") if ($TEST_LEVEL_DBG>2);

			}

=DISABLED in creation of 'summary_examine_complex.6p_aug_len.raw.pl'. (III)

				# First, examine the triangular interaction involving a past segment. #
				# (MOVED from below and augmented, to fix a BUG) #
			my $br2br2prev_nn_events = $gs2prev_nn_events{$c_gs};
			(defined $br2br2prev_nn_events) or ($br2br2prev_nn_events = {});

			my ($lm_igs, $rm_igs); # The indices for the leftmost and rightmost 'past' gapped segments interacting with $c_gs and $n_gs. 
			foreach my $info_evpair (@{$on_nn_brs}) { # 3rd inner-inner foreach-loop (over nn event pairs). #
			    my ($br_c, $br_n, $indx_c_ev, $indx_n_ev) = @{$info_evpair};
			    ##my $nn_c = $br2nn{$br_c}; # It's probably unnecessary any more. #
			    my $nn_n = $br2nn{$br_n};
				# %br2nn; # = ($br_id => {$id_nearest_neighbor => 1, ...}, ...)

			    my $br2prev_nn_events = $br2br2prev_nn_events->{$br_c};
			    (defined $br2prev_nn_events) or next; # Skip if the event along $br_c in $c_gs has no past neighboring events. #
			    foreach my $br (keys %{$br2prev_nn_events}) { # Outer-innermost foreach-loop (over branches hosting previous events nearest-neighboring $br_c. #
				(defined $nn_n->{$br}) or next; # Skip if $br is not the nearest-neighbor of $br_n as well. #
				my $prev_nn_events = $br2prev_nn_events->{$br};
				foreach my $pev_info (@{$prev_nn_events}) { # Inner-innermost foreach-loop (over the past interacting events). #
				    my ($p_igs, $indx_p_iev, $c_gs2, $indx_c_ev2) = @{$pev_info};
				    ((defined $lm_igs) and ($lm_igs <= $p_igs)) or ($lm_igs = $p_igs);
				    ((defined $rm_igs) and ($p_igs <= $rm_igs)) or ($rm_igs = $p_igs);
				}				# End of the inner-innermost foreach-loop (over the past interacting events). #
			    }				# End of the outer-innermost foreach-loop (over branches hosting previous events nearest-neighboring $br_c). #
			}				# End of the 5th inner-inner foreach-loop (over nn event pairs). #

			if (defined $lm_igs) {
				# Interaction between $n_gs and $c_gs (II-b). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (II-b) ...\n\n") if ($TEST_LEVEL_DBG>2);

					# MODIFIED to FIX a BUG (i). #
			    my $counter = 0;
			    while ($lm_igs < $last_cluster->[0]) { $last_cluster = pop @clusters; $counter++; } # Get back to older clusters until the updated $last_cluster accommodates $lm_igs. #
			    ##while ($lm_igs < $last_cluster->[0]) { $last_cluster = pop @clusters; } # Get back to older clusters until the updated $last_cluster accommodates $lm_igs. #

			    my ($gs_lm_in_lc, $gs_rm_in_lc) = @{$last_cluster}[0, $#{$last_cluster}];

				# Update the $last_cluster and the $indx_c_gs. #
			    push @{$last_cluster}, ($gs_rm_in_lc + 1 .. $n_gs);

					# MODIFIED to FIX a BUG (ii). #
			    ($counter>0) and (push @clusters, $last_cluster);
			    ##push @clusters, $last_cluster;
			    $indx_c_gs = $c_gs - $gs_lm_in_lc;

   if ($TEST_LEVEL_DBG>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n";
	print STDOUT '    $indx_c_gs  = ', $indx_c_gs, " .\n\n\n";
   }

			    next;
			}

				# END of "(MOVED from below and augmented, to fix a BUG)" #

			if (@{$on_coinc_brs}>0) {

				# Interaction between $n_gs and $c_gs (I). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (I) ...\n\n") if ($TEST_LEVEL_DBG>2);

			    push @{$last_cluster}, ($last_gs_in_cls + 1 .. $n_gs);

   if ($TEST_LEVEL_DBG>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

			    next;
			}

			unless (@{$on_nn_brs}>0) {
				# In this case, this "next" segment is NEVER considered as interacting with the current segment.

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (B) ...\n\n") if ($TEST_LEVEL_DBG>2);
			    next;
			}

				## In the following, $n_gs is regardes as interacting with $c_gs
				## ONLY IF the nearest-neighboring events form a triangle encompassing a node.

		        my $n_self_nn_events = $set_self_nn_events[$n_gs]; # = {$br2 => {$br1 => \@set_eventpairs, ...}, ...}, where @{$set_eventpairs[$ep]} = ($n_gs, $indx_ev1, $n_gs, $indx_ev2) for the $ep-th pair of nearest-neighboring events. #

#=MOVED to above interaction type (I), to fix a BUG. (1)
			my $br2br2prev_nn_events = $gs2prev_nn_events{$c_gs};
			(defined $br2br2prev_nn_events) or ($br2br2prev_nn_events = {});
#=cut

			my $if_interact = 0;

			foreach my $info_evpair (@{$on_nn_brs}) { # 5th inner-inner foreach-loop (over nn event pairs). #
			    my ($br_c, $br_n, $indx_c_ev, $indx_n_ev) = @{$info_evpair};
			    my $nn_c = $br2nn{$br_c};
			    my $nn_n = $br2nn{$br_n};
				# %br2nn; # = ($br_id => {$id_nearest_neighbor => 1, ...}, ...)

				# (1) Examine the self-nearest-neighboring events in $n_gs. #

			    my $br2self_nn_n = $n_self_nn_events->{$br_n};
			    if (defined $br2self_nn_n) {
				foreach my $br (keys %{$br2self_nn_n}) { # 1st innermost foreach-loop (over branches self-neighboring $br_n). #
				    if (defined $nn_c->{$br}) {
					$if_interact = 1;
					last;
				    }
				}				# End of the 1st innermost foreach-loop (over branches self-neighboring $br_n). #
				$if_interact and last;
			    }

				# (2) Examine the self-nearest-neighboring events in $c_gs. #

			    my $br2self_nn_c = $c_self_nn_events->{$br_c};
			    if (defined $br2self_nn_c) {
				foreach my $br (keys %{$br2self_nn_c}) { # 2nd innermost foreach-loop (over branches self-neighboring $br_c). #
				    if (defined $nn_n->{$br}) {
					$if_interact = 1;
					last;
				    }
				}				# End of the 2nd innermost foreach-loop (over branches self-neighboring $br_c). #
				$if_interact and last;
			    }

#=MOVED TO ABOVE interaction type (I), and augmented, to fix a BUG (2).
				# (3) Examine the events in the segments before $c_gs that are nearest-neighborng the events in $c_gs. #
			    my $br2prev_nn_events = $br2br2prev_nn_events->{$br_c};
			    if (defined $br2prev_nn_events) {
				foreach my $br (keys %{$br2prev_nn_events}) { # 3rd innermost foreach-loop (over branches hosting previous events nearest-neighboring $br_c. #
				    if (defined $nn_n->{$br}) {
					$if_interact = 1;
					last;
				    }
				}				# End of the 3rd innermost foreach-loop (over branches hosting previous events nearest-neighboring $br_c). #
				$if_interact and last;
			    }
#=cut # END of "MOVED TO ABOVE interaction type (I), and augmented, to fix a BUG (2)."

			}				# End of the 5th inner-inner foreach-loop (over nn event pairs). #

			if ($if_interact) {
				# Interaction between $n_gs and $c_gs (II-a). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (II-a) ...\n\n") if ($TEST_LEVEL_DBG>2);

			    push @{$last_cluster}, ($last_gs_in_cls + 1 .. $n_gs);

   if ($TEST_LEVEL_DBG>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

			    next;

			} else {

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (C) ...\n\n") if ($TEST_LEVEL_DBG>2);
			}

=cut # END of "DISABLED in creation of 'summary_examine_complex.6p_aug_len.raw.pl'. (III)"

		    }						# End of the inner for-loop (over "next" gapped segments). #

		    $indx_c_gs++;

   if ($TEST_LEVEL_DBG>3) {
	print STDOUT '    | Updated $indx_c_gs  = ', $indx_c_gs, " | \n\n\n";
    }

		}				# End of the inner-middle while-loop (until the "last" cluster is exhausted). #

		my $last_gs_clustered = $last_cluster->[$#{$last_cluster}];
		($last_gs_clustered == $ct_gpd_segs-1) and last;

			# Create a new cluster. # 

		my @new_last_cluster = ($last_gs_clustered + 1);
		$last_cluster = \@new_last_cluster;
		push @clusters, $last_cluster;

   if ($TEST_LEVEL_DBG>3) {
	print STDOUT "    | Information updated by default: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

	    }					# End of the 1st middle while-loop (until all gapped segments are examined). #

   if ($TEST_LEVEL>2) {

	my $ct_cls = @clusters;

	print STDOUT "\n[ Resulting Clusters (#{clusters} = $ct_cls) ]\n\n";

	print STDOUT join ("\t", 'Indx_cls', 'constituent,gapped,segments'), "\n\n";

	for (my $cl = 0; $cl <$ct_cls; $cl++) { # Testing for-loop (over the clusters). #

	    print STDOUT $cl, "\t", join (',', @{$clusters[$cl]}), "\n";

	}				# End of the testing for-loop (over the clusters). #
	print STDOUT "\n\n";
   }

					### RESTART FROM HERE ###


=For the predecessor 'summary_examine_complex.6p.raw.pl'. (I) #
	    my $thrsh_spsize = $THRSHS_SPSIZE[$i_thr]; # If the {spacer size} <= $thrsh_spsize, then the two gapped segments will be put into the same cluster. #

	    my @clusters = ([0]);
	    my $last_end_coord = $gpd_segs_in_msa[0]->[2];

	    for (my $gs=1; $gs < $ct_gpd_segs; $gs++) { # 1st inner for-loop (over gapped segments in the subject MSA). #

		my ($beg_coord, $end_coord) = @{$gpd_segs_in_msa[$gs]}[1,2];

		my $size_spacer = $beg_coord - $last_end_coord - 1;

		if ($size_spacer <= $thrsh_spsize) {
		     push @{$clusters[$#clusters]}, $gs;
		} else {
		    push @clusters, [$gs];
		}
		$last_end_coord = $end_coord;
	    }					# End of the 1st inner for-loop (over gapped segments in the subject MSA). #


=cut # END of "For the predecessor 'summary_examine_complex.6p.raw.pl'. (I)" #

	    foreach my $cluster (@clusters) { # 2nd middle foreach-loop (over clusters in the subject MSA). #

		my $sbct_gpd_segs = @{$cluster};
		my $ct_A = 0;
		my $totct_clms = 0;
		my $tot_totct_indel = my $tot_diffct_indel = my $tot_totlen_indel = my $tot_difflen_indel = 0;

		foreach my $gs (@{$cluster}) { # Inner foreach-loop (over gapped segments in the subject cluster). #

			# => Each element of @gpd_segs_in_msa will become:
			# [$category, $beg_coord, $end_coord, $ct_clms, $totct_indel, $diffct_indel, $totlen_indel, $difflen_indel]
		    my ($category, $beg_coord, $end_coord, $ct_clms, $totct_indel, $diffct_indel, $totlen_indel, $difflen_indel) = @{$gpd_segs_in_msa[$gs]};

		    ($category eq 'A') and ($ct_A++);
		    $totct_clms += $ct_clms;
		    $tot_totct_indel += $totct_indel;
		    $tot_diffct_indel += $diffct_indel;
		    $tot_totlen_indel += $totlen_indel;
		    $tot_difflen_indel += $difflen_indel;

		}		# End of the inner foreach-loop (over gapped segments in the subject cluster). #

		my $pc_A = 100 * $ct_A /$sbct_gpd_segs;


			# MODIFIED in creation of 'summary_examine_complex.6p.raw.pl'. (IIb) #
		my $ctpair = sprintf "(%d,%d)", $tot_totct_indel, $tot_diffct_indel;
		my $bins_pcA = $ctpair2bins_pcA->{$ctpair};
		(defined $bins_pcA) or ($bins_pcA = $ctpair2bins_pcA->{'Others'});

		cluster_distributor_wgt ($pc_A, @{$bins_pcA}, @BOUNDS_PCA, $sbct_gpd_segs);
		cluster_distributor_wgt ($pc_A, @{$bins_pcA_total}, @BOUNDS_PCA, $sbct_gpd_segs);

=For the predecessor 'summary_examine_complex.5p.raw.pl'. (IIb) #
			### MAJOR MODIFICATION in creation of 'summary_examine_complex.5p.raw.pl'. ###
		distributor_2dim_wgt ($pc_A, $totct_clms, @{$bins_pcA_vs_totct_clms}, @BOUNDS_PCA, @BOUNDS_CT, $sbct_gpd_segs);
		distributor_2dim_wgt ($pc_A, $tot_totct_indel, @{$bins_pcA_vs_tot_totct_indel}, @BOUNDS_PCA, @BOUNDS_CT0, $sbct_gpd_segs);
		##distributor_2dim ($pc_A, $totct_clms, @{$bins_pcA_vs_totct_clms}, @BOUNDS_PCA, @BOUNDS_CT);
		##distributor_2dim ($pc_A, $tot_totct_indel, @{$bins_pcA_vs_tot_totct_indel}, @BOUNDS_PCA, @BOUNDS_CT0);

=cut # END of "For the predecessor 'summary_examine_complex.5p.raw.pl'. (IIb)" #

	    }			    # End of the 2nd middle foreach-loop (over clusters in the subject MSA). #

	}				# End of the 3rd main outer-middle for-loop (over the threshoulds for the spacer sizes). #

			### RESTART FROM HERE (1) ###

=Just a reminder...

our @BOUNDS_PCA = ($EPS, 20+$EPS, 40+$EPS, 60, 80, 100); # MODIFIED in creation of 'summary_examine_complex.6p.raw.pl'. #
my $CT_BOUNDS_PCA = @BOUNDS_PCA;

our @TOTCTS_INDELS = (1 .. 7);

our @CTPAIRS = (); # Element = sprintf ("(%d,%d)", $tot_totct_indel, $tot_diffct_indel)

foreach my $totct (@TOTCTS_INDELS) {
    foreach my $x (0 .. $totct) {
	my $diffct = $totct - 2 * $x;
	push @CTPAIRS, sprintf ("(%d,%d)", $totct, $diffct);
    }
}

(undef, my $tags_intvl_pcA) = init_2dim_distri (@BOUNDS_PCA, @BOUNDS_PCA, "%1.0f");

our @TAGS_INTVL_PCA = @{$tags_intvl_pcA};

my @set_ctpair2bins_pcA = ();
my @set_bins_pcA_total = ();

for (my $i_thr = 0; $i_thr <$CT_THRSHS_SPSIZE; $i_thr++) {

    my %ctpair2bins_pcA;

    my @batch_pcA = ();
    foreach my $ctpair (@CTPAIRS, 'Others') {
	my @bins_pcA = ();
	push @batch_pcA, \@bins_pcA;
	$ctpair2bins_pcA{$ctpair} = \@bins_pcA;
    }
    my @bins_pcA_total = ();
    push @batch_pcA, \@bins_pcA_total;

    batch_initializer (@batch_pcA, @BOUNDS_PCA);

    push @set_ctpair2bins_pcA, \%ctpair2bins_pcA;
    push @set_bins_pcA_total, \@bins_pcA_total;
}

=cut # END of "Just a reminder..."



=Just a reminder... for the predecessor 'summary_examine_complex.5p.raw.pl'. (IIx) #

our @THRSHS_SPSIZE = (1 .. 8);
our $CT_THRSHS_SPSIZE = @THRSHS_SPSIZE;

my @set_bins_pcA_vs_totct_clms = my @set_bins_pcA_vs_tot_totct_indels = ();

for (my $i_thr = 0; $i_thr <$CT_THRSHS_SPSIZE; $i_thr++) {

	(my $bins_pcA_vs_totct_clms, undef) = init_2dim_distri (@BOUNDS_PCA, @BOUNDS_CT, "%1.0f");
	(my $bins_pcA_vs_tot_totct_indel, undef) = init_2dim_distri (@BOUNDS_PCA, @BOUNDS_CT0, "%1.0f");

	$set_bins_pcA_vs_totct_clms[$i_thr] = $bins_pcA_vs_totct_clms;
	$set_bins_pcA_vs_tot_totct_indel[$i_thr] = $bins_pcA_vs_tot_totct_indel;

}

=cut # END of "Just a reminder... for the predecessor 'summary_examine_complex.5p.raw.pl'. (IIx)" #

=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (VI)

			# Take the 2D distributions. #

	distributor_2dim ($max_totct_indel, $maxct_clms, @{$bins_max_totct_indels_vs_maxct_columns}, @BOUNDS_CT0, @BOUNDS_CT);
	distributor_2dim ($max_totct_indel, $max_totlen_indel, @{$bins_max_totct_indels_vs_max_totlen_indels}, @BOUNDS_CT0, @BOUNDS_CT);
	distributor_2dim ($max_totct_indel, $max_diffct_indel, @{$bins_max_totct_indels_vs_max_diffct_indels}, @BOUNDS_CT0, @BOUNDS_PNCT0);
	distributor_2dim ($max_totct_indel, $min_diffct_indel, @{$bins_max_totct_indels_vs_min_diffct_indels}, @BOUNDS_CT0, @BOUNDS_PNCT0);


#=Just a reminder... 
(my $bins_max_totct_indels_vs_maxct_columns_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_maxct_columns_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_totlen_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
(my $bins_max_totct_indels_vs_max_totlen_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");

(my $bins_max_totct_indels_vs_max_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_max_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

(my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
(my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");

#=cut # END of "Just a reminder..."

=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (VI)"



=DISABLED in creation of 'summary_examine_purge.raw.pl'.
		# ADDED in creation of 'summary_cmp_ref_vs_rec_complliment.pl'. #
	if (($id_class eq '2_4') or ($id_class eq '3_4')) {
	    $ct_skipped1++; # Skip the entries belonging to these classes (to avoid redundancy). #
	    next;
	}
=cut


=DISABLED in creation of 'summary_examine_purge.raw.pl'.
		# ADDED in creation of 'summary_classify_msa_errors.raw.pl'. #

	unless (($ct_equiv_indels ne 'undef') and (uc($ct_equiv_indels) ne 'N/A')) {
	    print STDOUT '[CAUTION] ($id_class, $session, $segment) = ', "($id_class, $session, $segment) has ", '$ct_equiv_indels = ', "$ct_equiv_indels: $! !\n\n";
	    $ct_skipped_x++;
	    next;
	}
=cut


	$ct_succ++;

=DISABLED in creation of 'summary_examine_complex.2.raw.pl'.
}		# End of outer while-loop (over erroneous MSA segments) #
=cut

	  ## ($ct_succ>=1000) and last; ### For TESTING the script. ###
	  ## ($ct_succ>=100) and last; ### For TESTING the script. ###
	  ## ($ct_succ>= 3) and last; ### For TESTING the script. ('Test1b') ###
	  ## ($ct_succ>= 1) and last; ### For DEBUGGING the script. ((1) and (2)) ###

}		# End of outer while-loop (until the last cluster runs out). # MODIFIED in creation of 'summary_examine_complex.2.raw.pl'. #

close ($fh_in1);
##close ($fh_out); ## DISABLED in creation of 'summary_examine_purge.raw.pl'. ##

print STDOUT "\n\n=== ALL the Processes ENDED !! ===\n\n";

print STDOUT "#{Anomalous input lines}    = $ct_anom .\n"; # MOVED from below in creation of 'summary_examine_complex.1.raw.pl'.
print STDOUT "Total#{input Segments}         = $totct_segs .\n";
print STDOUT "   #{Segments w/ anomalous type or diagnosis} = $ct_anom2 .\n"; # ADDED in creation of 'summary_examine_complex.1.raw.pl'.
print STDOUT "   #{Segments w/ inconsistent #{gapped segments}} = $ct_anom3 .\n\n"; # ADDED in creation of 'summary_examine_complex.2.raw.pl'.

=DISABLED in creation of 'summary_examine_complex.1.raw.pl'.
print STDOUT "#{Skipped segments (to avoid redundancy)} = $ct_skipped1 .\n";

		# MODIFIED in creation of 'summary_examine_purge.raw.pl'. #

print STDOUT '#{Skipped segments (w/ $size_block or $abs_shift fractional)} = ', $ct_w_fractional, " .\n";
=cut

=DISABLED in creation of 'summary_examine_purge.raw.pl'.
			# ADDED in creation of 'summary_classify_msa_errors.raw.pl'. #
print STDOUT '#{Skipped segments (w $ct_equiv_indels unavailable)} = ', $ct_skipped_x, " .\n";
print STDOUT "#{Anomalous input lines}    = $ct_anom .\n"; # MOVED from above in creation of 'summary_classify_msa_errors.raw.pl'.
print STDOUT "#{Input segments containing 'undef' differences} = $ct_w_undef .\n";
=cut

print STDOUT "   #{Input segments failed to be processed} = $ct_fail .\n";
print STDOUT "   #{Input segments successfully processed} = $ct_succ .\n\n";

unless (0 == $ct_fail) {
	print STDOUT "[ERROR '$program_name'] FAILED to process $ct_fail inputs !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}


		### MAJOR ADDITION in creation of 'summary_examine_complex.2_2D.raw.pl'. ###


print STDOUT "\n\n[ (i) ] Categories of the erroneous and correct segments:\n\n";
print STDOUT "Category A:   'Complex' errors + segments that are 'Too-Long' ,\n\n";
print STDOUT "Category B:   'Non-complex' errors + Correctly reconstructed segments .\n\n";

my $section = 0;

		### MAJOR MODIFICATION in creation of 'summary_examine_complex.6p.raw.pl'. (III) ### 

my @labels_out = ('(X,Y) \ %A', '', @TAGS_INTVL_PCA, 'PureA / PureB');
foreach my $intvl (@TAGS_INTVL_PCA) { ($intvl =~ m/^T/) or (push @labels_out, "$intvl (%)"); }
foreach my $intvl (@TAGS_INTVL_PCA) { ($intvl =~ m/^T/) or (push @labels_out, "$intvl (cum%)"); }

$section++;
print STDOUT "\n\n[$section] 2D-distribution, for each of the threshold values, of %{Category-A gapped segments}, abscissa) and (Tot[Tot#{columns}], Tot[#{ins} - #{del}]) (ordinates) for the artificial clusters of gapped segments:\n";

print STDOUT "\n( NOTE1: In each cell is the number of gapped segments in the specified class of artificial clusters. )\n";

print STDOUT "(NOTE2: X = Tot[#{ins} + #{del}], Y = Tot[#{ins} - #{del}]. )\n\n"; # MODIFIED on July 2 (Sat), 2016. #
##print STDOUT "(NOTE2: X = Tot[Tot#{indels}], Y = Tot[#{ins} - #{del}]. )\n\n";


for (my $i_thr = 0; $i_thr <$CT_THRSHS_SPSIZE; $i_thr++) { # 1st trailing for-loop (over the spacer-size thresholds). #

	my $thr = $THRSHS_SPSIZE[$i_thr];

	my $ctpair2bins_pcA = $set_ctpair2bins_pcA[$i_thr];
	my $bins_pcA_total = $set_bins_pcA_total[$i_thr];
	my $ct_bins = @{$bins_pcA_total};

	my %ctpair2ratio_AB; # For those count pairs with #{pure B} > 0.
	my %ctpair2ct_pureA; # For those count pairs with #{pure B} = 0.

	foreach my $ctpair (@CTPAIRS) {
	    my $bins_pcA = $ctpair2bins_pcA->{$ctpair};
	    my ($ct_pureB, $ct_pureA) = @{$bins_pcA}[0, $CT_BOUNDS_PCA];
	    if ($ct_pureB>0) {
		$ctpair2ratio_AB{$ctpair} = $ct_pureA/$ct_pureB;
	    } else {
		$ctpair2ct_pureA{$ctpair} = $ct_pureA;
	    }
	}

	my @std_ctpairs = sort { $ctpair2ratio_AB{$a} <=> $ctpair2ratio_AB{$b} } keys %ctpair2ratio_AB;
	push @std_ctpairs, sort { $ctpair2ct_pureA{$a} <=> $ctpair2ct_pureA{$b} } keys %ctpair2ct_pureA;

	print STDOUT "\n( For clustering segments with {spacer-size} <= $thr }\n\n";

	my @cum_freqs = ();
	for (1 .. $ct_bins) { push @cum_freqs, 0; }

	print STDOUT join ("\t", @labels_out), "\n\n";

	foreach my $ctpair (@std_ctpairs, 'Others') {
	    my $bins_pcA = $ctpair2bins_pcA->{$ctpair};
	    my ($ct_pureB, $ct_pureA) = @{$bins_pcA}[0, $CT_BOUNDS_PCA];
	    my $ratio_AB = ($ct_pureB>0) ? $ct_pureA/$ct_pureB : 'undef' ; 

	    my @freqs = ();
	    my $totct = 0;
	    for (my $i=0; $i<$ct_bins; $i++) {
		my $ct = $bins_pcA->[$i];
		$totct += $ct;

			# MODIFIED to fix a BUG. #
		my $totct_clm = $bins_pcA_total->[$i];
		my $freq = ($totct_clm > 0) ? 100 * $ct / $totct_clm : 'undef';
		#my $freq = ($totct_clm > 0) ? $ct / $totct_clm : 'undef'; # Another BUG fixed.
		###my $freq = $ct /$bins_pcA_total->[$i];

		$freqs[$i] = $freq;

			# MODIFIED to fix a BUG. #
		if (($freq eq 'undef') or ($cum_freqs[$i] eq 'undef')) {
		   $cum_freqs[$i] = 'undef';
		} else {
		   $cum_freqs[$i] += $freq;
		}
		##$cum_freqs[$i] += $freq;
	    }
	    my @out = (sprintf ("%-10s", $ctpair), '');
	    foreach my $ct ( @{$bins_pcA}, $totct) { push @out, sprintf ("%7d", $ct); }

			# MODIFIED to fix a BUG. #
	    push @out, (($ratio_AB eq 'undef') ? sprintf ("%13s", $ratio_AB) : sprintf ("%13.8f", $ratio_AB));
	    foreach my $freq (@freqs) { push @out, (($freq eq 'undef') ? sprintf ("%6s", $freq) : sprintf ("%6.2f%%", $freq)); }
	    foreach my $cfreq (@cum_freqs) { push @out, (($cfreq eq 'undef') ? sprintf ("%6s", $cfreq) : sprintf ("%6.2f%%", $cfreq)); }
	    ##push @out, sprintf ("%13.8f", $ratio_AB);
	    ##foreach my $freq (@freqs) { push @out, sprintf ("%6.2f%%", $freq); }
	    ##foreach my $cfreq (@cum_freqs) { push @out, sprintf ("%6.2f%%", $cfreq); }

	    print STDOUT join ("\t", @out), "\n";
	}

	print STDOUT "\n";

	my $totalct_sbjcts = 0;
	foreach my $total (@{$bins_pcA_total}) { $totalct_sbjcts += $total; }
	my ($totct_pureB, $totct_pureA) = @{$bins_pcA_total}[0, $CT_BOUNDS_PCA];

		# MODIFIED to fix a BUG. #
	my $ratio_AB_total = ($totct_pureB > 0) ? $totct_pureA/$totct_pureB : 'undef';
	##my $ratio_AB_total = $totct_pureA/$totct_pureB;

	my @out_total = (sprintf ("%-10s", 'Total'), '');
	foreach my $ct (@{$bins_pcA_total}, $totalct_sbjcts) { push @out_total, sprintf ("%7d", $ct); }

		# MODIFIED to fix a BUG. #
	push @out_total, (($ratio_AB_total eq 'undef') ? sprintf ("%13s", $ratio_AB_total) : sprintf ("%13.8f", $ratio_AB_total));
	for (my $i=0; $i<$ct_bins; $i++) { push @out_total, (($bins_pcA_total->[$i]>0) ? sprintf ("%6.2f%%", 100) : 'undef'); }
	##push @out_total, sprintf ("%13.8f", $ratio_AB_total);
	##for (1..$ct_bins) { push @out_total, sprintf ("%6.2f%%", 100); }

	print STDOUT join ("\t", @out_total), "\n\n";

}					# End of the 1st trailing for-loop (over the spacer-size thresholds). #
print STDOUT "\n\n";


=Just a reminder...

our @BOUNDS_PCA = ($EPS, 20+$EPS, 40+$EPS, 60, 80, 100); # MODIFIED in creation of 'summary_examine_complex.6p.raw.pl'. #
my $CT_BOUNDS_PCA = @BOUNDS_PCA;

our @TOTCTS_INDELS = (1 .. 7);

our @CTPAIRS = (); # Element = sprintf ("(%d,%d)", $tot_totct_indel, $tot_diffct_indel)

foreach my $totct (@TOTCTS_INDELS) {
    foreach my $x (0 .. $totct) {
	my $diffct = $totct - 2 * $x;
	push @CTPAIRS, sprintf ("(%d,%d)", $totct, $diffct);
    }
}

(undef, my $tags_intvl_pcA) = init_2dim_distri (@BOUNDS_PCA, @BOUNDS_PCA, "%1.0f");

our @TAGS_INTVL_PCA = @{$tags_intvl_pcA};

my @set_ctpair2bins_pcA = ();
my @set_bins_pcA_total = ();

for (my $i_thr = 0; $i_thr <$CT_THRSHS_SPSIZE; $i_thr++) {

    my %ctpair2bins_pcA;

    my @batch_pcA = ();
    foreach my $ctpair (@CTPAIRS, 'Others') {
	my @bins_pcA = ();
	push @batch_pcA, \@bins_pcA;
	$ctpair2bins_pcA{$ctpair} = \@bins_pcA;
    }
    my @bins_pcA_total = ();
    push @batch_pcA, \@bins_pcA_total;

    batch_initializer (@batch_pcA, @BOUNDS_PCA);

    push @set_ctpair2bins_pcA, \%ctpair2bins_pcA;
    push @set_bins_pcA_total, \@bins_pcA_total;
}

=cut # END of "Just a reminder..."


=For the predecessor 'summary_examine_complex.5p.raw.pl'. (III) #
		## MAJOR MODOFICATION in creation of 'summary_examine_complex.5.raw.pl'. (VII) ##

$section++;
print STDOUT "\n\n[$section] 2D-distribution, for each of the threshold values, of %{Category-A gapped segments}, abscissa) and Tot#{columns} (ordinates) for the artificial clusters of gapped segments:\n";

print STDOUT "\n( NOTE: In each cell is the number of gapped segments in the specified class of artificial clusters. )\n";

for (my $i_thr = 0; $i_thr <$CT_THRSHS_SPSIZE; $i_thr++) { # 1st trailing for-loop (over the spacer-size thresholds). #

	my $thr = $THRSHS_SPSIZE[$i_thr];
 	my $bins_pcA_vs_totct_clms = $set_bins_pcA_vs_totct_clms[$i_thr];

		# $set_bins_pcA_vs_totct_clms[$i_thr] = $bins_pcA_vs_totct_clms;

	print STDOUT "\n\n( For clustering segments with {spacer-size} <= $thr }\n\n";


	pr_file_2dim_distri_xl (@{$bins_pcA_vs_totct_clms}, @BOUNDS_PCA, @BOUNDS_CT, '%{Category-A}', 'Tot#{columns}', *STDOUT, "%1.0f", "%3d");

		# #(my $bins_pcA_vs_totct_clms, undef) = init_2dim_distri (@BOUNDS_PCA, @BOUNDS_CT, "%1.0f");

}					# End of the 1st trailing for-loop (over the spacer-size thresholds). #
print STDOUT "\n\n";

$section++;
print STDOUT "\n\n[$section] 2D-distribution, for each of the threshold values, of %{Category-A gapped segments}, abscissa) and Tot[Tot#{indels}] (ordinates) for the artificial clusters of gapped segments:\n";

print STDOUT "\n( NOTE: In each cell is the number of gapped segments in the specified class of artificial clusters. )\n";

for (my $i_thr = 0; $i_thr <$CT_THRSHS_SPSIZE; $i_thr++) { # 1st trailing for-loop (over the spacer-size thresholds). #

	my $thr = $THRSHS_SPSIZE[$i_thr];
 	my $bins_pcA_vs_tot_totct_indel = $set_bins_pcA_vs_tot_totct_indel[$i_thr];

		# $set_bins_pcA_vs_tot_totct_indel[$i_thr] = $bins_pcA_vs_tot_totct_indel;

	print STDOUT "\n\n( For clustering segments with {spacer-size} <= $thr }\n\n";


	pr_file_2dim_distri_xl (@{$bins_pcA_vs_tot_totct_indel}, @BOUNDS_PCA, @BOUNDS_CT0, '%{Category-A}', 'Tot[Tot#{indels}]', *STDOUT, "%1.0f", "%2d");

		# (my $bins_pcA_vs_tot_totct_indel, undef) = init_2dim_distri (@BOUNDS_PCA, @BOUNDS_CT0, "%1.0f");
}					# End of the 1st trailing for-loop (over the spacer-size thresholds). #
print STDOUT "\n\n";


#=Just a reminder... (II)

our @THRSHS_SPSIZE = (1 .. 8);
our $CT_THRSHS_SPSIZE = @THRSHS_SPSIZE;

#=cut # END of "Just a reminder... (II)"

=cut # END of "For the predecessor 'summary_examine_complex.5p.raw.pl'. (III)" #

=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (VII)

$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and #{columns} (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_maxct_columns_A}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max#{columns}', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_maxct_columns_B}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max#{columns}', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_maxct_columns_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
# (my $bins_max_totct_indels_vs_maxct_columns_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and the total length of indels (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_totlen_indels_A}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max(Totlen{indels})', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_totlen_indels_B}, @BOUNDS_CT0, @BOUNDS_CT, 'Max#{indels}', 'Max(Totlen{indels})', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_max_totlen_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");
# (my $bins_max_totct_indels_vs_max_totlen_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_CT, "%1d");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and Max(#{ins} - #{del}) (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_diffct_indels_A}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Max(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_max_diffct_indels_B}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Max(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_max_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
# (my $bins_max_totct_indels_vs_max_diffct_indels_B, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");


$section++;
print STDOUT "\n\n[$section] 2D-distribution of #{indels} (inferred with Dollo-parsimony, abscissa) and Min(#{ins} - #{del}) (ordinates) for the 'largest' gapped segments in the segments of categories A and B:\n\n";

print STDOUT "{ Category A }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_min_diffct_indels_A}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Min(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

print STDOUT "\n\n{ Category B }\n\n";

pr_file_2dim_distri_xl (@{$bins_max_totct_indels_vs_min_diffct_indels_B}, @BOUNDS_CT0, @BOUNDS_PNCT0, 'Max#{indels}', 'Min(#{ins} - #{del})', *STDOUT, "%1d", "%2d");

# (my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");
# (my $bins_max_totct_indels_vs_min_diffct_indels_A, undef) = init_2dim_distri (@BOUNDS_CT0, @BOUNDS_PNCT0, "%1d");


=cut # END of "For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (VII)"
###=For the predecessor 'summary_examine_complex.2_2D.raw.pl'. (VII) # A BUG fixed on July 13 (Wed),2016. #


print STDOUT "\n\n";


print STDOUT "\n\n=== ENDED Printing Out the Statistics ===\n\n";


	### The DEFINITION(s) of the SUBROUTINE(s). ###

# readin_dawg_tree2:
#	This modified version of 'readin_dawg_tree' copes with the new format
#	of the Dawg control file.
#	(CREATED in creation of 'compare_ref_vs_rec_compllimment.pl' on Aug 9 (Sat), 2014)
#	(=> IMPORTED from '../../../Full_Analyses/Scripts/examine_complex_via_mblks.2.pl'
#	in creation of 'summary_examine_complex.6p_aug.raw.pl'.)
#
# readin_dawg_tree:
#	This subroutine reads in the input tree used to guide
#	a simulation of sequence evolution via DAWG.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011) 
#
# INPUT: \$infile, where
#	$infile is the file recording input parameters of DAWG.
#
# OUTPUT: \@info_dawg_tree, where
#	@info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr).
#	See e.g. 'MyTreeManipu_II.pm' for the format of the output hash.
#
sub readin_dawg_tree2 (\$) {

	my ($infile) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$infile});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'open_input_file2' FAILED on the input file '${$infile}': $! \n";
	    return undef;
	}

	my $in_tree;
	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #

	    ($tmpin =~ s/^Tree\s+\=\s+//) or next; # MODIFIED in creation of 'readin_dawg_tree2'. #
	    ##($tmpin =~ s/^Tree\.Tree\s+\=\s+//) or next;
	    $in_tree = $tmpin;
	    last;
	}		# End of the main while-loop # 
	close ($fh_tmpin);

	unless ((defined $in_tree) and ($in_tree =~ m/\(/)) {
	    (defined $in_tree) or ($in_tree = 'undef');
	    $in_tree =~ s/\s+$//;
	    print STDERR "[ERROR 'readin_dawg_tree2'] NO TREE was extracted from the input file '${$infile}': $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

		## MODIFIED in creation of 'readin_dawg_tree2'. ##

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr; my %name2id_intnode; my $topnode_name;

	unless ( input_tree_file_dawg ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr, %name2id_intnode, $topnode_name) ) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'input_tree_file_dawg' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

		# (Just in case,) remove the interior node names (or bootstrap supports) from the list of branch attributes. #

	foreach my $br (keys %br_attr) {
	    my $attr = $br_attr{$br};
	    my @new_attr = ();
	    $new_attr[$INDX_LENGTH] = $attr->[$INDX_LENGTH];
	    $br_attr{$br} = \@new_attr;
	}

=for the predecessor 'readin_dawg_tree'
		# Process the input tree a bit #
	($in_tree =~ s/[^\)\;]+$//);
	($in_tree =~ m/\;$/) or ($in_tree .= ';');

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;
	unless ( input_tree_file ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr) ) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'input_tree_file' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}
=cut

	my @info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr);

	return \@info_dawg_tree;

}	# End of the subroutine 'readin_dawg_tree2' #


# selfcompare_indel_hosting_branches:
#	This subroutine compares a set of indel events with itself,
#	especially in terms of the branches along which the events occur.
#	[CREATED on July 14 (Thu), 2016, in creation of 'summary_examine_complex.6p_aug.raw.pl'.]
#
# INPUT: (\@eventset, \%equiv_br, \%br2nn), where
#	@{$eventset[$ev]} = ($br, $stat_ue, $beg, $end, $len) for the $ev th event in the set,
#	   with $br being the ID of the branch along which it occurred,
#	   with $stat_ue being the state ("presence"/"absence") at the upper-end of the branch,
#	   with $beg/$end/$len being the beginning coordinate/ending coordinate/length of the event;
#
#	%equiv_br = ($br_id => $equiv_br_id, ...) for branches with their equivalents;
#	%br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...).
#
# OUTPUT: (\@on_coinc_brs, \@on_nn_brs), where
#	
#	@{$on_coinc_brs[$k]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $k th pair of events along the same branch ($br1 or $br2), with
#	  $indx_ev1/2 being the index of the relevant 1st/2nd event (in the order stored in @eventset);
#
#	@{$on_nn_brs[$m]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $m th pair of events along the nearest-neighboring branches ($br1 and $br2), with
#	   $indx_ev1/2 being the index of the relevant 1st/2nd event (in the order stored in @eventset).
#
sub selfcompare_indel_hosting_branches (\@\%\%) {

	my ($eventset, $equiv_br, $br2nn) = @_;

	my $ct_events = @{$eventset};

	my @on_coinc_brs = my @on_nn_brs = ();

	for (my $ev2=1; $ev2 < $ct_events; $ev2++) { # Outer for-loop (over 'later' events). #

	    my ($br2, $state_ue2, $beg2, $end2, $len2) = @{$eventset->[$ev2]};
	    my $eq_br2 = $equiv_br->{$br2};
	    my $nn2 = $br2nn->{$br2};
	    (defined $nn2) or ($nn2 = {});

	    my $if_relv = 0;

	    for (my $ev1=0; $ev1 < $ev2; $ev1++) { # Inner for-loop (over 'earlier' events). #

		my ($br1, $state_ue1, $beg1, $end1, $len1) = @{$eventset->[$ev1]};

		if ( ($br1 == $br2)  or
			((defined $eq_br2) and ($br1 == $eq_br2)) ) {

		    push @on_coinc_brs, [$br1, $br2, $ev1, $ev2];
		    # $if_relv=1;
		} elsif (defined $nn2->{$br1}) {

		    push @on_nn_brs, [$br1, $br2, $ev1, $ev2];
		    # $if_relv=1;
		}
	    }					# End of the inner for-loop (over 'earlier' events). #

	    ## (!($if_relv) and ($state_ue2 eq $char_abs)) and (push @irrlv_insertions2, $ev2);

	}				# End of the outer for-loop (over 'later' events). #

	return (\@on_coinc_brs, \@on_nn_brs);

}	# End of the subroutine 'selfcompare_indel_hosting_branches'. #


# compare_indel_hosting_branches:
#	This subroutine compares a set of indel events with another set,
#	especially in terms of the branches along which the events occur.
#	[CREATED on July 14 (Thu), 2016, in creation of 'summary_examine_complex.6p_aug.raw.pl'.]
#
# INPUT: (\@eventset1, \@eventset2, \%equiv_br, \%br2nn, $char_abs, \%node2pa), where
#	@{$eventset1/2[$ev]} = ($br, $stat_ue, $beg, $end, $len) for the $ev th event in the 1st/2nd set,
#	   with $br being the ID of the branch along which it occurred,
#	   with $stat_ue being the state ("presence"/"absence") at the upper-end of the branch,
#	   with $beg/$end/$len being the beginning coordinate/ending coordinate/length of the event;
#
#	%equiv_br = ($br_id => $equiv_br_id, ...) for branches with their equivalents;
#	%br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...);
#	$char_abs is the string representing the absence state;
#	%node2pa = ($node_id => $parent_id, ...);
#
# OUTPUT: (\@on_coinc_brs, \@on_nn_brs, \@irrlv_insertions2), where
#	
#	@{$on_coinc_brs[$k]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $k th pair of events along the same branch ($br1 or $br2), with
#	  $indx_ev1/2 being the index of the relevant event in @eventset1/2;
#
#	@{$on_nn_brs[$m]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $m th pair of events along the nearest-neighboring branches ($br1 and $br2), with
#	   $indx_ev1/2 being the index of the relevant event in @eventset1/2;
#
#	@irrlv_insertions2 stores the indices of the insertions in @eventset2 along the branches that neither coincide with, nor is the nearest-neighbors of, any branches hosting the events in @eventset1.
#
sub compare_indel_hosting_branches (\@\@\%\%$\%) {

	my ($eventset1, $eventset2, $equiv_br, $br2nn, $char_abs, $node2pa) = @_;

	my $ct_events1 = @{$eventset1};
	my $ct_events2 = @{$eventset2};

	my @on_coinc_brs = my @on_nn_brs = my @irrlv_insertions2 = ();

	for (my $ev2=0; $ev2 < $ct_events2; $ev2++) { # Outer for-loop (over events in @{$eventset2}). #

	    my ($br2, $state_ue2, $beg2, $end2, $len2) = @{$eventset2->[$ev2]};
	    my $eq_br2 = $equiv_br->{$br2};
	    my $nn2 = $br2nn->{$br2};
	    (defined $nn2) or ($nn2 = {});

	    my $if_relv = 0;

	    for (my $ev1=0; $ev1 < $ct_events1; $ev1++) { # Inner for-loop (over events in @{$eventset1}). #

		my ($br1, $state_ue1, $beg1, $end1, $len1) = @{$eventset1->[$ev1]};

		if ( ($br1 == $br2)  or
			((defined $eq_br2) and ($br1 == $eq_br2)) ) {
		    push @on_coinc_brs, [$br1, $br2, $ev1, $ev2];
		    $if_relv=1;
		} elsif (defined $nn2->{$br1}) {
		    push @on_nn_brs, [$br1, $br2, $ev1, $ev2];
		    $if_relv=1;
		} elsif (!($if_relv) and ($state_ue2 eq $char_abs) ) {
				# ADDED on July 14 (Thu), 2016. #
		    my ($common_anc, $after_div1, $after_div2) = fetch_common_ancestors ($br1, $br2, %{$node2pa});
		    if (@{$after_div1} == 0) { # $br1 is ancestor of $br2.
			($state_ue1 eq $char_abs) and ($if_relv = 1); # event along $br1 is insertion. #
		    } elsif (@{$after_div2} == 0) { # $br2 is ancestor of $br1.
			$if_relv = 1;
		    }
		}
	    }					# End of the inner for-loop (over events in @{$eventset1}). #

	    (!($if_relv) and ($state_ue2 eq $char_abs)) and (push @irrlv_insertions2, $ev2);

	}				# End of the outer for-loop (over events in @{$eventset2}). #

	return (\@on_coinc_brs, \@on_nn_brs, \@irrlv_insertions2);

}	# End of the subroutine 'compare_indel_hosting_branches'. #


# broad_error_type:
#	This subroutine receives the name of an error-type,
#	and classifies it into a broad type.
#	(CREATED in creation of 'summary_classify_msa_errors.bulk_pars.pl' on Nov 17 (Mon), 2014.)
#
# INPUT: $type_error, which is the name of an error-type.
#
# OUTPUT: $br_type, which is the broad type accommodating $type_error.
#
sub broad_error_type ($) {

	my ($type_error) = @_;

	$type_error = uc($type_error);

	foreach my $brtype ('Shift', 'Complex', 'Merge', 'Purge', 'Split', 'Ex-nihilo') { # Main foreach-loop (over some broad error-types). #
	    if (substr ($type_error, 0, length ($brtype)) eq uc($brtype)) {
		return $brtype;
	    }
	} 				# End of the main foreach-loop (over some broad error-types). # 

	if ($type_error =~ m/^NEIGHBORING-BRANCH-MERGE/) {
	    return 'NB_Merge';

	} elsif ($type_error =~ m/^NEIGHBORING-BRANCH-SPLIT/) {
	    return 'NB_Split';

	} elsif ($type_error =~ m/^COLLAPSE-OF-INDEPENDENT-INSERTIONS/) {
	    return 'CII';

	} elsif ($type_error =~ m/^CREATION-OF-ERRONEOUS-INDEPENDENT-INSERTIONS/) {
	    return 'CEII';

	} elsif ($type_error =~ m/^INCOMPLETE-COLLAPSE-OF-INDEPENDENT-INSERTIONS/) {
	    return 'iCII';

	} elsif ($type_error =~ m/^INCOMPLETE-CREATION-OF-ERRONEOUS-INDEPENDENT-INSERTIONS/) {
	    return 'iCEII';

	} elsif ($type_error =~ m/^PARTIAL-NEIGHBORING-BRANCH-MERGE/) {
	    return 'pNB_Merge';

	} elsif ($type_error =~ m/^PARTIAL-NEIGHBORING-BRANCH-SPLIT/) {
	    return 'pNB_Split';

	} else {
	    return 'Other';
	}

		## @BROAD_TYPES = ('Shift', 'Complex', 'Merge', 'Purge', 'Split', 'Ex-nihilo', 'NB_Merge', 'NB_Split', 'CII', 'CEII', 'pNB_Merge', 'pNB_Split', 'iCII', 'iCEII', 'Other');

}		# End of the subroutine 'broad_error_type'. #


