#!/usr/bin/perl
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# ingr_stat_srds_in_correct_segs.ver0.6.pl:
#	This Perl script gathers "ingredients" for taking the statistics
#	of "Substitutional Residue Difference"s (SRDs)
#	on the MSA segments where the reference and reconstructed MSAs coincide.
#	The script was created by modifying 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'.
#	[CREATION STARTED on 2020/04/02, ENDED on 2020/04/dd.]
#
#	[NOTE (on 2020/04/02)]:
#	Although one of the input files, $inaln_file_ref and $inaln_file_rec, 
#	is actually NOT necessary,
#	they were kept intact, in order to minimize the program modification.
#
#
# ingr_correl_alnshear_cb_vs_srd.ver0.6.pl:
#	This Perl script gathers "ingredients" for examining the correlations (or associations)
#	between	"Alignment-Shear"s and "Substitutional Residue Difference"s (SRDs).
#	The script was created from the script, 'map_alnshear_cb_and_srd.ver0.6.pl',
#	by just adding a subroutine to tally SRDs at the positions in the "Alignment-Shear"s and at the "None-Shear" positions,
#	where each "position" is specified with (branch x site).
#	[CREATION STARTED on 2020/03/24, ENDED on 2020/03/25.]
#
#	[REVISED on 2020/03/26 & 27]:
#	Now, the subroutine also constructs distributions of observed & expected SRDs, on and off the Alignment-Shears.
#
#	[REVISED on 2020/03/29]:
#	Now, the subroutine sets an UPPER-BOUND of the segment-length,
#	and SKIPS analyses on those segments whose $ref_len or $rec_len is longer than the UPPER-BOUND.
#
# map_alnshear_cb_and_srd.ver0.6.pl:
#	This Perl script was created by REVISING the predecessor script, 'map_alnshear_and_srd.ver0.6.pl',
#	to change the strategy for creating position-shift MINI-classes,
#	in the same way that 'map_aln_shear_clm_based.ver0.6.pl' was created from 'map_aln_shear.ver0.6.pl'.
#	(For the specific strategy, refer to 'map_aln_shear_clm_based.ver0.6.pl'.)
#	[CREATION STARTED on 2020/03/23, ENDED on 2020/03/dd.]
#
# map_alnshear_and_srd.ver0.6.pl:
#       This Perl script receives a reconstructed multiple sequence alignment (MSA),
#       a reference MSA, and a phylogenetic tree, etc.,
#       and outputs an "Alignment-Shear Map" and a "Substitutional Residue-Difference" Map (abbr. SRD Map),
#       as well as some auxiliary outputs, for each segment.
#       The script was created by incorporating the main functions of 'srd_map.ver0.6.pl'
#       into the script, 'map_aln_shear.ver0.6.STRIPPED.pl'.
#	[CREATION STARTED on 2020/03/11, ENDED on 2020/03/dd.]
#
# map_aln_shear.ver0.6.STRIPPED.pl:
#	This Perl script was created from 'map_aln_shear.ver0.6.pl'
#	by STRIPPING OFF many UNnecessary Parts & subroutines.
#
# map_aln_shear.ver0.6.pl:
#	This Perl script receives a reconstructed multiple sequence alignment (MSA),
#	a reference MSA, and a phylogenetic tree, etc.,
#	and outputs a Position-Shift Map (described in K. Ezawa 2016a) and an "Alignment-Shear Map".
#	The "Alignment-Shear Map" marks the "alignment-shear"s,
#	each of which shows misalignment across a branch,
#	on the "table" whose columns represent sites and rows represent branches.
#	(The specific recipe for drawing the "Alignment-Shear Map" is described in K. Ezawa 2020c.)
#	The script was created by taking advantge of the script, 'classify_msa_errors_via_mblks.alpha2.pl'
#	(in "ComplLiMment_P/Sample_Scripts/").
#	[CREATION STARTED on 2020/03/07, ENDED on 2019/03/dd.]
#
###################################################################################
#	
#	ComplLiMment_P: A package to calculate the Complete-likelihood for a given Multiple-sequence alignment
#			(almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.6.1: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.6.1.1: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# classify_msa_errors_via_mblks.alpha2.pl:
#	This script was created from "classify_msa_errors_via_mblks.2.pl,"
#	by removing superfluous functions and commented-out portions,
#	in order to make the script less complicated.
#	[CREATED on Oct 30 (Fri), 2015.]
#
# classify_msa_errors_via_mblks.2.pl:
#	This script attempts to further classify some simple 'Complex' errors
#	that is either associated with a block "hard-link"ing another block
#	or associated with a pair of blocks.
#	The script was created by enhancing the script, 'classify_msa_errors_via_mblks.1.pl'.
#	[CREATION STARTED on Oct 5 (Tue), 2015, ENDED on Oct 23 (Fri), 2015.]
#
#	[MODOFIED on Oct 14 (Wed), 2015]
#	Now, the script will preview $commoner_shift_flank, and, if it is nonzero,
#	either the reference or reconstructed MSA will be shifted entirely,
#	to nullify $commoner_shift_flank before beginning the main processes.
#
#	[MODIFIED on Oct 15 (Thu), 2015]
#	The rather long blocks to display the position-shift map were now replaced by
#	the subroutine 'pr_map_shifts_respos'.
#
#	[MODIFIED on Oct 21 (Wed), 2015]
#	Now, the subroutine 'classify_error_assoc_w_sgl_cblk' outputs the penalty of the error as well.
#
#	[MODIFIED on Oct 24 (Sat), 2015]
#	BUGs were fixed in the subroutines, 'framework_cblock' and 'undo_mv_latter_of_cblock_pair'.
#
#	[MODIFIED on Oct 25 Sun), 2015]
#	(1) The subroutine, 'undo_sgl_bw_fw_move_on_2maps', was replaced with 'undo_sgl_bw_fw_move_on_2maps_2'.
#	The difference between the two in in how to handle "inpurities," i.e., gaps and sites with different values of shifts,
#	within the framework to be moved. (See the header of 'undo_sgl_bw_fw_move_on_2maps_2' for details.)
#	(2) The subroutine 'undo_mv_latter_of_cblock_pair' now handles any errors in satellite subroutines as
#		the situation hard to undo the move, instead of as its own error.
#	
# classify_msa_errors_via_mblks.1.pl:
#	This script uses the mini-block-organization of position shifts to classify the MSA errors.
#	It was created by disabling functions on the block-organization of position shifts.
#	(CREATION STARTED on Oct 24 (Fri), 2014, ENDED on Oct 10 (Mon), 2014.)
#
#	[MODIFIED on Oct 2 (Fri) through Oct 4 (Sun), 2015]
#	Now the script can handle a special case of iCII, which were overlooked before the modification.
#
# classify_msa_errors.1.pl:
#	This Perl script classifies MSA errors by comparing the reference MSA and a reconstructed MSA
#	in each segment that was erroneously reconstructed by an aligner.
#	It was created by taking advantage of the framework of 'compare_ref_vs_rec_prank_scores.stripped.pl'.
#	(CREATION STARTED on Oct 10 (Fri), 2014, ENDED on Oct xx (yyy), 2014.)
#
# compare_ref_vs_rec_prank_scores.stripped.pl:
#	This script is functionally identical to 'compare_ref_vs_rec_prank_scores.pl',
#	although it has many commented out commands and blocks stripped off.
#	(CREATED on Oct 7 (Tue), 2014.)
#
# compare_ref_vs_rec_prank_scores.pl:
#	This Perl script compares the reference MSA and a reconstructed MSA
#	in terms of the profile-profile pairwise score used in PRANK.
#	It was created by modifying 'compare_ref_vs_rec_mafft_consist.pl'.
#	(CREATION STARTED on Sep 18 (Thu), 2014, ENDED on Sep xx (yyy), 2014)
#
# compare_ref_vs_rec_mafft_consist.pl:
#	This Perl script compares the reference MSA and a reconstructed MSA
#	in terms of the consistency score used in (the refinement mode of) MAFFT.
#	It was created by modifying 'compare_ref_vs_rec_compllimment.pl'.
#	(CREATION STARTED on Sep 12 (Fri), 2014, ENDED on Sep xx (yyy), 2014)
#
# compare_ref_vs_rec_compllimment.pl:
#	This is a slightly modified version of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
#	The major difference from the predecessor is that it now accepts four input file paths,
#	instead of the path to an input directory as the predecessor did.
#	(CREATION STARTED on Aug 8 (Fri), 2014, ENDED on Aug 9 (Sat), 2014)
#
# compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl:
#	This revised version of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' uses the Perl package, "MyTreeMap_indels_spt_odr.pm,"
#	instead of "MyTreeMap_indels.pm" as the predecessor was using.
#	And it also underwent some additional minor modifications,
#	which were mainly imported from "br_infer_psm_indel_events_spt_odr_dawg.v2.pl"
#	in the "Proof_of_Principle2" session for the framework to enable indel likelihood analyses.
#	(CREATED on Aug 5 (Tue), 2014)
#
#	[REVISED on Aug 6 (Wed), 2014]
#	For the indel analyses (enumerating parsimonious indel scenarios
#	plus calculating the approximate probabilities of the gap configurations of the segments),
#	the script now uses the input tree as it is, 
#	even if it has a bifurcated root.
#
# compare_ref_vs_rec_abs_scoring.PhyML.pl:
#	This Perl script compares the reference MSA and a reconstructed MSA
#	in terms of the 'absolute score', which is the summation
#	of the log-likelihood of gap pattern and that of the residue pattern.
#	It was created by importing some parts from the script 'compare_ml_msa.bch.extJC.6t.PhyML.pl'.
#	

use strict;
use warnings;

	# MODIFIED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 28 (Tue), 2014). #
use MyArrayManipu_num qw(
	diff_nmarray
	nr_nmarray
	log_sum_exp
	binsearch_num
	);
##use MyArrayManipu_num qw( log_sum_exp );

	# MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 7 (Wed), 2015). #
use MyArrayManipu qw( diff_array binsearch_string );
##use MyArrayManipu qw( binsearch_string );

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext
	tfm_labels2indices
	);

use MyMalnIO qw(
	extract_fasta
	extract_alignment
	pr_ODEN_input
	pr_phylip_aln
	);
	## pr_clustal_like_lite_aa
use MyMalnManipu qw( rm_gapcolumn2 );

use MyTreeManipu_II qw(
	extract_sgltree_from_file
	input_tree_file output_tree
	);

	# ADDED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg
	);

	# ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl' (on Sep 18 (Thu), 2014). #
	# => MODIFIED in creation of 'classify_msa_errors.1.pl' (on Oct 19 (Sun), 2014). #
use MyTreeManipu_III qw( make_node2abs_depth  make_node2ct_exoffs );
##use MyTreeManipu_III qw( make_node2abs_depth );

use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );

use MyPAML_parser qw( retrieve_aa_freqs );
	##prepare_paml_ctrl 

use MyApBayes_Score_MSA_ExtJC qw(
	readin_set_trees3 log_akaike_weights
	);
	## apbayes_scoring_msa_extJC_f 

	### The following three modules are for calculating log-likelihoods of substitution patterns. ###


use MyPhyML_parser qw(
	est_resfreqs_frm_msa print_custom_sbst_model
	extract_phyml_statistics2
	fetch_clmwise_likelihoods_phyml
	);
	## extract_phyml_statistics


=DISABLED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/07).
	### ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl'. ###
use MyPRANK_Scores qw(
	calc_br2tr_mtrx_TN
	calc_prob_descresstats_cond_ancstat
	calc_prob_descresstats
	set_prank_align_hmm
	prank_score_2profile_aln
	);
=cut # End of "DISABLED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/07)."

	### ADDED in creation of 'classify_msa_errors.1.pl'. ###

use MyMalnComparison qw(
	assign_resnums_to_clms
	map_shifts_respos_bw_2msas

	merge_miniclasses_shift_respos
	find_trivial_blocks
	lump_together_similar_blocks
	);
	## ## lump_together_similar_blocks_proto
	##	(REMOVED from the list in creation of 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.) #
	## br_list_classes_shift_respos
	## list_classes_shift_respos
	## dpsm_infer_sep_patterns_for_classes
	## split_unnatural_miniclasses
	## reorganize_classes_shift_respos
	##	End of "(REMOVED from the list in creation of 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.)" #


use MyTreeMap_indels_spt_odr qw(
	extract_cls_gblocks
	msa_rep_gpatterns
	scan_gbcls_for_long_gaps
	create_initial_set_indels_spt_odr
	);
	## enum_psm_indel_scns_for_msa

use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);

use MyTaxonomy qw(
	fetch_ext_offsprings_nr
	fetch_common_ancestors
	); # ADDED on Oct 28 (Tue), 2014. #

	### END of "ADDED in creation of 'classify_msa_errors.1.pl'." ###

	### ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/07). (PM) ###

use MyTreeMap_indels_spt_odr_finer qw(
	enum_psm_indel_hsts_for_msa_finer_PRE
	enum_psm_indel_hsts_for_msa_finer_1ST_HF
	);

use MyDetect_purge_cands qw(
	mk_node2sw_gstats
	);

	### End of "ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/07). (PM)" ###

	### Added in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/09). (PM2) ##
use MyTreeMap_indels_sppl qw(
	create_br2eqbr
	);
	### End of "Added in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/09). (PM2)" ##

	### Added in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (PM3) ##
use MyTreeMap_sppl qw(
	wrapper_mk_node2exoffs_msa_rows
	);

use MyMapAlnShear qw(
	mk_map_alnshear_clmbased
	print_map_alnshear
	);
	# mk_map_alnshear ... REMOVED in the same way as 'map_aln_shear_clm_based.ver0.6.pl', in creation of 'map_alnshear_cb_and_srd.ver0.6.pl' (on 2020/03/23).
	# ADDED 'mk_map_alnshear_clmbased', in the same way as 'map_aln_shear_clm_based.ver0.6.pl', in creation of 'map_alnshear_cb_and_srd.ver0.6.pl' (on 2020/03/23).

	### End of "Added in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (PM3)" ##

		## ADDED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12). (PM) ##
use MyLikelihoods_respattern_msa_dna qw(
	calc_br2tr_mtrx_TN
	calc_node2ntfreqs_td
	calc_prob_descresstats_cond_ancstat2
	calc_joint_prob_ancresstats_and_ancstat2
	calc_clm_probs
	);

use MySRD_Map qw(
	mk_srd_map
	print_srd_map

	tally_srds_in_map
	print_sums_srds_in_map
	brwise_distribute_srds_in_map
	);
	# ADDED 'tally_assoc_srds_and_alnshears' & 'print_sums_assoc_srds_and_alnshears', in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/24). (PM) #
	# ADDED 'brwise_distribute_srds_assoc_w_alnshear' in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). (PM-REV) #
	#
	# REPLACED 'tally_assoc_srds_and_alnshears', 'print_sums_assoc_srds_and_alnshears' & 'brwise_distribute_srds_assoc_w_alnshear'
	# WITH 'tally_assoc_srds_in_map', 'print_sums_srds_in_map' & 'brwise_distribute_srds_in_map',
	# in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/03). (PM) #
	# => REPLACED 'tally_assoc_srds_in_map' WITH 'tally_srds_in_map', in DEBUGGING 'ingr_stat_srds_in_correct_segs.ver0.6.pl'  (on 2020/04/03). (PM) #


use MyANEX_IO qw(
	readin_dawg_tree2
	);
	# print_out_bch_br_parsimonize_dollo_indel_histories
		## END of "ADDED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12). (PM)" ##

		### (IMPORTED from 'map_aln_shear_clm_based.ver0.6.pl' in creation of 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.) (PM) ###

	### ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' (on 2020/03/19). (PM) ###
use MyPosShiftMap qw(
	list_minicls_shift_respos_clm_based
	);
	### End of "ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' (on 2020/03/19). (PM)" ###

	# ADDED for DEBUGGING on 2020/03/21. (PM2) #
use MyTreeMap_II_xgap qw( );
	# End of "ADDED for DEBUGGING on 2020/03/21. (PM2)" #

		### END of "(IMPORTED from 'map_aln_shear_clm_based.ver0.6.pl' in creation of 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.) (PM)" ###

		### (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (PM-REV) ###
use MyCluster_distribution qw(
	batch_initializer
	batch_pr_file_distri_xl_fmt
	);
		### END of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (PM-REV)" ###


=DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (SR-1)
sub summarize_info_cblock ($\@\@\@\%\%); # A BUG fixed on Oct 23 (Fri), 2015.
##sub summarize_info_cblocks ($\@\@\@\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 22 (Thu), 2015). #
=cut # END of "DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (SR-1)"

sub pr_map_shifts_respos (\*\@\@$$); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 15 (Thu), 2015). #
sub blockwise_framework ($$\@\@\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 12 (Mon), 2015). #
sub framework_cblock ($$\@\@\@\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 12 (Mon), 2015). #
sub scout_possible_undoing ($\%\@$); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 12 (Mon), 2015).

		# MODIFIED in DEBUGGING 'classify_msa_errors_via_mblks.2.pl' (on Oct 25 (Sun), 2015). #
sub undo_sgl_bw_fw_move_on_2maps_2 (\%$$$\@\@);
##sub undo_sgl_bw_fw_move_on_2maps (\%$$$\@\@); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 13 (Tue), 2015). #

sub rm_zero_shifts_at_ends (\@$\@\@\@\@\@\@\@$\$\@); # ADDED in creation of  'classify_msa_errors_via_mblks.2.pl' (on Oct 18 (Sun), 2015). #

=DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (SR-2)
sub undo_mv_latter_of_cblock_pair (\@$$$\@\@\%\%\%$); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 14 (Wed), 2015). #
=cut # END of "DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (SR-2)"

sub sub_enum_end_sequences ($$$\@\@\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 10 (Sat), 2015). #
sub enum_end_sequences ($\@\@\%\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 10 (Sat), 2015). #

=DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (SR-3)
sub classify_error_assoc_w_sgl_cblk ($\@$$$\%$\%\%\%\@\@\@\@\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 6 (Tue), 2015). #

sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$\@); # MODIFIED on Oct 14 (Wed), 2015. #
##sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$); # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 9 (Fri), 2015). #

sub cal_init_psm_cands_for_twin_msas (\%\%\%\%\%\@\@\@$$); # ADDED in creation of 'classify_msa_errors.1.pl' (on Oct 20 (Mon), 2014). #
=cut # END of "DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (SR-3)"

sub reorganize_set_indels (\@\@); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 25 (Sat), 2014). #
sub identify_equiv_indels (\%\%\@\@$$\%\%\%); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 28 (Tue), 2014). #
sub effective_distance ($$$\%\%\%$$); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 30 (Thu), 2014). #
sub find_ovlp_flanking_indels ($$\@\@\%$\%\%\%$$); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 29 (Wed)- 30 (Thu), 2014). #


=DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (SR-3+1)
sub register_association_cblk_and_indels ($$\@\%\%\@\@\@); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Oct 31 (Fri), 2014). #
=cut # END of "DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (SR-3+1)"

sub if_collapse_indep_ins (\@\@\%\%$); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Nov 4 (Tue), 2014). #
sub if_part_nbr_mrg_2ins (\@\@\@$$\%\%$); # ADDED in creation of 'classify_msa_errors_via_mblks.1.pl' (on Nov 12 (Wed), 2014). #


sub read_list_pw_lclhom (\$$); # ADDED in creation of 'compare_ref_vs_rec_mafft_consist.pl' (on Sep 15 (Mon), 2014). #

sub readin_list_segs2 (\$\@); # ADDED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##sub readin_list_segs (\$\%);
sub readin_simul_params (\$);
sub inspect_indel_events (\@\@$$$);

=DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12).
sub readin_dawg_tree2 (\$); # ADDED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##sub readin_dawg_tree (\$); # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011 
=cut # END of "DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12)."

sub ln_ml_sgl_seq2 (\$\%); # ADDED in creation of 'compare_apbayes_msa.bch.extJC.6.pl' on Mar 7 (Mon), 2011 #
sub add_dummy_seq (\@\@\%\%\%\%\%\@); # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 12 (Tue), 2011 #
####sub even_brlen_2seq_tree (\%$);
####sub parse_paml_main_output (\$);
##sub fetch_clmwise_lnlks (\$$);

$| = 1; # Flush the buffer of standard output everytime data flows into it.

my $program_path = $0;
my ($program_name) = ($program_path =~ m/([^\/]+)$/); 

my $plscript_name = $program_name; # Added to supress the error messages (when checking the script) on 2020/03/09. #

	### Fixed Parameters ###

	# (IMPORTED from 'srd_map.ver0.6.pl' in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11).) (GL1) #

our @RESIDUES = our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
##our @RESIDUES = our @NUCLEOTIDES = @MyPRANK_Scores::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our $SV_RATIO_R = 1; # Ratio of purine-type transition rate to transversion rate.
our $SV_RATIO_Y = 1; # Ratio of pyrimidine-type transition rate to transversion rate.

##my @cond_probs_gapped = (); # ADDED in creation of 'test_calc_jnt_prob_ancresstats2.pl'. #
##for (1 .. $CT_NUCLS) { push @cond_probs_gapped, 1; }

	# Equilibrium residue frequencies. #
my @root_resfreqs = (0.25, 0.25, 0.25, 0.25);
##my @root_resfreqs = (0.1, 0.2, 0.3, 0.4);

my @labels_resfreqs = ();
foreach my $res (@NUCLEOTIDES) { push @labels_resfreqs, "f($res)"; }

our $DIS_LEN_SRD = 30; # Maximum number of sites (i.e., SRDs) to be displayed in a line of the SRD Map.
our $CT_EFF_DGTS = 2;  # Number of effective digits with which each SRD will be displayed.

	# END of "(IMPORTED from 'srd_map.ver0.6.pl' in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11).) (GL1)" #


	# INPORTED from 'ml_scoring_msa.bch.pl' on Mar 15 (Tue), 2011 #
(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;
our $DIS_LEN = 60; # Maximum length of the sequences to be displayed in a line

$MyANEX_IO::DIS_LEN = $DIS_LEN; # Added in REVISION (on 2019/07/07). => IMPORTED from 'srd_map.ver0.6.pl' in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11).) (GL2)" #

##our $DIS_LEN_SHIFT = 50; # Maximum length of the run of shifts to be displayed in a line.
our $DIS_LEN_SHIFT = 30; # Maximum length of the run of shifts to be displayed in a line.

		# ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/09. (GL) #
our $CHAR_NOSHEAR = '.'; # A character indicating that the position (=site x branch) has NO alignment-shear.

our $DIS_LEN_ASMAP = 30; # The maximum number of sites displayed in a line of the Alignment-Shear Map.
our $MIN_WDTH_FLD_ASMAP = 6; # The minimum width of each field of the Alignment-Shear Map.
		# End of "ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/09. (GL)" #

##our $GAP = '-'; # Character representing a gap
our $CHAR_ABS =  our $GAP  = '-';
our $CHAR_PRES = our $NGAP = 'X';

	# (ADDED on Nov 29 (Mon), 2010) #
our $TOP_NODE  = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;

	# (ADDED on Mar 7 (Mon), 2011) #
our $TOT_BOOT = -1; # Because bootstrap values are NOT necessary for this analysis. 
our @INDICES_ATTR = ($INDX_LENGTH, $INDX_BOOT, $TOT_BOOT);
our $COMMENT = '#';  # Indicator of the comment line in the tree file

our $INDX_STRING = undef;
	# Specifies the index for the reference to a string in the element array of a set of non-redundant columns.
	# = undef if the element itself is the reference to a string.


	### Parameters to be adjusted ###

			# (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (GL) #
our @BOUNDS_SRD = (0.005, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.995); 
			# End of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (GL)" #


			# (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (GL3) #
	# The level of the amount of outputs:
	# 0, nothing; 1, very brief; 2, a bit brief; 3, a bit verbose; 5, very verbose.

our $OUTPUT_LEVEL = 0;
##our $OUTPUT_LEVEL = 1;
##our $OUTPUT_LEVEL = 2;
##our $OUTPUT_LEVEL = 3;
##our $OUTPUT_LEVEL = 6;

			# End of "(ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (GL3)" #

	# The level of tests to examine the behavior of the script (ADDED on Sep 1 (Sun), 2012):
	# 0, nothing; 1, very brief; 2, a bit brief; 3, a bit verbose; 5, very verbose.

		# For the main program: #
##our $TEST_LEVEL = 0;
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##$TEST_LEVEL = 6;

		# For the main program, the 2nd round. #
		# (ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' on Oct 8 (Thu), 2015.)
our $TEST_LEVEL2 = 0;
###our $TEST_LEVEL2 = 1;
##our $TEST_LEVEL2 = 2;
###our $TEST_LEVEL2 = 3;
##$TEST_LEVEL2 = 6;


		# For examining how new types of errors could be identified.
		# (ADDED on Oct 2 (Fri), 2015.)
our $TEST_LEVEL_newtype = 0;
##our $TEST_LEVEL_newtype = 5;

		# For the subroutine, 'cal_init_psm_cands_for_twin_msas':
		# (ADDED in creation of 'classify_msa_errors.1.pl' (on Oct 20 (Mon), 2014).) #
our $TEST_LEVEL_cipc = 0;
##our $TEST_LEVEL_cipc = 5;

		# For the subroutine, 'cal_init_psm_cands_for_sgl_msa':
		# (ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 9 (Fri), 2015).) #
our $TEST_LEVEL_cipc2 = 0;
##our $TEST_LEVEL_cipc2 = 5;

		# For the subroutine, 'undo_mv_latter_of_cblock_pair':
		# (ADDED on debugging of 'classify_msa_errors_via_mblks.2.pl' (on Oct 16 (Fri), 2015).) #
our $TEST_LEVEL_umlp = 0;
##our $TEST_LEVEL_umlp = 6;

	# The level of debugging output for 'extract_phyml_statistics2' (ADDED on Aug 26 (Sun), 2012)
our $DBG_LEVEL_EXPHYML = 0;

our $FLG_RMV_TMPOUT = 1; ## For actual (large-scale) analyses
##our $FLG_RMV_TMPOUT = 0; ## For TESTING the script ##
	# = 1 if you want to remove temporary directories and files under them,
	# = 0 if you want to keep them.


	# Parameters necessary for screening the erroneous segments before calculating the initial candidate of the parsimonious indel scenario(s)... To save the running time . #
	# (ADDED in creation of 'classify_msa_errors.1.pl'.)

		# Upper-bound on the accepted length of a gap-block cluster (i.e., gapped segment) in the reference/reconstructed MSA.
our $UPPER_BOUND_ANAL_LEN_GBCLS = $MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_LEN_GBCLS = 150;

		# Upper-bound on the accepted length of a run of gaps in the reference/reconstructed MSA.
our $UPPER_BOUND_ANAL_GAPLEN = $MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN = 120;


		### Upper-bound on the accepted length of an erroneous segment. ###
		### (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/29.) (GL) ###
our $UPPER_BOUND_ANAL_LEN_SEG = 300; 
		### End of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/29.) (GL)" ###

		### ADJUST THE FOLLOWING GLOBAL VARIABLES ONLY WHEN DEBUGGING ###

	## (MODIFIED in creation of 'classify_msa_errors.1.pl'.) ##

			# (MODIFIED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/19.) (GL) #
			# (=> IMPORTED to 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.) (GL) #

	## Level of test output for 'MyPosShiftMap.pm'.
	## (0, silent; 1, very brief; 2, brief; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

		# For 'list_minicls_shift_respos_clm_based'.
$MyPosShiftMap::TEST_LEVEL_lmc = 0;
#$MyPosShiftMap::TEST_LEVEL_lmc = 6;

			# (ADDED for DEBUGGING on 2020/03/21.) (GL1) #
		# For 'select_pcls_for_psp_blks'. #
$MyPosShiftMap::TEST_LEVEL_spfpb = 0; 
#$MyPosShiftMap::TEST_LEVEL_spfpb = 5; 
			# End of "(ADDED for DEBUGGING on 2020/03/21.) (GL1)" #

			# (ADDED for DEBUGGING on 2020/03/22.) (GL2) #
		# For 'enum_psm_clusterings_for_new_psps'. #
$MyPosShiftMap::TEST_LEVEL_epcfnp = 0; 
#$MyPosShiftMap::TEST_LEVEL_epcfnp = 5; 
			# End of "(ADDED for DEBUGGING on 2020/03/22.) (GL2)" #

=For the predecessor, 'map_aln_shear.ver0.6.pl'. (GL)
	## Level of test output for 'MyMalnComparison.pm'.
	## (0, silent; 1, very brief; 2, brief; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

		# For '(br_)list_classes_shift_respos'.
$MyMalnComparison::TEST_LEVEL_lc = 0;
##$MyMalnComparison::TEST_LEVEL_lc = 6;
=cut # End of "For the predecessor, 'map_aln_shear.ver0.6.pl'. (GL)"

	## Level of test output for 'MyTreeMap_II_xgap.pm'.
	## (0, silent; 1, very brief; 2, brief; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

        # For 'map_parsim_states_sankoff_xgap_clm'. #
$MyTreeMap_II_xgap::TEST_LEVEL_mpssxc = 0;
#$MyTreeMap_II_xgap::TEST_LEVEL_mpssxc = 5;

			# END of "(MODIFIED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/19.) (GL) #
			# (=> IMPORTED to 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.) (GL)" #


	## Level of test output for 'MyTreeMap_D.pm'.
	## (Basically, believe in the behaviours of the subroutines in the package.)
	## (ADDED in creation of 'classify_msa_errors.1.pl'.)

		# For 'map_parsim_states_dollo_msa'.
$MyTreeMap_D::TEST_LEVEL = 0;


	## Control the level of verboseness of the test message for 'MyTreeMap_indels_spt_odr.pm'.
	## (Basically, believe in the behaviours of the subroutines in the package.)
	## [RESURRECTED in creation of 'classify_msa_errors.1.pl'.]

$MyTreeMap_indels_spt_odr::TEST_LEVEL1 = 0; # For 'extract_cls_glbocks' and 'msa_rep_gpatterns'.
$MyTreeMap_indels_spt_odr::TEST_LEVEL_SCN = 0; # For 'scan_gbcls_for_long_gaps'.
$MyTreeMap_indels_spt_odr::TEST_LEVEL2i = 0; # For 'create_initial_set_indels_spt_odr'.


		### END of "ADJUST THE FOLLOWING GLOBAL VARIABLES ONLY WHEN DEBUGGING" ###


	### Process the Command-line Arguments and Options. ###

			# (MODIFIED in creation of 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/11.) (ARGV) #
unless (@ARGV > 3+1) {
	print STDERR "\nUsage:  [perl] $program_name {inaln_file_ref} {inaln_file_rec} {inlist_segs} {inlist_params} {out_top_dir} [options] \n";

				# (IMPORTED from 'srd_map.ver0.6.pl' on 2020/03/11.) (ARGV) #
	print STDERR "   where the [options] can consist of any (or none) of the following:\n\n";
	print STDERR "            --BASE_FREQS=f(A),f(G),f(C),f(T)   (default=", join (',', @root_resfreqs), "),\n"; 
	print STDERR "                     where f(N) is the frequency of nucleotide 'N' at the root;\n\n";
	print STDERR "            --TN_PARAMS=alpha1,alpha2,beta     (default=$SV_RATIO_R,$SV_RATIO_Y,1),\n";
	print STDERR "                     where alpha1 is the transitional substitution rate between A and G,\n";
	print STDERR "                           alpha2 is the transitional substitution rate between C and T,\n";
	##print STDERR "                           beta is the transversional substitution rate.\n";
	print STDERR "                           beta is the transversional substitution rate;\n\n"; # Modified in creation of 'srd_map.ver0.6.pl' on 2020/03/06. (ARGV)

			# (ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (ARGV) #
	print STDERR "            --DIS_LEN_SRD=integer     for the maximum number of sites in a line of the SRD Map (default=$DIS_LEN_SRD);\n\n";
	print STDERR "            --CT_EFF_DGTS=integer     for the number of effective digits when displaying each SRD (default=$CT_EFF_DGTS).\n";
			# End of "(ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (ARGV)" #
				# END of "(IMPORTED from 'srd_map.ver0.6.pl' on 2020/03/11.) (ARGV)" #
	die "$! .\n\n";
}
=For the predecessor, 'map_aln_shear.ver0.6.STRIPPED.pl'. (ARGV)
unless (@ARGV > 4) {
	print STDERR "\nUsage:  [perl] $program_name {inaln_file_ref} {inaln_file_rec} {inlist_segs} {inlist_params} {out_top_dir} \n";
	## [ --intree={input_tree_file (default: undef)}  ]\n";  # Became OBSOLETE in creation of 'classify_msa_errors_via_mblks.alpha.pl'. #
	die "$! .\n\n";
}
=cut # END of "For the predecessor, 'map_aln_shear.ver0.6.STRIPPED.pl'. (ARGV)"
			# END of "(MODIFIED in creation of 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/11.) (ARGV)" #

my ($inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params, $outtoptopdir) = @ARGV[0 .. 3+1];


	### Check the Command-Line Options. ###
	### (IMPORTED w/ modifications from 'srd_map.ver0.6.pl' on 2020/03/11.) (CL) ###

for (my $k=3; $k <= $#ARGV; $k++) { # Preliminary for-loop (over additional command-line arguments). #

	my $clarg = $ARGV[$k];
	unless ($clarg =~ s/^(\-)+//) { next; } # Skip if the argument lacks preceding '-'s. #

	my ($option, $value) = split /\=/, $clarg;
	unless ((defined $value) and ($value ne '')) {
	    print STDOUT "[CAUTION '$plscript_name'] Found an UNEXPECTED FORMAT of the commmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument in an unexpected format.
	}
	my $org_option = $option;
	$option = uc $option; # Capitalize the option, in advance.

=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (CL)
	if ($option eq 'MIN_WD_SIZE') {
	    $MIN_WDSIZE = $value;
	} elsif ($option eq 'MAX_WD_SIZE') {
	    $MAX_WDSIZE = $value;
	} elsif ($option eq 'THRSH_PV') {
	    $THRSH_PV = $value;
	} elsif ($option eq 'IF_MODPV') {
	    $IF_MODPV = $value;
	} elsif ($option eq 'THRSH_PV0') {
	    $THRSH_PV0 = $value;

	} elsif ($option eq 'BASE_FREQS') {
=cut # End of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (CL)"

	if ($option eq 'BASE_FREQS') { # MODIFIED in creation of 'srd_map.ver0.6.pl'. (on 2020/03/04). (CL)
	    my @new_root_resfreqs = split /\,/, $value;
	    if (@new_root_resfreqs == 4) {
		@root_resfreqs = @new_root_resfreqs;
	    } else {
		print STDOUT "[CAUTION '$plscript_name'] An UNEXPECTED NUMBER of residue frequencies are specified by the argument, ", $ARGV[$k], ": $! !\n";
		print STDOUT "                         ... So, Use the DEFAULT residue frquencies, ", '@root_resfreqs =(', join (', ', @root_resfreqs), ") ...\n\n";
	    }

	} elsif ($option eq 'TN_PARAMS') {

	    my ($alpha1, $alpha2, $beta) = split /\,/, $value;    
	    if ((defined $beta) and ($beta>0)) {
		$SV_RATIO_R = $alpha1/$beta;
		$SV_RATIO_Y = $alpha2/$beta;
	    } else {
		print STDOUT "[CAUTION '$plscript_name'] The Tamura-Nei parameter, Beta, is ZERO or UNDEFINED  within the argument, ", $ARGV[$k], ": $! !\n";
		print STDOUT "                          ... So, USE the DEFAULT T-N ratios, (\$SV_RATIO_R, \$SV_RATIO_Y) = ($SV_RATIO_R, $SV_RATIO_Y) ...\n\n\n";
	    }

		# (ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (CL) #
	} elsif ($option eq 'DIS_LEN_SRD') {
		$DIS_LEN_SRD = $value;

	} elsif  ($option eq 'CT_EFF_DGTS') {
		$CT_EFF_DGTS = $value;

		# End of "(ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (CL)" #

	} else {
	    print STDOUT "[CAUTION '$plscript_name'] Found an UNEXPECTED OPTION, '$org_option', within the cmmmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument with an unexpected option.
	}

}	# End of the preliminary for-loop (over additional command-line arguments. #

	### End of "Check the Command-Line Options. ###
	### (IMPORTED w/ modifications from 'srd_map.ver0.6.pl' on 2020/03/11.) (CL)" ###


		# Modified in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (CL) #

my $outtopdir = "$outtoptopdir/Outputs";
my $outauxdir = "$outtoptopdir/Auxiliary";

=For the predecessor, 'classify_msa_errors_via_mblks.alpha2.pl'.
my $outtopdir = "$outtoptopdir/Interim";
my $out_summ1 = "$outtoptopdir/tbl_summ_class_msa_errors.1.txt";
my $out_summ2 = "$outtoptopdir/tbl_summ_class_msa_errors.2.txt";
=cut # END of "For the predecessor, 'classify_msa_errors_via_mblks.alpha2.pl'."

		# End of "Modified in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (CL)" #


		# (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/24.) (OF) #
		# (=> MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02.) (OF) #
##my $out_main_exp = "$outtoptopdir/tbl_assoc_exp_srds_and_alnshears.txt";
##my $out_main_obs = "$outtoptopdir/tbl_assoc_obs_srds_and_alnshears.txt";
my $out_main_exp = "$outtoptopdir/tbl_exp_srds_in_correct_segs.txt";
my $out_main_obs = "$outtoptopdir/tbl_obs_srds_in_correct_segs.txt";
		# End of "(ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/24.) (OF) #
		# (=> MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02.) (OF)" #

		# (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (OF-REV) #
		# (=> MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02.) (OF2) #
##my $out_dstr_srds = "$outtoptopdir/distributions_obs_srds.txt";
my $out_dstr_srds = "$outtoptopdir/distributions_obs_srds_in_corr_segs.txt";
		# End of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (OF-REV) #
		# (=> MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02.) (OF2)" #

my $logdir = "$outtoptopdir/LogFiles";


my $intree = $inlist_params; # ADDED in creation of 'classify_msa_errors_via_mblks.alpha.pl'. #

=For the predecessor, 'classify_msa_errors_via_mblks.alpha2.pl'. (CL)

for (my $k=5; $k<@ARGV; $k++) { # MODIFIED in creation of 'classify_msa_errors_via_mblks.alpha2.pl'. #
##for (my $k=11; $k<@ARGV; $k++) {...} # MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 6 (Tue), 2015). #

	my $arg = $ARGV[$k];
	unless ($arg =~ s/^\-\-?//) {
	    print STDOUT "[WARNING '$program_name'] Found an UNEXPECTED command-line ARGUMENT '$arg': $! !\n";
	    print STDOUT "                 ---  I will just skip it ...\n\n";
	    next;
	}
	if ($arg =~ s/^intree\=// ) {

		# MODIFIED in creation of 'classify_msa_errors_via_mblks.alpha.pl'. #
	    if ($arg =~ m/\S/) {
		print STDOUT "[INFO '$program_name'] The argument \$intree = '$intree' will be changed to '$arg' ...\n\n";
 		$intree = $arg;
	    }
	    ##($arg =~ m/\S/) and ($intree = $arg);

	} else {
	    print STDOUT "[WARNING '$program_name'] Found an UNEXPECTED command-line ARGUMENT '--$arg': $! !\n";
	    print STDOUT "                 ---  I will just skip it ...\n\n";
	}

}
=cut # END of "For the predecessor, 'classify_msa_errors_via_mblks.alpha2.pl'. (CL)"


	### Input/Output files/directories ###

	## Interim files ##

my $name_interim_trees = "tree_whole_full.txt";

		# Modified in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (IO-files) #
##my $interim_trees = "$outtopdir/$name_interim_trees";
my $interim_trees = "$outauxdir/$name_interim_trees";
		# End of "Modified in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (IO-files)" #

		# ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (IO-files) #
my $name_br_identifiers = "lists_branch_identifiers.txt";
my $lists_br_identifiers = "$outauxdir/$name_br_identifiers";
		# End of "ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (IO-files)" #

=For the predecessor, 'classify_msa_errors_via_mblks.alpha2.pl'. (IO-dirs)
	# Top Temporary directory for PAML #
	# temporary directory (reference) = "$top_tmpdir_ref/s{seg_ID}/g{indx_gpattern}/t{indx_tree}"
	# temporary directory (reconstructed) = "$top_tmpdir_rec/s{seg_ID}/g{indx_gpattern}/t{indx_tree}"
my $top_tmpdir_ref = "$outtopdir/Tmp/Ref";
my $top_tmpdir_rec = "$outtopdir/Tmp/Rec";
=cut # END of "For the predecessor, 'classify_msa_errors_via_mblks.alpha2.pl'. (IO-dirs)"

	## Main Output Files (1) ##
	## (ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/10.) (O-files 1) ##

my $name_out_refaln = "seg_refaln.aln";
my $name_out_recaln = "seg_recaln.aln";

=For the predecessor, 'ingr_stat_srds_in_correct_segs.ver0.6.pl'. (OF)
#
##my $name_olist_psclasses = 'list_pos_shift_classes.txt'; # Added on 2020/03/11. #
##
##			# (IMPORTED from 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/23.) (OF) #
##my $name_oinfo_psps = "info_pos_shift_patterns.txt"; # Added on 2020/03/20. #
##my $name_olist_psp_blks = 'list_pos_shift_pattern_blocks.txt'; # Added on 2020/03/20. #
##			# End of "(IMPORTED from 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/23.) (OF)" #
##
##my $name_out_main1  = "pos_shift_map.txt";
##my $name_out_main2  = "aln_shear_map.txt";
#
=cut # End of "For the predecessor, 'ingr_stat_srds_in_correct_segs.ver0.6.pl'. (OF)"

	## End of "(ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/10.) (O-files 1)" ##

	## Main Output Files (2) ##
	## (IMPORTED from 'srd_map.ver0.6.pl' in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11).) (O-files 2) ##
my $name_out_exp = 'expected_SRD_Map.txt';
my $name_out_obs = 'observed_SRD_Map.txt';
	## End of "(IMPORTED from 'srd_map.ver0.6.pl' in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11).) (O-files 2)" ##

	## Check Necessary Files/Directories ##

foreach my $file ($intree) {
    (defined $file) or next; ## ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PAML.pl' (on Mar 22 (Tue), 2011) ##

    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 8 (Fri), 2014). #

foreach my $infile ($inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params) { # MODIFIED in creation of 'compare_ref_vs_rec_prank_scores.pl'. #
    unless (( -f $infile ) and ( -r $infile )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$infile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ; 
    }
}

##foreach my $outdir ($outtopdir, $logdir, $top_tmpdir_ref, $top_tmpdir_rec) {
foreach my $outdir ($outtopdir, $outauxdir, $logdir) { # Modified in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (O-dirs) #
    unless ( -d $outdir ) {
	unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create a TOP OUTPUT DIRECTORY '$outdir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
    }
    unless (( -r $outdir ) and ( -w $outdir )) {
	unless (0 == system ("chmod u+rw $outdir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to CHANGE the permission MODES of the top output directory '$outdir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<<<<<< Program name: $program_name (path: $program_path) >>>>>>>>\n\n";

		# (ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/09.) (DC) #
print STDOUT '$DIS_LEN = ', $DIS_LEN, " ,\n";
print STDOUT '$DIS_LEN_SHIFT = ', $DIS_LEN_SHIFT, " ,\n";
print STDOUT '$DIS_LEN_ASMAP = ', $DIS_LEN_ASMAP, " ,\n";
print STDOUT '$MIN_WDTH_FLD_ASMAP = ', $MIN_WDTH_FLD_ASMAP, " .\n\n";

print STDOUT '$CHAR_ABS = ', $CHAR_ABS, " ,\n";
print STDOUT '$CHAR_PRES = ', $CHAR_PRES, " ,\n";
print STDOUT '$CHAR_NOSHEAR = ', $CHAR_NOSHEAR, " .\n\n\n";
		# End of "(ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/09.) (DC)" #

print STDOUT '@BOUNDS_SRD = (', join (', ', @BOUNDS_SRD), ") .\n\n\n"; # ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26. (DC-REV) #


print STDOUT "Current working directory: $CURR_WORK_DIR .\n\n";

print STDOUT "Input files:\n";

if (defined $intree) { 
    print STDOUT "     (Tree)      $intree .\n\n";
}

	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 8 (Fri), 2014). #

print STDOUT "   (reference MSA)         $inaln_file_ref,\n";
print STDOUT "   (reconstructed MSA)     $inaln_file_rec,\n";
print STDOUT "   (list of segments)      $inlist_segs .\n\n";

print STDOUT "\nTop directory for the output file system: {topoutdir} = $outtopdir .\n\n";

			# Modified in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (DC2) #
print STDOUT "Directory to keep auxiliary outputs:  {outauxdir} = $outauxdir .\n\n";

print STDOUT "Interim output list of candidate trees:  {outauxdir}/$name_interim_trees .\n\n"; # 

##print STDOUT "Interim output list of candidate trees:  {topoutdir}/$name_interim_trees .\n";
			# END of "Modified in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (DC2)" #

			# ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (DC3) #

print STDOUT "Lists of Identifiers of Branches for each candidate tree:  {outauxdir}/$name_br_identifiers .\n\n";

##print STDOUT "Directories for the output regarding each segment:  {segoutdir} = {topoutdir}/s{seg_id}/ .\n\n";
print STDOUT "Directories for the output regarding each correct segment, which left-flanks the erroneous segment w/ ID = {seg_id}:  {segoutdir} = {topoutdir}/s{seg_id}/ .\n\n"; # Modified in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/02). (DC1) #

print STDOUT "   Output files regarding each segment:\n\n";

if ($OUTPUT_LEVEL > 1) { # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-A). #
    print STDOUT "     (reference MSA)          {segoutdir}/$name_out_refaln ,\n";
    print STDOUT "     (reconstructed MSA)      {segoutdir}/$name_out_recaln ,\n";
} # End of "if ($OUTPUT_LEVEL > 1) { # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-A). # ...}"

=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (DC1)
##
##if ($OUTPUT_LEVEL > 2) { # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-B). #
##    print STDOUT "     (Position-Shift Classes) {segoutdir}/$name_olist_psclasses ,\n"; # Added on 2020/03/11. #
##} # End of "if ($OUTPUT_LEVEL > 2) { # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-B). # ...}"
##
##if ($OUTPUT_LEVEL >=5) { # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-C). #
##			# (IMPORTED from 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/23.) (DC) #
##    print STDOUT "    (Position-Shift Patterns) {segoutdir}/$name_oinfo_psps ,\n"; # Added on 2020/03/20. #
##    print STDOUT "     (PS-Pattern Blocks)      {segoutdir}/$name_olist_psp_blks ,\n"; # Added on 2020/03/20. #
##			# End of "(IMPORTED from 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/23.) (DC)" #
##} # End of "if ($OUTPUT_LEVEL >=5) { # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-C). # ...}"
##
##if ($OUTPUT_LEVEL > 1) { # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-D). #
##    print STDOUT "     (Position-Shift Map)     {segoutdir}/$name_out_main1 ,\n";
##} # End of "if ($OUTPUT_LEVEL > 1) { # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-D)." #
##
=cut # End of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (DC1)"

if ($OUTPUT_LEVEL > 0) { # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-E). #

##    print STDOUT "     (Alginment-Shear Map)    {segoutdir}/$name_out_main2 ,\n\n"; # For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (DC2)

			# END of "ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (DC3)" #

			# IMPORTED w/ modifications from 'srd_map.ver0.6.pl', in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (DC3-ADD) #
    print STDOUT "      (Expected SRD Map)      {segoutdir}/$name_out_exp ,\n";
    print STDOUT "      (Observed SRD Map)      {segoutdir}/$name_out_obs .\n\n";
			# End of "IMPORTED w/ modifications from 'srd_map.ver0.6.pl', in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (DC3-ADD)" #

} # END of "if ($OUTPUT_LEVEL > 0) { # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-E). # ...}"

print STDOUT "\n"; # (Added in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC-F) #


	### ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/24). (DC1) ###

##print STDOUT "MAIN Outputs (on the association between SRDs and \"Alignment Shear\"s):\n\n";
print STDOUT "MAIN Outputs (on SRDs in the \"Correct\" segments):\n\n"; # MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/02). (DC2) #

print STDOUT "   (For Expected SRDs) $out_main_exp ,\n";
print STDOUT "   (For Observed SRDs) $out_main_obs .\n\n";

			# ADDEd in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). (DC-REV) #
##print STDOUT "Sub-MAIN Outputs (of distributions of Observed SRDs, on and off \"Alignment Shear\"s,  along each of branches):\n\n";
print STDOUT "Sub-MAIN Outputs (of distributions of Observed SRDs, in the \"Correct\" segments,  along each of branches):\n\n"; # MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/02). (DC3) #

print STDOUT "       $out_dstr_srds .\n\n";
			# End of "ADDEd in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). (DC-REV)" #

	### END of "ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/24). (DC1)" ###

=For the predecessor, 'classify_msa_errors_via_mblks.alpha2.pl'. (DC1)
print STDOUT "Temporary directories for PAML (obsolete):  {top_tmpdir}/s{seg_id}/g{indx_gpattern}/t{indx_tree}/,\n"; # MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
##print STDOUT "Temporary directories for PAML:  {top_tmpdir}/s{seg_id}/g{indx_gpattern}/t{indx_tree}/,\n";
print STDOUT "   where {top_tmpdir} = $top_tmpdir_ref for the reference MSA,\n";
print STDOUT "                      = $top_tmpdir_rec for the reconstructed MSA.\n\n";

	# MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 6 (Tue), 2015). #
print STDOUT "Summary outputs:\n";
print STDOUT "   (1st Round) $out_summ1 ,\n";
print STDOUT "   (2nd Round) $out_summ2 .\n\n";
##print STDOUT "Summary output: $out_summ .\n\n";
=cut # END of "For the predecessor, 'classify_msa_errors_via_mblks.alpha2.pl'. (DC1)"

print STDOUT "Directory for the log files: {log_dir} = $logdir .\n\n";
print STDOUT "Log files:\n";

		# MODIFIED in creation of 'classify_msa_errors.1.pl'. #
print STDOUT "   (for analyzing the reference & reconstructed MSAs for each segment) {log_dir}/log_seg_anal.\{seg_id\}.txt .\n\n";


=For the predecessor, 'classify_msa_errors_via_mblks.alpha2.pl'. (DC2)
if ($FLG_RMV_TMPOUT) {
	print STDOUT "* All the temporary outputs will be removed .\n\n";
} else {
	print STDOUT "* All the temporary outputs will be kept .\n\n";
}
=cut # End of "For the predecessor, 'classify_msa_errors_via_mblks.alpha2.pl'. (DC2)"

			# (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC3) #
print STDOUT "\nLevel of the amount of Outputs:\n\n";

print STDOUT '   $OUTPUT_LEVEL      = ', $OUTPUT_LEVEL, " .\n\n";
			# End of "(ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (DC3") #

print STDOUT "\nLevels of tests for debugging or behavior tests:\n\n";

print STDOUT '   $TEST_LEVEL        =  ', $TEST_LEVEL, " ,\n";

print STDOUT '   $TEST_LEVEL2       =  ', $TEST_LEVEL2, " ,\n"; # ADDED in creation of 'classify_msa_errors_via_mblks.2.pl' (on Oct 8 (Thu), 2015). #

print STDOUT '   $TEST_LEVEL_newtype   = ', $TEST_LEVEL_newtype,  " ,\n"; # ADDED on Oct 3 (Fri), 2015. #

print STDOUT '   $TEST_LEVEL_cipc   = ', $TEST_LEVEL_cipc,  " ,\n";

print STDOUT '   $TEST_LEVEL_cipc2   = ', $TEST_LEVEL_cipc2,  " ,\n"; # ADDED on Oct 9 (Fri), 2015. #
print STDOUT '   $TEST_LEVEL_umlp   = ', $TEST_LEVEL_umlp,  " ,\n"; # ADDED on Oct 16 (Fri), 2015.

print STDOUT '   $DBG_LEVEL_EXPHYML =  ', $DBG_LEVEL_EXPHYML, " ,\n";
print STDOUT '   $FLG_RMV_TMPOUT    =  ', $FLG_RMV_TMPOUT, " .\n\n";


	# MODOFIED in creation of 'classify_msa_errors.1.pl'. #

			# (MODIFIED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/19.)  (DC) #
			# (=> IMPORTED to 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.) (DC) #

print STDOUT "\nParameters in the Perl package 'MyPosShiftMap.pm':\n\n";

print STDOUT '$MyPosShiftMap::TEST_LEVEL_lmc  = ', $MyPosShiftMap::TEST_LEVEL_lmc , " .\n\n";

print STDOUT '$MyPosShiftMap::TEST_LEVEL_spfpb  = ', $MyPosShiftMap::TEST_LEVEL_spfpb , " .\n\n"; # ADDED for DEBUGGING on 2020/03/21. (DC1) #

print STDOUT '$MyPosShiftMap::TEST_LEVEL_epcfnp = ', $MyPosShiftMap::TEST_LEVEL_epcfnp , " .\n\n"; # ADDED for DEBUGGING on 2020/03/22. (DC3) #

			# ADDED for DEBUGGING on 2020/03/21. (DC2) #
print STDOUT "\nParameters in the Perl package 'MyTreeMap_II_xga.pm':\n\n";

print STDOUT '$MyTreeMap_II_xgap::TEST_LEVEL_mpssxc  = ', $MyTreeMap_II_xgap::TEST_LEVEL_mpssxc , " .\n\n"; 
			# END of "ADDED for DEBUGGING on 2020/03/21. (DC2)" #

=For the predecessor, 'map_aln_shear.ver0.6.pl'. (DC)
print STDOUT "\nParameters in the Perl package 'MyMalnComparison.pm':\n\n";

print STDOUT '$MyMalnComparison::TEST_LEVEL_lc  = ', $MyMalnComparison::TEST_LEVEL_lc , " .\n\n";
=cut # End of "For the predecessor, 'map_aln_shear.ver0.6.pl'. (DC)"

			# END of "(MODIFIED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/19.)  (DC) #
			# (=> IMPORTED to 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.) (DC)" #


		# (Further ADDED on Oct 20 (Mon), 2014.) #

print STDOUT "\nParameters in the Perl package 'MyTreeMap_indels_spt_odr.pm':\n\n";

print STDOUT '$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_LEN_GBCLS = ', $UPPER_BOUND_ANAL_LEN_GBCLS, " ,\n";

print STDOUT '$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN = ', $UPPER_BOUND_ANAL_GAPLEN, " .\n\n\n";


			### (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/27.) (DC-REV) ###

print STDOUT "< Substitution Model Parameters: >\n\n";

print STDOUT 'Freqs (', join (', ', @NUCLEOTIDES), ')_root = (', join (', ', @root_resfreqs), ") ,\n\n";
print STDOUT '($SV_RATIO_R, $SV_RATIO_Y) = ', "($SV_RATIO_R, $SV_RATIO_Y) .\n\n\n";

print STDOUT "< Other Parameters: >\n\n";

print STDOUT '$DIS_LEN_SRD = ', $DIS_LEN_SRD, " ,\n\n";
print STDOUT '$CT_EFF_DGTS = ', $CT_EFF_DGTS, " .\n\n\n";

			### End of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/27.) (DC-REV)" ###


			### (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/29.) (DC-REV2) ###
print STDOUT '$UPPER_BOUND_ANAL_LEN_SEG = ', $UPPER_BOUND_ANAL_LEN_SEG , " .\n\n\n";
			### End of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/29.) (DC-REV2)" ###


	### Read in the list of segments. ###

print STDOUT "\n<<<<< Now, read in the list of segments >>>>>\n\n";

our @LABELS_SEG_ATTR = ('Indx_spacer', 'Start_in_Ref', 'End_in_ref', 'Start_in_Rec', 'End_in_Rec', 'Diagnosis'); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
##our @LABELS_SEG_ATTR = ('seg_type', 'ref_len', 'rec_len', '#len_grp', 'alt_state');

# @{$list_segments[$s]} = @attrs gives information on the $s th segment, including its ID.
#
my $list_segments = readin_list_segs2 ($inlist_segs, @LABELS_SEG_ATTR);# MODIFIED  in creation of 'compare_ref_vs_rec_compllimment.pl'. #

# @{$list_segments[$s]} = ($seg_ID, @attrs) gives information on the $s th segment.
#
##my $list_segments = readin_list_segs ($inlist_segs, %LABEL_SATTR2INDX);
unless (defined $list_segments) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_list_segs2' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_segments = @{$list_segments};

print STDOUT "\n#{segments} = $ct_segments\n\n";
print STDOUT join ("\t", 'Seg_ID', @LABELS_SEG_ATTR[1..$#LABELS_SEG_ATTR]), "\n\n"; # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
##print STDOUT join ("\t", 'Seg_ID', @LABELS_SEG_ATTR), "\n\n";

for (my $s=0; $s<$ct_segments; $s++) { print STDOUT join ("\t",  @{$list_segments->[$s]}), "\n"; }
print STDOUT "\n\n";


	### Read in the reference MSA ###
	### (ADDED in creation of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' on Aug 26 (Sun), 2012) ###

print STDOUT "\n<<<<< Now, read in the reference MSA >>>>>\n\n";

(my $ref_msa, my $sequence_ids, undef) = extract_alignment ($inaln_file_ref); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##(my $ref_msa, my $sequence_ids, undef, undef) = extract_fasta ($inaln_file_ref);
unless (defined $ref_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the reference MSA '$inaln_file_ref' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_sequences = @{$sequence_ids};
unless ($ct_sequences > 1) {
	print STDOUT "[ERROR '$program_name'] The number of sequences ($ct_sequences) is TOO SMALL ! I will stop here ...\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $totlen_refaln = length ${$ref_msa->[0]};

print STDOUT "#{sequences in MSA} =  $ct_sequences .\n";
print STDOUT "#{columns in MSA}   = $totlen_refaln .\n\n";

	### Read in the reconstructed MSA ###
	### (ADDED in creation of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' on Aug 26 (Sun), 2012) ###

print STDOUT "\n<<<<< Now, read in the reconstructed MSA >>>>>\n\n";

(my $rec_msa, my $seq_ids2, undef) = extract_alignment ($inaln_file_rec); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##(my $rec_msa, my $seq_ids2, undef, undef) = extract_fasta ($inaln_file_rec);
unless (defined $rec_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the reconstructed MSA '$inaln_file_rec' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


unless ($ct_sequences == @{$seq_ids2}) {
	print STDOUT "[ERROR '$program_name'] THe number of sequences in the core alignment (", scalar (@{$seq_ids2}), ") CLASHES with that of the reference alignment ($ct_sequences) !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
for (my $k=0; $k<$ct_sequences; $k++) {
	my $seqid = $sequence_ids->[$k];
	my $seqid2 = $seq_ids2->[$k];
	unless ($seqid eq $seqid2) {
	    print STDOUT "[ERROR '$program_name'] The ID of the $k th sequence in the reconstructed MSA ($seqid2) does NOT MATCH that in the reference MSA ($seqid) !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
}

my $totlen_recaln = length ${$rec_msa->[0]};

print STDOUT "#{columns in MSA} = $totlen_recaln .\n\n"; 


	### Either retrieve or reconstruct a phylogenetic tree that our main process is based on. ###

my @std_set_cand_trees = ();
our %STATS_PHYML;
	# = ($key => scalar_value ($key is the string between '.' and ':', which ' ' replaced with '_'),
	# 'discrete_gamma' => [$ncate, $alpha] (or undef),
	# 'nuc_freqs' => [$fA, $fC, $fG, $fT] (or undef),
	# 'rel_rates' => [$r_AC, $r_AG, $r_AT, $r_CG, $r_CT, $r_GT] (or undef),
	# 'rate_mtrx' => \@mtrx, with $mtrx[$i]->[$j] giving the instantaneous rate from the $i-th to the $j-th nucleotide (in the order: A, C, G, T) (or undef)
	# );
our @REL_RATES = ();

if ((defined $intree) and ($intree =~ m/\S/)) {

	## Read in the input tree (ESSENTIAL for the 'beta' version) ##

    print STDOUT "\n<<<<< Reading-in the Input Tree >>>>>\n\n";

    # @{$info_dawg_tree} = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr)
    #
    my $info_dawg_tree = readin_dawg_tree2 ($intree); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
    ##my $info_dawg_tree = readin_dawg_tree ($intree);
    unless (defined $info_dawg_tree) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'readin_dawg_tree2' FAILED on the input tree file '$intree' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    @std_set_cand_trees = (['NA', 'NA', 'NA', 0, 0, @{$info_dawg_tree}]);

} else {

	print STDOUT "[ERROR '$program_name'] NO INPUT TREE FILE IS SPECIFIED !!\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;

}


	## Output the input tree to an interim file ##

print STDOUT "\n<<<<< Now, create an interim file recording the input trees >>>>>\n\n";
##print STDOUT "\n<<<<< Now, create an interim file recording the input tree >>>>>\n\n";

my $ct_cand_trees = @std_set_cand_trees;

unless ($ct_cand_trees == 1) { # ADDED in creation of 'classify_msa_errors.1.pl'. #

	print STDOUT "[ERROR '$program_name'] ILLEGAL INPUT!! ... There are $ct_cand_trees candidate trees, whereas this program accepts only a SINGLE INPUT TREE !!\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


my $fh_otree0  = open_output_file ($interim_trees, 0);
unless (defined $fh_otree0) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'open_output_file' FAILED on the interim tree file '$interim_trees' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
print $fh_otree0 "<< Candidate trees that are considerably likely (#{candidates} = $ct_cand_trees) >>\n\n";
##print $fh_otree0 "<< Candidate trees that are considerably likely (#{candidates} = 1) >>\n\n";


for (my $t=0; $t<$ct_cand_trees; $t++) { # Preliminary for-loop (over candidate trees). # ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 6 (Wed), 2014). #

	# REVISED in creation of 'compare_ml_msa.bch.extJC.6t.PAML.pl' (on Mar 22 (Tue), 2011) #
	# => Further REVISED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
	my ($ln_ml_msa, $node2pa0, $node2ch0, $name2id0, $id2name0, $br_attr0) = @{$std_set_cand_trees[$t]}[1,5..9];
	##my ($node2pa0, $node2ch0, $name2id0, $id2name0, $br_attr0) = @{$std_set_cand_trees[0]}[5..9];
	####my ($node2pa0, $node2ch0, $name2id0, $id2name0, $br_attr0) = @{$info_dawg_tree};
	my $topnode_id0 = $name2id0->{$TOP_NODE};

	print $fh_otree0 "< The $t th candidate >\n\n";
	##print $fh_otree0 "< The 0 th candidate >\n\n";
	print $fh_otree0 "{ Phylogenetic Tree (top node ID: $topnode_id0) }\n\n";

	unless ( output_tree ($topnode_id0, %{$node2ch0}, %{$id2name0}, %{$br_attr0}, *${$fh_otree0}, 0, 0) ) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'output_tree' FAILED to output the dawg guide-tree onto the interim tree file '$interim_trees' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

		# FIXED on Mar 24 (Thu), 2011 #
		# => REVISED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
	print $fh_otree0 "\n{ Log-likelihood of the tree for the MSA: $ln_ml_msa }\n\n\n";
	##print $fh_otree0 "\n{ Log-likelihood of the MSA under the tree: 0 }\n\n";

}	# Preliminary for-loop (over candidate trees). # ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 6 (Wed), 2014). #

close ($fh_otree0);


	## Calculate the Akaike weights of the candidate trees

print STDOUT "\n<<<<< Now, calculate the Akaike weights of the candidate trees and necessary inputs  >>>>>\n\n";

# $ln_akaike_wgts->[$c] is the logarithmic Akaike weight factor of the $c th candidate tree.
#
my $ln_akaike_wgts = log_akaike_weights (@std_set_cand_trees, 3);
unless (defined $ln_akaike_wgts) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'log_akaike_weights' FAILED on the tree list extracted from '$interim_trees' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

=for TESTING the script (to see if the final results correctly take into account the effects of the Akaike weights ) (probably, it will NOT influence the comparison of the scores among alternatives)
for (my $t=0; $t<$ct_cand_trees; $t++) { $ln_akaike_wgts->[$t]++; }
=cut


	## Display the candidate trees, and their Akaike weights.
	## Also, assign absolute depths to the nodes of each tree. 
	## (MODIFIED in creation of 'compare_ref_vs_rec_prank_scores.pl'.)

print STDOUT "\n<<<<< Now, Output candidate trees, their Akaike weights, and their attributes  >>>>>\n\n";

my @set_node2depth = (); # %{$set_node2depth[$t]} = ($node_id => $depth, ...) for all nodes in the $t th candidate tree.
my @set_node2ct_exoffs = (); # %{$set_node2ct_exoffs[$t]} = ($node_id => #{exteior offsprings}, ...) for all nodes in the $t th candidate tree.
my @set_node2seq_indx = (); # %{$set_node2seq_indx[$t]} = ($node_id => {index in @{$sequence_ids}}, ...) for all exterior nodes for the $t th candidate tree. # (ADDED on Oct 28 (Tue), 2014.) #

my @set_br2eq = (); # %{$set_br2eq[$t]} = ($br_id => $equiv_br_id, ...) for all branches with "equivalent"s. # ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/09). #

my @set_node2trstats2score = (); # %{$set_node2trstats2score[$t]} = ($node_id => {"$state_bf:$state_af" => $score, ...) for all interior nodes in the $t th candidate tree.

		# ADDED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12). (In-Tree-Beg) #
my @set_br2tr_mtrx = ();
my @set_node2ntfreqs = ();
		# End of "ADDED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12). (In-Tree-Beg)" #


		# ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (In-Tree-Beg) #

# my $lists_br_identifiers = "$outauxdir/$name_br_identifiers";

my $fh_brids = open_output_file ($lists_br_identifiers, 0);
unless (defined $fh_brids) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'open_output_file' FAILED on the interim tree file '$interim_trees' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
print $fh_brids "<< Branch-Identifiers of the Candidate trees (#{candidates} = $ct_cand_trees) >>\n\n";
		# END of "ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (In-Tree-Beg)" #

		# (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26).) (In-Tree-Beg-REV) #
my $ct_all_brs;
my %id2name;
		# End of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26).) (In-Tree-Beg-REV)" #

for (my $t = 0; $t<$ct_cand_trees; $t++) { # Preliminary for-loop (over candidate trees). #

	my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[$t]};
	my $topnode_id = $name2id->{$TOP_NODE};
	my $ln_awgt = $ln_akaike_wgts->[$t];

	print STDOUT "\< The $t th candidate tree (topnode ID: $topnode_id\; Log_Akaike_weight = $ln_awgt) >\n\n";

	print STDOUT "Sum (-log(Pv)) = $sum_neg_log_pv\n";
	print STDOUT "log (Prob(MSA|Tree)_ML) = $ln_ml_msa\n";
	print STDOUT "#{free parameters} = $ct_free_params\n";
	print STDOUT "AIC = $aic\n";
	print STDOUT "BIC = $bic\n\n";

	my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	unless ( output_tree ($topnode_id, %{$node2ch}, %{$id2name}, %{$br_attr}, *STDOUT, $fl_bootlabel, $fl_nodeid) ) {
	    print STDOUT "[ERROR '$program_name'] 'output_tree' FAILED to output the $t th candidate tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n";

	##print STDERR "\n\nNode_ID --> Parent_ID\n\n";
	##foreach my $nodeid (sort keys %{$node2pa}) { print STDERR "$nodeid  --> ", $node2pa->{$nodeid}, "\n"; }
	##print STDERR "\n\n";

	print STDOUT join ("\t", 'Branch_ID', 'length', 'P-value', 'parent_node_ID'), "\n\n";
	foreach my $brid (sort {$a <=> $b} keys %{$br_attr}) { # 1st inner foreach-loop (over branches). #
	    my $attr = $br_attr->{$brid};
	    my ($brlen, $pvalue) = @{$attr}[$INDX_LENGTH, $INDX_BOOT];
	    (defined $brlen) or ($brlen = 'NA');
	    (defined $pvalue) or ($pvalue = 'NA');
	    my $pa = $node2pa->{$brid};
	    print STDOUT join ("\t", $brid, $brlen, $pvalue, $pa), "\n";
	}				# End of the 1st inner foreach-loop (over branches). #
	print STDOUT "\n\n";

			# (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (In-Tree-REV) #
	if ($t == 0) {
	    $ct_all_brs = scalar keys %{$br_attr};
	    %id2name = %{$id2name};
	}
			# End of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (In-Tree-REV)" #


		## ADDED in creation of 'compare_ref_vs_rec_prank_scores.pl'. ##

		# Check if all the interior nodes (including the root) have 2 children each. #

	foreach my $node_id (sort {$a <=> $b} keys %{$node2ch}) { # 2nd inner foreach-loop (over interior nodes). #
	     my $children = $node2ch->{$node_id};
	     unless ((defined $children) and (@{$children}>0)) {
		print STDOUT "[WARNING '$program_name'] The node '$node_id' has an EMPTY (or UNDEFINED) children set: $! ! ... I will delete its entry in ", '%{$node2ch}', " ...\n\n";

		delete $node2ch->{$node_id};
		next;
	     }

	    unless (@{$children} == 2) {
		my $ct_ch = @{$children};
		print STDOUT "[ERROR '$program_name'] The node '$node_id' has $ct_ch children, instead of 2 children as required by this script: $! !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }
	}			# End of the 2nd inner foreach-loop (over interior nodes). #

	print STDOUT "\n ... Confirmed that all interior nodes of this tree have 2 child nodes each ...\n\n";


		# Assign depths to the nodes. #

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDOUT "[ERROR '$program_name'] 'make_node2abs_depth' FAILED on the $t th candidate tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n ... The subroutine 'make_node2abs_depth' Ended ...\n\n";

		# (MOVED to below with modification (on Oct 19 (Sun), 2014).) #
	##print STDOUT join ("\t", 'Node_ID', 'Depth'), "\n\n";
	##foreach my $node (sort {$a <=> $b} keys %{$node2depth}) { print STDOUT join ("\t", $node, $node2depth->{$node}), "\n"; }
	##print STDOUT "\n\n\n";

	$set_node2depth[$t] = $node2depth;


		# Assign #{exterior offsprings} to the nodes. #
		# (ADDED on Oct 19 (Sun), 2014.)

	my $node2ct_exoffs = make_node2ct_exoffs (%{$node2ch}, %{$node2depth});
	unless (defined $node2ct_exoffs) {
	    print STDOUT "[ERROR '$program_name'] 'make_node2ct_exoffs' FAILED on the $t th candidate tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n ... The subroutine 'make_node2ct_exoffs' Ended ...\n\n";

	$set_node2ct_exoffs[$t] = $node2ct_exoffs;


		# (MOVED from above with modification (on Oct 19 (Sun), 2014).) #

	print STDOUT join ("\t", 'Node_ID', 'Depth', '#{exteior_offsprings}'), "\n\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2depth}) { print STDOUT join ("\t", $node, $node2depth->{$node}, $node2ct_exoffs->{$node}), "\n"; }
	print STDOUT "\n\n\n";



		# Assign indices (for @{$sequence_ids}) to the exterior nodes. #
		# (ADDED on Oct 28 (Tue), 2014.) #

	print STDOUT join ("\t", 'Seq_indx', 'Seqname', 'Node_ID') , "\n\n";

	my %node2seq_indx;

	for (my $k=0; $k<$ct_sequences; $k++) { # 5th inner for-loop (over aligned sequences). #

	    my $seqname = $sequence_ids->[$k];
	    my $nodeid = $name2id->{$seqname};
	    unless (defined $nodeid) {
		print STDOUT "[ERROR '$program_name'] The $k th sequence (name: $seqname) has NO CORRESPONDING NODE in the $t th candidate tree !\n\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }

	    $node2seq_indx{$nodeid} = $k;

	   print STDOUT join ("\t", $k, $seqname, $nodeid), "\n";

	}			# End of the 5th inner for-loop (over aligned sequences). #
	print STDOUT "\n\n\n";

	$set_node2seq_indx[$t] = \%node2seq_indx;


		# Create %br2eq. #
		# (ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/09.) (In-Tree1) #

	my $br2eq = create_br2eqbr ($topnode_id, %{$node2ch});
	unless (defined $br2eq) {
	    print STDOUT "[ERROR '$program_name'] 'create_br2eqbr' FAILED on the $t th candidate tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	print STDOUT join ("\t", 'Br_ID', 'Equiv_br_ID'), "\n\n";

	foreach my $br (sort {$a <=> $b} keys %{$br2eq}) {
		print STDOUT join ("\t", $br, $br2eq->{$br}), "\n";
	}
	print STDOUT "\n\n\n";

	$set_br2eq[$t] = $br2eq;
	
		# End of "(ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/09.) (In-Tree1)" #


		# Compute & Record the Branch Identifiers. #
		# (ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/10.) (In-Tree2) #

	my ($br2exoffs_msa_rows, $node2depth_W, $nodeids_in_msa, $nodeid2msa_row)
		= wrapper_mk_node2exoffs_msa_rows ($topnode_id, %{$node2ch}, @{$sequence_ids}, %{$name2id});
	unless (defined $br2exoffs_msa_rows) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'wrapper_mk_node2exoffs_msa_rows' FAILED on the input MSA & the $t th candidate tree !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print STDOUT "\n ... The subroutine, 'wrapper_mk_node2exoffs_msa_rows', Ended on the $t th candidate tree ...\n\n") if ($TEST_LEVEL >1);


	print $fh_brids "\< The $t th candidate tree (topnode ID: $topnode_id) >\n\n";


	print $fh_brids "\n", join ("\t", 'br_ID', '[msa,rows,under,br]'), "\n\n";

	my %br2cnct_exoffs_msa_rows;
	foreach my $br (sort {$a <=> $b} keys %{$br2exoffs_msa_rows}) { # Preliminary foreach-loop (over branches). #
		my $cnct_exoffs_msa_rows = join (',', @{$br2exoffs_msa_rows->{$br}});
		$cnct_exoffs_msa_rows = "[$cnct_exoffs_msa_rows]";
		$br2cnct_exoffs_msa_rows{$br} = $cnct_exoffs_msa_rows;

		print $fh_brids join ("\t", $br, $cnct_exoffs_msa_rows), "\n";

	} # End of the preliminary foreach-loop (over branches). #

	print $fh_brids "\n\n";

		# END of "(ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/10.) (In-Tree2)" #


		## (ADDED in creation of 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/12.) (In-Tree3) ##

		# Assign a substitution transition matrix to each branch. #
	
	my $br2tr_mtrx = calc_br2tr_mtrx_TN (%{$br_attr}, @root_resfreqs, $SV_RATIO_R, $SV_RATIO_Y, 1);
	unless (defined $br2tr_mtrx) {
		print STDOUT "[ERROR '$program_name'] 'calc_br2tr_mtrx_TN' FAILED on the $t th candidate tree !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
   (print STDOUT " ... Subroutine, 'calc_br2tr_mtrx_TN', Ended on the $t th candidate tree ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>3) {

	print STDOUT "\n<< Preliminary output (1): Substitution Transition Matrices Assigned to the Branches >>\n\n";

	foreach my $br_id (sort {$a <=> $b} keys %{$br2tr_mtrx}) {	# 1st preliminary foreach-loop (over branches). #
	    my $tr_mtrx = $br2tr_mtrx->{$br_id};

	    print STDOUT ' ' x 3, "( Branch ID: $br_id )\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'From\To', @RESIDUES), "\n\n";
	    for (my $r=0; $r<$CT_NUCLS; $r++) { print STDOUT ' ' x 3, join ("\t", $RESIDUES[$r], @{$tr_mtrx->[$r]}), "\n"; }
	    print STDOUT "\n\n";
	}						# End of the 1st preliminary foreach-loop (over branches). #

   } # End of "if ($TEST_LEVEL>3) {...}"

	$set_br2tr_mtrx[$t] = $br2tr_mtrx;


		# Calculate the residue frequences at ALL nodes. #

	my $node2ntfreqs = calc_node2ntfreqs_td (@root_resfreqs, %{$br2tr_mtrx}, %{$node2ch}, $topnode_id);

	unless (defined $node2ntfreqs) {
		print STDOUT "[ERROR '$plscript_name'] The Preliminary Subroutine 'calc_node2ntfreqs_td' FAILED on the $t th candidate tree !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n\n";
		die ;
	}
   (print STDOUT " ... Subroutine, 'calc_node2ntfreqs_td', Ended on the $t th candidate tree ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>3) {

	print STDOUT "\n<< Contents of the Preliminary Output (2), ", '%{$node2ntfreqs}', " >>\n\n";

	print STDOUT join ("\t", 'Node_ID', @labels_resfreqs), "\n\n";

	foreach my $node_id (sort {$a <=> $b} keys %{$node2ntfreqs}) { # 2nd preliminary outer foreach-loop (over nodes). #

	    my $ntfreqs = $node2ntfreqs->{$node_id};

	    print STDOUT join ("\t", $node_id, @{$ntfreqs}), "\n\n";


	}			# End of the 2nd preliminary outer foreach-loop (over nodes). #
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL>3) {...}"

	$set_node2ntfreqs[$t] = $node2ntfreqs;

		## END of "(ADDED in creation of 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/12.) (In-Tree3)" ##

}			# end of the preliminary for-loop (over candidate trees). #

print STDOUT "\n";

close ($fh_brids); # ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (In-Tree-End) #

### exit 0; ### For TESTING the Upper part of the script



	### MAIN ANALYSES: Classify the MSA Errors in each Erroneous Segments by comparing Reference & Reconstructed segmental MSAs. ###
	## (MODIFIED in creation of 'classify_msa_errors.1.pl'.)

print STDOUT "\n\n<<<<<<< The Main Analyses: Classify the MSA Errors in each Erroneous Segments by comparing Reference & Reconstructed segmental MSAs. >>>>>>>\n\n";


		# (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/24.) (Pre-MAIN) #
my $fh_main_obs = open_output_file ($out_main_obs, 0);
unless (defined $fh_main_obs) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$out_main_obs' !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $fh_main_exp = open_output_file ($out_main_exp, 0);
unless (defined $fh_main_exp) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$out_main_exp' !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
		# End of "(ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/24.) (Pre-MAIN)" #

=DISABLED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (MAIN-BEG)

my $fh_out1 = open_output_file ($out_summ1, 0);
unless (defined $fh_out1) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$out_summ1' !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
my $fh_out2 = open_output_file ($out_summ2, 0);
unless (defined $fh_out2) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$out_summ2' !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
=cut # END of "DISABLED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (MAIN-BEG)"

my $ct_anom_segs = my $ct_in_segs = my $ct_succ_segs = my $ct_fail_segs = my $ct_skipped_segs = 0;
my $ct_skipped_segs2 = 0; # ADDED in DEBUGGING on Aug 11 (Mon), 2014. #

my $ct_skipped_segs3 = 0; # ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/29.

#my $list_segments = readin_list_segs ($inlist_segs, %LABEL_SATTR2INDX);
# my $ct_segments = @{$list_segments};

my %ct_maps2freq; # = (#{maps} => {frequency}, ...) # ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' (on 2020/03/21). # => IMPORTED to 'map_alnshear_cb_and_srd.ver0.6.pl' (on 2020/03/23). #


		# [ Initialize the branch-wise distributions of observed SRDs, on anf off the Alignment-Shears. #
		# (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (BEG-MAIN-REV) # ] ... for the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'.
		#
		# Initialize the branch-wise distributions of observed SRDs, in the Correct Segments.
		# (=> MAJOR REVISION in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/03.)  (BEG-MAIN) #

##my @brwise_distri_obs_srds_on_as = my @brwise_distri_obs_srds_off_as = ();
my @brwise_distri_obs_srds = ();
	# NOTE: The last element of each array is the Total distribution over all branches. #

for (0 .. $ct_all_brs) {
	##push @brwise_distri_obs_srds_on_as, [];
	##push @brwise_distri_obs_srds_off_as, [];
	push @brwise_distri_obs_srds, [];
}
##batch_initializer (@brwise_distri_obs_srds_on_as, @BOUNDS_SRD);
##batch_initializer (@brwise_distri_obs_srds_off_as, @BOUNDS_SRD);
batch_initializer (@brwise_distri_obs_srds, @BOUNDS_SRD);


my @br_ids_for_display = (); # MOVED from within the for-loop over the segments, in revision of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). #

		# END of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (BEG-MAIN-REV) #
		# (=> MAJOR REVISION in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/03.)  (BEG-MAIN)" #


	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
my $last_ref_coord_end = my $last_rec_coord_end = 0;
##my $ref_coord_beg = my $rec_coord_beg = 0; # The 'beginning' coordinates of the segment in the reference MSA and reconstructed MSA, respectively. #

$GAP = uc $GAP; # Make the gap character upper-case. (ADDED in creation of 'compare_ref_vs_rec_mafft_consist.pl'.) #

OUTER: # ADDED on Oct 14 (Wed), 2015. #
##for (my $s = 0; $s<$ct_segments; $s++) { # Main outer for-loop (over MSA segments) #
for (my $s = 0; $s<= $ct_segments; $s++) { # Main outer for-loop (over MSA segments) # MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02. (MO-for) #

		## MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). ##

			# (MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02.) (M0) #

	my ($segid, $ref_coord_beg, $ref_coord_end, $rec_coord_beg, $rec_coord_end, $diagnosis) = ($s < $ct_segments) ? 
		@{$list_segments->[$s]} :
	   ('Right_END', $totlen_refaln, $totlen_refaln, $totlen_recaln, $totlen_recaln, 'Right_END');

# my $totlen_refaln = length ${$ref_msa->[0]};
# my $totlen_recaln = length ${$rec_msa->[0]};

	    # @LABELS_SEG_ATTR = ('Indx_spacer', 'Start_in_Ref', 'End_in_ref', 'Start_in_Rec', 'End_in_Rec', 'Diagnosis'); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
	    # @{$list_segments[$s]} = @attrs gives information on the $s th segment, including its ID.

	my $prev_segid = ($s > 0) ? $list_segments->[$s-1][0] : 'Left_END';

	my $c_segid = "$prev_segid:$segid";

	print STDOUT "<<<<<< Right-flanking Erroneous-Segment ID: $segid  (Diagnosis: $diagnosis); Correct-Segment ID: '$c_segid' >>>>>>\n\n";

=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (M0)
	my ($segid, $ref_coord_beg, $ref_coord_end, $rec_coord_beg, $rec_coord_end, $diagnosis) = @{$list_segments->[$s]};


	    # @LABELS_SEG_ATTR = ('Indx_spacer', 'Start_in_Ref', 'End_in_ref', 'Start_in_Rec', 'End_in_Rec', 'Diagnosis'); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
	    # @{$list_segments[$s]} = @attrs gives information on the $s th segment, including its ID.

#
##	unless ( uc ($diagnosis) =~ m/DIFFERENT$/ ) { # Skip segments NOT diagnosed as having reconstructed MSA 'DIFFERENT'  from the reference.
##
##   (print STDOUT "\n[INFORMATION] The $s th segment (ID: $segid) will be SKIPPED, because its reference MSA (in [$ref_coord_beg\, $ref_coord_end\) ) and reconstructed MSA (in [$rec_coord_beg\, $rec_coord_end\) ) are diagnosed as '$diagnosis' !\n\n") if ($TEST_LEVEL>0);
##
##	    $ct_skipped_segs++;
##
##	    next;
##	}
##
##	print STDOUT "<<<<<< Segment ID: $segid  (Diagnosis: $diagnosis) >>>>>>\n\n";
#
=cut # END of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (M0)"

			# End of "(MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02.) (M0)" #

=For TARGETTED DEBUGGING.
    ##if (($TEST_LEVEL_newtype>0) and ($segid ne '13')) {...}
    if ($segid ne '4') {
	print STDOUT "... This segment is irrelevant, thus its analysis is skipped ...\n\n";
	next;
    }
=cut


		# [ Remove the right flanking regions, if at all. # ] ... For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. #
		# Keep the right-flanking regions, which constitute the "Correct-Segment". #
		# (MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02.) (M1) #

	print STDOUT "<<<< Prepare for segments of reference/reconstructed MSAs >>>>\n\n"; # MOVED from below, in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02. (M1) #

	my @ref_seg_msa = my @rec_seg_msa = (); # MOVED from below, in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02. (M2) #

	##if ($last_ref_coord_end < $ref_coord_beg) {
	    ##my $len_rflnk = $ref_coord_beg - $last_ref_coord_end;
	    my $ref_len = $ref_coord_beg - $last_ref_coord_end;

			# MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
	    for (my $k=0; $k < $ct_sequences; $k++) {
		##my $tmp = uc substr (${$ref_msa->[$k]}, 0, $len_rflnk, '');
		my $tmp = ($ref_len>0) ? uc (substr (${$ref_msa->[$k]}, 0, $ref_len, '')) : '';
		$ref_seg_msa[$k] = \$tmp; # ADDED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02. (M1a) #
	    }
	    ##for (my $k=0; $k < $ct_sequences; $k++) { substr (${$ref_msa->[$k]}, 0, $len_rflnk, ''); }
	##}

	##if ($last_rec_coord_end < $rec_coord_beg) {
	    ##my $len_rflnk = $rec_coord_beg - $last_rec_coord_end;
	    my $rec_len = $rec_coord_beg - $last_rec_coord_end;

			# MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
	    for (my $k=0; $k < $ct_sequences; $k++) {
		##my $tmp = uc substr (${$rec_msa->[$k]}, 0, $len_rflnk, '');
		my $tmp = ($rec_len>0) ? uc (substr (${$rec_msa->[$k]}, 0, $rec_len, '')) : '';
		$rec_seg_msa[$k] = \$tmp; # ADDED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02. (M1b) #
	    }
	    ##for (my $k=0; $k < $ct_sequences; $k++) { substr (${$rec_msa->[$k]}, 0, $len_rflnk, ''); }
	##}

=Set aside for potential use.
			# ADDED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/02). (M1c) #
	my $ref_len = $ref_coord_beg - $last_ref_coord_end;
	my $rec_len = $rec_coord_beg - $last_rec_coord_end;
			# End of "ADDED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/02). (M1c)" #
=cut # End of "Set aside for potential use."


	print STDOUT "(ref_len, rec_len) = ($ref_len, $rec_len)\n\n"; # MOVED from below, in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02. (M3) #

		# END of "(MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02.) (M1)" #

	##print STDOUT "<<<< Prepare for segments of reference/reconstructed MSAs >>>>\n\n"; # MOVED to above, in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02. (M1) #

			# MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/02). (M2a) #
	##my $ref_len = $ref_coord_end - $ref_coord_beg;
	##my $rec_len = $rec_coord_end - $rec_coord_beg;
	my $ref_len_rfe = $ref_coord_end - $ref_coord_beg;
	my $rec_len_rfe = $rec_coord_end - $rec_coord_beg;
			# End of "MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/02). (M2a)" #

	my ($ref_len0, $rec_len0) = ($ref_len, $rec_len); # Keep the Initial Lengths. ## (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/24.) ##

	##my @ref_seg_msa = my @rec_seg_msa = (); # MOVED to above, in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02. (M2) #

			## Discard the Right-Flanking Erroneous Segment, if at all. ##
			# (MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02.) (M2b) #

	for (my $k=0; $k<$ct_sequences; $k++) { #  1st middle for-loop #

	    my $seg_ref = ($ref_len_rfe>0) ? substr (${$ref_msa->[$k]}, 0, $ref_len_rfe, '') : '';
	    my $seg_rec = ($rec_len_rfe>0) ? substr (${$rec_msa->[$k]}, 0, $rec_len_rfe, '') : '';

=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (M1)
	    my $seg_ref = substr (${$ref_msa->[$k]}, 0, $ref_len, '');
	    my $seg_rec = substr (${$rec_msa->[$k]}, 0, $rec_len, '');

	    push @ref_seg_msa, \$seg_ref;
	    push @rec_seg_msa, \$seg_rec;
=cut # End of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (M1)"

	}	# 2nd of the 1st middle for-loop #

			# END of "(MODIFIED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02.) (M2b)" #

	##print STDOUT "(ref_len, rec_len) = ($ref_len, $rec_len)\n\n"; # MOVED to above, in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/02. (M3) #



		# Update the last end-coordinates. #

	$last_ref_coord_end = $ref_coord_end;
	$last_rec_coord_end = $rec_coord_end;

		## END of "MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014)." ##

		# Copy the reconstructed MSA. #
		# (ADDED in creation of 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/12.) #

	my @cp_rec_seg_msa = ();
	for (my $k=0; $k<$ct_sequences; $k++) {
	    my $cp_seq = ${$rec_seg_msa[$k]};
	    $cp_rec_seg_msa[$k] = \$cp_seq;
	}
		# End of "(ADDED in creation of 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/12.)" #

=for targetted DEBUGGING (on Sep 22 (Sat) & 23 (Sun), 2012)
	##unless ($segid eq '0040') {...}
	unless ($segid eq '0014') {
	    $ref_coord_beg += $ref_len;
	    $rec_coord_beg += $rec_len;
	    $ct_skipped_segs++;
	    ##print STDOUT "\n --- Sgment with ID ne '0040' is skipped ---\n\n";
	    print STDOUT "\n --- Sgment with ID ne '0014' is skipped ---\n\n";
	    next;
	}
=cut


		## Prepare for the files necessary for analyzing the segment. ##

		# MODIFIED in creation of 'classify_msa_errors.1.pl'. #

	my $logfile_seg_anal = "$logdir/log_seg_anal.$segid.txt";
	my $fh_log = open_output_file ($logfile_seg_anal, 0);
	unless (defined $fh_log) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the log file '$logfile_seg_anal' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

		## Create the Segment-wise Output Directory & Set the path to the Output Files. ##
		## (ADDED in creatio of 'map_aln_shear.ver0.6.pl' on 2020/03/10.) (MAIN-Ofiles) ##

	my $segoutdir = "$outtopdir/s$segid";

   if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-Z) #

## print STDOUT "Directories for the output regarding each segment:  {segoutdir} = {topoutdir}/s{seg_id}/ .\n\n";
	unless ( -d $segoutdir) {
	    unless (0 == system ("mkdir -p $segoutdir" ) ) {
		print STDOUT "[ERROR '$program_name'] FAILED to create a segment-wise output directory, '$segoutdir': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	}
	unless (( -r $segoutdir ) and ( -w $segoutdir )) {
	    unless (0 == system ("chmod -r u+rw $segoutdir" ) ) {
		print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the segment-wise output directory '$segoutdir': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	}
   } # END of "if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-Z) # ...}"

	my $ofile_refaln = "$segoutdir/$name_out_refaln";
	my $ofile_recaln = "$segoutdir/$name_out_recaln";

=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (M2)
#
##	my $olist_psclasses = "$segoutdir/$name_olist_psclasses"; # Added on 2020/03/11.
##
##			# (IMPORTED from 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/23.) #
##	my $oinfo_psps = "$segoutdir/$name_oinfo_psps"; # Added on 2020/03/20.
##	my $olist_psp_blks = "$segoutdir/$name_olist_psp_blks"; # Added on 2020/03/20.
##			# End of "(IMPORTED from 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/23.)" #
#
#
##	my $out_main1 = "$segoutdir/$name_out_main1";
##	my $out_main2 = "$segoutdir/$name_out_main2";
#
=cut # End of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (M2)" 

			# IMPORTED w/ modifications from 'srd_map.ver0.6.pl', in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (Main-Ofiles) #
	my $out_srd_exp = "$segoutdir/$name_out_exp";
	my $out_srd_obs = "$segoutdir/$name_out_obs";
			# End of "IMPORTED w/ modifications from 'srd_map.ver0.6.pl', in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11). (Main-Ofiles)" #

## print STDOUT "   Output files regarding each segment:\n\n";
## print STDOUT "     (reference MSA)          {segoutdir}/$name_out_refaln ,\n";
## print STDOUT "     (reconstructed MSA)      {segoutdir}/$name_out_recaln ,\n";
## print STDOUT "     (Position-Shift Map)     {segoutdir}/$name_out_main1 ,\n";
## print STDOUT "     (Alginment-Shear Map)    {segoutdir}/$name_out_main2 ,\n\n";

		## End of "(ADDED in creatio of 'map_aln_shear.ver0.6.pl' on 2020/03/10.) (MAIN-Ofiles)" ##


		## FURTHER ADDED in creation of 'classify_msa_errors.1.pl'. #

		# Retrieve the tree information. #

	# ($sum_neg_log_pv through $bic are unnecessary for this program.) #
	#
	my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[0]};
	##my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[$0]}; # A BUG fixed on Oct 15 (Wed), 2014. #

	my $topnode_id = $name2id->{$TOP_NODE};
	my $node2depth = $set_node2depth[0];
	my $node2ct_exoffs = $set_node2ct_exoffs[0]; # ADDED on Oct 19 (Sun), 2014. #
	my $node2seq_indx = $set_node2seq_indx[0]; # ADDED on Oct 28 (Tue), 2014. # 

	my $br2eq = $set_br2eq[0]; # ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/09). #

				# ADDED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12). #
	my $br2tr_mtrx = $set_br2tr_mtrx[0];
	my $node2ntfreqs = $set_node2ntfreqs[0];
				# End of "ADDED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12)." #

	print $fh_log "<<<<< Input Tree (Top_node = $topnode_id) >>>>>\n\n";


   if ($TEST_LEVEL>1) {

	my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	unless ( output_tree ($topnode_id, %{$node2ch}, %{$id2name}, %{$br_attr}, *${$fh_log}, $fl_bootlabel, $fl_nodeid) ) {
	    print STDOUT "[ERROR '$program_name'] 'output_tree' FAILED to output the input tree (for analysis of position shifts): $! !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print $fh_log "\n\n";
   }

   if ($TEST_LEVEL>3) {
	print $fh_log "< Depths assigned to the Nodes >\n\n";
	print $fh_log join ("\t", 'Node_ID', 'Depth', 'Name'), "\n\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2depth}) { print $fh_log join ("\t", $node, $node2depth->{$node}), "\n"; }
	print $fh_log "\n\n";
   }

		## ADDED in creation of 'classify_msa_errors.1.pl'. ##

	print $fh_log "<<<<< Input MSA >>>>>\n\n";

	print $fh_log "#{Sequences} = $ct_sequences .\n";
	print $fh_log "#{Sites in the segment}_ref = $ref_len ,\n";
	print $fh_log "#{Sites in the segment}_rec = $rec_len .\n\n\n";


   if ($OUTPUT_LEVEL>1) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-A) #

			## MODIFIED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (MAIN-InMSA) ##

	# my $ofile_refaln = "$segoutdir/$name_out_refaln";
	# my $ofile_recaln = "$segoutdir/$name_out_recaln";

	my $fh_o_ref = open_output_file ($ofile_refaln, 0);
	unless (defined $fh_o_ref) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file \"$ofile_refaln\": $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	my $fh_o_rec = open_output_file ($ofile_recaln, 0);
	unless (defined $fh_o_rec) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file \"$ofile_recaln\": $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}


	unless ( pr_ODEN_input (@{$sequence_ids}, @ref_seg_msa, $DIS_LEN, *${$fh_o_ref}) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reference MSA of the $s th segment !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	}

	close ($fh_o_ref); # Added on 2020/03/10.

	print $fh_log "<< Original Segment of the Reference Alignment: ... recorded in \"$ofile_refaln\" . >>\n\n\n";


	unless ( pr_ODEN_input (@{$sequence_ids}, @rec_seg_msa, $DIS_LEN, *${$fh_o_rec}) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reconstructed MSA of the $s th segment !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	}

	close ($fh_o_rec); # Added on 2020/03/10.

	print $fh_log "<< Original Segment of the Reconstructed Alignment: ... recorded in \"$ofile_recaln\" . >>\n\n\n";

			## END of "MODIFIED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (MAIN-InMSA)" ##

   } # END of "if ($OUTPUT_LEVEL>1) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-A) # ...}"

		## ADDED for DEBUGGING the script (on Aug 11 (Mon), 2014). ##

	unless (($ref_len>0) and ($rec_len>0)) {

	    print STDOUT "\n[INFORMATION] Either reference or reconstructed sub-MSA (or both) is NULL! ... I will skip the calculation of the MAFFT consistent scores for this segment pair... \n\n";

	    ##foreach my $fh_log ($fh_log_ref, $fh_log_rec) { # MODIFIED in creation of 'compare_ref_vs_rec_mafft_consist.pl'. #
	    ####foreach my $fh_log ($fh_log_sref, $fh_log_srec, $fh_log_gref, $fh_log_grec) {...}
		print $fh_log "\n[INFORMATION] Either reference or reconstructed sub-MSA (or both) is NULL! ... I will skip the calculation of the MAFFT consistent scores for this segment pair (ID: $segid) ... \n\n\n";
	    ##}

	    $ct_skipped_segs2++;


	    close ($fh_log); # MODIFIED in creation of 'classify_msa_errors.1.pl'. #
	    ##foreach my $fh_log ($fh_log_ref, $fh_log_rec) { close ($fh_log); } # ADDED on Sep 15 (Mon), 2014. #

	    next;
	}
		## END of "ADDED for DEBUGGING the script (on Aug 11 (Mon), 2014)." ##


			### NEWLY ADDED in creation of 'classify_msa_errors.1.pl'. ###

		## Preliminary Process (0): Map the residue numbers (1-based) onto the reference & reconstructed MSAs. ##
			## (ADDED on Oct 28 (Tue), 2014.) ##

	print STDOUT "<<<<< Preliminary (0): Map the residue numbers onto the reference & reconstructed MSAs... >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (0): Map the residue numbers onto the reference & reconstructed MSAs... >>>>>\n\n";

	# $set_clminfo_ref/rec->[$c][$k] is the residue number (1-based) of the $k th sequence at the $c th column of the reference/reconstructed MSA.
	#		it is ($r + 1/2) if the column is occupied by a gap between the $r th and the ($r+1) th residues.
	#
	my $set_clminfo_ref = assign_resnums_to_clms (@ref_seg_msa, $GAP, 0);
	unless (defined $set_clminfo_ref) {
	    print STDOUT "[ERROR '$program_name'] 'assign_resnums_to_clms' FAILED on the reference MSA for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $set_clminfo_rec = assign_resnums_to_clms (@rec_seg_msa, $GAP, 0);
	unless (defined $set_clminfo_rec) {
	    print STDOUT "[ERROR '$program_name'] 'assign_resnums_to_clms' FAILED on the reconstructed MSA for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

   if ($TEST_LEVEL>3) { # (O)

	my $ct_clms_ref = @{$set_clminfo_ref};
	my $ct_clms_rec = @{$set_clminfo_rec};

	print $fh_log "<< Output of 'assign_resnums_to_clms' >>\n\n";

	print $fh_log "[ For reference MSA (#{columns} = $ct_clms_ref) ]\n\n";

	print $fh_log join ("\t", 'Clm_no.', @{$sequence_ids}), "\n\n";

	for (my $c=0; $c<$ct_clms_ref; $c++) { print $fh_log join ("\t", $c, @{$set_clminfo_ref->[$c]}), "\n"; }
	print $fh_log "\n\n";

	print $fh_log "[ For reconstructed MSA (#{columns} = $ct_clms_rec) ]\n\n";

	print $fh_log join ("\t", 'Clm_no.', @{$sequence_ids}), "\n\n";

	for (my $c=0; $c<$ct_clms_rec; $c++) { print $fh_log join ("\t", $c, @{$set_clminfo_rec->[$c]}), "\n"; }
	print $fh_log "\n\n\n";

   } # END of "if ($TEST_LEVEL>3) { # (O) ...}"

			## END of "(ADDED on Oct 28 (Tue), 2014.)" ##


=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (Pre 1)

		## Preliminary Process (1): Map the position shifts (from reference to reconstructed) onto the Reconstructed MSA.

	print STDOUT "<<<<< Preliminary (1): Map the position shifts (from reference to reconstructed) onto the Reconstructed MSA... >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (1): Map the position shifts (from reference to reconstructed) onto the Reconstructed MSA... >>>>>\n\n";

	# $map_shifts_respos->[$k][$s] is the shift in the residue position from the reference MSA to the reconstructed one
	# for the $s th site of the $k th sequence, in the reconstructed MSA. It is $gap if occupied by a gap;
	# $shift_lf is the shift at the left-flanking column of the reconstructed MSA. It must be 0 (zero) by default;
	# $shift_rf is the shift at the right-flanking column of the reconstructed MSA. It must be (len(reconstructed) - len(reference)) by default.
	#
	my ($map_shifts_respos, $shift_lf, $shift_rf) = map_shifts_respos_bw_2msas (@ref_seg_msa, @rec_seg_msa, $GAP);
	unless (defined $map_shifts_respos) {
	    print STDOUT "[ERROR '$program_name'] 'map_shifts_respos_bw_2msas' FAILED on the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

   if ($TEST_LEVEL>1) { # (I) 
	print $fh_log "<< Output of 'map_shifts_respos_bw_2msas' >>\n\n";

	print $fh_log '($shift_lf, $shift_rf) = ', "($shift_lf, $shift_rf) .\n\n";

	print $fh_log "[ Shifts in the Reconstructed MSA ]\n\n";

		# Improved the display of a long segment. #

	unless ( pr_map_shifts_respos (*${$fh_log}, @{$sequence_ids}, @{$map_shifts_respos}, $GAP, $DIS_LEN_SHIFT) ) {
	    print STDOUT "[ERROR '$program_name'] 'pr_map_shifts_respos' FAILED on the RAW map of position-shifts between the reference and reconstructed MSAs (onto the latter) !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ; 
	}

   } # END of "if ($TEST_LEVEL>1) { # (I) ...}"


		### NEW PRELIMINARY PROCESS (1hf):
		## PRE-Calculate the commoner flanking shift, and,
		# if the commoner_flanking shift is not zero,
		# nullify it by shifting the entire MSA (either reference or reconstructed).
		## (ADDED as a DRASTIC IMPROVEMENT on Oct 14 (Wed), 2015)

	my $freq_shift_lf = my $freq_shift_rf = 0;

	foreach my $row_shifts (@{$map_shifts_respos}) { # Preliminary outer foreach-loop (over sequences). #
	    foreach my $shift (@{$row_shifts}) { # Preliminary inner foreach-loop (over sites). #
		($shift eq $GAP) and next;
		if ($shift == $shift_lf) { $freq_shift_lf++;
		} elsif ($shift == $shift_rf) { $freq_shift_rf++;
		}
	    }			# End of the preliminary inner foreach-loop (over sites). #
	}		# End of the preliminary outer foreach-loop (over sequences). #

	my $commoner_shift_flank0 = ($freq_shift_lf >= $freq_shift_rf) ? $shift_lf : $shift_rf ;

	if ($commoner_shift_flank0 != 0) {

		# Subtract $commoner_shift_flank0 from all position-shifts in the map. #

	    foreach my $row_shifts (@{$map_shifts_respos}) { # 1st new outer foreach-loop (over sequences). #
		for (my $s=0; $s<$rec_len; $s++) { ($row_shifts->[$s] eq $GAP) or ($row_shifts->[$s] -= $commoner_shift_flank0); }
	    }				# End of the 1st new outer foreach-loop (over sequences). #

			# Prepare auxiliary variables. #

	    my $abs0 = ($commoner_shift_flank0>0)  ? $commoner_shift_flank0 : - $commoner_shift_flank0 ;
	    my $pad = $GAP x $abs0;

	    my @resnums_pad_clm = ();
	    for (1 .. $ct_sequences) { push @resnums_pad_clm, 1/2; }


	    if ($commoner_shift_flank0 < 0) {
		# Pad gap columns on the left-end of the reconstructed MSA.

	(print $fh_log "\n[INFORMATION] The original ", '$commoner_shift_flank', " = $commoner_shift_flank0. Thus, we will shift the entire reconstructed MSA ...\n\n") if ($TEST_LEVEL>1);

		foreach my $seq (@rec_seg_msa) { ${$seq} = $pad . ${$seq}; } 
		##foreach my $seq (@{$rec_msa}) { ${$seq} = $pad . ${$seq}; } # A BUG fixed on Oct 14 (Wed), 2015.

		foreach my $seq (@cp_rec_seg_msa) { ${$seq} = $pad . ${$seq}; }  # ADDED in DEBUGGING of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12). #
		

		for (1 .. $abs0) { # 3rd new outer for-loop (over padded columns). #
		   my @cp = @resnums_pad_clm;
		   unshift @{$set_clminfo_rec}, \@cp;
		}		# End of the 3rd new outer for-loop (over padded columns). #

			# In this case, pad the position-shift map as well.

		my @padded_le_gaps = ();
		for (1 .. $abs0) { push @padded_le_gaps, $GAP; }
		##for (1 .. $abs0) { push @padded_le_gaps; } # A BUG fixed on Oct 14 (Wed), 2015. #
		foreach my $row_shifts (@{$map_shifts_respos}) { # 5th new outer for-loop (over sequences). #
		    @{$row_shifts} = (@padded_le_gaps, @{$row_shifts});
		}			# End of the 5th new outer for-loop (over sequences). #

		$rec_len += $abs0;
		$shift_rf -= $commoner_shift_flank0;

	    } else {
		# Pad gap columns on the left-end of the reference MSA.

	(print $fh_log "\n[INFORMATION] The original ", '$commoner_shift_flank', " = $commoner_shift_flank0. Thus, we will shift the entire reference MSA ...\n\n") if ($TEST_LEVEL>1);

		foreach my $seq (@ref_seg_msa) { ${$seq} = $pad . ${$seq}; }
		##foreach my $seq (@{$ref_msa}) { ${$seq} = $pad . ${$seq}; } # A BUG fixed on Oct 14 (Wed), 2015. 

		for (1 .. $abs0) { # 7th new outer for-loop (over padded columns). #
		    my @cp = @resnums_pad_clm;
		    unshift @{$set_clminfo_ref}, \@cp;
		}		# End of the 7th new outer for-loop (over padded columns). #

		$ref_len += $abs0;
		$shift_rf -= $commoner_shift_flank0;
	    }

   if ($TEST_LEVEL>1) { # (I') 

     if ($TEST_LEVEL < 3) { # (Added in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/09). #

	print $fh_log "[INFORMATION:] The Output of 'map_shifts_respos_bw_2msas' has been REVISED!! ... the result will be recorded in \"$out_main1\" ...\n\n";

     } else {

	print $fh_log "<< REVISED Output of 'map_shifts_respos_bw_2msas' >>\n\n";

	print $fh_log 'New ($shift_lf, $shift_rf) = ', "($shift_lf, $shift_rf) .\n\n";

	print $fh_log "[ New Shifts in the Reconstructed MSA ]\n\n";

	unless ( pr_map_shifts_respos (*${$fh_log}, @{$sequence_ids}, @{$map_shifts_respos}, $GAP, $DIS_LEN_SHIFT) ) {
	    print STDOUT "[ERROR '$program_name'] 'pr_map_shifts_respos' FAILED on the REVISED map of position-shifts between the reference and reconstructed MSAs (onto the latter) !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ; 
	}

     } # End of "if ($TEST_LEVEL < 3) { # (Added in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/09). # ...} else {...}"

   } # END of "if ($TEST_LEVEL>1) { # (I') ...}"


	} # END of "if ($commoner_shift_flank0 < 0) {...}"

=cut # END of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (Pre 1)"

=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (MAIN 1)

		## MAIN Process 1: Output the Position-Shift Map. ##
		## (ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/09.) ##

	print STDOUT "\n\n<<<<< MAIN PROCESS 1:  Outputting the Position-Shift Map (on the reconstructed MSA). >>>>>\n\n\n";
	print $fh_log "\n\n<<<<< MAIN PROCESS 1:  Outputting the Position-Shift Map (on the reconstructed MSA). >>>>>\n\n\n";

   if ($OUTPUT_LEVEL>1) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-D) #

	##my $fh_out_main1 = open_output_file ($out_main1, 1);
	my $fh_out_main1 = open_output_file ($out_main1, 0); # CORRECTED on 2020/03/10.
	unless (defined $fh_out_main1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_main1\" for the $s th segment !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print $fh_log " ... Subroutine, 'open_output_file', Ended on \"$out_main1\" for the $s th segment ...\n\n") if ($TEST_LEVEL>0);

	print $fh_out_main1 '($shift_lf, $shift_rf) = ', "($shift_lf, $shift_rf) .\n\n";

	print $fh_out_main1 "[ Shifts in the Reconstructed MSA ]\n\n";

	unless ( pr_map_shifts_respos (*${$fh_out_main1}, @{$sequence_ids}, @{$map_shifts_respos}, $GAP, $DIS_LEN_SHIFT) ) {
	    print STDOUT "[ERROR '$plscript_name'] The subroutine 'pr_map_shifts_respos' FAILED to print out the Position-Shift Map on the $s th segment: $? !\n";
	    print STDERR "[ERROR '$plscript_name'] $!\n\n";
	    die ;
	}
   ##(print $fh_out_main1 "\n ... The subroutine, 'pr_map_shifts_respos', ENDED to print out the Position-Shift Map on the $s th segment ...\n\n") if ($TEST_LEVEL >0);
   (print $fh_log "\n ... The subroutine, 'pr_map_shifts_respos', ENDED to print out the Position-Shift Map on the $s th segment ...\n\n") if ($TEST_LEVEL >0); # CORRECTED on 2020/03/10.

	close ($fh_out_main1);

   } else {

   (print $fh_log "\n ... Skipped the subroutine, 'pr_map_shifts_respos', because \$OUTPUT_LEVEL = $OUTPUT_LEVEL ...\n\n") if ($TEST_LEVEL >0);

   } # END of "if ($OUTPUT_LEVEL>1) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-D) # ...} else {...}"

		## END of "(ADDED in creation of 'map_aln_shear.ver0.6.pl' on 2020/03/09.)" ##

=cut # END of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (MAIN 1)"


=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (Flt2)

		## The 2nd Filtering: SKIP the Subsequent Analyses IF either $ref_len or $rec_len is TOO LONG. ##
		## (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/29.) (Post MAIN 1)

	unless (($ref_len <= $UPPER_BOUND_ANAL_LEN_SEG) and ($rec_len <= $UPPER_BOUND_ANAL_LEN_SEG)) {
	    print STDOUT "\n[INFORMATION] Either reference or reconstructed sub-MSA (or both) is LONGER THAN the UPPER-BOUND (=$UPPER_BOUND_ANAL_LEN_SEG)! ... I will skip the further analyses on this segment pair... \n\n";
	    print $fh_log "\n[INFORMATION] Either reference or reconstructed sub-MSA (or both) is LONGER THAN the UPPER-BOUND (=$UPPER_BOUND_ANAL_LEN_SEG)! ... I will skip the further analyses on this segment pair... \n\n";

	    $ct_skipped_segs3++;

	    close ($fh_log);

	    foreach my $fh3 ($fh_main_obs, $fh_main_exp) {
		print $fh3 "[ The $s th segment (\$ref_len0 = $ref_len0, \$rec_len0 = $rec_len0; #{maps} = UNKNOWN): ]\n\n";
		print $fh3 "# ... The Analyses were SKIPPED, because either reference or reconstructed sub-MSA (or both) is LONGER THAN the UPPER-BOUND (=$UPPER_BOUND_ANAL_LEN_SEG)...\n";
		print $fh3 "\n\n";
	    }

	    next;

	} # END of "unless (($ref_len <= $UPPER_BOUND_ANAL_LEN_SEG) and ($rec_len <= $UPPER_BOUND_ANAL_LEN_SEG)) {...}"

		## END of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/29.) (Post MAIN 1)"

=cut # END of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (Flt2)"



				## MOVED OUT OF "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (Pre 2 --- 5'')", in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/03). (Pre 2 --- 5'') ##

			# MOVED from the bottom of (Pre 6) in the predecessor, 'map_aln_shear.ver0.6.pl' (on 2020/03/20). #
	##my @br_ids_for_display = (); # MOVED to above the for-loop over the segments, in revision of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). # => The PLACE WAS CORRECTED on 2020/03/27.

   unless (@br_ids_for_display > 0) { # (ADDED in revision of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). # => The PLACE WAS CORRECTED on 2020/03/27. #

	for (my $r=0; $r < $ct_sequences; $r++) {
		my $seqname = $sequence_ids->[$r];
		my $brid = $name2id->{$seqname};
		$br_ids_for_display[$r] = $brid;
	}
	foreach my $br (sort {$a <=> $b} keys %{$br_attr}) {
		unless (defined $id2name->{$br}) { push @br_ids_for_display, $br; }
	}

   } # END of "unless (@br_ids_for_display > 0) { # (ADDED in revision of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). # => The PLACE WAS CORRECTED on 2020/03/27. # ...}"
	
	my @set_all_nodes_for_dsp = (@br_ids_for_display, $topnode_id);

			# END of "(IMPORTED w/ modifications from 'srd_map.ver0.6.pl' (on 2020/03/07).) (Pre 6-2)" #

	my ($lcl_beg_recmsa, $lcl_end_recmsa) = (0, $rec_len -1);
		##my $rec_len = $rec_coord_end - $rec_coord_beg;

				## END of "MOVED OUT OF "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (Pre 2 --- 5'')", in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/03). (Pre 2 --- 5'')" ##


=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (Pre 2 --- 5'')

			### (MODIFIED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/19-) (Pre 2 thru 5'' & M1.5) ###
			### (=> IMPORTED from 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/23.) (Pre 2 thru 5'' & M1.5) ###

		## Preliminary Process (2 --- 5''): Put together the mapped position shifts, to create MINI-classes of shifts & provide branches separating sequences in each class from the rest, & identify 'trivial' MINI-classes. ##
		## (ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (Pre 2 --- 5'') ##

	print STDOUT "<<<<< Preliminary (2 --- 5''): Put together the mapped position shifts, to create MINI-classes of shifts & provide branches separating sequences in each class from the rest, & identify 'trivial' MINI-classes.  >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (2 --- 5''): Put together the mapped position shifts, to create MINI-classes of shifts & provide branches separating sequences in each class from the rest, & identify 'trivial' MINI-classes.  >>>>>\n\n";


			# MOVED from the bottom of (Pre 6) in the predecessor, 'map_aln_shear.ver0.6.pl' (on 2020/03/20). #
	##my @br_ids_for_display = (); # MOVED to above the for-loop over the segments, in revision of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). # => The PLACE WAS CORRECTED on 2020/03/27.

   unless (@br_ids_for_display > 0) { # (ADDED in revision of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). # => The PLACE WAS CORRECTED on 2020/03/27. #

	for (my $r=0; $r < $ct_sequences; $r++) {
		my $seqname = $sequence_ids->[$r];
		my $brid = $name2id->{$seqname};
		$br_ids_for_display[$r] = $brid;
	}
	foreach my $br (sort {$a <=> $b} keys %{$br_attr}) {
		unless (defined $id2name->{$br}) { push @br_ids_for_display, $br; }
	}

   } # END of "unless (@br_ids_for_display > 0) { # (ADDED in revision of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). # => The PLACE WAS CORRECTED on 2020/03/27. # ...}"
	
	my @set_all_nodes_for_dsp = (@br_ids_for_display, $topnode_id);

			# END of "(IMPORTED w/ modifications from 'srd_map.ver0.6.pl' (on 2020/03/07).) (Pre 6-2)" #

	my ($lcl_beg_recmsa, $lcl_end_recmsa) = (0, $rec_len -1);
		##my $rec_len = $rec_coord_end - $rec_coord_beg;

			# END of "MOVED from the bottom of (Pre 6) in the predecessor, 'map_aln_shear.ver0.6.pl' (on 2020/03/20)." #


	my %pspattern2indx;
	my @set_pspatterns = ();
	my @setset_psm_clusterings = ();
	my @set_info_pspatterns = (\%pspattern2indx, \@set_pspatterns, \@setset_psm_clusterings);

	my ($setset_miniclasses_shift_respos, $setset_sep_patterns_for_miniclasses, $commoner_shift_flank,
	   $set_pspattern_blocks, $sets_pcls_assoc_w_psp_blks, $set_miniclass2trivial)
	    = list_minicls_shift_respos_clm_based (@{$map_shifts_respos}, $shift_lf, $shift_rf, $lcl_beg_recmsa, $lcl_end_recmsa,
		$topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$sequence_ids}, $CHAR_ABS, $CHAR_PRES, @set_info_pspatterns) ;
	unless (defined $set_miniclass2trivial) {
	    print STDOUT "[ERROR '$program_name'] 'list_minicls_shift_respos_clm_based' FAILED on the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $ct_psps = @set_pspatterns;
	my $ct_psp_blks = @{$set_pspattern_blocks};
	my $ct_maps_pcls_on_blks = @{$sets_pcls_assoc_w_psp_blks};

#	@{$setset_miniclasses_shift_respos->[$s][$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th MINI-class of the run of shifts of residue positions 
#	under the $s th set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@{$setset_sep_patterns_for_miniclasses->[$s][$mc]} = (\%node2parsstat, \%br2change, $ct_changes, $topnode_mc)
#	is for the pattern separating the sequences affected by the $mc th MINI-class of the run of shifts of residue positions
#	under the $s th set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	%node2parsstat = ($node_id => [$state], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.
#
#	$commoner_shift_flank is whichever is commoner between $shift_lf and $shift_rf (or $shift_lf if they tie).
#
#	@{$set_pspattern_blocks->[$bl]} = ($beg_bl, $end_bl, $len_bl, $indx_psp)
#	 is for the $bl th position-shift-pattern block, where
#	$beg_bl/$end_bl is the starting/ending coordinate of the block (in the 0-based convention),
#	$len_bl is the length (in terms of #{columns}) of the block,
#	and $indx_psp is the index of the position-shift pattern.
#	[ \@set_pspattern_blocks is the main output of 'chop_psmap_into_pspblocks' above.]
#
#	@pcls_assoc_w_psp_blks = @{$sets_pcls_assoc_w_psp_blks->[$s]} is the $s th set of
#	"optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	$pcls_assoc_w_psp_blks[$bl] is the index (in @{$setset_psm_clusterings[$indx_psp]} where $indx_psp = $set_pspattern_blocks->[$bl][3]) of the parsimonious clusterings "optimum" for the $psp th position-shift-pattern block.
#	[ \@sets_pcls_assoc_w_psp_blks is the main output of 'select_pcls_for_psp_blks' above.]
#
#	%{$set_miniclass2trivial->[$s]} = ($miniclass_id => 1, ...) for all "trivial" MINI-classes
#	under the $s th set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks.
#	[\%miniclass2trivial is the main output of 'identify_trivial_miniclasses' above.]

		## END of "(ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (Pre 2 --- 5'')" ##

			# (ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/21.) #

    (print STDOUT "\n#{maps (psp-blocks -> psm-clusterings)} = $ct_maps_pcls_on_blks .\n\n\n") if ($TEST_LEVEL >1);

	unless (defined $ct_maps2freq{$ct_maps_pcls_on_blks}) { $ct_maps2freq{$ct_maps_pcls_on_blks} = 0; }
	$ct_maps2freq{$ct_maps_pcls_on_blks}++;

			# End of "(ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/21.)" #

			## MAIN PROCESS 1.25-A: Outputting Information on the Position-Shift Patterns. ##
			## (ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M 1.2) ##

	print STDOUT "\n\n<<<<< MAIN PROCESS 1.25-A: Outputting Information on the Position-Shift Patterns (#{patterns} = $ct_psps) >>>>>\n\n\n";
	print $fh_log "\n\n<<<<< MAIN PROCESS 1.25-A: Outputting Information on the Position-Shift Patterns (#{patterns} = $ct_psps) >>>>>\n\n\n";

   if ($OUTPUT_LEVEL>=5) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-C-1) #

	# my $oinfo_psps = "$segoutdir/$name_oinfo_psps"; # Added on 2020/03/20.

	my $fh_oi_psp = open_output_file ($oinfo_psps, 0);
	unless (defined $fh_oi_psp) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file \"$oinfo_psps\": $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	print $fh_oi_psp "\n<<< Information on the Position-Shift Patterns (#{patterns} = $ct_psps) >>>\n\n\n";

	for (my $psp =0; $psp <$ct_psps; $psp++)  { # Outer for-loop (over position-shift-patterns). (M1.25-A) #

	    my $cnct_psp = join (',', @{$set_pspatterns[$psp]});

#	@${set_pspatterns[$psp]} = @pspattern is the $psp th position-shift pattern,
#	 where $pspattern[$k] is the position-shift of the $k th sequence,

	    print $fh_oi_psp "\n<< For the $psp th Position-Shift Pattern: '$cnct_psp' >\n\n\n";

	    my $set_psm_cls = $setset_psm_clusterings[$psp];
	    my $ct_pcls = @{$set_psm_cls};

	    print $fh_oi_psp "[[ Contents of ", '@{$setset_psm_clusterings[$psp]}', " (w/ \$psp=$psp,  #{psm-clusterings}=$ct_pcls): ]]\n\n\n";

	    for (my $pcl=0; $pcl<$ct_pcls; $pcl++) { # Middle for-loop (over parsimonious clusterings). #

		print $fh_oi_psp "< For the $pcl th parsimonious clustering: >\n\n";

#	@{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change, \@set_ps_miniclass_seeds) is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where 

		my ($node2parsstat, $br2change, $set_ps_miniclass_seeds) = @{$set_psm_cls->[$pcl]};

		print $fh_oi_psp '%{$node2parsstat} = ( # $node_id => [$char, $sum_penal],', "\n\n";

#	%node2parsstat = ($node_id => \@state_info), where
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#	$sum_penal is the minimum summation of the penalties across the subtree under the node,
#	BUT is undef if $char = $char_abs;

		foreach my $node (sort {$a <=> $b} keys %{$node2parsstat}) {
		    my ($char, $sum_penal) = @{$node2parsstat->{$node}};
		    unless (defined $sum_penal) { $sum_penal = 'undef'; }
		    print $fh_oi_psp "\t", $node, ' => ', "[$char, $sum_penal],\n";
		}
		print $fh_oi_psp "\t\t) .\n\n\n";

		print $fh_oi_psp '%{$br2change} = ( # $br_id => [$stat_ue, $stat_le],', "\n\n";

#	%br2change = ($branch_id => [$stat_ue, $stat_le], ...) for all branches along which the changes occur, where
#	$stat_ue/$stat_le is the state at the upper-end/lower-end of the branch ($branch_id);

		foreach my $br (sort {$a <=> $b} keys %{$br2change}) {
		    my ($stat_ue, $stat_le) = @{$br2change->{$br}};
		    print $fh_oi_psp "\t", $br, ' => ', "[$stat_ue, $stat_le],\n";
		}
		print $fh_oi_psp "\t\t) .\n\n\n";

		my $ct_mcss = @{$set_ps_miniclass_seeds};

		print $fh_oi_psp "[ Contents of ", '@{$$set_ps_miniclass_seeds}', " (#{miniclass-seeds}=$ct_mcss): ]\n\n";

		print $fh_oi_psp join ("\t", 'Seed_no.', 'Shift', '$topnode_mcs', '[indices,of,affected,sequences]', '#{relevant_changes}', '{br_id/state_ue,state_le}_1', '...'), "\n\n";
		for (my $mcs=0; $mcs<$ct_mcss; $mcs++) { # Inner for-loop (over miniclass-seeds). #

		    my ($shift, $topnode_mcs, $seqs_affected, $br2rlv_change) = @{$set_ps_miniclass_seeds->[$mcs]} ;

#	@{$set_ps_miniclass_seeds[$mcs]} = ($shift, $topnode_mcs, \@seqs_affected, \%br2rlv_change)
#	 is for the $mcs th seed of a position-shift miniclass, where
#	$shift is the position-shift characterizing the miniclass seed,
#	$topnode_mcs is the top-node fot the miniclass seed,
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts, and
#	%br2rlv_change = ($branch_id => [$state_upper_end, $state_lower_end], ...) ONLY for the branches that accommodate changes relevant to this miniclass seed.

		    my $cnct_seqs_afd = join (',', @{$seqs_affected});
		    my $ct_rlv_chngs = scalar (keys %{$br2rlv_change});
		    my @out = ($shift, $topnode_mcs, "[$cnct_seqs_afd]", $ct_rlv_chngs);

		    foreach my $br (sort {$a <=> $b} keys %{$br2rlv_change}) { # Innermost foreach-loop (over relevant branches). #
			my ($state_ue, $state_le) = @{$br2rlv_change->{$br}};
			push @out, "$br/$state_ue,$state_le";
		    } # End of the innermost foreach-loop (over relevant branches). #

		    print $fh_oi_psp join ("\t", $mcs, @out), "\n";

		}  # End of the inner for-loop (over miniclass-seeds). #

		print $fh_oi_psp "\n\n\n";

	    }  # End of the Middle for-loop (over parsimonious clusterings). #

	}  # End of the Outer for-loop (over position-shift-patterns). (M1.25-A) #

	close ($fh_oi_psp);

   } else {

	print STDOUT "\n ... Skipped, because \$OUTPUT_LEVEL = $OUTPUT_LEVEL ...\n\n\n";
	print $fh_log "\n ... Skipped, because \$OUTPUT_LEVEL = $OUTPUT_LEVEL ...\n\n\n";

   } # END of "if ($OUTPUT_LEVEL>=5) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-C-1) # ... } else {...}"

			## END of "(ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M 1.25-A)" ##


			## MAIN PROCESS 1.25-B: Outputting Information on the Posithin-Shift-Pattern Blocks. ##
			## (ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M 1.25-B) ##

	print STDOUT "\n\n<<<<< MAIN PROCESS 1.25-B: Outputting Information on the Position-Shift-Pattern Blocks (#{blocks} = $ct_psp_blks) >>>>>\n\n\n";
	print $fh_log "\n\n<<<<< MAIN PROCESS 1.25-B: Outputting Information on the Position-Shift-Pattern Blocks (#{blocks} = $ct_psp_blks) >>>>>\n\n\n";

   if ($OUTPUT_LEVEL>=5) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-C-2) #

	#	# my $olist_psclasses = "$segoutdir/$name_olist_psclasses"; # Added on 2020/03/11.

	# my $olist_psp_blks = "$segoutdir/$name_olist_psp_blks"; # Added on 2020/03/20.

	my $fh_opsp_blks = open_output_file ($olist_psp_blks, 0);
	unless (defined $fh_opsp_blks) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file \"$olist_psp_blks\": $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	print $fh_opsp_blks "\n<< Position-Shift-Pattern Blocks (#{blocks} = $ct_psp_blks) >>\n\n\n";

	print $fh_opsp_blks join ("\t", 'Block_no.', 'beg_bl', 'end_bl', 'len_bl', 'indx_psp', "'position,shift,pattern,...'"), "\n\n";

	for (my $bl = 0; $bl <$ct_psp_blks; $bl++) { # For-loop (over psp-blocks). #

	    my ($beg_bl, $end_bl, $len_bl, $indx_psp) = @{$set_pspattern_blocks->[$bl]} ;

	    my $cnct_psp = join (',', @{$set_pspatterns[$indx_psp]});

	    ##print $fh_opsp_blks join ("\t", $bl, $beg_bl, $end_bl, $len_bl, "'$cnct_psp'"), "\n"; # A minor BUG fixed on 2020/03/22.
	    print $fh_opsp_blks join ("\t", $bl, $beg_bl, $end_bl, $len_bl, $indx_psp, "'$cnct_psp'"), "\n";

#	@{$set_pspattern_blocks->[$bl]} = ($beg_bl, $end_bl, $len_bl, $indx_psp)
#	 is for the $bl th position-shift-pattern block, where
#	$beg_bl/$end_bl is the starting/ending coordinate of the block (in the 0-based convention),
#	$len_bl is the length (in terms of #{columns}) of the block,
#	and $indx_psp is the index of the position-shift pattern.

	} # End of the for-loop (over psp-blocks). #

	print $fh_opsp_blks "\n";

	close ($fh_opsp_blks);

   } else {

	print STDOUT "\n ... Skipped, because \$OUTPUT_LEVEL = $OUTPUT_LEVEL ...\n\n\n";
	print $fh_log "\n ... Skipped, because \$OUTPUT_LEVEL = $OUTPUT_LEVEL ...\n\n\n";

   } # END of "if ($OUTPUT_LEVEL>=5) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-C-2) # ... } else {...}"

			## END of "(ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M 1.25-B)" ##


			## MAIN PROCESS 1.5:  Outputting Information on the Position-Shift (MINI-)Classes. ##
			## (ADDED in REVISION of 'map_aln_shear.ver0.6.pl' on 2020/03/11.) (M1.5) ##
			## (=> Substantially MODIFIED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M1.5) ##

#	@pcls_assoc_w_psp_blks = @{$sets_pcls_assoc_w_psp_blks->[$s]} is the $s th set of
#	"optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	$pcls_assoc_w_psp_blks[$bl] is the index (in @{$setset_psm_clusterings[$indx_psp]} where $indx_psp = $set_pspattern_blocks->[$bl][3]) of the parsimonious clusterings "optimum" for the $psp th position-shift-pattern block.
#	[ \@sets_pcls_assoc_w_psp_blks is the main output of 'select_pcls_for_psp_blks' above.]

	# my $ct_maps_pcls_on_blks = @{$sets_pcls_assoc_w_psp_blks};


	print STDOUT "\n\n<<<<< MAIN PROCESS 1.5:  Outputting Information on Final Position-Shift MINI-Classes (#{maps (psp-blocks -> psm-clusterings)} = $ct_maps_pcls_on_blks) >>>>>\n\n\n";
	print $fh_log "\n\n<<<<< MAIN PROCESS 1.5:  Outputting Information on Final Position-Shift MINI-Classes (#{maps (psp-blocks -> psm-clusterings)} = $ct_maps_pcls_on_blks) >>>>>\n\n\n";

	my $fh_olist; # ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25. (MAIN-B-0) #

   if ($OUTPUT_LEVEL>2) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-B) #

	# my $olist_psclasses = "$segoutdir/$name_olist_psclasses"; # Added on 2020/03/11.

	##my $fh_olist = open_output_file ($olist_psclasses, 0);
	$fh_olist = open_output_file ($olist_psclasses, 0); # Modified in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25. #

	unless (defined $fh_olist) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file \"$olist_psclasses\": $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	print $fh_olist "\n<<<<< There are $ct_maps_pcls_on_blks  maps (psp-blocks -> psm-clusterings) for this segment. >>>>>\n\n\n"; # ADDED for mental ease (on 2020/03/22). (M1.5-HD) #

			# (ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M1.5-OM-BEG) #

    for (my $mp = 0; $mp <$ct_maps_pcls_on_blks; $mp++) { # Outermost for-loop (over the maps (psp-blocks -> psm-clusterings)). (M1.5-OM) #

	my $pcls_assoc_w_psp_blks = $sets_pcls_assoc_w_psp_blks->[$mp];
	##my $cnct_map = join (',', @{$pcls_assoc_w_psp_blks});

	print $fh_olist "<<< For the $mp th set of parsimonious-clusterings associated with the psp-blocks >>>\n\n\n";

	print $fh_olist join ("\t", 'Block_no.', 'Indx_psm_cl'), "\n\n";
	for (my $bl=0; $bl <$ct_psp_blks; $bl++) { print $fh_olist join ("\t", $bl, $pcls_assoc_w_psp_blks->[$bl]), "\n"; }
	print $fh_olist "\n\n";


#	@{$setset_miniclasses_shift_respos->[$s][$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th MINI-class of the run of shifts of residue positions 
#	under the $s th set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@{$setset_sep_patterns_for_miniclasses->[$s][$mc]} = (\%node2parsstat, \%br2change, $ct_changes, $topnode_mc)
#	is for the pattern separating the sequences affected by the $mc th MINI-class of the run of shifts of residue positions
#	under the $s th set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	%node2parsstat = ($node_id => [$state], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.
#
#	%{$set_miniclass2trivial->[$s]} = ($miniclass_id => 1, ...) for all "trivial" MINI-classes
#	under the $s th set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks.

	my $set_miniclasses_shift_respos = $setset_miniclasses_shift_respos->[$mp];
	my $set_sep_patterns_for_miniclasses = $setset_sep_patterns_for_miniclasses->[$mp];
	my %miniclass2trivial = %{$set_miniclass2trivial->[$mp]};

	my $ct_miniclasses_shift = @{$set_miniclasses_shift_respos};

			# END of "(ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M1.5-OM-BEG)" #


	print $fh_olist "<< Final MINI-Classes of Shifts (#{classes} = $ct_miniclasses_shift) >>\n\n";

	print $fh_olist join ("\t", 'Mini-class_no.', 'Shift', 'beg_site', 'end_site', 'beg_rg1,end_rg1;beg_rg2,end_rg2;...', 'IDs,of,affected,seqs'), "\n\n";

	for (my $mc=0; $mc<$ct_miniclasses_shift; $mc++) { # Outer testing for-loop (over mini-classes of shifts). #

	    my ($shift, $beg_site, $end_site, $affected_rgs, $indices_seqs_affected) = @{$set_miniclasses_shift_respos->[$mc]};

	    my $cnct_aff_rgs = '';
	    foreach my $region (@{$affected_rgs}) { $cnct_aff_rgs .= sprintf ("%d,%d;", @{$region}); } 

	    my @seqids_affected = ();
	    foreach my $indx (@{$indices_seqs_affected}) { push @seqids_affected, $sequence_ids->[$indx]; }
	    my $cnct_seqids_affected = join (',', @seqids_affected);

	    print $fh_olist join ("\t", $mc, $shift, $beg_site, $end_site, $cnct_aff_rgs, $cnct_seqids_affected), "\n";
	    
	}				# End of the outer testing for-loop (over mini-classes of shifts). #
	print $fh_olist "\n\n\n";


	print $fh_olist "<< Final content of ", '@{$set_sep_patterns_for_miniclasses}', " >>\n\n";

	for (my $mc=0; $mc<$ct_miniclasses_shift; $mc++) { # Outer testing for-loop (over mini-classes of position shifts). #

	    print $fh_olist "< For the $mc th final MINI-class >\n\n";

	    my $sep_patterns_for_class = $set_sep_patterns_for_miniclasses->[$mc];

##=For the predecessor, 'map_aln_shear.ver0.6.pl'. (M1.5-M-BEG)
##	    unless (@{$sep_patterns_for_class} == 2) {
##		my $ct_sep_patterns = @{$sep_patterns_for_class};
##		print STDOUT "[ERROR '$program_name'] ILLEGAL OUTPUT: The number of separation patterns (=$ct_sep_patterns) is NOT the expected value, 2 !\n\n";
##		print STDERR "[ERROR '$program_name'] $!\n\n";
##		die ;
##	    }
##
##	    for (my $pt=1; $pt<2; $pt++) { # Middle testing for-lop (over separation patterns). #
##	    ##for (my $pt=0; $pt<2; $pt++) {...} # Middle testing for-lop (over separation patterns). # ($pt = 0 is no longer necessary.) #
##
##
##		my ($node2parsstat, $br2change, $ct_changes) = @{$sep_patterns_for_class->[$pt]};
##=cut # End of "For the predecessor, 'map_aln_shear.ver0.6.pl'. (M1.5-M-BEG)"

			# (Added in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M1.5-M-Mid) #

#	@{$setset_sep_patterns_for_miniclasses->[$s][$mc]} = (\%node2parsstat, \%br2change, $ct_changes, $topnode_mc)

		my ($node2parsstat, $br2change, $ct_changes, $topnode_mc) = @{$sep_patterns_for_class};

		print $fh_olist "[ \$topnode_mc = $topnode_mc ]\n\n";

			# End of "(Added in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M1.5-M-Mid)" #

		##my $assigned = $assigned_states[$pt];
		##print $fh_olist "{ The $pt th separation pattern (when the state '$assigned' is assigned to the affected sequences) }\n\n";

		print $fh_olist "[ Separating branches needed (#{branches} = $ct_changes) ]\n\n";

		print $fh_olist join ("\t", 'Branch_ID', 'state_ue', 'state_le'), "\n\n";

		foreach my $br_id (sort {$a <=> $b} keys %{$br2change}) { print $fh_olist join ("\t", $br_id, @{$br2change->{$br_id}}), "\n"; }

		print $fh_olist "\n\n";

##=For the predecessor, 'map_aln_shear.ver0.6.pl'. (M1.5-M-END)
##	    }			# End of the middle testing for-loop (over separation patterns). #
##=cut # End of "For the predecessor, 'map_aln_shear.ver0.6.pl'. (M1.5-M-END)"

	    print $fh_olist "\n";
	}			# End of the outer testing for-loop (over mini-classes of position shifts). #
	

			# (MOVED from (Pre 5'') below, in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M1.5) #

	print $fh_olist "\n\n<< Result of Identifying 'trivial' MINI-classes >>\n\n";

	my @trivial_miniclasses = sort {$a <=> $b} keys %miniclass2trivial;

	print $fh_olist "{'Trivial' MINI-classes} = ", join (', ', @trivial_miniclasses), " .\n\n\n";

			# End of "(MOVED from (Pre 5'') below, in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M1.5)" #

			# (ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M1.5-OM-END) #
    }  # END of the Outermost for-loop (over the maps (psp-blocks -> psm-clusterings)). (M1.5-OM) #
			# End of "(ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M1.5-OM-END)" #

	close ($fh_olist); # MOVED from the bottom of "MAIN PROCESS 1.5:  Outputting Information on Final Position-Shift (MINI-)Classes." ##

   } else {

	print STDOUT "\n ... Skipped, because \$OUTPUT_LEVEL = $OUTPUT_LEVEL ...\n\n\n";
	print $fh_log "\n ... Skipped, because \$OUTPUT_LEVEL = $OUTPUT_LEVEL ...\n\n\n";

   } # END of "if ($OUTPUT_LEVEL > 2) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-B) # ... } else {...}"


			## END of "(ADDED in REVISION of 'map_aln_shear.ver0.6.pl' on 2020/03/11.) (M1.5) ##
			## (=> Substantially MODIFIED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/20.) (M1.5)" ##


##=REPLACED w/ Counterparts in "MyPosShiftMap.pm" & some adapted Test Outputs (on 2020/03/19-). (Pre 2 thru 5'' & M1.5)
##
##		## Preliminary Process (2): Put together the mapped position shifts, to create classes of shifts.
##
##	print STDOUT "<<<<< Preliminary (2): Put together the mapped position shifts into some Classes ... >>>>>\n\n";
##	print $fh_log "<<<<< Preliminary (2): Put together the mapped position shifts into some Classes ... >>>>>\n\n";
##
##	# @{$set_classes_shift_respos->[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
##	# is for the $cl th class of the run of shifts of residue positions, where
##	# $shift is the value of shift,
##	# $beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
##	# @{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
##	# @seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
##	#
##	# @{$set_miniclasses_shift_respos->[$mc]}, which follows the same format as @{$set_classes_shift_respos->[$cl]},
##	# is for the $mc th mini-class of the run of shifts of residue positions.
##	# Each mini-class consists of single runs of the same shift, beginning and ending at the same sites, and affecting one or more sequence(s).
##	#
##		# Further MODIFIED in the middle of creation of 'classify_msa_errors.1.pl' (on Oct 22 (Wed), 2016). #
##	my ($set_classes_shift_respos, $set_miniclasses_shift_respos, $commoner_shift_flank) = br_list_classes_shift_respos (@{$map_shifts_respos}, $shift_lf, $shift_rf, $GAP);
##	##	# MODIFIED in the middle of creation of 'classify_msa_errors.1.pl' (on Oct 16 (Thu), 2016). #
##	##my ($set_classes_shift_respos, $commoner_shift_flank) = br_list_classes_shift_respos (@{$map_shifts_respos}, $shift_lf, $shift_rf, $GAP);
##	####my ($set_classes_shift_respos, $commoner_shift_flank) = list_classes_shift_respos (@{$map_shifts_respos}, $shift_lf, $shift_rf, $GAP);
##	unless (defined $set_classes_shift_respos) {
##	    print STDOUT "[ERROR '$program_name'] 'br_list_classes_shift_respos' FAILED on the $s th segment !\n";
##	    print STDERR "[ERROR '$program_name'] $!\n\n";
##	    die ;
##	}
##
##	my $ct_classes_shift = @{$set_classes_shift_respos};
##	my $ct_miniclasses_shift = @{$set_miniclasses_shift_respos}; # ADDED on Oct 22 (Wed), 2014. #
##
##   if ($TEST_LEVEL>1) { # (IIa)
##
##	print $fh_log "<< Output of 'br_list_classes_shift_respos' >>\n\n";
##	##print $fh_log "<< Output of 'list_classes_shift_respos' >>\n\n";
##
##	print $fh_log '$commoner_shift_flank = ', "$commoner_shift_flank .\n\n\n";
##
##   } # END of "if ($TEST_LEVEL>1) { # (IIa) ...}"
##
##   if ($TEST_LEVEL>1) { # (IIb)
##
##
##			# ADDED on Oct 22 (Wed), 2014. (I) #
##
##	print $fh_log "[ MINI-Classes of Shifts (#{classes} = $ct_miniclasses_shift) ]\n\n";
##
##	print $fh_log join ("\t", 'Mini-class_no.', 'Shift', 'beg_site', 'end_site', 'beg_rg1,end_rg1;beg_rg2,end_rg2;...', 'IDs,of,affected,seqs'), "\n\n";
##
##	for (my $mc=0; $mc<$ct_miniclasses_shift; $mc++) { # Outer testing for-loop (over mini-classes of shifts). #
##
##	    my ($shift, $beg_site, $end_site, $affected_rgs, $indices_seqs_affected) = @{$set_miniclasses_shift_respos->[$mc]};
##
##	    my $cnct_aff_rgs = '';
##	    foreach my $region (@{$affected_rgs}) { $cnct_aff_rgs .= sprintf ("%d,%d;", @{$region}); } 
##
##	    my @seqids_affected = ();
##	    foreach my $indx (@{$indices_seqs_affected}) { push @seqids_affected, $sequence_ids->[$indx]; }
##	    my $cnct_seqids_affected = join (',', @seqids_affected);
##
##	    print $fh_log join ("\t", $mc, $shift, $beg_site, $end_site, $cnct_aff_rgs, $cnct_seqids_affected), "\n";
##	    
##	}				# End of the outer testing for-loop (over mini-classes of shifts). #
##	print $fh_log "\n\n\n";
##
##			# END of "ADDED on Oct 22 (Wed), 2014. (I)" #
##
##   } # END of "if ($TEST_LEVEL>2) { # (IIb) ...}"
##
##
##
##		## Preliminary Process (3'): For each class of MINI-shifts, parsimoniously infer the branches separating the affected sequences from the rest. #
##		# (ADDED on Oct 22 (Wed), 2014. (II))
##
##	print STDOUT "<<<<< Preliminary (3'): For each MINI-class of shifts, parsimoniously infer the branch(es) separating the affected sequences from the rest. >>>>>\n\n";
##	print $fh_log "<<<<< Preliminary (3'): For each MINI-class of shifts, parsimoniously infer the branch(es) separating the affected sequences from the rest. >>>>>\n\n";
##
##	my $set_sep_patterns_for_miniclasses = dpsm_infer_sep_patterns_for_classes (@{$set_miniclasses_shift_respos}, 4, @{$sequence_ids}, %{$name2id}, %{$node2ch}, %{$node2pa}, %{$node2depth}, $CHAR_ABS, $CHAR_PRES);
##
##	unless (defined $set_sep_patterns_for_miniclasses) {
##	    print STDOUT "[ERROR '$program_name'] Subroutine 'dpsm_infer_sep_patterns_for_classes' FAILED on the MINI-classes for the $s th segment !\n";
##	    print STDERR "[ERROR '$program_name'] $!\n\n";
##	    die ;
##	}
##
##   if ($TEST_LEVEL>2) { # (III') 
##
##	print $fh_log "<< Output of 'dpsm_infer_sep_patterns_for_classes', part 2: Content of ", '@{$set_sep_patterns_for_miniclasses}', " >>\n\n";
##
##	for (my $mc=0; $mc<$ct_miniclasses_shift; $mc++) { # Outer testing for-loop (over mini-classes of position shifts). #
##
##	    print $fh_log "< For the $mc th MINI-class >\n\n";
##
##	    my $sep_patterns_for_class = $set_sep_patterns_for_miniclasses->[$mc];
##	    unless (@{$sep_patterns_for_class} == 2) {
##		my $ct_sep_patterns = @{$sep_patterns_for_class};
##		print STDOUT "[ERROR '$program_name'] ILLEGAL OUTPUT: The number of separation patterns (=$ct_sep_patterns) is NOT the expected value, 2 !\n\n";
##		print STDERR "[ERROR '$program_name'] $!\n\n";
##		die ;
##	    }
##
##	    for (my $pt=1; $pt<2; $pt++) { # Middle testing for-lop (over separation patterns). #
##	    ##for (my $pt=0; $pt<2; $pt++) {...} # Middle testing for-lop (over separation patterns). # ($pt = 0 is no longer necessary.) #
##
##		my ($node2parsstat, $br2change, $ct_changes) = @{$sep_patterns_for_class->[$pt]};
##
##		##my $assigned = $assigned_states[$pt];
##		##print $fh_log "{ The $pt th separation pattern (when the state '$assigned' is assigned to the affected sequences) }\n\n";
##
##		print $fh_log "[ Separating branches needed (#{branches} = $ct_changes) ]\n\n";
##
##		print $fh_log join ("\t", 'Branch_ID', 'state_ue', 'state_le'), "\n\n";
##
##		foreach my $br_id (sort {$a <=> $b} keys %{$br2change}) { print $fh_log join ("\t", $br_id, @{$br2change->{$br_id}}), "\n"; }
##
##		print $fh_log "\n\n";
##	    }			# End of the middle testing for-loop (over separation patterns). #
##
##	    print $fh_log "\n";
##	}			# End of the outer testing for-loop (over mini-classes of position shifts). #
##	
##
##   } # END of "if ($TEST_LEVEL>2) { # (III')  ...}"
##
##		# END of "(ADDED on Oct 22 (Wed), 2014. (II))" 
##
##
##		# ADDITIONAL Preliminary Process: Split mini-classes each of which consists of unnaturally remote groups of sequences. #
##		# (ADDED on Oct 30 (Thu), 2014.) #
##
##	print STDOUT "<<<<< ADDITIONAL Preliminary Process (3.5'): Split mini-classes each of which consists of unnaturally remote sequences... >>>>>\n\n";
##	print $fh_log "<<<<< ADDITIONAL Preliminary Process (3.5'): Split mini-classes each of which consists of unnaturally remote sequences... >>>>>\n\n";
##
##	# @{$set_classes_shift_respos->[$cl]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
##	# is for the $cl th class of the run of shifts of residue positions, where
##	# $shift is the value of shift,
##	# $beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
##	# @{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
##	# @seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
##	#
##	# @{$set_miniclasses_shift_respos->[$mc]}, which follows the same format as @{$set_classes_shift_respos->[$cl]},
##	# is for the $mc th mini-class of the run of shifts of residue positions.
##	# Each mini-class consists of single runs of the same shift, beginning and ending at the same sites, and affecting one or more sequence(s).
##	#
##	# @{$set_sep_patterns_for_classes->[$cl][$pt]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $pt (=0 or 1) th pattern separating the sequences affected by the $cl th class of position shifts, where
##	# %node2parsstat = ($node_id => [$state, $ct_pres_ext], ...),
##	# %br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...), and
##	# $ct_changes is the number of "changes" necessary for separating the affected sequences.
##	# ($pt = 0 is when 'absence' is assigned to the affected sequences (i.e., they could be independent), and $pt = 1 is when 'presence' is assigned to them (i.e., they are regarded as sharing the same origin).)
##	#
##	my ($newset_miniclasses_shift_respos, $newset_sep_patterns_for_miniclasses) = split_unnatural_miniclasses (@{$set_miniclasses_shift_respos}, @{$set_sep_patterns_for_miniclasses}, $topnode_id, %{$node2pa}, %{$node2ch}, %{$node2depth}, $CHAR_PRES, $CHAR_ABS, %{$node2seq_indx});
##	unless (defined $newset_miniclasses_shift_respos) {
##	    print STDOUT "[ERROR '$program_name'] Subroutine 'split_unnatural_miniclasses' FAILED for the $s th segment !\n";
##	    print STDERR "[ERROR '$program_name'] $!\n\n";
##	    die ;
##	}
##	my $newct_miniclasses_shift = @{$newset_miniclasses_shift_respos};
##
##
##	if ($newct_miniclasses_shift == $ct_miniclasses_shift) {
##
##	    print $fh_log "... NO CHANGES were made ...\n\n\n";
##
##	} else {
##
##		# Replace the old set of MINI-classes with the new one. #
##
##	    ($set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses) = ($newset_miniclasses_shift_respos, $newset_sep_patterns_for_miniclasses);
##	    my $diffct_miniclasses = $newct_miniclasses_shift - $ct_miniclasses_shift;
##	    $ct_miniclasses_shift = $newct_miniclasses_shift;
##
##	    print $fh_log "... The MINI-classes increased by $diffct_miniclasses compared to the old set!! ...\n\n\n";
##
##	} # END of "if ($newct_miniclasses_shift == $ct_miniclasses_shift) {...} else {...}"
##
##		# END of "ADDED on Oct 30 (Thu), 2014." #
##
##
##			## MAIN PROCESS 1.5:  Outputting Information on Final Position-Shift (MINI-)Classes. ##
##			## [NOTE: It is a modification of the "if ($TEST_LEVEL>2) { # (III'') ...}"
##			## in "if ($newct_miniclasses_shift == $ct_miniclasses_shift) {...} else {...}" above.]
##			## (ADDED in REVISION of 'map_aln_shear.ver0.6.pl' on 2020/03/11.) (M1.5) ##
##
##
##	print STDOUT "\n\n<<<<< MAIN PROCESS 1.5:  Outputting Information on Final Position-Shift (MINI-)Classes (#{classes} = $ct_miniclasses_shift) >>>>>\n\n\n";
##	print $fh_log "\n\n<<<<< MAIN PROCESS 1.5:  Outputting Information on Final Position-Shift (MINI-)Classes (#{classes} = $ct_miniclasses_shift) >>>>>\n\n\n";
##
##	# my $olist_psclasses = "$segoutdir/$name_olist_psclasses"; # Added on 2020/03/11.
##
##	my $fh_olist = open_output_file ($olist_psclasses, 0);
##	unless (defined $fh_olist) {
##	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file \"$olist_psclasses\": $? !\n";
##	    print STDERR "[ERROR '$program_name'] $!\n\n";
##	    die ;
##	}
##
##	print $fh_olist "<< Final MINI-Classes of Shifts (#{classes} = $ct_miniclasses_shift) >>\n\n";
##
##	print $fh_olist join ("\t", 'Mini-class_no.', 'Shift', 'beg_site', 'end_site', 'beg_rg1,end_rg1;beg_rg2,end_rg2;...', 'IDs,of,affected,seqs'), "\n\n";
##
##	for (my $mc=0; $mc<$ct_miniclasses_shift; $mc++) { # Outer testing for-loop (over mini-classes of shifts). #
##
##	    my ($shift, $beg_site, $end_site, $affected_rgs, $indices_seqs_affected) = @{$set_miniclasses_shift_respos->[$mc]};
##
##	    my $cnct_aff_rgs = '';
##	    foreach my $region (@{$affected_rgs}) { $cnct_aff_rgs .= sprintf ("%d,%d;", @{$region}); } 
##
##	    my @seqids_affected = ();
##	    foreach my $indx (@{$indices_seqs_affected}) { push @seqids_affected, $sequence_ids->[$indx]; }
##	    my $cnct_seqids_affected = join (',', @seqids_affected);
##
##	    print $fh_olist join ("\t", $mc, $shift, $beg_site, $end_site, $cnct_aff_rgs, $cnct_seqids_affected), "\n";
##	    
##	}				# End of the outer testing for-loop (over mini-classes of shifts). #
##	print $fh_olist "\n\n\n";
##
##
##	print $fh_olist "<< Final content of ", '@{$set_sep_patterns_for_miniclasses}', " >>\n\n";
##
##	for (my $mc=0; $mc<$ct_miniclasses_shift; $mc++) { # Outer testing for-loop (over mini-classes of position shifts). #
##
##	    print $fh_olist "< For the $mc th final MINI-class >\n\n";
##
##	    my $sep_patterns_for_class = $set_sep_patterns_for_miniclasses->[$mc];
##	    unless (@{$sep_patterns_for_class} == 2) {
##		my $ct_sep_patterns = @{$sep_patterns_for_class};
##		print STDOUT "[ERROR '$program_name'] ILLEGAL OUTPUT: The number of separation patterns (=$ct_sep_patterns) is NOT the expected value, 2 !\n\n";
##		print STDERR "[ERROR '$program_name'] $!\n\n";
##		die ;
##	    }
##
##	    for (my $pt=1; $pt<2; $pt++) { # Middle testing for-lop (over separation patterns). #
##	    ##for (my $pt=0; $pt<2; $pt++) {...} # Middle testing for-lop (over separation patterns). # ($pt = 0 is no longer necessary.) #
##
##		my ($node2parsstat, $br2change, $ct_changes) = @{$sep_patterns_for_class->[$pt]};
##
##		##my $assigned = $assigned_states[$pt];
##		##print $fh_olist "{ The $pt th separation pattern (when the state '$assigned' is assigned to the affected sequences) }\n\n";
##
##		print $fh_olist "[ Separating branches needed (#{branches} = $ct_changes) ]\n\n";
##
##		print $fh_olist join ("\t", 'Branch_ID', 'state_ue', 'state_le'), "\n\n";
##
##		foreach my $br_id (sort {$a <=> $b} keys %{$br2change}) { print $fh_olist join ("\t", $br_id, @{$br2change->{$br_id}}), "\n"; }
##
##		print $fh_olist "\n\n";
##	    }			# End of the middle testing for-loop (over separation patterns). #
##
##	    print $fh_olist "\n";
##	}			# End of the outer testing for-loop (over mini-classes of position shifts). #
##	
##	##close ($fh_olist); # MOVED to below "Preliminary Process (5''): Identify 'trivial' MINI-classes."
##
##			## END of "(ADDED in REVISION of 'map_aln_shear.ver0.6.pl' on 2020/03/11.) (M1.5)" ##
##
##
##		## Preliminary Process (4'): Merge the MINI-classes of shifts. #
##		## (ADDED on Oct 22 (Wed)- 23 (Thu), 2014.)
##
##	print STDOUT "<<<<< Preliminary (4'): Merge the MINI-classes of shifts. >>>>>\n\n";
##	print $fh_log "<<<<< Preliminary (4): Merge the MINI-classes of shifts. >>>>>\n\n";
##
##	# @{$set_miniblocks_shift_respos->[$mb]}, which follows the same format as @{$set_blocks_shift_respos->[$bl]},
##	# is for the $mb th block created by merging the mini-classes of the run of shifts of residue positions.
##	#
##	my $set_miniblocks_shift_respos = merge_miniclasses_shift_respos (@{$set_miniclasses_shift_respos}, @{$set_sep_patterns_for_miniclasses}, @{$sequence_ids}, %{$name2id}, %{$id2name}, %{$node2ch}, %{$node2pa}, %{$node2depth}, $CHAR_ABS, $CHAR_PRES);
##	unless (defined $set_miniblocks_shift_respos) {
##	    print STDOUT "[ERROR '$program_name'] Subroutine 'merge_miniclasses_shift_respos' FAILED for the $s th segment !\n";
##	    print STDERR "[ERROR '$program_name'] $!\n\n";
##	    die ;
##	}
##
##	my $ct_miniblocks_srp = @{$set_miniblocks_shift_respos};
##
##   if ($TEST_LEVEL>1) { # (IV') 
##
##	print $fh_log "<< Output of 'merge_miniclasses_shift_respos': Content of ", '@{$set_miniblocks_shift_respos}', " (#{blocks} = $ct_miniblocks_srp) >>\n\n";
##
##	for (my $bl = 0; $bl<$ct_miniblocks_srp; $bl++) { # Outer testing for-loop (over mini-blocks). #
##
##	    my ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, $seqs_involved, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_miniblocks_shift_respos->[$bl]};
##
##	    print $fh_log "< The $bl th MINI-block (Shift = $shift; Region = \[$beg_blk, $end_blk\]; Total#{changes} = $ct_changes) >\n\n";
##	    print $fh_log "{Indices of the source classes} = ", join (', ', @{$indices_src_classes}), " .\n\n";
##
##	    my @seqids_invlv = ();
##	    foreach my $indx (@{$seqs_involved}) { push @seqids_invlv, $sequence_ids->[$indx]; }
##
##	    print $fh_log "{MRCA of involved sequences} = $id_node_mrca (depth=$depth_mrca) .\n\n";
##	    print $fh_log "{Involved sequences} = ", join (', ', @seqids_invlv), " .\n\n";
##
##	    my $ct_br = scalar keys %{$br2change};
##	    print $fh_log "{ There are $ct_br branches separating the involved sequences from the rest: }\n\n";
##	    print $fh_log join ("\t", 'Branch_ID', 'state_ue', 'state_le'), "\n\n";
##
##	    foreach my $br (sort { $a <=> $b } keys %{$br2change}) { print $fh_log join ("\t", $br, @{$br2change->{$br}}), "\n"; }
##	    ##foreach my $br (sort {$node2depth->{$a} <=> $node2depth->{$b}} keys %{$br2change}) { print $fh_log join ("\t", $br, @{$br2change->{$br}}), "\n"; }
##	    print $fh_log "\n\n";
##
##	    my $ct_excl = @{$excl_sbblocks};
##	    print $fh_log "{ There are $ct_excl Excluded sub-blocks: }\n\n";
##	    print $fh_log join ("\t", 'No.', 'beg_sbblk', 'end_sbblk', 'nodeid_mrca'), "\n\n";
##
##	    for (my $sb=0; $sb<$ct_excl; $sb++) { print $fh_log join ("\t", $sb, @{$excl_sbblocks->[$sb]}), "\n"; }
##	    print $fh_log "\n\n\n";
##
##	}			# End of the outer testing for-loop (over mini-blocks). #
##
##   } # END of "if ($TEST_LEVEL>1) { # (IV')  ...}"
##
##
##
##		## Preliminary Process (5'): Identify 'trivial' MINI-blocks. ##
##
##	print STDOUT "<<<<< Preliminary (5'): Identify 'trivial' MINI-blocks. >>>>>\n\n";
##	print $fh_log "<<<<< Preliminary (5'): Identify 'trivial' MINI-blocks. >>>>>\n\n";
##
##	# %{$block2trivial} = ($indx_block => 1, ...) for the blocks judged as 'trivial'.
##	#
##	my $miniblock2trivial = find_trivial_blocks (@{$set_miniblocks_shift_respos}, @{$sequence_ids}, $commoner_shift_flank, $shift_lf, $shift_rf, @{$map_shifts_respos}, $GAP);
##	unless (defined $miniblock2trivial) {
##	    print STDOUT "[ERROR '$program_name'] Subroutine 'find_trivial_blocks' FAILED on the MINI-blocks for the $s th segment !\n";
##	    print STDERR "[ERROR '$program_name'] $!\n\n";
##	    die ;
##	}
##
##   if ($TEST_LEVEL>1) { # (V')
##
##	print $fh_log "<< Output of 'find_trivial_blocks' on the MINI-blocks >>\n\n";
##
##	my @trivial_miniblocks = sort {$a <=> $b} keys %{$miniblock2trivial};
##
##	print $fh_log "{'Trivial' MINI-blocks} = ", join (', ', @trivial_miniblocks), " .\n\n\n";
##
##   } # END of "if ($TEST_LEVEL>1) { # (V') ...}"
##
##
##		## END of "(ADDED on Oct 22 (Wed) - 23 (Thu), 2014.)" #
##
##
##		## Preliminary Process (5''): Identify 'trivial' MINI-classes. ##
##		## (ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/08).) (Pre 5'') ##
##
##	print STDOUT "<<<<< Preliminary (5''): Identify 'trivial' MINI-classes. >>>>>\n\n";
##	print $fh_log "<<<<< Preliminary (5''): Identify 'trivial' MINI-classes. >>>>>\n\n";
##
##	my %miniclass2trivial; # = ($indx_mclass => 1, ...) for the mini-classes belonging to the 'trivial' mini-blocks.
##
##	foreach my $tr_mbl (keys %{$miniblock2trivial}) { # Foreach-loop for "Pre 5''"
##	    my ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, $seqs_involved, $br2change, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_miniblocks_shift_respos->[$tr_mbl]};
##	    foreach my $tr_mcl (@{$indices_src_classes}) { $miniclass2trivial{$tr_mcl} = 1; }
##	} # End of the foreach-loop for "Pre 5''". #
##
##			# (MODIFIED in REVISION of 'map_aln_shear.ver0.6.pl' on 2020/03/11.) (Pre 5'') #
##
##	print $fh_olist "\n\n<< Result of Identifying 'trivial' MINI-classes >>\n\n";
##
##	my @trivial_miniclasses = sort {$a <=> $b} keys %miniclass2trivial;
##
##	print $fh_olist "{'Trivial' MINI-classes} = ", join (', ', @trivial_miniclasses), " .\n\n\n";
##
##	close ($fh_olist); # MOVED from the bottom of "MAIN PROCESS 1.5:  Outputting Information on Final Position-Shift (MINI-)Classes." ##
##
##			# End of "(MODIFIED in REVISION of 'map_aln_shear.ver0.6.pl' on 2020/03/11.) (Pre 5'')" #
##
##		## END of "(ADDED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/08).) (Pre 5'')" ##
##
##
##=cut # END of "REPLACED w/ Counterparts in "MyPosShiftMap.pm" & some adapted Test Outputs (on 2020/03/19-). (Pre 2 thru 5'' & M1.5)"

			### END of "(MODIFIED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/19-) (Pre 2 thru 5'' & M1.5) ###
			### (=> IMPORTED from 'map_alnshear_and_srd.ver0.6.pl' on 2020/03/23.) (Pre 2 thru 5'' & M1.5) " ###

=cut # END of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (Pre 2 --- 5'')"



		### MODIFIED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/07). (Pre 6) ###

		## Preliminary Process (6): Identify gap-pattern blocks, calculate their Dollo parsimony scenarios, and the initial parsimony candidate scenario of each gapped segment in the segmental MSAs (reconstructed only). ##

	print STDOUT "<<<<< Preliminary (6): Identify gap-pattern blocks, calculate their Dollo parsimony scenarios, and the initial parsimony candidate scenario of each gapped segment in the segmental MSAs (reconstructed only). >>>>>\n\n";
	print $fh_log "<<<<< Preliminary (6): Identify gap-pattern blocks, calculate their Dollo parsimony scenarios, and the initial parsimony candidate scenario of each gapped segment in the segmental MSAs (reconstructed only). >>>>>\n\n";

			# (IMPORTED w/ modifications from 'srd_map.ver0.6.pl' (on 2020/03/07).) (Pre 6) #

	my ($set_gpatterns, $set_node2dparsstat, $set_br2change, $info_psm_indel_hsts0)
	  = enum_psm_indel_hsts_for_msa_finer_1ST_HF (%{$node2ch}, %{$node2pa}, %{$name2id}, @{$sequence_ids}, @rec_seg_msa, $CHAR_ABS, $CHAR_PRES, %{$node2depth}, %{$node2ct_exoffs});

	##unless ((defined $info_psm_indel_hsts0) and (@{$info_psm_indel_hsts0}>2) ) {
	unless (defined $info_psm_indel_hsts0) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'enum_psm_indel_hsts_for_msa_finer_1ST_HF' FAILED for the $s th segment !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n\n";
		die ;
	}
   (print $fh_log " ... Subroutine, 'enum_psm_indel_hsts_for_msa_finer_1ST_HF', Ended for the $s th segment ...\n\n") if ($TEST_LEVEL>0);

	unless (@{$info_psm_indel_hsts0}>2) {

	    print STDOUT "[INFORMATION] SKIPPED the Calculation for the $s th segment, because the reconstructed MSA contains (a) gap(s) or (a) gapped segment(s) that is (are) too long ...\n\n\n";

	    print $fh_log "SKIPPED the Calculation, because either reference or reconstructed MSA contains (a) gap(s) or (a) gapped segment(s) that is (are) too long ...\n\n\n";

	    close ($fh_log);

	    next;
	}
##=For the predecessor, 'srd_map.ver0.6.pl'.
##	unless (@{$info_gpatterns}>0) {
##
##	    print STDOUT "[INFORMATION] SKIPPED the Calculation for the $s th segment, because either reference or reconstructed MSA contains (a) gap(s) or (a) gapped segment(s) that is (are) too long ...\n\n\n";
##
##	    print $fh_log "SKIPPED the Calculation, because either reference or reconstructed MSA contains (a) gap(s) or (a) gapped segment(s) that is (are) too long ...\n\n\n";
##	    close ($fh_log);
##
##	    next;
##	}
##=cut # End of "For the predecessor, 'srd_map.ver0.6.pl'."


	my ($set_cls_gpblocks, $totcts0_indels_for_cls, $psm_hsts0_for_cls) = @{$info_psm_indel_hsts0};
	my $ct_cls = @{$set_cls_gpblocks};
	my $ct_gpd_segments = $ct_cls; # Added to suppress error messages (on 2019/06/18). #

   if ($TEST_LEVEL>2) { # (Pre 6) #

	print $fh_log "\n<< Main Contents of the Preliminary output (6) ", '@{$set_cls_gpblocks} & @{$psm_hsts0_for_cls}', " >>\n\n";

	for (my $cl = 0; $cl < $ct_cls; $cl++) { # The 6 th preliminary  for-loop (over gapped segments). #
	    my $cls_gpblks = $set_cls_gpblocks->[$cl];
	    my $ct_gpblks = @{$cls_gpblks};

	    print $fh_log ' ' x 2, "{ The $cl th gapped segment (#{gap-pattern blocks} = $ct_gpblks) }\n\n";
	    print $fh_log ' ' x 2, join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal)'"), "\n\n";
	    for (my $bl=0; $bl < $ct_gpblks; $bl++) { # 1st inner for-loop (over blocks). #
	        my ($indx_gp, $len, $beg, $end) = @{$cls_gpblks->[$bl]};
	        my $gpattern = ${$set_gpatterns->[$indx_gp]};
	        print $fh_log ' ' x 2, join ("\t", $bl, $len, $beg, $end, $indx_gp, "'$gpattern'"), "\n";
	    }				# End of the 1st inner for-loop (over blocks). #
	    print $fh_log "\n\n";

	    my ($br2indels0, $node2bw_gstats0) = @{$psm_hsts0_for_cls->[$cl]};

	    print $fh_log ' ' x 2, "( Map of INITIAL block-wise gap states onto the nodes )\n\n";
	    print $fh_log ' ' x 2, join ("\t", 'Node_ID', "'block_wise_gap_states0(concatenated)'"), "\n\n";

	    foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats0}) { # 2nd inner foreach-loop (over nodes). #
	        my $cnct_bw_gstats0 = join ('', @{$node2bw_gstats0->{$node}});
	        print $fh_log ' ' x 2, join ("\t", $node, "'$cnct_bw_gstats0'"), "\n";
	    }                           # End of the 2nd inner foreach-loop (over nodes). #
	    print $fh_log "\n\n";

	}		# End of the 6th preliminary for-loop (over gapped segments). #

   } # End of "if ($TEST_LEVEL>2) { # (Pre 6) # ...}"

			# End of "(IMPORTED w/ modifications from 'srd_map.ver0.6.pl' (on 2020/03/07).) (Pre 6)" #


##=MOVED to the top of (Pre 2 --- 5'') JUST AS in creation of 'map_aln_shear_clm_based.ver0.6.pl' (on 2020/03/20), in creation of 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.
##
##			# (IMPORTED w/ modifications from 'srd_map.ver0.6.pl' (on 2020/03/07).) (Pre 6-2) #
##
##	##my @br_ids_for_display = (); # MOVED to above the for-loop over the segments, in revision of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). #
##
##   unless (@br_ids_for_display > 0) { # (ADDED in revision of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). # THIS WAS DONE in a WRONG PLACE!!
##
##	for (my $r=0; $r < $ct_sequences; $r++) {
##		my $seqname = $sequence_ids->[$r];
##		my $brid = $name2id->{$seqname};
##		$br_ids_for_display[$r] = $brid;
##	}
##	foreach my $br (sort {$a <=> $b} keys %{$br_attr}) {
##		unless (defined $id2name->{$br}) { push @br_ids_for_display, $br; }
##	}
##   } # End of "unless (@br_ids_for_display > 0) { # (ADDED in revision of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). # ...}"
##
##	my @set_all_nodes_for_dsp = (@br_ids_for_display, $topnode_id);
##
##			# END of "(IMPORTED w/ modifications from 'srd_map.ver0.6.pl' (on 2020/03/07).) (Pre 6-2)" #
##
##	my ($lcl_beg_recmsa, $lcl_end_recmsa) = (0, $rec_len -1);
##		##my $rec_len = $rec_coord_end - $rec_coord_beg;
##=cut # END of "MOVED to the top of (Pre 2 --- 5'') JUST AS in creation of 'map_aln_shear_clm_based.ver0.6.pl' (on 2020/03/20), in creation of 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23."

		### END of "MODIFIED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/07). (Pre 6)" ###


=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (MAIN 2)

		### MODIFIED in creation of 'map_aln_shear_clm_based.ver0.6.pl' (on 2020/03/20). (M2) ###
		### => IMPORTED to 'map_alnshear_cb_and_srd.ver0.6.pl' (on 2020/03/23). (M2) ###

	print STDOUT "\n\n<<<<< MAIN PROCESS 2:  Creating & Outputting the Alignment-Shear Maps on the reconstructed MSA, one for each map (psp-blocks -> psm-clusterings). >>>>>\n\n\n";
	print $fh_log "\n\n<<<<< MAIN PROCESS 2:  Creating & Outputting the Alignment-Shear Maps on the reconstructed MSA, one for each map (psp-blocks -> psm-clusterings). >>>>>\n\n\n";

	my $fh_out_main2; # ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25. (MAIN-E-1-0) #

   if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-1-1) #

	##my $fh_out_main2 = open_output_file ($out_main2, 0); # A BUG fixed on 2020/03/25. #
	$fh_out_main2 = open_output_file ($out_main2, 0);
	unless (defined $fh_out_main2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_main2\" for the $s th segment !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print $fh_log " ... Subroutine, 'open_output_file', Ended on \"$out_main2\" for the $s th segment ...\n\n") if ($TEST_LEVEL>0);

   } # End of "if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-1-1) # ...}"

	my @set_maps_alnshear = (); # element = $map_alnshear # ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/24). (M2-1) #


   if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-1-2) #
	print $fh_out_main2 "\n<<<<< There are $ct_maps_pcls_on_blks  maps (psp-blocks -> psm-clusterings) for this segment. >>>>>\n\n\n"; # ADDED for mental ease (on 2020/03/22). (M2-HD) #

   } # End of "if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-1-2) # ...}"

    for (my $mp = 0; $mp <$ct_maps_pcls_on_blks; $mp++) { # Outermost for-loop (over the maps (psp-blocks -> psm-clusterings)). (M2-OM) #

	my $pcls_assoc_w_psp_blks = $sets_pcls_assoc_w_psp_blks->[$mp];
	my $cnct_map = join (', ', @{$pcls_assoc_w_psp_blks});

   if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-1-3) #
	print $fh_out_main2 "<<< For the $mp th set of parsimonious-clusterings associated with the psp-blocks  >>>\n\n\n";

	print $fh_out_main2 "Map (psp-block = 0, ..., $ct_psp_blks -1)  = ($cnct_map) .\n\n\n";

   } # End of "if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-1-3) # ...}"

	my $set_miniclasses_shift_respos = $setset_miniclasses_shift_respos->[$mp];
	my $set_sep_patterns_for_miniclasses = $setset_sep_patterns_for_miniclasses->[$mp];
	my %miniclass2trivial = %{$set_miniclass2trivial->[$mp]};

	my $ct_miniclasses_shift = @{$set_miniclasses_shift_respos};

	my ($map_alnshear, $br2linmap_alnshear)
	  ## =  mk_map_alnshear (@br_ids_for_display, $CHAR_ABS, $CHAR_PRES, $CHAR_NOSHEAR, $topnode_id, %{$node2pa}, ## A BUG Fixed on 2020/03/21. 
	  =  mk_map_alnshear_clmbased (@br_ids_for_display, $CHAR_ABS, $CHAR_PRES, $CHAR_NOSHEAR, $topnode_id, %{$node2pa},  # MAJOR MODIFICATION, in creation of 'map_aln_shear_clm_based.ver0.6.pl' (on 2020/03/21). #
		@{$set_miniclasses_shift_respos}, @{$set_sep_patterns_for_miniclasses}, %miniclass2trivial,
		@{$info_psm_indel_hsts0}, $lcl_beg_recmsa, $lcl_end_recmsa, %{$br_attr}, %{$br2eq});

	## my ($lcl_beg_recmsa, $lcl_end_recmsa) = (0, $rec_len -1);

	unless (defined $map_alnshear) {
	    print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_map_alnshear_clmbased' FAILED on the $mp th map in the $s th segment: $? !\n";
	    print STDERR "[ERROR '$plscript_name'] $!\n\n";
	    die ;
	}
   (print $fh_log "\n ... The subroutine, 'mk_map_alnshear_clmbased', ENDED on the $mp th map in the $s th segment ...\n\n") if ($TEST_LEVEL >0);


	$set_maps_alnshear[$mp] = $map_alnshear; # ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/24). (M2-2)


   if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-1-3+1) #
	unless ( print_map_alnshear (*${$fh_out_main2}, @br_ids_for_display, %{$id2name}, @{$map_alnshear}, $lcl_beg_recmsa, $lcl_end_recmsa, $DIS_LEN_ASMAP, $MIN_WDTH_FLD_ASMAP) ) {
	    print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_map_alnshear' FAILED to print out the Alignment-Shear Map in the $s th segment, under the $mp th map (psp-blocks -> psm-clusterings) : $? !\n";
	    print STDERR "[ERROR '$plscript_name'] $!\n\n";
	    die ;
	}
   (print $fh_log "\n ... The subroutine, 'print_map_alnshear', ENDED to print out the Alignment-Shear Map for the $s th segment, under the $mp th map ...\n\n") if ($TEST_LEVEL >0);

   } else {

   (print $fh_log "\n ... SKIPPED the subroutine, 'print_map_alnshear', because \$OUTPUT_LEVEL = $OUTPUT_LEVEL ... \n\n") if ($TEST_LEVEL >0);

   } # End of "if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-1-3+1) # ...} else {...}"


    }   # END of the Outermost for-loop (over the maps (psp-blocks -> psm-clusterings)). (M2-OM) #

   if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-1-5) #
	close ($fh_out_main2);

   } # End of "if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-1-5) # ...}"

##=For the predecessor, 'map_aln_shear.ver0.6.pl'. (M2)
##
##	print STDOUT "\n\n<<<<< MAIN PROCESS 2:  Creating & Outputting the Alignment-Shear Map on the reconstructed MSA. >>>>>\n\n\n";
##	print $fh_log "\n\n<<<<< MAIN PROCESS 2:  Creating & Outputting the Alignment-Shear Map on the reconstructed MSA. >>>>>\n\n\n";
##
##	my ($map_alnshear, $br2linmap_alnshear)
##	  =  mk_map_alnshear (@br_ids_for_display, $CHAR_ABS, $CHAR_PRES, $CHAR_NOSHEAR, $topnode_id, %{$node2pa}, 
##		@{$set_miniclasses_shift_respos}, @{$set_sep_patterns_for_miniclasses}, %miniclass2trivial,
##		@{$info_psm_indel_hsts0}, $lcl_beg_recmsa, $lcl_end_recmsa, %{$br_attr}, %{$br2eq});
##
##	## my ($lcl_beg_recmsa, $lcl_end_recmsa) = (0, $rec_len -1);
##
##	unless (defined $map_alnshear) {
##	    print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_map_alnshear' FAILED on the $s th segment: $? !\n";
##	    print STDERR "[ERROR '$plscript_name'] $!\n\n";
##	    die ;
##	}
##   (print $fh_log "\n ... The subroutine, 'mk_map_alnshear', ENDED on the $s th segment ...\n\n") if ($TEST_LEVEL >0);
##
##
##	##my $fh_out_main2 = open_output_file ($out_main2, 1);
##	my $fh_out_main2 = open_output_file ($out_main2, 0); # CORRECTED on 2020/03/10.
##	unless (defined $fh_out_main2) {
##	    print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_main2\" for the $s th segment !\n";
##	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
##	    die ;
##	}
##   (print $fh_log " ... Subroutine, 'open_output_file', Ended on \"$out_main2\" for the $s th segment ...\n\n") if ($TEST_LEVEL>0);
##
##
##	unless ( print_map_alnshear (*${$fh_out_main2}, @br_ids_for_display, %{$id2name}, @{$map_alnshear}, $lcl_beg_recmsa, $lcl_end_recmsa, $DIS_LEN_ASMAP, $MIN_WDTH_FLD_ASMAP) ) {
##	    print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_map_alnshear' FAILED to print out the Alignment-Shear Map in the $s th segment: $? !\n";
##	    print STDERR "[ERROR '$plscript_name'] $!\n\n";
##	    die ;
##	}
##   (print $fh_log "\n ... The subroutine, 'print_map_alnshear', ENDED to print out the Alignment-Shear Map for the $s th segment ...\n\n") if ($TEST_LEVEL >0);
##
##	close ($fh_out_main2);
##
##=cut # END of "For the predecessor, 'map_aln_shear.ver0.6.pl'. (M2)"

		### END of "MODIFIED in creation of 'map_aln_shear_clm_based.ver0.6.pl' (on 2020/03/20). (M2) ###
		### => IMPORTED to 'map_alnshear_cb_and_srd.ver0.6.pl' (on 2020/03/23). (M2)" ###

=cut # END of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (MAIN 2)"


##=Set Aside for Potential Use. (Pre 7')
##
##			## ADDED on Oct 23 (Thu), 2014. ##
##
##		## Preliminary Process (7'): Lump together some neighboring MINI-blocks of the same type to form "composite MINI-blocks". ##
##
##	print STDOUT "<<<<< Preliminary (7'): Lump together some neighboring MINI-blocks affecting the identical set of sequences. >>>>>\n\n";
##	print $fh_log "<<<<< Preliminary (7'): Lump together some neighboring MINI-blocks affecting the identical set of sequences. >>>>>\n\n";
##
##	# @{$composite_blocks->[$cb]} = ($beg_site, $end_site, $nodeid_mrca, \%br2change, \@constituents, \@involved_seqs, \@merger_types), where
##	# @constituents lists the indices of blocks that compose the $cb th composite block (including "composite blocks" each of which consists of a single block (except trivial blocks)).
##	# @involved_seqs lists the indices (in @{$sequence_ids}) of the sequences involved,
##	# $merger_types[$i] is the type of merger between the $i th and ($i+1) th constituent blocks
##	# (= 0 for 'gap-mediated', = 1/2 for 'non-gap-mediated' with  the $i/($i+1) th block incorporated).
##	#
##			# (MODIFIED on Oct 20 (Mon) - 21 (Tue), 2014.) #
##	my $composite_miniblocks = lump_together_similar_blocks (@{$set_miniblocks_shift_respos}, @{$sequence_ids}, %{$miniblock2trivial}, @{$map_shifts_respos}, $GAP, @{$set_cls_gblocks_ref}, @{$set_cls_gblocks_rec}, @{$set_br2change}, %{$node2pa}, %{$node2ch}, $topnode_id);
##	##my $composite_blocks = lump_together_similar_blocks_proto (@{$set_blocks_shift_respos}, @{$sequence_ids}, %{$block2trivial}, @{$map_shifts_respos}, $GAP);
##	unless (defined $composite_miniblocks) {
##	    print STDOUT "[ERROR '$program_name'] Subroutine 'lump_together_similar_blocks' FAILED for the $s th segment !\n";
##	    print STDERR "[ERROR '$program_name'] $!\n\n";
##	    die ;
##	}
##
##	my $ct_cmp_miniblks = @{$composite_miniblocks};
##
##   if ($TEST_LEVEL>1) { # (VII)
##
##	print $fh_log "<< Output of 'lump_together_similar_blocks': Content of ", '@{$composite_miniblocks}', " (#{composite_miniblocks} = $ct_cmp_miniblks) >>\n\n";
##
##	print $fh_log join ("\t", 'Indx_cmp_miniblock', 'beg_cmb', 'end_cmb', 'mrca', 'indices,constituent,miniblocks', 'list,position,shifts', 'merger,types', 'indices,involved,seqs'), "\n\n"; # MODIFIED on Oct 21 (Tue), 2014. #
##	##print $fh_log join ("\t", 'Indx_cmp_block', 'beg_cb', 'end_cb', 'mrca', 'indices,constituent,blocks', 'merger,types', 'indices,involved,seqs'), "\n\n";
##
##	for (my $cb=0; $cb<$ct_cmp_miniblks; $cb++) { # Testing for-loop (over composite mini-blocks). #
##
##	    my ($beg_cb, $end_cb, $mrca, $br2change, $constituents, $involved_seqs, $merger_types) = @{$composite_miniblocks->[$cb]};
##	    my $cnct_constituents = join (',', @{$constituents});
##	    my $cnct_involved_seqs = join (',', @{$involved_seqs}); 
##	    my $cnct_merger_types = (@{$merger_types}>0) ? join (',', @{$merger_types}) : 'n/a' ;
##
##		# MODIFIED on Oct 21 (Tue), 2014. #
##
##	    my @pos_shifts = ();
##				# A BUG fixed on Oct 23 (Thu), 2014. #
##	    foreach my $indx_blk (@{$constituents}) { push @pos_shifts, $set_miniblocks_shift_respos->[$indx_blk][0]; }
##	    ##foreach my $indx_blk (@{$constituents}) { push @pos_shifts, $set_blocks_shift_respos->[$indx_blk][0]; }
##	    my $cnct_pos_shifts = join (',', @pos_shifts);
##
##	    print $fh_log join ("\t", $cb, $beg_cb, $end_cb, $mrca, $cnct_constituents, $cnct_pos_shifts, $cnct_merger_types, $cnct_involved_seqs), "\n";
##	    ##print $fh_log join ("\t", $cb, $beg_cb, $end_cb, $mrca, $cnct_constituents, $cnct_merger_types, $cnct_involved_seqs), "\n";
##	    ####print $fh_log join ("\t", $cb, $beg_cb, $end_cb, $mrca, $br2change, $cnct_constituents, $cnct_merger_types, $cnct_involved_seqs), "\n"; # A BUG fixed. ($br2change was superfluous!!)
##
##	}			# End of the testing for-loop (over composite mini-blocks). #
##	print $fh_log "\n\n\n";
##
##   } # END of "if ($TEST_LEVEL>1) { # (VII) ...}"
##
##=cut # END of "Set Aside for Potential Use. (Pre 7')"


##=Set Aside for Potential Use. (Pre 8)
##
##				## ADDED on Oct 24 (Fri), 2014). ##
##
##		## Preliminary Process (8): Reorganize the list of insertions/deletions in the initial candidate of parsimonious scenarios, for reference and reconstructed MSAs.
##
##	print STDOUT "<<<<< Preliminary (8): Reorganize the list of insertions/deletions in the initial candidate of parsimonious scenarios, for reference and reconstructed MSAs. >>>>>\n\n";
##	print $fh_log "<<<<< Preliminary (8): Reorganize the list of insertions/deletions in the initial candidate of parsimonious scenarios, for reference and reconstructed MSAs. >>>>>\n\n";
##
##
##	print $fh_log "<<< (1) For Reference MSA >>>\n\n";
##
##	# @{$linset_gblocks_ref->[$lgb]} = ($cl, $gb, $indx_gp, $len_gb, $beg_gb, $end_gb) is for the gap-pattern block in the $lgb th place in the linear set, which is the $gb th block in the $cl th cluster (i.e., gapped segment) in the refefence MSA, where
##	# $indx_gp is the index of the gap-pattern in the @{$set_gpatterns}, and $beg/end_gb is the beginning/ending coordinate of the block (in the 0-based full-closed convention);
##	#
##	# %{$br2all_indels_ref} = ($branch_id => \@set_all_indels, ...), where
##	# [ @{$set_all_indels[$e]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk) is for the $e th (in the spatial order) indel event along the branch, with ] --- OBSOLETE as of Nov 4 (Tue), 2014.
##	# @{$set_all_indels[$e]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, $length_indel, \@indices_lin_blk) is for the $e th (in the spatial order) indel event along the branch, with # MODIFIED on Nov 4 (Tue), 2014.
##	# $beg/end_site being the coordinate (in the 0-based full-closed convention) of the beginning/ending-site of the event, and
##	# @indices_lin_blk storing the indices (in @{$linset_cls_gblocks_ref}) of the involved gap-pattern blocks;
##	#
##	# @{$set_indels_std_by_begsite_ref->[$i]} = ($beg_site, \@set_indels_beg_at_here) is for the $i th beginning site (in numerically ascending order), with
##	# @{$set_indels_beg_at_here[$e1]} = (exactly in the same format as @{$set_all_indels[$e]}) is for the $e1 th (in the order of occurrences) indel event beginning at $beg_site;
##	# @{$set_indels_std_by_endsite_ref->[$i]} = ($end_site, \@set_indels_end_at_here) is for the $i th ending site (in numerically ascending order), with
##	# @{$set_indels_end_at_here[$e2]} = (exactly in the same format as @{$set_all_indels[$e]}) is for the $e2 th (in the order of occurrences) indel event ending at $end_site.
##	#
##	my ($linset_gblocks_ref, $br2all_indels_ref, $set_indels_std_by_begsite_ref, $set_indels_std_by_endsite_ref) = reorganize_set_indels (@{$set_cls_gblocks_ref}, @{$init_psm_scn_for_cls_ref}); 
##	unless (defined $linset_gblocks_ref) {
##	    print STDOUT "[ERROR '$program_name'] Subroutine 'reorganize_set_indels' FAILED for the Reference MSA of the $s th segment !\n";
##	    print STDERR "[ERROR '$program_name'] $!\n\n";
##	    die ;
##	}
##
##	my $totct_gblks_ref = @{$linset_gblocks_ref};
##	my $totct_begsites_ref = @{$set_indels_std_by_begsite_ref};
##	my $totct_endsites_ref = @{$set_indels_std_by_endsite_ref};
##
##   if ($TEST_LEVEL>2) { # (VIIIa) 
##
##	print $fh_log "[[ Outputs of 'reorganize_set_indels' ]]\n\n";
##
##	print $fh_log '[ Content of @{$linset_gblocks_ref} (#{elements} = ', "$totct_gblks_ref) ]\n\n";
##
##	print $fh_log join ("\t", 'Indx_lin_gblk', 'Indx_cl', 'Indx_gblk', 'Indx_gpatt', 'Len_gb', 'Beg_gb', 'End_gb') , "\n\n";
##
##	for (my $lgb=0; $lgb<$totct_gblks_ref; $lgb++) { print $fh_log join ("\t", $lgb, @{$linset_gblocks_ref->[$lgb]}), "\n"; }
##	print $fh_log "\n\n\n";
##
##     if ($TEST_LEVEL>3) {
##
##	print $fh_log '[ Content of %{$br2all_indels_ref} (#{entries} = ', scalar (keys %{$br2all_indels_ref}), ") ]\n\n";
##
##	foreach my $br (sort {$a <=> $b} keys %{$br2all_indels_ref}) { # 2nd testing foreach-loop (over branches). #
##
##	    print $fh_log "{ Branch ID: $br }\n\n";
##
##	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014. #
##	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";
##
##	    foreach my $event (@{$br2all_indels_ref->{$br}}) {
##		my ($br_id2, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
##		##my ($br_id2, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
##		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
##		print $fh_log ' ' x 3, join ("\t", $br_id2, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
##		##print $fh_log ' ' x 3, join ("\t", $br_id2, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
##	    }
##	    print $fh_log "\n\n";
##	}					# End of the 2nd testing foreach-loop (over branches). #
##	print $fh_log "\n";
##
##	print $fh_log '[ Content of @{$set_indels_std_by_begsite_ref} (#{elements} = ', "$totct_begsites_ref) ]\n\n";
##
##	for (my $i=0; $i<$totct_begsites_ref; $i++) { # 3rd testing for-loop (over beginning sites). #
##
##	    my ($beg_site0, $events) = @{$set_indels_std_by_begsite_ref->[$i]};
##
##	    print $fh_log "{ The $i th beginning site (= $beg_site0) }\n\n";
##
##	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014.
##	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";
##
##	    foreach my $event (@{$events}) {
##		my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
##		##my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
##		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
##		print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
##		##print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
##	    }
##	    print $fh_log "\n\n";
##
##	}				# End of the 3rd testing for-loop (over beginning sites). #
##	print $fh_log "\n";
##
##	print $fh_log '[ Content of @{$set_indels_std_by_endsite_ref} (#{elements} = ', "$totct_endsites_ref) ]\n\n";
##
##	for (my $i=0; $i<$totct_endsites_ref; $i++) { # 4th testing for-loop (over ending sites). #
##
##	    my ($end_site0, $events) = @{$set_indels_std_by_endsite_ref->[$i]};
##
##	    print $fh_log "{ The $i th ending site (= $end_site0) }\n\n";
##
##	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014.
##	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";
##
##	    foreach my $event (@{$events}) {
##		my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
##		##my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
##		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
##		print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
##		##print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
##	    }
##	    print $fh_log "\n\n";
##
##	}				# End of the 4th testing for-loop (over ending sites). #
##	print $fh_log "\n";
##
##
##     }   # END of "if ($TEST_LEVEL>3) {...}"
##
##
##   } # END of "if ($TEST_LEVEL>2) { # (VIIIa) ...}"
##
##
##
##	print $fh_log "<<< (2) For Reconstructed MSA >>>\n\n";
##
##	# {Formats of the outputs are the identical to those for the reference MSA}
##	#
##	my ($linset_gblocks_rec, $br2all_indels_rec, $set_indels_std_by_begsite_rec, $set_indels_std_by_endsite_rec) = reorganize_set_indels (@{$set_cls_gblocks_rec}, @{$init_psm_scn_for_cls_rec}); 
##	unless (defined $linset_gblocks_rec) {
##	    print STDOUT "[ERROR '$program_name'] Subroutine 'reorganize_set_indels' FAILED for the Reconstructed MSA of the $s th segment !\n";
##	    print STDERR "[ERROR '$program_name'] $!\n\n";
##	    die ;
##	}
##
##	my $totct_gblks_rec = @{$linset_gblocks_rec};
##	my $totct_begsites_rec = @{$set_indels_std_by_begsite_rec};
##	my $totct_endsites_rec = @{$set_indels_std_by_endsite_rec};
##
##   if ($TEST_LEVEL>2) { # (VIIIb) 
##
##	print $fh_log "[[ Outputs of 'reorganize_set_indels' ]]\n\n";
##
##	print $fh_log '[ Content of @{$linset_gblocks_rec} (#{elements} = ', "$totct_gblks_rec) ]\n\n";
##
##	print $fh_log join ("\t", 'Indx_lin_gblk', 'Indx_cl', 'Indx_gblk', 'Indx_gpatt', 'Len_gb', 'Beg_gb', 'End_gb') , "\n\n";
##
##	for (my $lgb=0; $lgb<$totct_gblks_rec; $lgb++) { print $fh_log join ("\t", $lgb, @{$linset_gblocks_rec->[$lgb]}), "\n"; }
##	print $fh_log "\n\n\n";
##
##     if ($TEST_LEVEL>3) {
##
##	print $fh_log '[ Content of %{$br2all_indels_rec} (#{entries} = ', scalar (keys %{$br2all_indels_rec}), ") ]\n\n";
##
##	foreach my $br (sort {$a <=> $b} keys %{$br2all_indels_rec}) { # 2nd testing foreach-loop (over branches). #
##
##	    print $fh_log "{ Branch ID: $br }\n\n";
##
##	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014.
##	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";
##
##	    foreach my $event (@{$br2all_indels_rec->{$br}}) {
##		my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
##		##my ($br_id2, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
##		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
##		print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
##		##print $fh_log ' ' x 3, join ("\t", $br_id2, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
##	    }
##	    print $fh_log "\n\n";
##	}					# End of the 2nd testing foreach-loop (over branches). #
##	print $fh_log "\n";
##
##	print $fh_log '[ Content of @{$set_indels_std_by_begsite_rec} (#{elements} = ', "$totct_begsites_rec) ]\n\n";
##
##	for (my $i=0; $i<$totct_begsites_rec; $i++) { # 3rd testing for-loop (over beginning sites). #
##
##	    my ($beg_site0, $events) = @{$set_indels_std_by_begsite_rec->[$i]};
##
##	    print $fh_log "{ The $i th beginning site (= $beg_site0) }\n\n";
##
##	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014.
##	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";
##
##	    foreach my $event (@{$events}) {
##		my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
##		##my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
##		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
##		print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
##		##print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
##	    }
##	    print $fh_log "\n\n";
##
##	}				# End of the 3rd testing for-loop (over beginning sites). #
##	print $fh_log "\n";
##
##	print $fh_log '[ Content of @{$set_indels_std_by_endsite_rec} (#{elements} = ', "$totct_endsites_rec) ]\n\n";
##
##	for (my $i=0; $i<$totct_endsites_rec; $i++) { # 4th testing for-loop (over ending sites). #
##
##	    my ($end_site0, $events) = @{$set_indels_std_by_endsite_rec->[$i]};
##
##	    print $fh_log "{ The $i th ending site (= $end_site0) }\n\n";
##
##	    print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'length_indel', 'indices,lin,blk,involved'), "\n\n"; # MODIFIED on Nov 4 (Tue), 2014.
##	    ##print $fh_log ' ' x 3, join ("\t", 'Br_ID', 'beg_site', 'end_site', 'state_ue', 'state_le', 'indices,lin,blk,involved'), "\n\n";
##
##	    foreach my $event (@{$events}) {
##		my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $indices_lin_blk) = @{$event}; # MODIFIED on Nov 4 (Tue), 2014.
##		##my ($br_id, $beg_site, $end_site, $state_ue, $state_le, $indices_lin_blk) = @{$event};
##		my $cnct_indices_lin_blk = join (',', @{$indices_lin_blk});
##		print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $len_indel, $cnct_indices_lin_blk), "\n"; # MODIFIED on Nov 4 (Tue), 2014.
##		##print $fh_log ' ' x 3, join ("\t", $br_id, $beg_site, $end_site, $state_ue, $state_le, $cnct_indices_lin_blk), "\n";
##	    }
##	    print $fh_log "\n\n";
##
##	}				# End of the 4th testing for-loop (over ending sites). #
##	print $fh_log "\n";
##
##
##     }   # END of "if ($TEST_LEVEL>3) {...}"
##
##
##   } # END of "if ($TEST_LEVEL>2) { # (VIIIb) ...}"
##
##=cut # End of "Set Aside for Potential Use. (Pre 8)"



				## (IMPORTED w/ modifications from 'srd_map.ver0.6.pl', in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12).) (Pre 7,8,9,10) ##

	print STDOUT "\n<<< (Preliminary 7:) Calculate probabilities of the residue states of extant sequences on the descendant side of each branch, conditioned on the residue state at either end of the branch. >>>\n\n\n";
	print $fh_log "\n<<< (Preliminary 7:) Calculate probabilities of the residue states of extant sequences on the descendant side of each branch, conditioned on the residue state at either end of the branch. >>>\n\n\n";

	my ($node2set_sw_cond_probs, $node2set_sw_extcond_probs)
	 = calc_prob_descresstats_cond_ancstat2 (@cp_rec_seg_msa, @{$sequence_ids}, %{$name2id}, %{$node2ch}, %{$br2tr_mtrx}, $GAP);

	unless (defined $node2set_sw_cond_probs) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'calc_prob_descresstats_cond_ancstats2' FAILED on the $s th segment of the reconstructed MSA !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n\n";
		die ;
	}
    (print $fh_log " ... Subroutine, 'calc_prob_descresstats_cond_ancstat2', Ended on the $s th segment ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>=5) { # (Pre 7) #

	print $fh_log "\n<< Contents of the Preliminary Output (3-1): ", '%{$node2set_sw_cond_probs}', " >>\n\n";

	foreach my $node_id (sort {$a <=> $b} keys %{$node2set_sw_cond_probs}) { # 3rd-1st preliminary outer foreach-loop (over nodes). #

	    my $set_sw_cond_probs = $node2set_sw_cond_probs->{$node_id};
	    my $ct_sites = @{$set_sw_cond_probs};

	    print $fh_log ' ' x 2, "{ Node ID: $node_id }\n\n";
	    print $fh_log ' ' x 2, join ("\t", 'Site\Cond', @NUCLEOTIDES), "\n\n";

	    	# A BUG fixed on Jan 13 (Sat), 2018. #
	    for (my $s = 0; $s < $ct_sites; $s++) { print $fh_log ' ' x 2, join ("\t", $s, @{$set_sw_cond_probs->[$s]}), "\n"; }
	    ##for (my $s = 0; $s < $ct_sites; $s++) { print $fh_log ' ' x 2, join ("\t", $s, @{$set_sw_cond_probs}), "\n"; }

	    print $fh_log "\n\n";
	} # End of the 3rd-1st preliminary outer foreach-loop (over nodes). #

	print $fh_log "\n<< Contents of the Preliminary Output (3-1): ", '%{$node2set_sw_extcond_probs}', " >>\n\n";

	foreach my $node_id (sort {$a <=> $b} keys %{$node2set_sw_extcond_probs}) { # 3rd-2nd preliminary outer foreach-loop (over nodes). #

	    my $set_sw_extcond_probs = $node2set_sw_extcond_probs->{$node_id};
	    my $ct_sites = @{$set_sw_extcond_probs};

	    print $fh_log ' ' x 2, "{ Node ID: $node_id }\n\n";
	    print $fh_log ' ' x 2, join ("\t", 'Site\Cond', @NUCLEOTIDES), "\n\n";

	    	# A BUG fixed on Jan 13 (Sat), 2018. #
	    for (my $s = 0; $s < $ct_sites; $s++) { print $fh_log ' ' x 2, join ("\t", $s, @{$set_sw_extcond_probs->[$s]}), "\n"; }
	    ##for (my $s = 0; $s < $ct_sites; $s++) { print $fh_log ' ' x 2, join ("\t", $s, @{$set_sw_extcond_probs}), "\n"; }

	    print $fh_log "\n\n";
	} # End of the 3rd-2nd preliminary outer foreach-loop (over nodes). #

   } # END of "if ($TEST_LEVEL>=5) { # (Pre 7) # ...}"


	print STDOUT "\n<<< (Preliminary 8:) Calculating the joint probabilities of the residue state of the upper-end of each branch and the residue states of extant sequences on the ancestor side of the branch. >>>\n\n\n"; 
	print $fh_log "\n<<< (Preliminary 8:) Calculating the joint probabilities of the residue state of the upper-end of each branch and the residue states of extant sequences on the ancestor side of the branch. >>>\n\n\n"; 

	my ($br2set_sw_jnt_probs, $br2set_sw_extjnt_probs)
	    = calc_joint_prob_ancresstats_and_ancstat2 (@cp_rec_seg_msa, @{$sequence_ids}, %{$name2id}, %{$node2ch}, %{$br2tr_mtrx}, $GAP, @root_resfreqs, %{$node2set_sw_extcond_probs}) ;

	unless (defined $br2set_sw_jnt_probs) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'calc_prob_descresstats_cond_ancstats2' FAILED on the $s th segment of the reconstructed MSA !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n\n";
		die ;
	}
   (print $fh_log " ... Subroutine, 'calc_joint_prob_ancresstats_and_ancstat2', Ended on the $s th segment ...\n\n") if ($TEST_LEVEL> 1);


   if ($TEST_LEVEL>=5) { # (Pre 8) #
	print $fh_log "\n<< Contents of the Preliminary Output (5-1): ", '%{$br2set_sw_jnt_probs}', " >>\n\n";

	foreach my $br_id (sort {$a <=> $b} keys %{$br2set_sw_jnt_probs}) { # 5th-1st preliminary outer foreach-loop (over branches). #

	    my $set_sw_jnt_probs = $br2set_sw_jnt_probs->{$br_id};
	    my $ct_sites = @{$set_sw_jnt_probs};

	    print $fh_log ' ' x 2, "{ Branch ID: $br_id }\n\n";
	    print $fh_log ' ' x 2, join ("\t", 'Site\Jnt', @NUCLEOTIDES), "\n\n";

	    for (my $s = 0; $s < $ct_sites; $s++) { print $fh_log ' ' x 2, join ("\t", $s, @{$set_sw_jnt_probs->[$s]}), "\n"; }

	    print $fh_log "\n\n";
	}					# End of the 5th-1st preliminary outer foreach-loop (over branches). #


	print $fh_log "\n<< Contents of the Preliminary Output (5-2): ", '%{$br2set_sw_extjnt_probs}', " >>\n\n";

	foreach my $br_id (sort {$a <=> $b} keys %{$br2set_sw_extjnt_probs}) { # 5th-2nd preliminary outer foreach-loop (over branches). #

	    my $set_sw_extjnt_probs = $br2set_sw_extjnt_probs->{$br_id};
	    my $ct_sites = @{$set_sw_extjnt_probs};

	    print $fh_log ' ' x 2, "{ Branch ID: $br_id }\n\n";
	    print $fh_log ' ' x 2, join ("\t", 'Site\Jnt', @NUCLEOTIDES), "\n\n";

	    for (my $s = 0; $s < $ct_sites; $s++) { print $fh_log ' ' x 2, join ("\t", $s, @{$set_sw_extjnt_probs->[$s]}), "\n"; }

	    print $fh_log "\n\n";
	}					# End of the 5th-2nd preliminary outer foreach-loop (over branches). #

   } # End of "if ($TEST_LEVEL>=5) { # (Pre 8) # ...}"



	print STDOUT "\n<<< (Preliminary 9:) Calculating the probability of each column (under the given substitution model) >>>\n\n\n";
	print $fh_log "\n<<< (Preliminary 9:) Calculating the probability of each column (under the given substitution model) >>>\n\n\n";

	my $set_clm_probs = calc_clm_probs (@root_resfreqs, %{$node2set_sw_cond_probs}, $topnode_id);
	unless (defined $set_clm_probs) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'calc_clm_probs' FAILED on the $s th segment of the reconstructed MSA !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n\n";
		die ;
	}
   (print $fh_log " ... Subroutine, 'calc_clm_probs', Ended on the $s th segment ...\n\n") if ($TEST_LEVEL>1);

	my $ct_clms = @{$set_clm_probs};

   if ($TEST_LEVEL>=5) { # (Pre 9) #

	print $fh_log "\n<< Contents of the Preliminary output (6): ", '@{$set_clm_probs}', " (#{columns} = $ct_clms) >>\n\n";

	print $fh_log join ("\t", 'Column', 'Probability'), "\n\n";

	for (my $c = 0; $c < $ct_clms; $c++) { print $fh_log join ("\t", $c, $set_clm_probs->[$c]), "\n"; }
	print $fh_log "\n\n";

   } # End of "if ($TEST_LEVEL>=5) { # (Pre 9) # ...}"


##=DISABLED because it is SUPERFLUOUS!! (Pre 10)
##
##	print STDOUT "\n<<< (Preliminary 10:) Calculating the Dollo-parsimonious indel history that results in the reconstructed MSA >>>\n\n\n";
##	print $fh_log "\n<<< (Preliminary 10:) Calculating the Dollo-parsimonious indel history that results in the reconstructed MSA >>>\n\n\n";
##
##	my ($set_gpatterns, $set_node2dparsstat, $set_br2change, $info_psm_indel_hsts0)
##	  = enum_psm_indel_hsts_for_msa_finer_1ST_HF (%{$node2ch}, %{$node2pa}, %{$name2id}, @{$sequence_ids}, @cp_rec_seg_msa, $CHAR_ABS, $CHAR_PRES, %{$node2depth}, %{$node2ct_exoffs});
##
##	unless ((defined $info_psm_indel_hsts0) and (@{$info_psm_indel_hsts0}>2) ) {
##		print STDOUT "[ERROR '$program_name'] The subroutine 'enum_psm_indel_hsts_for_msa_finer_1ST_HF' FAILED on the input tree & the $s th segment of the reconstructed MSA !\n";
##		print STDERR "[ERROR '$plscript_name'] $! !\n\n";
##		die ;
##	}
##   (print $fh_log " ... Subroutine, 'enum_psm_indel_hsts_for_msa_finer_1ST_HF', Ended on the $s th segment ...\n\n") if ($TEST_LEVEL>1);
##
##	my ($set_cls_gpblocks, $totcts0_indels_for_cls, $psm_hsts0_for_cls) = @{$info_psm_indel_hsts0};
##	my $ct_cls = @{$set_cls_gpblocks};
##	my $ct_gpd_segments = $ct_cls; # Added to suppress error messages (on 2019/06/18). #
##
##   if ($TEST_LEVEL>3) { # (Pre 10) #
##	print $fh_log "\n<< Main Contents of the Preliminary output (7) ", '@{$set_cls_gpblocks} & @{$psm_hsts0_for_cls}', " >>\n\n";
##
##	for (my $cl = 0; $cl < $ct_cls; $cl++) { # The 7 th preliminary  for-loop (over gapped segments). #
##		my $cls_gpblks = $set_cls_gpblocks->[$cl];
##		my $ct_gpblks = @{$cls_gpblks};
##
##		print $fh_log ' ' x 2, "{ The $cl th gapped segment (#{gap-pattern blocks} = $ct_gpblks) }\n\n";
##		print $fh_log ' ' x 2, join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal)'"), "\n\n";
##		for (my $bl=0; $bl < $ct_gpblks; $bl++) { # 1st inner for-loop (over blocks). #
##		    my ($indx_gp, $len, $beg, $end) = @{$cls_gpblks->[$bl]};
##		    my $gpattern = ${$set_gpatterns->[$indx_gp]};
##		    print $fh_log ' ' x 2, join ("\t", $bl, $len, $beg, $end, $indx_gp, "'$gpattern'"), "\n";
##		}				# End of the 1st inner for-loop (over blocks). #
##		print $fh_log "\n\n";
##
##		my ($br2indels0, $node2bw_gstats0) = @{$psm_hsts0_for_cls->[$cl]};
##
##		print $fh_log ' ' x 2, "( Map of INITIAL block-wise gap states onto the nodes )\n\n";
##		print $fh_log ' ' x 2, join ("\t", 'Node_ID', "'block_wise_gap_states0(concatenated)'"), "\n\n";
##
##		foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats0}) { # 2nd inner foreach-loop (over nodes). #
##		    my $cnct_bw_gstats0 = join ('', @{$node2bw_gstats0->{$node}});
##		    print $fh_log ' ' x 2, join ("\t", $node, "'$cnct_bw_gstats0'"), "\n";
##		}                           # End of the 2nd inner foreach-loop (over nodes). #
##		print $fh_log "\n\n";
##		
##	}					# End of the 7th preliminary for-loop (over gapped segments). #
##   } # End of "if ($TEST_LEVEL>3) { # (Pre 10) # ...}"
##
##=cut # END of "DISABLED because it is SUPERFLUOUS!! (Pre 10)"

				## END of "(IMPORTED w/ modifications from 'srd_map.ver0.6.pl', in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12).) (Pre 7,8,9,10)" ##



				## (IMPORTED w/ modifications from 'srd_map.ver0.6.pl', in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11&12).) (MAIN 3) ##

	print STDOUT "\n\n<<<< MAIN PROCESS 3A:  Creating & Outputting the Observed Substitutional Residue-Difference Map of the reconstructed MSA. >>>>\n\n\n";
	print $fh_log "\n\n<<<< MAIN PROCESS 3A:  Creating & Outputting the Observed Substitutional Residue-Difference Map of the reconstructed MSA. >>>>\n\n\n";

	our $IF_OBSERVED = 1;

	# my ($beg_msa, $end_msa) = (0, $lenaln -1);

	my ($srd_map_obs, $br2sw_srds_obs)
	  =  mk_srd_map ($IF_OBSERVED, @br_ids_for_display, $CHAR_ABS, $CHAR_PRES,
		$topnode_id, %{$node2pa}, %{$node2ntfreqs}, %{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, @{$set_clm_probs},
		@{$info_psm_indel_hsts0}, $lcl_beg_recmsa, $lcl_end_recmsa, %{$br_attr}, %{$br2eq});
	unless (defined $srd_map_obs) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_srd_map' FAILED to create an OBSERVED SRD Map from the $s th segment of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'mk_srd_map', (w/ \$IF_OBSERVED = 1) ENDED on the $s th segment ...\n\n") if ($TEST_LEVEL >1);

   if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-2) #

	##my $out_srd_obs = "$segoutdir/$name_out_obs";

	my $fh_osrd_obs = open_output_file ($out_srd_obs, 0);
	unless (defined $fh_osrd_obs) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_srd_obs\" !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n\n";
		die ;
	}

	unless ( print_srd_map (*${$fh_osrd_obs}, @br_ids_for_display, %{$id2name}, @{$srd_map_obs}, $lcl_beg_recmsa, $lcl_end_recmsa, $CHAR_ABS, $DIS_LEN_SRD, $CT_EFF_DGTS) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_srd_map' FAILED to print out the OBSERVED SRD Map on the $s th segment of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'print_srd_map', ENDED to print out the OBSERVED SRD Map on the $s th segment of the reconstructed MSA ...\n\n") if ($TEST_LEVEL >1);

	close ($fh_osrd_obs);

   } else {

   (print $fh_log "\n ... SKIPPED The subroutine, 'print_srd_map', for the OBSERVED SRD Map, because \$OUTPUT_LEVEL = $OUTPUT_LEVEL ...\n\n") if ($TEST_LEVEL >1);

   } # END of "if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-2) # ...} else {...}"


	print STDOUT "\n\n<<<< MAIN PROCESS 3B:  Creating & Outputting the Expected Substitutional Residue-Difference Map of the reconstructed MSA. >>>>\n\n\n";
	print $fh_log "\n\n<<<< MAIN PROCESS 3B:  Creating & Outputting the Expected Substitutional Residue-Difference Map of the reconstructed MSA. >>>>\n\n\n";

	$IF_OBSERVED = 0;

	my ($srd_map_exp, $br2sw_srds_exp)
	  =  mk_srd_map ($IF_OBSERVED, @br_ids_for_display, $CHAR_ABS, $CHAR_PRES,
		$topnode_id, %{$node2pa}, %{$node2ntfreqs}, %{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, @{$set_clm_probs},
		@{$info_psm_indel_hsts0}, $lcl_beg_recmsa, $lcl_end_recmsa, %{$br_attr}, %{$br2eq});
	unless (defined $srd_map_exp) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_srd_map' FAILED to create an EXPECTED SRD Map from the $s th segment of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'mk_srd_map', (w/ \$IF_OBSERVED = 0) ENDED on the $s th segment ...\n\n") if ($TEST_LEVEL >1);

   if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-3) #

	##my $out_srd_exp = "$segoutdir/$name_out_exp";

	my $fh_osrd_exp = open_output_file ($out_srd_exp, 0);
	unless (defined $fh_osrd_exp) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_srd_exp\" !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n\n";
		die ;
	}

	unless ( print_srd_map (*${$fh_osrd_exp}, @br_ids_for_display, %{$id2name}, @{$srd_map_exp}, $lcl_beg_recmsa, $lcl_end_recmsa, $CHAR_ABS, $DIS_LEN_SRD, $CT_EFF_DGTS) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_srd_map' FAILED to print out the EXPECTED SRD Map on the $s th segment of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'print_srd_map', ENDED to print out the EXPECTED SRD Map on the $s th segment ...\n\n") if ($TEST_LEVEL >1);

	##close ($fh_osrd_obs); # A BUG Fixed on 2020/03/25. #
	close ($fh_osrd_exp); 

   } else {

   (print $fh_log "\n ... SKIPPED The subroutine, 'print_srd_map', for the EXPECTED SRD Map, because \$OUTPUT_LEVEL = $OUTPUT_LEVEL ...\n\n") if ($TEST_LEVEL >1);

   } # END of "if ($OUTPUT_LEVEL>0) { # (ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/25.) (MAIN-E-3) # ...} else {...}"

				## END of "(IMPORTED w/ modifications from 'srd_map.ver0.6.pl', in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/11&12).) (MAIN 3)" ##


				### MAJOR MODIFICATION in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/03. (MAIN FIN) ###

	print STDOUT "\n\n<<<< MAIN PROCESS FINAL A:  Tallying the OBSERVED Substitutional Residue-Differences in the current MSA segment.  >>>>\n\n\n";
	print $fh_log "\n\n<<<< MAIN PROCESS FINAL A:  Tallying the OBSERVED Substitutional Residue-Differences in the current MSA segment.  >>>>\n\n\n";

	my ($set_total_sums_srd_obs, $setset_brwise_sums_srd_obs)
		= tally_srds_in_map (@{$srd_map_obs}, $CHAR_ABS);
	unless (defined $setset_brwise_sums_srd_obs) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'tally_srds_in_map' FAILED on the OBSERVED SRDS in the correct-segment w/ ID, '$c_segid', of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'tally_srds_in_map', ENDED on the OBSERVED SRDS in the correct-segment w/ ID, '$c_segid' ...\n\n") if ($TEST_LEVEL >1);


	print $fh_main_obs "[ The Correct-Segment ID, '$c_segid' (\$ref_len0 = $ref_len0, \$rec_len0 = $rec_len0): ]\n\n";

	unless ( print_sums_srds_in_map ( *${$fh_main_obs}, @br_ids_for_display, %{$id2name}, 
			@{$set_total_sums_srd_obs}, @{$setset_brwise_sums_srd_obs}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_sums_assoc_srds_in_map' FAILED on the OBSERVED SRDS in the correct-segment w/ ID, '$c_segid', of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'print_sums_assoc_srds_in_map', ENDED on the OBSERVED SRDS in the correct-segment w/ ID, '$c_segid' ...\n\n") if ($TEST_LEVEL >1);

	print $fh_main_obs "\n\n";


	print STDOUT "\n\n<<<< MAIN PROCESS FINAL B:  Tallying the EXPECTED Substitutional Residue-Differences in the current MSA segment.  >>>>\n\n\n";
	print $fh_log "\n\n<<<< MAIN PROCESS FINAL B:  Tallying the EXPECTED Substitutional Residue-Differences in the current MSA segment.  >>>>\n\n\n";

	my ($set_total_sums_srd_exp, $setset_brwise_sums_srd_exp)
		= tally_srds_in_map (@{$srd_map_exp}, $CHAR_ABS);
	unless (defined $setset_brwise_sums_srd_exp) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'tally_srds_in_map' FAILED on the EXPECTED SRDS in the correct-segment w/ ID, '$c_segid', of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'tally_srds_in_map', ENDED on the EXPECTED SRDS in the correct-segment w/ ID, '$c_segid' ...\n\n") if ($TEST_LEVEL >1);


	print $fh_main_exp "[ The Correct-Segment ID, '$c_segid' (\$ref_len0 = $ref_len0, \$rec_len0 = $rec_len0): ]\n\n";

	unless ( print_sums_srds_in_map ( *${$fh_main_exp}, @br_ids_for_display, %{$id2name}, 
			@{$set_total_sums_srd_exp}, @{$setset_brwise_sums_srd_exp}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_sums_assoc_srds_in_map' FAILED on the EXPECTED SRDS in the correct-segment w/ ID, '$c_segid', of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'print_sums_assoc_srds_in_map', ENDED on the EXPECTED SRDS in the correct-segment w/ ID, '$c_segid' ...\n\n") if ($TEST_LEVEL >1);

	print $fh_main_exp "\n\n";

				### RESTART FROM HERE (MAIN FINAL) ###

	print STDOUT "\n\n<<<< MAIN PROCESS FINAL C:  Distributing the OBSERVED Substitutional Residue-Differences, in the current MSA segment.  >>>>\n\n\n";
	print $fh_log "\n\n<<<< MAIN PROCESS FINAL C:  Distributing the OBSERVED Substitutional Residue-Differences, in the current MSA segment.  >>>>\n\n\n";

	unless ( brwise_distribute_srds_in_map ( @{$srd_map_obs}, $CHAR_ABS, @brwise_distri_obs_srds, @BOUNDS_SRD ) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'brwise_distribute_srds_in_map' FAILED on the OBSERVED SRDS in the correct-segment w/ ID, '$c_segid', of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'brwise_distribute_srds_in_map', ENDED on the OBSERVED SRDS in the correct-segment w/ ID, '$c_segid' ...\n\n") if ($TEST_LEVEL >1);


=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (MAIN FIN)

			### ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/24). (MAIN FIN-A) ###

	print STDOUT "\n\n<<<< MAIN PROCESS FINAL A:  Tallying the OBSERVED Substitutional Residue-Differences in the \"Alignment-Shear\"s and at the \"None-Shear\" positions, & Outputting the Results.  >>>>\n\n\n";
	print $fh_log "\n\n<<<< MAIN PROCESS FINAL A:  Tallying the OBSERVED Substitutional Residue-Differences in the \"Alignment-Shear\"s and at the \"None-Shear\" positions, & Outputting the Results.  >>>>\n\n\n";

	# my @set_maps_alnshear = (); # element = $map_alnshear # ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/24). (M2-1) #

	my ($set_total_sums_srd_obs_vs_alnshear, $set_brwise_sums_srd_obs_vs_alnshear)
	     ##= tally_assoc_srds_and_alnshears (@{$srd_map_obs}, @set_maps_alnshear, $CHAR_ABS, $CHAR_NOSHEAR); # A BUG Fixed on 2020/03/25. #
	     = tally_assoc_srds_and_alnshears (@set_maps_alnshear, @{$srd_map_obs}, $CHAR_ABS, $CHAR_NOSHEAR); 
	unless (defined $set_brwise_sums_srd_obs_vs_alnshear) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'tally_assoc_srds_and_alnshears' FAILED on the OBSERVED SRDS in the $s th segment of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'tally_assoc_srds_and_alnshears', ENDED on the OBSERVED SRDS in the $s th segment ...\n\n") if ($TEST_LEVEL >1);


	print $fh_main_obs "[ The $s th segment (\$ref_len0 = $ref_len0, \$rec_len0 = $rec_len0; #{maps} = $ct_maps_pcls_on_blks): ]\n\n";

	unless ( print_sums_assoc_srds_and_alnshears ( *${$fh_main_obs}, @br_ids_for_display, %{$id2name}, 
			@{$set_total_sums_srd_obs_vs_alnshear}, @{$set_brwise_sums_srd_obs_vs_alnshear}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_sums_assoc_srds_and_alnshears' FAILED on the OBSERVED SRDS in the $s th segment of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'print_sums_assoc_srds_and_alnshears', ENDED on the OBSERVED SRDS in the $s th segment ...\n\n") if ($TEST_LEVEL >1);

	print $fh_main_obs "\n\n";


	print STDOUT "\n\n<<<< MAIN PROCESS FINAL B:  Tallying the EXPECTED Substitutional Residue-Differences in the \"Alignment-Shear\"s and at the \"None-Shear\" positions, & Outputting the Results.  >>>>\n\n\n";
	print $fh_log "\n\n<<<< MAIN PROCESS FINAL B:  Tallying the EXPECTED Substitutional Residue-Differences in the \"Alignment-Shear\"s and at the \"None-Shear\" positions, & Outputting the Results.  >>>>\n\n\n";


	my ($set_total_sums_srd_exp_vs_alnshear, $setset_brwise_sums_srd_exp_vs_alnshear)
	     ##= tally_assoc_srds_and_alnshears (@{$srd_map_exp}, @set_maps_alnshear, $CHAR_ABS, $CHAR_NOSHEAR); # A BUG Fixed on 2020/03/25. #
	     = tally_assoc_srds_and_alnshears (@set_maps_alnshear, @{$srd_map_exp}, $CHAR_ABS, $CHAR_NOSHEAR);
	unless (defined $setset_brwise_sums_srd_exp_vs_alnshear) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'tally_assoc_srds_and_alnshears' FAILED on the EXPECTED SRDS in the $s th segment of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'tally_assoc_srds_and_alnshears', ENDED on the EXPECTED SRDS in the $s th segment ...\n\n") if ($TEST_LEVEL >1);


	print $fh_main_exp "[ The $s th segment (\$ref_len0 = $ref_len0, \$rec_len0 = $rec_len0; #{maps} = $ct_maps_pcls_on_blks): ]\n\n";

	unless ( print_sums_assoc_srds_and_alnshears ( *${$fh_main_exp}, @br_ids_for_display, %{$id2name}, 
			@{$set_total_sums_srd_exp_vs_alnshear}, @{$setset_brwise_sums_srd_exp_vs_alnshear} ) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_sums_assoc_srds_and_alnshears' FAILED on the EXPECTED SRDS in the $s th segment of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'print_sums_assoc_srds_and_alnshears', ENDED on the EXPECTED SRDS in the $s th segment ...\n\n") if ($TEST_LEVEL >1);

	print $fh_main_exp "\n\n";


			### END of "ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/24). (MAIN FIN-A)" ###


			### ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). (MAIN FIN-C) ###

	print STDOUT "\n\n<<<< MAIN PROCESS FINAL C:  Distributing the OBSERVED Substitutional Residue-Differences, on & off the \"Alignment-Shear\"s.  >>>>\n\n\n";

	unless ( brwise_distribute_srds_assoc_w_alnshear ( @set_maps_alnshear, @{$srd_map_obs}, $CHAR_ABS, $CHAR_NOSHEAR,
			@brwise_distri_obs_srds_on_as, @brwise_distri_obs_srds_off_as, @BOUNDS_SRD ) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'brwise_distribute_srds_assoc_w_alnshear' FAILED on the OBSERVED SRDS in the $s th segment of the reconstructed MSA: $? !\n";
		print STDERR "[ERROR '$plscript_name'] $!\n\n";
		die ;
	}
   (print $fh_log "\n ... The subroutine, 'brwise_distribute_srds_assoc_w_alnshear', ENDED on the OBSERVED SRDS in the $s th segment ...\n\n") if ($TEST_LEVEL >1);

			### END of "ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/26). (MAIN FIN-C)" ###

=cut # END of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (MAIN FIN)"

				### END of "MAJOR MODIFICATION in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/03. (MAIN FIN)" ###


		## Close Temporary Filehandles, and Remove Unnecessary Temporary Files. ##
		## (MOVED from below with some modifications in creation of 'classify_msa_errors.1.pl'.)
		## ... Needs be Modified...

	close ($fh_log);

	print STDOUT "\n --- Finished analysis for the segment $segid ! ---\n\n";

	## close ($fh_log);

	$ct_succ_segs++;

}			# End of the main outer for-loop (over MSA segments) #


		# ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/24). (Post-MAIN) #
close ($fh_main_obs);
close ($fh_main_exp);
		# End of "ADDED in creation of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' (on 2020/03/24). (Post-MAIN)" #

=DISABLED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (MAIN-END)
	# MODIFIED in creation of 'classify_msa_errors_via_mblks.2.pl'. (on Oct 6 (Tue), 2015). #
close ($fh_out1);
close ($fh_out2);
##close ($fh_out);
=cut # END of "DISABLED in creation of 'map_aln_shear.ver0.6.pl' (on 2020/03/10). (MAIN-END)"

print STDOUT "\n\n === All the Analyses FINISHED !! ===\n\n";

print STDOUT "#{Segments skipped} = $ct_skipped_segs .\n";
##print STDOUT "#{Segments whose ComplLiMMent calculations skipped} = $ct_skipped_segs2 .\n";
print STDOUT "#{Segments skipped because either reference or reconstructed ones are NULL} = $ct_skipped_segs2 .\n"; # Modified in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/29.
print STDOUT "#{Segments skipped because they are too long} = $ct_skipped_segs3 .\n"; # ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/29. #

print STDOUT "#{Anomalous segments} = $ct_anom_segs .\n";
print STDOUT "#{Segments used as inputs} = $ct_in_segs .\n";
print STDOUT "#{Segments successfully processed} = $ct_succ_segs .\n";
print STDOUT "#{Segments failed to be processed} = $ct_fail_segs .\n\n";

unless (0 == $ct_fail_segs) {
	print STDOUT "[ERROR '$program_name'] FAILED to process $ct_fail_segs segments !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

=DISABLED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/03).
		# (ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/21.) #
		# (=> IMPORTED to 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.) #

# my %ct_maps2freq; # = (#{maps} => {frequency}, ...) # ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' (on 2020/03/21).

print STDOUT "\n[[ Frequencies of #{maps (psp-blocks -> psm-clusterings) in each segment}: ]]\n\n";

print STDOUT join ("\t", '#{maps}', 'frequency'), "\n\n";

foreach my $ct_maps (sort {$a <=> $b} keys %ct_maps2freq) {
	print STDOUT join ("\t", $ct_maps, $ct_maps2freq{$ct_maps}), "\n";
}
print STDOUT "\n\n";

		# END of "(ADDED in creation of 'map_aln_shear_clm_based.ver0.6.pl' on 2020/03/21.) #
		# (=> IMPORTED to 'map_alnshear_cb_and_srd.ver0.6.pl' on 2020/03/23.)" #
=cut # END of "DISABLED in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' (on 2020/04/03)."


		# (ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (FNL) #
		# (=> MAJOR REVISION in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/03.) (FNL) #

print STDOUT "\n<<< FINALLY, Print Out the Branch-wise Distributions of SRDs in the Correct Segments, to an output file. >>>\n\n\n";

		# (i) Compute the total distributions. #

my $tot_distri_obs_srds = $brwise_distri_obs_srds[$ct_all_brs];

my $ct_bds_srd = @BOUNDS_SRD;

for (my $n=0; $n <$ct_all_brs; $n++) { # Trailing outer for-loop (over all branches). #

    my $distri_srds_br = $brwise_distri_obs_srds[$n];

    for (my $i = 0; $i <= $ct_bds_srd; $i++) { # Inner for-loop (over all bins). #

	$tot_distri_obs_srds->[$i] += $distri_srds_br->[$i];

    } # End of the inner for-loop (over all bins). #

} # End of the trailing outer for-loop (over all branches). #


		# (ii) Prepare a set of column-labels. #

## my $ct_all_brs;
## my %id2name;

my @clm_labels = ();

for (my $n=0; $n <$ct_all_brs; $n++) {
	my $br_id = $br_ids_for_display[$n];
	my $node_name = $id2name{$br_id};
	$clm_labels[$n] = (defined $node_name) ? $node_name : $br_id;
}
$clm_labels[$ct_all_brs] = 'Total';

my @tags = ('Intervals', @clm_labels); # ADDED in DEBUGGING 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/27. #


		# (iii) Actually print out the distributions. #

my $fh_o_dstr = open_output_file ($out_dstr_srds, 0);
unless (defined $fh_o_dstr) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_dstr_srds\" !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}

print $fh_o_dstr "<<< Branch-wise Distributions of \"Observed\" SRDS >>>\n\n\n";

batch_pr_file_distri_xl_fmt (@brwise_distri_obs_srds, @BOUNDS_SRD, 'SRD', *${$fh_o_dstr}, "%5.3f", "%8.3f", 0, @tags) ; 

print $fh_o_dstr "\n\n";

close ($fh_o_dstr);

=For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (FNL)

print STDOUT "\n<<< FINALLY, Print Out the Branch-wise Distributions of SRDs, 'on' and 'off' the Alignment-Shears, to an output file. >>>\n\n\n";

		# (i) Compute the total distributions. #

my $tot_distri_on_as = $brwise_distri_obs_srds_on_as[$ct_all_brs];
my $tot_distri_off_as = $brwise_distri_obs_srds_off_as[$ct_all_brs];

my $ct_bds_srd = @BOUNDS_SRD;

for (my $n=0; $n <$ct_all_brs; $n++) { # Trailing outer for-loop (over all branches). #

    my $distri_on_as_br = $brwise_distri_obs_srds_on_as[$n];
    my $distri_off_as_br = $brwise_distri_obs_srds_off_as[$n];

    for (my $i = 0; $i <$ct_bds_srd; $i++) { # Inner for-loop (over all bins). #

	$tot_distri_on_as->[$i] += $distri_on_as_br->[$i];
	$tot_distri_off_as->[$i] += $distri_off_as_br->[$i];

    } # End of the inner for-loop (over all bins). #

} # End of the trailing outer for-loop (over all branches). #


		# (ii) Prepare a set of column-labels. #

## my $ct_all_brs;
## my %id2name;

my @clm_labels = ();

for (my $n=0; $n <$ct_all_brs; $n++) {
	my $br_id = $br_ids_for_display[$n];
	my $node_name = $id2name{$br_id};
	$clm_labels[$n] = (defined $node_name) ? $node_name : $br_id;
}
$clm_labels[$ct_all_brs] = 'Total';

my @tags = ('Intervals', @clm_labels); # ADDED in DEBUGGING 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/27. #


		# (iii) Actually print out the distributions. #

my $fh_o_dstr = open_output_file ($out_dstr_srds, 0);
unless (defined $fh_o_dstr) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_dstr_srds\" !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}

##print $fh_o_dstr "<<< Branch-wise Distributions of SRDS >>>\n\n\n";
print $fh_o_dstr "<<< Branch-wise Distributions of \"Observed\" SRDS >>>\n\n\n"; # CORRECTED on 2020/03/27. #

print $fh_o_dstr "< ON the \"Alignment-Shear\"s >\n\n\n";

##batch_pr_file_distri_xl_fmt (@brwise_distri_obs_srds_on_as, @BOUNDS_SRD, 'SRD', *${$fh_o_dstr}, "%5.3f", "%8.3f", 0, @clm_labels) ;
batch_pr_file_distri_xl_fmt (@brwise_distri_obs_srds_on_as, @BOUNDS_SRD, 'SRD', *${$fh_o_dstr}, "%5.3f", "%8.3f", 0, @tags) ; # CORRECTED in DEBUGGING 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/27. #

print $fh_o_dstr "\n\n\n";


print $fh_o_dstr "< OFF the \"Alignment-Shear\"s >\n\n\n";

##batch_pr_file_distri_xl_fmt (@brwise_distri_obs_srds_off_as, @BOUNDS_SRD, 'SRD', *${$fh_o_dstr}, "%5.3f", "%8.3f", 0, @clm_labels) ;
batch_pr_file_distri_xl_fmt (@brwise_distri_obs_srds_off_as, @BOUNDS_SRD, 'SRD', *${$fh_o_dstr}, "%5.3f", "%8.3f", 0, @tags) ; # CORRECTED in DEBUGGING 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/27. #

print $fh_o_dstr "\n\n";


close ($fh_o_dstr);

=cut # END of "For the predecessor, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'. (FNL)"

		# END of "(ADDED in REVISION of 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/26.) (FNL) #
		# (=> MAJOR REVISION in creation of 'ingr_stat_srds_in_correct_segs.ver0.6.pl' on 2020/04/03.) (FNL)" #




		### DEFINITIONS of the subroutines ADDED in creation of 'classify_msa_errors_via_mblks.2.pl'. ###

# pr_map_shifts_respos:
#	This subroutine prints out a position-shift map
#	by formatting it so that the map will be decently visible.
#	[CREATED on Oct 15 (Thu), 2015.]
#
# INPUTS: (\*${$fh}, \@sequence_ids, \@map_shifts_respos, $char_gap, $dis_len_shift), where
#	$fh is the reference to a file-handle;
#	$sequence_ids[$k] is the ID of the sequence in the $k th row of the MSAs;
#	$map_shifts_respos[$k]->[$s] is the position-shift at the intersection of the $k th row and the $s th column of the mapped MSA;
#	$char_gap is the character representing a gap;
#	$dis_len_shift is the maximum length of the run of shifts to be displayed in a line.
#
# OUTPUTS: 1 if the subroutine succeeded,
#	   0 if it failed.
#
sub pr_map_shifts_respos (\*\@\@$$) {

	my ($fh, $sequence_ids, $map_shifts_respos, $char_gap, $dis_len_shift) = @_;

	my $ct_sequeunces = @{$sequence_ids};
	unless ($ct_sequences>0) {
	    print STDOUT "[INFORMATION 'pr_map_shifts_respos'] NO SEQUENCES in the alignment !\n\n";
	    return 1;
	}
	unless (@{$map_shifts_respos} == $ct_sequences) {
	    print STDOUT "[ERROR 'pr_map_shifts_respos'] The number of sequences in the map (=", scalar (@{$map_shifts_respos}), ") CLASHES with that in the set of sequences (=$ct_sequences): $! !\n\n";
	    return 0;
	}

	my $map_len = @{$map_shifts_respos->[0]}; 

	my $quotient = int (($map_len -1) / $dis_len_shift);

	for (my $r=0; $r<=$quotient; $r++) {  # 1st outer for-loop (over segmented displayes). #

	    my $beg_dis = $r * $dis_len_shift;
	    my $end_dis = ($r == $quotient) ? $map_len - 1 : ($r+1) * $dis_len_shift - 1 ;
	    my $format = "%-10s\t" . "%5d" x (($r==$quotient) ? $map_len - $quotient * $dis_len_shift : $dis_len_shift) . "\n";

	    printf $fh $format, 'Seq_ID', ($beg_dis .. $end_dis);
	    print $fh "\n";

	    for (my $k=0; $k<$ct_sequences; $k++) {  # Middle for-loop (over sequences). #

		my $out_string = sprintf ("%-10s\t", $sequence_ids->[$k]);

		my $row_shifts = $map_shifts_respos->[$k];
		my $ct_sites = @{$row_shifts};

		my $end_dis2 = ($ct_sites-1 < $end_dis) ? $ct_sites-1 : $end_dis;
		for (my $s=$beg_dis; $s<= $end_dis2; $s++) { # Inner testing for-loop (over sequences). #
		   my $shift = $row_shifts->[$s];
		   $out_string .= ($shift eq $char_gap) ? sprintf ("%5s", "'$char_gap'") : sprintf ("%5d", $shift) ;
		}				# End of the inner testing for-loop (over sequences). #

		print $fh $out_string, "\n";
	    
	    }			# End of the middle for-loop (over sequences). #
	    print $fh "\n\n\n";

	}			# End of the 1st outer for-loop (over segmented displayes). #

	return 1;

}		# End of the subroutine 'pr_map_shifts_respos'. #


# blockwise_framework:
#	This satellite subroutine of 'framework_cblock' creates
#	the framework of an input position-shift block.
#	[CREATED on Oct 12 (Mon), 2015.]
#
# INPUT: ($beg_bl, $end_bl, \@seqs_involved, \@excl_sbblockks, \%node2ch, \%node2seq_indx), where
#
#	$beg_bl and $end_bl are the beginning- and ending-positions of the input block;
#	@seqs_involved stores the indices of the sequences involved with the input block;
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block;
#	%node2ch = ($node_id => \@child_nodes, ...) for internal nodes;
#	%node2seq_indx = ($node_id => $indx_seq, ...) for external nodes;
#
# OUTPUT: (\%frwk_bl), where
#	$frwk_bl{$indx_seq} = [$beg_frwk, $end_frwk], which determines the beginning- and ending-positions of the framework of the input block in the $indx_seq th sequence (row),
#	 	     = undef if the input block does not involve the $indx_seq th sequence. 
#
sub blockwise_framework ($$\@\@\%\%) { 

	my ($beg_bl, $end_bl, $seqs_involved, $excl_sbblocks, $node2ch, $node2seq_indx) = @_;

	my %frwk_bl;

	foreach my $indx_seq (@{$seqs_involved}) { $frwk_bl{$indx_seq} = [$beg_bl, $end_bl]; }

	foreach my $info_sbbl (@{$excl_sbblocks}) { # 2nd foreach-loop (over excluded sub-blocks). #

	    my ($beg_sbbl, $end_sbbl, $mrca_sbbl) = @{$info_sbbl};

	    (($beg_bl < $beg_sbbl) and ($end_sbbl < $end_bl)) and next; # Skip if the sub-block is completely within the block, because the framework itself will not be influenced by such a sub-block. #

	    my $exoffs = fetch_ext_offsprings_nr ($mrca_sbbl, %{$node2ch}) ;

	    foreach my $nodeid_exoff (@{$exoffs}) { # Inner foreach-loop (over external offsprings). #
		my $seqid_exoff = $node2seq_indx->{$nodeid_exoff};
		my $frwk_exoff = $frwk_bl{$seqid_exoff};

		(defined $frwk_exoff) or next; # Skip if the sequence is not involved with the block. #

		if ($beg_sbbl <= $beg_bl) {
		    if ($end_bl <= $end_sbbl) {
			delete $frwk_bl{$seqid_exoff}; # In this case, the sequence is completely excluded.
		    } else {
			$frwk_exoff->[0] = $end_sbbl + 1;
		    }
		} elsif ($end_bl <= $end_sbbl) {
		    $frwk_exoff->[1] = $beg_sbbl-1;
		}
	    }					# End of the inner foreach-loop (over external offsprings). #

	}					# End of the 2nd foreach-loop (over excluded sub-blocks). #

	return \%frwk_bl;

}	# End of the subroutine 'blockwise_framework'. #


# framework_cblock:
#	This subroutine creates the framework of the subject composite-block.
#	[CREATION STARTED on Oct 11 (Sun), 2015, ENDED on Oct 12 (Mon), 2015.]
#
# INPUT: ($ct_seqs, $indx_cb, \@set_composite_blocks, \@hard_linking, \@set_blocks_shift_respos, \%node2ch, \%node2seq_indx), where
#
#	$indx_cb is the index of the subject c-block;
#	@{$set_composite_blocks[$cb]} = ($beg_cb, $end_cb, $nodeid_mrca_cb, \%br2change_cb, \@constituent_blks, \@involved_seqs_cb, \@merger_types) is for the $cb th composite block. Here
#	@constituent_blks stores indices of the position-shift blocks that constitute the composite block;
#	
#	@hard_linking stores indices of the c-blocks that hard links $indx_cb;
#
#	@{$set_blocks_shift_respos[$bl]} = ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes) is for the $bl th position-shift block. Here
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block;
#
#	%node2ch = ($node_id => \@child_nodes, ...) for internal nodes;
#	%node2seq_indx = ($node_id => $indx_seq, ...) for external nodes;
#
# OUTPUT: (\@frwk_cb, \@info_constituents, \@blwise_hard_linking), where
#
#	$frwk_cb[$i]->{$indx_seq} = [$beg_frwk, $end_frwk], which determines the beginning- and ending-positions of the framework of the $i th constituent block in the $indx_seq th sequence (row),
#			 = undef if the $i th constituent block does not involve the $indx_seq th sequence.
#
#	@{$info_constituents[$i]} = ($shift, $beg_bl, $end_bl) is for the $i th constituent block of the $indx_cb th c-block;
#
#	@{$blwise_hard_linking[$i]} is an array, each of whose elements is [$indx_cb2, $bl2] for the block that is hard_linking the $i th constituent block.
#
sub framework_cblock ($$\@\@\@\%\%) {

	my ($ct_seqs, $indx_cb, $set_composite_blocks, $hard_linking, $set_blocks_shift_respos, $node2ch, $node2seq_indx) = @_;

	my ($beg_cb, $end_cb, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$set_composite_blocks->[$indx_cb]}; 

	my $ct_constituents = @{$constituent_blks};

		# First, construct the framework of the main c-block. #

	my @info_constituents = (); # element = [$shift, $beg_bl, $end_bl] for the constituent block. #
	my @frwk_cb = ();

	for (my $i=0; $i<$ct_constituents; $i++) { # 1st main outer for-loop (over constituent blocks). #
	    my $bl = $constituent_blks->[$i];

	    my ($shift, $beg_bl, $end_bl, $nodeid_mrca_bl, $depth_mrca_bl, $seqs_involved, $br2change_bl, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_blocks_shift_respos->[$bl]};
	    (defined $excl_sbblocks) or ($excl_sbblocks = []);

	    $info_constituents[$i] = [$shift, $beg_bl, $end_bl];

	    my $frwk_bl = blockwise_framework ($beg_bl, $end_bl, @{$seqs_involved}, @{$excl_sbblocks}, %{$node2ch}, %{$node2seq_indx});
	    unless (defined $frwk_bl) {
		print STDOUT "[ERROR 'framework_cblock'] The subroutine 'blockwise_framework' FAILED on the $i th constituent block (Index: $bl) of the input c-block: $! !\n\n";
		return (undef, undef, undef);
	    }

	    $frwk_cb[$i] = $frwk_bl;

	}				# End of the 1st main outer for-loop (over constituent blocks). #


		# Second, augument the framework with those of the hard-linking c-blocks. #

	my @blwise_hard_linking = (); # @{$blwise_hard_linking[$i]} is an array of [$indx_cb2, $bl2] for the blocks that hard-link the $i th constituent block. #
	for (1 .. $ct_constituents) { push @blwise_hard_linking, []; }

	my $ct_hard_linking = @{$hard_linking};

	for (my $i2 = 0; $i2 <$ct_hard_linking; $i2++) { # 2nd main outer foreach-loop (over c-blocks hard-linking the input c-block). #

	    my $indx_cb2 = $hard_linking->[$i2];

	    my ($beg_cb2, $end_cb2, $nodeid_mrca_cb2, $br2change_cb2, $constituent_blks2, $involved_seqs_cb2, $merger_types2) = @{$set_composite_blocks->[$indx_cb2]}; 

	    foreach my $bl2 (@{$constituent_blks2}) { # Middle foreach-loop (over constituent blocks of hard-linking). #

		my ($shift2, $beg_bl2, $end_bl2, $nodeid_mrca_bl2, $depth_mrca_bl2, $seqs_involved2, $br2change_bl2, $excl_sbblocks2, $ct_changes2, $indices_src_classes2) = @{$set_blocks_shift_respos->[$bl2]};

		my $indx_accom;
		for (my $i=0; $i<$ct_constituents; $i++) { # 1st inner foreach-loop (over constituent blocks of input). #

		    my ($shift, $beg_bl, $end_bl) = @{$info_constituents[$i]};

		    ($shift == $shift2) or next;
		    ##($shift and $shift2) or next; # A BUG fixed on Oct 24 (Sat), 2015. #

		    (($beg_bl <= $beg_bl2) and ($end_bl2 <= $end_bl)) or next;
		    $indx_accom = $i;
		    last;
		}				# End of the 1st inner foreach-loop (over constituent blocks of input). #

		unless (defined $indx_accom) { # There is no input block accommodating the hard-likning block... This must not happen!! #
		    print STDOUT "[ERROR 'framework_cblock'] The $bl2 th block of the $indx_cb2 th c-block, which should hard-link the input c-block ($indx_cb), is actually NOT accommodated by any block of the input c-block: $! !\n\n";
		    return (undef, undef, undef);
		}

		push @{$blwise_hard_linking[$indx_accom]}, [$indx_cb2, $bl2];

		my $frwk_bl2 = blockwise_framework ($beg_bl2, $end_bl2, @{$seqs_involved2}, @{$excl_sbblocks2}, %{$node2ch}, %{$node2seq_indx});
		unless (defined $frwk_bl2) {
		   print STDOUT "[ERROR 'framework_cblock'] The subroutine 'blockwise_framework' FAILED on the $bl2 th block, which is a constituent of the hard-linking block (index: $indx_cb2): $! !\n\n";
		   return (undef, undef, undef);
		}

		my $frwk_bl = $frwk_cb[$indx_accom];

		foreach my $indx_seq (keys %{$frwk_bl2}) { $frwk_bl->{$indx_seq} = $frwk_bl2->{$indx_seq}; }

	    }				# End of the middle foreach-loop (over constituent blocks of hard-linking). #

	}			# End of the 2nd main outer foreach-loop (over c-blocks hard-linking the input c-block). #

	return (\@frwk_cb, \@info_constituents, \@blwise_hard_linking);

}	# END of the subroutine 'framework_cblock'. #

# scout_possible_undoing:
#	This satellite subroutine of 'undo_mv_latter_of_cblock_pair' examines
#	the extent to which the "undoing" of the specified block can be performed.
#	[CREATED on Oct 12 (Mon),2015.]
#	[MODIFIED on Oct 13 (Tue), 2015.] The subroutine now outputs \@indices_seqs_incomplete_mvs as well.
#	[DEBUGGED on Oct 16 (Fri), 2015.]
#
# INPUT: ($shift_bl, \%frwk_bl, \@map_shifts_respos, $char_abs), where
#	$shift_bl is the position-shift of the specified block;
#	%frwk_bl = ($indx_seq => [$beg_fw, $end_fw], ...) gives the beginning- and ending-positions of the framework in the $indx_seq th sequence;
#	@map_shifts_respos gives the position-shift map immediately before the move is undone. More specifically,
#	$map_shifts_respos[$indx_seq]->[$s]  gives the position-shift of the site with the horozontal position = $s in the $indx_seq th row of the MSA;
#
# OUTPUT: ($actual_shift_bl, \@indices_seqs_incomplete_mvs), where
#	$actual_shift_bl is the shift that can be actually undone;
#	@indices_seqs_incomplete_mvs stores indices of the sequences that cannot be undone the moves completely.
#
sub scout_possible_undoing ($\%\@$) {

	my ($shift_bl, $frwk_bl, $map_shifts_respos, $char_abs) = @_;

	my $actual_shift_bl = $shift_bl;
	my @indices_seqs_incomplete_mvs = ();

	foreach my $indx_seq (sort {$a <=> $b} keys %{$frwk_bl}) { # Main foreach-loop (over sequences involved with the block). #

	    my ($beg_fw, $end_fw) = @{$frwk_bl->{$indx_seq}};
	    my $shifts_in_seq = $map_shifts_respos->[$indx_seq];

		    # MODIFIED for DEBUGGING (on Oct 16 (Fri), 2015) #
	    my ($shift_beg, $shift_end) = @{$shifts_in_seq}[$beg_fw, $end_fw];
	    unless (($shift_beg ne $char_abs) and ($shift_bl == $shift_beg) and ($shift_end ne $char_abs) and ($shift_bl == $shift_end)) {
	    ##unless (($shift_bl eq $shifts_in_seq->[$beg_fw]) and ($shift_bl eq $shifts_in_seq->[$end_fw])) {
		print STDOUT "[ERROR 'scout_possible_undoing'] NOT BOTH of the positions $beg_fw and $end_fw in the $indx_seq th sequence have the position-shift = $shift_bl: $! !\n\n";
		return undef;
	    }
	    ## ($actual_shift_bl == 0) and next; # The actual shift cannot change any more. ## DISABLED on Oct 16 (Fri), 2015. #

	    my $shift2 = 0;

	    if ($shift_bl>0) { # The block moves leftward if undone.

		for (my $s = $beg_fw-1; $s>= $beg_fw - $shift_bl; $s--) { # 1st inner for-loop (downwards over sites in the desired range). #
		##for (my $s = $beg_fw-1; $s>= $beg_fw-$actual_shift_bl; $s--) {...} # 1st inner for-loop (downwards over sites in the possible range). # MODIFIED on Oct 13 (Tue), 2015. #

		    (($s < 0) or ($shifts_in_seq->[$s] ne $char_abs)) and last; # Found an obstacle (or the left-end)!! 
		    ($shifts_in_seq->[$s] eq $char_abs) or last; # Another possible BUG fixed on Oct 16 (Fri), 2015. #
		    ##($shifts_in_seq->[$s] eq $char_abs) and last; # A BUG fixed on Oct 16 (Fri), 2015. #
		    $shift2++;
		}				# End of the 1st inner for-loop (downwards over sites in the possible range). #
		($actual_shift_bl > $shift2) and ($actual_shift_bl = $shift2);

	    } else { # The block moves rightward if undone. #

		for (my $s = $end_fw+1; $s<= $end_fw - $shift_bl; $s++) { # 2nd inner for-loop (upwards over sites in the desired range). #
		##for (my $s = $end_fw+1; $s<= $end_fw + $shift_bl; $s++) {...} # 2nd inner for-loop (upwards over sites in the desired range). # Another BUG fixed on Oct 16 (Fri), 2015. #
		##for (my $s = $end_fw+1; $s<= $end_fw+$actual_shift_bl; $s++) { # 2nd inner for-loop (upwards over sites in the possible range). # MODIFIED on Oct 13 (Tue), 2015. #

		    ((defined $shifts_in_seq->[$s]) and ($shifts_in_seq->[$s] ne $char_abs)) and last; # Found an obstacle!!
		    ##($shifts_in_seq->[$s] eq $char_abs) or last; # Another BUG fixed on Oct 16 (Fri), 2015. #
		    ##($shifts_in_seq->[$s] eq $char_abs) and last; # A BUG fixed on Oct 16 (Fri), 2015. #
		    $shift2--;
		}				# End of the 1st inner for-loop (upwards over sites in the possible range). #
		($actual_shift_bl < $shift2) and ($actual_shift_bl = $shift2);
	    }

	    ($shift2 == $shift_bl) or (push @indices_seqs_incomplete_mvs, $indx_seq);

	}			# End of the main foreach-loop (over sequences involved with the block). #

		# MODIFIED on Oct 13 (Tue), 2015. #
	return ($actual_shift_bl, \@indices_seqs_incomplete_mvs);
	##return $actual_shift_bl;

}	# End of the subroutine 'scout_possible_undoing'. #


# undo_sgl_bw_fw_move_on_2maps:
#	This satellite subroutine of 'undo_mv_latter_of_cblock_pair'
#	undoes the move of the framework of a single (elementary but not composite) position-shift block.
#	The results will be reflected on two position-shift maps,
#	one between the reference and the intermediate MSAs (on the latter),
#	and the other between the intermediate and the reconstructed MSAs (on the latter).
#
#	This subroutine allows for up to one excluded region (per sequence) within the framework.
#
#	[CREATED on Oct 13 (Tue), 2015.]
#
#	[MODOFIED on Oct 14 (Wed), 2015]
#	When the undone move extends the intermediate MSA to the right,
#	the subroutine now fills in the extended part of the intermediate MSA with gaps
#	if the sequence is not involved in the undone move.
#
# INPUT: (\%frwk_bl, $char_abs, $full_shift, $actual_shift, \@map_shifts_respos_intm, \@map_shifts_respos_rec_af), where
#
#	%frwk_bl = ($indx_seq (in MSA) => [$beg_fw, $end_fw], ...) defines the framework of the subjecet block (in the intermediate MSA before the move of this block is undone);
#
#	$char_abs is the character representing the absence of the residue. (It is the same as the 'gap' character);
#	$full_shift is the full value of the shift of the subject block between the reference and the reconstructed MSA;
#	$actual_shift is the value of the shift by which the move of the subject block is actually undone (from the reconstructed MSA to the intermediate MSA);
#
#	@map_shifts_respos_intm is the position-shift map between the reference and the intermediate MSAs. It is under construction.
#	@map_shifts_respos_rec_af is the position-shift map between the intermediate and the reconstructed MSAs. It is also under construction.
#	($map_shifts_respos_xxx[$indx_seq]->[$s] is the position-shift at the site of the $indx_seq th sequence with horozontal position = $s.);
#
# OUTPUT: $if_success, where
#	$if_success = 1 if the subroutine ended successfully,
#		    = 0 if it failed.
#
sub undo_sgl_bw_fw_move_on_2maps (\%$$$\@\@) {

	my ($frwk_bl, $char_abs, $full_shift, $actual_shift, $map_shifts_respos_intm, $map_shifts_respos_rec_af) = @_;

	my $remaining_shift = $full_shift - $actual_shift;

	my $end_msa_intm = my $end_msa_intm0 = @{$map_shifts_respos_intm->[0]} -1; # ADDED on Oct 14 (Wed), 2015. #

	foreach my $indx_seq (keys %{$frwk_bl}) { # Main foreach-loop (over sequences involved with the subject block). #

		my ($beg_fw_outer, $end_fw_outer) = @{$frwk_bl->{$indx_seq}};
		my $shifts_in_seq_intm   = $map_shifts_respos_intm->[$indx_seq];
		my $shifts_in_seq_rec_af = $map_shifts_respos_rec_af->[$indx_seq];

		($end_msa_intm < $end_fw_outer - $actual_shift) and ($end_msa_intm = $end_fw_outer - $actual_shift); # ADDED on Oct 14 (Wed), 2015. #

			# Determine the boundaries of the "inner" framework, if at all. #

		my $beg_fw_inner = $beg_fw_outer;
		for (my $s=$beg_fw_outer+1; $s<=$end_fw_outer; $s++) { # 1st inner for-loop (over sites within the framework). #

		    my $shift = $shifts_in_seq_intm->[$s];
		    ##my $shift = $shifts_in_seq_rec_af->[$s]; # A BUG fixed on Oct 16 (Fri), 2015. #

		    (($shift eq $char_abs) or ($shift != $full_shift)) and last;
		    $beg_fw_inner++;
		}				# End of the 1st inner for-loop (over sites within the framework). #

		my $end_fw_inner;
		if ($beg_fw_inner < $end_fw_outer) {
		    $end_fw_inner = $end_fw_outer;
		    for (my $s = $end_fw_outer-1; $s>$beg_fw_inner; $s--) { # 2nd inner for-loop (over sites within the framework minus the shell at the beginning). #

			my $shift = $shifts_in_seq_intm->[$s];
			##my $shift = $shifts_in_seq_rec_af->[$s]; # A BUG fixed on Oct 16 (Fri), 2015. #

			(($shift eq $char_abs) or ($shift != $full_shift)) and last;
			$end_fw_inner--;
		    }				# End of the 2nd inner for-loop (over sites within the framework minus the shell at the beginning). #
		}

			# Extert the actual change. #

		for (my $s = $beg_fw_outer; $s <= $beg_fw_inner; $s++) { # 3rd inner for-loop (over sites within the shell at the beginning). #
			$shifts_in_seq_intm->[$s-$actual_shift] = $remaining_shift;
			$shifts_in_seq_rec_af->[$s] = $actual_shift;

		}				# End of the 3rd inner for-loop (over sites within the shell at the beginning). #

		my @sites_new_gaps = ($actual_shift>0) ? ($beg_fw_inner - $actual_shift+1 .. $beg_fw_inner) : ($beg_fw_outer .. $beg_fw_outer - $actual_shift - 1);

		for my $s (@sites_new_gaps) { $shifts_in_seq_intm->[$s] = $char_abs; }

		if (defined $end_fw_inner) {
		    for (my $s = $end_fw_inner; $s <= $end_fw_outer; $s++) { # 6th inner for-loop (over sites within the shell at the end). #
			$shifts_in_seq_intm->[$s - $actual_shift] = $remaining_shift;
			$shifts_in_seq_rec_af->[$s] = $actual_shift;
		    }				# End of the 6th inner for-loop (over sites within the shell at the end). #

		    my @sites_new_gaps2 = ($actual_shift>0) ? ($end_fw_outer -$actual_shift+1 .. $end_fw_outer) : ($end_fw_inner .. $end_fw_inner - $actual_shift -1);
		    for my $s (@sites_new_gaps2) { $shifts_in_seq_intm->[$s] = $char_abs; }
		}
		
	}			# End of the main foreach-loop (over sequences involved with the subject block). #

	if ($end_msa_intm > $end_msa_intm0) { # ADDED on Oct 14 (Wed), 2015. #

			# Fill in the new yet unoccupied right-end with gaps. #

	    foreach my $shifts_row (@{$map_shifts_respos_intm}) { # Trailing foreach-loop (over sequences in the MSA). #
		my $ct_sites = @{$shifts_row};
		for (my $s = $ct_sites; $s <= $end_msa_intm; $s++) { $shifts_row->[$s] = $char_abs; }
	    }					# End of the trailing foreach-loop (over sequences in the MSA). #
	}

	return 1;

}		# End of the subroutine 'undo_sgl_bw_move_on_2maps'. #


# undo_sgl_bw_fw_move_on_2maps_2:
#	This subroutine is a modified version of 'undo_sgl_bw_fw_move_on_2maps'.
#	The major difference from the old version is in how to handle "inpurities", i.e., gaps and/or sites with different values of the shift.
#	Whereas the old version attemped to keep the inpurities intact (unmoved),
#	the new version moves them along with the framework, by changing the values of the shift by the same amount (= $actual_shift).
#	[CREATED in DEBUGGING on Oct 25 (Sun), 2015.]
#
# undo_sgl_bw_fw_move_on_2maps:
#	This satellite subroutine of 'undo_mv_latter_of_cblock_pair'
#	undoes the move of the framework of a single (elementary but not composite) position-shift block.
#	The results will be reflected on two position-shift maps,
#	one between the reference and the intermediate MSAs (on the latter),
#	and the other between the intermediate and the reconstructed MSAs (on the latter).
#
#	This subroutine allows for up to one excluded region (per sequence) within the framework.
#
#	[CREATED on Oct 13 (Tue), 2015.]
#
#	[MODOFIED on Oct 14 (Wed), 2015]
#	When the undone move extends the intermediate MSA to the right,
#	the subroutine now fills in the extended part of the intermediate MSA with gaps
#	if the sequence is not involved in the undone move.
#
# INPUT: (\%frwk_bl, $char_abs, $full_shift, $actual_shift, \@map_shifts_respos_intm, \@map_shifts_respos_rec_af), where
#
#	%frwk_bl = ($indx_seq (in MSA) => [$beg_fw, $end_fw], ...) defines the framework of the subjecet block (in the intermediate MSA before the move of this block is undone);
#
#	$char_abs is the character representing the absence of the residue. (It is the same as the 'gap' character);
#	$full_shift is the full value of the shift of the subject block between the reference and the reconstructed MSA;
#	$actual_shift is the value of the shift by which the move of the subject block is actually undone (from the reconstructed MSA to the intermediate MSA);
#
#	@map_shifts_respos_intm is the position-shift map between the reference and the intermediate MSAs. It is under construction.
#	@map_shifts_respos_rec_af is the position-shift map between the intermediate and the reconstructed MSAs. It is also under construction.
#	($map_shifts_respos_xxx[$indx_seq]->[$s] is the position-shift at the site of the $indx_seq th sequence with horozontal position = $s.);
#
# OUTPUT: $if_success, where
#	$if_success = 1 if the subroutine ended successfully,
#		    = 0 if it failed.
#
sub undo_sgl_bw_fw_move_on_2maps_2 (\%$$$\@\@) {

	my ($frwk_bl, $char_abs, $full_shift, $actual_shift, $map_shifts_respos_intm, $map_shifts_respos_rec_af) = @_;

	my $remaining_shift = $full_shift - $actual_shift;

	my $end_msa_intm = my $end_msa_intm0 = @{$map_shifts_respos_intm->[0]} -1; # ADDED on Oct 14 (Wed), 2015. #

	foreach my $indx_seq (keys %{$frwk_bl}) { # Main foreach-loop (over sequences involved with the subject block). #

		my ($beg_fw_outer, $end_fw_outer) = @{$frwk_bl->{$indx_seq}};
		my $shifts_in_seq_intm   = $map_shifts_respos_intm->[$indx_seq];
		my $shifts_in_seq_rec_af = $map_shifts_respos_rec_af->[$indx_seq];

		($end_msa_intm < $end_fw_outer - $actual_shift) and ($end_msa_intm = $end_fw_outer - $actual_shift); # ADDED on Oct 14 (Wed), 2015. #

			# Extert the actual change. #

		my @sites_old_shifts = ($beg_fw_outer .. $end_fw_outer);
	 	($actual_shift < 0) and (@sites_old_shifts = reverse (@sites_old_shifts));

		for my $s (@sites_old_shifts) { # 1st inner for-loop (over sites within the framework.) #

		    my $old_shift = $shifts_in_seq_intm->[$s];
		    $shifts_in_seq_intm->[$s-$actual_shift] = ($old_shift eq $char_abs) ? $char_abs : $old_shift - $actual_shift; 
		    $shifts_in_seq_rec_af->[$s] = ($old_shift eq $char_abs) ? $char_abs : $actual_shift;

		}			# End of the 1st inner for-loop (over sites within the framework.) #


		my @sites_new_gaps = ($actual_shift>0) ? ($end_fw_outer - $actual_shift + 1 .. $end_fw_outer) : ($beg_fw_outer .. $beg_fw_outer - $actual_shift - 1);

		for my $s (@sites_new_gaps) { # The 2nd inner for-loop (over sites to be filled with new gaps.) #

		    $shifts_in_seq_intm->[$s] = $char_abs;

		} # End of the 2nd inner for-loop (over sites to be filled with new gaps.) #

	}			# End of the main foreach-loop (over sequences involved with the subject block). #

	if ($end_msa_intm > $end_msa_intm0) { # ADDED on Oct 14 (Wed), 2015. #

			# Fill in the new yet unoccupied right-end with gaps. #

	    foreach my $shifts_row (@{$map_shifts_respos_intm}) { # Trailing foreach-loop (over sequences in the MSA). #
		my $ct_sites = @{$shifts_row};
		for (my $s = $ct_sites; $s <= $end_msa_intm; $s++) { $shifts_row->[$s] = $char_abs; }
	    }					# End of the trailing foreach-loop (over sequences in the MSA). #
	}

	return 1;

}		# End of the subroutine 'undo_sgl_bw_move_on_2maps_2'. #


# rm_zero_shifts_at_ends:
#	This satellite subroutine of 'undo_mve_latter_of_cblocks_pair' removes blocks with zero shift
#	from both ends of a subject c-block.
#	The subroutine modifies the input variables accordingly, instead of outputting the results of its intended process.
#	[CREATED on Oct 18 (Sun), 2015.]
#
# INPUT: (\@set_blocks_shift_respos, $indx_shift, \@constituent_blks, \@cp_constituent_blks, \@info_constituents, \@frwks_cb, \@merger_types, \@cp_merger_types, \@bw_hard_linking, $indx_cb, \$ct_constituents, \@cbs_hard_linking), where
#
#	$set_blocks_shift_respos[$bl]->[$indx_shift] gives the shift of the $bl th block in the set;
#
#	@constituent_blks stores the indices (in @set_blocks_shift_respos) of the blocks that constitute the subject c-block,
#	and @cp_constituent_blks is its copy;
#
#	$info_constituents[$i] and $frwks_cb[$i] give some information and the framework, respectively, of the $i th block in @constituent_blks;
#
#	$merger_types[$i] gives the type of the merger between the $i th and the ($i+1) th constituent blocks,
#	and @cp_merger_types is a copy of @merger_types;
#
#	[ $bw_hard_linking[$i]->[$indx_cb] gives the index of the c-block accommodating the block that hard-links the $i th constituent block; ] --- It was WRONG!!
#	$bw_hard_linking[$i]->[$k][$indx_cb] gives the index of the c-block accommodating the $k th block that hard-links the $i th constituent block. (@{$bw_hard_linking[$i]} is empty if the $i th constituent block is NOT hard-linked by any block);

#
#	$ct_constituents is the number of constituents;
#
#	@cbs_hard_linking stores the indices of the c-blocks that hard-link the subject c-block.
#
# OUTPUT: 1 if the subroutine ended successfully,
#	  0 if it failed.
#
sub rm_zero_shifts_at_ends (\@$\@\@\@\@\@\@\@$\$\@) {

	my ($set_blocks_shift_respos, $indx_shift, $constituent_blks, $cp_constituent_blks, $info_constituents,
	   $frwks_cb, $merger_types, $cp_merger_types, $bw_hard_linking, $indx_cb, $rf_ct_constituents, $cbs_hard_linking) = @_;

	my @list_manipulated = ($constituent_blks, $cp_constituent_blks, $info_constituents, $frwks_cb, $merger_types, $cp_merger_types, $bw_hard_linking);

		# First, remove the blocks with zero shift from the left-end of the subject c-block. #

	while (@{$constituent_blks}>0) { # 1st main while-loop (as long as the c-block is non-empty). #

	    my $bl_le = $constituent_blks->[0];
	    my $shift_le = $set_blocks_shift_respos->[$bl_le][$indx_shift];
	    ($shift_le != 0) and last;

	    foreach my $manipulated (@list_manipulated) { shift @{$manipulated}; }

	}			# End of the 1st main while-loop (as long as the c-block is non-empty). #


		# Second, remove the blocks with zero shift from the right-end of the subject c-block. #

	while (@{$constituent_blks}>0) { # 2nd main while-loop (as long as the c-block is non-empty). #

	    my $bl_re = $constituent_blks->[$#{$constituent_blks}];
	    my $shift_re = $set_blocks_shift_respos->[$bl_re][$indx_shift];
	    ($shift_re != 0) and last;

	    foreach my $manipulated (@list_manipulated) { pop @{$manipulated}; }

	}			# End of the 2nd main while-loop (as long as the c-block is non-empty). #

	if (@{$constituent_blks} == ${$rf_ct_constituents}) {

	   return 1; # If no modification was made, end the subroutine successfully.
	}

		# Reflect the modification to the remaining variables. #

	${$rf_ct_constituents} = scalar (@{$constituent_blks}); # Update #{constituents}. #

	my @new_cbs_hard_linking = ();
	my %already; # = ($cb => 1, ...) for the c-blocks already appeared in @{$bw_hard_linking}.

		# An WRONG single-foreach-loop was CORRECTED to the DOUBLE-foreach-loops. #
	foreach my $hard_linking_to_bl (@{$bw_hard_linking}) { # 3rd main outer foreach-loop (over constituent blocks). #
	    (defined $hard_linking_to_bl) or next;
	    foreach my $info_hl_bl (@{$hard_linking_to_bl}) { # Inner foreach-loop (over hard-linking blocks). #
	##foreach my $info_hl_bl (@{$bw_hard_linking}) { # 3rd main foreach-loop (over hard-linking c-blocks). # # This was WRONG!!

		my $cb = $info_hl_bl->[$indx_cb];
		(defined $already{$cb}) and next;

		push @new_cbs_hard_linking, $cb;
		$already{$cb} = 1;
	##}				# End of the 3rd main foreach-loop (over hard-linking c-blocks). # # This was WRONG!!
	    }					 # End of the inner foreach-loop (over hard-linking blocks). #
	}				# End of the 3rd main outer foreach-loop (over constituent blocks). #

	@{$cbs_hard_linking} = @new_cbs_hard_linking; # Update the list of hard-likning c-blocks. #

	return 1; # End the subroutine successfully. #

}	# End of the subroutine 'rm_zero_shifts_at_ends'. #


# sub_enum_end_sequences:
#	This satellite subroutine of 'enum_end_sequences' enumerates sequences involved with a single c-block
#	at a specified pair of end-points.
#	[CREATED on Oct 10 (Sat), 2015]
#
# INPUT: ($beg, $end, $indx_cblk, \@set_composite_blocks, \@set_shift_blocks, \%$node2ch, \%node2seq_indx), where
#
#	$beg/$end is the specified left/right-end points;
#
#	@{$set_composite_blocks[$indx_cblk]} =  ($beg_cb, $end_cb, $nodeid_mrca_cb, \%br2change_cb, \@constituent_blks, \@involved_seqs_cb, \@merger_types) is for the $indx_cblk th composite-block of residue-position shifts;
#
#	@{$set_shift_blocks[$bl]} = ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes) is for the $bl th position-shift block, where
#	$beg/end_blk is the beginning/ending site (0-based full-closed) of the block,
#	@seqs_involved stores the (0-based) indices of the sequences involved with the block,
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
#	$ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
#	 @indices_src_classes stores the indices of the classes that are used to construct the block;
#
#	%node2ch = ($node_id => \@child_nodes, ...) for internal nodes;
#	%node2seq_indx = ($node_id => $indx_seq (in the MSA), ...) for external nodes.
#
# OUTPUT: (\@seqs_left_end, \@seqs_right_end), where
#	@seqs_left/right_end stores the indices of sequences involved in the subject c-block
#	at the block's left/right-end. (Each list may be redundant.)
#
sub sub_enum_end_sequences ($$$\@\@\%\%) { 

	my ($beg, $end, $indx_cblk, $set_composite_blocks, $set_shift_blocks, $node2ch, $node2seq_indx) = @_;

	my ($beg_cb, $end_cb, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$set_composite_blocks->[$indx_cblk]};

	my @seqs_left_end = ();

    if ($beg_cb <= $beg) {
	foreach my $bl (@{$constituent_blks}) { #  1st main foreach-loop (over constituent shift-blocks). #

 	    my ($shift, $beg_bl, $end_bl, $node_mrca_bl, $depth_mrca_bl, $involved_seqs_bl, $br2change_bl, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_shift_blocks->[$bl]};

	    ($beg_bl == $beg) or next;

	    my @excl_seqs = ();

	    if ((defined $excl_sbblocks) and (@{$excl_sbblocks}>0)) {

		foreach my $info_sbblk (@{$excl_sbblocks}) { # Inner foreach-loop (over excluded sub-blocks). #

		    my ($beg_sbbl, $end_sbbl, $mrca_sbbl) = @{$info_sbblk};

		    ($beg_sbbl == $beg) or next;

		    my $excl_seqs = fetch_ext_offsprings_nr ($mrca_sbbl, %{$node2ch});

		    foreach my $nodeid (@{$excl_seqs}) { push @excl_seqs, $node2seq_indx->{$nodeid}; }

		}				# End of the inner foreach-loop (over excluded sub-blocks). #
	    }

	    my ($excluded, $truly_involved, $extra_excl) = (@excl_seqs>0) ?
			diff_nmarray (@{$involved_seqs_bl}, @excl_seqs) : ([], $involved_seqs_bl, []);

	    push @seqs_left_end, @{$truly_involved};

	}				# End of the 1st main foreach-loop (over constituent shift-blocks). #

    } # END of "if ($beg_cb <= $beg) {...}"


	my @seqs_right_end = ();

    if ($end <= $end_cb) {

	foreach my $bl (@{$constituent_blks}) { #  2nd main foreach-loop (over constituent shift-blocks). #

 	    my ($shift, $beg_bl, $end_bl, $node_mrca_bl, $depth_mrca_bl, $involved_seqs_bl, $br2change_bl, $excl_sbblocks, $ct_changes, $indices_src_classes) = @{$set_shift_blocks->[$bl]};

	    ($end_bl == $end) or next;

	    my @excl_seqs = ();

	    if ((defined $excl_sbblocks) and (@{$excl_sbblocks}>0)) {

		foreach my $info_sbblk (@{$excl_sbblocks}) { # Inner foreach-loop (over excluded sub-blocks). #

		    my ($beg_sbbl, $end_sbbl, $mrca_sbbl) = @{$info_sbblk};

		    ($end_sbbl == $end) or next;

		    my $excl_seqs = fetch_ext_offsprings_nr ($mrca_sbbl, %{$node2ch});

		    foreach my $nodeid (@{$excl_seqs}) { push @excl_seqs, $node2seq_indx->{$nodeid}; }

		}				# End of the inner foreach-loop (over excluded sub-blocks). #
	    }

	    my ($excluded, $truly_involved, $extra_excl) = (@excl_seqs>0) ?
			diff_nmarray (@{$involved_seqs_bl}, @excl_seqs) : ([], $involved_seqs_bl, []);

	    push @seqs_right_end, @{$truly_involved};

	}				# End of the 2nd main foreach-loop (over constituent shift-blocks). #

    } # END of "if ($end <= $end_cb) {...}"


	return (\@seqs_left_end, \@seqs_right_end);

}	# End of the subroutine 'sub_enum_end_sequences'. #


# enum_end_sequences:
#	This subroutine enumerates sequences at the left-end and the right-end of a composite-block of residue-position shifts.
#	[CREATION STARTED on Oct 9 (Fri), 2015, ENDED on Oct 10 (Sat), 2015.]
#
# INPUT: ($indx_cblk, \@set_composite_blocks, \@set_shift_blocks, \%cb2hard_linking, \%node2ch, \%node2seq_indx), where
#
#	@{$set_composite_blocks[$indx_cblk]} =  ($beg_cb, $end_cb, $nodeid_mrca_cb, \%br2change_cb, \@constituent_blks, \@involved_seqs_cb, \@merger_types) is for the $indx_cblk th composite-block of residue-position shifts;
#
#	@{$set_shift_blocks[$bl]} = ($shift, $beg_blk, $end_blk, $id_node_mrca, $depth_mrca, \@seqs_involved, \%br2change, \@excl_sbblocks, $ct_changes, \@indices_src_classes) is for the $bl th position-shift block, where
#	$beg/end_blk is the beginning/ending site (0-based full-closed) of the block,
#	@seqs_involved stores the (0-based) indices of the sequences involved with the block,
#	%br2change = ($br_id => [$stat_upper_end, $stat_lower_end], ...),
#	@{$excl_sbblocks[$sb]} = ($beg_site, $end_site, $mrca_node) is for the $sb th sub-blocks excluded from the block, with $mrca_node is the "Most Recent Common Ancestor" node of the sequences making up the sub-block, and
#	$ct_changes is the number of "changes" (including the exclusion of sub-blocks) necessary for separating the involved sequences,
#	 @indices_src_classes stores the indices of the classes that are used to construct the block;
#
#	%cb2hard_linking = ($indx_cb (for non-skipped) => [list of indices of (skipped and complex) composite-blocks that "hard-link" the $cb], ...);
#
#	%node2ch = ($node_id => \@child_nodes, ...) for internal nodes;
#	%node2seq_indx = ($node_id => $indx_seq (in the MSA), ...) for external nodes.
#
# OUTPUT: (\@seqs_left_end, \@seqs_right_end), where
#	@seqs_left/right_end stores the indices of sequences involved in the subject c-block
#	at the block's left/right-end.
#
sub enum_end_sequences ($\@\@\%\%\%) {

	my ($indx_cblk, $set_composite_blocks, $set_shift_blocks, $cb2hard_linking, $node2ch, $node2seq_indx) = @_;

	my ($beg_cb, $end_cb, $nodeid_mrca_cb, $br2change_cb, $constituent_blks, $involved_seqs_cb, $merger_types) = @{$set_composite_blocks->[$indx_cblk]};

	my ($seqs_left_end, $seqs_right_end) = sub_enum_end_sequences ($beg_cb, $end_cb, $indx_cblk, @{$set_composite_blocks}, @{$set_shift_blocks}, %{$node2ch}, %{$node2seq_indx});

	unless (defined $seqs_left_end) {
	    print STDOUT "[ERROR 'enum_end_sequences'] Satellite subroutine 'sub_enum_end_sequences' FAILED on the $indx_cblk th c-block: $! !\n\n";
	    return (undef, undef);
	}

	my $hard_linking = $cb2hard_linking->{$indx_cblk};

	if (defined $hard_linking) {
	    foreach my $cb2 (@{$hard_linking}) { # Foreach-loop (over c-blocks hard-linking $indx_cblk). #

		my ($seqs_left_end2, $seqs_right_end2) = sub_enum_end_sequences ($beg_cb, $end_cb, $cb2, @{$set_composite_blocks}, @{$set_shift_blocks}, %{$node2ch}, %{$node2seq_indx});
		unless (defined $seqs_left_end2) {
		    print STDOUT "[ERROR 'enum_end_sequences'] Satellite subroutine 'sub_enum_end_sequences' FAILED on the $cb2 th c-block (, which hard-links the $indx_cblk th c-block): $! !\n\n";
		    return (undef, undef);
		}

		(@{$seqs_left_end2}>0)  and (push @{$seqs_left_end}, @{$seqs_left_end2});
		(@{$seqs_right_end2}>0) and (push @{$seqs_right_end}, @{$seqs_right_end2});

	    }				# End of the foreach-loop (over c-blocks hard-linking $indx_cblk). #

	} # END of "if (defined $hard_linking) {...}"

	my $nr_seqs_left_end = nr_nmarray (@{$seqs_left_end});
	my $nr_seqs_right_end = nr_nmarray (@{$seqs_right_end});

	return ($nr_seqs_left_end, $nr_seqs_right_end);

}	# End of the subroutine 'enum_end_sequences'. #


# reorganize_set_indels:
#	This subroutine receives a raw set of gap-pattern blocks and a candidate of the parsimonious indel scenario(s) for a MSA,
#	and returns a set of them reorganized to ease the search for a set of indel events with a particular identifier (either branch ID, beginning site, or ending site).
#	(CREATED on Oct 24 (Fri), 2014.)
#
#	[MODIFIED on Nov 4 (Tue), 2014.]
#	The output format somewhat changed. Now, the information on each indel event includes the length of the event.
#
# INPUT: (\@set_cls_gblocks, \@cand_psm_scn_for_cls), where
#
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the MSA, with
#	$indx_gp specifying the gap-pattern (stored in @set_gpatterns),
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);
#
#	@{$cand_psm_scn_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the MSA, where 
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices (for @{$set_cls_gblocks[$cl]}) of blocks involved), and where
#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#	  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node.
#
# OUTPUT: (\@linset_gblocks, \%br2all_indels, \@set_indels_std_by_begsite, \@set_indels_std_by_endsite), where
#
#	@{$linset_gblocks[$lgb]} = ($cl, $gb, $indx_gp, $len_gb, $beg_gb, $end_gb) is for the gap-pattern block in the $lgb th place in the linear set, which is the $gb th block in the $cl th cluster (i.e., gapped segment) in the MSA, where
#	$indx_gp is the index of the gap-pattern in the @{$set_gpatterns}, and $beg/end_gb is the beginning/ending coordinate of the block (in the 0-based full-closed convention);
#
#	%br2all_indels = ($branch_id => \@set_all_indels, ...), where
#	[ @{$set_all_indels[$e]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk) is for the $e th (in the spatial order) indel event along the branch, with ] --- OBSOLETE as of Nov 4 (Tue), 2014.
#	@{$set_all_indels[$e]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, $length_indel, \@indices_lin_blk) is for the $e th (in the spatial order) indel event along the branch, with # MODIFIED on Nov 4 (Tue), 2014.
#	$beg/end_site being the coordinate (in the 0-based full-closed convention) of the beginning/ending-site of the event, and
#	@indices_lin_blk storing the indices (in @linset_cls_gblocks}) of the involved gap-pattern blocks;
#
#	@{$set_indels_std_by_begsite[$i]} = ($beg_site, \@set_indels_beg_at_here) is for the $i th beginning site (in numerically ascending order), with
#	@{$set_indels_beg_at_here[$e1]} = (exactly in the same format as @{$set_all_indels[$e]}) is for the $e1 th (in the order of occurrences) indel event beginning at $beg_site;
#	@{$set_indels_std_by_endsite[$i]} = ($end_site, \@set_indels_end_at_here) is for the $i th ending site (in numerically ascending order), with
#	@{$set_indels_end_at_here[$e2]} = (exactly in the same format as @{$set_all_indels[$e]}) is for the $e2 th (in the order of occurrences) indel event ending at $end_site.
#
sub reorganize_set_indels (\@\@) {

	my ($set_cls_gblocks, $cand_psm_scn_for_cls) = @_;

	my $ct_cls = @{$set_cls_gblocks};

	my @linset_gblocks = my @set_indels_std_by_begsite = my @set_indels_std_by_endsite = ();
	my %br2all_indels;

	my $baseline = 0;

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Main outer for-loop (over clusters (i.e., gapped segments)). #

	    my $gblocks_in_cls = $set_cls_gblocks->[$cl];

	    my $ct_gblks = @{$gblocks_in_cls};

	    for (my $gb=0; $gb<$ct_gblks; $gb++) { # 1st main middle for-loop (over blocks). #

		my ($indx_gp, $len_gb, $beg_gb, $end_gb) = @{$gblocks_in_cls->[$gb]};
		$linset_gblocks[$baseline + $gb] = [$cl, $gb, $indx_gp, $len_gb, $beg_gb, $end_gb];

	    }				# End of the 1st main middle for-loop (over blocks). #

	    my %begsite2indels_in_cls;
	    my %endsite2indels_in_cls;

	    my $br2indels_in_cls = $cand_psm_scn_for_cls->[$cl][0];

	    foreach my $br (sort {$b <=> $a} keys %{$br2indels_in_cls}) { # 2nd main middle foreach-loop (over branches with indels). #

		my $list_indels = $br2indels_in_cls->{$br};
		my @lin_sblist_indels = ();

		foreach my $info_event (@{$list_indels}) { # Main inner foreach-loop (over events). #

		    my ($state_ue, $state_le, $indices_blk) = @{$info_event};

		    my ($indx_lm, $indx_rm) = @{$indices_blk}[0, $#{$indices_blk}];

		    my $beg_site = $gblocks_in_cls->[$indx_lm][2];
		    my $end_site = $gblocks_in_cls->[$indx_rm][3];

		    my @indices_lin_blk = ();

				# MODIFIED on Nov 4 (Tue), 2014. #

		    my $len_indel=0;
		    foreach my $indx (@{$indices_blk}) { # Inner-inner foreach-loop (over blocks involved in the event). #
			push @indices_lin_blk, $indx + $baseline;
			$len_indel += $gblocks_in_cls->[$indx][1];
		    }				# End of the inner-inner foreach-loop (over blocks involved in the event). #

		    my @lin_info_event = ($br, $beg_site, $end_site, $state_ue, $state_le, $len_indel, \@indices_lin_blk);

		    ##foreach my $indx (@{$indices_blk}) { push @indices_lin_blk, $indx + $baseline; }
		    ##my @lin_info_event = ($br, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk);

		    push @lin_sblist_indels, \@lin_info_event;

		    my $indels_in_cls1 = $begsite2indels_in_cls{$beg_site};
		    (defined $indels_in_cls1) or ($indels_in_cls1 = $begsite2indels_in_cls{$beg_site} = []);
		    push @{$indels_in_cls1}, \@lin_info_event;

		    my $indels_in_cls2 = $endsite2indels_in_cls{$end_site};
		    (defined $indels_in_cls2) or ($indels_in_cls2 = $endsite2indels_in_cls{$end_site} = []);
		    push @{$indels_in_cls2}, \@lin_info_event;

		}				# End of the main inner foreach-loop (over events). #

		my $all_indels = $br2all_indels{$br};
		(defined $all_indels) or ($all_indels = $br2all_indels{$br} = []);
		push @{$all_indels}, (sort { $a->[1] <=> $b->[1] } @lin_sblist_indels);

	    }						# End of the 2nd main middle foreach-loop (over branches with indels). #

	    foreach my $beg_site (sort {$a <=> $b} keys %begsite2indels_in_cls) { push @set_indels_std_by_begsite, [$beg_site, $begsite2indels_in_cls{$beg_site} ]; }

	    foreach my $end_site (sort {$a <=> $b} keys %endsite2indels_in_cls) { push @set_indels_std_by_endsite, [$end_site, $endsite2indels_in_cls{$end_site} ]; }


	    $baseline += $ct_gblks; # UPDATE the baseline. #

	}				# End of the main outer for-loop (over clusters (i.e., gapped segments)). #


	return (\@linset_gblocks, \%br2all_indels, \@set_indels_std_by_begsite, \@set_indels_std_by_endsite);

}	# End of the subroutine 'reorganize_set_indels'. #


# identify_equiv_indels:
#	This subroutine receives sets of indels resulting in two sets of MSAs (MSA 1 and 2),
#	and identifies the indel events for MSA 1 that have "equivalent" events for MSA 2.
#	(CREATED on Oct 28 (Tue), 2014)
#
#	[CRITERION for the "equivalence"]
#	The two events are considered as "equivalent" if they insert the identical set of sub-sequences (in terms of the residue numbers) (for equivalent insertions),
#		[or if the sets of "deleted" sub-sequences (in the complement of the deleted sequences) are identical (for equivalent deletions). ] --- Obsolete.
#	   or if the sets of "deleted" sub-sequences (in the sibling lineage, and, if any, in the uncle/aunt lineage, of the deleted lineage) are identical (for equivalent deletions).
#	The subroutine doesn't care about the positions they are inserted/deleted, i.e., the positions of the aligned gaps in the remaining sequences.
#
# INPUT: (\%br2all_indels1, \%br2all_indels2, \@set_clminfo1, \@set_clminfo2, $char_abs, $ct_seqs, \%node2ch, \%node2seq_indx, \%node2pa), where
#	%br2all_indels1/2 = ($branch_id => \@set_all_indels1/2, ...) is for the MSA 1/2, where
#	@{$set_all_indels1/2[$e]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk) is for the $e th (in the spatial order) indel event along the branch, with
#	$beg/end_site being the (0-based full-closed) coordinate of the beginning/ending site of the event in the MSA,
#	$state_ue/le being the state at the upper-/lower-end of the branch;
#
#	$set_clminfo1/2[$c]->[$k] is the residue number (1-based) of the $k th sequence at the $c th column of MSA 1/2.
#		it is ($r + 1/2) if the column is occupied by a gap between the $r th and the ($r+1) th residues.
#
#	$char_abs is the character representing the 'absence' state;
#
#	$ct_seqs is the number of sequences in the MSA (the two MSAs should be composed of the same sequences);
#
#	%node2ch = ($node_id => \@chilld_nodes, ...) is for the interior nodes;
#
#	%node2seq_indx = ($node_id => $seq_indx, ...) for the exterior nodes, with
#	$seq_indx specifying the index specifying the order of the sequence (or its corresponding exterior node) in the MSA;
#
#	%node2pa = ($node_id => $parent_id, ...) is for all nodes except the root.
#
# OUTPUT: (\%event1_2_event2, \%event2_2_event1), where
#
#	%event1_2_event2 = ("$br1:$beg1:$end1" => "$br2:$beg2:$end2", ...) with
#	$br1/2 being the branch along which the equivalent event in MSA 1/2 occurred,
#	$beg1/2 being the beginning site coordinate of the equivalent event in MSA 1/2,
#	$end1/2 beging the end site coordinate of the  equivalent event in MSA 1/2;
#
#	%event2_2_event1 is just the inverse of %event2_2_event1.
#
sub identify_equiv_indels (\%\%\@\@$$\%\%\%) {

	my ($br2all_indels1, $br2all_indels2, $set_clminfo1, $set_clminfo2, $char_abs, $ct_seqs, $node2ch, $node2seq_indx, $node2pa) = @_;

	my @indices_full_seqs = (0 .. $ct_seqs-1); 
	##my @indices_full_seqs = (0 .. $ct_seqs); # A BUG fixed (on Oct 28 (Tue), 2014). #

	my %event1_2_event2;
	my %event2_2_event1;

	foreach my $br (sort {$a <=> $b} keys %{$br2all_indels1}) { # Main outer foreach-loop (over branches where indels occur). #

	    my $all_indels1 = $br2all_indels1->{$br};
	   (@{$all_indels1}>0) or next;

	    my $all_indels2 = $br2all_indels2->{$br};
	   ((defined $all_indels2) and (@{$all_indels2}>0)) or next; # In this case, there can NEVER be equivalent events along this branch. #

	    my $exoffs = fetch_ext_offsprings_nr ($br, %{$node2ch});
	    my @offs_seq_indices = ();
	    foreach my $nodeid (@{$exoffs}) { push @offs_seq_indices, $node2seq_indx->{$nodeid}; }

	    @offs_seq_indices = sort {$a <=> $b} @offs_seq_indices;
	    my $cmpl_seq_indices;

	    foreach my $indel1 (@{$all_indels1}) { # Main middle foreach-loop (over indels in MSA 1 along $br). #

		my ($br1, $beg1, $end1, $state_ue1, $state_le1, $indices_blk1) = @{$indel1};

		my $label_event1 = "$br1:$beg1:$end1";

		my $resnums_beg1 = $set_clminfo1->[$beg1];
		my $resnums_end1 = $set_clminfo1->[$end1];

		my $relv_seq_indices;
		if ($state_le1 eq $char_abs) { # For a 'deletion'.

		    unless (defined $cmpl_seq_indices) {
	    		##(my $common, my $ex1, $cmpl_seq_indices) = diff_nmarray (@offs_seq_indices, @indices_full_seqs);
			my $node_pa = $node2pa->{$br};
			my $sibl;
			foreach my $node (@{$node2ch->{$node_pa}}) { # 1st preliminary inner foreach-loop (over siblings). #
			    if ($node != $br) { $sibl = $node; last; }
			} 					# End of the 1st preliminary inner foreach-loop (over siblings). #
			my $exoffs_sibl = fetch_ext_offsprings_nr ($sibl, %{$node2ch});
			my @cmpl_seq_indices = ();
			foreach my $nodeid (@{$exoffs_sibl}) { push @cmpl_seq_indices, $node2seq_indx->{$nodeid}; }
			my $node_grpa = $node2pa->{$node_pa};
			if (defined $node_grpa) {
			     my $uncle;
			     foreach my $node (@{$node2ch->{$node_grpa}}) { # 2nd preliminary inner foreach-loop (over siblings). #
				if ($node != $node_pa) { $uncle = $node; last; }
			     }					# End of the 2nd preliminary inner foreach-loop (over siblings). #
			     my $exoffs_uncle = fetch_ext_offsprings_nr ($uncle, %{$node2ch});
			     foreach my $nodeid (@{$exoffs_uncle}) { push @cmpl_seq_indices, $node2seq_indx->{$nodeid}; }
			}
			@cmpl_seq_indices = sort { $a <=> $b } @cmpl_seq_indices;
			$cmpl_seq_indices = \@cmpl_seq_indices;
		    }

		    $relv_seq_indices = $cmpl_seq_indices;

		} else {		# For an 'insertion'.

		    $relv_seq_indices = \@offs_seq_indices;
		}

		foreach my $indel2 (@{$all_indels2}) { # Main inner foreach-loop (over indels in MSA 2 along $br). #

		    my ($br2, $beg2, $end2, $state_ue2, $state_le2, $indices_blk2) = @{$indel2};

		    my $resnums_beg2 = $set_clminfo2->[$beg2];
		    my $resnums_end2 = $set_clminfo2->[$end2];

		    my $if_equiv = 1;
		    foreach my $seq_indx (@{$relv_seq_indices}) { # Innermost foreach-loop (over relevant sequences). #

			my $resnum_end1 = int ($resnums_end1->[$seq_indx]);
			my $resnum_end2 = int ($resnums_end2->[$seq_indx]);

			if ($resnum_end1 != $resnum_end2) {
			   $if_equiv = 0;
			   last;
			}

			my $resnum_beg1 = $resnums_beg1->[$seq_indx];
			($resnum_beg1 == int ($resnum_beg1)) or ($resnum_beg1 = int ($resnum_beg1) + 1 );
			my $resnum_beg2 = $resnums_beg2->[$seq_indx];
			($resnum_beg2 == int ($resnum_beg2)) or ($resnum_beg2 = int ($resnum_beg2) + 1 );

			if ($resnum_beg1 != $resnum_beg2) {
			    $if_equiv = 0;
			    last;
			}
		    }				# End of the innermost foreach-loop (over relevant sequences). #

		    if ($if_equiv) {
				# The two events were found to be "equivalent" to each other. #

			my $label_event2 = "$br2:$beg2:$end2";
			$event1_2_event2{$label_event1} = $label_event2;
			$event2_2_event1{$label_event2} = $label_event1;
			last;
		    }

		}				# End of the main inner foreach-loop (over indels in MSA 2 along $br). #

	    }				# End of the main middle foreach-loop (over indeld in MSA 1 along $br). #


	}				# End of the main outer foreach-loop (over branches where indels occur). #

	return (\%event1_2_event2, \%event2_2_event1);

}		# End of the subroutine 'identify_equiv_indels'. #


# effective_distance:
#	This satellite subroutine of 'find_ovlp_flanking_indels'
#	calculates the 'effective distance' between an indel event and a set of 'indel' events.
#	(CREATED on Oct 30 (Thu), 2014)
#
#	[DEFINITION]
#	The 'effective distance' is defined as the number of branches separating the two sets of nodes with the 'presence' state,
#	one set is defined by the indel event, and the other is defined by the set of 'indel' events.
#	If the two sets of nodes with the 'present' state overlap at some nodes, the 'effective distance' = 0.
#	If the two sets are complementary to each other (and both are singly-connected), the 'effective distance' = 1.
#	If, for example, the two sets resulted from two independent insertions along sibling branches,
#	the 'effective distance' = 2 usually, but = 1 if the sibling branches are children of the bifurcated top-node (i.e., root).
#	
# INPUT: ($br1, $state_ue1, $nodeid_mrca2, \%br2change2, \%node2pa, \%node2ch, $topnode_id, $char_abs), where
#
#	$br1 is the branch along which the 1st event occurred;
#	$state_ue1 is the state at the upper-end of $br1;
#
#	$nodeid_mrca2 is the MRCA of the nodes with the 'presence' state in the 2nd set;
#	%br2change2 = ($branch_id => [$state_ue, $state_le], ...) is for the changes ('indel' events) defining the 2nd set;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but the top-node (i.e., root);
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#	$topnode_id is the ID of the top-node (i.e., root);
#
#	$char_abs is the character representing a gap.
#
# OUTPUT: $eff_dist, which is the effective distance calculated.
#
sub effective_distance ($$$\%\%\%$$) {

	my ($br1, $state_ue1, $nodeid_mrca2, $br2change2, $node2pa, $node2ch, $topnode_id, $char_abs) = @_;

	if ($state_ue1 ne $char_abs) { # The 1st event is a deletion.

	    if ($nodeid_mrca2 == $topnode_id) {
		    # The two sets will certainly overlap. #
		return 0;
	    } elsif ($nodeid_mrca2 == $br1) {
		    # The 2nd set resulted from an 'insertion' along $br1. #
		return 1;
	    }
	} else { # The 1st event is an insertion.

	    if ($nodeid_mrca2 == $br1) {
		# The 2nd set resulted from an 'insertion' along $br1. => The two set will certainly overlap. #
		return 0;
	    }
	}

	my ($common_ancestors, $lineage1, $lineage2) = fetch_common_ancestors ($br1, $nodeid_mrca2, %{$node2pa});

	if ($state_ue1 ne $char_abs) { ## Again, the 1st event is a deletion.##

	    if (@{$lineage1}>0) {
		    # MRCA2 is also an ancestor of $br1. => The two set will overlap.
		return 0;
	    } else {
		   # $br1 is an ancestor of MRCA2. => The two sets will be separated by a series of branches leading to MRCA2.

		my $eff_dist = 1 + scalar (@{$lineage2});
		return $eff_dist;
	    }
	}

		## The 1st event is an insertion. ##

	if (@{$lineage1} == 0) {
		# MRCA2 is a 'descendant' of $br1. => The two sets will certainly overlap!!
	    return 0;
	}

	if (@{$lineage2}>0) {
		# MRCA2 and $br1 are on separate lineages. #
	    my $eff_dist = scalar (@{$lineage1}) + scalar (@{$lineage2});

	    if ((@{$common_ancestors}==1) and (@{$node2ch->{$topnode_id}}==2)) {
			# When the MRCA of MRCA2 and $br1 is the top-node with only 2 children, the two branches at the top will be regarded as a single branch. #
		$eff_dist--;
	    }

	    return $eff_dist;
	}

		# MRCA2 is an ancestor of $br1. #

	my $ct_lin1 = @{$lineage1};

	for (my $k=0; $k<$ct_lin1; $k++) { # Main for-loop (over descendant lineage of MRCA2 down to $br1). #
	    (defined $br2change2->{$lineage1->[0]}) and last; # A 'deletion' occurred. => The remaining branches separate the two sets. #
	    shift @{$lineage1};
	}			# End of the main for-loop (over descendant lineage of MRCA2 down to $br1). #

	my $eff_dist = @{$lineage1};

	return $eff_dist;


}	# End of the subroutine 'effective_distance'. #


	    # ($ct_flnk_ovlp_indels_rec, $indels_lf_rec, $indels_rf_rec, $indels_le_rec, $indels_re_rec, $indels_other_ovlp_rec) = find_ovlp_flanking_indels ($beg_cb_rec, $end_cb_rec, @{$set_indels_std_by_begsite_rec}, @{$set_indels_std_by_endsite_rec}, %{$rec_event2ref_event}, $nodeid_mrca_cb, %{$br2change_cb}, %{$node2pa}, %{$node2ch}, $topnode_id, $CHAR_ABS);

# find_ovlp_flanking_indels:
#	This subroutine searches for indel events that either flank or begin or end within a specified interval.
#	But those events with equivalent events in the other MSA will be excluded.
#	(CREATED in creation of 'classify_msa_errors_via_mblks.1.pl' on Oct 29 (Wed), 2014.)
#
#	[MODIFIED on Nov 5 (Wed), 2014]:
#	The elongation condition of the flanking events now takes account of the contiguity of the current and the next events.
#
# INPUT: ($beg_intvl, $end_intvl, \@set_indels_std_by_begsite, \@set_indels_std_by_endsite, \%event2equiv_event, $nodeid_mrca_cb, \%br2change_cb, \%node2pa, \%node2ch, $topnode_id, $char_abs), where
#
#	$beg/end_intvl is the coordinate of the beginning/ending site of the interval (in the 0-based full-closed convention);
#	
#	@{$set_indels_std_by_begsite_ref[$i]} = ($beg_site, \@set_indels_beg_at_here) is for the $i th beginning site (in numerically ascending order), with
#	[ @{$set_indels_beg_at_here[$e1]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, \@indices_lin_blk) ] --- OBSOLETE as of Nov 4 (Tue), 2014.
#	@{$set_indels_beg_at_here[$e1]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, $length_event, \@indices_lin_blk)  # MODIFIED as of Nov 4 (Tue), 2014.
#		is for the $e1 th (in the order of occurrences) indel event beginning at $beg_site;
#
#	@{$set_indels_std_by_endsite_ref[$i]} = ($end_site, \@set_indels_end_at_here) is for the $i th ending site (in numerically ascending order), with
#	@{$set_indels_end_at_here[$e2]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $e2 th (in the order of occurrences) indel event ending at $end_site.
#
#	%event2equiv_event = ("$br_id:$beg_site:$end_site" => "$br_id_eq:$beg_site_eq:$end_site_eq", ....), with
#	$br_id/$beg_site/$end_site being the branch ID/beginning site/ending site of the indel event, and
#	$br_id_eq/$beg_site_eq/$end_site_eq being the branch ID/beginning site/ending site of the equivalent indel event;
#
#	$nodeid_mrca_cb is the MRCA of the sequences involving the composite-block;
#
#	%br2change_cb = ($branch_id => [$state_ue, $state_le], ...) is for the branches separating sequences involving the composite-block from the rest;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but the top-node (i.e., root);
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#	$topnode_id is the node ID of the top-node (i.e., root);
#
#	$char_abs is the character representing the 'absence' state.
#
# OUTPUT: ($ct_flnk_ovlp_indels, \@indels_lf, \@indels_rf, \@indels_le, \@indels_re, \@indels_other_ovlp), where
#
#	$ct_flnk_ovlp_indels is the total number of indel events that either flank or begin/end within the interval;
#
#	@{$indels_lf[$lf]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $lf th indel event left-flanking the interval, i.e., ending exactly at ($beg_intvl-1);
#
#	@{$indels_rf[$rf]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $rf th indel event right-flanking the interval, i.e., beginning exactly at ($end_intvl+1);
#
#	@{$indels_le[$le]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $le th indel event beginning exactly at $beg_intvl;
#
#	@{$indels_re[$re]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $re th indel event ending exactly at $end_intvl;
#
#	@{$indels_other_ovlp[$ot]} = (exactly in the same format as @{$set_indels_beg_at_here[$e1]}) is for the $ot th other indel event beginning or ending within the interval.
#
#	NOTE that the first 4 sets can overlap with each other, whereas the last one (@indels_other_ovlp) does not overlap with the others.
#
sub find_ovlp_flanking_indels ($$\@\@\%$\%\%\%$$) {

	my ($beg_intvl, $end_intvl, $set_indels_std_by_begsite, $set_indels_std_by_endsite, $event2equiv_event, $nodeid_mrca_cb, $br2change_cb, $node2pa, $node2ch, $topnode_id, $char_abs) = @_;

	my $ct_begs = @{$set_indels_std_by_begsite};

	unless ($ct_begs>0) { # There is NO indel events anyway. #

	    return (0, [], [], [], [], []);
	}

	my $ct_ends = @{$set_indels_std_by_endsite};

	my @full_indices_beg = (0 .. $ct_begs -1);
	my @full_indices_end = (0 .. $ct_ends -1);

	my %event2if_ovlp; # = ("$branch_id:$beg_site:$end_site" => 1, ...) for the events either lanking the interval or beginning/ending within the interval.

		## First, look for events flanking the interval. ##
	my @indels_lf = my @indels_rf = ();

	my ($fl_lf, $indx2_lf, $indx_lf) = binsearch_num ($beg_intvl-1, @{$set_indels_std_by_endsite}, 0, @full_indices_end);
	if ($fl_lf) {
	    my $indx_eff_lf = $indx_lf;
	    my ($eff_dist, $set_sbj_indels);
	    for (; $indx_eff_lf >=0; $indx_eff_lf--) { # Preliminary for-loop (leftward down to the left-most endsite). #

			# Basically, skip a set of events that the composite-block can 'freely' path through. #

		$set_sbj_indels = $set_indels_std_by_endsite->[$indx_eff_lf][1];
		(@{$set_sbj_indels}>1) and last;

			# MODIFIED on Nov 5 (Wed), 2014. #
		my ($br, $beg_site, $state_ue) = @{$set_sbj_indels->[0]}[0, 1, 3];
		##my ($br, $state_ue) = @{$set_sbj_indels->[0]}[0, 3];
		$eff_dist = effective_distance ($br, $state_ue, $nodeid_mrca_cb, %{$br2change_cb}, %{$node2pa}, %{$node2ch}, $topnode_id, $char_abs);

		($eff_dist<3) and last;

		(($indx_eff_lf == 0) or ($set_indels_std_by_endsite->[$indx_eff_lf-1][0] < $beg_site-1)) and last; # If there is no contiguous events on the left. (ADDED on Nov 5 (Wed), 2014.)

	    }				# End of the preliminary for-loop (leftward down to the left-most endsite). #

	    (($indx_eff_lf >= 0) and ((@{$set_sbj_indels}>1) or ($eff_dist<3))) or ($indx_eff_lf = $indx_lf); # If no decent set of indels cannot be found, use the literally flanking indels. (ADDED on Nov 5 (Wed), 2014.)
	    
	    if ($indx_eff_lf >=0) {
		foreach my $indel_info (@{$set_indels_std_by_endsite->[$indx_eff_lf][1]}) { # 1st main foreach-loop (over left-flanking events). #
		    my $label = join (':', @{$indel_info}[0,1,2]);
		    (defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #

		    push @indels_lf, $indel_info;

		    $event2if_ovlp{$label} = 1;
		}				# End of the 1st main foreach-loop (over left-flanking events). #
	    }
		# MODIFIED to capture all events ending on the right of the 'left-flanking' events (on Nov 5 (WEd), 2014) 
	    $indx_lf = $indx_eff_lf;
	}
	$indx_lf++; # Increment the index by 1 to make the event-ending-site(s) on the right of $beg_intvl-1.

	my ($fl_rf, $indx2_rf, $indx_rf) = binsearch_num ($end_intvl+1, @{$set_indels_std_by_begsite}, 0, @full_indices_beg);
	if ($fl_rf) {
	    my $indx_eff_rf = $indx_rf;
	    my ($set_sbj_indels, $eff_dist);
	    for (; $indx_eff_rf < $ct_begs; $indx_eff_rf++) { # Preliminary for-loop (rightward up to the right-most beginning-site). #
			# Basically, skip a set of events that the composite-block can 'freely' path through. #

		$set_sbj_indels = $set_indels_std_by_begsite->[$indx_eff_rf][1];
		(@{$set_sbj_indels}>1) and last;

			# MODIFIED on Nov 5 (Wed), 2014. #
		my ($br, $end_site, $state_ue) = @{$set_sbj_indels->[0]}[0, 2, 3];
		##my ($br, $state_ue) = @{$set_sbj_indels->[0]}[0, 3];
		$eff_dist = effective_distance ($br, $state_ue, $nodeid_mrca_cb, %{$br2change_cb}, %{$node2pa}, %{$node2ch}, $topnode_id, $char_abs);

		($eff_dist<3) and last;

		(($indx_eff_rf == $ct_begs-1) or ($end_site + 1 < $set_indels_std_by_begsite->[$indx_eff_rf+1][0])) and last; # If there is no contiguous events on the right. (ADDED on Nov 5 (Wed), 2014.)

	    }				# End of the preliminary for-loop (rightward up to the right-most beginning-site). #

	    (($indx_eff_rf < $ct_begs) and ((@{$set_sbj_indels}>1) or ($eff_dist<3))) or ($indx_eff_rf = $indx_rf); # If no decent set of indels cannot be found, use the literally flanking indels. (ADDED on Nov 5 (Wed), 2014.)

	    if ($indx_eff_rf < $ct_begs) {
		foreach my $indel_info (@{$set_indels_std_by_begsite->[$indx_eff_rf][1]}) { # 2nd main foreach-loop  (over right-flanking events). #
		    my $label = join (':', @{$indel_info}[0,1,2]);
		    (defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #

		    push @indels_rf, $indel_info;

		    $event2if_ovlp{$label} = 1;
		}				# End of the 2nd main foreach-loop (over right-flanking events). #
	    }

		# MODIFIED to capture all events beginning on the left of the 'right-flanking' events (on Nov 5 (WEd), 2014) 
	    $indx_rf = $indx_eff_rf -1; 
	    ##$indx_rf--; # Decrement the index by 1 to make the event-beginning-site(s) on the left of $end_intvl+1.
	}


		## Second, look for events beginning/ending right at the beginning/ending site. ##
	my @indels_le = my @indels_re = ();

	my ($fl_le, $indx2_le, $indx_le) = binsearch_num ($beg_intvl, @{$set_indels_std_by_begsite}, 0, @full_indices_beg);
	if ($fl_le) {
	    foreach my $indel_info (@{$set_indels_std_by_begsite->[$indx_le][1]}) { # 3rd main foreach-loop (over left-ending events). #
		my $label = join (':', @{$indel_info}[0,1,2]);
		(defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #

		push @indels_le, $indel_info;

		$event2if_ovlp{$label} = 1;
	    }				# End of the 3rd main foreach-loop (over left-ending events). #
	}
	$indx_le++; # Increment the indx by 1 to make the event-beginning-site(s) on the right of $beg_intvl.

	my ($fl_re, $indx2_re, $indx_re) = binsearch_num ($end_intvl, @{$set_indels_std_by_endsite}, 0, @full_indices_end);
	if ($fl_re) {
	    foreach my $indel_info (@{$set_indels_std_by_endsite->[$indx_re][1]}) { # 4th main foreach-loop (over right-ending events). #
		my $label = join (':', @{$indel_info}[0,1,2]);
		(defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #

		push @indels_re, $indel_info;

		$event2if_ovlp{$label} = 1;
	    }				# End of the 4th main foreach-loop (over right-ending events). #

	    $indx_re--; # Decrement the index by 1 to make the event-ending-site(s) on the left of $end_intvl.
	}

		## Third, look for other events beginning/ending within the interval. ##
	my @indels_other_ovlp = ();

			# First, those beginning within the interval.

	for (my $i=$indx_le; $i<=$indx_rf; $i++) { # 5th main foreach-loop (over events beginning within the interval). #
	    foreach my $indel_info (@{$set_indels_std_by_begsite->[$i][1]}) { # Inner foreach-loop. #
		my $label = join (':', @{$indel_info}[0,1,2]);
		(defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #
		(defined $event2if_ovlp{$label}) and next; # Skip if it's already in any of the lists.

		push @indels_other_ovlp, $indel_info;

		$event2if_ovlp{$label} = 1;
	    }						# End of the inner foreach-loop. #
	}				# End of the 5th main foreach-loop (over events beginning within the interval). #

			# Then, those ending within the interval. 

	for (my $i=$indx_lf; $i<=$indx_re; $i++) { # 6th main foreach-loop (over events ending within the interval). #
	    foreach my $indel_info (@{$set_indels_std_by_endsite->[$i][1]}) { # Inner foreach-loop. #
		my $label = join (':', @{$indel_info}[0,1,2]);
		(defined $event2equiv_event->{$label}) and next; # Skip if the event has an equivalent. #
		(defined $event2if_ovlp{$label}) and next; # Skip if it's already in any of the lists.

		push @indels_other_ovlp, $indel_info;

		$event2if_ovlp{$label} = 1;
	    }						# End of the inner foreach-loop. #
	}				# End of the 6th main foreach-loop (over events ending within the interval). #


		## Finally, Output the results. ##

	my $ct_flnk_ovlp_indels = scalar (keys %event2if_ovlp);

	return ($ct_flnk_ovlp_indels, \@indels_lf, \@indels_rf, \@indels_le, \@indels_re, \@indels_other_ovlp);

}		# End of the subroutine 'find_ovlp_flanking_indels'. #


# if_collapse_indep_ins:
#	This subroutine examines whether the gap configuration of the input reconstructed MSA is likely due to
#	a 'Collapse' of independent insertions in the reference MSA.
#	(CREATED on Nov 4 (Tue), 2014.)
#	(=> MODIFIED on Nov 10 (Mon), 2014, for DEBUGGING.)
#
#	[MODIFIED on Oct 4 (Sun), 2015]
#	The subroutine now regards the error as (i)CII even if 1 spuious deletion is missing,
#	as long as it involves 2 or more spurious indels are involved.
#
# INPUT: (\@indels_ref, \@indels_rec, \%node2pa, \%node2ch, $char_abs), where
#
#	@{$indels_ref[$f]} = ($branch_id, $beg_site, $end_site, $state_ue, $state_le, $length_indel, \@indices_lin_blk) is for the $f th indel event among those giving the reference MSA;
#	@{$indels_rec[$c]}, which follows the same format as @{$indels_ref[$f]}, is for the $c th indel event among those giving the reconstructed MSA;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but the top-node (i.e., root);
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#
#	$char_abs is a character representing an 'absence' state.
#
# OUTPUT: ($if_cii, \@indels_invlvd_rec), where
#	$if_cii = 1 if the input is likely explained by a 'Collapse of Independent Insertions' (CII),
#		= 0 otherwise, but if the subroutine successfully ended,
#		= undef if the subroutine failed.
#
#	@indels_invlvd_rec, which is basically the subset of @indels_rec, stores only those events
#	that were likely involved in the CII.
#
sub if_collapse_indep_ins (\@\@\%\%$) {

	my ($indels_ref, $indels_rec, $node2pa, $node2ch, $char_abs) = @_;

	unless (@{$indels_ref}==2) {
	    my $ct_indels_ref = @{$indels_ref};
	    print STDOUT "[ERROR 'if_collapse_indep_ins'] ILLEGAL INPUT: The number of indel events in the reference MSA is $ct_indels_ref, not 2 as expected: $! !\n\n";
	    return (undef, undef);
	}

	my ($br1, $state_ue1) = @{$indels_ref->[0]}[0, 3];
	my ($br2, $state_ue2) = @{$indels_ref->[1]}[0, 3];

	if (($state_ue1 ne $char_abs) and ($state_ue2 ne $char_abs)) {
		# Both events are deletions. Thus, a CII cannot occur.
	    return (0, []);
	}

	if ($br1 == $br2) {
		# A CII cannot occur. Even if the events collapsed, it should be called a 'purge'. #
	    return (0, []);
	}

	my ($common_anc, $lineage1, $lineage2) = fetch_common_ancestors ($br1, $br2, %{$node2pa});

	my @to_be_examined;
	my $mrca;
	if ($state_ue1 ne $char_abs) {
		# The 1st event is a deletion. => $br1 must be an ancestor of $br2.

	    if (@{$lineage1}>0) { # The two events can never collapse into the same block.
		return (0, []);
	    }
	    @to_be_examined = @{$lineage2};

	} elsif ($state_ue2 ne $char_abs) {
		# The 2nd event is a deletion. => $br2 must be an ancestor of $br1.

	   if (@{$lineage2}>0) { # The two events can never collapse into the same block.
		return (0, []);
	   }
	   @to_be_examined = @{$lineage1};

	} else {
		# The both events are insertions. => $br1 and $br2 must diverge into different lineages.

	    unless ((@{$lineage1}>0) and (@{$lineage2}>0)) {

			# MODIFIED for DEBUGGING on Nov 10 (Mon), 2014. #
			# (It turns out that, although rare, this case could actually occur.
			# ... In that case, the two events (that were examined) can never collapse into the same block.)
			#
		return (0, []);
		##print STDOUT "[ERROR 'if_collapse_indep_ins'] IMPOSSIBLE!!... The two events for reference MSA are insertions, yet the branches they occur does NOT seem to diverge into different lineages: $! !!\n\n";
		##return (undef, undef);
	    }

	   (@{$common_anc}>1) and ($mrca = $common_anc->[$#{$common_anc}]); # NOTE: $mrca is undef if MRCA is the top node (i.e., the root). #

	   shift @{$lineage1}; # The 1st branch after divergence must be removed.
	   shift @{$lineage2}; # Same here.
	   @to_be_examined = (@{$lineage1}, @{$lineage2});
	}

		# Preliminary process. #

	my %br2rec_event_info; # = ($branch_id => \@event_info, ...) for all events on the reconstructed MSA. #
	foreach my $event_info (@{$indels_rec}) { $br2rec_event_info{$event_info->[0]} = $event_info; }

		### Main process. ###

	my $maxct_indels_invlvd_rec = 0; # Maximum possible number of indels involved in the reconstructed MSA.  (ADDED for DEBUGGING on Oct 4 (Sun), 2015). #

	my @indels_invlvd_rec = ();

	if (defined $mrca) {
	   my $event_mrca = $br2rec_event_info{$mrca};
	   unless ((defined $event_mrca) and ($event_mrca->[3] eq $char_abs)) {
		# Either no event along the MRCA branch, or if the event is not an insertion. #
		return (0, []);
	   }
	   push @indels_invlvd_rec, $event_mrca;
	   $maxct_indels_invlvd_rec++; # ADDED for DEBUGGING on Oct 4 (Sun), 2015. #
	}

	foreach my $br (@to_be_examined) { # Main outer foreach-loop (over branches mediating the two nodes). # 
	    my $pa = $node2pa->{$br};
	    my $sibls = $node2ch->{$pa};
	    foreach my $sbl (@{$sibls}) { # Main inner foreach-loop (over siblings of $br). #

		($sbl == $br) and next;

		$maxct_indels_invlvd_rec++;

		my $event_sbl = $br2rec_event_info{$sbl};

		unless ((defined $event_sbl) and ($event_sbl->[3] ne $char_abs)) {
			# Either no event along the sibling branch, or if the event is not a deletion.

		    ##return (0, []); # This might have been too strict,
		    next; 		# hence was replaced with a more lenient condition (for DEBUGGING on Oct 4 (Sun), 2015).
		}
		push @indels_invlvd_rec, $event_sbl;

	    }			# End of the main inner foreach-loop (over siblings of $br). #
	}			# End of the main outer foreach-loop (over branches mediating the two nodes). #

		# Regard the error as (i)CII if 1 spurious deletion is missing.
		# (ADDED for DEBUGGING on Oct 4 (Sun), 2015.) #
	my $ct_missing_indels = $maxct_indels_invlvd_rec - scalar (@indels_invlvd_rec);
	unless (($ct_missing_indels == 0) or (($ct_missing_indels < 2) and (@indels_invlvd_rec>1)) ) { 
	    return (0, []);
	}
		# END of "# Regard the error as (i)CII if 1 spurious deletion is missing. #

	return (1, \@indels_invlvd_rec);

}	# End of the subroutine 'if_collapse_indep_ins'. #


# if_part_nbr_mrg_2ins:
#	This subroutine examines whether the three input events are consistent with
#	the error of the type 'partial-neighboring-branch-merge' of two insertions.
#	(CREATED on Nov 12 (Wed), 2014.)
#
#	CRITERIA:
#	(1) The three events must be on different branches (pre-guaranteed), 
#		and the branches must share a node.
#	(2) Starting from the shared node, event 1 and event 2 must be 'insertion's, and event 3 must be a 'deletion'.
#
# INPUT: (\@event1, \@event2, \@event3, $indx_br, $indx_state_ue,  \%equiv_branches, \%node2pa, $char_abs), where
#
#	($br1/2/3, $state_ue1/2/3) = @{$event1/2/3}[$indx_br, $indx_state_ue] are the branch along which the event 1/2/3 occurred,
#	and the state at the upper-end for the event 1/2/3;
#
#	%equib_branches = ($br_id => $equiv_br_id, ...) for the branches with their own 'equivalent's;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node (i.e., root);
#
#	$char_abs is a character representing an 'absence' state.
#
# OUTPUT: $if_part_nbr_mrg, which should be
#		= 1 if the events are consistent with the error type 'partial-neighboring-branch-merge',
#		= 0 otherwise,
#		= undef if the subroutine failed.
#
sub if_part_nbr_mrg_2ins (\@\@\@$$\%\%$) {

	my ($event1, $event2, $event3, $indx_br, $indx_state_ue, $equiv_branches, $node2pa, $char_abs) = @_;

	my ($br1, $state_ue1) = @{$event1}[$indx_br, $indx_state_ue];
	my ($br2, $state_ue2) = @{$event2}[$indx_br, $indx_state_ue];
	my ($br3, $state_ue3) = @{$event3}[$indx_br, $indx_state_ue];

	if ($br1 == $node2pa->{$br2}) { # $br1 must be the parent of $br2 and $br3.

	    ($br1 == $node2pa->{$br3}) or (return 0); 
	    if (($state_ue1 ne $char_abs) and ($state_ue2 eq $char_abs) and ($state_ue3 ne $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }

	} elsif ($br2 == $node2pa->{$br1}) { # $br2 must be the parent of $br1 and $br3.

	    ($br2 == $node2pa->{$br3}) or (return 0);
	    if (($state_ue1 eq $char_abs) and ($state_ue2 ne $char_abs) and ($state_ue3 ne $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }

	} elsif ($br3 == $node2pa->{$br1}) { # $br3 must be the parent of $br1 and $br2.

	    ($br3 == $node2pa->{$br2}) or (return 0);
	    if (($state_ue1 eq $char_abs) and ($state_ue2 eq $char_abs) and ($state_ue3 eq $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }

	}

	my $eq_br1 = $equiv_branches->{$br1};
	my $eq_br2 = $equiv_branches->{$br2};
	my $eq_br3 = $equiv_branches->{$br3};

		# NOTE: at most one of the equivalents will be defined. #

	if (defined $eq_br1) { # The equivalent of $br1 must be the parent of $br2 and $br3.

	    (($eq_br1 == $node2pa->{$br2}) and ($eq_br1 == $node2pa->{$br3})) or (return 0);
	    if (($state_ue1 eq $char_abs) and ($state_ue2 eq $char_abs) and ($state_ue3 ne $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }

	} elsif (defined $eq_br2) { # The equivalent of $br2 must be the parent of $br1 and $br3.

	    (($eq_br2 == $node2pa->{$br1}) and ($eq_br2 == $node2pa->{$br3})) or (return 0);
	    if (($state_ue1 eq $char_abs) and ($state_ue2 eq $char_abs) and ($state_ue3 ne $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }

	} elsif (defined $eq_br3) { # The equivalent of $br3 must be the parent of $br1 and $br2. 

	    (($eq_br3 == $node2pa->{$br1}) and ($eq_br3 == $node2pa->{$br2})) or (return 0);
	    if (($state_ue1 eq $char_abs) and ($state_ue2 eq $char_abs) and ($state_ue3 ne $char_abs)) {
		return 1;
	    } else {
		return 0;
	    }
	}

		# The three events do NOT share a node. #
	return 0;

}		# End of the subroutine 'if_part_nbr_mrg_2ins'. #



	### DEFINITIONS of the Subroutines ADDED in creation of 'compare_ref_vs_rec_mafft_consist.pl' or later. ###

# read_list_pw_lclhom:
#	This subroutine reads a list of pairwise local homology information,
#	which was used to calculate the consistency score of MSAs.
#	(CREATED in creation of 'compare_ref_vs_rec_mafft_consist.pl' on Sep 15 (Mon), 2014.)
#
# INPUT: (\$inlist, $ct_seqs), where
#	$inlist specifies the path to the input list file;
#	$ct_seqs is the number of sequences in the MSA.
#
# OUTPUT: \@set_info_pw_lclhom, where
#	@info_pw_lclhom = @{$set_info_pw_lclhom[$i]->[$j]} ($i< $j) stores information on the pairwise local homology
#	between the $i th and the $j th sequences in the MSA. More specifically,
#	@{$info_pw_lclhom->[$n]} = ($start1, $end1, $start2, $end2, $importance) is for the $n the gapless segment
#	in the pairwise local alignment(s) between the two sequences.
#
sub read_list_pw_lclhom (\$$) {

	my ($inlist, $ct_seqs) = @_;
	local $/ = "\n";

	my $fh_tmp = open_input_file2 (${$inlist});
	unless (defined $fh_tmp) {
	    print STDOUT "[ERROR 'read_list_pw_lclhom'] 'open_input_file2' FAILED : $! !\n\n";
	    return undef;
	}

	my $header = <$fh_tmp>;

	my @set_info_pw_lclhom = ();
	for (my $i=0; $i<$ct_seqs-1; $i++) { $set_info_pw_lclhom[$i] = []; }

	while (my $line = <$fh_tmp>) { # Main while-loop (over input lines). #

	    my @fields = split /\s+/, $line;
	    (@fields>6) or next;

	    my ($i, $j, $start1, $end1, $start2, $end2, $importance) = @fields;

	    unless ($i < $j) {
		$line =~ s/\s+$//;
		print STDOUT "[WARNING 'read_list_pw_lclhom'] ILLEGAL INPUT: Indx_seq1 (=$i) is NOT LESS THAN Indx_seq2 (=$j) for the line:\n";
		print STDOUT "'$line' : $! !\n\n";
		next;
	    }

	    unless (($i>=0) and ($j <$ct_seqs)) {
		$line =~ s/\s+$//;
		print STDOUT "[WARNING 'read_list_pw_lclhom'] ILLEGAL INPUT: EITHER Indx_seq1 (=$i) is Negative OR Indx_seq2 (=$j) is NOT LESS THAN #{sequences} (=$ct_seqs) for the line:\n";
		print STDOUT "'$line' : $! !\n\n";
		next;
	    }

	    my $row_info = $set_info_pw_lclhom[$i];
	    (defined $row_info) or ($row_info = $set_info_pw_lclhom[$i] = []);

	    my $info_pw_lclhom = $row_info->[$j];
	    (defined $info_pw_lclhom) or ($info_pw_lclhom = $row_info->[$j] = []);

	    push @{$info_pw_lclhom}, [$start1, $end1, $start2, $end2, $importance];

	}			# End of the main while-loop (over input lines). #

	close ($fh_tmp);

	return \@set_info_pw_lclhom;

}	# End of the subroutine 'read_list_pw_lclhom'. #

	### DEFINITIONS OF THE NEW SUBROUTINES (ADDED since Aug 26 (Sun), 2012) ###

# readin_list_segs:
#	This subroutine reads in a list of segments.
#
# INPUT: (\$inlist, \%label_sattr2indx), where
#	$inlist is the path to the input file.
#	%label_sattr2indx = ($label_seg_attr => $indx, ...)
#	with $indx being the index in the output list.
#
# OUTPUT: \@list_segments, where
#	@{$list_segments[$s]} = ($seg_ID, @attributes)
#	is the information on the $s th segment, with
#	$seg_ID being the ID, and
#	@attributes storing the attributes of the segment in the same order as in @LABELS_ATTR.
#
sub readin_list_segs (\$\%) {

	my ($inlist, $label_sattr2indx) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$inlist});
	unless (defined $fh_tmpin) {
	   print STDOUT "[ERROR 'readin_list_segs'] 'open_input_file2' FAILED on the input file '${$inlist}' $! !\n";
	    return undef;
	}

	my $head1 = <$fh_tmpin>;
	my ($ct_seqs)  = ($head1 =~ m/^\#Sequences\s+(\d+)/);
	my $head2 = <$fh_tmpin>;
	my ($ct_segs) = ($head2 =~ m/^\#Segments\s+(\d+)/);
	unless ((defined $ct_seqs) and (defined $ct_segs)) {
	    (defined $ct_seqs) or ($ct_seqs = 'undef');
	    (defined $ct_segs) or ($ct_segs = 'undef');
	    print STDOUT "[ERROR 'readin_list_segs'] ILLEGAL HEADER: (\$ct_seqs, \$ct_segs) = ($ct_seqs, $ct_segs): $! !\n";
	    return undef;
	}

	my @list_segments = ();

	while (my $in = <$fh_tmpin>) { # Main while-loop #

	    my @flds = split /\s+/, $in;   

	    my $segid = shift @flds;
	    $segid =~ s/^seg\#//;
	    my @seg_info = ($segid);

	    while (my $label = shift @flds) { # Inner while-loop #
		$label =~ s/\:$//;
		my $indx = $label_sattr2indx->{$label};
		my $attr = shift @flds;
	 	unless (defined $indx) {
		    print STDERR "[CAUTION 'readin_list_segs'] NO INDEX corresponds to the attribute-label '$label' (with value '$attr') in the segment '$segid': $! !\n";
		    next;
		}
		$seg_info[$indx] = $attr; 
	    }		# End of the inner while-loop #

	    push @list_segments, \@seg_info;

	}		# End of the main while-loop #

	close ($fh_tmpin);

	return \@list_segments;

}	# End of the subroutine 'readin_list_segs' #


# readin_list_segs2:
#	This subroutine reads in a list of segments in a format different from the input for 'readin_list_segs'.
#	(CREATED in creation of 'compare_ref_vs_rec_compllimment.pl' on Aug 9 (Sat), 2014)
#
# INPUT: (\$inlist, \@labels_seg_attr), where
#	$inlist is the path to the input file.
#	@labels_seg_attr stores the labels of the fields for the segment attributes (including the ID (or index)).
#
# OUTPUT: \@list_segments, where
#	@{$list_segments[$s]} = @attributes
#	is the information on the $s th segment, with
#	@attributes storing the attributes of the segment in the same order as in @labels_seg_attr.
#
sub readin_list_segs2 (\$\@) {

	my ($inlist, $labels_seg_attr) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$inlist});
	unless (defined $fh_tmpin) {
	   print STDOUT "[ERROR 'readin_list_segs2'] 'open_input_file2' FAILED on the input file '${$inlist}' $! !\n";
	    return undef;
	}

	my $head = <$fh_tmpin>;

	my $indices_attr = tfm_labels2indices (@{$labels_seg_attr}, $head);
	unless (defined $indices_attr) {
	    $head =~ s/\s+$//;
	    print STDOUT "[ERROR 'readin_list_segs2'] 'tfm_labels2indices' FAILED on\n";
	    print STDOUT "      the labels  = (", join (', ', @{$labels_seg_attr}), " ),\n";
	    print STDOUT "  and the header  = '$head' : $! !\n\n";
	    return undef;
	}

=set aside for later use
	my $new_indx_diag;
	for (my $i=0; $i<@{$labels_seg_attr}; $i++) { # Preliminary for-loop (over field labels). #
	    ($labels_seg_attr->[$i] eq 'Diagnosis') or next;
	    $new_indx_diag = $i;
	    last;
	}				# End of the preliminary for-loop (over field labels). #
=cut

	my $max_indx_fld = 0;
	foreach my $indx (@{$indices_attr}) { ($max_indx_fld < $indx) and ($max_indx_fld = $indx); }

	my @list_segments = ();

	while (my $tmpin = <$fh_tmpin>) { # Main while-loop (over input lines). #

	    $tmpin =~ s/\s+$//;
	    my @fields = split /\t/, $tmpin;

	    (@fields > $max_indx_fld) or next; # Skip lines with an insufficient number of fields.

	    my @attributes = @fields[@{$indices_attr}];

	    push @list_segments, \@attributes;

	}		# End of the main while-loop (over input lines). #

	close ($fh_tmpin);

	return \@list_segments;

}	# End of the subroutine 'readin_list_segs2' #


# readin_simul_params:
#	This subroutine reads simulation parameters,
#	except the input tree and those in the comment line(s),
#	out of an input list.
#
# INPUT: (\$inlist), where
#	$inlist is the path to an input list file.
#
# OUTPUT: \%hash_simul_params, where
#	%hash_simul_params = ($label => $value, ...) for the simulation parameters.
#
sub readin_simul_params (\$) {

	my ($inlist) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$inlist});
	unless (defined $fh_tmpin) {
	    print STDOUT "[ERROR 'readin_simul_params'] 'open_input_file2' FAILED on the input list '${$inlist}': $! !\n";
	    return undef;
	}

	my %hash_simul_params;

	while (my $in = <$fh_tmpin>) { # Main while-loop #

	    ($in =~ m/^\s*\#/) and next; # Skip a comment line. #

		# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
	    ($in =~ m/^Tree\s/) and next; # Skip the input tree. #
	    ##($in =~ m/^Tree\.Tree/) and next; # Skip the input tree. #

	    ($in =~ s/\s+$//);

	    my ($label, $value) = split /\s*\=\s*/, $in;
	    unless ((defined $value) and ($value =~ m/\S/)) {
		print STDOUT "[CAUTION 'readin_simul_params'] NO VALUE in the input line:\n";
		print STDOUT "'$in': $! !\n\n";
		next;
	    }

		# ADDED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
	    if ($value =~ s/^\"//) { ($value =~ s/\"$//); }
	    if ($label eq 'GapParams') { $value =~ s/^\{//; $value =~ s/\}$//; }

	    $hash_simul_params{$label} = $value;

	}			# End of the main while-loop #
	close ($fh_tmpin);

	return \%hash_simul_params;

}		# End of the subroutine 'readin_simul_params' #


# inspect_indel_events:
#	This subroutine examines whether all the indel events are shorter than or equal to the cutoff-lengths.
#
# INPUT: (\@set_cls_gblocks, \@sets_psm_scns_for_cls, $cutoff_len_ins, $cutoff_len_del, $char_abs), where
#
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)
#	is for the $gb th block in the $cl th cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordiate of the block (in the 0-based convention);
#
#	@{$set_psm_scns_in_cls[$cl]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th of the parsimonious scenarios realizing the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) if the map of gap states onto the interior nodes,
#	with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id;
#	
#	$cutoff_len_ins/del is the cutoff length for the insertion/deletion events;
#
#	$char_abs is the character representing the 'absence' state.
#
# OUTPUT: \@illegal_indels, where
#	@{$illegal_indels[$il]} = ($cl, $ps, $br_id, $id, $length) is for the $il th illegal indel event, with
#	$cl specifying a cluster,
#	$ps specifying a parsimonious scenario in the cluster,
#	$br_id specifying a branch,
#	$id specifying an illegal event in the node,
#	and $length is the length of the event.
#
#	The subroutine returns undef if it fails.
#
sub inspect_indel_events (\@\@$$$) {

	my ($set_cls_gblocks, $sets_psm_scns_for_cls, $cutoff_len_ins, $cutoff_len_del, $char_abs) = @_;

    (print STDOUT ' ' x 3, "< Subroutine 'inspect_indel_events' >\n\n") if ($TEST_LEVEL>1);

	my $ct_cls = @{$set_cls_gblocks};
	unless (@{$sets_psm_scns_for_cls} == $ct_cls) {
	    print STDOUT "[ERROR 'inspect_indel_events'] ILLEGAL INPUT: The size of ", '@{$sets_psm_scns_for_cls} = (', scalar (@{$sets_psm_scns_for_cls}), ") CLASHES with the number of gap-block clusters (=$ct_cls): $! !\n";
	    return undef;
	}

	my @illegal_indels = ();

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Outer for-loop (over clusters) #

	    my $cls_gblock = $set_cls_gblocks->[$cl];
	    my $psm_scns_for_cls = $sets_psm_scns_for_cls->[$cl];
	    my $ct_psm_scns = @{$psm_scns_for_cls};

    (print STDOUT ' ' x 3, "[ The $cl th cluster has $ct_psm_scns parsimonious scenarios of indel events ]\n\n") if ($TEST_LEVEL>2);

	    for (my $ps=0; $ps<$ct_psm_scns; $ps++) { # Middle for-loop (over parsiminous scenarios) #

    (print STDOUT ' ' x 5, '( The $ps th parsimonious scenario )', "\n\n") if ($TEST_LEVEL>3);

		my ($br2indels, $node2bw_gstat) = @{$psm_scns_for_cls->[$ps]};

		foreach my $br_id (sort {$a <=> $b} keys %{$br2indels}) { # Inner-middle foreach-loop (over branches) #
		    my $indel_events = $br2indels->{$br_id};
		    my $ct_events = @{$indel_events};

    (print STDOUT ' ' x 7, "( The branch '$br_id' has $ct_events nominal indel events )\n") if ($TEST_LEVEL>3);

		    for (my $id=0; $id<$ct_events; $id++) { # Inner for-loop (over indel events) #
			my ($stat_pa, $stat_node, $indices_blk) = (defined $indel_events->[$id]) ? @{$indel_events->[$id]} : ();
			unless (defined $stat_pa) {
			    print STDOUT "[WARNING 'inspect_indel_events'] The $id th event on the branch '$br_id' is EMPTY  for the $ps th indel scenario for the $cl th cluster: $! !\n";
			    next;
			}
			my $cutoff_len = ($stat_pa eq $char_abs) ? $cutoff_len_ins : $cutoff_len_del;
			my $indel_len = 0;
			foreach my $indx_blk (@{$indices_blk}) { $indel_len += $cls_gblock->[$indx_blk][1] ; }

			unless ($indel_len <= $cutoff_len) {
			    print ' ' x 7, "\n[WARNING 'inspect_indel_events'] The $id th indel event along the $br_id th branch is EXCEEDINGLY LONG (length = $indel_len)  for the $ps th scenario in the $cl th gap-block cluster: $! !\n\n";
			    push @illegal_indels, [$cl, $ps, $br_id, $id, $indel_len];
			}

		    }				# End of the inner for-loop (over indel events) #

		} 		# End of the inner-middle foreach-loop (over branches) #

	    }			# End of the middle for-loop (over parsimonious scenarios) #

	}			# End of the outer for-loop (over clusters) #

    (print STDOUT ' ' x 3, "--- End of the Subroutine 'inspect_indel_events'  ---\n\n") if ($TEST_LEVEL>1);

	return \@illegal_indels;

}	# End of the subroutine 'inspect_indel_events' #


=DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12). (DF)
# readin_dawg_tree2:
#	This modified version of 'readin_dawg_tree' copes with the new format
#	of the Dawg control file.
#	(CREATED in creation of 'compare_ref_vs_rec_compllimment.pl' on Aug 9 (Sat), 2014)
#
# readin_dawg_tree:
#	This subroutine reads in the input tree used to guide
#	a simulation of sequence evolution via DAWG.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011) 
#
# INPUT: \$infile, where
#	$infile is the file recording input parameters of DAWG.
#
# OUTPUT: \@info_dawg_tree, where
#	@info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr).
#	See e.g. 'MyTreeManipu_II.pm' for the format of the output hash.
#
sub readin_dawg_tree2 (\$) {

	my ($infile) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$infile});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'open_input_file2' FAILED on the input file '${$infile}': $! \n";
	    return undef;
	}

	my $in_tree;
	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #

	    ($tmpin =~ s/^Tree\s+\=\s+//) or next; # MODIFIED in creation of 'readin_dawg_tree2'. #
	    ##($tmpin =~ s/^Tree\.Tree\s+\=\s+//) or next;
	    $in_tree = $tmpin;
	    last;
	}		# End of the main while-loop # 
	close ($fh_tmpin);

	unless ((defined $in_tree) and ($in_tree =~ m/\(/)) {
	    (defined $in_tree) or ($in_tree = 'undef');
	    $in_tree =~ s/\s+$//;
	    print STDERR "[ERROR 'readin_dawg_tree2'] NO TREE was extracted from the input file '${$infile}': $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

		## MODIFIED in creation of 'readin_dawg_tree2'. ##

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr; my %name2id_intnode; my $topnode_name;

	unless ( input_tree_file_dawg ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr, %name2id_intnode, $topnode_name) ) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'input_tree_file_dawg' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

		# (Just in case,) remove the interior node names (or bootstrap supports) from the list of branch attributes. #

	foreach my $br (keys %br_attr) {
	    my $attr = $br_attr{$br};
	    my @new_attr = ();
	    $new_attr[$INDX_LENGTH] = $attr->[$INDX_LENGTH];
	    $br_attr{$br} = \@new_attr;
	}

	my @info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr);

	return \@info_dawg_tree;

}	# End of the subroutine 'readin_dawg_tree2' #



	### DEFINITIONS of the SUBROUTINES in the predecessor,
	### "../../../2nd_Stage.DNA/ML_ScoringF6t.PhyML/Scripts/compare_ml_msa.bch.extJC.6t.PhyML.pl".

# readin_dawg_tree:
#	This subroutine reads in the input tree used to guide
#	a simulation of sequence evolution via DAWG.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011) 
#
# INPUT: \$infile, where
#	$infile is the file recording input parameters of DAWG.
#
# OUTPUT: \@info_dawg_tree, where
#	@info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr).
#	See e.g. 'MyTreeManipu_II.pm' for the format of the output hash.
#
sub readin_dawg_tree (\$) {

	my ($infile) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$infile});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'readin_dawg_tree'] 'open_input_file2' FAILED on the input file '${$infile}': $! \n";
	    return undef;
	}

	my $in_tree;
	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #
	    ($tmpin =~ s/^Tree\.Tree\s+\=\s+//) or next;
	    $in_tree = $tmpin;
	    last;
	}		# End of the main while-loop # 
	close ($fh_tmpin);

	unless ((defined $in_tree) and ($in_tree =~ m/\(/)) {
	    (defined $in_tree) or ($in_tree = 'undef');
	    $in_tree =~ s/\s+$//;
	    print STDERR "[ERROR 'readin_dawg_tree'] NO TREE was extracted from the input file '${$infile}': $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}
		# Process the input tree a bit #
	($in_tree =~ s/[^\)\;]+$//);
	($in_tree =~ m/\;$/) or ($in_tree .= ';');

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;
	unless ( input_tree_file ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr) ) {
	    print STDERR "[ERROR 'readin_dawg_tree'] 'input_tree_file' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

	my @info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr);

	return \@info_dawg_tree;

}	# End of the subroutine 'readin_dawg_tree' #

=cut # END of "DISABLED in creation of 'map_alnshear_and_srd.ver0.6.pl' (on 2020/03/12). (DF)"


# ln_ml_sgl_seq2:
#	This revised subroutine of 'ln_ml_sgl_seq' also outputs 'column-wise' log ML values.
#	(IMPORTED from 'MyApBayes_Score_MSA_ExtJC.pm' and MODIFIED in creation of
#	'compare_apbayes_msa.bch.extJC.6.pl' on Mar 7 (Mon), 2011)
#
# ln_ml_sgl_seq:
#	This subroutine estimates the log-likelihood of obtaining a single sequence.
#	The log-likelihood is simply calculated as:
#	$lnl = Sum_$r log ($freq[$r]) ,
#	where $freq[$r] is the equilibrium frequency of the $r th amino acid residue.
#
# INPUT: (\$seq, \%aa2freq), where
#	$seq is the sequence;
#	%aa2freq = ($amino_acid (a single capital letter) => $equilibrium_frequency, ...).
#
# OUTPUT: ($lnl, \@clmwise_lnl), where
#	$lnl is the log-likelihood of the whole alignment;
#	$clmwise_lnl[$l] is the log-likelihood of the $l th column.
#
#	The subroutine returns (undef, undef) if it failed.
#	
sub ln_ml_sgl_seq2 (\$\%) {

	my ($seq, $aa2freq) = @_;

	my $cp_seq = uc ${$seq};
	my $len = length $cp_seq;

	my @clmwise_lnl = ();
	my $lnl = 0;
	for (my $l=0; $l<$len; $l++) { # For-loop #
	    my $res = substr ($cp_seq, 0, 1, '');
	    my $frq = $aa2freq->{$res};
	    unless (defined $frq) {
		print STDERR "[ERROR 'ln_ml_sgl_seq2'] The frequency is UNDEFINED for the residue '$res' at the $l th site: $! !\n\n";
		return (undef, undef);
	    }
	    my $clmwise = log $frq;
	    $lnl += $clmwise;
	    $clmwise_lnl[$l] = $clmwise; # ADDED in creation of 'compare_apbayes_msa.bch.extJC.6.pl' (on Mar 7 (Mon), 2011) #
	}		# End of the for-loop #

	return ($lnl, \@clmwise_lnl);

}	# End of the subroutine 'ln_ml_sgl_seq2' #


# add_dummy_seq:
#	This subroutine adds a 'dummy' sequence to a tree and an MSA consisting of two sequences,
#	and makes a 'virtual' 3-sequence tree and MSA.
#	The dummy is the exact copy of the 2nd sequence, which 'diverge'd $eps above its exterior node,
#	where $eps is the minimum length permitted by PhyML.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 11 (Mon) & 12 (Tue), 2011)
#
# INPUT: (\@seqnames, \@msa, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr, \@indices_attr), where
#	$seqnames[$k] is the name of the $k th sequence;
#	${$msa[$k]} is the $k th sequence embedded in the alignment;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#	%name2id = ($name => $node_id, ...) for exterior nodes and the top node;
#	%id2name = ($node_id => $name, ...) for exterior nodes;
#	%br_attr = ($node_id => \@branch_attributes, ...) for the nodes immediately downstream of the branches,
#	with @branch_attributes[@indices_attr] = ($branch_length, $bootstrap_support, Total#{bootstrap resamplings}).
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub add_dummy_seq (\@\@\%\%\%\%\%\@) {

	my ($seqnames, $msa, $node2pa, $node2ch, $name2id, $id2name, $br_attr, $indices_attr) = @_;

	unless ((@{$seqnames} == 2) and (@{$msa} == 2)) {
	    print STDERR "[ERROR 'add_dummy_seq'] ILLEGAL INPUT: Either the size of MSA (", scalar (@{$msa}), ") or the number of sequence names (", scalar (@{$seqnames}), ") is NOT 2, which they are supposed to be: $! !\n";
	    return 0;
	}

		## Manipulate the input tree ##

	my $eps = 0.0000000001; # Minimum branch length permitted in PhyML #
	my ($indx_len, $indx_boot, $tot_boot) = @{$indices_attr};

	my $max_nodeid = 0;
	foreach my $nodeid (keys %{$node2ch}) { ($nodeid > $max_nodeid) and ($max_nodeid = $nodeid); }
	foreach my $nodeid (keys %{$id2name}) { ($nodeid > $max_nodeid) and ($max_nodeid = $nodeid); }

		# Change the identity of the second sequence #
	my $name_seq2 = $seqnames->[1];
	my $nid_seq2 = $name2id->{$name_seq2};
	delete $id2name->{$nid_seq2}; # $nid_seq2 is no longer an exterior node

		# Create two new exterior nodes, one real and the other virtual #
	my $nid_sbl1 = $max_nodeid + 1;
	my $nid_sbl2 = $max_nodeid + 2;

	##my $name_sbl1 = $name_seq2;
	my $name_sbl2 = 'dummy';

	$name2id->{$name_seq2} = $nid_sbl1;
	$name2id->{$name_sbl2} = $nid_sbl2;

	$id2name->{$nid_sbl1} = $name_seq2;
	$id2name->{$nid_sbl2} = $name_sbl2;

		# Establish the relationships between seq2, and the new siblings #
	$node2ch->{$nid_seq2} = [$nid_sbl1, $nid_sbl2];
	$node2pa->{$nid_sbl1} = $nid_seq2;
	$node2pa->{$nid_sbl2} = $nid_seq2;

		# Slightly change the length of the branch above seq2, which is now the 'virtual' parent node #
	my $attr_br2 = $br_attr->{$nid_seq2};
	my $len2 = $attr_br2->[$indx_len];
	$attr_br2->[$indx_len] = ($len2 > $eps) ? $len2 - $eps : 0 ;

		# Fix the lengths of the 'virtual' child branches as $eps #
	foreach my $nid ($nid_sbl1, $nid_sbl2) { # foreach-loop #
	    my @attr = ();
	    $attr[$indx_len] = $eps; 
	    $br_attr->{$nid} = \@attr;
	}			# End of the foreach-loop #

		## Manipulate the input alignment ##
	my $seq2 = ${$msa->[1]};
	$msa->[2] = \$seq2;
	$seqnames->[2] = $name_sbl2;

	return 1;

}	# End of the subroutine 'add_dummy_seq' #


