#!/usr/bin/perl
#
# serial_extract_archives.pl:
#	This Perl script serially extracts some selected files in a gzip-archive.
#	(CREATION STARTED on Oct 1 (Wed), 2014, ENDED on Oct x (yyy), 2014)
#

use strict;
use warnings;

use MyFile_manipu qw(
	enlist_files
	open_input_file open_output_file list_sessions_ext
	);

sub peek_into_archive (\$$$$); ## IMPORTED from 'bch_pbs_cmp_compllimment.D.pl'.

(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;

$| = 1; # Flush the buffer for standard output whenever it receives output.

	# ADDED in creation of 'bch_pbs_cmp_ml_msa.extJC.6t.beta.pl' on Mar 15 (Tue), 2011 #
my $program_path = $0;
my ($program_name) = ($program_path =~ m/([^\/]+)$/); 

my $curr_work_dir = $CURR_WORK_DIR;

	## Input Files/Directories. ##

	# $in_archive = "$inarch_dir/$archive_pref$id_class$archive_suff", and 

##my $inarch_dir = "Outputs";
##my $inarch_dir = 'hogd_alt_lite_sbcl.1/Outputs';
#my $inarch_dir = 'TESTs/Test1e/Outputs';
#my $inarch_dir = 'TESTs/Test1a/Outputs';
#my $inarch_dir = '12primates/Outputs';
my $inarch_dir = '15mammals/Outputs';
#my $inarch_dir = '9fe_mammals/Outputs';

my $archive_pref = 'class_';
##my $archive_pref = 'hog_d';
my $archive_suff = '.tgz';

#my $extfile_pref = 'tbl_summ_apbayes_msa';
##my $extfile_pref = 'tbl_summ_cts_indels';
##my $extfile_pref = 'tbl_summ_class_msa_errors';
#my $extfile_pref = 'tbl_summ_class_msa_errors.1';
##my $extfile_pref = 'tbl_summ_class_msa_errors.2';
##my $extfile_pref = 'tbl_summ_examine_purge';
my $extfile_pref = 'tbl_summ_examine_purge';


##my $topout_dir = 'Outputs_lite';
##my $topout_dir = 'hogd_alt_lite_sbcl.1_lite/Outputs';
#my $topout_dir = 'TESTs/Test1e_lite/Outputs';
#my $topout_dir = 'TESTs/Test1a_lite/Outputs';
#my $topout_dir = '12primates_lite/Outputs';
my $topout_dir = '15mammals_lite/Outputs';
#my $topout_dir = '9fe_mammals_lite/Outputs';

my $midout_dir_pref = $archive_pref;
my $midout_dir_suff = '';

my $logfile = "$topout_dir/log_tar_xpzf.txt";


	## Examine the input/output files/directories. ##

unless (( -d $inarch_dir ) and ( -r $inarch_dir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input archive directory '$inarch_dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
}

my $list_id_classes = list_sessions_ext ($inarch_dir, $archive_pref, $archive_suff, '-');

unless (defined $list_id_classes) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the archive directory '$inarch_dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
}


foreach my $out_dir ($topout_dir) {
    unless ( -d $out_dir) {
	print STDOUT "[WARNING '$program_name'] CanNOT FIND an output directory '$out_dir' ! ... I will try to create one...\n\n";
	unless (0 == system (" mkdir -p $out_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the output directory '$out_dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
    unless ( -w $out_dir ) {
	print STDOUT "[WARNING '$program_name'] The output directory '$out_dir' is NOT WRITABLE ! ... I will try to change its permission mode...\n\n";
	unless (0 == system (" chmod -R u+w $out_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$out_dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<< Program name: $program_name (path: $program_path) >>\n\n";

print STDOUT "< Input/Output files/directories >\n\n";

print STDOUT "Current working directory: $curr_work_dir .\n\n";

print STDOUT "Main Input archives: $inarch_dir/$archive_pref", '{ID_class}', "$archive_suff ,\n";
print STDOUT "          where {ID_class} = ", join (', ', @{$list_id_classes}), " .\n\n";

print STDOUT "Prefix for the extracted files: '$extfile_pref' .\n\n";

print STDOUT "Output Directories: {outdir} = {topout_dir}/$midout_dir_pref", '{ID_class}', "$midout_dir_suff/ ,\n";
print STDOUT "              where {topout_dir} = $topout_dir ,\n";
print STDOUT "               and  {ID_class}'s are the same as those for the input archives.\n\n";

print STDOUT "Log file for 'tar -xpzf': $logfile .\n\n";


print STDOUT "\n< Now, extract the files from archives >\n\n\n";

my $ct_classes = my $ct_extfiles = 0;

foreach my $id_class (@{$list_id_classes}) { # Outer foreach-loop. (over archives (or classes)) #

	print STDOUT "<< Class: $id_class >>\n\n";

	$ct_classes++;

		# Examine the first archive. #

	my $in_archive = join ('', $inarch_dir, '/', $archive_pref, $id_class, $archive_suff);
		# $in_archive = "$inarch_dir/$archive_pref$id_class$archive_suff", and 

	unless (( -f $in_archive ) and ( -r $in_archive ) ) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input archive file '$in_archive' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}

	my $fh_tmp;
	unless ( open ($fh_tmp, "tar -tzf $in_archive | grep '$extfile_pref' |") ) {
	    print STDOUT "[ERROR '$program_name'] CanNOT open the pipe to input the result of \"tar -tzf $in_archive | grep '$extfile_pref'\" !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}

	my $list_to_be_extracted = '';
	my $sbct_extfiles = 0;

	while (my $line = <$fh_tmp>) { # Inner while-loop (over files to be extracted). #
	    ($line =~ m/\S/) or next;
	    $sbct_extfiles++;
	    $list_to_be_extracted .= $line;
	}

	close ($fh_tmp);

	$list_to_be_extracted =~ s/\n/\ /g;

	print STDOUT "#{files to be extracted} = $sbct_extfiles .\n\n";

	unless (0 == system ("tar -C $topout_dir -xpzf $in_archive $list_to_be_extracted > $logfile 2>&1" ) ) {
	    print STDOUT "[ERROR '$program_name'] 'system ( tar -xpzf ... )' FAILED on the class '$id_class' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	} 

	$ct_extfiles += $sbct_extfiles;

}			# END of the outer foreach-loop. (over archives (or classes)) #


print STDOUT "\n=== All Processes FINISHED !! ===\n\n";

print STDOUT "#{MSA Classes} = $ct_classes .\n";
print STDOUT "Total#{files extracted} = $ct_extfiles .\n\n";



	### DEFINITIONS OF THE SUBROUTINES ###

# peek_into_archive:
#	This subroutine looks into an input archive and outputs a list of entries.
#	The archive is supposed to be storing files with the following paths:
#	"$dir/$pref$id$suff",
#	where $id specifies the ID of the file.
#	(IMPORTED from 'bch_serial_reconst_dawg_msas_via_mafft.pl' in creation of 'bch_pbs_cmp_compllimment.D.pl'.)
#
# INPUT: (\$in_archive, $dir, $pref, $suff), where
#	$in_archive is the path to the input archive;
#	$dir/$pref/$suff are templates for the paths to the files stored in the archive.
#
# OUTPUT: \@list_ids, where
#	@list_ids stores the IDs of the archived files with the above paths.
#
sub peek_into_archive (\$$$$) {

	my ($in_archive, $dir, $pref, $suff) = @_;
	local $/ = "\n"; # Use the default input-record-separator. #

	my $fh_tmp;
	unless ( open ($fh_tmp, "tar -tzf ${$in_archive} |" ) ) {
	    print STDOUT "[ERROR 'peek_into_archive'] CanNOT fork or look into ${$in_archive}: $!\n";
	    return undef;
	}

	my @list_ids = ();

	my $ext_pref = "$dir/$pref"; # Extended prefix. #
	my $len_epref = length $ext_pref;
	my $len_suff = length $suff;

	while (my $tmp = <$fh_tmp>) { # Main while-loop (over input lines) #

	    $tmp =~ s/\s+$//; # Remove white spaces at the end of the entry (serving as a separator (and possibly a pad)).
	    ($ext_pref =~ m/^\.\//) or ($tmp =~ s/^\.\///); # Remove './' at the beginning of the entry unless $ext_pref begins with './'. (ADDED in creation of 'bch_serial_PAML.pl'.)

	    my $tmp_suff = substr ($tmp, - $len_suff, $len_suff, '');

	    ($tmp_suff eq $suff) or next; # Skip if the suffix doesn't match.

	    my $tmp_epref = substr ($tmp, 0, $len_epref, '');

	    ($tmp_epref eq $ext_pref) or next; # Skip if the extended prefix doesn't match.

	    push @list_ids, $tmp; # The remnant of the entry must be the ID.


	}			# End of the main while-loop (over input lines) #
	close ($fh_tmp);

	return \@list_ids;

}	# End of the subroutine 'peek_into_archive' #


