#!/usr/bin/perl
#
# pivot_table_summations_distr_srds.pl:
#	This Perl script reads in the main results of MULTIPLE ROUNDS of the Perl script, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl',
#	which are recorded in TWO distributions per MSA (one for each of 'on' and 'off' the alignment-shears),
#	and performs the PIVOT-TABLE SUMMATION-like operation on each type of those tables.
#	The script was created by modifying 'pivot_table_summations.pl'.
#	[CREATION STARTED on 2020/03/30, ENDED on 2020/04/dd.]
#
# pivot_table_summations.pl:
#	This Perl script reads in the main results of MULTIPLE ROUNDS of the Perl script, 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl',
#	which are recorded in a table per segment in MSA,
#	and performs the PIVOT-TABLE SUMMATION-like operation on those tables.
#	The script was created by taking advantage of the framework of 'tabulate_examine_complex.pl'.
#	[CREATION STARTED on 2020/03/29, ENDED on 2020/03/30.]
#
# tabulate_examine_complex.pl:
#	This perl script reads in the main results of the Perl script 'examine_complex_via_mblks.2.pl',
#	which are recorded in a table per MSA, and convert them into a single table.
#	The script was created by modifying the script 'tabulate_examine_purge_ctrls.pl'.
#	(CREATED on June 1 (Wed), 2016)
#
# tabulate_examine_purge_ctrls.pl:
#	This Perl script reads in the main results of the Perl script 'examine_purge_ctrls.2.pl',
#	which are recorded in a table per MSA, and convert them into a single table.
#	The script was created by modifying the script 'tabulate_examine_purge.pl'.
#	(CREATED on May 14 (Sat), 2016)
#
# tabulate_examine_purge.pl:
#	This Perl script reads in the main results of the Perl script 'examine_purges_via_mblks.2.pl',
#	which are recorded in a table per MSA, and convert them into a single table.
#	The script was created by modifying the script 'tabulate_classify_msa_errors.pl'.
#	(CREATED on May 11 (Wed), 2016)
#
# tabulate_classify_msa_errors.pl:
#	This Perl script reads in the main results of the Perl script 'classify_msa_errors_via_mblks.1.pl',
#	which are recorded in a table per MSA, and convert them into a single table.
#	The script was created by modifying the script 'tabulate_cmp_ref_vs_rec_gconf_stats.pl'.
#	(CREATED on Nov 16 (Sun), 2014)
#
# tabulate_cmp_ref_vs_rec_gconf_stats.pl:
#	This Perl script reads in the main results of the Perl script 'compare_ref_vs_rec_gconf_stats.pl',
#	which are recorded in a table per MSA, and convert them into a single table.
#	The script was created by slightly modifying the script 'tabulate_cmp_ref_vs_rec_cts_indels.pl'.
#	(CREATED on Nov 14 (Fri), 2014)
#
# tabulate_cmp_ref_vs_rec_cts_indels.pl:
#	This Perl script reads in the main results of the Perl script 'compare_ref_vs_rec_indel_cts.pl',
#	which are recorded in a table per MSA, and convert them into a single table.
#	The script was created by slightly modifying the script 'tabulate_cmp_ref_vs_rec_compllimment.pl'.
#	(CREATED on Oct 2 (Thu), 2014)
#
# tabulate_cmp_ref_vs_rec_compllimment.pl:
#	This Perl script reads in the main results of the Perl script 'compare_ref_vs_rec_compllimment.pl',
#	which are recorded in a table per MSA, and convert them into a single table.
#	The script was created by slightly modifying the script 'tabulate_cmp_ref_vs_rec_abs_scoring.pl'.
#	(CREATED on Oct 2 (Thu), 2014)
#
# tabulate_cmp_ref_vs_rec_abs_scoring.pl:
#	This Perl script reads in the main results of the Perl script 'compare_ref_vs_rec_abs_scoring.PhyML.pl',
#	which are recorded in a table per MSA,
#	and convert them into a single table.
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);
use MyArrayManipu qw( compare_arrays );

use MySpecFunc qw( erffc ); # ADDED in creation of 'pivot_table_summations.pl' (on 2020/03/30). (US) #

		# IMPORTED from 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl', in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30). (US) #
use MyCluster_distribution qw(
	batch_initializer
	batch_pr_file_distri_xl_fmt
	);
		# END of "IMPORTED from 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl', in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30). (US)" #

		# ADDED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30). (US2) #
use MyContingency_table qw(
	convert_ctbl_to_2d_distri
	);
		# End of "ADDED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30). (US2)" #


sub extract_seg_set_attributes (\$); # Further MODIFIED in creation of 'tabulate_examine_purge_ctrls.pl'. #
##sub extract_seg_attributes (\$); # MODIFIED in creation of 'tabulate_classify_msa_errors.pl'. #
##sub extract_seg_scores_ref_rec (\$);

$| = 1; # Flush the buffer everytime it receives information to be written through the standard out.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);


	## Test Output Levels. (IMPORTED from 'bch_serial_reconst_dawg_msas_via_mafft.pl' in creation of 'bch_pbs_cmp_compllimment.D.pl'.) ##
	## 0, silent; 1, very brief; 2, pretty brief; 3, rather verbose; 4, verbose; 5, very verbose, >5: extremely verbose. ##

our $TEST_LEVEL = 0;
##our $TEST_LEVEL = 5;

	## Parameters concerning input data files to be read in. ##

=DISABLED in creation of 'pivot_table_summations.pl' (on 2020/03/29).
##	# The following strings label the fields in the input files. #
####our @IN_LABELS = ('Wd_size', '#{wds}', 'indx_wd', 'br_ID', 'br_len', '#{ext_offs}', '#{diffs}_exp', '#{diffs}_obs', 'P-value'); # MODIFIED in creation of 'tabulate_examine_purge_ctrls.pl'. #
##
##our @IN_LABELS = ('Type', 'Diagnosis', '#{gpd_segs}', 'indx_gpd_seg', 'Beg_coord', 'End_coord', '#{clms}', '#{ins}', '#{del}', 'totlen_ins', 'totlen_del', '#{gaps}'); # MODIFIED in creation of 'tabulate_examine_complex.pl'. #
##
##our $CT_INLABELS = @IN_LABELS;
##
=cut # END of "DISABLED in creation of 'pivot_table_summations.pl' (on 2020/03/29)."


		## Boundaries of the subject values (most likely the SRDs) when computing the distribution.
		## BE SURE that they will PERFECTLY MATCH those used in 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl'!!
		##
		## (IMPORTED from 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl', in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30).) (GL) ##

our @BOUNDS_SBJ = (0.005, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.995);

		## END of "(IMPORTED from 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl', in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30).) (GL)" ##



	### Input/Output files/directories ###

my $input_msa_type = '12primates';
#my $input_msa_type = '15mammals';
#####my $input_msa_type = '9fe_mammals';

### $input_msa_type = 'TESTs/Test1a'; # For TESTING the script (Test1a). #


	# $infile1 = "$indir1/$infile_name1",
	# where $indir1 = "$intopdir1/$inmiddir_pref1$id_class$inmiddir_suff1/$inmidpath/$inmiddir_pref2$session_id$inmiddir_suff2/$inbottom_path"
	# with $id_class and $session_ids will be collected later.

##my $intopdir1 = "../${input_msa_type}_lite/Outputs";

		# MODIFIED in creation of 'pivot_table_summations.pl'. (In-D) #
		# => Further MODIFIED in creation of 'pivot_table_summations_distr_srds.pl'. (In-D) #
###my $intopdir1 = "../${input_msa_type}_lite1/Outputs"; # For 'tbl_yyy_obs_zzz.txt'.
###my $intopdir1 = "../${input_msa_type}_lite2/Outputs"; # For 'tbl_yyy_exp_zzz.txt'.
my $intopdir1 = "../${input_msa_type}_lite3/Outputs"; # For 'distributions_www.txt'.
		# End of "MODIFIED in creation of 'pivot_table_summations.pl'. (In-D) #
		# => Further MODIFIED in creation of 'pivot_table_summations_distr_srds.pl'. (In-D)" #

my $inmiddir_pref1 = 'class_';
##my $inmiddir_pref1 = 'hog_d';
my $inmiddir_suff1 = '';
my $inmidpath = 'Outputs';
my $inmiddir_pref2 = 'lclid_';
my $inmiddir_suff2 = '';
my $inbottom_path = '';
####my $infile_name1 = 'tbl_summ_examine_purge.txt'; # MODIFIED in creation of 'tabulate_examine_purge_ctrls.pl'. #
##my $infile_name1 = 'tbl_summ_examine_complex.txt'; # MODIFIED in creation of 'tabulate_examine_complex.pl'. #

		# MODIFIED in creation of 'pivot_table_summations.pl'. (In-F) #
		# => Further MODIFIED in creation of 'pivot_table_summations_distr_srds.pl'. (In-F) #
###my $infile_name1 = 'tbl_assoc_obs_srds_and_alnshears.txt'; # For 'xxx_lite1'.
###my $infile_name1 = 'tbl_assoc_exp_srds_and_alnshears.txt'; # For 'xxx_lite2'.
my $infile_name1 = 'distributions_obs_srds.txt'; # For 'xxx_lite3'.
		# End of "MODIFIED in creation of 'pivot_table_summations.pl'. (In-F) #
		# => Further MODIFIED in creation of 'pivot_table_summations_distr_srds.pl'. (In-F)" #


	# $outfile = "$outdir/$out_pref$type$out_suff"
my $outdir = $input_msa_type;

## $outdir = "TESTs/Test2d_P10"; ### For TESTING the script ###
##  $outdir = "TESTs/Test2"; ### For TESTING the script (Test2). ###

####my $outfile_name = 'tbl_examine_purge_ctrl.txt'; # MODIFIED in creation of 'tabulate_examine_purge_ctrls.pl'. #
##my $outfile_name = 'tbl_examine_complex.txt'; # MODIFIED in creation of 'tabulate_examine_complex.pl'. #

		# MODIFIED in creation of 'pivot_table_summations.pl'. (Out-F) #
###my $outfile_name = 'pv_tbl_summ_assoc_obs_srds_and_alnshears.txt';  # For 'xxx_lite1'.
###my $outfile_name = 'pv_tbl_summ_assoc_exp_srds_and_alnshears.txt';  # For 'xxx_lite2'.
my $outfile_name = 'pv_tbl_distributions_obs_srds.txt';  # For 'xxx_lite3'.
		# End of "MODIFIED in creation of 'pivot_table_summations.pl'. (Out-F)" #


	## Check files/directories ##

foreach my $topdir ($intopdir1) {
    unless (( -d $topdir ) and ( -r $topdir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable TOP INPUT DIRECTORY '$topdir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

=For the predecessor 'tabulate_cmp_ref_vs_rec_abs_scoring.pl'.
    foreach my $type (@TYPES) {
	my $dir = "$topdir/$type";
	unless (( -d $dir) and ( -r $dir)) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable MIDDLE INPUT DIRECTORY '$dir' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
    }
=cut

}

	# ADDED in creation of 'tabulate_cmp_ref_vs_rec_compllimment.pl'. #
my $list_id_classes = list_sessions_ext ($intopdir1, $inmiddir_pref1, $inmiddir_suff1, 'dr');
unless (defined $list_id_classes) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the top input directory '$intopdir1' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

## (@{$list_id_classes}>2) and (@{$list_id_classes} = @{$list_id_classes}[0, 1]); ## For TESTING the script. (Test2) ##

unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
	print STDERR "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$outdir': $? !\n";
	die "$! .\n\n";
    }
}


	### Declare the Beginning of the Script ###

print STDERR "\n<< Program name: $0 >>\n\n";

print STDERR '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n\n";


print STDERR "Input files:  {in_dir1}/$infile_name1,\n";

print STDERR "       where  {in_dir1} = {intopdir1}/$inmiddir_pref1", '{ID_class}', "$inmiddir_suff1/$inmidpath/$inmiddir_pref2", '{session_ID}', "$inmiddir_suff2/$inbottom_path ,\n";
print STDERR "       with {intopdir1} = $intopdir1 ,\n";
print STDOUT "             {ID_class} = ", join (', ', @{$list_id_classes}), " ,\n\n";
print STDERR "       and  {session_ID}'s will be determined in a class-specific manner.\n\n\n";

	# where $indir1 = "$intopdir1/$inmiddir_pref1$id_class$inmiddir_suff1/$inmidpath/$inmiddir_pref2$session_id$inmiddir_suff2/$inbottom_path"

print STDOUT "Output file: {out_dir}/$outfile_name,\n";
print STDOUT "       where {out_dir} = $outdir .\n\n";

## print STDOUT "\n", '# @IN_LABELS = (', join (', ', @IN_LABELS), ").\n\n\n"; ## DISABLED in creation of 'pivot_table_summations.pl' (on 2020/03/29).

				# ADDED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (DC) #

print STDOUT "\n\n<< Boundaries set for the Distributions of the 'Subject' quantities: >>\n\n";

print STDOUT '@BOUNDS_SBJ = (', join (', ', @BOUNDS_SBJ), ") .\n\n\n";

				# End of "ADDED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (DC)" #



	### Main Process ###


		## INITIALIZE some arrays and a hash. ##
		## (ADDED in creation of 'pivot_table_summations.pl' (on 2020/03/29).) (INIT-MAIN) ##

my @header_tags = (); # Stores the field tags at the top of each table.

		# MAJOR MODIFICATION in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (INIT-MAIN) #
##my @pvtable_summations = (); # Stores Pivot-table summations. #
my @pvtable_summations_on_as = (); # Stores Pivot-table summations of distributions of SRDs ON Alignment-Shears. #
my @pvtable_summations_off_as = (); # Stores Pivot-table summations of distributions of SRDs OFF Alignment-Shears. #
		# End of "MAJOR MODIFICATION in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (INIT-MAIN)" #

my @row_labels = (); # Stores the row labels of each table.

my %row_label2indx; # = ($row_label => $indx, ...), which specifies the order that the row-labels appear in the table.

		## END of "(ADDED in creation of 'pivot_table_summations.pl' (on 2020/03/29).) (INIT-MAIN)" ##

=MOVED to below the main outermost foreach-loop, in creation of 'pivot_table_summations.pl' (on 2020/03/29).
my $outfile = "$outdir/$outfile_name";
my $fh_out = open_output_file ($outfile, 1);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
=cut # END of "MOVED to below the main outermost foreach-loop, in creation of 'pivot_table_summations.pl' (on 2020/03/29)."


=DISABLED in creation of 'pivot_table_summations.pl' (on 2020/03/29).

	# MODIFIED in creation of 'tabulate_classify_msa_errors.pl'. #

print $fh_out join ("\t", 'ID_Class', 'Session_ID', 'Segment_ID', @IN_LABELS), "\n\n";

##=For the predecessor 'tabulate_cmp_ref_vs_rec_gconf_stats.pl'.
##print $fh_out join ("\t", 'ID_Class', 'Session_ID', 'Segment_ID');
####print $fh_out join ("\t", 'Type', '#Session_ID', 'segment_ID');
##
##foreach my $in_label (@IN_LABELS) { print $fh_out "\t${in_label}.ref\t${in_label}.rec"; }
##print $fh_out "\n\n";
##=cut
=cut # END of "DISABLED in creation of 'pivot_table_summations.pl' (on 2020/03/29)."

my $ct_in = my $ct_succ = my $ct_fail = 0;
my $totct_sessions = my $totct_segs = 0;
my $totct_out_entries = 0; # ADDED in creation of 'tabulate_examine_purge_ctrls.pl'. #

my $totct_diss = 0; # ADDED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). #

my $ct_errorlines = my $ct_headers = my $ct_totals = 0; # ADDED in creation of 'pivot_table_summations.pl'. #


foreach my $id_class (@{$list_id_classes}) { # Main outermost foreach-loop (over classes). #

   print STDOUT "\n<< Class ID: $id_class >>\n\n";

   my $inmiddir1 = join ('', $intopdir1, '/', $inmiddir_pref1, $id_class, $inmiddir_suff1, '/', $inmidpath);


=For the predecessor 'tabulate_cmp_ref_vs_rec_abs_scoring.pl'.
foreach my $type (@TYPES) {...} # Main outermost foreach-loop (over types) #

    print STDERR "\n<< Type: $type >>\n\n";

    my $inmiddir1 = join ('', $intopdir1, '/', $type, '/', $middir_name1);
    ##my $inmiddir2 = join ('', $intopdir2, '/', $type, '/', $middir_name2);
=cut

    foreach my $dir ($inmiddir1) {
    ##foreach my $dir ($inmiddir1, $inmiddir2) {
	unless (( -d $dir ) and ( -r $dir )) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable MIDDLE INPUT DIRECTORY '$dir' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ;
	}
    }


	# where $indir1 = "$intopdir1/$inmiddir_pref1$id_class$inmiddir_suff1/$inmidpath/$inmiddir_pref2$session_id$inmiddir_suff2/$inbottom_path"

		# MODIFIED in creation of 'tabulate_cmp_ref_vs_rec_compllimment.pl'. #
    my $sessions = list_sessions_ext ($inmiddir1, $inmiddir_pref2, $inmiddir_suff2, 'dr');
    ##my $sessions = list_sessions_ext ($inmiddir1, $indir_pref1, $indir_suff1, 'dr'); ## For the predecessor 'tabulate_cmp_ref_vs_rec_abs_scoring.pl'.
    unless (defined $sessions) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the middle input directory '$inmiddir1' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }

    ##(@{$sessions}>2) and (@{$sessions} = @{$sessions}[0,1]); ## For TESTING the script.
    ## (@{$sessions}>1) and (@{$sessions} = ($sessions->[0])); ## For TESTING the script. (Test2)

    my $ct_sessions = @{$sessions};
    ##$totct_sessions += $ct_sessions;

    print STDERR "\n\n --- Fetched $ct_sessions sessions belonging to the class '$id_class':\n\n";
    ##print STDERR "\n\n --- Fetched $ct_sessions sessions belonging to the type '$type':\n\n";
    print STDERR join (', ', @{$sessions}), " .\n\n";

    foreach my $session (@{$sessions}) { # Middle foreach-loop (over sessions) #

	print STDERR "< Session ID: $session >\n\n";
	$totct_sessions++;

		# where $indir1 = "$intopdir1/$inmiddir_pref1$id_class$inmiddir_suff1/$inmidpath/$inmiddir_pref2$session_id$inmiddir_suff2/$inbottom_path"
	my $indir1 = join ('', $inmiddir1, '/', $inmiddir_pref2, $session, $inmiddir_suff2, '/', $inbottom_path);
	##my $indir1 = join ('', $inmiddir1, '/', $indir_pref1, $session, $indir_suff1);
	##my $indir2 = join ('', $inmiddir2, '/', $indir_pref2, $session, $indir_suff2);

	foreach my $dir ($indir1) {
	##foreach my $dir ($indir1, $indir2) {
	    unless (( -d $dir ) and ( -r $dir )) {
		print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT DIRECTORY '$dir' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }
	}

		# $infile1 = "$indir1/$infile_name1",
	my $infile1 = "$indir1/$infile_name1";

	foreach my $file ($infile1) {
	    unless (( -f $file ) and ( -r $file )) {
		print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }
	}

		## Read-in an input file (table) for each MSA. ##

	##local $/ = "\n\\\\\n\n"; # A non-default input-record-separator #
	local $/ = "\n"; # Use a newline character as an input-record-separator. # 

	my $fh_in1 = open_input_file2 ($infile1);
	unless (defined $fh_in1) {
	    print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the input file '$infile1' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ;
	}

	my $pvtable_summations; # ADDED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M1)

	##my $ct_segs = 0;
	my $ct_diss = 0; # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M1a) #

	##while (my $in = <$fh_in1> ) { # Middle-inner while-loop (over segments) #
	while (my $in_line = <$fh_in1> ) { # Middle-inner while-loop (over input lines) # MODIFIED in creation of 'pivot_table_summations.pl' (on 2020/03/29). #

	    ##unless ($in_line =~ m/^\S/) { next; } # Skip if the line does NOT begin with a non-blank character. # A BUG FIXED on 2020/04/01!!
	    unless ($in_line =~ m/\S/) { next; } # Skip if the line contains NO non-blank characters. #

	    if ($in_line =~ m/^\<\<\</) { next; } # Skip the overall header of the input file. #

			# MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M2) #
	    if ($in_line =~ s/^\<\s*//) {
		$ct_in++;

		if ($in_line =~ m/^ON\s/) {
		    $pvtable_summations = \@pvtable_summations_on_as;

		} elsif ($in_line =~ m/^OFF\s/) {
		    $pvtable_summations = \@pvtable_summations_off_as;

		} else {
		    $in_line =~ s/\n$//;
		    print STDOUT "[ERROR '$program_name'] In the session, '$session', Encountered an UNEXPECTED INPUT LINE:\n";
		    print STDOUT "  '< $in_line' : $! !\n\n";
		    die ;
		}

		next;
	    }
=For the predecessor, 'pivot_table_summations.pl'. (M1)
	    if ($in_line =~ m/^\[\s*The\s+\d+\s*th\s+segment/) {
		$ct_in++;
		next;
	    }
=cut # End of "For the predecessor, 'pivot_table_summations.pl'. (M1)"
			# End of "MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M2)" #

	    if ($in_line =~ m/^\#/) { next; } # Skip the comment line. #


	    $in_line =~ s/\s+$//;

	    ##if ($in_line =~ m/^Branch_ID/) {
	    if ($in_line =~ m/^Intervals/) { # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M3a) #

		$ct_headers++;

		unless (@header_tags>0) {
		    ##@header_tags = split /\t/, $in_line;
		    @header_tags = split /\t\s*/, $in_line; # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M3b) #
		}
		next;

	    } # End of "if ($in_line =~ m/^Intervals/) {...}"
	    ##} # End of "if ($in_line =~ m/^Branch_ID/) {...}"

	    ##my @flds = split /\t/, $in_line;
	    my @flds = split /\t\s*/, $in_line; # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M3c) #
	    my $row_label = shift @flds;
	    $row_label =~ s/\s+$//; ## ADDED for Fixing a BUG (on 2020/04/01). ##

   if ($TEST_LEVEL > 3) {
	print STDERR "\n[ For the line w/ \$row_label = '$row_label': ]\n\n";
	print STDERR '$in_line = ', "\"$in_line\"\n\n";
	print STDERR '@flds = (', join (', ', @flds), ") .\n";
	print STDERR "\n\n";
   } # End of "if ($TEST_LEVEL > 3) {...}"


			## MODIFIED for DEBUGGING 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (Main) ##
	    my $indx;
	    if ($ct_totals < 1) {
			# Initialize the association between $row_label and $indx. #

		$indx = @row_labels;
		$row_labels[$indx] = $row_label;
		$row_label2indx{$row_label} = $indx;

	    } else { # if ($ct_totals >= 1) #
				# Use an already-existing association between $row_label and $indx.

		$indx = $row_label2indx{$row_label};
		unless (defined $indx) {
		    print STDOUT "[CAUTION '$0'] Found a ROW-LABEL, '$row_label', that corresponds to NO INDEX, in the input line:\n";
		    print STDOUT " '$in_line': $! !\n";
		    print STDOUT "       ... Just SKIP the line ...\n\n\n";

		    $ct_errorlines++;
		    next;
		}
	    }

	    my $row_summations = $pvtable_summations->[$indx];

	    if (defined $row_summations) {

		my $ct_flds = @flds;
	        for (my $k=0; $k < $ct_flds; $k++) { $row_summations->[$k] += $flds[$k]; }

	    } else {
		$pvtable_summations->[$indx] = \@flds; # Initialize the $indx th row. #
	    }

=DISABLED (on 2020/04/01), because there seems to be a BUG.
	    ##if ($ct_totals >0) {
	    if ($ct_totals >1) { # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M3d) #
			# The line is regarding the 2nd or later distribution (for either of 'on' and 'off')  to be processed. #

		my $indx = $row_label2indx{$row_label};
		unless (defined $indx) {
		    print STDOUT "[CAUTION '$0'] Found a ROW-LABEL, '$row_label', that corresponds to NO INDEX, in the input line:\n";
		    print STDOUT " '$in_line': $! !\n";
		    print STDOUT "       ... Just SKIP the line ...\n\n\n";

		    $ct_errorlines++;
		    next;
		}

			# Perform the pivot-table summations. #

		my $ct_flds = @flds;
		##my $row_summations = $pvtable_summations[$indx];
		my $row_summations = $pvtable_summations->[$indx]; # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M3e) #

		for (my $k=0; $k < $ct_flds; $k++) { $row_summations->[$k] += $flds[$k]; }

	    } else {
			# The line is still regarding the 1st distribution (for either 'on' or 'off') to be processed. #

			# DEBUGGED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). #
		##my $indx = @row_labels;
		my $indx;
		if ($ct_totals ==0) {
				# Create a new association between $row_label and $indx. #
		   $indx = @row_labels;
		   push @row_labels, $row_label;
		   $row_label2indx{$row_label} = $indx;

		} else { # if ($ct_totals == 1) #
				# Use an already-existing association between $row_label and $indx.
		    $indx = $row_label2indx{$row_label};
		    unless (defined $indx) {
			print STDOUT "[CAUTION '$0'] Found a ROW-LABEL, '$row_label', that corresponds to NO INDEX, in the input line:\n";
			print STDOUT " '$in_line': $! !\n";
			print STDOUT "       ... Just SKIP the line ...\n\n\n";

			$ct_errorlines++;
			next;
		    }
		}
		##$pvtable_summations[$indx] = \@flds;
		$pvtable_summations->[$indx] = \@flds; # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M3f) #
	    }
=cut # END of "DISABLED (on 2020/04/01), because there seems to be a BUG."

			## END of "MODIFIED for DEBUGGING 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (Main)" ##


	    if (uc ($row_label) eq 'TOTAL') {
		##$ct_segs++;
		$ct_diss++; # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M1b) #
		$ct_totals++;
	    }


## my @header_tags = (); # Stores the field tags at the top of each table.
## my @pvtable_summations_on_as = (); # Stores Pivot-table summations of distributions of SRDs ON Alignment-Shears. #
## my @pvtable_summations_off_as = (); # Stores Pivot-table summations of distributions of SRDs OFF Alignment-Shears. #
## my @row_labels = (); # Stores the row labels of each table.
## my %row_label2indx; # = ($row_label => $indx, ...), which specifies the order that the row-labels appear in the table.

## my $ct_errorlines = my $ct_headers = my $ct_totals = 0; # ADDED in creation of 'pivot_table_summations.pl'. #

=For the predecessor, 'tabulate_examine_complex.pl'.

##	    $ct_in++;
##
##			# MODIFIED in creation of 'tabulate_examine_purge_ctrls.pl'. #
##
##	    my ($seg_id, $set_lists_attributes) = extract_seg_set_attributes ($in);
##	    unless ((defined $seg_id) and (defined $set_lists_attributes) ) {
##		(defined $seg_id) or ($seg_id = 'undef');
##		print STDOUT "[ERROR '$program_name'] 'extract_seg_set_attributes' FAILED for the segment '$seg_id' !\n";
##		##$ct_fail_seg++;
##		##next;
##		print STDOUT "[ERROR '$program_name'] $!\n";
##		die ;
##	    }
##
##=For the predecessor 'tabulate_examine_purge.pl'.
##
##	    my ($seg_id, $list_attributes) = extract_seg_attributes ($in);
##	    unless ((defined $seg_id) and (defined $list_attributes) ) {
##		(defined $seg_id) or ($seg_id = 'undef');
##		print STDOUT "[ERROR '$program_name'] 'extract_seg_scores_ref_rec' FAILED for the segment '$seg_id' !\n";
##		##$ct_fail_seg++;
##		##next;
##		print STDOUT "[ERROR '$program_name'] $!\n";
##		die ;
##	    }
##=cut # END of "For the predecessor 'tabulate_examine_purge.pl'."
##
##	    $ct_segs++;
##
##	    my $ct_out_entries = @{$set_lists_attributes};
##
##			# MODIFIED in creation of 'tabulate_examine_purge_ctrls.pl'. #
##	    print STDERR "{ Segment ID: $seg_id (#{entries} = $ct_out_entries) }\n\n";
##	    ##print STDERR "{ Segment ID: $seg_id }\n\n";
##
##	    foreach my $list_attributes (@{$set_lists_attributes}) { # Inner foreach-loop (over the output entries) # The foreach-loop was added in creation of 'tabulate_examine_purge_ctrls.pl'. #
##
##		my @outinfo = ($id_class, $session, $seg_id, @{$list_attributes});
##
##		print $fh_out join ("\t", @outinfo), "\n";
##
##	    }					# End of the inner foreach-loop (over the output entries) # The foreach-loop was added in creation of 'tabulate_examine_purge_ctrls.pl'. #
##
=cut # END of "For the predecessor, 'tabulate_examine_complex.pl'."

	    $ct_succ++;

	    ##$totct_out_entries += $ct_out_entries; # ADDED in creation of 'tabulate_examine_purge_ctrls.pl' # DISABLED in creation of 'pivot_table_summations.pl' (on 2020/03/29). #

	}			# End of the middle-inner while-loop (over segments) #

	##print STDERR "\n --- $ct_segs segments were processed in the session '$session' ---\n\n";
	print STDERR "\n --- $ct_diss distributions were processed in the session '$session' ---\n\n"; # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M1c) #

	close ($fh_in1);

	##$totct_segs += $ct_segs;
	$totct_diss += $ct_diss; # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (M1d) #

	### ($totct_segs >= 2) and last; ### For TESTING the script ###

    } # Middle foreach-loop (over sessions) #

} # End of the main outermost foreach-loop (over classes) #


print STDERR "\n --- All the Input Files Have Been Processed !! ---\n\n"; # (Added in creation of 'pivot_table_summations.pl' on 2020/03/29.) #


		### Examine the contents of @pvtable_summations_on/off_as. ###
		### (ADDED in DEBUGGING 'pivot_table_summations_distr_srds.pl' on 2020/04/01). (Pre-FIN) ##

if ($TEST_LEVEL >2) {

    print STDERR "\n<< Examining the Contents of the Pivot-Table-Summed Distributions & Lables >>\n\n";

    print STDERR '[ Contents of @header_tags: ]', "\n\n";

    print STDERR join ("\t", 'Indx', 'Tag'), "\n\n";

    for (my $i=0; $i <@header_tags; $i++) {
    	my $tag = $header_tags[$i];
    	unless (defined $tag) { $tag = 'undef'; }
    	print STDERR join ("\t", $i, $tag), "\n";
    }
    print STDERR "\n\n";

    print STDERR '[ Contents of @row_labels: ]', "\n\n";

    print STDERR join ("\t", 'Indx_row', 'Label'), "\n\n";

    for (my $i=0; $i <@row_labels; $i++) {
    	my $rlabel = $row_labels[$i];
    	unless (defined $rlabel) { $rlabel = 'undef'; }
    	print STDERR join ("\t", $i, $rlabel), "\n";
    }
    print STDERR "\n\n";


    print STDERR '[ Contents of @pvtable_summations_on_as: ]', "\n\n";

    print STDERR join ("\t", 'Indx_row', '[summations,in,the,row]'), "\n\n";

    for (my $i=0; $i <@pvtable_summations_on_as; $i++) {
	my $row_summations = $pvtable_summations_on_as[$i];
	unless (defined $row_summations){
	    print STDERR join ("\t", $i, 'The_row_is_UNDEFINED'), "\n";
	    next;
	}
	my $ref = ref ($row_summations);
	unless (defined $ref) {
	    print STDERR join ("\t", $i, "NOT_A_Reference,BUT:$row_summations"), "\n";
	    next;
	}
	unless ($ref eq 'ARRAY') {
	    print STDERR join ("\t", $i, "NOT_AN_ARRAY_Reference,BUT_A_${ref}_Reference:$row_summations"), "\n";
	    next;
	}

	my @out = ();
	my $size_row = @{$row_summations};
	for (my $j=0; $j <$size_row; $j++) {
	    my $element = $row_summations->[$j];
	    unless (defined $element) { $element = 'undef'; }
	    $out[$j] = $element;
	}
	my $cnct_out = join (',', @out);
	print STDERR join ("\t", $i, "[$cnct_out]"), "\n";
    }
    print STDERR "\n\n";


    print STDERR '[ Contents of @pvtable_summations_off_as: ]', "\n\n";

    print STDERR join ("\t", 'Indx_row', '[summations,in,the,row]'), "\n\n";

    for (my $i=0; $i <@pvtable_summations_off_as; $i++) {
	my $row_summations = $pvtable_summations_off_as[$i];
	unless (defined $row_summations){
	    print STDERR join ("\t", $i, 'The_row_is_UNDEFINED'), "\n";
	    next;
	}
	my $ref = ref ($row_summations);
	unless (defined $ref) {
	    print STDERR join ("\t", $i, "NOT_A_Reference,BUT:$row_summations"), "\n";
	    next;
	}
	unless ($ref eq 'ARRAY') {
	    print STDERR join ("\t", $i, "NOT_AN_ARRAY_Reference,BUT_A_${ref}_Reference:$row_summations"), "\n";
	    next;
	}

	my @out = ();
	my $size_row = @{$row_summations};
	for (my $j=0; $j <$size_row; $j++) {
	    my $element = $row_summations->[$j];
	    unless (defined $element) { $element = 'undef'; }
	    $out[$j] = $element;
	}
	my $cnct_out = join (',', @out);
	print STDERR join ("\t", $i, "[$cnct_out]"), "\n";
    }
    print STDERR "\n\n";

## my @header_tags = (); # Stores the field tags at the top of each table.
## my @pvtable_summations_on_as = (); # Stores Pivot-table summations of distributions of SRDs ON Alignment-Shears. #
## my @pvtable_summations_off_as = (); # Stores Pivot-table summations of distributions of SRDs OFF Alignment-Shears. #
## my @row_labels = (); # Stores the row labels of each table.

} # END of "if ($TEST_LEVEL >2) {...}"

		### END of "(ADDED in DEBUGGING 'pivot_table_summations_distr_srds.pl' on 2020/04/01). (Pre-FIN)" ##


		## Convert the Pivot-table Summations to Branch-wise distributions. ##
		## (ADDED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30). (Pre-FIN) ##

print STDERR "\n ... Now, Converting the Pivot-table Summations to Branch-wise distributions...\n\n";

my ($fl_tot_row, $fl_tot_clm) = (1, 1); # This MUST Reflect the format of the input (branch-wise) distribution files. #

my $pvtable_summations_distri_on_as  = convert_ctbl_to_2d_distri (@pvtable_summations_on_as, $fl_tot_row, $fl_tot_clm);
my $pvtable_summations_distri_off_as = convert_ctbl_to_2d_distri (@pvtable_summations_off_as, $fl_tot_row, $fl_tot_clm);

		## End of "(ADDED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30). (Pre-FIN)" ##


		# MOVED from above the main outermost foreach-loop, in creation of 'pivot_table_summations.pl' (on 2020/03/29). #
my $outfile = "$outdir/$outfile_name";
##my $fh_out = open_output_file ($outfile, 1);
my $fh_out = open_output_file ($outfile, 0); # MODIFIED on 2020/03/30.
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
		# END of "MOVED from above the main outermost foreach-loop, in creation of 'pivot_table_summations.pl' (on 2020/03/29)." #


			## ADDED in creation of 'pivot_table_summations.pl' (on 2020/03/29). (FIN) ##
			##  => REVISED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30). (FIN) ##

print STDERR "\n ... Now, Printing Out the Results to '$outfile' ...\n\n";


## my @header_tags = (); # Stores the field tags at the top of each table.
## my @pvtable_summations_on_as = (); # Stores Pivot-table summations of distributions of SRDs ON Alignment-Shears. #
## my @pvtable_summations_off_as = (); # Stores Pivot-table summations of distributions of SRDs OFF Alignment-Shears. #
## my @row_labels = (); # Stores the row labels of each table.
## my %row_label2indx; # = ($row_label => $indx, ...), which specifies the order that the row-labels appear in the table.


				## (IMPORTED w/ modifications from 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl', in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30).) (FIN) ##

print $fh_out "<<< Branch-wise Distributions of \"Observed\" SRDS, summed over ALL reconstructed MSAs >>>\n\n\n";

print $fh_out "< ON the \"Alignment-Shear\"s >\n\n\n";

####batch_pr_file_distri_xl_fmt (@brwise_distri_obs_srds_on_as, @BOUNDS_SRD, 'SRD', *${$fh_o_dstr}, "%5.3f", "%8.3f", 0, @clm_labels) ;
##batch_pr_file_distri_xl_fmt (@brwise_distri_obs_srds_on_as, @BOUNDS_SRD, 'SRD', *${$fh_o_dstr}, "%5.3f", "%8.3f", 0, @tags) ; # CORRECTED in DEBUGGING 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/27. #
batch_pr_file_distri_xl_fmt (@{$pvtable_summations_distri_on_as}, @BOUNDS_SBJ, 'SRD', *${$fh_out}, "%5.3f", "%12.3f", 0, @header_tags) ; # => MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30). #

print $fh_out "\n\n\n";


print $fh_out "< OFF the \"Alignment-Shear\"s >\n\n\n";

####batch_pr_file_distri_xl_fmt (@brwise_distri_obs_srds_off_as, @BOUNDS_SRD, 'SRD', *${$fh_o_dstr}, "%5.3f", "%8.3f", 0, @clm_labels) ;
##batch_pr_file_distri_xl_fmt (@brwise_distri_obs_srds_off_as, @BOUNDS_SRD, 'SRD', *${$fh_o_dstr}, "%5.3f", "%8.3f", 0, @tags) ; # CORRECTED in DEBUGGING 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl' on 2020/03/27. #
batch_pr_file_distri_xl_fmt (@{$pvtable_summations_distri_off_as}, @BOUNDS_SBJ, 'SRD', *${$fh_out}, "%5.3f", "%12.3f", 0, @header_tags) ; # => MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30). #

print $fh_out "\n\n";

				## END of "(IMPORTED w/ modifications from 'ingr_correl_alnshear_cb_vs_srd.ver0.6.pl', in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30).) (FIN)" ##

=For the predecessor, 'pivot_table_summations.pl'. (FIN)

print $fh_out "\n[ Results Summed Over All Successfully Processed Segments in All MSAs: ]\n\n";

## my @header_tags = (); # Stores the field tags at the top of each table.
## my @pvtable_summations = (); # Stores Pivot-table summations. #
## my @row_labels = (); # Stores the row labels of each table.
## my %row_label2indx; # = ($row_label => $indx, ...), which specifies the order that the row-labels appear in the table.

print $fh_out join ("\t", @header_tags), "\n\n";

my $ct_rows = @row_labels;


for (my $r=0; $r <$ct_rows; $r++) { # 1st trailing for-loop (over rows of the table). #

	my $row_label = $row_labels[$r];
	my $row_summations = $pvtable_summations[$r];

	if (uc ($row_label) eq 'TOTAL') { print $fh_out "\n"; }

	print $fh_out join ("\t", $row_label, @{$row_summations}), "\n";

} # End of the 1st trailing for-loop (over rows of the table). #

print $fh_out "\n\n";

				# Compute & OUtput the means and standard deviations of SRDs for the branches and total. #
				# (Further ADDED in revision of 'pivot_table_summations.pl' (on 2020/03/30).) (FIN2) #

print $fh_out "\n[ Means and Standard Deviations of SRDs over ALL effective sites: ]\n\n";

print $fh_out join ("\t", $header_tags[0], 'Mean[SRD/site]_on_as', 'Std_Dev[SRD/site]_on_as', 'Mean[SRD/site]_off_as', 'Std_Dev[SRD/site]_off_as', 'Mean[on - off]', 'Std_Error[on - off]', 'Z[on - off]', 'Pv[on - off]'), "\n\n";

for (my $r=0; $r <$ct_rows; $r++) { # 2nd trailing for-loop (over rows of the table). #

	my $row_label = $row_labels[$r];
	if (uc ($row_label) eq 'TOTAL') { print $fh_out "\n"; }

	my ($sum_wgts_on_as, $sum_srds_on_as, $sum_sqsrds_on_as, $sum_wgts_off_as, $sum_srds_off_as, $sum_sqsrds_off_as)
		= @{$pvtable_summations[$r]};

	my $mean_srd_on_as = ($sum_wgts_on_as>0) ? $sum_srds_on_as/$sum_wgts_on_as : 'n/a' ;
	my $var_srd_on_as = ($sum_wgts_on_as>0) ? $sum_sqsrds_on_as/$sum_wgts_on_as - $mean_srd_on_as * $mean_srd_on_as : 'n/a';

	if (($sum_wgts_on_as>0) and ($var_srd_on_as < 0)) {
	    print STDOUT "[CAUTION] A NEGATIVE VALUE of \$var_srd_on_as (=$var_srd_on_as) for the row, '$row_label': $! !!\n";
	    print STDOUT "         .... I will set it to be ZERO ...\n\n";
	    $var_srd_on_as = 0;
	}

	my $std_dev_on_as = ($sum_wgts_on_as>0) ? sqrt ($var_srd_on_as) : 'n/a';

	my $mean_srd_off_as = ($sum_wgts_off_as>0) ? $sum_srds_off_as/$sum_wgts_off_as : 'n/a' ;
	my $var_srd_off_as = ($sum_wgts_off_as>0) ? $sum_sqsrds_off_as/$sum_wgts_off_as - $mean_srd_off_as * $mean_srd_off_as : 'n/a';

	if (($sum_wgts_off_as>0) and ($var_srd_off_as < 0)) {
	    print STDOUT "[CAUTION] A NEGATIVE VALUE of \$var_srd_off_as (=$var_srd_off_as) for the row, '$row_label': $! !!\n";
	    print STDOUT "         .... I will set it to be ZERO ...\n\n";
	    $var_srd_off_as = 0;
	}

	my $std_dev_off_as = ($sum_wgts_off_as>0) ? sqrt ($var_srd_off_as) : 'n/a';


	unless (($sum_wgts_on_as>0) and ($sum_wgts_off_as>0)) {

	    print $fh_out join ("\t", $row_label, $mean_srd_on_as, $std_dev_on_as, $mean_srd_off_as, $std_dev_off_as, 'n/a', 'n/a', 'n/a', 'n/a'), "\n";
	    next;

	} # End of "unless (($sum_wgts_on_as>0) and ($sum_wgts_off_as>0)) {...}"

	my $mean_on_minus_off = $mean_srd_on_as - $mean_srd_off_as;

	##my $se_on_minus_off = $var_srd_on_as/$mean_srd_on_as + $var_srd_off_as/$mean_srd_off_as; # An unbelievable BUG fixed on 2020/03/30.
	my $se_on_minus_off = $var_srd_on_as/$sum_wgts_on_as + $var_srd_off_as/$sum_wgts_off_as;

	$se_on_minus_off = sqrt ($se_on_minus_off);

	unless ($se_on_minus_off > 0) {

	    print $fh_out join ("\t", $row_label, $mean_srd_on_as, $std_dev_on_as, $mean_srd_off_as, $std_dev_off_as, $mean_on_minus_off, $se_on_minus_off, 'n/a', 'n/a'), "\n";
	    next;

	} # End of "unless ($se_on_minus_off > 0) {...}"

	my $z_on_minus_off = $mean_on_minus_off /$se_on_minus_off;

	my $abs_z = $z_on_minus_off;
	if ($abs_z < 0) { $abs_z = - $abs_z; }

	my $pv = erffc ($abs_z/sqrt(2));
	unless (defined $pv) {
	    print STDOUT "[CAUTION '$0'] The subroutine 'erffc' FAILED for the row, '$row_label': $! !\n";
	    $pv = 'n/a';
	}

	print $fh_out join ("\t", $row_label, $mean_srd_on_as, $std_dev_on_as, $mean_srd_off_as, $std_dev_off_as, $mean_on_minus_off, $se_on_minus_off, $z_on_minus_off, $pv), "\n";

} # End of the 2nd trailing for-loop (over rows of the table). #

print $fh_out "\n\n";
				# End of "(Further ADDED in revision of 'pivot_table_summations.pl' (on 2020/03/30).) (FIN2)" #

=cut # END of "For the predecessor, 'pivot_table_summations.pl'. (FIN)"


close ($fh_out);

			## END of "ADDED in creation of 'pivot_table_summations.pl' (on 2020/03/29). (FIN) ##
			##  => REVISED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/03/30). (FIN)" ##


print STDERR "\n\n=== ALL the Processes ENDED !! ===\n\n";

print STDERR "\n#{Inputs} = $ct_in .\n";

			# MODIFIED in creation of 'pivot_table_summations.pl' (on 2020/03/29). #

	# my $ct_errorlines = my $ct_headers = my $ct_totals = 0;
print STDERR "#{headers}  = $ct_headers .\n";
print STDERR "#{totals}   = $ct_totals .\n";
print STDERR "#{Erroneous lines} = $ct_errorlines .\n";
##print STDERR "#{Inputs successfully processed} = $ct_succ .\n";
##print STDERR "#{Inputs failed to be processed} = $ct_fail .\n";
			# END of "MODIFIED in creation of 'pivot_table_summations.pl' (on 2020/03/29)." #

print STDERR "Total#{input sessions} = $totct_sessions .\n";
##print STDERR "Total#{input Segments} = $totct_segs .\n\n";
print STDERR "Total#{input Distributions} = $totct_diss .\n\n"; # MODIFIED in creation of 'pivot_table_summations_distr_srds.pl' (on 2020/04/01). (FIN) #


##print STDERR "Total#{output entries} = $totct_out_entries .\n\n"; # ADDED in creation of 'tabulate_examine_purge_ctrls.pl'. # DISABLED in creation of 'pivot_table_summations.pl' (on 2020/03/29). #

## my $ct_in = my $ct_succ = my $ct_fail = 0;
## my $totct_sessions = my $totct_segs = 0;

unless (0 == $ct_fail) {
	print STDOUT "[ERROR '$program_name'] FAILED to process $ct_fail inputs !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}



	### DEFINITIONS of the Subroutines ###


=For the predecessor, 'tabulate_examine_complex.pl'.

# extract_seg_set_attributes:
#	This subroutine reads in attributes of multiple entries assigned to a segment.
#	It was created by modifying 'extract_seg_attributes'
#	in the Perl script 'tabulate_examine_purge.pl'.
#	(CREATED in creation of 'tabulate_examine_purge_ctrls.pl' on May 15 (Sun), 2016.)
#
# extract_seg_attributes:
#	This subroutine reads in attributes of a segment.
#	The subroutine was created by modifying 'extract_seg_scores_ref_rec'
#	in the Perl script 'tabulate_cmp_ref_vs_rec_gconf_stats.pl'.
#	(CREATED in creation of 'tabulate_classify_msa_errors.pl' on Nov 16 (Sun), 2014.)
#
# extract_seg_scores_ref_rec:
#	This subroutine reads in scores and some other information on
#	the reference and reconstructed MSAs for a segment.
#	The subroutine was created by modifying the subroutine 'extract_seg_scores'
#	in the Perl script 'summary_results_scoring.2D.pl'.
#
# extract_seg_scores:
#	This subroutine reads in IDs, ranks, and scores of the alternative MSAs
#	for a segment.
#
# INPUT: \$in_chunk.
#
# OUTPUT: ($seg_id, \@set_lists_attributes), where
#	$seg_id is the ID of the segment whose record is being extracted;
#	@list_attributes (= @{$set_lists_attributes[$indx_en]}) stores attributes of the $indx_en th entry
#	recorded in the input chunk associated with the segment.
#
sub extract_seg_set_attributes (\$) {

	my ($in_chunk) = @_;

	${$in_chunk} =~ s/^\s+//;
	${$in_chunk} =~ s/\s+\\\\\s+$//;

	my @lines = split /\n+/, ${$in_chunk};

	unless (@lines>=2) { # Further MODIFIED for DEBUGGING 'tabulate_examine_purge_ctrls.pl'. #
	##unless (@lines>=3) {...} # MODIFIED in creation of 'extract_seg_set_attributes'. #
	####unless (@lines==3) {...}
	    print STDERR "[ERROR 'extract_seg_set_attributes'] Found an ANOMALOUS INPUT CHUNK: $! !\n";
	    print STDERR "'${$in_chunk}' .\n\n";
	    return (undef, undef, undef);
	}

	my $title = shift @lines;
	my ($seg_id) = ($title =~ m/Segment\s+ID\:\s+(\S+)/ );
	unless (defined $seg_id) {
	    print STDERR "[ERROR 'extract_seg_set_attributes'] Found an ANOMALOUS TITLE: $! !\n";
	    print STDERR "'$title' .\n\n";
	    return (undef, undef, undef);
	}

	my $in_labels = shift @lines;
	unless (defined $in_labels) {
	    print STDERR "[ERROR 'extract_seg_set_attributes'] Found NO LINE of labels for the segment '$seg_id': $! !\n";
	    print STDERR "'$in_labels' .\n\n";
	    return ($seg_id, undef, undef);
	}

	if ((@lines == 1) and ($lines[0] =~ m/^\.\.\./)) {

	    print "[INFORMATION 'extract_seg_set_attributes'] The segment '$seg_id' has NO ENTRY for the reason: '$lines[0]' : $! !\n\n";

	    return ($seg_id, []);
	}

	my @set_lists_attributes = (); # ADDED in creation of 'extract_seg_set_attributes'. #

	while (@lines>0) { # Main while-loop (over entries). # The while-loop was ADDED in creation of 'extract_seg_set_attributes'. #

	   my $line = shift @lines;
	   my @list_attributes = split /\s+/, $line;

	   unless (@list_attributes >= $CT_INLABELS -2) {
		print STDERR "[CAUTION 'extract_seg_set_attributes'] Found an ANOMALOUS LINE for the segment '$seg_id': $! !\n";
		print STDERR "[CAUTION 'extract_seg_set_attributes'] '$line' .\n\n";

			# MODIFIED in creation of 'extract_seg_set_attributes'. #
		print STDOUT "[CAUTION 'extract_seg_set_attributes']   ... I will NOT store this 'entry' ...\n\n";
		next;
	       ##return ($seg_id, undef);
	   }

	   foreach my $i ($CT_INLABELS-2, $CT_INLABELS-1) { (defined $list_attributes[$i]) or ($list_attributes[$i] = 'None'); }

	   push @set_lists_attributes, \@list_attributes; # ADDED in creation of 'extract_seg_set_attributes'. #

	}		# End of the main while-loop (over entries). # The while-loop was ADDED in creation of 'extract_seg_set_attributes'. #

	return ($seg_id, \@set_lists_attributes); # MODIFIED in creation of 'extract_seg_set_attributes'. #
	##return ($seg_id, \@list_attributes);

}		# End of the subroutine 'extract_seg_set_attributes'. #

=cut # END of "For the predecessor, 'tabulate_examine_complex.pl'."


=For the predecessor 'tabulate_examine_purge.pl'.

	    ## ($seg_id, $list_attributes) = extract_seg_attributes ($in);

# extract_seg_attributes:
#	This subroutine reads in attributes of a segment.
#	The subroutine was created by modifying 'extract_seg_scores_ref_rec'
#	in the Perl script 'tabulate_cmp_ref_vs_rec_gconf_stats.pl'.
#	(CREATED in creation of 'tabulate_classify_msa_errors.pl' on Nov 16 (Sun), 2014.)
#
# extract_seg_scores_ref_rec:
#	This subroutine reads in scores and some other information on
#	the reference and reconstructed MSAs for a segment.
#	The subroutine was created by modifying the subroutine 'extract_seg_scores'
#	in the Perl script 'summary_results_scoring.2D.pl'.
#
# extract_seg_scores:
#	This subroutine reads in IDs, ranks, and scores of the alternative MSAs
#	for a segment.
#
# INPUT: \$in_chunk.
#
# OUTPUT: ($seg_id, \@list_attributes), where
#	$seg_id is the ID of the segment whose record is being extracted;
#	@list_attributes stores attributes of the segment recorded in the input chunk.
#
sub extract_seg_attributes (\$) {

	my ($in_chunk) = @_;

	${$in_chunk} =~ s/^\s+//;
	${$in_chunk} =~ s/\s+\\\\\s+$//;

	my @lines = split /\n+/, ${$in_chunk};

	unless (@lines==3) {
	    print STDERR "[ERROR 'extract_seg_attributes'] Found an ANOMALOUS INPUT CHUNK: $! !\n";
	    print STDERR "'${$in_chunk}' .\n\n";
	    return (undef, undef, undef);
	}
	my $title = shift @lines;
	my ($seg_id) = ($title =~ m/Segment\s+ID\:\s+(\S+)/ );
	unless (defined $seg_id) {
	    print STDERR "[ERROR 'extract_seg_attributes'] Found an ANOMALOUS TITLE: $! !\n";
	    print STDERR "'$title' .\n\n";
	    return (undef, undef, undef);
	}

	my $in_labels = shift @lines;
	unless (defined $in_labels) {
	    print STDERR "[ERROR 'extract_seg_attributes'] Found NO LINE of labels for the segment '$seg_id': $! !\n";
	    print STDERR "'$in_labels' .\n\n";
	    return ($seg_id, undef, undef);
	}

	my $line = shift @lines;
	my @list_attributes = split /\s+/, $line;

	unless (@list_attributes >= $CT_INLABELS -2) {
	    print STDERR "[CAUTION 'extract_seg_attributes'] Found an ANOMALOUS LINE for the segment '$seg_id': $! !\n";
	    print STDERR "'$line' .\n\n";
	    return ($seg_id, undef);
	}

	foreach my $i ($CT_INLABELS-2, $CT_INLABELS-1) { (defined $list_attributes[$i]) or ($list_attributes[$i] = 'None'); }

	return ($seg_id, \@list_attributes);

}		# End of the subroutine 'extract_seg_attributes'. #

=cut # END of "For the predecessor 'tabulate_examine_purge.pl'."


=For the predecessor 'tabulate_cmp_ref_vs_rec_gconf_stats.pl'.
# extract_seg_scores_ref_rec:
#	This subroutine reads in scores and some other information on
#	the reference and reconstructed MSAs for a segment.
#	The subroutine was created by modifying the subroutine 'extract_seg_scores'
#	in the Perl script 'summary_results_scoring.2D.pl'.
#
# extract_seg_scores:
#	This subroutine reads in IDs, ranks, and scores of the alternative MSAs
#	for a segment.
#
# INPUT: \$in_chunk.
#
# OUTPUT: ($seg_id, \@list_scores_ref, \@list_scores_rec), where
#	$seg_id is the ID of the segment whose record is being extracted;
#	@list_scores_ref/rec = ($tot_score, $gap_score, $avct_indels, $avct_null_clms)
#	for the subject segment of the reference/reconstructed MSA, with
#	$tot_score being the total score (log-likelihood), which is the summation of $gap_score and the score for the substitution pattern,
#	$gap_score being the score (log-likelihood) of the gap-pattern,
#	$avct_indels is the average count of the indels (in the parsimonious scenario),
#	$avct_null_clms is the average count of the null columns, each of which consists solely of gap characters.
#
sub extract_seg_scores_ref_rec (\$) {

	my ($in_chunk) = @_;

	${$in_chunk} =~ s/^\s+//;
	${$in_chunk} =~ s/\s+\\\\\s+$//;

	my @lines = split /\n+/, ${$in_chunk};
	unless (@lines==4) {
	##unless (@lines>2) {
	    print STDERR "[ERROR 'extract_seg_scores_ref_rec'] Found an ANOMALOUS INPUT CHUNK: $! !\n";
	    print STDERR "'${$in_chunk}' .\n\n";
	    return (undef, undef, undef);
	}
	my $title = shift @lines;
	my ($seg_id) = ($title =~ m/Segment\s+ID\:\s+(\S+)/ );
	unless (defined $seg_id) {
	    print STDERR "[ERROR 'extract_seg_scores_ref_rec'] Found an ANOMALOUS TITLE: $! !\n";
	    print STDERR "'$title' .\n\n";
	    return (undef, undef, undef);
	}

	my $in_labels = shift @lines;
	unless (defined $in_labels =~ m!^Ref/Rec!) {
	##unless (defined $in_labels =~ m/^\#Rank/) {
	    print STDERR "[ERROR 'extract_seg_scores_ref_rec'] Found an ANOMALOUS line of labels for the segment '$seg_id': $! !\n";
	    print STDERR "'$in_labels' .\n\n";
	    return ($seg_id, undef, undef);
	}

	my %label2scores; # ($label => \@list_scores, ...) where $label eq 'Ref' or 'Rec'. 
			  # with @list_scores = ($tot_score, $gap_score, $avct_indels, $avct_null_clms).
	while (my $line = shift @lines) { # Foreach-loop (over segments) #

	    my @list_scores = split /\s+/, $line;

	    unless (@list_scores == 6) { # MODIFIED in creation of 'tabulate_cmp_ref_vs_rec_gconf_stats.pl'. #
	    ##unless (@list_scores == 5) {...}

		print STDERR "[CAUTION 'extract_seg_scores_ref_rec'] Found an ANOMALOUS LINE for the segment '$seg_id': $! !\n";
		print STDERR "'$line' .\n\n";
		next;
	    }

	    my $label = shift @list_scores;

	    $label2scores{$label} = \@list_scores;

	}		# End of the foreach-loop (over segments) #

	my $list_scores_ref = $label2scores{'Ref'};
	my $list_scores_rec = $label2scores{'Rec'};

	return ($seg_id, $list_scores_ref, $list_scores_rec);

}		# End of the subroutine 'extract_seg_scores_ref_rec' #
=cut

