package MyANEX_MainX;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyANEX_MainX.pm:
#	This package is a collection of EXTRA subroutines
#	that perform some additional main processes of ANEX.
#	[CREATION STARTED on 2019/09/22, ENDED on 2019/mm/dd.]
#

use strict;
use warnings;


use MyTreeManipu_II qw( );
		# output_tree
		# extract_sgltree_from_file output_tree 

use MyArrayManipu_num qw(
	diff_nmarray
	);

use MyArrayManipu qw(
	compare_arrays
	); # Added in REVISION of 'list_cii_candidates' on 2019/09/30. #

=Set aside for later use.
use MyTaxonomy qw(
	fetch_common_ancestors
	);
	# if_ancestor fetch_ancestry calculate_abs_depth
	# fetch_ext_offsprings_nr fetch_all_offsprings_nr
=cut # End of "Set aside for later use."

use MyANEX_Supple qw(
	mk_orgs_bds_gb_coords
	check_consistency_bw_msa_and_clmset
	);

use MyTreeMap_indels_spt_odr2 qw(
	cal_init_psm_cands_for_sgl_msa
	cmp_gblocks3
	bch_br_parsimonize_dollo_indel_histories
	);

use MyGapPatternTopology qw(
	mk_inter_block_relations
	mk_lists_gblocks_w_spec_bounds
	cts_vincluding_vincluded_onx_gblks
	mk_list_gblock_sizes
	);

use MyANEX_Main qw(
	print_out_gblock_configs
	print_out_remng_attr_gblks
	);

use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	cmp_nmarrays
	partition_clms_into_gpbs
	partition_rowset_into_monophyl_sets
	list_cii_candidates
	filter_cii_candidates

	reverse_cii_on_clmset
	rearrange_clms_preserving_hs
	mk_list_gblocks
	reverse_spec_cii_in_wd

	print_out_set_clms
	);

sub cmp_nmarrays (\@\@);
sub partition_clms_into_gpbs (\@\%$);
sub partition_rowset_into_monophyl_sets (\@\@\%\%\%\%$\%);
sub list_cii_candidates (\@\%\%$\%\%$\%$);
sub filter_cii_candidates (\@\@\@); # ADDED on 2019/09/29.

##sub reverse_cii_on_clmset (\@$$\@);
sub reverse_cii_on_clmset (\@$$$\@); # CORRECTED on 2019/09/26.

sub rearrange_clms_preserving_hs (\@\%$);
sub mk_list_gblocks (\@$$\@\%\%$\%\@\@\%\%\%\%\%);
sub reverse_spec_cii_in_wd (\@$$$$\@\@\%\%$\%\@\@\%\%\%\%\%\@$); # ADDED on 2019/09/26.

sub print_out_set_clms (\@\*);

	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;



	# This specifies the level of tests on the MAIN subroutine, 'reverse_spec_cii_in_wd'. #
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)

our $TEST_LEVEL_MAIN = 1;
##our $TEST_LEVEL_MAIN = 2;
##our $TEST_LEVEL_MAIN = 3;
##our $TEST_LEVEL_MAIN = 5;

	# For 'list_cii_candidates'. #
our $TEST_LEVEL_lciic = 0;
#our $TEST_LEVEL_lciic = 6;

	# For 'mk_list_gblocks'. #
our $TEST_LEVEL_mklgb = 0;
#our $TEST_LEVEL_mklgb = 6;



### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


		### DEFINITIONS of the Subroutines. ###

# cmp_nmarrays:
#	This basic subroutine compares two arrays of numbers.
#	[CREATED on 2019/09/23.]
#
# ASSUMPTION: The elements of each array are sorted in ascending order.
#
# INPUTS: (\@array1, \@array2), where
#
#	@array1 and @array2 are arrays of numbers.
#
# OUTPUT: $sgn, where
#
#	$sgn = -1 if @array1 is "smaller than" @array2,
#	     = +1 if @array1 is "greater than" @array2,
#	     =  0 if the two arrays equal.
#
sub cmp_nmarrays (\@\@) {

	my ($array1, $array2) = @_;

	my $size1 = @{$array1};
	my $size2 = @{$array2};

	my $minsize = ($size1 < $size2) ? $size1 : $size2 ;

	for (my $i=0; $i < $minsize; $i++) { # Main for-loop (over comparable elements). #
	    my $ele1 = $array1->[$i];
	    my $ele2 = $array2->[$i];
	    if ($ele1 < $ele2) {
		return -1;
	    } elsif ($ele1 > $ele2) {
		return +1;
	    }
	} # End of the main for-loop (over comparable elements). #

		# If the order cannot be determined yet, compare their sizes. #

	if ($size1 < $size2) {
	    return -1;
	} elsif ($size1 > $size2) {
	    return +1;
	} else {
	    return 0;
	}

}	# End of the subroutine, 'cmp_nmarrays'. #


# partition_clms_into_gpbs:
#	This satellite subroutine of 'list_cii_candidates' partitions the set of MSA columns into gap-pattern blocks.
#	[CREATED on 2019/09/23.]
#
# INPUTS: (\@set_clms, \%set_msa_rows2br_up_or_lw, $char_abs), where
#
#	@set_clms is the input set of alignment columns, that is,
#	@column = @{$set_clms[$c]} specifies the $c th column;
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_clms is an output of 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$char_abs is a character representing the "absence" state (i.e., a gap);
#
# OUTPUTS: (\@set_gpblocks), where
#
#	@set_gpblocks lists the gap-pattern blocks, each of which is the block of contiguous columns with the same set of MSA rows w/ the "presence" state;
#	More precisely,
#	@{$set_gpblocks[$gpb]} = ($le_clm2, $re_clm2, \@set_pres_rows, $class) provides information on the $gpb th gap-pattern block, where
#	$le_clm2 and $re_clm2 are the indices of the columns (in @set_clms above) at the left- and right-ends of the block,
#	@set_pres_rows stores the indices of the MSA rows with the "presence" state,
#	$class = 'F', 'E', 'Mono', and 'CII-cand', if the column is filled with the "presence" state, if it is filled with the "absence" state, if the rows with the "presence" state constitute a monophyletic group (or its complement), and if none of the above apply, respectively.
#
#
sub partition_clms_into_gpbs (\@\%$) {

	my ($set_clms, $set_msa_rows2br_up_or_lw, $char_abs) = @_;

	my $ct_clms = @{$set_clms};
	unless ($ct_clms >0) {
	    return []; # If there are no columns, there should be no gap-pattern blocks, either.
	}

	my $ct_rows = @{$set_clms->[0]};

	my @set_gpblocks = ();

	for (my $c = 0; $c < $ct_clms; $c++) { # Main for-loop (over the columns). #

	    my $column = $set_clms->[$c];
	    my @set_pres_rows = ();
	    for (my $r=0; $r <$ct_rows; $r++) { # Inner for-loop (over the rows). #
		if ($column->[$r] ne $char_abs) { push @set_pres_rows, $r; }
	    }
	    my $ct_pres_rows = @set_pres_rows;

	    my $class;
	    if ($ct_pres_rows == 0) {
		$class = 'E';
	    } elsif ($ct_pres_rows == $ct_rows) {
		$class = 'F';
	    } else {
		my $cnct_pres_rows = join (',', @set_pres_rows);
		if (defined $set_msa_rows2br_up_or_lw->{$cnct_pres_rows}) {
		    $class = 'Mono';
		} else {
 		    $class = 'CII-cand';
		}
	    }

	    if (@set_gpblocks>0) {
		my $prev_gpb = $set_gpblocks[$#set_gpblocks];
		my ($le_clm2_prev, $re_clm2_prev, $set_pres_rows_prev, $class_prev) = @{$prev_gpb};

	# @{$set_gpblocks[$gpb]} = ($le_clm2, $re_clm2, \@set_pres_rows, $class) provides information on the $gpb th gap-pattern block, where

		my $if_eq_gp = 0;
		if ($class eq $class_prev) {
		    if (($class eq 'E') or ($class eq 'F')) {
			$if_eq_gp = 1;
		    } elsif ($ct_pres_rows == @{$set_pres_rows_prev}) {

			my $sgn = cmp_nmarrays (@set_pres_rows, @{$set_pres_rows_prev});
			if ($sgn == 0) { $if_eq_gp = 1; }
		    }
		}

		if ($if_eq_gp) {
		    $prev_gpb->[1] = $c;
		    next;
		}
	    } # End of "if (@set_gpblocks>0) {...}"

		  # Create a new gap-pattern block. #

	    push @set_gpblocks, [$c, $c, \@set_pres_rows, $class];

	} # End of the main for-loop (over the columns). #

	return \@set_gpblocks;

}	# End of the subroutine, 'partition_clms_into_gpbs'. #


# partition_rowset_into_monophyl_sets:
#	This satellite subroutine of 'list_cii_candidates' partitions
#	a set of MSA rows into a set of maximal monophyletic sets.
#	[CREATED on 2019/09/23.]
#
#	[REVISED on 2019/09/27]:
#	To the while-loop for descending the tree, ADDED a measure to prevent the loop from continuing forever.
#
# INPUTS: (\@set_sbj_rows, \@set_all_rows, \%node2exoffs_msa_rows, \%set_msa_rows2br_up_or_lw, 
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq), where
#
#	@set_sbj_rows stores the indices of the subject rows in the MSA;
#	@set_all_rows stores the indices of all the rows in the MSA;
#
#	%node2exoffs_msa_rows = ($node_id => \@exoffs_msa_rows, ...) for all nodes,
#	with @exoffs_msa_rows stores the indices of all MSA rows corresponding to the external offsprings of $node_id;
#	[This \%node2exoffs_msa_rows is the MAIN OUTPUT of 'wrapper_mk_node2exoffs_msa_rows' in "MyTreeMap_sppl.pm";
#	it is also the output of 'mk_node2exoffs_msa_rows' in "MyTreeMap_sppl.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$char_abs is a character representing the "absence" state (i.e., a gap);
#
#	%node2pa = ($node_id => $parent_id, ...) for the (non-root) nodes;
#	%node2ch = ($node_id => \@child_node_ids, ...) for the internal nodes;
#	$topnode_id is the ID of the top-node (i.e., root);
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for the branches with equivalents;
#	
# OUTPUTS: (\@set_monophyl_sets_sbj_rows, \@set_br_up_or_lw), where
#
#	@{$set_monophyl_sets_sbj_rows[$ms]} stores the indices of the subject MSA rows in the $ms th monophyletic set;
#	@{$set_br_up_or_lw[$ms]} = ($br, $up_or_lw) indicates that the $ms th monophyletic set of the subject MSA rows is on the $up_or_lw side of the branch $br;
#
sub partition_rowset_into_monophyl_sets (\@\@\%\%\%\%$\%) {

	my ($set_sbj_rows, $set_all_rows, $node2exoffs_msa_rows, $set_msa_rows2br_up_or_lw, 
	    $node2pa, $node2ch, $topnode_id, $br2eq) = @_;

	unless (@{$set_sbj_rows}>0) {
		# If the subject set is empty, it should be partitioned into NO monophyletic sets.
	    return ([], []); 
	}

	my @cp_set_sbj_rows = @{$set_sbj_rows};

	my $remng_set_sbj_rows = \@cp_set_sbj_rows;

	my @set_monophyl_sets_sbj_rows = my @set_br_up_or_lw = ();

	while (@{$remng_set_sbj_rows}>0) { # Main outer while-loop. #

	    my $cnct_remng_sbj_rows = join (',', @{$remng_set_sbj_rows});
	    my $br_up_or_lw_remng = $set_msa_rows2br_up_or_lw->{$cnct_remng_sbj_rows};
	    if (defined $br_up_or_lw_remng) {
		push @set_monophyl_sets_sbj_rows, $remng_set_sbj_rows;
		push @set_br_up_or_lw, $br_up_or_lw_remng;
		last;
	    }

	    my $seed_row = $remng_set_sbj_rows->[0];
	    my $curr_monophyl_set = [$seed_row];
	    my ($curr_br, $curr_up_or_lw) = @{$set_msa_rows2br_up_or_lw->{$seed_row}};

	    while ($curr_up_or_lw eq 'L') { # 1st middle while-loop. #

		my $pa = $node2pa->{$curr_br};
		if ($pa == $topnode_id) {
		    my $eqbr = $br2eq->{$curr_br};
		    if (defined $eqbr) {
			$curr_br = $eqbr;
			$curr_up_or_lw = 'U';
			last;
		    }

		    my $sibs = $node2ch->{$pa};
		    foreach my $sib (@{$sibs}) { # 1st inner foreach-loop. #
			if ($sib == $curr_br) { next; }
			my $exoffs_ud_sib = $node2exoffs_msa_rows->{$sib};
			my ($common, $must_be_empty, $exoffs_ov_sib) = diff_nmarray (@{$exoffs_ud_sib}, @{$set_all_rows});
			my ($common2, $ex_remng, $ex2) = diff_nmarray (@{$remng_set_sbj_rows}, @{$exoffs_ov_sib});
			if (@{$ex2} == 0) {
			    $curr_br = $sib;
			    $curr_up_or_lw = 'U';
			    $curr_monophyl_set = $exoffs_ov_sib;
			    last;
			}

		    } # End of "foreach my $sib (@{$sibs}) { # 1st inner foreach-loop. # ...}"

		    last; # This is necessary for avoiding an infinite-loop. #

		} else {

		    my $exoffs_ud_pa = $node2exoffs_msa_rows->{$pa};
		    my ($common2, $ex_remng, $ex2) = diff_nmarray (@{$remng_set_sbj_rows}, @{$exoffs_ud_pa});
		    if (@{$ex2} == 0) {
			$curr_br = $pa;
			$curr_monophyl_set = $exoffs_ud_pa;
		    } else {
			last; # The current monophyletic set canNOT be expanded any longer.
		    }

		} # End of "if ($pa == $topnode_id) {...} else {...}"

	    }  # End of "while ($curr_up_or_lw eq 'L') { # 1st middle while-loop. # ...}"

	   while ($curr_up_or_lw eq 'U') { # 2nd middle while-loop. #

		my $children = $node2ch->{$curr_br};
		unless (@{$children}>0) { last; } # End the loop if $curr_br is an external node. #

		if (@{$children} == 1) { # Go down the tree automatically. # (Added in minor revision on 2019/09/28.)
		    $curr_br = $children->[0];
		    next;
		}

		my $if_unchanged = 1; # Added in REVISION on 2019/09/28. (A) #
		foreach my $ch (@{$children}) { # 2nd inner foreach-loop. #
		    my $exoffs_ud_ch = $node2exoffs_msa_rows->{$ch};
		    my ($common, $must_be_empty, $exoffs_ov_ch) = diff_nmarray (@{$exoffs_ud_ch}, @{$set_all_rows});
		    my ($common2, $ex_remng, $ex2) = diff_nmarray (@{$remng_set_sbj_rows}, @{$exoffs_ov_ch});
		    if (@{$ex2} == 0) {
		        $curr_br = $ch;
		        $curr_monophyl_set = $exoffs_ov_ch;
			    # Added in REVISION on 2019/09/28. (B) #
			$if_unchanged = 0;
			last;
			    # End of "Added in REVISION on 2019/09/28. (B)" #
		    }
		} # End of "foreach my $ch (@{$children}) { # 2nd inner foreach-loop. # ...}"

		if ($if_unchanged) { last; } # If the monophyletic set remains unchanged, leave the loop. # (Added in REVISION on 2019/09/28. (C)) #

	   } # End of "while ($curr_up_or_lw eq 'U') { # 2nd middle while-loop. # ...}"

	   my $cnct_curr_monophyl_set = join (',', @{$curr_monophyl_set});
	   my $br_up_or_lw_curr = $set_msa_rows2br_up_or_lw->{$cnct_curr_monophyl_set};
	   unless (defined $br_up_or_lw_curr) {
		print STDOUT "[ERROR 'partition_rowset_into_monophyl_sets'] IMPOSSIBLE!! The current monophyletic set, ($cnct_curr_monophyl_set), is NOT associated with any (branch, up_or_lw) pair: $! !\n\n";
		return (undef, undef);
	   }

	   push @set_monophyl_sets_sbj_rows, $curr_monophyl_set;
	   push @set_br_up_or_lw, $br_up_or_lw_curr;

 			# UPDATE the remaining set of subject rows. #

	   my ($common3, $must_be_empty3, $remng_set_sbj_rows_new) = diff_nmarray (@{$curr_monophyl_set}, @{$remng_set_sbj_rows});
	   $remng_set_sbj_rows = $remng_set_sbj_rows_new;

	} # End of "while ($curr_up_or_lw eq 'L') { # Main outer while-loop. # ...}"

	return (\@set_monophyl_sets_sbj_rows, \@set_br_up_or_lw);

}	# End of the subroutine, 'partition_rowset_into_monophyl_sets'. #


# list_cii_candidates:
#	This subroutine creates a list of candidate regions for a(n) "(i-)CII",
#	or a(n) (incomplete-) collapse of independent insertions, in an input MSA.
#	[CREATION STARTED on 2019/09/22, ENDED on 2019/09/23.]
#
#	[REVISED on 2019/09/30]: Now, the subroutine extends each candidate region in BOTH directions
#	when it is flanked by gap-pattern blocks whose set of "presence" rows is included in that of the candidate region.
#
# INPUTS: (\@set_clms, \%node2exoffs_msa_rows, \%set_msa_rows2br_up_or_lw, $char_abs, 
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq,
#	$maxct_monophyl_sets), where
#
#	@set_clms is the input set of alignment columns, that is,
#	@column = @{$set_clms[$c]} specifies the $c th column;
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_clms is an output of 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#
#	%node2exoffs_msa_rows = ($node_id => \@exoffs_msa_rows, ...) for all nodes,
#	with @exoffs_msa_rows stores the indices of all MSA rows corresponding to the external offsprings of $node_id;
#	[This \%node2exoffs_msa_rows is the MAIN OUTPUT of 'wrapper_mk_node2exoffs_msa_rows' in "MyTreeMap_sppl.pm";
#	it is also the output of 'mk_node2exoffs_msa_rows' in "MyTreeMap_sppl.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$char_abs is a character representing the "absence" state (i.e., a gap);
#
#	%node2pa = ($node_id => $parent_id, ...) for the (non-root) nodes;
#	%node2ch = ($node_id => \@child_node_ids, ...) for the internal nodes;
#	$topnode_id is the ID of the top-node (i.e., root);
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for the branches with equivalents;
#	
#	$maxct_monophyl_sets is the largest number of monophyletic sets of MSA-rows
#	that each CII-candidate are allowed to have;
#
#
# OUTPUTS: (\@list_cii_cands, \@set_gpblocks), where
#
#	@list_cii_cands lists the CII-candidate regions identified;
#	More precisely,
#	@{$list_cii_cands[$cc]} = ($le_clm, $re_clm, \@set_monophyl_sets_pres_rows, \@set_br_up_or_lw, $le_gpb, $re_gpb) provides information on the $cc th CII-candidate region, where
#	$le_clm and $re_clm are the IDs of the columns (in @set_clms above) at the left- and right-ends of the candidate region,
#	@{$set_monophyl_sets_pres_rows[$ms]} stores the indices of the MSA rows (with the "presence" state) in the $ms th monophyletic set;
#	@{$set_br_up_or_lw[$ms]} = ($br, $up_or_lw) indicates that the $ms th monophyletic set of the MSA rows (with the "presence" state) is on the $up_or_lw side of the branch $br,
#	$le_gpb and $re_gpb are the indices of the gap-pattern blocks (in @set_gpblocks below) at the left- and right-ends of the candidate region;
#
#	[NOTE that @set_monophyl_sets_msa_rows and @set_br_up_or_lw MUST be of the same size,
#	and that their size MUST be <= $maxct_monophyl_sets];
#
#	@set_gpblocks lists the gap-pattern blocks, each of which is the block of contiguous columns with the same set of MSA rows w/ the "presence" state;
#	More precisely,
#	@{$set_gpblocks[$gpb]} = ($le_clm2, $re_clm2, \@set_pres_rows, $class) provides information on the $gpb th gap-pattern block, where
#	$le_clm2 and $re_clm2 are the indices of the columns (in @set_clms above) at the left- and right-ends of the block,
#	@set_pres_rows stores the indices of the MSA rows with the "presence" state,
#	$class = 'F', 'E', 'Mono', and 'CII-cand', if the column is filled with the "presence" state, if it is filled with the "absence" state, if the rows with the "presence" state constitute a monophyletic group (or its complement), and if none of the above apply, respectively.
#
sub list_cii_candidates (\@\%\%$\%\%$\%$) {

	my ($set_clms, $node2exoffs_msa_rows, $set_msa_rows2br_up_or_lw, $char_abs,
	    $node2pa, $node2ch, $topnode_id, $br2eq,
	    $maxct_monophyl_sets) = @_;

	my $ct_clms = @{$set_clms};
	unless ($ct_clms>0) {
	    return ([], []);
	}

	my $ct_rows = @{$set_clms->[0]};

   (print STDOUT "< Subroutine 'list_cii_candidates' (w/ #{columns}=$ct_clms, #{rows}=$ct_rows, \$maxct_monophyl_sets = $maxct_monophyl_sets) >\n\n") if ($TEST_LEVEL_lciic >0);

	my @set_all_rows = (0 .. $ct_rows-1);

# INPUTS: (\@set_clms, \%node2exoffs_msa_rows, \%set_msa_rows2br_up_or_lw, $char_abs, 
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq,
#	$maxct_monophyl_sets), where

	my ($set_gpblocks) = partition_clms_into_gpbs (@{$set_clms}, %{$set_msa_rows2br_up_or_lw}, $char_abs);
	unless (defined $set_gpblocks) {
	    print STDOUT "[ERROR 'list_cii_candidates'] The subroutine 'partition_clms_into_gpbs' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $ct_gpbs = @{$set_gpblocks};

   (print STDOUT "\n ... The subroutine 'partition_clms_into_gpbs' Ended (w/ #{gap-pattern blocks}=$ct_gpbs) ...\n\n") if ($TEST_LEVEL_lciic > 1);

	my @list_cii_cands = ();
	my %cnct_pres_rows2last_cii_cand;

	for (my $gpb=0; $gpb < $ct_gpbs; $gpb++) { # 1st Main Outer for-loop (over gap-pattern blocks). #

	    my ($le_clm2, $re_clm2, $set_pres_rows, $class) = @{$set_gpblocks->[$gpb]};

#	@{$set_gpblocks[$gpb]} = ($le_clm2, $re_clm2, \@set_pres_rows, $class) provides information on the $gpb th gap-pattern block, where

	    unless ($class eq 'CII-cand') { next; }

	    my $cnct_pres_rows = join (',', @{$set_pres_rows});

	    my ($set_monophyl_sets_pres_rows, $set_br_up_or_lw) = partition_rowset_into_monophyl_sets (@{$set_pres_rows}, @set_all_rows, %{$node2exoffs_msa_rows}, %{$set_msa_rows2br_up_or_lw}, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	    unless (defined $set_monophyl_sets_pres_rows) {
		print STDOUT "[ERROR 'list_cii_candidates'] The subroutine 'partition_rowset_into_monophyl_sets' FAILED on the $gpb th gap-pattern block (w/ ", '@{$set_pres_rows}', " = ($cnct_pres_rows): $! !\n\n";
		return (undef, $set_gpblocks);
	    }
	    my $ct_monophyl_sets = @{$set_monophyl_sets_pres_rows};

   (print STDOUT "\n ... The subroutine 'partition_rowset_into_monophyl_sets' Ended on the $gpb th gap-pattern block, whose ",  '@{$set_pres_rows}', " is ($cnct_pres_rows),   (w/ #{monophyletic sets}=$ct_monophyl_sets) ...\n\n") if ($TEST_LEVEL_lciic >1);

	    if ($ct_monophyl_sets > $maxct_monophyl_sets) {
		print STDOUT "[WARNING 'list_cii_candidates'] The number of monophyletic sets (=$ct_monophyl_sets) of the $gpb th gap-pattern block is LARGER THAN the UPPER-BOUND (=$maxct_monophyl_sets) ! ... Thus, this block will NOT be included in the Final CII-candidates ...  \n\n";
		next;
	    }

	    my $indx_last_cii_cand_w_eq_gp = $cnct_pres_rows2last_cii_cand{$cnct_pres_rows};
	    unless (defined $indx_last_cii_cand_w_eq_gp) {
			# Create a new CII-candidate. #

		push @list_cii_cands, [$le_clm2, $re_clm2, $set_monophyl_sets_pres_rows, $set_br_up_or_lw, $gpb, $gpb];
		$cnct_pres_rows2last_cii_cand{$cnct_pres_rows} = $#list_cii_cands;
		next;
	    }

	    my $last_cii_cand_w_eq_gp = $list_cii_cands[$indx_last_cii_cand_w_eq_gp];
	    my ($le_clm_last, $re_clm_last, $set_monophyl_sets_pres_rows_last, $set_br_up_or_lw_last, $le_gpb_last, $re_gpb_last) = @{$last_cii_cand_w_eq_gp};

	    my $if_merged = 1;
	    for (my $gpb2 = $re_gpb_last+1; $gpb2 < $gpb; $gpb2++) { # Inner for-loop (over intervening gap-pattern blocks). #

		my ($le_clm22, $re_clm22, $set_pres_rows2, $class2) = @{$set_gpblocks->[$gpb2]};
		if ($class2 eq 'F') {
		    $if_merged = 0;
		    last;
		} elsif ($class2 eq 'E') {
		    next;
		}

		my ($common, $ex1, $ex2) = diff_nmarray (@{$set_pres_rows}, @{$set_pres_rows2});
		unless (defined $common) {
		    print STDOUT "[ERROR 'list_cii_candidates'] The subroutine 'diff_nmarray' FAILED on the gap-pattern block pair: ($gpb, $gpb2): $! !\n\n";
		    return (undef, $set_gpblocks);
		}

		if (@{$ex2}>0) { # The intervening block has extra "presence" rows.
		    $if_merged = 0;
		    last;
		}

	    } # End of the inner for-loop (over intervening gap-pattern blocks). #

	    if ($if_merged) {
			# Merge the current gap-pattern block into the last CII-candidate. #
		$last_cii_cand_w_eq_gp->[1] = $re_clm2;
		$last_cii_cand_w_eq_gp->[5] = $gpb;
	    } else {
			# Create a new CII-candidate. #
		push @list_cii_cands, [$le_clm2, $re_clm2, $set_monophyl_sets_pres_rows, $set_br_up_or_lw, $gpb, $gpb];
		$cnct_pres_rows2last_cii_cand{$cnct_pres_rows} = $#list_cii_cands;
	    }


#	@{$list_cii_cands[$cc]} = ($le_clm, $re_clm, \@set_monophyl_sets_pres_rows, \@set_br_up_or_lw, $le_gpb, $re_gpb) provides information on the $cc th CII-candidate region, where


	} # End of the 1st Outer Main for-loop (over gap-pattern blocks). #

	my $ct_cii_cands = @list_cii_cands;

   (print STDOUT "\n#{CII-cands}_initial = $ct_cii_cands .\n\n\n") if ($TEST_LEVEL_lciic > 1);

			# Attempt to Extend each CII-candidates into both directions;
			# If the attempt succeeds,  ADD the new candidate to the output set.#
			# (ADDED in REVISION on 2019/09/30.) #

	my @extended_cii_cands = ();
	for (my $cc = 0; $cc < $ct_cii_cands; $cc++) { # 2nd Outer Main for-loop (over CII-candidates). #

	     
	    my ($le_clm, $re_clm, $set_monophyl_sets_pres_rows, $set_br_up_or_lw, $le_gpb, $re_gpb) = @{$list_cii_cands[$cc]};

#	@{$list_cii_cands[$cc]} = ($le_clm, $re_clm, \@set_monophyl_sets_pres_rows, \@set_br_up_or_lw, $le_gpb, $re_gpb) provides information on the $cc th CII-candidate region, where

	    my $set_pres_rows = $set_gpblocks->[$le_gpb][2];

			# Attempt to extend the region to the left. #

	    my ($new_le_clm, $new_le_gpb) = ($le_clm, $le_gpb);
	    for (my $gpb= $le_gpb-1; $gpb >= 0; $gpb--) { # 1st inner for-loop. #

		my ($le_clm2, $re_clm2, $set_pres_rows2, $class2) = @{$set_gpblocks->[$gpb]};
		if ($class2 eq 'F') {
		   last; # A gapless block will always END the extension.
		} elsif ($class2 eq 'E') {
		   next; # Just skip the null block. #
		}

		my ($common, $ex1, $ex2) = diff_nmarray (@{$set_pres_rows}, @{$set_pres_rows2});
		unless (defined $common) {
		    print STDOUT "[ERROR 'list_cii_candidates'] The subroutine 'diff_nmarray' FAILED on the $gpb th gap-pattern block against the $cc th CII-candidate: $! !\n\n";
		    return (undef, $set_gpblocks);
		}

		if (@{$ex2}>0) { last; } # Skip if some "presence" rows of the glock do NOT belong to the set of "presence" rows of the CII-candidate. #

		$new_le_clm = $le_clm2;
		$new_le_gpb = $gpb;
	    } # End of the 1st inner for-loop. #

			# Attempt to extend the region to the right. #

	    my ($new_re_clm, $new_re_gpb) = ($re_clm, $re_gpb);
	    for (my $gpb= $re_gpb+1; $gpb < $ct_gpbs; $gpb++) { # 2nd inner for-loop. #

		my ($le_clm2, $re_clm2, $set_pres_rows2, $class2) = @{$set_gpblocks->[$gpb]};
		if ($class2 eq 'F') {
		   last; # A gapless block will always END the extension.
		} elsif ($class2 eq 'E') {
		   next; # Just skip the null block. #
		}

		my ($common, $ex1, $ex2) = diff_nmarray (@{$set_pres_rows}, @{$set_pres_rows2});
		unless (defined $common) {
		    print STDOUT "[ERROR 'list_cii_candidates'] The subroutine 'diff_nmarray' FAILED on the $gpb th gap-pattern block against the $cc th CII-candidate: $! !\n\n";
		    return (undef, $set_gpblocks);
		}

		if (@{$ex2}>0) { last; } # Skip if some "presence" rows of the glock do NOT belong to the set of "presence" rows of the CII-candidate. #

		$new_re_clm = $re_clm2;
		$new_re_gpb = $gpb;
	    } # End of the 2nd inner for-loop. #

	    if (($new_le_clm < $le_clm) or ($re_clm < $new_re_clm)) {
		push @extended_cii_cands, [$new_le_clm, $new_re_clm, $set_monophyl_sets_pres_rows, $set_br_up_or_lw, $new_le_gpb, $new_re_gpb];
	    ## my ($le_clm, $re_clm, $set_monophyl_sets_pres_rows, $set_br_up_or_lw, $le_gpb, $re_gpb) = @{$list_cii_cands[$cc]};
	    }

	} # End of the 2nd Outer Main for-loop (over CII-candidates). #

	my $ct_ext_cii_cands = @extended_cii_cands;

   (print STDOUT "\n#{Extended CII-candidates} = $ct_ext_cii_cands .\n\n\n") if ($TEST_LEVEL_lciic >1);

	if ($ct_ext_cii_cands>0) {
	    my @indices_srt = (0, 1); my @methods_srt = ('+', '+');

	    @list_cii_cands = sort { compare_arrays (@{$a}, @{$b}, @indices_srt, @methods_srt) } (@list_cii_cands, @extended_cii_cands);
	    $ct_cii_cands = @list_cii_cands;
	}

			# END of "(ADDED in REVISION on 2019/09/30.)" #


   (print STDOUT "\n< Subroutine 'list_cii_candidates' Ended (w/ #{CII-candidates}_final=$ct_cii_cands, #{gap-pattern blocks}=$ct_gpbs) />\n\n") if ($TEST_LEVEL_lciic >0);

	return (\@list_cii_cands, $set_gpblocks);

}	# End of the subroutine, 'list_cii_candidates'. # 


# filter_cii_candidates:
#	This subroutine receives a list of CII-candidate regions,
#	which was typically output by 'list_cii_candidates',
#	and weeds out those candidates that are quite likely to have resulted from other causes than CII,
#	by examining the configurations of gap-blocks encompassing each candidate.
#	[CREATED on 2019/09/29.]
#
#	[NOTE (on 2019/09/29)]:
#	This version is extremely simplistic;
#	Hopefully, the future version will be able to CORRECTLY handle
#	(nearly) all possible cases of LIKELY CIIs.
#
#	[ [NOTE (on 2019/09/29)]: It seems that \@inter_block_relations was redundant,
#	because the desired horizontal relationships between the gap-blocks
#	are possible only when the blocks are non-interfering or siblings. ] ... For the 1st Attempt!!
#
# INPUTS: (\@list_cii_cands, \@bds_gblocks, \@inter_block_relations), where
#
#	@list_cii_cands lists the CII-candidate regions identified;
#	More precisely,
#	@{$list_cii_cands[$cc]} = ($le_clm, $re_clm, \@set_monophyl_sets_pres_rows, \@set_br_up_or_lw, $le_gpb, $re_gpb) provides information on the $cc th CII-candidate region, where
#	$le_clm and $re_clm are the IDs of the columns (in @set_clms above) at the left- and right-ends of the candidate region,
#	@{$set_monophyl_sets_pres_rows[$ms]} stores the indices of the MSA rows (with the "presence" state) in the $ms th monophyletic set;
#	@{$set_br_up_or_lw[$ms]} = ($br, $up_or_lw) indicates that the $ms th monophyletic set of the MSA rows (with the "presence" state) is on the $up_or_lw side of the branch $br,
#	$le_gpb and $re_gpb are the indices of the gap-pattern blocks (in @set_gpblocks below) at the left- and right-ends of the candidate region;
#	[NOTE that @set_monophyl_sets_msa_rows and @set_br_up_or_lw MUST be of the same size,
#	and that their size MUST be <= $maxct_monophyl_sets];
#	[Typically, \@list_cii_cands is a main output of 'list_cii_candidates'];
#
#	@{$bds_gblocks[$gb]} = ($le_gb, $re_gb) gives the left- and right-ends of the $gb th gap-block;
#	[Typically, \@bds_gblocks is a main output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$inter_block_relations[$gb1]->[$gb2] gives the relationship of the $gb2 th gap-block to the $gb1 th one;
#	[Typically, \@inter_block_relations is the main output of 'mk_inter_block_relations(2)' in "MyGapPatternTopology.pm"];
#
# OUTPUT: (\@selected_cii_cands), where
#
#	@selected_cii_cands lists the CII-candidates selected by this subroutine;
#	It conforms to the same format as the input argument, @list_cii_cands;
#
sub filter_cii_candidates (\@\@\@) {

	my ($list_cii_cands, $bds_gblocks, $inter_block_relations) = @_;

	my $ct_ccands = @{$list_cii_cands};
	my $ct_gbs = @{$bds_gblocks};

	my @selected_cii_cands = ();

	for (my $cc = 0; $cc < $ct_ccands; $cc++) { # Main Outer for-loop. #

	    my $info_cii_cand = $list_cii_cands->[$cc];
	    my ($le_ccand, $re_ccand) = @{$info_cii_cand}[0,1];

	    my @hequal_gbs = ();
	    for (my $gb = 0; $gb < $ct_gbs; $gb++) { # 1st Middle for-loop. #

		my ($le_gb, $re_gb) = @{$bds_gblocks->[$gb]};

		if (($le_gb == $le_ccand) and ($re_ccand == $re_gb)) {
		    push @hequal_gbs, $gb;
		}
	    } # End of the 1st Middle for-loop. #

	    my $ct_heq_gbs = @hequal_gbs;
	    if ($ct_heq_gbs < 2) {
		next;
	    }

	    my $if_ninf_pair = 0;

	    for (my $i = 0; $i < $ct_heq_gbs-1; $i++) { # 2nd Middle for-loop. #

		my $gb1 = $hequal_gbs[$i];
		##my ($le_gb1, $re_gb1) = @{$bds_gblocks->[$gb1]};
		my $rels_w_gb1 = $inter_block_relations->[$gb1];

		for (my $j = $i+1; $j < $ct_heq_gbs; $j++) { # Inner for-loop. #
		    my $gb2 = $hequal_gbs[$j];
		    ##my ($le_gb2, $re_gb2) = @{$bds_gblocks->[$gb2]};
		    my $rel = $rels_w_gb1->[$gb2];

		    if ($rel eq 'NIF') {
			$if_ninf_pair = 1;
			last;
		    }

		} # End of the Inner for-loop. #

		if ($if_ninf_pair) { last; }

	    } # End of the 2nd Middle for-loop. #

	    if ($if_ninf_pair) {
		push @selected_cii_cands, $info_cii_cand;
	    }

	} # End of the Main Outer for-loop. #


	return \@selected_cii_cands;

}	# End of the subroutine, 'filter_cii_candidates'. #

=The 1st Attempt.
# filter_cii_candidates:
#	This subroutine receives a list of CII-candidate regions,
#	which was typically output by 'list_cii_candidates',
#	and weeds out those candidates that are quite likely to have resulted from other causes than CII,
#	by examining the configurations of gap-blocks encompassing each candidate.
#	[CREATED on 2019/09/29.]
#
#	[NOTE (on 2019/09/29)]: It seems that \@inter_block_relations was redundant,
#	because the desired horizontal relationships between the gap-blocks
#	are possible only when the blocks are non-interfering or siblings.
#
# INPUTS: (\@list_cii_cands, \@bds_gblocks, \@inter_block_relations), where
#
#	@list_cii_cands lists the CII-candidate regions identified;
#	More precisely,
#	@{$list_cii_cands[$cc]} = ($le_clm, $re_clm, \@set_monophyl_sets_pres_rows, \@set_br_up_or_lw, $le_gpb, $re_gpb) provides information on the $cc th CII-candidate region, where
#	$le_clm and $re_clm are the IDs of the columns (in @set_clms above) at the left- and right-ends of the candidate region,
#	@{$set_monophyl_sets_pres_rows[$ms]} stores the indices of the MSA rows (with the "presence" state) in the $ms th monophyletic set;
#	@{$set_br_up_or_lw[$ms]} = ($br, $up_or_lw) indicates that the $ms th monophyletic set of the MSA rows (with the "presence" state) is on the $up_or_lw side of the branch $br,
#	$le_gpb and $re_gpb are the indices of the gap-pattern blocks (in @set_gpblocks below) at the left- and right-ends of the candidate region;
#	[NOTE that @set_monophyl_sets_msa_rows and @set_br_up_or_lw MUST be of the same size,
#	and that their size MUST be <= $maxct_monophyl_sets];
#	[Typically, \@list_cii_cands is a main output of 'list_cii_candidates'];
#
#	@{$bds_gblocks[$gb]} = ($le_gb, $re_gb) gives the left- and right-ends of the $gb th gap-block;
#	[Typically, \@bds_gblocks is a main output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$inter_block_relations[$gb1]->[$gb2] gives the relationship of the $gb2 th gap-block to the $gb1 th one;
#	[Typically, \@inter_block_relations is the main output of 'mk_inter_block_relations(2)' in "MyGapPatternTopology.pm"];
#
# OUTPUT: (\@selected_cii_cands), where
#
#	@selected_cii_cands lists the CII-candidates selected by this subroutine;
#	It conforms to the same format as the input argument, @list_cii_cands;
#
sub filter_cii_candidates (\@\@\@) {

	my ($list_cii_cands, $bds_gblocks, $inter_block_relations) = @_;

	my $ct_ccands = @{$list_cii_cands};
	my $ct_gbs = @{$bds_gblocks};

	my @selected_cii_cands = ();

	for (my $cc = 0; $cc < $ct_ccands; $cc++) { # Main Outer for-loop. #

	    my $info_cii_cand = $list_cii_cands->[$cc];
	    my ($le_ccand, $re_ccand) = @{$info_cii_cand}[0,1];

	    my @hincluding_gbs = ();
	    for (my $gb = 0; $gb < $ct_gbs; $gb++) { # 1st Middle for-loop. #

		my ($le_gb, $re_gb) = @{$bds_gblocks->[$gb]};

		unless (($le_gb <= $le_ccand) and ($re_ccand <= $re_gb)) {
		    next;
		}
		if (($le_gb == $le_ccand) and ($re_ccand == $re_gb)) { next; }

		push @hincluding_gbs, $gb;

	    } # End of the 1st Middle for-loop. #

	    my $ct_hincl_gbs = @hincluding_gbs;
	    if ($ct_hincl_gbs < 2) {
		push @selected_cii_cands, $info_cii_cand;
		next;
	    }

	    my $if_causing_pair = 0;

	    for (my $i = 0; $i < $ct_hincl_gbs-1; $i++) { # 2nd Middle for-loop. #

		my $gb1 = $hincluding_gbs[$i];
		my ($le_gb1, $re_gb1) = @{$bds_gblocks->[$gb1]};
		##my $rels_w_gb1 = $inter_block_relations->[$gb1];

		for (my $j = $i+1; $j < $ct_hincl_gbs; $j++) { # Inner for-loop. #
		    my $gb2 = $hincluding_gbs[$j];
		    my ($le_gb2, $re_gb2) = @{$bds_gblocks->[$gb2]};
		    ##my $rel = $rels_w_gb1->[$gb2];

		    if ($le_gb1 < $le_gb2) {
			if ($re_gb1 < $re_gb2) {
			    if (($le_gb2 == $le_ccand) and ($re_gb1 == $re_ccand)) {
				$if_causing_pair = 1;
				last;
			    }
			}
		    } elsif ($le_gb2 < $le_gb1) {
			if ($re_gb2 < $re_gb1) {
			    if (($le_gb1 == $le_ccand) and ($re_gb2 == $re_ccand)) {
				$if_causing_pair = 1;
				last;
			    }
			}
		    }

		} # End of the Inner for-loop. #

		if ($if_causing_pair) { last; }

	    } # End of the 2nd Middle for-loop. #

	    if ($if_causing_pair) { next; } # In this case, the CII-candidate is NOT likely to be genuine. #

	    push @selected_cii_cands, $info_cii_cand;

	} # End of the Main Outer for-loop. #


	return \@selected_cii_cands;

}	# End of the subroutine, 'filter_cii_candidates'. #
=cut # END of "The 1st Attempt."


# reverse_cii_on_clmset:
#	This subroutine "reverse"s a suspected CII (collapse of independent insertions)
#	on a set of columns, which represents an input MSA.
#	[CREATED on 2019/09/24.]
#
# INPUTS: (\@set_clms, $char_abs, $le_cii, $re_cii, \@set_monophyl_sets_pres_rows), where
#	
#	@set_clms is the input set of alignment columns, that is,
#	@column = @{$set_clms[$c]} specifies the $c th column;
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_clms is an output of 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#
#	$char_abs is a character representing the "absence" state (i.e., gap);
#
#	$le_cii an $re_cii are the indices of the columns at the left-end and right-end (inclusive), respectively, of the region where a CII is suspected;
#
#	@set_monophyl_sets_pres_rows stores some monophyletic sets of rows with the "presence" state;
#	More precisely,
#	@{$set_monophyl_sets_pres_rows[$ms]} stores the indices of the MSA rows (with the "presence" state) in the $ms th monophyletic set;
#
#	[Typically, $le_cii, $re_cii and \@set_monophyl_sets_pres_rows are components of an element of
#	@list_cii_cands, which is a main output of 'list_cii_candidates' above];
#
# OUTPUTS: (\@new_set_clms), where
#
#	@new_set_clms is the OUTPUT set of alignment columns AFTER the suspected CII is "reverse"d;
#	It conforms to the same format as the input @set_clms does;
#
sub reverse_cii_on_clmset (\@$$$\@) {

	my ($set_clms, $char_abs, $le_cii, $re_cii, $set_monophyl_sets_pres_rows) = @_;

	my $ct_clms = @{$set_clms};
	unless ($ct_clms >0) {
	    return []; # If the input MSA has NO columns, the output has NO columns, either.
	}
	my $ct_rows = @{$set_clms->[0]};

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }

	my $ct_mnp_sets = @{$set_monophyl_sets_pres_rows};


	my @set_clms_af_rvcii = ();
	for (my $s=0; $s < $ct_mnp_sets; $s++) { # Main outer for-loop (over the monophyletic sets). #

	    my $mnp_set_pres_rows = $set_monophyl_sets_pres_rows->[$s];

	    my @set_clms_rvcii = ();
	    for (my $c = $le_cii; $c <= $re_cii; $c++) { # Main middle for-loop (over the columns in the CII-suspect). #
		my $old_clm = $set_clms->[$c];
		my @new_clm = @null_clm;

		my $ct_pres = 0;
		foreach my $pres_row (@{$mnp_set_pres_rows}) { # Main inner for-loop. #
		    my $res = $old_clm->[$pres_row];
		    if ($res eq $char_abs) { next; }
		    $new_clm[$pres_row] = $res;
		    $ct_pres++;
		} # End of the main inner for-loop. #

		unless ($ct_pres>0) { next; } # Skip if there is no "presence" states. #

		push @set_clms_af_rvcii, \@new_clm ;
		
	    } # End of the Main middle for-loop (over the columns in the CII-suspect). #

	} # End of the Main outer for-loop (over the monophyletic sets). #

	my @set_clms_left = ();
	for (my $c =0; $c < $le_cii; $c++) { # 2nd outer for-loop. #
	    my @cp_clm = @{$set_clms->[$c]};
	    push @set_clms_left, \@cp_clm;
	} # End of the 2nd outer for-loop. #

	my @set_clms_right = ();
	for (my $c = $re_cii+1; $c < $ct_clms; $c++) { # 3rd outer for-loop. #
	    my @cp_clm = @{$set_clms->[$c]};
	    push @set_clms_right, \@cp_clm;
	} # End of the 3rd outer for-loop. #

	my @new_set_clms = (@set_clms_left, @set_clms_af_rvcii, @set_clms_right);

	return \@new_set_clms ;

}	# End of the subroutine, 'reverse_cii_on_clmset'. #


# rearrange_clms_preserving_hs:
#	This subroutine rearranges columns in an input MSA
#	according to a specific rule,
#	while preserving the homology structure of the MSA.
#	In the new set of gap-pattern blocks, contiguous blocks with an identical gap-pattern will be merged without fail.
#	[CREATED on 2019/09/24.]
#
#	[NOTE]
#	This subroutine was created based on the (Step II-2) in "preprocess_msa_dawg.alpha.pl",
#	which is at the "Sample_Scripts/" sub-directory of the package, "ComplLiMment_P".
#
# INPUTS: (\@old_set_clms, \%set_msa_rows2br_up_or_lw, $char_abs), where
#
#	@old_set_clms is the input set of alignment columns, that is,
#	@column = @{$old_set_clms[$c]} specifies the $c th column in the input MSA;
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_clms is an output of 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$char_abs is a character representing the "absence" state (i.e., a gap);
#
#
# OUTPUTS: (\@new_set_clms, \@new_set_gpblocks, \@new_indices_old_gpbs, \@old_indices_intm_gpbs, \@new_indices_intm_gpbs), where
#
#	@new_set_clms is the OUTPUT set of alignment columns AFTER the REARRANGEMENT;
#	It conforms to the same format as @set_clms does;
#
#	@new_set_gpblocks lists the gap-pattern blocks, each of which is the block of contiguous columns with the same set of MSA rows w/ the "presence" state, IN THE OUTPUT MSA (AFTER the REARRANGEMENT and merger);
#	More precisely,
#	@{$new_set_gpblocks[$gpb]} = ($le_clm2, $re_clm2, \@set_pres_rows, $class) provides information on the $gpb th gap-pattern block, where
#	$le_clm2 and $re_clm2 are the indices of the columns (in @set_clms above) at the left- and right-ends of the block,
#	@set_pres_rows stores the indices of the MSA rows with the "presence" state,
#	$class = 'F', 'E', 'Mono', and 'CII-cand', if the column is filled with the "presence" state, if it is filled with the "absence" state, if the rows with the "presence" state constitute a monophyletic group (or its complement), and if none of the above apply, respectively.
#
#	@new_indices_old_gpbs provides correspondence between the gap-pattern blocks in the input MSA
#	and those in the output MSA;
#	More precisely,
#	$ngpb = $new_indices_old_gpbs[$ogpb] is the index in @new_set_gpblocks assigned to the $ogpb th gap-pattern block in @old_set_gpblocks(, which is only available as an interim output);
#
#	@old_indices_intm_gpbs provides correspondence between the old gap-pattern blocks and the intermediate ones, which mainly indicates rearrangements among the old gap-pattern blocks;
#
#	@new_indices_intm_gpbs provides correspondence between the new gap-pattern blocks and the intermediate ones, which mainly indicates mergers among the intermediate gap-pattern blocks;
#
sub rearrange_clms_preserving_hs (\@\%$) {

	my ($old_set_clms, $set_msa_rows2br_up_or_lw, $char_abs) = @_;

	my $ct_clms_old = @{$old_set_clms};

	# For 'rearrange_clms_preserving_hs'. #
our $TEST_LEVEL_rearr_c_phs = 0;
#our $TEST_LEVEL_rearr_c_phs = 6;

   (print STDOUT "< Subroutine 'rearrange_clms_preserving_hs' (w/ #{columns}_OLD=$ct_clms_old) >\n\n") if ($TEST_LEVEL_rearr_c_phs >0);

	my ($old_set_gpblocks) = partition_clms_into_gpbs (@{$old_set_clms}, %{$set_msa_rows2br_up_or_lw}, $char_abs);
	unless (defined $old_set_gpblocks) {
	    print STDOUT "[ERROR 'rearrange_clms_preserving_hs'] The subroutine 'partition_clms_into_gpbs' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef);
	}

	my $ct_gpbs_old = @{$old_set_gpblocks};

   (print STDOUT "\n ... The subroutine 'rearrange_clms_preserving_hs' Ended (w/ #{gap-pattern blocks}_OLD=$ct_gpbs_old) ...\n\n") if ($TEST_LEVEL_rearr_c_phs >1);

	my @old_indices_intm_gpbs = (0 .. $ct_gpbs_old -1);
	my @new_indices_old_gpbs = ();

			# (Step 1): Rearrange ONLY the INDICES of the gap-pattern blocks. #

	my $ct_reorderings = 1; # Initialize the counter.

	while ($ct_reorderings > 0) { # 1st Main Outer while-loop (iteration). #

	    $ct_reorderings = 0; # Reset the counter. #

	    my @new_old_indices_intm_gpbs = (shift @old_indices_intm_gpbs);
	    while (@old_indices_intm_gpbs>0) { # Inner while-loop (over old gap-pattern blocks). #

		my $indx_ogpb_curr = shift @old_indices_intm_gpbs;
		my ($pres_rows_curr, $class_curr) = @{$old_set_gpblocks->[$indx_ogpb_curr]}[2,3];
		if ($class_curr eq 'E') {
		    $new_indices_old_gpbs[$indx_ogpb_curr] = undef;
		    next; # Discard a null gap-block. #
		} elsif ($class_curr eq 'F') {
			# Keep a gapless block as it is. #
		    push @new_old_indices_intm_gpbs, $indx_ogpb_curr;
		    next;
		}

		my $indx_ogpb_prev = pop @new_old_indices_intm_gpbs;

		my ($pres_rows_prev, $class_prev) = @{$old_set_gpblocks->[$indx_ogpb_prev]}[2,3];
		if ($class_prev eq 'F') {
			# Keep the order with a gapless block. #
		    push @new_old_indices_intm_gpbs, $indx_ogpb_prev, $indx_ogpb_curr;
		    next;
		}

		my ($shared, $ex_curr, $ex_prev) = diff_nmarray (@{$pres_rows_curr}, @{$pres_rows_prev});
		if ((@{$shared}>0) or ($pres_rows_prev->[0] > $pres_rows_curr->[0])) {

			# Keep the order. #
		    push @new_old_indices_intm_gpbs, $indx_ogpb_prev, $indx_ogpb_curr;
		} else {
			# Swap the orders of the two blocks. #
		    push @new_old_indices_intm_gpbs, $indx_ogpb_curr, $indx_ogpb_prev;
		    $ct_reorderings++;
		}

	    } # End of the inner while-loop (over old gap-pattern blocks). #

			# Update the set of (indices of) gap-pattern blocks. #
	    @old_indices_intm_gpbs = @new_old_indices_intm_gpbs;

	} # End of the 1st Main Outer while-loop (iteration). #

	my $ct_gpbs_intm = @old_indices_intm_gpbs;


			# (Step 2): According to the result of Step 1,
			#   Actually rearrange the gap-pattern blocks
			#   (, and merge some contiguous blocks if necessary).

	my @new_set_clms = my @new_set_gpblocks = ();

	my @new_indices_intm_gpbs = ();

	for (my $igpb = 0; $igpb < $ct_gpbs_intm; $igpb++) { # 2nd Main outer for-loop (over intermediate gap-pattern blocks). #

	    my $ogpb = $old_indices_intm_gpbs[$igpb];
	    
	    my $le_clm_new = scalar (@new_set_clms);
	    my ($le_clm_old, $re_clm_old, $set_pres_rows_curr, $class_curr) = @{$old_set_gpblocks->[$ogpb]};
#	@{$set_gpblocks[$gpb]} = ($le_clm2, $re_clm2, \@set_pres_rows, $class) provides information on the $gpb th gap-pattern block, where

	    push @new_set_clms, @{$old_set_clms}[$le_clm_old .. $re_clm_old];
	    my $re_clm_new = scalar (@new_set_clms) -1;

	    my $if_merge = 0;
	    if (@new_set_gpblocks>0) {

		my $prev_gpblock = $new_set_gpblocks[$#new_set_gpblocks];
	        my ($le_clm_prev, $re_clm_prev, $set_pres_rows_prev, $class_prev) = @{$prev_gpblock};

		if ($class_curr eq $class_prev) {
		    if ($class_curr eq 'F') {
				# If both blocks are gapless. #
			$if_merge = 1;
		    } else {
				# If both blocks have an identical gap-pattern. #
			my $sgn = cmp_nmarrays (@{$set_pres_rows_curr}, @{$set_pres_rows_prev});
			if ($sgn == 0) { $if_merge = 1; }
		    }
		}
		if ($if_merge) {
			# Merge the current gap-pattern block with the previous one. #
		    $prev_gpblock->[1] = $re_clm_new;
		}
	    } # End of "if (@new_set_gpblocks>0) {...}."

	    unless ($if_merge) {
			# Add the current gap-pattern block to the new set as an independent one. #
		push @new_set_gpblocks, [$le_clm_new, $re_clm_new, $set_pres_rows_curr, $class_curr];

#	@{$new_set_gpblocks[$gpb]} = ($le_clm2, $re_clm2, \@set_pres_rows, $class) provides information on the $gpb th gap-pattern block, where
	    }

	    my $ngpb = $#new_set_gpblocks;

	    $new_indices_intm_gpbs[$igpb] = $ngpb;
	    $new_indices_old_gpbs[$ogpb] = $ngpb;


	} # End of the 2nd Main Outer for-loop (over intermediate gap-pattern blocks). #

	my $ct_clms_new = @new_set_clms;
	my $ct_gpbs_new = @new_set_gpblocks;

   (print STDOUT "\n< Subroutine 'rearrange_clms_preserving_hs' Ended (w/ #{columns}_NEW=$ct_clms_new, #{gap-pattern blocks}_NEW=$ct_gpbs_new) />\n\n") if ($TEST_LEVEL_rearr_c_phs >0);

	return (\@new_set_clms, \@new_set_gpblocks, \@new_indices_old_gpbs, \@old_indices_intm_gpbs, \@new_indices_intm_gpbs);

}	# End of the subroutine, 'rearrange_clms_preserving_hs'. #


# mk_list_gblocks:
#	This subroutine receives a set of columns in an input MSA,
#	and creates a list of gap-blocks in the format output by 'revise_list_gblocks' in "MyANEX_IO.pm".
#	[CREATION STARTED on 2019/09/24, ENDED on 2019/09/25.]
#
#	[REVISED on 2019/09/29]:
#	The subroutine now returns also \@info_gpatterns, which is one of the outputs of 'cal_init_psm_cands_for_sgl_msa'.
#
#	[NOTE] The current version of this subroutine is quite UNefficient,
#	in the sense that it performs many processes that are superfluous.
#	Hopefully, the subroutine will be streamlined in the future.
#
# INPUTS: (\@set_clms, $char_abs, $char_pres, \@set_ingredients_indel_hsts0,
#	\%node2ch, \%node2pa, $topnode_id, \%name2id, \@seqnames, \@set_exnodes,
#	\%br2eq, \%node2depth, \%node2ct_exoffs, 
#	\%set_msa_rows2br_up_or_lw, \%nodeid2msa_row), where
#
#	@set_clms is the input set of alignment columns, that is,
#	@column = @{$set_clms[$c]} specifies the $c th column;
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_clms is an output of 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#
#	$char_abs is a character representing the "absence" state (i.e., a gap);
#	$char_pres is a character representing the "presence" state (i.e., any of the residues);
#
#	@set_ingredients_indel_hsts0 = (\%gpattern2indx0, \@set_gpatterns0, \@set_node2dparsstat0, \@set_br2change0), where
#	@info_gpatterns0 = (\@set_gpatterns0, \@set_node2dparsstat0, \@set_br2change0)
#	can be used as an input of 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm";
#	[For details on the component hash and arrays,
#	SEE the header of 'nonrec_mlt_shifts_gapblocks' in "MyANEX_Main.pm"];
#
#	%node2ch = ($node_id => \@child_node_ids, ...) is for the all interior nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) is for all the nodes except the top node;
#	$topnode_id is the ID of the top-node (i.e., root) of the tree;
#
#	%name2id = ($node_name => $node_id, ...) is for the exterior nodes and the top node (with $node_name = $TOP_NODE);
#	$seqnames[$k] is the name of the $k th sequence in the input MSA (reference/reconstructed)
#	(NOTE that these names must exactly match those used in %name2id);
#
#	@set_exnodes stores the IDs of all external nodes;
#
#	%br2eq = ($br_id => $equivalent_br_id, ...) for branches with equivalents;
#	[\%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm"];
#
#	%node2depth = ($node_id => $absolute_depth, ...) is for all the nodes;
#	[\%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm"];
#
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) is for all the nodes;
#	[\%node2ct_exoffs is an output of, e.g., 'make_node2ct_exoffs' in "MyTreeManipu_III.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $node_ID.
#	[\%nodeid2msa_row is an output of 'wrapper_mk_node2exoffs_msa_rows' or of 'associate_nodeids_w_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#
# [ OUTPUTS: (\@list_gblocks, \@new_set_clms, \@new_set_cls_gpbs, \@set_gpbs_at_spec_places, \@new_in_msa), where ] ... OBSOLETE as of 2019/09/29.
#
# OUTPUTS: (\@list_gblocks, \@new_set_clms, \@new_set_cls_gpbs, \@set_gpbs_at_spec_places, \@new_in_msa, \@info_gpatterns), where
#
#	@list_gblocks lists the gap-blocks in this input MSA;
#	@{$list_gblocks[$indx_gb]} = ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, \@affd_exnodes_org, \@affd_msa_rows),
#	provides information on the $indx_gb th gap-block.
#	More specifically,
#	$br_org is the branch ID according to the original ID assignment of the tree nodes (used for the input),
#	$up_or_lw = 'U'/'L' if the gap-block is on the upper-/lower-side of the branch,
#	$lm/rm_clm is the left-/right-most column of the gap-block,
#	@affd_exnodes_org stores the IDs of the affected external nodes according to the (aforementioned) original ID assignment, and
#	$cnct_affd_exnodes_org = join (',', @affd_exnodes_org),
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block, and
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#
#	@new_set_clms is the NEW set of columns AFTER the brief-parsimonization of the Initial Indel History;
#	It conforms to the same format as @set_clms does;
#
#	 \@new_set_cls_gpbs, \@set_gpbs_at_spec_places and \@new_in_msa are some key outputs of
#	'bch_br_parsimonize_dollo_indel_histories' in "MyMyTreeMap_indels_spt_odr2.pm";
#	More specifically,
#
#	@{$new_set_cls_gpbs[$gs]->[$ngpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $ngpb th gap-pattern block in the $gs th cluster (i.e., gapped segment) in the NEW input MSA (AFTER the brief-parsimonization), with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gpb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);
#
#	@{$sets_gpbs_at_spec_places[$gs]} is @gpbs_at_spec_places, which is one of the outputs of 'br_parsimonize_dollo_indel_histories' (in "MyMyTreeMap_indels_spt_odr2.pm") for the $gs th gapped segment, where
#	$gpbs_at_spec_places[$pl] is the index of the gap-pattern block that is in the $pl th place AFTER the brief-parsimonization;
#	
#	@new_in_msa is the NEW input MSA,
#	which was MODIFIED into coordinate with the brief-parsimonized initial indel history;
#	More specifically,
#	${$new_in_msa[$r]} represents the sequence embedded in the $r th row of the NEW input MSA,
#	represented ONLY with $char_abs and $char_pres;
#	[This may be useful for checking whether the @new_set_clms and/or @new_set_cls_gpbs are/is correct];
#
#			# ADDED on 2019/09/29. #
#	@info_gpatterns = (\@set_gpatterns, \@set_node2dparsstat, \@set_br2change)
#	provides information on the gap-patterns used for chopping the new MSA into the new gapped-segments;
#	It conforms to the same format as @info_gpatterns0 in the description of the INPUT;
#			# End of "ADDED on 2019/09/29." #
#
sub mk_list_gblocks (\@$$\@\%\%$\%\@\@\%\%\%\%\%) {

	my ($set_clms, $char_abs, $char_pres, $set_ingredients_indel_hsts0,
		$node2ch, $node2pa, $topnode_id, $name2id, $seqnames, $set_exnodes,
		$br2eq, $node2depth, $node2ct_exoffs, 
		$set_msa_rows2br_up_or_lw, $nodeid2msa_row) = @_;

	my $ct_clms = @{$set_clms};
	my $ct_rows = @{$seqnames};

	unless ($ct_clms >0) {
	    my @new_in_msa = ();
	    for (my $r=0; $r <$ct_rows; $r++) { my $tmp = ''; $new_in_msa[$r] = \$tmp; }

	    return ([], [], [], [], \@new_in_msa) ; # If there are NO columns, there should be NO gap-blocks, either.
	}

   (print STDOUT "< Subroutine 'mk_list_gblocks' (w/ #{columns}=$ct_clms, #{rows}=$ct_rows) >\n\n") if ($TEST_LEVEL_mklgb > 0);

	my @in_msa = ();
	for (my $r=0; $r <$ct_rows; $r++) {
	    my $initseq = '';
	    $in_msa[$r] = \$initseq;
	}
	for (my $c=0; $c <$ct_clms; $c++) {
	    my $clm = $set_clms->[$c];
	    for (my $r=0; $r <$ct_rows; $r++) {
		${$in_msa[$r]} .= ($clm->[$r] eq $char_abs) ? $char_abs : $char_pres ;
	    }
	}
	my @cp_in_msa = ();
	for (my $r=0; $r <$ct_rows; $r++) {
	    my $cp_row = ${$in_msa[$r]};
	    $cp_in_msa[$r] = \$cp_row;
	}

	my ($gpattern2indx0, $set_gpatterns0, $set_node2dparsstat0, $set_br2change0) = @{$set_ingredients_indel_hsts0};
		# @set_ingredients_indel_hsts0 = (\%gpattern2indx0, \@set_gpatterns0, \@set_node2dparsstat0, \@set_br2change0)


		# (Step I): Compute the Initial Candidate of the Parsimonious Indel Histories that can create the input MSA. #

	my @info_gpatterns0 = ($set_gpatterns0, $set_node2dparsstat0, $set_br2change0);


	my ($info_init_psm_hst, $info_gpatterns) = cal_init_psm_cands_for_sgl_msa (%{$node2ch}, %{$node2pa}, %{$node2depth}, %{$node2ct_exoffs}, %{$name2id}, @{$seqnames}, @in_msa, $char_abs, $char_pres, @info_gpatterns0);
	unless (defined $info_init_psm_hst) {
	    print STDOUT "[ERROR 'mk_list_gblocks'] The subroutine 'cal_init_psn_cands_for_sgl_msa' FAILED on the input MSA: $! !\n";
	    return (undef, undef, undef, undef, undef) ;
	}

	my ($set_cls_gpbs, $init_totcts_indels_for_cls, $init_psm_hsts_for_cls) = @{$info_init_psm_hst} ;
	my ($set_gpatterns, $set_node2dparsstat, $set_br2change) = @{$info_gpatterns} ;

	my $ct_gpd_segments = @{$set_cls_gpbs};
	my $ct_gp = @{$set_gpatterns};

   (print STDOUT "\n ... The subroutine, 'cal_init_psm_cands_for_sgl_msa', Ended (w/ #{gap-patterns}=$ct_gp, #{gapped segments}=$ct_gpd_segments) ...\n\n") if ($TEST_LEVEL_mklgb >1);


		# (Step II): Attempt to Parsimonize the Initial Candidate Indel History. #

	my ($new_info_init_psm_hst, $new_in_msa, $set_new_lists_gblocks, $sets_places_gpbs, $sets_gpbs_at_spec_places)
	   = bch_br_parsimonize_dollo_indel_histories (@cp_in_msa, @{$info_init_psm_hst}, @{$set_exnodes}, $topnode_id, %{$node2pa}, %{$node2ch}, $char_abs, $char_pres, %{$br2eq});
	unless (defined $new_info_init_psm_hst) {
	    print STDOUT "[ERROR 'mk_list_gblocks'] The subroutine 'bch_br_parsimonize_dollo_indel_histories' FAILED on the input MSA: $! !\n";
	    return (undef, undef, undef, undef, undef) ;
	}
   (print STDOUT "\n ... The subroutine, 'bch_br_parsimonize_dollo_indel_histories', Ended ...\n\n") if ($TEST_LEVEL_mklgb>1);
	
	my ($new_set_cls_gpbs, $new_init_totcts_indels_for_cls, $new_init_psm_hsts_for_cls) = @{$new_info_init_psm_hst} ;
	
	unless (@{$new_set_cls_gpbs} == $ct_gpd_segments) {
	    print STDOUT "[ERROR 'mk_list_gblocks'] The number of gapped segments AFTER the brief parsimonization (=", scalar (@{$new_set_cls_gpbs}), ") CLASHES that BEFORE the brief parsimonization (=$ct_gpd_segments): $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

			# (Post-Step II A): Create the new set of columns, one of the main outputs. #

	my $le_lm_gs = $set_cls_gpbs->[0][0][2];
	my @new_set_clms = ();

	my $re_prev_gseg = -1;

	for (my $gs = 0; $gs < $ct_gpd_segments; $gs++) { # Semi-Final Main Outer for-loop (over gapped segments). #

	    my $old_cls_gpbs = $set_cls_gpbs->[$gs];
	    my $indices_ogpbs_at_spec_places = $sets_gpbs_at_spec_places->[$gs];

	    my $ct_gpbs = @{$old_cls_gpbs};
	    my ($lm_ogpb, $rm_ogpb) = @{$old_cls_gpbs}[0, $ct_gpbs-1];
	    my $le_gseg = $lm_ogpb->[2];
	    my $re_gseg = $rm_ogpb->[3];

	    if ($re_prev_gseg +1 < $le_gseg) {
			# Incorporate the left-flanking gapless block. #
		push @new_set_clms, @{$set_clms}[$re_prev_gseg+1 .. $le_gseg-1];
	    }

	    for (my $pl = 0; $pl < $ct_gpbs; $pl++) { # Inner for-loop (over new places of gap-pattern blocks). #
		my $indx_ogpb = $indices_ogpbs_at_spec_places->[$pl];
		my $info_ogpb = $old_cls_gpbs->[$indx_ogpb];
		my ($indx_gp, $len_ogpb, $le_ogpb, $re_ogpb) = @{$info_ogpb};

			# Incorporate the old gap-pattern block at the present place. #
		push @new_set_clms, @{$set_clms}[$le_ogpb .. $re_ogpb];
		
	    } # End of the inner for-loop (over new places of gap-pattern blocks). #

	    $re_prev_gseg = $re_gseg; # Update $re_prev_gseg. #

	} # End of the Simi-Final Main Outer for-loop (over gapped segments). #

	if ($re_prev_gseg < $ct_clms-1) {
			# Incorporate the gapless block at the right-end, if at all. #
	   push @new_set_clms, @{$set_clms}[$re_prev_gseg+1 .. $ct_clms-1];
	}


			# (Post-Step II B): Create a list of Gap-Blocks, one of the main outputs. #

	my @list_gblocks = ();

	for (my $gs = 0; $gs < $ct_gpd_segments; $gs++) { # Final Main Outer for-loop (over gapped segments). #

	    my $new_cls_gpbs = $new_set_cls_gpbs->[$gs];
	    my $new_list_gblocks = $set_new_lists_gblocks->[$gs];
	    my $ct_gbs = @{$new_list_gblocks};

	    for (my $gb=0; $gb <$ct_gbs; $gb++) { # Inner for-loop (over gap-blocks). #

		my $info_gblock = $new_list_gblocks->[$gb];
		my ($br0, $up_or_lw0, $lm_gpb, $rm_gpb, $invlvd_gpbs, $affd_exnodes, $e) = @{$info_gblock};
		my @affd_exnodes = (@{$affd_exnodes}>1) ? sort {$a <=> $b} @{$affd_exnodes} : @{$affd_exnodes} ;
		my @affd_msa_rows = ();
		foreach my $exnode (@affd_exnodes) { push @affd_msa_rows, $nodeid2msa_row->{$exnode}; }
		if (@affd_msa_rows > 1) {
		    @affd_msa_rows = sort {$a <=> $b} @affd_msa_rows;
		}
		my $cnct_affd_exnodes = join (',', @affd_exnodes);
		my $cnct_affd_msa_rows = join (',', @affd_msa_rows);

		my $br_up_or_lw = $set_msa_rows2br_up_or_lw->{$cnct_affd_msa_rows};
		unless (defined $br_up_or_lw) {
		    print STDOUT "[ERROR 'mk_list_gblocks'] Somehow, NO BRANCH corresponds to the set of affected MSA rows, ($cnct_affd_msa_rows), for the $gb th gap-block in the $gs th gapped segment: $! !\n\n";
		    return (undef, undef, undef, undef, undef);
		}
		my ($br, $up_or_lw) = @{$br_up_or_lw};

		my $lm_clm = $new_cls_gpbs->[$lm_gpb][2];
		my $rm_clm = $new_cls_gpbs->[$rm_gpb][3];

		push @list_gblocks, [$br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows];

			# @{$list_gblocks[$indx_gb]} = ($br_org, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, \@affd_exnodes_org, \@affd_msa_rows),

	    } # End of the inner for-loop (over gap-blocks). #

	} # End of the Final Main Outer for-loop (over gapped segments). #

	my $ct_gbs = @list_gblocks;

	if ($ct_gbs >1) {

	    @list_gblocks = sort { cmp_gblocks3 (@{$a}, @{$b}, 2, 3, 6) } @list_gblocks ;
	}

   (print STDOUT "\n< Subroutine 'mk_list_gblocks' Ended (w/ #{gap-blocks}=$ct_gbs) />\n\n") if ($TEST_LEVEL_mklgb > 0);

	return (\@list_gblocks, \@new_set_clms, $new_set_cls_gpbs, $sets_gpbs_at_spec_places, $new_in_msa, $info_gpatterns); # MODIFIED on 2019/09/29. #
	##return (\@list_gblocks, \@new_set_clms, $new_set_cls_gpbs, $sets_gpbs_at_spec_places, $new_in_msa);

}	# End of the subroutine, 'mk_list_gblocks'. #


# reverse_spec_cii_in_wd:
#	This subroutine "reverse"s a specified CII-candidate
#	in a window of an input MSA.
#	It actually bundles the subroutines, 
#	'reverse_cii_on_clmset', 'rearrange_clms_preserving_hs' and 'mk_list_gblocks', 
#	all of which are above,
#	and also performs some subroutines to augment the resulting MSA with some auxiliary data-sets.
#	[CREATION STARTED on 2019/09/25, ENDED on 2019/09/27.]
#
#	[REVISED on 2019/09/29]: Now, the subroutine also outputs \@set_gpatterns.
#
#	[DEBUGGED on 2019/09/29]: CORRECTED the order among the arguments of 'check_consistency_bw_msa_and_clmset'.
#
# INPUTS: (\@set_clms_in, $char_abs, $char_pres, 
#	$le_cii, $re_cii, \@set_monophyl_sets_pres_rows,
#	\@set_ingredients_indel_hsts0,
#	\%node2ch, \%node2pa, $topnode_id, \%name2id, \@seqnames, \@set_exnodes,
#	\%br2eq, \%node2depth, \%node2ct_exoffs, 
#	\%set_msa_rows2br_up_or_lw, \%nodeid2msa_row,
#	\@nodeids_in_msa, $maxct_1way_shifts), where
#
#	@set_clms_in is the input set of alignment columns, that is,
#	@column = @{$set_clms[$c]} specifies the $c th column;
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_clms is an output of 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#
#	$char_abs is a character representing the "absence" state (i.e., a gap);
#	$char_pres is a character representing the "presence" state (i.e., any of the residues);
#
#	$le_cii an $re_cii are the indices of the columns at the left-end and right-end (inclusive), respectively, of the region where a CII is suspected;
#	@set_monophyl_sets_pres_rows stores some monophyletic sets of rows with the "presence" state;
#	More precisely,
#	@{$set_monophyl_sets_pres_rows[$ms]} stores the indices of the MSA rows (with the "presence" state) in the $ms th monophyletic set;
#	($le_cii, $re_cii and \@set_monophyl_sets_pres_rows will be used as inputs of 'reverse_cii_on_clmset' above);
#	[Typically, $le_cii, $re_cii and \@set_monophyl_sets_pres_rows are components of an element of
#	@list_cii_cands, which is a main output of 'list_cii_candidates' above];
#
#	\@set_ingredients_indel_hsts0,
#	\%node2ch, \%node2pa, $topnode_id, \%name2id, \@seqnames, \@set_exnodes,
#	\%br2eq, \%node2depth, \%node2ct_exoffs, 
#	\%set_msa_rows2br_up_or_lw & \%nodeid2msa_row
#	will be used as inputs of 'mk_list_gblocks' above;
#	[For details, See the header of 'mk_list_gblocks'];
#
#	$nodeids_in_msa[$r] is the node ID corresponding to the $r th row of the input MSA;
#	[Typically, \@nodeids_in_msa and \%nodeid2msa_row are the outputs of 'associate_nodeids_w_msa_rows' in "MyTreeMap_sppl.pm"];
#
#	$maxct_1way_shifts specifies the upper-bound of the number of shifts in 1 way, i.e., either to the left or to the right;
#
# OUTPUTS: ($flag_end, 
#
#	 ## [ \@new_set_cls_gpbs, \@set_gpbs_at_spec_places, # Added on 2019/09/27. ] ... OBSOLETE as of 2019/09/29.
#	 \@new_set_cls_gpbs, \@set_gpbs_at_spec_places, \@set_gpatterns, # => MODIFIED on 2019/09/29.
#
#	\@new_curr_set_clms, \@new_curr_bds_gblks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_gblks,
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#	\@new_list_gblocks), where
#
#	$flag_end = 1 if the subroutine ended without any particular problems,
#		  = 0 if it failed;
#
#			# Added on 2019/09/27. (Hd-out) #
#	\@new_set_cls_gpbs and \@set_gpbs_at_spec_places are auxiliary (or subsidiary) outputs of 'mk_list_gblocks' above;
#	More precisely, they are some key outputs of 'bch_br_parsimonize_dollo_indel_histories' in "MyMyTreeMap_indels_spt_odr2.pm";
#	Specifically,
#
#	@{$new_set_cls_gpbs[$gs]->[$ngpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $ngpb th gap-pattern block in the $gs th cluster (i.e., gapped segment) in the NEW input MSA (AFTER the brief-parsimonization), with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gpb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);
#
#	@{$sets_gpbs_at_spec_places[$gs]} is @gpbs_at_spec_places, which is one of the outputs of 'br_parsimonize_dollo_indel_histories' (in "MyMyTreeMap_indels_spt_odr2.pm") for the $gs th gapped segment, where
#	$gpbs_at_spec_places[$pl] is the index of the gap-pattern block that is in the $pl th place AFTER the brief-parsimonization;
#			# End of "Added on 2019/09/27. (Hd-out)" #
#
#	${$sets_gpatterns[$indx_gp]} gives a string representing the $indx_gp th gap-pattern that occurred when constructing gapped-segments; # ADDED on 2019/09/29. (Hd-out2) #
#
#	The arguments from \@new_curr_set_clms up to \@new_list_gblocks constitutes the same line-up as those for 'cvsplit_gblock' in "MyANEX_Main.pm";
#	[See the header of 'cvsplit_gblock' in "MyANEX_Main.pm" for details];
#
sub reverse_spec_cii_in_wd (\@$$$$\@\@\%\%$\%\@\@\%\%\%\%\%\@$) {

	my ($set_clms_in, $char_abs, $char_pres, 
	   $le_cii, $re_cii, $set_monophyl_sets_pres_rows,
	   $set_ingredients_indel_hsts0,
	   $node2ch, $node2pa, $topnode_id, $name2id, $seqnames, $set_exnodes,
	   $br2eq, $node2depth, $node2ct_exoffs, 
	   $set_msa_rows2br_up_or_lw, $nodeid2msa_row,
	   $nodeids_in_msa, $maxct_1way_shifts) = @_;

# INPUTS: (\@set_clms_in, $char_abs, $char_pres, 
#	$le_cii, $re_cii, \@set_monophyl_sets_pres_rows,
#	\@set_ingredients_indel_hsts0,
#	\%node2ch, \%node2pa, $topnode_id, \%name2id, \@seqnames, \@set_exnodes,
#	\%br2eq, \%node2depth, \%node2ct_exoffs, 
#	\%set_msa_rows2br_up_or_lw, \%nodeid2msa_row,
#	\@nodeids_in_msa, $maxct_1way_shifts), where


	my $ct_clms_in = @{$set_clms_in};
	my $ct_monophyl_sets = @{$set_monophyl_sets_pres_rows};

   (print STDOUT "<< Subroutine 'reverse_spec_cii_in_wd' (w/ #{columns}_in=$ct_clms_in, [\$le_cii, \$re_cii] = [$le_cii, $re_cii], #{monophyletic sets}=$ct_monophyl_sets, (\$char_abs, \$char_pres) = ('$char_abs', '$char_pres'), \$maxct_1way_shifts = $maxct_1way_shifts) >>\n\n") if ($TEST_LEVEL_MAIN >0);

   if ($TEST_LEVEL_MAIN > 2) { # (TOP) #

	print STDOUT "\n[[ Major Input Data-Sets: ]]\n\n";

	print STDOUT "[ The Input Set of Columns BEFORE the Whole Process of 'reverse_spec_cii_in_wd' (, i.e., ", '@{$set_clms_in}', "): ]\n\n";

	unless ( print_out_set_clms (@{$set_clms_in}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_spec_cii_in_wd'] The subroutine 'print_out_set_clms' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_spec_cii_in_wd' ...\n\n";
	}


	print STDOUT "\n", '[ Contents of @{$set_monophyl_sets_pres_rows}', " (#{monophyl_sets}=$ct_monophyl_sets): ]\n\n";

	print STDOUT join ("\t", 'Set_No.', '[indices,of,constituent,msa,rows]'), "\n\n";

	for (my $mnp=0; $mnp< $ct_monophyl_sets; $mnp++) {
		my $cnct_monophyl_set = join (',', @{$set_monophyl_sets_pres_rows->[$mnp]});
		print STDOUT join ("\t", $mnp, "[$cnct_monophyl_set]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_MAIN > 2) { # (TOP) # ...}."
	

		# (Main 1): "Reverse" the specified CII-suspect. #

	my ($set_clms_intm1)
	   = reverse_cii_on_clmset (
		@{$set_clms_in}, $char_abs, $le_cii, $re_cii, @{$set_monophyl_sets_pres_rows}
		);
	unless (defined $set_clms_intm1) {
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] The subroutine 'reverse_cii_on_clmset' FAILED: $! !\n\n";
	    return (0, undef, undef, undef);
	}
	my $ct_clms_intm1 = @{$set_clms_intm1};

    (print STDOUT "\n ... The subroutine 'reverse_cii_on_clmset' Ended on the Input set of columns (w/ #{columns}_intm1=$ct_clms_intm1) ...\n\n") if ($TEST_LEVEL_MAIN >1);


   if ($TEST_LEVEL_MAIN >= 5) { # (Main 1) #

	print STDOUT "[[ The Output of Main 1, i.e., 'reverse_cii_on_clmset': ]]\n\n";
	print STDOUT '[ The Contents of @{$set_clms_intm1}: ]', "\n\n";

	unless ( print_out_set_clms (@{$set_clms_intm1}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_spec_cii_in_wd'] The subroutine 'print_out_set_clms' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_spec_cii_in_wd' ...\n\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_MAIN >= 5) { # (Main 1) # ...}"


		# (Main 2): Rearrange the resulting set of columns while preserving the homology-structure. #
	my ($set_clms_intm2, $set_gpblocks_intm2, $intm2_indices_intm1_gpbs, $intm1_indices_intm_gpbs, $intm2_indices_intm_gpbs)
	  = rearrange_clms_preserving_hs (@{$set_clms_intm1}, %{$set_msa_rows2br_up_or_lw}, $char_abs);
	unless (defined $set_clms_intm2) {
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] The subroutine 'rearrange_clms_preserving_hs' FAILED: $! !\n\n";
	    return (0, undef, undef, undef);
	}
	my $ct_clms_intm2 = @{$set_clms_intm2};
	my $ct_gpbs_intm2 = @{$set_gpblocks_intm2};

    (print STDOUT "\n ... The subroutine 'rearrange_clms_preserving_hs' Ended on the 1st Intermediate set of columns (w/ #{columns}_intm2=$ct_clms_intm2, #{gap-pattern blocks}_intm2=$ct_gpbs_intm2) ...\n\n") if ($TEST_LEVEL_MAIN >1);


   if ($TEST_LEVEL_MAIN >= 5) { # (Main 2-a) #

	print STDOUT "[[ The Outputs of Main 2, i.e., 'rearrange_clms_preserving_hs': ]]\n\n";

	print STDOUT '[ The Contents of @{$set_clms_intm2}: ]', "\n\n";

	unless ( print_out_set_clms (@{$set_clms_intm2}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_spec_cii_in_wd'] The subroutine 'print_out_set_clms' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_spec_cii_in_wd' ...\n\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_MAIN >= 5) { # (Main 2-a) # ...}"

   if ($TEST_LEVEL_MAIN > 5) { # (Main 2-b) #

	print STDOUT '[ The contents of @{$set_gpblocks_intm2}: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_gpb', 'le_gpb', 're_gpb', 'Class', '[rows,with,presence,state]'), "\n\n";

	for (my $gpb = 0; $gpb < $ct_gpbs_intm2; $gpb++) {

	    my ($le_gpb, $re_gpb, $set_pres_rows, $class) = @{$set_gpblocks_intm2->[$gpb]};
	    my $cnct_pres_rows = join (',', @{$set_pres_rows});

	    print STDOUT join ("\t", $gpb, $le_gpb, $re_gpb, $class, "[$cnct_pres_rows]"), "\n";

	}
	print STDOUT "\n\n";

	##print STDOUT '[ The contents of @{$intm2_indices_intm1_gpbs}, @{$intm1_indices_intm_gpbs} & @{$intm2_indices_intm_gpbs}: ]', "\n\n";

	print STDOUT "\n[ Relationships between old ('intm1' here), new ('intm2' here) indices, as well as their intermediates ('intm' here), of gap-pattern-blocks (#{gap-pattern-blocks}_intm2=$ct_gpbs_intm2) : ]\n\n";

			### RESTART FROM HERE (on 2019/09/27). ###
	
	my @out_intm2_indices_intm1_gpbs = my @out_intm1_indices_intm_gpbs = my @out_intm2_indices_intm_gpbs = ();

	foreach my $pair ( [$intm2_indices_intm1_gpbs, \@out_intm2_indices_intm1_gpbs],
			  [$intm1_indices_intm_gpbs, \@out_intm1_indices_intm_gpbs],
			  [$intm2_indices_intm_gpbs, \@out_intm2_indices_intm_gpbs],
		) {
		my ($indices_gpbs, $out_indices_gpbs) = @{$pair};
		foreach my $indx (@{$indices_gpbs}) { push @{$out_indices_gpbs}, ((defined $indx) ? $indx : 'undef'); }
	}

	print STDOUT '@{$intm2_indices_intm1_gpbs} = (', join (', ', @out_intm2_indices_intm1_gpbs), ") .\n\n";

	print STDOUT '@{$intm1_indices_intm_gpbs} = (', join (', ', @out_intm1_indices_intm_gpbs), ") .\n\n";

	print STDOUT '@{$intm2_indices_intm_gpbs} = (', join (', ', @out_intm2_indices_intm_gpbs), ") .\n\n";

	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_MAIN > 5) { # (Main 2-b) # ...}"


		# (Main 3): Create a List of Gap-Blocks while attempting to Briefly Parsimonize the Initial Candidate Indel History.

	my ($new_list_gblocks, $new_curr_set_clms, $new_set_cls_gpbs, $sets_gpbs_at_spec_places, $new_dicot_msa, $info_gpatterns) # MODIFIED on 2019/09/29. #
	##my ($new_list_gblocks, $new_curr_set_clms, $new_set_cls_gpbs, $sets_gpbs_at_spec_places, $new_dicot_msa)
	   = mk_list_gblocks (
		@{$set_clms_intm2}, $char_abs, $char_pres, @{$set_ingredients_indel_hsts0},
		%{$node2ch}, %{$node2pa}, $topnode_id, %{$name2id}, @{$seqnames}, @{$set_exnodes},
		%{$br2eq}, %{$node2depth}, %{$node2ct_exoffs},
		%{$set_msa_rows2br_up_or_lw}, %{$nodeid2msa_row}
		);
	unless (defined $new_list_gblocks) {
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] The subroutine 'mk_list_gblocks' FAILED: $! !\n\n";
	    return (0, undef, undef, undef);
	}
	my $ct_new_gbs = @{$new_list_gblocks};
	my $ct_new_clms = @{$new_curr_set_clms};

    (print STDOUT "\n ... The subroutine 'mk_list_gblocks' Ended on the 2nd Intermediate set of columns (w/ #{new gap-blocks} = $ct_new_gbs, #{new columns} = $ct_new_clms) ...\n\n") if ($TEST_LEVEL_MAIN >1);

	my ($set_gpatterns, $set_node2dparsstat, $set_br2change) = @{$info_gpatterns} ; # ADDED on 2019/09/29. #



			# Check the consistency between the new MSA and the new set of columns. #

 	my ($if_consistent, $list_inconsistent_cells) = check_consistency_bw_msa_and_clmset (@{$new_dicot_msa}, @{$new_curr_set_clms}, $char_abs, $char_abs); # A BUG Fixed on 2019/09/29. #
 	##my ($if_consistent, $list_inconsistent_cells) = check_consistency_bw_msa_and_clmset (@{$new_curr_set_clms}, @{$new_dicot_msa}, $char_abs, $char_abs);
	unless (defined $if_consistent) {
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] The subroutine 'check_consistency_bw_msa_and_clmset' FAILED on the outputs of 'mk_list_gblocks': $! !\n\n";
	    return (0, undef, undef, undef);
	}
	unless ($if_consistent == 1) {
	    my @out_list_inconsist_cells = ();
	    foreach my $ic_cell (@{$list_inconsistent_cells}) {
		push @out_list_inconsist_cells, sprintf ("[%d, %d]", @{$ic_cell});
	    }
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] The subroutine 'check_consistency_bw_msa_and_clmset' REVEALED that ", '@{$new_curr_set_clms} and @{$new_dicot_msa}', " are INconsistent with each other: $! !!\n\n";
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] Inconsistent cells (in the form of [column, row]):\n\n";
	    print STDOUT "\t", join ('; ', @out_list_inconsist_cells), " .\n\n"; 
	    return (0, undef, undef, undef);
	}
    (print STDOUT "\n ... The subroutine 'check_consistency_bw_msa_and_clmset' judged that ", '@{$new_curr_set_clms} and @{$new_dicot_msa}', " are Consistent with each other ...\n\n") if ($TEST_LEVEL_MAIN >1);


   if ($TEST_LEVEL_MAIN > 3) { # (Main 3-a) #

	print STDOUT "[[ The Outputs of Main 3, i.e., 'mk_list_gblocks' (except ", '@{$new_curr_set_clms}', ", which will be shown later): ]]\n\n";

	print STDOUT '[ Contents of @{$new_list_gblocks}: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_new_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$new_list_gblocks->[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_MAIN > 3) { # (Main 3-a) # ...}"


   if ($TEST_LEVEL_MAIN > 5) { # (Main 3-b) #

	my $ct_new_gsegs = @{$new_set_cls_gpbs};
	print STDOUT "#{gapped segments}_NEW = $ct_new_gsegs .\n\n";

	print STDOUT '[ Contents of @{$new_set_cls_gpbs}: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_gseg', 'indx_gpb_in_gseg', 'indx_gpattern', 'len_gpb', 'le_gpb', 're_gpb', "'gap-pattern'"), "\n\n";

	for (my $gs = 0; $gs < $ct_new_gsegs; $gs++) {

	    my $cl_gpbs = $new_set_cls_gpbs->[$gs];
	    my $ct_gpbs = @{$cl_gpbs};

	    for (my $gpb = 0; $gpb < $ct_gpbs; $gpb++) {
		my $indx_gp = $cl_gpbs->[$gpb][0];
		my $gpattern = ${$set_gpatterns->[$indx_gp]};
		print STDOUT join ("\t", $gs, $gpb, @{$cl_gpbs->[$gpb]}, "'$gpattern'"), "\n";
	    }
	    print STDOUT "\n";
	}
	print STDOUT "\n";

	print STDOUT '[ Contents of @{$sets_gpbs_at_spec_places}: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_gseg', '#{gap-pattern-blocks}', '[gap-pattern-blocks,at,specific,places]'), "\n\n";

	for (my $gs = 0; $gs < $ct_new_gsegs; $gs++) {

	    my $gpbs_at_spec_places = $sets_gpbs_at_spec_places->[$gs];

	    my $ct_gpbs = @{$gpbs_at_spec_places};
	    my $cnct_gpbs_at_spls = join (',', @{$gpbs_at_spec_places});

	    print STDOUT join ("\t", $gs, $ct_gpbs, "[$cnct_gpbs_at_spls]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_MAIN > 5) { # (Main 3-b) # ...}"


		### (POST-MAIN): Create some other Auxiliary Data-Sets. ###

		# (Post-Main 1): Create @new_curr_bds_gblks & @rows_affd_by_new_gblks. #

   (print STDOUT "\n", 'Manually Create: @new_curr_bds_gblks & @rows_affd_by_new_gblks ...', "\n\n") if ($TEST_LEVEL_MAIN>2);

	my @new_curr_bds_gblks = my @rows_affd_by_new_gblks = ();
	for (my $ngb = 0; $ngb <$ct_new_gbs; $ngb++) { # 1st trailing for-loop (over new gap-blocks). #

		my ($br, $up_or_lw, $le_ngb, $re_ngb, $cnct_affd_exnodes, $cnct_affd_msa_rows, $affd_exnodes, $affd_msa_rows) = @{$new_list_gblocks->[$ngb]};

#	@{$list_gblocks[$indx_gb]} = ($br_org, $up_or_lw, $le_gb, $re_gb, $cnct_affd_exnodes_org, $cnct_affd_msa_rows, \@affd_exnodes_org, \@affd_msa_rows),

		$new_curr_bds_gblks[$ngb] = [$le_ngb, $re_ngb];
		$rows_affd_by_new_gblks[$ngb] = $affd_msa_rows;

	} # End of the 1st trailing for-loop (over new gap-blocks). #


		# (Post-Main 2): Create \@new_curr_gblks_w_spec_lb & \@new_curr_gblks_w_spec_rb. #

	my ($le_wd, $re_wd) = (0, $ct_new_clms-1);

   (print STDOUT "\n", 'Manually Set: ($le_wd, $re_wd) = ', "($le_wd, $re_wd) .\n\n") if ($TEST_LEVEL_MAIN>2);

	my ($new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb)
	  = mk_lists_gblocks_w_spec_bounds ($le_wd, $re_wd, @new_curr_bds_gblks);
	unless (defined $new_curr_gblks_w_spec_lb) {
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] The subroutine 'mk_lists_gblocks_w_spec_bounds' FAILED: $! !\n\n";
	    return (0, undef, undef, undef);
	}
    (print STDOUT "\n ... The subroutine 'mk_lists_gblocks_w_spec_bounds' Ended on ", '@new_curr_bds_gblks', " ...\n\n") if ($TEST_LEVEL_MAIN >1);


		# (Post-Main 3): Create \@new_inter_block_relations. #

	my ($indx_br, $indx_up_or_lw) = (0, 1);

	my ($new_inter_block_relations)
	  = mk_inter_block_relations (
		@{$new_list_gblocks}, $indx_br, $indx_up_or_lw, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}
		);
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] The subroutine 'mk_inter_block_relations' FAILED: $! !\n\n";
	    return (0, undef, undef, undef);
	}
    (print STDOUT "\n ... The subroutine 'mk_inter_block_relations' Ended on ", '@{$new_list_gblocks}', " ...\n\n") if ($TEST_LEVEL_MAIN >1);


		# (Post-Main 5): Create \@new_cts_vincluding, \@new_cts_vincluded & \@new_cts_onx. #

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	  = cts_vincluding_vincluded_onx_gblks (@{$new_inter_block_relations});
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED: $! !\n\n";
	    return (0, undef, undef, undef);
	}
    (print STDOUT "\n ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@{$new_inter_block_relations}', " ...\n\n") if ($TEST_LEVEL_MAIN >1);


		# (Post-Main 6): Create \@new_gblock_sizes. #

	my ($new_gblock_sizes)
	  = mk_list_gblock_sizes (@new_curr_bds_gblks, 0, 1, @{$new_inter_block_relations});
	unless (defined $new_gblock_sizes) {
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] The subroutine 'mk_list_gblock_sizes' FAILED: $! !\n\n";
	    return (0, undef, undef, undef);
	}
    (print STDOUT "\n ... The subroutine 'mk_list_gblock_sizes' Ended on ", '@new_curr_bds_gblks & @{$new_inter_block_relations}', " ...\n\n") if ($TEST_LEVEL_MAIN >1);


		# (Post-Main 7): Create \@new_curr_bl_coords, \@new_bds_gb_coords & \@new_gb_coords_wd_frm, #

	my ($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	  = mk_orgs_bds_gb_coords (@new_curr_bds_gblks, $le_wd, $re_wd, @{$new_inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_curr_bl_coords) {
	    print STDOUT "[ERROR 'reverse_spec_cii_in_wd'] The subroutine 'mk_orgs_bds_gb_coords' FAILED: $! !\n\n";
	    return (0, undef, undef, undef);
	}
    (print STDOUT "\n ... The subroutine 'mk_orgs_bds_gb_coords' Ended on ", '@new_curr_bds_gblks & @{$new_inter_block_relations}', " ...\n\n") if ($TEST_LEVEL_MAIN >1);


		# (FINAL): Print out the Main Final Outputs. #
	
   if ($TEST_LEVEL_MAIN >3) {

	print STDOUT "\n[[ Main FINAL OUTPUTS: ]\n";

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of 'reverse_spec_cii_in_wd': ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @new_curr_bds_gblks, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_spec_cii_in_wd'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_spec_cii_in_wd' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_MAIN >3) {...}"


   if ($TEST_LEVEL_MAIN > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of 'reverse_spec_cii_in_wd': ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$new_gblock_sizes}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_gblks, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_spec_cii_in_wd'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_spec_cii_in_wd' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_MAIN > 3) {...}"


	my $flag_end = 1;
	my $ct_clms_out = @{$new_curr_set_clms};
	my $ct_gbs_out = @{$new_list_gblocks};

   (print STDOUT "<< Subroutine 'reverse_spec_cii_in_wd' ENDED (w/ \$flag_end=$flag_end, #{columns}_out=$ct_clms_out, #{gap-blocks}_out=$ct_gbs_out) />/>\n\n") if ($TEST_LEVEL_MAIN >0);

	return ($flag_end, 

	    ## $new_set_cls_gpbs, $sets_gpbs_at_spec_places, # Added on 2019/09/27. #
	    $new_set_cls_gpbs, $sets_gpbs_at_spec_places, $set_gpatterns, # => MODIFIED on 2019/09/29.

	    $new_curr_set_clms, \@new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, \@rows_affd_by_new_gblks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	    $new_list_gblocks);

}	# End of the subroutine, 'reverse_spec_cii_in_wd'. #


# print_out_set_clms:
#	This satellite subroutine of 'reverse_spec_cii_in_wd'
#	prints out a set of MSA columns. 
#	Its main part was IMPORTED from 'print_out_gblock_configs' in "MyANEX_Main.pm".
#	[CREATED on 2019/09/26.]
#
# INPUTS: (\@curr_set_clms, \*${$fh_out}), where
#
#	@curr_set_clms is a set of columns in the current MSA;
#	@column = @{$curr_set_clms[$c]} specifies the $c th column;
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_clms is an output of 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#
#	$fh_out is the reference to an output file-handle;
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any particular problems,
#		= 0 if it failed.
#
sub print_out_set_clms (\@\*) {

	my ($curr_set_clms, $fh_out) = @_;

	my $ct_clms = @{$curr_set_clms};
	my $ct_rows = ($ct_clms>0) ? @{$curr_set_clms->[0]} : 0;

	print $fh_out '( Contents of @{$curr_set_clms}: )', "\n\n";
	print $fh_out join ("\t", 'Column\Row:', 0 .. $ct_rows -1), "\n\n";

	for (my $c = 0; $c < $ct_clms; $c++) {
	    print $fh_out join ("\t", $c, @{$curr_set_clms->[$c]}), "\n"; # A BUG Fixed on 2019/07/10. #
	    ##print $fh_out join ("\t", $c, @{$curr_set_clms->[0]}), "\n";
	}
	print $fh_out "\n\n";

	return 1;

}	# End of the subroutine, 'print_out_set_clms'. #


1;		# a "spell" which makes the module "use"-able and/or "require"-able #


