package MySRD_Map;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019? Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MySRD_Map.pm:
#	This package is a collection of subroutines
#	necessary for creating the Substitutional Residue-Difference Map (in short, "SRD Map").
#	[CREATION Started on 2020/03/05, ENDED on 2020/03/dd.]
#

use strict;
use warnings;


		# (ADDED in creation of 'brwise_distribute_srds_assoc_w_alnshear' on 2020/03/26.) (US) #
use MyCluster_distribution qw(
	cluster_distributor_wgt
	);
		# End of "(ADDED in creation of 'brwise_distribute_srds_assoc_w_alnshear' on 2020/03/26.) (US)" #

=Set aside for potential use. (1)

use MyTaxonomy qw(
	if_ancestor fetch_ancestry calculate_abs_depth
	fetch_ext_offsprings_nr fetch_all_offsprings_nr
	);

use MyTreeManipu_II qw( );
	# extract_sgltree_from_file output_tree 

use MyTreeManipu_III qw( make_node2abs_depth );

use MyLikelihoods_respattern_msa_dna qw(
	calc_node2ntfreqs_td
	calc_prob_descresstats_cond_ancstat2
	calc_joint_prob_ancresstats_and_ancstat2
	calc_clm_probs
	);

use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );
use MyPAML_parser qw( retrieve_aa_freqs prepare_paml_ctrl );

use MyProbDistri qw(
	binomial_cum_via_betai
	);

use MyArrayManipu qw(
	merge_integer_intervals
	);

	# Added in creation of 'detect_purge_cands2' (on 2019/06/15). #
use MyArrayManipu_num qw(
	union_nmarray
	binsearch_num
	); # Added 'binsearch_num' in revision of 'detect_purge_cands2' (on 2019/06/16).
	# End of "Added in creation of 'detect_purge_cands2' (on 2019/06/15)." #
=cut # End of "Set aside for potential use. (1)"

use MyDetect_purge_cands qw(
	calc_br2cw_expcts_sbsts_ave
	calc_br2cw_obscts_sbsts
	mk_node2sw_gstats
	);
	# sum_arrays_over_branches
	# calc_br2cw_expcts_obscts_sbsts
	#
	# ADDED 'mk_node2sw_gstats' on 2020/03/07. #


use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	mk_srd_map
	print_srd_map

	tally_assoc_srds_and_alnshears
	print_sums_assoc_srds_and_alnshears
	tally_srds_in_map
	print_sums_srds_in_map
	
	brwise_distribute_srds_assoc_w_alnshear
	brwise_distribute_srds_in_map
	);

sub mk_srd_map ($\@$$$\%\%\%\%\%\@\@$$\%\%);
sub print_srd_map (\*\@\%\@$$$$$);

sub tally_assoc_srds_and_alnshears (\@\@$$) ; # ADDED on 2020/03/24.
sub print_sums_assoc_srds_and_alnshears (\*\@\%\@\@) ; # ADDED on 2020/03/24.
##sub tally_assoc_srds_in_map (\@$); # ADDED on 2020/04/03.
sub tally_srds_in_map (\@$); # => CORRECTED on 2020/04/03.
sub print_sums_srds_in_map (\*\@\%\@\@); # ADDED on 2020/04/03.

sub brwise_distribute_srds_assoc_w_alnshear (\@\@$$\@\@\@); # ADDED on 2020/03/26.
sub brwise_distribute_srds_in_map (\@$\@\@); # ADDED on 2020/04/03.

	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

=Set aside for potential use. (2)
our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our %NUCL2INDX; # = ($nucleotide => {index in @NUCLEOTIDES}, ...)
for (my $k=0; $k<$CT_NUCLS; $k++) { $NUCL2INDX{$NUCLEOTIDES[$k]} = $k; }
=cut # End of "Set aside for potential use. (2)"



	## The following parameters specify the verboseness of the test output.
	## (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

	# For 'mk_srd_map'. #
our $TEST_LEVEL_mk = 0;
#our $TEST_LEVEL_mk = 5;
##our $TEST_LEVEL_mk = 6;

	# For 'print_srd_map'. #
our $TEST_LEVEL_pr = 0;
#our $TEST_LEVEL_pr = 5;
##our $TEST_LEVEL_pr = 6;



### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.
#


		##### DEFINITIONS of the Subroutines. #####


# mk_srd_map:
#	This subroutine receives some variables characterizing a phylogenetic tree,
#	and some outputs of the likelihood analysis (under a given substitution model) 
#	performed on a given multiple sequence alignment (MSA),
#	then it  creates a Substitutional Residue-Difference Map (abbr. "SRD Map").
#	The script was created by taking advantage of some portions of
#	'detect_purge_cands2' (in "MyDetect_purge_cands.pm").
#	[CREATION STARTED on 2020/03/05, ENDED on 2020/03/05.]
#	
# INPUT: ($if_observed, @br_ids_for_display, $char_abs, $char_pres, $topnode_id, \%node2pa, \%node2ntfreqs, \%br2tr_mtrx, \%br2set_sw_jnt_probs, \%node2set_sw_cond_probs, \@set_clm_probs, \@info_psm_hst0, $beg_msa, $end_msa, \%br_attr, \%br2equiv), where
#
#	$if_observed	= 1 if creating the map of Observed SRDs,
#			= 0 if creating the map of Expected SRDs.
#
#	@br_ids_for_display stores ALL branch IDs
#	(, each of which is actually the ID of the node at the lower-end of the branch,)
#	in the order to be displayed as the rows of the SRD Map.
#
#	$char_abs/pres is the character that represents an "absence"/"presence" state.
#
#	%node2pa = ($node_id => $parent_id, ...) for the nodes except the root.
#
#	%node2ntfreqs = ($node_id => \@ntfreqs, ...), where
#	$ntfreqs[$i] is the frequency of $NUCLEOTIDES[$i] at the node, $node_id.
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	%br2set_sw_jnt_probs = ($br_id => \@set_sw_jnt_probs, ...) for ALL branches, where 
#	$set_sw_jnt_probs[$s]->[$i] is the probability that
#	all the extant sequences on the "ancestral side" of $br_id show the residue states that are actually found in the $s th site,
#	in addition to that $NUCLEOTIDES[$i] occupies the $s th site of the UPPER-end node of $br_id.
#	(It is an output of 'calc_joint_prob_ancresstats_and_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#	But $set_sw_cond_probs[$s] = [] (empty) if $node_id refers to an external node and if the node has its $s th site occupied by a gap.
#	(It is an output of 'calc_prob_descresstats_cond_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	$set_clm_probs[$c] is the occurrence probability of the $c th column
#	under the given substitution model.
#	(It is an output of 'calc_clm_probs' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	@info_psm_hst0 = (\@set_cls_gblocks, \@totcts0_indels_for_cls, \@psm_hst0_for_cls),
#	which is an output of 'cal_init_psm_cands_for_sgl_msa' in 'MyTreeMap_indels_spt_odr2.pm',
#	gives information on the Dollo parsimonious indel history giving rise to the input MSA gap configuration, where
#	
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$totcts0_indels_for_cls[$cl] is the total number of insertion/deletion events in the Dollo parsimonious indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA,
#	
#	@{$psm_hst0_for_cls[$cl]} = (\%br2indels0, \%node2blwise_gstat0) is for the Dollo parsimonious indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels0 = ($branch_id => \@list_indels0, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels0[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat0 = ($node_id => \@blwise_gapstats0, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats0[$bl] is the gap-state ($char_pres/abs) of the $bl th block at the node.
#
#	$beg/end_msa is the beginning/ending coordinates of the input MSA (in the 0-based full-closed convention).
#
#	%br_attr = ($br_id => \@attr, ...) for all branches, where $attr[$INDX_LENGTH] is the branch length.
#
#	%br2equiv = ($br_id => $equiv_br_id, ...) for those branches with equivalents.
#

# OUTPUTS: (\@srd_map, \%br2sw_srds), where
#
#	$srd_map[$n]->[$s] = the substitutional residue-difference (SRD) (of the specified type) at the site $s and along the branch,  $br_ids_for_display[$n];
#
#	%br2sw_srds = ($br_id => \@set_sw_srds, ...) for ALL branches, where
#	$set_sw_srds[$s] is the SRD (of the specified type) at the $s th site (more precisely, in the $s th column) of the input MSA, along the branch, $br_id.
#
#	[NOTE]: BOTH $srd_map[$n]->[$s] and $br2sw_srds{$br_id}->[$s] are SET to be $char_abs, meaning "missing" (or "not applicable"),
#		if either end (or both ends) of $br_id (= $br_ids_for_display[$n]) show(s) the initial gap-state of $char_abs at the $s th site.
#
#
sub mk_srd_map ($\@$$$\%\%\%\%\%\@\@$$\%\%) {

	my ($if_observed, $br_ids_for_display, $char_abs, $char_pres, 
	   $topnode_id, $node2pa, $node2ntfreqs, $br2tr_mtrx, $br2set_sw_jnt_probs, $node2set_sw_cond_probs,
	    $set_clm_probs, $info_psm_hst0, $beg_msa, $end_msa, $br_attr, $br2equiv) = @_;

   (print STDOUT "< 'mk_srd_map' (\$if_observed = $if_observed; absence/presence: '$char_abs'/'$char_pres'; $beg_msa <= {MSA coordinate} <= $end_msa) >\n\n") if ($TEST_LEVEL_mk>0);

	my $ct_brs = @{$br_ids_for_display};
	my $ct_brs2 = scalar (keys %{$br2tr_mtrx});

	unless ($ct_brs == $ct_brs2) {
	    ##print STDOUT "[ERROR 'print_srd_map'] The number of branches to be displayed (=$ct_brs) CLASHES with the number of branches in ", '%{$br2tr_mtrx}', " (=$ct_brs2) : $! !\n\n";
	    print STDOUT "[ERROR 'mk_srd_map'] The number of branches to be displayed (=$ct_brs) CLASHES with the number of branches in ", '%{$br2tr_mtrx}', " (=$ct_brs2) : $! !\n\n"; # CORRECTED on 2020/04/03. #
	    return (undef, undef);
	}

	my @set_all_nodes = (@{$br_ids_for_display}, $topnode_id);

   if ($TEST_LEVEL_mk >2) {
	print STDOUT "[ Secondary inputs ]\n\n";
	print STDOUT '@{$br_ids_for_display} = (', join (', ', @{$br_ids_for_display}), ") .\n\n";
	print STDOUT '@set_all_nodes = (', join (',', @set_all_nodes), ").\n\n\n";
   }


			# (Pre-1) Prepare the site-wise gap states (of the Dollo-parsimonious indel history) at every node. #

	# %{$node2sw_gstats0} = ($node_id => \@sw_gstats0, ...) for all nodes, where
	# $sw_gstats0[$s] is the gap state of the $s th site on the node $node_id
	# (usually according to the Dollo-parsimony principle).
	#
	my $node2sw_gstats0 = mk_node2sw_gstats (@{$info_psm_hst0}, $char_abs, $char_pres, $beg_msa, $end_msa, @set_all_nodes);
	unless (defined $node2sw_gstats0) {
	    print STDOUT "[ERROR 'mk_srd_map'] The subroutine 'mk_node2sw_gstats' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
   (print STDOUT " --- (Pre-1) Subroutine 'mk_node2sw_gstats' ended ...\n\n") if ($TEST_LEVEL_mk >1);

   if ($TEST_LEVEL_mk >3) {

	print STDOUT "[ Content of ", '%{$Node2sw_gstats0}', ", which is the output of 'mk_node2sw_gstats' ]\n\n";
	print STDOUT 'Site\Node', "\t", join (', ', @set_all_nodes), "\n\n";
	for (my $s = $beg_msa; $s <= $end_msa; $s++) { # Outer for-loop (over sites). #
	    my @sw_gstats_on_nodes = ();
	    foreach my $node (@set_all_nodes) {
		my $gstat = $node2sw_gstats0->{$node}[$s];
		(defined $gstat) or ($gstat = '?');
		push @sw_gstats_on_nodes, $gstat;
	    }
	    print STDOUT $s, "\t", join (', ', @sw_gstats_on_nodes), "\n";

	}				# End of the outer for-loop (over sites). #

	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL_mk >3) {...}"


		# Main Process 1: Compute RAW %br2sw_srds. #

   (print STDOUT "[ Main Process 1: Compute RAW \%br2sw_srds. ]\n\n") if ($TEST_LEVEL_mk >1);

	my  %br2sw_srds;

	if ($if_observed) { # When creating the "Observed" SRD Map. #

	    my $br2set_cw_obscts_sbsts = calc_br2cw_obscts_sbsts (%{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, @{$set_clm_probs}) ;
	    unless (defined $br2set_cw_obscts_sbsts) {
	        print STDOUT "[ERROR 'mk_srd_map'] The subroutine 'calc_br2cw_obscts_sbsts' FAILED: $! !\n\n";
	        return (undef, undef, undef);
	    }
   (print STDOUT " --- (Pre-2) Subroutine 'calc_br2cw_obscts_sbsts' ended ...\n\n") if ($TEST_LEVEL_mk >1);

	    %br2sw_srds = %{$br2set_cw_obscts_sbsts};

	} else { # When creating the "Expected" SRD Map. #

	    my ($br2set_cw_expcts_sbsts_ave, $br2set_cw_expcts_sbsts_ue, $br2set_cw_expcts_sbsts_le)
	        = calc_br2cw_expcts_sbsts_ave (%{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, %{$node2ntfreqs}, %{$node2pa});
	    unless (defined $br2set_cw_expcts_sbsts_ave) {
	        print STDOUT "[ERROR 'mk_srd_map'] The subroutine 'calc_br2cw_expcts_sbsts_ave' FAILED: $! !\n\n";
	        return (undef, undef, undef);
	    }
   (print STDOUT " --- (Main 1) Subroutine 'calc_br2cw_expcts_sbsts_ave' ended ...\n\n") if ($TEST_LEVEL_mk >1);

	    %br2sw_srds = %{$br2set_cw_expcts_sbsts_ave};

	} # End of "if ($if_observed) {...} else {...}".

   if ($TEST_LEVEL_mk >3) {
	my $name_subroutine = ($if_observed) ? 'calc_br2cw_obscts_sbsts' : 'calc_br2cw_expcts_sbsts_ave' ;

	print STDOUT "[ Content of ", '%$br2sw_srds', ", which is the main output of '$name_subroutine': ]\n\n";

	print STDOUT join ("\t", 'Column\Branch', @{$br_ids_for_display}), "\n\n";

	for (my $c = $beg_msa; $c <= $end_msa; $c++) { # Outer for-loop (over columns). #
	    my @srds_along_brs = ();
	    foreach my $br (@{$br_ids_for_display}) {
		my $sw_srd = $br2sw_srds{$br}->[$c];
		(defined $sw_srd) or ($sw_srd = 'undef');
		push @srds_along_brs, $sw_srd;
	    }
	    print STDOUT join ("\t", $c, @srds_along_brs), "\n";

	} # End of outer for-loop (over columns). #

	print STDOUT "\n\n";

=Alternative idea
	foreach my $br (@{$br_ids_for_display}) {
	    my $sw_srds = $br2sw_srds{$br};
	    print STDOUT "{ Branch ID: $br }\n\n";
	    print STDOUT join ("\t", 'Column(i.e.,site)', '#{srds}'), "\n\n";
	    for (my $s = $beg_msa; $s < $end_msa; $s++) { print STDOUT join ("\t", $s, $sw_srds->[$s]), "\n"; }
	    print STDOUT "\n\n";
	}
=cut

   } # END of "if ($TEST_LEVEL_mk >3) {...}"


		# Main Process 2: Refine %br2sw_srds. #

   (print STDOUT "[ Main Process 2: Refine \%br2sw_srds. ]\n\n") if ($TEST_LEVEL_mk >1);

	foreach my $br (@{$br_ids_for_display}) { # Main outer foreach-loop (over all branches) #

	    my $set_sw_srds = $br2sw_srds{$br};

	    ##my $eq_br = $br2equiv->{$br};
	    my $brlen = $br_attr->{$br}[$INDX_LENGTH];
	    ##if ((defined $eq_br) and ($brlen == 0)) {
	    if ($brlen == 0) {
			# For a null artificial branch,
			#  set all relevant entries to be $char_abs.

		for (my $s = $beg_msa; $s <= $end_msa; $s++) { $set_sw_srds->[$s] = $char_abs; }

		next;
	    }

	    my $node_ue = $node2pa->{$br};
	    my $sw_gstats0_le = $node2sw_gstats0->{$br};
	    my $sw_gstats0_ue = $node2sw_gstats0->{$node_ue};

	    for (my $s = $beg_msa; $s <= $end_msa; $s++) {  # 2nd middle for-loop (over relevant sites). #

		my $gstat0_le = $sw_gstats0_le->[$s];
		my $gstat0_ue = $sw_gstats0_ue->[$s];

		if (($gstat0_le eq $char_abs) or ($gstat0_ue eq $char_abs)) { $set_sw_srds->[$s] = $char_abs; }

	    } # End of the 2nd middle for-loop (over relevant sites). #

	}  # End of the Main outer foreach-loop (over all branches) #


		# Main Process 3: Compute @srd_map. #

   (print STDOUT "[ Main Process 3: Compute \@srd_map. ]\n\n") if ($TEST_LEVEL_mk >1);

	my @srd_map = ();

	for (my $i=0; $i <$ct_brs; $i++) {
	    my $br_id = $br_ids_for_display->[$i];

	    my @set_sw_srds = @{$br2sw_srds{$br_id}}; # Here, an array, which is a value of %br2sw_srds, is copied deliberately, to avoid inadvertent manipulation of @srd_map after it is output.

	    $srd_map[$i] = \@set_sw_srds;
	}

   (print STDOUT "\n\n< 'mk_srd_map' Ended />\n\n") if ($TEST_LEVEL_mk >0);

	return  (\@srd_map, \%br2sw_srds);

}	# End of the subroutine, 'mk_srd_map'. #


# print_srd_map:
#	This subroutine prints out a Substitutional Residue-Difference Map
#	(, which was typically created via 'mk_srd_map',)
#	by formatting it so that the map will be decently visible.
#	This subroutine was created by refering to the subroutine, 'pr_map_shifts_respos',
#	in the script, 'classify_msa_errors_via_mblks.alpha2.pl',
#	in the directory, "$HOME/Ezawa_works_UH/Main_Project/Secondary_Projects/OpenSource/Exported/ComplLiMment_P/Sample_Scripts/".
#	[CREATION STARTED on 2020/03/05, ENDED on 2020/03/06.]
#
# pr_map_shifts_respos:
#	This subroutine prints out a position-shift map
#	by formatting it so that the map will be decently visible.
#	[CREATED on Oct 15 (Thu), 2015.]
#
# [ INPUTS: (\*${$fh}, \@sequence_ids, \@map_shifts_respos, $char_gap, $dis_len_shift), where ] ... for the predecessor, 'pr_map_shifts_respos'.
#
# INPUTS: (\*${$fh}, \@br_ids_for_display, \%node_id2name, \@srd_map, $beg_msa, $end_msa, $char_abs, $dis_len_srd, $ct_eff_dgts), where
#
#	@br_ids_for_display stores ALL branch IDs
#	(, each of which is actually the ID of the node at the lower-end of the branch,)
#	in the order to be displayed as the rows of the SRD Map.
#
#	%node_id2name = ($node_ID => $node_name, ...)  for all NAMED nodes (NOT including top_node_ID => $TOP_NODE).
#
#	$srd_map[$n]->[$s] = the substitutional residue-difference (SRD) (of a specified type) at the site $s and along the branch,  $br_ids_for_display[$n];
#
#	$beg/end_msa is the beginning/ending coordinates of the input MSA (in the 0-based full-closed convention).
#
#	$char_abs is the character that represents an "absence" state.
#
#	$dis_len_srd (positive integer) is the maximum length of the run of SRDs to be displayed in a line.
#
#	$ct_eff_dgts (positive integer) is the number of effective digits of each SRD to be displayed.
#
#
# OUTPUTS: 1 if the subroutine succeeded,
#	   0 if it failed.
#
sub print_srd_map (\*\@\%\@$$$$$) {

	my ($fh, $br_ids_for_display, $node_id2name, $srd_map, $beg_msa, $end_msa, $char_abs, $dis_len_srd, $ct_eff_dgts) = @_;

	my $ct_brs = @{$br_ids_for_display};
	unless ($ct_brs >0) {
	    print STDOUT "[INFORMATION 'print_srd_map'] NO SEQUENCES in the alignment !\n\n";
	    return 1;
	}

	unless (@{$srd_map} == $ct_brs) {
	    print STDOUT "[ERROR 'print_srd_map'] The number of rows in the map (=", scalar (@{$srd_map}), ") CLASHES with the number of branches to be displayed (=$ct_brs): $! !\n\n";
	    return 0;
	}

	my $in_array_len = @{$srd_map->[0]}; 
	unless ($in_array_len > $end_msa) {
	    print STDOUT "[ERROR 'print_srd_map'] The length of the input array, ", '@{$srd_map->[0]}', " is NOT LARGER THAN \$end_msa (=$end_msa): $! !\n\n";
	    return 0;
	}


		# Create a set of row-labels. #

	my @row_labels = ();
	my $maxlen_rlbl = 0;
	for (my $n=0; $n <$ct_brs; $n++) {
	    my $br_id  = $br_ids_for_display->[$n];
	    my $node_name = $node_id2name->{$br_id};
	    my $rlabel = (defined $node_name) ? $node_name : $br_id ;

	    $row_labels[$n] = $rlabel;

	    my $len = length ($rlabel);
	    if ($maxlen_rlbl < $len) { $maxlen_rlbl = $len; }
	}

	####my $wdth_rlbl = ($maxlen_rlbl > 6) ? $maxlen_rlbl : 6;
	##my $wdth_rlbl = ($maxlen_rlbl > length ('Branch/Site') ) ? $maxlen_rlbl : length ('Branch/Site') ; # MODIFIED on 2020/03/11. #
	my $wdth_rlbl = ($maxlen_rlbl > length ('Branch\Site') ) ? $maxlen_rlbl : length ('Branch\Site') ; # => CORRECTED on 2020/03/11. #

		# Set the maginification ratio. #

	my $magratio = 1;
	for (my $i = 0; $i < $ct_eff_dgts; $i++) { $magratio *= 10; }

	my $wdth_fld = ($ct_eff_dgts + 3 > length ($end_msa) +2 ) ? $ct_eff_dgts + 3 : length ($end_msa) + 2;

		# Set the $quotient #

	my $map_len = $end_msa - $beg_msa + 1;

	my $quotient = int (($map_len -1) / $dis_len_srd);

	#    sprintf ("%-${maxlen_rlbl}s", $row_labels[$n]);


	for (my $r=0; $r<=$quotient; $r++) {  # 1st outer for-loop (over segmented displayes). #

	    my $beg_dis = $r * $dis_len_srd;
	    my $end_dis = ($r == $quotient) ? $map_len - 1 : ($r+1) * $dis_len_srd - 1 ;
	    $beg_dis += $beg_msa;
	    $end_dis += $beg_msa;

	    ##my $format = "%-10s\t" . "%5d" x (($r==$quotient) ? $map_len - $quotient * $dis_len_srd : $dis_len_srd) . "\n"; # For the predecessor, 'pr_map_shifts_respos'. #
	    my $format_clbl = "%-${wdth_rlbl}s\t" . "%${wdth_fld}d" x (($r==$quotient) ? $map_len - $quotient * $dis_len_srd : $dis_len_srd) . "\n";
	    ##my $format = "%-${wdth_rlbl}s\t" . "%${wdth_fld}s" x (($r==$quotient) ? $map_len - $quotient * $dis_len_srd : $dis_len_srd) . "\n";


	    ######printf $fh $format, 'Seq_ID', ($beg_dis .. $end_dis);
	    ####printf $fh $format_clbl, 'Seq_ID', ($beg_dis .. $end_dis); # CORRECTED on 2020/03/09.
	    ##printf $fh $format_clbl, 'Branch/Site', ($beg_dis .. $end_dis); # MODIFIED on 2020/03/11.
	    printf $fh $format_clbl, 'Branch\Site', ($beg_dis .. $end_dis); # => CORRECTED on 2020/03/11.
	    print $fh "\n";

	    for (my $k=0; $k<$ct_brs; $k++) {  # Middle for-loop (over branches to be displayed). #

		##my $out_string = sprintf ("%-${wdth_rlbl}s\t", $br_ids_for_display->[$k]);
		my $out_string = sprintf ("%-${wdth_rlbl}s\t", $row_labels[$k]); # CORRECTED on 2020/03/09. #

		my $row_srds = $srd_map->[$k];
		my $ct_sites = @{$row_srds};

		my $end_dis2 = ($ct_sites-1 < $end_dis) ? $ct_sites-1 : $end_dis;

		for (my $s=$beg_dis; $s<= $end_dis2; $s++) { # Inner for-loop (over sites). #

		   my $srd = $row_srds->[$s];
		   my $out_srd;
		   if ($srd eq $char_abs) {
			$out_srd = $srd;
		   } else {
			$out_srd = int ($magratio * $srd + 1/2) / $magratio;
			if ($out_srd > 0) { $out_srd =~ s/^0//; }
		   }
		   ##$out_string .= ($shift eq $char_gap) ? sprintf ("%5s", "'$char_gap'") : sprintf ("%5d", $shift) ; # For the predecessor, 'pr_map_shifts_respos'. #
		   $out_string .= sprintf ("%${wdth_fld}s", $out_srd) ;

		}				# End of the inner for-loop (over sites). #

		print $fh $out_string, "\n";
	    
	    }			# End of the middle for-loop (over branches to be displayed). #
	    print $fh "\n\n\n";

	}			# End of the 1st outer for-loop (over segmented displayes). #

	return 1;

}	# End of the subroutine, 'print_srd_map'. #


=Used as a Reference to 'print_srd_map'.
	# (IMPORTED from 'classify_msa_errors_via_mblks.alpha2.pl'
	# in "$HOME/Ezawa_works_UH/Main_Project/Secondary_Projects/OpenSource/Exported/ComplLiMment_P/Sample_Scripts/".)

# pr_map_shifts_respos:
#	This subroutine prints out a position-shift map
#	by formatting it so that the map will be decently visible.
#	[CREATED on Oct 15 (Thu), 2015.]
#
# INPUTS: (\*${$fh}, \@sequence_ids, \@map_shifts_respos, $char_gap, $dis_len_shift), where
#	$fh is the reference to a file-handle;
#	$sequence_ids[$k] is the ID of the sequence in the $k th row of the MSAs;
#	$map_shifts_respos[$k]->[$s] is the position-shift at the intersection of the $k th row and the $s th column of the mapped MSA;
#	$char_gap is the character representing a gap;
#	$dis_len_shift is the maximum length of the run of shifts to be displayed in a line.
#
# OUTPUTS: 1 if the subroutine succeeded,
#	   0 if it failed.
#
sub pr_map_shifts_respos (\*\@\@$$) {

	my ($fh, $sequence_ids, $map_shifts_respos, $char_gap, $dis_len_shift) = @_;

	my $ct_sequeunces = @{$sequence_ids};
	unless ($ct_sequences>0) {
	    print STDOUT "[INFORMATION 'pr_map_shifts_respos'] NO SEQUENCES in the alignment !\n\n";
	    return 1;
	}
	unless (@{$map_shifts_respos} == $ct_sequences) {
	    print STDOUT "[ERROR 'pr_map_shifts_respos'] The number of sequences in the map (=", scalar (@{$map_shifts_respos}), ") CLASHES with that in the set of sequences (=$ct_sequences): $! !\n\n";
	    return 0;
	}

	my $map_len = @{$map_shifts_respos->[0]}; 

	my $quotient = int (($map_len -1) / $dis_len_shift);

	for (my $r=0; $r<=$quotient; $r++) {  # 1st outer for-loop (over segmented displayes). #

	    my $beg_dis = $r * $dis_len_shift;
	    my $end_dis = ($r == $quotient) ? $map_len - 1 : ($r+1) * $dis_len_shift - 1 ;
	    my $format = "%-10s\t" . "%5d" x (($r==$quotient) ? $map_len - $quotient * $dis_len_shift : $dis_len_shift) . "\n";

	    printf $fh $format, 'Seq_ID', ($beg_dis .. $end_dis);
	    print $fh "\n";

	    for (my $k=0; $k<$ct_sequences; $k++) {  # Middle for-loop (over sequences). #

		my $out_string = sprintf ("%-10s\t", $sequence_ids->[$k]);

		my $row_shifts = $map_shifts_respos->[$k];
		my $ct_sites = @{$row_shifts};

		my $end_dis2 = ($ct_sites-1 < $end_dis) ? $ct_sites-1 : $end_dis;
		for (my $s=$beg_dis; $s<= $end_dis2; $s++) { # Inner testing for-loop (over sequences). #
		   my $shift = $row_shifts->[$s];
		   $out_string .= ($shift eq $char_gap) ? sprintf ("%5s", "'$char_gap'") : sprintf ("%5d", $shift) ;
		}				# End of the inner testing for-loop (over sequences). #

		print $fh $out_string, "\n";
	    
	    }			# End of the middle for-loop (over sequences). #
	    print $fh "\n\n\n";

	}			# End of the 1st outer for-loop (over segmented displayes). #

	return 1;

}		# End of the subroutine 'pr_map_shifts_respos'. #

=cut # End of "Used as a Reference to 'print_srd_map'."


# tally_assoc_srds_and_alnshears:
#	This subroutine computes the average sums of 1's, SRDs, and {SRD^2}s,
#	[ in and out of "Alignment-Shear"s. ] ... OBSOLETE as of 2020/03/25.
#	'on' and 'off' "Alignment-Shear"s.
#	[CREATED on 2020/03/24.]
#
#	[NOTE]:
#	This subroutine gives the same weight, 1/#{Alignment-Shear Maps},
#	to each of the Alignment-Shear Maps (typically corresponding to the map (psp-blocks -> psm-clusterings)).
#
# INPUTS: (\@set_maps_alnshear, \@srd_map, $char_abs, $char_noshear), where
#
#	@map_alnshear = @{$set_maps_alnshear[$mp]} is for the $mp th Alignment-Shear Map, where
#	$map_alnshear[$n]->[$s] = the Alignment-Shear status of the position specified by the site $s and the branch,  $br_ids_for_display[$n];
#	The "Alignment-Shear status" is:
#		 $char_abs if either end (or both ends) of the branch has (have) $char_abs in the initial gap-state,
#		 join (',', {list of mini-classes whose boundaries includ the position}),
#		 $char_noshear otherwise.
#	[Typically, \@map_alnshear is a main output of 'mk_map_alnshear_clmbased' (in "MyMapAlnShear.pm").]
# 
#	$srd_map[$n]->[$s] = the substitutional residue-difference (SRD) (of a specified type) at the site $s and along the branch,  $br_ids_for_display[$n];
#	[Typically, \@srd_map is a main output of 'mk_srd_map' above.]
#
#	$char_abs is the character to represent an "absence" state;
#	$char_noshear is the character indicating that the position (= site x branch) has NO shear. 
#
# OUTPUTS: (\@set_total_sums_srd_vs_alnshear, \@setset_brwise_sums_srd_vs_alnshear), where
#
#	@set_total_sums_srd_vs_alnshear
#	 = ($total_avsum_wgts_on_as, $total_avsum_srds_on_as, $total_avsum_sq_srds_on_as,
#		  $total_avsum_wgts_off_as, $total_avsum_srds_off_as, $total_avsum_sq_srds_off_as)
#	gives the total of the statistics over all branches, and
#
#	@{$setset_brwise_sums_srd_vs_alnshear[$n]}
#	 = ($bw_avsum_wgts_on_as, $bw_avsum_srds_on_as, $bw_avsum_sq_srds_on_as,
#		  $bw_avsum_wgts_off_as, $bw_avsum_srds_off_as, $bw_avsum_sq_srds_off_as)
#	gives the statistics along the branch, $br_ids_for_display[$n].
#
#	In both cases, 'avsum' stands for the "Average Sum(mation)",
#			'wgts' stands for the "weights" (, each of which is 1/scalar (@set_maps_alnshear) for each Alignment-shear Map),
#			'srds' stands for the "Substitutional Residue-Difference"s (SRDs),
#			'sq_' stands for the "square",
#			and 'as' stands for the "Alignment-Shear".
#
sub tally_assoc_srds_and_alnshears (\@\@$$) {

	my ($set_maps_alnshear, $srd_map, $char_abs, $char_noshear) = @_;

	my $ct_maps = @{$set_maps_alnshear};

	unless ($ct_maps>0) {
	    print STDOUT "[ERROR 'tally_assoc_srds_and_alnshears'] ILLEGAL INPUT: NO Alignment-Shear MAPS in the input, ", '@{$set_maps_alnshear}', ": $! !\n\n";
	    return (undef, undef);
	}

	my $ct_brs = @{$srd_map};


		# Initialize the Outputs. #

	my @set_total_sums_srd_vs_alnshear = (0, 0, 0, 0, 0, 0);
		#  = ($total_avsum_wgts_in_as, $total_avsum_srds_in_as, $total_avsum_sq_srds_in_as,
		#	  $total_avsum_wgts_outof_as, $total_avsum_srds_outof_as, $total_avsum_sq_srds_outof_as).

	my $ct_total_sums = @set_total_sums_srd_vs_alnshear;

	my @setset_brwise_sums_srd_vs_alnshear = ();
	for (my $n=0; $n <$ct_brs; $n++) {  # Preliminary for-loop (over the branches). #
	    my @cp_set = @set_total_sums_srd_vs_alnshear;
	    $setset_brwise_sums_srd_vs_alnshear[$n] = \@cp_set;
	}

		# Main Process 1: Tally the SRDs in and out of "Alignment-Shear"s, along each branch. #

	for (my $mp=0; $mp < $ct_maps; $mp++) { # 1st Main outer for-loop (over the Alignment-Shear Maps). #

	    my $map_alnshear = $set_maps_alnshear->[$mp];

	    for (my $n=0; $n <$ct_brs; $n++) { # Main middle for-loop (over the branches). #

		my $set_brwise_sums_vs_alnshear = $setset_brwise_sums_srd_vs_alnshear[$n];

		my $brwise_sums = $set_brwise_sums_vs_alnshear; # Abbreviate the long name.

		my $linmap_alnshear_br = $map_alnshear->[$n];

		##my $linmap_srd = $srd_map->[$n];
		my $linmap_srd_br = $srd_map->[$n]; # Modified on 2020/03/25.

		##my $ct_sites = @{$alnshear_along_br};
		my $ct_sites = @{$linmap_alnshear_br}; # CORRECTED on 2020/03/25.

		for (my $s=0; $s <$ct_sites; $s++) { # Main inner for-loop (over the sites). #

		    my $as_status = $linmap_alnshear_br->[$s];

		    if ($as_status eq $char_abs) { next; }

		    ##my $srd = $linmap_srd->[$s];
		    my $srd = $linmap_srd_br->[$s]; # Modified on 2020/03/25.
		    my $sq_srd = $srd * $srd;

		    if ($as_status eq $char_noshear) {
				# Out of Alignment-Shears. #
			$brwise_sums->[3] += 1;
			$brwise_sums->[3+1] += $srd;
			$brwise_sums->[5] += $sq_srd;
		    } else {
				# In an Alignment-Shear. #
			$brwise_sums->[0] += 1;
			$brwise_sums->[1] += $srd;
			$brwise_sums->[2] += $sq_srd;
		    }
		    
		} # End of the main inner for-loop (over the sites). #

	    } # End of the main middle for-loop (over the branches). #

	} # End of the 1st Main outer for-loop (over the Alignment-Shear Maps). #
	

		# Main Process 2: Divide the branch-wise summations by #{Alignment-Shear Maps},
		#			while totalling the summations #

	my $total_sums = \@set_total_sums_srd_vs_alnshear; # Abbreriate the long name.

	for (my $n=0; $n <$ct_brs; $n++) { # 2nd Main outer for-loop (over the branches). #

	    my $set_brwise_sums_vs_alnshear = $setset_brwise_sums_srd_vs_alnshear[$n];

	    my $brwise_sums = $set_brwise_sums_vs_alnshear; # Abbreviate the long name.

	    for (my $sm =0; $sm <$ct_total_sums; $sm++) { # Inner for-loop (over the summations). #
		$total_sums->[$sm] += $brwise_sums->[$sm];
		$brwise_sums->[$sm] /= $ct_maps;
	    } # End of the inner for-loop (over the summations). #

	} # 2nd Main outer for-loop (over the branches). #


		# Main Process 3: Divide the total summations by #{Alignment-Shear Maps},

	for (my $sm =0; $sm <$ct_total_sums; $sm++) { # Inner for-loop (over the summations). #
		$total_sums->[$sm] /= $ct_maps;
	} # End of the inner for-loop (over the summations). #


	return (\@set_total_sums_srd_vs_alnshear, \@setset_brwise_sums_srd_vs_alnshear);

}	# End of the subroutine, 'tally_assoc_srds_and_alnshears'. #


# print_sums_assoc_srds_and_alnshears:
#	This subroutine prints out the output of 'tally_assoc_srds_and_alnshears' above.
#	The subroutine was created by taking advantage of the 1st half of 'print_srd_map'.
#	[CREATION STARTED on 2020/03/24, ENDED on 2020/03/25.]
#
#
# INPUTS: (\*${$fh}, \@br_ids_for_display, \%node_id2name, 
#		\@set_total_sums_srd_vs_alnshear, \@setset_brwise_sums_srd_vs_alnshear), where
#
#	@br_ids_for_display stores ALL branch IDs
#	(, each of which is actually the ID of the node at the lower-end of the branch,)
#	in the order to be displayed as the rows of the SRD Map.
#
#	%node_id2name = ($node_ID => $node_name, ...)  for all NAMED nodes (NOT including top_node_ID => $TOP_NODE).
#
#	@set_total_sums_srd_vs_alnshear
#	 = ($total_avsum_wgts_on_as, $total_avsum_srds_on_as, $total_avsum_sq_srds_on_as,
#		  $total_avsum_wgts_off_as, $total_avsum_srds_off_as, $total_avsum_sq_srds_off_as)
#	gives the total of the statistics over all branches, and
#
#	@{$setset_brwise_sums_srd_vs_alnshear[$n]}
#	 = ($bw_avsum_wgts_on_as, $bw_avsum_srds_on_as, $bw_avsum_sq_srds_on_as,
#		  $bw_avsum_wgts_off_as, $bw_avsum_srds_off_as, $bw_avsum_sq_srds_off_as)
#	gives the statistics along the branch, $br_ids_for_display[$n].
#
#	[Typically, \@set_total_sums_srd_vs_alnshear & \@setset_brwise_sums_srd_vs_alnshear
#			are the main outputs of 'tally_assoc_srds_and_alnshears';
#	For more details on the elements of these arrays, refer to 'tally_assoc_srds_and_alnshears'.]
#
#
# OUTPUTS: 1 if the subroutine succeeded,
#	   0 if it failed.
#
sub print_sums_assoc_srds_and_alnshears (\*\@\%\@\@) {

	my ($fh, $br_ids_for_display, $node_id2name, $set_total_sums_srd_vs_alnshear, $setset_brwise_sums_srd_vs_alnshear) = @_;

	my $ct_brs = @{$br_ids_for_display};
	unless ($ct_brs >0) {
	    print STDOUT "[INFORMATION 'print_sums_assoc_srds_and_alnshears'] NO SEQUENCES in the alignment !\n\n";
	    return 1;
	}


		# Preliminary: Create a set of row-labels. #

	my @row_labels = ();
	for (my $n=0; $n <$ct_brs; $n++) { # Preliminary for-loop (over the branches). #
	    my $br_id  = $br_ids_for_display->[$n];
	    my $node_name = $node_id2name->{$br_id};
	    my $rlabel = (defined $node_name) ? $node_name : $br_id ;

	    $row_labels[$n] = $rlabel;
	} # End of the preliminary for-loop (over the branches). #


		# MAIN: Print out the Main inputs. #

#	@{$setset_brwise_sums_srd_vs_alnshear[$n]}
#	 = ($bw_avsum_wgts_on_as, $bw_avsum_srds_on_as, $bw_avsum_sq_srds_on_as,
#		  $bw_avsum_wgts_off_as, $bw_avsum_srds_off_as, $bw_avsum_sq_srds_off_as)
#	gives the statistics along the branch, $br_ids_for_display[$n].
#
#	In both cases, 'avsum' stands for the "Average Sum(mation)",
#			'wgts' stands for the "weights" (, each of which is 1/scalar (@set_maps_alnshear) for each Alignment-shear Map),
#			'srds' stands for the "Substitutional Residue-Difference"s (SRDs),
#			'sq_' stands for the "square",
#			and 'as' stands for the "Alignment-Shear".


	print $fh join ("\t", 'Branch_ID', 'Ave{Sum(weights)}_on_as', 'Ave{Sum(SRDs)}_on_as', 'Ave{Sum(SRD^2s)}_on_as',
				'Ave{Sum(weights)}_off_as', 'Ave{Sum(SRDs)}_off_as', 'Ave{Sum(SRD^2s)}_off_as') , "\n\n";

	for (my $n=0; $n <$ct_brs; $n++) { # Main for-loop (over the branches). #

	    print $fh join ("\t", $row_labels[$n], @{$setset_brwise_sums_srd_vs_alnshear->[$n]}), "\n";

	} # End of the Main for-loop (over the branches). #

	print $fh "\n";

	print $fh join ("\t", 'Total', @{$set_total_sums_srd_vs_alnshear}), "\n\n";

	return 1;

}	# End of the subroutine, 'print_sums_assoc_srds_and_alnshears'. #


# tally_srds_in_map:
#	This subroutine computes the sums of 1's, SRDs, and {SRD^2}s,
#	in an input SRD map.
#	It was created by stripping some functions off the predecessor,
#	'tally_assoc_srds_and_alnshears'.
#	[CREATED on 2020/04/03.]
#
# tally_assoc_srds_and_alnshears:
#	This subroutine computes the average sums of 1's, SRDs, and {SRD^2}s,
#	[ in and out of "Alignment-Shear"s. ] ... OBSOLETE as of 2020/03/25.
#	'on' and 'off' "Alignment-Shear"s.
#	[CREATED on 2020/03/24.]
#
#	[NOTE]:
#	This subroutine gives the same weight, 1/#{Alignment-Shear Maps},
#	to each of the Alignment-Shear Maps (typically corresponding to the map (psp-blocks -> psm-clusterings)).
#
# [ INPUTS: (\@set_maps_alnshear, \@srd_map, $char_abs, $char_noshear), where ] ... for the predecessor, 'tally_assoc_srds_and_alnshears'.
#
# INPUTS: (\@srd_map, $char_abs), where
#
#	$srd_map[$n]->[$s] = the substitutional residue-difference (SRD) (of a specified type) at the site $s and along the branch,  $br_ids_for_display[$n];
#	[NOTE]: $srd_map[$n]->[$s] is SET to be $char_abs, meaning "missing" (or "not applicable"),
#		if either end (or both ends) of $br_id (= $br_ids_for_display[$n]) show(s) the initial gap-state of $char_abs at the $s th site.
#	[Typically, \@srd_map is a main output of 'mk_srd_map' above.]
#
#	$char_abs is the character to represent an "absence" state;
#
# OUTPUTS: (\@set_total_sums_srd, \@setset_brwise_sums_srd), where
#
#	@set_total_sums_srd
#	 = ($total_sum_wgts, $total_sum_srds, $total_sum_sq_srds)
#	gives the total of the statistics over all branches, and
#
#	@{$setset_brwise_sums_srd[$n]}
#	 = ($bw_sum_wgts, $bw_sum_srds, $bw_sum_sq_srds)
#	gives the statistics along the branch, $br_ids_for_display[$n].
#
#	In both cases, 'sum' stands for the "Sum(mation)",
#			'wgts' stands for the "weights" (, each of which is 1 for each relevant site),
#			'srds' stands for the "Substitutional Residue-Difference"s (SRDs),
#			'sq_' stands for the "square".
#
# sub tally_assoc_srds_in_map (\@$) { # A BUG fixed on 2020/04/03. #
#
sub tally_srds_in_map (\@$) {

	my ($srd_map, $char_abs) = @_;

	my $ct_brs = @{$srd_map};


		# Initialize the Outputs. #

	my @set_total_sums_srd = (0, 0, 0);
		#  = ($total_sum_wgts, $total_sum_srds, $total_sum_sq_srds).

	my $ct_total_sums = @set_total_sums_srd;

	my @setset_brwise_sums_srd = ();
	for (my $n=0; $n <$ct_brs; $n++) {  # Preliminary for-loop (over the branches). #
	    my @cp_set = @set_total_sums_srd;
	    $setset_brwise_sums_srd[$n] = \@cp_set;
	}


		# Main Process 1: Tally the SRDs along each branch. #

	for (my $n=0; $n <$ct_brs; $n++) { # 1st Main Outer for-loop (over the branches). #

		my $set_brwise_sums_srd = $setset_brwise_sums_srd[$n];

		my $brwise_sums = $set_brwise_sums_srd; # Abbreviate the long name.

		my $linmap_srd_br = $srd_map->[$n];

		my $ct_sites = @{$linmap_srd_br};

		for (my $s=0; $s <$ct_sites; $s++) { # Main inner for-loop (over the sites). #

		    my $srd = $linmap_srd_br->[$s];
		    if ($srd eq $char_abs) { next; }

		    my $sq_srd = $srd * $srd;

		    $brwise_sums->[0] += 1;
		    $brwise_sums->[1] += $srd;
		    $brwise_sums->[2] += $sq_srd;
		    
		} # End of the main inner for-loop (over the sites). #

	} # End of the 1st Main Outer for-loop (over the branches). #


		# Main Process 2: Total the branch-wise summations.

	my $total_sums = \@set_total_sums_srd; # Abbreriate the long name.

	for (my $n=0; $n <$ct_brs; $n++) { # 2nd Main outer for-loop (over the branches). #

	    my $set_brwise_sums_srd = $setset_brwise_sums_srd[$n];

	    my $brwise_sums = $set_brwise_sums_srd; # Abbreviate the long name.

	    for (my $sm =0; $sm <$ct_total_sums; $sm++) { # Inner for-loop (over the summations). #
		$total_sums->[$sm] += $brwise_sums->[$sm];
		##$brwise_sums->[$sm] /= $ct_maps; # For the predecessor, 'tally_assoc_srds_and_alnshears'.
	    } # End of the inner for-loop (over the summations). #

	} # 2nd Main outer for-loop (over the branches). #


=For the predecessor, 'tally_assoc_srds_and_alnshears'.
		# Main Process 3: Divide the total summations by #{Alignment-Shear Maps},

	for (my $sm =0; $sm <$ct_total_sums; $sm++) { # Inner for-loop (over the summations). #
		$total_sums->[$sm] /= $ct_maps;
	} # End of the inner for-loop (over the summations). #
=cut # End of "For the predecessor, 'tally_assoc_srds_and_alnshears'."


	return (\@set_total_sums_srd, \@setset_brwise_sums_srd);

}	# End of the subroutine, 'tally_srds_in_map'. #


# print_sums_srds_in_map:
#	This subroutine prints out the output of 'tally_srds_in_map' above.
#	It was created by stripping some functions off the predecessor, 'print_sums_assoc_srds_and_alnshears'.
#	[CREATED on 2020/04/03.]
#
# print_sums_assoc_srds_and_alnshears:
#	This subroutine prints out the output of 'tally_assoc_srds_and_alnshears' above.
#	The subroutine was created by taking advantage of the 1st half of 'print_srd_map'.
#	[CREATION STARTED on 2020/03/24, ENDED on 2020/03/25.]
#
#
# [ INPUTS: (\*${$fh}, \@br_ids_for_display, \%node_id2name, 
#		\@set_total_sums_srd_vs_alnshear, \@setset_brwise_sums_srd_vs_alnshear), where ] ... for the predecessor, 'print_sums_assoc_srds_and_alnshears'.
#
# INPUTS: (\*${$fh}, \@br_ids_for_display, \%node_id2name, 
#		\@set_total_sums_srd, \@setset_brwise_sums_srd), where
#
#	@br_ids_for_display stores ALL branch IDs
#	(, each of which is actually the ID of the node at the lower-end of the branch,)
#	in the order to be displayed as the rows of the SRD Map.
#
#	%node_id2name = ($node_ID => $node_name, ...)  for all NAMED nodes (NOT including top_node_ID => $TOP_NODE).
#
#	@set_total_sums_srd
#	 = ($total_sum_wgts, $total_sum_srds, $total_sum_sq_srds)
#	gives the total of the statistics over all branches, and
#
#	@{$setset_brwise_sums_srd[$n]}
#	 = ($bw_avsum_wgts, $bw_avsum_srds, $bw_avsum_sq_srds)
#	gives the statistics along the branch, $br_ids_for_display[$n].
#
#	[Typically, \@set_total_sums_srd & \@setset_brwise_sums_srd
#			are the main outputs of 'tally_srds_in_map' above;
#	For more details on the elements of these arrays, refer to 'tally_srds_in_map'.]
#
#
# OUTPUTS: 1 if the subroutine succeeded,
#	   0 if it failed.
#
sub print_sums_srds_in_map (\*\@\%\@\@) {

	my ($fh, $br_ids_for_display, $node_id2name, $set_total_sums_srd, $setset_brwise_sums_srd) = @_;

	my $ct_brs = @{$br_ids_for_display};
	unless ($ct_brs >0) {
	    print STDOUT "[INFORMATION 'print_sums_srds_in_map'] NO SEQUENCES in the alignment !\n\n";
	    return 1;
	}


		# Preliminary: Create a set of row-labels. #

	my @row_labels = ();
	for (my $n=0; $n <$ct_brs; $n++) { # Preliminary for-loop (over the branches). #
	    my $br_id  = $br_ids_for_display->[$n];
	    my $node_name = $node_id2name->{$br_id};
	    my $rlabel = (defined $node_name) ? $node_name : $br_id ;

	    $row_labels[$n] = $rlabel;
	} # End of the preliminary for-loop (over the branches). #


		# MAIN: Print out the Main inputs. #

	print $fh join ("\t", 'Branch_ID', 'Sum(weights)', 'Sum(SRDs)', 'Sum(SRD^2s)') , "\n\n";

	for (my $n=0; $n <$ct_brs; $n++) { # Main for-loop (over the branches). #

	    print $fh join ("\t", $row_labels[$n], @{$setset_brwise_sums_srd->[$n]}), "\n";

	} # End of the Main for-loop (over the branches). #

	print $fh "\n";

	print $fh join ("\t", 'Total', @{$set_total_sums_srd}), "\n\n";

	return 1;

}	# End of the subroutine, 'print_sums_srds_in_map'. #


# brwise_distribute_srds_assoc_w_alnshear:
#	This subroutine distributes SRDs on each branch, 'on' and 'off' "Alignment-Shear"s,
#	into the corresponding specified sets of bins.
#	The subroutine was created by taking advantage of the framework of 'tally_assoc_srds_and_alnshears'.
#	[CREATED on 2020/03/26.]
#
#	[NOTE]: If there are more than one Alignment-Shear (AS) Map,
#		each AS map contributes the weight of 1/#{AS Maps} to each relevant branch x site.
#
# tally_assoc_srds_and_alnshears:
#	This subroutine computes the average sums of 1's, SRDs, and {SRD^2}s,
#	[ in and out of "Alignment-Shear"s. ] ... OBSOLETE as of 2020/03/25.
#	'on' and 'off' "Alignment-Shear"s.
#	[CREATED on 2020/03/24.]
#
#	[NOTE]:
#	This subroutine gives the same weight, 1/#{Alignment-Shear Maps},
#	to each of the Alignment-Shear Maps (typically corresponding to the map (psp-blocks -> psm-clusterings)).
#
# INPUTS: (\@set_maps_alnshear, \@srd_map, $char_abs, $char_noshear,
#	\@brwise_distri_srds_on_as, \@brwise_distri_srds_off_as, \@bounds_srd), where
#
#	@map_alnshear = @{$set_maps_alnshear[$mp]} is for the $mp th Alignment-Shear Map, where
#	$map_alnshear[$n]->[$s] = the Alignment-Shear status of the position specified by the site $s and the branch,  $br_ids_for_display[$n];
#	The "Alignment-Shear status" is:
#		 $char_abs if either end (or both ends) of the branch has (have) $char_abs in the initial gap-state,
#		 join (',', {list of mini-classes whose boundaries includ the position}),
#		 $char_noshear otherwise.
#	[Typically, \@map_alnshear is a main output of 'mk_map_alnshear_clmbased' (in "MyMapAlnShear.pm").]
# 
#	$srd_map[$n]->[$s] = the substitutional residue-difference (SRD) (of a specified type) at the site $s and along the branch,  $br_ids_for_display[$n];
#	[Typically, \@srd_map is a main output of 'mk_srd_map' above.]
#
#	$char_abs is the character to represent an "absence" state;
#	$char_noshear is the character indicating that the position (= site x branch) has NO shear. 
#
#	@{$brwise_distri_srds_on/off_as[$n]} (for 0 <= $n < #{branches}) is the set of bins constituting a distribution of SRDs 'on'/'off' the "Alignment-Shear" along the branch, $br_ids_for_display[$n],
#	@{$brwise_distri_srds_on/off_as[#{branches}]} is the set of bins constituting a TOTAL distribution over all branches.
#
#	@bounds_srd is a set of boundaries used for distributing SRDs.
#
#
# OUTPUTS: 1 if the subroutine succeeded,
#	   0 if it failed.
#
sub brwise_distribute_srds_assoc_w_alnshear (\@\@$$\@\@\@) {

	my ($set_maps_alnshear, $srd_map, $char_abs, $char_noshear,
	    $brwise_distri_srds_on_as, $brwise_distri_srds_off_as, $bounds_srd) = @_;

	my $ct_maps = @{$set_maps_alnshear};

	unless ($ct_maps>0) {
	    print STDOUT "[ERROR 'brwise_distribute_srds_assoc_w_alnshear'] ILLEGAL INPUT: NO Alignment-Shear MAPS in the input, ", '@{$set_maps_alnshear}', ": $! !\n\n";
	    return 0;
	}

	my $wgt  = 1/$ct_maps;

	my $ct_brs = @{$srd_map};

	##my $tot_distri_srds_on_as = $brwise_distri_srds_on_as->[$ct_brs];
	##my $tot_distri_srds_off_as = $brwise_distri_srds_off_as->[$ct_brs];


		# Main Process: Distribute the SRDs on and off "Alignment-Shear"s, along each branch. #

	for (my $mp=0; $mp < $ct_maps; $mp++) { # Main outer for-loop (over the Alignment-Shear Maps). #

	    my $map_alnshear = $set_maps_alnshear->[$mp];

	    for (my $n=0; $n <$ct_brs; $n++) { # Main middle for-loop (over the branches). #

		my $linmap_alnshear_br = $map_alnshear->[$n];
		my $linmap_srd_br = $srd_map->[$n];

		my $distri_srds_on_as_br = $brwise_distri_srds_on_as->[$n];
		my $distri_srds_off_as_br = $brwise_distri_srds_off_as->[$n];

		my $ct_sites = @{$linmap_alnshear_br};

		for (my $s=0; $s <$ct_sites; $s++) { # Main inner for-loop (over the sites). #

		    my $as_status = $linmap_alnshear_br->[$s];

		    if ($as_status eq $char_abs) { next; }

		    my $srd = $linmap_srd_br->[$s];

		    if ($as_status eq $char_noshear) {
				# Off Alignment-Shears. #
			cluster_distributor_wgt ($srd, @{$distri_srds_off_as_br}, @{$bounds_srd}, $wgt);

		    } else {
				# On an Alignment-Shear. #
			cluster_distributor_wgt ($srd, @{$distri_srds_on_as_br}, @{$bounds_srd}, $wgt);
		    }
		    
		} # End of the main inner for-loop (over the sites). #

	    } # End of the main middle for-loop (over the branches). #

	} # End of the Main outer for-loop (over the Alignment-Shear Maps). #

	return 1;

}	# End of the subroutine, 'brwise_distribute_srds_assoc_w_alnshear'. #


# brwise_distribute_srds_in_map:
#	This subroutine distributes SRDs on each branch, in an input SRD map,
#	into the corresponding specified sets of bins.
#	It was created by stripping some functions off the predecessor, 'brwise_distribute_srds_assoc_w_alnshear'.
#	[CREATED on 2020/04/03.]
#
# brwise_distribute_srds_assoc_w_alnshear:
#	This subroutine distributes SRDs on each branch, 'on' and 'off' "Alignment-Shear"s,
#	into the corresponding specified sets of bins.
#	The subroutine was created by taking advantage of the framework of 'tally_assoc_srds_and_alnshears'.
#	[CREATED on 2020/03/26.]
#
#	[NOTE]: If there are more than one Alignment-Shear (AS) Map,
#		each AS map contributes the weight of 1/#{AS Maps} to each relevant branch x site.
#
# tally_assoc_srds_and_alnshears:
#	This subroutine computes the average sums of 1's, SRDs, and {SRD^2}s,
#	[ in and out of "Alignment-Shear"s. ] ... OBSOLETE as of 2020/03/25.
#	'on' and 'off' "Alignment-Shear"s.
#	[CREATED on 2020/03/24.]
#
#	[NOTE]:
#	This subroutine gives the same weight, 1/#{Alignment-Shear Maps},
#	to each of the Alignment-Shear Maps (typically corresponding to the map (psp-blocks -> psm-clusterings)).
#
# [ INPUTS: (\@set_maps_alnshear, \@srd_map, $char_abs, $char_noshear,
#	\@brwise_distri_srds_on_as, \@brwise_distri_srds_off_as, \@bounds_srd), where ] ... for the predecessor, 'brwise_distribute_srds_assoc_w_alnshear'.
#
# INPUTS: (\@srd_map, $char_abs, \@brwise_distri_srds, \@bounds_srd), where
#
#	$srd_map[$n]->[$s] = the substitutional residue-difference (SRD) (of a specified type) at the site $s and along the branch,  $br_ids_for_display[$n];
#	[NOTE]: $srd_map[$n]->[$s] is SET to be $char_abs, meaning "missing" (or "not applicable"),
#		if either end (or both ends) of $br_id (= $br_ids_for_display[$n]) show(s) the initial gap-state of $char_abs at the $s th site.
#	[Typically, \@srd_map is a main output of 'mk_srd_map' above.]
#
#	$char_abs is the character to represent an "absence" state;
#
#	@{$brwise_distri_srds[$n]} (for 0 <= $n < #{branches}) is the set of bins constituting a distribution of SRDs along the branch, $br_ids_for_display[$n], in the SRD maps,
#	@{$brwise_distri_srds[#{branches}]} is the set of bins constituting a TOTAL distribution over all branches.
#
#	@bounds_srd is a set of boundaries used for distributing SRDs.
#
#
# OUTPUTS: 1 if the subroutine succeeded,
#	   0 if it failed.
#
sub brwise_distribute_srds_in_map (\@$\@\@) {

	my ($srd_map, $char_abs, $brwise_distri_srds, $bounds_srd) = @_;

	##my $ct_maps = @{$set_maps_alnshear};
	##my $wgt  = 1/$ct_maps;
	my $wgt  = 1; # Modified in creation of 'brwise_distribute_srds_in_map'.

	my $ct_brs = @{$srd_map};


		# Main Process: Distribute the SRDs in the map, along each branch. #

	for (my $n=0; $n <$ct_brs; $n++) { # Main Outer for-loop (over the branches). #

		my $linmap_srd_br = $srd_map->[$n];

		my $distri_srds_br = $brwise_distri_srds->[$n];

		my $ct_sites = @{$linmap_srd_br};

		for (my $s=0; $s <$ct_sites; $s++) { # Main inner for-loop (over the sites). #

		    my $srd = $linmap_srd_br->[$s];

		    if ($srd eq $char_abs) { next; }

		    cluster_distributor_wgt ($srd, @{$distri_srds_br}, @{$bounds_srd}, $wgt);
		    
		} # End of the main inner for-loop (over the sites). #

	} # End of the Main Outer for-loop (over the branches). #

	return 1;

}	# End of the subroutine, 'brwise_distribute_srds_in_map'. #


1;		# a "spell" which makes the module "use"-able and/or "require"-able #


