package MyTreeMap_indels_ML_sppl_hs;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2014 Kiyoshi Ezawa
#	Version 0.5.1: Copyright (C) 2014 Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.6.1: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.6.1.1: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.6.2: Copyright (C) 2019 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeMap_indels_ML_sppl_hs.pm:
#	This Perl package was created by modifying some subroutines in its predecessor, "MyTreeMap_indels_ML_sppl.pm",
#	in accompany with the creation of "MyTreeMap_indels_ML_hs_hs.pm" from its predecessor, "MyTreeMap_indels_ML_hs.pm".
#	The ONLY DIFFERENCE from the predecessor is that this package uses the output of the package "MyTreeMap_indels_spt_odr_hs.pm",
#	which means that the subroutines in this package follows the VIRTUAL ORDER among the gap-pattern blocks
#	specified by @vplaces_gpbs and @gpbs_at_spec_vplaces accompanying the main output of the major subroutines in "MyTreeMap_indels_spt_odr_hs.pm".
#	[CREATION STARTED on 2019/05/06, ENDED on 2019/05/dd.]
#
#	AS OF 2019/05/07,
#	the subroutines in this package are supplied almost exclusively to the package "MyTreeMap_indels_ML_hs_hs.pm".
#
# MyTreeMap_indels_ML_sppl.pm:
#	This Perl package contains some subroutines
#	that supplements the functions of the subroutines in the packages,
#	"MyTreeMap_indels_ML.pm", "MyTreeMap_indels_ML_hs.pm", and maybe
#	in "../MyPerlModules_ANEX/MyTreeMap_indels_ML_hs_finer.pm".
#
#	Especially, this package IMPORTED the subroutines
#	that were added to "MyTreeMap_indels_ML_hs_finer.pm"
#	in its modification from Dec 30, 2017 till Jan 04, 2018.
#	Here is the quotation of the note on the modification:
#
#	"[MODIFICATION STARTED on Dec 30 (Sat), 2017, ENDED on Jan 04 (Thr), 2018. ]
#
#	This subroutine now adds a correction to the contribution from each "indel history"
#	involving either the deletions of contiguous sub-sequences along the two child branches of a bifurcated root,
#	or an insertion along a parent branch (with only one child branch)
#	followed by a deletion of a sub-sequence immediately next to the inserted sub-sequence
#	along the child branch."
#
#	[CREATION STARTED on 2019/04/18, ENDED on 2019/04/dd.]
#

use strict;
use warnings;

use MyArrayManipu_num qw( diff_nmarray log_sum_exp );
	## union_nmarray

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

=set aside for later use (IMPORTED from "MyTreeMap_indels_ML_hs.pm") (I)

use MyArrayManipu qw( diff_array union_array nr_array);

use MyMalnIO qw( pr_ODEN_input );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);
use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm

=cut # END of "set aside for later use (IMPORTED from "MyTreeMap_indels_ML_hs.pm") (I)"

=Set aside for later use (IMPORTED from "MyTreeMap_indels_ML_hs_finer.pm") (I)
	# ADDED in creation of 'MyTreeMap_indels_ML_hs_finer.pm'. #
use MyTreeMap_indels_spt_odr_finer qw( );

use MyTreeMap_indels_ML_hs qw(
	search_paired_indels2
	search_leftmost_events2
	search_rightmost_events2
	cal_lnlk1_sgl_indel_cls2
	cal_lnlk1_pair_indel_cls2
	);
	# END of "ADDED in creation of 'MyTreeMap_indels_ML_hs_finer.pm'." #
=cut # END of "Set aside for later use (IMPORTED from "MyTreeMap_indels_ML_hs_finer.pm") (I)"


use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	cal_corr_fac_from_biroot_hs
	cal_corr_fac_from_uninode_hs
	cal_corr_fac_lnmfac1_indel_hst_hs
	);
	# search_biroot_uninodes # For the predecessor, "MyTreeMap_indels_ML_sppl.pm".

	# calc_lnmfac_gc_partit_nw_for_cl
	# mk_node2set_empty_blks2
	# cal_sumint_diff_xrate_in_idb
	# cal_lnmfac1_gpdseg_br_in_idb2
	# cal_lnmfac1_indel_hst_in_idb2
	# apprx1_tot_lnlk_gpattern_msa2_finer


sub if_all_absent_hs ($$\@\%\@\@$);
## sub search_biroot_uninodes ($\%\%); # For the predecessor, "MyTreeMap_indels_ML_sppl.pm".
sub find_delpairs_hs (\@\%\%\@\@$);
sub cal_corr_fac_from_biroot_hs (\@\@\%\%\%\@\@$$$\@\%);
sub find_insdelpairs_hs (\@\%\%\@\@$);
sub cal_corr_fac_from_uninode_hs (\@\@\%\%\%\@\@$$$\@\%);
sub cal_corr_fac_lnmfac1_indel_hst_hs (\@\@\@\%\%\%\%\@\@\%$$$\@\%$);


		# IMPORTED from "../MyPerlModules_ANEX/MyTreeMap_indels_ML_hs_finer.pm" (II) #

	# This specifies the level of tests on the main subroutines,
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)

##our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;


	# (Omitted the lines specifying some global variables in between.) #


	# For the portions of the subroutines involved in the calculation of correction factors
	# (e.g., 'cal_corr_fac_from_biroot', 'cal_corr_fac_from_uninode', 'cal_corr_fac_lnmfac1_indel_hst').
our $TEST_LEVEL_corr = 0;
#our $TEST_LEVEL_corr = 6;

		# END of "IMPORTED from "../MyPerlModules_ANEX/MyTreeMap_indels_ML_hs_finer.pm" (II)" #


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


=set aside for later use (IMPORTED from "MyTreeMap_indels_ML_hs.pm") (II)

	## These two indices specify where the information is stored
	## in %node2pars_stats = ($node_id => \@stat_info, ...) #
	#
our $INDX_STAT = 0; # The state (or the character) mapped onto the node.
our $INDX_SBTOT_PEN = 1; # The subtotal of penalties on the subtree under $node_id. 

	# This parameter is important to decide whether we should employ
	# the direct summation over indel lengths or its complement,
	# when calculating a 'composite' frequency.

our $MID_INDEL_LEN = 20; # An indel longer than this is regarded as 'long' => direct summation will be calculated.
## our $MID_INDEL_LEN = 10;

=cut # END of "set aside for later use (IMPORTED from "MyTreeMap_indels_ML_hs.pm") (II)"


=set aside for later use (IMPORTED from "../MyPerlModules_ANEX/MyTreeMap_indels_ML_hs_finer.pm") (III)

	## This variable sets an upper-bound for the gap-length.
	# IF a gap-block cluster contains a run of gaps longer than this variable,
	# the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses on the cluster.
	# The variable will be used in the new subroutine, 'scan_gbcls_for_long_gaps'.
	# (ADDED on Mar 10 (Mon), 2014.)
	#
	# NOTE: It should be synchronized with the synonymous global variable in the module 'MyTreeMap_indels.pm'.
	#
our $UPPER_BOUND_ANAL_GAPLEN = $MyTreeMap_indels_spt_odr_finer::UPPER_BOUND_ANAL_GAPLEN;
##our $UPPER_BOUND_ANAL_GAPLEN = 150;

	# This variable sets an upper-bound for the length of the gap-block cluster itself.
	# (ADDED on Mar 11 (Tue), 2014.):w
	#
	# NOTE: It should be synchronized with the synonymous global variable in the module 'MyTreeMap_indels.pm'.
	#
our $UPPER_BOUND_ANAL_LEN_GBCLS = $MyTreeMap_indels_spt_odr_finer::UPPER_BOUND_ANAL_LEN_GBCLS;
##our $UPPER_BOUND_ANAL_LEN_GBCLS = 200;

=cut # END of "set aside for later use (IMPORTED from "../MyPerlModules_ANEX/MyTreeMap_indels_ML_hs_finer.pm") (III)"


	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###

		## Subroutines IMPORTED FROM "../MyPerlModules_ANEX/MyTreeMap_indels_ML_hs_finer.pm" (on 2019/04/18). ##

	## Subroutines necessary for the corrections of multiplication factors
	## for the "insertion followed by deletion" type indel histories and their relatives.


# if_all_absent_hs:
#	This satellite subroutine of 'find_delpairs_hs' (& 'find_insdelpairs_hs') was created by modifying its predecessor, 'if_all_absent'.
#	The ONLY major DIFFERENCE from the precessor is that
#	this subroutine follows the "virtual" order of gap-pattern blocks specified by two arrays, @vplaces_gpbs and @gpbs_at_spec_vplaces,
#	whereas the predecessor follows the original order of the blocks, specified by their indices.
#	[CREATED on 2019/05/08.]
#
# if_all_absent:
#	This satellite subroutine of 'find_delpairs' (& 'find_insdelpairs')
#	examines whether a specified set of contiguous blocks 
#	are all occupied by the 'absence' state at all specified nodes.
#	[CREATED on Jan 1 (Mon), 2018]
#	[A FATAL BUG FIXED on 2019/04/18.]
#
# [ INPUT: ($indx_beg, $indx_end, \@set_nodes, \%node2bw_gstats, $char_abs), where ] ... for the predecessor, 'if_all_absent'.
# INPUT: ($vp_beg, $vp_end, \@set_nodes, \%node2bw_gstats, \@vplaces_gpbs, \@gpbs_at_spec_vplaces, $char_abs), where
#
#	[ $indx_beg/end is the index of the beginning/end of the specified set of contiguous blocks; ] ... for the predecessor, 'if_all_absent'.
#	$vp_beg/end is the virtual place of the beginning/end of the specified set of contiguous blocks;
#	@set_nodes lists the specified nodes;
#
#	%node2bw_gstats_in_idb = ($node_id => \@bw_gstats_in_idb, ...) is the map of gap states within the indel block onto the
#	nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id (, or $char_pres if the gap-pattern block (at the node) does not belong to the indel block);
#
#	[ADDED in creation of 'if_all_absent_hs'.]
#	@vplaces_gpbs and @gpbs_at_spec_vplaces specifiy the order among the gap-pattern blocks underlying this parsimonious history. Specifically,
#	$vplaces_gpbs[$gpb] specifies what place the $gpb th gap-pattern block is in,
#	$gpbs_at_spec_vplaces[$k] gives the index of the gap-pattern block in the $k th place.
#	[END of "ADDED in creation of 'if_all_absent_hs'."]
#
#	$char_abs is the character representing an "absence" state.
#
# OUTPUT: $if_all_absent = 1 if all specified blocks are in the 'absent' state at all specified nodes,
#			 = 0 otherwise.
#
##sub if_all_absent ($$\@\%$) {
#	my ($indx_beg, $indx_end, $set_nodes, $node2bw_gstats, $char_abs) = @_;
#
sub if_all_absent_hs ($$\@\%\@\@$) {

	my ($vp_beg, $vp_end, $set_nodes, $node2bw_gstats, $vplaces_gpbs, $gpbs_at_spec_vplaces, $char_abs) = @_; # MODIFIED in creation of 'if_all_absent_hs'. #

	$vp_end++; # Modified in creation of 'if_all_absent_hs'. #
	##$indx_end++;

	foreach my $node (@{$set_nodes}) { # Outer foreach-loop (over specified nodes). #

	    my $bw_gstats = $node2bw_gstats->{$node};

			# MAJOR MODIFICATION in creation of 'if_all_absent_hs'. #
	    for (my $vp = $vp_beg; $vp < $vp_end; $vp++) { # Inner for-loop (over specified interval of virtual places). #
		my $b = $gpbs_at_spec_vplaces->[$vp];

	    ##for (my $b = $indx_beg; $b < $indx_end; $b++) { # Inner for-loop (over specified blocks). #
		($bw_gstats->[$b] ne $char_abs) and (return 0); # A FATAL BUG fixed on 2019/04/18!!
		##($bw_gstats->[$b] eq $char_abs) and (return 0);
	    ##}			# End of the inner for-loop (over specified blocks). #
	    }			# End of the inner for-loop (over specified interval of virtual places). #

	}				# End of the outer foreach-loop (over specified nodes). #

	return 1;

}		# End of 'if_all_absent_hs'. #


=For the predecessor 'MyTreeMap_indels_ML_sppl.pm'.

# search_biroot_uninodes:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa2_finer'
#	searches a phylogenetic tree for a bifurcated root
#	and uni-furcated nodes (other than the root).
#	[CREATED in MODIFICATION of "MyTreeMap_indels_ML_hs_finer.pm" on Dec 31 (Sun), 2017.]
#
# INPUT: ($topnode_id, \%node2pa, \%node2ch), where
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2pa = ($node_id => $parent_id, ...) for the non-root nodes;
#	%node2ch = ($node_id => \@child_node_ids, ...) for the internal nodes;
#
# OUTPUT: (\@info_biroot, \@list_uninodes), where
#
#	@info_biroot = ($id_root, $id_child1, $id_child2) for the bifurcated root,
#		      = () if the root is NOT bifurcated;
#
#	@{$list_uninodes[$u]} = ($id_parent, $id_uninode, $id_child) for the $u th unifurcated node (other than the root),
#	$id_uninode, whose parent and the sole child are $id_parent and $id_child, respectively.
#
sub search_biroot_uninodes ($\%\%) {

	my ($topnode_id, $node2pa, $node2ch) = @_;

	my @info_biroot = ();
	my $ch_root = $node2ch->{$topnode_id};
	if (@{$ch_root} == 2) {
	    @info_biroot = ($topnode_id, @{$ch_root});
	}

	my @list_uninodes = ();

	foreach my $node (sort {$a <=> $b} keys %{$node2ch}) { # Main foreach-loop (over internal nodes) #

		# A BUG fixed on Jan 4 (Thu), 2018. #
	    ($node == $topnode_id) and next;
	    ##($node == $topnode_id) or next;

	    my $ch = $node2ch->{$node};
	    (defined $ch) or next;
	    (@{$ch} == 1) or next;
	    my $id_child = $ch->[0];
	    my $id_parent = $node2pa->{$node};

	    push @list_uninodes, [$id_parent, $node, $id_child];

	}	# End of the main foreach-loop (over internal nodes). #

	return (\@info_biroot, \@list_uninodes);

}	# End of 'search_biroot_uninodes'. #

=cut # END of "For the predecessor 'MyTreeMap_indels_ML_sppl.pm'."


# find_delpairs_hs:
#	This satellite subroutine of 'cal_corr_fac_from_biroot_hs' was created by modifying its predecessor, 'find_delpairs'.
#	The ONLY major DIFFERENCE from the precessor is that
#	this subroutine follows the "virtual" order of gap-pattern blocks specified by two arrays, @vplaces_gpbs and @gpbs_at_spec_vplaces,
#	whereas the predecessor follows the original order of the blocks, specified by their indices.
#	[CREATED on 2019/05/08.]
#
# find_delpairs:
#	This satellite subroutine of 'cal_corr_fac_from_biroot'
#	searches for the pairs of deletions (of contiguous sub-sequences) separated by a bifurcated root.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm' on Jan 1 (Mon), 2018]
#
# [ INPUT: (\@info_biroot, \%br2indels, \%node2blwise_gstats, $char_abs), where ] ... for the predecessor, 'find_delpairs'.
# INPUT: (\@info_biroot, \%br2indels, \%node2blwise_gstats, \@vplaces_gpbs, \@gpbs_at_spec_vplaces, $char_abs), where
#
#	@{$info_biroot} = ($id_root, $id_child1, $id_child2) for the bifurcated root,
#		      = () if the root is NOT bifurcated;
#
#	Detailes on other input variables are described in the header of 'cal_corr_fac_lnmfac1_indel_hst'.
#
# OUTPUT: \@delpairs, where
#
#	@{$delpairs[$dp]} = ($info_del_ch1, $info_del_ch2] is for the $dp th deletion pair, with
#	$info_del_ch1/2 = [$stat_pa, $stat_node, \@indices_blk] is the information for the deletion along the child branch 1/2.
#	The subroutine returns undef if it fails.
#
##sub find_delpairs (\@\%\%$) {
#	my ($info_biroot, $br2indels, $node2bw_gstats, $char_abs) = @_;
#
sub find_delpairs_hs (\@\%\%\@\@$) {

	my ($info_biroot, $br2indels, $node2bw_gstats, $vplaces_gpbs, $gpbs_at_spec_vplaces, $char_abs) = @_; # MODIFIED in creation of 'find_delpairs_hs'. #

	my ($id_root, $id_ch1, $id_ch2) = @{$info_biroot};
	my $indels_ch1 = $br2indels->{$id_ch1};
	my $indels_ch2 = $br2indels->{$id_ch2};

	unless ((defined $indels_ch1) and (@{$indels_ch1}>0) and (defined $indels_ch2) and (@{$indels_ch2}>0)) {
	    return []; # A pair of deletions can NEVER be formed.
	}

#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})

	my @dels_ch1 = my @dels_ch2 = ();

	my $ct_indels_ch1 = @{$indels_ch1};
	my $ct_indels_ch2 = @{$indels_ch2};

	for (my $e=0; $e <$ct_indels_ch1; $e++) { # 1st preliminary for-loop. #
		my $indel = $indels_ch1->[$e];
		((@{$indel}>0) and ($indel->[0] ne $char_abs)) and (push @dels_ch1, $indel); 
	}				# End of the 1st preliminary for-loop. #

	unless (@dels_ch1>0) {
	    return []; # A pair of deletions can NEVER be formed.
	}

	for (my $e=0; $e <$ct_indels_ch2; $e++) { # 2nd preliminary for-loop. #
		my $indel = $indels_ch2->[$e];
		((@{$indel}>0) and ($indel->[0] ne $char_abs)) and (push @dels_ch2, $indel); 
	}				# End of the 2nd preliminary for-loop. #

	unless (@dels_ch2>0) {
	    return []; # A pair of deletions can NEVER be formed.
	}

		# MODIFIED in creation of 'find_delpairs_hs'. (1) #

	my @prestd_dels_ch1 = my @prestd_dels_ch2 = ();

	foreach my $del (@dels_ch1) { # 3rd preliminary foreach-loop (over deletions along $id_ch1).
	    my ($stat1, $stat2, $invlvd_gpbs) = @{$del};
	    my @std_invlvd_gpbs = (@{$invlvd_gpbs} <2) ? @{$invlvd_gpbs} :
				sort {$vplaces_gpbs->[$a] <=> $vplaces_gpbs->[$b]} @{$invlvd_gpbs} ;
	    push @prestd_dels_ch1, [$stat1, $stat2, \@std_invlvd_gpbs];
	} # End of the 3rd preliminary foreach-loop (over deletions along $id_ch1).

	foreach my $del (@dels_ch2) { # 5th preliminary foreach-loop (over deletions along $id_ch2).
	    my ($stat1, $stat2, $invlvd_gpbs) = @{$del};
	    my @std_invlvd_gpbs = (@{$invlvd_gpbs} <2) ? @{$invlvd_gpbs} :
				sort {$vplaces_gpbs->[$a] <=> $vplaces_gpbs->[$b]} @{$invlvd_gpbs} ;
	    push @prestd_dels_ch2, [$stat1, $stat2, \@std_invlvd_gpbs];
	} # End of the 5th preliminary foreach-loop (over deletions along $id_ch2).

	my @std_dels_ch1 = (@prestd_dels_ch1<2) ? @prestd_dels_ch1 :
			sort {$vplaces_gpbs->[$a->[2][0]] <=> $vplaces_gpbs->[$b->[2][0]]} @prestd_dels_ch1;
	my @std_dels_ch2 = (@prestd_dels_ch2<2) ? @prestd_dels_ch2 :
			sort {$vplaces_gpbs->[$a->[2][0]] <=> $vplaces_gpbs->[$b->[2][0]]} @prestd_dels_ch2;
	##my @std_dels_ch1 = sort {$a->[2][0] <=> $b->[2][0]} @dels_ch1;
	##my @std_dels_ch2 = sort {$a->[2][0] <=> $b->[2][0]} @dels_ch2;

		# END of "MODIFIED in creation of 'find_delpairs_hs'. (1)" #

	my @delpairs = ();

	my $del1 = shift @std_dels_ch1;
	my $del2 = shift @std_dels_ch2;

	while ((defined $del1) and (defined $del2)) { # Main outer while-loop (as long as we have pairs to examine). #
	    my $indices_blk1 = $del1->[2];
	    my $indices_blk2 = $del2->[2];
	    my ($first_blk1, $last_blk1) = @{$indices_blk1}[0, $#{$indices_blk1}];
	    my ($first_blk2, $last_blk2) = @{$indices_blk2}[0, $#{$indices_blk2}];

			# ADDED in creation of 'find_delpairs_hs'. (1) #
	    my $vp_first_blk1 = $vplaces_gpbs->[$first_blk1];
	    my $vp_last_blk1 = $vplaces_gpbs->[$last_blk1];
	    my $vp_first_blk2 = $vplaces_gpbs->[$first_blk2];
	    my $vp_last_blk2 = $vplaces_gpbs->[$last_blk2];
			# END of "ADDED in creation of 'find_delpairs_hs'. (1)" #

	    if ($vp_last_blk1 < $vp_last_blk2) { # Modified in creation of 'find_delpairs_hs'. (2a)
	    ##if ($last_blk1 < $last_blk2) {

		if ($vp_last_blk1 + 1 >= $vp_first_blk2) { # Modified in creation of 'find_delpairs_hs'. (3a)
		##if ($last_blk1 + 1 == $first_blk2) {
		    push @delpairs, [$del1, $del2];

		} elsif ($vp_last_blk1 < $vp_first_blk2) { # Modified in creation of 'find_delpairs_hs'. (3b)
		##} elsif ($last_blk1 < $first_blk2) {

			# MODIFIED in creation of 'find_delpairs_hs'. (5) #
		    my $if_all_absent = if_all_absent_hs ($vp_last_blk1 + 1, $vp_first_blk2-1, @{$info_biroot}, %{$node2bw_gstats}, @{$vplaces_gpbs}, @{$gpbs_at_spec_vplaces}, $char_abs);
		    ##my $if_all_absent = if_all_absent ($last_blk1 + 1, $first_blk2-1, @{$info_biroot}, %{$node2bw_gstats}, $char_abs);
		    $if_all_absent and (push @delpairs, [$del1, $del2]);
		}

		$del1 = shift @std_dels_ch1; # Update the first deletion. #

	    } elsif ($vp_last_blk2 < $vp_last_blk1) { # Modified in creation of 'find_delpairs_hs' (2b)
	    ##} elsif ($last_blk2 < $last_blk1) {

		if ($vp_last_blk2 + 1 >= $vp_first_blk1) { # Modified in creation of 'find_delpairs_hs'. (6a)
		##if ($last_blk2 + 1 == $first_blk1) {
		    push @delpairs, [$del1, $del2];

		} elsif ($vp_last_blk2 < $vp_first_blk1) { # Modified in creation of 'find_delpairs_hs'. (6b)
		##} elsif ($last_blk2 < $first_blk1) {


			# MODIFIED in creation of 'find_delpairs_hs'. (7) #
		    my $if_all_absent = if_all_absent_hs ($vp_last_blk2 + 1, $vp_first_blk1-1, @{$info_biroot}, %{$node2bw_gstats}, @{$vplaces_gpbs}, @{$vplaces_gpbs}, $char_abs);
		    ##my $if_all_absent = if_all_absent ($last_blk2 + 1, $first_blk1-1, @{$info_biroot}, %{$node2bw_gstats}, $char_abs);
		    $if_all_absent and (push @delpairs, [$del1, $del2]);
		}

		$del2 = shift @std_dels_ch2; # Update the second deletion. #

	    } else {
		# This case is extremely inconceivable. If it occurs, however, update both deletions.
		$del1 = shift @std_dels_ch1;
		$del2 = shift @std_dels_ch2;
	    }

	}				# End of the main outer while-loop (as long as we have pairs to examine). #

	return \@delpairs;

}		# End of 'find_delpairs_hs'. #


# cal_corr_fac_from_biroot_hs:
#	This satellite subroutine of 'cal_corr_fac_lnmfac1_indel_hst_hs' was created by modifying the predecessor, 'cal_corr_fac_from_biroot'.
#	The ONLY major DIFFERENCE from the precessor is that
#	this subroutine follows the "virtual" order of gap-pattern blocks specified by two arrays, @vplaces_gpbs and @gpbs_at_spec_vplaces,
#	whereas the predecessor follows the original order of the blocks, specified by their indices.
#	Basically, the DIFFERENCE was implemented by replacing the subroutine, 'find_delpairs', with 'find_delpairs_hs'.
#	[CREATED on 2019/05/07.]
#
#	[MAJOR REVISION on 2019/05/09.]
#	Because "MyTreeMap_indels_spt_odr_hs.pm" regards the relevent indel history with different orders as TWO SEPARATE histories,
#	the factor of log (2) is not necessary any longer.
#	Thus, the factor WAS REMOVED in thie revision.
#
# cal_corr_fac_from_biroot:
#	This satellite subroutine of 'cal_corr_fac_lnmfac1_indel_hst' computes
#	the correction factor contributed from a pair of deletions separated by a bifurcated root.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm', from Dec 31 (Sun), 2017 till Jan 2 (Tue), 2018.]
#
#	[NOTE] This initial version of this subroutine deals with the deletion pair at either end of the MSA
#	in the same way as the pair in the bulk.
#
#
# [ INPUT: (\@info_biroot, \@cls_gpb, \%br2atrr, \%br2indels, \%node2blwise_gstats, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs), where ] ... for the predecessor 'cal_corr_fac_from_biroot'.
# INPUT: (\@info_biroot, \@cls_gpb, \%br2atrr, \%br2indels, \%node2blwise_gstats, \@vplaces_gpbs, \@gpbs_at_spec_vplaces, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs), where
#
#	@{$info_biroot} = ($id_root, $id_child1, $id_child2) for the bifurcated root,
#		      = () if the root is NOT bifurcated;
#
#	Detailes on other input variables are described in the header of 'cal_corr_fac_lnmfac1_indel_hst'.
#
# OUTPUT: $corr_fac1,
#	which is the logarithm of the multiplicative correlation factor for the contribution 
#	of the pair of deletions separated by the bifurcated root.
#	It is undefined if the subroutine failed.
#
##sub cal_corr_fac_from_biroot (\@\@\%\%\%$$$\@\%) {
#	my ($info_biroot, $cls_gpblks, $br2attr, $br2indels, $node2bw_gstats, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs) = @_;
#
sub cal_corr_fac_from_biroot_hs (\@\@\%\%\%\@\@$$$\@\%) {

	my ($info_biroot, $cls_gpblks, $br2attr, $br2indels, $node2bw_gstats, $vplaces_gpbs, $gpbs_at_spec_vplaces, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs) = @_;

	my $ct_gpblks = @{$cls_gpblks};

	my ($id_root, $id_ch1, $id_ch2) = @{$info_biroot};

	my $delpairs = find_delpairs_hs (@{$info_biroot}, %{$br2indels}, %{$node2bw_gstats}, @{$vplaces_gpbs}, @{$gpbs_at_spec_vplaces}, $char_abs);
	##my $delpairs = find_delpairs (@{$info_biroot}, %{$br2indels}, %{$node2bw_gstats}, $char_abs);
	unless (defined $delpairs) {
	    print STDOUT "[ERROR 'cal_corr_fac_from_biroot_hs'] The subroutine 'find_delpairs_hs' FAILED: $! !\n\n";
	    return undef;
	}
	unless (@{$delpairs}>0) {
	    return 0; # A pair of deletions can NEVER be formed, thus no correction is necessary.
	}
	my $ct_delpairs = @{$delpairs};

   (print STDOUT ' ' x 5, "< 'cal_corr_fac_from_biroot_hs'   (ID{root} = $id_root; ID{child node 1} = $id_ch1; ID{child node 2} = $id_ch2; #{delpairs} = $ct_delpairs) >\n\n") if ($TEST_LEVEL_corr>2);

	my ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end) = @{$params_bds};

	my $len_br1 = $br2attr->{$id_ch1}[$INDX_LENGTH];
	my $len_br2 = $br2attr->{$id_ch2}[$INDX_LENGTH];

	my $exponent_per_site = ($totrate_ins + $totrate_del) * ($len_br1 + $len_br2);

   if ($TEST_LEVEL_corr>3) {
	print STDOUT ' ' x 5, "[ Main inputs and secondary input ]\n\n";
	print STDOUT ' ' x 5, '$cutoff_del = ', $cutoff_del, ",\n";
	print STDOUT ' ' x 5, '$totrate_ins = ', $totrate_ins, ",\n";
	print STDOUT ' ' x 5, '$totrate_del = ', $totrate_del, ",\n";
	print STDOUT ' ' x 5, '$len_br1 = ', $len_br1, ",\n\n";
	print STDOUT ' ' x 5, '$len_br2 = ', $len_br2, ",\n\n";
	print STDOUT ' ' x 5, '$exponent_per_site (= ($totrate_ins + $totrate_del) * ($len_br1 + $len_br2)) = ', $exponent_per_site, "\n\n";
   } # END of "if ($TEST_LEVEL_corr>3) {...}"

	my $log_mfacs_del1 = $br2log_mfacs->{$id_ch1}[1];
	my $log_mfacs_del2 = $br2log_mfacs->{$id_ch2}[1];

#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,

	my $corr_fac1 = 0;

   (print STDOUT ' ' x5, "[ Calculating correction factors from $ct_delpairs deletion pairs ]\n\n") if ($TEST_LEVEL_corr>3);

	foreach my $delpair (@{$delpairs}) { # Main outer foreach-loop (over deletion pairs). #

	    my ($info_del1, $info_del2) = @{$delpair};
	    my $blks_del1 = $info_del1->[2];
	    my $blks_del2 = $info_del2->[2];

	    my $len_del1 = my $len_del2 = 0;

	    foreach my $blk (@{$blks_del1}) { $len_del1 += $cls_gpblks->[$blk][1]; }
	    foreach my $blk (@{$blks_del2}) { $len_del2 += $cls_gpblks->[$blk][1]; }

#	@{$cls_gpblocks[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is
#	for the $gpb th gap-pattern-block in the input cluster, with
#	$len_gpb being the length (in terms of #{columns}) of the block,

	    my @elements = ();

	    my $upper_bound = ($len_del1 > $len_del2) ? $cutoff_del - $len_del1 : $cutoff_del - $len_del2;
	    $upper_bound++;

   if ($TEST_LEVEL_corr>= 5) {
	print STDOUT ' ' x 5, "{deleted blocks 1} = ", join (', ', @{$blks_del1}), " (total length: $len_del1);\n";
	print STDOUT ' ' x 5, "{deleted blocks 2} = ", join (', ', @{$blks_del2}), " (total length: $len_del2);\n";
	print STDOUT ' ' x 5, "Summation over #{'invisible' residues} in [0, $upper_bound) .\n\n";
   }

	    ($upper_bound>0) or next; # Ignore the pair. #

	    for (my $x = 0; $x <$upper_bound; $x++) {
		my $element = (-$x * $exponent_per_site) + $log_mfacs_del1->[$len_del1 + $x] + $log_mfacs_del2->[$len_del2 + $x];
		push @elements, $element;

    (print STDOUT ' ' x 7, '$x = ', $x, ",\t", '$element = ', $element, ",\n") if ($TEST_LEVEL_corr>5); 
	    }

	    my $numer = (@elements==1) ? $elements[0] : log_sum_exp (@elements);
	    unless (defined $numer) {
		print STDOUT "[ERROR 'cal_corr_fac_from_biroot_hs'] The subroutine 'log_sum_exp' FAILED: $! !\n\n";
		return undef;
	    }

			# MAJOR REVISION on 2019/05/09.
	    my $corr_fac11 = $numer - ($log_mfacs_del1->[$len_del1] + $log_mfacs_del2->[$len_del2]) ;
	    ##my $corr_fac11 = ($numer + log (2)) - ($log_mfacs_del1->[$len_del1] + $log_mfacs_del2->[$len_del2]) ;

	    $corr_fac1 +=  $corr_fac11;

    if ($TEST_LEVEL_corr>= 5) {
	print STDOUT ' ' x 5, 'Log(total contribution from the delpairs with 0-', $upper_bound, ' invisible residues) (= $numer) = ', $numer, ",\n";
	print STDOUT ' ' x 5, 'Log(multiple factor from the original deletion 1) (= $log_mfacs_del1->[$len_del1]) = ', $log_mfacs_del1->[$len_del1], ",\n";
	print STDOUT ' ' x 5, 'Log(multiple factor from the original deletion 2) (= $log_mfacs_del2->[$len_del2]) = ', $log_mfacs_del2->[$len_del2], ",\n\n";
    } # END of "if ($TEST_LEVEL_corr>= 5) {...}".
    if ($TEST_LEVEL_corr > 3) {
		# Major revision on 2019/05/09.
	print STDOUT ' ' x 5, '$corr_fac11 (= $numer - ($log_mfacs_del1->[$len_del1] + $log_mfacs_del2->[$len_del2])) = ', $corr_fac11, ",\n";
	##print STDOUT ' ' x 5, '$corr_fac11 (= ($numer + log (2)) - ($log_mfacs_del1->[$len_del1] + $log_mfacs_del2->[$len_del2])) = ', $corr_fac11, ",\n";
	print STDOUT ' ' x 5, 'Current $corr_fac1 = ', $corr_fac1, ".\n\n";
    } # END of "if ($TEST_LEVEL_corr > 3) {...}".

	}			# End of the main outer foreach-loop (over deletion pairs). #

			### RESTART FROM HERE (2a) ###

   (print STDOUT ' ' x 5, "< 'cal_corr_fac_from_biroot_hs'   ended  (with", '$corr_fac1 = ', "$corr_fac1) />\n\n") if ($TEST_LEVEL_corr>2);

	return $corr_fac1;

}	# End of 'cal_corr_fac_from_biroot_hs'. #


# find_insdelpairs_hs:
#	This satellite subroutine of 'cal_corr_fac_from_uninode_hs' was created by modifying its predecessor, 'find_insdelpairs'.
#	The ONLY major DIFFERENCE from the precessor is that
#	this subroutine follows the "virtual" order of gap-pattern blocks specified by two arrays, @vplaces_gpbs and @gpbs_at_spec_vplaces,
#	whereas the predecessor follows the original order of the blocks, specified by their indices.
#	[CREATED on 2019/05/08.]
#
# find_insdelpairs:
#	This satellite subroutine of 'cal_corr_fac_from_uninode'
#	searches for pairs of an insertion and a subsequent deletion (of contiguous sub-sequences) separated by a unifurcated node.
#	The subroutine was created by modifying the sister subroutine 'find_delpairs'.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm' on Jan 2 (Tue), 2018]
#
# find_delpairs:
#	This satellite subroutine of 'cal_corr_fac_from_biroot'
#	searches for the pairs of deletions (of contiguous sub-sequences) separated by a bifurcated root.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm' on Jan 1 (Mon), 2018]
#
# [ INPUT: (\@info_uninode, \%br2indels, \%node2blwise_gstats, $char_abs), where ] ... for the predecessor, 'find_insdelpairs'.
# INPUT: (\@info_uninode, \%br2indels, \%node2blwise_gstats, \@vplaces_gpbs, \@gpbs_at_spec_vplaces, $char_abs), where
#
#	@{$info_uninode} = ($id_parent, $id_uninode, $id_child) for a unifurcated node (other than the root),
#	$id_uninode, whose parent and the sole child are $id_parent and $id_child, respectively;
#
#	Detailes on other input variables are described in the header of 'cal_corr_fac_lnmfac1_indel_hst'.
#
# OUTPUT: \@insdelpairs, where
#
#	@{$insdelpairs[$idp]} = ($info_ins, $info_del_ch] is for the $idp th insertion-deletion pair, where
#	$info_ins = [$stat_pa, $stat_node, \@indices_blk] is the information for the insertion along the branch immediately above the unifurcated node, and
#	$info_del_ch = [$stat_pa, $stat_node, \@indices_blk] is the information for the deletion along the branch immediately below the unifurcated node.
#	The subroutine returns undef if it fails.
#
##sub find_insdelpairs (\@\%\%$) {
#	my ($info_uninode, $br2indels, $node2bw_gstats, $char_abs) = @_;
#
sub find_insdelpairs_hs (\@\%\%\@\@$) {

	my ($info_uninode, $br2indels, $node2bw_gstats, $vplaces_gpbs, $gpbs_at_spec_vplaces, $char_abs) = @_; # MODIFIED in creation of 'find_insdelpairs_hs'. #

	my ($id_pa, $id_uninode, $id_ch) = @{$info_uninode};
	my $indels_sbj = $br2indels->{$id_uninode};
	my $indels_ch = $br2indels->{$id_ch};

	unless ((defined $indels_sbj) and (@{$indels_sbj}>0) and (defined $indels_ch) and (@{$indels_ch}>0)) {
	    return []; # A pair of an insertion and a subsequent deletion can NEVER be formed.
	}

#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})

	my @ins_sbj = my @dels_ch = ();

	my $ct_indels_sbj = @{$indels_sbj};
	my $ct_indels_ch = @{$indels_ch};

	for (my $e=0; $e <$ct_indels_sbj; $e++) { # 1st preliminary for-loop. #
		my $indel = $indels_sbj->[$e];
		((@{$indel}>0) and ($indel->[0] eq $char_abs)) and (push @ins_sbj, $indel); 
	}				# End of the 1st preliminary for-loop. #

	unless (@ins_sbj>0) {
	    return []; # A pair of an insertion and a subsequent deletion can NEVER be formed.
	}

	for (my $e=0; $e <$ct_indels_ch; $e++) { # 2nd preliminary for-loop. #
		my $indel = $indels_ch->[$e];
		((@{$indel}>0) and ($indel->[0] ne $char_abs)) and (push @dels_ch, $indel); 
	}				# End of the 2nd preliminary for-loop. #

	unless (@dels_ch>0) {
	    return []; # A pair of an insertion and a subsequent deletion can NEVER be formed.
	}

		# MODIFIED in creation of 'find_insdelpairs_hs'. (1) #

	my @prestd_ins_sbj = my @prestd_dels_ch = ();

	foreach my $ins (@ins_sbj) { # 3rd preliminary foreach-loop (over insertions along $id_uninode).
	    my ($stat1, $stat2, $invlvd_gpbs) = @{$ins};
	    my @std_invlvd_gpbs = (@{$invlvd_gpbs}<2) ? @{$invlvd_gpbs} :
			sort {$vplaces_gpbs->[$a] <=> $vplaces_gpbs->[$b]} @{$invlvd_gpbs};
	    push @prestd_ins_sbj, [$stat1, $stat2, \@std_invlvd_gpbs];
	} # End of the 3rd preliminary foreach-loop (over insertions along $id_uninode).

	foreach my $del (@dels_ch) { # 5th preliminary foreach-loop (over deletions along $id_ch).
	    my ($stat1, $stat2, $invlvd_gpbs) = @{$del};
	    my @std_invlvd_gpbs = (@{$invlvd_gpbs}<2) ? @{$invlvd_gpbs} :
			sort {$vplaces_gpbs->[$a] <=> $vplaces_gpbs->[$b]} @{$invlvd_gpbs};
	    push @prestd_dels_ch, [$stat1, $stat2, \@std_invlvd_gpbs];
	} # End of the 5th preliminary foreach-loop (over deletions along $id_ch).

	my @std_ins_sbj = (@prestd_ins_sbj<2) ? @prestd_ins_sbj :
		sort {$vplaces_gpbs->[$a->[2][0]] <=> $vplaces_gpbs->[$b->[2][0]]} @prestd_ins_sbj;
	my @std_dels_ch = (@prestd_dels_ch<2) ? @prestd_dels_ch :
		sort {$vplaces_gpbs->[$a->[2][0]] <=> $vplaces_gpbs->[$b->[2][0]]} @prestd_dels_ch;
	##my @std_ins_sbj = sort {$a->[2][0] <=> $b->[2][0]} @ins_sbj;
	##my @std_dels_ch = sort {$a->[2][0] <=> $b->[2][0]} @dels_ch;

		# END of "MODIFIED in creation of 'find_insdelpairs_hs'. (1)" #

	my @insdelpairs = ();

	my $ins = shift @std_ins_sbj;
	my $del = shift @std_dels_ch;

	while ((defined $ins) and (defined $del)) { # Main outer while-loop (as long as we have pairs to examine). #

	    my $indices_blk1 = $ins->[2];
	    my $indices_blk2 = $del->[2];
	    my ($first_blk1, $last_blk1) = @{$indices_blk1}[0, $#{$indices_blk1}];
	    my ($first_blk2, $last_blk2) = @{$indices_blk2}[0, $#{$indices_blk2}];

		# ADDED in creation of 'find_insdelpairs_hs'. (1) #
	    my $vp_first_blk1 = $vplaces_gpbs->[$first_blk1];
	    my $vp_last_blk1 = $vplaces_gpbs->[$last_blk1];
	    my $vp_first_blk2 = $vplaces_gpbs->[$first_blk2];
	    my $vp_last_blk2 = $vplaces_gpbs->[$last_blk2];
		# END of "ADDED in creation of 'find_insdelpairs_hs'. (1)" #

	    if ($vp_last_blk1 < $vp_last_blk2) { # Modified in creation of 'find_insdelpairs_hs'. (2a)
	    ##if ($last_blk1 < $last_blk2) {

		if ($vp_last_blk1 + 1 >= $vp_first_blk2) { # Modified in creation of 'find_insdelpairs_hs'. (3a)
		##if ($last_blk1 + 1 == $first_blk2) {
		    push @insdelpairs, [$ins, $del];

		} elsif ($vp_last_blk1 < $vp_first_blk2) { # Modified in creation of 'find_insdelpairs_hs'. (3b)
		##} elsif ($last_blk1 < $first_blk2) {

			# MODIFIED in creation of 'find_insdelpairs_hs'. (5) #
		    my $if_all_absent = if_all_absent_hs ($vp_last_blk1 + 1, $vp_first_blk2-1, @{$info_uninode}, %{$node2bw_gstats}, @{$vplaces_gpbs}, @{$gpbs_at_spec_vplaces}, $char_abs);
		    ##my $if_all_absent = if_all_absent ($last_blk1 + 1, $first_blk2-1, @{$info_uninode}, %{$node2bw_gstats}, $char_abs);
		    $if_all_absent and (push @insdelpairs, [$ins, $del]);
		}

		$ins = shift @std_ins_sbj; # Update the insertion. #

	    } elsif ($vp_last_blk2 < $vp_last_blk1) { # Modified in creation of 'find_insdelpairs_hs'. (2b)
	    ##} elsif ($last_blk2 < $last_blk1) {

		if ($vp_last_blk2 + 1 >= $vp_first_blk1) { # Modified in creation of 'find_insdelpairs_hs'. (6a)
		##if ($last_blk2 + 1 == $first_blk1) {
		    push @insdelpairs, [$ins, $del];

		} elsif ($vp_last_blk2 < $vp_first_blk1) { # Modified in creation of 'find_insdelpairs_hs'. (6b)
		##} elsif ($last_blk2 < $first_blk1) {

			# MODIFIED in creation of 'find_insdelpairs_hs'. (7) #
		    my $if_all_absent = if_all_absent_hs ($vp_last_blk2 + 1, $vp_first_blk1-1, @{$info_uninode}, %{$node2bw_gstats}, @{$vplaces_gpbs}, @{$gpbs_at_spec_vplaces}, $char_abs);
		    ##my $if_all_absent = if_all_absent ($last_blk2 + 1, $first_blk1-1, @{$info_uninode}, %{$node2bw_gstats}, $char_abs);
		    $if_all_absent and (push @insdelpairs, [$ins, $del]);
		}

		$del = shift @std_dels_ch; # Update the deletion. #

	    } else {
		# This case is extremely inconceivable. If it occurs, however, update both deletions.
		$ins = shift @std_ins_sbj;
		$del = shift @std_dels_ch;
	    }

	}				# End of the main outer while-loop (as long as we have pairs to examine). #

	return \@insdelpairs;

}		# End of 'find_insdelpairs_hs'. #


# cal_corr_fac_from_uninode_hs:
#	This satellite subroutine of 'cal_corr_fac_lnmfac1_indel_hst_hs' was
#	created by modifying its predecessor, 'cal_corr_fac_from_uninode'.
#	The ONLY major DIFFERENCE from the precessor is that
#	this subroutine follows the "virtual" order of gap-pattern blocks specified by two arrays, @vplaces_gpbs and @gpbs_at_spec_vplaces,
#	whereas the predecessor follows the original order of the blocks, specified by their indices.
#	Basically, the DIFFERENCE was implemented by replacing the subroutine, 'find_insdelpairs', with 'find_insdelpairs_hs'.
#	[CREATED on 2019/05/07.]
#	
#	[MAJOR REVISION on 2019/05/09.]
#	Because "MyTreeMap_indels_spt_odr_hs.pm" regards the relevent indel history with different orders as TWO SEPARATE histories,
#	the factor of log (2) is not necessary any longer.
#	Thus, the factor WAS REMOVED in this revision.
#
# cal_corr_fac_from_uninode:
#	This satellite subroutine of 'cal_corr_fac_lnmfac1_indel_hst' computes
#	the correction factor contributed from a pair of an insertion and a subsequent deletion
#	separated by a unifurcated node.
#	This subroutine was created by modifying the sister subroutine, 'cal_corr_fac_from_biroot'.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm', on Jan 2 (Tue), 2018.]
#
#	[NOTE] This initial version of this subroutine deals with the deletion pair at either end of the MSA
#	in the same way as the pair in the bulk.
#
# cal_corr_fac_from_biroot:
#	This satellite subroutine of 'cal_corr_fac_lnmfac1_indel_hst' computes
#	the correction factor contributed from a pair of deletions separated by a bifurcated root.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm', from Dec 31 (Sun), 2017 till Jan 2 (Tue), 2018.]
#
# [ INPUT: (\@info_uninode, \@cls_gpb, \%br2atrr, \%node2pa, \%br2indels, \%node2blwise_gstats, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs), where ] ... for the predecessor 'cal_corr_fac_from_uninode'.
# INPUT: (\@info_uninode, \@cls_gpb, \%br2atrr, \%node2pa, \%br2indels, \%node2blwise_gstats, \@vplaces_gpbs, \@gpbs_at_spec_vplaces, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs), where
#
#	@{$info_uninode} = ($id_parent, $id_uninode, $id_child) for a unifurcated node (other than the root),
#	$id_uninode, whose parent and the sole child are $id_parent and $id_child, respectively;
#
#	Detailes on other input variables are described in the header of 'cal_corr_fac_lnmfac1_indel_hst_hs'.
#
# OUTPUT: $corr_fac2,
#	which is the logarithm of the multiplicative correlation factor for the contribution 
#	of the pairs of an insertion and a subsequent deletion separated by the unifurcated node.
#	It is undefined if the subroutine failed.
#
##sub cal_corr_fac_from_uninode (\@\@\%\%\%$$$\@\%) {
#	my ($info_uninode, $cls_gpblks, $br2attr, $br2indels, $node2bw_gstats, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs) = @_;
#
sub cal_corr_fac_from_uninode_hs (\@\@\%\%\%\@\@$$$\@\%) {

	my ($info_uninode, $cls_gpblks, $br2attr, $br2indels, $node2bw_gstats, $vplaces_gpbs, $gpbs_at_spec_vplaces, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs) = @_;

	my $ct_gpblks = @{$cls_gpblks};

	my ($id_pa, $id_uninode, $id_ch) = @{$info_uninode};

	my $insdelpairs = find_insdelpairs_hs (@{$info_uninode}, %{$br2indels}, %{$node2bw_gstats},@{$vplaces_gpbs}, @{$gpbs_at_spec_vplaces}, $char_abs);
	##my $insdelpairs = find_insdelpairs (@{$info_uninode}, %{$br2indels}, %{$node2bw_gstats}, $char_abs);
	unless (defined $insdelpairs) {
	    print STDOUT "[ERROR 'cal_corr_fac_from_uninode_hs'] The subroutine 'find_insdelpairs_hs' FAILED: $! !\n\n";
	    return undef;
	}
	unless (@{$insdelpairs}>0) {
	    return 0; # A pair of an insertion & a subsequent deletion can NEVER be formed, thus no correction is necessary.
	}
	my $ct_insdelpairs = @{$insdelpairs};

   (print STDOUT ' ' x 5, "< 'cal_corr_fac_from_uninode_hs'   (ID{uni-furcated node} = $id_uninode; ID{child node} = $id_ch; #{insdelpairs} = $ct_insdelpairs) >\n\n") if ($TEST_LEVEL_corr>2);

	my ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end) = @{$params_bds};

	my $len_br_ch = $br2attr->{$id_ch}[$INDX_LENGTH];

	my $exponent_per_site = ($totrate_ins + $totrate_del) * $len_br_ch;

   if ($TEST_LEVEL_corr>3) {
	print STDOUT ' ' x 5, "[ Main inputs and secondary input ]\n\n";
	print STDOUT ' ' x 5, '$cutoff_ins = ', $cutoff_ins, ",\n";
	print STDOUT ' ' x 5, '$cutoff_del = ', $cutoff_del, ",\n";
	print STDOUT ' ' x 5, '$totrate_ins = ', $totrate_ins, ",\n";
	print STDOUT ' ' x 5, '$totrate_del = ', $totrate_del, ",\n";
	print STDOUT ' ' x 5, '$len_br_ch = ', $len_br_ch, ",\n\n";
	print STDOUT ' ' x 5, '$exponent_per_site (= ($totrate_ins + $totrate_del) * $len_br_ch) = ', $exponent_per_site, "\n\n";
   }

	my $log_mfacs_ins_sbj = $br2log_mfacs->{$id_uninode}[0];
	my $log_mfacs_del_ch = $br2log_mfacs->{$id_ch}[1];

#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,

	my $corr_fac2 = 0;

   (print STDOUT ' ' x5, "[ Calculating correction factors from $ct_insdelpairs insertion-deletion pairs ]\n\n") if ($TEST_LEVEL_corr>3);

	foreach my $insdelpair (@{$insdelpairs}) { # Main outer foreach-loop (over insertion-deletion pairs). #

	    my ($info_ins, $info_del) = @{$insdelpair};
	    my $blks_ins = $info_ins->[2];
	    my $blks_del = $info_del->[2];

	    my $len_ins = my $len_del = 0;

	    foreach my $blk (@{$blks_ins}) { $len_ins += $cls_gpblks->[$blk][1]; }
	    foreach my $blk (@{$blks_del}) { $len_del += $cls_gpblks->[$blk][1]; }

#	@{$cls_gpblocks[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is
#	for the $gpb th gap-pattern-block in the input cluster, with
#	$len_gpb being the length (in terms of #{columns}) of the block,

	    my @elements = ();

	    my $upper_bound1 = $cutoff_ins - $len_ins;
	    my $upper_bound2 = $cutoff_del - $len_del;
	    my $upper_bound = ($upper_bound1 < $upper_bound2) ? $upper_bound1 : $upper_bound2;
	    $upper_bound++;

   if ($TEST_LEVEL_corr>= 5) {
	print STDOUT ' ' x 5, "{inserted blocks} = ", join (', ', @{$blks_ins}), " (total length: $len_ins);\n";
	print STDOUT ' ' x 5, "{deleted blocks} = ", join (', ', @{$blks_del}), " (total length: $len_del);\n";
	print STDOUT ' ' x 5, "Summation over #{'invisible' residues} in [0, $upper_bound) .\n\n";
   }

	    ($upper_bound>0) or next; # Ignore the pair. #

	    for (my $x = 0; $x <$upper_bound; $x++) {
		my $element = (-$x * $exponent_per_site) + $log_mfacs_ins_sbj->[$len_ins + $x] + $log_mfacs_del_ch->[$len_del + $x];
		push @elements, $element;

    (print STDOUT ' ' x 7, '$x = ', $x, ",\t", '$element = ', $element, ",\n") if ($TEST_LEVEL_corr>5); 
	    }

	    my $numer = (@elements==1) ? $elements[0] : log_sum_exp (@elements);
	    unless (defined $numer) {
		print STDOUT "[ERROR 'cal_corr_fac_from_uninode_hs'] The subroutine 'log_sum_exp' FAILED: $! !\n\n";
		return undef;
	    }


			# MAJOR REVISION on 2019/05/09. #
	    my $corr_fac22 = $numer - ($log_mfacs_ins_sbj->[$len_ins] + $log_mfacs_del_ch->[$len_del]) ;
	    ##my $corr_fac22 = ($numer + log (2)) - ($log_mfacs_ins_sbj->[$len_ins] + $log_mfacs_del_ch->[$len_del]) ;

	    $corr_fac2 +=  $corr_fac22;

    if ($TEST_LEVEL_corr>= 5) {
	print STDOUT ' ' x 5, 'Log(total contribution from the insdelpairs with 0-', $upper_bound, ' invisible residues) (= $numer) = ', $numer, ",\n";
	print STDOUT ' ' x 5, 'Log(multiple factor from the original insertion) (= $log_mfacs_ins_sbj->[$len_ins]) = ', $log_mfacs_ins_sbj->[$len_ins], ",\n";
	print STDOUT ' ' x 5, 'Log(multiple factor from the original deletion) (= $log_mfacs_del_ch->[$len_del]) = ', $log_mfacs_del_ch->[$len_del], ",\n\n";
    }
    if ($TEST_LEVEL_corr > 3) {
			# Major revision on 2019/05/09. #
	print STDOUT ' ' x 5, '$corr_fac22 (= $numer - ($log_mfacs_ins_sbj->[$len_ins] + $log_mfacs_del_ch->[$len_del])) = ', $corr_fac22, ",\n";
	##print STDOUT ' ' x 5, '$corr_fac22 (= ($numer + log (2)) - ($log_mfacs_ins_sbj->[$len_ins] + $log_mfacs_del_ch->[$len_del])) = ', $corr_fac22, ",\n";
	print STDOUT ' ' x 5, 'Current $corr_fac2 = ', $corr_fac2, ".\n\n";
    }

	}			# End of the main outer foreach-loop (over insertion-deletion pairs). #

			### RESTART FROM HERE (2b) ###

   (print STDOUT ' ' x 5, "< 'cal_corr_fac_from_uninode_hs'  ended  (with", '$corr_fac2 = ', "$corr_fac2) />\n\n") if ($TEST_LEVEL_corr>2);

	return $corr_fac2;

}	# End of 'cal_corr_fac_from_uninode_hs'. #


# cal_corr_fac_lnmfac1_indel_hst_hs:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa2_hs_finer' (in "../MyPerlModules_ANEX/MyTreeMap_indels_ML_hs_hs_finer.pm" ... not yet created),
#	or of 'apprx1_tot_lnlk_gpattern_msa2_hs' (in "./MyTreeMap_indels_ML_hs_hs.pm"),
#	was created by modifying its predecessor, 'cal_corr_fac_lnmfac1_indel_hst'.
#	The ONLY major DIFFERENCE from the precessor is that
#	this subroutine follows the "virtual" order of gap-pattern blocks specified by two arrays, @vplaces_gpbs and @gpbs_at_spec_vplaces,
#	whereas the predecessor follows the original order of the blocks, specified by their indices.
#	Basically, the DIFFERENCE was implemented by replacing the main subroutines, 'cal_corr_fac_from_biroot' and 'cal_corr_fac_from_uninode', with 'cal_corr_fac_from_biroot_hs' and 'cal_corr_fac_from_uninode_hs', respectively.
#	[CREATED on 2019/05/07.]
#
# cal_corr_fac_lnmfac1_indel_hst:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa2_finer' (in "../MyPerlModules_ANEX/MyTreeMap_indels_ML_hs_finer.pm"),
#	or of 'apprx1_tot_lnlk_gpattern_msa2' (in the NEW VERSION of "./MyTreeMap_indels_ML_hs.pm"),
#	computes the correction factor for the multiplication factor contributed from an indel history 
#	involving at least a pair of an insertion and a subsequent deletion separated by a uni-furcated node,
#	and/or a pair of deletions separated by a bifurcated root.
#
#	[NOTE] This initial version of this subroutine deals with the indel history at either end of the MSA
#	in the same way as dealing with the history in the bulk.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm', from Dec 31 (Sun), 2017 till ...]
#	[MODIFIED immediately after being IMPORTED to "MyTreeMap_indels_ML_sppl.pm" on 2019/04/18.]
#
# [ INPUT: (\@info_biroot, \@list_uninodes, \@cls_gpb, \%br2atrr, \%node2pa, \%br2indels, \%node2blwise_gstats, \%node2empty_gpbs0, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs, $delta_log_prior_rootseq), where ] ... for the predecessor 'cal_corr_fac_lnmfac1_indel_hst'.
# INPUT: (\@info_biroot, \@list_uninodes, \@cls_gpb, \%br2atrr, \%node2pa, \%br2indels, \%node2blwise_gstats, \@vplaces_gpbs, \@gpbs_at_spec_vplaces,
#	\%node2empty_gpbs0, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs, $delta_log_prior_rootseq), where
#
#	@{$info_biroot} = ($id_root, $id_child1, $id_child2) for the bifurcated root,
#		      = () if the root is NOT bifurcated;
#	@{$list_uninodes[$u]} = ($id_parent, $id_uninode, $id_child) for the $u th unifurcated node (other than the root),
#	$id_uninode, whose parent and the sole child are $id_parent and $id_child, respectively;
#
#	@{$cls_gpblocks[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is
#	for the $gpb th gap-pattern-block in the input cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gpb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention;
#
#	%br2attr = ($br_id => \@attr, ...) for the branches, with
#	$br_id being the ID of the node at the lower-end of the branch, and
#	$attr[$INDX_LENGTH] giving the length of the branch (in terms of #{substitutions}/site);
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the nodes except the top node;
#	$len_msa is the length of the input MSA;
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions within the gapped segment onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event;
#
#	%node2bw_gstats = ($node_id => \@bw_gstats_in_idb, ...) is the map of gap states within the gapped segment onto the
#	nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id (, or $char_pres if the gap-pattern block (at the node) does not belong to the indel block);
#
#	[ADDED in creation of 'cal_corr_fac_lnmfac1_indel_hst_hs'.]
#	@vplaces_gpbs and @gpbs_at_spec_vplaces specifiy the order among the gap-pattern blocks underlying this parsimonious history. Specifically,
#	$vplaces_gpbs[$gpb] specifies what place the $gpb th gap-pattern block is in,
#	$gpbs_at_spec_vplaces[$k] gives the index of the gap-pattern block in the $k th place.
#	[END of "ADDED in creation of 'cal_corr_fac_lnmfac1_indel_hst_hs'."]
#
#	%node2empty_gpbs0 = ($node_id => \@empty_gpbs0, ...) for the nodes involved in the indel block,
#	with @empty_gpbs0 listing the indices of the gap-pattern-blocks that are empty at the node; # ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2'. #
#
#	$totrate_ins/del being the total rate of insertions/deletions (per site) (relative to the substitution rate),
#
#	@params_boundaries = ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end), with
#
#	$cutoff_ins/del being the cut-off length for the insertions/deletions that is common along the MSA,
#
#	$if_5p/3p_end = 1 if the gap-block cluster extends into the 5'-/3'-end (=0 otherwise).
#
#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,
#	$log_mfacs_ins/del_bd[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertions/deletions of $l sites on either boundary of the MSA,
#	$log_mfacs_pair[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites in the bulk of the MSA,
#	$log_mfacs_pair_bd[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites on either boundary of the MSA.
#
#	$delta_log_prior_rootseq_in_idb is the increment of the logarithm of the multiplicative contribution from the indel block to the prior probability of the root sequence state. # ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2'. #
#
#
# [ OUTPUT: $corr_fac, 
#	which is the logarithm of the multiplicative correlation factor for the multiplication factor ] ... OBSOLETE As of 2019/04/18.
#
# OUTPUT: ($corr_fac, \@sub_corr_facs), where
#	$curr_fac is the logarithm of the multiplicative correlation factor for the multiplication factor
#	contributed from this indel history. It is undefined if the subroutine failed;
#
#	@sub_corr_facs stores all the contributions to $corr_fac, each from either a bifurcated root or from a uni-furcated node (other than the root); 
#	
#
##sub cal_corr_fac_lnmfac1_indel_hst (\@\@\@\%\%\%\%\%$$$\@\%$) {
#	my ($info_biroot, $list_uninodes, $cls_gpblks, $br2attr, $node2pa, $br2indels, $node2bw_gstats, $node2empty_gpbs0, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs, $delta_log_prior_rootseq_in_idb) = @_;
#
sub cal_corr_fac_lnmfac1_indel_hst_hs (\@\@\@\%\%\%\%\@\@\%$$$\@\%$) {

	my ($info_biroot, $list_uninodes, $cls_gpblks, $br2attr, $node2pa, $br2indels, $node2bw_gstats, $vplaces_gpbs, $gpbs_at_spec_vplaces,
	$node2empty_gpbs0, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs, $delta_log_prior_rootseq_in_idb) = @_; # MODIFIED in creation of 'cal_corr_fac_lnmfac1_indel_hst_hs'. #

	##my $ct_gpblks = @{$cls_gpblks};
	my $ct_uninodes = @{$list_uninodes};

    (print STDOUT ' ' x 3, "< Subroutine 'cal_corr_fac_lnmfac1_indel_hst_hs' (#{uni-furcated nodes} = $ct_uninodes) >\n\n") if (($TEST_LEVEL_corr>0) and ((@{$info_biroot}>0) or ($ct_uninodes>0))) ;

	my $corr_fac = 0;
	my @sub_corr_facs = (); # ADDED in MODIFICATION on 2019/04/18.

	if (@{$info_biroot}>0) {
			# MODIFIED in creation of 'cal_corr_fac_lnmfac1_indel_hst_hs'. (1) #
	    my $corr_fac1 = cal_corr_fac_from_biroot_hs (@{$info_biroot}, @{$cls_gpblks}, %{$br2attr}, %{$br2indels}, %{$node2bw_gstats}, @{$vplaces_gpbs}, @{$gpbs_at_spec_vplaces}, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, %{$br2log_mfacs});
	    ##my $corr_fac1 = cal_corr_fac_from_biroot (@{$info_biroot}, @{$cls_gpblks}, %{$br2attr}, %{$br2indels}, %{$node2bw_gstats}, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, %{$br2log_mfacs});

	    unless (defined $corr_fac1) {
		print STDOUT "[ERROR 'cal_corr_fac_lnmfac1_indel_hst_hs'] The subroutine 'cal_corr_fac_from_biroot_hs' FAILED: $! !\n\n";
		return (undef, undef); # MODIFIED on 2019/04/18.
		##return undef;
	    }

	    $corr_fac += $corr_fac1;

	    push @sub_corr_facs, $corr_fac1; # ADDED in MODIFICATION on 2019/04/18.

   if ($TEST_LEVEL_corr>1) {
	print STDOUT ' ' x 3, "{Correction factor from the bifurcated root} = $corr_fac1,\n";
	print STDOUT ' ' x 3, 'Current $corr_fac = ', $corr_fac, ".\n\n";
   } 
	} # END of "if (@{$info_biroot}>0) {...}"


	for (my $u=0; $u<$ct_uninodes; $u++) { # Main for-loop (over uni-furcated nodes). #

	    my $info_uninode = $list_uninodes->[$u];

			# MODIFIED in creation of 'cal_corr_fac_lnmfac1_indel_hst_hs'. (2) #
	    my $corr_fac2 = cal_corr_fac_from_uninode_hs (@{$info_uninode}, @{$cls_gpblks}, %{$br2attr}, %{$br2indels}, %{$node2bw_gstats}, @{$vplaces_gpbs}, @{$gpbs_at_spec_vplaces}, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, %{$br2log_mfacs});
	    ##my $corr_fac2 = cal_corr_fac_from_uninode (@{$info_uninode}, @{$cls_gpblks}, %{$br2attr}, %{$br2indels}, %{$node2bw_gstats}, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, %{$br2log_mfacs});

	    unless (defined $corr_fac2) {
		print STDOUT "[ERROR 'cal_corr_fac_lnmfac1_indel_hst_hs'] The subroutine 'cal_corr_fac_from_uninode_hs' FAILED for the $u th uni-furcated node: $! !\n\n";
		return (undef, \@sub_corr_facs); # MODIFIED on 2019/04/18.
		##return undef;
	    }
	    $corr_fac += $corr_fac2;

	    push @sub_corr_facs, $corr_fac2; # ADDED in MODIFICATION on 2019/04/18.

   if ($TEST_LEVEL_corr>1) {
	print STDOUT ' ' x 3, "{Correction factor from the $u th uni-furcated node} = $corr_fac2,\n";
	print STDOUT ' ' x 3, 'Current $corr_fac = ', $corr_fac, ".\n\n";
   }

	}		# End of the main for-loop (over uni-furcated nodes). #


    (print STDOUT "\n", ' ' x 3, "< Subroutine 'cal_corr_fac_lnmfac1_indel_hst_hs' Ended (with ", '$corr_fac = ', "$corr_fac) />\n\n") if (($TEST_LEVEL_corr>0) and ((@{$info_biroot}>0) or ($ct_uninodes>0))) ;

	return ($corr_fac, \@sub_corr_facs); # MODIFIED on 2019/04/18.
	##return $corr_fac;

}	# End of 'cal_corr_fac_lnmfac1_indel_hst_hs'. #


		## END of "Subroutines IMPORTED FROM "../MyPerlModules_ANEX/MyTreeMap_indels_ML_hs_finer.pm" (on 2019/04/18)." ##




1;	# a "spell" which makes the module "use"-able and/or "require"-able #


